

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Constructs.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
// Forward declarations.
//====================================================================
class color4f;
class color4b;


//====================================================================
// Floating-point representation of color.
//====================================================================
class VW_COMMON_API color4f
{
public:
	//------
	// Data.
	//------
	float  r;
	float  g;
	float  b;
	float  a;


public:
	//-----------------------
	// Convenience constants.
	//-----------------------
	static const color4f  ZERO;
	static const color4f  ONE;
	static const color4f  BLACK;
	static const color4f  DARK_GRAY;
	static const color4f  GRAY;
	static const color4f  LIGHT_GRAY;
	static const color4f  WHITE;
	static const color4f  RED;
	static const color4f  GREEN;
	static const color4f  BLUE;
	static const color4f  CYAN;
	static const color4f  MAGENTA;
	static const color4f  YELLOW;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline color4f()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline color4f(float _r, float _g, float _b, float _a = 1.f) {
		Set(_r, _g, _b, _a);
	}

	__forceinline color4f(const float* p) {
		Set(p);
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline color4f(const color4f& rhs) {
		this->operator=(rhs);
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline color4f& operator=(const color4f& rhs) {
		r = rhs.r;
		g = rhs.g;
		b = rhs.b;
		a = rhs.a;
		return *this;
	}

	//------------------------
	// Conversion constructor.
	//------------------------
	__forceinline color4f(const color4b& rhs) {
		this->operator=(rhs);
	}

	//-----------------------
	// Conversion assignment.
	//-----------------------
	color4f& operator=(const color4b& rhs);

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(float _r, float _g, float _b, float _a = 1.f) {
		r = _r;
		g = _g;
		b = _b;
		a = _a;
	}

	__forceinline void  Set(const float* p) {
		r = p[0];
		g = p[1];
		b = p[2];
		a = p[3];
	}

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline float*        ToPtr()       { return &r; }
	__forceinline const float*  ToPtr() const { return &r; }

	__forceinline operator float*()             { return ToPtr(); }
	__forceinline operator const float*() const { return ToPtr(); }

	//----------
	// Addition.
	//----------
	__forceinline color4f&  operator+=(const color4f& rhs) {
		return  this->Add(rhs);
	}

	__forceinline color4f&  Add(const color4f& rhs) {
		r += rhs.r;
		g += rhs.g;
		b += rhs.b;
		a += rhs.a;
		return *this;
	}

	__forceinline color4f  operator+(const color4f& rhs) const {
		return color4f(r + rhs.r, g + rhs.g, b + rhs.b, a + rhs.a);
	}

	__forceinline color4f&  Add(const color4f& lhs, const color4f& rhs) {
		r = lhs.r + rhs.r;
		g = lhs.g + rhs.g;
		b = lhs.b + rhs.b;
		a = lhs.a + rhs.a;
		return *this;
	}

	//-------------
	// Subtraction.
	//-------------
	__forceinline color4f&  operator-=(const color4f& rhs) {
		return  this->Sub(rhs);
	}

	__forceinline color4f&  Sub(const color4f& rhs) {
		r -= rhs.r;
		g -= rhs.g;
		b -= rhs.b;
		a -= rhs.a;
		return *this;
	}

	__forceinline color4f  operator-(const color4f& rhs) const {
		return color4f(r - rhs.r, g - rhs.g, b - rhs.b, a - rhs.a);
	}

	__forceinline color4f&  Sub(const color4f& lhs, const color4f& rhs) {
		r = lhs.r - rhs.r;
		g = lhs.g - rhs.g;
		b = lhs.b - rhs.b;
		a = lhs.a - rhs.a;
		return *this;
	}

	//----------------
	// Multiplication.
	//----------------
	__forceinline color4f&  operator*=(const color4f& rhs) {
		return  this->Mul(rhs);
	}

	__forceinline color4f&  Mul(const color4f& rhs) {
		r *= rhs.r;
		g *= rhs.g;
		b *= rhs.b;
		a *= rhs.a;
		return *this;
	}

	__forceinline color4f  operator*(const color4f& rhs) const {
		return color4f(r * rhs.r, g * rhs.g, b * rhs.b, a * rhs.a);
	}

	__forceinline color4f&  Mul(const color4f& lhs, const color4f& rhs) {
		r = lhs.r * rhs.r;
		g = lhs.g * rhs.g;
		b = lhs.b * rhs.b;
		a = lhs.a * rhs.a;
		return *this;
	}

	//------------------------------------
	// Division (no divide-by-zero check).
	//------------------------------------
	__forceinline color4f&  operator/=(const color4f& rhs) {
		return  this->Div(rhs);
	}

	__forceinline color4f&  Div(const color4f& rhs) {
		r /= rhs.r;
		g /= rhs.g;
		b /= rhs.b;
		a /= rhs.a;
		return *this;
	}

	__forceinline color4f  operator/(const color4f& rhs) const {
		return color4f(r / rhs.r, g / rhs.g, b / rhs.b, a / rhs.a);
	}

	__forceinline color4f&  Div(const color4f& lhs, const color4f& rhs) {
		r = lhs.r / rhs.r;
		g = lhs.g / rhs.g;
		b = lhs.b / rhs.b;
		a = lhs.a / rhs.a;
		return *this;
	}

	//--------------------------
	// Multiplication by scalar.
	//--------------------------
	__forceinline color4f&  operator*=(float num) {
		return  this->Mul(num);
	}

	__forceinline color4f&  Mul(float num) {
		r *= num;
		g *= num;
		b *= num;
		a *= num;
		return *this;
	}

	__forceinline color4f  operator*(float num) const {
		return color4f(r * num, g * num, b * num, a * num);
	}

	__forceinline color4f&  Mul(const color4f& lhs, float num) {
		r = lhs.r * num;
		g = lhs.g * num;
		b = lhs.b * num;
		a = lhs.a * num;
		return *this;
	}

	//----------------------------------------------
	// Division by scalar (no divide-by-zero check).
	//----------------------------------------------
	__forceinline color4f&  operator/=(float num) {
		return  this->Div(num);
	}

	__forceinline color4f&  Div(float num) {
		return  this->Mul(1 / num);
	}

	__forceinline color4f  operator/(float num) const {
		num = 1 / num;
		return color4f(r * num, g * num, b * num, a * num);
	}

	__forceinline color4f&  Div(const color4f& lhs, float num) {
		return  this->Mul(lhs, 1 / num);
	}

	//----------
	// Negation.
	//----------
	__forceinline color4f&  Neg() {
		return  this->Neg(*this);
	}

	__forceinline color4f&  Neg(const color4f& src) {
		return  this->Mul(src, -1);
	}

	//------------------
	// Unary operations.
	//------------------
	__forceinline color4f  operator-() const {
		return color4f(-r, -g, -b, -a);
	}

	__forceinline color4f&        operator+()       { return *this; }
	__forceinline const color4f&  operator+() const { return *this; }

	//-------------
	// Comparisons.
	//-------------
	__forceinline bool  operator==(const color4f& rhs) const {
		return ((r == rhs.r) && (g == rhs.g) && (b == rhs.b) && (a == rhs.a));
	}
	__forceinline bool  operator!=(const color4f& rhs) const {
		return !(*this == rhs);
	}

	//-----------
	// Utilities.
	//-----------
	__forceinline float  Luminance() const {
		return (0.299f * r + 0.587f * g + 0.114f * b);
	}
};


//====================================================================
// Multiplication by scalar (redirection).
//====================================================================
__forceinline  color4f __fastcall operator*(float num, const color4f& rhs) {
	return  rhs * num;
}


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const color4f& c);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, color4f& c);


//====================================================================
// 4-byte representation of color.
// Consistent with ARGB in-register ordering.
//====================================================================
class VW_COMMON_API color4b
{
public:
	//------
	// Data.
	//------
	ubyte  b;
	ubyte  g;
	ubyte  r;
	ubyte  a;


public:
	//-----------------------
	// Convenience constants.
	//-----------------------
	static const color4b  ZERO;
	static const color4b  ONE;
	static const color4b  BLACK;
	static const color4b  DARK_GRAY;
	static const color4b  GRAY;
	static const color4b  LIGHT_GRAY;
	static const color4b  WHITE;
	static const color4b  RED;
	static const color4b  GREEN;
	static const color4b  BLUE;
	static const color4b  CYAN;
	static const color4b  MAGENTA;
	static const color4b  YELLOW;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline color4b()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline color4b(ubyte _r, ubyte _g, ubyte _b, ubyte _a = 255) {
		Set(_r, _g, _b, _a);
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline color4b(const color4b& rhs) {
		this->operator=(rhs);
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline color4b& operator=(const color4b& rhs) {
		reinterpret_cast<uint&>(*this) = reinterpret_cast<const uint&>(rhs);
		return *this;
	}

	//------------------------
	// Conversion constructor.
	//------------------------
	__forceinline color4b(const color4f& rhs) {
		this->operator=(rhs);
	}

	//-----------------------
	// Conversion assignment.
	//-----------------------
	color4b& operator=(const color4f& rhs);

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(ubyte _r, ubyte _g, ubyte _b, ubyte _a = 255) {
		b = _b;
		g = _g;
		r = _r;
		a = _a;
	}

	//-----------------------------------
	// Conversion to/from 32-bit integer.
	//-----------------------------------
	__forceinline color4b& operator=(uint rhs) { reinterpret_cast<uint&>(*this) = rhs; return *this; }
	__forceinline operator uint() const { return reinterpret_cast<const uint&>(*this); }

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline ubyte*        ToPtr()       { return reinterpret_cast<ubyte*>(this); }
	__forceinline const ubyte*  ToPtr() const { return reinterpret_cast<const ubyte*>(this); }

	__forceinline operator ubyte*()             { return ToPtr(); }
	__forceinline operator const ubyte*() const { return ToPtr(); }

	//-------------
	// Comparisons.
	//-------------
	__forceinline bool  operator==(const color4b& rhs) const {
		return ((b == rhs.b) && (g == rhs.g) && (r == rhs.r) && (a == rhs.a));
	}
	__forceinline bool  operator!=(const color4b& rhs) const {
		return !(*this == rhs);
	}

	//-----------------
	// RGB order swaps.
	//-----------------
	__forceinline void     FlipRGBOrder() { Swap(b, r); }
	__forceinline color4b  FlippedRGBOrder() const { return color4b(b, g, r, a); }
};


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const color4b& c);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, color4b& c);


//====================================================================
// Color component conversions.
//====================================================================
__forceinline ubyte __fastcall ConvertColorComponent(float c)
{
	return ubyte(255.f * c);
}

__forceinline float __fastcall ConvertColorComponent(ubyte c)
{
	return float(1. / 255.) * float(c);
}


//====================================================================
// Conversion assignment implementations.
//====================================================================
__forceinline  color4f&  color4f::operator=(const color4b& rhs)
{
	r = ConvertColorComponent(rhs.r);
	g = ConvertColorComponent(rhs.g);
	b = ConvertColorComponent(rhs.b);
	a = ConvertColorComponent(rhs.a);
	return *this;
}


//====================================================================
__forceinline  color4b&  color4b::operator=(const color4f& rhs)
{
	b = ConvertColorComponent(rhs.b);
	g = ConvertColorComponent(rhs.g);
	r = ConvertColorComponent(rhs.r);
	a = ConvertColorComponent(rhs.a);
	return *this;
}


//********************************************************************
}//namespace vertexware


