

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\CompilerConfig.h"
#include "BaseTools\Types.h"
#include "BaseTools\Exception.h"
#include <string>
#include <vector>
#include <ostream>


//********************************************************************
namespace vertexware {


//====================================================================
VW_DECLARE_EXCEPTION( CommandLineErrorException )


//====================================================================
// Parses command-line into command and arguments.
//====================================================================
class VW_COMMON_API CommandLineParser
{
public:
	static const char* s_szDefaultPackedArgumentSeparators;


public:
	//----------------
	enum eArgumentType
	{
		ArgumentType_Option = 0,
		ArgumentType_Assignment,
		ArgumentType_String,

		ArgumentType_Count,
		ArgumentType_Force32Bits = 0x7FFFFFFF
	};


	//------------------------------
	class VW_COMMON_API NameValueRef
	{
	private:
		const std::string& name;
		const std::string& value;

	public:
		NameValueRef(const std::string& n, const std::string& v) : name(n), value(v) {}

		const std::string& Name() const { return name; }
		const std::string& Value() const { return value; }
	};


	//--------------------------
	class VW_COMMON_API Argument
	{
	private:
		uint           clPosition;
		eArgumentType  type;
		std::string    value1;
		std::string    value2;

	public:
		Argument()
		{ }

		Argument(uint pos, eArgumentType t, const std::string& v1 = "", const std::string& v2 = "")
			: value1(v1), value2(v2)
		{
			clPosition = pos;
			type = t;
		}

		bool  operator==(eArgumentType t) const { return type == t; }
		bool  operator!=(eArgumentType t) const { return !(*this == t); }

		std::string  MakeFullString() const;

		uint  GetCmdLinePosition() const { return clPosition; }

		eArgumentType       Type() const { return type; }
		const std::string&  Option() const { return value1; }
		NameValueRef        Assignment() const { return NameValueRef(value1, value2); }
		const std::string&  String() const { return value1; }
	};


public:
	typedef  std::vector<std::string>  STRING_VEC;
	typedef  std::vector<Argument>     ARGUMENT_VEC;


public:
	CommandLineParser();

	void  SetSpecialCharacters(char quoteChar = '\'', char quoteEscapeChar = '$',
					char scopeChar = '.', char optionChar = '-', char assignChar = '=');

	void  GetSpecialCharacters(char& quoteChar, char& quoteEscapeChar,
					char& scopeChar, char& optionChar, char& assignChar) const;

	//---------------------------------
	// Clears last parsed command-line.
	//---------------------------------
	void  ClearLastCommandLine();

	//---------------------------------------------
	// Clears last results. Automatically called by
	// Parse() if error occurs.
	//---------------------------------------------
	void  ClearLastResults();

	//--------------------------------------------------
	// Clears last results and last parsed command-line.
	// Automatically called before parsing begins.
	//--------------------------------------------------
	void  ClearAll();

	//--------------------------------------------
	// Sets parsing options.
	// If <bIgnoreObjectChain> is false, parser
	// accepts commands in form  [OBJECT.] COMMAND
	// Example: object1.subobject1.command
	//--------------------------------------------
	void  SetParseFlags(bool bIgnoreObjectChain = true, bool bStripDoubleQuotesOnStdCmd = true);

	//---------------------------------------------------------------
	// Parses given command-line.
	// Returns true on success. If error is found in command-line
	// and <exceptionOnError> is true, exception is thrown with
	// message describing the error, otherwise it just returns false.
	//---------------------------------------------------------------
	bool  Parse(const std::string& commandLine, bool exceptionOnError = true);

	bool  Parse(int argc, const char** args, bool exceptionOnError = true);

	bool  Parse(int argc, char** args, bool exceptionOnError = true) {
		return Parse(argc, (const char**)(args), exceptionOnError);
	}

	//-----------------------------------
	// Access to results of last Parse().
	//-----------------------------------
	const std::string&  GetLastCommandLine() const { return lastCommandLine; }

	const std::string&  GetCommand() const { return commandIdentifier; }
	void  GetCommand(STRING_VEC& objectIdentifierChain, std::string& commandIdentifier) const;

	uint             GetNumArguments() const { return arguments.size(); }
	const Argument&  GetArgument(uint i) const;

	bool  GetOptionArgument(const std::string& s, Argument& arg) const;
	bool  GetAssignmentArgument(const std::string& s, Argument& arg) const;
	bool  GetStringArgument(const std::string& s, Argument& arg) const;

	bool  GetOption(const std::string& option) const;
	bool  GetAssignment(const std::string& name, std::string& value) const;
	bool  GetString(const std::string& s) const;

	//--------------------------------------------------------------
	// Extract typed argument.
	// Returns true if argument is found.
	// If argument is found but conversion from string to value
	// fails, argument is classified as not-found.
	// <outValue> is set to <defaultValue> if argument is not found.
	//--------------------------------------------------------------
	bool  ExtractArgumentPointer(const std::string& name, void*& outValue, void* defaultValue = 0) const;
	bool  ExtractArgumentInteger(const std::string& name, int& outValue, int defaultValue = 0) const;
	bool  ExtractArgumentUInteger(const std::string& name, uint& outValue, uint defaultValue = 0) const;
	bool  ExtractArgumentDouble(const std::string& name, double& outValue, double defaultValue = 0) const;
	bool  ExtractArgumentBool(const std::string& name, bool& outValue, bool defaultValue = false) const;

	//-----------------------------------
	// Extracts separator-packed values.
	// Returns true if argument is found.
	//-----------------------------------
	bool  GetPackedAssignment(const std::string& name, std::vector<std::string>& outValues,
						const std::string& separators = s_szDefaultPackedArgumentSeparators) const;

	// returns true on any error
	bool  ExtractPackedUIntegers(const std::string& name, std::vector<uint>& outValues,
						const std::string& separators = s_szDefaultPackedArgumentSeparators) const;

	//------------
	// Debug help.
	//------------
	void  DumpLastResults(std::ostream& os) const;

	//-----------
	// Utilities.
	//-----------
	std::string  FormatOptionHelp(const std::string& option,
						const std::string& helpText = "", uint tabLevel = 0) const;

	std::string  FormatAssignmentHelp(const std::string& name, const std::string& value,
						const std::string& helpText = "", uint tabLevel = 0) const;

	std::string  FormatStringHelp(const std::string& str,
						const std::string& helpText = "", uint tabLevel = 0) const;


private:
	// special characters
	char  quoteChar;
	char  quoteEscapeChar;
	char  scopeChar;
	char  optionChar;
	char  assignChar;

	// parsing flags
	bool  bIgnoreObjectChain; //object.object.object.command
	bool  bStripDoubleQuotesOnStdCmd;

	std::string   lastCommandLine;

	// parser products
	STRING_VEC    objectIdentifierChain;
	std::string   commandIdentifier;
	ARGUMENT_VEC  arguments;


private:
	bool  ExtractIdentifier(std::string& s, std::string& out,
				bool canStartWithDigit, bool canStartWithUnderscore, bool canStartWithHypen);
	bool  ExtractOption(std::string& s, std::string& out);
	bool  ExtractQuotedString(std::string& s, std::string& out);
	bool  ExtractString(std::string& s, std::string& out);
};


//====================================================================
VW_COMMON_API
std::ostream& operator<<(std::ostream& os, const CommandLineParser::Argument& arg);


//********************************************************************
}//namespace vertexware


