

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Memory.h"
#include "private\vw_common_exports.h"
#include <map>
#include <string>


//********************************************************************
namespace vertexware {


//====================================================================
// Prototype pattern.
// Defines cloning interface.
//====================================================================
class VW_COMMON_API Prototype
{
public:
	virtual ~Prototype() { }

	virtual  Prototype*  Clone() const = 0;
};


//====================================================================
// Clone factory.
// Keeps prototype-objects & creates prototype-clones based on given key.
// Registered prototypes are owned by factory (deleted on unregistration).
//====================================================================
template<typename KeyType>
class CloneFactory
{
public:
	CloneFactory()
	{ }

	virtual ~CloneFactory()
	{
		UnregisterAllPrototypes();
	}


	//-------------------
	// Prototype queries.
	//-------------------
	bool  IsPrototypeRegistered(const KeyType& key) const
	{
		PROTOTYPE_CITER it = prototypes.find(key);
		return (it != prototypes.end());
	}
	const Prototype*  GetPrototype(const KeyType& key) const
	{
		PROTOTYPE_CITER it = prototypes.find(key);
		if(it == prototypes.end()) return 0;
		return  it->second.Get();
	}


	//---------------------------------
	// Prototype registration.
	// <proto> must be created on heap.
	//---------------------------------
	bool  RegisterPrototype(const KeyType& key, Prototype* proto, bool bOverwrite = false)
	{
		if(IsPrototypeRegistered(key)) {
			if(! bOverwrite) return false;
			UnregisterPrototype(key);
		}
		prototypes.insert(PROTOTYPE_MAP::value_type(key, AutoPtr<Prototype>(proto)));
		OnRegister(key, proto);
		return true;
	}
	void  UnregisterPrototype(const KeyType& key)
	{
		PROTOTYPE_ITER it = prototypes.find(key);
		if(it != prototypes.end()) {
			OnUnregister(key, it->second.Get());
			prototypes.erase(it);
		}
	}
	void  UnregisterAllPrototypes()
	{
		PROTOTYPE_ITER it = prototypes.begin();
		while(it != prototypes.end()) {
			OnUnregister(it->first, it->second.Get());
			++it;
		}
		prototypes.clear();
	}


	//---------------------------------------------
	// Clone creation.
	// Caller is responsible for clone destruction.
	//---------------------------------------------
	Prototype*  Manufacture(const KeyType& key) const
	{
		PROTOTYPE_CITER it = prototypes.find(key);
		if(it == prototypes.end()) return 0;
		if(! QueryAllowManufacture(key)) return 0;
		Prototype* clone = it->second.Get()->Clone();
		OnManufacture(key, clone);
		return clone;
	}


protected:
	//----------------------------------------
	// Event notification for derived classes.
	//----------------------------------------
	virtual  void  OnRegister(const KeyType& key, const Prototype* proto) { }
	virtual  void  OnUnregister(const KeyType& key, const Prototype* proto) { }
	virtual  void  OnManufacture(const KeyType& key, const Prototype* clone) const { }


	//----------------------------------
	// Return true to allow manufacture.
	//----------------------------------
	virtual  bool  QueryAllowManufacture(const KeyType& key) const { return true; }


private:
	typedef  std::map<KeyType, AutoPtr<Prototype> >   PROTOTYPE_MAP;
	typedef  PROTOTYPE_MAP::iterator                  PROTOTYPE_ITER;
	typedef  PROTOTYPE_MAP::const_iterator            PROTOTYPE_CITER;

	PROTOTYPE_MAP   prototypes;
};


//********************************************************************
}//namespace vertexware


