

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\CompilerConfig.h"
#include "BaseTools\Serializers.h"
#include <algorithm>


//********************************************************************
namespace vertexware {


//====================================================================
// Simplifies clear operation on container-adapters.
// Used for std::stack, std::queue, std::priority_queue.
//====================================================================
template<typename ContainerType>
__forceinline
void __fastcall ClearContainerAdapter(ContainerType& c)
{
	while(! c.empty()) c.pop();
}


//====================================================================
// <TValue> is specified as additional template argument so that
// std::map<AutoPtr<TValue>> can be passed for <m>.
// Returns m.end() if not found.
//====================================================================
template<typename TMap, typename TValue>
TMap::const_iterator __fastcall
MapFindKeyByValue(const TMap& m, const TValue& val, TMap::key_type* pOutKey = 0)
{
	TMap::const_iterator it = m.begin();
	while(it != m.end()) {
		if(it->second == val) {
			SafeAssign(pOutKey, it->first);
			return it;
		}
		++it;
	}
	return m.end();
}


//====================================================================
// Extracts all map keys into another container.
//====================================================================
template<typename TMap, typename TOutputContainer>
void __fastcall MapExtractKeys(const TMap& m, TOutputContainer& outKeys)
{
	outKeys.clear();

	TMap::const_iterator it = m.begin();
	while(it != m.end()) {
		outKeys.push_back(it->first);
		++it;
	}
}


//====================================================================
// AliasMap (or DualKeyMap) has equal key_type & referent_type.
// If <k> is found, stores <k>'s value into <out> and returns true,
// otherwise copies <k> into <out> and returns false.
//====================================================================
template<typename TMap>
bool __fastcall MapResolveAlias(const TMap& m, const TMap::key_type& k, TMap::key_type& out)
{
	TMap::const_iterator it = m.find(k);
	if(it != m.end()) {
		out = it->second;
		return true;
	}
	out = k;
	return false;
}


//====================================================================
// Returns true if found.
// If <pOutIndex> != 0, the index of found item is assigned to it.
// If <pOutIter> != 0, the iterator pointing to found item is assigned to it.
//====================================================================
template<typename ContainerType>
bool __fastcall FindInContainer(ContainerType& c, const ContainerType::value_type& val,
								uint* pOutIndex = 0, ContainerType::iterator* pOutIter = 0)
{
	uint index = 0;
	ContainerType::iterator it = c.begin();
	while(it != c.end()) {
		ContainerType::value_type& object = *it;
		if(object == val) {
			SafeAssign(pOutIndex, index);
			SafeAssign(pOutIter, it);
			return true;
		}
		++it;
		++index;
	}
	return false;
}


//====================================================================
template<typename ContainerType>
bool __fastcall FindInContainer(const ContainerType& c, const ContainerType::value_type& val,
								uint* pOutIndex = 0, ContainerType::const_iterator* pOutIter = 0)
{
	uint index = 0;
	ContainerType::const_iterator it = c.begin();
	while(it != c.end()) {
		const ContainerType::value_type& object = *it;
		if(object == val) {
			SafeAssign(pOutIndex, index);
			SafeAssign(pOutIter, it);
			return true;
		}
		++it;
		++index;
	}
	return false;
}


//====================================================================
// Erases first matching object from container.
// Returns true if erased.
//====================================================================
template<typename ContainerType>
bool __fastcall EraseFromContainer(ContainerType& c,
						const ContainerType::value_type& val)
{
	ContainerType::iterator it = c.begin();
	while(it != c.end()) {
		ContainerType::value_type& object = *it;
		if(object == val) {
			c.erase(it);
			return true;
		}
		++it;
	}
	return false;
}


//====================================================================
// Erases all matching objects from container.
// Returns erase-count.
//====================================================================
template<typename ContainerType>
uint __fastcall EraseAllFromContainer(ContainerType& c,
						const ContainerType::value_type& val)
{
	uint eraseCount = 0;
	while(EraseFromContainer(c, val)) eraseCount++;
	return eraseCount;
}


//====================================================================
// Erases first matching object from container.
// Returns true if erased.
//
// struct UnaryPredicateType {
//     bool operator()(const ObjectType& object) const {
//         if(<object> matches criterion) return true;
//         return false;
//     }
// };
//====================================================================
template<typename UnaryPredicateType, typename ContainerType>
bool __fastcall EraseFromContainer_If(ContainerType& c,
						UnaryPredicateType unaryPredicate = UnaryPredicateType())
{
	ContainerType::iterator it = c.begin();
	while(it != c.end()) {
		ContainerType::value_type& object = *it;
		if(unaryPredicate(object)) {
			c.erase(it);
			return true;
		}
		++it;
	}
	return false;
}


//====================================================================
// Erases all matching objects from container.
// Returns erase-count.
//====================================================================
template<typename UnaryPredicateType, typename ContainerType>
uint __fastcall EraseAllFromContainer_If(ContainerType& c,
						UnaryPredicateType unaryPredicate = UnaryPredicateType())
{
	uint eraseCount = 0;
	while(EraseFromContainer_If(c, unaryPredicate)) eraseCount++;
	return eraseCount;
}


//====================================================================
// Used for containers of pointers to owned objects.
// Deletes all objects in the container, then clears the container.
//====================================================================
template<typename PtrContainerType>
void __fastcall DeletePtrContainer(PtrContainerType& c)
{
	PtrContainerType::iterator it = c.begin();
	while(it != c.end()) {
		PtrContainerType::value_type pObject = *it++;
		delete pObject;
	}
	c.clear();
}


//====================================================================
// Used for containers of pointers to owned objects.
// Deletes first matching object from container.
// Returns true if deleted.
//====================================================================
template<typename PtrContainerType>
bool __fastcall DeleteFromPtrContainer(PtrContainerType& c,
									PtrContainerType::value_type ptr)
{
	PtrContainerType::iterator it = c.begin();
	while(it != c.end()) {
		PtrContainerType::value_type pObject = *it;
		if(pObject == ptr) {
			delete ptr;
			c.erase(it);
			return true;
		}
		++it;
	}
	return false;
}


//====================================================================
// Deletes all matching objects from container.
// Returns delete-count.
//====================================================================
template<typename PtrContainerType>
uint __fastcall DeleteAllFromPtrContainer(PtrContainerType& c,
									PtrContainerType::value_type ptr)
{
	uint deleteCount = 0;
	while(DeleteFromPtrContainer(c, ptr)) deleteCount++;
	return deleteCount;
}


//====================================================================
// Used for containers of pointers to owned objects.
// Deletes first matching object from container.
// Returns true if deleted.
//
// struct UnaryPredicateType {
//     bool operator()(const ObjectType* pObject) const {
//         if(<pObject> matches criterion) return true;
//         return false;
//     }
// };
//====================================================================
template<typename UnaryPredicateType, typename PtrContainerType>
bool __fastcall DeleteFromPtrContainer_If(PtrContainerType& c,
						UnaryPredicateType unaryPredicate = UnaryPredicateType())
{
	PtrContainerType::iterator it = c.begin();
	while(it != c.end()) {
		PtrContainerType::value_type pObject = *it;
		if(unaryPredicate(pObject)) {
			delete pObject;
			c.erase(it);
			return true;
		}
		++it;
	}
	return false;
}


//====================================================================
// Deletes all matching objects from container.
// Returns delete-count.
//====================================================================
template<typename UnaryPredicateType, typename PtrContainerType>
uint __fastcall DeleteAllFromPtrContainer_If(PtrContainerType& c,
						UnaryPredicateType unaryPredicate = UnaryPredicateType())
{
	uint deleteCount = 0;
	while(DeleteFromPtrContainer_If(c, unaryPredicate)) deleteCount++;
	return deleteCount;
}


//====================================================================
// Used for containers of unique values (like <set>).
// Most likely these will contain pointers.
// If <value> is found, push_back is not executed.
// Returns true if added.
//====================================================================
template<typename ContainerType>
bool __fastcall UniqueContainerPushBack(ContainerType& c,
								const ContainerType::value_type& value)
{
	ContainerType::iterator it = std::find(c.begin(), c.end(), value);
	if(it != c.end()) return false;
	c.push_back(value);
	return true;
}


//====================================================================
// Returns 1 if inserted.
// Returns 0 if prepended.
// Returns -1 if not inserted.
//
// struct SearchPredicateType {
//     bool operator()(const ContainerType::value_type& value) const;
//     SearchPredicateType(const ContainerType::value_type& initValue);
// };
//====================================================================
template<typename SearchPredicateType, typename ContainerType>
bool __fastcall InsertBefore_If(ContainerType& c,
								const ContainerType::value_type& value,
								const ContainerType::value_type& before,
								bool bPrependIfNotFound = true)
{
	ContainerType::iterator it = std::find_if(c.begin(), c.end(), SearchPredicateType(before));
	if(it != c.end()) {
		c.insert(it, value);
		return 1;
	}
	if(bPrependIfNotFound) {
		c.insert(c.begin(), value);
		return 0;
	}
	return -1;
}


//====================================================================
// Returns 1 if inserted.
// Returns 0 if appended.
// Returns -1 if not inserted.
//====================================================================
template<typename SearchPredicateType, typename ContainerType>
bool __fastcall InsertAfter_If(ContainerType& c,
								const ContainerType::value_type& value,
								const ContainerType::value_type& after,
								bool bAppendIfNotFound = true)
{
	ContainerType::iterator it = std::find_if(c.begin(), c.end(), SearchPredicateType(after));
	it = (it != c.end()) ? ++it : it;
	if(it != c.end()) {
		c.insert(it, value);
		return 1;
	}
	if(bAppendIfNotFound) {
		c.insert(c.end(), value);
		return 0;
	}
	return -1;
}


//====================================================================
// Serializes flat-sequence-container (std::vector, std::list, std::deque).
// Container must contain simple-data-types or serializable-objects.
// WARNING: Be careful when serializing objects that contain pointers.
//====================================================================
template<typename ContainerType>
void __fastcall SerializeFlatSequenceContainer(std::ostream& os, const ContainerType& c)
{
	StreamExceptionSentry ses(os);

	// write container-size
	os <= c.size();

	// write container-elements
	ContainerType::const_iterator it = c.begin();
	while(it != c.end()) {
		ContainerType::const_reference val = *it++;
		os <= val;
	}
}


//====================================================================
// Deserializes flat-sequence-container.
// Container-element-type must be default-constructible.
//====================================================================
template<typename ContainerType>
void __fastcall DeserializeFlatSequenceContainer(std::istream& is, ContainerType& c)
{
	// clear container
	c.clear();

	try {
		StreamExceptionSentry ses(is);

		// read container-size
		uint size = 0;
		is >= size;
		if(! is) return;

		// read container-elements
		ContainerType::value_type val;
		for(uint i = 0; i < size; i++) {
			is >= val;
			c.push_back(val);
		}
	}
	catch(std::ios_base::failure&) {
		// clear on error
		c.clear();
		throw;
	}
}


//********************************************************************
}//namespace vertexware


