

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include <string>
#include <ostream>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace serializer_parameters {


//====================================================================
VW_COMMON_API
bool __fastcall IsMachineBigEndian();


//====================================================================
// WARNING: Not MT-safe.
// Defaults are set to match the machine byte-order, which is
// little-endian on IA32 machines.
// This affects only built-in-type serializers, so all others types
// should be serialized through built-in-types.
//====================================================================
VW_COMMON_API
void __fastcall SetOutputStreamByteOrder(bool bBigEndian = IsMachineBigEndian());


//====================================================================
VW_COMMON_API
bool __fastcall IsOutputStreamByteOrderBigEndian();


//====================================================================
class VW_COMMON_API OutputStreamByteOrderSentry
{
private:
	bool  bWasBigEndian;

public:
	OutputStreamByteOrderSentry(bool bBigEndian = IsMachineBigEndian()) {
		bWasBigEndian = IsOutputStreamByteOrderBigEndian();
		SetOutputStreamByteOrder(bBigEndian);
	}
	~OutputStreamByteOrderSentry() {
		SetOutputStreamByteOrder(bWasBigEndian);
	}
};


//====================================================================
VW_COMMON_API
void __fastcall SetInputStreamByteOrder(bool bBigEndian = IsMachineBigEndian());


//====================================================================
VW_COMMON_API
bool __fastcall IsInputStreamByteOrderBigEndian();


//====================================================================
class VW_COMMON_API InputStreamByteOrderSentry
{
private:
	bool  bWasBigEndian;

public:
	InputStreamByteOrderSentry(bool bBigEndian = IsMachineBigEndian()) {
		bWasBigEndian = IsInputStreamByteOrderBigEndian();
		SetInputStreamByteOrder(bBigEndian);
	}
	~InputStreamByteOrderSentry() {
		SetInputStreamByteOrder(bWasBigEndian);
	}
};


//********************************************************************
}//namespace serializer_parameters


//====================================================================
// Serialization / deserialization of built-in types.
//====================================================================
VW_COMMON_API __forceinline
std::ostream& __fastcall WriteByteBlock(std::ostream &os, const byte* block, uint byteCount) {
	return os.write(block, byteCount);
}


//====================================================================
VW_COMMON_API __forceinline
std::istream& __fastcall ReadByteBlock(std::istream &is, byte* block, uint byteCount) {
	return is.read(block, byteCount);
}


//====================================================================
VW_COMMON_API __forceinline
std::ostream& __fastcall WriteByteBlock(std::ostream &os, const ubyte* block, uint byteCount) {
	return WriteByteBlock(os, reinterpret_cast<const byte*>(block), byteCount);
}


//====================================================================
VW_COMMON_API __forceinline
std::istream& __fastcall ReadByteBlock(std::istream &is, ubyte* block, uint byteCount) {
	return ReadByteBlock(is, reinterpret_cast<byte*>(block), byteCount);
}


//====================================================================
// Auto-flips byte-order within each block if required by
// active input/output byte-order.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall WriteByteBlocks(std::ostream &os, const byte* blocks,
									uint byteCountPerBlock, uint blockCount);


//====================================================================
VW_COMMON_API
std::istream& __fastcall ReadByteBlocks(std::istream &is, byte* blocks,
									uint byteCountPerBlock, uint blockCount);


//====================================================================
VW_COMMON_API __forceinline
std::ostream& __fastcall WriteByteBlocks(std::ostream &os, const ubyte* blocks,
									uint byteCountPerBlock, uint blockCount) {
	return WriteByteBlocks(os, reinterpret_cast<const byte*>(blocks),
							byteCountPerBlock, blockCount);
}


//====================================================================
VW_COMMON_API __forceinline
std::istream& __fastcall ReadByteBlocks(std::istream &is, ubyte* blocks,
									uint byteCountPerBlock, uint blockCount) {
	return ReadByteBlocks(is, reinterpret_cast<byte*>(blocks),
							byteCountPerBlock, blockCount);
}


//====================================================================
// Footprint: 1 byte
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const ubyte& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, ubyte& x);


//====================================================================
// Footprint: 1 byte
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const byte& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, byte& x);


//====================================================================
// Footprint: 2 bytes
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const ushort& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, ushort& x);


//====================================================================
// Footprint: 2 bytes
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const short& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, short& x);


//====================================================================
// Footprint: 4 bytes
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const uint& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, uint& x);


//====================================================================
// Footprint: 4 bytes
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const int& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, int& x);


//====================================================================
// Footprint: 4 bytes
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const ulong& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, ulong& x);


//====================================================================
// Footprint: 4 bytes
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const long& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, long& x);


//====================================================================
// Footprint: 4 bytes
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const float& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, float& x);


//====================================================================
// Footprint: 8 bytes
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const double& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, double& x);


//====================================================================
// Footprint: 1 byte
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const bool& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, bool& x);


//====================================================================
// Serialization / deserialization of std::string.
// Footprint:
// a) 1 x UINT denotes string-length
// b) string-length x UBYTE
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const std::string& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, std::string& x);


//====================================================================
// Serialization / deserialization of UID.
// Footprint:
// a) 4 x UINT
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const UID& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, UID& x);


//********************************************************************
}//namespace vertexware


