

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Debug.h"
#include "private\vw_common_exports.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Singleton pattern.
// Ensures that at most one object of given type exists.
// Usage example:
//     class System : public Singleton<System> {};
//====================================================================
template<typename T>
class Singleton
{
protected:
	Singleton() {
		VW_ASSERT(! s_pInstance,
					std::string(VW_GET_CLASS_NAME(T)) + " singleton already created.");
		s_pInstance = static_cast<T*>(this);
	}
	~Singleton() {
		VW_ASSERT(s_pInstance, "Destructing non-existent " +
					std::string(VW_GET_CLASS_NAME(T)) + " singleton.");
		s_pInstance = 0;
	}


public:
	//-----------------------------
	// Access to one & only object.
	//-----------------------------
	static  T&  Instance() {
		VW_ASSERT(s_pInstance,
					std::string(VW_GET_CLASS_NAME(T)) + " singleton is not created.");
		return  *s_pInstance;
	}
	static  T*  InstancePtr() {
		VW_ASSERT(s_pInstance,
					std::string(VW_GET_CLASS_NAME(T)) + " singleton is not created.");
		return  s_pInstance;
	}


private:
	static  T*  s_pInstance;
};


//====================================================================
template<typename T>
T*  Singleton<T>::s_pInstance = 0;


//====================================================================
// Non-template version.
// Template version does NOT work across DLLs.
//====================================================================
#define  VW_DECLARE_SINGLETON( ClassName ) \
	public: \
		static  ClassName&  Instance() { \
			VW_ASSERT(s_pInstance##ClassName, \
					std::string(VW_GET_CLASS_NAME(ClassName)) + " singleton is not created."); \
			return *s_pInstance##ClassName; \
		} \
		static  ClassName*  InstancePtr() { \
			VW_ASSERT(s_pInstance##ClassName, \
					std::string(VW_GET_CLASS_NAME(ClassName)) + " singleton is not created."); \
			return s_pInstance##ClassName; \
		} \
	private: \
		static  ClassName*  s_pInstance##ClassName;


//====================================================================
#define  VW_IMPLEMENT_SINGLETON( ClassName ) \
	ClassName*  ClassName::s_pInstance##ClassName = 0;


//====================================================================
#define  VW_IMPLEMENT_SINGLETON_CTOR( ClassName ) \
	VW_ASSERT(! s_pInstance##ClassName, \
			std::string(VW_GET_CLASS_NAME(ClassName)) + " singleton already created."); \
	s_pInstance##ClassName = static_cast<ClassName*>(this);


//====================================================================
#define  VW_IMPLEMENT_SINGLETON_DTOR( ClassName ) \
	VW_ASSERT(s_pInstance##ClassName, "Destructing non-existent " + \
			std::string(VW_GET_CLASS_NAME(ClassName)) + " singleton."); \
	s_pInstance##ClassName = 0;


//********************************************************************
}//namespace vertexware

