

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\CompilerConfig.h"
#include "BaseTools\Types.h"
#include "BaseTools\Time.h"
#include <fstream>
#include <sstream>
#include <vector>


//********************************************************************
namespace vertexware {


//====================================================================
// Output stream manipulator: prints newline without executing flush.
//====================================================================
__forceinline  std::ostream& __cdecl nl(std::ostream& os) {
	return  os.put('\n');
}


//====================================================================
// Output stream manipulator: end-paragraph (double-newline & flush).
//====================================================================
__forceinline  std::ostream& __cdecl endpara(std::ostream& os) {
	return  os.put('\n') << std::endl;
}


//====================================================================
// Output stream manipulator: prints space.
//====================================================================
__forceinline  std::ostream& __cdecl spc(std::ostream& os) {
	return  os.put(' ');
}


//====================================================================
// Output stream manipulator: prints tab.
//====================================================================
__forceinline  std::ostream& __cdecl tab(std::ostream& os) {
	return  os.put('\t');
}


//====================================================================
// Base-class for output stream effectors.
//====================================================================
class VW_COMMON_API OStreamEffector
{
public:
	virtual ~OStreamEffector() {
	}
	__forceinline std::ostream& ApplyTo(std::ostream& os) {
		DoEffect(os);
		return os;
	}

protected:
	virtual void DoEffect(std::ostream& os) = 0;
};


//====================================================================
// Stream manipulator: applies any output stream effector to stream.
//====================================================================
__forceinline  std::ostream& __cdecl operator<<(std::ostream& os, OStreamEffector& effect) {
	return  effect.ApplyTo(os);
}


//====================================================================
// Output stream effector: prints separator line.
//====================================================================
class VW_COMMON_API SeparatorLineEffector : public OStreamEffector
{
public:
	SeparatorLineEffector(uint lineLength, char fillCharacter)
	{
		lineLen  = lineLength;
		fillChar = fillCharacter;
	}
protected:
	virtual void DoEffect(std::ostream& os)
	{
		for(uint i = lineLen; i > 0; i--) os << fillChar;
		os << std::endl;
	}
private:
	uint  lineLen;
	char  fillChar;
};


//====================================================================
__forceinline  SeparatorLineEffector __fastcall
separator_line(uint lineLength = 70, char fillCharacter = '-')
{
	return SeparatorLineEffector(lineLength, fillCharacter);
}


//====================================================================
// Output stream effector: prints fill-characters.
//====================================================================
class VW_COMMON_API FillEffector : public OStreamEffector
{
public:
	FillEffector(uint fillLength, char fillCharacter)
	{
		fillLen  = fillLength;
		fillChar = fillCharacter;
	}
protected:
	virtual void DoEffect(std::ostream& os)
	{
		for(uint i = fillLen; i > 0; i--) os << fillChar;
	}
private:
	uint  fillLen;
	char  fillChar;
};


//====================================================================
__forceinline  FillEffector __fastcall
fill(uint fillLength = 1, char fillCharacter = ' ')
{
	return FillEffector(fillLength, fillCharacter);
}


//====================================================================
// FillEffector specialization, prints multiple tabs.
//====================================================================
class VW_COMMON_API TabEffector : public FillEffector
{
public:
	TabEffector(uint fillLength) : FillEffector(fillLength, '\t') {}
};


//====================================================================
__forceinline  TabEffector __fastcall
tab(uint fillLength = 1)
{
	return TabEffector(fillLength);
}


//====================================================================
// Output stream effector: prints title.
//====================================================================
class VW_COMMON_API TitleEffector : public OStreamEffector
{
public:
	TitleEffector(const std::string& title, uint lineLength,
							char fillCharacter, bool bCenter) : sTitle(title)
	{
		lineLen  = lineLength;
		fillChar = fillCharacter;
		center   = bCenter;
	}
	TitleEffector& operator()(const std::string& title)
	{
		sTitle = title;
		return *this;
	}
protected:
	virtual void DoEffect(std::ostream& os)
	{
		FillEffector feLN(lineLen, fillChar);
		os << feLN << nl;
		if(center) {
			uint wsLen = (sTitle.length() < lineLen) ? (lineLen - sTitle.length()) / 2 : 0;
			os << fill(wsLen, ' ');
		}
		os << sTitle << nl << feLN << std::endl;
	}
private:
	std::string  sTitle;
	uint         lineLen;
	char         fillChar;
	bool         center;
};


//====================================================================
__forceinline  TitleEffector __fastcall
title(const std::string& title, uint lineLength,
		char fillCharacter = '-', bool bCenter = true)
{
	return TitleEffector(title, lineLength, fillCharacter, bCenter);
}


//====================================================================
// Output stream effector: prints caption.
//====================================================================
class VW_COMMON_API CaptionEffector : public OStreamEffector
{
public:
	CaptionEffector(const std::string& caption,
					uint fenceLength, char fillCharacter,
					uint whitespaceLength) :
		sCaption(caption)
	{
		fenceLen = fenceLength;
		fillChar = fillCharacter;
		wsLen    = whitespaceLength;
	}
	CaptionEffector& operator()(const std::string& caption)
	{
		sCaption = caption;
		return *this;
	}
protected:
	virtual void DoEffect(std::ostream& os)
	{
		FillEffector feFENCE(fenceLen, fillChar);
		FillEffector feWS(wsLen, ' ');
		os << feFENCE << feWS << sCaption << feWS << feFENCE << std::endl;
	}
private:
	std::string  sCaption;
	uint         fenceLen;
	char         fillChar;
	uint         wsLen;
};


//====================================================================
__forceinline  CaptionEffector __fastcall
caption(const std::string& caption, uint fenceLength = 5,
		char fillCharacter = '-', uint whitespaceLength = 1)
{
	return CaptionEffector(caption, fenceLength, fillCharacter, whitespaceLength);
}


//====================================================================
// Output stream.
// Defines additional states.
//====================================================================
class VW_COMMON_API OutputStream : public std::ostream
{
public:
	enum eOutputType
	{
		OutputType_Info = 0,
		OutputType_Warning,
		OutputType_Error,
		OutputType_Attention,
		OutputType_Query,
		OutputType_Echo,
		OutputType_Debug,

		OutputType_Count,
		OutputType_Force32Bits = 0x7FFFFFFF
	};


public:
	virtual void  SetCurrentOutputType(uint outputType = OutputType_Info) {}


protected:
	explicit OutputStream(std::streambuf* buf) : std::basic_ostream<char>(buf) {
	}
	virtual ~OutputStream() {
		flush();
	}
};


//====================================================================
// Output stream effector: sets output type for OutputStream.
// It has no effect on non-OutputStream streams.
//====================================================================
class VW_COMMON_API SetOutputTypeEffector : public OStreamEffector
{
public:
	SetOutputTypeEffector(uint _outputType) : outputType(_outputType) {}

protected:
	virtual void DoEffect(std::ostream& os) {
		OutputStream* p = dynamic_cast<OutputStream*>(&os);
		if(p) p->SetCurrentOutputType(outputType);
	}
private:
	uint  outputType;
};


//====================================================================
__forceinline  SetOutputTypeEffector __fastcall
set_output_type(uint outputType) { return SetOutputTypeEffector(outputType); }


//====================================================================
// Predefined SetCurrentUserDataEffectors.
//====================================================================
VW_COMMON_API extern  SetOutputTypeEffector  e_info;
VW_COMMON_API extern  SetOutputTypeEffector  e_warning;
VW_COMMON_API extern  SetOutputTypeEffector  e_error;
VW_COMMON_API extern  SetOutputTypeEffector  e_attention;
VW_COMMON_API extern  SetOutputTypeEffector  e_query;
VW_COMMON_API extern  SetOutputTypeEffector  e_echo;
VW_COMMON_API extern  SetOutputTypeEffector  e_debug;


//====================================================================
// Controlled stream.
// Supports automatic output to multiple streams.
//====================================================================
class VW_COMMON_API ControlledOutputStream : public OutputStream
{
protected:
	typedef  std::vector<std::ostream*>  OSTREAM_PTR_VEC;
	typedef  OSTREAM_PTR_VEC::iterator   OSTREAM_PTR_VEC_ITER;


private:
	//==========================
	// Controlled stream buffer.
	//==========================
	class VW_COMMON_API StreamBuffer : public std::stringbuf
	{
	public:
		OSTREAM_PTR_VEC  targets;

	protected:
		//--------------------------
		// std::stringbuf overrides.
		//--------------------------
		virtual int sync();
	};


public:
	ControlledOutputStream() : OutputStream(&buffer) {}

	virtual void  SetCurrentOutputType(uint outputType);

	virtual void  AddOutputTarget(std::ostream& os);
	virtual void  RemoveOutputTarget(std::ostream& os);
	virtual void  RemoveAllOutputTargets();


private:
	StreamBuffer  buffer;


protected:
	OSTREAM_PTR_VEC&  GetOutputTargets() { return buffer.targets; }

	//--------------------
	// std::ios overrides.
	//--------------------
	virtual StreamBuffer* rdbuf() const {
		return const_cast<StreamBuffer*>(&buffer);
	}
};


//====================================================================
// Log-file output-stream.
// Log-file with overridable prolog & epilog.
//====================================================================
class VW_COMMON_API LogFileStream : public std::ostream
{
public:
	LogFileStream(const std::string& logName, const std::string& fileName);
	virtual ~LogFileStream();

	void  Close();


protected:
	void  OutputBlockPE(bool bProlog, const time::DateTimeSeparators& dts = time::DateTimeSeparators());

	const std::string&  GetLogName() const { return logName; }
	const std::string&  GetFileName() const { return fileName; }


	virtual  void  OnOpen() {
		OutputBlockPE(true);
		*this << nl;
	}

	virtual  void  OnClose() {
		*this << nl;
		OutputBlockPE(false);
	}


private:
	std::string   logName;
	std::string   fileName;
	std::filebuf  filebuf;
};


//====================================================================
// After applying this sentry to any standard stream, stream will
// throw standard library exception std::ios_base::failure
// if an error occurs. Valid only until the end of definition scope.
//
// Example:
//     std::ofstream  os;
//     //no exceptions thrown on errors
//     {
//         try {
//             vertexware::StreamExceptionSentry sentry(os);
//             //now throws exceptions on errors
//             ...
//             //do stuff with stream
//         }
//         catch(std::ios_base::failure) {
//         }
//     }
//     //no exceptions thrown on errors
//
//====================================================================
class VW_COMMON_API StreamExceptionSentry
{
public:
	explicit StreamExceptionSentry(std::ios& stream, std::ios_base::io_state activeExceptions = All) :
		targetStream(stream)
	{
		lastActiveExceptions = targetStream.exceptions();
		targetStream.exceptions(activeExceptions);
	}
	~StreamExceptionSentry() {
		targetStream.exceptions(lastActiveExceptions);
	}


public:
	static  const  std::ios_base::io_state  Eof;
	static  const  std::ios_base::io_state  Fail;
	static  const  std::ios_base::io_state  Bad;
	static  const  std::ios_base::io_state  All;


private:
	std::ios&                targetStream;
	std::ios_base::io_state  lastActiveExceptions;
};


//====================================================================
// Guards standard stream format flags.
//====================================================================
class VW_COMMON_API StreamFormatFlagsSentry
{
public:
	explicit StreamFormatFlagsSentry(std::ios& stream) : targetStream(stream) {
		lastFlags = targetStream.flags();
	}
	~StreamFormatFlagsSentry() {
		targetStream.flags(lastFlags);
	}


private:
	std::ios&                targetStream;
	std::ios_base::fmtflags  lastFlags;
};


//====================================================================
// Simpler variable output for quick debugging.
//====================================================================
#define  VW_OUTVAR(ostream, var) \
			ostream << e_debug << #var << " = " << var << std::endl << e_info;


//********************************************************************
}//namespace vertexware


