

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Exception.h"
#include "BaseTools\SyncPrimitives.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace time {


//====================================================================
class VW_COMMON_API DateTimeSeparators
{
private:
	char  dateSep;
	char  timeSep;
	char  dtSep;


public:
	DateTimeSeparators() { Set(); }
	DateTimeSeparators(char date, char time, char dt) { Set(date, time, dt); }

	void  Set(char date = '-', char time = ':', char dt = '/')
	{
		dateSep = date;
		timeSep = time;
		dtSep   = dt;
	}

	char  GetDateSep() const { return dateSep; }
	char  GetTimeSep() const { return timeSep; }
	char  GetDTSep() const { return dtSep; }
};


//====================================================================
VW_COMMON_API
void __fastcall  GetFormattedSystemDate(std::string& out,
								bool shortYear = false, bool weekDay = false,
								const DateTimeSeparators& dts = DateTimeSeparators());


//====================================================================
VW_COMMON_API
void __fastcall  GetFormattedSystemTime(std::string& out,
								bool msecs = false,
								const DateTimeSeparators& dts = DateTimeSeparators());


//====================================================================
VW_COMMON_API
void  __fastcall  GetFormattedSystemDateTime(std::string& out,
								bool shortYear = false, bool weekDay = false,
								bool msecs = false,
								const DateTimeSeparators& dts = DateTimeSeparators());


//====================================================================
VW_COMMON_API
void __fastcall  GetSystemDate(uint& year, uint& month, uint& day, uint& weekDay);


//====================================================================
VW_COMMON_API
void __fastcall  GetSystemTime(uint& hour, uint& min, uint& sec, uint& msec);


//====================================================================
VW_COMMON_API
void __fastcall  GetSystemDateTime(uint& year, uint& month, uint& day, uint& weekDay,
									uint& hour, uint& min, uint& sec, uint& msec);


//********************************************************************
}//namespace time


//====================================================================
// Measures time directly by using the CPU-counter.
//====================================================================
class VW_COMMON_API HighResolutionTimer
{
public:
	//-----------------------------------------------------
	// Calls Start internally.
	// Throws <SystemResourceException> on failure.
	//-----------------------------------------------------
	HighResolutionTimer();

	void  Start() {
		::QueryPerformanceCounter(&liStartTicks);
	}

	double  GetElapsedSeconds() {
		LARGE_INTEGER  liStopTicks;
		::QueryPerformanceCounter(&liStopTicks);
		return (double(liStopTicks.QuadPart - liStartTicks.QuadPart) * dOneOverTicksPerSec);
	}

	uint  GetElapsedTicks() {
		LARGE_INTEGER  liStopTicks;
		::QueryPerformanceCounter(&liStopTicks);
		return (liStopTicks.QuadPart - liStartTicks.QuadPart);
	}

	void  GetElapsedTicks(uint& highPart, uint& lowPart);

	uint  GetTicksPerSecond() {
		return (liTicksPerSec.QuadPart);
	}


public:
	static bool __fastcall GetPerformanceCounterInfo(LARGE_INTEGER* pliTicksPerSec = 0,
														double* pdTicksPerSec = 0);


private:
	LARGE_INTEGER  liStartTicks;
	LARGE_INTEGER  liTicksPerSec;
	double         dOneOverTicksPerSec;
};


//====================================================================
// Measures thread-time exactly (independent of system-load).
// Use for loop-profiling or for obtaining the total running time
// of the thread. System does not update internal structures
// too frequently, so don't use it for very-short sections
// of code - it will probably return 0.
//====================================================================
class VW_COMMON_API ThreadTimer
{
public:
	ThreadTimer() {
		Start();
	}

	void  Start() {
		FILETIME  dummy;
		::GetThreadTimes(GetCurrentThread(), &dummy, &dummy,
			reinterpret_cast<FILETIME*>(&startKernel), reinterpret_cast<FILETIME*>(&startUser));
	}

	//-----------------------------------------------------------
	// Returns total thread-time.
	// If specified, output parameters return kernel & user time.
	//-----------------------------------------------------------
	double  GetElapsedSeconds(double* pOutKernelTime = 0, double* pOutUserTime = 0);


private:
	ULARGE_INTEGER  startKernel;
	ULARGE_INTEGER  startUser;
};


//====================================================================
// Low-priority alarm-timer bound to the window.
// Windows-message-loop must be active.
//====================================================================
class WindowAlarmTimer;


//====================================================================
class VW_COMMON_API WindowAlarmHandler
{
public:
	virtual void OnAlarm(WindowAlarmTimer& wat) = 0;
};


//====================================================================
class VW_COMMON_API WindowAlarmTimer
{
public:
	explicit WindowAlarmTimer(HWND _hwnd = 0) {
		hwnd = _hwnd;
		timerID = 0;
		pWindowAlarmHandler = 0;
	}
	~WindowAlarmTimer() {
		StopPeriodic();
	}
	void  SetHWND(HWND _hwnd) {
		hwnd = _hwnd;
	}
	bool  StartPeriodic(WindowAlarmHandler& wah, uint msecsPeriod) {
		if(! hwnd) return false;
		pWindowAlarmHandler = &wah;
		timerID = ::SetTimer(hwnd, reinterpret_cast<UINT>(this), msecsPeriod, TimerMapper);
		return (timerID != 0);
	}
	void  StopPeriodic() {
		if(timerID) {
			::KillTimer(hwnd, reinterpret_cast<UINT>(this));
			timerID = 0;
			pWindowAlarmHandler = 0;
		}
	}


private:
	static void CALLBACK TimerMapper(HWND, UINT, UINT_PTR eventID, DWORD) {
		WindowAlarmTimer* wat = reinterpret_cast<WindowAlarmTimer*>(eventID);
		WindowAlarmHandler* wah = wat->pWindowAlarmHandler;
		if(wah) {
			wah->OnAlarm(*wat);
		}
	}


private:
	HWND  hwnd;
	UINT  timerID;
	WindowAlarmHandler*  pWindowAlarmHandler;
};


//====================================================================
// Alarm-timer & alarm-handler interface.
// Alarm-timer runs in system-created-thread, which means that
// alarm-handler's work-function is executed by one of the
// system-created-threads (though not always the same one).
//====================================================================
class AlarmTimer;


//====================================================================
class VW_COMMON_API AlarmHandler
{
public:
	//----------------------------------------------------------------
	// Return true if work-function takes relatively long time to execute.
	//----------------------------------------------------------------
	virtual bool QueryHintLongExecution() { return false; }

	//-------------------------------------------
	// Work-function.
	// Return true to deactivate periodic alarms.
	// (you cannot call StopPeriodic() directly)
	//-------------------------------------------
	virtual bool OnAlarm(const AlarmTimer& at) = 0;
};


//====================================================================
class VW_COMMON_API AlarmTimer
{
public:
	AlarmTimer();
	~AlarmTimer();

	//--------------------------------------------
	// Returns true if timer is currently running.
	//--------------------------------------------
	bool  IsRunning();

	//---------------------------------------------------
	// Fires one-shot alarm.
	// <msecs> indicates how many millisecons should pass
	// before the alarm fires. Returns false on fail.
	//---------------------------------------------------
	bool  FireOnce(AlarmHandler& ah, uint msecs);

	//---------------------------------------------------
	// Starts periodic alarm.
	// <msecsFirst> indicates how many millisecons should
	// pass before the first alarm fires. <msecsPeriod>
	// indicates the period between consecutive alarms.
	// Returns false on fail.
	//----------------------------------------------------
	bool  StartPeriodic(AlarmHandler& ah, uint msecsFirst, uint msecsPeriod);

	//---------------------------------------------------
	// Stops periodic alarm.
	// Function blocks until all queued asynchronous
	// alarm-handler-calls are completed. Never call from
	// alarm-handler' work-function directly, otherwise
	// deadlock will occur.
	//---------------------------------------------------
	void  StopPeriodic();


private:
	bool   Activate(AlarmHandler& ah, uint msecsFirst, uint msecsPeriod);
	void   Deactivate(bool bWaitAlarmHandlerCompletion);

	static void __stdcall  AlarmHandlerMapper(void* userData, BOOLEAN);


private:
	CriticalSection  csHandle;
	CriticalSection  csDeactivator;
	AlarmHandler*    pAlarmHandler;
	HANDLE           hSystemTimer;
	bool             bOneShot;
};


//********************************************************************
}//namespace vertexware

