

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\String.h"
#include "BaseTools\Debug.h"
#include "private\vw_common_exports.h"
#include <vector>


//********************************************************************
namespace vertexware {


//====================================================================
class VW_COMMON_API VariableProxy
{
public:
	//---------------------------
	struct VW_COMMON_API Observer
	{
		virtual ~Observer() {}
		virtual void OnValueChanged(const VariableProxy& vp) {}
	};
	static  Observer  s_nullObserver;

	// setup global observer
	static void      __fastcall SetGlobalObserver(Observer* ob = 0) { s_pGlobalObserver = ob; }
	static Observer* __fastcall GetGlobalObserver() { return s_pGlobalObserver; }


public:
	VariableProxy();
	virtual ~VariableProxy();

	void  SetDeveloperModeOnly(bool bDevModeOnly = false) { bDeveloperModeOnly = bDevModeOnly; }
	bool  IsDeveloperModeOnly() const { return bDeveloperModeOnly; }

	void                SetName(const std::string& s) { name = s; }
	const std::string&  GetName() const { return name; }

	void                SetVarDescription(const std::string& s) { varDesc = s; }
	const std::string&  GetVarDescription() const { return varDesc; }

	// setup per-object observer
	void       SetObserver(Observer* ob = 0) { pObserver = ob; }
	Observer*  GetObserver() const { return pObserver; }

	//----------------------------------------------------------
	// If per-object observer is available it is used,
	// Else if global observer is available it is used.
	//----------------------------------------------------------
	void  NotifyObserverAboutValueChange();

	//------------------------------
	// Implemented by derived-class.
	//------------------------------
	virtual  void         SetValueFromString(const std::string& s) = 0;
	virtual  std::string  GetValueAsString() const = 0;

	virtual  VariableProxy*  Clone() const = 0;
	virtual  std::string     GetTargetTypeName() const = 0;
	virtual  std::string     GetTargetRestrictions() const { return ""; }
	virtual  std::string     GetTargetObjectName() const = 0;


private:
	static  Observer*  s_pGlobalObserver;
	Observer*          pObserver;
	bool               bDeveloperModeOnly;
	std::string        name;
	std::string        varDesc;
};


//====================================================================
template<typename TTarget>
class TVariableProxy : public VariableProxy
{
public:
	TVariableProxy(TTarget* _pTarget = 0,
			const std::string& _name = "", const std::string& _varDescription = "",
			bool _bDevModeOnly = false, Observer* _pObserver = 0)
	{
		pTarget = _pTarget;
		SetDeveloperModeOnly(_bDevModeOnly);
		SetName(_name);
		SetVarDescription(_varDescription);
		SetObserver(_pObserver);
	}

	void      SetTarget(TTarget* _pTarget = 0) { pTarget = _pTarget; }
	TTarget*  GetTarget() const { return pTarget; }

	void  SetValue(const TTarget& value) {
		if(pTarget) {
			*pTarget = value;
			NotifyObserverAboutValueChange();
		}
	}

	virtual  void  SetValueFromString(const std::string& s) {
		if(pTarget) {
			ToValue(s, *pTarget, *pTarget);
			SetValue(*pTarget);
		}
	}

	virtual  std::string  GetValueAsString() const {
		if(! pTarget) return "Unknown";
		return ToString(*pTarget);
	}

	virtual  VariableProxy*  Clone() const {
		return new TVariableProxy(pTarget, GetName(), GetVarDescription(),
									IsDeveloperModeOnly(), GetObserver());
	}

	virtual  std::string  GetTargetTypeName() const {
		return VW_GET_CLASS_NAME(TTarget);
	}

	virtual  std::string  GetTargetObjectName() const {
		if(! pTarget) return "NA";
		return VW_GET_OBJECT_NAME(*pTarget);
	}


protected:
	TTarget*  pTarget;
};


//====================================================================
// Typedef is sufficient, this is only to keep the generated code
// within the library-module.
//====================================================================
#define  VW_DECLARE_T_VARIABLE_PROXY(ClassName, TTarget) \
\
class VW_COMMON_API ClassName : public TVariableProxy<TTarget> \
{ \
public: \
	ClassName(TTarget* pTarget = 0, \
			const std::string& name = "", const std::string& varDescription = "", \
			bool bDevModeOnly = false, Observer* pObserver = 0) \
		: \
		TVariableProxy<TTarget>(pTarget, name, varDescription, bDevModeOnly, pObserver) \
	{ } \
};


//====================================================================
VW_DECLARE_T_VARIABLE_PROXY( VariableProxyVoidPointer, void* );
VW_DECLARE_T_VARIABLE_PROXY( VariableProxyBool, bool );
VW_DECLARE_T_VARIABLE_PROXY( VariableProxyInt, int );
VW_DECLARE_T_VARIABLE_PROXY( VariableProxyUInt, uint );
VW_DECLARE_T_VARIABLE_PROXY( VariableProxyFloat, float );
VW_DECLARE_T_VARIABLE_PROXY( VariableProxyDouble, double );
VW_DECLARE_T_VARIABLE_PROXY( VariableProxyString, std::string );


//====================================================================
#undef  VW_DECLARE_T_VARIABLE_PROXY


//====================================================================
template<typename TTarget>
class TVariableProxyRangeChecked : public TVariableProxy<TTarget>
{
public:
	TVariableProxyRangeChecked(TTarget* _pTarget = 0,
			TTarget _minValue = TTarget(0), TTarget _maxValue = TTarget(0),
			const std::string& _name = "", const std::string& _varDescription = "",
			bool _bDevModeOnly = false, Observer* _pObserver = 0)
		:
		TVariableProxy<TTarget>(_pTarget, _name, _varDescription, _bDevModeOnly, _pObserver)
	{
		SetMinValue(_minValue);
		SetMaxValue(_maxValue);
	}

	void  SetMinValue(TTarget _minValue = TTarget(0)) {
		minValue = _minValue;
		if(maxValue < minValue) maxValue = minValue;
	}
	void  SetMaxValue(TTarget _maxValue = TTarget(0)) {
		maxValue = _maxValue;
		if(minValue > maxValue) minValue = maxValue;
	}

	TTarget  GetMinValue() const { return minValue; }
	TTarget  GetMaxValue() const { return maxValue; }

	void  SetValueRangeChecked(TTarget value) {
		if(pTarget) {
			*pTarget = value;
			if(value <= minValue) *pTarget = minValue;
			if(value >= maxValue) *pTarget = maxValue;
			NotifyObserverAboutValueChange();
		}
	}

	virtual  void  SetValueFromString(const std::string& s) {
		if(pTarget) {
			ToValue(s, *pTarget, *pTarget);
			SetValueRangeChecked(*pTarget);
		}
	}

	virtual  VariableProxy*  Clone() const {
		return new TVariableProxyRangeChecked(pTarget, minValue, maxValue,
											GetName(), GetVarDescription(),
											IsDeveloperModeOnly(), GetObserver());
	}

	virtual  std::string  GetTargetRestrictions() const {
		return "range=[" + ToString(minValue) + ", " + ToString(maxValue) + "]";
	}


protected:
	TTarget  minValue;
	TTarget  maxValue;
};


//====================================================================
// Typedef is sufficient, this is only to keep the generated code
// within the library-module.
//====================================================================
#define  VW_DECLARE_T_VARIABLE_PROXY_RANGE_CHECKED(ClassName, TTarget) \
\
class VW_COMMON_API ClassName : public TVariableProxyRangeChecked<TTarget> \
{ \
public: \
	ClassName(TTarget* pTarget = 0, \
			TTarget minValue = TTarget(0), TTarget maxValue = TTarget(0), \
			const std::string& name = "", const std::string& varDescription = "", \
			bool bDevModeOnly = false, Observer* pObserver = 0) \
		: \
		TVariableProxyRangeChecked<TTarget>(pTarget, minValue, maxValue, \
							name, varDescription, bDevModeOnly, pObserver) \
	{ } \
};


//====================================================================
VW_DECLARE_T_VARIABLE_PROXY_RANGE_CHECKED( VariableProxyRangeCheckedInt, int );
VW_DECLARE_T_VARIABLE_PROXY_RANGE_CHECKED( VariableProxyRangeCheckedUInt, uint );
VW_DECLARE_T_VARIABLE_PROXY_RANGE_CHECKED( VariableProxyRangeCheckedFloat, float );
VW_DECLARE_T_VARIABLE_PROXY_RANGE_CHECKED( VariableProxyRangeCheckedDouble, double );


//====================================================================
#undef  VW_DECLARE_T_VARIABLE_PROXY_RANGE_CHECKED


//====================================================================
// Registered variables are copied to internal objects.
// Do not change the names after the variables are registered.
//====================================================================
class VW_COMMON_API VariableRegistry
{
public:
	VariableRegistry(const std::string& registryName = "");
	virtual ~VariableRegistry();

	void                SetRegistryName(const std::string& s) { registryName = s; }
	const std::string&  GetRegistryName() const { return registryName; }

	// returns false if variable with the same name is already registered
	bool  RegisterVariable(const VariableProxy& vp);
	void  UnregisterVariable(const std::string& name);
	void  UnregisterAllVariables();

	void  ForceValueRanges();

	// returns 0 if not found
	VariableProxy*  GetVariable(const std::string& name);

	uint            GetNumVariables();
	VariableProxy&  GetVariable(uint i);


private:
	VariableProxy*  Find(const std::string& name);
	void            SortVariables();


private:
	typedef  std::vector<VariableProxy*>  VAR_VEC;
	typedef  VAR_VEC::iterator            VAR_ITER;
	typedef  VAR_VEC::const_iterator      VAR_CITER;

	VAR_VEC      variables;
	std::string  registryName;
};


//********************************************************************
}//namespace vertexware

