

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Color.h"
#include "BaseTools\AutoCleanup.h"
#include "private\vw_common_exports.h"
#include <iosfwd>
#include <vector>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace win_utils {


//====================================================================
// GDI / GDI+.
//
// NOTES:
// GDI+ creates additional private thread after calling GdiplusStartup().
// Thread is destroyed after calling GdiplusShutdown().
//====================================================================
class VW_COMMON_API GdiplusGuard
{
public:
	GdiplusGuard();
	~GdiplusGuard();


private:
	ULONG_PTR  token;
};


//====================================================================
// Use to find-out wheter the required font is installed.
//====================================================================
VW_COMMON_API
void __fastcall GdiplusGetFontFamilyNames(std::vector<std::string>& outFontFamilyNames);


//====================================================================
// Colors.
//====================================================================
VW_COMMON_API
__forceinline Gdiplus::Color __fastcall GdiplusConvertColor(color4b c)
{
	return Gdiplus::Color(c.a, c.r, c.g, c.b);
}


//====================================================================
VW_COMMON_API
__forceinline color4b __fastcall GdiplusConvertColor(Gdiplus::Color c)
{
	return color4b(c.GetR(), c.GetG(), c.GetB(), c.GetA());
}


//====================================================================
enum eSystemColor
{
	SystemColor_Desktop = 0,

	SystemColor_Count,
	SystemColor_Force32Bits = 0x7FFFFFFF
};


//====================================================================
VW_COMMON_API
Gdiplus::Color __fastcall GetSystemColor(eSystemColor sc);


//====================================================================
class VW_COMMON_API Point : public POINT
{
public:
	__forceinline Point() {
	}

	__forceinline Point(LONG _x, LONG _y) {
		this->Set(_x, _y);
	}
	__forceinline void Set(LONG _x, LONG _y) {
		x = _x;
		y = _y;
	}

	__forceinline Point(const POINT& p) {
		this->operator=(p);
	}
	__forceinline Point& operator=(const POINT& p) {
		*(static_cast<POINT*>(this)) = p;
		return *this;
	}

	// <lparam> contains mouse coordinates from WM_MOUSE* messages
	__forceinline Point(const LPARAM& lparam) {
		*this = lparam;
	}
	__forceinline Point& operator=(const LPARAM& lparam) {
		x = static_cast<short>(LOWORD(lparam));
		y = static_cast<short>(HIWORD(lparam));
		return *this;
	}

	__forceinline Point& operator+=(const Point& rhs) {
		x += rhs.x;
		y += rhs.y;
		return *this;
	}
	__forceinline Point operator+(const Point& rhs) const {
		return Point(x + rhs.x, y + rhs.y);
	}

	__forceinline Point& operator-=(const Point& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		return *this;
	}
	__forceinline Point operator-(const Point& rhs) const {
		return Point(x - rhs.x, y - rhs.y);
	}

	bool  HitTestRectangle(LONG rx, LONG ry, LONG rw, LONG rh) const {
		if(x < rx  ||  x >= rx + rw) return false;
		if(y < ry  ||  y >= ry + rh) return false;
		return true;
	}
};


//====================================================================
// Resources.
//====================================================================
enum eStockCursor
{
	StockCursor_Normal = 0,
	StockCursor_IBeam,
	StockCursor_Wait,
	StockCursor_Cross,
	StockCursor_Up,
	StockCursor_SizeNWSE,
	StockCursor_SizeNESW,
	StockCursor_SizeWE,
	StockCursor_SizeNS,
	StockCursor_SizeAll,
	StockCursor_No,
	StockCursor_Hand,
	StockCursor_AppStarting,

	StockCursor_Count,
	StockCursor_Force32Bits = 0x7FFFFFFF
};


//====================================================================
VW_COMMON_API
HCURSOR __fastcall LoadStockCursor(eStockCursor stockCursorType);


//====================================================================
// Message-loops.
//====================================================================
enum  eDispatchResult
{
	DispatchResult_MessageNotAvailable = 0,
	DispatchResult_FoundQuitMessage    = 1,
	DispatchResult_MessageDispatched   = 2,

	DispatchResult_Force32Bits = 0x7FFFFFFF //not used
};


//====================================================================
// Dispatches messages for all windows owned by the calling thread.
//====================================================================
__forceinline eDispatchResult __fastcall DispatchOneMessage(MSG& msg)
{
	if(::PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
		if(msg.message == WM_QUIT) {
			return DispatchResult_FoundQuitMessage;
		}
		::TranslateMessage(&msg);
		::DispatchMessage(&msg);
		return DispatchResult_MessageDispatched;
	}
	return DispatchResult_MessageNotAvailable;
}


//====================================================================
// Dispatches all available messages, then sleeps until a new
// message arrives.
//====================================================================
VW_COMMON_API
void __fastcall RunStandardMessageLoop();


//====================================================================
// COM.
//====================================================================
class VW_COMMON_API COMGuard
{
public:
	COMGuard();
	~COMGuard();


private:
	bool  bCleanup;
};


//====================================================================
template<typename T>
__forceinline void __fastcall SafeRelease(T* &pIUnknown)
{
	if(pIUnknown) {
		pIUnknown->Release();
		pIUnknown = 0;
	}
}


//====================================================================
struct VW_COMMON_API CleanupHandlerIUnknown
{
	__forceinline  void  operator()(IUnknown& target) {
		target.Release();
	}
};


//====================================================================
struct VW_COMMON_API AutoCleanupIUnknown : public AutoCleanup<IUnknown, CleanupHandlerIUnknown>
{
	AutoCleanupIUnknown(IUnknown* pTarget = 0, bool bEnabled = true)
		: AutoCleanup<IUnknown, CleanupHandlerIUnknown>(pTarget, bEnabled)
	{ }
};


//====================================================================
template<typename T>
class COMPtr
{
public:
	COMPtr(T* _p = 0, bool bAddRef = false) {
		p = _p;
		if(p  &&  bAddRef) p->AddRef();
	}

	COMPtr(const COMPtr<T>& rhs) {
		p = rhs.p;
		if(p) p->AddRef();
	}

	~COMPtr() {
		if(p) p->Release();
	}

	COMPtr&  operator=(T* _p) {
		Assign(_p);
		return *this;
	}

	COMPtr&  operator=(const COMPtr<T>& rhs) {
		Assign(rhs.p);
		return *this;
	}

	void  Release() {
		if(p) p->Release();
		p = 0;
	}

	void  Assign(T* _p = 0) {
		if(p == _p) return;
		if(p) p->Release();
		p = _p;
		if(p) p->AddRef();
	}

	T*    Get() const { return p; }
	void  Set(T* _p = 0) { p = _p; }

	T&   operator*() const { return *p; }
	T**  operator&() { return &p; }
	T*   operator->() const { return p; }

	operator  T*() const { return p; }

	bool  operator!() const { return (! p); }
	bool  operator==(T* _p) const { return (p == _p); }
	bool  operator!=(T* _p) const { return (p != _p); }


private:
	T*  p;
};


//====================================================================
class VW_COMMON_API Guid : public GUID
{
public:
	Guid(); //set all to 0

	Guid(DWORD data1, WORD data2, WORD data3,
		BYTE data40, BYTE data41, BYTE data42, BYTE data43,
		BYTE data44, BYTE data45, BYTE data46, BYTE data47);

	Guid(const GUID& rhs) {
		this->operator=(rhs);
	}
	Guid& operator=(const GUID& rhs) {
		*(static_cast<GUID*>(this)) = rhs;
		return *this;
	}

	Guid(const UID& rhs) {
		this->operator=(rhs);
	}
	Guid& operator=(const UID& rhs) {
		CopyFrom(rhs);
		return *this;
	}

	bool operator==(const Guid& rhs) const;
	bool operator!=(const Guid& rhs) const { return !(*this == rhs); }

	// vertexware-UID <===> windows-GUID
	void  CopyFrom(const UID& uid);
	void  CopyTo(UID& uid) const;
};


//********************************************************************
}//namespace win_utils


//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const win_utils::Guid& x);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, win_utils::Guid& x);


//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<<(std::ostream &os, const GUID& x);


//********************************************************************
}//namespace vertexware


