

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Memory.h"
#include "BaseTools\Debug.h"
#include "BaseTools\WinUtils.h"
#include "D3D8Tools\ErrorHandling.h"
#include "private\vw_direct3d8_exports.h"
#include <d3d8.h>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace d3d8_tools {


//====================================================================
template<typename TVertex>
class AdaptiveVertexBuffer : public AdaptiveBufferBase
{
public:
	typedef  TVertex  VertexType;


public:
	// Usage, FVF and Pool are used when calling CreateVertexBuffer internally
	explicit AdaptiveVertexBuffer(IDirect3DDevice8& _device,
							DWORD Usage = D3DUSAGE_WRITEONLY, DWORD FVF = 0,
							D3DPOOL Pool = D3DPOOL_DEFAULT,
							uint minSize = 0, uint minStableCycles = 0)
		:
		device(_device)
	{
		usage = Usage;
		fvf   = FVF;
		pool  = Pool;
		iD3DVertexBuffer = 0;

		SetMinSize(minSize);
		SetMinStableCycles(minStableCycles);
	}

	virtual ~AdaptiveVertexBuffer()
	{
		DoDeallocate();
	}

	// d3d-vertex-buffer properties
	DWORD    GetUsage() const { return usage; }
	DWORD    GetFVF() const { return fvf; }
	D3DPOOL  GetPool() const { return pool; }

	IDirect3DVertexBuffer8*  GetD3DVertexBuffer() const { return iD3DVertexBuffer; }

	// lock/unlock d3d-vertex-buffer
	__forceinline  HRESULT  Lock(UINT OffsetToLock, UINT SizeToLock, BYTE** ppbData, DWORD Flags)
	{
		VW_ASSERT(iD3DVertexBuffer, "Locking non-allocated D3D-vertex-buffer.");
		return iD3DVertexBuffer->Lock(OffsetToLock, SizeToLock, ppbData, Flags);
	}

	__forceinline  HRESULT  Unlock()
	{
		VW_ASSERT(iD3DVertexBuffer, "Unlocking non-allocated D3D-vertex-buffer.");
		return iD3DVertexBuffer->Unlock();
	}


protected:
	// create d3d-vertex-buffer
	virtual  void  DoAllocate(uint newSize)
	{
		HRESULT hr = device.CreateVertexBuffer(newSize * sizeof(TVertex),
									usage, fvf, pool, &iD3DVertexBuffer);
		CheckForD3DError<D3D8Exception>(hr, "IDirect3DDevice8::CreateVertexBuffer");
	}

	// release d3d-vertex-buffer
	virtual  void  DoDeallocate()
	{
		win_utils::SafeRelease(iD3DVertexBuffer);
	}


private:
	IDirect3DDevice8&        device;
	DWORD                    usage;
	DWORD                    fvf;
	D3DPOOL                  pool;
	IDirect3DVertexBuffer8*  iD3DVertexBuffer;
};


//********************************************************************
}//namespace d3d8_tools


//********************************************************************
}//namespace vertexware


