

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Singleton.h"
#include "BaseTools\HeapTracked.h"
#include "BaseTools\NamedParameter.h"
#include "BaseTools\Exception.h"
#include <vector>
#include <map>


//********************************************************************
namespace vertexware {


//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, DataCategory_UserDefined );


//====================================================================
class DataEncoder;
class DataDecoder;


//====================================================================
class VW_COMMON_API DataCodecBase : public HeapTracked
{
friend class DataEncoder;
friend class DataDecoder;

public:
	virtual ~DataCodecBase() {}

	uint                GetDataCategory() const { return dataCategory; }
	const std::string&  GetCodecName() const { return codecName; }
	const std::string&  GetFormatList() const { return formatList; }

private:
	DataCodecBase(uint dataCategory, const std::string& codecName,
									const std::string& formatList);

private:
	uint         dataCategory;
	std::string  codecName;
	std::string  formatList;
};


//====================================================================
// Base class for data-encoders.
// Encoder-name must be unique within data-category.
//====================================================================
class VW_COMMON_API DataEncoder : public DataCodecBase
{
protected:
	// format-list must be whitespace separated
	DataEncoder(uint dataCategory,
				const std::string& encoderName,
				const std::string& formatList);

public:
	//-------------------------------------------------------------
	// Return 0 on any kind of error.
	// If format is recognized: parse parameter-list and save data.
	// Assume default values for non-specified parameters.
	//-------------------------------------------------------------
	virtual  bool  SaveData(const void* data,
							const std::string& fileName,
							const std::string& fileFormat,
							const NamedParameterList& parameterList) = 0;
};


//====================================================================
// Base class for data-decoders.
// Decoder-name must be unique within data-category.
//====================================================================
class VW_COMMON_API DataDecoder : public DataCodecBase
{
protected:
	// format-list must be whitespace separated
	DataDecoder(uint dataCategory,
				const std::string& decoderName,
				const std::string& formatList);

public:
	//--------------------------------------------------------
	// Return 0 on any kind of error.
	// If format is recognized: parse parameter-list, create
	// suitable data-object, load data and return data-object.
	// Assume default values for non-specified parameters.
	//--------------------------------------------------------
	virtual  void*  LoadData(const std::string& fileName,
							const NamedParameterList& parameterList) = 0;
};


//********************************************************************
namespace data_codec_parameters {


//====================================================================
// Name...: FileOverwrite
// Type...: bool
// Range..: NA
// Default: false
// Support: Supported by all encoders
// Notes..: If <false> encoder must not overwrite existing file.
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_FileOverwrite );
VW_COMMON_API  VW_DECLARE_CONSTANT( bool, Value_FileOverwrite_Default );


//====================================================================
// Name...: ByteOrderBigEndian
// Type...: bool
// Range..: NA
// Default: machine-native (false on IA32)
// Support: Optionally supported by encoders
// Notes..: If <true> encoder should write values using the
//          big-endian byte-order where applicable.
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_ByteOrderBigEndian );
VW_COMMON_API  VW_DECLARE_CONSTANT( bool, Value_ByteOrderBigEndian_Default );


//====================================================================
// Name...: UseCompression
// Type...: bool
// Range..: NA
// Default: true
// Support: Optionally supported by lossless-compression encoders
// Notes..: NA.
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_UseCompression );
VW_COMMON_API  VW_DECLARE_CONSTANT( bool, Value_UseCompression_Default );


//====================================================================
// Name...: CompressionLevel
// Type...: uint
// Range..: [0-100]
// Default: 100
// Support: Optionally supported by lossless-compression encoders
// Notes..: 100 represents the highest possible compression.
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_CompressionLevel );
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, Value_CompressionLevel_Min );
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, Value_CompressionLevel_Max );
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, Value_CompressionLevel_Default );


//====================================================================
// Name...: Quality
// Type...: uint
// Range..: [0 - 100]
// Default: 100
// Support: Optionally supported by lossy-compression encoders
// Notes..: 100 represents the highest possible data quality.
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_Quality );
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, Value_Quality_Min );
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, Value_Quality_Max );
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, Value_Quality_Default );


//********************************************************************
}//namespace data_codec_parameters


//====================================================================
struct CodecParameterFactory;


//====================================================================
// Data-codec manager (singleton)
//====================================================================
class VW_COMMON_API DataCodecManager
{
VW_DECLARE_SINGLETON( DataCodecManager );

public:
	DataCodecManager();
	~DataCodecManager();


	//----------------------------------------------------------------
	// Save / load data through registered codecs. If <pParameterList>
	// is 0 category's default parameter-list is used.
	//----------------------------------------------------------------
	bool   SaveData(uint dataCategory,
					const void* data,
					const std::string& fileName,
					const std::string& fileFormat,
					const NamedParameterList* pParameterList = 0);

	void*  LoadData(uint dataCategory,
					const std::string& fileName,
					const NamedParameterList* pParameterList = 0);


	//------------------------------------------------------------
	// Codec management. All registered objects must be created on
	// heap. Once registered, codec is owned by DataCodecManager.
	//------------------------------------------------------------
	void  RegisterDataCategory(uint dataCategory);
	bool  IsDataCategoryRegistered(uint dataCategory);

	void  RegisterEncoder(DataEncoder* pEncoder);
	void  RegisterDecoder(DataDecoder* pDecoder);

	bool  IsEncoderRegistered(uint dataCategory, const std::string& encoderName);
	bool  IsDecoderRegistered(uint dataCategory, const std::string& decoderName);

	void  PrioritizeEncoders(uint dataCategory,
							const std::string& higherPriorityEncoderName,
							const std::string& lowerPriorityEncoderName);
	void  PrioritizeDecoders(uint dataCategory,
							const std::string& higherPriorityDecoderName,
							const std::string& lowerPriorityDecoderName);

	void  UnregisterEncoder(uint dataCategory, const std::string& encoderName);
	void  UnregisterDecoder(uint dataCategory, const std::string& decoderName);

	void  UnregisterAllEncoders(uint dataCategory);
	void  UnregisterAllDecoders(uint dataCategory);

	void  UnregisterAllEncoders();
	void  UnregisterAllDecoders();

	uint                GetNumEncoders(uint dataCategory);
	const DataEncoder&  GetEncoder(uint dataCategory, uint i);

	uint                GetNumDecoders(uint dataCategory);
	const DataDecoder&  GetDecoder(uint dataCategory, uint i);

	void  DumpRegisteredEncoders(uint dataCategory, std::ostream& os, uint tabLevel = 0);
	void  DumpRegisteredDecoders(uint dataCategory, std::ostream& os, uint tabLevel = 0);


	//----------------------------------------------
	// Default file-format / file-extension mapping.
	//----------------------------------------------

	// empty string for <fileExtension> erases mapping
	void  SetDefaultFileFormatExtension(uint dataCategory, const std::string& fileFormat,
						const std::string& fileExtension = "");

	// returns true if found
	bool  GetDefaultFileFormatExtension(uint dataCategory, const std::string& fileFormat,
						std::string& fileExtension, const std::string& failValueExtension);

	void  ClearDefaultFileFormatExtensions(uint dataCategory);
	void  ClearDefaultFileFormatExtensions();


	//------------------------------
	// Default parameter management.
	//------------------------------
	void  SetEncoderDefaultParamList(uint dataCategory, const NamedParameterList& npl);
	void  SetDecoderDefaultParamList(uint dataCategory, const NamedParameterList& npl);

	void  SetEncoderDefaultParamList(uint dataCategory, const CodecParameterFactory& cpf);
	void  SetDecoderDefaultParamList(uint dataCategory, const CodecParameterFactory& cpf);

	const NamedParameterList&  GetEncoderDefaultParamList(uint dataCategory) const;
	const NamedParameterList&  GetDecoderDefaultParamList(uint dataCategory) const;

	//-------------------------------------------------------------------------
	// Initial parameter management (used to initialize other parameter lists).
	//-------------------------------------------------------------------------
	void  ResetEncoderInitialParamList();
	void  ResetDecoderInitialParamList();

	void  SetEncoderInitialParamList(const NamedParameterList& npl);
	void  SetDecoderInitialParamList(const NamedParameterList& npl);

	const NamedParameterList&  GetEncoderInitialParamList() const;
	const NamedParameterList&  GetDecoderInitialParamList() const;


private:
	typedef  std::vector<DataEncoder*>  ENC_VEC;
	typedef  ENC_VEC::iterator          ENC_ITER;
	typedef  ENC_VEC::const_iterator    ENC_CITER;

	typedef  std::map<uint, ENC_VEC>      CAT_ENC_MAP;
	typedef  CAT_ENC_MAP::iterator        CAT_ENC_ITER;
	typedef  CAT_ENC_MAP::const_iterator  CAT_ENC_CITER;

	typedef  std::vector<DataDecoder*>  DEC_VEC;
	typedef  DEC_VEC::iterator          DEC_ITER;
	typedef  DEC_VEC::const_iterator    DEC_CITER;

	typedef  std::map<uint, DEC_VEC>      CAT_DEC_MAP;
	typedef  CAT_DEC_MAP::iterator        CAT_DEC_ITER;
	typedef  CAT_DEC_MAP::const_iterator  CAT_DEC_CITER;

	typedef  std::map<uint, NamedParameterList>  CAT_NPL_MAP;
	typedef  CAT_NPL_MAP::iterator               CAT_NPL_ITER;
	typedef  CAT_NPL_MAP::const_iterator         CAT_NPL_CITER;

	typedef  std::map<std::string, std::string>  FMT_EXT_MAP;
	typedef  FMT_EXT_MAP::iterator               FMT_EXT_ITER;
	typedef  FMT_EXT_MAP::const_iterator         FMT_EXT_CITER;

	typedef  std::map<uint, FMT_EXT_MAP>      CAT_FMT_EXT_MAP;
	typedef  CAT_FMT_EXT_MAP::iterator        CAT_FMT_EXT_ITER;
	typedef  CAT_FMT_EXT_MAP::const_iterator  CAT_FMT_EXT_CITER;

	CAT_ENC_MAP  catEncoderMap;
	CAT_DEC_MAP  catDecoderMap;

	CAT_NPL_MAP  encCatDefParamListMap;
	CAT_NPL_MAP  decCatDefParamListMap;

	CAT_FMT_EXT_MAP  catFmtExtMap;

	NamedParameterList  encInitialParamList;
	NamedParameterList  decInitialParamList;


private:
	DataEncoder*  FindEncoder(uint dataCategory, const std::string& encoderName);
	DataDecoder*  FindDecoder(uint dataCategory, const std::string& decoderName);

	ENC_VEC*  FindEncoders(uint dataCategory);
	DEC_VEC*  FindDecoders(uint dataCategory);
};


//====================================================================
struct VW_COMMON_API CodecParameterFactory
{
	//--------------------------------------------------------------
	// If <bEncoder> is true, fill the list with encoder parameters,
	// otherwise fill the list with decoder parameters.
	// <npl> is initialized with default general parameters defined
	// in data_codec_parameters namespace.
	//--------------------------------------------------------------
	virtual  void  FillParameterList(uint dataCategory,
									bool bEncoder,
									NamedParameterList& npl) const = 0;
};


//********************************************************************
}//namespace vertexware


