

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "DataObjects\SignalBase.h"
#include "DataObjects\ImageBase.h"
#include "DataObjects\VolumeBase.h"
#include "BaseTools\Memory.h"
#include "BaseTools\NoCopy.h"
#include "BaseTools\Debug.h"
#include <vector>


//********************************************************************
namespace vertexware {


//====================================================================
class VW_COMMON_API MipmapChain : public NoCopy
{
public:
	__forceinline static uint __fastcall ComputeDimensionAtLevel(uint level, uint dimLevel0) {
		return ClampedLow<uint>((dimLevel0 >> level), 1);
	}

	eSampleFormat  GetSampleFormat() const { return sampleFormat; }
	uint           GetNumLevels() const { return numLevels; }


protected:
	MipmapChain(eSampleFormat sampleFormat);

	// passing 0 sets numLevels to 0
	void  SetMaxDimension(uint maxDim = 0);


private:
	eSampleFormat  sampleFormat;
	uint           numLevels;
};


//====================================================================
class VW_COMMON_API SignalMipmapChain : public MipmapChain
{
public:
	SignalMipmapChain(eSampleFormat sampleFormat);

	void  Clear();

	//----------------------------------------------------------------
	// If sample-formats match and <bCopyLevel0OnFormatMatch> is false,
	// chain will only keep internal reference to <srcLevel0>.
	// This avoids extra copy when mipmap-chain is used only as an
	// intermediate data-source. Returns false on error.
	//----------------------------------------------------------------
	bool  Build(const SignalBase& srcLevel0, bool bCopyLevel0OnFormatMatch = true);


	const SignalBase&  GetLevel(uint level) const {
		VW_ASSERT(level < levels.size(), "Level-index out of range.");
		return *levels[level].Get();
	}


private:
	std::vector<AutoPtr<SignalBase> >  levels;
};


//====================================================================
class VW_COMMON_API ImageMipmapChain : public MipmapChain
{
public:
	ImageMipmapChain(eSampleFormat sampleFormat);

	void  Clear();

	//----------------------------------------------------------------
	// If sample-formats match and <bCopyLevel0OnFormatMatch> is false,
	// chain will only keep internal reference to <srcLevel0>.
	// This avoids extra copy when mipmap-chain is used only as an
	// intermediate data-source. Returns false on error.
	//----------------------------------------------------------------
	bool  Build(const ImageBase& srcLevel0, bool bCopyLevel0OnFormatMatch = true);


	const ImageBase&  GetLevel(uint level) const {
		VW_ASSERT(level < levels.size(), "Level-index out of range.");
		return *levels[level].Get();
	}


private:
	std::vector<AutoPtr<ImageBase> >  levels;
};


//====================================================================
class VW_COMMON_API VolumeMipmapChain : public MipmapChain
{
public:
	VolumeMipmapChain(eSampleFormat sampleFormat);

	void  Clear();

	//----------------------------------------------------------------
	// If sample-formats match and <bCopyLevel0OnFormatMatch> is false,
	// chain will only keep internal reference to <srcLevel0>.
	// This avoids extra copy when mipmap-chain is used only as an
	// intermediate data-source. Returns false on error.
	//----------------------------------------------------------------
	bool  Build(const VolumeBase& srcLevel0, bool bCopyLevel0OnFormatMatch = true);


	const VolumeBase&  GetLevel(uint level) const {
		VW_ASSERT(level < levels.size(), "Level-index out of range.");
		return *levels[level].Get();
	}


private:
	std::vector<AutoPtr<VolumeBase> >  levels;
};


//********************************************************************
}//namespace vertexware


