

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "DataObjects\Sample.h"
#include "BaseTools\Debug.h"
#include "BaseTools\NoCopy.h"


//********************************************************************
namespace vertexware {


//====================================================================
// SignalBase provides sample-type independent operations.
// Size can be 0, but samples must not be accessed in that case.
//====================================================================
class VW_COMMON_API SignalBase : public NoCopy
{
public:
	// destructor
	virtual ~SignalBase();

	// sample format
	bool           IsValidSampleFormat(eSampleFormat sampleFormat) const;
	void           SetSampleFormat(eSampleFormat sampleFormat);
	eSampleFormat  GetSampleFormat() const { return sampleFormat; }

	// storage configuration
	virtual  void  ReserveSampleStorage(uint numSamples);
	virtual  void  ReleaseSampleStorage();

	// storage info
	__forceinline  uint  GetNumSamples() const { return numSamples; }
	__forceinline  uint  GetSampleByteSize() const { return sampleByteSize; }
	__forceinline  uint  GetTotalByteSize() const { return totalByteSize; }

	bool  CheckDimensionMatch(const SignalBase& sig) const;
	bool  CheckSampleByteSizeMatch(const SignalBase& sig) const;
	bool  CheckTotalByteSizeMatch(const SignalBase& sig) const;
	bool  CheckSampleFormatMatch(const SignalBase& sig) const;

	// storage access
	__forceinline  ubyte*        GetSamplesB() { return pSamples; }
	__forceinline  const ubyte*  GetSamplesB() const { return pSamples; }

	__forceinline  ubyte*  SampleB(uint sampleIndex) {
		CheckAccess1D(sampleIndex);
		return (pSamples + sampleByteSize * sampleIndex);
	}
	__forceinline  const ubyte*  SampleB(uint sampleIndex) const {
		return const_cast<SignalBase*>(this)->SampleB(sampleIndex);
	}


protected:
	// constructor
	SignalBase(uint sampleByteSize);

	// debug helpers
	__forceinline  void  CheckAccess1D(uint sampleIndex) const {
		VW_ASSERT(numSamples, "Accessing 0-size signal.");
		VW_ASSERT(sampleIndex < numSamples, "Sample-index out of range.");
	}

	// low-level setup used by reserve/release overrides
	void  AllocateStorage(uint numSamples, ubyte* pExternal = 0);
	void  DeallocateStorage();


private:
	eSampleFormat  sampleFormat;
	uint           sampleByteSize;
	uint           numSamples;
	uint           totalByteSize;
	bool           bExternalAllocation;
	ubyte*         pSamples;
};


//====================================================================
// Adapts externally allocated storage so that it appears as SignalBase.
// Reserve is not allowed. External storage must be tightly packed.
//====================================================================
class VW_COMMON_API SignalBaseFromExternalStorageAdapter : public SignalBase
{
public:
	SignalBaseFromExternalStorageAdapter(eSampleFormat sampleFormat);

	void  SetExternalStorage(ubyte* pStorage, uint numSamples);

	// storage configuration (reserve throws exception if called)
	virtual  void  ReserveSampleStorage(uint);
	virtual  void  ReleaseSampleStorage() {}
};


//********************************************************************
}//namespace vertexware


