

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\CommandInterpreter.h"
#include "BaseTools\VariableProxy.h"
#include "BaseTools\Profiler.h"
#include "DataObjects\SignalIO.h"
#include "Framework\Application.h"
#include "Framework\FrameObservers.h"
#include "Framework\StandardCommandInterpreter.h"
#include "HIDControllers\Keyboard.h"
#include "HIDControllers\Mouse.h"
#include "HIDControllers\OrientationTracker.h"
#include "HIDControllers\PositionTracker.h"
#include "HIDControllers\Glove.h"
#include "HIDControllers\GameController.h"
#include "GUI\GUI_Component.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace framework_commands {


//====================================================================
VW_COMMON_API
void __fastcall CreateAndRegisterAllCommands(
								std_command_interpreter::Interpreter& ci,
								gui::Component* screenShotCmd_AutoHideComponent = 0,
								uint hidConfigCmd_HIDReconfigCode = 0,
								uint videoConfigCmd_VideoReconfigCode = 0);


//====================================================================
class VW_COMMON_API ExitCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionWaitCloseConfirm;


public:
	ExitCmd();

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
class VW_COMMON_API CrashCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionAssert;
	static const char* s_szOptionVerify;
	static const char* s_szOptionThrow;
	static const char* s_szOptionZeroPtrRead;
	static const char* s_szOptionZeroPtrWrite;
	static const char* s_szOptionHardCleanupConfigLoop;
	static const char* s_szOptionNoHardCleanupConfigLoop;
	static const char* s_szOptionHardRethrowExit;
	static const char* s_szOptionNoHardRethrowExit;
	static const char* s_szAssignLogMsgOnMainWindowCleanup;


public:
	CrashCmd();

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
class VW_COMMON_API VSyncCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionOn;
	static const char* s_szOptionOff;
	static const char* s_szAssignDynamic;


public:
	VSyncCmd();

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
// Registered callback is auto-deleted by command destructor.
//====================================================================
class VW_COMMON_API ScreenShotCmd : public Command, public FrameObserver
{
public:
	//==============================================================
	// Called only if the command is not executed in immediate mode.
	//==============================================================
	class VW_COMMON_API ScreenShotCallback
	{
	public:
		virtual ~ScreenShotCallback() {}

		virtual  void  PreScreenShot() = 0;
		virtual  void  PostScreenShot() = 0;
	};

	//==============================================================
	// Auto-hide-target is automatically hidden while the screenshot
	// is being taken. It is restored to visible state if it was
	// originally visible.
	//==============================================================
	class VW_COMMON_API ScreenShotCallback_AutoHideComponent : public ScreenShotCallback
	{
	public:
		ScreenShotCallback_AutoHideComponent(gui::Component* _pAutoHideTarget = 0);

		void             SetAutoHideTarget(gui::Component* _pAutoHideTarget = 0);
		gui::Component*  GetAutoHideTarget() const { return pAutoHideTarget; }

		virtual  void  PreScreenShot();
		virtual  void  PostScreenShot();

	private:
		gui::Component* pAutoHideTarget;
		bool            bAutoHideTargetVisible;
	};


public:
	static const char* s_szDefaultName;
	static const char* s_szOptionImmediate;
	static const char* s_szOptionOverwrite;
	static const char* s_szOptionColor;
	static const char* s_szOptionDepth;
	static const char* s_szOptionStencil;
	static const char* s_szOptionAll;
	static const char* s_szOptionFront;


public:
	ScreenShotCmd(ScreenShotCallback* pCallback = 0);
	virtual ~ScreenShotCmd();

	void                 SetCallback(ScreenShotCallback* _pCallback = 0) { pCallback = _pCallback; }
	ScreenShotCallback*  GetCallback() const { return pCallback; }

	void  SetSerialNumberWidth(uint width = 4); //set 0 to disable
	uint  GetSerialNumberWidth() const { return iSerialNumberWidth; }

	void  SetSerialNumber(uint sn = 0) { iSerialNumber = sn; }
	uint  GetSerialNumber() const { return iSerialNumber; }

	void  SetAllowFileOverwrite(bool bAllowOverwrite = false) { bAllowFileOverwrite = bAllowOverwrite; }
	bool  GetAllowFileOverwrite() const { return bAllowFileOverwrite; }

	void                SetFileFormat(const std::string& fmt = signal_codec_parameters::Value_ImageFileFormat_Default) { fileFormat = fmt; }
	const std::string&  GetFileFormat() const { return fileFormat; }

	void                SetBaseFilename(const std::string& s = "screenshot") { baseFilename = s; }
	const std::string&  GetBaseFilename() const { return baseFilename; }

	// used only if file-format's default extension is not available
	void                SetFileExtension(const std::string& s = "image") { fileExtension = s; }
	const std::string&  GetFileExtension() const { return fileExtension; }

	void  SetBufferTypeSuffixes(const std::string& color = "color",
				const std::string& depth = "depth", const std::string& stencil = "stencil");

	void  GetBufferTypeSuffixes(std::string& color,
				std::string& depth, std::string& stencil) const;

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);

	//-----------------------------
	// FrameCallback implementation
	//-----------------------------
	virtual  void  OnFrame(bool& bAutoUnregister, uint frameNumber, double secsFrameTimeDelta);


private:
	void  CaptureScreenShot();


private:
	ScreenShotCallback*  pCallback;

	bool         bAllowFileOverwrite;
	uint         iSerialNumber;
	uint         iSerialNumberWidth;
	std::string  fileExtension;
	std::string  fileFormat;
	std::string  baseFilename;
	std::string  colorSuffix;
	std::string  depthSuffix;
	std::string  stencilSuffix;

	// collected options
	std::ostream*  pOStream;

	uint  iTargetFrame;
	uint  frameBufferFlagsColor;
	uint  frameBufferFlagsDepthStencil;
	bool  bDepthBuffer;
	bool  bStencilBuffer;
	bool  bColorBuffer;
};


//====================================================================
class VW_COMMON_API HIDConfigCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionNoDIAtInit;
	static const char* s_szOptionDIExclusiveModeFS;
	static const char* s_szAssignDIVersion;
	static const char* s_szValueLast;


public:
	HIDConfigCmd(uint hidReconfigCode = 0);

	void  SetHIDReconfigCode(uint rc = 0) { iHIDReconfigCode = rc; }

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);


private:
	uint  iHIDReconfigCode;
};


//====================================================================
class VW_COMMON_API HIDControlCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionUseDI;
	static const char* s_szOptionNoDI;
	static const char* s_szOptionDumpInfo;
	static const char* s_szAssignDeviceCreationID;
	static const char* s_szAssignControllerNumber;

	// mouse specific
	static const char* s_szAssignMouseSyncWithSystemCursor;

	// orientation-tracker specific
	static const char* s_szOptionOTReset;
	static const char* s_szOptionOTResetHead;
	static const char* s_szAssignOTrackerHPRModifier;
	static const char* s_szAssignOTrackerReverseModifiers;
	static const char* s_szAssignOTrackerReverseHead;
	static const char* s_szAssignOTrackerReversePitch;
	static const char* s_szAssignOTrackerReverseRoll;


public:
	HIDControlCmd();

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);


private:
	HumanInterfaceDevice*  FindHID(uint creationID);


protected:
	// by-type dispatch
	virtual void  UpdateController(std::ostream& os, const CommandLineParser& clp,
								uint controllerNumber, HIDController& hidc);

	// type-specific processors
	virtual void  Update(std::ostream& os, const CommandLineParser& clp,
						uint controllerNumber, Keyboard& keyboard);

	virtual void  Update(std::ostream& os, const CommandLineParser& clp,
						uint controllerNumber, Mouse& mouse);

	virtual void  Update(std::ostream& os, const CommandLineParser& clp,
						uint controllerNumber, OrientationTracker& tracker);

	virtual void  Update(std::ostream& os, const CommandLineParser& clp,
						uint controllerNumber, PositionTracker& tracker);

	virtual void  Update(std::ostream& os, const CommandLineParser& clp,
						uint controllerNumber, Glove& glove);

	virtual void  Update(std::ostream& os, const CommandLineParser& clp,
						uint controllerNumber, GameController& gameController);
};


//====================================================================
class VW_COMMON_API VideoConfigCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionShowUnfilteredProfile;
	static const char* s_szOptionReconfig;
	static const char* s_szOptionNoProfileCheck;
	static const char* s_szOptionNoAutoSelectClosestFBC;

	// listings
	static const char* s_szOptionListRenderers;
	static const char* s_szOptionListFrameBufferConfigs;
	static const char* s_szOptionListDisplayModes;
	static const char* s_szOptionListLegend;

	// renderer selection
	static const char* s_szAssignRendererNumber;
	static const char* s_szAssignRendererPluginName;
	static const char* s_szAssignRendererInterfaceName;

	// display-mode / window-properties
	static const char* s_szAssignWidth;
	static const char* s_szAssignHeight;
	static const char* s_szAssignBitsPerPixel;
	static const char* s_szAssignRefreshFrequency;
	static const char* s_szAssignWindowResizable;

	// frame-buffer
	static const char* s_szAssignFullscreen;
	static const char* s_szAssignHardwareRasterizer;
	static const char* s_szAssignBackBufferCount;
	static const char* s_szAssignRGBA;
	static const char* s_szAssignDepthStencil;
	static const char* s_szAssignAccumRGBA;

	// video-hints
	static const char* s_szAssignVSync;
	static const char* s_szAssignFSDestinationAlpha;
	static const char* s_szAssignPureDeviceMode;
	static const char* s_szAssignVertexProcessingMode;
	static const char* s_szValueVPM_Hardware;
	static const char* s_szValueVPM_Software;
	static const char* s_szValueVPM_Mixed;
	static const char* s_szValueVPM_Default;


public:
	VideoConfigCmd(uint videoReconfigCode = 0);

	void  SetVideoReconfigCode(uint rc = 0) { iVideoReconfigCode = rc; }

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);


private:
	void  DumpVRHeader(std::ostream& os, const VideoRendererInfo& vrInfo);
	bool  FindInfoByNumber(uint number, VideoRendererInfo& vrInfo, VideoProfile& unfilteredProfile);
	bool  FindInfoByPluginName(const std::string& pluginName, VideoRendererInfo& vrInfo, VideoProfile& unfilteredProfile);
	bool  FindInfoByVRInterfaceName(const std::string& vrInterfaceName, VideoRendererInfo& vrInfo, VideoProfile& unfilteredProfile);
	bool  ExtractRGBA(const CommandLineParser& clp, const std::string& name, uint outVals[4]);
	bool  ExtractUIntegers(const CommandLineParser& clp, const std::string& name, uint* pOutVals, uint numVals);


private:
	uint  iVideoReconfigCode;
};


//====================================================================
class VW_COMMON_API GUIFontCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionListSystemFonts;
	static const char* s_szOptionUpdateResource;

	// font-selection
	static const char* s_szAssignFontResourceName;
	static const char* s_szAssignFontNumber;

	// texture resource
	static const char* s_szAssignTextureWidth;
	static const char* s_szAssignTextureMultiLevel;
	static const char* s_szAssignAsciiSubsetOnly;

	// font-family
	static const char* s_szAssignFontFamily;
	static const char* s_szAssignFontBold;
	static const char* s_szAssignFontItalic;

	// additional font properties
	static const char* s_szAssignDefaultSpacingMono;
	static const char* s_szAssignDefaultSpacing;

	// image generation parameters
	static const char* s_szAssignIGScaleEmHeight;
	static const char* s_szAssignIGGlyphOffsetH;
	static const char* s_szAssignIGGlyphOffsetV;
	static const char* s_szAssignIGGlyphSizeModifyH;
	static const char* s_szAssignIGGlyphSizeModifyV;

	// glyph properties
	static const char* s_szAssignAutoCalcWidthEmptyGlyph;
	static const char* s_szAssignAutoCalcAddPixelsLeft;
	static const char* s_szAssignAutoCalcAddPixelsRight;
	static const char* s_szAssignAutoCalcAddPixelsTop;
	static const char* s_szAssignAutoCalcAddPixelsBottom;


public:
	GUIFontCmd();

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
class VW_COMMON_API GUIAliasCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionListObjects;
	static const char* s_szOptionClearAll;
	static const char* s_szOptionClear;
	static const char* s_szOptionTile;
	static const char* s_szOptionFont;
	static const char* s_szOptionShader;
	static const char* s_szOptionCursor;
	static const char* s_szAssignRefName;
	static const char* s_szAssignAliasName;


public:
	GUIAliasCmd();

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
class VW_COMMON_API GUIStatsCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionRun;
	static const char* s_szOptionStop;
	static const char* s_szOptionReset;
	static const char* s_szOptionDump;


public:
	GUIStatsCmd();

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);


private:
	GUIRendererStatsCollector  collector;

	uint frameCount;

	gui::Renderer::Statistics  statsMin;
	gui::Renderer::Statistics  statsMax;
	gui::Renderer::Statistics  statsAvg;
};


//====================================================================
class VW_COMMON_API VariableCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionInfo;


public:
	VariableCmd();

	void               SetVariableRegistry(VariableRegistry* vr = 0); //pass 0 to use global registry
	VariableRegistry&  GetVariableRegistry() { return *pVarReg; }

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);


private:
	void  DumpVar(std::ostream& os, const VariableProxy& vp);


private:
	VariableRegistry* pVarReg;
};


//====================================================================
class VW_COMMON_API ProfilerCmd : public Command
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionClear;
	static const char* s_szOptionReset;
	static const char* s_szOptionDump;
	static const char* s_szOptionPrintEntryCounts;
	static const char* s_szOptionPrintNodeTimes;
	static const char* s_szOptionPrintBranchTimes;
	static const char* s_szOptionPrintNodePercentages;
	static const char* s_szOptionRun;
	static const char* s_szOptionStop;
	static const char* s_szAssignRecursionHandler;
	static const char* s_szAssignAutoReset;
	static const char* s_szAssignMaxDepth;


public:
	ProfilerCmd();

	void       SetProfiler(Profiler* p = 0); //pass 0 to use global profiler
	Profiler&  GetProfiler() { return *pProfiler; }

	void                     SetProfilerDumpDesc(const ProfilerDumpDesc& pdd = ProfilerDumpDesc()) { profilerDumpDesc = pdd; }
	const ProfilerDumpDesc&  GetProfilerDumpDesc() const { return profilerDumpDesc; }

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);


private:
	Profiler*         pProfiler;
	ProfilerDumpDesc  profilerDumpDesc;
};


//********************************************************************
}//namespace framework_commands


//********************************************************************
}//namespace vertexware


