

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Singleton.h"
#include "BaseTools\Exception.h"
#include "Framework\FrameworkExceptions.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class  Plugin;
class  PluginInfo;
class  Application;
class  GenericExtension;
class  HumanInterfaceDevice;
class  HIDController;
class  VideoRendererInfo;
class  VideoRenderer;
class  VideoSubsystem;
class  VideoProfile;
class  VideoConfig;
class  DirectInputConfig;
class  ConsoleBuffers;
class  FrameworkProfiler;
class  NamedParameter;
class  VariableRegistry;


//********************************************************************
namespace gui {


//====================================================================
class Manager;


//********************************************************************
}//namespace gui


//====================================================================
// All observers are auto-unregistered by their destructor.
// If <bAutoUnregister> is set to true, kernel unregisters
// observer after the function returns.
//====================================================================
class VW_COMMON_API FrameObserver
{
public:
	virtual ~FrameObserver();

	//-----------------------------------------------------------------
	// Called at start of each frame. <bAutoUnregister> default = false
	//-----------------------------------------------------------------
	virtual  void  OnFrame(bool& bAutoUnregister,
							uint frameNumber, double secsFrameTimeDelta) = 0;
};


//====================================================================
class VW_COMMON_API ReconfigObserver
{
public:
	virtual ~ReconfigObserver();

	virtual  void  OnVideoReconfig() {}
	virtual  void  OnVSyncReconfig(bool bVSyncEnabled) {}
	virtual  void  OnHIDReconfig() {}
};


//====================================================================
struct VW_COMMON_API MainWindowProperties
{
	HICON        hIcon;
	std::string  text;

	MainWindowProperties();
};


//====================================================================
struct VW_COMMON_API MainWindowGeometry
{
	bool  bMaximized;
	uint  positionX;
	uint  positionY;
	uint  minSizeX;
	uint  minSizeY;

	MainWindowGeometry();
};


//====================================================================
// Kernel interface parameter-sets
//====================================================================
enum eKernelLevel
{
	KernelLevel_KernelSetup = 0,
	KernelLevel_PluginSetup,
	KernelLevel_ConfigLoop,
	KernelLevel_FrameLoop,

	KernelLevel_Count,
	KernelLevel_Force32Bits = 0x7FFFFFFF
};


//====================================================================
struct VW_COMMON_API KernelThreadInfo
{
	HANDLE  threadHandle; //read-only
	uint    threadID;     //read-only
	int     priority;
	bool    bEnablePriorityBoost;

	KernelThreadInfo();

	void  Dump(std::ostream& os, uint tabLevel = 0) const;
};


//====================================================================
// Kernel never samples the device if it is put in asynchronous-mode.
// This flags control only synchronous-mode auto sampling.
// By default all flags are on.
//====================================================================
struct VW_COMMON_API HIDAutoSamplingConfig
{
	bool bAutoSampleSystemKeyboard;
	bool bAutoSampleSystemMouse;
	bool bAutoSampleNonSystemDevices;

	HIDAutoSamplingConfig();
};


//====================================================================
// For convenience, default constructor calls
// Kernel::GetExecutionParams() and copies the values from
// kernel-current values, pass false to ctor to disable this.
//====================================================================
struct VW_COMMON_API KernelExecutionParams
{
	HIDAutoSamplingConfig  hidAutoSamplingConfig;
	bool                   bAutoDispatchGUIEvents;
	bool                   bGUIEventsPreFrameUpdateStates;
	bool                   bAutoClearGUIStats;
	bool                   bAutoDrawGUI;
	bool                   bAutoEndGUIRenderer;
	bool                   bAutoPresentBackBuffer;
	bool                   bEnableHardCleanupConfigLoop;
	bool                   bEnableHardRethrowExit;
	bool                   bLogWinMsgOnMainWindowCleanup;
	bool                   bWaitCloseConfirmationOnExit;
	bool                   bDeveloperMode; //read-only

	KernelExecutionParams(bool bInitFromKernelCurrent = true);

	void  GetKernelCurrent();
	void  SetKernelCurrent() const;

	void  Dump(std::ostream& os, uint tabLevel = 0) const;
};


//====================================================================
// Kernel checks known exception types and fills appropriate flags for convenience.
//====================================================================
struct VW_COMMON_API FailureDesc
{
	eKernelLevel  origin;

	bool  bCPlusPlus;
	bool  bLibrary;
	bool  bKernel;
	bool  bGUI;
	bool  bGenericExtension;
	bool  bApplication;
	bool  bVideoSubsystem;
	bool  bVideoRenderer;
	bool  bInputSubsystem;
	bool  bGenericSubsystem;

	FailureDesc(eKernelLevel origin = eKernelLevel(0),
				const Exception* el = 0,
				const std::exception* es = 0);
};


//====================================================================
// Kernel interface
//====================================================================
class VW_COMMON_API Kernel
{
VW_DECLARE_SINGLETON( Kernel );

friend VW_COMMON_API std::ostream& __fastcall GetConOut();
friend VW_COMMON_API std::ostream& __fastcall GetLogOut();

public:
	//---------------------------
	// Kernel commands & queries.
	//---------------------------
	virtual  void  SetExecutionParams(const KernelExecutionParams& kep) = 0;
	virtual  void  GetExecutionParams(KernelExecutionParams& kep) = 0;

	virtual  void  SetKernelThreadInfo(const KernelThreadInfo& kti) = 0;
	virtual  void  GetKernelThreadInfo(KernelThreadInfo& kti) = 0;

	virtual  eKernelLevel  GetActiveKernelLevel() = 0;
	virtual  bool          IsAppWindowActive() = 0;

	virtual  void  SetMainWindowProperties(const MainWindowProperties& mwp) = 0;
	virtual  void  GetMainWindowProperties(MainWindowProperties& mwp) = 0;

	virtual  void  RequestShutdown() = 0;
	virtual  void  RequestHIDReconfig(const DirectInputConfig& diConfig, uint reconfigCode = 0) = 0;
	virtual  void  RequestVideoReconfig(const std::string& vrPluginName, const VideoConfig& vc, uint reconfigCode = 0) = 0;
	virtual  void  RequestSubsystemReconfig(uint reconfigCode = 0) = 0;
	virtual  void  RequestGenericReconfig(uint reconfigCode = 0) = 0;

	void  RequestHIDReconfig(uint reconfigCode = 0);
	void  RequestVideoReconfig(uint reconfigCode = 0);

	// return true if parameter is recognized
	virtual  bool  SetPluginParameter(const std::string& pluginName, const NamedParameter& np) = 0;
	virtual  bool  GetPluginParameter(const std::string& pluginName, const std::string& parameterName, NamedParameter& out) = 0;

	virtual  uint               GetNumGenericExtensions() = 0;
	virtual  GenericExtension&  GetGenericExtension(uint i) = 0;

	virtual  GenericExtension*  FindGenericExtension(const std::string* pPluginName = 0,
										const std::string* pGenericExtensionInterfaceName = 0) = 0;

	virtual  Application*  GetApplicationPtr() = 0;
	Application&           GetApplication();

	virtual  uint  GetNumVideoRenderers() = 0;
	virtual  void  GetVideoRendererInfo(uint i, VideoRendererInfo& vrInfo) = 0;
	virtual  void  GetVideoRendererUnfilteredProfile(uint i, VideoProfile& profile) = 0;

	// returns false if failed (if <vrInterfaceName> is empty all renderers are checked in order)
	bool  ConstructClosestSafeVideoConfig(
						const std::string& vrInterfaceName, const VideoConfig& refVideoConfig,
						std::string& outVideoRendererPluginName, VideoConfig& outVideoConfig);

	virtual  VideoRenderer*  GetActiveVideoRendererPtr() = 0;
	VideoRenderer&           GetActiveVideoRenderer();
	virtual  bool            GetActiveVideoRendererInfo(VideoRendererInfo& vrInfo) = 0;
	virtual  bool            GetActiveVideoRendererUnfilteredProfile(VideoProfile& profile) = 0;

	virtual  uint  GetNumHIDDrivers() = 0;
	virtual  void  GetHIDDriverInfo(uint i, PluginInfo& pi) = 0;

	virtual  uint                   GetNumHumanInterfaceDevices() = 0;
	virtual  HumanInterfaceDevice&  GetHumanInterfaceDevice(uint i) = 0;

	uint            GetNumHIDControllers(const std::string& hidControllerInterfaceName);
	HIDController*  GetHIDController(const std::string& hidControllerInterfaceName, uint i);

	virtual  void   GetActiveDirectInputConfig(DirectInputConfig& diConfig) = 0;

	virtual  void   ResetFrameNumber(uint frameNumber = 0) = 0;
	virtual  uint   GetFrameNumber() = 0;
	virtual  bool   IsFrameInProgress() = 0;
	virtual  double GetFrameTimeDeltaSecs() = 0;

	virtual  FrameworkProfiler&  GetFrameworkProfiler() = 0;
	virtual  ConsoleBuffers&     GetConsoleBuffers() = 0;
	virtual  VariableRegistry&   GetVariableRegistry() = 0;

	virtual  void  RegisterFrameObserver(FrameObserver& fo) = 0;
	virtual  void  UnregisterFrameObserver(FrameObserver& fo) = 0;
	virtual  void  UnregisterAllFrameObservers() = 0;

	virtual  void  RegisterReconfigObserver(ReconfigObserver& ro) = 0;
	virtual  void  UnregisterReconfigObserver(ReconfigObserver& ro) = 0;
	virtual  void  UnregisterAllReconfigObservers() = 0;

	virtual  void  NotifyVSyncChanged() = 0;


protected:
	//-------------------------
	// Derived-class interface.
	//-------------------------
	Kernel();
	virtual ~Kernel();

	static void __fastcall SetOutputStreams(std::ostream& conOut, std::ostream& logOut);

	static void __fastcall GUIManager_ConnectApplication(bool bConnect);
	static void __fastcall GUIManager_NotifyApplicationWindowActive(bool bActive);
	static void __fastcall GUIManager_NotifyApplicationWindowCreated();

	static void __fastcall GUIManager_BeginConnectHumanInterfaceDevices();
	static void __fastcall GUIManager_ConnectHumanInterfaceDevice(HumanInterfaceDevice& hid);
	static void __fastcall GUIManager_EndConnectHumanInterfaceDevices();
	static void __fastcall GUIManager_DisconnectHumanInterfaceDevices();
	static void __fastcall GUIManager_SetTolerateMissingInputDevice(bool bTolerate);

	static void __fastcall GUIManager_ConnectVideoRenderer(VideoRenderer& vr);
	static void __fastcall GUIManager_DisconnectVideoRenderer();
	static void __fastcall GUIManager_CreateVideoResources(bool bPostReset);
	static void __fastcall GUIManager_ReleaseVideoResources(bool bPreReset);
	static void __fastcall GUIManager_OnWindowResize(uint width, uint height);

	static void __fastcall VideoRenderer_ConnectVideoSubsystem(VideoRenderer& vr, VideoSubsystem& vs);
	static void __fastcall VideoRenderer_DisconnectVideoSubsystem(VideoRenderer& vr);
	static void __fastcall VideoRenderer_CreateVideoResources(VideoRenderer& vr, bool bPostReset);
	static void __fastcall VideoRenderer_ReleaseVideoResources(VideoRenderer& vr, bool bPreReset);
	static void __fastcall VideoRenderer_OnWindowResize(VideoRenderer& vr, uint width, uint height);


private:
	static  std::ostream*  s_pConOut;
	static  std::ostream*  s_pLogOut;
};


//====================================================================
// Access to console-style facilities.
//====================================================================
VW_COMMON_API __forceinline std::ostream& __fastcall GetConOut()
{
	return *Kernel::s_pConOut;
}


//====================================================================
VW_COMMON_API __forceinline std::ostream& __fastcall GetLogOut()
{
	return *Kernel::s_pLogOut;
}


//********************************************************************
}//namespace vertexware


