

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Memory.h"
#include "BaseTools\NoCopy.h"
#include "VideoRendererObjects\VRO_Textures.h"
#include "GLTools\States.h"
#include "private\vw_opengl_exports.h"
#include <gl\gl.h>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gl_tools {


//====================================================================
class VW_OPENGL_API TextureBindingSentry : public NoCopy
{
public:
	TextureBindingSentry(GLenum target, GLenum bindingPName,
						GLuint textureNameToBind);
	~TextureBindingSentry();


private:
	GLenum  target;
	GLuint  lastBoundName;
};


//====================================================================
struct VW_OPENGL_API TextureBindingSentry1D : public TextureBindingSentry
{
	TextureBindingSentry1D(GLuint textureName) :
		TextureBindingSentry(GL_TEXTURE_1D, GL_TEXTURE_BINDING_1D, textureName) {}
};


//====================================================================
struct VW_OPENGL_API TextureBindingSentry2D : public TextureBindingSentry
{
	TextureBindingSentry2D(GLuint textureName) :
		TextureBindingSentry(GL_TEXTURE_2D, GL_TEXTURE_BINDING_2D, textureName) {}
};


//====================================================================
struct VW_OPENGL_API TextureBindingSentry3D : public TextureBindingSentry
{
	TextureBindingSentry3D(GLuint textureName);
};


//====================================================================
class VW_OPENGL_API TextureUnit
{
public:
	// returns number of texture units supported by active driver
	static uint __fastcall GetNumUnits();

	// returns active texture unit
	static uint __fastcall GetActiveUnitIndex();

	// activation-wrappers / enable/disable-wrappers
	static void __fastcall ActivateUnit(uint unitIndex);

	static void __fastcall EnableActiveUnit(GLenum target) { glEnable(target); }
	static void __fastcall DisableActiveUnit();

	static void __fastcall EnableUnit(uint unitIndex, GLenum target);
	static void __fastcall DisableUnit(uint unitIndex);

	static void __fastcall DisableAllUnitsAndActivateUnit0();


public:
	//----------------------------
	// Active texture-unit sentry.
	//----------------------------
	class VW_OPENGL_API ActiveUnitSentry
	{
	public:
		ActiveUnitSentry(uint newActiveUnitIndex);
		~ActiveUnitSentry();

	private:
		bool  bNeedsChange;
		uint  lastActiveUnitIndex;
	};


public:
	//------------------------------------------------------
	// <textureUnitIndex> is 0-based. Throws exception if
	// <textureUnitIndex> is out-of-range for active driver.
	//------------------------------------------------------
	TextureUnit(uint textureUnitIndex = 0);

	uint  GetIndex() const { return unitIndex; }
	uint  GetName() const { return unitName; }

	void  Enable(GLenum target) const { EnableUnit(unitIndex, target); }
	void  Disable() const { DisableUnit(unitIndex); }

	void  Activate() const { ActivateUnit(unitIndex); }


private:
	uint    unitIndex;
	GLenum  unitName;
};


//====================================================================
class VW_OPENGL_API TextureCoordArray
{
public:
	// returns number of texture array supported by active driver
	static uint __fastcall GetNumArrays() { return TextureUnit::GetNumUnits(); }

	// returns active texture-coord array
	static uint __fastcall GetActiveArrayIndex();

	static void __fastcall ActivateArray(uint arrayIndex);

	static void __fastcall EnableActiveArray() { glEnableClientState(GL_TEXTURE_COORD_ARRAY); }
	static void __fastcall DisableActiveArray() { glDisableClientState(GL_TEXTURE_COORD_ARRAY); }

	static void __fastcall EnableArray(uint arrayIndex);
	static void __fastcall DisableArray(uint arrayIndex);

	static void __fastcall DisableAllArraysAndActivateArray0();


public:
	//-----------------------------------
	// Active texture-coord-array sentry.
	//-----------------------------------
	class VW_OPENGL_API ActiveArraySentry
	{
	public:
		ActiveArraySentry(uint newActiveArrayIndex);
		~ActiveArraySentry();

	private:
		bool  bNeedsChange;
		uint  lastActiveArrayIndex;
	};


public:
	//-------------------------------------------------------
	// <textureArrayIndex> is 0-based. Throws exception if
	// <textureArrayIndex> is out-of-range for active driver.
	// Array-index must be < number of texture-units.
	//-------------------------------------------------------
	TextureCoordArray(uint textureArrayIndex = 0);

	uint  GetIndex() const { return arrayIndex; }

	void  Enable() { EnableArray(arrayIndex); }
	void  Disable() { DisableArray(arrayIndex); }

	void  Activate() { ActivateArray(arrayIndex); }


private:
	uint  arrayIndex;
};


//====================================================================
class VW_OPENGL_API TextureBase : public NoCopy
{
public:
	// returns native storage order
	static bool __fastcall IsNativeLayerStorageOrderBackToFront() { return false; }
	static bool __fastcall IsNativeRowStorageOrderBottomToTop() { return true; }

	// returns rough estimate of largest supported texture size
	static uint __fastcall GetMaxTextureSize();

	// returns 0 if mapping is undefined
	static GLint __fastcall MapSampleFormatToGLInternalFormat(eSampleFormat sampleFormat);

	// returns SampleFormat_Invalid if mapping is undefined
	static eSampleFormat __fastcall MapGLInternalFormatToSampleFormat(GLint internalFormat);

	// returns false if mapping is undefined
	static bool __fastcall MapSampleFormatToGLUploadParameters(eSampleFormat sampleFormat,
											GLenum& outDataFormat, GLenum& outDataType,
											bool& outNeedsConversionToARGB32);


public:
	virtual ~TextureBase();

	GLenum  GetTarget() const { return target; }
	GLenum  GetBindingPName() const { return bindingPName; }
	GLint   GetInternalFormat() const { return internalFormat; }
	GLuint  GetName() const { return name; }

	void  BindToActiveTextureUnit() const { glBindTexture(target, name); }


protected:
	TextureBase(GLenum target, GLenum bindingPName, GLint internalFormat);

	static bool __fastcall CheckRealizedTextureWidth(GLenum target, uint level, uint expectedWidth);

	// throws exception (should be called from ctor)
	void  ValidateInternalFormat();

	void  RealizePriority(uint newPriority);
	bool  RealizeBaseLevel(uint newBaseLevel);


private:
	GLenum  target;
	GLenum  bindingPName;
	GLint   internalFormat;
	GLuint  name;
};


//====================================================================
// Dimensions must be powers-of-2.
//====================================================================
class VW_OPENGL_API Texture1D : public video_renderer_objects::Texture1D,
								public gl_tools::TextureBase
{
public:
	// sample-format mapping to GL-internal-format is enforced
	Texture1D(eSampleFormat sampleFormat, bool bMultiLevel,
				uint widthLevel0);

	// GL-internal-format mapping to sample-format is NOT enforced
	Texture1D(GLint internalFormat, bool bMultiLevel,
				uint widthLevel0);


protected:
	//--------------------------------------
	// video_renderer_objects::Texture1D implementation.
	//--------------------------------------
	virtual  void  UpdatePriority() { RealizePriority(GetPriority()); }
	virtual  bool  UpdateBaseLevel() { return RealizeBaseLevel(GetBaseLevel()); }

	virtual  bool  DoLoadLevel(uint level, const SignalBase& src,
							uint dataWidth,
							uint dstOffsetX,
							uint srcOffsetX);


private:
	void  Allocate();
};


//====================================================================
class VW_OPENGL_API Texture2D : public video_renderer_objects::Texture2D,
								public gl_tools::TextureBase
{
public:
	// sample-format mapping to GL-internal-format is enforced
	Texture2D(eSampleFormat sampleFormat, bool bMultiLevel,
				uint heightLevel0, uint widthLevel0);

	// GL-internal-format mapping to sample-format is NOT enforced
	Texture2D(GLint internalFormat, bool bMultiLevel,
				uint heightLevel0, uint widthLevel0);


protected:
	//--------------------------------------
	// video_renderer_objects::Texture2D implementation.
	//--------------------------------------
	virtual  void  UpdatePriority() { RealizePriority(GetPriority()); }
	virtual  bool  UpdateBaseLevel() { return RealizeBaseLevel(GetBaseLevel()); }

	virtual  bool  DoLoadLevel(uint level, const ImageBase& src,
								bool bSrcIsStoredBottomToTop,
								uint dataHeight, uint dataWidth,
								uint dstOffsetY, uint dstOffsetX,
								uint srcOffsetY, uint srcOffsetX);


private:
	void  Allocate();
};


//====================================================================
class VW_OPENGL_API Texture3D : public video_renderer_objects::Texture3D,
								public gl_tools::TextureBase
{
public:
	// sample-format mapping to GL-internal-format is enforced
	Texture3D(eSampleFormat sampleFormat, bool bMultiLevel,
				uint depthLevel0, uint heightLevel0, uint widthLevel0);

	// GL-internal-format mapping to sample-format is NOT enforced
	Texture3D(GLint internalFormat, bool bMultiLevel,
				uint depthLevel0, uint heightLevel0, uint widthLevel0);


protected:
	//--------------------------------------
	// video_renderer_objects::Texture3D implementation.
	//--------------------------------------
	virtual  void  UpdatePriority() { RealizePriority(GetPriority()); }
	virtual  bool  UpdateBaseLevel() { return RealizeBaseLevel(GetBaseLevel()); }

	virtual  bool  DoLoadLevel(uint level, const VolumeBase& src,
							bool bSrcIsStoredBackToFront, bool bSrcIsStoredBottomToTop,
							uint dataDepth, uint dataHeight, uint dataWidth,
							uint dstOffsetZ, uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetZ, uint srcOffsetY, uint srcOffsetX);


private:
	void  Allocate();
};


//********************************************************************
}//namespace gl_tools


//********************************************************************
}//namespace vertexware


