

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Component.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//********************************************************************
namespace layouters {


//====================================================================
enum eSizeFlag
{
	SF_Target = 0,
	SF_TargetMul,
	SF_TargetAdd,
	SF_Fixed,
	SF_Variable,
	SF_Auto,
	SF_Force32Bits = 0x7FFFFFFF
};


//====================================================================
enum eAlignFlag
{
	AF_Near = 0,
	AF_Far,
	AF_Center,
	AF_Target,
	AF_Force32Bits = 0x7FFFFFFF
};


//====================================================================
class VW_COMMON_API Cell
{
public:
	enum { DefaultWidth = 10 };

	//-----------------
	// Cell properties.
	//-----------------
	gui::Component*  target;
	eSizeFlag        sizeFlagX; //valid-flags: SF_*
	eSizeFlag        sizeFlagY; //all flags except SF_Variable are interpreted as SF_Target
	float            sizeX;
	float            sizeY;
	eAlignFlag       alignX;
	eAlignFlag       alignY;
	gui::Component*  alignTargetX;
	gui::Component*  alignTargetY;

	// Output data
	float  outputOffsetX;
	float  outputSizeX;


public:
	//----------------
	// Cell interface.
	//----------------
	Cell(gui::Component* target = 0,
		eSizeFlag sizeFlagX = SF_Target, float sizeX = 1,
		eAlignFlag alignX = AF_Near, gui::Component* alignTargetX = 0,
		eSizeFlag sizeFlagY = SF_Target, float sizeY = 1,
		eAlignFlag alignY = AF_Center, gui::Component* alignTargetY = 0);
};


//====================================================================
class VW_COMMON_API Row
{
public:
	enum { DefaultHeight = 32 };

	//----------------
	// Row properties.
	//----------------
	gui::Component*  target;
	eSizeFlag        sizeFlagY; //valid-flags: SF_*
	float            sizeY;

	// Output data
	float  outputOffsetY;
	float  outputSizeY;


public:
	//---------------
	// Row interface.
	//---------------
	Row(gui::Component* target = 0,
		eSizeFlag sizeFlagY = SF_Auto,
		float sizeY = 1);

	uint         NumCells() const { return cells.size(); }
	const Cell&  operator[](uint i) const { return cells[i]; }
	Cell&        operator[](uint i) { return cells[i]; }

	Cell*        FindCell(gui::Component& target);
	const Cell*  FindCell(gui::Component& target) const { return const_cast<Row*>(this)->FindCell(target); }

	Row&  Clear();
	Row&  Remove(const Cell& c);
	Row&  Remove(uint i);
	Row&  Append(const Cell& c);
	Row&  Prepend(const Cell& c);
	Row&  InsertBefore(const Cell& c, const Cell& before);
	Row&  InsertBefore(const Cell& c, uint iBefore);
	Row&  InsertAfter(const Cell& c, const Cell& after);
	Row&  InsertAfter(const Cell& c, uint iAfter);
	Row&  Append(const Row& r);
	Row&  Prepend(const Row& r);

	// Standard cell appends
	Row&  Target(gui::Component* target, eSizeFlag sizeFlagX = SF_Target, float sizeX = 1,
				eAlignFlag alignX = AF_Near, gui::Component* alignTargetX = 0,
				eAlignFlag alignY = AF_Center, gui::Component* alignTargetY = 0);

	Row&  Spacer(float sizeX = Cell::DefaultWidth, eSizeFlag sizeFlagX = SF_Fixed);
	Row&  AutoSpacer();

	// If requested cell is not available it appends AutoSpacer and returns it
	Cell&  GetCell(uint i);
	Cell&  LastCell();

	// Calculates horizonatal output-data, <totalWidth> is NOT forced (garbage-in, garbage-out)
	void  ProcessX(float totalWidth = 0);


protected:
	//--------------
	// Cell storage.
	//--------------
	typedef  std::vector<Cell>          CONTAINER;
	typedef  CONTAINER::iterator        CONTAINER_ITER;
	typedef  CONTAINER::const_iterator  CONTAINER_CITER;


protected:
	CONTAINER  cells;
};


//====================================================================
class VW_COMMON_API RowLayouter : public gui::Component
{
public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_COMMON_API GeomParams
	{
		vec2  preferredSize;
		bool  bAutoFitToParentOnRecalcLayout;
		vec2  autoFitPaddingTL;
		vec2  autoFitPaddingBR;

		GeomParams();
	};


public:
	//-----------------------
	// RowLayouter interface.
	//-----------------------
	RowLayouter(gui::Component* parent = 0);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	uint        NumRows() const { return rows.size(); }
	const Row&  operator[](uint i) const { return rows[i]; }
	Row&        operator[](uint i) { return rows[i]; }

	Row*        FindRow(gui::Component& target);
	const Row*  FindRow(gui::Component& target) const { return const_cast<RowLayouter*>(this)->FindRow(target); }

	RowLayouter&  Clear();
	RowLayouter&  Remove(const Row& r);
	RowLayouter&  Append(const Row& r);
	RowLayouter&  Prepend(const Row& r);
	RowLayouter&  InsertBefore(const Row& r, const Row& before);
	RowLayouter&  InsertAfter(const Row& r, const Row& after);

	// Standard row appends
	Row&  TargetHeightRow(gui::Component* target, eSizeFlag sizeFlagY = SF_Target, float sizeY = 1);
	Row&  FixedHeightRow(float sizeY = Row::DefaultHeight);
	Row&  VariableHeightRow(float sizeY = 1);
	Row&  AutoHeightRow();

	// If requested row is not available it appends AutoHeightRow and returns it
	Row&  GetRow(uint i);
	Row&  LastRow();

	// Calculates h/v output-data
	void  Process(float totalWidth = 0, float totalHeight = 0);

	// Calculates vertical output-data, <totalHeight> is NOT forced (garbage-in, garbage-out)
	void  ProcessY(float totalHeight = 0);


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();


protected:
	//------------
	// Properties.
	//------------
	GeomParams  geomParams;

	//-------------
	// Row storage.
	//-------------
	typedef  std::vector<Row>           CONTAINER;
	typedef  CONTAINER::iterator        CONTAINER_ITER;
	typedef  CONTAINER::const_iterator  CONTAINER_CITER;

	CONTAINER  rows;
};


//====================================================================
// Spacer helper.
//====================================================================
struct VW_COMMON_API Spacer
{
	enum { DefaultLength = 12 };

	const gui::Component*  pBefore;
	float                  length;

	Spacer(const gui::Component* _pBefore = 0, float _length = DefaultLength)
	{
		pBefore = _pBefore;
		length = _length;
	}
	bool operator==(const Spacer& rhs) const { return (pBefore == rhs.pBefore); }
	bool operator!=(const Spacer& rhs) const { return !(*this == rhs); }
};


//====================================================================
// Layouts children into array & auto-resizes self to minimum
// possible size.
//====================================================================
class VW_COMMON_API ArrayLayouter : public gui::Component
{
public:
	enum {
		DefaultAutoSpacing = 2,
		DefaultAutoSpacingHead = 0,
		DefaultAutoSpacingTail = 0
	};

	//----------------
	// Parameter sets.
	//----------------
	struct VW_COMMON_API GeomParams
	{
		bool  bVertical;
		bool  bResizeControlsToMax;
		float autoSpacing;
		float autoSpacingHead;
		float autoSpacingTail;

		GeomParams();
	};


public:
	//-------------------------
	// ArrayLayouter interface.
	//-------------------------
	ArrayLayouter(gui::Component* parent = 0);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	void  ClearSpacers();
	void  AddSpacer(const Spacer& s = Spacer());
	void  RemoveSpacer(const gui::Component* pBefore = 0);

	// Calculate minimum-possible-sizes & resize self
	void  Process();


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcLayout();

	//------------------
	// Internal helpers.
	//------------------
	Spacer*  FindSpacer(const gui::Component& before);


protected:
	//------------
	// Properties.
	//------------
	GeomParams           geomParams;
	std::vector<Spacer>  spacers;
};


//********************************************************************
}//namespace layouters


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


