

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Base.h"
#include "GUI\GUI_Component.h"
#include "GUI\GUI_Cursor.h"
#include "GUI\GUI_Shader.h"
#include "BaseTools\Singleton.h"
#include "BaseTools\Memory.h"
#include <vector>
#include <map>


//********************************************************************
namespace vertexware {


//====================================================================
class Kernel;


//********************************************************************
namespace gui {


//====================================================================
// GUI manager.
//====================================================================
class VW_COMMON_API Manager
{
VW_DECLARE_SINGLETON( Manager );

friend class Kernel;
friend class Component;
friend class BaseShader;
friend class Cursor;


public:
	Manager();
	~Manager();

	//-------------------------
	// Information / debugging.
	//-------------------------
	void  DumpComponentHierarchy(std::ostream& os, uint tabLevel = 0,
								bool bTopLevelOnly = false,
								bool bPrintDepth = true, bool bPrintAttribs = true);

	void  DumpCycleQueue(std::ostream& os, uint tabLevel = 0);

	void  DumpRegisteredTileTextures(std::ostream& os, uint tabLevel = 0);
	void  DumpRegisteredFonts(std::ostream& os, uint tabLevel = 0);
	void  DumpRegisteredShaders(std::ostream& os, uint tabLevel = 0);
	void  DumpRegisteredCursors(std::ostream& os, uint tabLevel = 0);

	//------------------------------------------------------
	// Internal recaches and relinks. Called automatically
	// by manager at every frame. Can be used by application
	// as pre-flush after change.
	//------------------------------------------------------
	void  SyncObjectStates();

	//---------------------------------------------
	// Per-frame tasks.
	// Should be called by application every frame.
	//---------------------------------------------
	void  DispatchInputEvents();
	void  Draw();

	//-----------------------------------------------------------
	// Services (only one provider per service can by registered)
	// Service-providers are NOT owned by manager.
	//-----------------------------------------------------------
	MessageHandler*  GetServiceProvider(uint serviceID);
	bool             RegisterServiceProvider(uint serviceID, MessageHandler& mh);
	bool             UnregisterServiceProvider(uint serviceID, MessageHandler& mh);
	bool             UnregisterServiceProvider(uint serviceID);
	void             UnregisterAllServiceProviders();

	bool  SendMessageToServiceProvider(uint serviceID, Message& msg);
	bool  BroadcastMessage(Message& msg); //to all attached components

	//---------------------------------------------------------------------------
	// Component hierarchy.
	// All created components are tracked and owned by manager,
	// i.e. they are auto-deleted by manager's destructor.
	// Components are divided into layers, and z-ordered inside the layer.
	// Layer is controlled by application through component's interface
	// and cannot be changed directly by user actions. Layer with higher-index
	// has priority over layer with lower-index when input events are dispatched.
	// Layers with lower-index are drawn before layers with higher-index.
	// Z-order is dynamically updated by user's interaction with components.
	// Input-event dispatch and drawing-order used for layers also apply to
	// z-order.
	//---------------------------------------------------------------------------
	void  DeleteAllComponents();

	uint        GetNumTopLevelComponents() { return componentsTopLevel.size(); }
	Component&  GetTopLevelComponent(uint i) { return *componentsTopLevel[i]; }

	// top-level cycle queue
	void  SetCycleQueueProperties(bool bEnableAutoCycle = true,
									bool bAutoInsertAtTopLevelCreate = true,
									bool bAutoInsertAtHierarchyChange = true);

	void  GetCycleQueueProperties(bool& bEnableAutoCycle,
									bool& bAutoInsertAtTopLevelCreate,
									bool& bAutoInsertAtHierarchyChange);

	void  CycleQueueClear();
	void  CycleQueueInsert(Component& c);
	void  CycleQueueRemove(Component& c);
	void  CycleQueueOrder(Component& before, Component& after);
	void  CycleQueueOrderToEnd(Component& c);
	void  CycleQueueOrderToStart(Component& c);
	void  CycleQueueInsert(Application& app);
	void  CycleQueueRemove(Application& app);
	void  CycleQueueOrder(Component& before, Application& app);
	void  CycleQueueOrder(Application& app, Component& after);
	void  CycleQueueOrderToEnd(Application& app);
	void  CycleQueueOrderToStart(Application& app);

	void  CycleQueueActivate(bool bForward);

	// per-layer z-order
	void  BringToFront(Component& c);
	void  SendToBack(Component& c);
	void  SetZOrder(Component& front, Component& back);

	void  ReportDirtyGeometry(Component& c);
	void  UpdateDirtyGeometry();
	void  UpdateAllGeometry();

	__forceinline  void  ForceHitTestAtNextFrame() { bForceHitTest = true; }

	//--------------
	// Video states.
	//--------------
	VideoRenderer*  GetVideoRenderer() { return pVideoRenderer; }
	gui::Renderer*  GetGUIRenderer() { return pGUIRenderer; }
	uint            GetWindowWidth() { return windowWidth; }
	uint            GetWindowHeight() { return windowHeight; }
	vec2            GetWindowSize() { return vec2(windowWidth, windowHeight); }

	// returns true if video-renderer is not connected
	bool  IntersectsWindow(Component& c);

	//---------------------------------------------------
	// Video resources.
	// All registered resources are owned by manager.
	// Same resource cannot be registered under multiple
	// names, name-aliasing is provided for that purpose.
	//---------------------------------------------------
	void  PrepareVideoRendererResources(bool bRecreateAllTileTextures = false,
										bool bRecreateAllFonts = false);

	// tile-texures
	void  RegisterTileTexture(const std::string& name, TileTexture* pTexture);
	void  RegisterTileTexture(const std::string& name, AutoPtr<TileTexture>& apTexture);
	void  UnregisterTileTexture(const std::string& name);
	void  UnregisterAllTileTextures();
	void  GetRegisteredTileTextureNames(std::vector<std::string>& outNames);
	bool  IsRegisteredTileTexture(TileTexture& texture, std::string* pOutName = 0);

	void  AddTileTextureAlias(const std::string& alias, const std::string& refName);
	void  ClearTileTextureAlias(const std::string& alias);
	void  ClearAllTileTextureAliases();
	void  GetTileTextureAliases(std::vector<std::string>& outAliases);
	bool  ResolveTileTextureAlias(const std::string& alias, std::string& refName);

	const TileTexture*  GetTileTexture(const std::string& name);

	void  RequestTileTextureRecache(VideoResourceUser& vru);

	// fonts
	void  RegisterFont(const std::string& name, Font* pFont);
	void  RegisterFont(const std::string& name, AutoPtr<Font>& apFont);
	void  RegisterFont(const std::string& name, AutoPtr<FontTexture>& apFontTexture);
	void  UnregisterFont(const std::string& name);
	void  UnregisterAllFonts();
	void  GetRegisteredFontNames(std::vector<std::string>& outNames);
	bool  IsRegisteredFont(Font& font, std::string* pOutName = 0);

	void  AddFontAlias(const std::string& alias, const std::string& refName);
	void  ClearFontAlias(const std::string& alias);
	void  ClearAllFontAliases();
	void  GetFontAliases(std::vector<std::string>& outAliases);
	bool  ResolveFontAlias(const std::string& alias, std::string& refName);

	const Font*  GetFont(const std::string& name);

	void  RequestFontRecache(VideoResourceUser& vru);

	//----------------------------------------------------------------
	// Shaders & cursors.
	// All created shaders and cursors are tracked and owned
	// by manager, i.e. they are auto-deleted by manager's destructor.
	//----------------------------------------------------------------

	// shaders
	void  DeleteAllShaders();
	void  DeleteShader(const std::string& shaderName);

	void  AddShaderAlias(const std::string& alias, const std::string& refName);
	void  ClearShaderAlias(const std::string& alias);
	void  ClearAllShaderAliases();
	void  GetShaderAliases(std::vector<std::string>& outAliases);
	bool  ResolveShaderAlias(const std::string& alias, std::string& refName);

	uint               GetNumShaders() { return shaders.size(); }
	BaseShader&        GetShader(uint i) { return *shaders[i]; }
	const BaseShader*  GetShader(const std::string& name);

	void  RequestShaderRecache(Component& c);

	// cursors
	void  DeleteAllCursors();
	void  DeleteCursor(const std::string& cursorName);

	void  AddCursorAlias(const std::string& alias, const std::string& refName);
	void  ClearCursorAlias(const std::string& alias);
	void  ClearAllCursorAliases();
	void  GetCursorAliases(std::vector<std::string>& outAliases);
	bool  ResolveCursorAlias(const std::string& alias, std::string& refName);

	uint           GetNumCursors() { return cursors.size(); }
	Cursor&        GetCursor(uint i) { return *cursors[i]; }
	const Cursor*  GetCursor(const std::string& name);

	void  RequestCursorRecache(Component& c);

	//---------------
	// Cursor states.
	//---------------
	void  DispatchGlobalCursorParams();

	void  ShowGUICursor(bool bShow = true);
	bool  IsShownGUICursor() const { return bShowGUICursor; }

	void  ShowSystemCursor(bool bShow = true);
	bool  IsShownSystemCursor() const { return bShowSystemCursor; }

	const Cursor&  GetActiveCursor() { return *pActiveCursor; }

	//---------------------
	// Input device states.
	//---------------------
	Keyboard*  GetKeyboard() { return pKeyboard; }
	ubyte      IsKeyDown(eKey kc) { return pKeyboard->KeyDown(kc); }
	ubyte      IsKeyDown() { return pKeyboard->KeyDown(); }
	ubyte      IsKeyUp(eKey kc) { return pKeyboard->KeyUp(kc); }
	ubyte      IsKeyUp() { return pKeyboard->KeyUp(); }
	ubyte      IsKeyPressed(eKey kc) { return pKeyboard->KeyPressed(kc); }
	ubyte      IsKeyPressed() { return pKeyboard->KeyPressed(); }
	ubyte      IsKeyModifierPressed(eKeyModifier km) { return pKeyboard->KeyModifierPressed(km); }
	ubyte      IsKeyModifierPressed() { return pKeyboard->KeyModifierPressed(); }

	Mouse*  GetMouse() { return pMouse; }
	ubyte   IsButtonDown(eMouseButton bc) { return pMouse->ButtonDown(bc); }
	ubyte   IsButtonDown() { return pMouse->ButtonDown(); }
	ubyte   IsButtonUp(eMouseButton bc) { return pMouse->ButtonUp(bc); }
	ubyte   IsButtonUp() { return pMouse->ButtonUp(); }
	ubyte   IsButtonPressed(eMouseButton bc) { return pMouse->ButtonPressed(bc); }
	ubyte   IsButtonPressed() { return pMouse->ButtonPressed(); }
	int     GetMouseWheelDelta() { return pMouse->GetDeltaWheel(); }
	vec2    GetMousePosition() { return vec2(pMouse->GetPositionX(), pMouse->GetPositionY()); }

	//---------------
	// Event control.
	//---------------
	void  SetKeyboardRepeat(double secsDelay = 0.25f, double secsPause = 0.025f);
	void  GetKeyboardRepeat(double& secsDelay, double& secsPause);

	void  SetMouseRepeat(double secsDelay = 0.25f, double secsPause = 0.025f);
	void  GetMouseRepeat(double& secsDelay, double& secsPause);

	void  SetMouseDoubleClick(double secsMaxToSecondClick = 0.25f);
	void  GetMouseDoubleClick(double& secsMaxToSecondClick);

	void  SetCycleQueueKeys(eKey kcCycle = Key_GraveAccent,
							eKeyModifier kmCycle = KeyModifier_Count,
							eKeyModifier kmBackward = KeyModifier_Shift);

	void  GetCycleQueueKeys(eKey& kcCycle, eKeyModifier& kmCycle, eKeyModifier& kmBackward);

	void  SetBehaviorOnWindowActivate(
								bool bIgnoreMouseClickAfterWindowActivate = true,
								bool bIgnoreMouseClickAfterWindowCreate = false);

	void  GetBehaviorOnWindowActivate(
								bool& bIgnoreMouseClickAfterWindowActivate,
								bool& bIgnoreMouseClickAfterWindowCreate);

	//-------------------------
	// Input control & queries.
	//-------------------------
	void  EnableApplicationMessageDispatch(bool bEnable = true);

	void  TimeNotifyRegister(Component& c, bool bRegister);
	void  TimeNotifyRegister(BaseShader& s, bool bRegister);
	void  TimeNotifyRegister(Cursor& c, bool bRegister);

	double  GetTimeDelta() { return secsTimeDelta; }

	double  GetMouseHoverTime() { return secsMouseHover; }
	void    ResetMouseHoverTime(double secsMouseHover = 0.0f);

	// returns last-target if called before mouse-event dispatch
	Component*  GetMouseOverTarget();
	bool        IsMouseOverTargetApplication() { return (GetMouseOverTarget() == pComponentToApplicationAdapter); }
	vec2        GetMousePositionDelta() { return GetMousePosition() - mousePosLast; }

	void  ActivateTopLevel(Component& c, bool bAutoBringToFront = true);
	void  ActivateApplication();

	Component*  GetActiveTopLevel();
	bool        IsApplicationActive();

	// modal target must be top-level component or application
	Component*  GetModalTarget() { return pModalTarget; }
	bool        IsModalTargetApplication() { return (pModalTarget == pComponentToApplicationAdapter); }
	bool        IsRunningModal() { return ToBool(pModalTarget); }
	void        RunModal(Component& c);
	void        EndModal(Component& c);
	void        RunModal(Application& app);
	void        EndModal(Application& app);

	Component*  GetMouseCaptureTarget() { return pMouseCaptureTarget; }
	bool        IsMouseCapturedByApplication() { return (pMouseCaptureTarget == pComponentToApplicationAdapter); }
	bool        IsMouseCaptured() { return (pMouseCaptureTarget != 0); }
	bool        RequestMouseCapture(Component& c);
	void        ReleaseMouseCapture(Component& c);
	bool        RequestMouseCapture(Application& app);
	void        ReleaseMouseCapture(Application& app);

	Component*  GetKeyboardFocusTarget() { return pKeyboardFocusTarget; }
	bool        IsKeyboardFocusTargetApplication() { return (pKeyboardFocusTarget == pComponentToApplicationAdapter); }
	bool        RequestKeyboardFocus(Component& c);
	void        ReleaseKeyboardFocus(Component& c);
	bool        RequestKeyboardFocus(Application& app);

	Component*    GetCommandTarget() { return pCommandTarget; }
	eCommandArea  GetCommandArea() { return commandArea; }
	bool          IsCommandModeActive() { return (pCommandTarget != 0); }
	void          BreakCommandMode();

	//---------------------------------
	// General depth-first-loop support
	//---------------------------------
	struct DepthFirstPathMarker
	{
		Component*  component;
		uint        iNextChild;
		bool        bUnprocessed;

		DepthFirstPathMarker(Component* _component) {
			component = _component;
			iNextChild = 0;
			bUnprocessed = true;
		}
	};
	typedef  std::vector<DepthFirstPathMarker>  DEPTHFIRSTPATHMARKER_VECTOR;

	//----------------------------------------------------
	// Depth-first-loop support extended for hit-test-loop
	//----------------------------------------------------
	struct DepthFirstPathMarkerHitTest : public DepthFirstPathMarker
	{
		int  iParentHitRecord;

		DepthFirstPathMarkerHitTest(Component* _component, uint _iParentHitRecord = -1)
			: DepthFirstPathMarker(_component)
		{
			iParentHitRecord = _iParentHitRecord;
		}
	};
	typedef  std::vector<DepthFirstPathMarkerHitTest>  DEPTHFIRSTPATHMARKERHITTEST_VECTOR;

	//---------------------
	// Hit-test information
	//---------------------
	struct HitRecord
	{
		Component*  component;
		int         depth; //0-based
		bool        bDescendantHit;

		HitRecord()
		{ }

		explicit HitRecord(Component* _component, int _depth = 0, bool _bDescendantHit = false)
		{
			component = _component;
			depth = _depth;
			bDescendantHit = _bDescendantHit;
		}

		void  Clear()
		{
			component = 0;
			depth = 0;
			bDescendantHit = false;
		}
	};
	typedef  std::vector<HitRecord>  HITRECORD_VECTOR;


private:
	//-------------
	// Type aliases
	//-------------
	typedef  std::vector<std::string>  STR_VECTOR;

	typedef  std::vector<VideoResourceUser*>  VRUSER_PTR_VECTOR;
	typedef  VRUSER_PTR_VECTOR::iterator      VRUSER_PTR_ITER;

	typedef  std::vector<BaseShader*>     SHADER_PTR_VECTOR;
	typedef  SHADER_PTR_VECTOR::iterator  SHADER_PTR_ITER;

	typedef  std::vector<Cursor*>         CURSOR_PTR_VECTOR;
	typedef  CURSOR_PTR_VECTOR::iterator  CURSOR_PTR_ITER;

	typedef  std::map<uint, MessageHandler*>  SERVICE_PROVIDER_MAP;
	typedef  SERVICE_PROVIDER_MAP::iterator   SERVICE_PROVIDER_ITER;

	typedef  std::map<std::string, std::string>  STRSTR_MAP;
	typedef  STRSTR_MAP::iterator                STRSTR_ITER;

	typedef  AutoPtr<TileTexture>                 TILETEX_APTR;
	typedef  std::map<std::string, TILETEX_APTR>  TILETEX_MAP;
	typedef  TILETEX_MAP::iterator                TILETEX_ITER;

	typedef  AutoPtr<Font>                     FONT_APTR;
	typedef  std::map<std::string, FONT_APTR>  FONT_MAP;
	typedef  FONT_MAP::iterator                FONT_ITER;

	//------------
	// Null target
	//------------
	class NullTarget : public gui::Component
	{
	public:
		NullTarget();

	protected:
		virtual  bool  HandleMessage(gui::Message& msg) { return false; }
	};

	//--------------------
	// Application adapter
	//--------------------
	class ComponentToApplicationAdapter : public gui::Component
	{
	public:
		ComponentToApplicationAdapter();

		void  ConnectApplication(bool bConnect);
		void  EnableMessageDispatch(bool bEnable);

	protected:
		virtual  bool  HandleMessage(gui::Message& msg);

	private:
		Application*  pApp;
		bool          bEnableMessageDispatch;
	};

	//-----------------
	// Internal helpers
	//-----------------
	void  DisplaySystemCursor(bool bDisplay);
	void  NotifyVideoRendererChanged();
	void  NotifyWindowResize();
	void  NotifyTimeDelta();
	void  NotifyServiceProvidersPreEventDispatch();
	void  NotifyServiceProvidersPostEventDispatch();
	void  UpdateComponentHierarchyChanges();
	void  UpdateComponentLayers();
	void  UpdateComponentQueues();
	void  DispatchKeyboardEvents();
	void  NotifyKeyboardKey(eKey keyCode, bool bDown, uint repeatCount, char ch);
	void  DispatchMouseEvents();
	void  NotifyMouseExit(const vec2& mousePos);
	void  NotifyMouseOver(const vec2& mousePos);
	void  ReportMouseOver(Component& c, bool bOver, bool bOverDirect, const vec2& mousePos);
	void  NotifyMouseMove(Component& c, const vec2& mousePos);
	void  NotifyMouseHover(Component& c, const vec2& mousePos);
	void  NotifyMouseWheel(Component& c, const vec2& mousePos);
	void  NotifyMouseButtons(const vec2& mousePos);
	void  NotifyActivate(const COMPONENT_PTR_VECTOR& c, bool bByMouse, const vec2& mousePos);
	void  NotifyDeactivate(const COMPONENT_PTR_VECTOR& c, Component* cIgnore, bool bByMouse);
	void  NotifyDeactivate(Component& c, bool bByMouse);
	bool  IsModalTargetDescendant(Component& c);
	void  EndMouseCapture();
	void  EndCommandMode();
	void  ChangeKeyboardFocusTarget(Component& c, bool bByMouse, eMouseButton buttonCode = MouseButton_Count);
	void  SetupKeyboardForFocusTarget();
	void  ResetKeyboardRepeat();
	void  ResetMouseRepeat();
	void  ResetMouseDoubleClick();
	void  ResetMouseButtonEvents();
	void  ProcessMouseButtons();
	void  InvalidateHitTestCache();
	void  HitTestCollect(const vec2& testPointAbs, HITRECORD_VECTOR& out, HitRecord& outDeep);

	void  HitTestCollect(const vec2& testPointAbs, HITRECORD_VECTOR& out, HitRecord& outDeep,
						Component& topLevel, HITRECORD_VECTOR& tempHitRecords,
						DEPTHFIRSTPATHMARKERHITTEST_VECTOR& tempPathMarkersHitTest);

	void  ExtractHitRecordDifferences();
	void  InvalidateMouseClickedBranchCache();
	void  SetupMouseClickedBranchCache(COMPONENT_PTR_VECTOR& clickedTargets, eMouseButton clickedButton);
	void  ExtractMouseClickedBranch();
	void  ExtractMouseClickedBranchDifferences();
	void  DispatchMouseClickToBranches(const vec2& mousePos);
	void  EraseFromHitTestCache(Component& c);
	void  NullifyHitTestCache(HitRecord& hr, Component& c);
	void  NullifyHitTestCache(HITRECORD_VECTOR& hrv, Component& c);
	void  EraseFromMouseClickedBranchCache(Component& c);
	void  SetupActiveCursor(Component& c, eCommandArea cursorArea, const vec2& mousePos);

	//---------------------------------------
	// Automatic object-manager notifications
	//---------------------------------------
	void  NotifyObjectCreateDelete(BaseShader& s, bool bCreate);
	void  NotifyObjectCreateDelete(Cursor& c, bool bCreate);
	void  NotifyObjectCreateDelete(Component& c, bool bCreate);
	void  NotifyComponentHierarchyChanged(Component& c);
	void  NotifyComponentLayerChanged(Component& c);
	void  NotifyComponentOverlayChanged(Component& c);
	void  NotifyComponentVisibilityChanged(Component& c);
	void  NotifyNameChanged(BaseShader& s);
	void  NotifyNameChanged(Cursor& c);
	void  NotifyGlobalCursorParamsChanged();

	//--------------------------------------------
	// Application notifications, called by kernel
	//--------------------------------------------
	void  ConnectApplication(bool bConnect);
	void  NotifyApplicationWindowActive(bool bActive);
	void  NotifyApplicationWindowCreated();

	//------------------------------------
	// HID notifications, called by kernel
	//------------------------------------
	void  BeginConnectHumanInterfaceDevices();
	void  ConnectHumanInterfaceDevice(HumanInterfaceDevice& hid);
	void  EndConnectHumanInterfaceDevices();
	void  DisconnectHumanInterfaceDevices();
	void  SetTolerateMissingInputDevice(bool bTolerate);

	//--------------------------------------
	// Video notifications, called by kernel
	//--------------------------------------
	void  ConnectVideoRenderer(VideoRenderer& vr);
	void  DisconnectVideoRenderer();
	void  CreateVideoResources(bool bPostReset);
	void  ReleaseVideoResources(bool bPreReset);
	void  OnWindowResize(uint width, uint height);

	//-----------------
	// Resource updates
	//-----------------
	void  DoCreateVRRTileTextures(bool bRecreateAll);
	void  DoCreateVRRFonts(bool bRecreateAll);

	//-----------------------
	// Link recache protocols
	//-----------------------
	void  CollectAllVRUsers(VRUSER_PTR_VECTOR& vru);

	void  VRUser_CacheTileTextureHandles();
	void  VRUser_PostCacheTileTextureHandles();
	void  VRUser_TileTextureListChanged();
	void  VRUser_CacheFontHandles();
	void  VRUser_PostCacheFontHandles();
	void  VRUser_FontListChanged();

	void  Component_CacheCursorHandles();
	void  Component_PostCacheCursorHandles();
	void  Component_CursorListChanged();
	void  Component_CacheShaderHandles();
	void  Component_PostCacheShaderHandles();
	void  Component_ShaderListChanged();

	void  Internal_CacheCursorHandles();
	void  Internal_InvalidateCursorHandles();


private:
	// update control
	bool  bAllowMainContainerUpdateAtNotifyComponent;
	bool  bAllowMainContainerUpdateAtNotifyShader;
	bool  bAllowMainContainerUpdateAtNotifyCursor;

	bool  bComponentLayersChanged;
	bool  bComponentHierarchyChanged;

	bool  bTileTextureObjectsChanged;
	bool  bFontObjectsChanged;
	bool  bShaderObjectsChanged;
	bool  bCursorObjectsChanged;

	bool  bTileTextureListChanged;
	bool  bFontListChanged;

	bool  bNeedFullTileTextureRecache;
	bool  bNeedFullFontRecache;
	bool  bNeedFullShaderRecache;
	bool  bNeedFullCursorRecache;
	bool  bNeedInternalCursorRecache;
	bool  bNeedGlobalCursorParamsUpdate;
	bool  bNeedFullGeometryUpdate;
	bool  bNeedFullWindowResizeUpdate;

	// recache queues
	VRUSER_PTR_VECTOR     tileTextureRecacheQueue;
	VRUSER_PTR_VECTOR     fontRecacheQueue;
	COMPONENT_PTR_VECTOR  shaderRecacheQueue;
	COMPONENT_PTR_VECTOR  cursorRecacheQueue;

	// service providers
	SERVICE_PROVIDER_MAP  mapServiceProviders;

	// component hierarchy
	NullTarget*                     pNullTarget;
	ComponentToApplicationAdapter*  pComponentToApplicationAdapter;

	COMPONENT_PTR_VECTOR  componentsTopLevel;
	COMPONENT_PTR_VECTOR  componentsAll;
	COMPONENT_PTR_VECTOR  componentsDrawQueue;
	COMPONENT_PTR_VECTOR  componentsDirtyGeometry;
	COMPONENT_PTR_VECTOR  componentsWindowResize;
	COMPONENT_PTR_VECTOR  componentsCycleQueue;

	// queue-construction & hit-test temporary storage
	COMPONENT_PTR_VECTOR                tsOverlays;
	DEPTHFIRSTPATHMARKER_VECTOR         tsDepthFirstPathMarkers;
	DEPTHFIRSTPATHMARKERHITTEST_VECTOR  tsDepthFirstPathMarkersHitTest;
	HITRECORD_VECTOR                    tsHitRecords;

	// video states
	VideoRenderer*  pVideoRenderer;
	gui::Renderer*  pGUIRenderer;
	bool            bVideoRendererReadyToCreateResources;
	bool            bRendererCreatesManagedResourcesOnly;
	uint            windowWidth;
	uint            windowHeight;

	// video-resources & gui-objects storage
	TILETEX_MAP        mapTileTextures;
	STRSTR_MAP         mapTileTextureAliases;
	FONT_MAP           mapFonts;
	STRSTR_MAP         mapFontAliases;
	SHADER_PTR_VECTOR  shaders;
	STRSTR_MAP         mapShaderAliases;
	CURSOR_PTR_VECTOR  cursors;
	STRSTR_MAP         mapCursorAliases;
	STR_VECTOR         vrrCreateListTileTextures;
	STR_VECTOR         vrrCreateListFonts;

	// cursor states
	bool           bShowGUICursor;
	bool           bShowSystemCursor;
	bool           bVisibleSystemCursor;
	Cursor*        pNullCursor;
	const Cursor*  cursorHandles[CommandArea_Count];
	const Cursor*  pActiveCursor;

	// input devices
	bool       bTolerateMissingInputDevice;
	Keyboard*  pKeyboard;
	Mouse*     pMouse;

	// event control
	bool          bIgnoreMouseClickAfterWindowActivate;
	bool          bIgnoreMouseClickAfterWindowCreate;
	double        fKeyboardRepeatDelay;
	double        fKeyboardRepeatPause;
	double        fMouseRepeatDelay;
	double        fMouseRepeatPause;
	double        fDoubleClickLimit;

	bool          bCycleQueueEnableAutoCycle;
	bool          bCycleQueueAutoInsertAtCreate;
	bool          bCycleQueueAutoInsertAtHierarchyChange;
	eKey          kcCycleQueue;
	eKeyModifier  kmCycleQueue;
	eKeyModifier  kmCycleQueueBackward;

	// input states
	COMPONENT_PTR_VECTOR  timeNotifyComponents;
	SHADER_PTR_VECTOR     timeNotifyShaders;
	CURSOR_PTR_VECTOR     timeNotifyCursors;

	double  secsTimeDelta;
	double  secsMouseHover;

	Component*    pCommandTarget;
	eCommandArea  commandArea;
	eMouseButton  commandButton;
	vec2          grabPosScreen;
	vec2          grabPosTargetOriginal;
	vec2          grabSizeTargetOriginal;

	COMPONENT_PTR_VECTOR  modalTargetStack;
	Component*            pModalTarget;
	Component*            pMouseCaptureTarget;
	Component*            pKeyboardFocusTarget;
	Component*            pMouseButtonNotifyTarget;
	Component*            pForcedKeyboardTarget;

	vec2  mousePosLast;
	bool  bLastMousePosInsideAppWindow;
	bool  bIgnoreMouseButtonEvents;
	bool  bForceHitTest;
	bool  bAllowExternalTopLevelActivate;
	bool  bTargetNeedsCharacterData;

	eKey   kRepeatKeyCode;
	char   kRepeatChar;
	uint   kRepeatCount;
	float  kRepeatTimeLast;
	float  kRepeatTimeAccum;

	eMouseButton  mRepeatButtonCode;
	uint          mRepeatCount;
	float         mRepeatTimeLast;
	float         mRepeatTimeAccum;

	eMouseButton  mDoubleClickButtonCode;
	bool          mDoubleClickEvent;
	float         mDoubleClickTimeAccum;

	uint          mNumEvents;
	uint          mNumClicks;
	eMouseButton  mFirstClickedButtonCurr;
	eMouseButton  mFirstClickedButtonLast;
	ubyte         mButtonDown[MouseButton_Count];
	ubyte         mButtonUp[MouseButton_Count];

	eKey  keyDown[Key_Count];
	eKey  keyUp[Key_Count];
	char  keyDownChar[Key_Count];
	char  keyUpChar[Key_Count];

	HITRECORD_VECTOR  hitRecordsLast;
	HITRECORD_VECTOR  hitRecordsCurr;
	HITRECORD_VECTOR  hitRecordsDiffExit;
	HITRECORD_VECTOR  hitRecordsDiffEnter;
	HitRecord         hitRecordDeepLast;
	HitRecord         hitRecordDeepCurr;

	COMPONENT_PTR_VECTOR  mouseClickedBranchLast;
	COMPONENT_PTR_VECTOR  mouseClickedBranchCurr;
	COMPONENT_PTR_VECTOR  mouseClickedBranchDiffLastNotCurr;
	COMPONENT_PTR_VECTOR  mouseClickedBranchDiffCurrNotLast;
};


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


