

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_BehaviorBases.h"
#include "private\vw_gui_cs0_exports.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
// Internal hierarchy:
// * ComboBox
//   * ComboBoxButton (Open)
//	 * ComboBoxPane [overlay]
//	   * ComboBoxPaneButton (Up)
//	   * ComboBoxPaneButton (Down)
//====================================================================
class ComboBoxItem;
class ComboBoxPane;
class ComboBoxPaneButton;
class ComboBox;
class ComboBoxButton;


//====================================================================
class VW_GUI_CS0_API ComboBoxItem
{
public:
	ComboBoxItem(const std::string& text = "ComboBoxItem",
				int id = 0);

	const std::string&  GetText() const { return text; }
	int                 GetID() const { return id; }


protected:
	std::string  text;
	int          id;
};


//====================================================================
class VW_GUI_CS0_API ComboBoxPaneButton : public gui::PressButton
{
friend class ComboBoxPane;

protected:
	ComboBoxPaneButton(ComboBoxPane& comboBoxPane);

	//---------------------
	// Component overrides.
	//---------------------
	virtual  bool  OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos);
	virtual  void  OnTimeDelta(double secs);

	//----------------------------
	// PressButton implementation.
	//----------------------------
	virtual  void  ProcessButtonPress();


protected:
	//------------
	// Properties.
	//------------
	ComboBoxPane&  comboBoxPane;

	//--------
	// States.
	//--------
	bool    bClickedBeforeAutoScroll;
	bool    bFirstAutoScrollEmit;
	double  secsAccumTime;
};


//====================================================================
class VW_GUI_CS0_API ComboBoxPane : public gui::Component
{
friend class ComboBoxPaneButton;
friend class ComboBox;

public:
	//-----------------------------------------------
	// Additional methods intended for use by shader.
	//-----------------------------------------------
	const ComboBox&            GetComboBox() const { return comboBox; }
	const ComboBoxPaneButton&  GetButtonBackward() const { return *pButtonBackward; }
	const ComboBoxPaneButton&  GetButtonForward() const { return *pButtonForward; }

	bool  GetButtonsVisible() const { return pButtonBackward->IsVisible(); }

	// return false if no items are visible
	bool  GetVisibleItemRange(int& first, int& last) const;

	// return < 0 if there is no hot-item
	int   GetHotItem() const { return firstVisibleItem + hotItem; }
	int   GetHotItemOffset() const { return hotItem; }


protected:
	ComboBoxPane(ComboBox& comboBox);

	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();
	virtual  void  UpdateGeometryCache();
	virtual  void  DrawSelf(gui::Renderer& gr);
	virtual  bool  OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos);
	virtual  bool  OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos);
	virtual  bool  OnMouseWheel(int delta, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  void  OnChangedVisible();

	//------------------
	// Internal helpers.
	//------------------
	void  CalcVisibleItemRange();
	void  UpdateVisibleItemRange(int scroll);
	void  ScrollItems(const ComboBoxPaneButton& button);
	int   CalcHotItem(const vec2& mousePos);


protected:
	//----------------
	// Sub-components.
	//----------------
	ComboBoxPaneButton*  pButtonForward;
	ComboBoxPaneButton*  pButtonBackward;

	//------------
	// Properties.
	//------------
	ComboBox&  comboBox;

	//--------
	// States.
	//--------
	int  firstVisibleItem;
	int  lastVisibleItem;
	int  hotItem;
};


//====================================================================
class VW_GUI_CS0_API ComboBoxButton : public gui::PushButton
{
friend class ComboBox;

protected:
	ComboBoxButton(ComboBox& comboBox);

	//---------------------------
	// PushButton implementation.
	//---------------------------
	virtual  void  ProcessButtonRelease();


protected:
	//------------
	// Properties.
	//------------
	ComboBox&  comboBox;
};


//====================================================================
// Items are owned by combo-box.
// Feature support:
// CanDisable.
//====================================================================
class VW_GUI_CS0_API ComboBox : public gui::Component, public gui::CommandSender
{
friend class ComboBoxPane;

public:
	//--------------------------------------------------
	class VW_GUI_CS0_API Shader : public gui::BaseShader
	{
	public:
		//------------------
		// Shader interface.
		//------------------
		Shader(const std::string& shaderName = "") : BaseShader(shaderName) {}

		virtual  void  Shade(const gui_cs0::ComboBox& cbox, gui::Renderer& gr) const = 0;
		virtual  void  Shade(const gui_cs0::ComboBoxPane& pane, gui::Renderer& gr) const = 0;

		//---------------------------
		// BaseShader implementation.
		//---------------------------
		virtual  void  Shade(const gui::Component& c, gui::Renderer& gr) const {}
	};
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);


public:
	enum ePanePosition
	{
		PanePosition_Top = 0,
		PanePosition_Bottom,
		PanePosition_Center,

		PanePosition_Count,
		PanePosition_Force32Bits = 0x7FFFFFFF
	};

	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		uint           maxVisibleItems;
		ePanePosition  panePosition;
		vec2           preferredSize;
		float          openButtonWidth;
		bool           bOpenButtonRightAlign; //otherwise left-align
		float          paneButtonHeight;
		float          textHeightBox;
		float          textPaddingBox;
		float          textHeightPane;
		float          textPaddingPane;
		float          itemHeightPane;

		GeomParams();
	};


public:
	//--------------------
	// ComboBox interface.
	//--------------------
	ComboBox(gui::Component* parent = 0,
			gui::MessageHandler* commandTarget = 0,
			uint commandID = 0);

	virtual ~ComboBox();

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	// set negative delay to disable auto-scroll
	void    SetAutoScrollDelay(double secs = 0.75) { autoScrollDelay = secs; }
	double  GetAutoScrollDelay() const { return autoScrollDelay; }

	void    SetAutoScrollRepeat(double secs = 0.2) { autoScrollRepeat = secs; }
	double  GetAutoScrollRepeat() const { return autoScrollRepeat; }

	void    SetPopupHoverDelay(double secs = 0.5) { popupHoverDelay = secs; }
	double  GetPopupHoverDelay() const { return popupHoverDelay; }

	void   OpenPane(bool bOpen);
	bool   IsPaneOpened() const { return pPane->IsVisible(); }

	//-------
	// Items.
	//-------
	void  SelectItem(ComboBoxItem& cbi);
	void  SelectItem(uint i);
	void  SelectPreviousItem();
	void  SelectNextItem();

	ComboBoxItem*        GetSelectedItem()       { return pSelectedItem; }
	const ComboBoxItem*  GetSelectedItem() const { return pSelectedItem; }

	// returns -1 if no items are available
	int   GetSelectedItemIndex() const { return iSelectedItem; }
	bool  IsSelected(const ComboBoxItem& cbi) const { return (pSelectedItem == &cbi); }

	void  AddItem(ComboBoxItem* cbi) { if(cbi) AddItem(*cbi); }
	void  AddItem(ComboBoxItem& cbi);
	void  RemoveItem(ComboBoxItem& cbi);
	void  RemoveAllItems();
	void  DeleteItem(ComboBoxItem& cbi);
	void  DeleteAllItems();
	bool  OwnsItem(const ComboBoxItem& cbi) const;

	uint                 GetNumItems() const { return items.size(); }
	ComboBoxItem&        GetItem(uint i) { return *items[i]; }
	const ComboBoxItem&  GetItem(uint i) const { return *items[i]; }

	//-----------------------------------------------
	// Additional methods intended for use by shader.
	//-----------------------------------------------
	const ComboBoxButton&  GetButtonOpen() const { return *pButtonOpen; }

	bool                GetShowSelectedItemPopup() const { return (bNeedsSelectedItemPopup && bShowSelectedItemPopup); }
	const std::string&  GetSelectedItemText() const { return selectedItemText; }
	float               GetSelectedItemTextWidth() const { return fullSelectedTextWidth; }
	uint                GetSelectedItemVisibleTextLength() const { return selectedItemVisibleTextLength; }
	uint                GetItemVisibleTextLengthPane(uint i) const { return itemVisibleTextLengthsPane[i]; }
	float               GetItemFullTextWidthPane(uint i) const { return itemFullTextWidthsPane[i]; }


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();
	virtual  void  UpdateGeometryCache();
	virtual  void  DrawSelf(gui::Renderer& gr);
	virtual  void  PostCacheShaderHandles();
	virtual  bool  OnDeactivate(bool bDeactivateByMouse);
	virtual  bool  OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos);
	virtual  bool  OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  void  OnChangedEnabled();

	//------------------
	// Internal helpers.
	//------------------
	void  UpdateSelectedItemCache();
	void  UpdatePaneItemCache();


protected:
	//----------------
	// Sub-components.
	//----------------
	ComboBoxButton*  pButtonOpen;
	ComboBoxPane*    pPane;

	//------------
	// Properties.
	//------------
	double      autoScrollDelay;
	double      autoScrollRepeat;
	double      popupHoverDelay;
	GeomParams  geomParams;

	//--------
	// States.
	//--------
	typedef  std::vector<ComboBoxItem*>  ITEM_VEC;
	typedef  ITEM_VEC::iterator          ITEM_ITER;
	typedef  ITEM_VEC::const_iterator    ITEM_CITER;

	ITEM_VEC       items;
	ComboBoxItem*  pSelectedItem;
	int            iSelectedItem;
	bool           bActiveSelectedItemPopup;

	//-------
	// Cache.
	//-------
	std::vector<uint>   itemVisibleTextLengthsPane;
	std::vector<float>  itemFullTextWidthsPane;
	const Shader*       pComboBoxShader;
	bool                bShowSelectedItemPopup;
	bool                bNeedsSelectedItemPopup;
	float               maxWidthSelectedText;
	std::string         selectedItemText;
	uint                selectedItemVisibleTextLength;
	float               fullSelectedTextWidth;
};


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


