

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI_CS0\Buttons.h"
#include "GUI_CS0\ShaderNames.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
// Expected hierarchy:
// * MenuBar
//   * MenuTitle
//	   * MenuPane
//	     * MenuItem
//       * MenuSeparator
//	     * MenuItem
//         * MenuPane
//           * MenuItem
//           * MenuSeparator
//           ...
//       ...
//   ...
//====================================================================
class MenuBar;
class MenuTitle;
class MenuPane;
class MenuItem;
class MenuSeparator;


//====================================================================
// Feature support:
// Can dock to top area of the screen.
// Dock-sensor is at top-left corner.
//====================================================================
class VW_GUI_CS0_API MenuBar : public gui::Component
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);


public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		float  height;
		float  paddingLeft;
		float  paddingRight;
		float  spacingButtonTitle;
		float  spacingTitles;
		float  dockSensorWidth;

		GeomParams();
	};

	//----------------------------------------------------------------------
	static IconButton::GeomParams __fastcall DefaultGeomParams_IconButton();


public:
	//-------------------
	// MenuBar interface.
	//-------------------
	MenuBar();

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	void  SetMenuBarShaderNames(
				const std::string& menuBar = DefaultShaderName(Shader_MenuBar),
				const std::string& iconButtonBarControl = DefaultShaderName(Shader_MenuBar_IconButton_BarControl));

	void  SetMenuBarSubcomponentLayout(
						bool bShowCloseButton = false);

	// pass empty string to disable tooltip
	void  SetMenuBarSubcomponentToolTipText(
						const std::string& buttonClose = "Close");

	void  EnableUndock(bool bEnable);
	bool  IsEnabledUndock() { return bEnableUndock; }

	void  SetMenuBarPosition(bool bDock, const vec2& position = vec2::ZERO);
	bool  IsDocked() const { return bDocked; }

	bool  IsMovingInDockSensorArea() const { return bMovingInDockSensorArea; }


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();
	virtual  void  UpdateGeometryCache();
	virtual  void  OnWindowResize(uint width, uint height);
	virtual  gui::eCommandArea  HitTestCommandAreaSelf(const vec2& testPoint) const;
	virtual  bool  OnCommand(uint commandID);

	virtual  bool  OnGeometryUpdateByManager(bool& bConfirmEvent,
								gui::eGeometryUpdateEvent updateEvent,
								gui::eCommandArea commandArea, eMouseButton buttonCode,
								const vec2& newPosition, const vec2& newSize);

	virtual  bool  OnActivate(bool bActivateByMouse, eMouseButton buttonCode,
							const vec2& mousePosition, gui::Component* pClickedDescendant,
							bool& bAutoBringToFront, bool& bAutoActivate);

	virtual  bool  OnDeactivate(bool bDeactivateByMouse);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  bool  OnMouseButton(gui::MessageHandler* target, bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  void  OnChangedVisible();

	//------------------
	// Internal helpers.
	//------------------
	void  CloseMenu();
	void  RealizeDock(bool bDock, const vec2& undockedPosition);


protected:
	//----------------
	// Sub-components.
	//----------------
	IconButton*  pButtonClose;

	//------------
	// Properties.
	//------------
	GeomParams  geomParams;
	bool        bEnableUndock;

	//--------
	// States.
	//--------
	bool  bDocked;
	bool  bMovingInDockSensorArea;

	//-------
	// Cache.
	//-------
	gui::Rectangle  rDockSensor;
};


//====================================================================
// Feature support:
// CanDisable.
//====================================================================
class VW_GUI_CS0_API MenuTitle : public gui::Component
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);


public:
	//-------
	// Flags.
	//-------
	enum eFlag
	{
		Flag_TitleOpened = 1,
		Flag_TitleActive = 2
	};


public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		float  textHeight;
		vec2   paneOffset;

		GeomParams();
	};


public:
	//---------------------
	// MenuTitle interface.
	//---------------------
	MenuTitle(MenuBar* parent, const std::string& text = "MenuTitle");

	void  AttachToMenuBar(MenuBar& menuBar);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	void                SetText(const std::string& text);
	const std::string&  GetText() const { return text; }

	void  SetTitleOpened(bool bOpened) { SetFlagsOnOff(Flag_TitleOpened, bOpened); }
	bool  IsTitleOpened() const { return ToBool(IsFlagOn(Flag_TitleOpened)); }

	void  SetTitleActive(bool bActive) { SetFlagsOnOff(Flag_TitleActive, bActive); }
	bool  IsTitleActive() const { return ToBool(IsFlagOn(Flag_TitleActive)); }


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();
	virtual  bool  OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  bool  OnMsgFlagsOnOff(uint flags, bool bOn);
	virtual  bool  OnMsgPreCommandExecute(uint commandID);
	virtual  void  OnChangedFlags();


protected:
	//------------
	// Properties.
	//------------
	GeomParams   geomParams;
	std::string  text;
};


//====================================================================
class VW_GUI_CS0_API MenuPane : public gui::Component
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);


public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		vec2   padding;
		float  minWidth;

		GeomParams();
	};


public:
	//--------------------
	// MenuPane interface.
	//--------------------
	MenuPane();
	MenuPane(MenuTitle& parent);
	MenuPane(MenuItem& parent);

	void  AttachToMenuTitle(MenuTitle& menuTitle);
	void  AttachToMenuItem(MenuItem& menuItem);

	// when used as context-menu call AttachToManager(), then ShowAsContextMenu()
	void  ShowAsContextMenu(bool bShow, const vec2& position = vec2::ZERO);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();
	virtual  void  OnWindowResize(uint width, uint height);
	virtual  void  OnChangedVisible();
	virtual  bool  OnDeactivate(bool bDeactivateByMouse);
	virtual  bool  OnMsgPreCommandExecute(uint commandID);


protected:
	//------------
	// Properties.
	//------------
	GeomParams   geomParams;


private:
	void  Init();
};


//====================================================================
class VW_GUI_CS0_API MenuSeparator : public gui::Component
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);


public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		float  height;

		GeomParams();
	};


public:
	//-------------------------
	// MenuSeparator interface.
	//-------------------------
	MenuSeparator(MenuPane* parent = 0);

	void  AttachToMenuPane(MenuPane& menuPane);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcSize();


protected:
	//------------
	// Properties.
	//------------
	GeomParams  geomParams;
};


//====================================================================
// Feature support:
// CanDisable.
//====================================================================
class VW_GUI_CS0_API MenuItem : public gui::Component,
								public gui::CommandSender
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);


public:
	//-------
	// Flags.
	//-------
	enum eFlag
	{
		Flag_On       = 1,
		Flag_Selected = 2
	};


public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		float  textHeight;
		float  itemHeight;
		float  padding;
		float  spacingIconText;
		float  iconSize;
		vec2   paneOffset;

		GeomParams();
	};

	//------------
	// Item style.
	//------------
	enum eStyle
	{
		Style_Default = 0,
		Style_OnOff,
		Style_Selection,

		Style_Count,
		Style_Force32Bits = 0x7FFFFFFF
	};


public:
	//--------------------
	// MenuItem interface.
	//--------------------
	MenuItem(MenuPane* parent = 0,
			gui::MessageHandler* commandTarget = 0,
			uint commandID = 0,
			const std::string& text = "MenuItem",
			eStyle style = Style_Default,
			uint groupID = 0);

	void  AttachTo(MenuPane& menuPane);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	void                SetText(const std::string& text);
	const std::string&  GetText() const { return text; }

	void    SetStyle(eStyle style = Style_Default);
	eStyle  GetStyle() const { return style; }

	void    SetGroupID(uint _groupID = 0) { groupID = _groupID; }
	uint    GetGroupID() const { return groupID; }

	void    SetOn(bool bOn);
	bool    IsOn() const { return ToBool(IsFlagOn(Flag_On)); }

	void    SetSelected(bool bSelected);
	bool    IsSelected() const { return ToBool(IsFlagOn(Flag_Selected)); }
	void    Select() { SetSelected(true); }


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  OnAttachedChild(gui::Component& child);
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();
	virtual  bool  OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  bool  OnMsgOnOff(bool bOn);
	virtual  bool  OnMsgSelect(bool bSelect, uint groupID);


protected:
	//------------
	// Properties.
	//------------
	GeomParams   geomParams;
	std::string  text;
	eStyle       style;
	uint         groupID;
};


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


