

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Component.h"
#include "GUI\GUI_Shader.h"
#include "private\vw_gui_cs0_exports.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
// Expected hierarchy:
//
// * PropertyFrame
//   * PropertyTab
//     * PropertyPage
//       * [controls]
//   * PropertyTab
//     * PropertyPage
//       * [controls]
//   ...
//====================================================================
class PropertyFrame;
class PropertyTab;
class PropertyPage;


//====================================================================
class VW_GUI_CS0_API PropertyFrame : public gui::Component
{
public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		float  tabOffsetLeft;
		float  tabOffsetLeftSelected;
		float  tabSpacing;
		float  tabSpacingSelected;
		vec2   tabSelectedExtend;
		float  tabTextHeight;
		float  tabPadding;
		vec2   tabMinSize;

		GeomParams();
	};


public:
	//-------------------------
	// PropertyFrame interface.
	//-------------------------
	PropertyFrame(gui::Component* parent = 0);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	void  EnableAutoUpdateControlsOnSelect(bool bEnable = true) { bAutoUpdateControlsOnSelect = bEnable; }
	bool  IsEnabledAutoUpdateControlsOnSelect() const { return bAutoUpdateControlsOnSelect; }

	// updates all attached property-pages
	void  UpdateControlDependencies(bool bSelectedOnly = true);
	void  ApplyControlValues(bool bSelectedOnly = true);


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  OnAttachedChild(gui::Component& child);
	virtual  void  RecalcLayout();


protected:
	//------------
	// Properties.
	//------------
	GeomParams  geomParams;
	bool        bAutoUpdateControlsOnSelect;
};


//====================================================================
// Notifies target about tab selection.
//====================================================================
class VW_GUI_CS0_API PropertyTab : public gui::Component, public gui::CommandSender
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);


public:
	//-------
	// Flags.
	//-------
	enum eFlag
	{
		Flag_Selected = 1
	};


public:
	//-----------------------
	// PropertyTab interface.
	//-----------------------
	PropertyTab(PropertyFrame* parent = 0,
				gui::MessageHandler* commandTarget = 0,
				uint commandID = 0,
				const std::string& text = "Tab");

	void  AttachToPropertyFrame(PropertyFrame& parent) { parent.Attach(*this); }

	void                SetText(const std::string& text);
	const std::string&  GetText() const { return text; }

	void  SetSelected(bool bSelected) { SetFlagsOnOff(Flag_Selected, bSelected); }
	bool  IsSelected() const { return ToBool(IsFlagOn(Flag_Selected)); }

	void  Select();
	void  UpdateControlDependencies();
	void  ApplyControlValues();

	//-----------------------------------------------
	// Additional methods intended for use by shader.
	//-----------------------------------------------
	float        GetTextHeight() const { return (pFrame ? pFrame->GetGeomParams().tabTextHeight : 1); }
	const vec2&  GetTextSize() const { return textSize; }


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  OnAttachedToParent();
	virtual  void  OnDetachedFromParent();
	virtual  void  OnAttachedChild(gui::Component& child);
	virtual  void  OnDetachedChild(gui::Component& child);
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();
	virtual  bool  OnCommand(uint commandID);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  bool  OnMsgSelect(bool bSelect, uint groupID);


protected:
	//------------
	// Properties.
	//------------
	std::string  text;

	//-------
	// Cache.
	//-------
	PropertyFrame*  pFrame;
	PropertyPage*   pPage;
	vec2            textSize;
	vec2            unselectedSize;
};


//====================================================================
class VW_GUI_CS0_API PropertyPage : public gui::Component
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);

friend class PropertyTab;

public:
	//------------------------
	// PropertyPage interface.
	//------------------------
	PropertyPage(PropertyTab* parent = 0);

	void  AttachToPropertyTab(PropertyTab& parent) { parent.Attach(*this); }

	//-----------------------------------------------
	// Additional methods intended for use by shader.
	//-----------------------------------------------
	void  GetTabPosition(float& tabOffset, float& tabLength) const;


protected:
	// methods intended for use by PropertyTab
	void  SetTabPosition(float tabOffset, float tabLength);

	//------------------------------------------
	// Generic way to update the control states.
	//------------------------------------------
	virtual  void  UpdateControlDependencies() {}
	virtual  void  ApplyControlValues() {}


private:
	//-------
	// Cache.
	//-------
	float  tabOffset;
	float  tabLength;
};


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


