

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_BehaviorBases.h"
#include "BaseTools\ScrollGadget.h"
#include "private\vw_gui_cs0_exports.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
// Internal hierarchy:
// * ScrollBar
//   * ScrollBarThumb
//	 * ScrollBarButton (Forward)
//	 * ScrollBarButton (Backward)
//====================================================================
class ScrollBarThumb;
class ScrollBarButton;
class ScrollBar;


//====================================================================
// Feature support:
// CanDisable.
//====================================================================
class VW_GUI_CS0_API ScrollBarThumb : public gui::Thumb
{
friend class ScrollBar;

protected:
	ScrollBarThumb(gui_cs0::ScrollBar& sbar, bool bReportVerticalOffset);

	//----------------------
	// Thumb implementation.
	//----------------------
	virtual  void  ProcessThumbGrabBegin();
	virtual  void  ProcessThumbGrabOffset(float offset);
	virtual  void  ProcessThumbGrabEnd() {}


protected:
	//------------
	// Properties.
	//------------
	ScrollBar&  sbar;
};


//====================================================================
// Feature support:
// CanDisable.
//====================================================================
class VW_GUI_CS0_API ScrollBarButton : public gui::PressButton
{
friend class ScrollBar;

public:
	enum eButtonID
	{
		ButtonID_Up = 0,
		ButtonID_Down,
		ButtonID_Left,
		ButtonID_Right,

		ButtonID_Count,
		ButtonID_Force32Bits = 0x7FFFFFFF
	};

	//---------------------------
	// ScrollBarButton interface.
	//---------------------------
	eButtonID  GetButtonID() const { return buttonID; }
	bool       IsButtonForward() const { return (buttonID == ButtonID_Down  ||  buttonID == ButtonID_Right); }


protected:
	ScrollBarButton(ScrollBar& sbar, eButtonID id);

	//----------------------------
	// PressButton implementation.
	//----------------------------
	virtual  void  ProcessButtonPress();


protected:
	//------------
	// Properties.
	//------------
	ScrollBar&  sbar;
	eButtonID   buttonID;
};


//====================================================================
// Feature support:
// CanDisable. CanFocus.
//====================================================================
class VW_GUI_CS0_API ScrollBar : public gui::Component,
									public gui::CommandSender,
									private ScrollGadget
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);

friend class ScrollBarThumb;
friend class ScrollBarButton;

public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		vec2   preferredSize;
		float  buttonLength;
		float  minThumbLength;

		GeomParams(bool bVertical = true);
	};


public:
	//---------------------
	// ScrollBar interface.
	//---------------------
	ScrollBar(gui::Component* parent = 0,
				gui::MessageHandler* commandTarget = 0,
				uint commandID = 0,
				bool bVertical = true);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	void  EnableMaxPixelsPerStep(bool bEnable = true);
	bool  IsEnabledMaxPixelsPerStep() const { return ScrollGadget::IsEnabledMaxPixelsPerStep(); }

	void  ActivateAutoEnableDisable(bool bActivate = true);
	bool  IsActiveAutoEnableDisable() const { return bActiveAutoEnableDisable; }

	void  SetDeltaPerMouseWheelUnit(int delta = 3);
	int   GetDeltaPerMouseWheelUnit() const { return deltaPerWheelUnit; }

	void         SetPosition(float x, float y) { Component::SetPosition(x, y); }
	void         SetPosition(const vec2& pos) { Component::SetPosition(pos); }
	const vec2&  GetPosition() const { return Component::GetPosition(); }

	//---------------------------
	// Logical scroll parameters.
	//---------------------------
	void  ResetLogicalScrollParameters();

	uint  GetTotalScrollSize() const { return ScrollGadget::GetTotalSize(); }
	void  SetTotalScrollSize(uint total = 1);

	uint  GetVisibleScrollSize() const { return ScrollGadget::GetVisibleSize(); }
	void  SetVisibleScrollSize(uint visible = 1);

	bool  IsScrollSizeVisibleLessThanTotal() const { return ScrollGadget::IsVisibleLessThanTotal(); }

	uint  GetScrollPosition() const { return ScrollGadget::GetPosition(); }
	uint  GetMaxScrollPosition() const { return ScrollGadget::GetMaxPosition(); }
	uint  GetLastVisibleScrollPosition() const { return ScrollGadget::GetLastVisiblePosition(); }
	bool  IsScrollPositionAtMaximum() const { return ScrollGadget::IsPositionAtMaximum(); }

	void  SetScrollPosition(int pos = 0);
	void  SetScrollPositionToMax();
	void  OffsetScrollPosition(int offset);

	//-----------------------------------------------
	// Additional methods intended for use by shader.
	//-----------------------------------------------
	bool                    IsVertical() const { return bVertical; }
	const ScrollBarButton&  GetButtonBackward() const { return *pButtonBackward; }
	const ScrollBarButton&  GetButtonForward() const { return *pButtonForward; }
	const ScrollBarThumb&   GetThumb() const { return *pThumb; }

	void  GetDrawRectangles(gui::Rectangle& rBase,
							gui::Rectangle& rThumb,
							gui::Rectangle& rButtonBackward,
							gui::Rectangle& rButtonForward) const;


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcSize();
	virtual  void  RecalcLayout();
	virtual  void  UpdateGeometryCache();
	virtual  bool  OnMouseWheel(int delta, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  bool  OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData);
	virtual  bool  OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch);
	virtual  void  OnChangedEnabled();
	virtual  bool  OnMsgScroll(bool bForward, gui::eScrollDistance distance, int offset);

	//-------------------------
	// Sub-component interface.
	//-------------------------
	void  ReportThumbGrabBegin();
	void  ReportThumbGrabOffset(float offset);

	//------------------
	// Internal helpers.
	//------------------
	uint  GetThumbLength() const { return ScrollGadget::GetThumbLength(); }
	uint  GetThumbOffset() const { return ScrollGadget::GetThumbOffset(); }
	void  SetBarLength(float barLength) { ScrollGadget::SetBarLength(ClampedLow<float>(1, barLength)); }

	void  DoEnable(bool bEnable);
	void  UpdateBar();
	void  UpdateThumbGeometry();


protected:
	//----------------
	// Sub-components.
	//----------------
	ScrollBarButton*  pButtonForward;
	ScrollBarButton*  pButtonBackward;
	ScrollBarThumb*   pThumb;

	//------------
	// Properties.
	//------------
	bool        bVertical;
	GeomParams  geomParams;

	//--------
	// States.
	//--------
	bool   bActiveAutoEnableDisable;
	int    deltaPerWheelUnit;
	int    thumbPixelOffsetAtGrab;
};


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


