

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "Framework\HIDController.h"


//********************************************************************
namespace vertexware {


//====================================================================
enum eGloveSensor
{
	GloveSensor_Invalid = -1,

	// wrist
	GloveSensor_WristYaw = 0,
	GloveSensor_WristPitch,
	GloveSensor_WristRoll,

	// abduction (angle between fingers)
	GloveSensor_AbductionThumbIndex,
	GloveSensor_AbductionIndexMiddle,
	GloveSensor_AbductionMiddleRing,
	GloveSensor_AbductionRingPinky,

	// thumb
	GloveSensor_ThumbRoll,
	GloveSensor_ThumbJointInner,
	GloveSensor_ThumbJointOuter,

	// index-finger
	GloveSensor_IndexJointInner,
	GloveSensor_IndexJointMiddle,
	GloveSensor_IndexJointOuter,

	// middle-finger
	GloveSensor_MiddleJointInner,
	GloveSensor_MiddleJointMiddle,
	GloveSensor_MiddleJointOuter,

	// ring-finger
	GloveSensor_RingJointInner,
	GloveSensor_RingJointMiddle,
	GloveSensor_RingJointOuter,

	// pinky-finger
	GloveSensor_PinkyJointInner,
	GloveSensor_PinkyJointMiddle,
	GloveSensor_PinkyJointOuter,

	GloveSensor_Count,
	GloveSensor_Force32Bits = 0x7FFFFFFF
};


//====================================================================
// Sensors-readings unsupported by glove must be filled with
// sensible values. Gloves can be roughly categorized in 2 classes:
// a) gesture-gloves b) precision-gloves
// Gesture-class gloves are not suitable for precise measurement,
// but they provide enough information for gesture recognition.
// Precision-class gloves support precise angle-measurement.
// Valid range for all angles is [-180, 180]. Certainly an overkill,
// even physically impossible in some cases.
// An alternative would be to define specific ranges for each sensor
// which would unnecessarily complicate things, so lets keep it simple.
// Measurements are reported in angles (specified in degrees),
// raw-values and flexures. Flexures are intended for use by
// gesture-recognition algorithms. They represent full sensor range
// normalized into [0, 1] (minimum maps to 0, maximum maps to 1).
// If precise angle measurement is not supported, driver must
// report some sensible angle-values based on available information.
// When hand is in "flat" position, all angles are 0.
// Sensor-specific notes:
// Wrist:
//    Orientation angles are specified in standard RHS.
// Finger-joints:
//    Positive angle indicate that finger is bent towards the palm.
//    Note that negative angle is physically possible.
// Abduction:
//    Angle 0 indicates that fingers are completely closed together.
//    Positive angle indicates that fingers are apart.
//    Negative angle is physically impossible and should not be used.
//====================================================================
class VW_COMMON_API Glove : public HIDController
{
public:
	static std::string __fastcall GetSensorName(eGloveSensor sensor);


public:
	virtual ~Glove();

	//------------------------------
	// HIDController implementation.
	//------------------------------
	virtual  std::string  GetControllerInterfaceName() const;
	virtual  void         DumpControllerInfo(std::ostream& os, uint tabLevel) const;
	virtual  void         ResetInputReadings();
	virtual  void         PreSampleUpdate();
	virtual  void         PostSampleUpdate();
	virtual  void         OnBeginAsyncSampleRead();
	virtual  void         OnEndAsyncSampleRead();


	//-----------------
	// Glove interface.
	//-----------------

	// hardware capabilities
	bool   IsLeftHanded() const { return bLeftHanded; }
	bool   IsSupportedAngleMeasurement() const { return bSupportsAngleMeasurement; }
	bool   IsSupportedSensor(eGloveSensor sensor) const { return bSupportsSensor[sensor]; }

	// flexure calibration
	void   ResetFlexureAutoCalibration();
	void   ActivateFlexureAutoCalibration(bool bActivate, bool bResetOnActivate = true);
	bool   IsActiveFlexureAutoCalibration() const { return bFlexureAutoCalibration; }

	void   SetMinFlexure(eGloveSensor sensor, float minFlexure = 0.0f);
	float  GetMinFlexure(eGloveSensor sensor) const { return minFlexures[sensor]; }

	void   SetMaxFlexure(eGloveSensor sensor, float maxFlexure = 1.0f);
	float  GetMaxFlexure(eGloveSensor sensor) const { return maxFlexures[sensor]; }

	// flexure normalization
	void   SetDefaultEqualValueFlexureNormalizations();

	void   SetEqualValueFlexureNormalization(eGloveSensor sensor, float norm);
	float  GetEqualValueFlexureNormalization(eGloveSensor sensor) const { return equalValueFlexureNorm[sensor]; }

	// setup modifier-maps used by gesture-gloves to derive angle-values
	void   SetDefaultFlexureToAngleMaps();

	void   SetFlexureToAngleScale(eGloveSensor sensor, float scaleDeg);
	float  GetFlexureToAngleScale(eGloveSensor sensor) const { return flexureToAngleScale[sensor]; }

	void   SetFlexureToAngleBias(eGloveSensor sensor, float biasDeg);
	float  GetFlexureToAngleBias(eGloveSensor sensor) const { return flexureToAngleBias[sensor]; }

	// sampling values
	float  GetAngle(eGloveSensor sensor) const { return angles[sensor]; }
	float  GetNormalizedFlexure(eGloveSensor sensor) const { return normalizedFlexures[sensor]; }
	float  GetFlexure(eGloveSensor sensor) const { return flexures[sensor]; }
	float  GetRawValue(eGloveSensor sensor) const { return rawValues[sensor]; }


private:
	// hardware capabilities
	bool  bLeftHanded;
	bool  bSupportsAngleMeasurement;
	bool  bSupportsSensor[GloveSensor_Count];

	// operational states
	bool  bFlexureAutoCalibration;

	// helper maps (used by gesture-gloves to derive angle-values)
	float  flexureToAngleScale[GloveSensor_Count];
	float  flexureToAngleBias[GloveSensor_Count];

	// sensor readings
	float  angles[GloveSensor_Count];
	float  normalizedFlexures[GloveSensor_Count];
	float  flexures[GloveSensor_Count];
	float  rawValues[GloveSensor_Count];

	// flexure normalizers
	float  equalValueFlexureNorm[GloveSensor_Count];
	float  minFlexures[GloveSensor_Count];
	float  maxFlexures[GloveSensor_Count];
	float  oneOverRangeFlexures[GloveSensor_Count];


private:
	void  CalculateNormalizedFlexures();
	void  UpdateFlexureNormalizers();
	void  UpdateFlexureNormalizer(eGloveSensor sensor);


protected:
	//-------------------------
	// Derived-class interface.
	//-------------------------
	Glove();

	void  ReportHandedness(bool bLeftHanded);

	void  ReportHardwareCaps_Measurement(bool bSupportsAngleMeasurement);
	void  ReportHardwareCaps_Sensor(eGloveSensor sensor, bool bSupported);


	// angle in degrees
	__forceinline  void  SetAngle(eGloveSensor sensor, float angleDeg) {
		angles[sensor] = angleDeg;
	}

	// uses flexure-to-angle maps to derive angle-values
	void  DeriveAnglesFromFlexures();

	// map full sensor range to [0, 1]
	__forceinline  void  SetFlexure(eGloveSensor sensor, float flexure) {
		flexures[sensor] = flexure;
	}

	// value reported by native-driver (set to 0 if sensor is unavailable)
	__forceinline  void  SetRawValue(eGloveSensor sensor, float rawValue) {
		rawValues[sensor] = rawValue;
	}
};


//====================================================================
#define  VW_INTERFACE_NAME_Glove  "Glove"


//********************************************************************
}//namespace vertexware


