

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "HIDControllers\Glove.h"
#include <vector>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace static_gestures {


//====================================================================
// Threshold range is [0, 1]
// Threshold-unflexed <= threshold-flexed
//====================================================================
class VW_COMMON_API Thresholds
{
public:
	Thresholds();

	float  GetUnflexed(eGloveSensor sensor) const { return unflexed[sensor]; }
	float  GetFlexed(eGloveSensor sensor) const { return flexed[sensor]; }
	float  GetCenter(eGloveSensor sensor) const { return center[sensor]; }

	void   SetValues(eGloveSensor sensor, float threshUnflexed, float threshFlexed, float threshCenter);
	void   ResetDefaultValues();


private:
	float  unflexed[GloveSensor_Count];
	float  flexed[GloveSensor_Count];
	float  center[GloveSensor_Count];
};


//====================================================================
// Conditions are defined in terms of normalized-flexure NF:
//    Unflexed - NF in range [0, threshold-unflexed]
//    Normal   - NF in range <threshold-unflexed, threshold-flexed>
//    Flexed   - NF in range [threshold-flexed, 1]
//    Low      - NF in range [0, threshold-center>
//    High     - NF in range [threshold-center, 1]
//====================================================================
enum eFlexure
{
	// 3-area division
	Flexure_Unflexed = 0,
	Flexure_Neutral,
	Flexure_Flexed,

	// 2-area division
	Flexure_Low,
	Flexure_High,

	Flexure_Count,
	Flexure_Force32Bits = 0x7FFFFFFF
};


//====================================================================
class VW_COMMON_API Condition
{
public:
	Condition(eGloveSensor _sensor = eGloveSensor(0), eFlexure _flexure = Flexure_Neutral) {
		sensor  = _sensor;
		flexure = _flexure;
	}


public:
	eGloveSensor  sensor;
	eFlexure      flexure;
};


//====================================================================
class VW_COMMON_API Gesture
{
public:
	Gesture(uint _id = 0, const std::string& _gestureName = "Unnamed");

	void                SetName(const std::string& s) { gestureName = s; }
	const std::string&  GetName() const { return gestureName; }

	uint  GetID() const { return id; }
	void  SetID(uint _id) { id = _id; }

	void  ClearConditions() { conditions.clear(); }
	void  RemoveCondition(eGloveSensor sensor);
	void  AddCondition(const Condition& cond);
	void  AddCondition(eGloveSensor sensor, eFlexure flexure) { AddCondition(Condition(sensor, flexure)); }

	uint              GetNumConditions() const { return conditions.size(); }
	const Condition&  GetCondition(uint i) const { return conditions[i]; }


private:
	typedef  std::vector<Condition>          VEC_CONDITIONS;
	typedef  VEC_CONDITIONS::iterator        ITER_CONDITIONS;
	typedef  VEC_CONDITIONS::const_iterator  CITER_CONDITIONS;

	VEC_CONDITIONS  conditions;
	uint            id;
	std::string     gestureName;
};


//====================================================================
class VW_COMMON_API GestureAnalyzer
{
public:
	GestureAnalyzer();

	// sets active thresholds, object is copied into anaylzer
	void  SetThresholds(const Thresholds& t);

	// standard gestures
	void  RegisterDefaultGestures();

	// Gesture-object is copied into analyzer. If an object with equal
	// gesture-id is already registered, it is replaced with new object.
	void  Register(const Gesture& g);

	// unregister matching gestures
	void  Unregister(uint gestureID);
	void  Unregister(const std::string& gestureName);
	void  UnregisterAll();

	// returns 0 if not registered
	const Gesture*  FindRegisteredGesture(uint gestureID) const;
	const Gesture*  FindRegisteredGesture(const std::string& gestureName) const;

	// returns true if current-glove-state satisfies gesture conditions
	bool  IsGestureActive(Glove& glove, const Gesture& gesture) const;
	bool  IsGestureActive(Glove& glove, uint gestureID) const;
	bool  IsGestureActive(Glove& glove, const std::string& gestureName) const;

	// returns 0 if current-glove-state does not match any of
	// the remaining gestures, otherwise it returns next matching gesture
	const Gesture*  GetFirstActiveGesture(Glove& glove) const;
	const Gesture*  GetNextActiveGesture(Glove& glove) const;


private:
	typedef  std::vector<Gesture>         VEC_GESTURE;
	typedef  VEC_GESTURE::iterator        ITER_GESTURE;
	typedef  VEC_GESTURE::const_iterator  CITER_GESTURE;

	VEC_GESTURE  gestures;
	Thresholds   thresholds;

	mutable  uint  nextTestIndex;
};


//********************************************************************
namespace standard_gestures {


//====================================================================
// Finger gestures use only inner-finger-joints.
//====================================================================
enum eID
{
	// 5-finger
	ID_Grab = 0,
	ID_Stop,
	ID_ThumbUp,

	// 4-finger
	ID_Fist,
	ID_FlatHand,
	ID_IndexPoint,
	ID_NotIndex,
	ID_DoublePoint,
	ID_TriplePoint,
	ID_PinkyPoint,
	ID_UpYours,
	ID_NotUpYours,
	ID_Howzit,
	ID_NotHowzit,
	ID_NotRing,

	// wrist
	ID_RollNeutral,
	ID_RollLeft,
	ID_RollRight,
	ID_PitchNeutral,
	ID_PitchUp,
	ID_PitchDown,

	ID_Count,
	ID_Force32Bits = 0x7FFFFFFF
};


//====================================================================
// 5-finger gesture
// Flexed..: all
// Neutral.: none
// Unflexed: none
//====================================================================
struct VW_COMMON_API Grab : public Gesture {
	Grab(const std::string& name = "Grab");
};


//====================================================================
// 5-finger gesture
// Flexed..: none
// Neutral.: none
// Unflexed: all
//====================================================================
struct VW_COMMON_API Stop : public Gesture {
	Stop(const std::string& name = "Stop");
};


//====================================================================
// 5-finger gesture
// Flexed..: index, middle, ring, pinky
// Neutral.: none
// Unflexed: thumb
//====================================================================
struct VW_COMMON_API ThumbUp : public Gesture {
	ThumbUp(const std::string& name = "ThumbUp");
};


//====================================================================
// 4-finger gesture
// Flexed..: all
// Neutral.: none
// Unflexed: none
//====================================================================
struct VW_COMMON_API Fist : public Gesture {
	Fist(const std::string& name = "Fist");
};


//====================================================================
// 4-finger gesture
// Flexed..: none
// Neutral.: none
// Unflexed: all
//====================================================================
struct VW_COMMON_API FlatHand : public Gesture {
	FlatHand(const std::string& name = "FlatHand");
};


//====================================================================
// 4-finger gesture
// Flexed..: middle, ring, pinky
// Neutral.: none
// Unflexed: index
//====================================================================
struct VW_COMMON_API IndexPoint : public Gesture {
	IndexPoint(const std::string& name = "IndexPoint");
};


//====================================================================
// 4-finger gesture
// Flexed..: index
// Neutral.: none
// Unflexed: middle, ring, pinky
//====================================================================
struct VW_COMMON_API NotIndex : public Gesture {
	NotIndex(const std::string& name = "NotIndex");
};


//====================================================================
// 4-finger gesture
// Flexed..: ring, pinky
// Neutral.: none
// Unflexed: index, middle
//====================================================================
struct VW_COMMON_API DoublePoint : public Gesture {
	DoublePoint(const std::string& name = "DoublePoint");
};


//====================================================================
// 4-finger gesture
// Flexed..: pinky
// Neutral.: none
// Unflexed: index, middle, ring
//====================================================================
struct VW_COMMON_API TriplePoint : public Gesture {
	TriplePoint(const std::string& name = "TriplePoint");
};


//====================================================================
// 4-finger gesture
// Flexed..: index, middle, ring
// Neutral.: none
// Unflexed: pinky
//====================================================================
struct VW_COMMON_API PinkyPoint : public Gesture {
	PinkyPoint(const std::string& name = "PinkyPoint");
};


//====================================================================
// 4-finger gesture
// Flexed..: index, ring, pinky
// Neutral.: none
// Unflexed: middle
//====================================================================
struct VW_COMMON_API UpYours : public Gesture {
	UpYours(const std::string& name = "UpYours");
};


//====================================================================
// 4-finger gesture
// Flexed..: middle
// Neutral.: none
// Unflexed: index, ring, pinky
//====================================================================
struct VW_COMMON_API NotUpYours : public Gesture {
	NotUpYours(const std::string& name = "NotUpYours");
};


//====================================================================
// 4-finger gesture
// Flexed..: middle, ring
// Neutral.: none
// Unflexed: index, pinky
//====================================================================
struct VW_COMMON_API Howzit : public Gesture {
	Howzit(const std::string& name = "Howzit");
};


//====================================================================
// 4-finger gesture
// Flexed..: index, pinky
// Neutral.: none
// Unflexed: middle, ring
//====================================================================
struct VW_COMMON_API NotHowzit : public Gesture {
	NotHowzit(const std::string& name = "NotHowzit");
};


//====================================================================
// 4-finger gesture
// Flexed..: ring
// Neutral.: none
// Unflexed: index, middle, pinky
//====================================================================
struct VW_COMMON_API NotRing : public Gesture {
	NotRing(const std::string& name = "NotRing");
};


//====================================================================
// wrist gesture
// Flexed..: none
// Neutral.: wrist-roll
// Unflexed: none
//====================================================================
struct VW_COMMON_API RollNeutral : public Gesture {
	RollNeutral(const std::string& name = "RollNeutral");
};


//====================================================================
// wrist gesture
// Flexed..: none
// Neutral.: none
// Unflexed: wrist-roll
//====================================================================
struct VW_COMMON_API RollLeft : public Gesture {
	RollLeft(const std::string& name = "RollLeft");
};


//====================================================================
// wrist gesture
// Flexed..: wrist-roll
// Neutral.: none
// Unflexed: none
//====================================================================
struct VW_COMMON_API RollRight : public Gesture {
	RollRight(const std::string& name = "RollRight");
};


//====================================================================
// wrist gesture
// Flexed..: none
// Neutral.: wrist-pitch
// Unflexed: none
//====================================================================
struct VW_COMMON_API PitchNeutral : public Gesture {
	PitchNeutral(const std::string& name = "PitchNeutral");
};


//====================================================================
// wrist gesture
// Flexed..: none
// Neutral.: none
// Unflexed: wrist-pitch
//====================================================================
struct VW_COMMON_API PitchUp : public Gesture {
	PitchUp(const std::string& name = "PitchUp");
};


//====================================================================
// wrist gesture
// Flexed..: wrist-pitch
// Neutral.: none
// Unflexed: none
//====================================================================
struct VW_COMMON_API PitchDown : public Gesture {
	PitchDown(const std::string& name = "PitchDown");
};


//********************************************************************
}//namespace standard_gestures


//********************************************************************
}//namespace static_gestures


//********************************************************************
}//namespace vertexware


