

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "Framework\HIDController.h"
#include "HIDControllers\KeyboardCodes.h"


//********************************************************************
namespace vertexware {


//====================================================================
class VW_COMMON_API Keyboard : public HIDController
{
public:
	virtual ~Keyboard();

	//------------------------------
	// HIDController implementation.
	//------------------------------
	virtual  std::string  GetControllerInterfaceName() const;
	virtual  void         DumpControllerInfo(std::ostream& os, uint tabLevel) const;
	virtual  void         ResetInputReadings();
	virtual  void         PreSampleUpdate();
	virtual  void         PostSampleUpdate();
	virtual  void         OnBeginAsyncSampleRead();
	virtual  void         OnEndAsyncSampleRead();


	//-------------------------------------------------------
	// Keyboard interface.
	// It is safe to use Key_Count & KeyModifier_Count values
	// as keys, their states are always 0.
	//-------------------------------------------------------

	// hardware capabilities
	bool  IsSystemKeyboard() const { return bSystemKeyboard; }

	// tests key-modifier state
	__forceinline  ubyte  KeyModifierPressed(eKeyModifier keyModifier) const {
		return IsPressed(keyModifierStates[keyModifier]);
	}
	bool  KeyModifierPressed() const { return (numKeyModifiersPressed > 0); }
	uint  NumKeyModifiersPressed() const { return numKeyModifiersPressed; }

	// tests current key-state
	__forceinline  ubyte  KeyPressed(eKey keyCode) const {
		return IsPressed(keyStates[keyCode]);
	}
	bool  KeyPressed() const { return (numKeysPressed > 0); }
	uint  NumKeysPressed() const { return numKeysPressed; }

	// tests key-state transitions
	__forceinline  ubyte  KeyDown(eKey keyCode) const {
		ubyte state = keyStates[keyCode];
		return IsPressed(state)  &&  ! WasPressed(state);
	}
	bool  KeyDown() const { return (numKeysDown > 0); }
	uint  NumKeysDown() const { return numKeysDown; }

	__forceinline  ubyte  KeyUp(eKey keyCode) const {
		ubyte state = keyStates[keyCode];
		return WasPressed(state)  &&  ! IsPressed(state);
	}
	bool  KeyUp() const { return (numKeysUp > 0); }
	uint  NumKeysUp() const { return numKeysUp; }

	// Enable if key-code to mapping calls will be used.
	// Disable for faster performance.
	void  EnableCharacterMapping(bool bEnable);
	bool  IsEnabledCharacterMapping() const { return bEnableCharacterMapping; }

	// Translate key-code into displayable ascii character [32-126],
	// based on the current keyboard state (shift, control, alt keys).
	// Returns false if no translation is available
	virtual  bool  MapKeyCodeToCharacter(eKey keyCode, char& ch) const = 0;


private:
	//------------------
	// Internal helpers.
	//------------------
	enum {
		CurrentStateBit = 0x80,
		CurrentStateInv = 0x7F,
		LastStateBit    = 0x01,
		StateBitShift   = 0x07,
		Force8Bits      = 0x7F //not used
	};

	__forceinline static ubyte  IsPressed(ubyte keyState) {
		return keyState & CurrentStateBit;
	}
	__forceinline static ubyte  WasPressed(ubyte keyState) {
		return keyState & LastStateBit;
	}


private:
	enum { KeyStateStorageSize = Key_Count + 1 };
	enum { KeyModifierStateStorageSize = KeyModifier_Count + 1 };

	// hardware capabilities
	bool  bSystemKeyboard;

	// operational states
	bool  bEnableCharacterMapping;

	// input readings
	ubyte  keyStates[KeyStateStorageSize];
	ubyte  keyModifierStates[KeyModifierStateStorageSize];

	uint   numKeyModifiersPressed;
	uint   numKeysPressed;
	uint   numKeysDown;
	uint   numKeysUp;


protected:
	//-------------------------
	// Derived-class interface.
	//-------------------------
	Keyboard(bool bSystemKeyboard = false);

	__forceinline  void  SetKeyPressed(eKey keyCode) {
		keyStates[keyCode] |= CurrentStateBit;
	}
	__forceinline  void  ResetKeyPressed(eKey keyCode) {
		keyStates[keyCode] &= CurrentStateInv;
	}

	void  ResetKeyStates();
	void  ClearCurrentKeyStates();
	void  ShiftKeyStates();
	void  ShiftCopyKeyStates();
	void  ResetKeyModifierCache();
	void  UpdateKeyModifierCache();
	void  ResetNumKeysCache();
	void  UpdateNumKeysCache();

	// called after character mapping is enabled/disabled
	virtual  void  OnChangedProcessingMode() = 0;
};


//====================================================================
#define  VW_INTERFACE_NAME_Keyboard  "Keyboard"


//********************************************************************
}//namespace vertexware


