

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "Framework\HIDController.h"


//********************************************************************
namespace vertexware {


//====================================================================
// WARNING: Codes must start with 0. Code-holes are NOT allowed.
//====================================================================
enum eMouseButton
{
	MouseButton_Left = 0,
	MouseButton_Right,
	MouseButton_Middle,
	MouseButton_X1,
	MouseButton_X2,

	MouseButton_Count,
	MouseButton_Force32Bits = 0x7FFFFFFF
};


//====================================================================
// X-Axis: move-left = negative, move-right = positive.
// Y-Axis: move-up = negative, move-down = positive.
// Wheel-Axis: roll-to-user = negative, roll-from-user = positive.
//====================================================================
class VW_COMMON_API Mouse : public HIDController
{
public:
	virtual ~Mouse();

	//------------------------------
	// HIDController implementation.
	//------------------------------
	virtual  std::string  GetControllerInterfaceName() const;
	virtual  void         DumpControllerInfo(std::ostream& os, uint tabLevel) const;
	virtual  void         ResetInputReadings();
	virtual  void         PreSampleUpdate();
	virtual  void         PostSampleUpdate();
	virtual  void         OnBeginAsyncSampleRead();
	virtual  void         OnEndAsyncSampleRead();


	//------------------------------------------
	// Mouse interface.
	// It is safe to use MouseButton_Count value
	// as button, it's state is always 0.
	//------------------------------------------

	// hardware capabilities
	bool  IsSystemMouse() const { return bSystemMouse; }
	bool  CanDisplaySystemCursor() const { return bCanDisplaySystemCursor; }
	bool  CanSyncWithSystemCursor() const { return bCanSyncWithSystemCursor; }

	// operation modifiers
	bool  IsReversedDeltaX() const { return (iRMReverseX < 0); }
	void  SetReverseDeltaX(bool bReverse);

	bool  IsReversedDeltaY() const { return (iRMReverseY < 0); }
	void  SetReverseDeltaY(bool bReverse);

	bool  IsReversedDeltaWheel() const { return (iRMReverseWheel < 0); }
	void  SetReverseDeltaWheel(bool bReverse);

	// sensitivity can be negative
	void  SetRelativeModeSensitivity(float x = 1.0f, float y = 1.0f);
	void  GetRelativeModeSensitivity(float& x, float& y) const;

	// acceleration & speed cannot be negative, only used if sync-with-system-cursor is not active
	void   SetAbsoluteModeSpeed(float speed = 1.0f);
	float  GetAbsoluteModeSpeed() const { return fAMSpeed; }

	void  SetAbsoluteModeAccelerationLevel(uint level = 0); //[0-2]
	uint  GetAbsoluteModeAccelerationLevel() const { return iAMAccelerationLevel; }

	void  SetAbsoluteModeAccelerationThresholds(float t1 = 64.0f, float t2 = 256.0f);
	void  GetAbsoluteModeAccelerationThresholds(float& t1, float& t2) const;

	// operational states
	bool  IsActiveSyncWithSystemCursor() const { return bActiveSyncWithSystemCursor; }
	void  ActivateSyncWithSystemCursor(bool bActivate);

	bool  IsActiveRelativeMode() const { return bActiveRelativeMode; }
	void  ActivateRelativeMode(bool bActivate);

	// absolute position data (valid only if relative-mode is NOT active)
	__forceinline  bool  IsPositionInsideAppWindowClientArea() const { return bPositionInsideClientArea; }
	__forceinline  int   GetPositionX() const { return xPosition; }
	__forceinline  int   GetPositionY() const { return yPosition; }

	// relative movement data (valid only if relative-mode is active)
	__forceinline  int  GetDeltaX() const { return xDelta; }
	__forceinline  int  GetDeltaY() const { return yDelta; }

	// wheel data
	__forceinline  int  GetDeltaWheel() const { return wheelDelta; }

	// tests current button-state
	__forceinline  ubyte  ButtonPressed(eMouseButton buttonCode) const {
		return IsPressed(buttonStates[buttonCode]);
	}
	bool  ButtonPressed() const { return (numButtonsPressed > 0); }
	uint  NumButtonsPressed() const { return numButtonsPressed; }

	// tests button-state transitions
	__forceinline  ubyte  ButtonDown(eMouseButton buttonCode) const {
		ubyte state = buttonStates[buttonCode];
		return IsPressed(state)  &&  ! WasPressed(state);
	}
	bool  ButtonDown() const { return (numButtonsDown > 0); }
	uint  NumButtonsDown() const { return numButtonsDown; }

	__forceinline  ubyte  ButtonUp(eMouseButton buttonCode) const {
		ubyte state = buttonStates[buttonCode];
		return WasPressed(state)  &&  ! IsPressed(state);
	}
	bool  ButtonUp() const { return (numButtonsUp > 0); }
	uint  NumButtonsUp() const { return numButtonsUp; }


private:
	//------------------
	// Internal helpers.
	//------------------
	enum {
		CurrentStateBit = 0x80,
		CurrentStateInv = 0x7F,
		LastStateBit    = 0x01,
		StateBitShift   = 0x07,
		Force8Bits      = 0x7F //not used
	};

	__forceinline static ubyte  IsPressed(ubyte keyState) {
		return keyState & CurrentStateBit;
	}
	__forceinline static ubyte  WasPressed(ubyte keyState) {
		return keyState & LastStateBit;
	}


private:
	enum { ButtonStateStorageSize = MouseButton_Count + 1 };

	// hardware capabilities
	bool   bSystemMouse;
	bool   bCanDisplaySystemCursor;
	bool   bCanSyncWithSystemCursor;

	// operational states
	bool   bActiveSyncWithSystemCursor;
	bool   bActiveRelativeMode;

	// operation modifiers
	float  fRMSensitivityX;
	float  fRMSensitivityY;

	float  fAMSpeed;
	float  fAMAccelerationThresholds[2];
	uint   iAMAccelerationLevel;

	int    iRMReverseX;
	int    iRMReverseY;
	int    iRMReverseWheel;

	// input readings
	bool   bPositionInsideClientArea;
	int    xPosition;
	int    yPosition;
	int    xDelta;
	int    yDelta;
	int    wheelDelta;
	ubyte  buttonStates[ButtonStateStorageSize];

	uint   numButtonsPressed;
	uint   numButtonsDown;
	uint   numButtonsUp;


protected:
	//-------------------------
	// Derived-class interface.
	//-------------------------
	Mouse(bool bSystemMouse = false);

	// system-mices only
	__forceinline  void  SetCanDisplaySystemCursor(bool bCanDisplay) {
		bCanDisplaySystemCursor = bCanDisplay;
	}

	__forceinline  void  SetCanSyncWithSystemCursor(bool bCanSync) {
		bCanSyncWithSystemCursor = bCanSync;
	}

	// apply modifiers to raw-values
	void  ApplyAbsoluteModeModifiers(float rawX, float rawY, float& finalX, float& finalY);
	void  ApplyRelativeModeModifiers(float rawX, float rawY, float& finalX, float& finalY);

	// report position in client-area coordinates (absolute mode only)
	__forceinline  void  SetPosition(int x, int y, bool bInsideAppWindowClientArea) {
		bPositionInsideClientArea = bInsideAppWindowClientArea;
		xPosition = x;
		yPosition = y;
	}

	// report movement (relative mode only)
	__forceinline  void  ResetDelta(int x = 0, int y = 0) {
		xDelta = x * iRMReverseX;
		yDelta = y * iRMReverseY;
	}
	__forceinline  void  AccumDelta(int x, int y) {
		xDelta += x * iRMReverseX;
		yDelta += y * iRMReverseY;
	}

	// report wheel movement
	__forceinline  void  ResetWheelDelta(int delta = 0) {
		wheelDelta = delta * iRMReverseWheel;
	}
	__forceinline  void  AccumWheelDelta(int delta) {
		wheelDelta += delta * iRMReverseWheel;
	}

	// report button states
	__forceinline  void  SetButtonPressed(eMouseButton buttonCode) {
		buttonStates[buttonCode] |= CurrentStateBit;
	}

	void  ResetAxisStates();
	void  ResetButtonStates();
	void  ClearCurrentButtonStates();
	void  ShiftButtonStates();
	void  ShiftCopyButtonStates();
	void  ResetNumButtonsCache();
	void  UpdateNumButtonsCache();

	//------------------
	// Template methods.
	//------------------
	virtual  void  OnInputModeChanged() = 0;
	virtual  void  OnSyncWithSystemCursorChanged() = 0;
};


//====================================================================
#define  VW_INTERFACE_NAME_Mouse  "Mouse"


//********************************************************************
}//namespace vertexware


