

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "MathTools\MatrixBase.h"
#include "MathTools\Vector3.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class mat2;
class mat4;


//====================================================================
// 3 x 3 matrix.
// Storage form is row-major.
//====================================================================
class VW_COMMON_API mat3
{
public:
	//------
	// Data.
	//------
	union {
		float  mm[3][3];
		float  m[9];
		struct {
			float  m00, m01, m02;
			float  m10, m11, m12;
			float  m20, m21, m22;
		};
	};

	//-----------------------
	// Convenience constants.
	//-----------------------
	static const mat3  ZERO;
	static const mat3  IDENTITY;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline mat3()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline mat3(float _00, float _01, float _02,
							float _10, float _11, float _12,
							float _20, float _21, float _22) {
		m00 = _00;
		m01 = _01;
		m02 = _02;
		m10 = _10;
		m11 = _11;
		m12 = _12;
		m20 = _20;
		m21 = _21;
		m22 = _22;
	}

	__forceinline mat3(const float* a) {
		m[0] = a[0];
		m[1] = a[1];
		m[2] = a[2];
		m[3] = a[3];
		m[4] = a[4];
		m[5] = a[5];
		m[6] = a[6];
		m[7] = a[7];
		m[8] = a[8];
	}

	__forceinline mat3(const vec3& row0, const vec3& row1, const vec3& row2) {
		m00 = row0.x;
		m01 = row0.y;
		m02 = row0.z;
		m10 = row1.x;
		m11 = row1.y;
		m12 = row1.z;
		m20 = row2.x;
		m21 = row2.y;
		m22 = row2.z;
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline mat3(const mat3& rhs) {
		m00 = rhs.m00;
		m01 = rhs.m01;
		m02 = rhs.m02;
		m10 = rhs.m10;
		m11 = rhs.m11;
		m12 = rhs.m12;
		m20 = rhs.m20;
		m21 = rhs.m21;
		m22 = rhs.m22;
	}

	__forceinline mat3(const mat2& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m02 = 0;
		m10 = reinterpret_cast<const float*>(&rhs)[2];
		m11 = reinterpret_cast<const float*>(&rhs)[3];
		m12 = 0;
		m20 = 0;
		m21 = 0;
		m22 = 1;
	}

	__forceinline mat3(const mat4& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m02 = reinterpret_cast<const float*>(&rhs)[2];
		m10 = reinterpret_cast<const float*>(&rhs)[4];
		m11 = reinterpret_cast<const float*>(&rhs)[5];
		m12 = reinterpret_cast<const float*>(&rhs)[6];
		m20 = reinterpret_cast<const float*>(&rhs)[8];
		m21 = reinterpret_cast<const float*>(&rhs)[9];
		m22 = reinterpret_cast<const float*>(&rhs)[10];
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline mat3& operator=(const mat3& rhs) {
		m00 = rhs.m00;
		m01 = rhs.m01;
		m02 = rhs.m02;
		m10 = rhs.m10;
		m11 = rhs.m11;
		m12 = rhs.m12;
		m20 = rhs.m20;
		m21 = rhs.m21;
		m22 = rhs.m22;
		return *this;
	}

	__forceinline mat3& operator=(const mat2& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m02 = 0;
		m10 = reinterpret_cast<const float*>(&rhs)[2];
		m11 = reinterpret_cast<const float*>(&rhs)[3];
		m12 = 0;
		m20 = 0;
		m21 = 0;
		m22 = 1;
		return *this;
	}

	__forceinline mat3& operator=(const mat4& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m02 = reinterpret_cast<const float*>(&rhs)[2];
		m10 = reinterpret_cast<const float*>(&rhs)[4];
		m11 = reinterpret_cast<const float*>(&rhs)[5];
		m12 = reinterpret_cast<const float*>(&rhs)[6];
		m20 = reinterpret_cast<const float*>(&rhs)[8];
		m21 = reinterpret_cast<const float*>(&rhs)[9];
		m22 = reinterpret_cast<const float*>(&rhs)[10];
		return *this;
	}

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(float _00, float _01, float _02,
							float _10, float _11, float _12,
							float _20, float _21, float _22) {
		m00 = _00;
		m01 = _01;
		m02 = _02;
		m10 = _10;
		m11 = _11;
		m12 = _12;
		m20 = _20;
		m21 = _21;
		m22 = _22;
	}

	__forceinline void  Set(const float* a) {
		m[0] = a[0];
		m[1] = a[1];
		m[2] = a[2];
		m[3] = a[3];
		m[4] = a[4];
		m[5] = a[5];
		m[6] = a[6];
		m[7] = a[7];
		m[8] = a[8];
	}

	__forceinline void  Set(const vec3& row0, const vec3& row1, const vec3& row2) {
		m00 = row0.x;
		m01 = row0.y;
		m02 = row0.z;
		m10 = row1.x;
		m11 = row1.y;
		m12 = row1.z;
		m20 = row2.x;
		m21 = row2.y;
		m22 = row2.z;
	}

	//---------------
	// Special forms.
	//---------------
	__forceinline void  SetIdentity() {
		m00 = 1;
		m01 = 0;
		m02 = 0;
		m10 = 0;
		m11 = 1;
		m12 = 0;
		m20 = 0;
		m21 = 0;
		m22 = 1;
	}

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline float*        ToPtr()       { return &m00; }
	__forceinline const float*  ToPtr() const { return &m00; }

	__forceinline operator float*()             { return ToPtr(); }
	__forceinline operator const float*() const { return ToPtr(); }

	//------------------
	// Grid-like access.
	//------------------
	__forceinline float*        operator[](uint i)       { return &mm[i][0]; }
	__forceinline const float*  operator[](uint i) const { return &mm[i][0]; }

	__forceinline float*        operator[](int i)       { return &mm[i][0]; }
	__forceinline const float*  operator[](int i) const { return &mm[i][0]; }

	//------------
	// Row access.
	//------------
	__forceinline vec3&        Row0()       { return reinterpret_cast<vec3&>(mm[0][0]); }
	__forceinline const vec3&  Row0() const { return reinterpret_cast<const vec3&>(mm[0][0]); }

	__forceinline vec3&        Row1()       { return reinterpret_cast<vec3&>(mm[1][0]); }
	__forceinline const vec3&  Row1() const { return reinterpret_cast<const vec3&>(mm[1][0]); }

	__forceinline vec3&        Row2()       { return reinterpret_cast<vec3&>(mm[2][0]); }
	__forceinline const vec3&  Row2() const { return reinterpret_cast<const vec3&>(mm[2][0]); }

	__forceinline vec3&        Row(uint i)       { return reinterpret_cast<vec3&>(mm[i][0]); }
	__forceinline const vec3&  Row(uint i) const { return reinterpret_cast<const vec3&>(mm[i][0]); }

	__forceinline vec3&        Row(int i)       { return reinterpret_cast<vec3&>(mm[i][0]); }
	__forceinline const vec3&  Row(int i) const { return reinterpret_cast<const vec3&>(mm[i][0]); }

	//-------------------
	// Column generators.
	//-------------------
	__forceinline vec3  Column0() const {
		return vec3(m00, m10, m20);
	}
	__forceinline void  Column0(vec3& dst) const {
		dst.Set(m00, m10, m20);
	}

	__forceinline vec3  Column1() const {
		return vec3(m01, m11, m21);
	}
	__forceinline void  Column1(vec3& dst) const {
		dst.Set(m01, m11, m21);
	}

	__forceinline vec3  Column2() const {
		return vec3(m02, m12, m22);
	}
	__forceinline void  Column2(vec3& dst) const {
		dst.Set(m02, m12, m22);
	}

	__forceinline vec3  Column(uint i) const {
		return vec3(mm[0][i], mm[1][i], mm[2][i]);
	}
	__forceinline void  Column(uint i, vec3& dst) const {
		dst.Set(mm[0][i], mm[1][i], mm[2][i]);
	}

	__forceinline vec3  Column(int i) const {
		return vec3(mm[0][i], mm[1][i], mm[2][i]);
	}
	__forceinline void  Column(int i, vec3& dst) const {
		dst.Set(mm[0][i], mm[1][i], mm[2][i]);
	}

	//----------
	// Addition.
	//----------
	__forceinline mat3&  operator+=(const mat3& rhs) {
		m00 += rhs.m00;
		m01 += rhs.m01;
		m02 += rhs.m02;
		m10 += rhs.m10;
		m11 += rhs.m11;
		m12 += rhs.m12;
		m20 += rhs.m20;
		m21 += rhs.m21;
		m22 += rhs.m22;
		return *this;
	}

	__forceinline mat3&  Add(const mat3& rhs) {
		m00 += rhs.m00;
		m01 += rhs.m01;
		m02 += rhs.m02;
		m10 += rhs.m10;
		m11 += rhs.m11;
		m12 += rhs.m12;
		m20 += rhs.m20;
		m21 += rhs.m21;
		m22 += rhs.m22;
		return *this;
	}

	__forceinline mat3  operator+(const mat3& rhs) const {
		return mat3(
				m00 + rhs.m00, m01 + rhs.m01, m02 + rhs.m02,
				m10 + rhs.m10, m11 + rhs.m11, m12 + rhs.m12,
				m20 + rhs.m20, m21 + rhs.m21, m22 + rhs.m22);
	}

	__forceinline mat3&  Add(const mat3& lhs, const mat3& rhs) {
		m00 = lhs.m00 + rhs.m00;
		m01 = lhs.m01 + rhs.m01;
		m02 = lhs.m02 + rhs.m02;
		m10 = lhs.m10 + rhs.m10;
		m11 = lhs.m11 + rhs.m11;
		m12 = lhs.m12 + rhs.m12;
		m20 = lhs.m20 + rhs.m20;
		m21 = lhs.m21 + rhs.m21;
		m22 = lhs.m22 + rhs.m22;
		return *this;
	}

	//-------------
	// Subtraction.
	//-------------
	__forceinline mat3&  operator-=(const mat3& rhs) {
		m00 -= rhs.m00;
		m01 -= rhs.m01;
		m02 -= rhs.m02;
		m10 -= rhs.m10;
		m11 -= rhs.m11;
		m12 -= rhs.m12;
		m20 -= rhs.m20;
		m21 -= rhs.m21;
		m22 -= rhs.m22;
		return *this;
	}

	__forceinline mat3&  Sub(const mat3& rhs) {
		m00 -= rhs.m00;
		m01 -= rhs.m01;
		m02 -= rhs.m02;
		m10 -= rhs.m10;
		m11 -= rhs.m11;
		m12 -= rhs.m12;
		m20 -= rhs.m20;
		m21 -= rhs.m21;
		m22 -= rhs.m22;
		return *this;
	}

	__forceinline mat3  operator-(const mat3& rhs) const {
		return mat3(
				m00 - rhs.m00, m01 - rhs.m01, m02 - rhs.m02,
				m10 - rhs.m10, m11 - rhs.m11, m12 - rhs.m12,
				m20 - rhs.m20, m21 - rhs.m21, m22 - rhs.m22);
	}

	__forceinline mat3&  Sub(const mat3& lhs, const mat3& rhs) {
		m00 = lhs.m00 - rhs.m00;
		m01 = lhs.m01 - rhs.m01;
		m02 = lhs.m02 - rhs.m02;
		m10 = lhs.m10 - rhs.m10;
		m11 = lhs.m11 - rhs.m11;
		m12 = lhs.m12 - rhs.m12;
		m20 = lhs.m20 - rhs.m20;
		m21 = lhs.m21 - rhs.m21;
		m22 = lhs.m22 - rhs.m22;
		return *this;
	}

	//--------------------------
	// Multiplication by matrix.
	//--------------------------
	__forceinline mat3&  operator*=(const mat3& rhs) {
		return this->Mul(mat3(*this), rhs);
	}

	__forceinline mat3  operator*(const mat3& rhs) const {
		mat3 temp;
		temp.Mul(*this, rhs);
		return temp;
	}

	__forceinline mat3&  PreMul(const mat3& lhs) {
		return this->Mul(lhs, mat3(*this));
	}

	__forceinline mat3&  PostMul(const mat3& rhs) {
		return this->Mul(mat3(*this), rhs);
	}

	mat3&  Mul(const mat3& lhs, const mat3& rhs);

	//--------------------------
	// Multiplication by scalar.
	//--------------------------
	__forceinline mat3&  operator*=(float num) {
		m00 *= num;
		m01 *= num;
		m02 *= num;
		m10 *= num;
		m11 *= num;
		m12 *= num;
		m20 *= num;
		m21 *= num;
		m22 *= num;
		return *this;
	}

	__forceinline mat3&  Mul(float num) {
		m00 *= num;
		m01 *= num;
		m02 *= num;
		m10 *= num;
		m11 *= num;
		m12 *= num;
		m20 *= num;
		m21 *= num;
		m22 *= num;
		return *this;
	}

	__forceinline mat3  operator*(float num) const {
		return mat3(
				m00 * num, m01 * num, m02 * num,
				m10 * num, m11 * num, m12 * num,
				m20 * num, m21 * num, m22 * num);
	}

	__forceinline mat3&  Mul(const mat3& lhs, float num) {
		m00 = lhs.m00 * num;
		m01 = lhs.m01 * num;
		m02 = lhs.m02 * num;
		m10 = lhs.m10 * num;
		m11 = lhs.m11 * num;
		m12 = lhs.m12 * num;
		m20 = lhs.m20 * num;
		m21 = lhs.m21 * num;
		m22 = lhs.m22 * num;
		return *this;
	}

	//----------------------------------------------
	// Division by scalar (no divide-by-zero check).
	//----------------------------------------------
	__forceinline mat3&  operator/=(float num) {
		return this->Mul(1 / num);
	}

	__forceinline mat3&  Div(float num) {
		return this->Mul(1 / num);
	}

	__forceinline mat3  operator/(float num) const {
		num = 1 / num;
		return mat3(
				m00 * num, m01 * num, m02 * num,
				m10 * num, m11 * num, m12 * num,
				m20 * num, m21 * num, m22 * num);
	}

	__forceinline mat3&  Div(const mat3& lhs, float num) {
		return this->Mul(lhs, 1 / num);
	}

	//----------
	// Negation.
	//----------
	__forceinline mat3&  Neg() {
		return this->Mul(*this, -1);
	}

	__forceinline mat3&  Neg(const mat3& src) {
		return this->Mul(src, -1);
	}

	//------------------
	// Unary operations.
	//------------------
	__forceinline mat3  operator-() const {
		return mat3(
				-m00, -m01, -m02,
				-m10, -m11, -m12,
				-m20, -m21, -m22);
	}

	__forceinline mat3&        operator+()       { return *this; }
	__forceinline const mat3&  operator+() const { return *this; }

	//-----------
	// Transpose.
	//-----------
	__forceinline mat3&  Transpose() {
		float  temp;
		Swap(m01, m10, temp);
		Swap(m02, m20, temp);
		Swap(m12, m21, temp);
		return *this;
	}

	__forceinline mat3  Transposed() const {
		return mat3(m00, m10, m20,
						m01, m11, m21,
						m02, m12, m22);
	}

	__forceinline mat3&  Transposed(const mat3& src) {
		Set(src.m00, src.m10, src.m20,
			src.m01, src.m11, src.m21,
			src.m02, src.m12, src.m22);
		return *this;
	}

	//---------
	// Inverse.
	//---------
	__forceinline mat3&  Invert(bool* pFailed = 0) {
		return this->Inverse(mat3(*this), pFailed);
	}

	__forceinline mat3  Inverse(bool* pFailed = 0) const {
		mat3  temp;
		return temp.Inverse(*this, pFailed);
	}

	mat3&  Inverse(const mat3& src, bool* pFailed = 0);

	//------------------------------
	// Inverse (exception versions).
	//------------------------------
	__forceinline mat3&  InvertE() {
		return this->InverseE(mat3(*this));
	}

	__forceinline mat3  InverseE() const {
		mat3  temp;
		return temp.InverseE(*this);
	}

	__forceinline mat3&  InverseE(const mat3& src) {
		bool failed;
		this->Inverse(src, &failed);
		if(failed) {
			VW_THROW(MatrixNotInvertibleException, "Uninvertible matrix.");
		}
		return *this;
	}

	//-------
	// Trace.
	//-------
	__forceinline float  Trace() const {
		return  (m00 + m11 + m22);
	}

	//-------------
	// Determinant.
	//-------------
	float  Determinant() const {
		return (m00 * (m11 * m22 - m12 * m21) +
				m01 * (m12 * m20 - m10 * m22) +
				m02 * (m10 * m21 - m11 * m20));
	}

	//-------------
	// Comparisons.
	//-------------
	bool  operator==(const mat3& rhs) const;

	__forceinline bool  operator!=(const mat3& rhs) const {
		return !(*this == rhs);
	}

	//----------------------------------------------------
	// Noise filtering (zero-out all elements close to 0).
	//----------------------------------------------------
	__forceinline mat3&  NoiseFilter(const mat3& src, float tolerance =
										math_constants::MatrixNoiseFilterDefaultTolerance) {
		vertexware::NoiseFilter(this->ToPtr(), 9, tolerance);
		return *this;
	}

	__forceinline  mat3 __fastcall NoiseFiltered(float tolerance =
										math_constants::MatrixNoiseFilterDefaultTolerance) const {
		mat3 temp;
		vertexware::NoiseFilter(temp.ToPtr(), this->ToPtr(), 9, tolerance);
		return temp;
	}
};


//====================================================================
// Multiplication by scalar (redirection).
//====================================================================
__forceinline  mat3 __fastcall operator*(float num, const mat3& rhs) {
	return rhs * num;
}


//====================================================================
// Row-column dot product.
//====================================================================
__forceinline float __fastcall Dot(const mat3& lhs, uint row, const mat3& rhs, uint column)
{
	return
	lhs.mm[row][0] * rhs.mm[0][column] +
	lhs.mm[row][1] * rhs.mm[1][column] +
	lhs.mm[row][2] * rhs.mm[2][column];
}


//====================================================================
// Row-column dot product.
// Usage example:
//    mat3 a, b;
//    float d = Dot01(a, b); // takes dot of row0 from a & column1 from b
//====================================================================
#define  VW_IMPLEMENT_MATRIX33_DOT(row, column) \
	__forceinline float __fastcall Dot##row##column(const mat3& lhs, const mat3& rhs) { \
		return \
		lhs.mm[row][0] * rhs.mm[0][column] + \
		lhs.mm[row][1] * rhs.mm[1][column] + \
		lhs.mm[row][2] * rhs.mm[2][column]; \
	}


VW_IMPLEMENT_MATRIX33_DOT(0, 0)
VW_IMPLEMENT_MATRIX33_DOT(0, 1)
VW_IMPLEMENT_MATRIX33_DOT(0, 2)
VW_IMPLEMENT_MATRIX33_DOT(1, 0)
VW_IMPLEMENT_MATRIX33_DOT(1, 1)
VW_IMPLEMENT_MATRIX33_DOT(1, 2)
VW_IMPLEMENT_MATRIX33_DOT(2, 0)
VW_IMPLEMENT_MATRIX33_DOT(2, 1)
VW_IMPLEMENT_MATRIX33_DOT(2, 2)


#undef  VW_IMPLEMENT_MATRIX33_DOT


//====================================================================
// Multiplication matrix * vector.
//====================================================================
__forceinline vec3& __fastcall Mul(vec3& dst, const mat3& m, const vec3& v)
{
	dst.Set(m.Row0().Dot(v), m.Row1().Dot(v), m.Row2().Dot(v));
	return dst;
}


//====================================================================
__forceinline vec3 __fastcall operator*(const mat3& m, const vec3& v)
{
	return
	vec3(m.Row0().Dot(v), m.Row1().Dot(v), m.Row2().Dot(v));
}


//====================================================================
// Multiplication vector * matrix.
//====================================================================
__forceinline vec3& __fastcall Mul(vec3& dst, const vec3& v, const mat3& m)
{
	dst.Set(v.x * m.m00 + v.y * m.m10 + v.z * m.m20,
			v.x * m.m01 + v.y * m.m11 + v.z * m.m21,
			v.x * m.m02 + v.y * m.m12 + v.z * m.m22);
	return dst;
}


//====================================================================
__forceinline vec3 __fastcall operator*(const vec3& v, const mat3& m)
{
	return
	vec3(v.x * m.m00 + v.y * m.m10 + v.z * m.m20,
			v.x * m.m01 + v.y * m.m11 + v.z * m.m21,
			v.x * m.m02 + v.y * m.m12 + v.z * m.m22);
}


//====================================================================
// Element-access indirectors. Useful for MV/VM matrix analyzers.
//====================================================================
template<typename TTranspose>
struct Matrix33Accessor
{ };


//====================================================================
template<>
struct Matrix33Accessor<no_transpose_t>
{
	mat3& m;

	Matrix33Accessor(mat3& _m) : m(_m) {}
	__forceinline float&  operator()(uint row, uint column) { return m.mm[row][column]; };
};


//====================================================================
template<>
struct Matrix33Accessor<transpose_t>
{
	mat3& m;

	Matrix33Accessor(mat3& _m) : m(_m) {}
	__forceinline float&  operator()(uint row, uint column) { return m.mm[column][row]; };
};


//====================================================================
template<typename TTranspose>
struct Matrix33ConstAccessor
{ };


//====================================================================
template<>
struct Matrix33ConstAccessor<no_transpose_t>
{
	const mat3& m;

	Matrix33ConstAccessor(const mat3& _m) : m(_m) {}
	__forceinline float  operator()(uint row, uint column) { return m.mm[row][column]; };
};


//====================================================================
template<>
struct Matrix33ConstAccessor<transpose_t>
{
	const mat3& m;

	Matrix33ConstAccessor(const mat3& _m) : m(_m) {}
	__forceinline float  operator()(uint row, uint column) { return m.mm[column][row]; };
};


//====================================================================
// Fill indirectors. Useful for MV/VM matrix generators.
//====================================================================
template<typename TTranspose>
__forceinline void __fastcall Fill(const TTranspose&, mat3& m,
									float _00, float _01, float _02,
									float _10, float _11, float _12,
									float _20, float _21, float _22);


//====================================================================
template<>
__forceinline void __fastcall Fill(const no_transpose_t&, mat3& m,
									float _00, float _01, float _02,
									float _10, float _11, float _12,
									float _20, float _21, float _22)
{
	m.m00 = _00;
	m.m01 = _01;
	m.m02 = _02;
	m.m10 = _10;
	m.m11 = _11;
	m.m12 = _12;
	m.m20 = _20;
	m.m21 = _21;
	m.m22 = _22;
}


//====================================================================
template<>
__forceinline void __fastcall Fill(const transpose_t&, mat3& m,
									float _00, float _01, float _02,
									float _10, float _11, float _12,
									float _20, float _21, float _22)
{
	m.m00 = _00;
	m.m01 = _10;
	m.m02 = _20;
	m.m10 = _01;
	m.m11 = _11;
	m.m12 = _21;
	m.m20 = _02;
	m.m21 = _12;
	m.m22 = _22;
}


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const mat3& m);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, mat3& m);


//********************************************************************
}//namespace vertexware


