

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "MathTools\MatrixBase.h"
#include "MathTools\Vector4.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class mat2;
class mat3;


//====================================================================
// 4 x 4 matrix.
// Storage form is row-major.
//====================================================================
class VW_COMMON_API mat4
{
public:
	//------
	// Data.
	//------
	union {
		float  mm[4][4];
		float  m[16];
		struct {
			float  m00, m01, m02, m03;
			float  m10, m11, m12, m13;
			float  m20, m21, m22, m23;
			float  m30, m31, m32, m33;
		};
	};

	//-----------------------
	// Convenience constants.
	//-----------------------
	static const mat4  ZERO;
	static const mat4  IDENTITY;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline mat4()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline mat4(float _00, float _01, float _02, float _03, float _10, float _11, float _12, float _13,
							float _20, float _21, float _22, float _23, float _30, float _31, float _32, float _33)
	{
		m00 = _00;
		m01 = _01;
		m02 = _02;
		m03 = _03;
		m10 = _10;
		m11 = _11;
		m12 = _12;
		m13 = _13;
		m20 = _20;
		m21 = _21;
		m22 = _22;
		m23 = _23;
		m30 = _30;
		m31 = _31;
		m32 = _32;
		m33 = _33;
	}

	__forceinline mat4(const float* a) {
		m[0]  = a[0];
		m[1]  = a[1];
		m[2]  = a[2];
		m[3]  = a[3];
		m[4]  = a[4];
		m[5]  = a[5];
		m[6]  = a[6];
		m[7]  = a[7];
		m[8]  = a[8];
		m[9]  = a[9];
		m[10] = a[10];
		m[11] = a[11];
		m[12] = a[12];
		m[13] = a[13];
		m[14] = a[14];
		m[15] = a[15];
	}

	__forceinline mat4(const vec4& row0, const vec4& row1,
			const vec4& row2, const vec4& row3)
	{
		m00 = row0.x;
		m01 = row0.y;
		m02 = row0.z;
		m03 = row0.w;
		m10 = row1.x;
		m11 = row1.y;
		m12 = row1.z;
		m13 = row1.w;
		m20 = row2.x;
		m21 = row2.y;
		m22 = row2.z;
		m23 = row2.w;
		m30 = row3.x;
		m31 = row3.y;
		m32 = row3.z;
		m33 = row3.w;
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline mat4(const mat4& rhs) {
		m00 = rhs.m00;
		m01 = rhs.m01;
		m02 = rhs.m02;
		m03 = rhs.m03;
		m10 = rhs.m10;
		m11 = rhs.m11;
		m12 = rhs.m12;
		m13 = rhs.m13;
		m20 = rhs.m20;
		m21 = rhs.m21;
		m22 = rhs.m22;
		m23 = rhs.m23;
		m30 = rhs.m30;
		m31 = rhs.m31;
		m32 = rhs.m32;
		m33 = rhs.m33;
	}

	__forceinline mat4(const mat2& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m02 = 0;
		m03 = 0;
		m10 = reinterpret_cast<const float*>(&rhs)[2];
		m11 = reinterpret_cast<const float*>(&rhs)[3];
		m12 = 0;
		m13 = 0;
		m20 = 0;
		m21 = 0;
		m22 = 0;
		m23 = 0;
		m30 = 0;
		m31 = 0;
		m32 = 0;
		m33 = 1;
	}

	__forceinline mat4(const mat3& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m02 = reinterpret_cast<const float*>(&rhs)[2];
		m03 = 0;
		m10 = reinterpret_cast<const float*>(&rhs)[3];
		m11 = reinterpret_cast<const float*>(&rhs)[4];
		m12 = reinterpret_cast<const float*>(&rhs)[5];
		m13 = 0;
		m20 = reinterpret_cast<const float*>(&rhs)[6];
		m21 = reinterpret_cast<const float*>(&rhs)[7];
		m22 = reinterpret_cast<const float*>(&rhs)[8];
		m23 = 0;
		m30 = 0;
		m31 = 0;
		m32 = 0;
		m33 = 1;
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline mat4& operator=(const mat4& rhs) {
		m00 = rhs.m00;
		m01 = rhs.m01;
		m02 = rhs.m02;
		m03 = rhs.m03;
		m10 = rhs.m10;
		m11 = rhs.m11;
		m12 = rhs.m12;
		m13 = rhs.m13;
		m20 = rhs.m20;
		m21 = rhs.m21;
		m22 = rhs.m22;
		m23 = rhs.m23;
		m30 = rhs.m30;
		m31 = rhs.m31;
		m32 = rhs.m32;
		m33 = rhs.m33;
		return *this;
	}

	__forceinline mat4& operator=(const mat2& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m02 = 0;
		m03 = 0;
		m10 = reinterpret_cast<const float*>(&rhs)[2];
		m11 = reinterpret_cast<const float*>(&rhs)[3];
		m12 = 0;
		m13 = 0;
		m20 = 0;
		m21 = 0;
		m22 = 0;
		m23 = 0;
		m30 = 0;
		m31 = 0;
		m32 = 0;
		m33 = 1;
		return *this;
	}

	__forceinline mat4& operator=(const mat3& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m02 = reinterpret_cast<const float*>(&rhs)[2];
		m03 = 0;
		m10 = reinterpret_cast<const float*>(&rhs)[3];
		m11 = reinterpret_cast<const float*>(&rhs)[4];
		m12 = reinterpret_cast<const float*>(&rhs)[5];
		m13 = 0;
		m20 = reinterpret_cast<const float*>(&rhs)[6];
		m21 = reinterpret_cast<const float*>(&rhs)[7];
		m22 = reinterpret_cast<const float*>(&rhs)[8];
		m23 = 0;
		m30 = 0;
		m31 = 0;
		m32 = 0;
		m33 = 1;
		return *this;
	}

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(float _00, float _01, float _02, float _03, float _10, float _11, float _12, float _13,
							float _20, float _21, float _22, float _23, float _30, float _31, float _32, float _33)
	{
		m00 = _00;
		m01 = _01;
		m02 = _02;
		m03 = _03;
		m10 = _10;
		m11 = _11;
		m12 = _12;
		m13 = _13;
		m20 = _20;
		m21 = _21;
		m22 = _22;
		m23 = _23;
		m30 = _30;
		m31 = _31;
		m32 = _32;
		m33 = _33;
	}

	__forceinline void  Set(const float* a) {
		m[0]  = a[0];
		m[1]  = a[1];
		m[2]  = a[2];
		m[3]  = a[3];
		m[4]  = a[4];
		m[5]  = a[5];
		m[6]  = a[6];
		m[7]  = a[7];
		m[8]  = a[8];
		m[9]  = a[9];
		m[10] = a[10];
		m[11] = a[11];
		m[12] = a[12];
		m[13] = a[13];
		m[14] = a[14];
		m[15] = a[15];
	}

	__forceinline void  Set(const vec4& row0, const vec4& row1,
				const vec4& row2, const vec4& row3)
	{
		m00 = row0.x;
		m01 = row0.y;
		m02 = row0.z;
		m03 = row0.w;
		m10 = row1.x;
		m11 = row1.y;
		m12 = row1.z;
		m13 = row1.w;
		m20 = row2.x;
		m21 = row2.y;
		m22 = row2.z;
		m23 = row2.w;
		m30 = row3.x;
		m31 = row3.y;
		m32 = row3.z;
		m33 = row3.w;
	}

	//---------------
	// Special forms.
	//---------------
	__forceinline void  SetIdentity() {
		m00 = 1;
		m01 = 0;
		m02 = 0;
		m03 = 0;
		m10 = 0;
		m11 = 1;
		m12 = 0;
		m13 = 0;
		m20 = 0;
		m21 = 0;
		m22 = 1;
		m23 = 0;
		m30 = 0;
		m31 = 0;
		m32 = 0;
		m33 = 1;
	}

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline float*        ToPtr()       { return &m00; }
	__forceinline const float*  ToPtr() const { return &m00; }

	__forceinline operator float*()             { return ToPtr(); }
	__forceinline operator const float*() const { return ToPtr(); }

	//------------------
	// Grid-like access.
	//------------------
	__forceinline float*        operator[](uint i)       { return &mm[i][0]; }
	__forceinline const float*  operator[](uint i) const { return &mm[i][0]; }

	__forceinline float*        operator[](int i)       { return &mm[i][0]; }
	__forceinline const float*  operator[](int i) const { return &mm[i][0]; }

	//------------
	// Row access.
	//------------
	__forceinline vec4&        Row0()       { return reinterpret_cast<vec4&>(mm[0][0]); }
	__forceinline const vec4&  Row0() const { return reinterpret_cast<const vec4&>(mm[0][0]); }

	__forceinline vec4&        Row1()       { return reinterpret_cast<vec4&>(mm[1][0]); }
	__forceinline const vec4&  Row1() const { return reinterpret_cast<const vec4&>(mm[1][0]); }

	__forceinline vec4&        Row2()       { return reinterpret_cast<vec4&>(mm[2][0]); }
	__forceinline const vec4&  Row2() const { return reinterpret_cast<const vec4&>(mm[2][0]); }

	__forceinline vec4&        Row3()       { return reinterpret_cast<vec4&>(mm[3][0]); }
	__forceinline const vec4&  Row3() const { return reinterpret_cast<const vec4&>(mm[3][0]); }

	__forceinline vec4&        Row(uint i)       { return reinterpret_cast<vec4&>(mm[i][0]); }
	__forceinline const vec4&  Row(uint i) const { return reinterpret_cast<const vec4&>(mm[i][0]); }

	__forceinline vec4&        Row(int i)       { return reinterpret_cast<vec4&>(mm[i][0]); }
	__forceinline const vec4&  Row(int i) const { return reinterpret_cast<const vec4& >(mm[i][0]); }

	//-------------------
	// Column generators.
	//-------------------
	__forceinline vec4  Column0() const {
		return vec4(m00, m10, m20, m30);
	}
	__forceinline void  Column0(vec4& dst) const {
		dst.Set(m00, m10, m20, m30);
	}

	__forceinline vec4  Column1() const {
		return vec4(m01, m11, m21, m31);
	}
	__forceinline void  Column1(vec4& dst) const {
		dst.Set(m01, m11, m21, m31);
	}

	__forceinline vec4  Column2() const {
		return vec4(m02, m12, m22, m32);
	}
	__forceinline void  Column2(vec4& dst) const {
		dst.Set(m02, m12, m22, m32);
	}

	__forceinline vec4  Column3() const {
		return vec4(m03, m13, m23, m33);
	}
	__forceinline void  Column3(vec4& dst) const {
		dst.Set(m03, m13, m23, m33);
	}

	__forceinline vec4  Column(uint i) const {
		return vec4(mm[0][i], mm[1][i], mm[2][i], mm[3][i]);
	}
	__forceinline void  Column(uint i, vec4& dst) const {
		dst.Set(mm[0][i], mm[1][i], mm[2][i], mm[3][i]);
	}

	__forceinline vec4  Column(int i) const {
		return vec4(mm[0][i], mm[1][i], mm[2][i], mm[3][i]);
	}
	__forceinline void  Column(int i, vec4& dst) const {
		dst.Set(mm[0][i], mm[1][i], mm[2][i], mm[3][i]);
	}

	//----------
	// Addition.
	//----------
	__forceinline mat4&  operator+=(const mat4& rhs) {
		m00 += rhs.m00;
		m01 += rhs.m01;
		m02 += rhs.m02;
		m03 += rhs.m03;
		m10 += rhs.m10;
		m11 += rhs.m11;
		m12 += rhs.m12;
		m13 += rhs.m13;
		m20 += rhs.m20;
		m21 += rhs.m21;
		m22 += rhs.m22;
		m23 += rhs.m23;
		m30 += rhs.m30;
		m31 += rhs.m31;
		m32 += rhs.m32;
		m33 += rhs.m33;
		return *this;
	}

	__forceinline mat4&  Add(const mat4& rhs) {
		m00 += rhs.m00;
		m01 += rhs.m01;
		m02 += rhs.m02;
		m03 += rhs.m03;
		m10 += rhs.m10;
		m11 += rhs.m11;
		m12 += rhs.m12;
		m13 += rhs.m13;
		m20 += rhs.m20;
		m21 += rhs.m21;
		m22 += rhs.m22;
		m23 += rhs.m23;
		m30 += rhs.m30;
		m31 += rhs.m31;
		m32 += rhs.m32;
		m33 += rhs.m33;
		return *this;
	}

	__forceinline mat4  operator+(const mat4& rhs) const {
		return mat4(
				m00 + rhs.m00, m01 + rhs.m01, m02 + rhs.m02, m03 + rhs.m03,
				m10 + rhs.m10, m11 + rhs.m11, m12 + rhs.m12, m13 + rhs.m13,
				m20 + rhs.m20, m21 + rhs.m21, m22 + rhs.m22, m23 + rhs.m23,
				m30 + rhs.m30, m31 + rhs.m31, m32 + rhs.m32, m33 + rhs.m33);
	}

	__forceinline mat4&  Add(const mat4& lhs, const mat4& rhs) {
		m00 = lhs.m00 + rhs.m00;
		m01 = lhs.m01 + rhs.m01;
		m02 = lhs.m02 + rhs.m02;
		m03 = lhs.m03 + rhs.m03;
		m10 = lhs.m10 + rhs.m10;
		m11 = lhs.m11 + rhs.m11;
		m12 = lhs.m12 + rhs.m12;
		m13 = lhs.m13 + rhs.m13;
		m20 = lhs.m20 + rhs.m20;
		m21 = lhs.m21 + rhs.m21;
		m22 = lhs.m22 + rhs.m22;
		m23 = lhs.m23 + rhs.m23;
		m30 = lhs.m30 + rhs.m30;
		m31 = lhs.m31 + rhs.m31;
		m32 = lhs.m32 + rhs.m32;
		m33 = lhs.m33 + rhs.m33;
		return *this;
	}

	//-------------
	// Subtraction.
	//-------------
	__forceinline mat4&  operator-=(const mat4& rhs) {
		m00 -= rhs.m00;
		m01 -= rhs.m01;
		m02 -= rhs.m02;
		m03 -= rhs.m03;
		m10 -= rhs.m10;
		m11 -= rhs.m11;
		m12 -= rhs.m12;
		m13 -= rhs.m13;
		m20 -= rhs.m20;
		m21 -= rhs.m21;
		m22 -= rhs.m22;
		m23 -= rhs.m23;
		m30 -= rhs.m30;
		m31 -= rhs.m31;
		m32 -= rhs.m32;
		m33 -= rhs.m33;
		return *this;
	}

	__forceinline mat4&  Sub(const mat4& rhs) {
		m00 -= rhs.m00;
		m01 -= rhs.m01;
		m02 -= rhs.m02;
		m03 -= rhs.m03;
		m10 -= rhs.m10;
		m11 -= rhs.m11;
		m12 -= rhs.m12;
		m13 -= rhs.m13;
		m20 -= rhs.m20;
		m21 -= rhs.m21;
		m22 -= rhs.m22;
		m23 -= rhs.m23;
		m30 -= rhs.m30;
		m31 -= rhs.m31;
		m32 -= rhs.m32;
		m33 -= rhs.m33;
		return *this;
	}

	__forceinline mat4  operator-(const mat4& rhs) const {
		return mat4(
				m00 - rhs.m00, m01 - rhs.m01, m02 - rhs.m02, m03 - rhs.m03,
				m10 - rhs.m10, m11 - rhs.m11, m12 - rhs.m12, m13 - rhs.m13,
				m20 - rhs.m20, m21 - rhs.m21, m22 - rhs.m22, m23 - rhs.m23,
				m30 - rhs.m30, m31 - rhs.m31, m32 - rhs.m32, m33 - rhs.m33);
	}

	__forceinline mat4&  Sub(const mat4& lhs, const mat4& rhs) {
		m00 = lhs.m00 - rhs.m00;
		m01 = lhs.m01 - rhs.m01;
		m02 = lhs.m02 - rhs.m02;
		m03 = lhs.m03 - rhs.m03;
		m10 = lhs.m10 - rhs.m10;
		m11 = lhs.m11 - rhs.m11;
		m12 = lhs.m12 - rhs.m12;
		m13 = lhs.m13 - rhs.m13;
		m20 = lhs.m20 - rhs.m20;
		m21 = lhs.m21 - rhs.m21;
		m22 = lhs.m22 - rhs.m22;
		m23 = lhs.m23 - rhs.m23;
		m30 = lhs.m30 - rhs.m30;
		m31 = lhs.m31 - rhs.m31;
		m32 = lhs.m32 - rhs.m32;
		m33 = lhs.m33 - rhs.m33;
		return *this;
	}

	//--------------------------
	// Multiplication by matrix.
	//--------------------------
	__forceinline mat4&  operator*=(const mat4& rhs) {
		return this->Mul(mat4(*this), rhs);
	}

	__forceinline mat4  operator*(const mat4& rhs) const {
		mat4 temp;
		temp.Mul(*this, rhs);
		return temp;
	}

	__forceinline mat4&  PreMul(const mat4& lhs) {
		return this->Mul(lhs, mat4(*this));
	}

	__forceinline mat4&  PostMul(const mat4& rhs) {
		return this->Mul(mat4(*this), rhs);
	}

	mat4&  Mul(const mat4& lhs, const mat4& rhs);

	//--------------------------
	// Multiplication by scalar.
	//--------------------------
	__forceinline mat4&  operator*=(float num) {
		m00 *= num;
		m01 *= num;
		m02 *= num;
		m03 *= num;
		m10 *= num;
		m11 *= num;
		m12 *= num;
		m13 *= num;
		m20 *= num;
		m21 *= num;
		m22 *= num;
		m23 *= num;
		m30 *= num;
		m31 *= num;
		m32 *= num;
		m33 *= num;
		return *this;
	}

	__forceinline mat4&  Mul(float num) {
		m00 *= num;
		m01 *= num;
		m02 *= num;
		m03 *= num;
		m10 *= num;
		m11 *= num;
		m12 *= num;
		m13 *= num;
		m20 *= num;
		m21 *= num;
		m22 *= num;
		m23 *= num;
		m30 *= num;
		m31 *= num;
		m32 *= num;
		m33 *= num;
		return *this;
	}

	__forceinline mat4  operator*(float num) const {
		return mat4(
				m00 * num, m01 * num, m02 * num, m03 * num,
				m10 * num, m11 * num, m12 * num, m13 * num,
				m20 * num, m21 * num, m22 * num, m23 * num,
				m30 * num, m31 * num, m32 * num, m33 * num);
	}

	__forceinline mat4&  Mul(const mat4& lhs, float num) {
		m00 = lhs.m00 * num;
		m01 = lhs.m01 * num;
		m02 = lhs.m02 * num;
		m03 = lhs.m03 * num;
		m10 = lhs.m10 * num;
		m11 = lhs.m11 * num;
		m12 = lhs.m12 * num;
		m13 = lhs.m13 * num;
		m20 = lhs.m20 * num;
		m21 = lhs.m21 * num;
		m22 = lhs.m22 * num;
		m23 = lhs.m23 * num;
		m30 = lhs.m30 * num;
		m31 = lhs.m31 * num;
		m32 = lhs.m32 * num;
		m33 = lhs.m33 * num;
		return *this;
	}

	//----------------------------------------------
	// Division by scalar (no divide-by-zero check).
	//----------------------------------------------
	__forceinline mat4&  operator/=(float num) {
		return this->Mul(1 / num);
	}

	__forceinline mat4&  Div(float num) {
		return this->Mul(1 / num);
	}

	__forceinline mat4  operator/(float num) const {
		num = 1 / num;
		return mat4(
				m00 * num, m01 * num, m02 * num, m03 * num,
				m10 * num, m11 * num, m12 * num, m13 * num,
				m20 * num, m21 * num, m22 * num, m23 * num,
				m30 * num, m31 * num, m32 * num, m33 * num);
	}

	__forceinline mat4&  Div(const mat4& lhs, float num) {
		return this->Mul(lhs, 1 / num);
	}

	//----------
	// Negation.
	//----------
	__forceinline mat4&  Neg() {
		return this->Mul(*this, -1);
	}

	__forceinline mat4&  Neg(const mat4& src) {
		return this->Mul(src, -1);
	}

	//------------------
	// Unary operations.
	//------------------
	__forceinline mat4  operator-() const {
		return mat4(
				-m00, -m01, -m02, -m03,
				-m10, -m11, -m12, -m13,
				-m20, -m21, -m22, -m23,
				-m30, -m31, -m32, -m33);
	}

	__forceinline mat4&        operator+()       { return *this; }
	__forceinline const mat4&  operator+() const { return *this; }

	//-----------
	// Transpose.
	//-----------
	__forceinline mat4&  Transpose() {
		float  temp;
		Swap(m01, m10, temp);
		Swap(m02, m20, temp);
		Swap(m03, m30, temp);
		Swap(m12, m21, temp);
		Swap(m13, m31, temp);
		Swap(m23, m32, temp);
		return *this;
	}

	__forceinline mat4  Transposed() const {
		return mat4(m00, m10, m20, m30,
						m01, m11, m21, m31,
						m02, m12, m22, m32,
						m03, m13, m23, m33);
	}

	__forceinline mat4&  Transposed(const mat4& src) {
		Set(src.m00, src.m10, src.m20, src.m30,
			src.m01, src.m11, src.m21, src.m31,
			src.m02, src.m12, src.m22, src.m32,
			src.m03, src.m13, src.m23, src.m33);
		return *this;
	}

	//---------
	// Inverse.
	//---------
	__forceinline mat4&  Invert(bool* pFailed = 0) {
		return this->Inverse(*this, pFailed);
	}

	__forceinline mat4  Inverse(bool* pFailed = 0) const {
		mat4  temp;
		return temp.Inverse(*this, pFailed);
	}

	// NC-version is faster, but leaves <src> argument in undefined state.
	// NC stands for "No-Copy" or "Non-Constant".
	mat4&  InverseNC(mat4& src, bool* pFailed = 0);

	__forceinline  mat4&  Inverse(const mat4& src, bool* pFailed = 0) {
		mat4  temp = src;
		return this->InverseNC(temp, pFailed);
	}

	//------------------------------
	// Inverse (exception versions).
	//------------------------------
	__forceinline mat4&  InvertE() {
		return this->InverseE(*this);
	}

	__forceinline mat4  InverseE() const {
		mat4  temp;
		return temp.InverseE(*this);
	}

	__forceinline mat4&  InverseNCE(mat4& src) {
		bool failed;
		this->InverseNC(src, &failed);
		if(failed) {
			VW_THROW(MatrixNotInvertibleException, "Uninvertible matrix.");
		}
		return *this;
	}

	__forceinline mat4&  InverseE(const mat4& src) {
		mat4  temp = src;
		return this->InverseNCE(temp);
	}

	//-------
	// Trace.
	//-------
	__forceinline float  Trace() const {
		return  (m00 + m11 + m22 + m33);
	}

	//-------------
	// Determinant.
	//-------------
	float  Determinant() const;

	//-------------
	// Comparisons.
	//-------------
	bool  operator==(const mat4& rhs) const;

	__forceinline bool  operator!=(const mat4& rhs) const {
		return !(*this == rhs);
	}

	//----------------------------------------------------
	// Noise filtering (zero-out all elements close to 0).
	//----------------------------------------------------
	__forceinline mat4&  NoiseFilter(const mat4& src, float tolerance =
										math_constants::MatrixNoiseFilterDefaultTolerance) {
		vertexware::NoiseFilter(this->ToPtr(), 16, tolerance);
		return *this;
	}

	__forceinline  mat4 __fastcall NoiseFiltered(float tolerance =
										math_constants::MatrixNoiseFilterDefaultTolerance) const {
		mat4 temp;
		vertexware::NoiseFilter(temp.ToPtr(), this->ToPtr(), 16, tolerance);
		return temp;
	}
};


//====================================================================
// Multiplication by scalar (redirection).
//====================================================================
__forceinline  mat4 __fastcall operator*(float num, const mat4& rhs) {
	return rhs * num;
}


//====================================================================
// Row-column dot product.
//====================================================================
__forceinline float __fastcall Dot(const mat4& lhs, uint row, const mat4& rhs, uint column)
{
	return
	lhs.mm[row][0] * rhs.mm[0][column] +
	lhs.mm[row][1] * rhs.mm[1][column] +
	lhs.mm[row][2] * rhs.mm[2][column] +
	lhs.mm[row][3] * rhs.mm[3][column];
}


//====================================================================
// Row-column dot product.
// Usage example:
//    mat4 a, b;
//    float d = Dot01(a, b); // takes dot of row0 from a & column1 from b
//====================================================================
#define  VW_IMPLEMENT_MATRIX44_DOT(row, column) \
	__forceinline float __fastcall Dot##row##column(const mat4& lhs, const mat4& rhs) { \
		return \
		lhs.mm[row][0] * rhs.mm[0][column] + \
		lhs.mm[row][1] * rhs.mm[1][column] + \
		lhs.mm[row][2] * rhs.mm[2][column] + \
		lhs.mm[row][3] * rhs.mm[3][column]; \
	}


VW_IMPLEMENT_MATRIX44_DOT(0, 0)
VW_IMPLEMENT_MATRIX44_DOT(0, 1)
VW_IMPLEMENT_MATRIX44_DOT(0, 2)
VW_IMPLEMENT_MATRIX44_DOT(0, 3)
VW_IMPLEMENT_MATRIX44_DOT(1, 0)
VW_IMPLEMENT_MATRIX44_DOT(1, 1)
VW_IMPLEMENT_MATRIX44_DOT(1, 2)
VW_IMPLEMENT_MATRIX44_DOT(1, 3)
VW_IMPLEMENT_MATRIX44_DOT(2, 0)
VW_IMPLEMENT_MATRIX44_DOT(2, 1)
VW_IMPLEMENT_MATRIX44_DOT(2, 2)
VW_IMPLEMENT_MATRIX44_DOT(2, 3)
VW_IMPLEMENT_MATRIX44_DOT(3, 0)
VW_IMPLEMENT_MATRIX44_DOT(3, 1)
VW_IMPLEMENT_MATRIX44_DOT(3, 2)
VW_IMPLEMENT_MATRIX44_DOT(3, 3)


#undef  VW_IMPLEMENT_MATRIX44_DOT


//====================================================================
// Multiplication matrix * vector.
//====================================================================
__forceinline vec4& __fastcall Mul(vec4& dst, const mat4& m, const vec4& v)
{
	dst.Set(m.Row0().Dot(v), m.Row1().Dot(v), m.Row2().Dot(v), m.Row3().Dot(v));
	return dst;
}


//====================================================================
__forceinline vec4 __fastcall operator*(const mat4& m, const vec4& v)
{
	return
	vec4(m.Row0().Dot(v), m.Row1().Dot(v), m.Row2().Dot(v), m.Row3().Dot(v));
}


//====================================================================
// Multiplication vector * matrix.
//====================================================================
__forceinline vec4& __fastcall Mul(vec4& dst, const vec4& v, const mat4& m)
{
	dst.Set(v.x * m.m00 + v.y * m.m10 + v.z * m.m20 + v.w * m.m30,
			v.x * m.m01 + v.y * m.m11 + v.z * m.m21 + v.w * m.m31,
			v.x * m.m02 + v.y * m.m12 + v.z * m.m22 + v.w * m.m32,
			v.x * m.m03 + v.y * m.m13 + v.z * m.m23 + v.w * m.m33);
	return dst;
}


//====================================================================
__forceinline vec4 __fastcall operator*(const vec4& v, const mat4& m)
{
	return
	vec4(v.x * m.m00 + v.y * m.m10 + v.z * m.m20 + v.w * m.m30,
			v.x * m.m01 + v.y * m.m11 + v.z * m.m21 + v.w * m.m31,
			v.x * m.m02 + v.y * m.m12 + v.z * m.m22 + v.w * m.m32,
			v.x * m.m03 + v.y * m.m13 + v.z * m.m23 + v.w * m.m33);
}


//====================================================================
// Element-access indirectors. Useful for MV/VM matrix analyzers.
//====================================================================
template<typename TTranspose>
struct Matrix44Accessor
{ };


//====================================================================
template<>
struct Matrix44Accessor<no_transpose_t>
{
	mat4& m;

	Matrix44Accessor(mat4& _m) : m(_m) {}
	__forceinline float&  operator()(uint row, uint column) { return m.mm[row][column]; };
};


//====================================================================
template<>
struct Matrix44Accessor<transpose_t>
{
	mat4& m;

	Matrix44Accessor(mat4& _m) : m(_m) {}
	__forceinline float&  operator()(uint row, uint column) { return m.mm[column][row]; };
};


//====================================================================
template<typename TTranspose>
struct Matrix44ConstAccessor
{ };


//====================================================================
template<>
struct Matrix44ConstAccessor<no_transpose_t>
{
	const mat4& m;

	Matrix44ConstAccessor(const mat4& _m) : m(_m) {}
	__forceinline float  operator()(uint row, uint column) { return m.mm[row][column]; };
};


//====================================================================
template<>
struct Matrix44ConstAccessor<transpose_t>
{
	const mat4& m;

	Matrix44ConstAccessor(const mat4& _m) : m(_m) {}
	__forceinline float  operator()(uint row, uint column) { return m.mm[column][row]; };
};


//====================================================================
// Fill indirectors. Useful for MV/VM matrix generators.
//====================================================================
template<typename TTranspose>
__forceinline void __fastcall Fill(const TTranspose&, mat4& m,
									float _00, float _01, float _02, float _03,
									float _10, float _11, float _12, float _13,
									float _20, float _21, float _22, float _23,
									float _30, float _31, float _32, float _33);


//====================================================================
template<>
__forceinline void __fastcall Fill(const no_transpose_t&, mat4& m,
									float _00, float _01, float _02, float _03,
									float _10, float _11, float _12, float _13,
									float _20, float _21, float _22, float _23,
									float _30, float _31, float _32, float _33)
{
	m.m00 = _00;
	m.m01 = _01;
	m.m02 = _02;
	m.m03 = _03;
	m.m10 = _10;
	m.m11 = _11;
	m.m12 = _12;
	m.m13 = _13;
	m.m20 = _20;
	m.m21 = _21;
	m.m22 = _22;
	m.m23 = _23;
	m.m30 = _30;
	m.m31 = _31;
	m.m32 = _32;
	m.m33 = _33;
}


//====================================================================
template<>
__forceinline void __fastcall Fill(const transpose_t&, mat4& m,
									float _00, float _01, float _02, float _03,
									float _10, float _11, float _12, float _13,
									float _20, float _21, float _22, float _23,
									float _30, float _31, float _32, float _33)
{
	m.m00 = _00;
	m.m01 = _10;
	m.m02 = _20;
	m.m03 = _30;
	m.m10 = _01;
	m.m11 = _11;
	m.m12 = _21;
	m.m13 = _31;
	m.m20 = _02;
	m.m21 = _12;
	m.m22 = _22;
	m.m23 = _32;
	m.m30 = _03;
	m.m31 = _13;
	m.m32 = _23;
	m.m33 = _33;
}


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const mat4& m);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, mat4& m);


//********************************************************************
}//namespace vertexware


