

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "MathTools\VectorBase.h"
#include "MathTools\MathProcessors.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class vec3;
class vec4;


//====================================================================
// 2-element vector.
//====================================================================
class VW_COMMON_API vec2
{
public:
	//------
	// Data.
	//------
	union {
		float  v[2];
		struct {
			float  x;
			float  y;
		};
	};

	//-----------------------
	// Convenience constants.
	//-----------------------
	static const vec2  ZERO;
	static const vec2  ONE;
	static const vec2  INFINITY;
	static const vec2  UNIT_X;
	static const vec2  UNIT_Y;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline vec2()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline vec2(float _x, float _y) {
		x = _x;
		y = _y;
	}

	__forceinline explicit vec2(const float* a) {
		x = a[0];
		y = a[1];
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline vec2(const vec2& rhs) {
		x = rhs.x;
		y = rhs.y;
	}

	__forceinline vec2(const vec3& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
	}

	__forceinline vec2(const vec4& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline vec2& operator=(const vec2& rhs) {
		x = rhs.x;
		y = rhs.y;
		return *this;
	}

	vec2& operator=(const vec3& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		return *this;
	}

	vec2& operator=(const vec4& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		return *this;
	}

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(float _x, float _y) {
		x = _x;
		y = _y;
	}
	__forceinline void  Set(const float* a) {
		x = a[0];
		y = a[1];
	}

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline float*        ToPtr()       { return &x; }
	__forceinline const float*  ToPtr() const { return &x; }

	__forceinline operator float*()             { return ToPtr(); }
	__forceinline operator const float*() const { return ToPtr(); }

	//----------
	// Addition.
	//----------
	__forceinline vec2&  operator+=(const vec2& rhs) {
		x += rhs.x;
		y += rhs.y;
		return *this;
	}

	__forceinline vec2&  Add(const vec2& rhs) {
		x += rhs.x;
		y += rhs.y;
		return *this;
	}

	__forceinline vec2  operator+(const vec2& rhs) const {
		return vec2(x + rhs.x, y + rhs.y);
	}

	__forceinline vec2&  Add(const vec2& lhs, const vec2& rhs) {
		x = lhs.x + rhs.x;
		y = lhs.y + rhs.y;
		return *this;
	}

	//-------------
	// Subtraction.
	//-------------
	__forceinline vec2&  operator-=(const vec2& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		return *this;
	}

	__forceinline vec2&  Sub(const vec2& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		return *this;
	}

	__forceinline vec2  operator-(const vec2& rhs) const {
		return vec2(x - rhs.x, y - rhs.y);
	}

	__forceinline vec2&  Sub(const vec2& lhs, const vec2& rhs) {
		x = lhs.x - rhs.x;
		y = lhs.y - rhs.y;
		return *this;
	}

	//----------------
	// Multiplication.
	//----------------
	__forceinline vec2&  operator*=(const vec2& rhs) {
		x *= rhs.x;
		y *= rhs.y;
		return *this;
	}

	__forceinline vec2&  Mul(const vec2& rhs) {
		x *= rhs.x;
		y *= rhs.y;
		return *this;
	}

	__forceinline vec2  operator*(const vec2& rhs) const {
		return vec2(x * rhs.x, y * rhs.y);
	}

	__forceinline vec2&  Mul(const vec2& lhs, const vec2& rhs) {
		x = lhs.x * rhs.x;
		y = lhs.y * rhs.y;
		return *this;
	}

	//------------------------------------
	// Division (no divide-by-zero check).
	//------------------------------------
	__forceinline vec2&  operator/=(const vec2& rhs) {
		x /= rhs.x;
		y /= rhs.y;
		return *this;
	}

	__forceinline vec2&  Div(const vec2& rhs) {
		x /= rhs.x;
		y /= rhs.y;
		return *this;
	}

	__forceinline vec2  operator/(const vec2& rhs) const {
		return vec2(x / rhs.x, y / rhs.y);
	}

	__forceinline vec2&  Div(const vec2& lhs, const vec2& rhs) {
		x = lhs.x / rhs.x;
		y = lhs.y / rhs.y;
		return *this;
	}

	//--------------------------
	// Multiplication by scalar.
	//--------------------------
	__forceinline vec2&  operator*=(float num) {
		x *= num;
		y *= num;
		return *this;
	}

	__forceinline vec2&  Mul(float num) {
		x *= num;
		y *= num;
		return *this;
	}

	__forceinline vec2  operator*(float num) const {
		return vec2(x * num, y * num);
	}

	__forceinline vec2&  Mul(const vec2& lhs, float num) {
		x = lhs.x * num;
		y = lhs.y * num;
		return *this;
	}

	//----------------------------------------------
	// Division by scalar (no divide-by-zero check).
	//----------------------------------------------
	__forceinline vec2&  operator/=(float num) {
		num = 1 / num;
		x *= num;
		y *= num;
		return *this;
	}

	__forceinline vec2&  Div(float num) {
		num = 1 / num;
		x *= num;
		y *= num;
		return *this;
	}

	__forceinline vec2  operator/(float num) const {
		num = 1 / num;
		return vec2(x * num, y * num);
	}

	__forceinline vec2&  Div(const vec2& lhs, float num) {
		num = 1 / num;
		x = lhs.x * num;
		y = lhs.y * num;
		return *this;
	}

	//----------
	// Negation.
	//----------
	__forceinline vec2&  Neg() {
		x *= -1;
		y *= -1;
		return *this;
	}

	__forceinline vec2&  Neg(const vec2& src) {
		x = - src.x;
		y = - src.y;
		return *this;
	}

	//------------------
	// Unary operations.
	//------------------
	__forceinline vec2  operator-() const {
		return vec2(-x, -y);
	}

	__forceinline vec2&        operator+()       { return *this; }
	__forceinline const vec2&  operator+() const { return *this; }

	//----------------------------
	// Absolute minimum component.
	//----------------------------
	__forceinline float  AbsMin() const {
		return Min(Abs(x), Abs(y));
	}

	//----------------------------
	// Absolute maximum component.
	//----------------------------
	__forceinline float  AbsMax() const {
		return Max(Abs(x), Abs(y));
	}

	//--------
	// Length.
	//--------
	__forceinline float  Length() const {
		return Sqrt<float>(x * x + y * y);
	}

	//----------------
	// Squared length.
	//----------------
	__forceinline float  LengthSquared() const {
		return (x * x + y * y);
	}

	//---------
	// Unitize.
	//---------
	__forceinline vec2&  Unitize() {
		float len = Length();
		if(len > 0) {
			len = 1 / len;
			x *= len;
			y *= len;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	__forceinline vec2  Unit() const {
		float len = Length();
		if(len > 0) {
			len = 1 / len;
			return vec2(x * len, y * len);
		}
		return INFINITY;
	}

	__forceinline vec2&  Unit(const vec2& src) {
		float len = src.Length();
		if(len > 0) {
			len = 1 / len;
			x = src.x * len;
			y = src.y * len;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	//-------------
	// Dot product.
	//-------------
	__forceinline float  Dot(const vec2& rhs) const {
		return (x * rhs.x + y * rhs.y);
	}

	//------------------------------------------------
	// Cross products (counter-clockwise & clockwise).
	//------------------------------------------------
	__forceinline vec2  CrossCCW() const {
		return vec2(-y, x);
	}
	__forceinline vec2&  CrossCCW(const vec2& src) {
		x = - src.y;
		y =   src.x;
		return *this;
	}

	__forceinline vec2  CrossCW() const {
		return vec2(y, -x);
	}
	__forceinline vec2&  CrossCW(const vec2& src) {
		x =   src.y;
		y = - src.x;
		return *this;
	}

	//------------
	// Projection.
	//------------
	__forceinline vec2&  ProjectToUnit(const vec2& to) {
		*this = to * this->Dot(to);
		return *this;
	}
	__forceinline vec2   ProjectionToUnit(const vec2& to) const {
		return to * this->Dot(to);
	}
	__forceinline vec2&   ProjectionToUnit(const vec2& to, const vec2& src) {
		return this->Mul(to, src.Dot(to));
	}

	__forceinline vec2&  ProjectTo(const vec2& to) {
		*this = to * (this->Dot(to) / to.LengthSquared());
		return *this;
	}
	__forceinline vec2   ProjectionTo(const vec2& to) const {
		return to * (this->Dot(to) / to.LengthSquared());
	}
	__forceinline vec2&   ProjectionTo(const vec2& to, const vec2& src) {
		return this->Mul(to, src.Dot(to) / to.LengthSquared());
	}

	//----------
	// Distance.
	//----------
	__forceinline float  Distance(const vec2& rhs) const {
		return (*this - rhs).Length();
	}

	//------------------
	// Squared distance.
	//------------------
	__forceinline float  DistanceSquared(const vec2& rhs) const {
		return (*this - rhs).LengthSquared();
	}

	//------------
	// Zero tests.
	//------------
	__forceinline bool  IsZero() const {
		return ((x == 0) && (y == 0));
	}
	bool  IsZero(float tolerance) const;

	__forceinline bool  HasZeroComponent() const {
		return ((x == 0) || (y == 0));
	}
	bool  HasZeroComponent(float tolerance) const;

	//-------------
	// Angle tests.
	//-------------
	__forceinline bool  Orthogonal(const vec2& rhs, float maxCosAngle =
									math_constants::VectorOrthogonalTestDefaultMaxCosAngle) const {
		return (Abs(this->Dot(rhs)) <= maxCosAngle);
	}
	__forceinline bool  Parallel(const vec2& rhs, float minCosAngle =
									math_constants::VectorParallelTestDefaultMinCosAngle) const {
		return (Abs(this->Dot(rhs)) >= minCosAngle);
	}

	//-------------
	// Comparisons.
	//-------------
	__forceinline bool  operator==(const vec2& rhs) const {
		return ((x == rhs.x) && (y == rhs.y));
	}
	__forceinline bool  operator!=(const vec2& rhs) const {
		return !(*this == rhs);
	}
	__forceinline bool  Equal(const vec2& rhs, float toleranceRadius =
								math_constants::VectorEqualTestDefaultToleranceRadius) const {
		return (DistanceSquared(rhs) <= toleranceRadius * toleranceRadius);
	}

	//----------------------------------------------------
	// Noise filtering (zero-out all elements close to 0).
	//----------------------------------------------------
	__forceinline vec2&  NoiseFilter(const vec2& src, float tolerance =
										math_constants::VectorNoiseFilterDefaultTolerance) {
		vertexware::NoiseFilter(this->ToPtr(), 2, tolerance);
		return *this;
	}

	__forceinline  vec2 __fastcall NoiseFiltered(float tolerance =
										math_constants::VectorNoiseFilterDefaultTolerance) const {
		vec2 temp;
		vertexware::NoiseFilter(temp.ToPtr(), this->ToPtr(), 2, tolerance);
		return temp;
	}

	//------------------------------------------
	// Rotation (counter-clockwise & clockwise).
	//------------------------------------------
	__forceinline vec2&  RotateCCW(float angleDeg) {
		*this = RotationCCW(angleDeg);
		return *this;
	}
	__forceinline vec2  RotationCCW(float angleDeg) const {
		float  sinA, cosA;
		SinCosDeg(angleDeg, sinA, cosA);
		return vec2(x * cosA - y * sinA, x * sinA + y * cosA);
	}
	__forceinline vec2&  RotationCCW(float angleDeg, const vec2& src) {
		float  sinA, cosA;
		SinCosDeg(angleDeg, sinA, cosA);
		x = src.x * cosA - src.y * sinA;
		y = src.x * sinA + src.y * cosA;
		return *this;
	}

	__forceinline vec2&  RotateCW(float angleDeg) {
		*this = RotationCW(angleDeg);
		return *this;
	}
	__forceinline vec2  RotationCW(float angleDeg) const {
		return RotationCCW(- angleDeg);
	}
	__forceinline vec2&  RotationCW(float angleDeg, const vec2& src) {
		return RotationCCW(- angleDeg, src);
	}
};


//====================================================================
// Clamping.
//====================================================================
__forceinline  vec2 __fastcall ComponentAbs(const vec2& val)
{
	return vec2(Abs(val.x), Abs(val.y));
}


//====================================================================
__forceinline  vec2 __fastcall ComponentMin(const vec2& v1, const vec2& v2) {
	return vec2(Min(v1.x, v2.x), Min(v1.y, v2.y));
}


//====================================================================
__forceinline  vec2 __fastcall ComponentMax(const vec2& v1, const vec2& v2) {
	return vec2(Max(v1.x, v2.x), Max(v1.y, v2.y));
}


//====================================================================
__forceinline  vec2 __fastcall ComponentClamp(const vec2& v, const vec2& low, const vec2& high) {
	return ComponentMax(ComponentMin(v, high), low);
}


//====================================================================
// Multiplication by scalar (redirection).
//====================================================================
__forceinline  vec2 __fastcall operator*(float num, const vec2& rhs) {
	return rhs * num;
}


//====================================================================
// Orthonormalization.
//====================================================================
VW_COMMON_API
void __fastcall OrthoNormalize(vec2& v0, vec2& v1);


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const vec2& v);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, vec2& v);


//********************************************************************
}//namespace vertexware


