

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "MathTools\VectorBase.h"
#include "MathTools\MathProcessors.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class vec2;
class vec3;


//====================================================================
// 4-element vector.
//====================================================================
class VW_COMMON_API vec4
{
public:
	//------
	// Data.
	//------
	union {
		float  v[4];
		struct {
			float  x;
			float  y;
			float  z;
			float  w;
		};
	};

	//-----------------------
	// Convenience constants.
	//-----------------------
	static const vec4  ZERO;
	static const vec4  ONE;
	static const vec4  INFINITY;
	static const vec4  UNIT_X;
	static const vec4  UNIT_Y;
	static const vec4  UNIT_Z;
	static const vec4  UNIT_W;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline vec4()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline vec4(float _x, float _y, float _z, float _w) {
		x = _x;
		y = _y;
		z = _z;
		w = _w;
	}

	__forceinline explicit vec4(const float* a) {
		x = a[0];
		y = a[1];
		z = a[2];
		w = a[3];
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline vec4(const vec4& rhs) {
		x = rhs.x;
		y = rhs.y;
		z = rhs.z;
		w = rhs.w;
	}

	__forceinline vec4(const vec2& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		z = 0;
		w = 1;
	}

	__forceinline vec4(const vec3& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		z = reinterpret_cast<const float*>(&rhs)[2];
		w = 1;
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline vec4& operator=(const vec4& rhs) {
		x = rhs.x;
		y = rhs.y;
		z = rhs.z;
		w = rhs.w;
		return *this;
	}

	vec4& operator=(const vec2& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		z = 0;
		w = 1;
		return *this;
	}

	vec4& operator=(const vec3& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		z = reinterpret_cast<const float*>(&rhs)[2];
		w = 1;
		return *this;
	}

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(float _x, float _y, float _z, float _w) {
		x = _x;
		y = _y;
		z = _z;
		w = _w;
	}

	__forceinline void  Set(const float* a) {
		x = a[0];
		y = a[1];
		z = a[2];
		w = a[3];
	}

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline float*        ToPtr()       { return &x; }
	__forceinline const float*  ToPtr() const { return &x; }

	__forceinline operator float*()             { return ToPtr(); }
	__forceinline operator const float*() const { return ToPtr(); }

	//----------
	// Addition.
	//----------
	__forceinline vec4&  operator+=(const vec4& rhs) {
		x += rhs.x;
		y += rhs.y;
		z += rhs.z;
		w += rhs.w;
		return *this;
	}

	__forceinline vec4&  Add(const vec4& rhs) {
		x += rhs.x;
		y += rhs.y;
		z += rhs.z;
		w += rhs.w;
		return *this;
	}

	__forceinline vec4  operator+(const vec4& rhs) const {
		return vec4(x + rhs.x, y + rhs.y, z + rhs.z, w + rhs.w);
	}

	__forceinline vec4&  Add(const vec4& lhs, const vec4& rhs) {
		x = lhs.x + rhs.x;
		y = lhs.y + rhs.y;
		z = lhs.z + rhs.z;
		w = lhs.w + rhs.w;
		return *this;
	}

	//-------------
	// Subtraction.
	//-------------
	__forceinline vec4&  operator-=(const vec4& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		z -= rhs.z;
		w -= rhs.w;
		return *this;
	}

	__forceinline vec4&  Sub(const vec4& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		z -= rhs.z;
		w -= rhs.w;
		return *this;
	}

	__forceinline vec4  operator-(const vec4& rhs) const {
		return vec4(x - rhs.x, y - rhs.y, z - rhs.z, w - rhs.w);
	}

	__forceinline vec4&  Sub(const vec4& lhs, const vec4& rhs) {
		x = lhs.x - rhs.x;
		y = lhs.y - rhs.y;
		z = lhs.z - rhs.z;
		w = lhs.w - rhs.w;
		return *this;
	}

	//----------------
	// Multiplication.
	//----------------
	__forceinline vec4&  operator*=(const vec4& rhs) {
		x *= rhs.x;
		y *= rhs.y;
		z *= rhs.z;
		w *= rhs.w;
		return *this;
	}

	__forceinline vec4&  Mul(const vec4& rhs) {
		x *= rhs.x;
		y *= rhs.y;
		z *= rhs.z;
		w *= rhs.w;
		return *this;
	}

	__forceinline vec4  operator*(const vec4& rhs) const {
		return vec4(x * rhs.x, y * rhs.y, z * rhs.z, w * rhs.w);
	}

	__forceinline vec4&  Mul(const vec4& lhs, const vec4& rhs) {
		x = lhs.x * rhs.x;
		y = lhs.y * rhs.y;
		z = lhs.z * rhs.z;
		w = lhs.w * rhs.w;
		return *this;
	}

	//------------------------------------
	// Division (no divide-by-zero check).
	//------------------------------------
	__forceinline vec4&  operator/=(const vec4& rhs) {
		x /= rhs.x;
		y /= rhs.y;
		z /= rhs.z;
		w /= rhs.w;
		return *this;
	}

	__forceinline vec4&  Div(const vec4& rhs) {
		x /= rhs.x;
		y /= rhs.y;
		z /= rhs.z;
		w /= rhs.w;
		return *this;
	}

	__forceinline vec4  operator/(const vec4& rhs) const {
		return vec4(x / rhs.x, y / rhs.y, z / rhs.z, w / rhs.w);
	}

	__forceinline vec4&  Div(const vec4& lhs, const vec4& rhs) {
		x = lhs.x / rhs.x;
		y = lhs.y / rhs.y;
		z = lhs.z / rhs.z;
		w = lhs.w / rhs.w;
		return *this;
	}

	//--------------------------
	// Multiplication by scalar.
	//--------------------------
	__forceinline vec4&  operator*=(float num) {
		x *= num;
		y *= num;
		z *= num;
		w *= num;
		return *this;
	}

	__forceinline vec4&  Mul(float num) {
		x *= num;
		y *= num;
		z *= num;
		w *= num;
		return *this;
	}

	__forceinline vec4  operator*(float num) const {
		return vec4(x * num, y * num, z * num, w * num);
	}

	__forceinline vec4&  Mul(const vec4& lhs, float num) {
		x = lhs.x * num;
		y = lhs.y * num;
		z = lhs.z * num;
		w = lhs.w * num;
		return *this;
	}

	//----------------------------------------------
	// Division by scalar (no divide-by-zero check).
	//----------------------------------------------
	__forceinline vec4&  operator/=(float num) {
		num = 1 / num;
		x *= num;
		y *= num;
		z *= num;
		w *= num;
		return *this;
	}

	__forceinline vec4&  Div(float num) {
		num = 1 / num;
		x *= num;
		y *= num;
		z *= num;
		w *= num;
		return *this;
	}

	__forceinline vec4  operator/(float num) const {
		num = 1 / num;
		return vec4(x * num, y * num, z * num, w * num);
	}

	__forceinline vec4&  Div(const vec4& lhs, float num) {
		num = 1 / num;
		x = lhs.x * num;
		y = lhs.y * num;
		z = lhs.z * num;
		w = lhs.w * num;
		return *this;
	}

	//----------
	// Negation.
	//----------
	__forceinline vec4&  Neg() {
		x *= -1;
		y *= -1;
		z *= -1;
		w *= -1;
		return *this;
	}

	__forceinline vec4&  Neg(const vec4& src) {
		x = - src.x;
		y = - src.y;
		z = - src.z;
		w = - src.w;
		return *this;
	}

	//------------------
	// Unary operations.
	//------------------
	__forceinline vec4  operator-() const {
		return vec4(-x, -y, -z, -w);
	}

	__forceinline vec4&        operator+()       { return *this; }
	__forceinline const vec4&  operator+() const { return *this; }

	//-------------
	// Dot product.
	//-------------
	__forceinline float  Dot(const vec4& rhs) const {
		return (x * rhs.x + y * rhs.y + z * rhs.z + w * rhs.w);
	}

	//------------
	// Zero tests.
	//------------
	__forceinline bool  IsZero() const {
		return ((x == 0) && (y == 0) && (z == 0) && (w == 0));
	}
	bool  IsZero(float tolerance) const;

	__forceinline bool  HasZeroComponent() const {
		return ((x == 0) || (y == 0) || (z == 0) || (w == 0));
	}
	bool  HasZeroComponent(float tolerance) const;

	//-------------
	// Comparisons.
	//-------------
	__forceinline bool  operator==(const vec4& rhs) const {
		return ((x == rhs.x) && (y == rhs.y) && (z == rhs.z) && (w == rhs.w));
	}
	__forceinline bool  operator!=(const vec4& rhs) const {
		return !(*this == rhs);
	}

	//----------------------------------------------------
	// Noise filtering (zero-out all elements close to 0).
	//----------------------------------------------------
	__forceinline vec4&  NoiseFilter(const vec4& src, float tolerance =
										math_constants::VectorNoiseFilterDefaultTolerance) {
		vertexware::NoiseFilter(this->ToPtr(), 4, tolerance);
		return *this;
	}

	__forceinline  vec4 __fastcall NoiseFiltered(float tolerance =
										math_constants::VectorNoiseFilterDefaultTolerance) const {
		vec4 temp;
		vertexware::NoiseFilter(temp.ToPtr(), this->ToPtr(), 4, tolerance);
		return temp;
	}

	//-------------------------
	// Normalize (divide by w).
	//-------------------------
	__forceinline vec4&  Normalize() {
		if(w != 0) {
			float num = 1 / w;
			x *= num;
			y *= num;
			z *= num;
			w = 1;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	__forceinline vec4  Normalized() const {
		if(w != 0) {
			float num = 1 / w;
			return vec4(x * num, y * num, z * num, 1);
		}
		return INFINITY;
	}

	__forceinline vec4&  Normalized(const vec4& src) {
		if(src.w != 0) {
			float num = 1 / src.w;
			x = src.x * num;
			y = src.y * num;
			z = src.z * num;
			w = 1;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}
};


//====================================================================
// Clamping.
//====================================================================
__forceinline  vec4 __fastcall ComponentAbs(const vec4& val)
{
	return vec4(Abs(val.x), Abs(val.y), Abs(val.z), Abs(val.w));
}


//====================================================================
__forceinline  vec4 __fastcall ComponentMin(const vec4& v1, const vec4& v2) {
	return vec4(Min(v1.x, v2.x), Min(v1.y, v2.y), Min(v1.z, v2.z), Min(v1.w, v2.w));
}


//====================================================================
__forceinline  vec4 __fastcall ComponentMax(const vec4& v1, const vec4& v2) {
	return vec4(Max(v1.x, v2.x), Max(v1.y, v2.y), Max(v1.z, v2.z), Max(v1.w, v2.w));
}


//====================================================================
__forceinline  vec4 __fastcall ComponentClamp(const vec4& v, const vec4& low, const vec4& high) {
	return ComponentMax(ComponentMin(v, high), low);
}


//====================================================================
// Multiplication by scalar (redirection).
//====================================================================
__forceinline  vec4 __fastcall operator*(float num, const vec4& rhs) {
	return rhs * num;
}


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const vec4& v);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, vec4& v);


//********************************************************************
}//namespace vertexware


