

//--------------------------------------------------------------------
#include "DataGlove.h"
#include "BaseTools\String.h"


//********************************************************************
namespace {


//====================================================================
// Returns -1 if sensor cannot be mapped.
//====================================================================
static int __fastcall MapSensorToHW(eGloveSensor sensor)
{
	switch(sensor) {
		case GloveSensor_WristYaw: return FD_WRISTBEND;
		case GloveSensor_WristPitch: return FD_PITCH;
		case GloveSensor_WristRoll: return FD_ROLL;

		case GloveSensor_AbductionThumbIndex: return FD_THUMBINDEX;
		case GloveSensor_AbductionIndexMiddle: return FD_INDEXMIDDLE;
		case GloveSensor_AbductionMiddleRing: return FD_MIDDLERING;
		case GloveSensor_AbductionRingPinky: return FD_RINGLITTLE;

		case GloveSensor_ThumbRoll: return FD_THUMBPALM;
		case GloveSensor_ThumbJointInner: return FD_THUMBNEAR;
		case GloveSensor_ThumbJointOuter: return FD_THUMBFAR;

		case GloveSensor_IndexJointInner: return FD_INDEXNEAR;
		case GloveSensor_IndexJointMiddle: return FD_INDEXFAR;

		case GloveSensor_MiddleJointInner: return FD_MIDDLENEAR;
		case GloveSensor_MiddleJointMiddle: return FD_MIDDLEFAR;

		case GloveSensor_RingJointInner: return FD_RINGNEAR;
		case GloveSensor_RingJointMiddle: return FD_RINGFAR;

		case GloveSensor_PinkyJointInner: return FD_LITTLENEAR;
		case GloveSensor_PinkyJointMiddle: return FD_LITTLEFAR;
	}
	return -1;
}


//====================================================================
// Returns GloveSensor_Invalid if sensor cannot be mapped.
//====================================================================
static eGloveSensor __fastcall MapSensorFromHW(int sensor)
{
	switch(sensor) {
		case FD_THUMBNEAR: return GloveSensor_ThumbJointInner;
		case FD_THUMBFAR: return GloveSensor_ThumbJointOuter;
		case FD_THUMBINDEX: return GloveSensor_AbductionThumbIndex;
		case FD_INDEXNEAR: return GloveSensor_IndexJointInner;
		case FD_INDEXFAR: return GloveSensor_IndexJointMiddle;
		case FD_INDEXMIDDLE: return GloveSensor_AbductionIndexMiddle;
		case FD_MIDDLENEAR: return GloveSensor_MiddleJointInner;
		case FD_MIDDLEFAR: return GloveSensor_MiddleJointMiddle;
		case FD_MIDDLERING: return GloveSensor_AbductionMiddleRing;
		case FD_RINGNEAR: return GloveSensor_RingJointInner;
		case FD_RINGFAR: return GloveSensor_RingJointMiddle;
		case FD_RINGLITTLE: return GloveSensor_AbductionRingPinky;
		case FD_LITTLENEAR: return GloveSensor_PinkyJointInner;
		case FD_LITTLEFAR: return GloveSensor_PinkyJointMiddle;
		case FD_THUMBPALM: return GloveSensor_ThumbRoll;
		case FD_WRISTBEND: return GloveSensor_WristYaw;
		case FD_PITCH: return GloveSensor_WristPitch;
		case FD_ROLL: return GloveSensor_WristRoll;
	}
	return GloveSensor_Invalid;
}


//********************************************************************
}//namespace


//====================================================================
DataGlove::DataGlove(fdGlove* _gloveHandle, uint _iComPort)
{
	// init parent-hid
	SetParentHID(*this);

	GetLogOut() << "Creating DataGlove.\n";

	gloveHandle = _gloveHandle;
	iComPort = _iComPort;

	// collect device information
	iModel = static_cast<EfdGloveTypes>(fdGetGloveType(gloveHandle));
	iHand = static_cast<EfdGloveHand>(fdGetGloveHand(gloveHandle));

	// currently fixed at 18
	iNumSensors = fdGetNumSensors(gloveHandle);

	std::string  sModelType;
	switch(iModel)
	{
		case FD_GLOVE7:
			sModelType = "DataGlove 5";
			break;

		case FD_GLOVE7W:
			sModelType = "DataGlove 5 Wireless";
			break;

		case FD_GLOVE16:
			sModelType = "DataGlove 16";
			break;

		case FD_GLOVE16W:
			sModelType = "DataGlove 16 Wireless";
			break;

		default:
			sModelType = "Unknown";
			break;
	}

	std::string  sModelSubType;
	switch(iHand)
	{
		case FD_HAND_LEFT:
			sModelSubType = "Left handed";
			break;

		case FD_HAND_RIGHT:
			sModelSubType = "Right handed";
			break;

		default:
			sModelSubType = "Unknown";
			break;
	}

	// fdGetGloveInfo() returns meaningless string, so just extract driver-info
	char  szDriverInfo[32];
	fdGetDriverInfo(gloveHandle, reinterpret_cast<ubyte*>(szDriverInfo));

	// fill device info
	DeviceInfo().SetDeviceDescription("Glove");
	DeviceInfo().SetVendor("5DT");
	DeviceInfo().SetModelType(sModelType);
	DeviceInfo().SetModelSubType(sModelSubType);
	DeviceInfo().SetHardwarePort("COM" + ToString(iComPort));
	DeviceInfo().SetAdditionalInfo(szDriverInfo);
	DeviceInfo().SetDirectInputOnly(false);
	DeviceInfo().SetCanUseDirectInput(false);

	//-----------------------------
	// Report hardware capabilities
	//-----------------------------
	ReportHardwareCaps_Measurement(false);

	switch(iModel) {
		case FD_GLOVE7:
		case FD_GLOVE7W: {
			int  availableSensorsHW[] =
			{
				FD_THUMBNEAR,
				FD_INDEXNEAR,
				FD_MIDDLENEAR,
				FD_RINGNEAR,
				FD_LITTLENEAR,
				FD_PITCH,
				FD_ROLL
			};
			ReportAvailableSensors(availableSensorsHW, sizeof(availableSensorsHW) / sizeof(int));
			break;
		}
		case FD_GLOVE16:
		case FD_GLOVE16W: {
			int  availableSensorsHW[] =
			{
				FD_THUMBNEAR,
				FD_THUMBFAR,
				FD_THUMBINDEX,
				FD_INDEXNEAR,
				FD_INDEXFAR,
				FD_INDEXMIDDLE,
				FD_MIDDLENEAR,
				FD_MIDDLEFAR,
				FD_MIDDLERING,
				FD_RINGNEAR,
				FD_RINGFAR,
				FD_RINGLITTLE,
				FD_LITTLENEAR,
				FD_LITTLEFAR
			};
			ReportAvailableSensors(availableSensorsHW, sizeof(availableSensorsHW) / sizeof(int));
			break;
		}
	}

	// reserve operational storage
	rawSensorValues.Allocate(iNumSensors);

	InitSensorMaps();

	// setup base-class flags
	OnUseDirectInput(false);
}


//====================================================================
DataGlove::~DataGlove()
{
	fdClose(gloveHandle);
}


//====================================================================
void  DataGlove::ConnectToAppWindow(TopLevelWindow* pAppWindow,
									HWND hwndAppWindow, bool bFullscreen)
{
	//nothing to do
}


//====================================================================
void  DataGlove::DisconnectFromAppWindow()
{
	//nothing to do
}


//====================================================================
bool  DataGlove::SampleInputData()
{
	PreSampleUpdate();

	// poll raw driver values
	fdGetSensorRawAll(gloveHandle, rawSensorValues);

	// report raw driver values
	{for(uint iSensor = 0; iSensor < GloveSensor_Count; iSensor++) {
		int iHWSensor = sensorMapToHW[iSensor];
		if(iHWSensor >= 0) {
			SetRawValue(eGloveSensor(iSensor), rawSensorValues[iHWSensor]);
		}
		else {
			SetRawValue(eGloveSensor(iSensor), 0);
		}
	}}

	// report flexure values (copy values from finger-middle-joints to finger-outer-joints)
	SetFlexure(GloveSensor_WristYaw, 0.5f);
	SetFlexure(GloveSensor_WristPitch, NormRawValue(GetRawValue(GloveSensor_WristPitch)));
	SetFlexure(GloveSensor_WristRoll, NormRawValue(GetRawValue(GloveSensor_WristRoll)));

	SetFlexure(GloveSensor_AbductionThumbIndex, NormRawValue(GetRawValue(GloveSensor_AbductionThumbIndex)));
	SetFlexure(GloveSensor_AbductionIndexMiddle, NormRawValue(GetRawValue(GloveSensor_AbductionIndexMiddle)));
	SetFlexure(GloveSensor_AbductionMiddleRing, NormRawValue(GetRawValue(GloveSensor_AbductionMiddleRing)));
	SetFlexure(GloveSensor_AbductionRingPinky, NormRawValue(GetRawValue(GloveSensor_AbductionRingPinky)));

	SetFlexure(GloveSensor_ThumbRoll, 0.5f);
	SetFlexure(GloveSensor_ThumbJointInner, NormRawValue(GetRawValue(GloveSensor_ThumbJointInner)));
	SetFlexure(GloveSensor_ThumbJointOuter, NormRawValue(GetRawValue(GloveSensor_ThumbJointOuter)));

	SetFlexure(GloveSensor_IndexJointInner, NormRawValue(GetRawValue(GloveSensor_IndexJointInner)));
	SetFlexure(GloveSensor_IndexJointMiddle, NormRawValue(GetRawValue(GloveSensor_IndexJointMiddle)));
	SetFlexure(GloveSensor_IndexJointOuter, GetFlexure(GloveSensor_IndexJointMiddle));

	SetFlexure(GloveSensor_MiddleJointInner, NormRawValue(GetRawValue(GloveSensor_MiddleJointInner)));
	SetFlexure(GloveSensor_MiddleJointMiddle, NormRawValue(GetRawValue(GloveSensor_MiddleJointMiddle)));
	SetFlexure(GloveSensor_MiddleJointOuter, GetFlexure(GloveSensor_MiddleJointMiddle));

	SetFlexure(GloveSensor_RingJointInner, NormRawValue(GetRawValue(GloveSensor_RingJointInner)));
	SetFlexure(GloveSensor_RingJointMiddle, NormRawValue(GetRawValue(GloveSensor_RingJointMiddle)));
	SetFlexure(GloveSensor_RingJointOuter, GetFlexure(GloveSensor_RingJointMiddle));

	SetFlexure(GloveSensor_PinkyJointInner, NormRawValue(GetRawValue(GloveSensor_PinkyJointInner)));
	SetFlexure(GloveSensor_PinkyJointMiddle, NormRawValue(GetRawValue(GloveSensor_PinkyJointMiddle)));
	SetFlexure(GloveSensor_PinkyJointOuter, GetFlexure(GloveSensor_PinkyJointMiddle));

	// report angles (use standard derivation based on flexure)
	DeriveAnglesFromFlexures();

	PostSampleUpdate();
	return true;
}


//====================================================================
void  DataGlove::OnUseDirectInput(bool bUseDI)
{
	HumanInterfaceDevice::UpdateUsingDirectInput(false);
}


//====================================================================
void  DataGlove::ReportAvailableSensors(const int* pHWSensors, uint count)
{
	for(uint i = 0; i < count; i++) {
		eGloveSensor sensor = MapSensorFromHW(pHWSensors[i]);

		if(sensor != GloveSensor_Invalid) {
			ReportHardwareCaps_Sensor(sensor, true);
		}
	}
}


//====================================================================
void  DataGlove::InitSensorMaps()
{
	sensorMapToHW.Allocate(GloveSensor_Count);
	for(uint i = 0; i < GloveSensor_Count; i++) {
		sensorMapToHW[i] = MapSensorToHW(eGloveSensor(i));
	}
}

