

//--------------------------------------------------------------------
#include "InterSenseOrientationTracker.h"
#include "MathTools\TransformTools.h"
#include "BaseTools\String.h"
#include "Framework\KernelMacros.h"

using namespace vertexware;


//====================================================================
InterSenseOrientationTracker::InterSenseOrientationTracker(
										ISD_TRACKER_HANDLE _trackerHandle)
{
	// init parent-hid
	SetParentHID(*this);

	olog << "Creating InterSenseOrientationTracker.\n";

	trackerHandle = _trackerHandle;
	stationNumber = 0; //always 0 for now
	modifierH = modifierP = modifierR = 1;

	// collect device information
	ISD_GetTrackerConfig(trackerHandle, &trackerInfo, FALSE);

	std::string  sModelType;
	switch(trackerInfo.TrackerType)
	{
		case ISD_PRECISION_SERIES:
			sModelType = "PrecisionSeries";
			break;

		case ISD_INTERTRAX_SERIES:
			sModelType = "InterTraxSeries";
			break;

		default:
			sModelType = "Unknown";
			break;
	}

	std::string  sModelSubType;
	switch(trackerInfo.TrackerModel)
	{
		case ISD_IS300:
			sModelSubType = "IS300";
			break;

		case ISD_IS600:
			sModelSubType = "IS600";
			break;

		case ISD_IS900:
			sModelSubType = "IS900";
			break;

		case ISD_INTERTRAX:
			sModelSubType = "InterTrax";
			break;

		case ISD_INTERTRAX_2:
			sModelSubType = "InterTrax2";
			break;

		case ISD_INTERTRAX_LS:
			sModelSubType = "InterTraxLS";
			break;

		case ISD_INTERTRAX_LC:
			sModelSubType = "InterTraxLC";
			break;

		case ISD_ICUBE2:
			sModelSubType = "InertiaCube2";
			break;

		case ISD_ICUBE2_PRO:
			sModelSubType = "InertiaCube2Pro";
			break;

		case ISD_IS1200:
			sModelSubType = "IS1200";
			break;

		default:
			sModelSubType = "Unknown";
			break;
	}

	std::string  sHWPort;
	switch(trackerInfo.Interface)
	{
		case ISD_INTERFACE_SERIAL:
			sHWPort = "COM";
			sHWPort.append(ToString(trackerInfo.Port));
			break;

		case ISD_INTERFACE_USB:
			sHWPort = "USB";
			break;

		default:
			sHWPort = "Unknown";
			break;
	}

	// fill device info
	DeviceInfo().SetDeviceDescription("Tracker");
	DeviceInfo().SetVendor("InterSense");
	DeviceInfo().SetModelType(sModelType);
	DeviceInfo().SetModelSubType(sModelSubType);
	DeviceInfo().SetHardwarePort(sHWPort);
	DeviceInfo().SetDirectInputOnly(false);
	DeviceInfo().SetCanUseDirectInput(false);

	//-------------------------------------------------------------
	// Report hardware capabilities
	// According to isense.dll version 3.26 docs:
	// 1) only serial-version of InterTrax supports full-reset call
	// 2) all versions support head-reset call
	// 3) only precision-series supports quat sampling-mode
	//-------------------------------------------------------------
	bool bCanResetOrientation =
			(trackerInfo.TrackerType == ISD_INTERTRAX_SERIES)  &&
			(trackerInfo.Interface == ISD_INTERFACE_SERIAL);

	bool bCanResetHead = true;
	bool bCanResetPitch = false;
	bool bCanResetRoll = false;
	bool bCanSampleEuler = true;

	bool bCanSampleQuaternion =
			(trackerInfo.TrackerType == ISD_PRECISION_SERIES);

	ReportHardwareCaps_ResetControls(bCanResetOrientation, bCanResetHead, bCanResetPitch, bCanResetRoll);
	ReportHardwareCaps_SamplingMode(SamplingMode_Euler, bCanSampleEuler);
	ReportHardwareCaps_SamplingMode(SamplingMode_Quaternion, bCanSampleQuaternion);

	SetSamplingMode(SamplingMode_Euler);
	AdjustInternalHPRModifiers();

	// RevisionCheckPoint:
	// Maybe activate HPR-modifier by default if that configuration matches
	// native-hardware-mode. Manual does not say anything about the order
	// which is probably HPR or PHR. For now assume HPR.

	EnableHPRModifier(false);

	// setup base-class flags
	OnUseDirectInput(false);
}


//====================================================================
InterSenseOrientationTracker::~InterSenseOrientationTracker()
{
	ISD_CloseTracker(trackerHandle);
}


//====================================================================
void  InterSenseOrientationTracker::ConnectToAppWindow(TopLevelWindow* _pAppWindow,
										HWND hwndAppWindow, bool bFullscreen)
{
	//nothing to do
}


//====================================================================
void  InterSenseOrientationTracker::DisconnectFromAppWindow()
{
	//nothing to do
}


//====================================================================
bool  InterSenseOrientationTracker::SampleInputData()
{
	PreSampleUpdate();

	bool bSamplingOK = false;
	if(ISD_GetData(trackerHandle, &trackerData)) {

		const float* data = trackerData.Station[stationNumber].Orientation;

		switch(GetSamplingMode()) {

			case SamplingMode_Euler: {
				// reverse some values to match the standard RHS
				SetOrientation(modifierH * data[0], modifierP * data[1], modifierR * data[2]);
				bSamplingOK = true;
				break;
			}

			case SamplingMode_Quaternion: {
				// RevisionCheckPoint:
				// Maybe convert "intersense quat" to rotation-quat.
				// Their "quat" could be direction-axis + roll-angle.
				// Probably not, but this needs to be verified on precision-series.

				SetOrientation(quat(data[0], data[1], data[2], data[3]));
				bSamplingOK = true;
				break;
			}
		}
	}

	if(bSamplingOK) {
		PostSampleUpdate();
	}
	return bSamplingOK;
}


//====================================================================
void  InterSenseOrientationTracker::OnUseDirectInput(bool bUseDI)
{
	HumanInterfaceDevice::UpdateUsingDirectInput(false);
}


//====================================================================
bool  InterSenseOrientationTracker::DoResetOrientation()
{
	// Ignore return-value for library version 3.26
	// ISD_ResetAngles() sometimes returns FALSE,
	// but the call always succeeds for serial-InterTrax ?!
	ISD_ResetAngles(trackerHandle, 0, 0, 0);
	return true;
}


//====================================================================
bool  InterSenseOrientationTracker::DoResetHead()
{
	return ToBool(ISD_ResetHeading(trackerHandle, GetDriverStationNumber()));
}


//====================================================================
bool  InterSenseOrientationTracker::DoResetPitch()
{
	return false;
}


//====================================================================
bool  InterSenseOrientationTracker::DoResetRoll()
{
	return false;
}


//====================================================================
bool  InterSenseOrientationTracker::DoChangeSamplingMode(eSamplingMode mode)
{
	// Intertrax series supports only Euler.
	// Precision series supports Euler & Quaternion.
	if(IsSupportedSamplingMode(SamplingMode_Quaternion)) {

		// get current config
		ISD_STATION_INFO_TYPE  stationInfo;
		ISD_GetStationConfig(trackerHandle, &stationInfo, GetDriverStationNumber(), FALSE);

		// update config
		stationInfo.AngleFormat = ISD_EULER;
		if(mode == SamplingMode_Quaternion) {
			stationInfo.AngleFormat = ISD_QUATERNION;
		}
		ISD_SetStationConfig(trackerHandle, &stationInfo, GetDriverStationNumber(), FALSE);
	}
	return true;
}


//====================================================================
void  InterSenseOrientationTracker::AdjustInternalHPRModifiers()
{
	// defaults: adjusted for InterTraxSeries-InterTrax2-USB
	modifierH = -1;
	modifierP = -1;
	modifierR = +1;

	// RevisionCheckPoint:
	// Add model-specific modifiers if necessary.
}


