

//--------------------------------------------------------------------
#include "GenericGameController.h"
#include "BaseTools\Stream.h"
#include "Framework\KernelMacros.h"

using namespace vertexware;


//====================================================================
#define  DEVICE_DESC_PREFIX  "GameController"


//====================================================================
GenericGameController::GenericGameController(DirectInputProxy8& dip,
							const DirectInputDeviceInstanceInfo8& didii8)
{
	// init parent-hid
	SetParentHID(*this);

	olog << "Creating GenericGameController.\n";

	pDirectInputProxy8 = &dip;
	diDeviceInfo = didii8;
	iDIDevice = 0;
	hwndAppWindow = 0;
	bDIDeviceConnectedToWindow = false;
	bUsingExclusiveMode = false;

	// fill device info
	DeviceInfo().SetDeviceDescription(DEVICE_DESC_PREFIX);
	DeviceInfo().SetDirectInputOnly(true);
	DeviceInfo().SetCanUseDirectInput(true);

	DirectInputDeviceInstanceInfo didii;
	didii8.CopyInfo(didii);
	DeviceInfo().SetDirectInputDeviceInstanceInfo(didii);

	CreateDIDevice();
	CalcAxisModifiers();

	// setup base-class flags
	HumanInterfaceDevice::UpdateUsingDirectInput(true);
}


//====================================================================
GenericGameController::~GenericGameController()
{
	DisconnectFromAppWindow();
	ReleaseDIDevice();
}


//====================================================================
void  GenericGameController::ConnectToAppWindow(TopLevelWindow* _pAppWindow,
									HWND _hwndAppWindow, bool bFullscreen)
{
	hwndAppWindow = _hwndAppWindow;
	bUsingExclusiveMode = false;

	// setup direct-input device behavior
	if(iDIDevice) {
		bDIDeviceConnectedToWindow = false;

		iDIDevice->Unacquire();

		DWORD dwFlags = DISCL_FOREGROUND;
		if(bFullscreen  &&  pDirectInputProxy8->IsAllowedFullscreenExclusiveMode()) {
			dwFlags |= DISCL_EXCLUSIVE;
		}
		else {
			dwFlags |= DISCL_NONEXCLUSIVE;
		}
		olog << "Setting DirectInput cooperative level: " << pDirectInputProxy8->GetCooperativeLevelDesc(dwFlags) << nl;

		HRESULT hr = iDIDevice->SetCooperativeLevel(hwndAppWindow, dwFlags);
		if(SUCCEEDED(hr)) {
			bUsingExclusiveMode = ToBool(dwFlags & DISCL_EXCLUSIVE);
			bDIDeviceConnectedToWindow = true;

			// try acquire
			iDIDevice->Acquire();
		}
		else {
			olog << e_error << "SetCooperativeLevel failed.\n" << e_info;
		}
	}
}


//====================================================================
void  GenericGameController::DisconnectFromAppWindow()
{
	if(iDIDevice) {
		iDIDevice->Unacquire();
	}
	bDIDeviceConnectedToWindow = false;
	hwndAppWindow = 0;
}


//====================================================================
bool  GenericGameController::SampleInputData()
{
	PreSampleUpdate();

	if(bDIDeviceConnectedToWindow) {

		// get direct-input joystick state
		iDIDevice->Poll();
		HRESULT hr = iDIDevice->GetDeviceState(sizeof(diJoystickState), &diJoystickState);

		// handle errors
		if(FAILED(hr)) {
			switch(hr) {
				case DIERR_NOTACQUIRED:
				case DIERR_INPUTLOST:
					// device is not acquired or data stream was lost, try to reacquire
					iDIDevice->Unacquire();
					iDIDevice->Acquire();
					break;

				case E_PENDING:
					return false;
			}
			ResetInputReadings();
			return false;
		}

		// report standard axes states
		{for(uint i = 0; i < Axis_Count; i++) {

			eAxis a = eAxis(i);
			if(! HasAxis(a)) continue;

			float pos = NormalizeAxisValue(*(&diJoystickState.lX + i), axisRangeVals[i]);

			SetAxisPosition(a, ApplyAxisModifiers(pos,
				GetAxisDeadZone(a), GetAxisSaturation(a), axisModsOneOverRange[i]));
		}}

		// report slider states
		{for(uint i = 0; i < GetNumSliders(); i++) {

			float pos = NormalizeAxisValue(diJoystickState.rglSlider[i], sliderRangeVals[i]);

			SetSliderPosition(i, ApplyAxisModifiers(pos,
				GetSliderDeadZone(i), GetSliderSaturation(i), sliderModsOneOverRange[i]));
		}}

		// report pov-hat states
		{for(uint i = 0; i < GetNumPOVHats(); i++) {

			// Position is indicated in hundredths of a degree clockwise from north.
			// Center position is normally reported as 1.
			// Some drivers report the centered position of the POV indicator as 65535

			DWORD  state = diJoystickState.rgdwPOV[i];
			float  positionDeg = -1;

			if(state & 0x0000FFFF != 0x0000FFFF) {
				positionDeg = state * 0.01f;
			}
			SetPOVHatPosition(i, positionDeg);
		}}

		// report button states
		{for(uint i = 0; i < GetNumButtons(); i++) {

			if(diJoystickState.rgbButtons[i] & 0x80) {
				SetButtonPressed(i);
			}
		}}
	}

	// update transitions
	PostSampleUpdate();

	// sampling succeeded
	return true;
}


//====================================================================
void  GenericGameController::OnUseDirectInput(bool bUseDI)
{
	HumanInterfaceDevice::UpdateUsingDirectInput(true);
}


//====================================================================
void  GenericGameController::OnInputModifiersChanged()
{
	CalcAxisModifiers();
}


//====================================================================
void  GenericGameController::CreateDIDevice()
{
	iDIDevice = 0;

	// create direct-input device
	HRESULT hr = pDirectInputProxy8->GetDirectInputObject()->CreateDevice(
									diDeviceInfo.guidInstance, &iDIDevice, 0);

	if(SUCCEEDED(hr)) {
		hr = iDIDevice->SetDataFormat(&pDirectInputProxy8->GetDataFormatJoystick());

		if(SUCCEEDED(hr)) {
			ExtractHardwareProperties();
			SetupHardwareProperties();
			return;
		}
		olog << e_error << "Cannot set DirectInput data-format.\n" << e_info;
	}
	else {
		iDIDevice = 0;
		olog << e_error << "Cannot create DirectInput device.\n" << e_info;
	}
	win_utils::SafeRelease(iDIDevice);
}


//====================================================================
void  GenericGameController::ReleaseDIDevice()
{
	if(iDIDevice) {
		iDIDevice->Unacquire();
	}
	win_utils::SafeRelease(iDIDevice);
}


//====================================================================
void  GenericGameController::ExtractHardwareProperties()
{
	if(! iDIDevice) return;

	// hardware port
	DIPropString dipsHWPort(0, DIPH_DEVICE);
	if(SUCCEEDED(iDIDevice->GetProperty(DIPROP_GETPORTDISPLAYNAME, dipsHWPort))) {

		std::string sHWPort;
		if(UnicodeToAscii(dipsHWPort.wsz, sHWPort)) {
			DeviceInfo().SetHardwarePort(sHWPort);
		}
	}

	// game-controller id
	std::string deviceDesc = DEVICE_DESC_PREFIX;
	DIPropDWord dipdwJoystickID(0, DIPH_DEVICE);
	if(SUCCEEDED(iDIDevice->GetProperty(DIPROP_JOYSTICKID, dipdwJoystickID))) {

		// Game-controller ID-s in ControlPanel->GamingOptions start at 1,
		// but DirectInput returns ID-s starting at 0. Use the convention
		// established by GamingOptions

		uint gameControllerID = dipdwJoystickID.dwData + 1;
		deviceDesc.append(".").append(ToString(gameControllerID));
	}
	DeviceInfo().SetDeviceDescription(deviceDesc);

	// enum available device gadgets
	DeviceObjectsEnumStats  stats;
	HRESULT hr = iDIDevice->EnumObjects(DIEnumDeviceObjectsCallback, &stats, DIDFT_ALL);
	if(hr == DI_OK) {

		// clamp values to max supported by DirectInput joystick data-format
		ClampHigh<uint>(stats.numSliders, 2);
		ClampHigh<uint>(stats.numPOVHats, 4);
		ClampHigh<uint>(stats.numButtons, 32);

		ReportHardwareCaps_Axes(stats.bX, stats.bY, stats.bZ, stats.bRX, stats.bRY, stats.bRZ);
		ReportHardwareCaps_Gadgets(stats.numSliders, stats.numPOVHats, stats.numButtons);
	}
	else {
		olog << e_error << "Cannot obtain hardware capabilites.\n" << e_info;
	}

	// axis properties
	{for(uint i = 0; i < Axis_Count; i++) {
		GetDIPropertyRange(axisRangeVals[i], DIPROP_RANGE, MapAxisToDIOffset(eAxis(i)));
	}}

	{for(uint i = 0; i < GetNumSliders(); i++) {
		GetDIPropertyRange(sliderRangeVals[i], DIPROP_RANGE, DIJOFS_SLIDER(i));
	}}
}


//====================================================================
void  GenericGameController::SetupHardwareProperties()
{
	// DeadZone/Saturation properties are implemented manually;
	// Setup device properties so that DirectInput works with maximum
	// possible dynamic range at each axis.
	// Enable DirecInput internal-calibration.

	{for(uint i = 0; i < Axis_Count; i++) {
		SetDIPropertyValue(0, DIPROP_DEADZONE, MapAxisToDIOffset(eAxis(i)));
		SetDIPropertyValue(10000, DIPROP_SATURATION, MapAxisToDIOffset(eAxis(i)));
	}}

	{for(uint i = 0; i < GetNumSliders(); i++) {
		SetDIPropertyValue(0, DIPROP_DEADZONE, DIJOFS_SLIDER(i));
		SetDIPropertyValue(10000, DIPROP_SATURATION, DIJOFS_SLIDER(i));
	}}

	SetDIPropertyValue(DIPROPCALIBRATIONMODE_COOKED, DIPROP_CALIBRATIONMODE, 0, DIPH_DEVICE);
}


//====================================================================
bool  GenericGameController::GetDIPropertyRange(float* rangeVals, REFGUID rguidProp,
											DWORD dwObj, DWORD dwHow)
{
	rangeVals[RangeVal_Min] = 0;
	rangeVals[RangeVal_Max] = 0;
	rangeVals[RangeVal_TwoOverRange] = 2;

	DIPropRange dipr(dwObj, dwHow);
	if(SUCCEEDED(iDIDevice->GetProperty(rguidProp, dipr))) {

		rangeVals[RangeVal_Min] = dipr.lMin;
		rangeVals[RangeVal_Max] = dipr.lMax;

		float range = dipr.lMax - dipr.lMin;
		if(range) {
			rangeVals[RangeVal_TwoOverRange] = 2.0f / range;
		}
		return true;
	}
	return false;
}


//====================================================================
bool  GenericGameController::GetDIPropertyValue(DWORD& val, REFGUID rguidProp,
											DWORD dwObj, DWORD dwHow)
{
	val = 0;

	DIPropDWord dipdw(dwObj, dwHow);
	if(SUCCEEDED(iDIDevice->GetProperty(rguidProp, dipdw))) {

		val = dipdw.dwData;
		return true;
	}
	return false;
}


//====================================================================
bool  GenericGameController::SetDIPropertyValue(DWORD val, REFGUID rguidProp,
											DWORD dwObj, DWORD dwHow)
{
	DIPropDWord dipdw(dwObj, dwHow);
	dipdw.dwData = val;

	if(SUCCEEDED(iDIDevice->SetProperty(rguidProp, dipdw))) return true;
	return false;
}


//====================================================================
BOOL  CALLBACK  GenericGameController::DIEnumDeviceObjectsCallback(
						const DIDEVICEOBJECTINSTANCE* pdoi, void* pData)
{
	DeviceObjectsEnumStats* stats = static_cast<DeviceObjectsEnumStats*>(pData);

	// collect gadget statistics
	if(pdoi->guidType == GUID_XAxis) {
		stats->bX = true;
	}
	else if(pdoi->guidType == GUID_YAxis) {
		stats->bY = true;
	}
	else if(pdoi->guidType == GUID_ZAxis) {
		stats->bZ = true;
	}
	else if(pdoi->guidType == GUID_RxAxis) {
		stats->bRX = true;
	}
	else if(pdoi->guidType == GUID_RyAxis) {
		stats->bRY = true;
	}
	else if(pdoi->guidType == GUID_RzAxis) {
		stats->bRZ = true;
	}
	else if(pdoi->guidType == GUID_Slider) {
		stats->numSliders++;
	}
	else if(pdoi->guidType == GUID_POV) {
		stats->numPOVHats++;
	}
	else if(pdoi->guidType == GUID_Button) {
		stats->numButtons++;
	}
	return DIENUM_CONTINUE;
}


//====================================================================
GenericGameController::DeviceObjectsEnumStats::DeviceObjectsEnumStats()
{
	bX = bY = bZ = bRX = bRY = bRZ = false;
	numSliders = numPOVHats = numButtons = 0;
}


//====================================================================
DWORD __fastcall GenericGameController::MapAxisToDIOffset(eAxis a)
{
	switch(a) {
		case Axis_X: return DIJOFS_X;
		case Axis_Y: return DIJOFS_Y;
		case Axis_Z: return DIJOFS_Z;
		case Axis_RX: return DIJOFS_RX;
		case Axis_RY: return DIJOFS_RY;
		case Axis_RZ: return DIJOFS_RZ;
	}
	return 0;
}


//====================================================================
void   GenericGameController::CalcAxisModifiers()
{
	{for(uint i = 0; i < Axis_Count; i++) {
		eAxis a = eAxis(i);
		float range = Clamped<float>(GetAxisSaturation(a) - GetAxisDeadZone(a), 0, 1);
		axisModsOneOverRange[i] = range ? (1 / range) : 0;
	}}

	{for(uint i = 0; i < 2; i++) {
		float range = Clamped<float>(GetSliderSaturation(i) - GetSliderDeadZone(i), 0, 1);
		sliderModsOneOverRange[i] = range ? (1 / range) : 0;
	}}
}


