

//--------------------------------------------------------------------
#include "BaseTools\CommandLineParser.h"
#include "BaseTools\FileSystem.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"

#include "vw_common_autolink.h"
using namespace vertexware;

#include <iostream>
using namespace std;


//--------------------------------------------------------------------
class  CopyDirectoryWalker : public file_system::FileSystemWalker
{
private:
	std::ostream& log;
	std::string   srcDir;
	std::string   dstDir;
	bool          bOverwrite;

	uint          numWarnings;
	uint          numErrors;


protected:
	virtual void OnBeginWalk(const std::string& pathName)
	{
		numWarnings = 0;
		numErrors = 0;
	}

	virtual void OnEndWalk()
	{
		log << separator_line();
		log << numWarnings << " warnings." << endl;
		log << numErrors << " errors." << endl;
	}

	virtual void OnFile(uint depth, uint fileAttributes,
						const std::string& fileName,
						const std::string& fileNameWithPath)
	{
		std::string  dstFileNameWithPath(dstDir);
		dstFileNameWithPath.append(fileNameWithPath, srcDir.size(), std::string::npos);

		if(file_system::FileExists(dstFileNameWithPath)  &&  ! bOverwrite) {
			log << "Warning: skipping overwrite of existing file '" << dstFileNameWithPath << "'.'" << endl;
			numWarnings ++;
			return;
		}

		cout << "copying '" << fileNameWithPath << "'." << endl;

		if(! file_system::CopyFile(fileNameWithPath, dstFileNameWithPath, true)) {
			log << "Error: failed to copy file '" << fileNameWithPath << "'." << endl;
			numErrors ++;
		}
	}

	virtual bool OnDirectory(uint depth, uint fileAttributes,
								const std::string& directoryName,
								const std::string& directoryNameWithPath)
	{
		std::string  dstDirectoryNameWithPath(dstDir);
		dstDirectoryNameWithPath.append(directoryNameWithPath, srcDir.size(), std::string::npos);

		if(! file_system::DirectoryExists(dstDirectoryNameWithPath)) {
			if(! file_system::CreateDirectory(dstDirectoryNameWithPath)) {

				log << "Error: failed to create directory '";
				log << dstDirectoryNameWithPath << "'" << endl;
				numErrors ++;

				CancelWalk();
				log << "Cannot continue. Aborting." << endl;
			}
		}
		return true; //always descend
	}


public:
	CopyDirectoryWalker(std::ostream& _log,
						const std::string& _srcDir, const std::string& _dstDir,
						bool _bOverwrite) :
		log(_log), srcDir(_srcDir), dstDir(_dstDir), bOverwrite(_bOverwrite)
	{
	}
};


//--------------------------------------------------------------------
int main(int argc, char* argv[])
{
	try {
		CommandLineParser  clp;
		clp.Parse(argc, argv);

		// construct log-file
		std::string logFilename;
		clp.GetAssignment("log", logFilename);

		if(logFilename.empty()) {
			logFilename = "recursive-copy.log";
		}
		LogFileStream logFile(logFilename, logFilename);

		// construct log object
		ControlledOutputStream  log;
		log.AddOutputTarget(logFile);
		log.AddOutputTarget(cout);

		// print usage
		log << endl;
		log << separator_line() << "recursive-copy v1.0 by petar.basic.2002\n";
		log << "arguments: -src='source directory'\n";
		log << "           -dst='destination directory'\n";
		log << "           -log='log file'\n";
		log << "           -overwrite\n";
		log << separator_line();

		// extract source dir
		std::string  srcDir;
		if(! clp.GetAssignment("src", srcDir)) {
			log << "ERROR: missing argument <src>." << endl;
			return 0;
		}
		file_system::ForceFullPathName(srcDir, true);

		if(! file_system::DirectoryExists(srcDir)) {
			log << "ERROR: source directory '" << srcDir << "' does not exist." << endl;
			return 0;
		}

		// extract destination dir
		std::string  dstDir;
		if(! clp.GetAssignment("dst", dstDir)) {
			log << "ERROR: missing argument <dst>." << endl;
			return 0;
		}
		file_system::ForceFullPathName(dstDir, true);

		if(! file_system::DirectoryExists(dstDir)) {
			if(! file_system::CreateDirectory(dstDir)) {
				log << "ERROR: cannot create destination directory '" << dstDir << "'." << endl;
				return 0;
			}
			else {
				log << "created destination directory '" << dstDir << "'." << endl;
			}
		}

		// extract overwrite flag
		bool bOverwrite = clp.GetOption("overwrite");

		// echo arguments
		log << "echoing arguments..." << endl;
		log << "source directory: " << srcDir << endl;
		log << "destination directory: " << dstDir << endl;
		log << "log file: " << logFilename << endl;
		log << "overwrite: " << ToString(bOverwrite) << endl;
		log << separator_line();

		// do copy
		CopyDirectoryWalker cdw(log, srcDir, dstDir, bOverwrite);
		if(! cdw.Walk(srcDir)) {
			log << "ERROR: cannot open source directory." << endl;
		}
	}
	catch(CommandLineErrorException& e) {
		cout << "ERROR: " << e.GetMessage() << endl;
	}
	catch(Exception& e) {
		cout << "ERROR: " << e.MakeDescription() << endl;
	}
	return 0;
}

