

//--------------------------------------------------------------------
#include "BaseTools\CommandLineParser.h"
#include "BaseTools\FileSystem.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"
#include "BaseTools\WinUtils.h"
#include "DataObjects\SignalIO.h"
#include "DataObjects\SignalCodecGdiplus.h"
#include "DataObjects\SignalCodecVWDP.h"
#include "DataObjects\ImageOperations.h"

#include "vw_common_autolink.h"
using namespace vertexware;

#include <iostream>
using namespace std;


//--------------------------------------------------------------------
int main(int argc, char* argv[])
{
	try {
		CommandLineParser  clp;
		clp.Parse(argc, argv);

		// print usage
		cout << endl;
		cout << separator_line() << "resource-converter v1.0 by petar.basic.2002\n";
		cout << "arguments: -src='source file'\n";
		cout << "           -dst='destination file'\n";
		cout << "           -overwrite\n";
		cout << "           -format='destination sample format'\n";
		cout << "           -origin-bottom\n";
		cout << "           -big-endian\n";
		cout << separator_line();

		// extract flags
		bool bOverwrite = clp.GetOption("overwrite");
		bool bOriginBottom = clp.GetOption("origin-bottom");
		bool bBigEndian = clp.GetOption("big-endian");

		eSampleFormat sampleFormat = SampleFormat_Invalid;
		std::string sSampleFormat;
		if(clp.GetAssignment("format", sSampleFormat)) {
			sampleFormat = eSampleFormat(sample_formats::MapNameToSampleFormat(sSampleFormat));
			if(sampleFormat == SampleFormat_Invalid) {
				cout << "ERROR: invalid sample format specified." << endl;
				return 0;
			}
		}

		// extract source file
		std::string  srcFile, srcFileName;
		if(! clp.GetAssignment("src", srcFileName)) {
			cout << "ERROR: missing argument <src>." << endl;
			return 0;
		}
		srcFile = srcFileName;
		file_system::ForceFullPathName(srcFile, false);

		// extract destination file
		std::string  dstFile;
		if(! clp.GetAssignment("dst", dstFile)) {
			cout << "ERROR: missing argument <dst>." << endl;
			return 0;
		}
		file_system::ForceFullPathName(dstFile, false);

		// echo arguments
		cout << "echoing arguments..." << endl;
		cout << "source file: " << srcFile << endl;
		cout << "destination file: " << dstFile << endl;
		cout << "overwrite: " << ToString(bOverwrite) << endl;
		cout << "format: " << sample_formats::MapSampleFormatToName(sampleFormat) << endl;
		cout << "origin-bottom: " << ToString(bOriginBottom) << endl;
		cout << "big-endian: " << ToString(bBigEndian) << endl;
		cout << separator_line();

		if(! file_system::FileExists(srcFile)) {
			cout << "ERROR: source file does not exist." << endl;
			return 0;
		}
		if(file_system::FileExists(dstFile)  &&  ! bOverwrite) {
			cout << "Destination file exists. Overwrite flag is not specified. Exiting." << endl;
			return 0;
		}

		// create signal-operations-manager
		SignalOperationsManager  som;

		// create data-codec-manager
		using namespace data_codec_parameters;
		using namespace signal_codec_parameters;

		DataCodecManager  dcm;

		dcm.RegisterDataCategory(DataCategory_Image);
		dcm.RegisterDecoder(new ImageDecoderGdiplus);
		dcm.RegisterDecoder(new ImageDecoderVWDP);
		dcm.RegisterEncoder(new ImageEncoderVWDP);

		NamedParameterList saveNPL;
		saveNPL.Insert(NamedParameter(Param_OriginBottom, bOriginBottom));
		saveNPL.Insert(NamedParameter(Param_ByteOrderBigEndian, bBigEndian));
		saveNPL.Insert(NamedParameter(Param_FileOverwrite, bOverwrite));

		dcm.SetEncoderDefaultParamList(DataCategory_Image, saveNPL);
		dcm.SetDecoderDefaultParamList(DataCategory_Image, SignalCodecParamFactory());

		// image resource ?
		AutoPtr<ImageBase> apImg(LoadImage(srcFile));
		if(apImg) {
			cout << "recognized source-file as image." << endl;

			AutoPtr<ImageBase> apConvImg(apImg, false);
			if(sampleFormat != SampleFormat_Invalid) {
				apConvImg.Assign(ManufactureImage(sampleFormat));
				if(! apConvImg) {
					cout << "ERROR: failed to create destination-format image." << endl;
					return 0;
				}
				if(! TransferSamples2D(*apConvImg, *apImg, true)) {
					cout << "ERROR: failed to convert image to destination-format." << endl;
					return 0;
				}
			}

			// save converted image
			cout << "writing destination-file." << endl;
			SaveImage(*apConvImg, dstFile, signal_codec_parameters::Value_ImageFileFormat_VWDP);
			return 0;
		}
	}
	catch(CommandLineErrorException& e) {
		cout << "ERROR: " << e.GetMessage() << endl;
	}
	catch(Exception& e) {
		cout << "ERROR: " << e.MakeDescription() << endl;
	}
	return 0;
}

