

//--------------------------------------------------------------------
#include "private\vs_direct3d8.h"
#include "PluginParams\PluginParams_vs_direct3d8.h"
#include "BaseTools\WinUtils.h"
#include "BaseTools\Memory.h"
#include "BaseTools\Stream.h"
#include "BaseTools\CompilerConfig.h"
#include "DataObjects\ImageOperations.h"
#include "Framework\KernelMacros.h"
#include "D3D8Tools\ErrorHandling.h"
#include "D3D8Tools\Surfaces.h"
#include <sstream>

using namespace vertexware;


//********************************************************************
namespace {


//====================================================================
void __fastcall ExtractFrameBufferConfigurations(IDirect3D8* iObject,
										VideoProfile& outVideoProfile,
										bool bFullscreen, bool bHardwareRasterizer)
{
	// prepare constant properties
	FrameBufferConfig fbc;

	fbc.fullscreen = bFullscreen;
	fbc.hardwareRasterizer = bHardwareRasterizer;

	D3DDEVTYPE devType = (bHardwareRasterizer ? D3DDEVTYPE_HAL : D3DDEVTYPE_REF);

	// ignore legacy display-formats & legacy back-buffer-formats
	fbc.bitsRed   = 8;
	fbc.bitsGreen = 8;
	fbc.bitsBlue  = 8;
	fbc.bitsAlpha = 8;

	D3DFORMAT displayFormat    = D3DFMT_X8R8G8B8;
	D3DFORMAT backBufferFormat = D3DFMT_A8R8G8B8;

	// accumulation-buffer not supported by d3d
	fbc.bitsAccumRed   = 0;
	fbc.bitsAccumGreen = 0;
	fbc.bitsAccumBlue  = 0;
	fbc.bitsAccumAlpha = 0;

	// check device-type support
	if(iObject->CheckDeviceType(D3DADAPTER_DEFAULT, devType, displayFormat,
								backBufferFormat, ! fbc.fullscreen) != D3D_OK) return;

	// create temporary list
	std::vector<FrameBufferConfig>  fbcVec;

	// extract available depth-stencil formats
	D3DFORMAT dsFormat;
	for(uint iDSMode = 1; iDSMode <= 4; iDSMode++) {
		switch(iDSMode) {
			case 1:
				dsFormat = D3DFMT_D16;
				fbc.bitsDepth   = 16;
				fbc.bitsStencil = 0;
				break;
			case 2:
				dsFormat = D3DFMT_D24X8;
				fbc.bitsDepth   = 24;
				fbc.bitsStencil = 0;
				break;
			case 3:
				dsFormat = D3DFMT_D24S8;
				fbc.bitsDepth   = 24;
				fbc.bitsStencil = 8;
				break;
			case 4:
				dsFormat = D3DFMT_D32;
				fbc.bitsDepth   = 32;
				fbc.bitsStencil = 0;
				break;
		}

		// check depth-stencil support / match
		if(iObject->CheckDeviceFormat(D3DADAPTER_DEFAULT, devType, displayFormat,
									D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE,
									dsFormat) != D3D_OK) continue;

		if(iObject->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, devType, displayFormat,
									backBufferFormat, dsFormat) != D3D_OK) continue;

		// put configuration into temporary list
		fbcVec.push_back(fbc);
	}

	// add configurations for different back-buffer-counts
	const uint maxBackBuffers = 2;
	for(uint iBBC = 1; iBBC <= maxBackBuffers; iBBC++) {

		std::vector<FrameBufferConfig>::iterator it = fbcVec.begin();
		while(it != fbcVec.end()) {
			FrameBufferConfig& fbc = *it++;

			fbc.backBufferCount = iBBC;
			outVideoProfile.Add(fbc);
		}
	}
}


//====================================================================
bool __fastcall ExtractDepthStencilFormat(const FrameBufferConfig& fbc, D3DFORMAT& dsFormat)
{
	if(fbc.bitsDepth == 16  &&  fbc.bitsStencil == 0) {
		dsFormat = D3DFMT_D16;
		return true;
	}
	if(fbc.bitsDepth == 24  &&  fbc.bitsStencil == 0) {
		dsFormat = D3DFMT_D24X8;
		return true;
	}
	if(fbc.bitsDepth == 24  &&  fbc.bitsStencil == 8) {
		dsFormat = D3DFMT_D24S8;
		return true;
	}
	if(fbc.bitsDepth == 32  &&  fbc.bitsStencil == 0) {
		dsFormat = D3DFMT_D32;
		return true;
	}
	return false;
}


//====================================================================
void __fastcall SetupSwapParameters(IDirect3D8* iObject, D3DDEVTYPE devType,
							D3DPRESENT_PARAMETERS& pp, const VideoHints& vh)
{
	bool bSupportsReadScanline = false;
	bool bSupportsDiscardWithFSDestinationAlpha = false;

	D3DCAPS8 devCaps;
	if(iObject->GetDeviceCaps(D3DADAPTER_DEFAULT, devType, &devCaps) == D3D_OK) {
		bSupportsReadScanline = ToBool(devCaps.Caps & D3DCAPS_READ_SCANLINE);
		bSupportsDiscardWithFSDestinationAlpha = ToBool(devCaps.Caps3 &
													D3DCAPS3_ALPHA_FULLSCREEN_FLIP_OR_DISCARD);
	}

	// Update SwapEffect and PresentationInterval.
	// Force single back-buffer if copy-semantics swap-effect is used.
	if(pp.Windowed) {

		if(vh.vSync  &&  bSupportsReadScanline) {
			pp.SwapEffect = D3DSWAPEFFECT_COPY_VSYNC;
			pp.BackBufferCount = 1;
		}
		else {
			pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		}
		pp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
	}
	else {
		if(bSupportsDiscardWithFSDestinationAlpha || (! vh.fullscreenDestinationAlpha)) {

			// most efficient if available or allowed
			pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		}
		else {
			pp.SwapEffect = vh.vSync ? D3DSWAPEFFECT_COPY_VSYNC : D3DSWAPEFFECT_COPY;
			pp.BackBufferCount = 1;
		}
		pp.FullScreen_PresentationInterval = vh.vSync ?
									D3DPRESENT_INTERVAL_ONE : D3DPRESENT_INTERVAL_IMMEDIATE;
	}
}


//====================================================================
std::string __fastcall GetSwapEffectString(D3DSWAPEFFECT swapEffect)
{
	switch(swapEffect) {
		case D3DSWAPEFFECT_DISCARD: return "discard";
		case D3DSWAPEFFECT_FLIP: return "flip";
		case D3DSWAPEFFECT_COPY: return "copy";
		case D3DSWAPEFFECT_COPY_VSYNC: return "copy-vsync";
	}
	return "unknown";
}


//====================================================================
std::string __fastcall GetPresentationIntervalString(UINT presentInterval)
{
	switch(presentInterval) {
		case D3DPRESENT_INTERVAL_DEFAULT: return "default";
		case D3DPRESENT_INTERVAL_ONE: return "one";
		case D3DPRESENT_INTERVAL_TWO: return "two";
		case D3DPRESENT_INTERVAL_THREE: return "three";
		case D3DPRESENT_INTERVAL_FOUR: return "four";
		case D3DPRESENT_INTERVAL_IMMEDIATE: return "immediate";
	}
	return "unknown";
}


//====================================================================
void __fastcall DumpSwapParameters(const D3DPRESENT_PARAMETERS& pp,
									std::ostream& os, uint tabLevel = 0)
{
	os << tab(tabLevel) << "SwapEffect: " << GetSwapEffectString(pp.SwapEffect) << nl;;
	os << tab(tabLevel) << "FullScreen_PresentationInterval: " << GetPresentationIntervalString(pp.FullScreen_PresentationInterval) << std::endl;
}


//====================================================================
bool __fastcall TranslateConfiguration(IDirect3D8* iObject,
							D3DDEVTYPE& devType, D3DPRESENT_PARAMETERS& pp,
							const VideoConfig& vc)
{
	// extract device-type
	devType = (vc.frameBufferConfig.hardwareRasterizer ? D3DDEVTYPE_HAL : D3DDEVTYPE_REF);

	// ignore legacy display-modes
	if(vc.displayModeConfig.bitsPerPixel != 32) return false;

	// ignore legacy back-buffer-formats
	if((vc.frameBufferConfig.bitsRed != 8) ||
		(vc.frameBufferConfig.bitsGreen != 8) ||
		(vc.frameBufferConfig.bitsBlue != 8) ||
		(vc.frameBufferConfig.bitsAlpha != 8))
	{
		return false;
	}

	// accumulation-buffer not supported by d3d
	if(vc.frameBufferConfig.bitsAccumRed ||
		vc.frameBufferConfig.bitsAccumGreen ||
		vc.frameBufferConfig.bitsAccumBlue ||
		vc.frameBufferConfig.bitsAccumAlpha)
	{
		return false;
	}

	pp.BackBufferWidth = vc.displayModeConfig.width;
	pp.BackBufferHeight = vc.displayModeConfig.height;
	pp.BackBufferFormat = D3DFMT_A8R8G8B8;
	pp.BackBufferCount = vc.frameBufferConfig.backBufferCount;
	pp.MultiSampleType = D3DMULTISAMPLE_NONE;
	pp.hDeviceWindow = 0;
	pp.Windowed = (vc.frameBufferConfig.fullscreen ? FALSE : TRUE);
	pp.EnableAutoDepthStencil = TRUE;
	pp.Flags = 0;

	// depth-stencil parameters
	if(! ExtractDepthStencilFormat(vc.frameBufferConfig, pp.AutoDepthStencilFormat)) {
		return false;
	}

	// refresh parameters
	if(pp.Windowed) {
		pp.FullScreen_RefreshRateInHz = 0;
	}
	else {
		if(vc.displayModeConfig.refreshFrequency != 0) {
			pp.FullScreen_RefreshRateInHz = vc.displayModeConfig.refreshFrequency;
		}
		else {
			pp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
		}
	}
	SetupSwapParameters(iObject, devType, pp, vc.videoHints);

	// success
	return true;
}


//====================================================================
DWORD __fastcall TranslateVertexProcessingMode(VideoHints::eVertexProcessingMode vpm)
{
	switch(vpm) {
		case VideoHints::VertexProcessingMode_Software:
			return D3DCREATE_SOFTWARE_VERTEXPROCESSING;

		case VideoHints::VertexProcessingMode_Mixed:
			return D3DCREATE_MIXED_VERTEXPROCESSING;

		case VideoHints::VertexProcessingMode_Hardware:
			return D3DCREATE_HARDWARE_VERTEXPROCESSING;
	}
	return 0;
}


//====================================================================
VideoHints::eVertexProcessingMode __fastcall TranslateVertexProcessingFlag(DWORD vpf)
{
	DWORD vpfMask = D3DCREATE_SOFTWARE_VERTEXPROCESSING |
					D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_MIXED_VERTEXPROCESSING;

	switch(vpf & vpfMask) {
		case D3DCREATE_SOFTWARE_VERTEXPROCESSING:
			return VideoHints::VertexProcessingMode_Software;

		case D3DCREATE_HARDWARE_VERTEXPROCESSING:
			return VideoHints::VertexProcessingMode_Hardware;

		case D3DCREATE_MIXED_VERTEXPROCESSING:
			return VideoHints::VertexProcessingMode_Mixed;
	}
	return VideoHints::VertexProcessingMode_Default;
}


//====================================================================
bool __fastcall GetNextVertexProcessingFlag(DWORD& flag)
{
	DWORD vpfMask = D3DCREATE_SOFTWARE_VERTEXPROCESSING |
					D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_MIXED_VERTEXPROCESSING;

	// fallback semantics
	switch(flag & vpfMask) {
		case D3DCREATE_HARDWARE_VERTEXPROCESSING:
			flag = D3DCREATE_MIXED_VERTEXPROCESSING;
			return true;

		case D3DCREATE_MIXED_VERTEXPROCESSING:
			flag = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
			return true;
	}
	return false;
}


//********************************************************************
}//namespace


//====================================================================
bool  vs_direct3d8::SetPluginParameter(const NamedParameter& np)
{
	if(np.GetName() == VW_PLUGIN_PARAM_NAME_VS_DIRECT3D8_DestroyWindowBeforeD3DObjectRelease) {
		if(np.GetValueType() == NamedParameter::ValueType_Bool) {
			np.GetValue(bDestroyWindowBeforeD3DObjectRelease);
			return true;
		}
	}
	return false;
}


//====================================================================
bool  vs_direct3d8::GetPluginParameter(const std::string& parameterName, NamedParameter& out) const
{
	if(parameterName == VW_PLUGIN_PARAM_NAME_VS_DIRECT3D8_DestroyWindowBeforeD3DObjectRelease) {
		out.SetValue(bDestroyWindowBeforeD3DObjectRelease);
		return true;
	}
	return false;
}


//====================================================================
vs_direct3d8::vs_direct3d8()
{
	pWindow = 0;
	iObject = 0;
	iDevice = 0;

	bDestroyWindowBeforeD3DObjectRelease = true;
}


//====================================================================
vs_direct3d8::~vs_direct3d8()
{
	CleanupWindow();
}


//====================================================================
std::string  vs_direct3d8::GetVideoSubsystemInterfaceName()
{
	return VW_INTERFACE_NAME_VideoSubsystemDirect3D8;
}


//====================================================================
bool  vs_direct3d8::DoBuildVideoProfile(std::ostream& log,
										VideoProfile& outVideoProfile)
{
	//-----------
	// create d3d
	bool bReleaseD3DObject = (iObject == 0);
	if(! CreateD3DObject()) {
		log << "Cannot create Direct3D8 object." << nl;
		return false;
	}
	bool bOK = true;

	//--------------------
	// obtain general-info
	D3DADAPTER_IDENTIFIER8 adapterID;
	if(iObject->GetAdapterIdentifier(D3DADAPTER_DEFAULT, D3DENUM_NO_WHQL_LEVEL,
															&adapterID) == D3D_OK)
	{
		// extract general-info
		std::ostringstream oss;
		oss << HIWORD(adapterID.DriverVersion.HighPart) << '.';
		oss << LOWORD(adapterID.DriverVersion.HighPart) << '.';
		oss << HIWORD(adapterID.DriverVersion.LowPart) << '.';
		oss << LOWORD(adapterID.DriverVersion.LowPart);

		win_utils::Guid guid = adapterID.DeviceIdentifier;
		outVideoProfile.SetDriverInfo(adapterID.Description, oss.str(), guid);
	}
	else {
		bOK = false;
		log << "Cannot access Direct3D8 adapter-info." << nl;
	}

	//----------------------
	// collect display-modes
	uint  numModes = iObject->GetAdapterModeCount(D3DADAPTER_DEFAULT);
	for(uint i = 0; i < numModes; i++) {

		D3DDISPLAYMODE dm;
		if(iObject->EnumAdapterModes(D3DADAPTER_DEFAULT, i, &dm) != D3D_OK) continue;

		if(dm.Format != D3DFMT_X8R8G8B8) continue;
		uint bitsPerPixel = 32;

		DisplayModeConfig dmc(dm.Width, dm.Height, bitsPerPixel, dm.RefreshRate);

		// save safe-modes
		if(DisplayModeConfigFilter::IsSafeDisplayModeConfig(dmc)) {
			outVideoProfile.Add(dmc);
		}
	}

	//-----------------------------
	// collect frame-buffer-configs
	ExtractFrameBufferConfigurations(iObject, outVideoProfile, true, true);
	ExtractFrameBufferConfigurations(iObject, outVideoProfile, true, false);
	ExtractFrameBufferConfigurations(iObject, outVideoProfile, false, true);
	ExtractFrameBufferConfigurations(iObject, outVideoProfile, false, false);

	//------------
	// release d3d
	if(bReleaseD3DObject) {
		win_utils::SafeRelease(iObject);
	}

	// report result
	return bOK;
}


//====================================================================
VideoSubsystemWindow*  vs_direct3d8::InitializeWindow(const VideoConfig& _videoConfig,
													bool _bAutoShowWindow)
{
	// check arguments
	if(pWindow) {
		VW_THROW(VideoSubsystemException, "Window already initialized.");
	}

	videoConfig = _videoConfig;

	// fullscreen mode
	if(videoConfig.frameBufferConfig.fullscreen) {

		videoConfig.resizable = false;

		// create and setup window
		pWindow = new Window(*this, true);
		pWindow->CreateSystemWindow(0, WS_POPUP, WS_EX_APPWINDOW | WS_EX_TOPMOST);

		// set exact size because CreateDevice() will not send WM_SIZE
		pWindow->CenterOnDesktop(videoConfig.displayModeConfig.width,
								videoConfig.displayModeConfig.height);
	}

	// windowed mode
	else {
		videoConfig.displayModeConfig.refreshFrequency = 0;

		// adjust size and style
		DWORD style = Window::GetStandardWindowStyle(videoConfig.resizable);
		DWORD styleEx = WS_EX_APPWINDOW;

		uint  totalWindowWidth, totalWindowHeight;
		Window::AdjustWindowSize(style, styleEx,
								videoConfig.displayModeConfig.width,
								videoConfig.displayModeConfig.height,
								totalWindowWidth, totalWindowHeight, true);

		// create and setup window
		pWindow = new Window(*this, false);
		pWindow->CreateSystemWindow(0, style, styleEx);
		pWindow->CenterOnDesktop(totalWindowWidth, totalWindowHeight);
	}

	// additional window setup
	pWindow->EnableSystemCommandEmulation(false);

	// setup rendering device
	try {
		CreateD3DDevice();

		// check realized refresh-frequency
		if(videoConfig.frameBufferConfig.fullscreen) {
			D3DDISPLAYMODE dm;
			if(iDevice->GetDisplayMode(&dm) == D3D_OK) {
				videoConfig.displayModeConfig.refreshFrequency = dm.RefreshRate;
			}
		}
	}
	catch(VideoSubsystemRecoverableException&) {
		CleanupWindow();
		throw;
	}

	// ready to render
	SetOperationalState(OperationalState_Ready);

	if(_bAutoShowWindow) {
		pWindow->Show();
	}
	StabilizeColorBuffers();

	return pWindow;
}


//====================================================================
VideoSubsystemWindow*  vs_direct3d8::GetWindow()
{
	return pWindow;
}


//====================================================================
void  vs_direct3d8::CleanupWindow()
{
	if(! pWindow) return;

	// release d3d-device-interface
	olog << "Releasing Direct3D8Device." << nl;
	win_utils::SafeRelease(iDevice);

	while(pWindow  ||  iObject) {

		// destroy rendering window
		if(pWindow  &&  (bDestroyWindowBeforeD3DObjectRelease  ||  ! iObject)) {
			olog << "Destroying main window." << nl;
			SafeDelete(pWindow);
		}

		// release d3d-object-interface
		if(iObject) {
			olog << "Releasing Direct3D8." << nl;
			win_utils::SafeRelease(iObject);
		}
	}

	// update state
	olog << "Window-cleanup finished." << nl;
	SetOperationalState(OperationalState_WaitingInit);
}


//====================================================================
void  vs_direct3d8::PresentBackBuffer()
{
	if(iDevice->Present(0, 0, 0, 0) != D3D_OK) {
		SetOperationalState(OperationalState_WaitingReset);
	}
}


//====================================================================
void  vs_direct3d8::ErrorCheck()
{
	if(iDevice->TestCooperativeLevel() != D3D_OK) {
		SetOperationalState(OperationalState_WaitingReset);
	}
}


//====================================================================
void  vs_direct3d8::Reset()
{
	SetOperationalState(OperationalState_WaitingReset);

	// is device ready for reset ?
	HRESULT hrTest = iDevice->TestCooperativeLevel();
	switch(hrTest)
	{
		case D3DERR_DEVICELOST:
			// cannot reset yet, try later
			break;

		case D3D_OK:
		case D3DERR_DEVICENOTRESET: {

			// try reseting
			HRESULT hrReset = iDevice->Reset(&presentationParameters);

			if(hrReset == D3D_OK) {
				// ready to continue
				SetOperationalState(OperationalState_Ready);
			}
			else {
				// cannot recover
				d3d8_tools::D3DErrorCheckET<VideoSubsystemException>(hrReset, "IDirect3DDevice8::Reset");
			}
			break;
		}

		default:
			// unexpected error code
			d3d8_tools::D3DErrorCheckET<VideoSubsystemException>(hrTest, "IDirect3DDevice8::Reset");
			break;
	}
}


//====================================================================
void  vs_direct3d8::UpdateVSync(bool bEnable)
{
	if(videoConfig.videoHints.vSync != bEnable) {
		videoConfig.videoHints.vSync = bEnable;

		SetOperationalState(OperationalState_WaitingReset);
		SetupSwapParameters(iObject, deviceType, presentationParameters, videoConfig.videoHints);

		olog << "Presentation parameters:" << nl;
		DumpSwapParameters(presentationParameters, olog, 1);
	}
}


//====================================================================
bool  vs_direct3d8::CaptureFrameBuffer(AutoPtr<ImageBase>& out,
										uint frameBufferFlags,
										bool bForceColorSampleFormat)
{
	out.SafeDelete();
	if(! iDevice) return false;

	IDirect3DDevice8& dev = *iDevice;

	// process stereo flags
	switch(frameBufferFlags  &  FrameBufferFlags_StereoMask) {

		case FrameBufferFlags_Mono: {
			break;
		}
		default: {
			// stereo not implemented
			return false;
		}
	}

	// process swap-chain flags
	bool bReadFrontBuffer = false;

	switch(frameBufferFlags  &  FrameBufferFlags_SwapChainMask) {

		case FrameBufferFlags_Back: {
			bReadFrontBuffer = false;
			break;
		}
		case FrameBufferFlags_Front: {
			bReadFrontBuffer = true;
			break;
		}
		default: {
			return false;
		}
	}

	uint width = pWindow->GetClientAreaWidth();
	uint height = pWindow->GetClientAreaHeight();

	// process component flags
	eSampleFormat  outSampleFormat = SampleFormat_Invalid;

	switch(frameBufferFlags  &  FrameBufferFlags_ComponentMask) {

		case FrameBufferFlags_Color: {
			if(bReadFrontBuffer) {
				// copy pixels to temporary surface
				win_utils::COMPtr<IDirect3DSurface8> apTempSurf;

				// must use screen-size
				D3DDISPLAYMODE  displayMode;
				if(dev.GetDisplayMode(&displayMode) != D3D_OK) return false;

				if(dev.CreateImageSurface(displayMode.Width, displayMode.Height, D3DFMT_A8R8G8B8, &apTempSurf) != D3D_OK) return false;
				if(dev.GetFrontBuffer(apTempSurf.Get()) != D3D_OK) return false;

				// copy pixels to output-image
				outSampleFormat = SampleFormat_A8R8G8B8;
				out.Assign(ManufactureImage(outSampleFormat, displayMode.Height, displayMode.Width));
				if(! d3d8_tools::CopySurfaceToImage(*apTempSurf, *out)) {
					out.SafeDelete();
					return false;
				}

				// when running in windowed mode D3D captures the whole screen;
				// crop image to frame-buffer-area
				if(displayMode.Height != height  ||  displayMode.Width != width) {

					RECT rect;
					GetWindow()->GetVisibleClientAreaScreenCoords(rect);

					AutoPtr<ImageBase> apCroppedImage(ManufactureImage(outSampleFormat, height, width));
					Fill_ARGB(*apCroppedImage, color4b::GRAY);

					int  visibleWidth = Clamped<int>(rect.right - rect.left, 0, width);
					int  visibleHeight = Clamped<int>(rect.bottom - rect.top, 0, height);

					int  dstOffsetX = 0;
					int  dstOffsetY = 0;
					int  srcOffsetX = Clamped<int>(rect.left, 0, displayMode.Width);
					int  srcOffsetY = Clamped<int>(rect.top, 0, displayMode.Height);

					if(visibleWidth != width) {
						if(rect.left == 0) {
							// left screen-side
							dstOffsetX = Clamped<int>(width - visibleWidth, 0, width);
						}
					}
					if(visibleHeight != height) {
						if(rect.top == 0) {
							// top screen-side
							dstOffsetY = Clamped<int>(height - visibleHeight, 0, height);
						}
					}

					if(visibleHeight  &&  visibleWidth) {
						CopySamples2D(*apCroppedImage, *out, visibleHeight, visibleWidth,
										dstOffsetY, dstOffsetX, srcOffsetY, srcOffsetX, false);
					}
					out = apCroppedImage;
				}
			}
			else {
				// cannot lock back-buffer, so first copy pixels to temporary surface
				win_utils::COMPtr<IDirect3DSurface8> apTempSurf;
				win_utils::COMPtr<IDirect3DSurface8> apBBSurf;

				if(dev.GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &apBBSurf) != D3D_OK) return false;
				if(dev.CreateImageSurface(width, height, D3DFMT_A8R8G8B8, &apTempSurf) != D3D_OK) return false;
				if(dev.CopyRects(apBBSurf, 0, 0, apTempSurf, 0) != D3D_OK) return false;

				// copy pixels to output-image
				outSampleFormat = SampleFormat_A8R8G8B8;
				out.Assign(ManufactureImage(outSampleFormat, height, width));
				if(! d3d8_tools::CopySurfaceToImage(*apTempSurf, *out)) {
					out.SafeDelete();
					return false;
				}
			}
			break;
		}
		case FrameBufferFlags_Depth:
		case FrameBufferFlags_Stencil: {
			// Standard depth-stencil surfaces cannot be locked and
			// cannot be copied by D3D8. The only exception is D3DFMT_D16_LOCKABLE
			// which is generally unsupported by hardware and not used by this framework.
			return false;
		}
	}

	// pixels ready for export
	return (out != 0);
}


//====================================================================
IDirect3D8*  vs_direct3d8::GetD3DObject()
{
	return iObject;
}


//====================================================================
IDirect3DDevice8*  vs_direct3d8::GetD3DDevice()
{
	return iDevice;
}


//====================================================================
bool  vs_direct3d8::CreateD3DObject()
{
	if(iObject) return true;
	iObject = Direct3DCreate8(D3D_SDK_VERSION);
	return (iObject != 0);
}


//====================================================================
void  vs_direct3d8::CreateD3DDevice()
{
	// create d3d-object
	if(! CreateD3DObject()) {
		VW_THROW(VideoSubsystemException, "Cannot create Direct3D8 object.");
	}

	// check current display-mode match for non-fullscreen
	if(! videoConfig.frameBufferConfig.fullscreen) {

		D3DDISPLAYMODE  adapterDM;
		iObject->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &adapterDM);

		if(adapterDM.Format != D3DFMT_X8R8G8B8) {
			VW_THROW(VideoSubsystemRecoverableException,
					"Cannot operate in non-32-bit display-mode.");
		}
		if(videoConfig.displayModeConfig.bitsPerPixel != 32) {
			VW_THROW(VideoSubsystemRecoverableException,
					"Cannot operate in non-32-bit display-mode.");
		}
	}

	// translate parameters
	if(! TranslateConfiguration(iObject, deviceType, presentationParameters, videoConfig)) {

		VW_THROW(VideoSubsystemRecoverableException,
				"Requested frame-buffer-configuration is invalid.");
	}

	// check hint support
	D3DCAPS8 devCaps;
	HRESULT hr = iObject->GetDeviceCaps(D3DADAPTER_DEFAULT, deviceType, &devCaps);
	d3d8_tools::D3DErrorCheckET<VideoSubsystemException>(hr, "IDirect3D8::GetDeviceCaps");

	if(! (devCaps.DevCaps & D3DDEVCAPS_PUREDEVICE)) {
		videoConfig.videoHints.pureDeviceMode = false;
	}

	// map vertex-processing default, depending on rasterizer class
	if(videoConfig.videoHints.vertexProcessingMode == VideoHints::VertexProcessingMode_Default) {

		if(videoConfig.frameBufferConfig.hardwareRasterizer) {
			videoConfig.videoHints.vertexProcessingMode = VideoHints::VertexProcessingMode_Hardware;
		}
		else {
			videoConfig.videoHints.vertexProcessingMode = VideoHints::VertexProcessingMode_Software;
		}
	}

	// try different vertex-processing flags
	DWORD vertProcFlag = TranslateVertexProcessingMode(videoConfig.videoHints.vertexProcessingMode);
	while(true) {

		// add pure-device flag if requested and legal
		if(videoConfig.videoHints.pureDeviceMode) {
			if(vertProcFlag & D3DCREATE_HARDWARE_VERTEXPROCESSING) {
				vertProcFlag |= D3DCREATE_PUREDEVICE;
			}
		}

		// create rendering-device
		hr = iObject->CreateDevice(D3DADAPTER_DEFAULT, deviceType, pWindow->GetHWND(),
									vertProcFlag, &presentationParameters, &iDevice);

		if(hr == D3D_OK) {
			// success, extract realized hints
			videoConfig.videoHints.pureDeviceMode = ToBool(vertProcFlag & D3DCREATE_PUREDEVICE);
			videoConfig.videoHints.vertexProcessingMode = TranslateVertexProcessingFlag(vertProcFlag);

			olog << "Presentation parameters:" << nl;
			DumpSwapParameters(presentationParameters, olog, 1);
			return;
		}
		else {
			iDevice = 0;
		}
		if(! GetNextVertexProcessingFlag(vertProcFlag)) break;
	}

	// all bets are off, throw recoverable exception
	d3d8_tools::D3DErrorCheckET<VideoSubsystemRecoverableException>(hr, "IDirect3D8::CreateDevice");
}


//====================================================================
void  vs_direct3d8::StabilizeColorBuffers()
{
	D3DVIEWPORT8 vp;
	vp.X = 0;
	vp.Y = 0;
	vp.Width = pWindow->GetClientAreaWidth();
	vp.Height = pWindow->GetClientAreaHeight();
	vp.MinZ = 0.0;
	vp.MaxZ = 1.0;

	iDevice->SetViewport(&vp);

	for(uint i = 1; i <= presentationParameters.BackBufferCount + 1; i++) {
		iDevice->Clear(0, 0, D3DCLEAR_TARGET, D3DCOLOR_ARGB(255, 0, 0, 0), 0, 0);
		PresentBackBuffer();
	}
}


//====================================================================
void  vs_direct3d8::Window::OnResized(eResizeAction action)
{
	// Reset is necessary. Avoid changes if window is minimized.
	if(action != TopLevelWindow::ResizeAction_Minimize) {

		if(subsys.iDevice  &&  subsys.presentationParameters.Windowed) {

			// update backbuffer size, but never set dimensions to zero
			if(GetClientAreaWidth() != 0  &&  GetClientAreaHeight() != 0)
			{
				subsys.videoConfig.displayModeConfig.width = GetClientAreaWidth();
				subsys.videoConfig.displayModeConfig.height = GetClientAreaHeight();

				subsys.presentationParameters.BackBufferWidth = GetClientAreaWidth();
				subsys.presentationParameters.BackBufferHeight = GetClientAreaHeight();
			}
			subsys.SetOperationalState(OperationalState_WaitingReset);
		}
	}

	// propagate to base-class
	VideoSubsystemWindow::OnResized(action);
}

