

//--------------------------------------------------------------------
#include "private\vs_opengl.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Memory.h"
#include "BaseTools\PlatformInfo.h"
#include "DataObjects\ImageOperations.h"
#include "Framework\KernelMacros.h"
#include "GLTools\Extensions.h"
#include "GLTools\ErrorHandling.h"
#include "GLTools\States.h"
#include <sstream>

using namespace vertexware;


//********************************************************************
namespace {


//====================================================================
class DesktopDCSentry
{
public:
	DesktopDCSentry() {
		hdc = ::GetWindowDC(0);
	}
	~DesktopDCSentry() {
		if(hdc) ::ReleaseDC(0, hdc);
	}
	operator HDC() const { return hdc; }

private:
	HDC hdc;
};


//====================================================================
class PFDMaxIndex
{
public:
	PFDMaxIndex(HDC desktopDC) {
		maxIndex = ::DescribePixelFormat(desktopDC, 1, sizeof(PIXELFORMATDESCRIPTOR), 0);
	}
	operator int() const { return maxIndex; }

private:
	int maxIndex;
};


//====================================================================
__forceinline bool __fastcall IsAccelerated(const PIXELFORMATDESCRIPTOR& pfd)
{
	if(pfd.dwFlags & PFD_GENERIC_FORMAT) return false;
	return true;
}


//====================================================================
bool __fastcall IsLegacyPixelFormat(const PIXELFORMATDESCRIPTOR& pfd)
{
	if(pfd.iPixelType != PFD_TYPE_RGBA) return true;

	if((pfd.cRedBits != 8) || (pfd.cGreenBits != 8) ||
		(pfd.cBlueBits != 8) || (pfd.cAlphaBits != 8)) return true;

	if(pfd.dwFlags & PFD_NEED_PALETTE) return true;
	if(pfd.dwFlags & PFD_NEED_SYSTEM_PALETTE) return true;
	if(pfd.dwFlags & PFD_SUPPORT_GDI) return true;
	if(pfd.dwFlags & PFD_DRAW_TO_BITMAP) return true;
	if(pfd.dwFlags & PFD_SWAP_LAYER_BUFFERS) return true;

	if(! (pfd.dwFlags & PFD_DRAW_TO_WINDOW)) return true;
	if(! (pfd.dwFlags & PFD_SUPPORT_OPENGL)) return true;
	if(! (pfd.dwFlags & PFD_DOUBLEBUFFER)) return true;

	// WARNING: disabling native-stereo formats
	if(pfd.dwFlags & PFD_STEREO) return true;

	return false;
}


//====================================================================
int __fastcall FindPixelFormatIndex(const FrameBufferConfig& fbc,
									PIXELFORMATDESCRIPTOR& pfdOut)
{
	// set not-found
	int result = 0;

	DesktopDCSentry  desktopDC;
	if(! desktopDC) return result;

	PFDMaxIndex pfdMaxIndex(desktopDC);
	if(! pfdMaxIndex) return result;

	// iterate through available pixel-formats
	for(int pfdIndex = 1; pfdIndex <= pfdMaxIndex; pfdIndex++) {

		PIXELFORMATDESCRIPTOR pfd;
		if(! ::DescribePixelFormat(desktopDC, pfdIndex, sizeof(pfd), &pfd)) continue;

		// skip legacy formats
		if(IsLegacyPixelFormat(pfd)) continue;

		// match frame-buffer-config
		if(
			(fbc.hardwareRasterizer == IsAccelerated(pfd)) &&
			(fbc.backBufferCount == 1) &&

			(fbc.bitsRed   == pfd.cRedBits) &&
			(fbc.bitsGreen == pfd.cGreenBits) &&
			(fbc.bitsBlue  == pfd.cBlueBits) &&
			(fbc.bitsAlpha == pfd.cAlphaBits) &&

			(fbc.bitsDepth   == pfd.cDepthBits) &&
			(fbc.bitsStencil == pfd.cStencilBits) &&

			(fbc.bitsAccumRed   == pfd.cAccumRedBits) &&
			(fbc.bitsAccumGreen == pfd.cAccumGreenBits) &&
			(fbc.bitsAccumBlue  == pfd.cAccumBlueBits) &&
			(fbc.bitsAccumAlpha == pfd.cAccumAlphaBits)
		)
		{
			// found matching pixel-format
			result = pfdIndex;
			pfdOut = pfd;

			// prefer swap-exchange over swap-copy if available
			if((pfd.dwFlags  &  PFD_SWAP_EXCHANGE)  ||  !(pfd.dwFlags  &  PFD_SWAP_COPY)) {
				break; //found swap-exchange
			}
			// keep seeking
		}
	}
	return result;
}


//********************************************************************
}//namespace


//====================================================================
vs_opengl::vs_opengl()
{
	pWindow = 0;
	hglrc = 0;

	InvalidateExtensionPointers();
}


//====================================================================
vs_opengl::~vs_opengl()
{
	CleanupWindow();
}


//====================================================================
std::string  vs_opengl::GetVideoSubsystemInterfaceName()
{
	return VW_INTERFACE_NAME_VideoSubsystemOpenGL;
}


//====================================================================
bool  vs_opengl::DoBuildVideoProfile(std::ostream& log,
									VideoProfile& outVideoProfile)
{
	bool bOK = true;

	//--------------------
	// obtain general-info
	uint adapterID = 0;
	bool bFoundPrimary = false;

	while(true) {
		DISPLAY_DEVICE dd;
		::ZeroMemory(&dd, sizeof(dd));
		dd.cb = sizeof(dd);

		if(! ::EnumDisplayDevices(0, adapterID ++, &dd, 0)) break;

		if(dd.StateFlags  &  DISPLAY_DEVICE_PRIMARY_DEVICE) {
			outVideoProfile.SetDriverInfo(dd.DeviceString, "unknown", win_utils::Guid());

			bFoundPrimary = true;
			break;
		}
	}
	if(! bFoundPrimary) {
		bOK = false;
		log << "Cannot access primary-display-adapter info." << nl;
	}

	//----------------------
	// collect display-modes
	DWORD dmIndex = 0;
	while(true) {

		DEVMODE dm;
		::ZeroMemory(&dm, sizeof(dm));
		dm.dmSize = sizeof(dm);
		if(! ::EnumDisplaySettings(0, dmIndex ++, &dm)) break;

		DisplayModeConfig dmc(dm.dmPelsWidth, dm.dmPelsHeight,
								dm.dmBitsPerPel, dm.dmDisplayFrequency);
		// save safe-modes
		if(DisplayModeConfigFilter::IsSafeDisplayModeConfig(dmc)) {
			outVideoProfile.Add(dmc);
		}
	}

	//-----------------------------
	// collect frame-buffer-configs
	VideoProfile::FBC_VEC  fbcVecTemp;

	DesktopDCSentry desktopDC;
	if(desktopDC) {

		PFDMaxIndex pfdMaxIndex(desktopDC);
		if(pfdMaxIndex) {

			// iterate through available pixel-formats
			for(int pfdIndex = 1; pfdIndex <= pfdMaxIndex; pfdIndex++) {

				PIXELFORMATDESCRIPTOR pfd;
				if(::DescribePixelFormat(desktopDC, pfdIndex, sizeof(pfd), &pfd)) {

					// skip legacy formats
					if(IsLegacyPixelFormat(pfd)) continue;

					// fill frame-buffer-config
					FrameBufferConfig  fbc;

					fbc.fullscreen = true;
					fbc.hardwareRasterizer = IsAccelerated(pfd);
					fbc.backBufferCount = 1;

					fbc.bitsRed   = pfd.cRedBits;
					fbc.bitsGreen = pfd.cGreenBits;
					fbc.bitsBlue  = pfd.cBlueBits;
					fbc.bitsAlpha = pfd.cAlphaBits;

					fbc.bitsDepth   = pfd.cDepthBits;
					fbc.bitsStencil = pfd.cStencilBits;

					fbc.bitsAccumRed   = pfd.cAccumRedBits;
					fbc.bitsAccumGreen = pfd.cAccumGreenBits;
					fbc.bitsAccumBlue  = pfd.cAccumBlueBits;
					fbc.bitsAccumAlpha = pfd.cAccumAlphaBits;

					// save configuration
					fbcVecTemp.push_back(fbc);
				}
			}
		}
	}

	// add configurations to profile
	for(uint i = 0; i < fbcVecTemp.size(); i++) {
		outVideoProfile.Add(fbcVecTemp[i]);
	}
	for(uint j = 0; j < fbcVecTemp.size(); j++) {
		fbcVecTemp[j].fullscreen = false;
		outVideoProfile.Add(fbcVecTemp[j]);
	}

	// report result
	return bOK;
}


//====================================================================
VideoSubsystemWindow*  vs_opengl::InitializeWindow(const VideoConfig& _videoConfig,
													bool _bAutoShowWindow)
{
	// check arguments
	if(pWindow) {
		VW_THROW(VideoSubsystemException, "Window already initialized.");
	}

	videoConfig = _videoConfig;

	// fullscreen mode
	if(videoConfig.frameBufferConfig.fullscreen) {

		videoConfig.resizable = false;

		// switch display-mode
		if(! dmss.Switch(videoConfig.displayModeConfig)) {
			VW_THROW(VideoSubsystemRecoverableException,
					"Cannot switch to requested display-mode.");
		}

		// check realized refresh-frequency
		platform_info::GetDesktopDisplayMode(0, 0, 0,
							&videoConfig.displayModeConfig.refreshFrequency);

		// create and setup window
		pWindow = new Window(*this, true);
		pWindow->CreateSystemWindow(0, WS_POPUP, WS_EX_APPWINDOW | WS_EX_TOPMOST);
		pWindow->FillDesktop();
	}

	// windowed mode
	else {
		videoConfig.displayModeConfig.refreshFrequency = 0;

		// adjust size and style
		DWORD style = Window::GetStandardWindowStyle(videoConfig.resizable);
		DWORD styleEx = WS_EX_APPWINDOW;

		uint  totalWindowWidth, totalWindowHeight;
		Window::AdjustWindowSize(style, styleEx,
								videoConfig.displayModeConfig.width,
								videoConfig.displayModeConfig.height,
								totalWindowWidth, totalWindowHeight, true);

		// create and setup window
		pWindow = new Window(*this, false);
		pWindow->CreateSystemWindow(0, style, styleEx);
		pWindow->CenterOnDesktop(totalWindowWidth, totalWindowHeight);
	}

	// additional window setup
	pWindow->EnableSystemCommandEmulation(false);

	// setup rendering context
	try {
		CreateRenderingContext();
		BindRCToCurrentThread();
	}
	catch(VideoSubsystemRecoverableException&) {
		CleanupWindow();
		throw;
	}

	// ready to render
	SetOperationalState(OperationalState_Ready);

	if(_bAutoShowWindow) {
		pWindow->Show();
	}
	PostWindowInitialize();

	// v-sync control is available through extension
	SetupVSync(videoConfig.videoHints.vSync);

	return pWindow;
}


//====================================================================
VideoSubsystemWindow*  vs_opengl::GetWindow()
{
	return pWindow;
}


//====================================================================
void  vs_opengl::CleanupWindow()
{
	if(! pWindow) return;

	PreWindowCleanup();

	// release rendering context
	olog << "Releasing rendering-context." << nl;
	UnbindRCFromCurrentThread();
	DestroyRenderingContext();

	// destroy window
	olog << "Destroying main window." << nl;
	SafeDelete(pWindow);

	// restore display-mode if necessary
	olog << "Restoring display mode." << nl;
	dmss.Restore();

	// update state
	olog << "Window-cleanup finished." << nl;
	SetOperationalState(OperationalState_WaitingInit);
}


//====================================================================
void  vs_opengl::PresentBackBuffer()
{
	if(::SwapBuffers(pWindow->GetHDC()) != TRUE) {
		SetOperationalState(OperationalState_WaitingReset);
	}
}


//====================================================================
void  vs_opengl::ErrorCheck()
{
	if(gl_tools::GLErrorCheck("vs_opengl::ErrorCheck", 0, &errorDesc)) {
		olog << errorDesc << std::endl;
		errorDesc.erase();
	}
}


//====================================================================
void  vs_opengl::Reset()
{
	if(pWindow->IsFullscreen()) {

		// OnWindowActivate handler failed to switch display-mode,
		// but report error here, it seems cleaner than throwing
		// exception in the middle of message-loop

		VW_THROW(VideoSubsystemException,
				"Cannot switch to requested display-mode.");
	}
	// else nothing to do, just confirm ready-state
	SetOperationalState(OperationalState_Ready);
}


//====================================================================
void  vs_opengl::UpdateVSync(bool bEnable)
{
	if(videoConfig.videoHints.vSync != bEnable) {
		videoConfig.videoHints.vSync = bEnable;

		SetupVSync(bEnable);
	}
}


//====================================================================
bool  vs_opengl::CaptureFrameBuffer(AutoPtr<ImageBase>& out,
									uint frameBufferFlags,
									bool bForceColorSampleFormat)
{
	out.SafeDelete();
	if(! hglrc) return false;

	// process stereo flags
	switch(frameBufferFlags  &  FrameBufferFlags_StereoMask) {

		case FrameBufferFlags_Mono: {
			break;
		}
		default: {
			// stereo not implemented
			return false;
		}
	}

	// process swap-chain flags
	bool bReadFrontBuffer = false;

	switch(frameBufferFlags  &  FrameBufferFlags_SwapChainMask) {

		case FrameBufferFlags_Back: {
			bReadFrontBuffer = false;
			break;
		}
		case FrameBufferFlags_Front: {
			bReadFrontBuffer = true;
			break;
		}
		default: {
			return false;
		}
	}

	// process component flags
	eSampleFormat  outSampleFormat = SampleFormat_Invalid;
	GLenum         readFormat = GL_BGRA;
	GLenum         readType = GL_UNSIGNED_BYTE;

	switch(frameBufferFlags  &  FrameBufferFlags_ComponentMask) {

		case FrameBufferFlags_Color: {
			if(bReadFrontBuffer) {
				outSampleFormat = SampleFormat_X8R8G8B8;
				readFormat = GL_BGRA;
				readType = GL_UNSIGNED_BYTE;
			}
			else {
				outSampleFormat = SampleFormat_A8R8G8B8;
				readFormat = GL_BGRA;
				readType = GL_UNSIGNED_BYTE;
			}
			break;
		}
		case FrameBufferFlags_Depth: {
			if(bReadFrontBuffer) return false;

			outSampleFormat = SampleFormat_ZB32;
			readFormat = GL_DEPTH_COMPONENT;
			readType = GL_FLOAT;
			break;
		}
		case FrameBufferFlags_Stencil: {
			if(bReadFrontBuffer) return false;

			outSampleFormat = SampleFormat_L8;
			readFormat = GL_STENCIL_INDEX;
			readType = GL_UNSIGNED_BYTE;
			break;
		}
		default: {
			return false;
		}
	}

	// prepare output storage
	uint width = pWindow->GetClientAreaWidth();
	uint height = pWindow->GetClientAreaHeight();

	out.Assign(ManufactureImage(outSampleFormat, height, width));
	if(! out) return false;

	// setup read-buffer & pixel-storage control
	gl_tools::ReadBufferSentry  rbs(bReadFrontBuffer ? GL_FRONT : GL_BACK);
	gl_tools::PixelPackSentry   pps(GL_FALSE, 1, 0, 0);

	// do read pixels & fix row-storage order
	glReadPixels(0, 0, width, height, readFormat, readType, out->GetSamplesB());
	FlipRows(*out);

	// force color output if requested
	bool bOK = true;

	if(bForceColorSampleFormat) {
		switch(outSampleFormat) {

			case SampleFormat_ZB32:
			case SampleFormat_WB32: {
				AutoPtr<ImageBase> outColor(ManufactureImage(SampleFormat_L8, height, width));
				if(! TransferSamples2D(*outColor, *out, false)) bOK = false;
				out = outColor;
				break;
			}
		}
	}
	if(! bOK) {
		out.SafeDelete();
	}

	// pixels ready for export
	return bOK;
}


//====================================================================
vs_opengl::DisplayModeSwitchSentry::DisplayModeSwitchSentry(const DisplayModeConfig& dmc)
{
	bSwitched = false;
	Switch(dmc);
}


//====================================================================
bool  vs_opengl::DisplayModeSwitchSentry::Switch(const DisplayModeConfig& dmc)
{
	Restore();

	DEVMODE dm;
	::ZeroMemory(&dm, sizeof(dm));
	dm.dmSize = sizeof(dm);

	dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
	dm.dmPelsWidth = dmc.width;
	dm.dmPelsHeight = dmc.height;
	dm.dmBitsPerPel = dmc.bitsPerPixel;

	if(dmc.refreshFrequency != 0) {
		dm.dmFields |= DM_DISPLAYFREQUENCY;
		dm.dmDisplayFrequency = dmc.refreshFrequency;
	}

	bSwitched = (::ChangeDisplaySettings(&dm, CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL);
	return bSwitched;
}


//====================================================================
void  vs_opengl::DisplayModeSwitchSentry::Restore()
{
	if(bSwitched) {
		::ChangeDisplaySettings(0, 0);
		bSwitched = false;
	}
}


//====================================================================
void  vs_opengl::CreateRenderingContext()
{
	// pure-device-mode not supported, cannot determine vertex-processing-mode,
	// fullscreen destination alpha always supported
	videoConfig.videoHints.fullscreenDestinationAlpha = true;
	videoConfig.videoHints.pureDeviceMode = false;
	videoConfig.videoHints.vertexProcessingMode = VideoHints::VertexProcessingMode_Default;

	// check current display-mode match for non-fullscreen
	if(! videoConfig.frameBufferConfig.fullscreen) {
		if(videoConfig.displayModeConfig.bitsPerPixel != 32) {
			VW_THROW(VideoSubsystemRecoverableException,
					"Cannot operate in non-32-bit display-mode.");
		}
	}

	// find matching pixel-format
	PIXELFORMATDESCRIPTOR  pfd;
	int  pixelFormatIndex = FindPixelFormatIndex(videoConfig.frameBufferConfig, pfd);

	if((! pixelFormatIndex) ||
		IsLegacyPixelFormat(pfd) ||
		(videoConfig.frameBufferConfig.backBufferCount != 1))
	{
		VW_THROW(VideoSubsystemRecoverableException,
				"Requested frame-buffer-configuration is invalid.");
	}

	// setup pixel-format
	if(! ::SetPixelFormat(pWindow->GetHDC(), pixelFormatIndex, &pfd)) {
		VW_THROW(VideoSubsystemRecoverableException, "Cannot set pixel-format.");
	}

	// create OpenGL-rendering-context
	hglrc = ::wglCreateContext(pWindow->GetHDC());
	if(! hglrc) {
		VW_THROW(VideoSubsystemRecoverableException, "Cannot create OpenGL-rendering-context.");
	}
}


//====================================================================
void  vs_opengl::DestroyRenderingContext()
{
	// destroy OpenGL-rendering-context
	if(hglrc) {
		::wglDeleteContext(hglrc);
		hglrc = 0;
	}
}


//====================================================================
void  vs_opengl::SetupVSync(bool bEnable)
{
	if(wglSwapIntervalEXT  &&  wglGetSwapIntervalEXT) {

		int lastSwapInterval = wglGetSwapIntervalEXT();
		wglSwapIntervalEXT(bEnable ? 1 : 0);

		olog << "Updated WGL-swap-interval from "
					<< lastSwapInterval << " to "
					<< wglGetSwapIntervalEXT() << std::endl;
	}
	else {
		olog << "WGL_EXT_swap_control not available." << std::endl;
	}
}


//====================================================================
bool  vs_opengl::IsRCBoundToCurrentThread()
{
	// check if OpenGL-rendering-context is bound to calling-thread
	return (::wglGetCurrentContext() == hglrc);
}


//====================================================================
void  vs_opengl::BindRCToCurrentThread()
{
	// bind OpenGL-rendering-context to calling-thread
	if(! ::wglMakeCurrent(pWindow->GetHDC(), hglrc)) {
		VW_THROW(VideoSubsystemException,
				"Cannot bind OpenGL-rendering-context to thread.");
	}
}


//====================================================================
void  vs_opengl::UnbindRCFromCurrentThread()
{
	// unbind OpenGL-rendering-context from calling-thread
	if(IsRCBoundToCurrentThread()) {
		::wglMakeCurrent(0, 0);
	}
}


//====================================================================
void  vs_opengl::StabilizeColorBuffers()
{
	glViewport(0, 0, pWindow->GetClientAreaWidth(), pWindow->GetClientAreaHeight());

	glClearColor(0, 0, 0, 1);
	for(uint i = 1; i <= 2; i++) {
		glClear(GL_COLOR_BUFFER_BIT);
		PresentBackBuffer();
	}
}


//====================================================================
void  vs_opengl::InitExtensionPointers()
{
	using namespace vertexware;
	using namespace gl_tools;
	wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXT) GetExtensionProc("wglSwapIntervalEXT");
	wglGetSwapIntervalEXT = (PFNWGLGETSWAPINTERVALEXT) GetExtensionProc("wglGetSwapIntervalEXT");
}


//====================================================================
void  vs_opengl::InvalidateExtensionPointers()
{
	wglSwapIntervalEXT = 0;
	wglGetSwapIntervalEXT = 0;
}


//====================================================================
void  vs_opengl::PostWindowInitialize()
{
	StabilizeColorBuffers();
	InitExtensionPointers();
}


//====================================================================
void  vs_opengl::PreWindowCleanup()
{
	InvalidateExtensionPointers();
}


//====================================================================
void  vs_opengl::Window::OnActivated(bool bActive)
{
	// fullscreen complications
	if(IsResizeHandlerActive()  &&  IsFullscreen()) {

		if(bActive) {
			if(! subsys.dmss.IsSwitched()) {
				// dynamically activated fullscreen
				if(! subsys.dmss.Switch(subsys.videoConfig.displayModeConfig)) {
					subsys.SetOperationalState(VideoSubsystem::OperationalState_WaitingReset);
				}
			}
		}
		else {
			// dynamically deactivated fullscreen (ALT-TAB or similar combo)
			// WARNING: Commands must be executed in given order,
			// otherwise errors will occur in size-handling.
			// Minimize the window first, then restore display settings.
			Minimize();
			subsys.dmss.Restore();
		}
	}

	// propagate to base-class
	VideoSubsystemWindow::OnActivated(bActive);
}


//====================================================================
void  vs_opengl::Window::OnResized(eResizeAction action)
{
	// Avoid changes if window is minimized.
	if(action != TopLevelWindow::ResizeAction_Minimize)
	{
		subsys.videoConfig.displayModeConfig.width = GetClientAreaWidth();
		subsys.videoConfig.displayModeConfig.height = GetClientAreaHeight();
	}

	// propagate to base-class
	VideoSubsystemWindow::OnResized(action);
}

