

//--------------------------------------------------------------------
#include "BaseTools\Debug.h"
#include "BaseTools\CommandLineParser.h"
#include "BaseTools\String.h"
#include "BaseTools\Constructs.h"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
static const char*  gc_whitespaceChars  = " \t";


//====================================================================
template<typename T>
inline bool __fastcall ExtractArgument(const CommandLineParser& clp,
							const std::string& name, T& outValue, T defaultValue = T(0))
{
	bool bFound = false;
	std::string sVal;

	if(clp.GetAssignment(name, sVal)) {
		bFound = ToValue(sVal, outValue);
	}
	if(! bFound) {
		outValue = defaultValue;
	}
	return bFound;
}


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
const char*  CommandLineParser::s_szDefaultPackedArgumentSeparators = ".,:;";


//====================================================================
std::string  CommandLineParser::Argument::MakeFullString() const
{
	switch(type) {
		case CommandLineParser::ArgumentType_Option:
			return Option();

		case CommandLineParser::ArgumentType_Assignment:
			return Assignment().Name() + "=" + Assignment().Value();

		case CommandLineParser::ArgumentType_String:
			return String();
	}
	return "";
}


//====================================================================
CommandLineParser::CommandLineParser()
{
	SetSpecialCharacters();
	SetParseFlags();
}


//====================================================================
void  CommandLineParser::SetSpecialCharacters(char _quoteChar, char _quoteEscapeChar,
									char _scopeChar, char _optionChar, char _assignChar)
{
	quoteChar = _quoteChar;
	quoteEscapeChar = _quoteEscapeChar;
	scopeChar = _scopeChar;
	optionChar = _optionChar;
	assignChar = _assignChar;
}


//====================================================================
void  CommandLineParser::GetSpecialCharacters(char& _quoteChar, char& _quoteEscapeChar,
							char& _scopeChar, char& _optionChar, char& _assignChar) const
{
	_quoteChar = quoteChar;
	_quoteEscapeChar = quoteEscapeChar;
	_scopeChar = scopeChar;
	_optionChar = optionChar;
	_assignChar = assignChar;
}


//====================================================================
void  CommandLineParser::ClearLastCommandLine()
{
	lastCommandLine.erase();
}


//====================================================================
void  CommandLineParser::ClearLastResults()
{
	objectIdentifierChain.clear();
	commandIdentifier.erase();
	arguments.clear();
}


//====================================================================
void  CommandLineParser::ClearAll()
{
	ClearLastCommandLine();
	ClearLastResults();
}


//====================================================================
void  CommandLineParser::SetParseFlags(bool _bIgnoreObjectChain,
										bool _bStripDoubleQuotesOnStdCmd)
{
	bIgnoreObjectChain = _bIgnoreObjectChain;
	bStripDoubleQuotesOnStdCmd = _bStripDoubleQuotesOnStdCmd;
}


//====================================================================
bool  CommandLineParser::Parse(const std::string& commandLine, bool exceptionOnError)
{
	ClearAll();

	lastCommandLine = commandLine;
	std::string s = lastCommandLine;

	try {
		// eat leading whitespace
		EatLeading(s, gc_whitespaceChars);

		// extract command-sequence
		if(bIgnoreObjectChain) {

			// standard command / quoted_string or string
			if(! ExtractQuotedString(s, commandIdentifier)) {
				if(! ExtractString(s, commandIdentifier)) {
					VW_THROW(CommandLineErrorException, "Expected command");
				}
			}
			if(bStripDoubleQuotesOnStdCmd) {
				TrimQuotes(commandIdentifier);
			}
		}
		else {
			// object_id.object_id.command_id
			while(true) {
				std::string id;
				if(! ExtractIdentifier(s, id, false, false, false)) {
					VW_THROW(CommandLineErrorException, "Expected identifier");
				}
				if(s.empty()  ||  MatchNext(s, gc_whitespaceChars)) {
					// found end of command-sequence
					commandIdentifier = id;
					break;
				}
				if(! EatNext(s, scopeChar)) {
					VW_THROW(CommandLineErrorException, "Expected scope specifier");
				}
				objectIdentifierChain.push_back(id);
			}
		}

		// extract arguments
		int argPosition = -1;
		std::string  curr, next;

		while(s.size()) {
			uint wsCount = EatLeading(s, gc_whitespaceChars);

			argPosition++;

			// try option
			if(ExtractOption(s, curr)) {

				// try assignment
				if(EatNext(s, assignChar)) {

					// process assignment rhs
					if(! ExtractQuotedString(s, next)) {
						if(! ExtractString(s, next)) {
							// allow null-assignment
							next.erase();
						}
					}
					// found assignment
					arguments.push_back(Argument(argPosition, ArgumentType_Assignment, curr, next));
					continue;
				}
				else if(s.empty()  ||  MatchNext(s, gc_whitespaceChars)) {
					// found option
					arguments.push_back(Argument(argPosition, ArgumentType_Option, curr));
					continue;
				}
				else {
					VW_THROW(CommandLineErrorException, "Expected whitespace or assignment");
				}
			}

			// try quoted string first, than any string
			bool  bFoundQuotedString = ExtractQuotedString(s, curr);
			if(bFoundQuotedString  ||  ExtractString(s, curr)) {

				// found string, must be whitespace-separated from previous entity
				if(wsCount == 0) {

					// putback extracted string for proper error report
					if(bFoundQuotedString) {
						curr.insert(0, quoteChar);
						curr.append(1, quoteChar);
					}
					s.insert(0, curr);

					VW_THROW(CommandLineErrorException, "Expected whitespace before string");
				}
				arguments.push_back(Argument(argPosition, ArgumentType_String, curr));
				continue;
			}
		}
	}
	catch(CommandLineErrorException& e) {
		// error occured, cleanup accumulated data
		ClearLastResults();

		if(exceptionOnError) {
			// report error position
			std::string errorMessage = e.GetMessage();
			errorMessage.append(" at: ").append(s);
			errorMessage.append("\nFull command-line was: ").append(lastCommandLine);
			e.SetMessage(errorMessage);

			throw;
		}
		return false;
	}
	return true;
}


//====================================================================
bool  CommandLineParser::Parse(int argCount, const char* args[], bool exceptionOnError)
{
	std::string  commandLine;
	for(uint i = 0; i < argCount; i++) {
		commandLine.append(args[i]);
		commandLine.append(1, ' ');
	}
	return  Parse(commandLine, exceptionOnError);
}


//====================================================================
bool  CommandLineParser::ExtractIdentifier(std::string& s, std::string& out,
					bool canStartWithDigit, bool canStartWithUnderscore, bool canStartWithHypen)
{
	uint pos = 0;
	while(pos < s.size()) {
		char ch = s[pos];

		if(pos == 0) {
			bool bOK =	isalpha(ch)  ||
						(canStartWithDigit  &&  isdigit(ch))  ||
						(canStartWithUnderscore  &&  ch == '_')  ||
						(canStartWithHypen  &&  ch == '-');
			if(! bOK) break; //error
		}
		else if(! isalpha(ch)  &&  ! isdigit(ch)  &&  ch != '-'  &&  ch != '_') {
			break; //error
		}
		pos++;
	}

	if(pos >= 1) {
		out = s.substr(0, pos);
		s.erase(0, pos);
		return true;
	}
	return false;
}


//====================================================================
bool  CommandLineParser::ExtractOption(std::string& s, std::string& out)
{
	if(s.size() <= 1  ||  s[0] != optionChar) return false;
	if(ExtractIdentifier(s.substr(1), out, false, true, true)) {
		s.erase(0, out.size() + 1);
		return true;
	}
	return false;
}


//====================================================================
bool  CommandLineParser::ExtractQuotedString(std::string& s, std::string& out)
{
	if(s.empty()  ||  s[0] != quoteChar) return false;

	out.erase();

	bool inEscape = false;
	uint pos = 1;

	while(pos < s.size()) {
		char ch = s[pos++];

		if(inEscape) {
			if(ch != quoteEscapeChar  &&  ch != quoteChar) {
				VW_THROW(CommandLineErrorException, "Bad escape sequence");
			}
			inEscape = false;
		}
		else if(ch == quoteEscapeChar) {
			inEscape = true;
			continue;
		}
		else if(ch == quoteChar) {
			s.erase(0, pos);
			return true;
		}
		out.append(1, ch);
	}
	VW_THROW(CommandLineErrorException, "Quote not closed");
	return false;
}


//====================================================================
bool  CommandLineParser::ExtractString(std::string& s, std::string& out)
{
	if(s.empty()) return false;

	std::string stopChars = gc_whitespaceChars;

	uint pos = s.find_first_of(stopChars);
	if(pos == 0) return false;

	out = s.substr(0, pos);
	s.erase(0, pos);
	return true;
}


//====================================================================
void  CommandLineParser::GetCommand(STRING_VEC& outObjectIdentifierChain,
									std::string& outCommandIdentifier) const
{
	outObjectIdentifierChain = objectIdentifierChain;
	outCommandIdentifier     = commandIdentifier;
}


//====================================================================
const CommandLineParser::Argument&  CommandLineParser::GetArgument(uint i) const
{
	VW_ASSERT(i < arguments.size(), "Index out of range.");
	return  arguments[i];
}


//====================================================================
bool  CommandLineParser::GetOptionArgument(const std::string& s, Argument& arg) const
{
	for(uint i = 0; i < GetNumArguments(); i++) {
		const Argument& a = GetArgument(i);
		if(a == ArgumentType_Option  &&  a.Option() == s) {
			arg = a;
			return true;
		}
	}
	return false;
}


//====================================================================
bool  CommandLineParser::GetAssignmentArgument(const std::string& s, Argument& arg) const
{
	for(uint i = 0; i < GetNumArguments(); i++) {
		const Argument& a = GetArgument(i);
		if(a == ArgumentType_Assignment  &&  a.Assignment().Name() == s) {
			arg = a;
			return true;
		}
	}
	return false;
}


//====================================================================
bool  CommandLineParser::GetStringArgument(const std::string& s, Argument& arg) const
{
	for(uint i = 0; i < GetNumArguments(); i++) {
		const Argument& a = GetArgument(i);
		if(a == ArgumentType_String  &&  a.String() == s) {
			arg = a;
			return true;
		}
	}
	return false;
}


//====================================================================
bool  CommandLineParser::GetOption(const std::string& option) const
{
	Argument arg;
	return GetOptionArgument(option, arg);
}


//====================================================================
bool  CommandLineParser::GetAssignment(const std::string& name, std::string& value) const
{
	Argument arg;
	if(GetAssignmentArgument(name, arg)) {
		value = arg.Assignment().Value();
		return true;
	}
	return false;
}


//====================================================================
bool  CommandLineParser::GetString(const std::string& s) const
{
	Argument arg;
	return GetStringArgument(s, arg);
}


//====================================================================
bool  CommandLineParser::ExtractArgumentPointer(const std::string& name,
									void*& outValue, void* defaultValue) const
{
	return ExtractArgument<void*>(*this, name, outValue, defaultValue);
}


//====================================================================
bool  CommandLineParser::ExtractArgumentInteger(const std::string& name,
									int& outValue, int defaultValue) const
{
	return ExtractArgument<int>(*this, name, outValue, defaultValue);
}


//====================================================================
bool  CommandLineParser::ExtractArgumentUInteger(const std::string& name,
									uint& outValue, uint defaultValue) const
{
	return ExtractArgument<uint>(*this, name, outValue, defaultValue);
}


//====================================================================
bool  CommandLineParser::ExtractArgumentDouble(const std::string& name,
									double& outValue, double defaultValue) const
{
	return ExtractArgument<double>(*this, name, outValue, defaultValue);
}


//====================================================================
bool  CommandLineParser::ExtractArgumentBool(const std::string& name,
									bool& outValue, bool defaultValue) const
{
	return ExtractArgument<bool>(*this, name, outValue, defaultValue);
}


//====================================================================
bool  CommandLineParser::GetPackedAssignment(const std::string& name,
			std::vector<std::string>& outValues, const std::string& separators) const
{
	outValues.clear();

	std::string sVal;
	if(! GetAssignment(name, sVal)) return false;

	while(sVal.length()) {
		uint pos = sVal.find_first_of(separators);

		if(pos == std::string::npos) {
			outValues.push_back(sVal);
			break;
		}
		else {
			if(pos != 0) {
				outValues.push_back(sVal.substr(0, pos));
			}
			sVal.erase(0, pos + 1);
		}
	}
	return true;
}


//====================================================================
bool  CommandLineParser::ExtractPackedUIntegers(const std::string& name,
			std::vector<uint>& outValues, const std::string& separators) const
{
	outValues.clear();

	std::vector<std::string>  sValues;
	if(! GetPackedAssignment(name, sValues, separators)) return false;

	// return on first conversion error, keep accumulated output
	for(uint i = 0; i < sValues.size(); i++) {
		uint value;
		if(! ToValue<uint>(sValues[i], value)) {
			return false;
		}
		outValues.push_back(value);
	}
	return true;
}


//====================================================================
std::ostream& operator<<(std::ostream& os, const CommandLineParser::Argument& arg)
{
	switch(arg.Type()) {
		case CommandLineParser::ArgumentType_Option:
			os << "option: " << arg.Option();
			break;
		case CommandLineParser::ArgumentType_Assignment:
			os << "assign: " << arg.Assignment().Name() << " = " << arg.Assignment().Value();
			break;
		case CommandLineParser::ArgumentType_String:
			os << "string: " << arg.String();
			break;
	}
	return os;
}


//====================================================================
void  CommandLineParser::DumpLastResults(std::ostream& os) const
{
	if(! bIgnoreObjectChain) {
		os << "Object-chain: ";
		uint max = objectIdentifierChain.size() - 1;
		for(uint i = 0; i <= max; i++) {
			os << objectIdentifierChain[i];
			if(i != max) os << '.';
		}
		os << std::endl;
	}
	os << "Command: " << commandIdentifier << std::endl;
	
	for(uint i = 0; i < arguments.size(); i++) {
		os << "Argument " << i << ": " << arguments[i] << std::endl;
	}
}


//====================================================================
std::string  CommandLineParser::FormatOptionHelp(const std::string& option,
						const std::string& helpText, uint tabLevel) const
{
	std::string s(tabLevel, '\t');
	s.append(1, optionChar).append(option);
	if(helpText.length()) s.append(", ").append(helpText);
	return s;
}


//====================================================================
std::string  CommandLineParser::FormatAssignmentHelp(const std::string& name,
						const std::string& value, const std::string& helpText, uint tabLevel) const
{
	std::string s(tabLevel, '\t');
	s.append(1, optionChar).append(name).append(1, assignChar).append(value);
	if(helpText.length()) s.append(", ").append(helpText);
	return s;
}


//====================================================================
std::string  CommandLineParser::FormatStringHelp(const std::string& str,
						const std::string& helpText, uint tabLevel) const
{
	std::string s(tabLevel, '\t');
	s.append(str);
	if(helpText.length()) s.append(", ").append(helpText);
	return s;
}


//********************************************************************
}//namespace vertexware


