

//--------------------------------------------------------------------
#include "BaseTools\ConsoleBuffers.h"
#include "BaseTools\STLExtensions.h"


//====================================================================
#define  DEFAULT_LINE_LIMIT      300
#define  DEFAULT_HISTORY_LIMIT   50


//********************************************************************
namespace vertexware {


//====================================================================
const uint ConsoleOutputBuffer::s_defaultLineLimit = DEFAULT_LINE_LIMIT;


//====================================================================
ConsoleOutputBuffer::ConsoleOutputBuffer(uint _lineLimit)
{
	SetLineLimit(_lineLimit);
	SetTabWidth();
	Clear();
}


//====================================================================
ConsoleOutputBuffer::~ConsoleOutputBuffer()
{
}


//====================================================================
void  ConsoleOutputBuffer::AddObserver(Observer& ob)
{
	CriticalSectionSentry css(cs);
	UniqueContainerPushBack(observers, &ob);
	NotifyObservers();
}


//====================================================================
void  ConsoleOutputBuffer::RemoveObserver(Observer& ob)
{
	CriticalSectionSentry css(cs);
	EraseFromContainer(observers, &ob);
}


//====================================================================
void  ConsoleOutputBuffer::SetLineLimit(uint _lineLimit)
{
	CriticalSectionSentry css(cs);
	// must have at least one line
	lineLimit = ClampedLow<uint>(_lineLimit, 1);
	DiscardLines();
	NotifyObservers();
}


//====================================================================
void  ConsoleOutputBuffer::SetTabWidth(uint _tabWidth)
{
	CriticalSectionSentry css(cs);
	tabWidth = ClampedLow<uint>(_tabWidth, 1);
}


//====================================================================
void  ConsoleOutputBuffer::Clear()
{
	CriticalSectionSentry css(cs);
	lines.clear();
	columnMap.clear();
	BreakLine(0);
	CacheMaxColumns();
	NotifyObservers();
}


//====================================================================
void  ConsoleOutputBuffer::Append(const std::string& s, uint outputType)
{
#ifndef VW_BUILD_DEBUG
	// ignore debug output in non-debug builds
	if(outputType == OutputStream::OutputType_Debug) return;
#endif

	// auto-break on different output-type
	bool bLineAdded = false;

	if(lines.back().outputType != outputType) {
		if(lines.back().text.length()) {
			BreakLine(outputType);
			bLineAdded = true;
		}
		else {
			lines.back().outputType = outputType;
		}
	}

	if(s.empty()  &&  ! bLineAdded) return; //no changes

	if(s.length()) {
		// tabs can take a lot of space
		const uint reserveFactor = 2;
		std::string data;
		data.reserve(s.length() * reserveFactor);

		// expand tabs and strip newlines
		for(uint i = 0; i < s.length(); i++) {
			char ch = s[i];

			switch(ch) {
				case '\t':
					data.append(tabWidth, ' ');
					break;

				case '\n':
					if(data.length()) {
						AppendToLastLine(data);
					}
					BreakLine(outputType);
					data.erase();
					data.reserve(s.length() * reserveFactor);
					break;

				default:
					data.append(1, ch);
					break;
			}
		}
		if(data.length()) {
			AppendToLastLine(data);
		}
	}
	CacheMaxColumns();
	NotifyObservers();
}


//====================================================================
void  ConsoleOutputBuffer::DiscardLines()
{
	while(lines.size() > lineLimit) {
		ReleaseColumnCount(lines.front().text.length());
		lines.pop_front();
	}
}


//====================================================================
void  ConsoleOutputBuffer::BreakLine(uint outputType)
{
	lines.push_back(Line(std::string(), outputType));
	AddColumnCount(0);
	DiscardLines();
}


//====================================================================
void  ConsoleOutputBuffer::AppendToLastLine(const std::string& s)
{
	std::string& lastLine = lines.back().text;
	ReleaseColumnCount(lastLine.length());

	lastLine.append(s);
	AddColumnCount(lastLine.length());
}


//====================================================================
void  ConsoleOutputBuffer::NotifyObservers()
{
	for(uint i = 0; i < observers.size(); i++) {
		observers[i]->OnChanged(*this);
	}
}


//====================================================================
const uint  ConsoleBuffers::s_defaultCommandHistoryLimit = DEFAULT_HISTORY_LIMIT;
const uint  ConsoleBuffers::s_defaultOutputLineLimit = DEFAULT_LINE_LIMIT;


//====================================================================
ConsoleBuffers::ConsoleBuffers()
	:
	outputBuffer(s_defaultOutputLineLimit)
{
	SetDefaultOutputType();
	commandHistoryLimit = s_defaultCommandHistoryLimit;
}


//====================================================================
void  ConsoleBuffers::SetMaxCommandsInHistory(uint numCommands)
{
	CriticalSectionSentry css(csCommandHistory);

	commandHistoryLimit = numCommands;
	while(commandHistory.size() > commandHistoryLimit) {
		commandHistory.pop_front();
	}
}


//====================================================================
void  ConsoleBuffers::ClearCommandHistory()
{
	CriticalSectionSentry css(csCommandHistory);
	commandHistory.clear();
}


//====================================================================
void  ConsoleBuffers::PushCommandInHistory(const std::string& s)
{
	if(s.empty()) return;

	CriticalSectionSentry css(csCommandHistory);
	commandHistory.push_back(s);
	if(commandHistory.size() > commandHistoryLimit) {
		commandHistory.pop_front();
	}
}


//====================================================================
const std::string&  ConsoleBuffers::GetCommandFromHistory(uint i) const
{
	VW_ASSERT(i < commandHistory.size(), "Input-command index out of range.");
	return commandHistory[i];
}


//********************************************************************
}//namespace vertexware


