

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\CompilerConfig.h"
#include "BaseTools\Types.h"
#include "BaseTools\Debug.h"
#include "BaseTools\SyncPrimitives.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\Stream.h"
#include "private\vw_common_exports.h"
#include <vector>
#include <deque>
#include <map>
#include <string>
#include <sstream>


//********************************************************************
namespace vertexware {


//====================================================================
// Line-based text-buffer.
// Keeps each line as a separate entry. Per-line output-type is
// available. All newline characters are stripped.
// MT-safe if Lock/Operation/Unlock sequences are used.
//====================================================================
class VW_COMMON_API ConsoleOutputBuffer
{
public:
	class VW_COMMON_API Observer {
	public:
		virtual ~Observer() {}
		virtual void OnChanged(const ConsoleOutputBuffer& cob) = 0;
	};


public:
	static const uint s_defaultLineLimit;


public:
	explicit ConsoleOutputBuffer(uint lineLimit = s_defaultLineLimit);
	virtual ~ConsoleOutputBuffer();

	//----------------------------------
	// MT-safe helpers.
	// Expected method sequence:
	//     Lock()
	//     Operation1(), Operation2()...
	//     Unlock()
	//----------------------------------
	void  Lock() const { cs.Enter(); }
	void  Unlock() const { cs.Leave(); }

	//---------------------------------------------
	// Request change notifications (auto-locking).
	//---------------------------------------------
	void  AddObserver(Observer& ob);
	void  RemoveObserver(Observer& ob);

	void  SetLineLimit(uint lineLimit);
	uint  GetLineLimit() const { return lineLimit; }

	void  SetTabWidth(uint tabWidth = 4);
	uint  GetTabWidth() const { return tabWidth; }

	//-----------------------------------------------------------------
	// Clear() deletes all lines and sets output-type to default value.
	//-----------------------------------------------------------------
	void  Clear();

	//------------------------------------------------------------------
	// If the new output-type is different from current output-type
	// and the current line is not empty, line-break is prefixed to <s>.
	//------------------------------------------------------------------
	void  Append(const std::string& s, uint outputType = 0);

	//------------------------
	// Retrieve buffered data.
	//------------------------
	uint  GetMaxColumns() const {
		return maxColumns;
	}
	uint  GetNumLines() const {
		return lines.size();
	}
	const std::string&  GetLine(uint index) const {
		VW_ASSERT(index < lines.size(), "Index out of range.");
		return lines[index].text;
	}
	const std::string&  GetLine(uint index, uint& outputType) const {
		VW_ASSERT(index < lines.size(), "Index out of range.");
		outputType = lines[index].outputType;
		return lines[index].text;
	}


private:
	//--------------
	// Storage unit.
	//--------------
	struct Line {
		std::string  text;
		uint   outputType;

		explicit Line(const std::string& txt = std::string(), uint ot = 0) :
			text(txt), outputType(ot) { }
	};


private:
	void  DiscardLines();
	void  BreakLine(uint outputType);
	void  AppendToLastLine(const std::string& s);

	void  AddColumnCount(uint colCount) { columnMap[colCount]++; }
	void  ReleaseColumnCount(uint colCount)
	{
		COLUMN_MAP_ITER it = columnMap.find(colCount);
		if((--(it->second)) == 0) columnMap.erase(it);
	}
	void  CacheMaxColumns() { maxColumns = columnMap.begin()->first; }

	void  NotifyObservers();


private:
	mutable CriticalSection  cs;
	std::vector<Observer*>   observers;

	uint  lineLimit;
	uint  columnLimit;
	uint  tabWidth;

	std::deque<Line> lines;
	uint             maxColumns;

	typedef  std::map<uint, uint, std::greater<uint> >  COLUMN_MAP;
	typedef  COLUMN_MAP::iterator  COLUMN_MAP_ITER;
	COLUMN_MAP  columnMap; //[column-count, line-count]
};


//====================================================================
// Console-output-stream.
// Provides familiar std::ostream interface for output operations.
// Multiple console-output-streams in multiple threads can share
// single console-output-buffer (MT-safe).
//====================================================================
class VW_COMMON_API ConsoleOutputStream : public OutputStream
{
public:
	explicit ConsoleOutputStream(ConsoleOutputBuffer& conOutBuffer,
										uint initialOutputType = 0)
		:
		OutputStream(&buffer),
		buffer(conOutBuffer, initialOutputType)
	{ }

	virtual void  SetCurrentOutputType(uint outputType) {
		if(buffer.currentOutputType != outputType) {
			flush(); //calls ConsoleOutputStreamBuffer::sync()
			buffer.currentOutputType = outputType;
		}
	}


private:
	//==============================================
	// Console-output-stream-buffer.
	// MT-safe indirection to console-output-buffer.
	//==============================================
	class ConsoleOutputStreamBuffer : public std::stringbuf {
	public:
		ConsoleOutputBuffer&  cob;
		uint  currentOutputType;
	public:
		explicit ConsoleOutputStreamBuffer(ConsoleOutputBuffer& _cob,
												uint initialOutputType)
			: cob(_cob), currentOutputType(initialOutputType)
		{ }
	protected:
		// std::stringbuf override
		virtual int sync() {
			cob.Lock();
			cob.Append(str(), currentOutputType);
			cob.Unlock();
			str(std::string());
			return 0; //success
		}
	}
	buffer;


protected:
	// std::ios override
	virtual ConsoleOutputStreamBuffer* rdbuf() const {
		return const_cast<ConsoleOutputStreamBuffer*>(&buffer);
	}
};


//====================================================================
// Console-buffers.
//====================================================================
class VW_COMMON_API ConsoleBuffers
{
public:
	static const uint s_defaultCommandHistoryLimit;
	static const uint s_defaultOutputLineLimit;


public:
	explicit ConsoleBuffers();

	//-------
	// Input.
	//-------
	void  SetMaxCommandsInHistory(uint numCommands = ConsoleBuffers::s_defaultCommandHistoryLimit);
	uint  GetMaxCommandsInHistory() const { return commandHistoryLimit; }

	void  ClearCommandHistory();

	// stored only if not empty
	void  PushCommandInHistory(const std::string& s);

	// Command history access: most recent command has largest index
	// Expected call sequence:
	// Lock(), Operation1(), Operation2(), ..., Unlock()
	void                LockCommandHistory() const { csCommandHistory.Enter(); }
	uint                GetNumCommandsInHistory() const { return commandHistory.size(); }
	const std::string&  GetCommandFromHistory(uint i) const;
	void                UnlockCommandHistory() const { csCommandHistory.Leave(); }

	//--------
	// Output.
	//--------
	void  SetMaxOutputLines(uint numLines = ConsoleBuffers::s_defaultOutputLineLimit) { outputBuffer.SetLineLimit(numLines); }
	uint  GetMaxOutputLines() const { return outputBuffer.GetLineLimit(); }

	void  SetOutputTabWidth(uint tabWidth = 4) { outputBuffer.SetTabWidth(tabWidth); }
	uint  GetOutputTabWidth() const { return outputBuffer.GetTabWidth(); }

	void  ClearOutput() { outputBuffer.Clear(); }

	virtual  void         SetDefaultOutputType(uint outputType = 0) { defaultOutputType = outputType; }
	ConsoleOutputBuffer&  GetOutputBuffer() { return outputBuffer; }

	//--------------------------------------------------------------
	// Create new output-stream hooked to console-output-buffer.
	// Example usage:
	//     Console console;
	//     ConsoleOutputStream* cos1 = console.CreateOutputStream();
	//     *cos1 << "hello to buffer from stream 1";
	//     delete cos1;
	//--------------------------------------------------------------
	__forceinline  ConsoleOutputStream*  CreateOutputStream() {
		return new ConsoleOutputStream(outputBuffer, defaultOutputType);
	}


private:
	ConsoleOutputBuffer  outputBuffer;
	uint                 defaultOutputType;

	mutable CriticalSection  csCommandHistory;
	std::deque<std::string>  commandHistory;
	uint                     commandHistoryLimit;
};


//********************************************************************
}//namespace vertexware

