

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Debug.h"
#include "private\vw_common_exports.h"


//********************************************************************
namespace vertexware {


//====================================================================
template<typename TDst, typename TSrc>
__forceinline  TDst*  __fastcall  PointerCast(TSrc* pSrc)
{
	return const_cast<TDst*>(reinterpret_cast<const TDst*>(pSrc));
}


//====================================================================
template<typename TDst, typename TSrc>
struct DynamicCastPtr
{
	TDst* ptr;

	__forceinline operator TDst*() { return ptr; }
	__forceinline TDst* operator->() { return ptr; }
	__forceinline TDst& operator*() { return *ptr; }

	__forceinline  DynamicCastPtr(TSrc* src)
	{
		ptr = dynamic_cast<TDst*>(src);
	}
	__forceinline  DynamicCastPtr(TSrc& src)
	{
		ptr = dynamic_cast<TDst*>(&src);
	}
};


//====================================================================
template<typename TDst, typename TSrc>
struct ConstDynamicCastPtr
{
	const TDst* ptr;

	__forceinline operator const TDst*() { return ptr; }
	__forceinline const TDst* operator->() { return ptr; }
	__forceinline const TDst& operator*() { return *ptr; }

	__forceinline  ConstDynamicCastPtr(const TSrc* src)
	{
		ptr = dynamic_cast<const TDst*>(src);
	}
	__forceinline  ConstDynamicCastPtr(const TSrc& src)
	{
		ptr = dynamic_cast<const TDst*>(&src);
	}
};


//====================================================================
// Used for assignment to function's output-parameters of pointer type
// which are optional.
//====================================================================
template<typename T>
__forceinline  void __fastcall SafeAssign(T* dst, const T& src)
{
	if(dst) *dst = src;
}


//====================================================================
// Used to eliminate compiler performance warning (conversion to bool).
//====================================================================
template<typename T>
__forceinline  bool __fastcall ToBool(const T& t)
{
	return (t ? true : false);
}


//====================================================================
// Variable sign flip.
//====================================================================
template <typename T>
__forceinline  void __fastcall FlipSign(T& val) {
	val = - val;
}


//====================================================================
// Boolean variable flip.
//====================================================================
__forceinline  bool& __fastcall Flip(bool& a)
{
	a = ! a;
	return a;
}


//====================================================================
// Change active value.
//====================================================================
template<typename T>
__forceinline  void __fastcall ChangeActiveValue(T& activeValue, const T& value1, const T& value2)
{
	activeValue = (activeValue != value1) ? value1 : value2;
}


//====================================================================
// Packed-flag helpers.
//====================================================================
__forceinline  void  SetFlagOn(uint& flags, uint f)
{
	flags |= f;
}


//====================================================================
__forceinline  void  SetFlagOff(uint& flags, uint f)
{
	flags &= ~f;
}


//====================================================================
__forceinline  void  SetFlagOnOff(uint& flags, uint f, bool bOn)
{
	bOn ? SetFlagOn(flags, f) : SetFlagOff(flags, f);
}


//====================================================================
__forceinline  uint  IsFlagOn(uint flags, uint f)
{
	return (flags & f);
}


//====================================================================
__forceinline  uint  IsFlagOff(uint flags, uint f)
{
	return ! IsFlagOn(flags, f);
}


//====================================================================
// Variable swaps.
//====================================================================
template<typename T>
__forceinline  void __fastcall Swap(T& a, T&b)
{
	T  t = a;
	a    = b;
	b    = t;
}


//====================================================================
template<typename T>
__forceinline  void  __fastcall Swap(T& a, T&b, T& temp)
{
	temp = a;
	a    = b;
	b    = temp;
}


//====================================================================
// Order flips.
//====================================================================
template<typename T>
__forceinline  void __fastcall FlipOrder2(T* p)
{
	Swap(*p, *(p + 1));
}


//====================================================================
template<typename T>
__forceinline  void __fastcall CopyWithFlippedOrder2(T* dst, const T* src)
{
	dst[0] = src[1];
	dst[1] = src[0];
}


//====================================================================
template<typename T>
__forceinline  void __fastcall FlipOrder4(T* p)
{
	T  temp;
	Swap(*(p + 0), *(p + 3), temp);
	Swap(*(p + 1), *(p + 2), temp);
}


//====================================================================
template<typename T>
__forceinline  void __fastcall CopyWithFlippedOrder4(T* dst, const T* src)
{
	dst[0] = src[3];
	dst[1] = src[2];
	dst[2] = src[1];
	dst[3] = src[0];
}


//====================================================================
template<typename T>
__forceinline  void __fastcall FlipOrder8(T* p)
{
	T  temp;
	Swap(*(p + 0), *(p + 7), temp);
	Swap(*(p + 1), *(p + 6), temp);
	Swap(*(p + 2), *(p + 5), temp);
	Swap(*(p + 3), *(p + 4), temp);
}


//====================================================================
template<typename T>
__forceinline  void __fastcall CopyWithFlippedOrder8(T* dst, const T* src)
{
	dst[0] = src[7];
	dst[1] = src[6];
	dst[2] = src[5];
	dst[3] = src[4];
	dst[4] = src[3];
	dst[5] = src[2];
	dst[6] = src[1];
	dst[7] = src[0];
}


//====================================================================
template<typename T>
__forceinline  void __fastcall FlipOrderN(T* p, uint count)
{
	uint halfCount = count / 2;
	T* pRev = p + (count - 1);
	for(uint i = 0; i < halfCount; i++) {
		Swap(*(p + i), *(pRev - i));
	}
}


//====================================================================
// Sign / Abs / Minimum / Maximum
//====================================================================
template <typename T>
__forceinline  T __fastcall Sign(const T& val) {
	if(val > 0) return  1;
	if(val < 0) return -1;
	return 0;
}


//====================================================================
template<typename T>
__forceinline  T __fastcall Abs(const T& val) {
	return (val >= 0) ? val : - val;
}


//====================================================================
template<typename T>
__forceinline const T& __fastcall Min(const T& val1, const T& val2)
{
	return (val1 <= val2) ? val1 : val2;
}


//====================================================================
template<typename T>
__forceinline const T& __fastcall Min(const T& val1, const T& val2, const T& val3)
{
	return Min(Min(val1, val2), val3);
}


//====================================================================
template<typename T>
__forceinline const T& __fastcall Max(const T& val1, const T& val2)
{
	return (val1 >= val2) ? val1 : val2;
}


//====================================================================
template<typename T>
__forceinline const T& __fastcall Max(const T& val1, const T& val2, const T& val3)
{
	return Max(Max(val1, val2), val3);
}


//====================================================================
// Linear interpolation, a = [0, 1]
// val = val1 + (val2 - val1) * a
//====================================================================
template<typename TInterpolated, typename TInterpolator>
__forceinline  TInterpolated __fastcall
InterpolateLinear(const TInterpolated& val1, const TInterpolated& val2, TInterpolator a)
{
	return static_cast<TInterpolated>(val1 + (val2 - val1) * a);
}


//====================================================================
// Internal float to integer conversion performs round-down.
//====================================================================
template<typename TInteger, typename TFloatingPoint>
__forceinline  TInteger __fastcall Round(TFloatingPoint val)
{
	return static_cast<TInteger>(val + TFloatingPoint(0.5));
}


//====================================================================
// Value-range related.
//====================================================================
template<typename T>
__forceinline  bool  __fastcall InRange(const T& val, T lowLimit, T highLimit)
{
	if(val >= lowLimit  &&  val <= highLimit) return true;
	return false;
}


//====================================================================
template<typename TInteger, typename TFloatingNormalizer = float>
inline
TInteger __fastcall RangeMapInt(const TInteger& val, TInteger rangeMin, TInteger rangeMax,
								TInteger newRangeMin, TInteger newRangeMax)
{
	TInteger  range = rangeMax - rangeMin;
	if(range == TInteger(0)) return newRangeMin;
	TFloatingNormalizer normalizedVal = TFloatingNormalizer(val - rangeMin) / TFloatingNormalizer(range);
	return Round<TInteger>(normalizedVal * (newRangeMax - newRangeMin) + newRangeMin);
}


//====================================================================
// For RangeMax parameters, specify the actual maximum value
// (e.g. for 8-bit range specify 255 as RangeMax).
//====================================================================
template<typename TUIntLow, uint LowPrecisionRangeMax,
		typename TUIntHigh, uint HighPrecisionRangeMax,
		typename TFloatingNormalizer = float>
struct RangeMapperUInt
{
	__forceinline static TUIntLow __fastcall MapHighToLow(TUIntHigh highPrecisionRangeVal)
	{
		return Round<TUIntLow>((TFloatingNormalizer(LowPrecisionRangeMax) *
		(TFloatingNormalizer(highPrecisionRangeVal) / TFloatingNormalizer(HighPrecisionRangeMax))));
	}

	__forceinline static TUIntHigh __fastcall MapLowToHigh(TUIntLow lowPrecisionRangeVal)
	{
		return Round<TUIntHigh>((TFloatingNormalizer(HighPrecisionRangeMax) *
		(TFloatingNormalizer(lowPrecisionRangeVal) / TFloatingNormalizer(LowPrecisionRangeMax))));
	}
};


//====================================================================
template<typename T>
__forceinline  void  __fastcall ClampLow(T& val, T lowLimit)
{
	if(val < lowLimit) val = lowLimit;
}


//====================================================================
template<typename T>
__forceinline  const T __fastcall  ClampedLow(T val, T lowLimit)
{
	if(val <= lowLimit) return lowLimit;
	return val;
}


//====================================================================
template<typename T>
__forceinline  void __fastcall  ClampHigh(T& val, T highLimit)
{
	if(val > highLimit) val = highLimit;
}


//====================================================================
template<typename T>
__forceinline  const T __fastcall  ClampedHigh(T val, T highLimit)
{
	if(val >= highLimit) return highLimit;
	return val;
}


//====================================================================
template<typename T>
__forceinline  void __fastcall  Clamp(T& val, T lowLimit, T highLimit)
{
	if     (val < lowLimit)  val = lowLimit;
	else if(val > highLimit) val = highLimit;
}


//====================================================================
template<typename T>
__forceinline  const T __fastcall  Clamped(T val, T lowLimit, T highLimit)
{
	if     (val <= lowLimit)  return lowLimit;
	else if(val >= highLimit) return highLimit;
	return  val;
}


//====================================================================
// Snaps.
// Low  versions snap to low  boundary if differences are equal.
// High versions snap to high boundary if differences are equal.
//====================================================================
template<typename T>
__forceinline  void __fastcall  GetSnapPoints(T val, T gridSize, T gridOffset,
										T& outLowSnapPoint, T& outHighSnapPoint)
{
	T  gridDistance  = (val - gridOffset) % gridSize;
	outLowSnapPoint  = val - gridDistance;
	outHighSnapPoint = val + (gridSize - gridDistance);
}


//====================================================================
template<typename T>
__forceinline  T __fastcall  SnapLow(T val, T lowSnapPoint, T highSnapPoint)
{
	if(Abs(val - lowSnapPoint) <= Abs(highSnapPoint - val)) return lowSnapPoint;
	return highSnapPoint;
}


//====================================================================
template<typename T>
__forceinline  T __fastcall  SnapHigh(T val, T lowSnapPoint, T highSnapPoint)
{
	if(Abs(val - lowSnapPoint) < Abs(highSnapPoint - val)) return lowSnapPoint;
	return highSnapPoint;
}


//====================================================================
// Arbitrary step increments and decrements.
//====================================================================
template<typename TVAL, typename TINC>
__forceinline  TVAL& __fastcall PreIncrement(TVAL& val, TINC inc)
{
	val += inc;
	return  val;
}


//====================================================================
template<typename TVAL, typename TINC>
__forceinline  TVAL __fastcall PostIncrement(TVAL& val, TINC inc)
{
	TVAL  last = val;
	val += inc;
	return  last;
}


//====================================================================
template<typename TVAL, typename TDEC>
__forceinline  TVAL& __fastcall PreDecrement(TVAL& val, TDEC dec)
{
	val -= dec;
	return  val;
}


//====================================================================
template<typename TVAL, typename TDEC>
__forceinline  TVAL __fastcall PostDecrement(TVAL& val, TDEC dec)
{
	TVAL  last = val;
	val -= dec;
	return  last;
}


//====================================================================
// Basic loops.
//====================================================================
template<typename T>
__forceinline  void __fastcall Copy(T* dst, const T* src, uint count)
{
	VW_ASSERT(dst  &&  src, "Invalid argument.");

	const T* dstEnd = dst + count;
	while(dst < dstEnd) {
		*dst++ = *src++;
	}
}


//====================================================================
template<typename T>
__forceinline  void __fastcall ReverseCopy(T* dst, const T* src, const uint count)
{
	VW_ASSERT(dst  &&  src, "Invalid argument.");

	src += (count - 1);
	const T* dstEnd = dst + count;
	while(dst < dstEnd) {
		*dst++ = *src--;
	}
}


//====================================================================
template<typename T>
__forceinline  void __fastcall Fill(T* dst, const T& val, uint count)
{
	VW_ASSERT(dst, "Invalid argument.");

	const T* dstEnd = dst + count;
	while(dst < dstEnd) {
		*dst++ = val;
	}
}


//====================================================================
// Tests if value is present in array.
// Array must be zero-terminated.
//====================================================================
template<typename T>
__forceinline  bool __fastcall IsValueInArray(const T& value, const T* array)
{
	while(T arrayValue = *array++) {
		if(arrayValue == value) return true;
	}
	return false;
}


//********************************************************************
}//namespace vertexware


