

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Exception.h"
#include "BaseTools\SyncPrimitives.h"
#include "private\vw_common_exports.h"
#include <string>


//********************************************************************
namespace vertexware {


//====================================================================
VW_COMMON_API
bool __fastcall CheckBuiltInTypeSizes(std::ostream& os);


//====================================================================
// Source-code-location string construction.
//====================================================================
VW_COMMON_API
std::string __fastcall FormatSourceCodeLocation(const std::string& file, uint line,
												bool bForceWinKernelName = false);


//====================================================================
// Class-name & object-name construction.
//====================================================================
VW_COMMON_API
std::string __fastcall FormatClassName(const std::string& nameFromTypeInfo);


//====================================================================
VW_COMMON_API
std::string __fastcall FormatObjectName(const std::string& nameFromTypeInfo,
										const void* objectAddress);


//====================================================================
#define  VW_GET_CLASS_NAME(ObjectReferenceOrClass) \
	(vertexware::FormatClassName(typeid(ObjectReferenceOrClass).name()))


//====================================================================
#define  VW_GET_OBJECT_NAME(ObjectReference) \
	(vertexware::FormatObjectName(typeid(ObjectReference).name(), &ObjectReference))


//====================================================================
// Shows debug dialog with the following options:
//     abort  ===> returns true
//     retry  ===> returns false
//     ignore ===> returns false & sets <ignoreAlways> to true
//====================================================================
VW_COMMON_API
bool __fastcall ShowDialogAbortRetryIgnore(bool& outIgnoreAlways,
						uint defaultButton,
						const std::string& caption, const std::string& message,
						const std::string& file, uint line);


//====================================================================
// Shows debug dialog with the following options:
//     cancel ===> returns true
//     ok     ===> returns false
//====================================================================
VW_COMMON_API
bool __fastcall ShowDialogCancelOk(uint defaultButton,
						const std::string& caption, const std::string& message);


//====================================================================
// Shows debug dialog with no options.
//====================================================================
VW_COMMON_API
void __fastcall ShowDialogOk(const std::string& caption, const std::string& message);


//====================================================================
// Expression check.
// Choosing abort breaks into debugger.
// Ignored in release builds.
//====================================================================
#ifdef VW_BUILD_DEBUG
#define  VW_ASSERT(expression, message) \
	if(!(expression)) { \
		static bool ignoreAlways = false; \
		if(vertexware::ShowDialogAbortRetryIgnore(ignoreAlways, \
						1, "Assert Failed", message, __FILE__, __LINE__)) \
		{ __asm int 3 } \
	}
#else//VW_BUILD_RELEASE
#define  VW_ASSERT(expression, message)
#endif


//====================================================================
// Expression check with external ignore-flag.
// Set <ignoreFlagBool> to false at init.
// Choosing abort breaks into debugger.
// Ignored in release builds.
//====================================================================
#ifdef VW_BUILD_DEBUG
#define  VW_ASSERT_EXTERNAL_IGNORE_FLAG(expression, message, ignoreFlagBool) \
	if(!(expression)) { \
		if(vertexware::ShowDialogAbortRetryIgnore(ignoreFlagBool, \
						1, "Assert Failed", message, __FILE__, __LINE__)) \
		{ __asm int 3 } \
	}
#else//VW_BUILD_RELEASE
#define  VW_ASSERT_EXTERNAL_IGNORE_FLAG(expression, message, ignoreFlagBool)
#endif


//====================================================================
// Expression check.
// Choosing abort breaks into debugger.
// Works in release builds.
//====================================================================
#define  VW_VERIFY(expression, message) \
	if(!(expression)) { \
		static bool ignoreAlways = false; \
		if(vertexware::ShowDialogAbortRetryIgnore(ignoreAlways, \
						1, "Verify Failed", message, __FILE__, __LINE__)) \
		{ __asm int 3 } \
	}


//====================================================================
// Re-entrant assert. Returns from function if busy.
//====================================================================
#ifdef VW_BUILD_DEBUG
#define  VW_REENTRANT_ASSERT_RETURNS_VOID(expression, message) \
	{	static vertexware::Event ev(true, false, "ReentrantAssertEvent_" + \
						vertexware::FormatSourceCodeLocation(__FILE__, __LINE__, true)); \
		bool bTimeout; \
		if(! ev.Wait(0, bTimeout) || bTimeout) return; \
		VW_ASSERT(expression, message); \
		ev.Set(); \
	}
#else//VW_BUILD_RELEASE
#define  VW_REENTRANT_ASSERT_RETURNS_VOID(expression, message)
#endif


//====================================================================
// Re-entrant assert. Returns value from function if busy.
//====================================================================
#ifdef VW_BUILD_DEBUG
#define  VW_REENTRANT_ASSERT_RETURNS_VALUE(expression, message, returnValue) \
	{	static vertexware::Event ev(true, false, "ReentrantAssertEvent_" + \
						vertexware::FormatSourceCodeLocation(__FILE__, __LINE__, true)); \
		bool bTimeout; \
		if(! ev.Wait(0, bTimeout) || bTimeout) return(returnValue); \
		VW_ASSERT(expression, message); \
		ev.Set(); \
	}
#else//VW_BUILD_RELEASE
#define  VW_REENTRANT_ASSERT_RETURNS_VALUE(expression, message, returnValue)
#endif


//====================================================================
// Shows exception info.
//====================================================================
__forceinline  void __fastcall ShowException(Exception& e) {
	ShowDialogOk("Exception Caught", e.MakeDescription());
}


//====================================================================
// Automatically prepend class-name to exception-message.
//====================================================================
#define  VW_THROW_CN(ExceptionClass, Message) \
	VW_THROW(ExceptionClass, (VW_GET_CLASS_NAME(*this) + std::string(": ") + Message))


//********************************************************************
}//namespace vertexware

