

//--------------------------------------------------------------------
#include "BaseTools\DynamicLinkLibrary.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\FileSystem.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"
#include "BaseTools\Debug.h"


//********************************************************************
namespace vertexware {


//====================================================================
DLLImportEntry::DLLImportEntry(const std::string& _symbol, void* _pHook)
	:
	symbol(_symbol)
{
	pHook = static_cast<void**>(_pHook);
}


//====================================================================
bool  DLLImportEntry::ImportFrom(const DynamicLinkLibrary& dll)
{
	VW_ASSERT(pHook, "Invalid argument.");
	VW_ASSERT(symbol.length(), "Invalid argument.");

	return (DLLGetSymbolAddress(dll, symbol, *pHook) != 0);
}


//====================================================================
bool  DLLImportTable::ImportSymbolsFrom(const DynamicLinkLibrary& dll,
										uint* pOutNumImportedSymbols)
{
	uint numImported = 0;

	ENTRY_LIST_ITER it = entries.begin();
	while(it != entries.end()) {
		DLLImportEntry& e = *it++;

		if(e.ImportFrom(dll)) {
			numImported ++;
		}
	}
	SafeAssign(pOutNumImportedSymbols, numImported);
	return (numImported == entries.size());
}


//====================================================================
void  DLLImportTable::InvalidateSymbolHooks()
{
	ENTRY_LIST_ITER it = entries.begin();
	while(it != entries.end()) {
		DLLImportEntry& e = *it++;

		e.InvalidateHook();
	}
}


//====================================================================
void  DLLImportTable::Dump(std::ostream& os)
{
	uint numImported = 0;

	ENTRY_LIST_ITER it = entries.begin();
	while(it != entries.end()) {
		DLLImportEntry& e = *it++;

		if(e.GetHookValue()) {
			numImported ++;
			os << ToString(e.GetHookValue());
		}
		else {
			os << fill(10, ' ');
		}
		os << "  |  " << e.GetSymbol() << std::endl;
	}
	os << "Imported " << numImported << " of " << GetNumEntries() << " symbols." << std::endl;
}


//====================================================================
DynamicLinkLibrary::DynamicLinkLibrary()
{
	hModule = 0;
	bImportTableCleared = true;

	bAutoUnload = true;
}


//====================================================================
DynamicLinkLibrary::DynamicLinkLibrary(const std::string& _fileName)
{
	fileName = _fileName;
	file_system::SplitFileName(fileName, pathName, baseName, extension);

	hModule = 0;
	bImportTableCleared = true;

	bAutoUnload = true;
}


//====================================================================
DynamicLinkLibrary::~DynamicLinkLibrary()
{
	if(bAutoUnload) {
		Unload();
	}
}


//====================================================================
void   DynamicLinkLibrary::SafeCopyFrom(const DynamicLinkLibrary& dll)
{
	Unload();

	// copy data members
	fileName  = dll.fileName;
	pathName  = dll.pathName;
	baseName  = dll.baseName;
	extension = dll.extension;
	hModule   = dll.hModule;

	bImportTableCleared = dll.bImportTableCleared;
	importTable = dll.importTable;

	// set data-members
	bAutoUnload = true;

	// increment reference-count if necessary
	if(dll.IsLoaded()) {
		::LoadLibrary(fileName.c_str());
	}
}


//====================================================================
std::string  DynamicLinkLibrary::GetFormattedName(bool bAddExtension) const
{
	std::string s("[");
	s.append(GetBaseName());
	if(bAddExtension) s.append(".").append(GetExtension());
	s.append("]");
	return s;
}


//====================================================================
bool  DynamicLinkLibrary::IsLoaded() const
{
	return (hModule != 0);
}


//====================================================================
bool  DynamicLinkLibrary::Load()
{
	if(hModule) return true;

	hModule = ::LoadLibrary(fileName.c_str());
	if(hModule) {
		PostLoad();
		return true;
	}
	return false;
}


//====================================================================
bool  DynamicLinkLibrary::Unload()
{
	if(hModule) {
		PreUnload();
		InvalidateSymbolHooks();
		BOOL bUnloaded = ::FreeLibrary(hModule);
		hModule = 0;
		return (bUnloaded == TRUE);
	}
	return true;
}


//====================================================================
void  DynamicLinkLibrary::SetAutoUnload(bool bEnable)
{
	bAutoUnload = bEnable;
}


//====================================================================
void*  DynamicLinkLibrary::GetSymbolAddress(const std::string& symbolName) const
{
	if(! hModule) return 0;
	void* pSymbol = ::GetProcAddress(hModule, symbolName.c_str());
	return pSymbol;
}


//====================================================================
bool  DynamicLinkLibrary::ImportSymbols(uint* pOutNumImportedSymbols)
{
	VW_ASSERT(IsLoaded(), "DLL is not loaded.");

	uint numImportedSymbols = 0;

	FillImportTable();
	bool bAllImported = importTable.ImportSymbolsFrom(*this, &numImportedSymbols);
	PostImport(bAllImported, numImportedSymbols);

	SafeAssign(pOutNumImportedSymbols, numImportedSymbols);
	return bAllImported;
}


//====================================================================
void  DynamicLinkLibrary::InvalidateSymbolHooks()
{
	FillImportTable();
	importTable.InvalidateSymbolHooks();
}


//====================================================================
void  DynamicLinkLibrary::DumpImportTable(std::ostream& os)
{
	FillImportTable();
	importTable.Dump(os);
}


//====================================================================
void  DynamicLinkLibrary::ClearImportTable()
{
	bImportTableCleared = true;
	importTable.Clear();
}


//====================================================================
void  DynamicLinkLibrary::FillImportTable()
{
	if(bImportTableCleared) {
		bImportTableCleared = false;
		FillImportTable(importTable);
	}
}


//********************************************************************
}//namespace vertexware


