

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include <string>
#include <list>
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class DynamicLinkLibrary;


//====================================================================
class VW_COMMON_API DLLImportEntry
{
public:
	//------------------------------------------------------------
	// <pHook> is not declared as void** because an explicit cast
	// from pointer-to-variable-pointer to pointer-to-void-pointer
	// would be required when writing each import entry, which
	// is inconvenient.
	//
	// WARNING: Be sure to pass the adress-of-hook-variable,
	//          not the address-the-hook-variable-is-pointing-to.
	//
	// Example:
	//     typedef  int(* PFN_MY_FUNCTION)();
	//     PFN_MY_FUNCTION  myHook;
	//
	//     // causes access violation
	//     DLLImportEntry entry("myFunction", myHook);
	//
	//     // the right way (note the additional &).
	//     DLLImportEntry entry("myFunction", &myHook);
	//------------------------------------------------------------
	DLLImportEntry(const std::string& symbol, void* pHook);

	const std::string&  GetSymbol() const { return symbol; }
	void*               GetHookValue() const { return *pHook; }

	bool  ImportFrom(const DynamicLinkLibrary& dll);
	void  InvalidateHook() { *pHook = 0; }


private:
	std::string  symbol;
	void**       pHook;
};


//====================================================================
class VW_COMMON_API DLLImportTable
{
public:
	void  Clear() { entries.clear(); }

	void  Add(const DLLImportEntry& entry) { entries.push_back(entry); }

	void  Add(const std::string& symbol, void* pHook) { entries.push_back(DLLImportEntry(symbol, pHook)); }

	uint  GetNumEntries() { return entries.size(); }

	//-------------------------------------------------------------
	// Imports all available symbols. If symbol is unavailable,
	// appropriate hook will be set to 0.
	// Returns false if the number of successfully imported symbols
	// is less than total number of table entries.
	// Only failures are logged.
	//-------------------------------------------------------------
	bool  ImportSymbolsFrom(const DynamicLinkLibrary& dll, uint* pOutNumImportedSymbols = 0);
	void  InvalidateSymbolHooks();

	void  Dump(std::ostream& os);


private:
	typedef  std::list<DLLImportEntry>   ENTRY_LIST;
	typedef  ENTRY_LIST::iterator        ENTRY_LIST_ITER;
	typedef  ENTRY_LIST::const_iterator  ENTRY_LIST_CITER;

	ENTRY_LIST  entries;
};


//====================================================================
class VW_COMMON_API DynamicLinkLibrary
{
public:
	DynamicLinkLibrary(const std::string& fileName);
	virtual ~DynamicLinkLibrary();

	DynamicLinkLibrary(const DynamicLinkLibrary& dll) { SafeCopyFrom(dll); }
	DynamicLinkLibrary& operator=(const DynamicLinkLibrary& dll) { SafeCopyFrom(dll); return *this; }

	//------------------------------------------
	// Unloads current DLL.
	// Increments reference-count of cloned DLL.
	// Sets auto-unload to true.
	// Clones other information.
	//------------------------------------------
	void   SafeCopyFrom(const DynamicLinkLibrary& dll);

	const std::string&  GetFileName() const { return fileName; }
	const std::string&  GetPathName() const { return pathName; }
	const std::string&  GetBaseName() const { return baseName; }
	const std::string&  GetExtension() const { return extension; }

	std::string         GetFormattedName(bool bAddExtension = false) const;

	HMODULE             GetHandle() const { return hModule; }
	bool                IsLoaded() const;
	bool                Load();
	bool                Unload();

	//------------------------------------------------
	// Disable when handling exceptions thrown by DLL.
	//------------------------------------------------
	void   SetAutoUnload(bool bEnable);

	void*  GetSymbolAddress(const std::string& symbolName) const;

	bool   ImportSymbols(uint* pOutNumImportedSymbols = 0);
	void   InvalidateSymbolHooks();

	void   DumpImportTable(std::ostream& os);
	void   ClearImportTable();


protected:
	//-------------------------
	// Derived-class interface.
	//-------------------------
	DynamicLinkLibrary();

	//------------------
	// Template methods.
	//------------------
	virtual  void  PostLoad() {}
	virtual  void  PreUnload() {}
	virtual  void  FillImportTable(DLLImportTable& table) {}
	virtual  void  PostImport(bool bAllImported, uint numImportedSymbols) {}


private:
	void  FillImportTable();


private:
	// cloned data
	std::string   fileName;
	std::string   pathName;
	std::string   baseName;
	std::string   extension;
	HMODULE       hModule;

	bool            bImportTableCleared;
	DLLImportTable  importTable;

	// non-cloned data
	bool  bAutoUnload;
};


//====================================================================
// DLL-function access with automatic type-casting.
//====================================================================
template<typename T>
__forceinline T __fastcall
DLLGetSymbolAddress(const DynamicLinkLibrary& dll, const std::string& symbolName, T& symbolHook)
{
	symbolHook = static_cast<T>( dll.GetSymbolAddress(symbolName) );
	return symbolHook;
}


//====================================================================
template<typename T>
__forceinline T __fastcall
DLLGetSymbolAddress(const DynamicLinkLibrary* pdll, const std::string& symbolName, T& symbolHook)
{
	if(pdll) {
		symbolHook = static_cast<T>( pdll->GetSymbolAddress(symbolName) );
	}
	else {
		symbolHook = 0;
	}
	return symbolHook;
}


//********************************************************************
}//namespace vertexware


