

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Constructs.h"
#include "private\vw_common_exports.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Element processor templates.
// <loopDepth> is 0-based.
// Virtuals provide more convenient derivation. Example:
//     struct MyProcessor : public ElementProcessorInplace<TElement> {
//         // override desired methods only
//     };
// Since ProcessElements*() family receives element-processor class
// as a template argument, all methods will be called directly.
// Default empty operations are inlined, so they will compile to
// 0 assembly instructions.
// However, processor-class does not have to derive from
// base-processor-class, just define all necessary methods.
//====================================================================
template<typename TElement>
struct ElementProcessorInplace
{
	typedef  TElement  ElementType;

	__forceinline  virtual void  PreLoop(uint loopDepth) {}
	__forceinline  virtual void  Process(TElement*) {}
	__forceinline  virtual void  PostLoop(uint loopDepth) {}
};


//====================================================================
template<typename TDstElement, typename TSrcElement>
struct ElementProcessorDstSrc
{
	typedef  TDstElement  DstElementType;
	typedef  TSrcElement  SrcElementType;

	__forceinline  virtual void  PreLoop(uint loopDepth) {}
	__forceinline  virtual void  Process(TDstElement*, TSrcElement*) {}
	__forceinline  virtual void  PostLoop(uint loopDepth) {}
};


//====================================================================
// ProcessElements*() family of functions are used for looping
// through memory layouts. Each element is processed by user-supplied
// processor-object. Strides can be negative.
//====================================================================
template<typename TElement, typename TProcessorInplace>
inline
void __fastcall ProcessElements1D(
					TProcessorInplace& processor,
					uint numElements,
					TElement* pElements, int elemStride)
{
	VW_ASSERT(pElements, "Invalid argument.");

	TElement*  pElementsEnd = pElements + elemStride * numElements;

	processor.PreLoop(0);
	while(pElements != pElementsEnd) {

		processor.Process(PointerCast<TProcessorInplace::ElementType>(
								PostIncrement(pElements, elemStride)));
	}
	processor.PostLoop(0);
}


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TProcessorDstSrc>
inline
void __fastcall ProcessElements1D(
					TProcessorDstSrc& processor,
					uint numElements,
					TDstElement* pDst, int dstElemStride,
					TSrcElement* pSrc, int srcElemStride)
{
	VW_ASSERT(pDst && pSrc, "Invalid argument.");

	TDstElement*  pDstEnd = pDst + dstElemStride * numElements;

	processor.PreLoop(0);
	while(pDst != pDstEnd) {

		processor.Process(PointerCast<TProcessorDstSrc::DstElementType>(
								PostIncrement(pDst, dstElemStride)),
							PointerCast<TProcessorDstSrc::SrcElementType>(
								PostIncrement(pSrc, srcElemStride)));
	}
	processor.PostLoop(0);
}


//====================================================================
template<typename TElement, typename TProcessorInplace>
inline
void __fastcall ProcessElements2D(
					TProcessorInplace& processor,
					uint numRows, uint numColumns,
					TElement* pElements, int rowStride, int columnStride)
{
	VW_ASSERT(pElements, "Invalid argument.");

	TElement*  pRow        = pElements;
	TElement*  pRowEnd     = pRow + rowStride * numRows;
	int        skipColumns = columnStride * numColumns;

	processor.PreLoop(0);
	while(pRow != pRowEnd) {

		TElement*  pColumn    = PostIncrement(pRow, rowStride);
		TElement*  pColumnEnd = pColumn + skipColumns;

		processor.PreLoop(1);
		while(pColumn != pColumnEnd) {

			processor.Process(PointerCast<TProcessorInplace::ElementType>(
									PostIncrement(pColumn, columnStride)));
		}
		processor.PostLoop(1);
	}
	processor.PostLoop(0);
}


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TProcessorDstSrc>
inline
void __fastcall ProcessElements2D(
					TProcessorDstSrc& processor,
					uint numRows, uint numColumns,
					TDstElement* pDst, int dstRowStride, int dstColumnStride,
					TSrcElement* pSrc, int srcRowStride, int srcColumnStride)
{
	VW_ASSERT(pDst && pSrc, "Invalid argument.");

	TDstElement*  pDstRow        = pDst;
	TDstElement*  pDstRowEnd     = pDstRow + dstRowStride * numRows;
	int           dstSkipColumns = dstColumnStride * numColumns;
	TSrcElement*  pSrcRow        = pSrc;

	processor.PreLoop(0);
	while(pDstRow != pDstRowEnd) {

		TDstElement*  pDstColumn    = PostIncrement(pDstRow, dstRowStride);
		TDstElement*  pDstColumnEnd = pDstColumn + dstSkipColumns;
		TSrcElement*  pSrcColumn    = PostIncrement(pSrcRow, srcRowStride);

		processor.PreLoop(1);
		while(pDstColumn != pDstColumnEnd) {

			processor.Process(PointerCast<TProcessorDstSrc::DstElementType>(
									PostIncrement(pDstColumn, dstColumnStride)),
								PointerCast<TProcessorDstSrc::SrcElementType>(
									PostIncrement(pSrcColumn, srcColumnStride)));
		}
		processor.PostLoop(1);
	}
	processor.PostLoop(0);
}


//====================================================================
template<typename TElement, typename TProcessorInplace>
inline
void __fastcall ProcessElements3D(
					TProcessorInplace& processor,
					uint numLayers, uint numRows, uint numColumns,
					TElement* pElements, int layerStride, int rowStride, int columnStride)
{
	VW_ASSERT(pElements, "Invalid argument.");

	TElement*  pLayer      = pElements;
	TElement*  pLayerEnd   = pLayer + layerStride * numLayers;
	int        skipRows    = rowStride * numRows;
	int        skipColumns = elemStride * numColumns;

	processor.PreLoop(0);
	while(pLayer != pLayerEnd) {

		TElement*  pRow    = PostIncrement(pLayer, layerStride);
		TElement*  pRowEnd = pRow + skipRows;

		processor.PreLoop(1);
		while(pRow != pRowEnd) {

			TElement*  pColumn    = PostIncrement(pRow, rowStride);
			TElement*  pColumnEnd = pColumn + skipColumns;

			processor.PreLoop(2);
			while(pColumn != pColumnEnd) {

				processor.Process(PointerCast<TProcessorInplace::ElementType>(
										PostIncrement(pColumn, columnStride)));
			}
			processor.PostLoop(2);
		}
		processor.PostLoop(1);
	}
	processor.PostLoop(0);
}


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TProcessorDstSrc>
inline
void __fastcall ProcessElements3D(
					TProcessorDstSrc& processor,
					uint numLayers, uint numRows, uint numColumns,
					TDstElement* pDst, int dstLayerStride, int dstRowStride, int dstColumnStride,
					TSrcElement* pSrc, int srcLayerStride, int srcRowStride, int srcColumnStride)
{
	VW_ASSERT(pDst && pSrc, "Invalid argument.");

	TDstElement*  pDstLayer      = pDst;
	TDstElement*  pDstLayerEnd   = pDstLayer + dstLayerStride * numLayers;
	int           dstSkipRows    = dstRowStride * numRows;
	int           dstSkipColumns = dstColumnStride * numColumns;
	TSrcElement*  pSrcLayer      = pSrc;

	processor.PreLoop(0);
	while(pDstLayer != pDstLayerEnd) {

		TDstElement*  pDstRow    = PostIncrement(pDstLayer, dstLayerStride);
		TDstElement*  pDstRowEnd = pDstRow + dstSkipRows;
		TSrcElement*  pSrcRow    = PostIncrement(pSrcLayer, srcLayerStride);

		processor.PreLoop(1);
		while(pDstRow != pDstRowEnd) {

			TDstElement*  pDstColumn    = PostIncrement(pDstRow, dstRowStride);
			TDstElement*  pDstColumnEnd = pDstColumn + dstSkipColumns;
			TSrcElement*  pSrcColumn    = PostIncrement(pSrcRow, srcRowStride);

			processor.PreLoop(2);
			while(pDstColumn != pDstColumnEnd) {

				processor.Process(PointerCast<TProcessorDstSrc::DstElementType>(
										PostIncrement(pDstColumn, dstColumnStride)),
									PointerCast<TProcessorDstSrc::SrcElementType>(
										PostIncrement(pSrcColumn, srcColumnStride)));
			}
			processor.PostLoop(2);
		}
		processor.PostLoop(1);
	}
	processor.PostLoop(0);
}


//********************************************************************
namespace element_processors {


//====================================================================
template<typename TElement>
struct Copy_1 : public ElementProcessorDstSrc<TElement, TElement>
{
	__forceinline  void  Process(TElement* dst, const TElement* src) {
		*dst = *src;
	}
};


//====================================================================
template<typename TElement>
struct Copy_N : public ElementProcessorDstSrc<TElement, TElement>
{
	uint  copyCount;

	Copy_N(uint _copyCount) {
		copyCount = _copyCount;
	}

	__forceinline  void  Process(TElement* dst, const TElement* src) {
		for(uint i = 0; i < copyCount; i++) {
			dst[i] = src[i];
		}
	}
};


//====================================================================
template<typename TElement>
struct Process_Swap1 : public ElementProcessorInplace<TElement>
{
	int  offset;

	Process_Swap1(int _offset) {
		offset = _offset;
	}

	__forceinline  void  Process(TElement* elem) {
		Swap(*elem, elem[offset]);
	}
};


//====================================================================
template<typename TElement>
struct Swap_N : public ElementProcessorDstSrc<TElement, TElement>
{
	uint  swapCount;

	Swap_N(uint _swapCount) {
		swapCount = _swapCount;
	}

	__forceinline  void  Process(TElement* elem1, TElement* elem2) {
		for(uint i = 0; i < swapCount; i++) {
			Swap(elem1[i], elem2[i]);
		}
	}
};


//====================================================================
template<typename TElement>
struct Process_FlipOrder2 : public ElementProcessorInplace<TElement>
{
	__forceinline  void  Process(TElement* elem) {
		Swap(elem[0], elem[1]);
	}
};


//====================================================================
template<typename TElement>
struct Copy_FlippedOrder2 : public ElementProcessorDstSrc<TElement, TElement>
{
	__forceinline  void  Process(TElement* dst, const TElement* src) {
		dst[0] = src[1];
		dst[1] = src[0];
	}
};


//====================================================================
template<typename TElement>
struct Process_FlipOrder3 : public ElementProcessorInplace<TElement>
{
	__forceinline  void  Process(TElement* elem) {
		Swap(elem[0], elem[2]);
	}
};


//====================================================================
template<typename TElement>
struct Copy_FlippedOrder3 : public ElementProcessorDstSrc<TElement, TElement>
{
	__forceinline  void  Process(TElement* dst, const TElement* src) {
		dst[0] = src[2];
		dst[1] = src[1];
		dst[2] = src[0];
	}
};


//====================================================================
template<typename TElement>
struct Process_FlipOrder4 : public ElementProcessorInplace<TElement>
{
	__forceinline  void  Process(TElement* elem) {
		TElement  temp;
		Swap(elem[0], elem[3], temp);
		Swap(elem[1], elem[2], temp);
	}
};


//====================================================================
template<typename TElement>
struct Copy_FlippedOrder4 : public ElementProcessorDstSrc<TElement, TElement>
{
	__forceinline  void  Process(TElement* dst, const TElement* src) {
		dst[0] = src[3];
		dst[1] = src[2];
		dst[2] = src[1];
		dst[3] = src[0];
	}
};


//====================================================================
template<typename TElement>
struct Process_FlipOrderN : public ElementProcessorInplace<TElement>
{
	uint count;

	Process_FlipOrderN(uint _count) {
		count = _count;
	}

	__forceinline  void  Process(TElement* elem) {
		FlipOrderN<TElement>(elem, count);
	}
};


//====================================================================
template<typename TElement>
struct Process_Fill1 : public ElementProcessorInplace<TElement>
{
	TElement  fillValue0;

	Process_Fill1(const TElement& v0) {
		fillValue0 = v0;
	}

	__forceinline  void  Process(TElement* elem) {
		*elem = fillValue0;
	}
};


//====================================================================
template<typename TElement>
struct Process_Fill2 : public ElementProcessorInplace<TElement>
{
	TElement  fillValue0;
	TElement  fillValue1;

	Process_Fill2(const TElement& v0, const TElement& v1) {
		fillValue0 = v0;
		fillValue1 = v1;
	}

	__forceinline  void  Process(TElement* elem) {
		elem[0] = fillValue0;
		elem[1] = fillValue1;
	}
};


//====================================================================
template<typename TElement>
struct Process_Fill3 : public ElementProcessorInplace<TElement>
{
	TElement  fillValue0;
	TElement  fillValue1;
	TElement  fillValue2;

	Process_Fill3(const TElement& v0, const TElement& v1, const TElement& v2) {
		fillValue0 = v0;
		fillValue1 = v1;
		fillValue2 = v2;
	}

	__forceinline  void  Process(TElement* elem) {
		elem[0] = fillValue0;
		elem[1] = fillValue1;
		elem[2] = fillValue2;
	}
};


//====================================================================
template<typename TElement>
struct Process_Fill4 : public ElementProcessorInplace<TElement>
{
	TElement  fillValue0;
	TElement  fillValue1;
	TElement  fillValue2;
	TElement  fillValue3;

	Process_Fill4(const TElement& v0, const TElement& v1, const TElement& v2, const TElement& v3) {
		fillValue0 = v0;
		fillValue1 = v1;
		fillValue2 = v2;
		fillValue3 = v3;
	}

	__forceinline  void  Process(TElement* elem) {
		elem[0] = fillValue0;
		elem[1] = fillValue1;
		elem[2] = fillValue2;
		elem[3] = fillValue3;
	}
};


//====================================================================
template<typename TElement>
struct Process_FillN : public ElementProcessorInplace<TElement>
{
	uint      fillCount;
	TElement  fillValue;

	Process_FillN(uint _fillCount, uint _fillValue) {
		fillCount = _fillCount;
		fillValue = _fillValue;
	}

	__forceinline  void  Process(TElement* elem) {
		for(uint i = 0; i < fillCount; i++) {
			elem[i] = fillValue;
		}
	}
};


//====================================================================
template<typename TElement>
struct Process_HighPassFilter : public ElementProcessorInplace<TElement>
{
	TElement  thresholdValue;
	TElement  newThresholdValue;
	TElement  lowValue;

	Process_HighPassFilter(const TElement& threshold,
							const TElement& newThreshold, const TElement& low)
	{
		thresholdValue = threshold;
		newThresholdValue = newThreshold;
		lowValue = low;
	}

	__forceinline  void  Process(TElement* elem)
	{
		TElement testValue = *elem;
		if(testValue < thresholdValue) *elem = lowValue;
		else if(testValue == thresholdValue) *elem = newThresholdValue;
	}
};


//====================================================================
template<typename TElement>
struct Process_LowPassFilter : public ElementProcessorInplace<TElement>
{
	TElement  thresholdValue;
	TElement  newThresholdValue;
	TElement  highValue;

	Process_LowPassFilter(const TElement& threshold,
							const TElement& newThreshold, const TElement& high)
	{
		thresholdValue = threshold;
		newThresholdValue = newThreshold;
		highValue = high;
	}

	__forceinline  void  Process(TElement* elem)
	{
		TElement testValue = *elem;
		if(testValue > thresholdValue) *elem = highValue;
		else if(testValue == thresholdValue) *elem = newThresholdValue;
	}
};


//====================================================================
template<typename TElement>
struct Process_ThresholdSegmentation : public ElementProcessorInplace<TElement>
{
	TElement  thresholdValue;
	TElement  lowValue;
	TElement  midValue;
	TElement  highValue;

	Process_ThresholdSegmentation(const TElement& threshold, const TElement& low,
								const TElement& mid, const TElement& high)
	{
		thresholdValue = threshold;
		lowValue = low;
		midValue = mid;
		highValue = high;
	}

	__forceinline  void  Process(TElement* elem)
	{
		TElement testValue = *elem;
		if(testValue < thresholdValue) *elem = lowValue;
		else if(testValue > thresholdValue) *elem = highValue;
		else *elem = midValue;
	}
};


//********************************************************************
}//namespace element_processors


//********************************************************************
}//namespace vertexware


