

//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Memory.h"
#include "BaseTools\FileSystem.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\String.h"
#include "BaseTools\PlatformInfo.h"
#include <fstream>


//********************************************************************
namespace {


//====================================================================
bool __fastcall GetFileSystemEntryInfo(const std::string& s, WIN32_FIND_DATAW& fd)
{
	if(s.empty()) return false;

	std::wstring unicodeFileName;
	if(! vertexware::file_system::ConvertPathName(s, unicodeFileName)) return false;

	HANDLE  hSearch = ::FindFirstFileExW(unicodeFileName.c_str(), FindExInfoStandard, &fd,
															FindExSearchNameMatch, 0, 0);
	if(hSearch == INVALID_HANDLE_VALUE) {
		return false; //entry not found
	}
	::FindClose(hSearch);
	return true;
}


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//********************************************************************
namespace file_system {


//====================================================================
void __fastcall SplitFileName(const std::string& fileName, std::string& pathName,
								std::string& baseName, std::string& extension)
{
	pathName.erase();
	baseName.erase();
	extension.erase();

	uint baseNameStart = 0;
	uint pathSepPos = fileName.rfind(GetPathSeparator());

	if(pathSepPos != std::string::npos) {
		baseNameStart = pathSepPos + 1;
		pathName.assign(fileName.substr(0, baseNameStart));
	}

	if(baseNameStart >= fileName.length()) return;
	baseName.assign(fileName.substr(baseNameStart));

	uint extSepPos = baseName.rfind(GetExtensionSeparator());

	bool bHasExtension =
				(extSepPos != std::string::npos) &&
				(extSepPos != baseName.length() - 1) &&
				(extSepPos != 0);

	if(bHasExtension) {
		extension.assign(baseName.substr(extSepPos + 1));
		baseName.erase(extSepPos);
	}
}


//====================================================================
std::string __fastcall GetFileNameExtension(const std::string& fileName)
{
	uint dotPos = fileName.rfind('.');
	if((dotPos != std::string::npos) && (dotPos != fileName.length() - 1) && (dotPos != 0)) {
		return fileName.substr(dotPos + 1);
	}
	return "";
}


//====================================================================
void __fastcall RemoveTrailingPathSeparator(std::string& s)
{
	if(s.length()  &&  s[s.length() - 1] == GetPathSeparator()) {
		s.erase(s.length() - 1);
	}
}


//====================================================================
void __fastcall ForceTrailingPathSeparator(std::string& s)
{
	if(s.length()  &&  s[s.length() - 1] != GetPathSeparator()) {
		s.append(1, GetPathSeparator());
	}
}


//====================================================================
void __fastcall ForceFullPathName(std::string& pathName, bool bForceTrailingPathSep)
{
	if(! IsAbsolutePathName(pathName)) {
		std::string  proccesDirectory;
		file_system::GetProcessDirectory(proccesDirectory, true);
		pathName.insert(0, proccesDirectory);
	}
	if(bForceTrailingPathSep) {
		ForceTrailingPathSeparator(pathName);
	}
}


//====================================================================
std::string __fastcall GetFullPathName(const std::string& pathName, bool bForceTrailingPathSep)
{
	std::string fullPath = pathName;
	ForceFullPathName(fullPath, bForceTrailingPathSep);
	return fullPath;
}


//====================================================================
bool __fastcall IsAbsolutePathName(const std::string& pathName)
{
	// absolute starts with "?:" or "\"
	return
	(pathName.length() >= 2  &&  pathName[1] == GetVolumeSeparator())  ||
	(pathName.length() >= 1  &&  pathName[0] == GetPathSeparator());
}


//====================================================================
void __fastcall SplitPathName(const std::string& pathName, std::vector<std::string>& subNames)
{
	subNames.clear();

	std::string  s = pathName;
	while(s.length()) {
		uint sepPos = s.find(GetPathSeparator());

		if(sepPos != std::string::npos) {
			if(sepPos > 0) {
				subNames.push_back(s.substr(0, sepPos));
			}
			s.erase(0, sepPos + 1);
		}
		else {
			subNames.push_back(s);
			break;
		}
	}
}


//====================================================================
bool __fastcall ExpandPathName(const std::string& pathName, std::string& out,
							bool bForceTrailingPathSep)
{
	// analyze path
	std::vector<std::string>  pathSubNames;
	SplitPathName(pathName, pathSubNames);

	// construct new path
	std::vector<std::string>  newPathSubNames;
	newPathSubNames.reserve(pathSubNames.size());

	bool bOK = true;
	uint minPathDepth = IsAbsolutePathName(pathName) ? 1 : 0;

	for(uint i = 0; i < pathSubNames.size(); i++) {
		const std::string& s = pathSubNames[i];

		// ignore
		if(s == ".") {
			continue;
		}

		// backtrack if possible
		else if(s == "..") {
			if(newPathSubNames.size() <= minPathDepth) {
				bOK = false;
				break;
			}
			newPathSubNames.pop_back();
		}

		// copy sub-name
		else {
			newPathSubNames.push_back(s);
		}
	}

	// construct path string
	out.erase();
	if(bOK) {
		for(uint i = 0; i < newPathSubNames.size(); i++) {
			out.append(newPathSubNames[i]);
			if(i < newPathSubNames.size() - 1) {
				out.append(1, GetPathSeparator());
			}
		}
		if(bForceTrailingPathSep) {
			ForceTrailingPathSeparator(out);
		}
	}
	return bOK;
}


//====================================================================
bool __fastcall ConvertShortPathNameToLongPathName(const std::string& shortPathName,
												std::string& outLongPathName)
{
	std::wstring ws;
	if(! ConvertPathName(shortPathName, ws)) return false;

	uint len = ::GetLongPathNameW(ws.c_str(), 0, 0);
	if(! len) return false;
	Buffer<wchar_t> wBuf(len);
	if(! ::GetLongPathNameW(ws.c_str(), wBuf, len)) return false;
	ws.assign(wBuf);

	return ConvertPathName(ws, outLongPathName);
}


//====================================================================
bool __fastcall ConvertPathName(const std::string& src, std::wstring& dst)
{
	std::string  preparedSrc;
	preparedSrc.reserve(4 + src.length());
	preparedSrc.append("\\\\?\\").append(src);
	return AsciiToUnicode(preparedSrc, dst);
}


//====================================================================
bool __fastcall ConvertPathName(const std::wstring& src, std::string& dst)
{
	std::wstring  preparedSrc;
	if(src.length() >= 4  &&  src.substr(0, 4) == L"\\\\?\\") {
		preparedSrc = src.substr(4);
	}
	else {
		preparedSrc = src;
	}
	return UnicodeToAscii(preparedSrc, dst);
}


//====================================================================
bool __fastcall FileExists(const std::string& s)
{
	// cannot contain wildcards
	if(s.find_first_of("?*") != std::string::npos) return false;

	WIN32_FIND_DATAW  fd;
	return GetFileSystemEntryInfo(s, fd);
}


//====================================================================
bool __fastcall DeleteFile(const std::string& s)
{
	std::wstring ws;
	if(! ConvertPathName(s, ws)) return false;
	return ToBool(::DeleteFileW(ws.c_str()));
}


//====================================================================
bool __fastcall IfFileExistsDeleteOnOverwrite(const std::string& s, bool bOverwrite)
{
	if(file_system::FileExists(s)) {
		if(! bOverwrite) return false;
		return file_system::DeleteFile(s);
	}
	return true;
}


//====================================================================
bool __fastcall CopyFile(const std::string& src, const std::string& dst, bool bOverwrite)
{
	std::wstring  wSrc;
	if(! ConvertPathName(src, wSrc)) return false;
	std::wstring  wDst;
	if(! ConvertPathName(dst, wDst)) return false;

	return ToBool(::CopyFileW(wSrc.c_str(), wDst.c_str(), ! bOverwrite));
}


//====================================================================
bool __fastcall MoveFile(const std::string& src, const std::string& dst,
						bool bAllowOverwriteFile)
{
	std::wstring wSrc, wDst;
	if(! ConvertPathName(src, wSrc)) return false;
	if(! ConvertPathName(dst, wDst)) return false;

	DWORD flags = MOVEFILE_COPY_ALLOWED | MOVEFILE_WRITE_THROUGH;
	if(bAllowOverwriteFile) flags |= MOVEFILE_REPLACE_EXISTING;

	return (::MoveFileExW(wSrc.c_str(), wDst.c_str(), flags) != 0);
}


//====================================================================
bool __fastcall DirectoryExists(const std::string& s)
{
	// cannot contain wildcards
	if(s.find_first_of("?*") != std::string::npos) return false;

	if(s.empty()) return false;

	// must not contain trailing path-separator
	std::string name = s;
	RemoveTrailingPathSeparator(name);

	// prepare volume name (special case)
	if(name.length() == 2) {
		if(! isalpha(name[0])  ||  name[1] != ':') return false;
		name.append("\\*");
	}

	WIN32_FIND_DATAW  fd;
	return GetFileSystemEntryInfo(name, fd);
}


//====================================================================
bool __fastcall DeleteDirectory(const std::string& s)
{
	std::wstring ws;
	if(! ConvertPathName(s, ws)) return false;
	return ToBool(::RemoveDirectoryW(ws.c_str()));
}


//====================================================================
bool __fastcall CreateDirectory(const std::string& s)
{
	std::wstring ws;
	if(! ConvertPathName(s, ws)) return false;
	return ToBool(::CreateDirectoryW(ws.c_str(), 0));
}


//====================================================================
bool __fastcall CreateDirectoryPath(const std::string& s)
{
	std::vector<std::string> subNames;
	SplitPathName(s, subNames);

	if(subNames.empty()) return false;

	std::string pathName;
	for(uint i = 0; i < subNames.size(); i++) {

		pathName.append(subNames[i]).append(1, GetPathSeparator());

		if(! DirectoryExists(pathName)) {
			if(! CreateDirectory(pathName)) return false;
		}
	}
	return true;
}


//====================================================================
bool __fastcall GetProcessDirectory(std::string& s, bool bForceTrailingPathSep)
{
	s.erase();

	uint len = ::GetCurrentDirectoryW(0, 0);
	if(! len) return false;
	Buffer<wchar_t> buf(len);
	if(! ::GetCurrentDirectoryW(len, buf)) return false;

	if(! ConvertPathName(std::wstring(buf), s)) return false;

	if(bForceTrailingPathSep) {
		ForceTrailingPathSeparator(s);
	}
	return true;
}


//====================================================================
bool __fastcall SetProcessDirectory(const std::string& s)
{
	std::string processDir = s;
	ForceFullPathName(processDir, false);

	if(! DirectoryExists(processDir)) return false;

	std::wstring wProcessDir;
	if(! ConvertPathName(processDir, wProcessDir)) return false;

	return (::SetCurrentDirectoryW(wProcessDir.c_str()) != 0);
}


//====================================================================
bool __fastcall GetUserHomeDirectory(std::string& s, bool bForceTrailingPathSep)
{
	s.erase();

	// get %HOMEDRIVE% + %HOMEPATH%
	std::string sHomeDrive, sHomePath;

	if(! platform_info::ExpandEnvironmentVariable("%HOMEDRIVE%", sHomeDrive)) return false;
	if(! platform_info::ExpandEnvironmentVariable("%HOMEPATH%", sHomePath)) return false;

	s = sHomeDrive + sHomePath;

	if(bForceTrailingPathSep) {
		ForceTrailingPathSeparator(s);
	}
	return true;
}


//====================================================================
bool __fastcall GetUserProfileDirectory(std::string& s, bool bForceTrailingPathSep)
{
	s.erase();

	// get %USERPROFILE%
	if(! platform_info::ExpandEnvironmentVariable("%USERPROFILE%", s)) return false;

	if(bForceTrailingPathSep) {
		ForceTrailingPathSeparator(s);
	}
	return true;
}


//====================================================================
bool __fastcall GetTempDirectory(std::string& s, bool bForceTrailingPathSep)
{
	s.erase();

	uint len = ::GetTempPath(0, 0);
	if(! len) return false;
	Buffer<char> buf(len);
	if(! ::GetTempPath(len, buf)) return false;

	if(! ConvertShortPathNameToLongPathName(buf.ToPtr(), s)) return false;

	if(bForceTrailingPathSep) {
		ForceTrailingPathSeparator(s);
	}
	return true;
}


//====================================================================
bool __fastcall CreateTempFile(std::string& outPathName,
								const std::string& tempDir, const std::string& prefix)
{
	std::wstring  wTempDir, wPrefix;
	if(! ConvertPathName(tempDir, wTempDir)) return false;
	if(! AsciiToUnicode(prefix, wPrefix)) return false;

	Buffer<wchar_t> buf(tempDir.length() + wPrefix.length() + MAX_PATH + 2);
	if(::GetTempFileNameW(wTempDir.c_str(), wPrefix.c_str(), 0, buf) != 0) {
		if(! ConvertPathName(buf.ToPtr(), outPathName)) return false;
		return true;
	}
	return false;
}


//====================================================================
bool  FileSystemWalker::Walk(const std::string& pathName)
{
	// prepare path-name
	std::string  fixedPathName = pathName;
	file_system::ForceTrailingPathSeparator(fixedPathName);

	std::wstring fixedUnicodePathName;
	if(! file_system::ConvertPathName(fixedPathName, fixedUnicodePathName)) return false;

	// do walk
	bKeepWalking = true;

	OnBeginWalk(fixedPathName);
	bool bOpenedRootDir = WalkOneLevel(0, fixedPathName, fixedUnicodePathName);
	OnEndWalk();

	return bOpenedRootDir;
}


//====================================================================
bool  FileSystemWalker::WalkOneLevel(uint depth,
									std::string& pathName, std::wstring& unicodePathName)
{
	if(! bKeepWalking) return true;

	// open system resource
	unicodePathName.append(1, L'*'); //required by windows

	WIN32_FIND_DATAW  fd;
	HANDLE  hSearch = ::FindFirstFileExW(unicodePathName.c_str(), FindExInfoStandard, &fd,
															FindExSearchNameMatch, 0, 0);
	unicodePathName.erase(unicodePathName.length() - 1);

	if(hSearch == INVALID_HANDLE_VALUE) {
		// failed to open directory
		return false;
	}

	// enumerate current-level entries
	do {
		// extract attributes
		uint  attrib = 0;

		if(fd.dwFileAttributes  &  FILE_ATTRIBUTE_ARCHIVE)  attrib |= FileAttribute_Archive;
		if(fd.dwFileAttributes  &  FILE_ATTRIBUTE_HIDDEN)   attrib |= FileAttribute_Hidden;
		if(fd.dwFileAttributes  &  FILE_ATTRIBUTE_READONLY) attrib |= FileAttribute_ReadOnly;
		if(fd.dwFileAttributes  &  FILE_ATTRIBUTE_SYSTEM)   attrib |= FileAttribute_System;

		// construct entry names
		std::wstring  unicodeEntryName(fd.cFileName);
		std::string   entryName;
		UnicodeToAscii(unicodeEntryName, entryName);

		std::string  entryNameWithPath(pathName);
		entryNameWithPath.append(entryName);

		// report entry
		if(fd.dwFileAttributes  &  FILE_ATTRIBUTE_DIRECTORY) {

			// found directory, discard special entries
			if(entryName != "."  &&  entryName != "..") {

				// found standard directory
				if(OnDirectory(depth, attrib, entryName, entryNameWithPath)) {

					// push paths
					pathName.append(entryName).append(1, '\\');
					unicodePathName.append(unicodeEntryName).append(1, L'\\');

					// recurse search
					WalkOneLevel(depth + 1, pathName, unicodePathName);

					// pop paths
					pathName.erase(pathName.length() - entryName.length() - 1);
					unicodePathName.erase(unicodePathName.length() - unicodeEntryName.length() - 1);
				}
			}
		}
		else {
			// found non-directory
			OnFile(depth, attrib, entryName, entryNameWithPath);
		}

	} while(bKeepWalking  &&  ::FindNextFileW(hSearch, &fd));

	// close system resource
	::FindClose(hSearch);

	// report success
	return true;
}


//********************************************************************
}//namespace file_system


//********************************************************************
}//namespace vertexware


