

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "private\vw_common_exports.h"
#include <string>
#include <vector>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace file_system {


//====================================================================
// Separators
//====================================================================
VW_COMMON_API __forceinline char __fastcall GetPathSeparator() { return '\\'; }
VW_COMMON_API __forceinline char __fastcall GetVolumeSeparator() { return ':'; }
VW_COMMON_API __forceinline char __fastcall GetExtensionSeparator() { return '.'; }


//====================================================================
// Splits full file-name into path-name, base-name and file-extension.
//====================================================================
VW_COMMON_API
void __fastcall SplitFileName(const std::string& fileName, std::string& pathName,
							std::string& baseName, std::string& extension);


//====================================================================
// Extracts file-extension (part after last extension-separator).
//====================================================================
VW_COMMON_API
std::string __fastcall GetFileNameExtension(const std::string& fileName);


//====================================================================
// Removes trailing path-separator if it exists.
//====================================================================
VW_COMMON_API
void __fastcall RemoveTrailingPathSeparator(std::string& s);


//====================================================================
// Appends trailing path-separator if it's missing.
//====================================================================
VW_COMMON_API
void __fastcall ForceTrailingPathSeparator(std::string& s);


//====================================================================
// Retrieve full path-name. If <pathName> is in relative form,
// current process directory is prepended to <pathName>.
//====================================================================
VW_COMMON_API
void __fastcall ForceFullPathName(std::string& pathName,
								bool bForceTrailingPathSep = false);


//====================================================================
VW_COMMON_API
std::string __fastcall GetFullPathName(const std::string& pathName,
								bool bForceTrailingPathSep = false);


//====================================================================
// Returns true if <pathName> is in absolute form.
//====================================================================
VW_COMMON_API
bool __fastcall IsAbsolutePathName(const std::string& pathName);


//====================================================================
// Splits <pathName> into parts separated by path-separator or volume-separator.
//====================================================================
VW_COMMON_API
void __fastcall SplitPathName(const std::string& pathName,
								std::vector<std::string>& subNames);


//====================================================================
// Expands special path identifiers { ".", ".." }
// Returns false on error.
//====================================================================
VW_COMMON_API
bool __fastcall ExpandPathName(const std::string& pathName, std::string& out,
								bool bForceTrailingPathSep = false);


//====================================================================
// Expands short path-name to long path-name.
// Returns false on error.
//====================================================================
VW_COMMON_API
bool __fastcall ConvertShortPathNameToLongPathName(const std::string& shortPathName,
												std::string& outLongPathName);


//====================================================================
// Converts file-system name between unicode/ascii form.
// Unicode form is adjusted to match W-functions required by Windows,
// that is "\\?\" is prepended to name.
//====================================================================
VW_COMMON_API
bool __fastcall ConvertPathName(const std::string& src, std::wstring& dst);


//====================================================================
VW_COMMON_API
bool __fastcall ConvertPathName(const std::wstring& src, std::string& dst);


//====================================================================
// Returns true if file exists.
// Requires absolute path-name.
//====================================================================
VW_COMMON_API
bool __fastcall FileExists(const std::string& s);


//====================================================================
// Returns false if failed.
// Requires absolute path-name.
//====================================================================
VW_COMMON_API
bool __fastcall DeleteFile(const std::string& s);


//====================================================================
// Returns false if file exists and <bOverwrite> is false
// or file-system failed to delete the file.
// Requires absolute path-name.
//====================================================================
VW_COMMON_API
bool __fastcall IfFileExistsDeleteOnOverwrite(const std::string& s, bool bOverwrite);


//====================================================================
// Returns false if failed.
// Requires absolute path-names.
//====================================================================
VW_COMMON_API
bool __fastcall CopyFile(const std::string& src,
						const std::string& dst, bool bOverwrite = false);


//====================================================================
// Moves file or directory. Directory cannot be moved to a different volume.
// <bAllowOverwriteFile> is valid for files only.
// Returns false if failed.
// Requires absolute path-names.
//====================================================================
VW_COMMON_API
bool __fastcall MoveFile(const std::string& src, const std::string& dst,
						bool bAllowOverwriteFile = false);


//====================================================================
// Returns true if directory exists.
// Requires absolute path-name.
//====================================================================
VW_COMMON_API
bool __fastcall DirectoryExists(const std::string& s);


//====================================================================
// Returns false if failed.
// Directory must be empty for successful deletion.
// Requires absolute path-name.
//====================================================================
VW_COMMON_API
bool __fastcall DeleteDirectory(const std::string& s);


//====================================================================
// Returns false if failed.
// Requires absolute path-name.
//====================================================================
VW_COMMON_API
bool __fastcall CreateDirectory(const std::string& s);


//====================================================================
// Returns false if failed. Returns true if path exists.
// Requires absolute path-name.
//====================================================================
VW_COMMON_API
bool __fastcall CreateDirectoryPath(const std::string& s);


//====================================================================
// Obtains current process directory.
// Returns false if failed.
//====================================================================
VW_COMMON_API
bool __fastcall GetProcessDirectory(std::string& s, bool bForceTrailingPathSep = false);


//====================================================================
// Sets current process directory.
// Returns false if failed.
//====================================================================
VW_COMMON_API
bool __fastcall SetProcessDirectory(const std::string& s);


//====================================================================
// Obtains user's home directory.
// Returns false if failed.
//====================================================================
VW_COMMON_API
bool __fastcall GetUserHomeDirectory(std::string& s, bool bForceTrailingPathSep = false);


//====================================================================
// Obtains user's profile directory.
// Returns false if failed.
//====================================================================
VW_COMMON_API
bool __fastcall GetUserProfileDirectory(std::string& s, bool bForceTrailingPathSep = false);


//====================================================================
// Obtains user's temp directory.
// Returns false if failed.
//====================================================================
VW_COMMON_API
bool __fastcall GetTempDirectory(std::string& s, bool bForceTrailingPathSep = false);


//====================================================================
// Creates file with unique name in specified directory.
// Only 3 letters from <prefix> are used for filename.
// Returns false if failed.
//====================================================================
VW_COMMON_API
bool __fastcall CreateTempFile(std::string& outPathName,
								const std::string& tempDir, const std::string& prefix);


//====================================================================
// File-system depth-first directory-walker.
//====================================================================
class VW_COMMON_API FileSystemWalker
{
public:
	enum eFileAttribute
	{
		FileAttribute_Archive  = 1,
		FileAttribute_Hidden   = 2,
		FileAttribute_ReadOnly = 4,
		FileAttribute_System   = 8,

		FileAttribute_Force32Bits = 0x7FFFFFFF
	};


public:
	virtual ~FileSystemWalker() {}

	//-----------------------------------------------------------
	// Walks through file-system subtree specified by <pathName>.
	// <pathName> must be a directory.
	// Returns true if <pathName> can be opened, false otherwise.
	//-----------------------------------------------------------
	bool  Walk(const std::string& pathName);


private:
	bool  WalkOneLevel(uint depth, std::string& pathName, std::wstring& unicodePathName);


private:
	bool  bKeepWalking;


protected:
	//--------------------------
	// Cancels the current walk.
	//--------------------------
	void  CancelWalk() { bKeepWalking = false; }

	//------------------
	// Template methods.
	//------------------
	virtual void OnBeginWalk(const std::string& pathName)
	{ }

	virtual void OnEndWalk()
	{ }

	virtual void OnFile(uint depth, uint fileAttributes,
						const std::string& fileName,
						const std::string& path)
	{ }

	//--------------------------------------------------------------
	// Template method: return true to descend into given directory.
	//--------------------------------------------------------------
	virtual bool OnDirectory(uint depth, uint fileAttributes,
								const std::string& directoryName,
								const std::string& path)
	{ return false; }
};


//********************************************************************
}//namespace file_system


//********************************************************************
}//namespace vertexware


