

//--------------------------------------------------------------------
#include "BaseTools\Memory.h"
#include "BaseTools\Debug.h"
#include <sstream>
#include <new.h>


//********************************************************************
namespace vertexware {


//====================================================================
NewThrowsExceptionSentry::NewThrowsExceptionSentry()
{
	pfnLastInstalledNewHandler = _set_new_handler(NewHandler);
}


//====================================================================
NewThrowsExceptionSentry::~NewThrowsExceptionSentry()
{
	PFN_NEW_HANDLER pfnHandler = _set_new_handler(pfnLastInstalledNewHandler);

	// check for new-handler corruption
	VW_VERIFY(pfnHandler == NewHandler,
			"vertexware-new-handler was unhooked by non-vertexware code.");
}


//====================================================================
int __cdecl NewThrowsExceptionSentry::NewHandler(uint numBytesRequested)
{
	std::ostringstream  oss;
	oss << "Failed to allocate " << numBytesRequested << " bytes";

	VW_THROW(vertexware::NoMemoryException, oss.str());

	return 0; //do not try to allocate again
}


//====================================================================
AdaptiveBufferBase::AdaptiveBufferBase()
{
	minSize = 0;
	minStableCycles = 0;

	currSize = 0;

	numStableCycles = 0;
	lastCycleMaxSize = 0;
	activeCycleMaxSize = 0;
}


//====================================================================
AdaptiveBufferBase::~AdaptiveBufferBase()
{
	// cannot call DoDeallocate
}


//====================================================================
void  AdaptiveBufferBase::ForceMinSize()
{
	DoReserve(minSize);
	numStableCycles = 0;
}


//====================================================================
void  AdaptiveBufferBase::SetMinSize(uint _minSize)
{
	minSize = _minSize;

	if(currSize < minSize) {
		DoReserve(minSize);
	}
	numStableCycles = 0;
}


//====================================================================
void  AdaptiveBufferBase::SetMinStableCycles(uint newMinStableCycles)
{
	minStableCycles = newMinStableCycles;
}


//====================================================================
void  AdaptiveBufferBase::BeginReserveCycle()
{
	lastCycleMaxSize = activeCycleMaxSize;
	activeCycleMaxSize = 0;
}


//====================================================================
void  AdaptiveBufferBase::ReserveSpace(uint newSize)
{
	activeCycleMaxSize = Max(activeCycleMaxSize, newSize);
	if(currSize < newSize) {
		DoReserve(ClampedLow<uint>(newSize, minSize));
	}
}


//====================================================================
void  AdaptiveBufferBase::EndReserveCycle()
{
	if(lastCycleMaxSize == activeCycleMaxSize) {
		numStableCycles ++;
	}
	else {
		numStableCycles = 0;
	}
	if(numStableCycles < minStableCycles) return;
	if(activeCycleMaxSize >= currSize) return;

	// deallocate unused storage
	if(activeCycleMaxSize > minSize) {
		DoReserve(activeCycleMaxSize);
	}
}


//====================================================================
void  AdaptiveBufferBase::DoReserve(uint newSize)
{
	DoDeallocate();
	currSize = newSize;
	if(currSize) DoAllocate(currSize);
}


//********************************************************************
}//namespace vertexware


