

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Exception.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\Debug.h"
#include "private\vw_common_exports.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Byte-order flips.
//====================================================================
VW_COMMON_API
__forceinline  void __fastcall FlipByteOrder(ushort& x)
{
	x = (x << 8) | (x >> 8);
}


//====================================================================
VW_COMMON_API __forceinline  void __fastcall FlipByteOrder(short& x) {
	FlipByteOrder(reinterpret_cast<ushort&>(x));
}


//====================================================================
VW_COMMON_API
__forceinline  void __fastcall FlipByteOrder(uint& x)
{
	x = (x << 24) | (x >> 24) | ((x & 0x0000FF00) << 8) | ((x & 0x00FF0000) >> 8);
}


//====================================================================
VW_COMMON_API __forceinline  void __fastcall FlipByteOrder(int& x) {
	FlipByteOrder(reinterpret_cast<uint&>(x));
}


//====================================================================
VW_COMMON_API __forceinline  void __fastcall FlipByteOrder(ulong& x) {
	FlipByteOrder(reinterpret_cast<uint&>(x));
}


//====================================================================
VW_COMMON_API __forceinline  void __fastcall FlipByteOrder(long& x) {
	FlipByteOrder(reinterpret_cast<uint&>(x));
}


//====================================================================
VW_COMMON_API __forceinline  void __fastcall FlipByteOrder(float& x) {
	FlipByteOrder(reinterpret_cast<uint&>(x));
}


//====================================================================
VW_COMMON_API __forceinline  void __fastcall FlipByteOrder(double& x) {
	FlipOrder8<ubyte>(reinterpret_cast<ubyte*>(&x));
}


//====================================================================
VW_COMMON_API __forceinline  void __fastcall FlipByteOrderN(ubyte* x, uint count) {
	FlipOrderN<ubyte>(x, count);
}


//====================================================================
// Memory allocation control.
// Standard new operator returns zero if memory allocation fails.
// Creating one of this sentries causes new operator to throw an
// exception if memory allocation fails. Valid only until the end
// of definition scope.
// Example:
//
//     int* p = new int[1000000000]; //returns 0
//     {
//         NewThrowsExceptionSentry ntes;
//         try {
//             int* q = new int[1000000000]; //throws exception
//         }
//         catch(NoMemoryException& e) {
//         }
//     }
//     int* r = new int[1000000000]; //returns 0
// 
//====================================================================
class VW_COMMON_API NewThrowsExceptionSentry
{
public:
	NewThrowsExceptionSentry();
	~NewThrowsExceptionSentry();


private:
	static int __cdecl NewHandler(uint numBytesRequested);


private:
	typedef int (__cdecl *PFN_NEW_HANDLER)(uint);
	PFN_NEW_HANDLER  pfnLastInstalledNewHandler;
};


//====================================================================
VW_DECLARE_EXCEPTION( NoMemoryException )


//====================================================================
// BufferStandardInterface provides client with array-style access
// interface, and frees the buffer-implementation from redefining it.
// It also provides subset of STL container interface.
// However, TBufferImp must implement the following members
// accessible to BufferStandardInterface by either declaring them
// public or declaring BufferStandardInterface<TBufferImp, T> as friend:
//
// 1) uint      BufferImpGetSize() const;
// 2) T*        BufferImpGetStorage();
// 3) const T*  BufferImpGetStorage();
//
// Derivation example:
//     class BufferImp : public BufferStandardInterface<BufferImp, int>
//     {
//         friend class BufferStandardInterface<BufferImp, int>;
//
//         uint        BufferImpGetSize() const {...}
//         int*        BufferImpGetStorage() {...}
//         const int*  BufferImpGetStorage() {...}
//     };
//====================================================================
template<typename TBufferImp, typename T>
class BufferStandardInterface
{
public:
	typedef  T         value_type;
	typedef  T*        iterator;
	typedef  const T*  const_iterator;
	typedef  T&        reference;
	typedef  const T&  const_reference;


public:
	uint  SizeInBytes() const { return TBufferImpCast().BufferImpGetSize() * sizeof(T); }

	//----------------
	// Pointer access.
	//----------------
	operator        T*()       { return TBufferImpCast().BufferImpGetStorage(); }
	operator  const T*() const { return TBufferImpCast().BufferImpGetStorage(); }

	T*        ToPtr()       { return TBufferImpCast().BufferImpGetStorage(); }
	const T*  ToPtr() const { return TBufferImpCast().BufferImpGetStorage(); }

	//---------------
	// STL interface.
	//---------------
	reference  operator[](uint i) {
		VW_ASSERT(i < TBufferImpCast().BufferImpGetSize(), "Buffer-index out of range.");
		return TBufferImpCast().BufferImpGetStorage()[i];
	}
	const_reference  operator[](uint i) const {
		VW_ASSERT(i < TBufferImpCast().BufferImpGetSize(), "Buffer-index out of range.");
		return TBufferImpCast().BufferImpGetStorage()[i];
	}

	reference  operator[](int i) {
		VW_ASSERT(i < TBufferImpCast().BufferImpGetSize(), "Buffer-index out of range.");
		return TBufferImpCast().BufferImpGetStorage()[i];
	}
	const_reference  operator[](int i) const {
		VW_ASSERT(i < TBufferImpCast().BufferImpGetSize(), "Buffer-index out of range.");
		return TBufferImpCast().BufferImpGetStorage()[i];
	}

	iterator        begin()       { return TBufferImpCast().BufferImpGetStorage(); }
	const_iterator  begin() const { return TBufferImpCast().BufferImpGetStorage(); }

	iterator        end()       { return TBufferImpCast().BufferImpGetStorage() + TBufferImpCast().BufferImpGetSize(); }
	const_iterator  end() const { return TBufferImpCast().BufferImpGetStorage() + TBufferImpCast().BufferImpGetSize(); }

	uint  size() const { return TBufferImpCast().BufferImpGetSize(); }


private:
	__forceinline        TBufferImp&  TBufferImpCast()       { return *static_cast<TBufferImp*>(this); }
	__forceinline  const TBufferImp&  TBufferImpCast() const { return *static_cast<const TBufferImp*>(this); }
};


//====================================================================
// Built-in array wrapper (size must be specified at compile-time).
// Usage example:
//     Array<int, 1024> a;
//====================================================================
template<typename T, uint ArraySize>
class Array : public BufferStandardInterface<Array, T>
{
friend class BufferStandardInterface<Array, T>;

private:
	uint      BufferImpGetSize() const { return ArraySize; }
	T*        BufferImpGetStorage() { return elements; }
	const T*  BufferImpGetStorage() const { return elements; }


private:
	T  elements[ArraySize];
};


//====================================================================
// Dynamically allocated buffer (size can be specified at run-time).
// Usage example:
//     int bufferSize = 1024;
//     Buffer<int> b(bufferSize);
//====================================================================
template<typename T>
class Buffer : public BufferStandardInterface<Buffer, T>
{
friend class BufferStandardInterface<Buffer, T>;

private:
	uint      BufferImpGetSize() const { return numElements; }
	T*        BufferImpGetStorage() { return pElements; }
	const T*  BufferImpGetStorage() const { return pElements; }


public:
	//--------------------------------------------------
	// Constructor.
	// Storage is NOT initialized for built-in types.
	// User-defined types must have default-constructor.
	//--------------------------------------------------
	explicit Buffer(uint size = 0) {
		pElements = 0;
		if(size > 0) pElements = new T[size];
		numElements = size;
	}

	//------------
	// Destructor.
	//------------
	~Buffer() {
		Deallocate();
	}

	//------------
	// Allocation.
	//------------
	void  Allocate(uint size) {
		Deallocate();
		if(size > 0) pElements = new T[size];
		numElements = size;
	}

	//--------------
	// Deallocation.
	//--------------
	void  Deallocate() {
		if(pElements) {
			delete [] pElements;
			pElements = 0;
		}
	}


private:
	uint  numElements;
	T*    pElements;
};


//====================================================================
// Auto-resizes to maximum-size requested by reserve-cycles,
// after the reservation requests become stable. Stability condition
// is controlled by user-parameter.
//====================================================================
class VW_COMMON_API AdaptiveBufferBase
{
public:
	virtual ~AdaptiveBufferBase();

	// adaptive allocation control
	void  ForceMinSize();

	void  SetMinSize(uint newMinSize);
	void  SetMinStableCycles(uint newMinStableCycles = 0);

	__forceinline uint  GetMinSize() const { return minSize; }
	__forceinline uint  GetSize() const { return currSize; }

	// reserve cycle
	void  BeginReserveCycle();
	void  ReserveSpace(uint newSize);
	void  EndReserveCycle();


private:
	// adaptive-allocation params
	uint  minSize;
	uint  minStableCycles;

	// current size
	uint  currSize;

	// cycle statistics
	uint  numStableCycles;
	uint  lastCycleMaxSize;
	uint  activeCycleMaxSize;


private:
	void  DoReserve(uint newSize);


protected:
	explicit AdaptiveBufferBase();

	// implemented by derived class
	// WARNING: be sure to call DoDeallocate in derived-class's dtor
	virtual  void  DoAllocate(uint newSize) = 0;
	virtual  void  DoDeallocate() = 0;
};


//====================================================================
template<typename T>
class AdaptiveBuffer : public AdaptiveBufferBase,
						public BufferStandardInterface<AdaptiveBuffer, T>
{
friend class BufferStandardInterface<AdaptiveBuffer, T>;

private:
	uint      BufferImpGetSize() const { return GetSize(); }
	T*        BufferImpGetStorage() { return pElements; }
	const T*  BufferImpGetStorage() const { return pElements; }


public:
	typedef  T  ElementType;


public:
	explicit AdaptiveBuffer(uint minSize = 0, uint minStableCycles = 0)
	{
		pElements = 0;

		SetMinSize(minSize);
		SetMinStableCycles(minStableCycles);
	}

	virtual ~AdaptiveBuffer()
	{
		DoDeallocate();
	}


protected:
	virtual  void  DoAllocate(uint newSize) { pElements = new T[newSize]; }
	virtual  void  DoDeallocate() { SafeDelete(pElements); }


private:
	T*  pElements;
};


//====================================================================
// Safe-delete automatically zeroes-out the pointer to resource,
// after deleting the resource.
//====================================================================
template<typename T>
__forceinline  void __fastcall SafeDelete(T* &p) {
	if(p) {
		delete p;
		p = 0;
	}
}


//====================================================================
template<typename T>
__forceinline  void __fastcall SafeDeleteArray(T* &p) {
	if(p) {
		delete [] p;
		p = 0;
	}
}


//====================================================================
// Auto-pointer.
// Compatible with STL containers.
// Provides simple automatic resource management.
// Use only for holding pointers to single objects, do NOT store
// pointers to arrays into it, or memory leaks will occur.
//====================================================================
template<typename T>
class AutoPtr
{
public:
	typedef  T  element_type;


public:
	//------------------------------------------------------------
	// Constructor.
	// Usage example: AutoPtr<int> ap(new int);
	// This will not compile by design: AutoPtr<int> ap = new int;
	//------------------------------------------------------------
	explicit AutoPtr(T* p = 0, bool bOwnsResource = true) {
		pResource = p;
		bOwner = bOwnsResource  &&  (p != 0);
	}

	//----------------------------------------
	// Copy constructor.
	// Transfers resource ownership.
	// Source object still points to resource.
	//----------------------------------------
	AutoPtr(const AutoPtr& rhs) {
		bOwner = rhs.bOwner;
		pResource = rhs.Release();
	}

	//------------------------------------------------
	// Destructor.
	// Deletes resource if it is owned by this object.
	//------------------------------------------------
	~AutoPtr() {
		if(bOwner) {
			delete pResource;
		}
	}

	//----------------------------------------
	// Copy assignment.
	// Transfers resource ownership.
	// Source object still points to resource.
	//----------------------------------------
	AutoPtr&  operator=(const AutoPtr& rhs) {
		if(this != &rhs) {
			if(pResource != rhs.pResource) {
				if(bOwner) {
					delete pResource;
				}
				bOwner = rhs.bOwner;
			}
			else if(rhs.bOwner) {
				bOwner = true;
			}
			pResource = rhs.Release();
		}
		return (*this);
	}

	//--------------------------------
	// Assigns naked-pointer.
	// Returns assigned naked-pointer.
	//--------------------------------
	T*  Assign(T* p, bool bOwnsResource = true) {
		this->operator=(AutoPtr(p, bOwnsResource));
		return pResource;
	}

	//------------------------------------------------
	// Deletes resource if it is owned by this object.
	// Object now points to zero.
	// Usage example:
	// AutoPtr<int> ap(new int);
	// ap = AutoPtr<int>(0);  //long version
	// ap.SafeDelete();       //short version
	//------------------------------------------------
	void  SafeDelete() {
		if(bOwner) {
			bOwner = false;
			if(pResource) {
				delete pResource;
			}
		}
		pResource = 0;
	}

	//--------------------------------------------------------
	// Removes resource ownership.
	// Returns naked-pointer. Object still points to resource.
	//--------------------------------------------------------
	T*  Release() const {
		bOwner = false;
		return pResource;
	}

	//----------------------
	// Naked-pointer access.
	//----------------------
	T*  Get() const { return pResource; }
	operator  T*() const { return pResource; }

	//---------------
	// State queries.
	//---------------
	bool  IsValid() const { return (pResource != 0); }
	bool  IsZero() const { return (pResource == 0); }
	bool  IsOwner() const { return bOwner; }

	//--------------------------
	// Dereference operator.
	// Usage example:
	// AutoPtr<int> ap(new int);
	// *ap = 1234;
	//--------------------------
	T&   operator*() const { return *pResource; }

	//----------------------------
	// Delegation operator.
	// Usage example:
	// struct B { int x; int y; };
	// AutoPtr<B> ap(new B);
	// ap->x = 1;
	// ap->y = 2;
	//----------------------------
	T*   operator->() const { return pResource; }

	//---------------------------
	// Automatic type conversion.
	// Usage example:
	// AutoPtr<int> ap(new int);
	// if(  ap) {}
	// if(! ap) {}
	//---------------------------
	bool operator!() const { return (pResource == 0); }

	//-----------------------------
	// Comparisons AutoPtr/AutoPtr.
	// Usage example:
	// AutoPtr<int> ap1(new int);
	// AutoPtr<int> ap2(new int);
	// if(ap1 == ap2) {}
	// if(ap1 != ap2) {}
	//-----------------------------
	bool  operator==(const AutoPtr& rhs) const { return (pResource == rhs.pResource); }
	bool  operator!=(const AutoPtr& rhs) const { return (pResource != rhs.pResource); }

	//-----------------------------------
	// Comparisons AutoPtr/naked-pointer.
	// Usage example:
	// AutoPtr<int> ap(new int);
	// int* p = 0;
	// if(ap == p) {}
	// if(ap != p) {}
	//-----------------------------------
	friend  bool  operator==(const AutoPtr& a, const T* p) { return (a.pResource == p); }
	friend  bool  operator==(const T* p, const AutoPtr& a) { return (a.pResource == p); }
	friend  bool  operator!=(const AutoPtr& a, const T* p) { return (a.pResource != p); }
	friend  bool  operator!=(const T* p, const AutoPtr& a) { return (a.pResource != p); }


private:
	T*            pResource;
	mutable bool  bOwner;
};


//====================================================================
// Inheritance casts for AutoPtr.
//====================================================================
template<typename DerivedType, typename BaseType>
__forceinline  AutoPtr<DerivedType> __fastcall StaticCast(AutoPtr<BaseType>& src)
{
	return  AutoPtr<DerivedType>(
								static_cast<DerivedType*>(src.Release())
								);
}


//====================================================================
// Reference-style-access pointer.
// Completely unsafe.
//====================================================================
template<typename T>
class Reference
{
public:
	Reference(T* p = 0) { SetPtr(p); }

	void  SetPtr(T* p = 0) { pObject = p; }

	T*        GetPtr() { return pObject; }
	const T*  GetPtr() const { return pObject; }

	T&        GetRef() { return *pObject; }
	const T&  GetRef() const { return *pObject; }

	operator T&() { return GetRef(); }
	operator const T&() const { return GetRef(); }


private:
	T*  pObject;
};


//********************************************************************
}//namespace vertexware


