

//--------------------------------------------------------------------
#include "BaseTools\NamedParameter.h"
#include "BaseTools\STLExtensions.h"


//====================================================================
#define  ASSERT_VALUE_TYPE(ValueType) \
	VW_ASSERT(valueType == ValueType, \
			"Extracting wrong value-type from NamedParameter: " + parameterName);


//====================================================================
#define  IMPLEMENT_CTOR \
	parameterName = _parameterName; \
	SetValue(val);


//====================================================================
#define  IMPLEMENT_GET_VALUE(ValueType) \
	const NamedParameter* np = Find(parameterName); \
	if(np  &&  np->GetValueType() == ValueType) { \
		np->GetValue(val); \
		return true; \
	} \
	return false;


//====================================================================
#define  DEFAULT_LIST_RESERVE_SIZE  10


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
class ParameterNamePredicate
{
public:
	ParameterNamePredicate(const std::string& parameterName) : targetName(parameterName) {
	}

	bool operator()(const NamedParameter& np) const {
		return (np.GetName() == targetName);
	}

private:
	std::string  targetName;
};


//====================================================================
struct ParameterSortPredicate
{
	bool operator()(const NamedParameter& lhs, const NamedParameter& rhs) const {
		return (lhs.GetName() < rhs.GetName());
	}
};


//====================================================================
struct ParameterBinarySearchPredicate
{
	bool operator()(const NamedParameter& np, const std::string& refName) const {
		return (np.GetName() < refName);
	}
};


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
NamedParameter::NamedParameter()
{
	valueType = ValueType_Unknown;
}


//====================================================================
NamedParameter::NamedParameter(const std::string& _parameterName, void* const& val)
{
	IMPLEMENT_CTOR;
}


//====================================================================
NamedParameter::NamedParameter(const std::string& _parameterName, const bool& val)
{
	IMPLEMENT_CTOR;
}


//====================================================================
NamedParameter::NamedParameter(const std::string& _parameterName, const int& val)
{
	IMPLEMENT_CTOR;
}


//====================================================================
NamedParameter::NamedParameter(const std::string& _parameterName, const uint& val)
{
	IMPLEMENT_CTOR;
}


//====================================================================
NamedParameter::NamedParameter(const std::string& _parameterName, const double& val)
{
	IMPLEMENT_CTOR;
}


//====================================================================
NamedParameter::NamedParameter(const std::string& _parameterName, const std::string& val)
{
	IMPLEMENT_CTOR;
}


//====================================================================
void*  NamedParameter::GetValuePointer() const
{
	ASSERT_VALUE_TYPE(ValueType_Pointer);
	return valPointer;
}


//====================================================================
bool  NamedParameter::GetValueBool() const
{
	ASSERT_VALUE_TYPE(ValueType_Bool);
	return valBool;
}


//====================================================================
int  NamedParameter::GetValueInteger() const
{
	ASSERT_VALUE_TYPE(ValueType_Integer);
	return valInteger;
}


//====================================================================
uint  NamedParameter::GetValueUInteger() const
{
	ASSERT_VALUE_TYPE(ValueType_UInteger);
	return valUInteger;
}


//====================================================================
double  NamedParameter::GetValueDouble() const
{
	ASSERT_VALUE_TYPE(ValueType_Double);
	return valDouble;
}


//====================================================================
std::string  NamedParameter::GetValueString() const
{
	ASSERT_VALUE_TYPE(ValueType_String);
	return valString;
}


//====================================================================
void  NamedParameter::SetValuePointer(void* val)
{
	valueType = ValueType_Pointer;
	valPointer = val;
}


//====================================================================
void  NamedParameter::SetValueBool(bool val)
{
	valueType = ValueType_Bool;
	valBool = val;
}


//====================================================================
void  NamedParameter::SetValueInteger(int val)
{
	valueType = ValueType_Integer;
	valInteger = val;
}


//====================================================================
void  NamedParameter::SetValueUInteger(uint val)
{
	valueType = ValueType_UInteger;
	valUInteger = val;
}


//====================================================================
void  NamedParameter::SetValueDouble(double val)
{
	valueType = ValueType_Double;
	valDouble = val;
}


//====================================================================
void  NamedParameter::SetValueString(const std::string& val)
{
	valueType = ValueType_String;
	valString = val;
}


//====================================================================
NamedParameterList::NamedParameterList()
{
	parameters.reserve(DEFAULT_LIST_RESERVE_SIZE);
}


//====================================================================
NamedParameterList::~NamedParameterList()
{
}


//====================================================================
bool  NamedParameterList::Insert(const NamedParameter& np, bool bReplaceIfExists)
{
	// check for duplicate names
	while(Find(np.GetName())) {
		if(bReplaceIfExists) {
			Remove(np.GetName());
		}
		else {
			return false;
		}
	}

	// store parameter object
	parameters.push_back(np);
	SortParameters();
	return true;
}


//====================================================================
void  NamedParameterList::Remove(const std::string& parameterName)
{
	PARAM_ITER it = std::find_if(parameters.begin(), parameters.end(),
								ParameterNamePredicate(parameterName));
	if(it != parameters.end()) {
		parameters.erase(it);
		SortParameters();
	}
}


//====================================================================
void  NamedParameterList::Clear()
{
	parameters.clear();
	parameters.reserve(DEFAULT_LIST_RESERVE_SIZE);
}


//====================================================================
bool  NamedParameterList::GetValue(const std::string& parameterName, void*& val) const
{
	IMPLEMENT_GET_VALUE(NamedParameter::ValueType_Pointer);
}


//====================================================================
bool  NamedParameterList::GetValue(const std::string& parameterName, bool& val) const
{
	IMPLEMENT_GET_VALUE(NamedParameter::ValueType_Bool);
}


//====================================================================
bool  NamedParameterList::GetValue(const std::string& parameterName, int& val) const
{
	IMPLEMENT_GET_VALUE(NamedParameter::ValueType_Integer);
}


//====================================================================
bool  NamedParameterList::GetValue(const std::string& parameterName, uint& val) const
{
	IMPLEMENT_GET_VALUE(NamedParameter::ValueType_UInteger);
}


//====================================================================
bool  NamedParameterList::GetValue(const std::string& parameterName, double& val) const
{
	IMPLEMENT_GET_VALUE(NamedParameter::ValueType_Double);
}


//====================================================================
bool  NamedParameterList::GetValue(const std::string& parameterName, std::string& val) const
{
	IMPLEMENT_GET_VALUE(NamedParameter::ValueType_String);
}


//====================================================================
NamedParameter*  NamedParameterList::Find(const std::string& parameterName)
{
	PARAM_ITER it;

	// search sorted sequence O(log(N))
	it = std::lower_bound(parameters.begin(), parameters.end(),
						parameterName, ParameterBinarySearchPredicate());

	if(it == parameters.end()) return 0;
	NamedParameter& np = *it;
	return (np.GetName() == parameterName) ? &np : 0;
}


//====================================================================
void  NamedParameterList::SortParameters()
{
	std::sort(parameters.begin(), parameters.end(), ParameterSortPredicate());
}


//********************************************************************
}//namespace vertexware


