

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Exception.h"
#include "BaseTools\String.h"
#include "BaseTools\Memory.h"
#include "private\vw_common_exports.h"
#include <string>
#include <vector>


//********************************************************************
namespace vertexware {


//====================================================================
// Named-parameter.
//====================================================================
class VW_COMMON_API NamedParameter
{
public:
	enum eValueType
	{
		ValueType_Unknown = 0,
		ValueType_Pointer,
		ValueType_Bool,
		ValueType_Integer,
		ValueType_UInteger,
		ValueType_Double,
		ValueType_String,

		ValueType_Count,
		ValueType_Force32Bits = 0x7FFFFFFF
	};


public:
	NamedParameter();
	NamedParameter(const std::string& parameterName, void* const& val);
	NamedParameter(const std::string& parameterName, const bool& val);
	NamedParameter(const std::string& parameterName, const int& val);
	NamedParameter(const std::string& parameterName, const uint& val);
	NamedParameter(const std::string& parameterName, const double& val);
	NamedParameter(const std::string& parameterName, const std::string& val);

	void                SetName(const std::string& s) { parameterName = s; }
	const std::string&  GetName() const { return parameterName; }

	eValueType  GetValueType() const { return valueType; }

	void  GetValue(void*& val) const { val = GetValuePointer(); }
	void  GetValue(bool& val) const { val = GetValueBool(); }
	void  GetValue(int& val) const { val = GetValueInteger(); }
	void  GetValue(uint& val) const { val = GetValueUInteger(); }
	void  GetValue(double& val) const { val = GetValueDouble(); }
	void  GetValue(std::string& val) const { val = GetValueString(); }

	void*        GetValuePointer() const;
	bool         GetValueBool() const;
	int          GetValueInteger() const;
	uint         GetValueUInteger() const;
	double       GetValueDouble() const;
	std::string  GetValueString() const;

	void  SetValue(void* const& val) { SetValuePointer(val); }
	void  SetValue(const bool& val) { SetValueBool(val); }
	void  SetValue(const int& val) { SetValueInteger(val); }
	void  SetValue(const uint& val) { SetValueUInteger(val); }
	void  SetValue(const double& val) { SetValueDouble(val); }
	void  SetValue(const std::string& val) { SetValueString(val); }

	void  SetValuePointer(void* val);
	void  SetValueBool(bool val);
	void  SetValueInteger(int val);
	void  SetValueUInteger(uint val);
	void  SetValueDouble(double val);
	void  SetValueString(const std::string& val);


private:
	std::string  parameterName;
	eValueType   valueType;

	union {
		void*   valPointer;
		bool    valBool;
		int     valInteger;
		uint    valUInteger;
		double  valDouble;
	};
	std::string  valString; //cannot put in union, has copy constructor
};


//====================================================================
// Named-parameter list.
// Defines named-parameter storage and extraction interface.
// All parameters are cloned when inserted into list.
//====================================================================
class VW_COMMON_API NamedParameterList
{
public:
	NamedParameterList();
	~NamedParameterList();

	// returns false if parameter with the same name is already inserted
	bool  Insert(const NamedParameter& np, bool bReplaceIfExists = true);
	void  Remove(const std::string& parameterName);
	void  Clear();

	// extraction by ordinal
	uint  GetNumParameters() const { return parameters.size(); }
	const NamedParameter&  GetParameter(uint i) const { return parameters[i]; }

	// extraction by name and type, returns 0 if not found
	const NamedParameter*  GetParameter(const std::string& parameterName) const {
		return Find(parameterName);
	}

	// direct value-extraction, name & type must match, returns false otherwise
	bool  GetValue(const std::string& parameterName, void*& val) const;
	bool  GetValue(const std::string& parameterName, bool& val) const;
	bool  GetValue(const std::string& parameterName, int& val) const;
	bool  GetValue(const std::string& parameterName, uint& val) const;
	bool  GetValue(const std::string& parameterName, double& val) const;
	bool  GetValue(const std::string& parameterName, std::string& val) const;

	// direct value-extraction, returns fail-value if not found
	void*  GetValuePointer(const std::string& parameterName, void* failValue = 0) const {
		void* v;
		return GetValue(parameterName, v) ? v : failValue;
	}
	bool  GetValueBool(const std::string& parameterName, bool failValue = false) const {
		bool v;
		return GetValue(parameterName, v) ? v : failValue;
	}
	int  GetValueInteger(const std::string& parameterName, int failValue = 0) const {
		int v;
		return GetValue(parameterName, v) ? v : failValue;
	}
	uint  GetValueUInteger(const std::string& parameterName, uint failValue = 0) const {
		uint v;
		return GetValue(parameterName, v) ? v : failValue;
	}
	double  GetValueDouble(const std::string& parameterName, double failValue = 0) const {
		double v;
		return GetValue(parameterName, v) ? v : failValue;
	}
	std::string  GetValueString(const std::string& parameterName, const std::string failValue = "") const {
		std::string v;
		return GetValue(parameterName, v) ? v : failValue;
	}


private:
	NamedParameter*  Find(const std::string& parameterName);

	const NamedParameter*  Find(const std::string& parameterName) const {
		return const_cast<NamedParameterList*>(this)->Find(parameterName);
	}

	void  SortParameters();


private:
	typedef  std::vector<NamedParameter>  PARAM_VEC;
	typedef  PARAM_VEC::iterator          PARAM_ITER;
	typedef  PARAM_VEC::const_iterator    PARAM_CITER;

	PARAM_VEC  parameters;
};


//********************************************************************
}//namespace vertexware

