

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\CompilerConfig.h"
#include "BaseTools\Types.h"
#include "BaseTools\Constructs.h"
#include "private\vw_common_exports.h"
#include <ios>
#include <string>
#include <vector>


//********************************************************************
namespace vertexware {


//====================================================================
// Output-form for numbers.
//====================================================================
class VW_COMMON_API NumberOutputForm
{
public:
	enum eFPNumberFormat
	{
		Automatic  = 0,
		Fixed      = std::ios_base::fixed,
		Scientific = std::ios_base::scientific
	};


public:
	NumberOutputForm();

	void  NumberWidth(int wid = 0) { numberWidth = wid; }
	int   NumberWidth() const { return numberWidth; }

	void  FPNumberPrecision(int prec = 6) { fpNumberPrecision = prec; }
	int   FPNumberPrecision() const { return fpNumberPrecision; }

	void  FPNumberFormat(eFPNumberFormat fmt = Automatic) { fpNumberFormat = fmt; }
	int   FPNumberFormat() const { return fpNumberFormat; }


protected:
	int  numberWidth;
	int  fpNumberPrecision;
	int  fpNumberFormat;
};


//====================================================================
// Output-form for number sequences.
//====================================================================
class VW_COMMON_API NumberSequenceOutputForm : public NumberOutputForm
{
public:
	NumberSequenceOutputForm();

	void  Multiline(bool b = false) { bMultiline = b; }
	bool  Multiline() const { return bMultiline; }

	void  SequenceProlog(const std::string& s = "[") { sequenceProlog = s; }
	const std::string& SequenceProlog() const { return sequenceProlog; }

	void  NumberSeparator(const std::string& s = ", ") { numberSeparator = s; }
	const std::string& NumberSeparator() const { return numberSeparator; }

	void  SequenceEpilog(const std::string& s = "]") { sequenceEpilog = s; }
	const std::string& SequenceEpilog() const { return sequenceEpilog; }

	//-------------------------------------------------------------
	// Fill <outProlog> & return true to specify per-number-prolog.
	//-------------------------------------------------------------
	virtual  bool  GetNumberProlog(uint numberPosition, std::string& outProlog) const
	{ return false; }

	//-------------------------------------------------------------
	// Fill <outEpilog> & return true to specify per-number-epilog.
	//-------------------------------------------------------------
	virtual  bool  GetNumberEpilog(uint numberPosition, std::string& outEpilog) const
	{ return false; }


protected:
	bool         bMultiline;
	std::string  sequenceProlog;
	std::string  numberSeparator;
	std::string  sequenceEpilog;
};


//====================================================================
// Form bound to specific sequence.
//====================================================================
class VW_COMMON_API NumberSequenceBoundOutputForm
{
public:
	NumberSequenceBoundOutputForm(const NumberSequenceOutputForm& _form,
							uint _elemCount, uint _lineLength) : form(_form) {
		elemCount  = _elemCount;
		lineLength = ClampedLow<uint>(_lineLength, 1);
	}

public:
	enum eValueType
	{
		ValueType_Int = 0,
		ValueType_UInt,
		ValueType_Double,

		ValueType_Count,
		ValueType_Force32Bits = 0x7FFFFFFF
	};

	virtual  eValueType  GetValueType() const = 0;
	virtual  int         GetValueInt(uint i) const { return 0; }
	virtual  uint        GetValueUInt(uint i) const { return 0; }
	virtual  double      GetValueDouble(uint i) const { return 0; }

public:
	const NumberSequenceOutputForm&  form;
	uint  elemCount;
	uint  lineLength;
};


//====================================================================
// Outputs supplied number-sequence bound-form.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<<(std::ostream& os,
									const NumberSequenceBoundOutputForm& boundForm);


//====================================================================
// Typed bound-output-form derivation macros.
//====================================================================
#define  VW_DECLARE_NUMBER_SEQUENCE_BOUND_OUTPUT_FORM( ClassName, ValueType ) \
	public: \
		ClassName(const NumberSequenceOutputForm& form, \
					const ValueType* values, uint count, uint lineLength) : \
			NumberSequenceBoundOutputForm(form, count, lineLength), pValues(values) {} \
		const ValueType& GetValue(uint i) const { return pValues[i]; } \
	private: \
		const ValueType* pValues;


//====================================================================
// Output float-s as double-s.
//====================================================================
class VW_COMMON_API NumberSequenceBoundOutputForm_float : public NumberSequenceBoundOutputForm
{
VW_DECLARE_NUMBER_SEQUENCE_BOUND_OUTPUT_FORM( NumberSequenceBoundOutputForm_float, float );

public:
	virtual  eValueType  GetValueType() const { return ValueType_Double; }
	virtual  double      GetValueDouble(uint i) const { return GetValue(i); }
};


//====================================================================
// Output ubyte-s as uint-s.
//====================================================================
class VW_COMMON_API NumberSequenceBoundOutputForm_ubyte : public NumberSequenceBoundOutputForm
{
VW_DECLARE_NUMBER_SEQUENCE_BOUND_OUTPUT_FORM( NumberSequenceBoundOutputForm_ubyte, ubyte );

public:
	virtual  eValueType  GetValueType() const { return ValueType_UInt; }
	virtual  uint        GetValueUInt(uint i) const { return GetValue(i); }
};


//====================================================================
// Output-form derivation macros.
// All output-forms derived from NumberSequenceOutputForm must have
// default constructor available.
//====================================================================
#define  VW_DECLARE_NUMBER_SEQUENCE_OUTPUT_FORM( OutputFormName ) \
	public: \
	/* Allow copy from other suitable forms. */ \
	OutputFormName& operator=(const NumberSequenceOutputForm& ref) { \
		static_cast<NumberSequenceOutputForm&>(*this) = ref; \
		return *this; \
	} \
	/* Current output form (used by operator<<). */ \
	static  void __fastcall SetCurrentOutputForm(const NumberSequenceOutputForm* p = 0) { \
		if(p) s_pCurrentOutputForm = p; \
		else  s_pCurrentOutputForm = &s_defaultOutputForm; \
	} \
	static  const NumberSequenceOutputForm& __fastcall GetCurrentOutputForm() { \
		return *s_pCurrentOutputForm; \
	} \
	private: \
	static  const  OutputFormName             s_defaultOutputForm; \
	static  const  NumberSequenceOutputForm*  s_pCurrentOutputForm;


//====================================================================
#define  VW_IMPLEMENT_NUMBER_SEQUENCE_OUTPUT_FORM( OutputFormName ) \
	const OutputFormName             OutputFormName::s_defaultOutputForm; \
	const NumberSequenceOutputForm*  OutputFormName::s_pCurrentOutputForm = &s_defaultOutputForm;


//********************************************************************
}//namespace vertexware


