

//--------------------------------------------------------------------
#include "BaseTools\Profiler.h"
#include "BaseTools\Memory.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Debug.h"
#include "BaseTools\String.h"
#include "BaseTools\STLExtensions.h"
#include <iomanip>


//====================================================================
#define  BLOCK_NAME_SPACE_RESERVE  128


//********************************************************************
namespace vertexware {


//====================================================================
ProfilerDumpDesc::ProfilerDumpDesc()
{
	linePrefix = " ";
	categorySeparator = " | ";
	valueSeparator = "  ";
	levelIndent = 2;
	numLinesPerColorCodedBlock = 3;
	bPrintEntryCounts = true;
	bPrintNodeTimes = true;
	bPrintBranchTimes = true;
	bPrintNodePercentages = true;
	entryCountWidth = 3;
	timeWidth = 8;
	timePrecision = 6;
	percentageWidth = 7;
	percentagePrecision = 3;

	ConstructHeaderLines();
}


//====================================================================
void  ProfilerDumpDesc::ConstructHeaderLines()
{
	headerLines.clear();
	headerLines.push_back(std::string(linePrefix.length(), ' '));
	headerLines.push_back(linePrefix);

	if(bPrintEntryCounts) {
		ConstructHeaderLines(headerLines[0], headerLines[1], "EntryCounts", entryCountWidth);
	}
	if(bPrintNodeTimes) {
		ConstructHeaderLines(headerLines[0], headerLines[1], "NodeTimes", timeWidth);
	}
	if(bPrintBranchTimes) {
		ConstructHeaderLines(headerLines[0], headerLines[1], "BranchTimes", timeWidth);
	}
	if(bPrintNodePercentages) {
		ConstructHeaderLines(headerLines[0], headerLines[1], "NodePercentages", percentageWidth);
	}
	headerLines[0].append("BlockNames");
}


//====================================================================
void  ProfilerDumpDesc::ConstructHeaderLines(std::string& line0, std::string& line1,
										const std::string& title, uint valueWidth) const
{
	int  valueSeparatorWidth = valueSeparator.length();
	int  maxLen = valueWidth * 3 + valueSeparatorWidth * 2;

	std::string  padt(Max<int>(maxLen - title.length(), 0), ' ');
	std::string  padr(Max<int>(valueWidth - 3, 0), ' ');
	std::string  vsep(valueSeparatorWidth, ' ');

	line0.append(title).append(padt).append(categorySeparator);

	line1.append("Min").append(padr).append(vsep);
	line1.append("Avg").append(padr).append(vsep);
	line1.append("Max").append(padr).append(categorySeparator);
}


//====================================================================
ProfilerNode::ProfilerNode(const std::string& _blockName, ProfilerNode* _parent)
	:
	blockName(_blockName)
{
	PropagateResetStats();
	PropagateResetIterAccum();

	iNextSearch = 0;

	parent = _parent;
	if(parent) parent->AddChild(*this);
}


//====================================================================
ProfilerNode::~ProfilerNode()
{
	for(uint i = 0; i < children.size(); i++) {
		delete children[i];
	}
}


//====================================================================
ProfilerNode*  ProfilerNode::FindChild(const std::string& blockName)
{
	uint max = children.size();
	if(! max) return 0;

	// cyclic search
	uint i = iNextSearch;
	while(true) {

		ProfilerNode* child = children[i];
		if(child->blockName == blockName) {

			iNextSearch = i + 1;
			if(iNextSearch >= max) iNextSearch = 0;
			return child;
		}
		i++;
		if(i >= max) i = 0;
		if(i == iNextSearch) break;
	}
	return 0;
}


//====================================================================
ProfilerNode*  ProfilerNode::FindDescendant(const std::string& blockName)
{
	for(uint i = 0; i < children.size(); i++) {
		ProfilerNode* c = children[i];
		if(c->blockName == blockName) return c;
		c = c->FindDescendant(blockName);
		if(c) return c;
	}
	return 0;
}


//====================================================================
std::string  ProfilerNode::FormatStatistics(const ProfilerDumpDesc& pdd, uint level) const
{
	using namespace std;
	std::ostringstream  oss;

	oss << pdd.linePrefix;

	// entry points
	if(pdd.bPrintEntryCounts) {
		oss << setw(pdd.entryCountWidth) << entryCountMin << pdd.valueSeparator;
		oss << setw(pdd.entryCountWidth) << entryCountAvg << pdd.valueSeparator;
		oss << setw(pdd.entryCountWidth) << entryCountMax;
		oss << pdd.categorySeparator;
	}

	oss.setf(std::ios::floatfield, std::ios::fixed);
	oss.precision(pdd.timePrecision);

	// node times
	if(pdd.bPrintNodeTimes) {
		oss << setw(pdd.timeWidth) << timeNodeMin << pdd.valueSeparator;
		oss << setw(pdd.timeWidth) << timeNodeAvg << pdd.valueSeparator;
		oss << setw(pdd.timeWidth) << timeNodeMax;
		oss << pdd.categorySeparator;
	}

	// branch times
	if(pdd.bPrintBranchTimes) {
		oss << setw(pdd.timeWidth) << timeBranchMin << pdd.valueSeparator;
		oss << setw(pdd.timeWidth) << timeBranchAvg << pdd.valueSeparator;
		oss << setw(pdd.timeWidth) << timeBranchMax;
		oss << pdd.categorySeparator;
	}

	oss.precision(pdd.percentagePrecision);

	// node percentages
	if(pdd.bPrintNodePercentages) {
		oss << setw(pdd.percentageWidth) << percentageNodeMin << pdd.valueSeparator;
		oss << setw(pdd.percentageWidth) << percentageNodeAvg << pdd.valueSeparator;
		oss << setw(pdd.percentageWidth) << percentageNodeMax;
		oss << pdd.categorySeparator;
	}

	FillEffector fe(pdd.levelIndent * level, ' ');
	oss << fe << blockName;

	return oss.str();
}


//====================================================================
void  ProfilerNode::DumpStatistics(std::ostream& os, uint& lineNum, const ProfilerDumpDesc& pdd, uint level) const
{
	if(pdd.numLinesPerColorCodedBlock) {
		bool bEven = (lineNum % (pdd.numLinesPerColorCodedBlock * 2) >= pdd.numLinesPerColorCodedBlock);
		if(bEven) os << e_info;
		else      os << e_attention;
	}

	os << FormatStatistics(pdd, level) << nl;
	lineNum++;

	for(uint i = 0; i < children.size(); i++) {
		children[i]->DumpStatistics(os, lineNum, pdd, level + 1);
	}
}


//====================================================================
void  ProfilerNode::GetStats(uint _entryCountMin, uint _entryCountMax, uint _entryCountAvg,
							double _timeNodeMin, double _timeNodeMax, double _timeNodeAvg,
							double _timeBranchMin, double _timeBranchMax, double _timeBranchAvg,
							float _percentageNodeMin, float _percentageNodeMax, float _percentageNodeAvg) const
{
	_entryCountMin = entryCountMin;
	_entryCountMax = entryCountMax;
	_entryCountAvg = entryCountAvg;

	_timeNodeMin = timeNodeMin;
	_timeNodeMax = timeNodeMax;
	_timeNodeAvg = timeNodeAvg;

	_timeBranchMin = timeBranchMin;
	_timeBranchMax = timeBranchMax;
	_timeBranchAvg = timeBranchAvg;

	_percentageNodeMin = percentageNodeMin;
	_percentageNodeMax = percentageNodeMax;
	_percentageNodeAvg = percentageNodeAvg;
}


//====================================================================
void  ProfilerNode::AddChild(ProfilerNode& child)
{
	children.push_back(&child);
	iNextSearch = 0;
}


//====================================================================
void  ProfilerNode::PropagateResetIterAccum()
{
	entryCountIterAccum = 0;
	timeNodeIterAccum = 0;

	for(uint i = 0; i < children.size(); i++) {
		children[i]->PropagateResetIterAccum();
	}
}


//====================================================================
void  ProfilerNode::PropagateProcessIterAccum(double oneHOverFrameTime)
{
	entryCountMin = Min(entryCountMin, entryCountIterAccum);
	entryCountMax = Max(entryCountMax, entryCountIterAccum);
	entryCountTotal += entryCountIterAccum;
	entryCountIterAccum = 0;

	double percentageNode = timeNodeIterAccum * oneHOverFrameTime;
	percentageNodeTotal += percentageNode;
	percentageNodeMin = Min(percentageNodeMin, percentageNode);
	percentageNodeMax = Max(percentageNodeMax, percentageNode);

	timeNodeTotal += timeNodeIterAccum;
	timeNodeMin = Min(timeNodeMin, timeNodeIterAccum);
	timeNodeMax = Max(timeNodeMax, timeNodeIterAccum);
	timeNodeIterAccum = 0;

	for(uint i = 0; i < children.size(); i++) {
		children[i]->PropagateProcessIterAccum(oneHOverFrameTime);
	}
}


//====================================================================
void  ProfilerNode::PropagateResetStats()
{
	entryCountTotal = 0;
	entryCountMin = UINT_MAX;
	entryCountMax = 0;
	entryCountAvg = 0;

	timeNodeTotal = 0;
	timeNodeMin = FLT_MAX;
	timeNodeMax = 0;
	timeNodeAvg = 0;

	timeBranchMin = 0;
	timeBranchMax = 0;
	timeBranchAvg = 0;

	percentageNodeTotal = 0;
	percentageNodeMin = 100;
	percentageNodeMax = 0;
	percentageNodeAvg = 0;

	for(uint i = 0; i < children.size(); i++) {
		children[i]->PropagateResetStats();
	}
}


//====================================================================
void  ProfilerNode::PropagateProcessStats(uint numIterations, const ProfilerNode& root)
{
	if(! numIterations) return;
	double dOneOverNumIter = 1.0 / double(numIterations);

	entryCountAvg = entryCountTotal * dOneOverNumIter;
	timeNodeAvg = timeNodeTotal * dOneOverNumIter;
	percentageNodeAvg = percentageNodeTotal * dOneOverNumIter;

	// process descendants & accum descendant time
	double descendantTimeMin = 0;
	double descendantTimeMax = 0;
	double descendantTimeAvg = 0;

	for(uint i = 0; i < children.size(); i++) {
		ProfilerNode* c = children[i];

		c->PropagateProcessStats(numIterations, root);

		descendantTimeMin += c->timeBranchMin;
		descendantTimeMax += c->timeBranchMax;
		descendantTimeAvg += c->timeBranchAvg;
	}

	// calc node times
	timeBranchMin = timeNodeMin + descendantTimeMin;
	timeBranchMax = timeNodeMax + descendantTimeMax;
	timeBranchAvg = timeNodeAvg + descendantTimeAvg;
}


//====================================================================
Profiler::Profiler()
{
	blockName.reserve(BLOCK_NAME_SPACE_RESERVE);

	pRootNode = 0;
	ClearNodes();

	SetMaxExpectedStackDepth();
	SetAutoResetThreshold();

	bEnabled = false;
	bRecursionHandlerEnabled = false;
	bUnsafe = false;
	elapsedSeconds = 0;
	bRootEntered = false;
	bRecursionWarningIgnoreFlag = false;
}


//====================================================================
Profiler::~Profiler()
{
	SafeDelete(pRootNode);
}


//====================================================================
void  Profiler::RegisterObserver(Observer& ob)
{
	UniqueContainerPushBack(observers, &ob);
}


//====================================================================
void  Profiler::UnregisterObserver(Observer& ob)
{
	EraseFromContainer(observers, &ob);
}


//====================================================================
void  Profiler::Enable(bool _bEnable)
{
	bEnabled = _bEnable;

	ResetStack();
	bUnsafe = true;

	for(uint i = 0; i < observers.size(); i++) {
		observers[i]->OnProfilerEnabled(bEnabled);
	}
}


//====================================================================
void  Profiler::EnableRecursionHandler(bool bEnable)
{
	bRecursionHandlerEnabled = bEnable;
	ClearNodes();
}


//====================================================================
void  Profiler::SetMaxExpectedStackDepth(uint maxDepth)
{
	maxExpectedStackDepth = Max<uint>(maxDepth, 2);
}


//====================================================================
void  Profiler::ClearNodes()
{
	bUnsafe = true;
	bRecursionWarningIgnoreFlag = false;

	// create new root
	SafeDelete(pRootNode);
	pRootNode = new ProfilerNode("Root", 0);

	// prepare internals
	ResetStack();
	ResetStatistics();
}


//====================================================================
void  Profiler::ResetStatistics()
{
	bAutoReset = (autoResetThreshold != 0);
	elapsedSeconds = 0;

	pRootNode->PropagateResetStats();
	pRootNode->PropagateResetIterAccum();
}


//====================================================================
void  Profiler::ProcessStatistics()
{
	pRootNode->PropagateProcessStats(pRootNode->entryCountTotal, *pRootNode);
}


//====================================================================
void  Profiler::DumpStatistics(std::ostream& os, const ProfilerDumpDesc& pdd) const
{
	double irMin, irAvg, irMax;
	CalcIterationRate(irMin, irAvg, irMax);

	os << "RecursionHandlerEnabled.........: " << ToString(IsEnabledRecursionHandler()) << nl;
	os << "AutoResetThreshold..............: " << GetAutoResetThreshold() << nl;
	os << "IterationCount..................: " << GetIterationCount() << nl;
	os << "ElapsedSeconds..................: " << GetElapsedSeconds() << nl;
	os << "IterationsPerSecond(Min/Avg/Max): ";

	const char* sep = " / ";
	uint  numDigits = 3;
	os << ToStringFix(irMin, 0, numDigits) << sep;
	os << ToStringFix(irAvg, 0, numDigits) << sep;
	os << ToStringFix(irMax, 0, numDigits) << "\n\n";

	if(GetIterationCount()) {
		for(uint i = 0; i < pdd.headerLines.size(); i++) {
			os << pdd.headerLines[i] << nl;
		}
		uint lineNum = 0;
		pRootNode->DumpStatistics(os, lineNum, pdd, 0);
		os << nl << e_info;
	}
}


//====================================================================
uint  Profiler::GetIterationCount() const
{
	return pRootNode->entryCountTotal;
}


//====================================================================
double  Profiler::GetElapsedSeconds() const
{
	return elapsedSeconds;
}


//====================================================================
void  Profiler::CalcIterationRate(double& iterPerSecMin, double& iterPerSecAvg, double& iterPerSecMax) const
{
	uint iterCount = GetIterationCount();
	if(iterCount) {
		iterPerSecMin = CalcIterRate(pRootNode->timeBranchMax);
		iterPerSecAvg = CalcIterRate(pRootNode->timeBranchAvg);
		iterPerSecMax = CalcIterRate(pRootNode->timeBranchMin);
	}
	else {
		iterPerSecMin = 0;
		iterPerSecAvg = 0;
		iterPerSecMax = 0;
	}
}


//====================================================================
void  Profiler::EnterRootNode()
{
	bUnsafe = false;

	VW_ASSERT(! bRootEntered, "Profile-stack recursive root-entry.");
	VW_ASSERT(profileStack.size() == 1, "Profile-stack push/pop mismatch.");

#ifdef VW_BUILD_DEBUG
	bRootEntered = true;
#endif

	if(! pRootNode->entryCountTotal) {
		elapsedSeconds = 0;
	}
	pRootNode->entryCountIterAccum++;

	timer.Start();
	ActiveStackEntry()->startTime = timer.GetElapsedSeconds();
}


//====================================================================
void  Profiler::ExitRootNode()
{
#ifdef VW_BUILD_DEBUG
	bRootEntered = false;
#endif

	if(bUnsafe) {
		ResetStack();
		bUnsafe = false;
		return;
	}

	VW_ASSERT(profileStack.size() == 1, "Profile-stack push/pop mismatch.");

	double secsDelta = timer.GetElapsedSeconds() - ActiveStackEntry()->startTime;
	elapsedSeconds += secsDelta;

	double oneHOverFrameTime = (secsDelta > 0) ? 100.0 / secsDelta : 0;

	pRootNode->timeNodeIterAccum += secsDelta;
	pRootNode->PropagateProcessIterAccum(oneHOverFrameTime);

	if(bAutoReset  &&  pRootNode->entryCountTotal >= autoResetThreshold) {
		ResetStatistics();
		bAutoReset = false;
	}
}


//====================================================================
void  Profiler::EnterNode(const char* pBlockName)
{
	VW_ASSERT(pBlockName  &&  pBlockName[0] != 0, "Passing empty block-name to profiler.");
	blockName.assign(pBlockName);

	if(bUnsafe) return;

	ProfilerNode* node = 0;

	// check for recursive block entry
	if(bRecursionHandlerEnabled) {
		node = TraceStackForRecursion(blockName);
	}

	// search active node for child
	if(! node) {
		node = ActiveNode()->FindChild(blockName);
		if(! node) {
			// not found, create new node
			node = new ProfilerNode(blockName, ActiveNode());
		}
	}

	// update stack
	profileStack.push_back(ProfileStackEntry(node, timer.GetElapsedSeconds()));

	// update active node states
	ProfilerNode* pActiveNode = ActiveNode();
	pActiveNode->entryCountIterAccum++;

	// unhandled recursion warning
	VW_ASSERT_EXTERNAL_IGNORE_FLAG(
			bRecursionHandlerEnabled  ||  (profileStack.size() <= maxExpectedStackDepth),
			"Possible recursion in profile-stack.\nEnable recursion-handler, then clear and restart profiler.",
			bRecursionWarningIgnoreFlag);
}


//====================================================================
void  Profiler::ExitNode()
{
	if(bUnsafe) return;

	ProfilerNode* pActiveNode = ActiveNode();

	// ignore in case push/pop mismatch occurs
	if(pActiveNode == pRootNode) return;

	// update active node states
	double timeDelta = timer.GetElapsedSeconds() - ActiveStackEntry()->startTime;
	pActiveNode->timeNodeIterAccum += timeDelta;

	// update stack
	profileStack.pop_back();

	// update previous node states
	ActiveNode()->timeNodeIterAccum -= timeDelta;
}


//====================================================================
ProfilerNode*  Profiler::TraceStackForRecursion(const std::string& blockName)
{
	ProfilerNode* node = ActiveNode();
	while(node != pRootNode) {
		if(node->GetBlockName() == blockName) return node;
		node = node->GetParent();
	}
	return 0;
}


//====================================================================
void  Profiler::ResetStack()
{
#ifdef VW_BUILD_DEBUG
	bRootEntered = false;
#endif

	profileStack.clear();
	profileStack.push_back(ProfileStackEntry(pRootNode, 0));
}


//********************************************************************
}//namespace vertexware

