

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\Time.h"
#include <vector>
#include <cfloat>


//********************************************************************
namespace vertexware {


//====================================================================
class Profiler;
class ProfilerNode;


//====================================================================
struct VW_COMMON_API ProfilerDumpDesc
{
	std::vector<std::string>  headerLines; //without newline
	std::string               linePrefix;
	std::string               categorySeparator;
	std::string               valueSeparator;
	uint                      levelIndent;
	uint                      numLinesPerColorCodedBlock; //0 to disable color-coding
	bool                      bPrintEntryCounts;
	bool                      bPrintNodeTimes;
	bool                      bPrintBranchTimes;
	bool                      bPrintNodePercentages;
	uint                      entryCountWidth;     //num-columns
	uint                      timeWidth;           //num-columns
	uint                      timePrecision;       //num-digits after decimal-point
	uint                      percentageWidth;     //num-columns
	uint                      percentagePrecision; //num-digits after decimal-point

	ProfilerDumpDesc();

	void  ConstructHeaderLines(); //call after properties are modified


private:
	void  ConstructHeaderLines(std::string& line0, std::string& line1,
							const std::string& title, uint valueWidth) const;
};


//====================================================================
class VW_COMMON_API ProfilerNode
{
friend class Profiler;

public:
	~ProfilerNode();

	__forceinline  const std::string&  GetBlockName() const { return blockName; }

	__forceinline  ProfilerNode*        GetParent() { return parent; }
	__forceinline  const ProfilerNode*  GetParent() const { return parent; }
	__forceinline  uint                 GetNumChildren() const { return children.size(); }

	ProfilerNode*        FindChild(const std::string& blockName);
	const ProfilerNode*  FindChild(const std::string& blockName) const { return const_cast<ProfilerNode*>(this)->FindChild(blockName); }

	ProfilerNode*        FindDescendant(const std::string& blockName);
	const ProfilerNode*  FindDescendant(const std::string& blockName) const { return const_cast<ProfilerNode*>(this)->FindDescendant(blockName); }

	std::string  FormatStatistics(const ProfilerDumpDesc& pdd = ProfilerDumpDesc(), uint level = 0) const;
	void         DumpStatistics(std::ostream& os, uint& lineNum, const ProfilerDumpDesc& pdd = ProfilerDumpDesc(), uint level = 0) const;

	void  GetStats(uint entryCountMin, uint entryCountMax, uint entryCountAvg,
					double timeNodeMin, double timeNodeMax, double timeNodeAvg,
					double timeBranchMin, double timeBranchMax, double timeBranchAvg,
					float percentageNodeMin, float percentageNodeMax, float percentageNodeAvg) const;


private:
	ProfilerNode(const std::string& blockName, ProfilerNode* parent = 0);

	void  AddChild(ProfilerNode& child);

	void  PropagateResetIterAccum();
	void  PropagateProcessIterAccum(double oneHOverFrameTime);

	void  PropagateResetStats();
	void  PropagateProcessStats(uint numIterations, const ProfilerNode& root);


private:
	std::string                 blockName;
	ProfilerNode*               parent;
	std::vector<ProfilerNode*>  children;
	uint                        iNextSearch;

	uint    entryCountIterAccum;
	uint    entryCountTotal;
	uint    entryCountMin;
	uint    entryCountMax;
	uint    entryCountAvg;

	double  timeNodeIterAccum;
	double  timeNodeTotal;
	double  timeNodeMin;
	double  timeNodeMax;
	double  timeNodeAvg;

	double  timeBranchMin;
	double  timeBranchMax;
	double  timeBranchAvg;

	double  percentageNodeTotal;
	double  percentageNodeMin;
	double  percentageNodeMax;
	double  percentageNodeAvg;
};


//====================================================================
class VW_COMMON_API Profiler
{
public:
	struct VW_COMMON_API Observer
	{
		virtual ~Observer() {}
		virtual void OnProfilerEnabled(bool bEnabled) = 0;
	};


public:
	Profiler();
	~Profiler();

	void  RegisterObserver(Observer& ob);
	void  UnregisterObserver(Observer& ob);

	void               Enable(bool bEnable);
	__forceinline bool IsEnabled() const { return bEnabled; }

	// larger overhead when enabled (clears nodes internally)
	void  EnableRecursionHandler(bool bEnable = false);
	bool  IsEnabledRecursionHandler() const { return bRecursionHandlerEnabled; }

	// used to display warning if recursion-handler is inactive (abs-min=2, debug-build-only)
	void  SetMaxExpectedStackDepth(uint maxDepth = 20);
	uint  GetMaxExpectedStackDepth() const { return maxExpectedStackDepth; }

	// pass 0 to disable (used to avoid memory-allocation noise in results)
	void  SetAutoResetThreshold(uint iterationCount = 10) { autoResetThreshold = iterationCount; }
	uint  GetAutoResetThreshold() const { return autoResetThreshold; }

	void  ClearNodes();
	void  ResetStatistics();
	void  ProcessStatistics(); //process node stats, call before dump or external processing
	void  DumpStatistics(std::ostream& os, const ProfilerDumpDesc& pdd = ProfilerDumpDesc()) const;

	uint    GetIterationCount() const;
	double  GetElapsedSeconds() const;
	void    CalcIterationRate(double& iterPerSecMin, double& iterPerSecAvg, double& iterPerSecMax) const;

	const ProfilerNode&  GetRootNode() const { return *pRootNode; }

	// block entry/exit points
	void  EnterRootNode();
	void  ExitRootNode();

	void  EnterNode(const char* pBlockName);
	void  ExitNode();


private:
	// returns != 0 if recursive block entry is detected
	ProfilerNode*  TraceStackForRecursion(const std::string& blockName);
	void           ResetStack();

	//-------------
	// Trace entry.
	//-------------
	struct ProfileStackEntry
	{
		ProfilerNode*  node;
		double         startTime;

		ProfileStackEntry(ProfilerNode* _node, double _startTime) {
			node = _node;
			startTime = _startTime;
		}
	};

	__forceinline  ProfilerNode*       ActiveNode() { return profileStack.back().node; }
	__forceinline  ProfileStackEntry*  ActiveStackEntry() { return &profileStack.back(); }

	__forceinline  static double CalcIterRate(double time) { return (time > 0) ? (1 / time) : 0; }


private:
	std::vector<Observer*>  observers;

	bool  bEnabled;
	bool  bRecursionHandlerEnabled;
	uint  autoResetThreshold;
	uint  maxExpectedStackDepth;
	bool  bRootEntered;
	bool  bRecursionWarningIgnoreFlag;

	bool    bAutoReset;
	bool    bUnsafe;
	double  elapsedSeconds;

	ProfilerNode*                   pRootNode;
	std::vector<ProfileStackEntry>  profileStack;
	std::string                     blockName;
	HighResolutionTimer             timer;
};


//====================================================================
class VW_COMMON_API ProfileRootBlock
{
public:
	__forceinline ~ProfileRootBlock()
	{
		if(p.IsEnabled()) p.ExitRootNode();
	}

protected:
	__forceinline ProfileRootBlock(Profiler& _p)
		:
		p(_p)
	{
		if(p.IsEnabled()) p.EnterRootNode();
	}

private:
	Profiler& p;
};


//====================================================================
class VW_COMMON_API ProfileBlock
{
protected:
	__forceinline ~ProfileBlock()
	{
		if(p.IsEnabled()) p.ExitNode();
	}

public:
	__forceinline ProfileBlock(Profiler& _p, const char* pBlockName)
		:
		p(_p)
	{
		if(p.IsEnabled()) p.EnterNode(pBlockName);
	}

private:
	Profiler& p;
};


//********************************************************************
}//namespace vertexware


