

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Constructs.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Keeps and calculates scrolling parameters.
// Total-size represents the size of the contents.
// Visible-size represents the size of the current viewport.
// Positions start at 0.
//====================================================================
class VW_COMMON_API ScrollGadget
{
public:
	ScrollGadget();
	virtual ~ScrollGadget();

	void  ResetLogicalParameters();
	void  ResetDrawingParameters();

	// call before extracting calculated drawing-parameters for thumb
	void  UpdateThumbParameters();

	//--------------------
	// Logical parameters.
	//--------------------
	uint  GetTotalSize() const { return total; }
	void  SetTotalSize(uint total = 1);

	uint  GetVisibleSize() const { return visible; }
	void  SetVisibleSize(uint visible = 1);

	bool  IsVisibleLessThanTotal() const { return (visible < total); }

	uint  GetPosition() const { return pos; }
	uint  GetMaxPosition() const;
	uint  GetLastVisiblePosition() const;
	bool  IsPositionAtMaximum() const { return (uint(pos) == GetMaxPosition()); }

	void  SetPosition(int pos = 0);
	void  SetPositionToMax() { pos = GetMaxPosition(); }
	void  OffsetPosition(int offset) { SetPosition(pos + offset); }

	//--------------------
	// Drawing parameters.
	//--------------------
	// these are important; used in thumb calculations
	uint  GetBarLength() const { return barLength; }
	void  SetBarLength(uint len = 1) { barLength = ClampedLow<uint>(len, 1); }

	uint  GetMinThumbLength() const { return minThumbLength; }
	void  SetMinThumbLength(uint minThumb = 1) { minThumbLength = ClampedLow<uint>(minThumb, 1); }

	uint  GetMaxPixelsPerStep() const { return maxPixelsPerStep; }
	void  SetMaxPixelsPerStep(uint maxPixels = 1) { maxPixelsPerStep = ClampedLow<uint>(maxPixels, 1); }

	bool  IsEnabledMaxPixelsPerStep() const { return bEnableMaxPixelsPerStep; }
	void  EnableMaxPixelsPerStep(bool bEnable) { bEnableMaxPixelsPerStep = bEnable; }

	// these are calculated by UpdateThumbParameters(); main purpose of this class
	uint  GetThumbLength() const { return thumbLength; }
	uint  GetThumbOffset() const { return thumbOffset; }

	// these are not important; can by used for storing additional information
	int   GetBarPosX() const { return barPosX; }
	void  SetBarPosX(int x) { barPosX = x; }

	int   GetBarPosY() const { return barPosY; }
	void  SetBarPosY(int y) { barPosY = y; }

	uint  GetBarWidth() const { return barWidth; }
	void  SetBarWidth(uint wid = 1) { barWidth = ClampedLow<uint>(wid, 1); }


private:
	// logical
	uint  total;
	uint  visible;
	int   pos;

	// drawing
	uint  barLength;
	uint  minThumbLength;
	uint  maxPixelsPerStep;
	bool  bEnableMaxPixelsPerStep;

	uint  thumbOffset;
	uint  thumbLength;

	int   barPosX;
	int   barPosY;
	uint  barWidth;
};


//********************************************************************
}//namespace vertexware


