

//--------------------------------------------------------------------
#include "BaseTools\Serializers.h"
#include "BaseTools\Memory.h"
#include "BaseTools\Constructs.h"


//--------------------------------------------------------------------
#define  DESERIALIZER_STRING_BUFFER_LENGTH  128


//********************************************************************
namespace vertexware {


//********************************************************************
namespace serializer_parameters {


//====================================================================
__forceinline bool __fastcall IsMachineBigEndian()
{
	// assume IA32 machine
	return false;
}


//====================================================================
// Set defaults to match machine.
//====================================================================
bool  bOutputStreamByteOrderBigEndian = IsMachineBigEndian();
bool  bInputStreamByteOrderBigEndian  = IsMachineBigEndian();
bool  bFlipByteOrderOnWrite = false;
bool  bFlipByteOrderOnRead  = false;


//====================================================================
void __fastcall SetOutputStreamByteOrder(bool bBigEndian)
{
	bOutputStreamByteOrderBigEndian = bBigEndian;
	bFlipByteOrderOnWrite = (bOutputStreamByteOrderBigEndian != IsMachineBigEndian());
}


//====================================================================
bool __fastcall IsOutputStreamByteOrderBigEndian()
{
	return bOutputStreamByteOrderBigEndian;
}


//====================================================================
void __fastcall SetInputStreamByteOrder(bool bBigEndian)
{
	bInputStreamByteOrderBigEndian = bBigEndian;
	bFlipByteOrderOnRead = (bInputStreamByteOrderBigEndian != IsMachineBigEndian());
}


//====================================================================
bool __fastcall IsInputStreamByteOrderBigEndian()
{
	return bInputStreamByteOrderBigEndian;
}


//********************************************************************
}//namespace serializer_parameters


//====================================================================
std::ostream& __fastcall WriteByteBlocks(std::ostream &os, const byte* blocks,
									uint byteCountPerBlock, uint blockCount)
{
	if(serializer_parameters::bFlipByteOrderOnWrite) {
		Buffer<byte>  buffer(byteCountPerBlock);
		for(uint i = 0; i < blockCount; i++) {
			ReverseCopy(buffer.ToPtr(), PostIncrement(blocks, byteCountPerBlock),
															byteCountPerBlock);
			os.write(buffer, byteCountPerBlock);
		}
	}
	else {
		os.write(blocks, byteCountPerBlock * blockCount);
	}
	return os;
}


//====================================================================
std::istream& __fastcall ReadByteBlocks(std::istream &is, byte* blocks,
									uint byteCountPerBlock, uint blockCount)
{
	if(serializer_parameters::bFlipByteOrderOnRead) {
		Buffer<byte>  buffer(byteCountPerBlock);
		for(uint i = 0; i < blockCount; i++) {
			is.read(buffer, byteCountPerBlock);
			ReverseCopy(PostIncrement(blocks, byteCountPerBlock), buffer.ToPtr(),
															byteCountPerBlock);
		}
	}
	else {
		is.read(blocks, byteCountPerBlock * blockCount);
	}
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const ubyte& x)
{
	return os.write(reinterpret_cast<const char*>(&x), 1);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, ubyte& x)
{
	return is.read(reinterpret_cast<char*>(&x), 1);
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const byte& x)
{
	return os.write(reinterpret_cast<const char*>(&x), 1);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, byte& x)
{
	return is.read(reinterpret_cast<char*>(&x), 1);
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const ushort& x)
{
	ushort a = x;
	if(serializer_parameters::bFlipByteOrderOnWrite) FlipByteOrder(a);
	return os.write(reinterpret_cast<const char*>(&a), 2);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, ushort& x)
{
	is.read(reinterpret_cast<char*>(&x), 2);
	if(serializer_parameters::bFlipByteOrderOnRead) FlipByteOrder(x);
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const short& x)
{
	short a = x;
	if(serializer_parameters::bFlipByteOrderOnWrite) FlipByteOrder(a);
	return os.write(reinterpret_cast<const char*>(&a), 2);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, short& x)
{
	is.read(reinterpret_cast<char*>(&x), 2);
	if(serializer_parameters::bFlipByteOrderOnRead) FlipByteOrder(x);
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const uint& x)
{
	uint a = x;
	if(serializer_parameters::bFlipByteOrderOnWrite) FlipByteOrder(a);
	return os.write(reinterpret_cast<const char*>(&a), 4);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, uint& x)
{
	is.read(reinterpret_cast<char*>(&x), 4);
	if(serializer_parameters::bFlipByteOrderOnRead) FlipByteOrder(x);
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const int& x)
{
	int a = x;
	if(serializer_parameters::bFlipByteOrderOnWrite) FlipByteOrder(a);
	return os.write(reinterpret_cast<const char*>(&a), 4);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, int& x)
{
	return is.read(reinterpret_cast<char*>(&x), 4);
	if(serializer_parameters::bFlipByteOrderOnRead) FlipByteOrder(x);
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const ulong& x)
{
	ulong a = x;
	if(serializer_parameters::bFlipByteOrderOnWrite) FlipByteOrder(a);
	return os.write(reinterpret_cast<const char*>(&a), 4);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, ulong& x)
{
	is.read(reinterpret_cast<char*>(&x), 4);
	if(serializer_parameters::bFlipByteOrderOnRead) FlipByteOrder(x);
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const long& x)
{
	long a = x;
	if(serializer_parameters::bFlipByteOrderOnWrite) FlipByteOrder(a);
	return os.write(reinterpret_cast<const char*>(&a), 4);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, long& x)
{
	is.read(reinterpret_cast<char*>(&x), 4);
	if(serializer_parameters::bFlipByteOrderOnRead) FlipByteOrder(x);
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const float& x)
{
	float a = x;
	if(serializer_parameters::bFlipByteOrderOnWrite) FlipByteOrder(a);
	return os.write(reinterpret_cast<const char*>(&a), 4);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, float& x)
{
	is.read(reinterpret_cast<char*>(&x), 4);
	if(serializer_parameters::bFlipByteOrderOnRead) FlipByteOrder(x);
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const double& x)
{
	double a = x;
	if(serializer_parameters::bFlipByteOrderOnWrite) FlipByteOrder(a);
	return os.write(reinterpret_cast<const char*>(&a), 8);
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, double& x)
{
	is.read(reinterpret_cast<char*>(&x), 8);
	if(serializer_parameters::bFlipByteOrderOnRead) FlipByteOrder(x);
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const bool& x)
{
	ubyte a = x ? 1 : 0;
	return os <= a;
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, bool& x)
{
	ubyte a;
	is >= a;
	x = a ? true : false;
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const std::string& x)
{
	os <= uint(x.size());
	return os.write(reinterpret_cast<const char*>(x.c_str()), x.size());
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, std::string& x)
{
	x.erase();

	uint  len;
	is >= len;

	const uint buflen = DESERIALIZER_STRING_BUFFER_LENGTH;
	char  buf[buflen];

	while(len > 0) {
		const uint chunk = (len > buflen) ? buflen : len;
		len -= chunk;
		is.read(buf, chunk);
		x.append(buf, chunk);
	}
	return is;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const UID& x)
{
	return os <= x[0] <= x[1] <= x[2] <= x[3];
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, UID& x)
{
	return is >= x[0] >= x[1] >= x[2] >= x[3];
}


//********************************************************************
}//namespace vertexware


