

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Debug.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Single-pass iterator.
// Wrapper around STL-container iterators.
//====================================================================
template<typename Container>
class SinglePassIterator
{
private:
	Container&  c;
	Container::iterator  it;

public:
	SinglePassIterator(Container& _c) : c(_c), it(c.begin())
	{ }

	//-------------------------------------
	// Usage example:
	//    ContainerType c;
	//    IteratorType  it(c);
	//    while(! it.End()) {
	//       ValueType& val = it.GetNext();
	//       use val;
	//    }
	//-------------------------------------
	bool End() { return (it == c.end()); }
	bool More() { return ! End(); }

	Container::value_type&  GetNext() { return *it++; }

	//---------------------------------------
	// Shorthand-notation for More() & End().
	//---------------------------------------
	operator bool() { return More(); }
	bool operator!() { return End(); }

	//-------------------------------------------
	// Usage example:
	//    ContainerType c;
	//    ValueType*    pVal;
	//    IteratorType  it(c);
	//    while(it.GetNext(&pVal)) { use *pVal; }
	//-------------------------------------------
	bool  GetNext(Container::value_type** pOutValue)
	{
		VW_ASSERT(pOutValue, "Bad argument");
		if(it == c.end()) return false;
		*pOutValue = &(*it);
		++it;
		return true;
	}
};


//====================================================================
template<typename Container>
class SinglePassConstIterator
{
private:
	const Container&  c;
	Container::const_iterator  it;

public:
	SinglePassConstIterator(const Container& _c) : c(_c), it(c.begin())
	{ }

	bool End() { return (it == c.end()); }
	bool More() { return ! End(); }
	const Container::value_type&  GetNext() { return *it++; }

	operator bool() { return More(); }
	bool operator!() { return End(); }

	bool  GetNext(const Container::value_type** pOutValue)
	{
		ASSERT(pOutValue, "Bad argument");
		if(it == c.end()) return false;
		*pOutValue = &(*it);
		++it;
		return true;
	}
};


//********************************************************************
}//namespace vertexware


//====================================================================
// Nested iterator macros.
//====================================================================
#define  VW_DECLARE_NESTED_SINGLEPASSITERATOR(IteratorName, ContainerClass, ContainerObject) \
	typedef  SinglePassIterator<ContainerClass>  IteratorName; \
	IteratorName  Get##IteratorName() { return IteratorName(ContainerObject); }


//====================================================================
#define  VW_DECLARE_NESTED_SINGLEPASSCONSTITERATOR(IteratorName, ContainerClass, ContainerObject) \
	typedef  SinglePassConstIterator<ContainerClass>  IteratorName; \
	IteratorName  Get##IteratorName() const { return IteratorName(ContainerObject); }

