

//--------------------------------------------------------------------
#include "BaseTools\Stream.h"
#include "BaseTools\PlatformInfo.h"
#include "BaseTools\Debug.h"
#include <algorithm>


//********************************************************************
namespace vertexware {


//====================================================================
const  std::ios_base::io_state   StreamExceptionSentry::Eof  = std::ios_base::eofbit;
const  std::ios_base::io_state   StreamExceptionSentry::Fail = std::ios_base::failbit;
const  std::ios_base::io_state   StreamExceptionSentry::Bad  = std::ios_base::badbit;
const  std::ios_base::io_state   StreamExceptionSentry::All  =
			StreamExceptionSentry::Eof | StreamExceptionSentry::Fail | StreamExceptionSentry::Bad;


//====================================================================
SetOutputTypeEffector  e_info(OutputStream::OutputType_Info);
SetOutputTypeEffector  e_warning(OutputStream::OutputType_Warning);
SetOutputTypeEffector  e_error(OutputStream::OutputType_Error);
SetOutputTypeEffector  e_query(OutputStream::OutputType_Query);
SetOutputTypeEffector  e_attention(OutputStream::OutputType_Attention);
SetOutputTypeEffector  e_debug(OutputStream::OutputType_Debug);
SetOutputTypeEffector  e_echo(OutputStream::OutputType_Echo);


//====================================================================
void  ControlledOutputStream::SetCurrentOutputType(uint outputType)
{
	uint numTargets = buffer.targets.size();
	if(! numTargets) return;

	flush(); //flush accumulated data before changing output-type

	for(uint i = 0; i < numTargets; i++) {
		OutputStream* p = dynamic_cast<OutputStream*>(buffer.targets[i]);
		if(p) p->SetCurrentOutputType(outputType);
	}
}


//====================================================================
void  ControlledOutputStream::AddOutputTarget(std::ostream& os)
{
	flush();
	OSTREAM_PTR_VEC_ITER it = std::find(buffer.targets.begin(),
									buffer.targets.end(), &os);
	if(it != buffer.targets.end()) return; //already added
	buffer.targets.push_back(&os);
}


//====================================================================
void  ControlledOutputStream::RemoveOutputTarget(std::ostream& os)
{
	flush();
	OSTREAM_PTR_VEC_ITER it = std::find(buffer.targets.begin(),
									buffer.targets.end(), &os);
	if(it == buffer.targets.end()) return; //not found
	buffer.targets.erase(it);
}


//====================================================================
void  ControlledOutputStream::RemoveAllOutputTargets()
{
	flush();
	buffer.targets.clear();
}


//====================================================================
int  ControlledOutputStream::StreamBuffer::sync()
{
	if(targets.size()) {
		const std::string& s = str();

		for(uint i = 0; i < targets.size(); i++) {
			std::ostream* p = targets[i];

			p->write(s.c_str(), s.length());
			p->flush();
		}
	}
	str(std::string());
	return 0; //assume success for all targets
}


//====================================================================
LogFileStream::LogFileStream(const std::string& _logName, const std::string& _fileName)
	:
	std::basic_ostream<char>(&filebuf),
	logName(_logName),
	fileName(_fileName)
{
	// allow empty filename
	if(fileName.empty()) {
		setstate(std::ios::failbit);
	}
	else {
		if(filebuf.open(fileName.c_str(), std::ios::out | std::ios::trunc) == 0) {
			setstate(std::ios::failbit);
		}
	}
	if(filebuf.is_open()) {
		OnOpen();
	}
}


//====================================================================
LogFileStream::~LogFileStream()
{
	Close();
}


//====================================================================
void  LogFileStream::Close()
{
	if(filebuf.is_open()) {
		OnClose();
		if(filebuf.close() == 0) {
			setstate(std::ios::failbit);
		}
	}
}


//====================================================================
void  LogFileStream::OutputBlockPE(bool bProlog, const time::DateTimeSeparators& dts)
{
	const char* sep = "/";

	std::ostringstream  oss;
	oss << logName;

	if(bProlog) oss << " log opened @ ";
	else		oss << " log closed @ ";

	std::string sTime;
	time::GetFormattedSystemDateTime(sTime, false, true, false, dts);

	oss << sTime;
	oss << sep << platform_info::GetMachineName();
	oss << sep << platform_info::GetUserName();

	std::string  block = oss.str();

	const char fillChar = '-';

	*this << separator_line(block.length(), fillChar);
	*this << block << nl;
	*this << separator_line(block.length(), fillChar);
}


//********************************************************************
}//namespace vertexware


