

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "private\vw_common_exports.h"
#include <string>


//********************************************************************
namespace vertexware {


//====================================================================
// Converts name to form that is acceptable for windows-kernel.
//====================================================================
VW_COMMON_API
void __fastcall ForceWinKernelName(std::string& s);


//====================================================================
// Critical-section.
//====================================================================
class VW_COMMON_API CriticalSection
{
public:
	CriticalSection() {
		::InitializeCriticalSection(&cs);
	}
	~CriticalSection() {
		::DeleteCriticalSection(&cs);
	}
	void  Enter() {
		::EnterCriticalSection(&cs);
	}
	bool  TryEnter() {
		return (::TryEnterCriticalSection(&cs) != 0);
	}
	void  Leave() {
		::LeaveCriticalSection(&cs);
	}


private:
	CRITICAL_SECTION  cs;
};


//====================================================================
// Critical-section sentry.
// Provides automatic enter/leave for current scope.
// Example usage:
//     CriticalSection cs;
//     {
//        CriticalSectionSentry css(cs);
//        // do guarded stuff
//     }
//====================================================================
class CriticalSectionSentry
{
private:
	CriticalSection&  cs;

public:
	explicit CriticalSectionSentry(CriticalSection& _cs) : cs(_cs) { cs.Enter(); }
	~CriticalSectionSentry() { cs.Leave(); }
};


//====================================================================
// Event.
//====================================================================
class VW_COMMON_API Event
{
public:
	Event(bool initiallySet = false, bool manualReset = false,
			const std::string& name = "", bool bForceWinKernelName = true);
	~Event();

	HANDLE              GetHandle() const { return handle; }
	const std::string&  GetName() const { return name; }

	bool  Set()   { return (::SetEvent(handle)   ? true : false); }
	bool  Reset() { return (::ResetEvent(handle) ? true : false); }

	//------------------------
	// Return true on success.
	//------------------------
	bool  WaitInfinite();

	//--------------------------------------------
	// On error, returns false with <bTimeout> set
	// to false. On success <bTimeout> is set to
	// false if event was set before timeout
	// of <msces> elapsed.
	//--------------------------------------------
	bool  Wait(uint msecs, bool& bTimeout);


protected:
	HANDLE       handle;
	std::string  name;
};


//====================================================================
// Single instance flag.
//====================================================================
class VW_COMMON_API SingleInstanceFlagSentry
{
public:
	SingleInstanceFlagSentry(const std::string& name, bool bForceWinKernelName = true);
	~SingleInstanceFlagSentry();

	bool  IsAcquired() { return (handle != 0); }

	const std::string&  GetName() const { return name; }


private:
	HANDLE       handle;
	std::string  name;
};


//********************************************************************
}//namespace vertexware


