

//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Debug.h"
#include "BaseTools\Thread.h"
#include "BaseTools\String.h"


//********************************************************************
namespace vertexware {


//====================================================================
Thread::Thread()
{
	threadHandle = 0;
	threadID = 0;

	priority = THREAD_PRIORITY_NORMAL;
	bEnablePriorityBoost = true;

	exitCode = 0;
	bContinue = false;
}


//====================================================================
Thread::~Thread()
{
	VW_VERIFY(! threadHandle,
			"Destroying thread-object while thread is still executing:"
			" ThreadID = " + ToString(threadID));
}


//====================================================================
void  Thread::SetPriority(int _priority)
{
	priority = _priority;
	if(IsRunning()) {
		::SetThreadPriority(threadHandle, priority);
		priority = ::GetThreadPriority(threadHandle);
	}
}


//====================================================================
void  Thread::EnablePriorityBoost(bool bEnable)
{
	bEnablePriorityBoost = bEnable;
	if(IsRunning()) {
		BOOL bBoostDisabled = bEnable ? FALSE : TRUE;
		::SetThreadPriorityBoost(threadHandle, bBoostDisabled);
		::GetThreadPriorityBoost(threadHandle, &bBoostDisabled);
		bEnablePriorityBoost = bBoostDisabled ? false : true;
	}
}


//====================================================================
void  Thread::Run()
{
	VW_VERIFY(! threadHandle, "Thread is already running.");

	// reset exit events & flags
	if(! exitEvent.Reset()) {
		VW_THROW(SystemResourceException, "Failed to reset exit-signal.");
	}
	bContinue = true;

	// create new thread and run it immediately
	threadHandle = reinterpret_cast<HANDLE>(
						_beginthreadex(0, 0, WorkFunctionMapper, this, 0, &threadID));

	// calling thread continues here
	if(! threadHandle) {
		VW_THROW(SystemResourceException, "_beginthreadex() failed.");
	}

	// realize thread properties
	SetPriority(priority);
	EnablePriorityBoost(bEnablePriorityBoost);
}


//====================================================================
bool  Thread::IsRunning()
{
	if(threadHandle) return true;
	return false;
}


//====================================================================
bool  Thread::WaitReady()
{
	return readyEvent.WaitInfinite();
}


//====================================================================
void  Thread::SignalExit()
{
	if(threadHandle) {
		bContinue = false;
		exitEvent.Set();
	}
}


//====================================================================
void  Thread::WaitExit(bool bCloseThreadHandle)
{
	if(threadHandle) {
		::WaitForSingleObject(threadHandle, INFINITE);
		if(bCloseThreadHandle) {
			::CloseHandle(threadHandle);
			threadHandle = 0;
		}
	}
}


//====================================================================
uint  Thread::GetLastExitCode()
{
	return exitCode;
}


//====================================================================
bool  Thread::PostStandardMessage(UINT message, WPARAM wparam, LPARAM lparam)
{
	if(threadHandle) {
		if(::PostThreadMessage(threadID, message, wparam, lparam)) {
			return true;
		}
	}
	return false;
}


//====================================================================
bool  Thread::WaitMessageQueueReady(uint maxRetries, uint msecSleepInterval)
{
	bool bInfRetries = ! maxRetries;

	if(threadHandle) {
		while(! PostStandardNullMessage()) {
			if(bInfRetries  ||  maxRetries) {
				maxRetries--;
				::Sleep(msecSleepInterval);
			}
			else {
				return false;
			}
		}
		return true;
	}
	return false;
}


//====================================================================
bool  Thread::Suspend()
{
	VW_VERIFY(threadHandle, "Thread is not running.");

	DWORD retVal = ::SuspendThread(threadHandle);
	return (retVal != 0xFFFFFFFF);
}


//====================================================================
bool  Thread::Resume()
{
	VW_VERIFY(threadHandle, "Thread is not running.");

	DWORD retVal = ::ResumeThread(threadHandle);
	return (retVal != 0xFFFFFFFF);
}


//====================================================================
uint __stdcall Thread::WorkFunctionMapper(void* v)
{
	// created thread begins here
	Thread* pThread = reinterpret_cast<Thread*>(v);

	// execute work-function
	pThread->exitCode = pThread->DoWork();

	// cleanup after exit
	pThread->threadHandle = 0;
	pThread->threadID = 0;

	// created thread ends after next return.
	// OS calls _endthreadex(exitCode) internally
	return  pThread->exitCode;
}


//====================================================================
bool  Thread::IsExitSignaled()
{
	bool bTimeout = false;
	if(! exitEvent.Wait(0, bTimeout)) {
		// error occcured, force exit
		return true;
	}
	return (! bTimeout);
}


//====================================================================
void  Thread::WaitExitSignal()
{
	exitEvent.WaitInfinite();
}


//====================================================================
void  Thread::SignalReady()
{
	readyEvent.Set();
}


//********************************************************************
}//namespace vertexware

