

//--------------------------------------------------------------------
#include "BaseTools\Time.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\Debug.h"
#include <sstream>
#include <iomanip>


//********************************************************************
namespace {


//====================================================================
void __fastcall FormatDate(::SYSTEMTIME& st, std::ostream& os,
							bool shortYear, bool weekDay,
							const vertexware::time::DateTimeSeparators& dts)
{
	using namespace std;

	if(shortYear) {
		os << setw(2) << (st.wYear % 100) << dts.GetDateSep();
	}
	else {
		os << setw(4) << st.wYear << dts.GetDateSep();
	}
	os << setw(2) << st.wMonth << dts.GetDateSep();
	os << setw(2) << st.wDay;

	if(weekDay) {
		const char* p = 0;

		switch(st.wDayOfWeek) {
			case 0:  p = "sun"; break;
			case 1:  p = "mon"; break;
			case 2:  p = "tue"; break;
			case 3:  p = "wed"; break;
			case 4:  p = "thu"; break;
			case 5:  p = "fri"; break;
			case 6:  p = "sat"; break;
			default: p = "???"; break;
		}
		os << dts.GetDateSep() << p;
	}
}


//====================================================================
void __fastcall FormatTime(::SYSTEMTIME& st, std::ostream& os, bool msecs,
						   const vertexware::time::DateTimeSeparators& dts)
{
	using namespace std;

	os << setw(2) << st.wHour << dts.GetTimeSep();
	os << setw(2) << st.wMinute << dts.GetTimeSep();
	os << setw(2) << st.wSecond;

	if(msecs) {
		os << dts.GetTimeSep() << setw(3) << st.wMilliseconds;
	}
}


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//********************************************************************
namespace time {


//====================================================================
void __fastcall GetFormattedSystemDate(std::string& out,
							bool shortYear, bool weekDay,
							const DateTimeSeparators& dts)
{
	using namespace std;

	::SYSTEMTIME st;
	::GetLocalTime(&st);

	ostringstream  oss;
	oss.fill('0');
	FormatDate(st, oss, shortYear, weekDay, dts);
	out.assign(oss.str());
}


//====================================================================
void __fastcall GetFormattedSystemTime(std::string& out,
							bool msecs,
							const DateTimeSeparators& dts)
{
	using namespace std;

	::SYSTEMTIME st;
	::GetLocalTime(&st);

	ostringstream  oss;
	oss.fill('0');
	FormatTime(st, oss, msecs, dts);
	out.assign(oss.str());
}


//====================================================================
void __fastcall GetFormattedSystemDateTime(std::string& out,
								bool shortYear, bool weekDay, bool msecs,
								const DateTimeSeparators& dts)
{
	using namespace std;

	::SYSTEMTIME st;
	::GetLocalTime(&st);

	ostringstream  oss;
	oss.fill('0');
	FormatDate(st, oss, shortYear, weekDay, dts);
	oss << dts.GetDTSep();
	FormatTime(st, oss, msecs, dts);
	out.assign(oss.str());
}


//====================================================================
void __fastcall GetSystemDate(uint& year, uint& month, uint& day, uint& weekDay)
{
	::SYSTEMTIME st;
	::GetLocalTime(&st);

	year    = st.wYear;
	month   = st.wMonth;
	day     = st.wDay;
	weekDay = st.wDayOfWeek;
}


//====================================================================
void __fastcall GetSystemTime(uint& hour, uint& min, uint& sec, uint& msec)
{
	::SYSTEMTIME st;
	::GetLocalTime(&st);

	hour = st.wHour;
	min  = st.wMinute;
	sec  = st.wSecond;
	msec = st.wMilliseconds;
}


//====================================================================
void __fastcall GetSystemDateTime(uint& year, uint& month, uint& day, uint& weekDay,
								uint& hour, uint& min, uint& sec, uint& msec)
{
	::SYSTEMTIME st;
	::GetLocalTime(&st);

	year    = st.wYear;
	month   = st.wMonth;
	day     = st.wDay;
	weekDay = st.wDayOfWeek;
	hour    = st.wHour;
	min     = st.wMinute;
	sec     = st.wSecond;
	msec    = st.wMilliseconds;
}


//********************************************************************
}//namespace time


//====================================================================
HighResolutionTimer::HighResolutionTimer()
{
	double dTicksPerSec;
	if(! GetPerformanceCounterInfo(&liTicksPerSec, &dTicksPerSec)) {
		VW_THROW(SystemResourceException,
				"Hardware does not support high-resolution performance counter.");
	}

	if(dTicksPerSec == 0) dOneOverTicksPerSec = 0;
	else                  dOneOverTicksPerSec = 1 / dTicksPerSec;

	Start(); //auto-start
}


//====================================================================
void  HighResolutionTimer::GetElapsedTicks(uint& highPart, uint& lowPart)
{
	LARGE_INTEGER  liStopTicks;
	::QueryPerformanceCounter(&liStopTicks);
	liStopTicks.QuadPart -= liStartTicks.QuadPart;
	highPart = liStopTicks.HighPart;
	lowPart  = liStopTicks.LowPart;
}


//====================================================================
bool __fastcall HighResolutionTimer::GetPerformanceCounterInfo(
							LARGE_INTEGER* pliTicksPerSec, double* pdTicksPerSec)
{
	// check performance counter support
	LARGE_INTEGER  liTicksPerSec;
	if(::QueryPerformanceFrequency(&liTicksPerSec)) {

		// convert type
		double dTicksPerSec = double(liTicksPerSec.QuadPart);

		// check for bad frequency
		if(liTicksPerSec.QuadPart > 0) {
			SafeAssign(pliTicksPerSec, liTicksPerSec);
			SafeAssign(pdTicksPerSec, dTicksPerSec);
			return true;
		}
	}
	// counter not available or bad frequency reported
	if(pliTicksPerSec) pliTicksPerSec->QuadPart = 0;
	SafeAssign(pdTicksPerSec, 0.0);
	return false;
}


//====================================================================
double  ThreadTimer::GetElapsedSeconds(double* pOutKernelTime, double* pOutUserTime)
{
	FILETIME        dummy;
	ULARGE_INTEGER  endKernel, endUser;

	::GetThreadTimes(GetCurrentThread(), &dummy, &dummy,
		reinterpret_cast<FILETIME*>(&endKernel), reinterpret_cast<FILETIME*>(&endUser));

	const double secsPerUnit = 1e-7;

	double  kernelTime = (endKernel.QuadPart - startKernel.QuadPart) * secsPerUnit;
	SafeAssign(pOutKernelTime, kernelTime);

	double  userTime = (endUser.QuadPart - startUser.QuadPart) * secsPerUnit;
	SafeAssign(pOutUserTime, userTime);

	return (kernelTime + userTime);
}


//====================================================================
AlarmTimer::AlarmTimer()
{
	pAlarmHandler = 0;
	hSystemTimer = 0;
}


//====================================================================
AlarmTimer::~AlarmTimer()
{
	StopPeriodic();
}


//====================================================================
bool  AlarmTimer::IsRunning()
{
	CriticalSectionSentry css(csHandle);
	return  ToBool(hSystemTimer);
}


//====================================================================
bool  AlarmTimer::FireOnce(AlarmHandler& ah, uint msecs)
{
	return  Activate(ah, msecs, 0);
}


//====================================================================
bool  AlarmTimer::StartPeriodic(AlarmHandler& ah, uint msecsFirst, uint msecsPeriod)
{
	return  Activate(ah, msecsFirst, ClampedLow<uint>(msecsPeriod, 1));
}


//====================================================================
void  AlarmTimer::StopPeriodic()
{
	// 1) If creator-thread enters into deactivation process first,
	//    alarm-thread will refrain from auto-deactivating and
	//    just exit the work-function, which provides the
	//    return-condition for creator-thread.
	//
	// 2) If alarm-thread enters into deactivation process first,
	//    creator-thread must ensure that alarm-thread exits from
	//    work-function before this method returns. Since this
	//    method is also called from alarm-destructor, this condition
	//    ensures that alarm-thread will not reference destroyed
	//    alarm-object. It does so by blocking until alarm-thread
	//    auto-deactivates. After that a call to Deactivate()
	//    discovers that alarm-object is deactivated and does
	//    nothing. Safe-state is now established, and this method
	//    returns.
	//
	csDeactivator.Enter();
	Deactivate(true);
	csDeactivator.Leave();
}


//====================================================================
bool  AlarmTimer::Activate(AlarmHandler& ah, uint msecsFirst, uint msecsPeriod)
{
	CriticalSectionSentry css(csHandle);
	if(hSystemTimer) return false; //busy

	ULONG flags = (ah.QueryHintLongExecution() ? WT_EXECUTELONGFUNCTION : 0);

	if(::CreateTimerQueueTimer(&hSystemTimer, 0, AlarmHandlerMapper, this,
								msecsFirst, msecsPeriod, flags))
	{
		pAlarmHandler = &ah;
		bOneShot = (msecsPeriod == 0);
		return true;
	}
	hSystemTimer = 0;
	pAlarmHandler = 0;
	return false;
}


//====================================================================
void  AlarmTimer::Deactivate(bool bWaitAlarmHandlerCompletion)
{
	CriticalSectionSentry css(csHandle);
	if(hSystemTimer) {
		::DeleteTimerQueueTimer(0, hSystemTimer,
						(bWaitAlarmHandlerCompletion ? INVALID_HANDLE_VALUE : 0));
		hSystemTimer = 0;
		pAlarmHandler = 0;
	}
}


//====================================================================
void __stdcall AlarmTimer::AlarmHandlerMapper(void* userData, BOOLEAN)
{
	AlarmTimer* pAlarmTimer = reinterpret_cast<AlarmTimer*>(userData);

	// execute work-function
	bool bStopRequested = pAlarmTimer->pAlarmHandler->OnAlarm(*pAlarmTimer);

	// auto-deactivate
	if(bStopRequested  ||  pAlarmTimer->bOneShot) {

		if(pAlarmTimer->csDeactivator.TryEnter()) {
			pAlarmTimer->Deactivate(false);
			pAlarmTimer->csDeactivator.Leave();

			// IMPORTANT: this function must not reference
			// <pAlarmTimer> variable after it gets here.
		}
	}
}


//********************************************************************
}//namespace vertexware

