

//--------------------------------------------------------------------
#include "BaseTools\TopLevelWindow.h"
#include "BaseTools\Debug.h"
#include "BaseTools\Exception.h"
#include "BaseTools\String.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\Memory.h"
#include "BaseTools\PlatformInfo.h"
#include "MathTools\MathBase.h"
#include <sstream>


//====================================================================
#define  ABSOLUTE_MINIMUM_WINDOW_SIZE   10
#define  GADGET_SIZE                    20


//********************************************************************
namespace vertexware {


//====================================================================
MessageHook::~MessageHook()
{
	if(TopLevelWindow::GetMessageHook() == this) {
		TopLevelWindow::SetMessageHook(0);
	}
}


//====================================================================
MessageHook*  TopLevelWindow::s_messageHook = 0;


//====================================================================
void  __fastcall TopLevelWindow::SetMessageHook(MessageHook* hook)
{
	s_messageHook = hook;
}


//====================================================================
MessageHook* __fastcall TopLevelWindow::GetMessageHook()
{
	return s_messageHook;
}


//====================================================================
TopLevelWindow::TopLevelWindow(bool bUsingGdiplus)
{
	gpg = 0;
	SetUseGdiplus(bUsingGdiplus);

	bPopup = false;
	hwnd = 0;
	hdc = 0;
	windowText = VW_GET_CLASS_NAME(*this);
	bActive = false;
	bAlwaysOnTop = false;
	bUsingCompositingEffects = false;
	SWPMoveFlags = 0;

	bDisableMessageDispatchMouseMove = false;
	bDisableMessageDispatchMouseWheel = false;
	bDisableMessageDispatchMouseButtons = false;
	bDisableMessageDispatchKeyboardKeyStates = false;
	bDisableMessageDispatchKeyboardCharEvents = false;

	pMouseMessageHook = 0;
	pKeyboardMessageHook = 0;

	positionX = 0;
	positionY = 0;
	maximizedPositionX = 0;
	maximizedPositionY = 0;
	width = 1;
	height = 1;
	clientAreaWidth = 1;
	clientAreaHeight = 1;
	minWidth  = 1;
	maxWidth  = 1;
	minHeight = 1;
	maxHeight = 1;

	bMouseCursorVisible = true;
	hActiveMouseCursor = 0;

	bEnableCloseThroughAltF4 = true;
	bEnableMenuThroughF10 = true;
	bEnableMenuThroughAltSpace = true;

	bSystemCommandEmulation = false;

	bMoving = false;
	bSizing = false;
	bSizingBottom = false;
	bSizingRight = false;
	bMouseLeaveTrackingInactive = true;

	currentMouseArea = WindowArea_Outside;
	currentMouseAreaSubCode = 0;

	SetForceRepaintOnMove();
	SetMaximizedPosition();
	SetMinSize();
	SetMaxSize();
}


//====================================================================
TopLevelWindow::~TopLevelWindow()
{
	SetUseGdiplus(false);

	if(hwnd) {
		DestroySystemWindow();
	}
}


//====================================================================
void  TopLevelWindow::SetIcon(HICON hIcon)
{
	::SetClassLongPtr(hwnd, GCLP_HICON, reinterpret_cast<LONG_PTR>(hIcon));
}


//====================================================================
HICON  TopLevelWindow::GetIcon() const
{
	return reinterpret_cast<HICON>(::GetClassLongPtr(hwnd, GCLP_HICON));
}


//====================================================================
void  TopLevelWindow::SetText(const std::string& s)
{
	windowText = s;
	::SetWindowText(hwnd, windowText.c_str());
}


//====================================================================
bool  TopLevelWindow::Activate()
{
	return (::SetForegroundWindow(hwnd) != 0);
}


//====================================================================
void  TopLevelWindow::SetAlwaysOnTop(bool bFlag)
{
	bAlwaysOnTop = bFlag;
	if(bAlwaysOnTop) {
		::SetWindowPos(GetHWND(), HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
	else {
		::SetWindowPos(GetHWND(), HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
}


//====================================================================
void  TopLevelWindow::EnableMessageDispatchMouseMove(bool bEnable)
{
	bDisableMessageDispatchMouseMove = ! bEnable;
}


//====================================================================
void  TopLevelWindow::EnableMessageDispatchMouseWheel(bool bEnable)
{
	bDisableMessageDispatchMouseWheel = ! bEnable;
}


//====================================================================
void  TopLevelWindow::EnableMessageDispatchMouseButtons(bool bEnable)
{
	bDisableMessageDispatchMouseButtons = ! bEnable;
}


//====================================================================
void  TopLevelWindow::EnableMessageDispatchKeyboardKeyStates(bool bEnable)
{
	bDisableMessageDispatchKeyboardKeyStates = ! bEnable;
}


//====================================================================
void  TopLevelWindow::EnableMessageDispatchKeyboardCharEvents(bool bEnable)
{
	bDisableMessageDispatchKeyboardCharEvents = ! bEnable;
}


//====================================================================
void  TopLevelWindow::SetMouseMessageHook(MouseMessageHook* hook)
{
	pMouseMessageHook = hook;
}


//====================================================================
void  TopLevelWindow::SetKeyboardMessageHook(KeyboardMessageHook* hook)
{
	pKeyboardMessageHook = hook;
}


//====================================================================
bool  TopLevelWindow::EnableCompositingEffects(bool bEnable)
{
	if(bEnable) {
		bUsingCompositingEffects = AddWindowStyles(0, WS_EX_LAYERED);
		return bUsingCompositingEffects;
	}
	// disable
	bUsingCompositingEffects = ! RemoveWindowStyles(0, WS_EX_LAYERED);
	return (! bUsingCompositingEffects);
}


//====================================================================
void  TopLevelWindow::SetOpacity(ubyte opacity)
{
	if(bUsingCompositingEffects) {
		::SetLayeredWindowAttributes(hwnd, 0, opacity, LWA_ALPHA);
	}
}


//====================================================================
void  TopLevelWindow::SetColorKey(color4b colorKey)
{
	if(bUsingCompositingEffects) {
		::SetLayeredWindowAttributes(hwnd, RGB(colorKey.r, colorKey.g, colorKey.b),
											0, LWA_COLORKEY);
	}
}


//====================================================================
void  TopLevelWindow::SetOpacityAndColorKey(ubyte opacity, color4b colorKey)
{
	if(bUsingCompositingEffects) {
		if(opacity == 255) {
			SetColorKey(colorKey);
		}
		else {
			::SetLayeredWindowAttributes(hwnd, RGB(colorKey.r, colorKey.g, colorKey.b),
									opacity, LWA_COLORKEY | LWA_ALPHA);
		}
	}
}


//====================================================================
void  TopLevelWindow::SetForceRepaintOnMove(bool bEnable)
{
	if(bEnable) {
		SWPMoveFlags |= SWP_NOCOPYBITS;
	}
	else {
		SWPMoveFlags &= ~SWP_NOCOPYBITS;
	}
}


//====================================================================
void  TopLevelWindow::SetPosition(int x, int y)
{
	::SetWindowPos(hwnd, NULL, x, y, 0, 0, SWP_NOSIZE | SWP_NOREPOSITION | SWPMoveFlags);
}


//====================================================================
void  TopLevelWindow::UpdatePosition(int deltaX, int deltaY)
{
	if(CanMoveWindow()) {
		int x, y;
		GetCurrentPosition(x, y);
		SetPosition(x + deltaX, y + deltaY);
	}
}


//====================================================================
void  TopLevelWindow::SetMaximizedPosition(int x, int y)
{
	if(hwnd  &&  IsMaximized()) {
		SetPosition(maximizedPositionX, maximizedPositionY);
	}
	else {
		maximizedPositionX = x;
		maximizedPositionY = y;
	}
}


//====================================================================
void  TopLevelWindow::GetCurrentPosition(int& x, int& y) const
{
	if(IsMaximized()) {
		x = maximizedPositionX;
		y = maximizedPositionY;
	}
	else {
		x = positionX;
		y = positionY;
	}
}


//====================================================================
void  TopLevelWindow::SetSize(uint w, uint h)
{
	MinMaxClamp(w, h);
	::SetWindowPos(hwnd, NULL, 0, 0, w, h, SWP_NOMOVE | SWP_NOREPOSITION);
}


//====================================================================
void  TopLevelWindow::UpdateSize(int deltaW, int deltaH)
{
	if(CanSizeWindow()) {
		SetSize(ClampedLow((int(width)  + deltaW), 0),
				ClampedLow((int(height) + deltaH), 0));
	}
}


//====================================================================
void  TopLevelWindow::SetMinSize(uint w, uint h)
{
	minWidth  = ClampedLow<uint>(w, ABSOLUTE_MINIMUM_WINDOW_SIZE);
	minHeight = ClampedLow<uint>(h, ABSOLUTE_MINIMUM_WINDOW_SIZE);

	if(width < minWidth  ||  height < minHeight) {
		SetSize(ClampedLow<uint>(width, minWidth), ClampedLow<uint>(height, minHeight));
	}
}


//====================================================================
void  TopLevelWindow::GetMinSize(uint& w, uint& h) const
{
	w = minWidth;
	h = minHeight;
}


//====================================================================
void  TopLevelWindow::SetMaxSize(uint w, uint h)
{
	maxWidth  = w ? w : UINT_MAX;
	maxHeight = h ? h : UINT_MAX;

	if(width > maxWidth  ||  height > maxHeight) {
		SetSize(ClampedHigh<uint>(width, maxWidth), ClampedHigh<uint>(height, maxHeight));
	}
}


//====================================================================
void  TopLevelWindow::GetMaxSize(uint& w, uint& h) const
{
	w = maxWidth;
	h = maxHeight;
}


//====================================================================
void  TopLevelWindow::GetSize(int& w, int& h) const
{
	w = width;
	h = height;
}


//====================================================================
void  TopLevelWindow::GetClientAreaSize(int& w, int& h) const
{
	w = clientAreaWidth;
	h = clientAreaHeight;
}


//====================================================================
void  TopLevelWindow::SetPositionAndSize(int x, int y, uint w, uint h)
{
	MinMaxClamp(w, h);
	::SetWindowPos(hwnd, NULL, x, y, w, h, SWP_NOREPOSITION);
}


//====================================================================
void  TopLevelWindow::CenterOnDesktop(uint w, uint h)
{
	uint desktopW, desktopH;
	platform_info::GetDesktopDisplayMode(&desktopW, &desktopH, 0, 0);

	Clamp<uint>(w, minWidth,  Min(maxWidth,  desktopW));
	Clamp<uint>(h, minHeight, Min(maxHeight, desktopH));

	SetPositionAndSize((desktopW - w) / 2, (desktopH - h) / 2, w, h);
}


//====================================================================
void  TopLevelWindow::FillDesktop()
{
	uint desktopW, desktopH;
	platform_info::GetDesktopDisplayMode(&desktopW, &desktopH, 0, 0);

	SetPositionAndSize(0, 0, desktopW, desktopH);
}


//====================================================================
void  TopLevelWindow::GetVisibleClientAreaScreenCoords(RECT& rect) const
{
	uint workAreaW, workAreaH, workAreaL, workAreaT;
	platform_info::GetWorkArea(&workAreaW, &workAreaH, &workAreaL, &workAreaT);

	rect.left = 0;
	rect.top = 0;
	::ClientToScreen(GetHWND(), reinterpret_cast<POINT*>(&rect.left));

	rect.right = GetClientAreaWidth();
	rect.bottom = GetClientAreaHeight();
	::ClientToScreen(GetHWND(), reinterpret_cast<POINT*>(&rect.right));

	ClampLow<LONG>(rect.left, workAreaL);
	ClampHigh<LONG>(rect.right, workAreaL + workAreaW);
	ClampLow<LONG>(rect.top, workAreaT);
	ClampHigh<LONG>(rect.bottom, workAreaT + workAreaH);
}


//====================================================================
void  TopLevelWindow::ShowMouseCursor(bool bShow)
{
	if(bShow) {
		if(! bMouseCursorVisible) ::ShowCursor(TRUE);
	}
	else {
		if(bMouseCursorVisible) ::ShowCursor(FALSE);
	}
	bMouseCursorVisible = bShow;
}


//====================================================================
void  TopLevelWindow::SetMouseCursorShape(eWindowArea area, HCURSOR hcursor)
{
	VW_ASSERT(area < WindowArea_Count, "Invalid enumeration value.");
	if(hcursor) {
		hMouseCursors[area] = hcursor;
	}
	else {
		hMouseCursors[area] = win_utils::LoadStockCursor(win_utils::StockCursor_Normal);
	}
	// re-cache
	hActiveMouseCursor = hMouseCursors[currentMouseArea];

	// register with system
	if(bSystemCommandEmulation) {
		if(area == currentMouseArea) {
			::SetCursor(hActiveMouseCursor);
		}
	}
	else {
		if(area == WindowArea_Client) {
			SetWindowClassMouseCursor(hMouseCursors[WindowArea_Client]);
		}
	}
}


//====================================================================
void  TopLevelWindow::ResetMouseCursorShapes()
{
	hMouseCursors[WindowArea_Outside] =
							win_utils::LoadStockCursor(win_utils::StockCursor_Normal);

	hMouseCursors[WindowArea_Move] =
							win_utils::LoadStockCursor(win_utils::StockCursor_SizeAll);

	hMouseCursors[WindowArea_SizeBottom] =
							win_utils::LoadStockCursor(win_utils::StockCursor_SizeNS);

	hMouseCursors[WindowArea_SizeRight] =
							win_utils::LoadStockCursor(win_utils::StockCursor_SizeWE);

	hMouseCursors[WindowArea_SizeBottomRight] =
							win_utils::LoadStockCursor(win_utils::StockCursor_SizeNWSE);

	hMouseCursors[WindowArea_Minimize] =
							win_utils::LoadStockCursor(win_utils::StockCursor_Hand);

	hMouseCursors[WindowArea_MaximizeRestore] =
							win_utils::LoadStockCursor(win_utils::StockCursor_Hand);

	hMouseCursors[WindowArea_AlwaysOnTop] =
							win_utils::LoadStockCursor(win_utils::StockCursor_Hand);

	hMouseCursors[WindowArea_Close] =
							win_utils::LoadStockCursor(win_utils::StockCursor_Hand);

	hMouseCursors[WindowArea_Command] =
							win_utils::LoadStockCursor(win_utils::StockCursor_Hand);

	hMouseCursors[WindowArea_Client] =
							win_utils::LoadStockCursor(win_utils::StockCursor_Normal);
}


//====================================================================
void  TopLevelWindow::SynthesizeMouseMove()
{
	win_utils::Point p;
	GetMousePositionClient(p);
	LPARAM lparam = ((p.y << 16) & 0xFFFF0000) | (p.x & 0x0000FFFF);
	PostMessage(hwnd, WM_MOUSEMOVE, 0, lparam);
}


//====================================================================
void  TopLevelWindow::Show()
{
	::ShowWindow(hwnd, SW_SHOW);
	::UpdateWindow(hwnd);
}


//====================================================================
void  TopLevelWindow::ShowMinimized(bool bActivate)
{
	::ShowWindow(hwnd, bActivate ? SW_SHOWMINIMIZED : SW_SHOWMINNOACTIVE);
}


//====================================================================
void  TopLevelWindow::ShowMaximized()
{
	::ShowWindow(hwnd, SW_SHOWMAXIMIZED);
	::UpdateWindow(hwnd);
}


//====================================================================
void  TopLevelWindow::Hide()
{
	::ShowWindow(hwnd, SW_HIDE);
}


//====================================================================
void  TopLevelWindow::Restore()
{
	::ShowWindow(hwnd, SW_RESTORE);
}


//====================================================================
void  TopLevelWindow::Minimize()
{
	::ShowWindow(hwnd, SW_MINIMIZE);
}


//====================================================================
void  TopLevelWindow::Maximize()
{
	::ShowWindow(hwnd, SW_MAXIMIZE);
}


//====================================================================
void  TopLevelWindow::CreateSystemWindow(UINT classStyle, DWORD windowStyle,
											DWORD windowStyleEx)
{
	VW_ASSERT(! hwnd, "System-window already created.");

	HINSTANCE    hInst = ::GetModuleHandle(NULL);
	std::string  sClassName = VW_GET_CLASS_NAME(*this);

	// register window class
	WNDCLASSEX wcex;
	::ZeroMemory(&wcex, sizeof(wcex));
	wcex.cbSize = sizeof(WNDCLASSEX); 
	wcex.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC | classStyle;
	wcex.lpfnWndProc = MessageHandlerMapper;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra	= 0;
	wcex.hInstance = hInst;
	wcex.hIcon = 0;
	wcex.hCursor = 0;
	wcex.hbrBackground = 0;
	wcex.lpszMenuName = 0;
	wcex.lpszClassName = sClassName.c_str();
	wcex.hIconSm = 0;

	if(! ::RegisterClassEx(&wcex)) {
		DWORD err = GetLastError();
		if(err != ERROR_CLASS_ALREADY_EXISTS) {
			VW_THROW(SystemResourceException, "RegisterClassEx() failed.");
		}
	}

	// extract states from style-flags
	bPopup = ToBool(windowStyle  &  WS_POPUP);
	bAlwaysOnTop = ToBool(windowStyleEx  &  WS_EX_TOPMOST);

	// fix style-flags
	windowStyle   |= WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
	windowStyleEx &= ~(WS_EX_LAYERED | WS_EX_TRANSPARENT);

	// create window
	hwnd = ::CreateWindowEx(
			windowStyleEx,
			wcex.lpszClassName, wcex.lpszClassName,
			windowStyle,
			CW_USEDEFAULT, CW_USEDEFAULT,
			CW_USEDEFAULT, CW_USEDEFAULT,
			0, 0, hInst, 0);
	if(! hwnd) {
		VW_THROW(SystemResourceException, "CreateWindowEx() failed.");
	}

	// cache parameters
	hdc = ::GetDC(hwnd);
	if(! hdc) {
		VW_THROW(SystemResourceException, "GetDC() failed.");
	}
	CacheGeometryParameters();

	// setup message-handler-mapper pointer
	::SetWindowLongPtr(hwnd, GWL_USERDATA, reinterpret_cast<LONG_PTR>(this));

	// setup mouse-cursor shapes
	ResetMouseCursorShapes();
	SetWindowClassMouseCursor(hMouseCursors[WindowArea_Client]);
	hActiveMouseCursor = hMouseCursors[WindowArea_Outside];

	// default settings
	SetText(VW_GET_OBJECT_NAME(*this));
	SetPositionAndSize(100, 100, 400, 200);
}


//====================================================================
void  TopLevelWindow::DestroySystemWindow()
{
	VW_ASSERT(hwnd, "System-window not created.");

	// remove message-handler-mapper pointer
	::SetWindowLongPtr(hwnd, GWL_USERDATA, 0);

	// release all resources
	::ReleaseDC(hwnd, hdc);
	hdc = 0;

	::DestroyWindow(hwnd);
	hwnd = 0;

	std::string  sClassName = VW_GET_CLASS_NAME(*this);
	::UnregisterClass(sClassName.c_str(), ::GetModuleHandle(NULL));
}


//====================================================================
DWORD __fastcall TopLevelWindow::GetStandardWindowStyle(bool bResizable)
{
	DWORD style = (WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX);
	if(bResizable) style |= (WS_MAXIMIZEBOX | WS_SIZEBOX);
	return style;
}


//====================================================================
void __fastcall TopLevelWindow::AdjustWindowSize(DWORD style, DWORD styleEx,
									uint& inOutClientWidth, uint& inOutClientHeight,
									uint& outTotalWidth, uint& outTotalHeight,
									bool bClampToWorkAreaSize)
{
	ClampLow<uint>(inOutClientWidth, 1);
	ClampLow<uint>(inOutClientHeight, 1);

	RECT r;
	r.left   = 0;
	r.right  = inOutClientWidth;
	r.top    = 0;
	r.bottom = inOutClientHeight;

	::AdjustWindowRectEx(&r, style, FALSE, styleEx);

	outTotalWidth = r.right - r.left;
	outTotalHeight = r.bottom - r.top;

	if(bClampToWorkAreaSize) {
		uint workAreaW, workAreaH;
		platform_info::GetWorkArea(&workAreaW, &workAreaH, 0, 0);

		if(outTotalWidth > workAreaW) {
			uint diff = outTotalWidth - workAreaW;
			outTotalWidth -= diff;
			inOutClientWidth -= diff;
		}
		if(outTotalHeight > workAreaH) {
			uint diff = outTotalHeight - workAreaH;
			outTotalHeight -= diff;
			inOutClientHeight -= diff;
		}
	}
}


//====================================================================
void  TopLevelWindow::SetUseGdiplus(bool bUse)
{
	if(bUse) {
		if(! gpg) gpg = new win_utils::GdiplusGuard();
	}
	else {
		SafeDelete(gpg);
	}
}


//====================================================================
bool  TopLevelWindow::MessageHandler(LRESULT& outResult, HWND hwnd, UINT msg,
										WPARAM wparam, LPARAM lparam)
{
	// NOTES: keep order of cases switch-construct-optimized.
	switch(msg) {

		case WM_MOVE: {
			CacheGeometryParameters();
			OnMoved();
			return true;
		}
		case WM_SIZE: {
			CacheGeometryParameters();

			if(bSystemCommandEmulation) {
				switch(wparam) {
					case SIZE_MAXIMIZED:
						OnResized(ResizeAction_Maximize);
						break;
					case SIZE_MINIMIZED:
						OnResized(ResizeAction_Minimize);
						break;
					case SIZE_RESTORED:
						OnResized(ResizeAction_Resize);
						break;
				}
			}
			else {
				switch(wparam) {
					case SIZE_MAXIMIZED:
						OnResized(ResizeAction_Maximize);
						break;
					case SIZE_MINIMIZED:
						OnResized(ResizeAction_Minimize);
						break;
					case SIZE_RESTORED:
						OnResized(ResizeAction_Resize);
						break;
				}
			}
			return true;
		}
		case WM_ACTIVATE: {
			bActive = (LOWORD(wparam) != WA_INACTIVE);
			bool bMinimized = (HIWORD(wparam) != 0);

			// this combination can only cause problems
			if(bMinimized  &&  bActive) return true;

			// send notification
			OnActivated(bActive);
			return true;
		}
		case WM_PAINT: {
			OnPaint();
			// stop further generation of WM_PAINT messages
			::ValidateRect(hwnd, NULL);
			return true;
		}
		case WM_CLOSE: {
			OnClose();
			return true;
		}
		case WM_ERASEBKGND: {
			outResult = 1; //halt further processing
			return true;
		}
		case WM_SYSCOLORCHANGE: {
			OnSystemColorsChanged();
			OnPaint();
			return true;
		}
		case WM_SETTINGCHANGE: {
			OnSystemSettingsChanged();
			return true;
		}
		case WM_SETCURSOR: {
			if(bSystemCommandEmulation  &&  bMouseCursorVisible) {
				::SetCursor(hActiveMouseCursor);
				outResult = TRUE; //halt further processing
				return true;
			}
			return false;
		}
		case WM_DISPLAYCHANGE: {
			OnDisplaySettingsChanged();
			OnPaint();
			return true;
		}
		case WM_GETMINMAXINFO: {
			LPMINMAXINFO p = reinterpret_cast<LPMINMAXINFO>(lparam);

			if(bSystemCommandEmulation) {
				if(maxWidth != UINT_MAX) {
					p->ptMaxSize.x = maxWidth;
				}
				if(maxHeight != UINT_MAX) {
					p->ptMaxSize.y = maxHeight;
				}
			}
			else {
				p->ptMinTrackSize.x = minWidth;
				p->ptMinTrackSize.y = minHeight;
				if(maxWidth != UINT_MAX) {
					p->ptMaxTrackSize.x = maxWidth;
					p->ptMaxSize.x      = maxWidth;
				}
				if(maxHeight != UINT_MAX) {
					p->ptMaxTrackSize.y = maxHeight;
					p->ptMaxSize.y      = maxHeight;
				}
			}
			p->ptMaxPosition.x = maximizedPositionX;
			p->ptMaxPosition.y = maximizedPositionY;
			return true;
		}
		case WM_NCHITTEST: {
			if(bSystemCommandEmulation) {
				outResult = HTCLIENT;
				return true;
			}
			return false;
		}
		case WM_KEYDOWN: {
			if(bDisableMessageDispatchKeyboardKeyStates) return true;

			// hook has priority
			if(pKeyboardMessageHook  &&  pKeyboardMessageHook->OnKeyDown(wparam, lparam)) {
				return true;
			}
			OnKeyDown(wparam, lparam);
			return true;
		}
		case WM_KEYUP: {
			if(bDisableMessageDispatchKeyboardKeyStates) return true;

			// hook has priority
			if(pKeyboardMessageHook  &&  pKeyboardMessageHook->OnKeyUp(wparam, lparam)) {
				return true;
			}
			OnKeyUp(wparam, lparam);
			return true;
		}
		case WM_CHAR: {
			if(bDisableMessageDispatchKeyboardCharEvents) return true;

			char ch = static_cast<char>(wparam);

			// hook has priority
			if(pKeyboardMessageHook  &&  pKeyboardMessageHook->OnChar(ch)) {
				return true;
			}
			OnChar(ch);
			return true;
		}
		case WM_SYSKEYDOWN: {
			switch(wparam) {
				case VK_F4:
					// block message if disabled
					if(! bEnableCloseThroughAltF4) return true;
					break;

				case VK_F10:
					// block message if disabled
					if(! bEnableMenuThroughF10) return true;
					break;
			}
			return false;
		}
		case WM_SYSCOMMAND: {
			switch(wparam) {
				case SC_KEYMENU:
					switch(lparam) {
						case VK_SPACE:
							// block message if disabled
							if(! bEnableMenuThroughAltSpace) return true;
							break;
					}
					break;
			}
			return false;
		}
		case WM_MOUSEMOVE: {
			if(bDisableMessageDispatchMouseMove) return true;

			win_utils::Point posClient(lparam);

			// hook has priority
			if(pMouseMessageHook  &&
				pMouseMessageHook->OnMouseMove(posClient, wparam)) {
				return true;
			}
			if(bSystemCommandEmulation) {

				// emulate window-movement
				if(bMoving) {
					win_utils::Point  mousePosScreen = posClient;
					::ClientToScreen(hwnd, &mousePosScreen);

					win_utils::Point  diff = mousePosScreen - lastMousePosScreen;
					lastMousePosScreen = mousePosScreen;

					int lastX, lastY;
					GetCurrentPosition(lastX, lastY);

					int newX = lastX + diff.x;
					int newY = lastY + diff.y;

					// realize movement if necessary
					if(newX != lastX  ||  newY != lastY) {
						SetPosition(newX, newY);
					}
					return true;
				}

				// emulate window-sizing
				if(bSizing) {
					int newWidth = width;
					if(bSizingRight) {
						newWidth = ClampedLow<int>(posClient.x + sizeGripOffsetClient.x, 0);
					}

					int newHeight = height;
					if(bSizingBottom) {
						newHeight = ClampedLow<int>(posClient.y + sizeGripOffsetClient.y, 0);
					}

					// realize new size if necessary
					if(newWidth != width  ||  newHeight != height) {
						SetSize(newWidth, newHeight);
					}
					return true;
				}

				// setup mouse-leave tracking
				if(bMouseLeaveTrackingInactive) {
					ActivateMouseLeaveTracking();
				}
				// determine window-area
				uint newMouseAreaSubCode = 0;
				eWindowArea newMouseArea = MouseHitTest(posClient, newMouseAreaSubCode);

				if(newMouseArea != currentMouseArea  ||  newMouseAreaSubCode != currentMouseAreaSubCode) {
					currentMouseArea = newMouseArea;
					currentMouseAreaSubCode = newMouseAreaSubCode;
					ChangeActiveMouseCursor();
					OnMouseEnter(currentMouseArea, currentMouseAreaSubCode);
				}
				if(currentMouseArea != WindowArea_Client) {
					return true;
				}
			}
			// report mouse-move
			OnMouseMove(posClient, wparam);
			return true;
		}
		case WM_LBUTTONDOWN: {
			if(bDisableMessageDispatchMouseButtons) return true;

			win_utils::Point posClient(lparam);

			// hook has priority
			if(pMouseMessageHook  &&
				pMouseMessageHook->OnMouseButtonLeft(true, posClient, wparam)) {
				return true;
			}
			if(bSystemCommandEmulation) {
				switch(currentMouseArea) {
					case WindowArea_Move: {
						if(CanMoveWindow()) {
							bMoving = true;
							EnterSizeMove(posClient);
						}
						return true;
					}
					case WindowArea_SizeBottom:
					case WindowArea_SizeRight:
					case WindowArea_SizeBottomRight: {
						if(CanSizeWindow()) {
							bSizing = true;
							EnterSizeMove(posClient);
							ExtractSizingStates(posClient);
						}
						return true;
					}
					case WindowArea_Minimize: {
						Minimize();
						return true;
					}
					case WindowArea_MaximizeRestore: {
						if(::IsZoomed(hwnd)) {
							Restore();
						}
						else {
							Maximize();
						}
						SynthesizeMouseMove();
						return true;
					}
					case WindowArea_AlwaysOnTop: {
						SetAlwaysOnTop(! bAlwaysOnTop);
						OnAlwaysOnTop(bAlwaysOnTop);
						return true;
					}
					case WindowArea_Close: {
						::PostMessage(hwnd, WM_CLOSE, 0, 0);
						return true;
					}
					case WindowArea_Command: {
						OnCommandAreaActivated(currentMouseAreaSubCode);
						return true;
					}
					// WindowArea_Client
				}
			}
			// report mouse-button-left
			OnMouseButtonLeft(true, posClient, wparam);
			return true;
		}
		case WM_LBUTTONUP: {
			if(bDisableMessageDispatchMouseButtons) return true;

			win_utils::Point posClient(lparam);

			// hook has priority
			if(pMouseMessageHook  &&
				pMouseMessageHook->OnMouseButtonLeft(false, posClient, wparam)) {
				return true;
			}
			// report mouse-button-left
			if(bSystemCommandEmulation) {
				if(ExitSizeMove()) return true;
			}
			OnMouseButtonLeft(false, posClient, wparam);
			return true;
		}
		case WM_RBUTTONDOWN: {
			if(bDisableMessageDispatchMouseButtons) return true;

			win_utils::Point posClient(lparam);

			// hook has priority
			if(pMouseMessageHook  &&
				pMouseMessageHook->OnMouseButtonRight(true, posClient, wparam)) {
				return true;
			}
			// report mouse-button-right
			OnMouseButtonRight(true, posClient, wparam);
			return true;
		}
		case WM_RBUTTONUP: {
			if(bDisableMessageDispatchMouseButtons) return true;

			win_utils::Point posClient(lparam);

			// hook has priority
			if(pMouseMessageHook  &&
				pMouseMessageHook->OnMouseButtonRight(false, posClient, wparam)) {
				return true;
			}
			// report mouse-button-right
			OnMouseButtonRight(false, posClient, wparam);
			return true;
		}
		case WM_MBUTTONDOWN: {
			if(bDisableMessageDispatchMouseButtons) return true;

			win_utils::Point posClient(lparam);

			// hook has priority
			if(pMouseMessageHook  &&
				pMouseMessageHook->OnMouseButtonMiddle(true, posClient, wparam)) {
				return true;
			}
			// report mouse-button-middle
			OnMouseButtonMiddle(true, posClient, wparam);
			return true;
		}
		case WM_MBUTTONUP: {
			if(bDisableMessageDispatchMouseButtons) return true;

			win_utils::Point posClient(lparam);

			// hook has priority
			if(pMouseMessageHook  &&
				pMouseMessageHook->OnMouseButtonMiddle(false, posClient, wparam)) {
				return true;
			}
			// report mouse-button-middle
			OnMouseButtonMiddle(false, posClient, wparam);
			return true;
		}
		case WM_MOUSEWHEEL: {
			if(bDisableMessageDispatchMouseWheel) return true;

			int delta = static_cast<short>(HIWORD(wparam)) / WHEEL_DELTA;
			win_utils::Point posClient(lparam);

			// hook has priority
			if(pMouseMessageHook  &&
				pMouseMessageHook->OnMouseWheel(delta, posClient, wparam)) {
				return true;
			}
			// report mouse-wheel
			OnMouseWheel(delta, posClient, wparam);
			return true;
		}
		case WM_MOUSELEAVE: {
			if(bSystemCommandEmulation) {
				bMouseLeaveTrackingInactive = true;
				currentMouseArea = WindowArea_Outside;
				currentMouseAreaSubCode = 0;
				OnMouseEnter(currentMouseArea, currentMouseAreaSubCode);
				return true;
			}
			return false;
		}
	}
	return false;
}


//====================================================================
void  TopLevelWindow::EnableCloseThroughAltF4(bool bEnable)
{
	bEnableCloseThroughAltF4 = bEnable;
}


//====================================================================
void  TopLevelWindow::EnableMenuThroughF10(bool bEnable)
{
	bEnableMenuThroughF10 = bEnable;
}


//====================================================================
void  TopLevelWindow::EnableMenuThroughAltSpace(bool bEnable)
{
	bEnableMenuThroughAltSpace = bEnable;
}


//====================================================================
void  TopLevelWindow::EnableSystemCommandEmulation(bool bEnable)
{
	if(bSystemCommandEmulation) {
		ExitSizeMove();
	}
	bSystemCommandEmulation = bEnable  &&  bPopup;
	bMouseLeaveTrackingInactive = true;
}


//====================================================================
TopLevelWindow::eWindowArea  TopLevelWindow::MouseHitTest(const win_utils::Point& mousePosClient,
														uint& commandAreaID)
{
	int b = GADGET_SIZE;

	if(mousePosClient.HitTestRectangle(0, 0, b, b)) {
		return WindowArea_Move;
	}
	if(mousePosClient.HitTestRectangle(width - b, height - b, b, b)) {
		return WindowArea_SizeBottomRight;
	}
	if(mousePosClient.HitTestRectangle(width - 3 * b, 0, b, b)) {
		return WindowArea_Minimize;
	}
	if(mousePosClient.HitTestRectangle(width - 2 * b, 0, b, b)) {
		return WindowArea_MaximizeRestore;
	}
	if(mousePosClient.HitTestRectangle(b, 0, b, b)) {
		return WindowArea_AlwaysOnTop;
	}
	if(mousePosClient.HitTestRectangle(width - b, 0, b, b)) {
		return WindowArea_Close;
	}
	return WindowArea_Client;
}


//====================================================================
std::wstring  TopLevelWindow::GetGeometryStringWindowW()
{
	int x, y;
	GetCurrentPosition(x, y);

	std::wostringstream  woss;
	woss << width << L"x" << height << L"@" << x << L"." << y;
	return  woss.str();
}


//====================================================================
std::wstring  TopLevelWindow::GetGeometryStringClientW()
{
	std::wostringstream  woss;
	woss << GetClientAreaWidth() << L"x" << GetClientAreaHeight();
	return  woss.str();
}


//====================================================================
void  TopLevelWindow::CacheGeometryParameters()
{
	RECT r;

	// total-size and position
	::GetWindowRect(hwnd, &r);

	if(IsMaximized()) {
		maximizedPositionX = r.left;
		maximizedPositionY = r.top;
	}
	else {
		positionX = r.left;
		positionY = r.top;
	}
	width  = r.right  - r.left;
	height = r.bottom - r.top;

	// client-area-size
	::GetClientRect(hwnd, &r);
	clientAreaWidth  = r.right;
	clientAreaHeight = r.bottom;
}


//====================================================================
void  TopLevelWindow::OnSystemColorsChanged()
{
}


//====================================================================
void  TopLevelWindow::OnSystemSettingsChanged()
{
}


//====================================================================
void  TopLevelWindow::OnDisplaySettingsChanged()
{
}


//====================================================================
void  TopLevelWindow::OnActivated(bool bActive)
{
}


//====================================================================
void  TopLevelWindow::OnClose()
{
	PostQuit();
}


//====================================================================
void  TopLevelWindow::OnMoved()
{
}


//====================================================================
void  TopLevelWindow::OnResized(eResizeAction action)
{
}


//====================================================================
void  TopLevelWindow::OnPaint()
{
	VW_REENTRANT_ASSERT_RETURNS_VOID(gpg, "Gdiplus not initialized.");
	using namespace Gdiplus;

	int b = GADGET_SIZE;

	// background
	Graphics g(hdc);
	g.Clear(Color::Black);

	SolidBrush brush(Color::Red);
	Pen pen(Color::Red);

	if(bPopup  &&  bSystemCommandEmulation) {
		// draw default system-command frames
		int rw = width - 1;
		int rh = height - 1;
		int rb = b - 1;
		g.DrawRectangle(&pen, 0, 0, rw, rh);
		g.DrawRectangle(&pen, 0, 0, rb, rb);
		g.DrawRectangle(&pen, b, 0, rb, rb);
		g.DrawRectangle(&pen, width - b, 0, rb, rb);
		g.DrawRectangle(&pen, width - b * 2, 0, rb, rb);
		g.DrawRectangle(&pen, width - b * 3, 0, rb, rb);
		g.DrawRectangle(&pen, width - b, height - b, rb, rb);
	}

	// prepare text resources
	Font font(FontFamily::GenericSansSerif(), 16, 0, UnitPixel);
	StringFormat sf;
	sf.SetAlignment(StringAlignmentCenter);

	// print current window-text
	std::wstring ws;
	AsciiToUnicode(GetText(), ws);

	sf.SetLineAlignment(StringAlignmentFar);
	RectF rect(0, 0, width, height / 2);
	g.DrawString(ws.c_str(), ws.length(), &font, rect, &sf, &brush);

	// print window-geometry info
	ws = bPopup ? GetGeometryStringWindowW() : GetGeometryStringClientW();

	rect.Y = height / 2;
	sf.SetLineAlignment(StringAlignmentNear);
	g.DrawString(ws.c_str(), ws.length(), &font, rect, &sf, &brush);
}


//====================================================================
void  TopLevelWindow::OnMouseEnter(eWindowArea area, uint areaSubCode)
{
}


//====================================================================
void  TopLevelWindow::OnAlwaysOnTop(bool bOnTop)
{
}


//====================================================================
void  TopLevelWindow::OnCommandAreaActivated(uint areaSubCode)
{
}


//====================================================================
#define  CALL_HOOK(DispatchPoint) \
	if(s_messageHook) { \
		s_messageHook->Process(DispatchPoint, MessageHook::WinMsg(hwnd, msg, wparam, lparam), result, bProcessed); \
	}


//====================================================================
LRESULT CALLBACK  TopLevelWindow::MessageHandlerMapper(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	// setup common result value
	LRESULT  result     = 0;
	bool     bProcessed = false;

	// top-level-window processing
	TopLevelWindow* w = reinterpret_cast<TopLevelWindow*>(::GetWindowLongPtr(hwnd, GWL_USERDATA));
	if(w) {

		CALL_HOOK(MessageHook::DispatchPoint_PreMessageHandler);
		if(bProcessed) return result;

		bProcessed = w->MessageHandler(result, hwnd, msg, wparam, lparam);

		CALL_HOOK(MessageHook::DispatchPoint_PostMessageHandler);
		if(bProcessed) return result;
	}

	// def-window-proc processing
	CALL_HOOK(MessageHook::DispatchPoint_PreDefWindowProc);
	if(bProcessed) return result;

	result = ::DefWindowProc(hwnd, msg, wparam, lparam);
	bProcessed = true;

	CALL_HOOK(MessageHook::DispatchPoint_PostDefWindowProc);
	return result;
}


//====================================================================
#undef  CALL_HOOK


//====================================================================
bool  TopLevelWindow::AddWindowStyles(DWORD style, DWORD styleEx)
{
	LONG r = ::SetWindowLongPtr(hwnd, GWL_STYLE, ::GetWindowLongPtr(hwnd, GWL_STYLE) | style);
	if(! r) return false;
	r = ::SetWindowLongPtr(hwnd, GWL_EXSTYLE, ::GetWindowLongPtr(hwnd, GWL_EXSTYLE) | styleEx);
	return ToBool(r);
}


//====================================================================
bool  TopLevelWindow::RemoveWindowStyles(DWORD style, DWORD styleEx)
{
	LONG r = ::SetWindowLongPtr(hwnd, GWL_STYLE, ::GetWindowLongPtr(hwnd, GWL_STYLE) & ~style);
	if(! r) return false;
	r = ::SetWindowLongPtr(hwnd, GWL_EXSTYLE, ::GetWindowLongPtr(hwnd, GWL_EXSTYLE) & ~styleEx);
	return ToBool(r);
}


//====================================================================
bool  TopLevelWindow::GetMousePositionClient(win_utils::Point& mousePosClient)
{
	bool bInside = false;
	::GetCursorPos(&mousePosClient);

	int x, y;
	GetCurrentPosition(x, y);
	if(mousePosClient.HitTestRectangle(x, y, width, height)) {
		bInside = true;
	}
	::ScreenToClient(hwnd, &mousePosClient);
	return bInside;
}


//====================================================================
void  TopLevelWindow::ActivateMouseLeaveTracking()
{
	TRACKMOUSEEVENT tme;
	tme.cbSize = sizeof(tme);
	tme.dwFlags = TME_LEAVE;
	tme.hwndTrack = GetHWND();
	tme.dwHoverTime = 0;
	::TrackMouseEvent(&tme);

	bMouseLeaveTrackingInactive = false;
}


//====================================================================
void  TopLevelWindow::ChangeActiveMouseCursor()
{
	switch(currentMouseArea) {

		case WindowArea_Move:
			if(! CanMoveWindow()) {
				hActiveMouseCursor = hMouseCursors[WindowArea_Client];
				return;
			}
			break;

		case WindowArea_SizeBottom:
		case WindowArea_SizeRight:
		case WindowArea_SizeBottomRight:
			if(! CanSizeWindow()) {
				hActiveMouseCursor = hMouseCursors[WindowArea_Client];
				return;
			}
			break;
	}
	hActiveMouseCursor = hMouseCursors[currentMouseArea];
}


//====================================================================
void  TopLevelWindow::SetWindowClassMouseCursor(HCURSOR hcursor)
{
	::SetClassLongPtr(hwnd, GCLP_HCURSOR, reinterpret_cast<LONG_PTR>(hcursor));
}


//====================================================================
void  TopLevelWindow::EnterSizeMove(const win_utils::Point& mousePosClient)
{
	lastMousePosScreen = mousePosClient;
	::ClientToScreen(hwnd, &lastMousePosScreen);
	::SetCapture(hwnd);
}


//====================================================================
bool  TopLevelWindow::ExitSizeMove()
{
	if(bMoving  ||  bSizing) {
		bMoving = false;
		bSizing = false;
		::ReleaseCapture();
		return true;
	}
	return false;
}


//====================================================================
void  TopLevelWindow::ExtractSizingStates(const win_utils::Point& mousePosClient)
{
	switch(currentMouseArea) {
		case WindowArea_SizeBottom:
			bSizingBottom = true;
			bSizingRight  = false;
			break;

		case WindowArea_SizeRight:
			bSizingBottom = false;
			bSizingRight  = true;
			break;

		case WindowArea_SizeBottomRight:
			bSizingBottom = true;
			bSizingRight  = true;
			break;
	}
	sizeGripOffsetClient.x = width  - mousePosClient.x;
	sizeGripOffsetClient.y = height - mousePosClient.y;
}


//====================================================================
bool  TopLevelWindow::IsZoomedToWorkAreaSize()
{
	uint workAreaW, workAreaH;
	platform_info::GetWorkArea(&workAreaW, &workAreaH, 0, 0);

	return (width == workAreaW  &&  height == workAreaH);
}


//====================================================================
void  TopLevelWindow::MinMaxClamp(uint& w, uint& h)
{
	Clamp<uint>(w, minWidth, maxWidth);
	Clamp<uint>(h, minHeight, maxHeight);
}


//********************************************************************
}//namespace vertexware

