

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Color.h"
#include "BaseTools\WinUtils.h"
#include "private\vw_common_exports.h"
#include <string>


//********************************************************************
namespace vertexware {


//====================================================================
class TopLevelWindow;


//====================================================================
class VW_COMMON_API KeyboardMessageHook
{
public:
	virtual ~KeyboardMessageHook() {}

	//-------------------------------------
	// Return false to continue processing.
	//-------------------------------------
	virtual  bool  OnKeyDown(WPARAM virtKeyCode, LPARAM keyData)
	{ return false; }

	virtual  bool  OnKeyUp(WPARAM virtKeyCode, LPARAM keyData)
	{ return false; }

	virtual  bool  OnChar(char ch)
	{ return false; }
};


//====================================================================
class VW_COMMON_API MouseMessageHook
{
public:
	virtual ~MouseMessageHook() {}

	//-------------------------------------
	// Return false to continue processing.
	//-------------------------------------
	virtual  bool  OnMouseMove(const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ return false; }

	// <delta> is normalized (already divided by WHEEL_DELTA)
	virtual  bool  OnMouseWheel(int delta,
								const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ return false; }

	virtual  bool  OnMouseButtonLeft(bool bDown,
								const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ return false; }

	virtual  bool  OnMouseButtonMiddle(bool bDown,
								const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ return false; }

	virtual  bool  OnMouseButtonRight(bool bDown,
								const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ return false; }
};


//====================================================================
// Debugging hooks.
//====================================================================
class VW_COMMON_API MessageHook
{
public:
	//-------------------------
	struct VW_COMMON_API WinMsg
	{
		HWND    hwnd;
		UINT    msg;
		WPARAM  wparam;
		LPARAM  lparam;

		WinMsg(HWND _hwnd, UINT _msg, WPARAM _wparam, LPARAM _lparam)
		{
			hwnd = _hwnd;
			msg = _msg;
			wparam = _wparam;
			lparam = _lparam;
		}
	};

	//-----------------
	enum eDispatchPoint
	{
		DispatchPoint_PreMessageHandler = 0,
		DispatchPoint_PostMessageHandler,
		DispatchPoint_PreDefWindowProc,
		DispatchPoint_PostDefWindowProc
	};


public:
	virtual ~MessageHook();

	virtual  void  Process(eDispatchPoint dp, const WinMsg& wm,
							LRESULT& result, bool& bProcessed) = 0;
};


//====================================================================
class VW_COMMON_API TopLevelWindow
{
public:
	static void         __fastcall SetMessageHook(MessageHook* hook);
	static MessageHook* __fastcall GetMessageHook();


public:
	//--------------
	enum eWindowArea
	{
		WindowArea_Outside = 0,
		WindowArea_Move,
		WindowArea_SizeBottom,
		WindowArea_SizeRight,
		WindowArea_SizeBottomRight,
		WindowArea_Minimize,
		WindowArea_MaximizeRestore,
		WindowArea_AlwaysOnTop,
		WindowArea_Close,
		WindowArea_Command,
		WindowArea_Client,

		WindowArea_Count,
		WindowArea_Force32Bits = 0x7FFFFFFF
	};

	//----------------
	enum eResizeAction
	{
		ResizeAction_Minimize = 0,
		ResizeAction_Resize,
		ResizeAction_Maximize,

		ResizeAction_Count,
		ResizeAction_Force32Bits = 0x7FFFFFFF //not used
	};


public:
	virtual  ~TopLevelWindow();

	HWND   GetHWND() const { return hwnd; }
	HDC    GetHDC() const { return hdc;  }

	void   SetIcon(HICON hIcon = 0);
	HICON  GetIcon() const;

	void                SetText(const std::string& s);
	const std::string&  GetText() const { return windowText; }

	bool  Activate();
	bool  IsActive() const { return bActive; }

	void  SetAlwaysOnTop(bool bFlag);
	bool  IsAlwaysOnTop() const { return bAlwaysOnTop; }

	bool  IsPopup() const { return bPopup; }

	//---------------
	// Message-hooks.
	//---------------
	void  EnableMessageDispatchMouseMove(bool bEnable);
	void  EnableMessageDispatchMouseWheel(bool bEnable);
	void  EnableMessageDispatchMouseButtons(bool bEnable);
	void  EnableMessageDispatchKeyboardKeyStates(bool bEnable);
	void  EnableMessageDispatchKeyboardCharEvents(bool bEnable);

	void  SetMouseMessageHook(MouseMessageHook* hook = 0);
	void  SetKeyboardMessageHook(KeyboardMessageHook* hook = 0);

	//-------------------------------------------------------------------
	// Transparency: pixels colored with <colorKey> will be transparent.
	// Translucency: 0 = completely transparent, 255 = completely opaque.
	//-------------------------------------------------------------------
	bool  EnableCompositingEffects(bool bEnable); //returns false if failed
	void  SetOpacity(ubyte opacity = 255);
	void  SetColorKey(color4b colorKey = color4b::BLACK);
	void  SetOpacityAndColorKey(ubyte opacity = 255, color4b colorKey = color4b::BLACK);

	void  SetForceRepaintOnMove(bool bEnable = false);

	void  SetPosition(int x, int y);
	void  UpdatePosition(int deltaX, int deltaY);
	void  SetMaximizedPosition(int x = 0, int y = 0);
	int   GetPositionX() const { return positionX; }
	int   GetPositionY() const { return positionY; }
	int   GetMaximizedPositionX() const { return maximizedPositionX; }
	int   GetMaximizedPositionY() const { return maximizedPositionY; }
	void  GetCurrentPosition(int& x, int& y) const; //maximized or standard

	void  SetSize(uint w, uint h);
	void  UpdateSize(int deltaW = 0, int deltaH = 0);
	void  SetMinSize(uint w = 0, uint h = 0); //0 means absolute minimum
	void  GetMinSize(uint& w, uint& h) const;
	void  SetMaxSize(uint w = 0, uint h = 0); //0 means no-limit
	void  GetMaxSize(uint& w, uint& h) const;
	uint  GetWidth() const { return width; }
	uint  GetHeight() const { return height; }
	void  GetSize(int& w, int& h) const;
	uint  GetClientAreaWidth() const { return clientAreaWidth; }
	uint  GetClientAreaHeight() const { return clientAreaHeight; }
	void  GetClientAreaSize(int& w, int& h) const;

	void  SetPositionAndSize(int x, int y, uint w, uint h);
	void  CenterOnDesktop(uint w, uint h);
	void  FillDesktop();
	void  GetVisibleClientAreaScreenCoords(RECT& rect) const;

	void  ShowMouseCursor(bool bShow);
	void  SetMouseCursorShape(eWindowArea area = WindowArea_Client, HCURSOR hcursor = 0); //0 means default
	void  ResetMouseCursorShapes(); //all to default
	void  SynthesizeMouseMove();

	void  Show();
	void  ShowMinimized(bool bActivate = false);
	void  ShowMaximized();
	void  Hide();

	void  Minimize();
	void  Maximize();
	void  Restore();

	bool  IsMinimized() const { return ToBool(::IsIconic(hwnd)); }
	bool  IsMaximized() const { return ToBool(::IsZoomed(hwnd)); }

	//---------------------------------------------------
	// Nudge window's message-pump in case it is waiting.
	//---------------------------------------------------
	void  PostNullMessage() { if(hwnd) ::PostMessage(hwnd, WM_NULL, 0, 0); }


protected:
	//-------------
	// Constructor.
	//-------------
	TopLevelWindow(bool bUsingGdiplus = true);

	//--------------------------
	// System-window management.
	//--------------------------
	void   CreateSystemWindow(UINT classStyle, DWORD windowStyle, DWORD windowStyleEx);
	void   DestroySystemWindow();

	static DWORD __fastcall GetStandardWindowStyle(bool bResizable);

	static void  __fastcall AdjustWindowSize(DWORD style, DWORD styleEx,
									uint& inOutClientWidth, uint& inOutClientHeight,
									uint& outTotalWidth, uint& outTotalHeight,
									bool bClampToWorkAreaSize = true);

	void   SetUseGdiplus(bool bUse);

	//--------------------------------------------------------
	// Windows-message handler.
	// Override to handle additional messages, but if message
	// is not handled always call this-one from override.
	// If message is handled, put result code into <outResult>
	// and return true, otherwise just return false.
	// <outResult> is initially set to 0 for convenience.
	//--------------------------------------------------------
	virtual  bool  MessageHandler(LRESULT& outResult, HWND hwnd, UINT msg,
									WPARAM wparam, LPARAM lparam);

	//-----------------------
	// Native-system-command.
	//-----------------------
	void  EnableCloseThroughAltF4(bool bEnable);
	void  EnableMenuThroughF10(bool bEnable);
	void  EnableMenuThroughAltSpace(bool bEnable);

	//--------------------------
	// System-command emulation.
	//--------------------------
	void  EnableSystemCommandEmulation(bool bEnable); //popup only

	eWindowArea  GetCurrentMouseArea() const { return currentMouseArea; }
	uint         GetCurrentMouseAreaSubCode() const { return currentMouseAreaSubCode; }

	bool  IsMouseInArea(eWindowArea area) const { return (currentMouseArea == area); }
	bool  IsMouseInCommandArea(uint areaSubCode) const { return IsMouseInArea(WindowArea_Command)
													&&  (currentMouseAreaSubCode == areaSubCode); }

	//----------------------------------------------------------------
	// When using system-command-emulation with popup-style
	// windows, this should be overriden to test for command-area hit.
	//----------------------------------------------------------------
	virtual  eWindowArea  MouseHitTest(const win_utils::Point& mousePosClient, uint& areaSubCode);

	//-------------------------
	// Window-geometry helpers.
	//-------------------------
	std::wstring  GetGeometryStringWindowW();
	std::wstring  GetGeometryStringClientW();

	void   CacheGeometryParameters();

	//-----------------------------------
	// Common Win32-function redirectors.
	//-----------------------------------
	void  PostRepaint() { ::InvalidateRect(hwnd, 0, FALSE); }
	void  PostQuit() { if(hwnd) ::PostQuitMessage(0); }

	int   IsKeyDown(int virtKeyCode) { return (::GetKeyState(virtKeyCode) & 0x8000); }
	int   IsKeyToggled(int virtKeyCode) { return (::GetKeyState(virtKeyCode) & 0x0001); }

	//-------------------------
	// Standard event-handlers.
	//-------------------------
	virtual  void  OnSystemColorsChanged();
	virtual  void  OnSystemSettingsChanged();
	virtual  void  OnDisplaySettingsChanged();
	virtual  void  OnActivated(bool bActive);
	virtual  void  OnClose();
	virtual  void  OnMoved();
	virtual  void  OnResized(eResizeAction action);
	virtual  void  OnPaint();

	//------------------------------------------------------
	// System-command-emulation event-handlers (popup only).
	//------------------------------------------------------
	virtual  void  OnMouseEnter(eWindowArea area, uint areaSubCode);
	virtual  void  OnAlwaysOnTop(bool bOnTop);
	virtual  void  OnCommandAreaActivated(uint areaSubCode);

	//---------------------------------------------------------------
	// Input event-handlers.
	// Key-flags for mouse-messages are WPARAM in WM_MOUSE* messages.
	//---------------------------------------------------------------
	virtual  void  OnMouseMove(const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ }
	virtual  void  OnMouseWheel(int delta,
								const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ }
	virtual  void  OnMouseButtonLeft(bool bDown,
								const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ }
	virtual  void  OnMouseButtonMiddle(bool bDown,
								const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ }
	virtual  void  OnMouseButtonRight(bool bDown,
								const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
	{ }

	virtual  void  OnKeyDown(WPARAM virtKeyCode, LPARAM keyData)
	{ }
	virtual  void  OnKeyUp(WPARAM virtKeyCode, LPARAM keyData)
	{ }
	virtual  void  OnChar(char ch)
	{ }


private:
	//--------------------------------
	// Windows-message-handler mapper.
	//--------------------------------
	static LRESULT CALLBACK MessageHandlerMapper(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);

	//------------------
	// Internal helpers.
	//------------------
	bool  AddWindowStyles(DWORD style, DWORD styleEx);
	bool  RemoveWindowStyles(DWORD style, DWORD styleEx);

	bool  GetMousePositionClient(win_utils::Point& mousePosClient);
	void  ActivateMouseLeaveTracking();
	void  ChangeActiveMouseCursor();
	void  SetWindowClassMouseCursor(HCURSOR hcursor);

	void  EnterSizeMove(const win_utils::Point& mousePosClient);
	bool  ExitSizeMove();
	void  ExtractSizingStates(const win_utils::Point& mousePosClient);
	bool  IsZoomedToWorkAreaSize();
	void  MinMaxClamp(uint& w, uint& h);

	__forceinline bool CanMoveWindow() { return ! IsMaximized()  ||  ! IsZoomedToWorkAreaSize(); }
	__forceinline bool CanSizeWindow() { return ! IsMaximized(); }


private:
	static MessageHook*  s_messageHook;

	win_utils::GdiplusGuard*  gpg;
	bool                      bPopup;
	HWND                      hwnd;
	HDC                       hdc;
	std::string               windowText;
	bool                      bActive;
	bool                      bAlwaysOnTop;
	bool                      bUsingCompositingEffects;
	UINT                      SWPMoveFlags;

	//---------------
	// Message-hooks.
	//---------------
	bool   bDisableMessageDispatchMouseMove;
	bool   bDisableMessageDispatchMouseWheel;
	bool   bDisableMessageDispatchMouseButtons;
	bool   bDisableMessageDispatchKeyboardKeyStates;
	bool   bDisableMessageDispatchKeyboardCharEvents;

	MouseMessageHook*     pMouseMessageHook;
	KeyboardMessageHook*  pKeyboardMessageHook;

	//-------------------
	// Position and size.
	//-------------------
	int    positionX;
	int    positionY;
	int    maximizedPositionX;
	int    maximizedPositionY;
	uint   width;
	uint   height;
	uint   clientAreaWidth;
	uint   clientAreaHeight;
	uint   minWidth;
	uint   maxWidth;
	uint   minHeight;
	uint   maxHeight;

	//---------------------
	// Mouse-cursor states.
	//---------------------
	bool     bMouseCursorVisible;
	HCURSOR  hActiveMouseCursor;
	HCURSOR  hMouseCursors[WindowArea_Count];

	//------------------------------
	// Native system-command states.
	//------------------------------
	bool  bEnableCloseThroughAltF4;
	bool  bEnableMenuThroughF10;
	bool  bEnableMenuThroughAltSpace;

	//----------------------------------------------
	// System-command-emulation states (popup only).
	//----------------------------------------------
	bool   bSystemCommandEmulation;
	bool   bMoving;
	bool   bSizing;
	bool   bSizingBottom;
	bool   bSizingRight;
	bool   bMouseLeaveTrackingInactive;

	eWindowArea  currentMouseArea;
	uint         currentMouseAreaSubCode;

	win_utils::Point  sizeGripOffsetClient;
	win_utils::Point  lastMousePosScreen;
};


//====================================================================
class VW_COMMON_API TopLevelWindowTestPopup : public TopLevelWindow
{
public:
	TopLevelWindowTestPopup() : TopLevelWindow(true)
	{
		CreateSystemWindow(0, WS_POPUP, WS_EX_APPWINDOW);
		EnableSystemCommandEmulation(true);
		SetForceRepaintOnMove(true);
		SetMinSize(80, 80);
		Show();
	};
};


//====================================================================
class VW_COMMON_API TopLevelWindowTestOverlapped : public TopLevelWindow
{
public:
	TopLevelWindowTestOverlapped() : TopLevelWindow(true)
	{
		CreateSystemWindow(0, WS_OVERLAPPEDWINDOW, WS_EX_APPWINDOW);
		SetMinSize(80, 80);
		SetPosition(200, 200);
		Show();
	};
};


//********************************************************************
}//namespace vertexware


