

//--------------------------------------------------------------------
#include "BaseTools\VariableProxy.h"
#include "BaseTools\STLExtensions.h"
#include "BaseTools\Debug.h"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
class VariableNamePredicate
{
public:
	VariableNamePredicate(const std::string& refName) : targetName(refName) {
	}

	bool operator()(const VariableProxy* vp) const {
		return (vp->GetName() == targetName);
	}

private:
	std::string  targetName;
};


//====================================================================
struct VariableSortPredicate
{
	bool operator()(const VariableProxy* lhs, const VariableProxy* rhs) const {
		return (lhs->GetName() < rhs->GetName());
	}
};


//====================================================================
struct VariableBinarySearchPredicate
{
	bool operator()(const VariableProxy* vp, const std::string& refName) const {
		return (vp->GetName() < refName);
	}
};


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
VariableProxy::Observer   VariableProxy::s_nullObserver;
VariableProxy::Observer*  VariableProxy::s_pGlobalObserver = 0;


//====================================================================
VariableProxy::VariableProxy()
{
	pObserver = 0;
	bDeveloperModeOnly = false;
}


//====================================================================
VariableProxy::~VariableProxy()
{
}


//====================================================================
void  VariableProxy::NotifyObserverAboutValueChange()
{
	if(pObserver) {
		pObserver->OnValueChanged(*this);
		return;
	}
	if(s_pGlobalObserver) {
		s_pGlobalObserver->OnValueChanged(*this);
	}
}


//====================================================================
VariableRegistry::VariableRegistry(const std::string& _registryName)
	: registryName(_registryName)
{
}


//====================================================================
VariableRegistry::~VariableRegistry()
{
	UnregisterAllVariables();
}


//====================================================================
bool  VariableRegistry::RegisterVariable(const VariableProxy& vp)
{
	// check for duplicate names
	if(Find(vp.GetName())) return false;

	// clone & store proxy object
	VariableProxy* vpClone = vp.Clone();
	if(! vpClone) return false;

	variables.push_back(vpClone);
	SortVariables();

	return true;
}


//====================================================================
void  VariableRegistry::UnregisterVariable(const std::string& name)
{
	if(DeleteAllFromPtrContainer_If(variables, VariableNamePredicate(name))) {
		SortVariables();
	}
}


//====================================================================
void  VariableRegistry::UnregisterAllVariables()
{
	DeletePtrContainer(variables);
}


//====================================================================
void  VariableRegistry::ForceValueRanges()
{
	for(uint i = 0; i < variables.size(); i++) {
		VariableProxy& vp = *variables[i];
		vp.SetValueFromString(vp.GetValueAsString());
	}
}


//====================================================================
VariableProxy*  VariableRegistry::GetVariable(const std::string& name)
{
	return Find(name);
}


//====================================================================
uint  VariableRegistry::GetNumVariables()
{
	return variables.size();
}


//====================================================================
VariableProxy&  VariableRegistry::GetVariable(uint i)
{
	VW_ASSERT(i < variables.size(), "Variable-index out of range.");
	return *variables[i];
}


//====================================================================
VariableProxy*  VariableRegistry::Find(const std::string& name)
{
	VAR_ITER it;

	// search sorted sequence O(log(N))
	it = std::lower_bound(variables.begin(), variables.end(),
						name, VariableBinarySearchPredicate());

	if(it == variables.end()) return 0;
	VariableProxy* vp = *it;
	return (vp->GetName() == name) ? vp : 0;
}


//====================================================================
void  VariableRegistry::SortVariables()
{
	std::sort(variables.begin(), variables.end(), VariableSortPredicate());
}


//********************************************************************
}//namespace vertexware


