

//--------------------------------------------------------------------
#include "BaseTools\WinUtils.h"
#include "BaseTools\Exception.h"
#include "BaseTools\Debug.h"
#include "BaseTools\Serializers.h"
#include "BaseTools\Memory.h"
#include "BaseTools\String.h"
#include <sstream>
#include <iomanip>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace win_utils {


//====================================================================
GdiplusGuard::GdiplusGuard()
{
	Gdiplus::GdiplusStartupInput  startupInput;
	if(Gdiplus::GdiplusStartup(&token, &startupInput, 0) != Gdiplus::Ok) {
		VW_THROW(SystemResourceException, "Gdiplus startup failed.");
	}
}


//====================================================================
GdiplusGuard::~GdiplusGuard()
{
	Gdiplus::GdiplusShutdown(token);
}


//====================================================================
void __fastcall GdiplusGetFontFamilyNames(std::vector<std::string>& outFontFamilyNames)
{
	outFontFamilyNames.clear();

	GdiplusGuard  gpg;
	using namespace Gdiplus;

	// collect available gdiplus font-family objects
	InstalledFontCollection  ifc;
	uint numFamilies = ifc.GetFamilyCount();

	Buffer<FontFamily>  fontFamilies(numFamilies);

	int numFound = 0;
	if(ifc.GetFamilies(numFamilies, fontFamilies, &numFound) != Gdiplus::Ok) return;

	// extract font-family names
	outFontFamilyNames.reserve(numFound);

	for(uint i = 0; i < numFound; i++) {
		FontFamily& ff = fontFamilies[i];

		wchar_t wFaceName[LF_FACESIZE];
		if(ff.GetFamilyName(wFaceName) != Gdiplus::Ok) continue;

		std::string faceName;
		UnicodeToAscii(wFaceName, faceName);

		outFontFamilyNames.push_back(faceName);
	}
}


//====================================================================
Gdiplus::Color __fastcall GetSystemColor(eSystemColor sc)
{
	static  int  sysColorTable[SystemColor_Count] =
	{
		COLOR_DESKTOP
	};
	DWORD c = ::GetSysColor(sysColorTable[sc]);
	return Gdiplus::Color(GetRValue(c), GetGValue(c), GetBValue(c));
}


//====================================================================
HCURSOR __fastcall LoadStockCursor(eStockCursor stockCursorType)
{
	static  LPCTSTR  oemResourceTable[StockCursor_Count] =
	{
		MAKEINTRESOURCE(OCR_NORMAL),
		MAKEINTRESOURCE(OCR_IBEAM),
		MAKEINTRESOURCE(OCR_WAIT),
		MAKEINTRESOURCE(OCR_CROSS),
		MAKEINTRESOURCE(OCR_UP),
		MAKEINTRESOURCE(OCR_SIZENWSE),
		MAKEINTRESOURCE(OCR_SIZENESW),
		MAKEINTRESOURCE(OCR_SIZEWE),
		MAKEINTRESOURCE(OCR_SIZENS),
		MAKEINTRESOURCE(OCR_SIZEALL),
		MAKEINTRESOURCE(OCR_NO),
		MAKEINTRESOURCE(OCR_HAND),
		MAKEINTRESOURCE(OCR_APPSTARTING)
	};
	return static_cast<HCURSOR>(::LoadImage(0, oemResourceTable[stockCursorType],
									IMAGE_CURSOR, 0, 0, LR_DEFAULTSIZE | LR_SHARED));
}


//====================================================================
void __fastcall RunStandardMessageLoop()
{
	while(true) {
		MSG  msg;
		switch(DispatchOneMessage(msg)) {

			case DispatchResult_MessageNotAvailable: {
				::WaitMessage();
				break;
			}
			case DispatchResult_FoundQuitMessage: {
				return;
			}
		}
	}
}


//====================================================================
COMGuard::COMGuard()
{
	HRESULT hr = CoInitialize(0);
	bCleanup = SUCCEEDED(hr);
}


//====================================================================
COMGuard::~COMGuard()
{
	if(bCleanup) {
		CoUninitialize();
	}
}


//====================================================================
Guid::Guid()
{
	::ZeroMemory(this, sizeof(*this));
}


//====================================================================
Guid::Guid(DWORD _data1, WORD _data2, WORD _data3,
			BYTE _data40, BYTE _data41, BYTE _data42, BYTE _data43,
			BYTE _data44, BYTE _data45, BYTE _data46, BYTE _data47)
{
	Data1 = _data1;
	Data2 = _data2;
	Data3 = _data3;
	Data4[0] = _data40;
	Data4[1] = _data41;
	Data4[2] = _data42;
	Data4[3] = _data43;
	Data4[4] = _data44;
	Data4[5] = _data45;
	Data4[6] = _data46;
	Data4[7] = _data47;
}


//====================================================================
bool Guid::operator==(const Guid& rhs) const
{
	if(Data1 != rhs.Data1) return false;
	if(Data2 != rhs.Data2) return false;
	if(Data3 != rhs.Data3) return false;
	for(uint i = 0; i < 8; i++) {
		if(Data4[i] != rhs.Data4[i]) return false;
	}
	return true;
}


//====================================================================
void Guid::CopyFrom(const UID& uid)
{
	Data1 = uid[0];
	Data2 = ushort(uid[1] >> 16);
	Data3 = ushort(uid[1] & 0x0000FFFF);
	for(uint i = 0; i < 4; i++) {
		uint shift = 24 - i * 8;
		Data4[i]     = ubyte((uid[2] >> shift) & 0x000000FF);
		Data4[i + 4] = ubyte((uid[3] >> shift) & 0x000000FF);
	}
}


//====================================================================
void Guid::CopyTo(UID& uid) const
{
	uid[0] = Data1;
	uid[1] = (uint(Data2) << 16) | uint(Data3);
	uid[2] = (uint(Data4[0]) << 24) | (uint(Data4[1]) << 16) | (uint(Data4[2]) << 8) | uint(Data4[3]);
	uid[3] = (uint(Data4[4]) << 24) | (uint(Data4[5]) << 16) | (uint(Data4[6]) << 8) | uint(Data4[7]);
}


//********************************************************************
}//namespace win_utils


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const win_utils::Guid& x)
{
	os <= x.Data1 <= x.Data2 <= x.Data3;
	for(uint i = 0; i < 8; i++) os <= x.Data4[i];
	return os;
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, win_utils::Guid& x)
{
	is >= x.Data1 >= x.Data2 >= x.Data3;
	for(uint i = 0; i < 8; i++) is >= x.Data4[i];
	return is;
}


//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<<(std::ostream &os, const GUID& x)
{
	const char sep = '-';

	std::ostringstream oss;
	oss.setf(std::ios::hex, std::ios::basefield);
	oss.setf(std::ios::uppercase);
	oss.fill('0');

	oss << '{';
	oss << std::setw(8) << x.Data1 << sep;
	oss << std::setw(4) << x.Data2 << sep;
	oss << std::setw(4) << x.Data3 << sep;
	for(uint i = 0; i < 2; i++) {
		oss << std::setw(2) << uint(x.Data4[i]);
	}
	oss << sep;
	for(; i < 8; i++) {
		oss << std::setw(2) << uint(x.Data4[i]);
	}
	oss << '}';

	os << oss.str();
	return os;
}


//********************************************************************
}//namespace vertexware


