

//--------------------------------------------------------------------
#include "DataObjects\DataCodecManager.h"
#include "BaseTools\Debug.h"
#include "BaseTools\STLExtensions.h"
#include "BaseTools\FileSystem.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Memory.h"


//====================================================================
// Configure VC60 warnings.
//====================================================================
// 'identifier' : decorated name length exceeded, name was truncated
#pragma warning ( disable : 4503 )


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
class DataCodecNamePredicate
{
public:
	DataCodecNamePredicate(const std::string& _targetName) : targetName(_targetName) { }
	bool  operator()(const DataCodecBase* dcb) const { return dcb->GetCodecName() == targetName; }
private:
	std::string targetName;
};


//====================================================================
template<typename TCodecPtrContainer>
TCodecPtrContainer::value_type __fastcall
GetCodecByName(TCodecPtrContainer& c, const std::string& codecName)
{
	TCodecPtrContainer::iterator it =
		std::find_if(c.begin(), c.end(), DataCodecNamePredicate(codecName));
	if(it != c.end()) return *it;
	return 0;
}


//====================================================================
// Returns number of found codecs {0, 1, 2}.
//====================================================================
template<typename TCodecPtrContainer>
uint __fastcall
SwapCodecPositions(TCodecPtrContainer& c, const std::string& higherPriorityCodecName,
										const std::string& lowerPriorityCodecName)
{
	TCodecPtrContainer::iterator  itHigher = c.end();
	TCodecPtrContainer::iterator  itLower = c.end();
	TCodecPtrContainer::iterator* pFirstFound = 0;

	// search for both codecs, mark first found
	TCodecPtrContainer::iterator it = c.begin();
	while(it != c.end()) {
		TCodecPtrContainer::value_type pCodec = *it;

		if(pCodec->GetCodecName() == higherPriorityCodecName) {
			itHigher = it;
			if(! pFirstFound) pFirstFound = &itHigher;
		}
		else if(pCodec->GetCodecName() == lowerPriorityCodecName) {
			itLower = it;
			if(! pFirstFound) pFirstFound = &itLower;
		}
		++it;
	}

	// check number of found codecs
	uint numFound = 0;
	if(itHigher != c.end()) numFound++;
	if(itLower != c.end()) numFound++;

	// swap codec positions if lower-priority-codec is found first
	if(numFound == 2  &&  pFirstFound == &itLower) {
		std::iter_swap(itHigher, itLower);
	}
	return numFound;
}


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
VW_DEFINE_CONSTANT( uint, DataCategory_UserDefined, 0x10000000 );


//====================================================================
DataCodecBase::DataCodecBase(uint _dataCategory, const std::string& _codecName,
											const std::string& _formatList)
	:
	dataCategory(_dataCategory), codecName(_codecName), formatList(_formatList)
{
	VW_ASSERT(codecName.length(), "Codec-name is empty.");
}


//====================================================================
DataEncoder::DataEncoder(uint dataCategory, const std::string& encoderName,
											const std::string& formatList)
	: DataCodecBase(dataCategory, encoderName, formatList)
{
}


//====================================================================
DataDecoder::DataDecoder(uint dataCategory, const std::string& decoderName,
											const std::string& formatList)
	: DataCodecBase(dataCategory, decoderName, formatList)
{
}


//********************************************************************
namespace data_codec_parameters {


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_FileOverwrite, "FileOverwrite" );
VW_DEFINE_CONSTANT( bool, Value_FileOverwrite_Default, false );


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_ByteOrderBigEndian, "ByteOrderBigEndian" );
VW_DEFINE_CONSTANT( bool, Value_ByteOrderBigEndian_Default, false );


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_UseCompression, "UseCompression" );
VW_DEFINE_CONSTANT( bool, Value_UseCompression_Default, true );


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_CompressionLevel, "CompressionLevel" );
VW_DEFINE_CONSTANT( uint, Value_CompressionLevel_Min, 0 );
VW_DEFINE_CONSTANT( uint, Value_CompressionLevel_Max, 100 );
VW_DEFINE_CONSTANT( uint, Value_CompressionLevel_Default, Value_CompressionLevel_Max );


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_Quality, "Quality" );
VW_DEFINE_CONSTANT( uint, Value_Quality_Min, 0 );
VW_DEFINE_CONSTANT( uint, Value_Quality_Max, 100 );
VW_DEFINE_CONSTANT( uint, Value_Quality_Default, Value_Quality_Max );


//********************************************************************
}//namespace data_codec_parameters


//====================================================================
VW_IMPLEMENT_SINGLETON( DataCodecManager );


//====================================================================
DataCodecManager::DataCodecManager()
{
	VW_IMPLEMENT_SINGLETON_CTOR( DataCodecManager );

	ResetEncoderInitialParamList();
	ResetDecoderInitialParamList();
}


//====================================================================
DataCodecManager::~DataCodecManager()
{
	UnregisterAllEncoders();
	UnregisterAllDecoders();

	VW_IMPLEMENT_SINGLETON_DTOR( DataCodecManager );
}


//====================================================================
bool  DataCodecManager::SaveData(uint dataCategory, const void* data,
						const std::string& fileName, const std::string& fileFormat,
						const NamedParameterList* pParameterList)
{
	// check arguments
	if(! data) return false;
	if(fileName.empty()) return false;
	if(fileFormat.empty()) return false;

	// iterate through all encoders in category until data is saved successfully
	ENC_VEC* encoders = FindEncoders(dataCategory);
	if(! encoders) {
		VW_THROW(InvalidUsageException, "Saving unregistered data-category data.");
	}

	const NamedParameterList& npl = pParameterList ? *pParameterList :
									GetEncoderDefaultParamList(dataCategory);

	for(uint i = 0; i < encoders->size(); i++) {
		if((*encoders)[i]->SaveData(data, fileName, fileFormat, npl)) {
			return true;
		}
	}
	// none of the registered encoders managed to save the file
	return false;
}


//====================================================================
void*  DataCodecManager::LoadData(uint dataCategory, const std::string& fileName,
						const NamedParameterList* pParameterList)
{
	// check arguments
	if(fileName.empty()) return false;
	if(! file_system::FileExists(file_system::GetFullPathName(fileName))) return 0;

	// iterate through all decoders until data is loaded successfully
	DEC_VEC* decoders = FindDecoders(dataCategory);
	if(! decoders) {
		VW_THROW(InvalidUsageException, "Loading unregistered data-category data.");
	}

	const NamedParameterList& npl = pParameterList ? *pParameterList :
									GetDecoderDefaultParamList(dataCategory);

	for(uint i = 0; i < decoders->size(); i++) {
		void* pData = (*decoders)[i]->LoadData(fileName, npl);
		if(pData) return pData;
	}
	// none of the registered decoders managed to load the file
	return 0;
}


//====================================================================
void  DataCodecManager::RegisterDataCategory(uint dataCategory)
{
	if(IsDataCategoryRegistered(dataCategory)) return;

	// reserve storage for new category codecs
	catEncoderMap[dataCategory] = ENC_VEC();
	catDecoderMap[dataCategory] = DEC_VEC();

	SetEncoderDefaultParamList(dataCategory, encInitialParamList);
	SetDecoderDefaultParamList(dataCategory, decInitialParamList);
}


//====================================================================
bool  DataCodecManager::IsDataCategoryRegistered(uint dataCategory)
{
	return (FindEncoders(dataCategory) != 0);
}


//====================================================================
void  DataCodecManager::RegisterEncoder(DataEncoder* pEncoder)
{
	// check arguments
	VW_ASSERT(pEncoder, "Invalid argument.");

	if(! pEncoder->IsOnHeap()) {
		VW_THROW(InvalidUsageException, "Encoder is not created on heap: " + pEncoder->GetCodecName());
	}
	if(! IsDataCategoryRegistered(pEncoder->GetDataCategory())) {
		VW_THROW(InvalidUsageException, "Encoder category is not registered: " + pEncoder->GetCodecName());
	}
	if(IsEncoderRegistered(pEncoder->GetDataCategory(), pEncoder->GetCodecName())) {
		VW_THROW(InvalidUsageException, "Encoder is already registered: " + pEncoder->GetCodecName());
	}

	// store encoder
	ENC_VEC* encoders = FindEncoders(pEncoder->GetDataCategory());
	VW_ASSERT(encoders, "Internal error.");
	encoders->push_back(pEncoder);
}


//====================================================================
void  DataCodecManager::RegisterDecoder(DataDecoder* pDecoder)
{
	// check arguments
	VW_ASSERT(pDecoder, "Invalid argument.");

	if(! pDecoder->IsOnHeap()) {
		VW_THROW(InvalidUsageException, "Decoder is not created on heap: " + pDecoder->GetCodecName());
	}
	if(! IsDataCategoryRegistered(pDecoder->GetDataCategory())) {
		VW_THROW(InvalidUsageException, "Decoder category is not registered: " + pDecoder->GetCodecName());
	}
	if(IsDecoderRegistered(pDecoder->GetDataCategory(), pDecoder->GetCodecName())) {
		VW_THROW(InvalidUsageException, "Decoder is already registered: " + pDecoder->GetCodecName());
	}

	// store decoder
	DEC_VEC* decoders = FindDecoders(pDecoder->GetDataCategory());
	VW_ASSERT(decoders, "Internal error.");
	decoders->push_back(pDecoder);
}


//====================================================================
bool  DataCodecManager::IsEncoderRegistered(uint dataCategory, const std::string& encoderName)
{
	return (FindEncoder(dataCategory, encoderName) != 0);
}


//====================================================================
bool  DataCodecManager::IsDecoderRegistered(uint dataCategory, const std::string& decoderName)
{
	return (FindDecoder(dataCategory, decoderName) != 0);
}


//====================================================================
void  DataCodecManager::PrioritizeEncoders(uint dataCategory,
							const std::string& higherPriorityEncoderName,
							const std::string& lowerPriorityEncoderName)
{
	ENC_VEC* encoders = FindEncoders(dataCategory);
	if(! encoders) return;
	SwapCodecPositions(*encoders, higherPriorityEncoderName, lowerPriorityEncoderName);
}


//====================================================================
void  DataCodecManager::PrioritizeDecoders(uint dataCategory,
							const std::string& higherPriorityDecoderName,
							const std::string& lowerPriorityDecoderName)
{
	DEC_VEC* decoders = FindDecoders(dataCategory);
	if(! decoders) return;
	SwapCodecPositions(*decoders, higherPriorityDecoderName, lowerPriorityDecoderName);
}


//====================================================================
void  DataCodecManager::UnregisterEncoder(uint dataCategory, const std::string& encoderName)
{
	ENC_VEC* encoders = FindEncoders(dataCategory);
	if(! encoders) return;
	DataCodecNamePredicate  dcnp(encoderName);
	DeleteFromPtrContainer_If(*encoders, dcnp);
}


//====================================================================
void  DataCodecManager::UnregisterDecoder(uint dataCategory, const std::string& decoderName)
{
	DEC_VEC* decoders = FindDecoders(dataCategory);
	if(! decoders) return;
	DataCodecNamePredicate  dcnp(decoderName);
	DeleteFromPtrContainer_If(*decoders, dcnp);
}


//====================================================================
void  DataCodecManager::UnregisterAllEncoders(uint dataCategory)
{
	ENC_VEC* encoders = FindEncoders(dataCategory);
	if(! encoders) return;
	DeletePtrContainer(*encoders);
}


//====================================================================
void  DataCodecManager::UnregisterAllDecoders(uint dataCategory)
{
	DEC_VEC* decoders = FindDecoders(dataCategory);
	if(! decoders) return;
	DeletePtrContainer(*decoders);
}


//====================================================================
void  DataCodecManager::UnregisterAllEncoders()
{
	CAT_ENC_ITER it = catEncoderMap.begin();
	while(it != catEncoderMap.end()) {
		DeletePtrContainer(it->second);
		++it;
	}
}


//====================================================================
void  DataCodecManager::UnregisterAllDecoders()
{
	CAT_DEC_ITER it = catDecoderMap.begin();
	while(it != catDecoderMap.end()) {
		DeletePtrContainer(it->second);
		++it;
	}
}


//====================================================================
uint  DataCodecManager::GetNumEncoders(uint dataCategory)
{
	ENC_VEC* encoders = FindEncoders(dataCategory);
	return (encoders ? encoders->size() : 0);
}


//====================================================================
const DataEncoder&  DataCodecManager::GetEncoder(uint dataCategory, uint i)
{
	ENC_VEC* encoders = FindEncoders(dataCategory);
	VW_ASSERT(encoders, "Unregistered data-category.");
	VW_ASSERT(i < encoders->size(), "Index out of range.");
	return *(*encoders)[i];
}


//====================================================================
uint  DataCodecManager::GetNumDecoders(uint dataCategory)
{
	DEC_VEC* decoders = FindDecoders(dataCategory);
	return (decoders ? decoders->size() : 0);
}


//====================================================================
const DataDecoder&  DataCodecManager::GetDecoder(uint dataCategory, uint i)
{
	DEC_VEC* decoders = FindDecoders(dataCategory);
	VW_ASSERT(decoders, "Unregistered data-category.");
	VW_ASSERT(i < decoders->size(), "Index out of range.");
	return *(*decoders)[i];
}


//====================================================================
void  DataCodecManager::DumpRegisteredEncoders(uint dataCategory,
												std::ostream& os, uint tabLevel)
{
	TabEffector te(tabLevel);
	uint count = GetNumEncoders(dataCategory);
	for(uint i = 0; i < count; i++) {
		const DataEncoder& de = GetEncoder(dataCategory, i);
		os << te << de.GetCodecName() << ": " << de.GetFormatList() << nl;
	}
}


//====================================================================
void  DataCodecManager::DumpRegisteredDecoders(uint dataCategory,
												std::ostream& os, uint tabLevel)
{
	TabEffector te(tabLevel);
	uint count = GetNumDecoders(dataCategory);
	for(uint i = 0; i < count; i++) {
		const DataDecoder& dd = GetDecoder(dataCategory, i);
		os << te << dd.GetCodecName() << ": " << dd.GetFormatList() << nl;
	}
}


//====================================================================
void  DataCodecManager::SetDefaultFileFormatExtension(uint dataCategory,
								const std::string& fileFormat, const std::string& fileExtension)
{
	CAT_FMT_EXT_ITER itCat = catFmtExtMap.find(dataCategory);

	if(itCat != catFmtExtMap.end()) {
		if(fileExtension.empty()) {

			itCat->second.erase(fileFormat);
			if(itCat->second.empty()) {
				catFmtExtMap.erase(dataCategory);
			}
		}
		else {
			itCat->second[fileFormat] = fileExtension;
		}
	}
	else {
		if(fileExtension.length()) {
			catFmtExtMap[dataCategory][fileFormat] = fileExtension;
		}
	}
}


//====================================================================
bool  DataCodecManager::GetDefaultFileFormatExtension(uint dataCategory,
								const std::string& fileFormat, std::string& fileExtension,
								const std::string& failValueExtension)
{
	fileExtension = failValueExtension;

	CAT_FMT_EXT_ITER itCat = catFmtExtMap.find(dataCategory);
	if(itCat != catFmtExtMap.end()) {

		FMT_EXT_ITER itFmt = itCat->second.find(fileFormat);
		if(itFmt != itCat->second.end()) {

			fileExtension = itFmt->second;
			return true;
		}
	}
	return false;
}


//====================================================================
void  DataCodecManager::ClearDefaultFileFormatExtensions(uint dataCategory)
{
	catFmtExtMap.erase(dataCategory);
}


//====================================================================
void  DataCodecManager::ClearDefaultFileFormatExtensions()
{
	catFmtExtMap.clear();
}


//====================================================================
void  DataCodecManager::SetEncoderDefaultParamList(uint dataCategory,
													const NamedParameterList& npl)
{
	if(! IsDataCategoryRegistered(dataCategory)) return;
	encCatDefParamListMap[dataCategory] = npl;
}


//====================================================================
void  DataCodecManager::SetDecoderDefaultParamList(uint dataCategory,
													const NamedParameterList& npl)
{
	if(! IsDataCategoryRegistered(dataCategory)) return;
	decCatDefParamListMap[dataCategory] = npl;
}


//====================================================================
void  DataCodecManager::SetEncoderDefaultParamList(uint dataCategory,
													const CodecParameterFactory& cpf)
{
	NamedParameterList  npl = encInitialParamList;
	cpf.FillParameterList(dataCategory, true, npl);
	SetEncoderDefaultParamList(dataCategory, npl);
}


//====================================================================
void  DataCodecManager::SetDecoderDefaultParamList(uint dataCategory,
													const CodecParameterFactory& cpf)
{
	NamedParameterList  npl = decInitialParamList;
	cpf.FillParameterList(dataCategory, false, npl);
	SetDecoderDefaultParamList(dataCategory, npl);
}


//====================================================================
const NamedParameterList&  DataCodecManager::GetEncoderDefaultParamList(uint dataCategory) const
{
	CAT_NPL_CITER it = encCatDefParamListMap.find(dataCategory);
	if(it != encCatDefParamListMap.end()) return (it->second);
	return encInitialParamList;
}


//====================================================================
const NamedParameterList&  DataCodecManager::GetDecoderDefaultParamList(uint dataCategory) const
{
	CAT_NPL_CITER it = decCatDefParamListMap.find(dataCategory);
	if(it != decCatDefParamListMap.end()) return (it->second);
	return decInitialParamList;
}


//====================================================================
void  DataCodecManager::ResetEncoderInitialParamList()
{
	using namespace data_codec_parameters;

	encInitialParamList.Clear();

	encInitialParamList.Insert(NamedParameter(Param_FileOverwrite, Value_FileOverwrite_Default));
	encInitialParamList.Insert(NamedParameter(Param_ByteOrderBigEndian, Value_ByteOrderBigEndian_Default));
	encInitialParamList.Insert(NamedParameter(Param_UseCompression, Value_UseCompression_Default));
	encInitialParamList.Insert(NamedParameter(Param_CompressionLevel, Value_CompressionLevel_Default));
	encInitialParamList.Insert(NamedParameter(Param_Quality, Value_Quality_Default));
}


//====================================================================
void  DataCodecManager::ResetDecoderInitialParamList()
{
	using namespace data_codec_parameters;

	decInitialParamList.Clear();
}


//====================================================================
void  DataCodecManager::SetEncoderInitialParamList(const NamedParameterList& npl)
{
	encInitialParamList = npl;
}


//====================================================================
void  DataCodecManager::SetDecoderInitialParamList(const NamedParameterList& npl)
{
	decInitialParamList = npl;
}


//====================================================================
const NamedParameterList&  DataCodecManager::GetEncoderInitialParamList() const
{
	return encInitialParamList;
}


//====================================================================
const NamedParameterList&  DataCodecManager::GetDecoderInitialParamList() const
{
	return decInitialParamList;
}


//====================================================================
DataEncoder*  DataCodecManager::FindEncoder(uint dataCategory, const std::string& encoderName)
{
	ENC_VEC* encoders = FindEncoders(dataCategory);
	if(! encoders) return 0;
	return GetCodecByName(*encoders, encoderName);
}


//====================================================================
DataDecoder*  DataCodecManager::FindDecoder(uint dataCategory, const std::string& decoderName)
{
	DEC_VEC* decoders = FindDecoders(dataCategory);
	if(! decoders) return 0;
	return GetCodecByName(*decoders, decoderName);
}


//====================================================================
DataCodecManager::ENC_VEC*  DataCodecManager::FindEncoders(uint dataCategory)
{
	CAT_ENC_ITER it = catEncoderMap.find(dataCategory);
	if(it != catEncoderMap.end()) return &(it->second);
	return 0;
}


//====================================================================
DataCodecManager::DEC_VEC*  DataCodecManager::FindDecoders(uint dataCategory)
{
	CAT_DEC_ITER it = catDecoderMap.find(dataCategory);
	if(it != catDecoderMap.end()) return &(it->second);
	return 0;
}


//********************************************************************
}//namespace vertexware


