

//--------------------------------------------------------------------
#include "DataObjects\DataPackage.h"
#include "BaseTools\Serializers.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Debug.h"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
const char*  gc_DPHMagicSequenceOpen     = "<VertexWareDataPackage//";
const char*  gc_DPHMagicSequenceClose    = "///>";
const uint   gc_DPHByteOrderBigEndian    = 0xFFFFFFFF;
const uint   gc_DPHByteOrderLittleEndian = 0x00000000;
const char*  gc_DPTMagicSequenceOpen     = "<///";
const char*  gc_DPTMagicSequenceClose    = "//VertexWareDataPackage>";


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
DataPackageHeader::DataPackageHeader()
{
	bPackageByteOrderBigEndian = serializer_parameters::IsMachineBigEndian();
	packageID = 0;
	payloadByteSize = 0;
	payloadType = 0;
}


//====================================================================
void  DataPackageHeader::Serialize(std::ostream& os) const
{
	serializer_parameters::OutputStreamByteOrderSentry  bos(bPackageByteOrderBigEndian);

	WriteByteBlock(os, gc_DPHMagicSequenceOpen, 24);
	os <= (bPackageByteOrderBigEndian ? gc_DPHByteOrderBigEndian : gc_DPHByteOrderLittleEndian);
	os <= packageID;
	os <= payloadByteSize;
	os <= payloadType;
	WriteByteBlock(os, gc_DPHMagicSequenceClose, 4);
}


//====================================================================
void  DataPackageHeader::Deserialize(std::istream& is)
{
	char  buf[24];

	// opening magic sequence
	ReadByteBlock(is, buf, 24);
	if(strncmp(buf, gc_DPHMagicSequenceOpen, 24)) {
		VW_THROW_CN(DeserializationException, "Invalid opening magic-sequence.");
	}

	// package byte-order
	uint  byteOrderFlags;
	is >= byteOrderFlags;
	if(byteOrderFlags == gc_DPHByteOrderBigEndian) {
		bPackageByteOrderBigEndian = true;
	}
	else if(byteOrderFlags == gc_DPHByteOrderLittleEndian) {
		bPackageByteOrderBigEndian = false;
	}
	else {
		VW_THROW_CN(DeserializationException, "Invalid byte-order specifier.");
	}

	serializer_parameters::InputStreamByteOrderSentry  bos(bPackageByteOrderBigEndian);

	// package & payload properties
	is >= packageID;
	is >= payloadByteSize;
	is >= payloadType;

	// closing magic sequence
	ReadByteBlock(is, buf, 4);
	if(strncmp(buf, gc_DPHMagicSequenceClose, 4)) {
		VW_THROW_CN(DeserializationException, "Invalid closing magic-sequence.");
	}
}


//====================================================================
DataPackageTrailer::DataPackageTrailer()
{
	bPackageByteOrderBigEndian = serializer_parameters::IsMachineBigEndian();
	packageID = 0;
}


//====================================================================
void  DataPackageTrailer::Serialize(std::ostream& os) const
{
	serializer_parameters::OutputStreamByteOrderSentry  bos(bPackageByteOrderBigEndian);

	WriteByteBlock(os, gc_DPTMagicSequenceOpen, 4);
	os <= packageID;
	WriteByteBlock(os, gc_DPTMagicSequenceClose, 24);
}


//====================================================================
void  DataPackageTrailer::Deserialize(std::istream& is)
{
	char  buf[24];

	// opening magic sequence
	ReadByteBlock(is, buf, 4);
	if(strncmp(buf, gc_DPTMagicSequenceOpen, 4)) {
		VW_THROW_CN(DeserializationException, "Invalid opening magic-sequence.");
	}

	serializer_parameters::InputStreamByteOrderSentry  bos(bPackageByteOrderBigEndian);

	// package identifier
	is >= packageID;

	// closing magic sequence
	ReadByteBlock(is, buf, 24);
	if(strncmp(buf, gc_DPTMagicSequenceClose, 24)) {
		VW_THROW_CN(DeserializationException, "Invalid closing magic-sequence.");
	}
}


//====================================================================
DataPackage::DataPackage(uint _payloadType)
{
	bPackageByteOrderBigEndianOnWrite = serializer_parameters::IsMachineBigEndian();
	packageID = 0;
	payloadType = _payloadType;

	bSkipHeaderOnRead = false;
	pExternalDPH = 0;
}


//====================================================================
void  DataPackage::SetSkipHeaderOnRead(bool bSkip, const DataPackageHeader* dph)
{
	VW_ASSERT((bSkip  &&  dph)  ||  (! bSkip  &&  ! dph), "Invalid argument.");

	bSkipHeaderOnRead = bSkip;
	pExternalDPH = dph;
}


//====================================================================
bool  DataPackage::SaveToFile(const std::string& fileName, bool bThrowExceptionsOnError)
{
	try {
		std::ofstream os(fileName.c_str(), std::ios::out | std::ios::trunc | std::ios::binary);
		if(! os) {
			VW_THROW(SerializationException, "Cannot open file: " + fileName);
		}
		os <= *this;
	}
	catch(std::ios_base::failure&) {
		if(bThrowExceptionsOnError) throw;
		return false;
	}
	catch(Exception&) {
		if(bThrowExceptionsOnError) throw;
		return false;
	}
	return true;
}


//====================================================================
bool  DataPackage::LoadFromFile(const std::string& fileName, bool bThrowExceptionsOnError)
{
	try {
		std::ifstream is(fileName.c_str(), std::ios::in | std::ios::binary);
		if(! is) {
			VW_THROW(DeserializationException, "Cannot open file: " + fileName);
		}
		is >= *this;
	}
	catch(std::ios_base::failure&) {
		if(bThrowExceptionsOnError) throw;
		return false;
	}
	catch(Exception&) {
		if(bThrowExceptionsOnError) throw;
		return false;
	}
	return true;
}


//====================================================================
void  DataPackage::Serialize(std::ostream& os) const
{
	StreamExceptionSentry  ses(os);
	serializer_parameters::OutputStreamByteOrderSentry  bos(bPackageByteOrderBigEndianOnWrite);

	// collect header data
	uint  payloadByteSize = 0;
	QueryPayloadByteSize(payloadByteSize);

	// write header
	DataPackageHeader  dph;
	dph.SetPackageByteOrder(bPackageByteOrderBigEndianOnWrite);
	dph.SetPackageID(packageID);
	dph.SetPayloadByteSize(payloadByteSize);
	dph.SetPayloadType(payloadType);
	os <= dph;

	// mark payload's starting position
	uint  payloadStartPos = os.tellp();

	// write payload
	SerializePayload(os);

	// validate payload byte-size
	if(uint(os.tellp()) - payloadStartPos != payloadByteSize) {
		VW_THROW_CN(SerializationException, "Payload byte-size mismatch.");
	}

	// write trailer
	DataPackageTrailer  dpt;
	dpt.SetPackageByteOrder(bPackageByteOrderBigEndianOnWrite);
	dpt.SetPackageID(packageID);
	os <= dpt;
}


//====================================================================
void  DataPackage::Deserialize(std::istream& is)
{
	StreamExceptionSentry  ses(is);

	// read header
	DataPackageHeader  dph;

	if(bSkipHeaderOnRead) {
		dph = *pExternalDPH;
	}
	else {
		is >= dph;
	}
	serializer_parameters::InputStreamByteOrderSentry  bos(dph.IsPackageByteOrderBigEndian());

	// check payload-type
	if(payloadType != dph.GetPayloadType()) {
		VW_THROW_CN(DeserializationException, "Payload-type mismatch.");
	}

	// make package-id available to derived class
	packageID = dph.GetPackageID();

	// mark payload's starting position
	uint  payloadStartPos = is.tellg();

	// read payload
	DeserializePayload(is, dph.GetPayloadByteSize());

	// validate payload byte-size
	if(uint(is.tellg()) - payloadStartPos != dph.GetPayloadByteSize()) {
		VW_THROW_CN(DeserializationException, "Payload byte-size mismatch.");
	}

	// read trailer
	DataPackageTrailer  dpt;
	dpt.SetPackageByteOrder(dph.IsPackageByteOrderBigEndian());
	is >= dpt;

	// check package-id match
	if(packageID != dpt.GetPackageID()) {
		VW_THROW_CN(DeserializationException, "Package-id mismatch.");
	}
}


//********************************************************************
}//namespace vertexware


