

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Serializable.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
// Data-package-header specification:
//        |                      |
// Bytes  |     Contents         |        Details
//        |                      |
//    24  |  magic sequence      |  "<VertexWareDataPackage//"
//     4  |  package byte-order  |  0x00000000 = little-endian, 0xFFFFFFFF = big-endian
//     4  |  package identifier  |  uint
//     4  |  payload byte-size   |  uint
//     4  |  payload type        |  uint
//     4  |  magic sequence      |  "///>"
//        |                      |
//    44 bytes total
// User-defined payload-types begin at 0x80000000
//====================================================================
class VW_COMMON_API DataPackageHeader : public Serializable
{
private:
	bool  bPackageByteOrderBigEndian;
	uint  packageID;
	uint  payloadByteSize;
	uint  payloadType;


public:
	// default information:
	// byte-order = machine-native, package-id = 0, byte-size = 0, type = 0
	DataPackageHeader();

	void  SetPackageByteOrder(bool bBigEndian) { bPackageByteOrderBigEndian = bBigEndian; }
	bool  IsPackageByteOrderBigEndian() const { return bPackageByteOrderBigEndian; }

	void  SetPackageID(uint id) { packageID = id; }
	uint  GetPackageID() const { return packageID; }

	void  SetPayloadByteSize(uint byteSize) { payloadByteSize = byteSize; }
	uint  GetPayloadByteSize() const { return payloadByteSize; }

	void  SetPayloadType(uint type) { payloadType = type; }
	uint  GetPayloadType() const { return payloadType; }

	// serializable implementation
	virtual  void  Serialize(std::ostream& os) const;
	virtual  void  Deserialize(std::istream& is);
};


//====================================================================
// Data-package-trailer specification:
//        |                      |
// Bytes  |     Contents         |        Details
//        |                      |
//     4  |  magic sequence      |  "<///"
//     4  |  package identifier  |  uint
//    24  |  magic sequence      |  "//VertexWareDataPackage>"
//        |                      |
//    32 bytes total
//====================================================================
class VW_COMMON_API DataPackageTrailer : public Serializable
{
private:
	bool  bPackageByteOrderBigEndian;
	uint  packageID;


public:
	// default information: byte-order = machine-native, package-id = 0
	DataPackageTrailer();

	void  SetPackageByteOrder(bool bBigEndian) { bPackageByteOrderBigEndian = bBigEndian; }
	bool  IsPackageByteOrderBigEndian() const { return bPackageByteOrderBigEndian; }

	void  SetPackageID(uint id) { packageID = id; }
	uint  GetPackageID() const { return packageID; }

	// serializable implementation
	virtual  void  Serialize(std::ostream& os) const;
	virtual  void  Deserialize(std::istream& is);
};


//====================================================================
// All DataPackages throw exceptions when using direct (de)serialization.
//====================================================================
class VW_COMMON_API DataPackage : public Serializable
{
public:
	virtual ~DataPackage() {}

	void  SetPackageByteOrderOnWrite(bool bBigEndian) { bPackageByteOrderBigEndianOnWrite = bBigEndian; }

	void  SetPackageID(uint id) { packageID = id; }
	uint  GetPackageID() const { return packageID; }

	// If <bSkip> is false <dph> must be 0, otherwise <dph> must point to
	// valid DataPackageHeader for current package. Header-skipping allows
	// client to read the header manually and if it is recognized,
	// deserialize the package without rewinding the stream back to header-start.
	void  SetSkipHeaderOnRead(bool bSkip = false, const DataPackageHeader* dph = 0);

	// save / load
	bool  SaveToFile(const std::string& fileName, bool bThrowExceptionsOnError = false);
	bool  LoadFromFile(const std::string& fileName, bool bThrowExceptionsOnError = false);

	// serializable implementation
	virtual  void  Serialize(std::ostream& os) const;
	virtual  void  Deserialize(std::istream& is);


private:
	bool  bPackageByteOrderBigEndianOnWrite;
	uint  packageID;
	uint  payloadType;

	bool                      bSkipHeaderOnRead;
	const DataPackageHeader*  pExternalDPH;


protected:
	// default payload writing parameters:
	// byte-order = machine-native, skip-header = false, package-id = 0
	DataPackage(uint payloadType);

	// Throw exceptions on any kind of error. Correct byte-order is activated by base-class
	// before payload (de)serialization. Stream-exception-sentry is also activated
	// for all std::ios errors.
	virtual  void  QueryPayloadByteSize(uint& payloadByteSize) const = 0;
	virtual  void  SerializePayload(std::ostream& os) const = 0;
	virtual  void  DeserializePayload(std::istream& is, uint payloadByteSize) = 0;
};


//********************************************************************
}//namespace vertexware


