

//--------------------------------------------------------------------
#include "DataObjects\ImageOperations.h"
#include "DataObjects\SignalOperations.h"
#include "DataObjects\SampleProcessors.h"
#include "MathTools\MathBase.h"
#include "BaseTools\ElementProcessors.h"
#include "BaseTools\String.h"
#include "BaseTools\WinUtils.h"
#include "BaseTools\Memory.h"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
template<typename TElementProcessor>
__forceinline
void __fastcall Call_ResampleElements2D_Nearest_ubyte_ubyte(
									TElementProcessor& copyProcessor,
									ImageBase& dst, const ImageBase& src)
{
	ResampleElements2D_Nearest<ubyte, ubyte>(
			copyProcessor,
			dst.GetSamplesB(), dst.GetRowByteSize(), dst.GetSampleByteSize(),
			dst.GetNumRows(), dst.GetNumColumns(),
			src.GetSamplesB(), src.GetRowByteSize(), src.GetSampleByteSize(),
			src.GetNumRows(), src.GetNumColumns());
}


//====================================================================
template<typename TLinearInterpolator2D>
__forceinline
void __fastcall Call_ResampleElements2D_Linear_PackedSample_UShort(
									TLinearInterpolator2D& interpolator,
									ImageBase& dst, const ImageBase& src)
{
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.GetSampleByteSize() == 2, "Invalid sample-format size.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");

	ResampleElements2D_Linear<ubyte, ubyte>(
			interpolator,
			dst.GetSamplesB(), dst.GetRowByteSize(), dst.GetSampleByteSize(),
			dst.GetNumRows(), dst.GetNumColumns(),
			src.GetSamplesB(), src.GetRowByteSize(), src.GetSampleByteSize(),
			src.GetNumRows(), src.GetNumColumns());
}


//====================================================================
class CheckerboardProcessor : public ElementProcessorInplace<ubyte>
{
private:
	uint   numValues;

	const ubyte*  values1;
	const ubyte*  values2;

	uint   vertBlockSize;
	uint   horzBlockSize;

	const ubyte*  vValues;
	const ubyte*  hValues;

	uint   vCount;
	uint   hCount;


public:
	CheckerboardProcessor(uint _numValues, const ubyte* _values1, const ubyte* _values2,
						uint _vertBlockSize, uint _horzBlockSize)
	{
		numValues = _numValues;
		values1 = _values1;
		values2 = _values2;
		vertBlockSize = _vertBlockSize;
		horzBlockSize = _horzBlockSize;
	}

	__forceinline  void  PreLoop(uint loopDepth) {
		if(loopDepth == 0) {
			vValues = values1;
			vCount = 0;
		}
		hValues = vValues;
		hCount = 0;
	}

	__forceinline  void  Process(ubyte* elem) {
		for(uint i = 0; i < numValues; i++) {
			elem[i] = hValues[i];
		}
		hCount++;
		if(hCount >= horzBlockSize) {
			hCount = 0;
			ChangeActiveValue(hValues, values1, values2);
		}
	}

	__forceinline  void  PostLoop(uint loopDepth) {
		vCount++;
		if(vCount >= vertBlockSize) {
			vCount = 0;
			ChangeActiveValue(vValues, values1, values2);
		}
	}
};


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
void __fastcall GdiplusFillBitmapData(Gdiplus::BitmapData& bitmapData,
						Gdiplus::PixelFormat pixelFormat, const ImageBase& image)
{
	bitmapData.Width  = image.GetNumColumns();
	bitmapData.Height = image.GetNumRows();
	bitmapData.Stride = image.GetRowByteSize();
	bitmapData.Scan0  = const_cast<ubyte*>(image.GetSamplesB());
	bitmapData.PixelFormat = pixelFormat;
}


//====================================================================
bool __fastcall GdiplusTransferBitmapPixels(Gdiplus::Bitmap& bitmap,
							Gdiplus::BitmapData& bitmapData, bool bToUserBuffer)
{
	if(! bitmapData.Width  ||  ! bitmapData.Height) return false;

	Gdiplus::Rect lockRect(0, 0, bitmapData.Width, bitmapData.Height);

	UINT  lockFlags = bToUserBuffer ? Gdiplus::ImageLockModeRead : Gdiplus::ImageLockModeWrite;
	lockFlags |= Gdiplus::ImageLockModeUserInputBuf;

	if(bitmap.LockBits(&lockRect, lockFlags, bitmapData.PixelFormat, &bitmapData) == Gdiplus::Ok) {
		bitmap.UnlockBits(&bitmapData);
		return true;
	}
	return false;
}


//====================================================================
ImageBase* __fastcall ManufactureImage(eSampleFormat sampleFormat,
									uint numRows, uint numColumns)
{
	switch(sample_formats::ExtractByteSize(sampleFormat))
	{
		case 1: return new Image8(sampleFormat, numRows, numColumns);
		case 2: return new Image16(sampleFormat, numRows, numColumns);
		case 3: return new Image24(sampleFormat, numRows, numColumns);
		case 4: return new Image32(sampleFormat, numRows, numColumns);
		case 6: return new Image48(sampleFormat, numRows, numColumns);
		case 8: return new Image64(sampleFormat, numRows, numColumns);
		case 12: return new Image96(sampleFormat, numRows, numColumns);
		case 16: return new Image128(sampleFormat, numRows, numColumns);
	}
	return 0;
}


//====================================================================
ImageBase* __fastcall CloneImage(const ImageBase& src)
{
	ImageBase* dst = ManufactureImage(src.GetSampleFormat(),
						src.GetNumRows(), src.GetNumColumns());
	if(! dst) return 0;
	CopySamples2D(*dst, src);
	return dst;
}


//====================================================================
void __fastcall CopySamples2D(ImageBase& dst, const ImageBase& src,
							uint rowCount, uint columnCount,
							uint dstRowOffset, uint dstColumnOffset,
							uint srcRowOffset, uint srcColumnOffset,
							bool bFlipRows)
{
	if(! rowCount) rowCount = src.GetNumRows();
	if(! columnCount) columnCount = src.GetNumColumns();

	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dstRowOffset + rowCount <= dst.GetNumRows(), "Destination-row-offset/row-count out of range.");
	VW_ASSERT(dstColumnOffset + columnCount <= dst.GetNumColumns(), "Destination-column-offset/column-count out of range.");
	VW_ASSERT(srcRowOffset + rowCount <= src.GetNumRows(), "Source-row-offset/row-count out of range.");
	VW_ASSERT(srcColumnOffset + columnCount <= src.GetNumColumns(), "Source-column-offset/column-count out of range.");

	// burst copy
	if(! bFlipRows && columnCount == dst.GetNumColumns() && columnCount == src.GetNumColumns()) {
		::CopyMemory(dst.SampleB(dstRowOffset, 0),
					src.SampleB(srcRowOffset, 0),
					src.GetRowByteSize() * rowCount);
	}
	// loop copy
	else {
		int   srcRow         = bFlipRows ? (srcRowOffset + rowCount - 1) : srcRowOffset;
		int   srcRowInc      = bFlipRows ? -1 : 1;
		uint  dstRow         = dstRowOffset;
		uint  dstRowEnd      = dstRow + rowCount;
		uint  numBytesPerRow = src.GetSampleByteSize() * columnCount;

		while(dstRow < dstRowEnd) {
			::CopyMemory(dst.SampleB(dstRow, dstColumnOffset),
						src.SampleB(srcRow, srcColumnOffset),
						numBytesPerRow);

			srcRow += srcRowInc;
			dstRow++;
		}
	}
}


//====================================================================
bool __fastcall TransferSamples2D(ImageBase& dst, const ImageBase& src, bool bAutoResize)
{
	SignalOperationsManager::TransferSamplesHandler* tsh =
		SignalOperationsManager::Instance().GetTransferHandler(dst.GetSampleFormat(), src.GetSampleFormat());

	if(! tsh) return false; //not supported

	if(bAutoResize) {
		dst.ReserveSampleStorage(src.GetNumRows(), src.GetNumColumns());
	}
	VW_ASSERT(dst.CheckDimensionMatch(src), "Dimensions mismatch.");

	TransferSamples1D(dst, src, *tsh, false);
	return true;
}


//====================================================================
void __fastcall Resample2D_Nearest(ImageBase& dst, const ImageBase& src)
{
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");

	switch(dst.GetSampleByteSize()) {
		case 1:
			Call_ResampleElements2D_Nearest_ubyte_ubyte(
				element_processors::Copy_1<ubyte>(), dst, src);
			break;

		case 2:
			Call_ResampleElements2D_Nearest_ubyte_ubyte(
				element_processors::Copy_1<ushort>(), dst, src);
			break;

		case 4:
			Call_ResampleElements2D_Nearest_ubyte_ubyte(
				element_processors::Copy_1<uint>(), dst, src);
			break;

		default:
			Call_ResampleElements2D_Nearest_ubyte_ubyte(
				element_processors::Copy_N<ubyte>(dst.GetSampleByteSize()), dst, src);
			break;
	}
}


//====================================================================
void __fastcall Resample2D_Linear_NFloats(ImageBase& dst, const ImageBase& src,
								uint componentOffset, uint numComponentsPerSample)
{
	uint numComponents = dst.GetSampleByteSize() / sizeof(float);
	uint byteOffset = componentOffset * sizeof(float);

	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");
	VW_ASSERT(numComponentsPerSample <= numComponents, "Num-components-per-sample out of range.");
	VW_ASSERT(componentOffset <= numComponents - numComponentsPerSample, "Component-offset out of range.");

	ResampleElements2D_Linear<ubyte, ubyte>(
			sample_processors::LinearInterpolator2D_NFloats(numComponentsPerSample),
			dst.GetSamplesB() + byteOffset, dst.GetRowByteSize(), dst.GetSampleByteSize(),
			dst.GetNumRows(), dst.GetNumColumns(),
			src.GetSamplesB() + byteOffset, src.GetRowByteSize(), src.GetSampleByteSize(),
			src.GetNumRows(), src.GetNumColumns());
}


//====================================================================
void __fastcall Resample2D_Linear_NBytes(ImageBase& dst, const ImageBase& src,
										uint byteOffset, uint numBytesPerSample)
{
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");
	VW_ASSERT(numBytesPerSample <= dst.GetSampleByteSize(), "Num-bytes-per-sample out of range.");
	VW_ASSERT(byteOffset <= dst.GetSampleByteSize() - numBytesPerSample, "Byte-offset out of range.");

	ResampleElements2D_Linear<ubyte, ubyte>(
			sample_processors::LinearInterpolator2D_NBytes(numBytesPerSample),
			dst.GetSamplesB() + byteOffset, dst.GetRowByteSize(), dst.GetSampleByteSize(),
			dst.GetNumRows(), dst.GetNumColumns(),
			src.GetSamplesB() + byteOffset, src.GetRowByteSize(), src.GetSampleByteSize(),
			src.GetNumRows(), src.GetNumColumns());
}


//====================================================================
void __fastcall Resample2D_Linear_PackedSample_UShort4444(ImageBase& dst, const ImageBase& src)
{
	Call_ResampleElements2D_Linear_PackedSample_UShort(
		sample_processors::LinearInterpolator2D_PackedSample_UShort4444(), dst, src);
}


//====================================================================
void __fastcall Resample2D_Linear_PackedSample_UShort1555(ImageBase& dst, const ImageBase& src)
{
	Call_ResampleElements2D_Linear_PackedSample_UShort(
		sample_processors::LinearInterpolator2D_PackedSample_UShort1555(), dst, src);
}


//====================================================================
bool __fastcall Resample2D(ImageBase& dst, const ImageBase& src,
										eResampleFilter filterType)
{
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");

	// nearest-point resampling
	if(filterType == ResampleFilter_Nearest) {

		Resample2D_Nearest(dst, src);
		return true;
	}

	// linear-interpolation resampling
	if(filterType == ResampleFilter_Linear) {

		switch(dst.GetSampleFormat()) {
			case SampleFormat_L8:
			case SampleFormat_A8:
			case SampleFormat_A8L8:
			case SampleFormat_A8R8G8B8:
				Resample2D_Linear_NBytes(dst, src, 0, dst.GetSampleByteSize());
				return true;

			case SampleFormat_X8R8G8B8:
				Resample2D_Linear_NBytes(dst, src, 0, 3);
				return true;

			case SampleFormat_A4R4G4B4:
				Resample2D_Linear_PackedSample_UShort4444(dst, src);
				return true;

			case SampleFormat_A1R5G5B5:
				Resample2D_Linear_PackedSample_UShort1555(dst, src);
				return true;

			case SampleFormat_ZB32:
			case SampleFormat_WB32:
				Resample2D_Linear_NFloats(dst, src, 0, dst.GetSampleByteSize() / sizeof(float));
				return true;
		}
	}

	// unsupported filter-type / sample-type
	return false;
}


//====================================================================
void __fastcall FlipRows(ImageBase& target)
{
	uint halfNumRows = target.GetNumRows() / 2;
	if(! halfNumRows) return;

	ProcessElements2D(element_processors::Swap_N<ubyte>(target.GetSampleByteSize()),
			halfNumRows, target.GetNumColumns(),
			target.GetSamplesB(), target.GetRowByteSize(), target.GetSampleByteSize(),
			target.SampleB(target.GetNumRows() - 1, 0), - target.GetRowByteSize(), target.GetSampleByteSize());
}


//====================================================================
void __fastcall FlipColumns(ImageBase& target)
{
	uint halfNumCols = target.GetNumColumns() / 2;
	if(! halfNumCols) return;

	// reverse row/column semantics
	ProcessElements2D(element_processors::Swap_N<ubyte>(target.GetSampleByteSize()),
			halfNumCols, target.GetNumRows(),
			target.GetSamplesB(), target.GetSampleByteSize(), target.GetRowByteSize(),
			target.GetSamplesB() + target.GetRowByteSize() - target.GetSampleByteSize(),
			- target.GetSampleByteSize(), target.GetRowByteSize());
}


//====================================================================
void __fastcall FillBlock(ImageBase& target, uint byteOffset, ubyte value,
						uint topRow, uint countRows, uint leftColumn, uint countColumns)
{
	VW_ASSERT(byteOffset < target.GetSampleByteSize(), "Byte-offset out of range.");

	VW_ASSERT((topRow < target.GetNumRows()) &&
			(topRow + countRows <= target.GetNumRows()) &&
			(leftColumn < target.GetNumColumns()) &&
			(leftColumn + countColumns <= target.GetNumColumns()),
			"Block-indices out of range");

	ProcessElements2D(element_processors::Process_Fill1<ubyte>(value),
						countRows, countColumns,
						target.SampleB(topRow, leftColumn) + byteOffset,
						target.GetRowByteSize(), target.GetSampleByteSize());
}


//====================================================================
void __fastcall GenerateCheckerboard(ImageBase& target, uint byteOffset,
									ubyte value1, ubyte value2,
									uint vertBlockSize, uint horzBlockSize)
{
	VW_ASSERT(byteOffset < target.GetSampleByteSize(), "Byte-offset out of range.");

	if(! target.GetNumSamples()) return;

	ClampLow<uint>(horzBlockSize, 1);
	ClampLow<uint>(vertBlockSize, 1);

	ProcessElements2D(CheckerboardProcessor(1, &value1, &value2,
											vertBlockSize, horzBlockSize),
						target.GetNumRows(), target.GetNumColumns(),
						target.GetSamplesB() + byteOffset,
						target.GetRowByteSize(), target.GetSampleByteSize());
}


//====================================================================
void __fastcall GenerateCheckerboard(ImageBase& target, uint tripletOffset,
							color4b rgb1, color4b rgb2,
							uint vertBlockSize, uint horzBlockSize)
{
	VW_ASSERT((target.GetSampleByteSize() >= 3)  &&
			(tripletOffset <= target.GetSampleByteSize() - 3),
			"Triplet-offset out of range.");

	if(! target.GetNumSamples()) return;

	ClampLow<uint>(horzBlockSize, 1);
	ClampLow<uint>(vertBlockSize, 1);

	ProcessElements2D(CheckerboardProcessor(3, rgb1.ToPtr(), rgb2.ToPtr(),
											vertBlockSize, horzBlockSize),
						target.GetNumRows(), target.GetNumColumns(),
						target.GetSamplesB() + tripletOffset,
						target.GetRowByteSize(), target.GetSampleByteSize());
}


//====================================================================
void __fastcall GenerateRGBATest(ImageBase& target)
{
	if(! target.GetNumSamples()) return;

	// create temporary image with suitable sample-format
	Image32  img(SampleFormat_A8R8G8B8, target.GetNumRows(), target.GetNumColumns());

	// prepare drawing entities
	win_utils::GdiplusGuard  gpg;

	Gdiplus::BitmapData  bitmapData;
	GdiplusFillBitmapData(bitmapData, PixelFormat32bppARGB, img);

	Gdiplus::Bitmap    bitmap(bitmapData.Width, bitmapData.Height, bitmapData.PixelFormat);
	Gdiplus::Graphics  gfx(&bitmap);

	Gdiplus::FontFamily  fontFamily(VW_WIDE_TEXT("Tahoma"));
	Gdiplus::Font        font(&fontFamily, Max<uint>(1, bitmapData.Height / 4),
								Gdiplus::FontStyleBold, Gdiplus::UnitPixel);

	Gdiplus::SolidBrush  redBrush(Gdiplus::Color(255, 255, 0, 0));
	Gdiplus::SolidBrush  greenBrush(Gdiplus::Color(255, 0, 255, 0));
	Gdiplus::SolidBrush  blueBrush(Gdiplus::Color(255, 0, 0, 255));
	Gdiplus::SolidBrush  blackBrush(Gdiplus::Color(255, 0, 0, 0));

	Gdiplus::StringFormat  sFmt;
	sFmt.SetAlignment(Gdiplus::StringAlignmentCenter);
	sFmt.SetLineAlignment(Gdiplus::StringAlignmentCenter);

	gfx.SetTextRenderingHint(Gdiplus::TextRenderingHintAntiAlias);

	Gdiplus::REAL hw = bitmapData.Width / 2.f;
	Gdiplus::REAL hh = bitmapData.Height / 2.f;

	// draw rgb areas
	gfx.FillRectangle(&redBrush, 0.f, 0.f, hw, hh);
	gfx.DrawString(VW_WIDE_TEXT("R"), 1, &font, Gdiplus::RectF(0, 0, hw, hh), &sFmt, &blackBrush);

	gfx.FillRectangle(&greenBrush, hw, 0.f, hw, hh);
	gfx.DrawString(VW_WIDE_TEXT("G"), 1, &font, Gdiplus::RectF(hw, 0, hw, hh), &sFmt, &blackBrush);

	gfx.FillRectangle(&blueBrush, 0.f, hh, hw, hh);
	gfx.DrawString(VW_WIDE_TEXT("B"), 1, &font, Gdiplus::RectF(0, hh, hw, hh), &sFmt, &blackBrush);

	// draw alpha area
	Gdiplus::RectF  alphaRect(hw, hh, hw, hh);
	Gdiplus::LinearGradientBrush  alphaBrush(alphaRect,
		Gdiplus::Color(255, 255, 255, 255), Gdiplus::Color(0, 255, 255, 255), 90, FALSE);
	gfx.SetCompositingMode(Gdiplus::CompositingModeSourceCopy);
	gfx.FillRectangle(&alphaBrush, alphaRect);
	gfx.SetCompositingMode(Gdiplus::CompositingModeSourceOver);
	gfx.DrawString(VW_WIDE_TEXT("A"), 1, &font, Gdiplus::RectF(hw, hh, hw, hh), &sFmt, &blackBrush);

	// transfer pixels from GDI+ to temporary image, then transfer samples to output image
	GdiplusTransferBitmapPixels(bitmap, bitmapData, true);
	TransferSamples2D(target, img, false);
}


//====================================================================
bool __fastcall CanGenerateFont(const std::string& fontFamilyName,
								bool bBold, bool bItalic)
{
	std::wstring  fontFamilyNameU;

	if(fontFamilyName.empty()) return false;
	if(! AsciiToUnicode(fontFamilyName, fontFamilyNameU)) return false;

	// determine font parameters and create test-font
	INT         fontStyle  = Gdiplus::FontStyleRegular;
	if(bBold)   fontStyle |= Gdiplus::FontStyleBold;
	if(bItalic) fontStyle |= Gdiplus::FontStyleItalic;

	AutoPtr<Gdiplus::FontFamily>  apFontFamily(new Gdiplus::FontFamily(fontFamilyNameU.c_str()));
	if(! apFontFamily->IsAvailable()) return false;
	if(! apFontFamily->IsStyleAvailable(fontStyle)) return false;
	return true;
}


//====================================================================
void __fastcall GenerateFontGlyphs(ImageBase& target,
								uint numGlyphs, const wchar_t* glyphs,
								const std::string& fontFamilyName, bool bBold, bool bItalic,
								uint numGlyphsPerRow, uint cellSize,
								float scaleEmHeight,
								float glyphOffsetH, float glyphOffsetV,
								float glyphSizeModifyH, float glyphSizeModifyV,
								bool bAntialias, bool bAlphaPremultiply,
								bool bDrawCells)
{
	VW_ASSERT(numGlyphs  &&  glyphs, "Invalid argument.");

	ClampLow<uint>(numGlyphsPerRow, 1);
	ClampLow<uint>(cellSize, 1);

	float maxTweak = cellSize;
	Clamp<float>(glyphOffsetH, -maxTweak, maxTweak);
	Clamp<float>(glyphOffsetV, -maxTweak, maxTweak);
	Clamp<float>(glyphSizeModifyH, -maxTweak, maxTweak);
	Clamp<float>(glyphSizeModifyV, -maxTweak, maxTweak);

	uint  numRows = numGlyphs / numGlyphsPerRow;
	if(numGlyphs % numGlyphsPerRow) numRows++;

	// create temporary image with suitable sample-format
	Image32  img(SampleFormat_A8R8G8B8, cellSize * numRows, cellSize * numGlyphsPerRow);

	// prepare drawing entities
	win_utils::GdiplusGuard  gpg;

	Gdiplus::BitmapData  bitmapData;
	GdiplusFillBitmapData(bitmapData, PixelFormat32bppARGB, img);

	Gdiplus::Bitmap    bitmap(bitmapData.Width, bitmapData.Height, bitmapData.PixelFormat);
	Gdiplus::Graphics  gfx(&bitmap);

	// determine font parameters and create font-family
	AutoPtr<Gdiplus::FontFamily>  apFontFamily;

	INT         fontStyle  = Gdiplus::FontStyleRegular;
	if(bBold)   fontStyle |= Gdiplus::FontStyleBold;
	if(bItalic) fontStyle |= Gdiplus::FontStyleItalic;

	if(CanGenerateFont(fontFamilyName, bBold, bItalic)) {

		std::wstring  fontFamilyNameU;
		AsciiToUnicode(fontFamilyName, fontFamilyNameU);

		apFontFamily.Assign(new Gdiplus::FontFamily(fontFamilyNameU.c_str()));
	}
	else {
		// fallback to generic family
		apFontFamily.Assign(const_cast<Gdiplus::FontFamily*>(Gdiplus::FontFamily::GenericMonospace()));
		apFontFamily.Release(); //generic font-family is owned by GDI+

		fontStyle = Gdiplus::FontStyleRegular;
	}
	VW_ASSERT(apFontFamily, "Internal error, failed to create font-family.");

	// compute em-height scaling factor
	float finalScaleEmHeight = float(apFontFamily->GetEmHeight(fontStyle)) /
					float(apFontFamily->GetCellAscent(fontStyle) + apFontFamily->GetCellDescent(fontStyle));

	finalScaleEmHeight *= scaleEmHeight;

	// create font, 1 cell should accomodate full font-height
	AutoPtr<Gdiplus::Font>  apFont(new Gdiplus::Font(apFontFamily,
									cellSize * finalScaleEmHeight, fontStyle, Gdiplus::UnitPixel));

	if(! apFont->IsAvailable()) {
		VW_THROW(SystemResourceException, "Failed to create gdiplus-font.");
	}

	// setup antialiasing
	gfx.SetTextRenderingHint(bAntialias ?
							Gdiplus::TextRenderingHintAntiAlias :
							Gdiplus::TextRenderingHintSingleBitPerPixelGridFit);

	Gdiplus::StringFormat  sFmt;
	sFmt.SetAlignment(Gdiplus::StringAlignmentCenter);
	sFmt.SetLineAlignment(Gdiplus::StringAlignmentCenter);

	gfx.Clear(Gdiplus::Color(255, 0, 0, 0));
	gfx.SetCompositingMode(Gdiplus::CompositingModeSourceOver);

	Gdiplus::Pen         cellPen(Gdiplus::Color(255, 255, 255, 255));
	Gdiplus::SolidBrush  glyphBrush(Gdiplus::Color(255, 255, 255, 255));

	Gdiplus::REAL  xPos  = 0;
	Gdiplus::REAL  yPos  = 0;
	Gdiplus::REAL  rSize = cellSize - 1;

	wchar_t  glyphIndex = 0;
	uint     currGlyphsPerRow = 0;

	// draw glyphs
	while(glyphIndex < numGlyphs) {
		if(bDrawCells) {
			Gdiplus::RectF  cellRect(xPos, yPos, rSize, rSize);
			gfx.DrawRectangle(&cellPen, cellRect);
		}
		Gdiplus::RectF  glyphRect(xPos + glyphOffsetH, yPos + glyphOffsetV,
								cellSize + glyphSizeModifyH, cellSize + glyphSizeModifyV);

		gfx.DrawString(glyphs + glyphIndex, 1, apFont, glyphRect, &sFmt, &glyphBrush);

		glyphIndex++;
		currGlyphsPerRow++;

		if(currGlyphsPerRow >= numGlyphsPerRow) {
			currGlyphsPerRow = 0;
			xPos = 0;
			yPos += cellSize;
		}
		else {
			xPos += cellSize;
		}
	}

	// transfer pixels from GDI+ to temporary image
	GdiplusTransferBitmapPixels(bitmap, bitmapData, true);

	// fill alpha channel with luminance
	CopySamples_3BytesBGRTo1ByteL(img, 3, img, 0);

	// set non-transparent RGB to full intensity
	if(! bAlphaPremultiply) {
		ThresholdSegmentationNBytes(img, 0, 3, 0, 0, 0, 255);
	}

	// transfer samples to output image
	TransferSamples2D(target, img, true);
}


//====================================================================
void __fastcall GenerateFontGlyphsAscii(ImageBase& target,
								uint firstAsciiGlyph, uint lastAsciiGlyph,
								const std::string& fontFamilyName, bool bBold, bool bItalic,
								uint numGlyphsPerRow, uint cellSize,
								float scaleEmHeight,
								float glyphOffsetH, float glyphOffsetV,
								float glyphSizeModifyH, float glyphSizeModifyV,
								bool bAntialias, bool bAlphaPremultiply,
								bool bDrawCells)
{
	Clamp<uint>(firstAsciiGlyph, 0, 255);
	Clamp<uint>(lastAsciiGlyph, firstAsciiGlyph, 255);

	uint numGlyphs = lastAsciiGlyph - firstAsciiGlyph + 1;
	Buffer<wchar_t>  asciiGlyphs(numGlyphs);

	for(uint i = 0; i < numGlyphs; i++) {
		asciiGlyphs[i] = firstAsciiGlyph + i;
	}

	GenerateFontGlyphs(target,
					numGlyphs, asciiGlyphs,
					fontFamilyName, bBold, bItalic,
					numGlyphsPerRow, cellSize,
					scaleEmHeight,
					glyphOffsetH, glyphOffsetV,
					glyphSizeModifyH, glyphSizeModifyV,
					bAntialias, bAlphaPremultiply,
					bDrawCells);
}


//********************************************************************
}//namespace vertexware


