

//--------------------------------------------------------------------
#include "DataObjects\MipmapChains.h"
#include "DataObjects\SignalOperations.h"
#include "DataObjects\ImageOperations.h"
#include "DataObjects\VolumeOperations.h"
#include "MathTools\MathBase.h"
#include "BaseTools\AutoCleanup.h"


//********************************************************************
namespace vertexware {


//====================================================================
MipmapChain::MipmapChain(eSampleFormat _sampleFormat)
{
	sampleFormat = _sampleFormat;
	numLevels = 0;
}


//====================================================================
void  MipmapChain::SetMaxDimension(uint maxDim)
{
	if(maxDim == 0) {
		numLevels = 0;
	}
	else {
		numLevels = 1 + Log2<uint>(maxDim);
	}
}


//====================================================================
SignalMipmapChain::SignalMipmapChain(eSampleFormat sampleFormat)
	:
	MipmapChain(sampleFormat)
{
}


//====================================================================
void  SignalMipmapChain::Clear()
{
	SetMaxDimension(0);
	levels.clear();
}


//====================================================================
bool  SignalMipmapChain::Build(const SignalBase& srcLevel0, bool bCopyLevel0OnFormatMatch)
{
	Clear();
	AutoCleanup<SignalMipmapChain> autoCleanup(this, true, CleanupHandlerMFN<SignalMipmapChain>(SignalMipmapChain::Clear));

	// setup base properties
	VW_ASSERT(IsPowerOf2(srcLevel0.GetNumSamples()), "Dimension is not power-of-2.");

	SetMaxDimension(srcLevel0.GetNumSamples());
	levels.reserve(GetNumLevels());

	// convert srcLevel0 to chain-format if necessary (implicitly checks conversion-support)
	AutoPtr<SignalBase>  apMatchingLevel0(const_cast<SignalBase*>(&srcLevel0), false);

	bool  bNeedsConversion = false;
	if(GetSampleFormat() != srcLevel0.GetSampleFormat())
	{
		bNeedsConversion = true;
		apMatchingLevel0.Assign(ManufactureSignal(GetSampleFormat(), srcLevel0.GetNumSamples()));
		if(! apMatchingLevel0) return false;
		if(! TransferSamples1D(*apMatchingLevel0, srcLevel0, false)) return false;
	}
	else if(bCopyLevel0OnFormatMatch)
	{
		apMatchingLevel0.Assign(ManufactureSignal(GetSampleFormat(), srcLevel0.GetNumSamples()));
		CopySamples1D(*apMatchingLevel0, srcLevel0);
	}

	// store level 0
	levels.push_back(apMatchingLevel0);

	AutoPtr<SignalBase>  apResampleSrcLevelN(const_cast<SignalBase*>(&srcLevel0), false);
	AutoPtr<SignalBase>  apResampleDstLevelN(ManufactureSignal(srcLevel0.GetSampleFormat()));
	AutoPtr<SignalBase>  apMatchingLevelN(bNeedsConversion ? ManufactureSignal(GetSampleFormat()) : 0);

	uint level = 1;
	while(true) {

		uint countLevelN = ComputeDimensionAtLevel(level, srcLevel0.GetNumSamples());

		// resample from higher-detail-level original-format source
		apResampleDstLevelN->ReserveSampleStorage(countLevelN);
		if(! Resample1D(*apResampleDstLevelN, *apResampleSrcLevelN, ResampleFilter_Linear)) {
			return false;
		}

		// convert to chain-sample-format if necessary
		if(bNeedsConversion) {
			TransferSamples1D(*apMatchingLevelN, *apResampleDstLevelN, true);
		}
		else {
			apMatchingLevelN = apResampleDstLevelN;
		}

		// store level N
		levels.push_back(apMatchingLevelN);

		// setup resample-destination as resample-source for next-level
		apResampleSrcLevelN = apResampleDstLevelN;

		// next level
		level++;
		if(level >= GetNumLevels()) break;

		// allocate new resample-destination object
		apResampleDstLevelN.Assign(ManufactureSignal(srcLevel0.GetSampleFormat()));
	}

	autoCleanup.Disable();
	return true;
}


//====================================================================
ImageMipmapChain::ImageMipmapChain(eSampleFormat sampleFormat)
	:
	MipmapChain(sampleFormat)
{
}


//====================================================================
void  ImageMipmapChain::Clear()
{
	SetMaxDimension(0);
	levels.clear();
}


//====================================================================
bool  ImageMipmapChain::Build(const ImageBase& srcLevel0, bool bCopyLevel0OnFormatMatch)
{
	Clear();
	AutoCleanup<ImageMipmapChain> autoCleanup(this, true, CleanupHandlerMFN<ImageMipmapChain>(ImageMipmapChain::Clear));

	// setup base properties
	VW_ASSERT(IsPowerOf2(srcLevel0.GetNumRows()) &&
				IsPowerOf2(srcLevel0.GetNumColumns()), "Dimension is not power-of-2.");

	SetMaxDimension(Max(srcLevel0.GetNumRows(), srcLevel0.GetNumColumns()));
	levels.reserve(GetNumLevels());

	// convert srcLevel0 to chain-format if necessary (implicitly checks conversion-support)
	AutoPtr<ImageBase>  apMatchingLevel0(const_cast<ImageBase*>(&srcLevel0), false);

	bool  bNeedsConversion = false;
	if(GetSampleFormat() != srcLevel0.GetSampleFormat())
	{
		bNeedsConversion = true;
		apMatchingLevel0.Assign(ManufactureImage(GetSampleFormat(),
								srcLevel0.GetNumRows(), srcLevel0.GetNumColumns()));
		if(! apMatchingLevel0) return false;
		if(! TransferSamples2D(*apMatchingLevel0, srcLevel0, false)) return false;
	}
	else if(bCopyLevel0OnFormatMatch)
	{
		apMatchingLevel0.Assign(ManufactureImage(GetSampleFormat(),
								srcLevel0.GetNumRows(), srcLevel0.GetNumColumns()));
		CopySamples2D(*apMatchingLevel0, srcLevel0);
	}

	// store level 0
	levels.push_back(apMatchingLevel0);

	AutoPtr<ImageBase>  apResampleSrcLevelN(const_cast<ImageBase*>(&srcLevel0), false);
	AutoPtr<ImageBase>  apResampleDstLevelN(ManufactureImage(srcLevel0.GetSampleFormat()));
	AutoPtr<ImageBase>  apMatchingLevelN(bNeedsConversion ? ManufactureImage(GetSampleFormat()) : 0);

	uint level = 1;
	while(true) {

		uint numRowsLevelN = ComputeDimensionAtLevel(level, srcLevel0.GetNumRows());
		uint numColumnsLevelN = ComputeDimensionAtLevel(level, srcLevel0.GetNumColumns());

		// resample from higher-detail-level original-format source
		apResampleDstLevelN->ReserveSampleStorage(numRowsLevelN, numColumnsLevelN);
		if(! Resample2D(*apResampleDstLevelN, *apResampleSrcLevelN, ResampleFilter_Linear)) {
			return false;
		}

		// convert to chain-sample-format if necessary
		if(bNeedsConversion) {
			TransferSamples2D(*apMatchingLevelN, *apResampleDstLevelN, true);
		}
		else {
			apMatchingLevelN = apResampleDstLevelN;
		}

		// store level N
		levels.push_back(apMatchingLevelN);

		// setup resample-destination as resample-source for next-level
		apResampleSrcLevelN = apResampleDstLevelN;

		// next level
		level++;
		if(level >= GetNumLevels()) break;

		// allocate new resample-destination object
		apResampleDstLevelN.Assign(ManufactureImage(srcLevel0.GetSampleFormat()));
	}

	autoCleanup.Disable();
	return true;
}


//====================================================================
VolumeMipmapChain::VolumeMipmapChain(eSampleFormat sampleFormat)
	:
	MipmapChain(sampleFormat)
{
}


//====================================================================
void  VolumeMipmapChain::Clear()
{
	SetMaxDimension(0);
	levels.clear();
}


//====================================================================
bool  VolumeMipmapChain::Build(const VolumeBase& srcLevel0, bool bCopyLevel0OnFormatMatch)
{
	Clear();
	AutoCleanup<VolumeMipmapChain> autoCleanup(this, true, CleanupHandlerMFN<VolumeMipmapChain>(VolumeMipmapChain::Clear));

	// setup base properties
	VW_ASSERT(IsPowerOf2(srcLevel0.GetNumLayers()) &&
				IsPowerOf2(srcLevel0.GetNumRows()) &&
				IsPowerOf2(srcLevel0.GetNumColumns()), "Dimension is not power-of-2.");

	SetMaxDimension(Max(srcLevel0.GetNumLayers(), srcLevel0.GetNumRows(), srcLevel0.GetNumColumns()));
	levels.reserve(GetNumLevels());

	// convert srcLevel0 to chain-format if necessary (implicitly checks conversion-support)
	AutoPtr<VolumeBase>  apMatchingLevel0(const_cast<VolumeBase*>(&srcLevel0), false);

	bool  bNeedsConversion = false;
	if(GetSampleFormat() != srcLevel0.GetSampleFormat())
	{
		bNeedsConversion = true;
		apMatchingLevel0.Assign(ManufactureVolume(GetSampleFormat(), srcLevel0.GetNumLayers(),
								srcLevel0.GetNumRows(), srcLevel0.GetNumColumns()));
		if(! apMatchingLevel0) return false;
		if(! TransferSamples3D(*apMatchingLevel0, srcLevel0, false)) return false;
	}
	else if(bCopyLevel0OnFormatMatch)
	{
		apMatchingLevel0.Assign(ManufactureVolume(GetSampleFormat(), srcLevel0.GetNumLayers(),
								srcLevel0.GetNumRows(), srcLevel0.GetNumColumns()));
		CopySamples3D(*apMatchingLevel0, srcLevel0);
	}

	// store level 0
	levels.push_back(apMatchingLevel0);

	// upload other levels
	AutoPtr<VolumeBase>  apResampleSrcLevelN(const_cast<VolumeBase*>(&srcLevel0), false);
	AutoPtr<VolumeBase>  apResampleDstLevelN(ManufactureVolume(srcLevel0.GetSampleFormat()));
	AutoPtr<VolumeBase>  apMatchingLevelN(bNeedsConversion ? ManufactureVolume(GetSampleFormat()) : 0);

	uint level = 1;
	while(true) {

		uint numLayersLevelN = ComputeDimensionAtLevel(level, srcLevel0.GetNumLayers());
		uint numRowsLevelN = ComputeDimensionAtLevel(level, srcLevel0.GetNumRows());
		uint numColumnsLevelN = ComputeDimensionAtLevel(level, srcLevel0.GetNumColumns());

		// resample from higher-detail-level original-format source
		apResampleDstLevelN->ReserveSampleStorage(numLayersLevelN, numRowsLevelN, numColumnsLevelN);
		if(! Resample3D(*apResampleDstLevelN, *apResampleSrcLevelN, ResampleFilter_Linear)) {
			return false;
		}

		// convert to texture-sample-format if necessary
		if(bNeedsConversion) {
			TransferSamples3D(*apMatchingLevelN, *apResampleDstLevelN, true);
		}
		else {
			apMatchingLevelN = apResampleDstLevelN;
		}

		// store level N
		levels.push_back(apMatchingLevelN);

		// setup resample-destination as resample-source for next-level
		apResampleSrcLevelN = apResampleDstLevelN;

		// next level
		level++;
		if(level >= GetNumLevels()) break;

		// allocate new resample-destination object
		apResampleDstLevelN.Assign(ManufactureVolume(srcLevel0.GetSampleFormat()));
	}

	autoCleanup.Disable();
	return true;
}


//********************************************************************
}//namespace vertexware


