

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "MathTools\MathBase.h"
#include "BaseTools\ElementProcessors.h"
#include "DataObjects\Sample.h"


//********************************************************************
namespace vertexware {


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TElementProcessor>
inline
void __fastcall ResampleElements1D_Nearest(
						TElementProcessor& copyProcessor,
						TDstElement* pDst, int dstElemStride,
						uint dstNumElements,
						const TSrcElement* pSrc, int srcElemStride,
						uint srcNumElements)
{
	VW_ASSERT(pDst && pSrc, "Invalid argument.");

	// compute normalized step per destination element, cache addressing parameters
	float  dstStep         = 1.0f / (dstNumElements + 1);
	float  dstHalfStep     = dstStep * 0.5f;
	float  srcNumElementsF = srcNumElements;
	uint   srcMaxIndex     = srcNumElements - 1;

	// spin element loop
	TDstElement*  pDstEnd = pDst + dstElemStride * dstNumElements;
	float         dstPos  = dstHalfStep;

	while(pDst != pDstEnd) {

		// round normalized coord & access nearest source element
		uint  srcIndex = ClampedHigh(Round<uint>(srcNumElementsF *
									PostIncrement(dstPos, dstStep)), srcMaxIndex);

		const TSrcElement*  pSrcNearest = pSrc + srcElemStride * srcIndex;

		// copy element contents
		copyProcessor.Process(PointerCast<TElementProcessor::DstElementType>(
									PostIncrement(pDst, dstElemStride)),
								PointerCast<TElementProcessor::SrcElementType>(
									pSrcNearest));
	}
}


//====================================================================
template<typename TDstElement, typename TSrcElement>
struct LinearInterpolator1D
{
	typedef  TDstElement  DstElementType;
	typedef  TSrcElement  SrcElementType;

	__forceinline virtual void Process(
					TDstElement* dst,
					const TSrcElement* srcL, float weightL,
					const TSrcElement* srcR, float weightR)
	{ }
};


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TLinearInterpolator1D>
inline
void __fastcall ResampleElements1D_Linear(
						TLinearInterpolator1D& linearInterpolator,
						TDstElement* pDst, int dstElemStride,
						uint dstNumElements,
						const TSrcElement* pSrc, int srcElemStride,
						uint srcNumElements)
{
	// compute normalized step per destination element, cache addressing parameters
	float  dstStep         = 1.0f / (dstNumElements + 1);
	float  dstHalfStep     = dstStep * 0.5f;
	float  srcNumElementsF = srcNumElements;
	uint   srcMaxIndex     = srcNumElements - 1;

	// spin element loop
	TDstElement*  pDstEnd = pDst + dstElemStride * dstNumElements;
	float         dstPos  = dstHalfStep;

	while(pDst != pDstEnd) {

		// floor normalized coord & access 2 consecutive elements
		float  srcIndexF = srcNumElementsF * PostIncrement(dstPos, dstStep);
		uint   srcIndex  = srcIndexF;

		const TSrcElement*  pSrcL = pSrc + srcElemStride * srcIndex;
		const TSrcElement*  pSrcR = pSrcL + srcElemStride;

		// clamp sampling area to last element
		if(srcIndex == srcMaxIndex) {
			pSrcR -= srcElemStride;
		}

		// compute element weights
		float  weightR = srcIndexF - srcIndex;
		float  weightL = 1.0f - weightR;

		// intepolate element contents
		linearInterpolator.Process(
			PointerCast<TLinearInterpolator1D::DstElementType>(
									PostIncrement(pDst, dstElemStride)),
			PointerCast<const TLinearInterpolator1D::SrcElementType>(pSrcL),
			weightL,
			PointerCast<const TLinearInterpolator1D::SrcElementType>(pSrcR),
			weightR);
	}
}


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TElementProcessor>
inline
void __fastcall ResampleElements2D_Nearest(
						TElementProcessor& copyProcessor,
						TDstElement* pDst, int dstRowStride, int dstColumnStride,
						uint dstNumRows, uint dstNumColumns,
						const TSrcElement* pSrc, int srcRowStride, int srcColumnStride,
						uint srcNumRows, uint srcNumColumns)
{
	VW_ASSERT(pDst && pSrc, "Invalid argument.");

	// compute normalized steps per destination element, cache addressing parameters
	float  dstStepRow        = 1.0f / (dstNumRows + 1);
	float  dstHalfStepRow    = dstStepRow * 0.5f;
	float  dstStepColumn     = 1.0f / (dstNumColumns + 1);
	float  dstHalfStepColumn = dstStepColumn * 0.5f;
	int    dstSkipRows       = dstRowStride * dstNumRows;
	int    dstSkipColumns    = dstColumnStride * dstNumColumns;
	float  srcNumRowsF       = srcNumRows;
	float  srcNumColumnsF    = srcNumColumns;
	uint   srcMaxRow         = srcNumRows - 1;
	uint   srcMaxColumn      = srcNumColumns - 1;

	// spin row loop
	TDstElement*  pDstRow    = pDst;
	TDstElement*  pDstRowEnd = pDstRow + dstSkipRows;
	float         dstPosRow  = dstHalfStepRow;

	while(pDstRow != pDstRowEnd) {

		// round normalized row coord & access nearest source row
		uint  srcRow = ClampedHigh(Round<uint>(srcNumRowsF *
									PostIncrement(dstPosRow, dstStepRow)), srcMaxRow);

		const TSrcElement*  pSrcRow = pSrc + srcRowStride * srcRow;

		// spin column loop
		TDstElement*  pDstColumn    = PostIncrement(pDstRow, dstRowStride);
		TDstElement*  pDstColumnEnd = pDstColumn + dstSkipColumns;
		float         dstPosColumn  = dstHalfStepColumn;

		while(pDstColumn != pDstColumnEnd) {

			// round normalized column coord & access nearest source column
			uint  srcColumn = ClampedHigh(Round<uint>(srcNumColumnsF *
									PostIncrement(dstPosColumn, dstStepColumn)), srcMaxColumn);

			const TSrcElement*  pSrcColumn = pSrcRow + srcColumnStride * srcColumn;

			// copy element contents
			copyProcessor.Process(PointerCast<TElementProcessor::DstElementType>(
										PostIncrement(pDstColumn, dstColumnStride)),
									PointerCast<TElementProcessor::SrcElementType>(
										pSrcColumn));
		}
	}
}


//====================================================================
template<typename TDstElement, typename TSrcElement>
struct LinearInterpolator2D
{
	typedef  TDstElement  DstElementType;
	typedef  TSrcElement  SrcElementType;

	__forceinline virtual void Process(
					TDstElement* dst,
					const TSrcElement* srcTL, float weightTL,
					const TSrcElement* srcTR, float weightTR,
					const TSrcElement* srcBL, float weightBL,
					const TSrcElement* srcBR, float weightBR)
	{ }
};


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TLinearInterpolator2D>
inline
void __fastcall ResampleElements2D_Linear(
						TLinearInterpolator2D& linearInterpolator,
						TDstElement* pDst, int dstRowStride, int dstColumnStride,
						uint dstNumRows, uint dstNumColumns,
						const TSrcElement* pSrc, int srcRowStride, int srcColumnStride,
						uint srcNumRows, uint srcNumColumns)
{
	VW_ASSERT(pDst && pSrc, "Invalid argument.");

	// compute normalized steps per destination element, cache addressing parameters
	float  dstStepRow        = 1.0f / (dstNumRows + 1);
	float  dstHalfStepRow    = dstStepRow * 0.5f;
	float  dstStepColumn     = 1.0f / (dstNumColumns + 1);
	float  dstHalfStepColumn = dstStepColumn * 0.5f;
	int    dstSkipRows       = dstRowStride * dstNumRows;
	int    dstSkipColumns    = dstColumnStride * dstNumColumns;
	float  srcNumRowsF       = srcNumRows;
	float  srcNumColumnsF    = srcNumColumns;
	uint   srcMaxRow         = srcNumRows - 1;
	uint   srcMaxColumn      = srcNumColumns - 1;

	// spin row loop
	TDstElement*  pDstRow    = pDst;
	TDstElement*  pDstRowEnd = pDstRow + dstSkipRows;
	float         dstPosRow  = dstHalfStepRow;

	while(pDstRow != pDstRowEnd) {

		// floor to integer coordinate & access 2 consecutive source rows
		float  srcRowF = srcNumRowsF * PostIncrement(dstPosRow, dstStepRow);
		uint   srcRow  = srcRowF;

		const TSrcElement*  pSrcRowT = pSrc + srcRowStride * srcRow;
		const TSrcElement*  pSrcRowB = pSrcRowT + srcRowStride;

		// clamp sampling area to last row
		if(srcRow == srcMaxRow) {
			pSrcRowB -= srcRowStride;
		}

		// compute row weights
		float  weightB = srcRowF - srcRow;
		float  weightT = 1.0f - weightB;

		// spin column loop
		TDstElement*  pDstColumn    = PostIncrement(pDstRow, dstRowStride);
		TDstElement*  pDstColumnEnd = pDstColumn + dstSkipColumns;
		float         dstPosColumn  = dstHalfStepColumn;

		while(pDstColumn != pDstColumnEnd) {

			// floor to integer coordinate & access 2 consecutive columns per row
			float  srcColumnF     = srcNumColumnsF * PostIncrement(dstPosColumn, dstStepColumn);
			uint   srcColumn      = srcColumnF;
			int    srcSkipColumns = srcColumnStride * srcColumn;

			const TSrcElement*  pSrcTL = pSrcRowT + srcSkipColumns;
			const TSrcElement*  pSrcTR = pSrcTL + srcColumnStride;

			const TSrcElement*  pSrcBL = pSrcRowB + srcSkipColumns;
			const TSrcElement*  pSrcBR = pSrcBL + srcColumnStride;

			// clamp sampling area to last column
			if(srcColumn == srcMaxColumn) {
				pSrcTR -= srcColumnStride;
				pSrcBR -= srcColumnStride;
			}

			// compute column weights
			float  weightR = srcColumnF - srcColumn;
			float  weightL = 1.0f - weightR;

			// intepolate element contents
			linearInterpolator.Process(
				PointerCast<TLinearInterpolator2D::DstElementType>(
											PostIncrement(pDstColumn, dstColumnStride)),
				PointerCast<const TLinearInterpolator2D::SrcElementType>(pSrcTL),
				weightT * weightL,
				PointerCast<const TLinearInterpolator2D::SrcElementType>(pSrcTR),
				weightT * weightR,
				PointerCast<const TLinearInterpolator2D::SrcElementType>(pSrcBL),
				weightB * weightL,
				PointerCast<const TLinearInterpolator2D::SrcElementType>(pSrcBR),
				weightB * weightR);
		}
	}
}


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TElementProcessor>
inline
void __fastcall ResampleElements3D_Nearest(
					TElementProcessor& copyProcessor,
					TDstElement* pDst, int dstLayerStride, int dstRowStride, int dstColumnStride,
					uint dstNumLayers, uint dstNumRows, uint dstNumColumns,
					const TSrcElement* pSrc, int srcLayerStride, int srcRowStride, int srcColumnStride,
					uint srcNumLayers, uint srcNumRows, uint srcNumColumns)
{
	VW_ASSERT(pDst && pSrc, "Invalid argument.");

	// compute normalized steps per destination element, cache addressing parameters
	float  dstStepLayer      = 1.0f / (dstNumLayers + 1);
	float  dstHalfStepLayer  = dstStepLayer * 0.5f;
	float  dstStepRow        = 1.0f / (dstNumRows + 1);
	float  dstHalfStepRow    = dstStepRow * 0.5f;
	float  dstStepColumn     = 1.0f / (dstNumColumns + 1);
	float  dstHalfStepColumn = dstStepColumn * 0.5f;
	int    dstSkipLayers     = dstLayerStride * dstNumLayers;
	int    dstSkipRows       = dstRowStride * dstNumRows;
	int    dstSkipColumns    = dstColumnStride * dstNumColumns;
	float  srcNumLayersF     = srcNumLayers;
	float  srcNumRowsF       = srcNumRows;
	float  srcNumColumnsF    = srcNumColumns;
	uint   srcMaxLayer       = srcNumLayers - 1;
	uint   srcMaxRow         = srcNumRows - 1;
	uint   srcMaxColumn      = srcNumColumns - 1;

	// spin layer loop
	TDstElement*  pDstLayer    = pDst;
	TDstElement*  pDstLayerEnd = pDstLayer + dstSkipLayers;
	float         dstPosLayer  = dstHalfStepLayer;

	while(pDstLayer != pDstLayerEnd) {

		// round normalized layer coord & access nearest source layer
		uint  srcLayer = ClampedHigh(Round<uint>(srcNumLayersF *
								PostIncrement(dstPosLayer, dstStepLayer)), srcMaxLayer);

		const TSrcElement*  pSrcLayer = pSrc + srcLayerStride * srcLayer;

		// spin row loop
		TDstElement*  pDstRow    = PostIncrement(pDstLayer, dstLayerStride);
		TDstElement*  pDstRowEnd = pDstRow + dstSkipRows;
		float         dstPosRow  = dstHalfStepRow;

		while(pDstRow != pDstRowEnd) {

			// round normalized row coord & access nearest source row
			uint  srcRow = ClampedHigh(Round<uint>(srcNumRowsF *
										PostIncrement(dstPosRow, dstStepRow)), srcMaxRow);

			const TSrcElement*  pSrcRow = pSrcLayer + srcRowStride * srcRow;

			// spin column loop
			TDstElement*  pDstColumn    = PostIncrement(pDstRow, dstRowStride);
			TDstElement*  pDstColumnEnd = pDstColumn + dstSkipColumns;
			float         dstPosColumn  = dstHalfStepColumn;

			while(pDstColumn != pDstColumnEnd) {

				// round normalized column coord & access nearest source column
				uint  srcColumn = ClampedHigh(Round<uint>(srcNumColumnsF *
										PostIncrement(dstPosColumn, dstStepColumn)), srcMaxColumn);

				const TSrcElement*  pSrcColumn = pSrcRow + srcColumnStride * srcColumn;

				// copy element contents
				copyProcessor.Process(PointerCast<TElementProcessor::DstElementType>(
											PostIncrement(pDstColumn, dstColumnStride)),
										PointerCast<TElementProcessor::SrcElementType>(
											pSrcColumn));
			}
		}
	}
}


//====================================================================
template<typename TDstElement, typename TSrcElement>
struct LinearInterpolator3D
{
	typedef  TDstElement  DstElementType;
	typedef  TSrcElement  SrcElementType;

	__forceinline virtual void Process(
					TDstElement* dst,
					const TSrcElement* srcNTL, float weightNTL,
					const TSrcElement* srcNTR, float weightNTR,
					const TSrcElement* srcNBL, float weightNBL,
					const TSrcElement* srcNBR, float weightNBR,
					const TSrcElement* srcFTL, float weightFTL,
					const TSrcElement* srcFTR, float weightFTR,
					const TSrcElement* srcFBL, float weightFBL,
					const TSrcElement* srcFBR, float weightFBR)
	{ }
};


//====================================================================
template<typename TDstElement, typename TSrcElement, typename TLinearInterpolator3D>
inline
void __fastcall ResampleElements3D_Linear(
					TLinearInterpolator3D& linearInterpolator,
					TDstElement* pDst, int dstLayerStride, int dstRowStride, int dstColumnStride,
					uint dstNumLayers, uint dstNumRows, uint dstNumColumns,
					const TSrcElement* pSrc, int srcLayerStride, int srcRowStride, int srcColumnStride,
					uint srcNumLayers, uint srcNumRows, uint srcNumColumns)
{
	VW_ASSERT(pDst && pSrc, "Invalid argument.");

	// compute normalized steps per destination element, cache addressing parameters
	float  dstStepLayer      = 1.0f / (dstNumLayers + 1);
	float  dstHalfStepLayer  = dstStepLayer * 0.5f;
	float  dstStepRow        = 1.0f / (dstNumRows + 1);
	float  dstHalfStepRow    = dstStepRow * 0.5f;
	float  dstStepColumn     = 1.0f / (dstNumColumns + 1);
	float  dstHalfStepColumn = dstStepColumn * 0.5f;
	int    dstSkipLayers     = dstLayerStride * dstNumLayers;
	int    dstSkipRows       = dstRowStride * dstNumRows;
	int    dstSkipColumns    = dstColumnStride * dstNumColumns;
	float  srcNumLayersF     = srcNumLayers;
	float  srcNumRowsF       = srcNumRows;
	float  srcNumColumnsF    = srcNumColumns;
	uint   srcMaxLayer       = srcNumLayers - 1;
	uint   srcMaxRow         = srcNumRows - 1;
	uint   srcMaxColumn      = srcNumColumns - 1;

	// spin layer loop
	TDstElement*  pDstLayer    = pDst;
	TDstElement*  pDstLayerEnd = pDstLayer + dstSkipLayers;
	float         dstPosLayer  = dstHalfStepLayer;

	while(pDstLayer != pDstLayerEnd) {

		// floor to integer coordinate & access 2 consecutive source layers
		float  srcLayerF = srcNumLayersF * PostIncrement(dstPosLayer, dstStepLayer);
		uint   srcLayer  = srcLayerF;

		const TSrcElement*  pSrcLayerN = pSrc + srcLayerStride * srcLayer;
		const TSrcElement*  pSrcLayerF = pSrcLayerN + srcLayerStride;

		// clamp sampling area to last layer
		if(srcLayer == srcMaxLayer) {
			pSrcLayerF -= srcLayerStride;
		}

		// compute layer weights
		float  weightF = srcLayerF - srcLayer;
		float  weightN = 1.0f - weightF;

		// spin row loop
		TDstElement*  pDstRow    = PostIncrement(pDstLayer, dstLayerStride);
		TDstElement*  pDstRowEnd = pDstRow + dstSkipRows;
		float         dstPosRow  = dstHalfStepRow;

		while(pDstRow != pDstRowEnd) {

			// floor to integer coordinate & access 2 consecutive source rows per layer
			float  srcRowF     = srcNumRowsF * PostIncrement(dstPosRow, dstStepRow);
			uint   srcRow      = srcRowF;
			int    srcSkipRows = srcRowStride * srcRow;

			const TSrcElement*  pSrcRowNT = pSrcLayerN + srcSkipRows;
			const TSrcElement*  pSrcRowNB = pSrcRowNT + srcRowStride;

			const TSrcElement*  pSrcRowFT = pSrcLayerF + srcSkipRows;
			const TSrcElement*  pSrcRowFB = pSrcRowFT + srcRowStride;

			// clamp sampling area to last row
			if(srcRow == srcMaxRow) {
				pSrcRowNB -= srcRowStride;
				pSrcRowFB -= srcRowStride;
			}

			// compute row weights
			float  weightB = srcRowF - srcRow;
			float  weightT = 1.0f - weightB;

			// spin column loop
			TDstElement*  pDstColumn    = PostIncrement(pDstRow, dstRowStride);
			TDstElement*  pDstColumnEnd = pDstColumn + dstSkipColumns;
			float         dstPosColumn  = dstHalfStepColumn;

			while(pDstColumn != pDstColumnEnd) {

				// floor to integer coordinate & access 2 consecutive columns per row
				float  srcColumnF     = srcNumColumnsF * PostIncrement(dstPosColumn, dstStepColumn);
				uint   srcColumn      = srcColumnF;
				int    srcSkipColumns = srcColumnStride * srcColumn;

				const TSrcElement*  pSrcNTL = pSrcRowNT + srcSkipColumns;
				const TSrcElement*  pSrcNTR = pSrcNTL + srcColumnStride;

				const TSrcElement*  pSrcNBL = pSrcRowNB + srcSkipColumns;
				const TSrcElement*  pSrcNBR = pSrcNBL + srcColumnStride;

				const TSrcElement*  pSrcFTL = pSrcRowFT + srcSkipColumns;
				const TSrcElement*  pSrcFTR = pSrcFTL + srcColumnStride;

				const TSrcElement*  pSrcFBL = pSrcRowFB + srcSkipColumns;
				const TSrcElement*  pSrcFBR = pSrcFBL + srcColumnStride;

				// clamp sampling area to last column
				if(srcColumn == srcMaxColumn) {
					pSrcNTR -= srcColumnStride;
					pSrcNBR -= srcColumnStride;
					pSrcFTR -= srcColumnStride;
					pSrcFBR -= srcColumnStride;
				}

				// compute column weights
				float  weightR = srcColumnF - srcColumn;
				float  weightL = 1.0f - weightR;

				// intepolate element contents
				float  weightTL = weightT * weightL;
				float  weightTR = weightT * weightR;
				float  weightBL = weightB * weightL;
				float  weightBR = weightB * weightR;

				linearInterpolator.Process(
					PointerCast<TLinearInterpolator3D::DstElementType>(
												PostIncrement(pDstColumn, dstColumnStride)),
					PointerCast<const TLinearInterpolator3D::SrcElementType>(pSrcNTL),
					weightN * weightTL,
					PointerCast<const TLinearInterpolator3D::SrcElementType>(pSrcNTR),
					weightN * weightTR,
					PointerCast<const TLinearInterpolator3D::SrcElementType>(pSrcNBL),
					weightN * weightBL,
					PointerCast<const TLinearInterpolator3D::SrcElementType>(pSrcNBR),
					weightN * weightBR,
					PointerCast<const TLinearInterpolator3D::SrcElementType>(pSrcFTL),
					weightF * weightTL,
					PointerCast<const TLinearInterpolator3D::SrcElementType>(pSrcFTR),
					weightF * weightTR,
					PointerCast<const TLinearInterpolator3D::SrcElementType>(pSrcFBL),
					weightF * weightBL,
					PointerCast<const TLinearInterpolator3D::SrcElementType>(pSrcFBR),
					weightF * weightBR);
			}
		}
	}
}


//********************************************************************
namespace sample_processors {


//====================================================================
template<typename TComponent>
struct LinearInterpolator1D_NComponents : public LinearInterpolator1D<TComponent, TComponent>
{
	uint numComponents;

	LinearInterpolator1D_NComponents(uint _numComponents) {
		numComponents = _numComponents;
	}

	__forceinline virtual void Process(TComponent* dst,
									const TComponent* srcL, float weightL,
									const TComponent* srcR, float weightR)
	{
		for(uint i = 0; i < numComponents; i++) {
			dst[i] = TComponent(
				weightL * srcL[i] + weightR * srcR[i]);
		}
	}
};


//====================================================================
typedef  LinearInterpolator1D_NComponents<float>  LinearInterpolator1D_NFloats;


//====================================================================
struct VW_COMMON_API LinearInterpolator1D_NBytes : public LinearInterpolator1D_NComponents<ubyte>
{
	LinearInterpolator1D_NBytes(uint _numComponents)
		: LinearInterpolator1D_NComponents<ubyte>(_numComponents)
	{ }

	__forceinline virtual void Process(ubyte* dst,
									const ubyte* srcL, float weightL,
									const ubyte* srcR, float weightR)
	{
		for(uint i = 0; i < numComponents; i++) {
			dst[i] = Round<ubyte, float>(
				weightL * srcL[i] + weightR * srcR[i]);
		}
	}
};


//====================================================================
template<typename TPackedSample>
struct LinearInterpolator1D_PackedSample4 :
							public LinearInterpolator1D<TPackedSample::PackageType, TPackedSample::PackageType>
{
	__forceinline virtual void Process(TPackedSample::PackageType* dst,
									const TPackedSample::PackageType* srcL, float weightL,
									const TPackedSample::PackageType* srcR, float weightR)
	{
		using namespace sample_formats;

		TPackedSample::PackageType regSrcL = *srcL;
		TPackedSample::PackageType regSrcR = *srcR;
		TPackedSample::PackageType regDst  = 0;

		TPackedSample::Pack0(regDst, Round<TPackedSample::PackageType, float>(
				weightL * TPackedSample::Unpack0(regSrcL) +
				weightR * TPackedSample::Unpack0(regSrcR)));

		TPackedSample::Pack1(regDst, Round<TPackedSample::PackageType, float>(
				weightL * TPackedSample::Unpack1(regSrcL) +
				weightR * TPackedSample::Unpack1(regSrcR)));

		TPackedSample::Pack2(regDst, Round<TPackedSample::PackageType, float>(
				weightL * TPackedSample::Unpack2(regSrcL) +
				weightR * TPackedSample::Unpack2(regSrcR)));

		TPackedSample::Pack3(regDst, Round<TPackedSample::PackageType, float>(
				weightL * TPackedSample::Unpack3(regSrcL) +
				weightR * TPackedSample::Unpack3(regSrcR)));

		*dst = regDst;
	}
};


//====================================================================
typedef  LinearInterpolator1D_PackedSample4<sample_formats::PackedSample_UShort4444>
								LinearInterpolator1D_PackedSample_UShort4444;


//====================================================================
typedef  LinearInterpolator1D_PackedSample4<sample_formats::PackedSample_UShort1555>
								LinearInterpolator1D_PackedSample_UShort1555;


//====================================================================
template<typename TComponent>
struct LinearInterpolator2D_NComponents : public LinearInterpolator2D<TComponent, TComponent>
{
	uint numComponents;

	LinearInterpolator2D_NComponents(uint _numComponents) {
		numComponents = _numComponents;
	}

	__forceinline virtual void Process(TComponent* dst,
									const TComponent* srcTL, float weightTL,
									const TComponent* srcTR, float weightTR,
									const TComponent* srcBL, float weightBL,
									const TComponent* srcBR, float weightBR)
	{
		for(uint i = 0; i < numComponents; i++) {
			dst[i] = TComponent(
				weightTL * srcTL[i] + weightTR * srcTR[i] +
				weightBL * srcBL[i] + weightBR * srcBR[i]);
		}
	}
};


//====================================================================
typedef  LinearInterpolator2D_NComponents<float>  LinearInterpolator2D_NFloats;


//====================================================================
struct VW_COMMON_API LinearInterpolator2D_NBytes : public LinearInterpolator2D_NComponents<ubyte>
{
	LinearInterpolator2D_NBytes(uint _numComponents)
		: LinearInterpolator2D_NComponents<ubyte>(_numComponents)
	{ }

	__forceinline virtual void Process(ubyte* dst,
									const ubyte* srcTL, float weightTL,
									const ubyte* srcTR, float weightTR,
									const ubyte* srcBL, float weightBL,
									const ubyte* srcBR, float weightBR)
	{
		for(uint i = 0; i < numComponents; i++) {
			dst[i] = Round<ubyte, float>(
				weightTL * srcTL[i] + weightTR * srcTR[i] +
				weightBL * srcBL[i] + weightBR * srcBR[i]);
		}
	}
};


//====================================================================
template<typename TPackedSample>
struct LinearInterpolator2D_PackedSample4 :
							public LinearInterpolator2D<TPackedSample::PackageType, TPackedSample::PackageType>
{
	__forceinline virtual void Process(TPackedSample::PackageType* dst,
									const TPackedSample::PackageType* srcTL, float weightTL,
									const TPackedSample::PackageType* srcTR, float weightTR,
									const TPackedSample::PackageType* srcBL, float weightBL,
									const TPackedSample::PackageType* srcBR, float weightBR)
	{
		using namespace sample_formats;

		TPackedSample::PackageType regSrcTL = *srcTL;
		TPackedSample::PackageType regSrcTR = *srcTR;
		TPackedSample::PackageType regSrcBL = *srcBL;
		TPackedSample::PackageType regSrcBR = *srcBR;
		TPackedSample::PackageType regDst   = 0;

		TPackedSample::Pack0(regDst, Round<TPackedSample::PackageType, float>(
				weightTL * TPackedSample::Unpack0(regSrcTL) +
				weightTR * TPackedSample::Unpack0(regSrcTR) +
				weightBL * TPackedSample::Unpack0(regSrcBL) +
				weightBR * TPackedSample::Unpack0(regSrcBR)));

		TPackedSample::Pack1(regDst, Round<TPackedSample::PackageType, float>(
				weightTL * TPackedSample::Unpack1(regSrcTL) +
				weightTR * TPackedSample::Unpack1(regSrcTR) +
				weightBL * TPackedSample::Unpack1(regSrcBL) +
				weightBR * TPackedSample::Unpack1(regSrcBR)));

		TPackedSample::Pack2(regDst, Round<TPackedSample::PackageType, float>(
				weightTL * TPackedSample::Unpack2(regSrcTL) +
				weightTR * TPackedSample::Unpack2(regSrcTR) +
				weightBL * TPackedSample::Unpack2(regSrcBL) +
				weightBR * TPackedSample::Unpack2(regSrcBR)));

		TPackedSample::Pack3(regDst, Round<TPackedSample::PackageType, float>(
				weightTL * TPackedSample::Unpack3(regSrcTL) +
				weightTR * TPackedSample::Unpack3(regSrcTR) +
				weightBL * TPackedSample::Unpack3(regSrcBL) +
				weightBR * TPackedSample::Unpack3(regSrcBR)));

		*dst = regDst;
	}
};


//====================================================================
typedef  LinearInterpolator2D_PackedSample4<sample_formats::PackedSample_UShort4444>
								LinearInterpolator2D_PackedSample_UShort4444;


//====================================================================
typedef  LinearInterpolator2D_PackedSample4<sample_formats::PackedSample_UShort1555>
								LinearInterpolator2D_PackedSample_UShort1555;


//====================================================================
template<typename TComponent>
struct LinearInterpolator3D_NComponents : public LinearInterpolator3D<TComponent, TComponent>
{
	uint numComponents;

	LinearInterpolator3D_NComponents(uint _numComponents) {
		numComponents = _numComponents;
	}

	__forceinline virtual void Process(TComponent* dst,
									const TComponent* srcNTL, float weightNTL,
									const TComponent* srcNTR, float weightNTR,
									const TComponent* srcNBL, float weightNBL,
									const TComponent* srcNBR, float weightNBR,
									const TComponent* srcFTL, float weightFTL,
									const TComponent* srcFTR, float weightFTR,
									const TComponent* srcFBL, float weightFBL,
									const TComponent* srcFBR, float weightFBR)
	{
		for(uint i = 0; i < numComponents; i++) {
			dst[i] = TComponent(
				weightNTL * srcNTL[i] + weightNTR * srcNTR[i] +
				weightNBL * srcNBL[i] + weightNBR * srcNBR[i] +
				weightFTL * srcFTL[i] + weightFTR * srcFTR[i] +
				weightFBL * srcFBL[i] + weightFBR * srcFBR[i]);
		}
	}
};


//====================================================================
typedef  LinearInterpolator3D_NComponents<float>  LinearInterpolator3D_NFloats;


//====================================================================
struct VW_COMMON_API LinearInterpolator3D_NBytes : public LinearInterpolator3D_NComponents<ubyte>
{
	LinearInterpolator3D_NBytes(uint _numComponents)
		: LinearInterpolator3D_NComponents<ubyte>(_numComponents)
	{ }

	__forceinline virtual void Process(ubyte* dst,
									const ubyte* srcNTL, float weightNTL,
									const ubyte* srcNTR, float weightNTR,
									const ubyte* srcNBL, float weightNBL,
									const ubyte* srcNBR, float weightNBR,
									const ubyte* srcFTL, float weightFTL,
									const ubyte* srcFTR, float weightFTR,
									const ubyte* srcFBL, float weightFBL,
									const ubyte* srcFBR, float weightFBR)
	{
		for(uint i = 0; i < numComponents; i++) {
			dst[i] = Round<ubyte, float>(
				weightNTL * srcNTL[i] + weightNTR * srcNTR[i] +
				weightNBL * srcNBL[i] + weightNBR * srcNBR[i] +
				weightFTL * srcFTL[i] + weightFTR * srcFTR[i] +
				weightFBL * srcFBL[i] + weightFBR * srcFBR[i]);
		}
	}
};


//====================================================================
template<typename TPackedSample>
struct LinearInterpolator3D_PackedSample4 :
							public LinearInterpolator3D<TPackedSample::PackageType, TPackedSample::PackageType>
{
	__forceinline virtual void Process(TPackedSample::PackageType* dst,
									const TPackedSample::PackageType* srcNTL, float weightNTL,
									const TPackedSample::PackageType* srcNTR, float weightNTR,
									const TPackedSample::PackageType* srcNBL, float weightNBL,
									const TPackedSample::PackageType* srcNBR, float weightNBR,
									const TPackedSample::PackageType* srcFTL, float weightFTL,
									const TPackedSample::PackageType* srcFTR, float weightFTR,
									const TPackedSample::PackageType* srcFBL, float weightFBL,
									const TPackedSample::PackageType* srcFBR, float weightFBR)
	{
		using namespace sample_formats;

		TPackedSample::PackageType regSrcNTL = *srcNTL;
		TPackedSample::PackageType regSrcNTR = *srcNTR;
		TPackedSample::PackageType regSrcNBL = *srcNBL;
		TPackedSample::PackageType regSrcNBR = *srcNBR;
		TPackedSample::PackageType regSrcFTL = *srcFTL;
		TPackedSample::PackageType regSrcFTR = *srcFTR;
		TPackedSample::PackageType regSrcFBL = *srcFBL;
		TPackedSample::PackageType regSrcFBR = *srcFBR;
		TPackedSample::PackageType regDst    = 0;

		TPackedSample::Pack0(regDst, Round<TPackedSample::PackageType, float>(
				weightNTL * TPackedSample::Unpack0(regSrcNTL) +
				weightNTR * TPackedSample::Unpack0(regSrcNTR) +
				weightNBL * TPackedSample::Unpack0(regSrcNBL) +
				weightNBR * TPackedSample::Unpack0(regSrcNBR) +
				weightFTL * TPackedSample::Unpack0(regSrcFTL) +
				weightFTR * TPackedSample::Unpack0(regSrcFTR) +
				weightFBL * TPackedSample::Unpack0(regSrcFBL) +
				weightFBR * TPackedSample::Unpack0(regSrcFBR)));

		TPackedSample::Pack1(regDst, Round<TPackedSample::PackageType, float>(
				weightNTL * TPackedSample::Unpack1(regSrcNTL) +
				weightNTR * TPackedSample::Unpack1(regSrcNTR) +
				weightNBL * TPackedSample::Unpack1(regSrcNBL) +
				weightNBR * TPackedSample::Unpack1(regSrcNBR) +
				weightFTL * TPackedSample::Unpack1(regSrcFTL) +
				weightFTR * TPackedSample::Unpack1(regSrcFTR) +
				weightFBL * TPackedSample::Unpack1(regSrcFBL) +
				weightFBR * TPackedSample::Unpack1(regSrcFBR)));

		TPackedSample::Pack2(regDst, Round<TPackedSample::PackageType, float>(
				weightNTL * TPackedSample::Unpack2(regSrcNTL) +
				weightNTR * TPackedSample::Unpack2(regSrcNTR) +
				weightNBL * TPackedSample::Unpack2(regSrcNBL) +
				weightNBR * TPackedSample::Unpack2(regSrcNBR) +
				weightFTL * TPackedSample::Unpack2(regSrcFTL) +
				weightFTR * TPackedSample::Unpack2(regSrcFTR) +
				weightFBL * TPackedSample::Unpack2(regSrcFBL) +
				weightFBR * TPackedSample::Unpack2(regSrcFBR)));

		TPackedSample::Pack3(regDst, Round<TPackedSample::PackageType, float>(
				weightNTL * TPackedSample::Unpack3(regSrcNTL) +
				weightNTR * TPackedSample::Unpack3(regSrcNTR) +
				weightNBL * TPackedSample::Unpack3(regSrcNBL) +
				weightNBR * TPackedSample::Unpack3(regSrcNBR) +
				weightFTL * TPackedSample::Unpack3(regSrcFTL) +
				weightFTR * TPackedSample::Unpack3(regSrcFTR) +
				weightFBL * TPackedSample::Unpack3(regSrcFBL) +
				weightFBR * TPackedSample::Unpack3(regSrcFBR)));

		*dst = regDst;
	}
};


//====================================================================
typedef  LinearInterpolator3D_PackedSample4<sample_formats::PackedSample_UShort4444>
								LinearInterpolator3D_PackedSample_UShort4444;


//====================================================================
typedef  LinearInterpolator3D_PackedSample4<sample_formats::PackedSample_UShort1555>
								LinearInterpolator3D_PackedSample_UShort1555;


//====================================================================
struct VW_COMMON_API CopyConverter_FP32Norm_To_UInt8 : public ElementProcessorDstSrc<ubyte, const ubyte>
{
	__forceinline  void  Process(ubyte* dst, const ubyte* src) {
		*dst = Round<ubyte, float>(*reinterpret_cast<const float*>(src) * 255.0f);
	}
};


//====================================================================
struct VW_COMMON_API CopyConverter_UInt8_To_FP32Norm : public ElementProcessorDstSrc<ubyte, const ubyte>
{
	__forceinline  void  Process(ubyte* dst, const ubyte* src) {
		*reinterpret_cast<float*>(dst) = *src * float(1.0 / 255.0);
	}
};


//====================================================================
struct VW_COMMON_API CopyConverter_B8G8R8_To_L8 : public ElementProcessorDstSrc<ubyte, const ubyte>
{
	__forceinline  void  Process(ubyte* dst, const ubyte* src) {
		*dst = Round<ubyte, float>(0.114f * src[0] + 0.587f * src[1] + 0.299f * src[2]);
	}
};


//====================================================================
struct VW_COMMON_API CopyConverter_L8_To_B8G8R8 : public ElementProcessorDstSrc<ubyte, const ubyte>
{
	__forceinline  void  Process(ubyte* dst, const ubyte* src) {
		dst[2] = dst[1] = dst[0] = *src;
	}
};


//====================================================================
struct VW_COMMON_API CopyConverter_UInt8888_To_UShort4444 : public ElementProcessorDstSrc<ushort, const uint>
{
	__forceinline  void  Process(ushort* dst, const uint* src) {
		using namespace sample_formats;
		ushort regDst;
		uint   regSrc = *src;
		PackedSample_UShort4444::Pack0(regDst, PackedSample_UInt8888::Unpack0(regSrc) >> 4);
		PackedSample_UShort4444::Pack1(regDst, PackedSample_UInt8888::Unpack1(regSrc) >> 4);
		PackedSample_UShort4444::Pack2(regDst, PackedSample_UInt8888::Unpack2(regSrc) >> 4);
		PackedSample_UShort4444::Pack3(regDst, PackedSample_UInt8888::Unpack3(regSrc) >> 4);
		*dst = regDst;
	}
};


//====================================================================
struct VW_COMMON_API CopyConverter_UShort4444_To_UInt8888 : public ElementProcessorDstSrc<uint, const ushort>
{
	typedef  RangeMapperUInt<ushort, 15, uint, 255>  RangeMapper_4_8;

	__forceinline  void  Process(uint* dst, const ushort* src) {
		using namespace sample_formats;
		uint   regDst;
		ushort regSrc = *src;
		PackedSample_UInt8888::Pack0(regDst,
			RangeMapper_4_8::MapLowToHigh(PackedSample_UShort4444::Unpack0(regSrc)));
		PackedSample_UInt8888::Pack1(regDst,
			RangeMapper_4_8::MapLowToHigh(PackedSample_UShort4444::Unpack1(regSrc)));
		PackedSample_UInt8888::Pack2(regDst,
			RangeMapper_4_8::MapLowToHigh(PackedSample_UShort4444::Unpack2(regSrc)));
		PackedSample_UInt8888::Pack3(regDst,
			RangeMapper_4_8::MapLowToHigh(PackedSample_UShort4444::Unpack3(regSrc)));
		*dst = regDst;
	}
};


//====================================================================
struct VW_COMMON_API CopyConverter_UInt8888_To_UShort1555 : public ElementProcessorDstSrc<ushort, const uint>
{
	__forceinline  void  Process(ushort* dst, const uint* src) {
		using namespace sample_formats;
		ushort regDst;
		uint   regSrc = *src;
		PackedSample_UShort1555::Pack0(regDst, PackedSample_UInt8888::Unpack0(regSrc) >> 7);
		PackedSample_UShort1555::Pack1(regDst, PackedSample_UInt8888::Unpack1(regSrc) >> 3);
		PackedSample_UShort1555::Pack2(regDst, PackedSample_UInt8888::Unpack2(regSrc) >> 3);
		PackedSample_UShort1555::Pack3(regDst, PackedSample_UInt8888::Unpack3(regSrc) >> 3);
		*dst = regDst;
	}
};


//====================================================================
struct VW_COMMON_API CopyConverter_UShort1555_To_UInt8888 : public ElementProcessorDstSrc<uint, const ushort>
{
	typedef  RangeMapperUInt<ushort, 1,  uint, 255>  RangeMapper_1_8;
	typedef  RangeMapperUInt<ushort, 31, uint, 255>  RangeMapper_5_8;

	__forceinline  void  Process(uint* dst, const ushort* src) {
		using namespace sample_formats;
		uint   regDst;
		ushort regSrc = *src;
		PackedSample_UInt8888::Pack0(regDst,
			RangeMapper_1_8::MapLowToHigh(PackedSample_UShort1555::Unpack0(regSrc)));
		PackedSample_UInt8888::Pack1(regDst,
			RangeMapper_5_8::MapLowToHigh(PackedSample_UShort1555::Unpack1(regSrc)));
		PackedSample_UInt8888::Pack2(regDst,
			RangeMapper_5_8::MapLowToHigh(PackedSample_UShort1555::Unpack2(regSrc)));
		PackedSample_UInt8888::Pack3(regDst,
			RangeMapper_5_8::MapLowToHigh(PackedSample_UShort1555::Unpack3(regSrc)));
		*dst = regDst;
	}
};


//====================================================================
struct VW_COMMON_API Process_Fill1NibbleHigh : public ElementProcessorInplace<ubyte>
{
	ubyte  fillValue;

	Process_Fill1NibbleHigh(ubyte value) {
		fillValue = value & 0x0F;
	}

	__forceinline  void  Process(ubyte* elem) {
		sample_formats::PackedSample_UByte44::Pack0(*elem, fillValue);
	}
};


//====================================================================
struct VW_COMMON_API Process_Fill1NibbleLow : public ElementProcessorInplace<ubyte>
{
	ubyte  fillValue;

	Process_Fill1NibbleLow(ubyte value) {
		fillValue = value & 0x0F;
	}

	__forceinline  void  Process(ubyte* elem) {
		sample_formats::PackedSample_UByte44::Pack1(*elem, fillValue);
	}
};


//====================================================================
template<typename UTPackage>
struct Process_FillBits : public ElementProcessorInplace<UTPackage>
{
	UTPackage  fillValue;
	UTPackage  bitMask;
	UTPackage  bitShift;

	Process_FillBits(UTPackage value, UTPackage _bitMask, UTPackage _bitShift) {
		fillValue = value & _bitMask;
		bitMask = _bitMask;
		bitShift = _bitShift;
	}

	__forceinline  void  Process(UTPackage* elem) {
		sample_formats::Pack(*elem, fillValue, bitMask, bitShift);
	}
};


//********************************************************************
}//namespace sample_processors


//********************************************************************
}//namespace vertexware


