

//--------------------------------------------------------------------
#include "DataObjects\SignalBase.h"
#include "BaseTools\Memory.h"


//********************************************************************
namespace vertexware {


//====================================================================
SignalBase::SignalBase(uint _sampleByteSize)
{
	VW_ASSERT(_sampleByteSize > 0, "Invalid sample-byte-size.");

	sampleByteSize = _sampleByteSize;
	numSamples = 0;
	totalByteSize = 0;
	bExternalAllocation = false;
	pSamples = 0;
}


//====================================================================
SignalBase::~SignalBase()
{
	ReleaseSampleStorage();
}


//====================================================================
bool  SignalBase::IsValidSampleFormat(eSampleFormat sampleFormat) const
{
	return (sample_formats::ExtractByteSize(sampleFormat) == sampleByteSize);
}


//====================================================================
void  SignalBase::SetSampleFormat(eSampleFormat _sampleFormat)
{
	VW_ASSERT(IsValidSampleFormat(_sampleFormat), "Setting invalid sample-format.");
	sampleFormat = _sampleFormat;
}


//====================================================================
void  SignalBase::AllocateStorage(uint _numSamples, ubyte* pExternal)
{
	DeallocateStorage();

	numSamples = _numSamples;
	totalByteSize = numSamples * sampleByteSize;

	bExternalAllocation = (pExternal != 0);

	if(bExternalAllocation) {
		pSamples = pExternal;
	}
	else if(numSamples) {
		pSamples = new ubyte[totalByteSize];
	}
}


//====================================================================
void  SignalBase::DeallocateStorage()
{
	totalByteSize = numSamples = 0;

	if(bExternalAllocation) {
		pSamples = 0;
	}
	else {
		SafeDeleteArray(pSamples);
	}
	bExternalAllocation = false;
}


//====================================================================
void  SignalBase::ReserveSampleStorage(uint _numSamples)
{
	if(numSamples != _numSamples) {
		AllocateStorage(_numSamples);
	}
}


//====================================================================
void  SignalBase::ReleaseSampleStorage()
{
	DeallocateStorage();
}


//====================================================================
bool  SignalBase::CheckDimensionMatch(const SignalBase& sig) const
{
	return (numSamples == sig.GetNumSamples());
}


//====================================================================
bool  SignalBase::CheckSampleByteSizeMatch(const SignalBase& sig) const
{
	return (sampleByteSize == sig.GetSampleByteSize());
}


//====================================================================
bool  SignalBase::CheckTotalByteSizeMatch(const SignalBase& sig) const
{
	return (totalByteSize == sig.GetTotalByteSize());
}


//====================================================================
bool  SignalBase::CheckSampleFormatMatch(const SignalBase& sig) const
{
	return (sampleFormat == sig.GetSampleFormat());
}


//====================================================================
SignalBaseFromExternalStorageAdapter::SignalBaseFromExternalStorageAdapter(
													eSampleFormat sampleFormat)
	:
	SignalBase(sample_formats::ExtractByteSize(sampleFormat))
{
	SetSampleFormat(sampleFormat);
}


//====================================================================
void  SignalBaseFromExternalStorageAdapter::SetExternalStorage(
									ubyte* pStorage, uint numSamples)
{
	if(pStorage) {
		VW_ASSERT(numSamples,
				"Invalid argument combination passed to SignalBase adapter.");

		AllocateStorage(numSamples, pStorage);
	}
	else {
		DeallocateStorage();
	}
}


//====================================================================
void  SignalBaseFromExternalStorageAdapter::ReserveSampleStorage(uint)
{
	VW_THROW(InvalidUsageException, "Cannot manage external storage through SignalBase adapter.");
}


//********************************************************************
}//namespace vertexware


