

//--------------------------------------------------------------------
#include "DataObjects\SignalCodecGdiplus.h"
#include "DataObjects\Image.h"
#include "DataObjects\SampleProcessors.h"
#include "DataObjects\ImageOperations.h"
#include "BaseTools\String.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\Memory.h"
#include "BaseTools\FileSystem.h"
#include <list>


//--------------------------------------------------------------------
// GDI+ pixel storage notes.
//
// RGB  pixels: XRGB in registers, BGRX in memory
// RGBA pixels: ARGB in registers, BGRA in memory
//--------------------------------------------------------------------


//====================================================================
#define  CODEC_NAME   "GDI+"
#define  FORMAT_LIST  "tiff jpeg png bmp"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
bool __fastcall TranslateFileFormatToMimeType(const std::string& fileFormat,
													std::string& outMimeType)
{
	using namespace signal_codec_parameters;

	if(fileFormat == Value_ImageFileFormat_TIFF) {
		outMimeType = "image/tiff";
		return true;
	}
	if(fileFormat == Value_ImageFileFormat_JPEG) {
		outMimeType = "image/jpeg";
		return true;
	}
	if(fileFormat == Value_ImageFileFormat_PNG) {
		outMimeType = "image/png";
		return true;
	}
	if(fileFormat == Value_ImageFileFormat_BMP) {
		outMimeType = "image/bmp";
		return true;
	}
	return false;
}


//====================================================================
bool __fastcall GdiplusGetEncoderFromMimeType(const std::string& mimeType, CLSID& outClsid)
{
	// unicode translation
	std::wstring  mimeTypeU;
	if(! AsciiToUnicode(mimeType, mimeTypeU)) return false;

	// obtain available encoder information
	UINT  numEncoders      = 0;
	UINT  encoderArraySize = 0;
	if(Gdiplus::GetImageEncodersSize(&numEncoders, &encoderArraySize) != Gdiplus::Ok) return false;

	Buffer<ubyte>  codecInfoBuffer(encoderArraySize);
	Gdiplus::ImageCodecInfo*  pImageCodecInfo = reinterpret_cast<Gdiplus::ImageCodecInfo*>(codecInfoBuffer.ToPtr());

	if(Gdiplus::GetImageEncoders(numEncoders, encoderArraySize, pImageCodecInfo) != Gdiplus::Ok) return false;

	// search for requested encoder
	for(UINT i = 0; i < numEncoders; i++) {
		if(std::wstring(pImageCodecInfo[i].MimeType) == mimeTypeU) {
			outClsid = pImageCodecInfo[i].Clsid;
			return true;
		}
	}
	// not found
	return false;
}


//====================================================================
// Construction of arguments for Gdiplus::Image::Save()
// Usage pattern:
//     Add, Add...Add, Lock, Pass_to_Gdiplus, Unlock
//====================================================================
class GdiplusEncoderParameterList
{
public:
	// add parameter
	void  Add(const Gdiplus::EncoderParameter& ep) {
		parameters.push_back(ep);
	}

	// add parameter indirect
	void  Add(const GUID& paramGuid, ULONG* addressOfValue)
	{
		Gdiplus::EncoderParameter  ep;
		ep.Guid = paramGuid;
		ep.NumberOfValues = 1;
		ep.Type = Gdiplus::EncoderParameterValueTypeLong;
		ep.Value = addressOfValue;
		Add(ep);
	}

	// construct temporary parameter-list compatible with Gdiplus::Image::Save
	Gdiplus::EncoderParameters*  Lock()
	{
		Unlock();
		if(parameters.empty()) return 0;

		// construct Gdiplus::EncoderParameters
		lockBuffer.Allocate(sizeof(Gdiplus::EncoderParameters) +
							sizeof(Gdiplus::EncoderParameter) * (parameters.size() - 1));
		Gdiplus::EncoderParameters*  p = reinterpret_cast<Gdiplus::EncoderParameters*>(lockBuffer.ToPtr());

		uint counter = 0;
		PARAM_ITER it = parameters.begin();
		while(it != parameters.end()) {
			p->Parameter[counter] = *it++;
			counter++;
		}
		p->Count = counter;
		return p;
	}

	// release temporary parameter-list
	void  Unlock() { lockBuffer.Deallocate(); }


private:
	typedef  std::list<Gdiplus::EncoderParameter>  PARAM_LIST;
	typedef  PARAM_LIST::iterator                  PARAM_ITER;

	PARAM_LIST     parameters;
	Buffer<ubyte>  lockBuffer;
};


//====================================================================
// Examine values supported by Gdiplus encoder-parameter.
//====================================================================
bool __fastcall GdiplusIsSupportedEncoderParameterValueLong(
						const Gdiplus::EncoderParameter& ep, ULONG value)
{
	if(ep.Type != Gdiplus::EncoderParameterValueTypeLong) return false;
	for(uint i = 0; i < ep.NumberOfValues; i++) {
		if(static_cast<ULONG*>(ep.Value)[i] == value) return true;
	}
	return false;
}


//====================================================================
bool __fastcall GdiplusGetEncoderParameterValueLongRange(
						const Gdiplus::EncoderParameter& ep, ULONG& outLow, ULONG& outHigh)
{
	if(ep.Type != Gdiplus::EncoderParameterValueTypeLongRange) return false;
	if(ep.NumberOfValues != 1) return false;
	outLow  = static_cast<ULONG*>(ep.Value)[0];
	outHigh = static_cast<ULONG*>(ep.Value)[1];
	return true;
}


//====================================================================
// Search for parameters supported by Gdiplus encoder.
//====================================================================
class GdiplusEncoderParameterInfo
{
public:
	// cache encoder-parameter information
	GdiplusEncoderParameterInfo(Gdiplus::Image& image, CLSID &encoderClsid)
	{
		if(uint paramListSize = image.GetEncoderParameterListSize(&encoderClsid)) {
			buffer.Allocate(paramListSize);
			image.GetEncoderParameterList(&encoderClsid, paramListSize,
					reinterpret_cast<Gdiplus::EncoderParameters*>(buffer.ToPtr()));
		}
	}

	// search for encoder-parameter, returns 0 if not found
	const Gdiplus::EncoderParameter*  Find(const GUID& paramGuid) const
	{
		const Gdiplus::EncoderParameters* pEncParams =
					reinterpret_cast<const Gdiplus::EncoderParameters*>(buffer.ToPtr());
		if(! pEncParams) return 0;
		for(uint i = 0; i < pEncParams->Count; i++) {
			if(pEncParams->Parameter[i].Guid == paramGuid) {
				return &(pEncParams->Parameter[i]);
			}
		}
		return 0;
	}

	// check parameter-value support
	bool  IsSupportedEncoderParameterValueLong(const GUID& paramGuid, ULONG value) const
	{
		if(const Gdiplus::EncoderParameter* p = Find(paramGuid)) {
			return GdiplusIsSupportedEncoderParameterValueLong(*p, value);
		}
		return false;
	}

	// extract parameter-value range
	bool  GetEncoderParameterValueLongRange(const GUID& paramGuid, ULONG& outLow, ULONG& outHigh) const {
		if(const Gdiplus::EncoderParameter* p = Find(paramGuid)) {
			return GdiplusGetEncoderParameterValueLongRange(*p, outLow, outHigh);
		}
		return false;
	}


private:
	Buffer<ubyte>  buffer;
};


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
ImageEncoderGdiplus::ImageEncoderGdiplus() :
	ImageEncoder(CODEC_NAME, FORMAT_LIST)
{
}


//====================================================================
// Handles only standard sample-formats.
// Image is saved as RGBA if source-image contains alpha and file-format
// supports alpha-channel, otherwise it is saved as RGB.
//====================================================================
bool  ImageEncoderGdiplus::SaveImage(const ImageBase& sourceImage,
									const std::string& fileName,
									const std::string& fileFormat,
									const NamedParameterList& parameterList)
{
	// translate file-format to gdiplus-supported mime-type
	std::string  mimeType;
	if(! TranslateFileFormatToMimeType(fileFormat, mimeType)) return false;

	// check arguments
	if(fileName.empty()) return false;
	if(! sourceImage.GetNumSamples()) return false;

	// check image-object format and convert image to output format
	Gdiplus::BitmapData  bitmapData;
	AutoPtr<ImageBase>   apOutputImage(new Image32);

	bool  bNeedsAlphaReplicate = false;

	switch(sourceImage.GetSampleFormat()) {

		case SampleFormat_L8:
		case SampleFormat_X8R8G8B8:
			bitmapData.PixelFormat = PixelFormat32bppRGB;
			apOutputImage->SetSampleFormat(SampleFormat_X8R8G8B8);
			break;

		case SampleFormat_A8:
			bNeedsAlphaReplicate = true;
			// fall-through

		case SampleFormat_A8L8:
		case SampleFormat_A4R4G4B4:
		case SampleFormat_A1R5G5B5:
		case SampleFormat_A8R8G8B8:
			bitmapData.PixelFormat = PixelFormat32bppARGB;
			apOutputImage->SetSampleFormat(SampleFormat_A8R8G8B8);
			break;

		default:
			// unsupported sample-format
			return false;
	}

	if(! TransferSamples2D(*apOutputImage, sourceImage)) {
		return false;
	}
	if(bNeedsAlphaReplicate) {
		CopySamples_1ByteLTo3BytesBGR(*apOutputImage, 0, *apOutputImage, 3);
	}

	// get gdiplus-encoder handle
	CLSID  encoderClsid;
	if(! GdiplusGetEncoderFromMimeType(mimeType, encoderClsid)) return false;

	// file-name processing
	std::string   fullPathName = file_system::GetFullPathName(fileName, false);
	std::wstring  fullPathNameU;
	if(! AsciiToUnicode(fullPathName, fullPathNameU)) return false;

	// extract export parameters
	using namespace data_codec_parameters;
	using namespace signal_codec_parameters;

	bool  bOverwrite = parameterList.GetValueBool(Param_FileOverwrite, Value_FileOverwrite_Default);
	bool  bUseCompression = parameterList.GetValueBool(Param_UseCompression, Value_UseCompression_Default);

	uint  iQuality = parameterList.GetValueUInteger(Param_Quality, Value_Quality_Default);
	Clamp(iQuality, Value_Quality_Min, Value_Quality_Max);

	// prepare structures for Gdiplus::Bitmap::LockBits
	bitmapData.Width  = apOutputImage->GetNumColumns();
	bitmapData.Height = apOutputImage->GetNumRows();
	bitmapData.Stride = apOutputImage->GetRowByteSize();
	bitmapData.Scan0  = const_cast<ubyte*>(apOutputImage->GetSamplesB());

	// transfer pixels to GDI+
	Gdiplus::Bitmap  bitmap(bitmapData.Width, bitmapData.Height, bitmapData.PixelFormat);
	GdiplusTransferBitmapPixels(bitmap, bitmapData, false);

	// prepare available & applicable encoder parameters
	GdiplusEncoderParameterInfo  epInfo(bitmap, encoderClsid);
	GdiplusEncoderParameterList  epList;

	// color depth (supported by TIFF)
	ULONG  encValColorDepth = apOutputImage->GetSampleByteSize() * 8;
	if(epInfo.IsSupportedEncoderParameterValueLong(Gdiplus::EncoderColorDepth, encValColorDepth)) {
		epList.Add(Gdiplus::EncoderColorDepth, &encValColorDepth);
	}

	// compression type (supported by TIFF)
	ULONG  encValCompressionType = bUseCompression ?
					Gdiplus::EncoderValueCompressionLZW : Gdiplus::EncoderValueCompressionNone;
	if(epInfo.IsSupportedEncoderParameterValueLong(Gdiplus::EncoderCompression, encValCompressionType)) {
		epList.Add(Gdiplus::EncoderCompression, &encValCompressionType);
	}

	// quality (supported by JPEG)
	ULONG  encValQuality;
	ULONG  qualityMin, qualityMax;
	if(epInfo.GetEncoderParameterValueLongRange(Gdiplus::EncoderQuality, qualityMin, qualityMax)) {
		encValQuality = RangeMapInt<ULONG, float>(iQuality, Value_Quality_Min, Value_Quality_Max, qualityMin, qualityMax);
		epList.Add(Gdiplus::EncoderQuality, &encValQuality);
	}

	// implement overwrite (gdiplus does not overwrite files)
	if(! file_system::IfFileExistsDeleteOnOverwrite(fullPathName, bOverwrite)) return false;

	// use gdiplus-encoder to save the image
	return
	(bitmap.Save(fullPathNameU.c_str(), &encoderClsid, epList.Lock()) == Gdiplus::Ok);
}


//====================================================================
ImageDecoderGdiplus::ImageDecoderGdiplus() :
	ImageDecoder(CODEC_NAME, FORMAT_LIST)
{
}


//====================================================================
ImageBase*  ImageDecoderGdiplus::LoadImage(const std::string& fileName,
											const NamedParameterList& parameterList)
{
	// file-name processing
	std::string   fullPathName = file_system::GetFullPathName(fileName, false);
	std::wstring  fullPathNameU;
	if(! AsciiToUnicode(fullPathName, fullPathNameU)) return 0;

	// load file by using gdiplus-decoders (discard non-bitmaps)
	Gdiplus::Bitmap bitmap(fullPathNameU.c_str(), FALSE);
	if(bitmap.GetLastStatus() != Gdiplus::Ok) return 0;
	if(bitmap.GetType() != Gdiplus::ImageTypeBitmap) return 0;

	// extract import parameters
	using namespace signal_codec_parameters;

	bool bOriginBottom = parameterList.GetValueBool(Param_OriginBottom, Value_OriginBottom_Default);

	// prepare structures for Gdiplus::Bitmap::LockBits
	Gdiplus::BitmapData  bitmapData;
	bitmapData.Width   = bitmap.GetWidth();
	bitmapData.Height  = bitmap.GetHeight();

	ImageBase* pImage = 0;

	// Always lock as (A)RGB pixel format, GDI+ then auto-converts the pixels
	// (lock as 16bppGrayScale fails)
	bool bGrayscale = ToBool(bitmap.GetFlags()  &  Gdiplus::ImageFlagsColorSpaceGRAY);
	bool bHasAlpha  = ToBool(bitmap.GetFlags()  &  Gdiplus::ImageFlagsHasAlpha);

	if(! bGrayscale  &&  bHasAlpha) {
		pImage = new Image32(SampleFormat_A8R8G8B8, bitmapData.Height, bitmapData.Width);
		bitmapData.PixelFormat = PixelFormat32bppARGB;
	}
	else {
		pImage = new Image32(SampleFormat_X8R8G8B8, bitmapData.Height, bitmapData.Width);
		bitmapData.PixelFormat = PixelFormat32bppRGB;
	}

	// fill remaining transfer parameters
	bitmapData.Stride = bOriginBottom ? (- pImage->GetRowByteSize()) : pImage->GetRowByteSize();
	bitmapData.Scan0  = bOriginBottom ? pImage->SampleB(pImage->GetNumRows() - 1, 0) : pImage->GetSamplesB();

	// transfer pixels from GDI+
	if(! GdiplusTransferBitmapPixels(bitmap, bitmapData, true)) {
		SafeDelete(pImage);
	}

	// if original file is grayscale, export as 8-bit grayscale
	if(pImage  &&  bGrayscale) {
		ImageBase* pGrayscaleImage = new Image8(SampleFormat_L8);
		if(TransferSamples2D(*pGrayscaleImage, *pImage)) {
			SafeDelete(pImage);
			pImage = pGrayscaleImage;
		}
		else {
			SafeDelete(pGrayscaleImage);
		}
	}

	// return loaded image or 0
	return pImage;
}


//********************************************************************
}//namespace vertexware


