

//--------------------------------------------------------------------
#include "DataObjects\SignalCodecVWDP.h"
#include "DataObjects\SignalIO.h"
#include "BaseTools\FileSystem.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Serializers.h"
#include "BaseTools\Memory.h"


//====================================================================
#define  CODEC_NAME          "VWDP"
#define  FORMAT_LIST_SIGNAL  "vwdp.signal"
#define  FORMAT_LIST_IMAGE   "vwdp.image"
#define  FORMAT_LIST_VOLUME  "vwdp.volume"


//********************************************************************
namespace vertexware {


//====================================================================
SignalEncoderVWDP::SignalEncoderVWDP() :
	SignalEncoder(CODEC_NAME, FORMAT_LIST_SIGNAL)
{
}


//====================================================================
bool  SignalEncoderVWDP::SaveSignal(const SignalBase& sourceSignal,
									const std::string& fileName,
									const std::string& fileFormat,
									const NamedParameterList& parameterList)
{
	// check arguments
	if(fileFormat != signal_codec_parameters::Value_SignalFileFormat_VWDP) return false;
	if(fileName.empty()) return false;
	if(! sourceSignal.GetNumSamples()) return false;

	// file-name processing
	std::string  fullPathName = file_system::GetFullPathName(fileName, false);

	// extract export parameters
	using namespace data_codec_parameters;

	bool  bOverwrite = parameterList.GetValueBool(Param_FileOverwrite, Value_FileOverwrite_Default);
	bool  bBigEndian = parameterList.GetValueBool(Param_ByteOrderBigEndian, Value_ByteOrderBigEndian_Default);

	// implement overwrite
	if(! file_system::IfFileExistsDeleteOnOverwrite(fullPathName, bOverwrite)) return false;

	// write signal-data
	DataPackageSignal  dps;
	dps.SetPackageByteOrderOnWrite(bBigEndian);
	dps.SetPayload(&sourceSignal);
	return dps.SaveToFile(fullPathName, false);
}


//====================================================================
SignalDecoderVWDP::SignalDecoderVWDP() :
	SignalDecoder(CODEC_NAME, FORMAT_LIST_SIGNAL)
{
}


//====================================================================
SignalBase*  SignalDecoderVWDP::LoadSignal(const std::string& fileName,
										const NamedParameterList& parameterList)
{
	// file-name processing
	std::string  fullPathName = file_system::GetFullPathName(fileName, false);

	// read signal-data
	DataPackageSignal  dps;
	if(dps.LoadFromFile(fullPathName, false)) {
		return dps.GetPayload(true);
	}
	return 0; //error
}


//====================================================================
ImageEncoderVWDP::ImageEncoderVWDP() :
	ImageEncoder(CODEC_NAME, FORMAT_LIST_IMAGE)
{
}


//====================================================================
bool  ImageEncoderVWDP::SaveImage(const ImageBase& sourceImage,
									const std::string& fileName,
									const std::string& fileFormat,
									const NamedParameterList& parameterList)
{
	// check arguments
	if(fileFormat != signal_codec_parameters::Value_ImageFileFormat_VWDP) return false;
	if(fileName.empty()) return false;
	if(! sourceImage.GetNumSamples()) return false;

	// file-name processing
	std::string  fullPathName = file_system::GetFullPathName(fileName, false);

	// extract export parameters
	using namespace data_codec_parameters;
	using namespace signal_codec_parameters;

	bool  bOriginBottom = parameterList.GetValueBool(Param_OriginBottom, Value_OriginBottom_Default);
	bool  bOverwrite = parameterList.GetValueBool(Param_FileOverwrite, Value_FileOverwrite_Default);
	bool  bBigEndian = parameterList.GetValueBool(Param_ByteOrderBigEndian, Value_ByteOrderBigEndian_Default);

	// implement overwrite
	if(! file_system::IfFileExistsDeleteOnOverwrite(fullPathName, bOverwrite)) return false;

	// write image-data
	DataPackageImage  dpi;
	dpi.SetPackageByteOrderOnWrite(bBigEndian);
	dpi.SetRowOrderOnWrite(bOriginBottom);
	dpi.SetPayload(&sourceImage);
	return dpi.SaveToFile(fullPathName, false);
}


//====================================================================
ImageDecoderVWDP::ImageDecoderVWDP() :
	ImageDecoder(CODEC_NAME, FORMAT_LIST_IMAGE)
{
}


//====================================================================
ImageBase*  ImageDecoderVWDP::LoadImage(const std::string& fileName,
										const NamedParameterList& parameterList)
{
	// file-name processing
	std::string  fullPathName = file_system::GetFullPathName(fileName, false);

	// extract import parameters
	using namespace signal_codec_parameters;

	bool bOriginBottom = parameterList.GetValueBool(Param_OriginBottom, Value_OriginBottom_Default);

	// read image-data
	DataPackageImage  dpi;
	dpi.SetRowOrderOnRead(bOriginBottom);
	if(dpi.LoadFromFile(fullPathName, false)) {
		return dpi.GetPayload(true);
	}
	return 0; //error
}


//====================================================================
VolumeEncoderVWDP::VolumeEncoderVWDP() :
	VolumeEncoder(CODEC_NAME, FORMAT_LIST_VOLUME)
{
}


//====================================================================
bool  VolumeEncoderVWDP::SaveVolume(const VolumeBase& sourceVolume,
									const std::string& fileName,
									const std::string& fileFormat,
									const NamedParameterList& parameterList)
{
	// check arguments
	if(fileFormat != signal_codec_parameters::Value_VolumeFileFormat_VWDP) return false;
	if(fileName.empty()) return false;
	if(! sourceVolume.GetNumSamples()) return false;

	// file-name processing
	std::string  fullPathName = file_system::GetFullPathName(fileName, false);

	// extract export parameters
	using namespace data_codec_parameters;
	using namespace signal_codec_parameters;

	bool  bOriginBack = parameterList.GetValueBool(Param_OriginBack, Value_OriginBack_Default);
	bool  bOriginBottom = parameterList.GetValueBool(Param_OriginBottom, Value_OriginBottom_Default);
	bool  bOverwrite = parameterList.GetValueBool(Param_FileOverwrite, Value_FileOverwrite_Default);
	bool  bBigEndian = parameterList.GetValueBool(Param_ByteOrderBigEndian, Value_ByteOrderBigEndian_Default);

	// implement overwrite
	if(! file_system::IfFileExistsDeleteOnOverwrite(fullPathName, bOverwrite)) return false;

	// write volume-data
	DataPackageVolume  dpv;
	dpv.SetPackageByteOrderOnWrite(bBigEndian);
	dpv.SetLayerOrderOnWrite(bOriginBack);
	dpv.SetRowOrderOnWrite(bOriginBottom);
	dpv.SetPayload(&sourceVolume);
	return dpv.SaveToFile(fullPathName, false);
}


//====================================================================
VolumeDecoderVWDP::VolumeDecoderVWDP() :
	VolumeDecoder(CODEC_NAME, FORMAT_LIST_VOLUME)
{
}


//====================================================================
VolumeBase*  VolumeDecoderVWDP::LoadVolume(const std::string& fileName,
										const NamedParameterList& parameterList)
{
	// file-name processing
	std::string  fullPathName = file_system::GetFullPathName(fileName, false);

	// extract import parameters
	using namespace signal_codec_parameters;

	bool bOriginBack = parameterList.GetValueBool(Param_OriginBack, Value_OriginBack_Default);
	bool bOriginBottom = parameterList.GetValueBool(Param_OriginBottom, Value_OriginBottom_Default);

	// read volume-data
	DataPackageVolume  dpv;
	dpv.SetLayerOrderOnRead(bOriginBack);
	dpv.SetRowOrderOnRead(bOriginBottom);
	if(dpv.LoadFromFile(fullPathName, false)) {
		return dpv.GetPayload(true);
	}
	return 0; //error
}


//********************************************************************
}//namespace vertexware


