

//--------------------------------------------------------------------
#include "DataObjects\SignalIO.h"
#include "DataObjects\Signal.h"
#include "DataObjects\SignalOperations.h"
#include "DataObjects\Image.h"
#include "DataObjects\ImageOperations.h"
#include "DataObjects\Volume.h"
#include "DataObjects\VolumeOperations.h"
#include "BaseTools\Serializers.h"
#include "BaseTools\Constructs.h"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
const uint  gc_DPHType_Signal                = 0x00000001;
const uint  gc_DPHType_Image                 = 0x00000002;
const uint  gc_DPHType_Volume                = 0x00000003;

const uint  gc_StorageFlags_UnusedMaskSignal = 0xFFFFFFFF;
const uint  gc_StorageFlags_UnusedMaskImage  = 0xFFFFFFFD;
const uint  gc_StorageFlags_UnusedMaskVolume = 0xFFFFFFF9;

const uint  gc_StorageFlags_LayerOrderMask   = 0x00000004;
const uint  gc_LayerOrder_FrontToBack        = 0x00000000;
const uint  gc_LayerOrder_BackToFront        = 0x00000004;

const uint  gc_StorageFlags_RowOrderMask     = 0x00000002;
const uint  gc_RowOrder_TopToBottom          = 0x00000000;
const uint  gc_RowOrder_BottomToTop          = 0x00000002;


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
DataPackageSignal::DataPackageSignal() :
	DataPackage(gc_DPHType_Signal)
{
}


//====================================================================
void  DataPackageSignal::SetPayload(const SignalBase* pSignal)
{
	apSignal.Assign(const_cast<SignalBase*>(pSignal));
	apSignal.Release();
}


//====================================================================
SignalBase*  DataPackageSignal::GetPayload(bool bRelease)
{
	if(bRelease) return apSignal.Release();
	return apSignal.Get();
}


//====================================================================
void  DataPackageSignal::QueryPayloadByteSize(uint& payloadByteSize) const
{
	if(! apSignal) {
		VW_THROW_CN(SerializationException, "Payload not ready.");
	}
	payloadByteSize = apSignal->GetTotalByteSize() + 16;
}


//====================================================================
void  DataPackageSignal::SerializePayload(std::ostream& os) const
{
	if(! apSignal  ||  ! apSignal->GetNumSamples()  ||
		! sample_formats::ExtractByteSize(apSignal->GetSampleFormat()))
	{
		VW_THROW_CN(SerializationException, "Payload not ready.");
	}

	// write signal properties
	os <= apSignal->GetNumSamples();
	os <= apSignal->GetSampleFormat();

	uint storageFlags = 0;
	os <= storageFlags;

	uint reserved = 0;
	os <= reserved;

	// write samples
	WriteByteBlocks(os, apSignal->GetSamplesB(), apSignal->GetSampleByteSize(), apSignal->GetNumSamples());
}


//====================================================================
void  DataPackageSignal::DeserializePayload(std::istream& is, uint payloadByteSize)
{
	apSignal.SafeDelete();

	// extract signal properties
	uint  numSamples = 0;
	is >= numSamples;

	uint  sampleFormat = 0;
	is >= sampleFormat;
	uint  sampleByteSize = sample_formats::ExtractByteSize(sampleFormat);

	uint  storageFlags = 0;
	is >= storageFlags;

	uint  reserved = 0;
	is >= reserved;

	// check signal properties
	if((! numSamples)  ||  (! sampleByteSize)  ||  (reserved != 0)  ||  (storageFlags != 0))
	{
		VW_THROW_CN(DeserializationException, "Corrupted field.");
	}

	// construct signal-object
	AutoPtr<SignalBase>  apTemp(ManufactureSignal(eSampleFormat(sampleFormat), numSamples));
	if(! apTemp) {
		VW_THROW_CN(DeserializationException, "Unsupported sample-byte-size.");
	}

	// read samples
	ReadByteBlock(is, apTemp->GetSamplesB(), apTemp->GetTotalByteSize());

	// transfer ownership
	apSignal = apTemp;
}


//====================================================================
DataPackageImage::DataPackageImage() :
	DataPackage(gc_DPHType_Image)
{
	bRowOrderBottomToTopOnWrite = false;
	bRowOrderBottomToTopOnRead = false;
}


//====================================================================
void  DataPackageImage::SetPayload(const ImageBase* pImage)
{
	apImage.Assign(const_cast<ImageBase*>(pImage));
	apImage.Release();
}


//====================================================================
ImageBase*  DataPackageImage::GetPayload(bool bRelease)
{
	if(bRelease) return apImage.Release();
	return apImage.Get();
}


//====================================================================
void  DataPackageImage::QueryPayloadByteSize(uint& payloadByteSize) const
{
	if(! apImage) {
		VW_THROW_CN(SerializationException, "Payload not ready.");
	}
	payloadByteSize = apImage->GetTotalByteSize() + 20;
}


//====================================================================
void  DataPackageImage::SerializePayload(std::ostream& os) const
{
	if(! apImage  ||  ! apImage->GetNumSamples()  ||
		! sample_formats::ExtractByteSize(apImage->GetSampleFormat()))
	{
		VW_THROW_CN(SerializationException, "Payload not ready.");
	}

	// write image properties
	os <= apImage->GetNumRows();
	os <= apImage->GetNumColumns();
	os <= apImage->GetSampleFormat();

	uint storageFlags = 0;
	storageFlags |= (bRowOrderBottomToTopOnWrite ? gc_RowOrder_BottomToTop : gc_RowOrder_TopToBottom);
	os <= storageFlags;

	uint reserved = 0;
	os <= reserved;

	// write samples
	const ubyte* pRow      = apImage->GetSamplesB();
	int          rowStride = apImage->GetRowByteSize();

	if(bRowOrderBottomToTopOnWrite) {
		pRow += rowStride * (apImage->GetNumRows() - 1);
		FlipSign(rowStride);
	}

	for(int row = 0; row < apImage->GetNumRows(); row++) {
		WriteByteBlocks(os, pRow, apImage->GetSampleByteSize(), apImage->GetNumColumns());
		pRow += rowStride;
	}
}


//====================================================================
void  DataPackageImage::DeserializePayload(std::istream& is, uint payloadByteSize)
{
	apImage.SafeDelete();

	// extract image properties
	uint  numRows = 0;
	is >= numRows;

	uint  numColumns = 0;
	is >= numColumns;

	uint  sampleFormat = 0;
	is >= sampleFormat;
	uint  sampleByteSize = sample_formats::ExtractByteSize(sampleFormat);

	uint  storageFlags = 0;
	is >= storageFlags;

	uint  reserved = 0;
	is >= reserved;

	// check image properties
	if((! numRows)  ||  (! numColumns)  ||  (! sampleByteSize)  ||
		(reserved != 0)  ||  (storageFlags  &  gc_StorageFlags_UnusedMaskImage))
	{
		VW_THROW_CN(DeserializationException, "Corrupted field.");
	}

	// construct image-object
	AutoPtr<ImageBase>  apTemp(ManufactureImage(eSampleFormat(sampleFormat), numRows, numColumns));
	if(! apTemp) {
		VW_THROW_CN(DeserializationException, "Unsupported sample-byte-size.");
	}

	// determine effective row order
	bool  bReverseRowOrder = false;

	switch(storageFlags  &  gc_StorageFlags_RowOrderMask) {
		case gc_RowOrder_TopToBottom:
			if(bRowOrderBottomToTopOnRead) bReverseRowOrder = true;
			break;

		case gc_RowOrder_BottomToTop:
			if(! bRowOrderBottomToTopOnRead) bReverseRowOrder = true;
			break;
	}

	// read samples
	ubyte* pRow      = apTemp->GetSamplesB();
	int    rowStride = apTemp->GetRowByteSize();

	if(bReverseRowOrder) {
		pRow += rowStride * (apTemp->GetNumRows() - 1);
		FlipSign(rowStride);
	}

	for(int row = 0; row < apTemp->GetNumRows(); row++) {
		ReadByteBlock(is, pRow, apTemp->GetRowByteSize());
		pRow += rowStride;
	}

	// transfer ownership
	apImage = apTemp;
}


//====================================================================
DataPackageVolume::DataPackageVolume() :
	DataPackage(gc_DPHType_Volume)
{
	bLayerOrderBackToFrontOnWrite = false;
	bLayerOrderBackToFrontOnRead = false;

	bRowOrderBottomToTopOnWrite = false;
	bRowOrderBottomToTopOnRead = false;
}


//====================================================================
void  DataPackageVolume::SetPayload(const VolumeBase* pVolume)
{
	apVolume.Assign(const_cast<VolumeBase*>(pVolume));
	apVolume.Release();
}


//====================================================================
VolumeBase*  DataPackageVolume::GetPayload(bool bRelease)
{
	if(bRelease) return apVolume.Release();
	return apVolume.Get();
}


//====================================================================
void  DataPackageVolume::QueryPayloadByteSize(uint& payloadByteSize) const
{
	if(! apVolume) {
		VW_THROW(SerializationException, "Payload not ready.");
	}
	payloadByteSize = apVolume->GetTotalByteSize() + 24;
}


//====================================================================
void  DataPackageVolume::SerializePayload(std::ostream& os) const
{
	if(! apVolume  ||  ! apVolume->GetNumSamples()  ||
		! sample_formats::ExtractByteSize(apVolume->GetSampleFormat()))
	{
		VW_THROW(SerializationException, "Payload not ready.");
	}

	// write volume properties
	os <= apVolume->GetNumLayers();
	os <= apVolume->GetNumRows();
	os <= apVolume->GetNumColumns();
	os <= apVolume->GetSampleFormat();

	uint storageFlags = 0;
	storageFlags |= (bLayerOrderBackToFrontOnWrite ? gc_LayerOrder_BackToFront : gc_LayerOrder_FrontToBack);
	storageFlags |= (bRowOrderBottomToTopOnWrite ? gc_RowOrder_BottomToTop : gc_RowOrder_TopToBottom);
	os <= storageFlags;

	uint reserved = 0;
	os <= reserved;

	// write samples
	const ubyte* pLayer      = apVolume->GetSamplesB();
	int          layerStride = apVolume->GetLayerByteSize();
	int          rowStride   = apVolume->GetRowByteSize();

	if(bLayerOrderBackToFrontOnWrite) {
		pLayer += layerStride * (apVolume->GetNumLayers() - 1);
		FlipSign(layerStride);
	}

	if(bRowOrderBottomToTopOnWrite) {
		pLayer += rowStride * (apVolume->GetNumRows() - 1);
		FlipSign(rowStride);
	}

	for(int layer = 0; layer < apVolume->GetNumLayers(); layer++) {
		const ubyte* pRow = pLayer;
		pLayer += layerStride;

		for(int row = 0; row < apVolume->GetNumRows(); row++) {
			WriteByteBlocks(os, pRow, apVolume->GetSampleByteSize(), apVolume->GetNumColumns());
			pRow += rowStride;
		}
	}
}


//====================================================================
void  DataPackageVolume::DeserializePayload(std::istream& is, uint payloadByteSize)
{
	apVolume.SafeDelete();

	// extract volume properties
	uint  numLayers = 0;
	is >= numLayers;

	uint  numRows = 0;
	is >= numRows;

	uint  numColumns = 0;
	is >= numColumns;

	uint  sampleFormat = 0;
	is >= sampleFormat;
	uint  sampleByteSize = sample_formats::ExtractByteSize(sampleFormat);

	uint  storageFlags = 0;
	is >= storageFlags;

	uint  reserved = 0;
	is >= reserved;

	// check image properties
	if((! numLayers)  ||  (! numRows)  ||  (! numColumns)  ||  (! sampleByteSize)  ||
		(reserved != 0)  ||  (storageFlags  &  gc_StorageFlags_UnusedMaskVolume))
	{
		VW_THROW_CN(DeserializationException, "Corrupted field.");
	}

	// construct volume-object
	AutoPtr<VolumeBase>  apTemp(ManufactureVolume(eSampleFormat(sampleFormat), numLayers, numRows, numColumns));
	if(! apTemp) {
		VW_THROW_CN(DeserializationException, "Unsupported sample-byte-size.");
	}

	// determine effective layer order
	bool  bReverseLayerOrder = false;

	switch(storageFlags  &  gc_StorageFlags_LayerOrderMask) {
		case gc_LayerOrder_FrontToBack:
			if(bLayerOrderBackToFrontOnRead) bReverseLayerOrder = true;
			break;

		case gc_LayerOrder_BackToFront:
			if(! bLayerOrderBackToFrontOnRead) bReverseLayerOrder = true;
			break;
	}

	// determine effective row order
	bool  bReverseRowOrder = false;

	switch(storageFlags  &  gc_StorageFlags_RowOrderMask) {
		case gc_RowOrder_TopToBottom:
			if(bRowOrderBottomToTopOnRead) bReverseRowOrder = true;
			break;

		case gc_RowOrder_BottomToTop:
			if(! bRowOrderBottomToTopOnRead) bReverseRowOrder = true;
			break;
	}

	// read samples
	ubyte* pLayer      = apTemp->GetSamplesB();
	int    layerStride = apTemp->GetLayerByteSize();
	int    rowStride   = apTemp->GetRowByteSize();

	if(bReverseLayerOrder) {
		pLayer += layerStride * (apTemp->GetNumLayers() - 1);
		FlipSign(layerStride);
	}

	if(bReverseRowOrder) {
		pLayer += rowStride * (apTemp->GetNumRows() - 1);
		FlipSign(rowStride);
	}

	for(int layer = 0; layer < apTemp->GetNumLayers(); layer++) {
		ubyte* pRow = pLayer;
		pLayer += layerStride;

		for(int row = 0; row < apTemp->GetNumRows(); row++) {
			ReadByteBlock(is, pRow, apTemp->GetRowByteSize());
			pRow += rowStride;
		}
	}

	// transfer ownership
	apVolume = apTemp;
}


//====================================================================
VW_DEFINE_CONSTANT( uint, DataCategory_Signal, 0x00000001 );
VW_DEFINE_CONSTANT( uint, DataCategory_Image,  0x00000002 );
VW_DEFINE_CONSTANT( uint, DataCategory_Volume, 0x00000003 );


//********************************************************************
namespace signal_codec_parameters {


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_OriginBottom, "OriginBottom" );
VW_DEFINE_CONSTANT( bool, Value_OriginBottom_Default, false );


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_OriginBack, "OriginBack" );
VW_DEFINE_CONSTANT( bool, Value_OriginBack_Default, false );


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_SignalFileFormat, "SignalFileFormat" );
VW_DEFINE_CONSTANT( const char*, Value_SignalFileFormat_VWDP, "VWDP" );
VW_DEFINE_CONSTANT( const char*, Value_SignalFileFormat_Default, Value_SignalFileFormat_VWDP );


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_ImageFileFormat, "ImageFileFormat" );
VW_DEFINE_CONSTANT( const char*, Value_ImageFileFormat_TIFF, "TIFF" );
VW_DEFINE_CONSTANT( const char*, Value_ImageFileFormat_JPEG, "JPEG" );
VW_DEFINE_CONSTANT( const char*, Value_ImageFileFormat_PNG, "PNG" );
VW_DEFINE_CONSTANT( const char*, Value_ImageFileFormat_BMP, "BMP" );
VW_DEFINE_CONSTANT( const char*, Value_ImageFileFormat_VWDP, "VWDP" );
VW_DEFINE_CONSTANT( const char*, Value_ImageFileFormat_Default, Value_ImageFileFormat_JPEG );


//====================================================================
VW_DEFINE_CONSTANT( const char*, Param_VolumeFileFormat, "VolumeFileFormat" );
VW_DEFINE_CONSTANT( const char*, Value_VolumeFileFormat_VWDP, "VWDP" );
VW_DEFINE_CONSTANT( const char*, Value_VolumeFileFormat_Default, Value_VolumeFileFormat_VWDP );


//********************************************************************
}//namespace signal_codec_parameters


//====================================================================
void  SignalCodecParamFactory::FillParameterList(uint dataCategory,
								bool bEncoder, NamedParameterList& npl) const
{
	using namespace signal_codec_parameters;
	npl.Insert(NamedParameter(Param_OriginBottom, bBottomToTop));
	npl.Insert(NamedParameter(Param_OriginBack, bBackToFront));
}


//====================================================================
SignalEncoder::SignalEncoder(const std::string& encoderName, const std::string& formatList) :
	DataEncoder(DataCategory_Signal, "SignalEncoder/" + encoderName, formatList)
{
}


//====================================================================
SignalDecoder::SignalDecoder(const std::string& decoderName, const std::string& formatList) :
	DataDecoder(DataCategory_Signal, "SignalDecoder/" + decoderName, formatList)
{
}


//====================================================================
bool __fastcall LoadSignalAs(SignalBase& destinationSignal, const std::string& fileName,
							const NamedParameterList* pParameterList)
{
	AutoPtr<SignalBase> apLoadedSignal(LoadSignal(fileName, pParameterList));
	if(apLoadedSignal) {
		return TransferSamples1D(destinationSignal, *apLoadedSignal);
	}
	return false;
}


//====================================================================
ImageEncoder::ImageEncoder(const std::string& encoderName, const std::string& formatList) :
	DataEncoder(DataCategory_Image, "ImageEncoder/" + encoderName, formatList)
{
}


//====================================================================
ImageDecoder::ImageDecoder(const std::string& decoderName, const std::string& formatList) :
	DataDecoder(DataCategory_Image, "ImageDecoder/" + decoderName, formatList)
{
}


//====================================================================
bool __fastcall LoadImageAs(ImageBase& destinationImage, const std::string& fileName,
							const NamedParameterList* pParameterList)
{
	AutoPtr<ImageBase> apLoadedImage(LoadImage(fileName, pParameterList));
	if(apLoadedImage) {
		return TransferSamples2D(destinationImage, *apLoadedImage);
	}
	return false;
}


//====================================================================
VolumeEncoder::VolumeEncoder(const std::string& encoderName, const std::string& formatList) :
	DataEncoder(DataCategory_Volume, "VolumeEncoder/" + encoderName, formatList)
{
}


//====================================================================
VolumeDecoder::VolumeDecoder(const std::string& decoderName, const std::string& formatList) :
	DataDecoder(DataCategory_Volume, "VolumeDecoder/" + decoderName, formatList)
{
}


//====================================================================
bool __fastcall LoadVolumeAs(VolumeBase& destinationVolume, const std::string& fileName,
								const NamedParameterList* pParameterList)
{
	AutoPtr<VolumeBase> apLoadedVolume(LoadVolume(fileName, pParameterList));
	if(apLoadedVolume) {
		return TransferSamples3D(destinationVolume, *apLoadedVolume);
	}
	return false;
}


//====================================================================
void __fastcall RegisterDefaultSignalFileExtensions()
{
	using namespace signal_codec_parameters;
	DataCodecManager& dcm = DataCodecManager::Instance();

	dcm.SetDefaultFileFormatExtension(DataCategory_Signal, Value_SignalFileFormat_VWDP, "vwdp.signal");
}


//====================================================================
void __fastcall RegisterDefaultImageFileExtensions()
{
	using namespace signal_codec_parameters;
	DataCodecManager& dcm = DataCodecManager::Instance();

	dcm.SetDefaultFileFormatExtension(DataCategory_Image, Value_ImageFileFormat_TIFF, "tiff");
	dcm.SetDefaultFileFormatExtension(DataCategory_Image, Value_ImageFileFormat_JPEG, "jpeg");
	dcm.SetDefaultFileFormatExtension(DataCategory_Image, Value_ImageFileFormat_PNG, "png");
	dcm.SetDefaultFileFormatExtension(DataCategory_Image, Value_ImageFileFormat_BMP, "bmp");
	dcm.SetDefaultFileFormatExtension(DataCategory_Image, Value_ImageFileFormat_VWDP, "vwdp.image");
}


//====================================================================
void __fastcall RegisterDefaultVolumeFileExtensions()
{
	using namespace signal_codec_parameters;
	DataCodecManager& dcm = DataCodecManager::Instance();

	dcm.SetDefaultFileFormatExtension(DataCategory_Volume, Value_VolumeFileFormat_VWDP, "vwdp.volume");
}


//********************************************************************
}//namespace vertexware


