

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "DataObjects\DataCodecManager.h"
#include "DataObjects\DataPackage.h"
#include "DataObjects\SignalBase.h"
#include "DataObjects\ImageBase.h"
#include "DataObjects\VolumeBase.h"
#include "BaseTools\Memory.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Signal-package specification:
//        |                        |
// Bytes  |     Contents           |        Details
//        |                        |
//    44  |  data-package-header   |  payload-type = 0x00000001
//     4  |  number-of-samples     |  uint >= 1
//     4  |  sample-format         |  uint, msb  8-bits = sample-byte-size >= 1,
//        |                        |        lsb 24-bits = component-layout
//     4  |  storage-flags         |  uint, must be 0
//     4  |  reserved              |  uint, must be 0
// size0  |  samples               |  size0 = number-of-samples x sample-size,
//        |                        |          byte-order specified in package-header
//        |                        |          is applied per-sample
//    32  |  data-package-trailer  |
//        |                        |
//    16 + size0 bytes payload
//    92 + size0 bytes total
//
// Standard sample-formats can be found in DataObjects\Samples.h
//====================================================================
class VW_COMMON_API DataPackageSignal : public DataPackage
{
public:
	DataPackageSignal();

	void         SetPayload(const SignalBase* pSignal = 0);
	SignalBase*  GetPayload(bool bRelease = true);


protected:
	// data-package implementation
	virtual  void  QueryPayloadByteSize(uint& payloadByteSize) const;
	virtual  void  SerializePayload(std::ostream& os) const;
	virtual  void  DeserializePayload(std::istream& is, uint payloadByteSize);


private:
	AutoPtr<SignalBase>  apSignal;
};


//====================================================================
// Image-package specification:
//        |                        |
// Bytes  |     Contents           |        Details
//        |                        |
//    44  |  data-package-header   |  payload-type = 0x00000002
//     4  |  number-of-rows        |  uint >= 1
//     4  |  number-of-columns     |  uint >= 1
//     4  |  sample-format         |  uint, msb  8-bits = sample-byte-size >= 1,
//        |                        |        lsb 24-bits = component-layout
//     4  |  storage-flags         |  uint, bit0 = row-order (0 = top-to-bottom, 1 = bottom-to-top)
//        |                        |        other bits must be 0
//     4  |  reserved              |  uint, must be 0
// size0  |  samples               |  size0 = number-of-rows x number-of-columns x sample-size,
//        |                        |          byte-order specified in package-header
//        |                        |          is applied per-sample
//    32  |  data-package-trailer  |
//        |                        |
//    20 + size0 bytes payload
//    96 + size0 bytes total
//
// Standard sample-formats can be found in DataObjects\Samples.h
//====================================================================
class VW_COMMON_API DataPackageImage : public DataPackage
{
public:
	// defaults: row-order top-to-bottom
	DataPackageImage();

	void  SetRowOrderOnWrite(bool bBottomToTop) { bRowOrderBottomToTopOnWrite = bBottomToTop; }
	void  SetRowOrderOnRead(bool bBottomToTop) { bRowOrderBottomToTopOnRead = bBottomToTop; }

	void        SetPayload(const ImageBase* pImage = 0);
	ImageBase*  GetPayload(bool bRelease = true);


protected:
	// data-package implementation
	virtual  void  QueryPayloadByteSize(uint& payloadByteSize) const;
	virtual  void  SerializePayload(std::ostream& os) const;
	virtual  void  DeserializePayload(std::istream& is, uint payloadByteSize);


private:
	bool  bRowOrderBottomToTopOnWrite;
	bool  bRowOrderBottomToTopOnRead;

	AutoPtr<ImageBase>  apImage;
};


//====================================================================
// Volume-package specification:
//        |                        |
// Bytes  |     Contents           |        Details
//        |                        |
//    44  |  data-package-header   |  payload-type = 0x00000003
//     4  |  number-of-layers      |  uint >= 1
//     4  |  number-of-rows        |  uint >= 1
//     4  |  number-of-columns     |  uint >= 1
//     4  |  sample-format         |  uint, msb  8-bits = sample-byte-size >= 1,
//        |                        |        lsb 24-bits = component-layout
//     4  |  storage-flags         |  uint, bit1 = row-order (0 = top-to-bottom, 1 = bottom-to-top)
//        |                        |        bit2 = layer-order (0 = front-to-back, 1 = back-to-front)
//        |                        |        other bits must be 0
//     4  |  reserved              |  uint, must be 0
// size0  |  samples               |  size0 = number-of-layers x number-of-rows x
//        |                        |          number-of-columns x sample-size,
//        |                        |          byte-order specified in package-header
//        |                        |          is applied per-sample
//    32  |  data-package-trailer  |
//        |                        |
//    24 + size0 bytes payload
//   100 + size0 bytes total
//
// Standard sample-formats can be found in DataObjects\Samples.h
//====================================================================
class VW_COMMON_API DataPackageVolume : public DataPackage
{
public:
	// defaults: layer-order front-to-back, row-order top-to-bottom
	DataPackageVolume();

	void  SetLayerOrderOnWrite(bool bBackToFront) { bLayerOrderBackToFrontOnWrite = bBackToFront; }
	void  SetLayerOrderOnRead(bool bBackToFront) { bLayerOrderBackToFrontOnRead = bBackToFront; }

	void  SetRowOrderOnWrite(bool bBottomToTop) { bRowOrderBottomToTopOnWrite = bBottomToTop; }
	void  SetRowOrderOnRead(bool bBottomToTop) { bRowOrderBottomToTopOnRead = bBottomToTop; }

	void         SetPayload(const VolumeBase* pVolume = 0);
	VolumeBase*  GetPayload(bool bRelease = true);


protected:
	// data-package implementation
	virtual  void  QueryPayloadByteSize(uint& payloadByteSize) const;
	virtual  void  SerializePayload(std::ostream& os) const;
	virtual  void  DeserializePayload(std::istream& is, uint payloadByteSize);


private:
	bool  bLayerOrderBackToFrontOnWrite;
	bool  bLayerOrderBackToFrontOnRead;

	bool  bRowOrderBottomToTopOnWrite;
	bool  bRowOrderBottomToTopOnRead;

	AutoPtr<VolumeBase>  apVolume;
};


//====================================================================
// Signal - 0x00000001
// Image  - 0x00000002
// Volume - 0x00000003
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, DataCategory_Signal );
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, DataCategory_Image );
VW_COMMON_API  VW_DECLARE_CONSTANT( uint, DataCategory_Volume );


//********************************************************************
namespace signal_codec_parameters {


//====================================================================
// Name...: OriginBottom
// Type...: bool
// Range..: NA
// Default: false
// Support: Supported by all image-decoders & volume-decoders.
//          Optionally supported by image-encoders & volume-encoders.
// Notes..: If <true> bottom-row of the signal is stored first,
//          otherwise top-row is stored first.
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_OriginBottom );
VW_COMMON_API  VW_DECLARE_CONSTANT( bool, Value_OriginBottom_Default );


//====================================================================
// Name...: OriginBack
// Type...: bool
// Range..: NA
// Default: false
// Support: Supported by all volume-decoders.
//          Optionally supported by volume-encoders.
// Notes..: If <true> back-layer of the signal is stored first,
//          otherwise front-layer is stored first.
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_OriginBack );
VW_COMMON_API  VW_DECLARE_CONSTANT( bool, Value_OriginBack_Default );


//====================================================================
// Name...: SignalFileFormat
// Type...: string
// Range..: NA
// Default: VWDP
// Support: Signal-codec can support any number of formats
// Notes..: NA
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_SignalFileFormat );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_SignalFileFormat_VWDP );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_SignalFileFormat_Default );


//====================================================================
// Name...: ImageFileFormat
// Type...: string
// Range..: NA
// Default: TIFF
// Support: Image-codec can support any number of formats
// Notes..: NA
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_ImageFileFormat );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_ImageFileFormat_TIFF );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_ImageFileFormat_JPEG );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_ImageFileFormat_PNG );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_ImageFileFormat_BMP );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_ImageFileFormat_VWDP );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_ImageFileFormat_Default );


//====================================================================
// Name...: VolumeFileFormat
// Type...: string
// Range..: NA
// Default: VWDP
// Support: Volume-codec can support any number of formats
// Notes..: NA
//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Param_VolumeFileFormat );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_VolumeFileFormat_VWDP );
VW_COMMON_API  VW_DECLARE_CONSTANT( const char*, Value_VolumeFileFormat_Default );


//********************************************************************
}//namespace signal_codec_parameters


//====================================================================
// Signal-codec parameter factory.
//====================================================================
class VW_COMMON_API SignalCodecParamFactory : public CodecParameterFactory
{
private:
	bool bBottomToTop;
	bool bBackToFront;

public:
	SignalCodecParamFactory(bool _bBottomToTop = false, bool _bBackToFront = false)
	{
		bBottomToTop = _bBottomToTop;
		bBackToFront = _bBackToFront;
	}
	virtual void FillParameterList(uint dataCategory, bool bEncoder, NamedParameterList& npl) const;
};


//====================================================================
// Base class for signal-encoders.
//====================================================================
class VW_COMMON_API SignalEncoder : public DataEncoder
{
protected:
	SignalEncoder(const std::string& encoderName,
				const std::string& formatList);

public:
	virtual  bool  SaveData(const void* data, const std::string& fileName,
							const std::string& fileFormat, const NamedParameterList& parameterList) {
		return  SaveSignal(*reinterpret_cast<const SignalBase*>(data), fileName, fileFormat, parameterList);
	}

	virtual  bool  SaveSignal(const SignalBase& sourceSignal,
							const std::string& fileName,
							const std::string& fileFormat,
							const NamedParameterList& parameterList) = 0;
};


//====================================================================
// Base class for signal-decoders.
//====================================================================
class VW_COMMON_API SignalDecoder : public DataDecoder
{
protected:
	SignalDecoder(const std::string& decoderName,
				const std::string& formatList);

public:
	virtual  void*  LoadData(const std::string& fileName,
							const NamedParameterList& parameterList) {
		return LoadSignal(fileName, parameterList);
	}

	virtual  SignalBase*  LoadSignal(const std::string& fileName,
									const NamedParameterList& parameterList) = 0;
};


//====================================================================
// Save signal through registered encoders.
//====================================================================
VW_COMMON_API __forceinline
bool __fastcall SaveSignal(const SignalBase& sourceSignal,
						const std::string& fileName,
						const std::string& fileFormat = signal_codec_parameters::Value_SignalFileFormat_Default,
						const NamedParameterList* pParameterList = 0)
{
	return  DataCodecManager::Instance().SaveData(DataCategory_Signal, &sourceSignal,
											fileName, fileFormat, pParameterList);
}


//====================================================================
// Load signal through registered decoders.
//====================================================================
VW_COMMON_API __forceinline
SignalBase* __fastcall LoadSignal(const std::string& fileName,
								const NamedParameterList* pParameterList = 0)
{
	return  reinterpret_cast<SignalBase*>(DataCodecManager::Instance().LoadData(
							DataCategory_Signal, fileName, pParameterList));
}


//====================================================================
// Loaded signal is auto-converted to <destinationSignal> sample-format.
//====================================================================
VW_COMMON_API
bool __fastcall LoadSignalAs(SignalBase& destinationSignal, const std::string& fileName,
							const NamedParameterList* pParameterList = 0);


//====================================================================
// Base class for image-encoders.
//====================================================================
class VW_COMMON_API ImageEncoder : public DataEncoder
{
protected:
	ImageEncoder(const std::string& encoderName,
				const std::string& formatList);

public:
	virtual  bool  SaveData(const void* data, const std::string& fileName,
							const std::string& fileFormat, const NamedParameterList& parameterList) {
		return  SaveImage(*reinterpret_cast<const ImageBase*>(data), fileName, fileFormat, parameterList);
	}

	virtual  bool  SaveImage(const ImageBase& sourceImage,
							const std::string& fileName,
							const std::string& fileFormat,
							const NamedParameterList& parameterList) = 0;
};


//====================================================================
// Base class for image-decoders.
//====================================================================
class VW_COMMON_API ImageDecoder : public DataDecoder
{
protected:
	ImageDecoder(const std::string& decoderName,
				const std::string& formatList);

public:
	virtual  void*  LoadData(const std::string& fileName,
							const NamedParameterList& parameterList) {
		return LoadImage(fileName, parameterList);
	}

	virtual  ImageBase*  LoadImage(const std::string& fileName,
								const NamedParameterList& parameterList) = 0;
};


//====================================================================
// Save image through registered encoders.
//====================================================================
VW_COMMON_API __forceinline
bool __fastcall SaveImage(const ImageBase& sourceImage,
						const std::string& fileName,
						const std::string& fileFormat = signal_codec_parameters::Value_ImageFileFormat_Default,
						const NamedParameterList* pParameterList = 0)
{
	return  DataCodecManager::Instance().SaveData(DataCategory_Image, &sourceImage,
											fileName, fileFormat, pParameterList);
}


//====================================================================
// Load image through registered decoders.
//====================================================================
VW_COMMON_API __forceinline
ImageBase* __fastcall LoadImage(const std::string& fileName,
								const NamedParameterList* pParameterList = 0)
{
	return  reinterpret_cast<ImageBase*>(DataCodecManager::Instance().LoadData(
							DataCategory_Image, fileName, pParameterList));
}


//====================================================================
// Loaded image is auto-converted to <destinationImage> sample-format.
//====================================================================
VW_COMMON_API
bool __fastcall LoadImageAs(ImageBase& destinationImage, const std::string& fileName,
							const NamedParameterList* pParameterList = 0);


//====================================================================
// Base class for volume-encoders.
//====================================================================
class VW_COMMON_API VolumeEncoder : public DataEncoder
{
protected:
	VolumeEncoder(const std::string& encoderName,
				const std::string& formatList);

public:
	virtual  bool  SaveData(const void* data, const std::string& fileName,
							const std::string& fileFormat, const NamedParameterList& parameterList) {
		return  SaveVolume(*reinterpret_cast<const VolumeBase*>(data), fileName, fileFormat, parameterList);
	}

	virtual  bool  SaveVolume(const VolumeBase& sourceVolume,
							const std::string& fileName,
							const std::string& fileFormat,
							const NamedParameterList& parameterList) = 0;
};


//====================================================================
// Base class for volume-decoders.
//====================================================================
class VW_COMMON_API VolumeDecoder : public DataDecoder
{
protected:
	VolumeDecoder(const std::string& decoderName,
				const std::string& formatList);

public:
	virtual  void*  LoadData(const std::string& fileName,
							const NamedParameterList& parameterList) {
		return LoadVolume(fileName, parameterList);
	}

	virtual  VolumeBase*  LoadVolume(const std::string& fileName,
									const NamedParameterList& parameterList) = 0;
};


//====================================================================
// Save volume through registered encoders.
//====================================================================
VW_COMMON_API __forceinline
bool __fastcall SaveVolume(const VolumeBase& sourceVolume,
						const std::string& fileName,
						const std::string& fileFormat = signal_codec_parameters::Value_VolumeFileFormat_Default,
						const NamedParameterList* pParameterList = 0)
{
	return  DataCodecManager::Instance().SaveData(DataCategory_Volume, &sourceVolume,
											fileName, fileFormat, pParameterList);
}


//====================================================================
// Load volume through registered decoders.
//====================================================================
VW_COMMON_API __forceinline
VolumeBase* __fastcall LoadVolume(const std::string& fileName,
							const NamedParameterList* pParameterList = 0)
{
	return  reinterpret_cast<VolumeBase*>(DataCodecManager::Instance().LoadData(
							DataCategory_Volume, fileName, pParameterList));
}


//====================================================================
// Loaded volume is auto-converted to <destinationVolume> sample-format.
//====================================================================
VW_COMMON_API
bool __fastcall LoadVolumeAs(VolumeBase& destinationVolume, const std::string& fileName,
							const NamedParameterList* pParameterList = 0);


//====================================================================
// Register default file-extensions with data-codec-manager.
//====================================================================
VW_COMMON_API void __fastcall RegisterDefaultSignalFileExtensions();
VW_COMMON_API void __fastcall RegisterDefaultImageFileExtensions();
VW_COMMON_API void __fastcall RegisterDefaultVolumeFileExtensions();


//********************************************************************
}//namespace vertexware


