

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "DataObjects\SignalBase.h"


//********************************************************************
namespace vertexware {


//====================================================================
// VolumeBase provides sample-type independent operations.
// Size can be 0x0x0, but samples must not be accessed in that case.
//====================================================================
class VW_COMMON_API VolumeBase : virtual public SignalBase
{
public:
	// storage configuration
	virtual  void  ReserveSampleStorage(uint numLayers, uint numRows, uint numColumns);
	virtual  void  ReserveSampleStorage(uint numSamples);
	virtual  void  ReleaseSampleStorage();

	// storage info
	__forceinline  uint  GetNumLayers() const { return numLayers; }
	__forceinline  uint  GetNumRows() const { return numRows; }
	__forceinline  uint  GetNumColumns() const { return numColumns; }
	__forceinline  uint  GetNumSamplesPerLayer() const { return numSamplesPerLayer; }
	__forceinline  uint  GetLayerByteSize() const { return layerByteSize; }
	__forceinline  uint  GetRowByteSize() const { return rowByteSize; }

	bool  CheckDimensionMatch(const VolumeBase& vol) const;

	// storage access
	using SignalBase::SampleB;

	__forceinline  ubyte*  SampleB(uint layer, uint row, uint column) {
		CheckAccess3D(layer, row, column);
		return (GetSamplesB() + GetSampleByteSize() *
				(layer * numSamplesPerLayer + row * numColumns + column));
	}
	__forceinline  const ubyte*  SampleB(uint layer, uint row, uint column) const {
		return const_cast<VolumeBase*>(this)->SampleB(layer, row, column);
	}


protected:
	// constructor
	VolumeBase(uint sampleByteSize);

	// debug helpers
	__forceinline  void  CheckAccess3D(uint layer, uint row, uint column) const {
		VW_ASSERT(GetNumSamples(), "Accessing 0x0x0 volume.");
		VW_ASSERT(layer < numLayers, "Layer-index out of range.");
		VW_ASSERT(row < numRows, "Row-index out of range.");
		VW_ASSERT(column < numColumns,  "Column-index out of range.");
	}

	// low-level setup used by reserve/release overrides
	void  SetupInternalVolumeProperties(uint numLayers, uint numRows, uint numColumns);


private:
	uint  numLayers;
	uint  numRows;
	uint  numColumns;
	uint  numSamplesPerLayer;
	uint  rowByteSize;
	uint  layerByteSize;
};


//====================================================================
// Adapts externally allocated storage so that it appears as VolumeBase.
// Reserve is not allowed. External storage must be tightly packed.
//====================================================================
class VW_COMMON_API VolumeBaseFromExternalStorageAdapter : public VolumeBase
{
public:
	VolumeBaseFromExternalStorageAdapter(eSampleFormat sampleFormat);

	void  SetExternalStorage(ubyte* pStorage, uint numLayers, uint numRows, uint numColumns);

	// storage configuration (reserve throws exception if called)
	virtual  void  ReserveSampleStorage(uint, uint, uint);
	virtual  void  ReserveSampleStorage(uint);
	virtual  void  ReleaseSampleStorage() {}
};


//********************************************************************
}//namespace vertexware


