

//--------------------------------------------------------------------
#include "DataObjects\VolumeOperations.h"
#include "DataObjects\SignalOperations.h"
#include "DataObjects\SampleProcessors.h"
#include "MathTools\MathBase.h"
#include "BaseTools\ElementProcessors.h"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
template<typename TElementProcessor>
__forceinline
void __fastcall Call_ResampleElements3D_Nearest_ubyte_ubyte(
									TElementProcessor& copyProcessor,
									VolumeBase& dst, const VolumeBase& src)
{
	ResampleElements3D_Nearest<ubyte, ubyte>(
			copyProcessor,
			dst.GetSamplesB(), dst.GetLayerByteSize(),
			dst.GetRowByteSize(), dst.GetSampleByteSize(),
			dst.GetNumLayers(), dst.GetNumRows(), dst.GetNumColumns(),
			src.GetSamplesB(), src.GetLayerByteSize(),
			src.GetRowByteSize(), src.GetSampleByteSize(),
			src.GetNumLayers(), src.GetNumRows(), src.GetNumColumns());
}


//====================================================================
template<typename TLinearInterpolator3D>
__forceinline
void __fastcall Call_ResampleElements3D_Linear_PackedSample_UShort(
									TLinearInterpolator3D& interpolator,
									VolumeBase& dst, const VolumeBase& src)
{
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.GetSampleByteSize() == 2, "Invalid sample-format size.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");

	ResampleElements3D_Linear<ubyte, ubyte>(
			interpolator,
			dst.GetSamplesB(), dst.GetLayerByteSize(),
			dst.GetRowByteSize(), dst.GetSampleByteSize(),
			dst.GetNumLayers(), dst.GetNumRows(), dst.GetNumColumns(),
			src.GetSamplesB(), src.GetLayerByteSize(),
			src.GetRowByteSize(), src.GetSampleByteSize(),
			src.GetNumLayers(), src.GetNumRows(), src.GetNumColumns());
}


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
VolumeBase* __fastcall ManufactureVolume(eSampleFormat sampleFormat,
									uint numLayers, uint numRows, uint numColumns)
{
	switch(sample_formats::ExtractByteSize(sampleFormat))
	{
		case 1: return new Volume8(sampleFormat, numLayers, numRows, numColumns);
		case 2: return new Volume16(sampleFormat, numLayers, numRows, numColumns);
		case 3: return new Volume24(sampleFormat, numLayers, numRows, numColumns);
		case 4: return new Volume32(sampleFormat, numLayers, numRows, numColumns);
		case 6: return new Volume48(sampleFormat, numLayers, numRows, numColumns);
		case 8: return new Volume64(sampleFormat, numLayers, numRows, numColumns);
		case 12: return new Volume96(sampleFormat, numLayers, numRows, numColumns);
		case 16: return new Volume128(sampleFormat, numLayers, numRows, numColumns);
	}
	return 0;
}


//====================================================================
VolumeBase* __fastcall CloneVolume(const VolumeBase& src)
{
	VolumeBase* dst = ManufactureVolume(src.GetSampleFormat(),
						src.GetNumLayers(), src.GetNumRows(), src.GetNumColumns());
	if(! dst) return 0;
	CopySamples3D(*dst, src);
	return dst;
}


//====================================================================
void __fastcall CopySamples3D(VolumeBase& dst, const VolumeBase& src,
						uint layerCount, uint rowCount, uint columnCount,
						uint dstLayerOffset, uint dstRowOffset, uint dstColumnOffset,
						uint srcLayerOffset, uint srcRowOffset, uint srcColumnOffset,
						bool bFlipLayers, bool bFlipRows)
{
	if(! layerCount) layerCount = src.GetNumLayers();
	if(! rowCount) rowCount = src.GetNumRows();
	if(! columnCount) columnCount = src.GetNumColumns();

	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dstLayerOffset + layerCount <= dst.GetNumLayers(), "Destination-offset-z/layer-count out of range.");
	VW_ASSERT(dstRowOffset + rowCount <= dst.GetNumRows(), "Destination-offset-y/row-count out of range.");
	VW_ASSERT(dstColumnOffset + columnCount <= dst.GetNumColumns(), "Destination-offset-x/column-count out of range.");
	VW_ASSERT(srcLayerOffset + layerCount <= src.GetNumLayers(), "Source-offset-z/layer-count out of range.");
	VW_ASSERT(srcRowOffset + rowCount <= src.GetNumRows(), "Source-offset-y/row-count out of range.");
	VW_ASSERT(srcColumnOffset + columnCount <= src.GetNumColumns(), "Source-offset-x/column-count out of range.");

	// burst copy
	if(! bFlipLayers && ! bFlipRows &&
		rowCount == dst.GetNumRows() && rowCount == src.GetNumRows() &&
		columnCount == dst.GetNumColumns() && columnCount == src.GetNumColumns())
	{
		::CopyMemory(dst.SampleB(dstLayerOffset, 0, 0),
					src.SampleB(srcLayerOffset, 0, 0),
					src.GetLayerByteSize() * layerCount);
	}
	// loop copy
	else {
		int   srcLayer       = bFlipLayers ? (srcLayerOffset + layerCount - 1) : srcLayerOffset;
		int   srcLayerInc    = bFlipLayers ? -1 : 1;
		int   srcRowStart    = bFlipRows ? (srcRowOffset + rowCount - 1) : srcRowOffset;
		int   srcRowInc      = bFlipRows ? -1 : 1;

		uint  dstLayer       = dstLayerOffset;
		uint  dstLayerEnd    = dstLayer + layerCount;
		uint  dstRowEnd      = dstRowOffset + rowCount;
		uint  numBytesPerRow = src.GetSampleByteSize() * columnCount;

		while(dstLayer < dstLayerEnd) {

			int   srcRow = srcRowStart;
			uint  dstRow = dstRowOffset;

			while(dstRow < dstRowEnd) {
				::CopyMemory(dst.SampleB(dstLayer, dstRow, dstColumnOffset),
							src.SampleB(srcLayer, srcRow, srcColumnOffset),
							numBytesPerRow);

				srcRow += srcRowInc;
				dstRow++;
			}
			srcLayer += srcLayerInc;
			dstLayer++;
		}
	}
}


//====================================================================
bool __fastcall TransferSamples3D(VolumeBase& dst, const VolumeBase& src, bool bAutoResize)
{
	SignalOperationsManager::TransferSamplesHandler* tsh =
		SignalOperationsManager::Instance().GetTransferHandler(dst.GetSampleFormat(), src.GetSampleFormat());

	if(! tsh) return false; //not supported

	if(bAutoResize) {
		dst.ReserveSampleStorage(src.GetNumLayers(), src.GetNumRows(), src.GetNumColumns());
	}
	VW_ASSERT(dst.CheckDimensionMatch(src), "Dimensions mismatch.");

	TransferSamples1D(dst, src, *tsh, false);
	return true;
}


//====================================================================
void __fastcall Resample3D_Nearest(VolumeBase& dst, const VolumeBase& src)
{
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");

	switch(dst.GetSampleByteSize()) {
		case 1:
			Call_ResampleElements3D_Nearest_ubyte_ubyte(
				element_processors::Copy_1<ubyte>(), dst, src);
			break;

		case 2:
			Call_ResampleElements3D_Nearest_ubyte_ubyte(
				element_processors::Copy_1<ushort>(), dst, src);
			break;

		case 4:
			Call_ResampleElements3D_Nearest_ubyte_ubyte(
				element_processors::Copy_1<uint>(), dst, src);
			break;

		default:
			Call_ResampleElements3D_Nearest_ubyte_ubyte(
				element_processors::Copy_N<ubyte>(dst.GetSampleByteSize()), dst, src);
			break;
	}
}


//====================================================================
void __fastcall Resample3D_Linear_NFloats(VolumeBase& dst, const VolumeBase& src,
								uint componentOffset, uint numComponentsPerSample)
{
	uint numComponents = dst.GetSampleByteSize() / sizeof(float);
	uint byteOffset = componentOffset * sizeof(float);

	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");
	VW_ASSERT(numComponentsPerSample <= numComponents, "Num-components-per-sample out of range.");
	VW_ASSERT(componentOffset <= numComponents - numComponentsPerSample, "Component-offset out of range.");

	ResampleElements3D_Linear<ubyte, ubyte>(
			sample_processors::LinearInterpolator3D_NFloats(numComponentsPerSample),
			dst.GetSamplesB() + byteOffset, dst.GetLayerByteSize(),
			dst.GetRowByteSize(), dst.GetSampleByteSize(),
			dst.GetNumLayers(), dst.GetNumRows(), dst.GetNumColumns(),
			src.GetSamplesB() + byteOffset, src.GetLayerByteSize(),
			src.GetRowByteSize(), src.GetSampleByteSize(),
			src.GetNumLayers(), src.GetNumRows(), src.GetNumColumns());
}


//====================================================================
void __fastcall Resample3D_Linear_NBytes(VolumeBase& dst, const VolumeBase& src,
										uint byteOffset, uint numBytesPerSample)
{
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");
	VW_ASSERT(numBytesPerSample <= dst.GetSampleByteSize(), "Num-bytes-per-sample out of range.");
	VW_ASSERT(byteOffset <= dst.GetSampleByteSize() - numBytesPerSample, "Byte-offset out of range.");

	ResampleElements3D_Linear<ubyte, ubyte>(
			sample_processors::LinearInterpolator3D_NBytes(numBytesPerSample),
			dst.GetSamplesB() + byteOffset, dst.GetLayerByteSize(),
			dst.GetRowByteSize(), dst.GetSampleByteSize(),
			dst.GetNumLayers(), dst.GetNumRows(), dst.GetNumColumns(),
			src.GetSamplesB() + byteOffset, src.GetLayerByteSize(),
			src.GetRowByteSize(), src.GetSampleByteSize(),
			src.GetNumLayers(), src.GetNumRows(), src.GetNumColumns());
}


//====================================================================
void __fastcall Resample3D_Linear_PackedSample_UShort4444(VolumeBase& dst, const VolumeBase& src)
{
	Call_ResampleElements3D_Linear_PackedSample_UShort(
		sample_processors::LinearInterpolator3D_PackedSample_UShort4444(), dst, src);
}


//====================================================================
void __fastcall Resample3D_Linear_PackedSample_UShort1555(VolumeBase& dst, const VolumeBase& src)
{
	Call_ResampleElements3D_Linear_PackedSample_UShort(
		sample_processors::LinearInterpolator3D_PackedSample_UShort1555(), dst, src);
}


//====================================================================
bool __fastcall Resample3D(VolumeBase& dst, const VolumeBase& src,
										eResampleFilter filterType)
{
	VW_ASSERT(dst.GetNumSamples(), "Destination is empty.");
	VW_ASSERT(src.GetNumSamples(), "Source is empty.");
	VW_ASSERT(dst.CheckSampleFormatMatch(src), "Sample-format mismatch.");

	// nearest-point resampling
	if(filterType == ResampleFilter_Nearest) {

		Resample3D_Nearest(dst, src);
		return true;
	}

	// linear-interpolation resampling
	if(filterType == ResampleFilter_Linear) {

		switch(dst.GetSampleFormat()) {
			case SampleFormat_L8:
			case SampleFormat_A8:
			case SampleFormat_A8L8:
			case SampleFormat_A8R8G8B8:
				Resample3D_Linear_NBytes(dst, src, 0, dst.GetSampleByteSize());
				return true;

			case SampleFormat_X8R8G8B8:
				Resample3D_Linear_NBytes(dst, src, 0, 3);
				return true;

			case SampleFormat_A4R4G4B4:
				Resample3D_Linear_PackedSample_UShort4444(dst, src);
				return true;

			case SampleFormat_A1R5G5B5:
				Resample3D_Linear_PackedSample_UShort1555(dst, src);
				return true;

			case SampleFormat_ZB32:
			case SampleFormat_WB32:
				Resample3D_Linear_NFloats(dst, src, 0, dst.GetSampleByteSize() / sizeof(float));
				return true;
		}
	}

	// unsupported filter-type / sample-type
	return false;
}


//====================================================================
void __fastcall CopyToLayer(VolumeBase& dst, uint dstLayerIndex, const SignalBase& src)
{
	VW_ASSERT(dst.GetLayerByteSize() == src.GetTotalByteSize(), "Byte-size mismatch.");
	VW_ASSERT(dstLayerIndex < dst.GetNumLayers(), "Destination-layer-index out of range.");

	::CopyMemory(dst.SampleB(dstLayerIndex, 0, 0), src.GetSamplesB(), src.GetTotalByteSize());
}


//====================================================================
void __fastcall CopyFromLayer(SignalBase& dst, const VolumeBase& src, uint srcLayerIndex)
{
	VW_ASSERT(src.GetLayerByteSize() == dst.GetTotalByteSize(), "Byte-size mismatch.");
	VW_ASSERT(srcLayerIndex < src.GetNumLayers(), "Source-layer-index out of range.");

	::CopyMemory(dst.GetSamplesB(), src.SampleB(srcLayerIndex, 0, 0), dst.GetTotalByteSize());
}


//====================================================================
void __fastcall FlipLayers(VolumeBase& target)
{
	uint halfNumLayers = target.GetNumLayers() / 2;
	if(! halfNumLayers) return;

	ProcessElements3D(element_processors::Swap_N<ubyte>(target.GetSampleByteSize()),
			halfNumLayers, target.GetNumRows(), target.GetNumColumns(),
			target.GetSamplesB(), target.GetLayerByteSize(),
			target.GetRowByteSize(), target.GetSampleByteSize(),
			target.SampleB(target.GetNumLayers() - 1, 0, 0), - target.GetLayerByteSize(),
			target.GetRowByteSize(), target.GetSampleByteSize());
}


//====================================================================
void __fastcall FlipRows(VolumeBase& target)
{
	uint halfNumRows = target.GetNumRows() / 2;
	if(! halfNumRows) return;

	// reverse layer/row semantics
	ProcessElements3D(element_processors::Swap_N<ubyte>(target.GetSampleByteSize()),
			halfNumRows, target.GetNumLayers(), target.GetNumColumns(),
			target.GetSamplesB(), target.GetRowByteSize(),
			target.GetLayerByteSize(), target.GetSampleByteSize(),
			target.SampleB(0, target.GetNumRows() - 1, 0), - target.GetRowByteSize(),
			target.GetLayerByteSize(), target.GetSampleByteSize());
}


//====================================================================
void __fastcall FlipColumns(VolumeBase& target)
{
	uint halfNumCols = target.GetNumColumns() / 2;
	if(! halfNumCols) return;

	// reverse row/column semantics
	ProcessElements3D(element_processors::Swap_N<ubyte>(target.GetSampleByteSize()),
			target.GetNumLayers(), halfNumCols, target.GetNumRows(),
			target.GetSamplesB(), target.GetLayerByteSize(),
			target.GetSampleByteSize(), target.GetRowByteSize(),
			target.GetSamplesB() + target.GetRowByteSize() - target.GetSampleByteSize(),
			target.GetLayerByteSize(), - target.GetSampleByteSize(), target.GetRowByteSize());
}


//********************************************************************
}//namespace vertexware


