

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "Framework\Plugin.h"
#include "Framework\Kernel.h"
#include "Framework\VideoRenderer.h"
#include "Framework\HIDController.h"
#include "HIDControllers\Keyboard.h"
#include "HIDControllers\Mouse.h"
#include "DataObjects\DataCodecManager.h"
#include "GUI\GUI_Base.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Application base-class.
//====================================================================
class VW_COMMON_API Application : public Plugin, public gui::MessageHandler
{
public:
	//-------------------------------------
	// Startup-variables supplied by kernel
	//-------------------------------------
	struct VW_COMMON_API StartupInfo
	{
		StartupInfo();

		void  DumpPaths(std::ostream& os, uint tabLevel = 0) const;

		std::string  commandLine;
		std::string  fullCommandLine;
		std::string  processDir;
		std::string  userHomeDir;
		std::string  userProfileDir;
		std::string  appProfileDir;
		std::string  appLogFilename;
	};

	//--------------------------------------------
	// Startup-variables controlled by application
	//--------------------------------------------
	struct VW_COMMON_API StartupResult
	{
		StartupResult();

		void  DumpPaths(std::ostream& os, uint tabLevel = 0) const;

		std::string  processDir;
		std::string  appProfileDir;
		std::string  appLogFilename;
		std::string  appWindowText;

		bool  bUseProfileDir;
		bool  bAutoCreateProfileDir;
		bool  bShowLogAndExit;
	};

	//-----------------------------------------
	// Init-variables controlled by application
	//-----------------------------------------
	struct VW_COMMON_API AutoInitInfo
	{
		AutoInitInfo();

		bool  bDeveloperMode;

		bool  bShowGUICursor;
		bool  bShowSystemCursor;

		bool  bAutoRegisterStandardSignalCodecs;
		bool  bAutoRegisterStandardImageCodecs;
		bool  bAutoRegisterStandardVolumeCodecs;

		bool  bAutoRegisterStandardTileTextures;
		bool  bAutoRegisterStandardFonts;
		bool  bAutoRegisterStandardCursors;

		bool  bDumpSystemFonts;
		bool  bDumpSignalCodecs;
		bool  bDumpImageCodecs;
		bool  bDumpVolumeCodecs;
		bool  bDumpTileTextures;
		bool  bDumpFonts;
		bool  bDumpCursors;
		bool  bDumpShaders;
		bool  bDumpComponentHierarchy;
		bool  bDumpCycleQueue;
	};


public:
	//---------------------------------------
	// Do NOT perform anything in destructor.
	// Exceptions: NOT allowed.
	//---------------------------------------
	virtual ~Application();

	//-------------------------
	// Return application UID.
	// Exceptions: NOT allowed.
	//-------------------------
	virtual  UID   GetApplicationUID();

	//-------------------------------------------------------------
	// Called once at startup.
	// <si.commandLine> is filtered version of <si.fullCommandLine>
	// i.e. kernel command-line-arguments are stripped-off.
	// Set <sr.bShowLogAndExit> to true to display log window and
	// stop further execution (can be used for displaying
	// application-specific command-line arguments).
	// Throw exception on error.
	// Exceptions: allowed.
	//-------------------------------------------------------------
	virtual  void  BeginStartup(const StartupInfo& si, StartupResult& sr);

	//----------------------------------------------------------------
	// Called once at startup.
	// <aii.bDeveloperMode> reflects command-line option if specified.
	// Modify it if necessary (e.g. always disable in release builds).
	// Throw exception on error.
	// Exceptions: allowed.
	//----------------------------------------------------------------
	virtual  void  QueryAutoInit(AutoInitInfo& aii);

	//--------------------------
	// Called once at startup.
	// Throw exception on error.
	// Exceptions: allowed.
	//--------------------------
	virtual  void  Initialize();

	//------------------------------------------------
	// Called once after startup protocol is finished,
	// before entering config-loop for the first time.
	// Throw exception on error.
	// Exceptions: allowed.
	//------------------------------------------------
	virtual  void  EndStartup();

	//-----------------------------------------------
	// Called once at shutdown.
	// <bCleanExit> is false if exception was thrown.
	// <fd> is valid only if <bCleanExit> is false.
	// Exceptions: NOT allowed.
	//-----------------------------------------------
	virtual  void  Cleanup(bool bCleanExit, const FailureDesc& fd);

	//-----------------------------------------
	// Return true to accept generic-extension.
	// Exceptions: NOT allowed.
	//-----------------------------------------
	virtual  bool  ConfirmGenericExtension(const std::string& pluginName,
							const std::string& genericExtensionInterfaceName);

	//-------------------------------------
	// Ask kernel for available extensions.
	// Exceptions: allowed.
	//-------------------------------------
	virtual  void  ConnectGenericExtensions();

	//--------------------------------------------
	// Return true to accept direct-input-version.
	// Exceptions: NOT allowed.
	//--------------------------------------------
	virtual  bool  ConfirmDirectInputVersion(uint directInputVersion);

	//------------------------------------------------------------
	// Return true to accept driver. If necessary, driver-priority
	// can be changed through <outDriverPriority>.
	// Drivers cannot have equal priorities.
	// Exceptions: NOT allowed.
	//------------------------------------------------------------
	virtual  bool  ConfirmHIDDriver(const std::string& pluginName,
									uint defaultDriverPriority,
									uint& outDriverPriority);

	//----------------------------------------------------------------------
	// Called once at startup & at each hid-reconfig request.
	// Only hid-drivers compatible with returned di-version will be used.
	// Set <diConfig.enableDirectInput> to false to disable direct-input.
	// Set <diConfig.useLastVersion> to true to use last available version,
	// otherwise fill <diConfig.directInputVersion> to request specific
	// version. If requested version is not available, direct-input will
	// be disabled. Set <diConfig.enableFullscreenExclusiveMode> to false
	// to disable direct-input-device exclusive-mode.
	// <diConfig.useAtDeviceInit> tells devices whether to use direct-input
	// after initialization (if device can operate in di/non-di mode).
	// <reconfigCode> contains value passed through RequestHIDReconfig().
	// Exceptions: NOT allowed.
	//----------------------------------------------------------------------
	virtual  void  QueryHIDConfig(DirectInputConfig& diConfig,
								uint iReconfigCount, uint reconfigCode);

	//----------------------------------------------------------------
	// Return true to ignore driver. Called once per hid-configuration
	// cycle. Can be used to temporarily disable specific driver.
	// Exceptions: NOT allowed.
	//----------------------------------------------------------------
	virtual  bool  QueryIgnoreHIDDriver(const std::string& pluginName);

	//-------------------------------------------------------------
	// Called after human-interface-devices are initialized.
	// Return true if application will be using given device at any
	// time. Returning false will cause immediate device release.
	// Exceptions: NOT allowed.
	//-------------------------------------------------------------
	virtual  void  BeginConfirmHumanInterfaceDevices();
	virtual  bool  ConfirmHumanInterfaceDevice(HumanInterfaceDevice& hid);
	virtual  void  EndConfirmHumanInterfaceDevices();

	//----------------------------------------------------
	// Called after human-interface-devices are confirmed.
	// Store the pointers to devices.
	// Exceptions: NOT allowed.
	//----------------------------------------------------
	virtual  void  BeginConnectHumanInterfaceDevices(const DirectInputConfig& diConfig);
	virtual  void  ConnectHumanInterfaceDevice(HumanInterfaceDevice& hid);
	virtual  void  EndConnectHumanInterfaceDevices();

	//----------------------------------------------------
	// Called after devices are connected to video-window.
	// Exceptions: NOT allowed.
	//----------------------------------------------------
	virtual  void  SetupAsyncHIDSampling();

	//------------------------------------------------------
	// Called before human-interface-devices are destructed.
	// After this call, devices must not be used,
	// so just invalidate the pointers to devices.
	// Exceptions: NOT allowed.
	//------------------------------------------------------
	virtual  void  DisconnectHumanInterfaceDevices();

	//----------------------------------------------------------------------
	// <fullProfile> should be filtered and stored into <outFilteredProfile>
	// to indicate configurations with which application can work.
	// Called once for each available video-subsystem at startup.
	// Exceptions: NOT allowed.
	//----------------------------------------------------------------------
	virtual  void  FilterVideoProfile(const VideoProfile& fullProfile,
									VideoProfile& outFilteredProfile);

	//--------------------------------------
	// Return true to accept video-renderer.
	// Exceptions: NOT allowed.
	//--------------------------------------
	virtual  bool  ConfirmVideoRenderer(const std::string& pluginName,
										const std::string& interfaceName);

	//----------------------------------------------------------------------
	// Called once at startup & at each video-reconfig request.
	// Ask kernel for available video-renderers. Store plugin-name
	// of chosen video-renderer into <vrPluginName> parameter.
	// Find suitable video-properties in video-renderer-info, and fill <vc>.
	// <reconfigCode> contains value passed through RequestVideoReconfig().
	// Exceptions: NOT allowed.
	//----------------------------------------------------------------------
	virtual  void  QueryVideoConfig(std::string& vrPluginName, VideoConfig& vc,
									uint iReconfigCount, uint reconfigCode);

	//-----------------------------------------
	// Called after main-window is constructed.
	// Update <mwp> & <mwg> if necessary.
	// Exceptions: NOT allowed.
	//-----------------------------------------
	virtual  void  PostCreateMainWindow(MainWindowProperties& mwp, MainWindowGeometry& mwg,
									const std::string& vrPluginName, const VideoConfig& vc,
									uint iCreateCount);

	//----------------------------------------
	// Called before main-window is destroyed.
	// Exceptions: NOT allowed.
	//----------------------------------------
	virtual  void  PreDestroyMainWindow(const MainWindowGeometry& mwg,
								const std::string& vrPluginName, const VideoConfig& vc);

	//---------------------------------------------------------
	// Called at every video-reconfig-independent VSync-update.
	// Exceptions: NOT allowed.
	//---------------------------------------------------------
	virtual  void  OnVSyncUpdate(bool bVSyncEnabled);

	//-------------------------------------------
	// Called after VideoRenderer is constructed.
	// Store the pointer to video-renderer.
	// Exceptions: NOT allowed.
	//-------------------------------------------
	virtual  void  ConnectVideoRenderer(VideoRenderer& vr);

	//--------------------------------------------------
	// Called before VideoRenderer is destructed.
	// After this call, video-renderer must not be used,
	// so just invalidate the pointer to video-renderer.
	// Exceptions: NOT allowed.
	//--------------------------------------------------
	virtual  void  DisconnectVideoRenderer();

	//----------------------------------------------------------
	// Called after each successful VideoRenderer init or reset.
	// Full protocols executed by kernel:
	//
	// 1) after VideoRenderer create
	//    ConnectVideoRenderer(...);
	//    CreateVideoResources(false);
	//    InitVideoRendererStates(false);
	//    OnWindowResize(...);
	//
	// 2) before VideoRenderer reset
	//    ReleaseVideoResources(true);
	//
	// 3) after VideoRenderer reset
	//    CreateVideoResources(true);
	//    InitVideoRendererStates(true);
	//    OnWindowResize(...);
	//
	// 4) before VideoRenderer shutdown
	//    ReleaseVideoResources(false);
	//    DisonnectVideoRenderer();
	//----------------------------------------------------------
	virtual  void  InitVideoRendererStates(bool bPostReset);

	//----------------------------------------------------------
	// Called after each successful VideoRenderer initialization
	// or reset. If <bPostReset> is false, video-subsystem has
	// just been created, so application should create and
	// initialize all required video-resources. If <bPostReset>
	// is true, video-subsystem has been successfully reset, so
	// only resources not persistent accross video-reset should
	// be recreated here.
	// Exceptions: allowed.
	//----------------------------------------------------------
	virtual  void  CreateVideoResources(bool bPostReset);

	//--------------------------------------------------------
	// Called before each VideoRenderer shutdown or reset.
	// If <bPreReset> is false, video-shutdown is in progress,
	// so application should cleanup all video-resources.
	// If <bPreReset> is true, video-subsystem is about to
	// reset itself, so application should cleanup only
	// resources that are not persistent across video-reset.
	// Also, video-subsystem may fail to reset itself if
	// certain class of resources are not released before
	// actual reset call (which one depends on underlying
	// low-level API), so release those here as well.
	// Exceptions: NOT allowed.
	//--------------------------------------------------------
	virtual  void  ReleaseVideoResources(bool bPreReset);

	//-----------------------------------------------------------------------
	// Called once at startup and after each generic-reconfig request.
	// <reconfigCode> contains value passed through RequestGenericReconfig().
	// Exceptions: allowed.
	//-----------------------------------------------------------------------
	virtual  void  GenericReconfig(uint iReconfigCount, uint reconfigCode);

	//-------------------------------------------------------------------------
	// Called once at startup and after each subsystem-reconfig request.
	// <reconfigCode> contains value passed through RequestSubsystemReconfig().
	// Exceptions: allowed.
	//-------------------------------------------------------------------------
	virtual  void  InitializeSubsystems(uint iReconfigCount, uint reconfigCode);
	virtual  void  CleanupSubsystems(uint iReconfigCount, uint reconfigCode);

	virtual  void  ConnectSubsystemsToAppWindow(HWND hwnd, bool bFullscreen);
	virtual  void  DisconnectSubsystemsFromAppWindow();

	// return true if all subsystems are ok
	virtual  bool  CheckSubsystems();
	virtual  void  FixSubsystems();

	//--------------------------------
	// Kernel-execution notifications.
	// Exceptions: NOT allowed.
	//--------------------------------
	virtual  void  OnKernelLevelChanged(eKernelLevel activeLevel);

	//----------------------------
	// Window-event notifications.
	// Exceptions: NOT allowed.
	//----------------------------
	virtual  void  OnWindowActive(bool bActive);
	virtual  void  OnWindowResize(uint width, uint height);

	//---------------------
	// Exceptions: allowed.
	//---------------------
	virtual  void  FrameBegin(double secsFrameTimeDelta);
	virtual  void  FrameSampleInput();
	virtual  void  FrameUpdateStates(double secsFrameTimeDelta);
	virtual  void  FramePreRenderVideo();
	virtual  void  FrameRenderVideo();
	virtual  void  FrameRenderArea(const gui::InterleavedRenderArea& area);
	virtual  void  FramePostRenderVideo();
	virtual  void  FrameEnd();

	//-----------------------------------------------------------
	// GUI event handlers, called through HandleMessage().
	// If HandleMessage() is overriden, make sure that base-class
	// version is called from override, in order to make these
	// handlers operational.
	//-----------------------------------------------------------
	virtual  bool  GUIOnCommand(uint commandID) { return false; }
	virtual  bool  GUIOnActivate(bool bActivate, bool bByMouse) { return false; }
	virtual  bool  GUIOnMouseCapture(bool bCaptureBegin) { return false; }
	virtual  bool  GUIOnMouseEnterExit(bool bEnter, const vec2& mousePos) { return false; }
	virtual  bool  GUIOnMouseQueryCursor(const gui::Cursor* &pCursor, const vec2& mousePos) { return false; }
	virtual  bool  GUIOnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos) { return false; }
	virtual  bool  GUIOnMouseWheel(int delta, const vec2& mousePos) { return false; }
	virtual  bool  GUIOnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos) { return false; }
	virtual  bool  GUIOnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData) { return false; }
	virtual  bool  GUIOnKeyboardFocus(bool bGotFocus) { return false; }
	virtual  bool  GUIOnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch) { return false; }


protected:
	Application();

	virtual  bool  HandleMessage(gui::Message& msg);
};


//====================================================================
class VW_COMMON_API ApplicationPluginDescriptor : public PluginDescriptor
{
public:
	virtual  Plugin*  CreatePlugin() {
		return CreateApplication();
	}

	virtual  Application*  CreateApplication() = 0;


protected:
	ApplicationPluginDescriptor(const PluginInfo& pi);
};


//====================================================================
// Application plugin access-point declaration.
//====================================================================
#define  VW_DECLARE_PLUGIN_ACCESS_POINT_APPLICATION( \
				ClassNamePlugin, StringPluginName, \
				UIntPluginVersionMajor, UIntPluginVersionMinor, \
				UIntPluginVersionRevision, UIntPluginVersionBuildFlags, \
				StringPluginDescription, StringPluginDeveloperInfo) \
\
struct ClassNamePlugin##__PluginDescriptor : public ApplicationPluginDescriptor \
{ \
	ClassNamePlugin##__PluginDescriptor() \
		: ApplicationPluginDescriptor(PluginInfo( \
				vertexware::library_info::LibraryVersion(), \
				PluginType_Application, \
				StringPluginName, \
				vertexware::PluginVersion(UIntPluginVersionMajor, UIntPluginVersionMinor, \
										UIntPluginVersionRevision, UIntPluginVersionBuildFlags), \
				VW_BUILD_TIMESTAMP_STRING, \
				StringPluginDescription, \
				StringPluginDeveloperInfo)) \
	{ } \
\
	virtual  Application*  CreateApplication() { \
		return new ClassNamePlugin; \
	} \
}; \
VW_DECLARE_PLUGIN_ACCESS_POINT( ClassNamePlugin##__PluginDescriptor )


//********************************************************************
}//namespace vertexware

