

//--------------------------------------------------------------------
#include "Framework\FrameObservers.h"
#include "Framework\VideoRenderer.h"
#include "GUI\GUI_Renderer.h"
#include "BaseTools\Constructs.h"


//********************************************************************
namespace vertexware {


//====================================================================
FrameStatisticsCollector::FrameStatisticsCollector()
{
	SetMinMaxStabilizationPeriod();
	SetSmoothingInterval();
	SetHistoryInterval();
	EnableHistoryEvaluation();
	Reset();

	Kernel::Instance().RegisterFrameObserver(*this);
	Kernel::Instance().RegisterReconfigObserver(*this);
}


//====================================================================
void  FrameStatisticsCollector::SetMinMaxStabilizationPeriod(double secs)
{
	secsMinMaxStabilizationPeriod = ClampedLow<double>(secs, 0);
}


//====================================================================
void  FrameStatisticsCollector::SetSmoothingInterval(double secsInterval)
{
	secsSmoothingInterval = Clamped<double>(secsInterval, 0.001, 10);
}


//====================================================================
void  FrameStatisticsCollector::SetHistoryInterval(double secsInterval)
{
	secsHistoryInterval = Clamped<double>(secsInterval, 0.001, 10);
}


//====================================================================
void  FrameStatisticsCollector::EnableHistoryEvaluation(bool bEnable)
{
	bEnableHistoryEvaluation = bEnable;
	Reset();
}


//====================================================================
void  FrameStatisticsCollector::Reset()
{
	frameStartTimes.clear();

	accumFrameCount = 0;
	accumFrameTime = 0;

	totalFrameStartCount = 0;
	totalFrameTime = 0;

	frameRateImmediate = 0;
	frameRateImmediateMin = FLT_MAX;
	frameRateImmediateMax = 0;

	frameRateSmooth = 0;
	frameRateSmoothMin = FLT_MAX;
	frameRateSmoothMax = 0;
}


//====================================================================
void  FrameStatisticsCollector::OnFrame(bool& bAutoUnregister, uint frameNumber, double secsFrameTimeDelta)
{
	totalFrameStartCount++;
	totalFrameTime += secsFrameTimeDelta;

	// compute immediate rate
	frameRateImmediate = (secsFrameTimeDelta > 0) ? (1 / secsFrameTimeDelta) : 0;

	// compute smooth rate by history evaluation
	if(bEnableHistoryEvaluation) {

		// store time-point
		frameStartTimes.push_back(totalFrameTime);

		// find last obsolete time-point
		double  refStartTime;
		uint    obsoleteCount = 0;

		TIME_ITER it = frameStartTimes.begin();
		while(it != frameStartTimes.end()) {

			double startTime = *it;
			double diff = totalFrameTime - startTime;

			if(diff <= secsHistoryInterval) {
				refStartTime = startTime;
				break;
			}
			obsoleteCount++;
			++it;
		}

		// discard obsolete time-points
		if(obsoleteCount) {
			frameStartTimes.erase(frameStartTimes.begin(), it);
		}

		// update exported variable
		double historyInterval = totalFrameTime - refStartTime;
		frameRateSmooth = (historyInterval > 0) ? ((frameStartTimes.size() - 1) / historyInterval) : 0;
	}

	// compute smooth rate by history evaluation
	else {
		accumFrameCount++;
		accumFrameTime += secsFrameTimeDelta;

		// restore average at every interval to speed-up the response
		// after the rate changes abruptly
		if(accumFrameTime >= secsSmoothingInterval) {

			frameRateSmooth = accumFrameCount / accumFrameTime;
			accumFrameCount = 0;
			accumFrameTime = 0;
		}
		else {
			// interpolate with immediate rate
			float factorSmooth = (frameRateSmooth > 0) ? ((frameRateSmooth - 1) / frameRateSmooth) : 0;
			frameRateSmooth = factorSmooth * frameRateSmooth + (1 - factorSmooth) * frameRateImmediate;
		}
	}

	// compute min/max
	if(totalFrameTime > secsMinMaxStabilizationPeriod) {

		frameRateImmediateMin = Min(frameRateImmediateMin, frameRateImmediate);
		frameRateImmediateMax = Max(frameRateImmediateMax, frameRateImmediate);

		frameRateSmoothMin = Min(frameRateSmoothMin, frameRateSmooth);
		frameRateSmoothMax = Max(frameRateSmoothMax, frameRateSmooth);
	}
}


//====================================================================
void  FrameStatisticsCollector::OnVideoReconfig()
{
	Reset();
}


//====================================================================
void  FrameStatisticsCollector::OnVSyncReconfig(bool bVSyncEnabled)
{
	Reset();
}


//====================================================================
GUIRendererStatsCollector::GUIRendererStatsCollector()
{
	bRunning = false;
	Reset();
}


//====================================================================
void  GUIRendererStatsCollector::Reset()
{
	frameCount = 0;
	sAcc.Clear();
	sMin.Clear(UINT_MAX);
	sMax.Clear();
	sAvg.Clear();
}


//====================================================================
void  GUIRendererStatsCollector::Run()
{
	if(bRunning) return;
	bRunning = true;

	Kernel::Instance().RegisterFrameObserver(*this);
}


//====================================================================
void  GUIRendererStatsCollector::Stop()
{
	if(! bRunning) return;
	bRunning = false;

	Kernel::Instance().UnregisterFrameObserver(*this);
	CalcAverage();
}


//====================================================================
void  GUIRendererStatsCollector::CalcAverage()
{
	sAvg.Average(sAcc, frameCount);
}


//====================================================================
void  GUIRendererStatsCollector::OnFrame(bool& bAutoUnregister, uint frameNumber, double secsFrameTimeDelta)
{
	gui::Renderer& gr = Kernel::Instance().GetActiveVideoRenderer().GetGUIRenderer();
	sAcc.Accum(gr.Stats());
	sMin.Min(sMin, gr.Stats());
	sMax.Max(sMax, gr.Stats());
	frameCount++;
}


//********************************************************************
}//namespace vertexware


