

//--------------------------------------------------------------------
#include "Framework\FrameworkCommands.h"
#include "Framework\FrameworkProfiler.h"
#include "BaseTools\ConsoleBuffers.h"
#include "DataObjects\SignalIO.h"
#include "GUI\GUI_Manager.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace framework_commands {


//====================================================================
void __fastcall CreateAndRegisterAllCommands(
								std_command_interpreter::Interpreter& ci,
								gui::Component* screenShotCmd_AutoHideComponent,
								uint hidConfigCmd_HIDReconfigCode,
								uint videoConfigCmd_VideoReconfigCode)
{
	ci.RegisterExternalCommand(new framework_commands::ExitCmd);
	ci.RegisterExternalCommand(new framework_commands::CrashCmd);
	ci.RegisterExternalCommand(new framework_commands::VSyncCmd);

	ci.RegisterExternalCommand(new framework_commands::ScreenShotCmd(
		new framework_commands::ScreenShotCmd::ScreenShotCallback_AutoHideComponent(screenShotCmd_AutoHideComponent)));

	ci.RegisterExternalCommand(new framework_commands::HIDConfigCmd(hidConfigCmd_HIDReconfigCode));
	ci.RegisterExternalCommand(new framework_commands::HIDControlCmd);
	ci.RegisterExternalCommand(new framework_commands::VideoConfigCmd(videoConfigCmd_VideoReconfigCode));
	ci.RegisterExternalCommand(new framework_commands::GUIFontCmd);
	ci.RegisterExternalCommand(new framework_commands::GUIAliasCmd);
	ci.RegisterExternalCommand(new framework_commands::GUIStatsCmd);
	ci.RegisterExternalCommand(new framework_commands::VariableCmd);
	ci.RegisterExternalCommand(new framework_commands::ProfilerCmd);
}


//====================================================================
const char* ExitCmd::s_szDefaultName            = "exit";
const char* ExitCmd::s_szOptionWaitCloseConfirm = "w";


//====================================================================
ExitCmd::ExitCmd()
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Request application shutdown");
}


//====================================================================
bool  ExitCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
							const std::string& rawCommandLine)
{
	Kernel& k = Kernel::Instance();

	// setup wait-close-confirmation
	if(clp.GetOption(s_szOptionWaitCloseConfirm)) {

		KernelExecutionParams kep;
		kep.bWaitCloseConfirmationOnExit = true;
		kep.SetKernelCurrent();
	}

	k.RequestShutdown();
	return true;
}


//====================================================================
void  ExitCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionWaitCloseConfirm, "wait close confirmation at exit\n", 1);
}


//====================================================================
const char* CrashCmd::s_szDefaultName                     = "crash";
const char* CrashCmd::s_szOptionAssert                    = "assert";
const char* CrashCmd::s_szOptionVerify                    = "verify";
const char* CrashCmd::s_szOptionThrow                     = "throw";
const char* CrashCmd::s_szOptionZeroPtrRead               = "zeroptrr";
const char* CrashCmd::s_szOptionZeroPtrWrite              = "zeroptrw";
const char* CrashCmd::s_szOptionHardCleanupConfigLoop     = "hccl";
const char* CrashCmd::s_szOptionNoHardCleanupConfigLoop   = "nohccl";
const char* CrashCmd::s_szOptionHardRethrowExit           = "hre";
const char* CrashCmd::s_szOptionNoHardRethrowExit         = "nohre";
const char* CrashCmd::s_szAssignLogMsgOnMainWindowCleanup = "logmwc";


//====================================================================
CrashCmd::CrashCmd()
{
	SetDeveloperModeOnly(true);
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Force application crash / setup kernel crash-parameters");
}


//====================================================================
bool  CrashCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	// assert
	if(clp.GetOption(s_szOptionAssert)) {
		VW_ASSERT(0, "Forced assert by crash-command.");
		return true;
	}

	// verify
	if(clp.GetOption(s_szOptionVerify)) {
		VW_ASSERT(0, "Forced verify by crash-command.");
		return true;
	}

	// throw
	if(clp.GetOption(s_szOptionThrow)) {
		VW_THROW(ApplicationException, "Forced throw by crash-command.");
		return true;
	}

	// zero-pointer
	if(clp.GetOption(s_szOptionZeroPtrRead)) {
		int* pZeroPtr = 0;
		int x = *pZeroPtr;
		return true;
	}
	if(clp.GetOption(s_szOptionZeroPtrWrite)) {
		int* pZeroPtr = 0;
		*pZeroPtr = 0;
		return true;
	}

	// kernel exception control
	KernelExecutionParams kep;

	if(clp.GetOption(s_szOptionHardCleanupConfigLoop)) {
		kep.bEnableHardCleanupConfigLoop = true;
	}
	if(clp.GetOption(s_szOptionNoHardCleanupConfigLoop)) {
		kep.bEnableHardCleanupConfigLoop = false;
	}
	if(clp.GetOption(s_szOptionHardRethrowExit)) {
		kep.bEnableHardRethrowExit = true;
	}
	if(clp.GetOption(s_szOptionNoHardRethrowExit)) {
		kep.bEnableHardRethrowExit = false;
	}
	bool bValue;
	if(clp.ExtractArgumentBool(s_szAssignLogMsgOnMainWindowCleanup, bValue)) {
		kep.bLogWinMsgOnMainWindowCleanup = bValue;
	}
	kep.SetKernelCurrent();

	return true;
}


//====================================================================
void  CrashCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionAssert, "force assert\n", 1);
	os << clp.FormatOptionHelp(s_szOptionVerify, "force verify\n", 1);
	os << clp.FormatOptionHelp(s_szOptionThrow, "throw exception\n", 1);
	os << clp.FormatOptionHelp(s_szOptionZeroPtrRead, "read-dereference zero-pointer\n", 1);
	os << clp.FormatOptionHelp(s_szOptionZeroPtrWrite, "write-dereference zero-pointer\n", 1);
	os << clp.FormatOptionHelp(s_szOptionHardCleanupConfigLoop, "enable hard cleanup by kernel-config-loop\n", 1);
	os << clp.FormatOptionHelp(s_szOptionNoHardCleanupConfigLoop, "disable hard cleanup by kernel-config-loop\n", 1);
	os << clp.FormatOptionHelp(s_szOptionHardRethrowExit, "enable hard error rethrow at exit\n", 1);
	os << clp.FormatOptionHelp(s_szOptionNoHardRethrowExit, "disable hard error rethrow at exit\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignLogMsgOnMainWindowCleanup, "boolean", "log windows-messages on main window cleanup\n", 1);

	os << "Default behavior:\n\tprint parameters\n";
}


//====================================================================
const char* VSyncCmd::s_szDefaultName   = "vsync";
const char* VSyncCmd::s_szOptionOn      = "on";
const char* VSyncCmd::s_szOptionOff     = "off";
const char* VSyncCmd::s_szAssignDynamic = "dynamic";


//====================================================================
VSyncCmd::VSyncCmd()
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Setup video-renderer vertical-synchronization");
}


//====================================================================
bool  VSyncCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
							const std::string& rawCommandLine)
{
	VideoRenderer* pvr = Kernel::Instance().GetActiveVideoRendererPtr();
	if(! pvr) return true;

	// setup dynamic update
	bool bValue;
	if(clp.ExtractArgumentBool(s_szAssignDynamic, bValue)) {
		VideoRenderer::EnableDynamicVSyncUpdate(bValue);
	}

	// update v-sync
	if(clp.GetOption(s_szOptionOn)) {
		pvr->UpdateVSync(true);
	}
	else if(clp.GetOption(s_szOptionOff)) {
		pvr->UpdateVSync(false);
	}

	// print parameters
	os << "VSync........: " << ToString(pvr->GetVideoConfig().videoHints.vSync) << nl;
	os << "DynamicUpdate: " << ToString(VideoRenderer::IsEnabledDynamicVSyncUpdate()) << nl;
	return true;
}


//====================================================================
void  VSyncCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionOn, "enable wait-for-vertical-sync\n", 1);
	os << clp.FormatOptionHelp(s_szOptionOff, "disable wait-for-vertical-sync\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignDynamic, "boolean", "enable/disable dynamic vsync-update\n", 1);

	os << "Default behavior:\n\tprint parameters\n";
}


//====================================================================
ScreenShotCmd::ScreenShotCallback_AutoHideComponent::ScreenShotCallback_AutoHideComponent(
												gui::Component* _pAutoHideTarget)
{
	SetAutoHideTarget(_pAutoHideTarget);
}


//====================================================================
void  ScreenShotCmd::ScreenShotCallback_AutoHideComponent::SetAutoHideTarget(
												gui::Component* _pAutoHideTarget)
{
	pAutoHideTarget = _pAutoHideTarget;
	bAutoHideTargetVisible = false;
}


//====================================================================
void  ScreenShotCmd::ScreenShotCallback_AutoHideComponent::PreScreenShot()
{
	if(pAutoHideTarget) {
		bAutoHideTargetVisible = pAutoHideTarget->IsVisible();
		pAutoHideTarget->ShowTopLevel(false);
	}
}


//====================================================================
void  ScreenShotCmd::ScreenShotCallback_AutoHideComponent::PostScreenShot()
{
	if(pAutoHideTarget  &&  bAutoHideTargetVisible) {
		pAutoHideTarget->ShowTopLevel(true, false);
	}
}


//====================================================================
const char* ScreenShotCmd::s_szDefaultName     = "screenshot";
const char* ScreenShotCmd::s_szOptionImmediate = "imm";
const char* ScreenShotCmd::s_szOptionOverwrite = "ow";
const char* ScreenShotCmd::s_szOptionColor     = "color";
const char* ScreenShotCmd::s_szOptionDepth     = "depth";
const char* ScreenShotCmd::s_szOptionStencil   = "stencil";
const char* ScreenShotCmd::s_szOptionAll       = "all";
const char* ScreenShotCmd::s_szOptionFront     = "front";


//====================================================================
ScreenShotCmd::ScreenShotCmd(ScreenShotCallback* _pCallback)
{
	pCallback = _pCallback;
	pOStream = 0;

	SetDefaultName(s_szDefaultName);
	SetShortDesc("Capture current state of frame-buffer");

	SetAllowFileOverwrite();
	SetSerialNumberWidth();
	SetSerialNumber();
	SetBaseFilename();
	SetFileExtension();
	SetFileFormat();
	SetBufferTypeSuffixes();
}


//====================================================================
ScreenShotCmd::~ScreenShotCmd()
{
	// auto-delete callback
	SafeDelete(pCallback);
}


//====================================================================
void  ScreenShotCmd::SetSerialNumberWidth(uint width)
{
	iSerialNumberWidth = Clamped<uint>(width, 0, 8);
}


//====================================================================
void  ScreenShotCmd::SetBufferTypeSuffixes(const std::string& color,
						const std::string& depth, const std::string& stencil)
{
	colorSuffix = color;
	depthSuffix = depth;
	stencilSuffix = stencil;
}


//====================================================================
void  ScreenShotCmd::GetBufferTypeSuffixes(std::string& color,
						std::string& depth, std::string& stencil) const
{
	color = colorSuffix;
	depth = depthSuffix;
	stencil = stencilSuffix;
}


//====================================================================
bool  ScreenShotCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	pOStream = &os;

	// collect options
	frameBufferFlagsDepthStencil = FrameBufferFlags_Back;
	frameBufferFlagsColor = FrameBufferFlags_Back;
	if(clp.GetOption(s_szOptionFront)) frameBufferFlagsColor = FrameBufferFlags_Front;

	bool bAllBuffers = clp.GetOption(s_szOptionAll);
	bDepthBuffer = clp.GetOption(s_szOptionDepth) || bAllBuffers;
	bStencilBuffer = clp.GetOption(s_szOptionStencil) || bAllBuffers;
	bColorBuffer = clp.GetOption(s_szOptionColor) || bAllBuffers;

	if(! bDepthBuffer  &&  ! bStencilBuffer) bColorBuffer = true;

	// check for immediate mode execution
	if(clp.GetOption(s_szOptionImmediate)) {
		CaptureScreenShot();
		return true;
	}

	// schedule frame-capture depending on back-buffer-count (flipping)
	VideoRenderer* pvr = Kernel::Instance().GetActiveVideoRendererPtr();
	if(! pvr) return true;

	iTargetFrame = Kernel::Instance().GetFrameNumber() + 1 +
					pvr->GetVideoConfig().frameBufferConfig.backBufferCount;

	// notify callback
	if(pCallback) {
		pCallback->PreScreenShot();
	}

	// register self as capture-callback
	Kernel::Instance().RegisterFrameObserver(*this);
	return true;
}


//====================================================================
void  ScreenShotCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionImmediate, "immediate-mode\n", 1);
	os << clp.FormatOptionHelp(s_szOptionOverwrite, "allow file-overwrite\n", 1);
	os << clp.FormatOptionHelp(s_szOptionColor, "capture color buffer\n", 1);
	os << clp.FormatOptionHelp(s_szOptionDepth, "capture depth buffer\n", 1);
	os << clp.FormatOptionHelp(s_szOptionStencil, "capture stencil buffer\n", 1);
	os << clp.FormatOptionHelp(s_szOptionAll, "capture all buffers\n", 1);
	os << clp.FormatOptionHelp(s_szOptionFront, "get front buffer instead of back (applies to color only)\n", 1);

	os << "Default behavior:\n\tscreenshot of color back-buffer\n";
	os << "Notes:\n";
	os << "\tStandard (delayed) screenshot captures the next frame.\n"
	      "\tIn the process of capturing the frame it calls registered\n"
	      "\tcallbacks which should by default hide the console while\n"
		  "\tthe capture is in progress. This procedure enables the user\n"
	      "\tto open the console, type the screenshot command, and obtain\n"
	      "\tthe clean screenshot without the console in front.\n"
	      "\tImmediate-mode screenshot captures the current frame and\n"
	      "\tbypasses the callbacks. This can be used if the console itself\n"
	      "\tshould be visible.\n";
}


//====================================================================
void  ScreenShotCmd::OnFrame(bool& bAutoUnregister, uint frameNumber, double secsFrameTimeDelta)
{
	if(frameNumber == iTargetFrame) {
		CaptureScreenShot();
		bAutoUnregister = true;
	}
}


//====================================================================
void  ScreenShotCmd::CaptureScreenShot()
{
	std::ostream& os = *pOStream;
	VideoRenderer* pvr = Kernel::Instance().GetActiveVideoRendererPtr();
	if(! pvr) return;

	// construct base-name
	std::ostringstream ossFilename;
	ossFilename << baseFilename;

	if(iSerialNumberWidth) {
		ossFilename << '.';
		ossFilename.width(iSerialNumberWidth);
		ossFilename.fill('0');
		ossFilename << iSerialNumber;
	}

	std::string fileext;
	DataCodecManager::Instance().GetDefaultFileFormatExtension(
					DataCategory_Image, fileFormat, fileext, fileExtension);
	if(fileext.length()) fileext.insert(0, ".");

	// construct codec parameter list
	NamedParameterList nplSave = DataCodecManager::Instance().GetEncoderDefaultParamList(DataCategory_Image);
	nplSave.Insert(NamedParameter(data_codec_parameters::Param_FileOverwrite, bAllowFileOverwrite));

	// capture color-buffer
	if(bColorBuffer) {
		AutoPtr<ImageBase> apImage;
		if(pvr->CaptureFrameBuffer(apImage, frameBufferFlagsColor | FrameBufferFlags_Color, true)) {
			std::string filename = ossFilename.str();
			if(colorSuffix.length()) filename.append("." + colorSuffix);
			filename.append(fileext);
			if(SaveImage(*apImage, filename, fileFormat, &nplSave)) {
				os << "Saved color-buffer to: " << filename << nl;
			}
		}
		else {
			os << e_error << "Failed to save color-buffer.\n" << e_info;
		}
	}

	// capture depth-buffer
	if(bDepthBuffer) {
		AutoPtr<ImageBase> apImage;
		if(pvr->CaptureFrameBuffer(apImage, frameBufferFlagsDepthStencil | FrameBufferFlags_Depth, true)) {
			std::string filename = ossFilename.str();
			if(depthSuffix.length()) filename.append("." + depthSuffix);
			filename.append(fileext);
			if(SaveImage(*apImage, filename, fileFormat, &nplSave)) {
				os << "Saved depth-buffer to: " << filename << nl;
			}
		}
		else {
			os << e_error << "Failed to save depth-buffer.\n" << e_info;
		}
	}

	// capture stencil-buffer
	if(bStencilBuffer) {
		AutoPtr<ImageBase> apImage;
		if(pvr->CaptureFrameBuffer(apImage, frameBufferFlagsDepthStencil | FrameBufferFlags_Stencil, true)) {
			std::string filename = ossFilename.str();
			if(stencilSuffix.length()) filename.append("." + stencilSuffix);
			filename.append(fileext);
			if(SaveImage(*apImage, filename, fileFormat, &nplSave)) {
				os << "Saved stencil-buffer to: " << filename << nl;
			}
		}
		else {
			os << e_error << "Failed to save stencil-buffer.\n" << e_info;
		}
	}

	// update serial number
	iSerialNumber++;

	// flush log-stream (async call)
	os << std::flush;

	// notify callback
	if(pCallback) {
		pCallback->PostScreenShot();
	}
}


//====================================================================
const char* HIDConfigCmd::s_szDefaultName             = "hid-config";
const char* HIDConfigCmd::s_szOptionNoDIAtInit        = "nodiinit";
const char* HIDConfigCmd::s_szOptionDIExclusiveModeFS = "exclusive";
const char* HIDConfigCmd::s_szAssignDIVersion         = "version";
const char* HIDConfigCmd::s_szValueLast               = "last";


//====================================================================
HIDConfigCmd::HIDConfigCmd(uint _hidReconfigCode)
{
	SetHIDReconfigCode(_hidReconfigCode);
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Configure human-interface-device subsystems");
}


//====================================================================
bool  HIDConfigCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	// config DirectInput
	DirectInputConfig diConfig;

	diConfig.enableDirectInput = true;
	diConfig.useLastVersion = true;
	diConfig.directInputVersion = 0;
	diConfig.enableFullscreenExclusiveMode = clp.GetOption(s_szOptionDIExclusiveModeFS);
	diConfig.useAtDeviceInit = ! clp.GetOption(s_szOptionNoDIAtInit);

	std::string sVal;
	if(clp.GetAssignment(s_szAssignDIVersion, sVal)) {

		bool bOK = (sVal == s_szValueLast);
		if(! bOK) {
			if(ToValue<uint>(sVal, diConfig.directInputVersion)) {
				bOK = true;
				if(diConfig.directInputVersion != 0) diConfig.useLastVersion = false;
				else                                 diConfig.enableDirectInput = false;
			}
		}
		if(bOK) {
			Kernel::Instance().RequestHIDReconfig(diConfig, iHIDReconfigCode);
			return true;
		}
	}

	// print parameters
	Kernel::Instance().GetActiveDirectInputConfig(diConfig);
	diConfig.Dump(os, 0);
	return true;
}


//====================================================================
void  HIDConfigCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionNoDIAtInit, "dont use DirectInput at hid-init\n", 1);
	os << clp.FormatOptionHelp(s_szOptionDIExclusiveModeFS, "enable DirectInput fullscreen exclusive mode\n", 1);

	os << "Assignments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignDIVersion,
						std::string("uinteger/{") + s_szValueLast + "}",
						"specify DirectInput version, 0 disables DI\n", 1);

	os << "Default behavior:\n\tprint parameters\n";
}


//====================================================================
const char* HIDControlCmd::s_szDefaultName            = "hid-control";
const char* HIDControlCmd::s_szOptionUseDI            = "di";
const char* HIDControlCmd::s_szOptionNoDI             = "nodi";
const char* HIDControlCmd::s_szOptionDumpInfo         = "info";
const char* HIDControlCmd::s_szAssignDeviceCreationID = "d";
const char* HIDControlCmd::s_szAssignControllerNumber = "c";

const char* HIDControlCmd::s_szAssignMouseSyncWithSystemCursor = "mouse-sync";

const char* HIDControlCmd::s_szOptionOTReset                  = "ot-reset";
const char* HIDControlCmd::s_szOptionOTResetHead              = "ot-reset-h";
const char* HIDControlCmd::s_szAssignOTrackerHPRModifier      = "ot-mod-hpr";
const char* HIDControlCmd::s_szAssignOTrackerReverseModifiers = "ot-mod-rev";
const char* HIDControlCmd::s_szAssignOTrackerReverseHead      = "ot-rev-h";
const char* HIDControlCmd::s_szAssignOTrackerReversePitch     = "ot-rev-p";
const char* HIDControlCmd::s_szAssignOTrackerReverseRoll      = "ot-rev-r";


//====================================================================
HIDControlCmd::HIDControlCmd()
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Control human-interface-devices & associated controllers");
}


//====================================================================
bool  HIDControlCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	// check UseDirectInput
	bool bSetupUseDI = false;
	bool bUseDI;
	if(clp.GetOption(s_szOptionUseDI)) {
		bUseDI = true;
		bSetupUseDI = true;
	}
	else if(clp.GetOption(s_szOptionNoDI)) {
		bUseDI = false;
		bSetupUseDI = true;
	}

	// find referenced controller
	uint  controllerNumber = 0;
	clp.ExtractArgumentUInteger(s_szAssignControllerNumber, controllerNumber);

	// find referenced device
	uint  deviceCreationID = 0;
	if(clp.ExtractArgumentUInteger(s_szAssignDeviceCreationID, deviceCreationID)) {

		HumanInterfaceDevice* hid = FindHID(deviceCreationID);
		if(hid) {
			// found matching device
			os << hid->GetDeviceInfo().MakeShortDesc() << nl;

			// setup DirectInput parameters
			if(bSetupUseDI) {
				hid->SetUseDirectInput(bUseDI);
			}
			os << "UsingDirectInput: " << ToString(hid->IsUsingDirectInput()) << nl;

			// dump info
			if(clp.GetOption(s_szOptionDumpInfo)) {
				hid->GetDeviceInfo().Dump(os, 0);
				hid->DumpDeviceCaps(os, 0);

				for(uint i = 0; i < hid->GetNumControllers(); i++) {
					HIDController& hidc = hid->GetController(i);
					os << "Controller " << i << ": " << hidc.GetControllerInterfaceName() << nl;
					hidc.DumpControllerInfo(os, 1);
				}
			}

			// setup controller-interface specific parameters
			if(controllerNumber < hid->GetNumControllers()) {
				UpdateController(os, clp, controllerNumber, hid->GetController(controllerNumber));
			}
			return true;
		}
	}

	// list devices
	os << "Available human-interface-devices:\n";
	for(uint i = 0; i < Kernel::Instance().GetNumHumanInterfaceDevices(); i++) {
		os << Kernel::Instance().GetHumanInterfaceDevice(i).GetDeviceInfo().MakeShortDesc() << nl;
	}
	return true;
}


//====================================================================
void  HIDControlCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionUseDI, "use DirectInput\n", 1);
	os << clp.FormatOptionHelp(s_szOptionNoDI, "dont use DirectInput\n", 1);

	os << "Assignments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignDeviceCreationID, "uinteger", "select device by creation-id\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignControllerNumber, "uinteger", "select controller by number (default = 0)\n", 1);

	os << "Mouse-controller arguments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignMouseSyncWithSystemCursor, "boolean", "synchronize absolute position with system-cursor\n", 1);

	os << "OrientationTracker-controller arguments:\n";
	os << clp.FormatOptionHelp(s_szOptionOTReset, "soft-reset orientation\n", 1);
	os << clp.FormatOptionHelp(s_szOptionOTResetHead, "soft-reset head\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignOTrackerHPRModifier, "boolean", "enable/disable hpr modifier\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignOTrackerReverseModifiers, "boolean", "enable/diable reverse modifiers\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignOTrackerReverseHead, "boolean", "reverse head\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignOTrackerReversePitch, "boolean", "reverse pitch\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignOTrackerReverseRoll, "boolean", "reverse roll\n", 1);

	os << "Default behavior:\n\tlist available human-interface-devices & controllers\n";
}


//====================================================================
HumanInterfaceDevice*  HIDControlCmd::FindHID(uint creationID)
{
	for(uint i = 0; i < Kernel::Instance().GetNumHumanInterfaceDevices(); i++) {
		HumanInterfaceDevice& hid = Kernel::Instance().GetHumanInterfaceDevice(i);
		if(hid.GetDeviceInfo().GetCreationID() == creationID) return &hid;
	}
	return 0;
}


//====================================================================
void  HIDControlCmd::UpdateController(std::ostream& os, const CommandLineParser& clp,
									uint controllerNumber, HIDController& hidc)
{
	// keyboard
	Keyboard* pKeyboard = InterfaceCast<Keyboard>(&hidc);
	if(pKeyboard) {
		Update(os, clp, controllerNumber, *pKeyboard);
	}

	// mouse
	Mouse* pMouse = InterfaceCast<Mouse>(&hidc);
	if(pMouse) {
		Update(os, clp, controllerNumber, *pMouse);
	}

	// orientation-tracker
	OrientationTracker* pOTracker = InterfaceCast<OrientationTracker>(&hidc);
	if(pOTracker) {
		Update(os, clp, controllerNumber, *pOTracker);
	}

	// position-tracker
	PositionTracker* pPTracker = InterfaceCast<PositionTracker>(&hidc);
	if(pPTracker) {
		Update(os, clp, controllerNumber, *pPTracker);
	}

	// glove
	Glove* pGlove = InterfaceCast<Glove>(&hidc);
	if(pGlove) {
		Update(os, clp, controllerNumber, *pGlove);
	}

	// game-controller
	GameController* pGameController = InterfaceCast<GameController>(&hidc);
	if(pGameController) {
		Update(os, clp, controllerNumber, *pGameController);
	}
}


//====================================================================
void  HIDControlCmd::Update(std::ostream& os, const CommandLineParser& clp,
							uint controllerNumber, Keyboard& keyboard)
{
}


//====================================================================
void  HIDControlCmd::Update(std::ostream& os, const CommandLineParser& clp,
							uint controllerNumber, Mouse& mouse)
{
	bool bUpdated = false;
	bool bVal = false;

	// update operational states
	if(clp.ExtractArgumentBool(s_szAssignMouseSyncWithSystemCursor, bVal)) {
		mouse.ActivateSyncWithSystemCursor(bVal);
		bUpdated = true;
	}

	// print operational states
	if(bUpdated) {
		os << "Mouse-controller " << controllerNumber << ":\n";
		os << "\tSyncWithSystemCursor: " << ToString(mouse.IsActiveSyncWithSystemCursor()) << nl;
	}
}


//====================================================================
void  HIDControlCmd::Update(std::ostream& os, const CommandLineParser& clp,
							uint controllerNumber, OrientationTracker& tracker)
{
	bool bUpdated = false;
	bool bVal = false;

	// update operational states
	if(clp.ExtractArgumentBool(s_szAssignOTrackerHPRModifier, bVal)) {
		tracker.EnableHPRModifier(bVal);
		bUpdated = true;
	}
	if(clp.ExtractArgumentBool(s_szAssignOTrackerReverseModifiers, bVal)) {
		tracker.EnableReverseModifiers(bVal);
		bUpdated = true;
	}
	if(clp.ExtractArgumentBool(s_szAssignOTrackerReverseHead, bVal)) {
		tracker.SetReverseHead(bVal);
		bUpdated = true;
	}
	if(clp.ExtractArgumentBool(s_szAssignOTrackerReversePitch, bVal)) {
		tracker.SetReversePitch(bVal);
		bUpdated = true;
	}
	if(clp.ExtractArgumentBool(s_szAssignOTrackerReverseRoll, bVal)) {
		tracker.SetReverseRoll(bVal);
		bUpdated = true;
	}

	// print operational states
	if(bUpdated) {
		os << "OrientationTracker-controller " << controllerNumber << ":\n";
		os << "\tEnableHPRModifier.....: " << ToString(tracker.IsEnabledHPRModifier()) << nl;
		os << "\tEnableReverseModifiers: " << ToString(tracker.AreEnabledReverseModifiers()) << nl;
		os << "\tReverseHead...........: " << ToString(tracker.IsReversedHead()) << nl;
		os << "\tReversePitch..........: " << ToString(tracker.IsReversedPitch()) << nl;
		os << "\tReverseRoll...........: " << ToString(tracker.IsReversedRoll()) << nl;
	}
}


//====================================================================
void  HIDControlCmd::Update(std::ostream& os, const CommandLineParser& clp,
							uint controllerNumber, PositionTracker& tracker)
{
}


//====================================================================
void  HIDControlCmd::Update(std::ostream& os, const CommandLineParser& clp,
							uint controllerNumber, Glove& glove)
{
}


//====================================================================
void  HIDControlCmd::Update(std::ostream& os, const CommandLineParser& clp,
							uint controllerNumber, GameController& gameController)
{
}


//====================================================================
const char* VideoConfigCmd::s_szDefaultName                  = "video-config";
const char* VideoConfigCmd::s_szOptionShowUnfilteredProfile  = "uprofile";
const char* VideoConfigCmd::s_szOptionReconfig               = "r";
const char* VideoConfigCmd::s_szOptionNoProfileCheck         = "nocheck";
const char* VideoConfigCmd::s_szOptionNoAutoSelectClosestFBC = "noauto";

const char* VideoConfigCmd::s_szOptionListRenderers          = "lrend";
const char* VideoConfigCmd::s_szOptionListFrameBufferConfigs = "lfbc";
const char* VideoConfigCmd::s_szOptionListDisplayModes       = "ldm";
const char* VideoConfigCmd::s_szOptionListLegend             = "legend";

const char* VideoConfigCmd::s_szAssignRendererNumber        = "rn";
const char* VideoConfigCmd::s_szAssignRendererPluginName    = "rp";
const char* VideoConfigCmd::s_szAssignRendererInterfaceName = "ri";

const char* VideoConfigCmd::s_szAssignWidth            = "w";
const char* VideoConfigCmd::s_szAssignHeight           = "h";
const char* VideoConfigCmd::s_szAssignBitsPerPixel     = "bpp";
const char* VideoConfigCmd::s_szAssignRefreshFrequency = "freq";
const char* VideoConfigCmd::s_szAssignWindowResizable  = "rsz";

const char* VideoConfigCmd::s_szAssignFullscreen         = "fs";
const char* VideoConfigCmd::s_szAssignHardwareRasterizer = "hwr";
const char* VideoConfigCmd::s_szAssignBackBufferCount    = "bbc";
const char* VideoConfigCmd::s_szAssignRGBA               = "bb";
const char* VideoConfigCmd::s_szAssignDepthStencil       = "ds";
const char* VideoConfigCmd::s_szAssignAccumRGBA          = "acc";

const char* VideoConfigCmd::s_szAssignVSync                = "vsync";
const char* VideoConfigCmd::s_szAssignFSDestinationAlpha   = "fsda";
const char* VideoConfigCmd::s_szAssignPureDeviceMode       = "pure";
const char* VideoConfigCmd::s_szAssignVertexProcessingMode = "vpm";
const char* VideoConfigCmd::s_szValueVPM_Hardware          = "hw";
const char* VideoConfigCmd::s_szValueVPM_Software          = "sw";
const char* VideoConfigCmd::s_szValueVPM_Mixed             = "mix";
const char* VideoConfigCmd::s_szValueVPM_Default           = "def";


//====================================================================
VideoConfigCmd::VideoConfigCmd(uint _videoReconfigCode)
{
	SetVideoReconfigCode(_videoReconfigCode);
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Configure video-subsystem");
}


//====================================================================
bool  VideoConfigCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	VideoConfig        vConfig;
	VideoConfig        vConfigActive;
	VideoRendererInfo  vrInfoActive;
	VideoRendererInfo  vrInfoSelected;
	VideoProfile       vUProfileSelected;
	std::string        sVal;

	VideoRenderer* pvrActive = Kernel::Instance().GetActiveVideoRendererPtr();

	// must have active renderer
	if(! pvrActive) {
		os << e_error << "Cannot find active renderer.\n" << e_info;
		return false;
	}

	// active params
	Kernel::Instance().GetActiveVideoRendererInfo(vrInfoActive);
	vConfigActive = pvrActive->GetVideoConfig();

	// list renderers
	if(clp.GetOption(s_szOptionListRenderers)) {
		os << "Available video-renderers:\n";
		os << "(Number/PluginName/VideoRendererInterfaceName)\n";

		for(uint i = 0; i < Kernel::Instance().GetNumVideoRenderers(); i++) {
			VideoRendererInfo  vrInfo;
			Kernel::Instance().GetVideoRendererInfo(i, vrInfo);
			const char* sep = "/";
			os << i << sep << vrInfo.GetPluginInfo().GetPluginName() << sep << vrInfo.GetVideoRendererInterfaceName() << nl;
		}
		return true;
	}

	// renderer selection
	bool bRendSelected = false;
	uint iRendererNumber = 0;

	if(clp.ExtractArgumentUInteger(s_szAssignRendererNumber, iRendererNumber)) {
		if(FindInfoByNumber(iRendererNumber, vrInfoSelected, vUProfileSelected)) bRendSelected = true;
	}
	else if(clp.GetAssignment(s_szAssignRendererPluginName, sVal)) {
		if(FindInfoByPluginName(sVal, vrInfoSelected, vUProfileSelected)) bRendSelected = true;
	}
	else if(clp.GetAssignment(s_szAssignRendererInterfaceName, sVal)) {
		if(FindInfoByVRInterfaceName(sVal, vrInfoSelected, vUProfileSelected)) bRendSelected = true;
	}
	if(! bRendSelected) {
		vrInfoSelected = vrInfoActive;
		Kernel::Instance().GetActiveVideoRendererUnfilteredProfile(vUProfileSelected);
	}

	// info requests
	bool bUProfile = clp.GetOption(s_szOptionShowUnfilteredProfile);
	bool bListFBC = clp.GetOption(s_szOptionListFrameBufferConfigs);
	bool bListDM = clp.GetOption(s_szOptionListDisplayModes);
	bool bListLegend = clp.GetOption(s_szOptionListLegend);

	// list frame-buffer-configurations & display-mode-configurations
	if(bListFBC  ||  bListDM) {
		DumpVRHeader(os, vrInfoSelected);
		os << (bUProfile ? "Unfiltered profile:" : "Application-filtered profile:") << nl;

		VideoProfile& vProfile = bUProfile ? vUProfileSelected : vrInfoSelected.GetVideoProfile();

		if(bListFBC) {
			os << "FrameBufferConfigurations:\n";
			vProfile.DumpFrameBufferConfigs(os, bListLegend, 1);
			if(bListDM) os << nl;
		}
		if(bListDM) {
			os << "DisplayModes:\n";
			vProfile.DumpDisplayModeProperties(os, 1);
		}
		return true;
	}

	// update config
	vConfig = vConfigActive;

	// display mode/window
	clp.ExtractArgumentUInteger(s_szAssignWidth, vConfig.displayModeConfig.width, vConfig.displayModeConfig.width);
	clp.ExtractArgumentUInteger(s_szAssignHeight, vConfig.displayModeConfig.height, vConfig.displayModeConfig.height);
	clp.ExtractArgumentUInteger(s_szAssignBitsPerPixel, vConfig.displayModeConfig.bitsPerPixel, vConfig.displayModeConfig.bitsPerPixel);
	clp.ExtractArgumentUInteger(s_szAssignRefreshFrequency, vConfig.displayModeConfig.refreshFrequency, vConfig.displayModeConfig.refreshFrequency);
	clp.ExtractArgumentBool(s_szAssignWindowResizable, vConfig.resizable, vConfig.resizable);

	// frame-buffer
	clp.ExtractArgumentBool(s_szAssignFullscreen, vConfig.frameBufferConfig.fullscreen, vConfig.frameBufferConfig.fullscreen);
	clp.ExtractArgumentBool(s_szAssignHardwareRasterizer, vConfig.frameBufferConfig.hardwareRasterizer, vConfig.frameBufferConfig.hardwareRasterizer);
	clp.ExtractArgumentUInteger(s_szAssignBackBufferCount, vConfig.frameBufferConfig.backBufferCount, vConfig.frameBufferConfig.backBufferCount);

	uint bitCountVals[4];
	if(ExtractRGBA(clp, s_szAssignRGBA, bitCountVals)) {
		vConfig.frameBufferConfig.bitsRed   = bitCountVals[0];
		vConfig.frameBufferConfig.bitsGreen = bitCountVals[1];
		vConfig.frameBufferConfig.bitsBlue  = bitCountVals[2];
		vConfig.frameBufferConfig.bitsAlpha = bitCountVals[3];
	}
	if(ExtractUIntegers(clp, s_szAssignDepthStencil, bitCountVals, 2)) {
		vConfig.frameBufferConfig.bitsDepth   = bitCountVals[0];
		vConfig.frameBufferConfig.bitsStencil = bitCountVals[1];
	}
	if(ExtractRGBA(clp, s_szAssignAccumRGBA, bitCountVals)) {
		vConfig.frameBufferConfig.bitsAccumRed   = bitCountVals[0];
		vConfig.frameBufferConfig.bitsAccumGreen = bitCountVals[1];
		vConfig.frameBufferConfig.bitsAccumBlue  = bitCountVals[2];
		vConfig.frameBufferConfig.bitsAccumAlpha = bitCountVals[3];
	}

	// video-hints
	clp.ExtractArgumentBool(s_szAssignVSync, vConfig.videoHints.vSync, vConfig.videoHints.vSync);
	clp.ExtractArgumentBool(s_szAssignFSDestinationAlpha, vConfig.videoHints.fullscreenDestinationAlpha, vConfig.videoHints.fullscreenDestinationAlpha);
	clp.ExtractArgumentBool(s_szAssignPureDeviceMode, vConfig.videoHints.pureDeviceMode, vConfig.videoHints.pureDeviceMode);

	if(clp.GetAssignment(s_szAssignVertexProcessingMode, sVal)) {
		if(sVal == s_szValueVPM_Hardware) vConfig.videoHints.vertexProcessingMode = VideoHints::VertexProcessingMode_Hardware;
		if(sVal == s_szValueVPM_Software) vConfig.videoHints.vertexProcessingMode = VideoHints::VertexProcessingMode_Software;
		if(sVal == s_szValueVPM_Mixed) vConfig.videoHints.vertexProcessingMode = VideoHints::VertexProcessingMode_Mixed;
		if(sVal == s_szValueVPM_Default) vConfig.videoHints.vertexProcessingMode = VideoHints::VertexProcessingMode_Default;
	}

	// check update conditions
	bool bConfigChanged =
		(vrInfoSelected.GetPluginInfo().GetPluginName() != vrInfoActive.GetPluginInfo().GetPluginName()) ||
		(vConfig != vConfigActive);

	// process config-update
	if(bConfigChanged) {

		bool bProfileCheck = ! clp.GetOption(s_szOptionNoProfileCheck);
		bool bAutoSelectClosestFBC = ! clp.GetOption(s_szOptionNoAutoSelectClosestFBC);
		bool bRequestReconfig = clp.GetOption(s_szOptionReconfig);

		// find closest frame-buffer-config
		const FrameBufferConfig* pClosestFBC = 0;
		if(bAutoSelectClosestFBC) {

			pClosestFBC = vrInfoSelected.GetVideoProfile().SelectClosestFrameBufferConfig(vConfig.frameBufferConfig);
			if(pClosestFBC) {
				vConfig.frameBufferConfig = *pClosestFBC;
			}
		}

		// check profile
		bool bPassedProfileCheck = true;
		if(bConfigChanged  &&  bProfileCheck) {
			bPassedProfileCheck = vrInfoSelected.GetVideoProfile().IsValid(vConfig);
		}

		// dump updated config if there are differences or profile-mismatches
		if(! bRequestReconfig  ||  ! bPassedProfileCheck) {
			DumpVRHeader(os, vrInfoSelected);
			vConfig.Dump(os, 0);
			if(! bPassedProfileCheck) {
				os << e_warning << "Updated video-configuration is invalid."  << nl << e_info;
			}
		}
		// reconfig (async call)
		else {
			Kernel::Instance().RequestVideoReconfig(
				vrInfoSelected.GetPluginInfo().GetPluginName(), vConfig, iVideoReconfigCode);
		}
		return true;
	}

	// print active parameters
	DumpVRHeader(os, vrInfoActive);
	vConfigActive.Dump(os, 0);
	return true;
}


//====================================================================
void  VideoConfigCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionShowUnfilteredProfile, "show unfiltered video-profile\n", 1);
	os << clp.FormatOptionHelp(s_szOptionReconfig, "reconfigure video-subsystem\n", 1);
	os << clp.FormatOptionHelp(s_szOptionNoProfileCheck, "do not check video-profile\n", 1);
	os << clp.FormatOptionHelp(s_szOptionNoAutoSelectClosestFBC, "do not auto-select closest frame-buffer-configuration\n", 1);
	os << clp.FormatOptionHelp(s_szOptionListRenderers, "list renderers\n", 1);
	os << clp.FormatOptionHelp(s_szOptionListFrameBufferConfigs, "list frame-buffer-configurations\n", 1);
	os << clp.FormatOptionHelp(s_szOptionListDisplayModes, "list display-modes\n", 1);
	os << clp.FormatOptionHelp(s_szOptionListLegend, "include legend in lists (where applicable)\n", 1);

	os << "Renderer-selection arguments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignRendererNumber, "uinteger", "select renderer by number\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignRendererPluginName, "string", "select renderer by plugin-name\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignRendererInterfaceName, "string", "select renderer by interface-name\n", 1);

	os << "Display mode/window arguments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignWidth, "uinteger", "framebuffer width\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignHeight, "uinteger", "framebuffer height\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignBitsPerPixel, "uinteger", "bits-per-pixel\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignRefreshFrequency, "uinteger", "refresh frequency\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignWindowResizable, "boolean", "window resizable\n", 1);

	os << "Frame-buffer arguments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignFullscreen, "boolean", "fullscreen mode\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignHardwareRasterizer, "boolean", "use hardware rasterizer\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignBackBufferCount, "uinteger", "number of back-buffers\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignRGBA, "R.G.B.A", "back-buffer layout (bit-counts)\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignDepthStencil, "D.S", "depth-stencil-buffer layout (bit-counts)\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignAccumRGBA, "R.G.B.A", "accumulation-buffer layout (bit-counts)\n", 1);

	os << "Video hints:\n";
	os << clp.FormatAssignmentHelp(s_szAssignVSync, "boolean", "enable vertical-synchronization\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignFSDestinationAlpha, "boolean", "enable fullscreen destination-alpha\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignPureDeviceMode, "boolean", "enable pure-device mode\n", 1);

	const char* sep = ", ";
	os << clp.FormatAssignmentHelp(s_szAssignVertexProcessingMode,
							std::string("{") +
							s_szValueVPM_Hardware + sep + s_szValueVPM_Software +
							sep + s_szValueVPM_Mixed + sep + s_szValueVPM_Default +
							"}",
							"vertex-processing-mode (hardware/software/mixed/default)\n", 1);

	os << "Default behavior:\n\tprint parameters\n";
	os << "Notes:\n\tReconfigure must be explicitly specified,\n"
	      "\totherwise updated configuration is just printed.\n";
}


//====================================================================
void  VideoConfigCmd::DumpVRHeader(std::ostream& os, const VideoRendererInfo& vrInfo)
{
	os << "VideoRendererPluginName...: " << vrInfo.GetPluginInfo().GetPluginName() << nl;
	os << "VideoRendererInterfaceName: " << vrInfo.GetVideoRendererInterfaceName() << nl;
}


//====================================================================
bool  VideoConfigCmd::FindInfoByNumber(uint number, VideoRendererInfo& vrInfo,
										VideoProfile& unfilteredProfile)
{
	for(uint i = 0; i < Kernel::Instance().GetNumVideoRenderers(); i++) {
		Kernel::Instance().GetVideoRendererInfo(i, vrInfo);
		if(i == number) {
			Kernel::Instance().GetVideoRendererUnfilteredProfile(i, unfilteredProfile);
			return true;
		}
	}
	return false;
}


//====================================================================
bool  VideoConfigCmd::FindInfoByPluginName(const std::string& pluginName,
						VideoRendererInfo& vrInfo, VideoProfile& unfilteredProfile)
{
	for(uint i = 0; i < Kernel::Instance().GetNumVideoRenderers(); i++) {
		Kernel::Instance().GetVideoRendererInfo(i, vrInfo);
		if(vrInfo.GetPluginInfo().GetPluginName() == pluginName) {
			Kernel::Instance().GetVideoRendererUnfilteredProfile(i, unfilteredProfile);
			return true;
		}
	}
	return false;
}


//====================================================================
bool  VideoConfigCmd::FindInfoByVRInterfaceName(const std::string& vrInterfaceName,
						VideoRendererInfo& vrInfo, VideoProfile& unfilteredProfile)
{
	for(uint i = 0; i < Kernel::Instance().GetNumVideoRenderers(); i++) {
		Kernel::Instance().GetVideoRendererInfo(i, vrInfo);
		if(vrInfo.GetVideoRendererInterfaceName() == vrInterfaceName) {
			Kernel::Instance().GetVideoRendererUnfilteredProfile(i, unfilteredProfile);
			return true;
		}
	}
	return false;
}


//====================================================================
bool  VideoConfigCmd::ExtractRGBA(const CommandLineParser& clp,
									const std::string& name, uint outVals[4])
{
	std::vector<uint>  packedVals;

	if(clp.ExtractPackedUIntegers(name, packedVals)) {
		switch(packedVals.size()) {

			case 1: {
				for(uint i = 0; i < 4; i++) outVals[i] = packedVals[0];
				return true;
			}
			case 4: {
				for(uint i = 0; i < 4; i++) outVals[i] = packedVals[i];
				return true;
			}
		}
	}
	return false;
}


//====================================================================
bool  VideoConfigCmd::ExtractUIntegers(const CommandLineParser& clp,
							const std::string& name, uint* pOutVals, uint numVals)
{
	std::vector<uint>  packedVals;
	if(clp.ExtractPackedUIntegers(name, packedVals)) {

		if(packedVals.size() == numVals) {
			for(uint i = 0; i < numVals; i++) pOutVals[i] = packedVals[i];
			return true;
		}
	}
	return false;
}


//====================================================================
const char* GUIFontCmd::s_szDefaultName           = "gui-font";
const char* GUIFontCmd::s_szOptionListSystemFonts = "sysfonts";
const char* GUIFontCmd::s_szOptionUpdateResource  = "u";

const char* GUIFontCmd::s_szAssignFontResourceName = "fr";
const char* GUIFontCmd::s_szAssignFontNumber       = "fn";

const char* GUIFontCmd::s_szAssignTextureWidth      = "texwidth";
const char* GUIFontCmd::s_szAssignTextureMultiLevel = "multilevel";
const char* GUIFontCmd::s_szAssignAsciiSubsetOnly   = "ascii-only";

const char* GUIFontCmd::s_szAssignFontFamily = "family";
const char* GUIFontCmd::s_szAssignFontBold   = "bold";
const char* GUIFontCmd::s_szAssignFontItalic = "italic";

const char* GUIFontCmd::s_szAssignDefaultSpacingMono = "spacing-mono";
const char* GUIFontCmd::s_szAssignDefaultSpacing     = "spacing";

const char* GUIFontCmd::s_szAssignIGScaleEmHeight    = "ig-scale-height";
const char* GUIFontCmd::s_szAssignIGGlyphOffsetH     = "ig-offset-h";
const char* GUIFontCmd::s_szAssignIGGlyphOffsetV     = "ig-offset-v";
const char* GUIFontCmd::s_szAssignIGGlyphSizeModifyH = "ig-size-h";
const char* GUIFontCmd::s_szAssignIGGlyphSizeModifyV = "ig-size-v";

const char* GUIFontCmd::s_szAssignAutoCalcWidthEmptyGlyph = "ac-width-empty";
const char* GUIFontCmd::s_szAssignAutoCalcAddPixelsLeft   = "ac-add-l";
const char* GUIFontCmd::s_szAssignAutoCalcAddPixelsRight  = "ac-add-r";
const char* GUIFontCmd::s_szAssignAutoCalcAddPixelsTop    = "ac-add-t";
const char* GUIFontCmd::s_szAssignAutoCalcAddPixelsBottom = "ac-add-b";


//====================================================================
GUIFontCmd::GUIFontCmd()
{
	SetDeveloperModeOnly(true);
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Setup gui fonts");
}


//====================================================================
bool  GUIFontCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	// list system fonts
	if(clp.GetOption(s_szOptionListSystemFonts)) {

		std::vector<std::string> fontFamilyNames;
		win_utils::GdiplusGetFontFamilyNames(fontFamilyNames);

		os << "Installed system fonts:\n";
		for(uint i = 0; i < fontFamilyNames.size(); i++) {
			os << tab << fontFamilyNames[i] << nl;
		}
		return true;
	}

	// cache texture-resource names
	gui::Manager& gm = gui::Manager::Instance();

	std::vector<std::string>  fontNames;
	gm.GetRegisteredFontNames(fontNames);

	// temp storage
	bool         bVal = false;
	uint         uVal = 0;
	int          iVal = 0;
	double       dVal = 0;
	std::string  sVal;

	// font selection
	int iFontNumber = -1;

	if(clp.ExtractArgumentUInteger(s_szAssignFontNumber, uVal)) {
		// find info by number
		if(uVal < fontNames.size()) iFontNumber = uVal;
	}
	else if(clp.GetAssignment(s_szAssignFontResourceName, sVal)) {
		// find info by resource-name
		for(uint i = 0; i < fontNames.size(); i++) {
			if(fontNames[i] == sVal) {
				iFontNumber = i;
				break;
			}
		}
	}

	gui::FontTexture::CreationParams  cParams;

	const gui::Font*        pFont = 0;
	const gui::FontTexture* pFontTex = 0;

	if(iFontNumber >= 0) {
		pFont = gm.GetFont(fontNames[iFontNumber]);
		pFontTex = dynamic_cast<const gui::FontTexture*>(pFont);
	}
	if(pFontTex) {
		pFontTex->GetCreationParameters(cParams);

		// update creation parameters
		if(clp.ExtractArgumentUInteger(s_szAssignTextureWidth, uVal)) {
			cParams.textureWidth = uVal;
		}
		if(clp.ExtractArgumentBool(s_szAssignTextureMultiLevel, bVal)) {
			cParams.bMultiLevel = bVal;
		}
		if(clp.ExtractArgumentBool(s_szAssignAsciiSubsetOnly, bVal)) {
			cParams.bAsciiSubsetOnly = bVal;
		}
		if(clp.GetAssignment(s_szAssignFontFamily, sVal)) {
			cParams.familyName = sVal;
		}
		if(clp.ExtractArgumentBool(s_szAssignFontBold, bVal)) {
			cParams.bBold = bVal;
		}
		if(clp.ExtractArgumentBool(s_szAssignFontItalic, bVal)) {
			cParams.bItalic = bVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignDefaultSpacingMono, dVal)) {
			cParams.defaultSpacingMono = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignDefaultSpacing, dVal)) {
			cParams.defaultSpacing = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignIGScaleEmHeight, dVal)) {
			cParams.igScaleEmHeight = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignIGGlyphOffsetH, dVal)) {
			cParams.igGlyphOffsetH = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignIGGlyphOffsetV, dVal)) {
			cParams.igGlyphOffsetV = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignIGGlyphSizeModifyH, dVal)) {
			cParams.igGlyphSizeModifyH = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignIGGlyphSizeModifyV, dVal)) {
			cParams.igGlyphSizeModifyV = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignAutoCalcWidthEmptyGlyph, dVal)) {
			cParams.autoCalcWidthEmptyGlyph = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignAutoCalcAddPixelsLeft, dVal)) {
			cParams.autoCalcAddPixelsLeft = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignAutoCalcAddPixelsRight, dVal)) {
			cParams.autoCalcAddPixelsRight = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignAutoCalcAddPixelsTop, dVal)) {
			cParams.autoCalcAddPixelsTop = dVal;
		}
		if(clp.ExtractArgumentDouble(s_szAssignAutoCalcAddPixelsBottom, dVal)) {
			cParams.autoCalcAddPixelsBottom = dVal;
		}

		// update resource
		if(clp.GetOption(s_szOptionUpdateResource)) {

			// create texture with updated parameters
			os << "Creating updated font-resource.\n";

			gui::FontTexture* pUpdatedFontTex = gui::FontTexture::CreateFontTexture(cParams);
			if(pUpdatedFontTex) {

				// register new texture
				gm.RegisterFont(fontNames[iFontNumber], pUpdatedFontTex);
				pFontTex = pUpdatedFontTex;

				// force sync object states
				gm.SyncObjectStates();
			}
			else {
				os << e_error << "Failed to create updated font-resource.\n" << e_info;
			}
			pFontTex->GetCreationParameters(cParams);
		}

		// print font parameters
		os << "Font-resource info: " << fontNames[iFontNumber] << nl;
		cParams.Dump(os, 1);
		return true;
	}

	// print available fonts
	os << "GUI font-resources:\n";

	for(uint i = 0; i < fontNames.size(); i++) {
		os << i << ") " << fontNames[i] << ": ";

		pFont = gm.GetFont(fontNames[i]);
		if(pFont) {

			os << pFont->GetFamilyName();
			if(pFont->IsBold()) os << " Bold";
			if(pFont->IsItalic()) os << " Italic";
		}
		os << nl;
	}
	return true;
}


//====================================================================
void  GUIFontCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionListSystemFonts, "list installed system fonts\n", 1);
	os << clp.FormatOptionHelp(s_szOptionUpdateResource, "update font-resource parameters\n", 1);

	os << "Font-selection arguments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignFontNumber, "uinteger", "select font by number\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignFontResourceName, "string", "select font by resource-name\n", 1);

	os << "Texture arguments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignTextureWidth, "uinteger", "texture-width\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignTextureMultiLevel, "string", "enable/disable resource multi-level\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignAsciiSubsetOnly, "string", "enable/disable ascii-subset only\n", 1);

	os << "Font family:\n";
	os << clp.FormatAssignmentHelp(s_szAssignFontFamily, "string", "font-family name\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignFontBold, "boolean", "enable/disable bold property\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignFontItalic, "boolean", "enable/disable italic property\n", 1);

	os << "Image generation:\n";
	os << clp.FormatAssignmentHelp(s_szAssignIGScaleEmHeight, "real", "scale em-height\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignIGGlyphOffsetH, "real", "set horizontal glyph-offset in pixels\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignIGGlyphOffsetV, "real", "set vertical glyph-offset in pixels\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignIGGlyphSizeModifyH, "real", "modify horizontal glyph-size in pixels\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignIGGlyphSizeModifyV, "real", "modify vertical glyph-size in pixels\n", 1);

	os << "Glyph boundary auto-calculation:\n";
	os << clp.FormatAssignmentHelp(s_szAssignAutoCalcWidthEmptyGlyph, "real", "empty-glyph width\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignAutoCalcAddPixelsLeft, "real", "glyph's left-boundary additional extension in pixels\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignAutoCalcAddPixelsRight, "real", "glyph's right-boundary additional extension in pixels\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignAutoCalcAddPixelsTop, "real", "glyph's top-boundary additional extension in pixels\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignAutoCalcAddPixelsBottom, "real", "glyph's bottom-boundary additional extension in pixels\n", 1);

	os << "Default behavior:\n\tprint font-resources registered with gui-manager\n";
	os << "Notes:\n\tUpdate must be explicitly specified,\n"
	      "\totherwise updated parameters are just printed.\n";
}


//====================================================================
const char* GUIAliasCmd::s_szDefaultName       = "gui-alias";
const char* GUIAliasCmd::s_szOptionListObjects = "objects";
const char* GUIAliasCmd::s_szOptionClearAll    = "clear-all";
const char* GUIAliasCmd::s_szOptionClear       = "clear";
const char* GUIAliasCmd::s_szOptionTile        = "tile";
const char* GUIAliasCmd::s_szOptionFont        = "font";
const char* GUIAliasCmd::s_szOptionShader      = "shader";
const char* GUIAliasCmd::s_szOptionCursor      = "cursor";
const char* GUIAliasCmd::s_szAssignRefName     = "ref";
const char* GUIAliasCmd::s_szAssignAliasName   = "alias";


//====================================================================
GUIAliasCmd::GUIAliasCmd()
{
	SetDeveloperModeOnly(true);
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Setup gui object aliases");
}


//====================================================================
bool  GUIAliasCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	gui::Manager& gm = gui::Manager::Instance();

	// cache arguments
	bool bList = clp.GetOption(s_szOptionListObjects);
	bool bClearAll = clp.GetOption(s_szOptionClearAll);
	bool bClear = clp.GetOption(s_szOptionClear);
	bool bTile = clp.GetOption(s_szOptionTile);
	bool bFont = clp.GetOption(s_szOptionFont);
	bool bShader = clp.GetOption(s_szOptionShader);
	bool bCursor = clp.GetOption(s_szOptionCursor);

	std::string sRefName;
	bool bRefName = clp.GetAssignment(s_szAssignRefName, sRefName);

	std::string sAliasName;
	bool bAliasName = clp.GetAssignment(s_szAssignAliasName, sAliasName);

	bool bArgComboValid = bClearAll || (bClear && bAliasName) || (bRefName && bAliasName);

	// execute changes
	if(bArgComboValid) {
		if(bClearAll) {
			if(bTile)   gm.ClearAllTileTextureAliases();
			if(bFont)   gm.ClearAllFontAliases();
			if(bShader) gm.ClearAllShaderAliases();
			if(bCursor) gm.ClearAllCursorAliases();
		}
		else if(bClear) {
			if(bTile)   gm.ClearTileTextureAlias(sAliasName);
			if(bFont)   gm.ClearFontAlias(sAliasName);
			if(bShader) gm.ClearShaderAlias(sAliasName);
			if(bCursor) gm.ClearCursorAlias(sAliasName);
		}
		else {
			if(bTile)   gm.AddTileTextureAlias(sAliasName, sRefName);
			if(bFont)   gm.AddFontAlias(sAliasName, sRefName);
			if(bShader) gm.AddShaderAlias(sAliasName, sRefName);
			if(bCursor) gm.AddCursorAlias(sAliasName, sRefName);
		}
	}

	// print all substitutions
	std::vector<std::string> names;

	if(bTile) {
		gm.GetTileTextureAliases(names);

		os << "Tile-texture aliases:\n";
		for(uint i = 0; i < names.size(); i++) {
			std::string refName;
			gm.ResolveTileTextureAlias(names[i], refName);
			os << i << ") " << names[i] << " ==> " << refName << nl;
		}
		if(bList) {
			os << "Tile-texture objects:\n";
			gm.GetRegisteredTileTextureNames(names);

			for(uint i = 0; i < names.size(); i++) {
				os << i << ") " << names[i] << nl;
			}
		}
	}
	if(bFont) {
		gm.GetFontAliases(names);

		os << "Font aliases:\n";
		for(uint i = 0; i < names.size(); i++) {
			std::string refName;
			gm.ResolveFontAlias(names[i], refName);
			os << i << ") " << names[i] << " ==> " << refName << nl;
		}
		if(bList) {
			os << "Font objects:\n";
			gm.GetRegisteredFontNames(names);

			for(uint i = 0; i < names.size(); i++) {
				os << i << ") " << names[i] << nl;
			}
		}
	}
	if(bShader) {
		gm.GetShaderAliases(names);

		os << "Shader aliases:\n";
		for(uint i = 0; i < names.size(); i++) {
			std::string refName;
			gm.ResolveShaderAlias(names[i], refName);
			os << i << ") " << names[i] << " ==> " << refName << nl;
		}
		if(bList) {
			os << "Shader objects:\n";
			for(uint i = 0; i < gm.GetNumShaders(); i++) {
				os << i << ") " << gm.GetShader(i).GetName() << nl;
			}
		}
	}
	if(bCursor) {
		gm.GetCursorAliases(names);

		os << "Cursor aliases:\n";
		for(uint i = 0; i < names.size(); i++) {
			std::string refName;
			gm.ResolveCursorAlias(names[i], refName);
			os << i << ") " << names[i] << " ==> " << refName << nl;
		}
		if(bList) {
			os << "Cursor objects:\n";
			for(uint i = 0; i < gm.GetNumCursors(); i++) {
				os << i << ") " << gm.GetCursor(i).GetName() << nl;
			}
		}
	}
	return true;
}


//====================================================================
void  GUIAliasCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionListObjects, "list all objects of referenced type\n", 1);
	os << clp.FormatOptionHelp(s_szOptionClearAll, "clear all aliases referenced by object-type\n", 1);
	os << clp.FormatOptionHelp(s_szOptionClear, "clear alias referenced by object-type & alias-name\n", 1);
	os << clp.FormatOptionHelp(s_szOptionTile, "specify tile-texture object-type\n", 1);
	os << clp.FormatOptionHelp(s_szOptionFont, "specify font object-type\n", 1);
	os << clp.FormatOptionHelp(s_szOptionShader, "specify shader object-type\n", 1);
	os << clp.FormatOptionHelp(s_szOptionCursor, "specify cursor object-type\n", 1);

	os << "Assignments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignRefName, "string", "reference-name\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignAliasName, "string", "alias-name\n", 1);
}


//====================================================================
const char* GUIStatsCmd::s_szDefaultName = "gui-stats";
const char* GUIStatsCmd::s_szOptionRun   = "run";
const char* GUIStatsCmd::s_szOptionStop  = "stop";
const char* GUIStatsCmd::s_szOptionReset = "reset";
const char* GUIStatsCmd::s_szOptionDump  = "dump";


//====================================================================
GUIStatsCmd::GUIStatsCmd()
{
	SetDeveloperModeOnly(true);
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Monitor/accumulate GUI-renderer statistics");
}


//====================================================================
bool  GUIStatsCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	// collect arguments
	bool bReset = clp.GetOption(s_szOptionReset);
	bool bDump = clp.GetOption(s_szOptionDump);
	bool bStop = clp.GetOption(s_szOptionStop);
	bool bRun = clp.GetOption(s_szOptionRun);

	bool bPrintLast = !(bReset || bDump || bStop || bRun);

	// execute
	if(bReset) {
		os << "GUI-statistics-collector reset.\n";
		collector.Reset();
	}
	if(bRun  &&  ! collector.IsRunning()) {
		os << "GUI-statistics-collector running.\n";
		collector.Run();
	}
	if(bStop  &&  collector.IsRunning()) {
		collector.Stop();
		os << "GUI-statistics-collector stopped.\n";
	}
	if(bDump) {
		os << "GUI-statistics:\n";
		os << "Min:\n";
		collector.GetStatsMin().Dump(os, 1);
		os << "Max:\n";
		collector.GetStatsMax().Dump(os, 1);
		os << "Average:\n";
		collector.GetStatsAverage().Dump(os, 1);
		os << "FrameCount: " << collector.GetFrameCount() << nl;
	}

	// print last stats
	if(bPrintLast) {
		os << "GUI-statistics collector running: " << collector.IsRunning() << nl;
		os << "Last-frame:\n";

		gui::Renderer& gr = Kernel::Instance().GetActiveVideoRenderer().GetGUIRenderer();
		gr.Stats().Dump(os, 1);
	}
	return true;
}


//====================================================================
void  GUIStatsCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionRun, "start statistics accumulation\n", 1);
	os << clp.FormatOptionHelp(s_szOptionStop, "stop statistics accumulation\n", 1);
	os << clp.FormatOptionHelp(s_szOptionReset, "reset accumulated statistics\n", 1);
	os << clp.FormatOptionHelp(s_szOptionDump, "dump accumulated statistics\n", 1);

	os << "Default behavior:\n\tprint last-frame statistics\n";
}


//====================================================================
const char* VariableCmd::s_szDefaultName = "var";
const char* VariableCmd::s_szOptionInfo  = "i";


//====================================================================
VariableCmd::VariableCmd()
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Control variables exported by application");
	SetVariableRegistry();
}


//====================================================================
void  VariableCmd::SetVariableRegistry(VariableRegistry* vr)
{
	if(vr) {
		pVarReg = vr;
	}
	else {
		pVarReg = &Kernel::Instance().GetVariableRegistry();
	}
}


//====================================================================
bool  VariableCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	VariableRegistry& reg = *pVarReg;
	CommandLineParser::Argument arg;

	// variable-designator
	bool         bVarRef = false;
	std::string  sVarRef;

	if(clp.GetNumArguments() >= 1) {
		arg = clp.GetArgument(0);
		if(arg == CommandLineParser::ArgumentType_String) {
			bVarRef = true;
			sVarRef = arg.String();
		}
	}

	// variable-value
	bool         bVarVal = false;
	std::string  sVarVal;

	if(clp.GetNumArguments() >= 2) {
		arg = clp.GetArgument(1);
		if(arg == CommandLineParser::ArgumentType_String) {
			bVarVal = true;
			sVarVal = arg.String();
		}
	}

	// collect referenced variables
	std::vector<VariableProxy*>  vars;
	VariableProxy* var = 0;

	if(bVarRef) {
		// check for exact match
		var = reg.GetVariable(sVarRef);

		// find variables with matching prefix
		if(var) {
			vars.push_back(var);
		}
		else {
			vars.reserve(reg.GetNumVariables());

			for(uint i = 0; i < reg.GetNumVariables(); i++) {
				VariableProxy& vp = reg.GetVariable(i);
				if(vp.GetName().find(sVarRef) == 0) {
					vars.push_back(&vp);
				}
			}
		}
	}

	// print detailed variable info
	if(var  &&  clp.GetOption(s_szOptionInfo)) {

		os << "Variable....: " << var->GetName() << nl;
		os << "Description.: " << var->GetVarDescription() << nl;
		os << "DevModeOnly.: " << ToString(var->IsDeveloperModeOnly()) << nl;
		os << "Type........: " << var->GetTargetTypeName() << nl;
		os << "Restrictions: " << var->GetTargetRestrictions() << nl;
		os << "Target......: " << var->GetTargetObjectName() << nl;
		os << "Value.......: " << var->GetValueAsString() << nl;
		return true;
	}

	// set new value
	if(var  &&  bVarVal) {
		KernelExecutionParams kep;

		if(var->IsDeveloperModeOnly()  &&  kep.bDeveloperMode) {
			os << "Variable: " << var->GetName() << " is developer-mode only." << nl;
		}
		else {
			var->SetValueFromString(sVarVal);
		}
		DumpVar(os, *var);
		return true;
	}

	if(bVarRef) {
		// list referenced variables
		for(uint i = 0; i < vars.size(); i++) {
			DumpVar(os, *vars[i]);
		}
	}
	else {
		// list all variables
		for(uint i = 0; i < reg.GetNumVariables(); i++) {
			DumpVar(os, reg.GetVariable(i));
		}
	}
	return true;
}


//====================================================================
void  VariableCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionInfo, "print detailed variable info\n", 1);

	os << "Arguments:\n";
	os << clp.FormatStringHelp("0) variable-name or prefix", "referenced variable(s)\n", 1);
	os << clp.FormatStringHelp("1) variable-type-specific value", "new value that will be assigned to referenced variable\n", 1);

	os << "Default behavior:\n\tprint all variables\n";
	os << "Notes:\n";
	os << "\tIf argument-1 is not specified, all variables with prefix\n"
	      "\tmatching the argument-0 are printed. To set new value\n"
		  "\ttype variable's full name for argument-0 and desired value\n"
		  "\tfor argument-1. To obtain detailed info on variable, add\n"
	      "\tinfo-switch after the full variable name.\n";
}


//====================================================================
void  VariableCmd::DumpVar(std::ostream& os, const VariableProxy& vp)
{
	os << vp.GetName() << " = " << vp.GetValueAsString() << nl;
}


//====================================================================
const char* ProfilerCmd::s_szDefaultName                = "profiler";
const char* ProfilerCmd::s_szOptionClear                = "clear";
const char* ProfilerCmd::s_szOptionReset                = "reset";
const char* ProfilerCmd::s_szOptionDump                 = "dump";
const char* ProfilerCmd::s_szOptionPrintEntryCounts     = "ec";
const char* ProfilerCmd::s_szOptionPrintNodeTimes       = "nt";
const char* ProfilerCmd::s_szOptionPrintBranchTimes     = "bt";
const char* ProfilerCmd::s_szOptionPrintNodePercentages = "np";
const char* ProfilerCmd::s_szOptionRun                  = "run";
const char* ProfilerCmd::s_szOptionStop                 = "stop";
const char* ProfilerCmd::s_szAssignRecursionHandler     = "rh";
const char* ProfilerCmd::s_szAssignAutoReset            = "auto-reset";
const char* ProfilerCmd::s_szAssignMaxDepth             = "max-depth";


//====================================================================
ProfilerCmd::ProfilerCmd()
{
	SetDeveloperModeOnly(true);
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Profiler control");
	SetProfiler();
}


//====================================================================
void  ProfilerCmd::SetProfiler(Profiler* p)
{
	if(p) {
		pProfiler = p;
	}
	else {
		pProfiler = &Kernel::Instance().GetFrameworkProfiler();
	}
}


//====================================================================
bool  ProfilerCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	Profiler& profiler = *pProfiler;

	// modify params
	std::string sVal;
	bool        bVal;
	uint        iVal;

	if(clp.GetAssignment(s_szAssignRecursionHandler, sVal)) {
		if(ToValue<bool>(sVal, bVal)) {
			profiler.EnableRecursionHandler(bVal);
		}
	}
	if(clp.GetAssignment(s_szAssignAutoReset, sVal)) {
		if(ToValue<bool>(sVal, iVal)) {
			profiler.SetAutoResetThreshold(iVal);
		}
	}
	if(clp.GetAssignment(s_szAssignMaxDepth, sVal)) {
		if(ToValue<bool>(sVal, iVal)) {
			profiler.SetMaxExpectedStackDepth(iVal);
		}
	}

	// collect arguments
	bool bClear = clp.GetOption(s_szOptionClear);
	bool bReset = clp.GetOption(s_szOptionReset);
	bool bDump = clp.GetOption(s_szOptionDump);
	bool bPrintEC = clp.GetOption(s_szOptionPrintEntryCounts);
	bool bPrintNT = clp.GetOption(s_szOptionPrintNodeTimes);
	bool bPrintBT = clp.GetOption(s_szOptionPrintBranchTimes);
	bool bPrintNP = clp.GetOption(s_szOptionPrintNodePercentages);
	bool bStop = clp.GetOption(s_szOptionStop)  &&  profiler.IsEnabled();
	bool bRun = clp.GetOption(s_szOptionRun)  &&  ! profiler.IsEnabled();

	// stop
	if(bStop) {
		profiler.Enable(false);
		os << "Profiler stopped.\n";
	}

	// clear/reset/dump
	if(bDump) {
		bool bPrintAll = !(bPrintEC || bPrintBT || bPrintNT || bPrintNP);

		profilerDumpDesc.bPrintEntryCounts = (bPrintAll || bPrintEC);
		profilerDumpDesc.bPrintNodeTimes = (bPrintAll || bPrintNT);
		profilerDumpDesc.bPrintBranchTimes = (bPrintAll || bPrintBT);
		profilerDumpDesc.bPrintNodePercentages = (bPrintAll || bPrintNP);

		profilerDumpDesc.ConstructHeaderLines();

		os << e_attention << "Profile statistics:\n" << e_info;
		profiler.ProcessStatistics();
		profiler.DumpStatistics(os, profilerDumpDesc);
	}
	if(bClear) {
		os << "Cleaning profiler nodes.\n";
		profiler.ClearNodes();
	}
	if(bReset) {
		os << "Reseting profiler statistics.\n";
		profiler.ResetStatistics();
	}

	// print parameters
	if(! bDump) {
		os << "RecursionHandlerEnabled: " << profiler.IsEnabledRecursionHandler() << nl;
		os << "MaxExpectedStackDepth..: " << profiler.GetMaxExpectedStackDepth() << nl;
		os << "AutoResetThreshold.....: " << profiler.GetAutoResetThreshold() << nl;
		os << "IterationCount.........: " << profiler.GetIterationCount() << nl;
		os << "ElapsedSeconds.........: " << profiler.GetElapsedSeconds() << nl;

		if(! bRun) {
			os << "Profiler running.......: " << profiler.IsEnabled() << nl;
		}
	}

	// run
	if(bRun) {
		os << "Profiler running.\n";
		profiler.Enable(true);
	}
	return true;
}


//====================================================================
void  ProfilerCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionClear, "clear profiler nodes\n", 1);
	os << clp.FormatOptionHelp(s_szOptionReset, "reset accumulated statistics\n", 1);
	os << clp.FormatOptionHelp(s_szOptionDump, "dump statistics (all by default)\n", 1);
	os << clp.FormatOptionHelp(s_szOptionPrintEntryCounts, "include entry-counts in dump\n", 1);
	os << clp.FormatOptionHelp(s_szOptionPrintNodeTimes, "include node-times in dump\n", 1);
	os << clp.FormatOptionHelp(s_szOptionPrintBranchTimes, "include branch-times in dump\n", 1);
	os << clp.FormatOptionHelp(s_szOptionPrintNodePercentages, "include node-percentages in dump\n", 1);
	os << clp.FormatOptionHelp(s_szOptionRun, "start profiling\n", 1);
	os << clp.FormatOptionHelp(s_szOptionStop, "stop profiling\n", 1);

	os << "Assignments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignRecursionHandler, "boolean", "enable/disable recursion-handler (larger overhead)\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignAutoReset, "uinteger", "specify auto-reset threshold (number of iterations)\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignMaxDepth, "uinteger", "specify maximum expected stack-depth (debug-only, used if recursion-handler is disabled)\n", 1);

	os << "Default behavior:\n\tprint profiler states\n";
}


//********************************************************************
}//namespace framework_commands


//********************************************************************
}//namespace vertexware


