

//--------------------------------------------------------------------
#include "Framework\HumanInterfaceDevice.h"
#include "Framework\HIDController.h"
#include "Framework\KernelMacros.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"
#include "BaseTools\Debug.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\Time.h"
#include "MathTools\MathBase.h"


//********************************************************************
namespace vertexware {


//====================================================================
bool  HumanInterfaceDevice::s_bUseDirectInputAtInit = true;


//====================================================================
HumanInterfaceDeviceInfo::HumanInterfaceDeviceInfo()
{
	const char* szUnknown = "Unknown";

	creationID = 0;
	creatorPluginName = szUnknown;

	deviceDescription = szUnknown;
	vendor = szUnknown;
	modelType = szUnknown;
	modelSubType = szUnknown;
	hwPort = szUnknown;
	additionalInfo = szUnknown;

	bDirectInputOnly = false;
	bCanUseDirectInput = false;

	directInputDeviceInstanceInfo.instanceName = szUnknown;
	directInputDeviceInstanceInfo.productName = szUnknown;
}


//====================================================================
void  HumanInterfaceDeviceInfo::Dump(std::ostream& os, uint tabLevel) const
{
	TabEffector  te(tabLevel);
	os << te << "CreationID........: " << creationID << nl;
	os << te << "CreatorPluginName.: " << creatorPluginName << nl;
	os << te << "DeviceDescription.: " << deviceDescription << nl;
	os << te << "Vendor............: " << vendor << nl;
	os << te << "ModelType.........: " << modelType << nl;
	os << te << "ModelSubType......: " << modelSubType << nl;
	os << te << "HardwarePort......: " << hwPort << nl;
	os << te << "AdditionalInfo....: " << additionalInfo << nl;
	os << te << "IsDirectInputOnly.: " << ToString(bDirectInputOnly) << nl;
	os << te << "CanUseDirectInput.: " << ToString(bCanUseDirectInput) << nl;
	os << te << "DInputInstanceGUID: " << directInputDeviceInstanceInfo.instanceGuid << nl;
	os << te << "DInputInstanceName: " << directInputDeviceInstanceInfo.instanceName << nl;
	os << te << "DInputProductGUID.: " << directInputDeviceInstanceInfo.productGuid << nl;
	os << te << "DInputProductName.: " << directInputDeviceInstanceInfo.productName << nl;
}


//====================================================================
std::string  HumanInterfaceDeviceInfo::MakeShortDesc() const
{
	const char* szSep = "/";

	std::ostringstream oss;
	oss << GetCreationID();
	oss << szSep << GetDeviceDescription();
	oss << szSep << GetCreatorPluginName();

	return oss.str();
}


//====================================================================
void  HumanInterfaceDeviceInfo::SetDirectInputOnly(bool bDIOnly)
{
	bDirectInputOnly = bDIOnly;
	if(bDirectInputOnly) bCanUseDirectInput = true;
}


//====================================================================
void  HumanInterfaceDeviceInfo::SetCanUseDirectInput(bool bCanUseDI)
{
	bCanUseDirectInput = bDirectInputOnly ? true : bCanUseDI;
}


//====================================================================
void __fastcall HumanInterfaceDevice::SetUseDirectInputAtInit(bool bUseDI)
{
	s_bUseDirectInputAtInit = bUseDI;
}


//====================================================================
HumanInterfaceDevice::HumanInterfaceDevice()
{
	bUsingDirectInput = false;

	samplingThread.Init(*this);
	bSampledDataLocked = false;
	bAsyncSamplingActive = false;

	bUpdateSampleRateParams = false;
	bEnableMaxSampleRate = false;
	maxSampleRate = 50;
	avgSampleRate = 0;
	avgSamplingTime = 0;
}


//====================================================================
HumanInterfaceDevice::~HumanInterfaceDevice()
{
}


//====================================================================
bool  HumanInterfaceDevice::SetHIDParameter(const NamedParameter& np)
{
	return false;
}


//====================================================================
const NamedParameter*  HumanInterfaceDevice::GetHIDParameter(const std::string& parameterName) const
{
	return 0;
}


//====================================================================
void  HumanInterfaceDevice::DumpDeviceCaps(std::ostream& os, uint tabLevel) const
{
}


//====================================================================
HIDController*  HumanInterfaceDevice::FindController(const std::string& controllerInterfaceName)
{
	for(uint i = 0; i < GetNumControllers(); i++) {
		HIDController& hidc = GetController(i);
		if(hidc.GetControllerInterfaceName() == controllerInterfaceName) {
			return &hidc;
		}
	}
	return 0;
}


//====================================================================
void  HumanInterfaceDevice::OnAppWindowActive(bool bActive)
{
}


//====================================================================
void  HumanInterfaceDevice::OnAppWindowGeometryChanged()
{
}


//====================================================================
void  HumanInterfaceDevice::OnStartAsyncSampling(
					HANDLE samplingThreadHandle, uint samplingThreadID)
{
}


//====================================================================
void  HumanInterfaceDevice::OnStopAsyncSampling(
					HANDLE samplingThreadHandle, uint samplingThreadID)
{
}


//====================================================================
bool  HumanInterfaceDevice::SetUseDirectInput(bool bUseDI)
{
	if(bUseDI == bUsingDirectInput) return true;
	if(bUseDI  &&  ! hidInfo.CanUseDirectInput()) return false;
	if(! bUseDI  &&  hidInfo.IsDirectInputOnly()) return false;

	OnUseDirectInput(bUseDI);
	return (bUseDI == bUsingDirectInput);
}


//====================================================================
void  HumanInterfaceDevice::StartAsyncSampling()
{
	UnlockSampledData();

	if(bAsyncSamplingActive) return;
	bAsyncSamplingActive = true;

	olog << "Starting async-sampling on hid: " << GetDeviceInfo().GetCreationID() << std::endl;

	samplingThread.Run();
}


//====================================================================
void  HumanInterfaceDevice::StopAsyncSampling()
{
	UnlockSampledData();

	if(! bAsyncSamplingActive) return;
	bAsyncSamplingActive = false;

	olog << "Stoping async-sampling on hid: " << GetDeviceInfo().GetCreationID() << std::endl;

	samplingThread.SignalExit();
	samplingThread.WaitExit();
}


//====================================================================
void  HumanInterfaceDevice::ActivateAsyncSampling(bool bActivate)
{
	if(bActivate) {
		StartAsyncSampling();
	}
	else {
		StopAsyncSampling();
	}
}


//====================================================================
void  HumanInterfaceDevice::LockSampledData()
{
	if(bSampledDataLocked) return;

	csUpdateInputSamples.Enter();
	bSampledDataLocked = true;

	// notify child controllers
	uint numControllers = GetNumControllers();
	for(uint i = 0; i < numControllers; i++) {
		GetController(i).OnBeginAsyncSampleRead();
	}
}


//====================================================================
void  HumanInterfaceDevice::UnlockSampledData()
{
	if(! bSampledDataLocked) return;

	// notify child controllers
	uint numControllers = GetNumControllers();
	for(uint i = 0; i < numControllers; i++) {
		GetController(i).OnEndAsyncSampleRead();
	}

	bSampledDataLocked = false;
	csUpdateInputSamples.Leave();
}


//====================================================================
void   HumanInterfaceDevice::EnableMaxSampleRate(bool bEnable)
{
	bEnableMaxSampleRate = bEnable;
	bUpdateSampleRateParams = true;
}


//====================================================================
void   HumanInterfaceDevice::SetMaxSampleRate(double maxSamplesPerSec)
{
	maxSampleRate = Clamped<double>(maxSamplesPerSec, 1, 1000);
	bUpdateSampleRateParams = true;
}


//====================================================================
uint  HumanInterfaceDevice::SamplingThread::DoWork()
{
	// notify about async-start
	hid->OnStartAsyncSampling(GetThreadHandle(), GetThreadID());

	HighResolutionTimer  hrt;

	// internal statistics & sleep-adjustment params
	double  targetLoopTime = 1;
	double  avgLoopTime = 1;
	double  biasLast = 1;
	double  biasAccum = 0;
	double  sleepTime = 0;
	double  numSamples = 0;
	double  lastCheckTime = 0;

	const  double  checkInterval = 2.0; //secs
	const  double  syncDiffPerc = 0.05;

	// sampling-loop
	hid->bUpdateSampleRateParams = true;

	while(QueryContinue()) {

		double timeLoopStart = hrt.GetElapsedSeconds();

		// sync synthesized-sample-rate with measured-sample-rate
		if(hid->bEnableMaxSampleRate) {
			if(timeLoopStart - lastCheckTime >= checkInterval) {
				double checkedSampleRate = Max(1.0, numSamples / checkInterval);

				if(Abs((hid->avgSampleRate - checkedSampleRate) / checkedSampleRate) >= syncDiffPerc) {
					hid->bUpdateSampleRateParams = true;
				}
				numSamples = 0;
				lastCheckTime = timeLoopStart;
			}
		}

		// update sample-rate parameters
		if(hid->bUpdateSampleRateParams) {
			hid->bUpdateSampleRateParams = false;

			targetLoopTime = 1 / hid->maxSampleRate;
			avgLoopTime = targetLoopTime;
			sleepTime = hid->bEnableMaxSampleRate ? targetLoopTime : 0;
		}

		// sample the device
		double timeSamplingStart = hrt.GetElapsedSeconds();

		hid->csUpdateInputSamples.Enter();
		hid->SampleInputData();
		hid->csUpdateInputSamples.Leave();

		double timeSamplingEnd = hrt.GetElapsedSeconds();

		numSamples++;

		// throttle sampling-rate if requested
		if(hid->bEnableMaxSampleRate) {

			// sleep some, 0 included (millisecs)
			if(sleepTime >= 0) {
				::Sleep(sleepTime * 1000);
			}

			// adjust sleep interval
			double diffTime = targetLoopTime - avgLoopTime;
			sleepTime += diffTime * 0.55;
		}

		// update statistics
		biasLast = 1 / Max(1.0, hid->avgSampleRate);
		if(hid->bEnableMaxSampleRate) biasLast *= biasLast;
		biasAccum = 1 - biasLast;

		hid->avgSamplingTime = biasAccum * hid->avgSamplingTime + biasLast * (timeSamplingEnd - timeSamplingStart);
		hid->avgSampleRate = avgLoopTime ? (1 / avgLoopTime) : math_constants::PlusInfinity;

		double timeLoopEnd = hrt.GetElapsedSeconds();
		avgLoopTime = biasAccum * avgLoopTime + biasLast * (timeLoopEnd - timeLoopStart);
	}

	// reset exported statistics
	hid->avgSampleRate = 0;
	hid->avgSamplingTime = 0;

	// notify about async-stop
	hid->OnStopAsyncSampling(GetThreadHandle(), GetThreadID());

	// exit thread
	return 0;
}


//====================================================================
SampleReadSentry::SampleReadSentry(HumanInterfaceDevice& _hid, bool _bAutoUnlockAsync)
	:
	hid(_hid),
	bAutoUnlockAsync(_bAutoUnlockAsync)
{
	if(hid.IsAsyncSamplingActive()) hid.LockSampledData();
	else                            hid.SampleInputData();
}


//====================================================================
SampleReadSentry::SampleReadSentry(HIDController& _hidc, bool _bAutoUnlockAsync)
	:
	hid(_hidc.GetParentHID()),
	bAutoUnlockAsync(_bAutoUnlockAsync)
{
	if(hid.IsAsyncSamplingActive()) hid.LockSampledData();
	else                            hid.SampleInputData();
}


//====================================================================
SampleReadSentry::~SampleReadSentry()
{
	if(bAutoUnlockAsync  &&  hid.IsAsyncSamplingActive()) {
		hid.UnlockSampledData();
	}
}


//====================================================================
AsyncSampleReadSentry::AsyncSampleReadSentry(HumanInterfaceDevice& _hid, bool _bAutoUnlock)
	:
	hid(_hid),
	bAutoUnlock(_bAutoUnlock)
{
	if(hid.IsAsyncSamplingActive()) hid.LockSampledData();
}


//====================================================================
AsyncSampleReadSentry::AsyncSampleReadSentry(HIDController& _hidc, bool _bAutoUnlock)
	:
	hid(_hidc.GetParentHID()),
	bAutoUnlock(_bAutoUnlock)
{
	if(hid.IsAsyncSamplingActive()) hid.LockSampledData();
}


//====================================================================
AsyncSampleReadSentry::~AsyncSampleReadSentry()
{
	if(bAutoUnlock  &&  hid.IsAsyncSamplingActive()) {
		hid.UnlockSampledData();
	}
}


//********************************************************************
}//namespace vertexware


