

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Exception.h"
#include "BaseTools\Thread.h"
#include "BaseTools\NamedParameter.h"
#include "BaseTools\WinUtils.h"
#include "Framework\Kernel.h"
#include "Framework\DirectInputProxy.h"


//********************************************************************
namespace vertexware {


//====================================================================
class HIDController;
class TopLevelWindow;


//====================================================================
class VW_COMMON_API HumanInterfaceDeviceInfo
{
public:
	HumanInterfaceDeviceInfo();

	void         Dump(std::ostream& os, uint tabLevel = 0) const;
	std::string  MakeShortDesc() const;

	// info access
	uint  GetCreationID() const { return creationID; }

	const std::string&  GetCreatorPluginName() const { return creatorPluginName; }
	const std::string&  GetDeviceDescription() const { return deviceDescription; }
	const std::string&  GetVendor() const { return vendor; }
	const std::string&  GetModelType() const { return modelType; }
	const std::string&  GetModelSubType() const { return modelSubType; }
	const std::string&  GetHardwarePort() const { return hwPort; }
	const std::string&  GetAdditionalInfo() const { return additionalInfo; }

	bool  IsDirectInputOnly() const { return bDirectInputOnly; }
	bool  CanUseDirectInput() const { return bCanUseDirectInput; }

	const DirectInputDeviceInstanceInfo&  GetDirectInputDeviceInstanceInfo() const { return directInputDeviceInstanceInfo; }

	// filled by kernel
	void  SetCreationID(uint id) { creationID = id; }
	void  SetCreatorPluginName(const std::string& s) { creatorPluginName = s; }

	// filled by driver-plugin
	void  SetDeviceDescription(const std::string& s) { deviceDescription = s; }
	void  SetVendor(const std::string& s) { vendor = s; }
	void  SetModelType(const std::string& s) { modelType = s; }
	void  SetModelSubType(const std::string& s) { modelSubType = s; }
	void  SetHardwarePort(const std::string& s) { hwPort = s; }
	void  SetAdditionalInfo(const std::string& s) { additionalInfo = s; }

	void  SetDirectInputOnly(bool bDIOnly); //set this flag before setting <CanUseDirectInput> flag
	void  SetCanUseDirectInput(bool bCanUseDI);

	void  SetDirectInputDeviceInstanceInfo(const DirectInputDeviceInstanceInfo& didii) { directInputDeviceInstanceInfo = didii; }


private:
	uint  creationID;

	std::string  creatorPluginName;
	std::string  deviceDescription;
	std::string  vendor;
	std::string  modelType;
	std::string  modelSubType;
	std::string  hwPort;
	std::string  additionalInfo;

	bool  bDirectInputOnly;
	bool  bCanUseDirectInput;

	DirectInputDeviceInstanceInfo  directInputDeviceInstanceInfo;
};


//====================================================================
class VW_COMMON_API HumanInterfaceDevice
{
public:
	static void __fastcall SetUseDirectInputAtInit(bool bUseDI = true);
	static bool __fastcall GetUseDirectInputAtInit() { return s_bUseDirectInputAtInit; }


public:
	virtual ~HumanInterfaceDevice();

	const HumanInterfaceDeviceInfo&  GetDeviceInfo() const { return hidInfo; }

	//-------------------------
	// HID specific parameters.
	//-------------------------

	// returns true if parameter is recognized
	virtual  bool  SetHIDParameter(const NamedParameter& np);

	// returns 0 if parameter is not recognized
	virtual  const NamedParameter*  GetHIDParameter(const std::string& parameterName) const;

	//---------------------------
	// Print device capabilities.
	//---------------------------
	virtual  void  DumpDeviceCaps(std::ostream& os, uint tabLevel = 0) const;

	//----------------------
	// HIDController access.
	//----------------------
	virtual  uint            GetNumControllers() = 0;
	virtual  HIDController&  GetController(uint i) = 0;

	HIDController*  FindController(const std::string& controllerInterfaceName);

	//------------------------------------------------------
	// Connect device to given window.
	// <pWindow> is 0 for devices created by device-drivers.
	// Exceptions NOT allowed.
	//------------------------------------------------------
	virtual  void  ConnectToAppWindow(TopLevelWindow* pWindow,
									HWND hwnd, bool bFullscreen) = 0;

	//-------------------------------
	// Disconnect device from window.
	// Exceptions NOT allowed.
	//-------------------------------
	virtual  void  DisconnectFromAppWindow() = 0;

	//-----------------------------------------------------------
	// Some devices might need notifications about window status.
	// Exceptions NOT allowed.
	//-----------------------------------------------------------
	virtual  void  OnAppWindowActive(bool bActive);
	virtual  void  OnAppWindowGeometryChanged();

	virtual  void  OnStartAsyncSampling(HANDLE samplingThreadHandle, uint samplingThreadID);
	virtual  void  OnStopAsyncSampling(HANDLE samplingThreadHandle, uint samplingThreadID);

	//---------------------------
	// DirectInput usage control.
	//---------------------------
	bool  IsUsingDirectInput() const { return bUsingDirectInput; }
	bool  SetUseDirectInput(bool bUseDI);

	//------------------------
	// Sample device data.
	// Exceptions NOT allowed.
	//------------------------
	virtual  bool  SampleInputData() = 0;

	//-----------------------------------------------------
	// Sampling-thread control (always call this functions
	// from single-thread). Times are specified in secs.
	// Rates are specified in samples-per-second.
	//-----------------------------------------------------
	void    StartAsyncSampling();
	void    StopAsyncSampling();
	void    ActivateAsyncSampling(bool bActivate); //calls start/stop internally
	bool    IsAsyncSamplingActive() { return bAsyncSamplingActive; }

	void    LockSampledData();
	void    UnlockSampledData();
	bool    IsSampledDataLocked() { return bSampledDataLocked; }

	void    EnableMaxSampleRate(bool bEnable);
	bool    IsEnabledMaxSampleRate() const { return bEnableMaxSampleRate; }

	void    SetMaxSampleRate(double maxSamplesPerSec);
	double  GetMaxSampleRate() const { return maxSampleRate; }

	double  GetAverageSampleRate() const { return avgSampleRate; }
	double  GetAverageSamplingTime() const { return avgSamplingTime; }


protected:
	HumanInterfaceDevice();

	HumanInterfaceDeviceInfo&  DeviceInfo() { return hidInfo; }

	void  UpdateUsingDirectInput(bool bUsingDI) { bUsingDirectInput = bUsingDI; }

	// update <using-direct-input> flag before returning
	virtual  void  OnUseDirectInput(bool bUseDI) = 0;


private:
	//------------------------
	// Device sampling thread.
	//------------------------
	class VW_COMMON_API SamplingThread : public Thread
	{
	public:
		SamplingThread() { hid = 0; }

		void  Init(HumanInterfaceDevice& _hid) { hid = &_hid; }

	protected:
		virtual  uint  DoWork();

	private:
		HumanInterfaceDevice*  hid;
	};
	friend class SamplingThread;


private:
	static  bool  s_bUseDirectInputAtInit;

	bool                      bUsingDirectInput;
	HumanInterfaceDeviceInfo  hidInfo;
	SamplingThread            samplingThread;
	CriticalSection           csUpdateInputSamples;
	bool                      bSampledDataLocked;
	bool                      bAsyncSamplingActive;

	volatile  bool    bUpdateSampleRateParams;
	volatile  bool    bEnableMaxSampleRate;
	volatile  double  maxSampleRate;
	double            avgSampleRate;
	double            avgSamplingTime;
};


//====================================================================
// This construct correctly handles sync/async sample reading,
// depending on device's sampling mode.
//
// { //open read block
//
//     HumanInterfaceDevice& hid1 = GetMyDevice();
//     SampleReadSentry srs1(hid1); //prepare data
//
//     // process sampled data from device/controller 1
//
//     HIDController& hidc2 = GetMyController();
//     SampleReadSentry srs2(hidc2);
//
//     // process sampled data from device/controller 2
//
// } //close read block
//====================================================================
class VW_COMMON_API SampleReadSentry
{
public:
	SampleReadSentry(HumanInterfaceDevice& hid, bool bAutoUnlockAsync = true);
	SampleReadSentry(HIDController& hidc, bool bAutoUnlockAsync = true);

	~SampleReadSentry();


private:
	HumanInterfaceDevice&  hid;
	bool                   bAutoUnlockAsync;
};


//====================================================================
// This construct correctly async sample reading. If device is in
// synchronous mode, it does exactly nothing.
//
// { //open read block
//
//     HumanInterfaceDevice& hid1 = GetMyDevice();
//     AsyncSampleReadSentry asrs1(hid1); //prepare data
//
//     // process sampled data from device/controller 1
//
//     HIDController& hidc2 = GetMyController();
//     AsyncSampleReadSentry asrs2(hidc2);
//
//     // process sampled data from device/controller 2
//
// } //close read block
//====================================================================
class VW_COMMON_API AsyncSampleReadSentry
{
public:
	AsyncSampleReadSentry(HumanInterfaceDevice& hid, bool bAutoUnlock = true);
	AsyncSampleReadSentry(HIDController& hidc, bool bAutoUnlock = true);

	~AsyncSampleReadSentry();


private:
	HumanInterfaceDevice&  hid;
	bool                   bAutoUnlock;
};


//********************************************************************
}//namespace vertexware


