

//--------------------------------------------------------------------
#include "Framework\Kernel.h"
#include "Framework\Application.h"
#include "Framework\HumanInterfaceDevice.h"
#include "Framework\VideoRenderer.h"
#include "Framework\FrameworkCommands.h"
#include "Framework\HIDController.h"
#include "Framework\DirectInputProxy.h"
#include "Framework\VideoProperties.h"
#include "BaseTools\Debug.h"
#include "GUI\GUI_Manager.h"


//********************************************************************
namespace vertexware {


//====================================================================
FrameObserver::~FrameObserver()
{
	Kernel::Instance().UnregisterFrameObserver(*this);
}


//====================================================================
ReconfigObserver::~ReconfigObserver()
{
	Kernel::Instance().UnregisterReconfigObserver(*this);
}


//====================================================================
MainWindowProperties::MainWindowProperties()
{
	hIcon = 0;
}


//====================================================================
MainWindowGeometry::MainWindowGeometry()
{
	bMaximized = false;
	positionX = 0;
	positionY = 0;
	minSizeX = 160;
	minSizeY = 160;
}


//====================================================================
std::ostream*  Kernel::s_pConOut = 0;
std::ostream*  Kernel::s_pLogOut = 0;


//====================================================================
KernelThreadInfo::KernelThreadInfo()
{
	threadHandle = 0;
	threadID = 0;
	priority = THREAD_PRIORITY_NORMAL;
	bEnablePriorityBoost = true;
}


//====================================================================
void  KernelThreadInfo::Dump(std::ostream& os, uint tabLevel) const
{
	TabEffector te(tabLevel);
	os << te << "ThreadHandle.......: " << threadHandle << nl;
	os << te << "ThreadID...........: " << threadID << nl;
	os << te << "Priority...........: " << priority << nl;
	os << te << "EnablePriorityBoost: " << ToString(bEnablePriorityBoost) << nl;
}


//====================================================================
HIDAutoSamplingConfig::HIDAutoSamplingConfig()
{
	bAutoSampleSystemKeyboard = true;
	bAutoSampleSystemMouse = true;
	bAutoSampleNonSystemDevices = true;
}


//====================================================================
KernelExecutionParams::KernelExecutionParams(bool bInitFromKernelCurrent)
{
	Kernel* kp = Kernel::InstancePtr();

	if(bInitFromKernelCurrent  &&  kp) {
		kp->GetExecutionParams(*this);
		return;
	}
	bAutoDispatchGUIEvents = true;
	bGUIEventsPreFrameUpdateStates = false;
	bAutoClearGUIStats = true;
	bAutoDrawGUI = true;
	bAutoEndGUIRenderer = true;
	bAutoPresentBackBuffer = true;
	bEnableHardCleanupConfigLoop = true;
	bEnableHardRethrowExit = false;
	bLogWinMsgOnMainWindowCleanup = false;
	bWaitCloseConfirmationOnExit = false;
	bDeveloperMode = false;
}


//====================================================================
void  KernelExecutionParams::GetKernelCurrent()
{
	Kernel* kp = Kernel::InstancePtr();
	if(kp) kp->GetExecutionParams(*this);
}


//====================================================================
void  KernelExecutionParams::SetKernelCurrent() const
{
	Kernel* kp = Kernel::InstancePtr();
	if(kp) kp->SetExecutionParams(*this);
}


//====================================================================
void  KernelExecutionParams::Dump(std::ostream& os, uint tabLevel) const
{
	TabEffector te(tabLevel);
	os << te << "AutoSampleSystemKeyboard.....: " << ToString(hidAutoSamplingConfig.bAutoSampleSystemKeyboard) << nl;
	os << te << "AutoSampleSystemMouse........: " << ToString(hidAutoSamplingConfig.bAutoSampleSystemMouse) << nl;
	os << te << "AutoSampleNonSystemDevices...: " << ToString(hidAutoSamplingConfig.bAutoSampleNonSystemDevices) << nl;
	os << te << "AutoDispatchGUIEvents........: " << ToString(bAutoDispatchGUIEvents) << nl;
	os << te << "GUIEventsPreFrameUpdateStates: " << ToString(bGUIEventsPreFrameUpdateStates) << nl;
	os << te << "AutoClearGUIStats............: " << ToString(bAutoClearGUIStats) << nl;
	os << te << "AutoDrawGUI..................: " << ToString(bAutoDrawGUI) << nl;
	os << te << "AutoEndGUIRenderer...........: " << ToString(bAutoEndGUIRenderer) << nl;
	os << te << "AutoPresentBackBuffer........: " << ToString(bAutoPresentBackBuffer) << nl;
	os << te << "EnableHardCleanupConfigLoop..: " << ToString(bEnableHardCleanupConfigLoop) << nl;
	os << te << "EnableHardRethrowExit........: " << ToString(bEnableHardRethrowExit) << nl;
	os << te << "LogWinMsgOnMainWindowCleanup.: " << ToString(bLogWinMsgOnMainWindowCleanup) << nl;
	os << te << "WaitCloseConfirmationOnExit..: " << ToString(bWaitCloseConfirmationOnExit) << nl;
	os << te << "DeveloperMode................: " << ToString(bDeveloperMode) << nl;
}


//====================================================================
FailureDesc::FailureDesc(eKernelLevel _origin, const Exception* el, const std::exception* es)
{
	origin = _origin;

	bCPlusPlus = false;
	bLibrary = false;
	bKernel = false;
	bGUI = false;
	bGenericExtension = false;
	bApplication = false;
	bVideoSubsystem = false;
	bVideoRenderer = false;
	bInputSubsystem = false;
	bGenericSubsystem = false;

#define    ES(f, E)   if(dynamic_cast<const E*>(es)) { f = true; }
#define    EL(f, E)   if(dynamic_cast<const E*>(el)) { f = true; }

	ES(bCPlusPlus, std::exception)
	EL(bLibrary, Exception)
	EL(bKernel, KernelException)
	EL(bGUI, GUIException)
	EL(bGenericExtension, GenericExtensionException)
	EL(bApplication, ApplicationException)
	EL(bInputSubsystem, DirectInputProxyException )
	EL(bInputSubsystem, HIDDriverException )
	EL(bInputSubsystem, HIDException )
	EL(bVideoSubsystem, VideoSubsystemException)
	EL(bVideoSubsystem, VideoSubsystemRecoverableException)
	EL(bVideoRenderer, VideoRendererException)
	EL(bGenericSubsystem, GenericSubsystemException)

#undef     ES
#undef     EL
}


//====================================================================
VW_IMPLEMENT_SINGLETON( Kernel );


//====================================================================
Kernel::Kernel()
{
	VW_IMPLEMENT_SINGLETON_CTOR( Kernel );
}


//====================================================================
Kernel::~Kernel()
{
	VW_IMPLEMENT_SINGLETON_DTOR( Kernel );
}


//====================================================================
Application&  Kernel::GetApplication()
{
	VW_ASSERT(GetApplicationPtr(), "No active application.");
	return *GetApplicationPtr();
}


//====================================================================
void  Kernel::RequestHIDReconfig(uint reconfigCode)
{
	// get active settings
	DirectInputConfig  diConfig;

	GetActiveDirectInputConfig(diConfig);

	// do request
	RequestHIDReconfig(diConfig, reconfigCode);
}


//====================================================================
void  Kernel::RequestVideoReconfig(uint reconfigCode)
{
	// get active settings
	std::string  vrPluginName;
	VideoConfig  videoConfig;

	VideoRenderer* vr = GetActiveVideoRendererPtr();
	if(vr) {
		vrPluginName = vr->GetPluginInfo().GetPluginName();
		videoConfig  = vr->GetVideoConfig();
	}

	// do request
	RequestVideoReconfig(vrPluginName, videoConfig, reconfigCode);
}


//====================================================================
bool  Kernel::ConstructClosestSafeVideoConfig(
				const std::string& vrInterfaceName, const VideoConfig& refVideoConfig,
				std::string& outVideoRendererPluginName, VideoConfig& outVideoConfig)
{
	VideoRendererInfo vrInfo;

	for(uint i = 0; i < GetNumVideoRenderers(); i++) {
		GetVideoRendererInfo(i, vrInfo);

		if(vrInterfaceName.empty()  ||  vrInfo.GetVideoRendererInterfaceName() == vrInterfaceName) {
			if(vrInfo.GetVideoProfile().ConstructClosestValidVideoConfig(refVideoConfig, outVideoConfig)) {
				outVideoRendererPluginName = vrInfo.GetPluginInfo().GetPluginName();
				return true;
			}
		}
	}
	outVideoRendererPluginName.erase();
	outVideoConfig = refVideoConfig;
	return false;
}


//====================================================================
VideoRenderer&  Kernel::GetActiveVideoRenderer()
{
	VW_ASSERT(GetActiveVideoRendererPtr(), "No active video-renderer.");
	return *GetActiveVideoRendererPtr();
}


//====================================================================
uint  Kernel::GetNumHIDControllers(const std::string& hidControllerInterfaceName)
{
	uint iCount = 0;

	// find interface-name/ordinal-number match
	for(uint i = 0; i < GetNumHumanInterfaceDevices(); i++) {
		HIDController* hidc = GetHumanInterfaceDevice(i).FindController(hidControllerInterfaceName);
		if(hidc) iCount++;
	}
	return iCount;
}


//====================================================================
HIDController*  Kernel::GetHIDController(const std::string& hidControllerInterfaceName, uint iOrdinal)
{
	HIDController* hidc = 0;
	uint iCount = 0;

	// find interface-name/ordinal-number match
	for(uint i = 0; i < GetNumHumanInterfaceDevices(); i++) {
		hidc = GetHumanInterfaceDevice(i).FindController(hidControllerInterfaceName);
		if(hidc) {
			if(iCount == iOrdinal) return hidc;
			iCount++;
		}
	}
	// not found
	return 0;
}


//====================================================================
void __fastcall Kernel::SetOutputStreams(std::ostream& conOut, std::ostream& logOut)
{
	s_pConOut = &conOut;
	s_pLogOut = &logOut;
}


//====================================================================
void __fastcall Kernel::GUIManager_ConnectApplication(bool bConnect)
{
	gui::Manager::Instance().ConnectApplication(bConnect);
}


//====================================================================
void __fastcall Kernel::GUIManager_NotifyApplicationWindowActive(bool bActive)
{
	gui::Manager::Instance().NotifyApplicationWindowActive(bActive);
}


//====================================================================
void __fastcall Kernel::GUIManager_NotifyApplicationWindowCreated()
{
	gui::Manager::Instance().NotifyApplicationWindowCreated();
}


//====================================================================
void __fastcall Kernel::GUIManager_BeginConnectHumanInterfaceDevices()
{
	gui::Manager::Instance().BeginConnectHumanInterfaceDevices();
}


//====================================================================
void __fastcall Kernel::GUIManager_ConnectHumanInterfaceDevice(HumanInterfaceDevice& hid)
{
	gui::Manager::Instance().ConnectHumanInterfaceDevice(hid);
}


//====================================================================
void __fastcall Kernel::GUIManager_EndConnectHumanInterfaceDevices()
{
	gui::Manager::Instance().EndConnectHumanInterfaceDevices();
}


//====================================================================
void __fastcall Kernel::GUIManager_DisconnectHumanInterfaceDevices()
{
	gui::Manager::Instance().DisconnectHumanInterfaceDevices();
}


//====================================================================
void __fastcall Kernel::GUIManager_SetTolerateMissingInputDevice(bool bTolerate)
{
	gui::Manager::Instance().SetTolerateMissingInputDevice(bTolerate);
}


//====================================================================
void __fastcall Kernel::GUIManager_ConnectVideoRenderer(VideoRenderer& vr)
{
	gui::Manager::Instance().ConnectVideoRenderer(vr);
}


//====================================================================
void __fastcall Kernel::GUIManager_DisconnectVideoRenderer()
{
	gui::Manager::Instance().DisconnectVideoRenderer();
}


//====================================================================
void __fastcall Kernel::GUIManager_CreateVideoResources(bool bPostReset)
{
	gui::Manager::Instance().CreateVideoResources(bPostReset);
}


//====================================================================
void __fastcall Kernel::GUIManager_ReleaseVideoResources(bool bPreReset)
{
	gui::Manager::Instance().ReleaseVideoResources(bPreReset);
}


//====================================================================
void __fastcall Kernel::GUIManager_OnWindowResize(uint width, uint height)
{
	gui::Manager::Instance().OnWindowResize(width, height);
}


//====================================================================
void __fastcall Kernel::VideoRenderer_ConnectVideoSubsystem(VideoRenderer& vr, VideoSubsystem& vs)
{
	vr.ConnectVideoSubsystem(vs);
}


//====================================================================
void __fastcall Kernel::VideoRenderer_DisconnectVideoSubsystem(VideoRenderer& vr)
{
	vr.DisconnectVideoSubsystem();
}


//====================================================================
void __fastcall Kernel::VideoRenderer_CreateVideoResources(VideoRenderer& vr, bool bPostReset)
{
	vr.CreateVideoResources(bPostReset);
}


//====================================================================
void __fastcall Kernel::VideoRenderer_ReleaseVideoResources(VideoRenderer& vr, bool bPreReset)
{
	vr.ReleaseVideoResources(bPreReset);
}


//====================================================================
void __fastcall Kernel::VideoRenderer_OnWindowResize(VideoRenderer& vr, uint width, uint height)
{
	vr.OnWindowResize(width, height);
}


//********************************************************************
}//namespace vertexware


