

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\LibraryInfo.h"
#include "BaseTools\Exception.h"
#include "BaseTools\NamedParameter.h"
#include "Framework\Kernel.h"
#include "Framework\FrameworkExceptions.h"
#include "private\vw_common_exports.h"


//********************************************************************
namespace vertexware {


//====================================================================
template<typename TTargetInterface, typename TSourceInterface>
__forceinline  TTargetInterface* __fastcall InterfaceCast(TSourceInterface* pInterface)
{
	return dynamic_cast<TTargetInterface*>(pInterface);
}


//====================================================================
template<typename TTargetInterface, typename TSourceInterface>
__forceinline  const TTargetInterface* __fastcall InterfaceCast(const TSourceInterface* pInterface)
{
	return dynamic_cast<const TTargetInterface*>(pInterface);
}


//====================================================================
class PluginDescriptor;


//====================================================================
struct VW_COMMON_API PluginVersion
{
	uint  major;
	uint  minor;
	uint  revision;
	uint  buildFlags;

	PluginVersion(uint major = 1, uint minor = 0, uint revision = 0, uint buildFlags = 0);

	bool  operator==(const PluginVersion& rhs) const;
	bool  operator!=(const PluginVersion& rhs) const { return !(*this == rhs); }

	std::string  MakeString() const;
};


//====================================================================
enum ePluginType
{
	PluginType_GenericExtension = 0,
	PluginType_Application,
	PluginType_DirectInputProxy,
	PluginType_HIDDriver,
	PluginType_VideoSubsystem,
	PluginType_VideoRenderer,

	PluginType_Count,
	PluginType_Force32Bits = 0x7FFFFFFF
};


//====================================================================
// <PluginName> is case-insensitive unique identifier
// regardless of plugin-type.
//====================================================================
class VW_COMMON_API PluginInfo
{
public:
	PluginInfo()
	{ }

	PluginInfo(const library_info::LibraryVersion& libraryVersion,
				ePluginType pluginType,
				const std::string& pluginName,
				const PluginVersion& pluginVersion,
				const std::string& buildTimestamp,
				const std::string& description,
				const std::string& developerInfo);

	PluginInfo(const PluginInfo& pi, ePluginType pluginType);

	const library_info::LibraryVersion&  GetLibraryVersion() const { return libraryVersion; }
	ePluginType           GetPluginType() const { return pluginType; }
	const std::string&    GetPluginName() const { return pluginName; }
	const PluginVersion&  GetPluginVersion() const { return pluginVersion; }
	const std::string&    GetBuildTimestamp() const { return buildTimestamp; }
	const std::string&    GetDescription() const { return description; }
	const std::string&    GetDeveloperInfo() const { return developerInfo; }

	bool  operator==(const PluginInfo& rhs) const;
	bool  operator!=(const PluginInfo& rhs) const { return !(*this == rhs); }

	void  Dump(std::ostream& os, uint tabLevel = 0) const;


private:
	std::string  FormatPluginType() const;


private:
	library_info::LibraryVersion  libraryVersion;
	ePluginType                   pluginType;
	std::string                   pluginName;
	PluginVersion                 pluginVersion;
	std::string                   buildTimestamp;
	std::string                   description;
	std::string                   developerInfo;
};


//====================================================================
class VW_COMMON_API Plugin
{
friend class PluginDescriptor;

public:
	virtual ~Plugin();

	const PluginInfo&  GetPluginInfo() const;

	// return true if parameter is recognized
	virtual  bool  SetPluginParameter(const NamedParameter& np);
	virtual  bool  GetPluginParameter(const std::string& parameterName, NamedParameter& out) const;


protected:
	Plugin();


private:
	void  SetPluginInfo(PluginInfo* p) { pPluginInfo = p; }


private:
	PluginInfo*  pPluginInfo;
};


//====================================================================
class VW_COMMON_API PluginDescriptor
{
public:
	virtual ~PluginDescriptor() {}

	const PluginInfo&  GetPluginInfo() const { return pluginInfo; }

	Plugin*  CreateAndSetupPlugin();

	// exceptions NOT allowed
	virtual  Plugin*  CreatePlugin() = 0;


protected:
	PluginDescriptor(const PluginInfo& pluginInfo);


private:
	PluginInfo  pluginInfo;
};


//====================================================================
#define  VW_PLUGIN_ACCESS_POINT_RETURN_TYPE        vertexware::PluginDescriptor&
#define  VW_PLUGIN_ACCESS_POINT_CALL_CONVENTION    __cdecl
#define  VW_PLUGIN_ACCESS_POINT_NAME_ID            vertexware_GetPluginDescriptor
#define  VW_PLUGIN_ACCESS_POINT_NAME              "vertexware_GetPluginDescriptor"


//====================================================================
typedef  VW_PLUGIN_ACCESS_POINT_RETURN_TYPE   (*VW_PFN_PLUGIN_ACCESS_POINT)();


//====================================================================
#define  VW_DECLARE_PLUGIN_ACCESS_POINT( PluginDescriptorClassName ) \
	extern "C" __declspec(dllexport) \
	VW_PLUGIN_ACCESS_POINT_RETURN_TYPE \
	VW_PLUGIN_ACCESS_POINT_CALL_CONVENTION \
	VW_PLUGIN_ACCESS_POINT_NAME_ID() \
	{ static PluginDescriptorClassName PluginDescriptorClassName##__object; \
	return PluginDescriptorClassName##__object; }


//********************************************************************
}//namespace vertexware


