

//--------------------------------------------------------------------
#include "Framework\StandardCommandInterpreter.h"
#include "BaseTools\STLExtensions.h"
#include "BaseTools\Debug.h"
#include "BaseTools\Constructs.h"
#include "Framework\Kernel.h"


//====================================================================
#define  MAX_LINE_LEN_AT_COMMAND_COMPLETION   60


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
struct UnaryPredicateCommandName
{
	std::string  commandName;

	UnaryPredicateCommandName(const std::string& _commandName) {
		commandName = _commandName;
	}

	bool operator()(const Command* pCommand) const {
		return (pCommand->GetName() == commandName);
	}
};


//====================================================================
struct SortPredicateCommandName
{
	bool operator()(const Command* lhs, const Command* rhs) const {
		return (lhs->GetName() < rhs->GetName());
	}
};


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//********************************************************************
namespace std_command_interpreter {


//====================================================================
Interpreter::Interpreter()
{
	SetConsoleBuffers();

	KernelExecutionParams kep;
	SetDeveloperModeActive(kep.bDeveloperMode);

	SetDebugDumpPreprocessedCommandLine();
	SetGlobalHelpSwitch();
	SetCommandNotRecognizedMessage();

	// auto-register internal commands
	RegisterInternalCommand(new DumpPreprocessedCommandLineCmd(*this));
	RegisterInternalCommand(new ListCommandsCmd(*this));
	RegisterInternalCommand(new HelpCmd(*this));
	RegisterInternalCommand(new AliasCmd(*this));
	RegisterInternalCommand(new ConsoleBuffersCmd(*this));
	RegisterInternalCommand(new HistoryCmd(*this));

	Rehash();
}


//====================================================================
Interpreter::~Interpreter()
{
	DeleteAllExternalCommands();
	DeleteAllInternalCommands();
}


//====================================================================
void  Interpreter::SetConsoleBuffers(ConsoleBuffers* _pConsoleBuffers)
{
	if(_pConsoleBuffers) {
		pConsoleBuffers = _pConsoleBuffers;
	}
	else {
		pConsoleBuffers = &Kernel::Instance().GetConsoleBuffers();
	}
}


//====================================================================
void  Interpreter::Rehash()
{
	SortCommands(internalCommands);
	SortCommands(externalCommands);
	MakeMessages();
}


//====================================================================
void  Interpreter::SetGlobalHelpSwitch(const std::string& s)
{
	globalHelpSwitch = s;
}


//====================================================================
void  Interpreter::SetCommandNotRecognizedMessage(const std::string& s)
{
	messageNotRecognizedHeader = s;
	MakeMessages();
}


//====================================================================
void  Interpreter::RegisterAlias(const std::string& alias, const std::string& target)
{
	if(alias.empty()) return;

	if(target.empty()) {
		UnregisterAlias(alias);
	}
	else {
		aliases[alias] = target;
	}
}


//====================================================================
void  Interpreter::UnregisterAlias(const std::string& alias)
{
	aliases.erase(alias);
}


//====================================================================
void  Interpreter::UnregisterAllAliases()
{
	aliases.clear();
}


//====================================================================
bool  Interpreter::FindAlias(const std::string& alias, std::string& outTarget) const
{
	outTarget.erase();

	STRSTR_CITER it = aliases.find(alias);
	if(it != aliases.end()) {
		outTarget = it->second;
		return true;
	}
	return false;
}


//====================================================================
void  Interpreter::GetAliases(std::vector<std::pair<std::string, std::string> >& out) const
{
	out.reserve(aliases.size());

	STRSTR_CITER it = aliases.begin();
	while(it != aliases.end()) {
		out.push_back(*it++);
	}
}


//====================================================================
void  Interpreter::RegisterExternalCommand(Command* cmd)
{
	if(! cmd) return;
	UniqueContainerPushBack(externalCommands, cmd);
	Rehash();
}


//====================================================================
void  Interpreter::UnregisterAllExternalCommands()
{
	externalCommands.clear();
}


//====================================================================
bool  Interpreter::UnregisterExternalCommand(Command& cmd)
{
	bool bErased = EraseFromContainer(externalCommands, &cmd);
	if(bErased) Rehash();
	return bErased;
}


//====================================================================
uint  Interpreter::UnregisterExternalCommand(const std::string& name)
{
	uint numErased = EraseAllFromContainer_If(externalCommands, UnaryPredicateCommandName(name));
	if(numErased) Rehash();
	return numErased;
}


//====================================================================
void  Interpreter::DeleteAllExternalCommands()
{
	DeletePtrContainer(externalCommands);
}


//====================================================================
bool  Interpreter::DeleteExternalCommand(Command& cmd)
{
	bool bDeleted = DeleteFromPtrContainer(externalCommands, &cmd);
	if(bDeleted) Rehash();
	return bDeleted;
}


//====================================================================
uint  Interpreter::DeleteExternalCommand(const std::string& name)
{
	uint numDeleted = DeleteAllFromPtrContainer_If(externalCommands, UnaryPredicateCommandName(name));
	if(numDeleted) Rehash();
	return numDeleted;
}


//====================================================================
void  Interpreter::RegisterInternalCommand(Command* cmd)
{
	if(! cmd) return;
	UniqueContainerPushBack(internalCommands, cmd);
	Rehash();
}


//====================================================================
void  Interpreter::UnregisterAllInternalCommands()
{
	internalCommands.clear();
}


//====================================================================
bool  Interpreter::UnregisterInternalCommand(Command& cmd)
{
	bool bErased = EraseFromContainer(internalCommands, &cmd);
	if(bErased) Rehash();
	return bErased;
}


//====================================================================
uint  Interpreter::UnregisterInternalCommand(const std::string& name)
{
	uint numErased = EraseAllFromContainer_If(internalCommands, UnaryPredicateCommandName(name));
	if(numErased) Rehash();
	return numErased;
}


//====================================================================
void  Interpreter::DeleteAllInternalCommands()
{
	DeletePtrContainer(internalCommands);
}


//====================================================================
bool  Interpreter::DeleteInternalCommand(Command& cmd)
{
	bool bDeleted = DeleteFromPtrContainer(internalCommands, &cmd);
	if(bDeleted) Rehash();
	return bDeleted;
}


//====================================================================
uint  Interpreter::DeleteInternalCommand(const std::string& name)
{
	uint numDeleted = DeleteAllFromPtrContainer_If(internalCommands, UnaryPredicateCommandName(name));
	if(numDeleted) Rehash();
	return numDeleted;
}


//====================================================================
Command*  Interpreter::FindExternalCommand(const std::string& name)
{
	CMD_ITER it = std::find_if(externalCommands.begin(), externalCommands.end(),
								UnaryPredicateCommandName(name));
	if(it != externalCommands.end()) return *it;
	return 0;
}


//====================================================================
Command*  Interpreter::FindInternalCommand(const std::string& name)
{
	CMD_ITER it = std::find_if(internalCommands.begin(), internalCommands.end(),
								UnaryPredicateCommandName(name));
	if(it != internalCommands.end()) return *it;
	return 0;
}


//====================================================================
Command*  Interpreter::FindCommand(const std::string& name)
{
	// internal commands have higher priority
	Command* c = FindInternalCommand(name);
	if(c) return c;

	c = FindExternalCommand(name);
	return c;
}


//====================================================================
Command*  Interpreter::FindCommand(const CMD_VEC& commands, const type_info& ti)
{
	for(uint i = 0; i < commands.size(); i++) {
		if(typeid(*commands[i]) == ti) return commands[i];
	}
	return 0;
}


//====================================================================
bool  Interpreter::ExpandAliasPrefix(const std::string& commandLine, std::string& out)
{
	// find longest matching alias
	STRSTR_CITER itLongestMatchingAlias = aliases.end();
	uint maxLen = 0;

	STRSTR_CITER it = aliases.begin();
	while(it != aliases.end()) {

		const std::string& sAlias = it->first;
		uint pos = commandLine.find(sAlias);

		if(pos == 0) {
			// to substitute, whitespace or end-of-string must be next
			bool bSubst = true;
			if(commandLine.length() > sAlias.length()) {
				char nextChar = commandLine[sAlias.length()];
				if(nextChar != ' '  &&  nextChar != '\t') {
					bSubst = false;
				}
			}
			if(bSubst  &&  (sAlias.length() > maxLen)) {
				maxLen = sAlias.length();
				itLongestMatchingAlias = it;
			}
		}
		++it;
	}
	if(itLongestMatchingAlias != aliases.end()) {
		// found alias prefix, replace with alias-target
		out.assign(itLongestMatchingAlias->second);
		out.append(commandLine, itLongestMatchingAlias->first.length(),
					commandLine.length() - itLongestMatchingAlias->first.length());
		return true;
	}
	// not found, copy
	out = commandLine;
	return false;
}


//====================================================================
void  Interpreter::SortCommands(CMD_VEC& commands)
{
	std::stable_sort(commands.begin(), commands.end(), SortPredicateCommandName());
}


//====================================================================
void  Interpreter::ExtractCommandsWithMatchingPrefix(std::vector<std::string>& out,
						const CMD_VEC& commands, const std::string& prefix)
{
	if(prefix.empty()) return;

	for(uint i = 0; i < commands.size(); i++) {
		Command& cmd = *commands[i];
		if(cmd.GetName().find(prefix) == 0) {
			out.push_back(cmd.GetName());
		}
	}
}


//====================================================================
void  Interpreter::ExtractAliasesWithMatchingPrefix(std::vector<std::string>& out,
												const std::string& prefix)
{
	if(prefix.empty()) return;

	STRSTR_CITER it = aliases.begin();
	while(it != aliases.end()) {
		if(it->first.find(prefix) == 0) {
			out.push_back(it->first);
		}
		++it;
	}
}


//====================================================================
void  Interpreter::ExtractLongestCommonPrefix(const std::vector<std::string>& vals,
												std::string& out)
{
	out.erase();

	while(true) {
		uint testIndex = out.size();
		char outCh = 0;
		bool bFirstTest = true;

		for(uint i = 0; i < vals.size(); i++) {

			if(vals[i].length() <= testIndex) {
				// found shortest sequence
				return;
			}
			char ch = vals[i][testIndex];

			if(bFirstTest) {
				outCh = ch;
				bFirstTest = false;
			}
			else if(outCh != ch) {
				// found first difference
				return;
			}
		}
		// full match, add value to output
		out.append(1, outCh);
	}
}


//====================================================================
void  Interpreter::MakeMessages()
{
	messageNotRecognized = messageNotRecognizedHeader;

	Command* pCmd = FindCommand(internalCommands, typeid(ListCommandsCmd));
	if(pCmd) {
		if(messageNotRecognized.length()) {
			messageNotRecognized.append("\n");
		}
		messageNotRecognized.append("Type '").append(pCmd->GetName());
		messageNotRecognized.append("' to list valid commands.");
	}
}


//====================================================================
void  Interpreter::PushCommandInHistory(const std::string& s)
{
	if(pConsoleBuffers) {
		pConsoleBuffers->PushCommandInHistory(s);
	}
}


//====================================================================
void  Interpreter::DumpCommandHelp(std::ostream& os, const CommandLineParser& clp, Command& cmd)
{
	os << caption("Command-help");

	os << "TargetCommand: " << VW_GET_CLASS_NAME(cmd) << nl;
	os << "CurrentName..: " << cmd.GetName() << nl;
	os << "DefaultName..: " << cmd.GetDefaultName() << nl;
	os << "DevModeOnly..: " << ToString(cmd.IsDeveloperModeOnly()) << nl;
	os << "Summary......: " << cmd.GetShortDesc() << nl;

	if(GetGlobalHelpSwitch().length()) {
		os << "HelpSwitch...: " << clp.FormatOptionHelp(GetGlobalHelpSwitch()) << nl;
	}
	cmd.DumpHelp(os, clp);
	os << std::flush;
}


//====================================================================
std::string  Interpreter::GetInterpreterDesc() const
{
	return "Standard-command-interpreter v1.0";
}


//====================================================================
bool  Interpreter::InterpretCommandLine(const std::string& origCommandLine,
										std::ostream& os)
{
	lastCompletionPrefix.erase();

	std::string  noWSCommandLine = origCommandLine;
	TrimWhitespace(noWSCommandLine);

	std::string  commandLine;
	bool bAliasExpanded = ExpandAliasPrefix(noWSCommandLine, commandLine);

	// debug dump if requested
	if(bAliasExpanded  &&  bDebugDumpPreprocessedCommandLine) {
		os << e_debug;
		os << "Expanded to: " << commandLine << std::endl;
		os << e_info;
	}

	Command* pCommand = 0;
	bool bInterpreted = false;
	bool bExecute = false;

	try {
		// pre-process command-line
		commandLineParser.Parse(commandLine, true);

		// debug dump if requested
		if(bDebugDumpPreprocessedCommandLine) {
			os << e_debug;
			commandLineParser.DumpLastResults(os);
			os << std::flush << e_info;
		}

		// find command with matching name
		const std::string& sCommandName = commandLineParser.GetCommand();
		if(sCommandName.length()) {

			pCommand = FindCommand(sCommandName);
			if(pCommand) {
				bExecute = true;

				// check help-option
				if(GetGlobalHelpSwitch().length()  &&
					commandLineParser.GetOption(GetGlobalHelpSwitch()))
				{
					DumpCommandHelp(os, commandLineParser, *pCommand);
					bInterpreted = true;
					bExecute = false;
				}

				// check developer-mode flag
				if(! bDeveloperModeActive) {
					if(pCommand->IsDeveloperModeOnly()) {
						os << "Command: " << pCommand->GetName() << " is developer-mode only." << std::endl;
						bExecute = false;
					}
				}
			}
			else {
				// not recognized
				os << messageNotRecognized << std::endl;
			}
		}
	}
	catch(CommandLineErrorException& e) {
		bExecute = false;
		os << e_error << "Command-line error:\n";
		os << e.GetMessage() << std::endl << e_info;
	}

	// command-specific parse and execute
	if(bExecute) {
		bInterpreted = pCommand->ParseAndExecute(os, commandLineParser, commandLine);
		os << std::flush;
	}

	commandLineParser.ClearAll();
	PushCommandInHistory(origCommandLine);

	return bInterpreted;
}


//====================================================================
bool  Interpreter::CompleteCommandLine(const std::string& commandLinePrefix,
										std::string& completedCommandLine,
										std::ostream& os)
{
	if(lastCompletionPrefix == commandLinePrefix) return false;
	lastCompletionPrefix = commandLinePrefix;

	bool bCompleted = false;

	// pre-process command-line
	if(commandLinePrefix.empty()) return false;

	if(! commandLineParser.Parse(commandLinePrefix, false)) {
		commandLineParser.ClearAll();
		return false;
	}

	// complete only if there is no arguments,
	// find command/alias with closest matching prefix
	if(! commandLineParser.GetNumArguments()) {

		std::vector<std::string>  matchingNames;

		ExtractCommandsWithMatchingPrefix(matchingNames, internalCommands, commandLinePrefix);
		ExtractCommandsWithMatchingPrefix(matchingNames, externalCommands, commandLinePrefix);
		ExtractAliasesWithMatchingPrefix(matchingNames, commandLinePrefix);

		switch(matchingNames.size()) {
			case 0: {
				// no matching names
				break;
			}
			case 1: {
				// unique matching name
				if(commandLinePrefix.length() < matchingNames[0].length()) {
					completedCommandLine = matchingNames[0];
					bCompleted = true;
				}
				break;
			}
			default: {
				// multiple matching names
				std::string s("Matching commands/aliases: ");
				std::string sLine;

				for(uint i = 0; i < matchingNames.size(); i++) {
					sLine.append(matchingNames[i]);
					if(i < matchingNames.size() - 1) {
						sLine.append(", ");
						if(sLine.length() >= MAX_LINE_LEN_AT_COMMAND_COMPLETION) {
							s.append(sLine).append(1, '\n');
							sLine.erase();
						}
					}
				}
				os << s << sLine << std::endl;

				// complete command to last common character
				ExtractLongestCommonPrefix(matchingNames, completedCommandLine);
				bCompleted = true;
				break;
			}
		}
	}
	commandLineParser.ClearAll();
	return bCompleted;
}


//====================================================================
const char* DumpPreprocessedCommandLineCmd::s_szDefaultName = "dump-cmdline";
const char* DumpPreprocessedCommandLineCmd::s_szOptionOn    = "on";
const char* DumpPreprocessedCommandLineCmd::s_szOptionOff   = "off";


//====================================================================
DumpPreprocessedCommandLineCmd::DumpPreprocessedCommandLineCmd(Interpreter& interpreter)
	: InternalCommand(interpreter)
{
	SetDeveloperModeOnly(true);
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Control dump of preprocessed command-line");
}


//====================================================================
bool  DumpPreprocessedCommandLineCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
											const std::string& rawCommandLine)
{
	// update dump-param
	if(clp.GetOption(s_szOptionOn)) {
		GetInterpreter().SetDebugDumpPreprocessedCommandLine(true);
	}
	else if(clp.GetOption(s_szOptionOff)) {
		GetInterpreter().SetDebugDumpPreprocessedCommandLine(false);
	}

	// print parameters
	os << "DumpPreProcessedCommandLine: " << ToString(GetInterpreter().GetDebugDumpPreprocessedCommandLine()) << nl;
	return true;
}


//====================================================================
void  DumpPreprocessedCommandLineCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionOn, "enable dump of preprocessed command-line\n", 1);
	os << clp.FormatOptionHelp(s_szOptionOff, "disable dump of preprocessed command-line\n", 1);

	os << "Default behavior:\n\tprint parameters\n";
}


//====================================================================
const char* ListCommandsCmd::s_szDefaultName    = "list-commands";
const char* ListCommandsCmd::s_szOptionExternal = "ext";
const char* ListCommandsCmd::s_szOptionInternal = "int";


//====================================================================
ListCommandsCmd::ListCommandsCmd(Interpreter& interpreter)
	: InternalCommand(interpreter)
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("List commands registered with command-interpreter");
}


//====================================================================
bool  ListCommandsCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	bool bDumpInternal = clp.GetOption(s_szOptionInternal);
	bool bDumpExternal = clp.GetOption(s_szOptionExternal);

	if(! bDumpInternal  &&  ! bDumpExternal) {
		bDumpInternal = true;
		bDumpExternal = true;
	}

	// dump developer-mode state
	os << "DeveloperMode: " << ToString(GetInterpreter().IsActiveDeveloperMode()) << nl;
	if(bDumpInternal || bDumpExternal) os << nl;

	// list internal commands
	if(bDumpInternal) {
		os << "Internal commands:\n";

		uint maxNameLen = 0;
		{for(uint i = 0; i < GetInterpreter().GetNumInternalCommands(); i++) {
			maxNameLen = Max<uint>(maxNameLen, GetListLinePrefixLen(GetInterpreter().GetInternalCommand(i)));
		}}
		{for(uint i = 0; i < GetInterpreter().GetNumInternalCommands(); i++) {
			os << MakeListLine(GetInterpreter().GetInternalCommand(i), maxNameLen) << nl;
		}}

		if(bDumpExternal) {
			os << nl;
		}
	}

	// list external commands
	if(bDumpExternal) {
		os << "External commands:\n";

		uint maxNameLen = 0;
		{for(uint i = 0; i < GetInterpreter().GetNumExternalCommands(); i++) {
			maxNameLen = Max<uint>(maxNameLen, GetListLinePrefixLen(GetInterpreter().GetExternalCommand(i)));
		}}
		{for(uint i = 0; i < GetInterpreter().GetNumExternalCommands(); i++) {
			os << MakeListLine(GetInterpreter().GetExternalCommand(i), maxNameLen) << nl;
		}}
	}
	return true;
}


//====================================================================
void  ListCommandsCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionInternal, "list internal commands\n", 1);
	os << clp.FormatOptionHelp(s_szOptionExternal, "list external commands\n", 1);

	os << "Default behavior:\n\tlist all commands\n";
}


//====================================================================
uint  ListCommandsCmd::GetListLinePrefixLen(Command& cmd)
{
	uint len = cmd.GetName().length() + 2; //tab
	if(cmd.IsDeveloperModeOnly()) len += 4;
	return len;
}


//====================================================================
std::string  ListCommandsCmd::MakeListLine(Command& cmd, uint maxPrefixLen)
{
	std::string s(2, ' '); //tab
	s.append(cmd.GetName());
	if(cmd.IsDeveloperModeOnly()) s.append("[DM]");
	uint wsCount = maxPrefixLen - s.length();
	s.append(", ");
	if(wsCount) s.append(wsCount, ' ');
	s.append(cmd.GetShortDesc());
	return s;
}


//====================================================================
const char* HelpCmd::s_szDefaultName            = "help";
const char* HelpCmd::s_szAssignGlobalHelpSwitch = "switch";


//====================================================================
HelpCmd::HelpCmd(Interpreter& interpreter)
	: InternalCommand(interpreter)
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Print help for any command");
}


//====================================================================
bool  HelpCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	// global help switch
	std::string sVal;
	if(clp.GetAssignment(s_szAssignGlobalHelpSwitch, sVal)) {
		GetInterpreter().SetGlobalHelpSwitch(sVal);
		os << "GlobalHelpSwitch: " << clp.FormatOptionHelp(GetInterpreter().GetGlobalHelpSwitch()) << nl;
		return true;
	}

	// help for target command
	if(clp.GetNumArguments()) {
		const CommandLineParser::Argument& arg = clp.GetArgument(0);
		if(arg.Type() == CommandLineParser::ArgumentType_String) {

			Command* cmd = GetInterpreter().GetCommand(arg.String());
			if(cmd) {
				GetInterpreter().DumpCommandHelp(os, clp, *cmd);
			}
			else {
				os << "Command: " << arg.String() << " not found.\n";
			}
			return true;
		}
	}

	// print usage
	GetInterpreter().DumpCommandHelp(os, clp, *this);
	os << "GlobalHelpSwitch: " << clp.FormatOptionHelp(GetInterpreter().GetGlobalHelpSwitch()) << nl;
	return true;
}


//====================================================================
void  HelpCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Assignments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignGlobalHelpSwitch, "string", "sets global help switch\n", 1);

	os << "Arguments:\n";
	os << clp.FormatStringHelp("0) target-name", "command whose help will be printed\n", 1);

	os << "Default behavior:\n\tprint usage\n";
}


//====================================================================
const char* AliasCmd::s_szDefaultName = "alias";
const char* AliasCmd::s_szOptionClear = "clear";


//====================================================================
AliasCmd::AliasCmd(Interpreter& interpreter)
	: InternalCommand(interpreter)
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Setup command shortcuts");
}


//====================================================================
bool  AliasCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
							const std::string& rawCommandLine)
{
	// clear all aliases
	if(clp.GetOption(s_szOptionClear)) {
		GetInterpreter().UnregisterAllAliases();
		os << "Unregistered all aliases.\n";
		return true;
	}

	// register/unregister new alias
	if(clp.GetNumArguments() == 1) {
		const CommandLineParser::Argument& arg = clp.GetArgument(0);
		if(arg.Type() == CommandLineParser::ArgumentType_Assignment) {

			// check for special cases (cannot alias AliasCommandName to something else)
			if(arg.Assignment().Name() == GetName()) {
				os << "Alias not allowed: " << GetName() << nl;
				return false;
			}
			GetInterpreter().RegisterAlias(arg.Assignment().Name(), arg.Assignment().Value());
			os << "Registered alias: " << arg.Assignment().Name() << nl;
			return true;
		}
	}

	// list aliases
	std::vector<std::pair<std::string, std::string> >  aliases;
	GetInterpreter().GetAliases(aliases);

	os << "Alias list:\n";
	for(uint i = 0; i < aliases.size(); i++) {
		os << "  " << aliases[i].first << "='" << aliases[i].second << "'\n";
	}
	return true;
}


//====================================================================
void  AliasCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionClear, "clear aliases\n", 1);

	os << "Assignments:\n";
	os << clp.FormatAssignmentHelp("alias", "target", "assign target to alias, empty target unregisters alias\n", 1);

	os << "Default behavior:\n\tlist aliases\n";
}


//====================================================================
const char* ConsoleBuffersCmd::s_szDefaultName        = "console-buffers";
const char* ConsoleBuffersCmd::s_szOptionClearOutput  = "clear-output";
const char* ConsoleBuffersCmd::s_szOptionClearHistory = "clear-history";
const char* ConsoleBuffersCmd::s_szAssignMaxOutput    = "max-output";
const char* ConsoleBuffersCmd::s_szAssignMaxHistory   = "max-history";


//====================================================================
ConsoleBuffersCmd::ConsoleBuffersCmd(Interpreter& interpreter)
	: InternalCommand(interpreter)
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Control console-buffers");
}


//====================================================================
bool  ConsoleBuffersCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	ConsoleBuffers* pConBuf = GetInterpreter().GetConsoleBuffers();
	if(! pConBuf) {
		os << "Console-buffers not hooked to interpreter.\n";
		return true;
	}

	// clear output
	if(clp.GetOption(s_szOptionClearOutput)) {
		pConBuf->ClearOutput();
		return true;
	}

	// clear history
	if(clp.GetOption(s_szOptionClearHistory)) {
		pConBuf->ClearCommandHistory();
		return true;
	}

	std::string  sVal;

	// limit number of output-lines
	uint  maxOutputLines = 0;
	if(clp.ExtractArgumentUInteger(s_szAssignMaxOutput, maxOutputLines)) {
		pConBuf->SetMaxOutputLines(maxOutputLines);
	}

	// limit number of commands in history
	uint  maxCommandHistory = 0;
	if(clp.ExtractArgumentUInteger(s_szAssignMaxHistory, maxCommandHistory)) {
		pConBuf->SetMaxCommandsInHistory(maxCommandHistory);
	}

	// print parameters
	os << "MaxOutputLines......: " << pConBuf->GetMaxOutputLines() << nl;
	os << "MaxCommandsInHistory: " << pConBuf->GetMaxCommandsInHistory() << nl;
	return true;
}


//====================================================================
void  ConsoleBuffersCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionClearOutput, "clear output buffer\n", 1);
	os << clp.FormatOptionHelp(s_szOptionClearHistory, "clear command history\n", 1);

	os << "Assignments:\n";
	os << clp.FormatAssignmentHelp(s_szAssignMaxOutput, "uinteger", "limit number of output-lines\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignMaxHistory, "uinteger", "limit number of commands in history\n", 1);

	os << "Default behavior:\n\tprint parameters\n";
}


//====================================================================
const char* HistoryCmd::s_szDefaultName    = "history";
const char* HistoryCmd::s_szOptionPrintAll = "a";


//====================================================================
HistoryCmd::HistoryCmd(Interpreter& interpreter)
	: InternalCommand(interpreter)
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Print command history");

	SetDefaultNumPrinted();
}


//====================================================================
void  HistoryCmd::SetDefaultNumPrinted(uint numPrinted)
{
	defaultNumPrinted = ClampedLow<uint>(numPrinted, 1);
}


//====================================================================
bool  HistoryCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	ConsoleBuffers* pConBuf = GetInterpreter().GetConsoleBuffers();
	if(! pConBuf) {
		os << "Console-buffers not hooked to interpreter.\n";
		return true;
	}

	// number of printed commands
	if(clp.GetNumArguments()) {
		const CommandLineParser::Argument& arg = clp.GetArgument(0);
		if(arg.Type() == CommandLineParser::ArgumentType_String) {
			uint value;
			if(ToValue<uint>(arg.String(), value)) {
				SetDefaultNumPrinted(value);
			}
		}
	}

	// print all
	bool bPrintAll = clp.GetOption(s_szOptionPrintAll);

	// do print history
	os << "Command-history:\n";

	pConBuf->LockCommandHistory();
	int numInHist = pConBuf->GetNumCommandsInHistory();

	if(numInHist) {

		int numPrinted = numInHist;
		if(! bPrintAll) ClampHigh<int>(numPrinted, defaultNumPrinted);

		for(uint i = numInHist - numPrinted; i < numInHist; i++) {
			os << i << ") " << pConBuf->GetCommandFromHistory(i) << nl;
		}
	}
	pConBuf->UnlockCommandHistory();
	return true;
}


//====================================================================
void  HistoryCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionPrintAll, "print all commands in history\n", 1);

	os << "Arguments:\n";
	os << clp.FormatStringHelp("0) uinteger", "set (default) number of commands printed\n", 1);

	os << "Default behavior:\n\tprint tail of command-history: " << defaultNumPrinted << nl;
}


//********************************************************************
}//namespace std_command_interpreter


//********************************************************************
}//namespace vertexware


