

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\CommandInterpreter.h"
#include "BaseTools\CommandLineParser.h"
#include "BaseTools\ConsoleBuffers.h"
#include <vector>
#include <map>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace std_command_interpreter {


//====================================================================
// Registered commands are owned by interpreter.
//====================================================================
class VW_COMMON_API Interpreter : public CommandInterpreter
{
public:
	Interpreter();
	virtual ~Interpreter();

	// if 0 uses kernel buffers
	void             SetConsoleBuffers(ConsoleBuffers* pConsoleBuffers = 0);
	ConsoleBuffers*  GetConsoleBuffers() { return pConsoleBuffers; }

	void  SetDeveloperModeActive(bool bActive) { bDeveloperModeActive = bActive; }
	bool  IsActiveDeveloperMode() const { return bDeveloperModeActive; }

	// updates internal structures, use if registered commands are changed externally
	void  Rehash();

	void                SetGlobalHelpSwitch(const std::string& s = "h");
	const std::string&  GetGlobalHelpSwitch() const { return globalHelpSwitch; }

	void                SetCommandNotRecognizedMessage(const std::string& s = "Command not recognized.");
	const std::string&  GetCommandNotRecognizedMessage() const { return messageNotRecognizedHeader; }

	void  SetDebugDumpPreprocessedCommandLine(bool bDump = false) { bDebugDumpPreprocessedCommandLine = bDump; }
	bool  GetDebugDumpPreprocessedCommandLine() const { return bDebugDumpPreprocessedCommandLine; }

	void  RegisterAlias(const std::string& alias, const std::string& target);
	void  UnregisterAlias(const std::string& alias);
	void  UnregisterAllAliases();

	bool  FindAlias(const std::string& alias, std::string& outTarget) const;
	void  GetAliases(std::vector<std::pair<std::string, std::string> >& out) const;

	void  RegisterExternalCommand(Command* cmd);

	void  UnregisterAllExternalCommands();
	bool  UnregisterExternalCommand(Command& cmd);
	uint  UnregisterExternalCommand(const std::string& name); //returns number of unregistered commands

	void  DeleteAllExternalCommands();
	bool  DeleteExternalCommand(Command& cmd);
	uint  DeleteExternalCommand(const std::string& name); //returns number of deleted commands

	uint            GetNumExternalCommands() const { return externalCommands.size(); }
	Command&        GetExternalCommand(uint i) { return *externalCommands[i]; }
	const Command&  GetExternalCommand(uint i) const { return const_cast<Interpreter*>(this)->GetExternalCommand(i); }
	Command*        GetExternalCommand(const std::string& name) { return FindExternalCommand(name); }
	const Command*  GetExternalCommand(const std::string& name) const { return const_cast<Interpreter*>(this)->GetExternalCommand(name); }

	uint            GetNumInternalCommands() const { return internalCommands.size(); }
	Command&        GetInternalCommand(uint i) { return *internalCommands[i]; }
	const Command&  GetInternalCommand(uint i) const { return const_cast<Interpreter*>(this)->GetInternalCommand(i); }
	Command*        GetInternalCommand(const std::string& name) { return FindInternalCommand(name); }
	const Command*  GetInternalCommand(const std::string& name) const { return const_cast<Interpreter*>(this)->GetInternalCommand(name); }

	Command*        GetCommand(const std::string& name) { return FindCommand(name); }
	const Command*  GetCommand(const std::string& name) const { return const_cast<Interpreter*>(this)->GetCommand(name); }

	void  DumpCommandHelp(std::ostream& os, const CommandLineParser& clp, Command& cmd);

	//-----------------------------------
	// CommandInterpreter implementation.
	//-----------------------------------
	virtual  std::string  GetInterpreterDesc() const;

	virtual  bool  InterpretCommandLine(const std::string& commandLine,
										std::ostream& os);

	virtual  bool  CompleteCommandLine(const std::string& commandLinePrefix,
										std::string& completedCommandLine,
										std::ostream& os);


private:
	typedef  std::vector<Command*>    CMD_VEC;
	typedef  CMD_VEC::iterator        CMD_ITER;
	typedef  CMD_VEC::const_iterator  CMD_CITER;

	typedef  std::map<std::string, std::string>  STRSTR_MAP;
	typedef  STRSTR_MAP::iterator                STRSTR_ITER;
	typedef  STRSTR_MAP::const_iterator          STRSTR_CITER;

	CMD_VEC     externalCommands;
	CMD_VEC     internalCommands;
	STRSTR_MAP  aliases;

	CommandLineParser  commandLineParser;
	ConsoleBuffers*    pConsoleBuffers;

	bool  bDebugDumpPreprocessedCommandLine;
	bool  bDeveloperModeActive;

	std::string  globalHelpSwitch;

	std::string  lastCompletionPrefix;
	std::string  messageNotRecognized;
	std::string  messageNotRecognizedHeader;


protected:
	void  RegisterInternalCommand(Command* cmd);

	void  UnregisterAllInternalCommands();
	bool  UnregisterInternalCommand(Command& cmd);
	uint  UnregisterInternalCommand(const std::string& name); //returns number of unregistered commands

	void  DeleteAllInternalCommands();
	bool  DeleteInternalCommand(Command& cmd);
	uint  DeleteInternalCommand(const std::string& name); //returns number of deleted commands

	Command*  FindExternalCommand(const std::string& name);
	Command*  FindInternalCommand(const std::string& name);
	Command*  FindCommand(const std::string& name);
	Command*  FindCommand(const CMD_VEC& commands, const type_info& ti);

	bool  ExpandAliasPrefix(const std::string& commandLine, std::string& out);
	void  SortCommands(CMD_VEC& commands);
	void  ExtractCommandsWithMatchingPrefix(std::vector<std::string>& out, const CMD_VEC& commands, const std::string& prefix);
	void  ExtractAliasesWithMatchingPrefix(std::vector<std::string>& out, const std::string& prefix);
	void  ExtractLongestCommonPrefix(const std::vector<std::string>& vals, std::string& out);
	void  MakeMessages();
	void  PushCommandInHistory(const std::string& s);
};


//====================================================================
class VW_COMMON_API InternalCommand : public Command
{
public:
	InternalCommand(Interpreter& _interpreter) : interpreter(_interpreter) {
	}

protected:
	Interpreter&  GetInterpreter() { return interpreter; }

private:
	Interpreter&  interpreter;
};


//====================================================================
class VW_COMMON_API DumpPreprocessedCommandLineCmd : public InternalCommand
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionOn;
	static const char* s_szOptionOff;


public:
	DumpPreprocessedCommandLineCmd(Interpreter& interpreter);

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
class VW_COMMON_API ListCommandsCmd : public InternalCommand
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionExternal;
	static const char* s_szOptionInternal;


public:
	ListCommandsCmd(Interpreter& interpreter);

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);


private:
	uint         GetListLinePrefixLen(Command& cmd);
	std::string  MakeListLine(Command& cmd, uint maxPrefixLen);
};


//====================================================================
class VW_COMMON_API HelpCmd : public InternalCommand
{
public:
	static const char* s_szDefaultName;
	static const char* s_szAssignGlobalHelpSwitch;


public:
	HelpCmd(Interpreter& interpreter);

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
class VW_COMMON_API AliasCmd : public InternalCommand
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionClear;


public:
	AliasCmd(Interpreter& interpreter);

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
class VW_COMMON_API ConsoleBuffersCmd : public InternalCommand
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionClearOutput;
	static const char* s_szOptionClearHistory;
	static const char* s_szAssignMaxOutput;
	static const char* s_szAssignMaxHistory;


public:
	ConsoleBuffersCmd(Interpreter& interpreter);

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);
};


//====================================================================
class VW_COMMON_API HistoryCmd : public InternalCommand
{
public:
	static const char* s_szDefaultName;
	static const char* s_szOptionPrintAll;


public:
	HistoryCmd(Interpreter& interpreter);

	void  SetDefaultNumPrinted(uint numPrinted = 10);
	uint  GetDefaultNumPrinted() const { return defaultNumPrinted; }

	virtual  bool  ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine);

	virtual  void  DumpHelp(std::ostream& os, const CommandLineParser& clp);


private:
	uint  defaultNumPrinted;
};


//********************************************************************
}//namespace std_command_interpreter


//********************************************************************
}//namespace vertexware


