

//--------------------------------------------------------------------
#include "Framework\VideoProperties.h"
#include "BaseTools\Debug.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\PlatformInfo.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"
#include <iomanip>
#include <algorithm>
#include <set>


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
struct FBCSelectionTarget
{
	const FrameBufferConfig*  fbc;
	int                       rankPrimary;
	int                       rankSecondary;

	FBCSelectionTarget(const FrameBufferConfig* _fbc = 0) : fbc(_fbc) {
		rankPrimary = 0;
		rankSecondary = 0;
	}

	friend bool __fastcall operator<(const FBCSelectionTarget& lhs, const FBCSelectionTarget& rhs)
	{
		return
		(lhs.rankPrimary < rhs.rankPrimary) ||
		(lhs.rankPrimary == rhs.rankPrimary  &&  lhs.rankSecondary < rhs.rankSecondary);
	}
};


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
VideoHints::VideoHints(bool _vSync, bool _fullscreenDestinationAlpha,
						bool _pureDeviceMode, eVertexProcessingMode _vertexProcessingMode)
{
	vSync = _vSync;
	fullscreenDestinationAlpha = _fullscreenDestinationAlpha;
	pureDeviceMode = _pureDeviceMode;
	vertexProcessingMode = _vertexProcessingMode;
}


//====================================================================
std::string  VideoHints::FormatVertexProcessingMode(eVertexProcessingMode vpm)
{
	switch(vpm) {
		case VertexProcessingMode_Default:
			return "Default";

		case VertexProcessingMode_Software:
			return "Software";

		case VertexProcessingMode_Mixed:
			return "Mixed";

		case VertexProcessingMode_Hardware:
			return "Hardware";
	}
	return "Unknown";
}


//====================================================================
bool  VideoHints::operator==(const VideoHints& rhs) const
{
	return
	(vSync == rhs.vSync) && (fullscreenDestinationAlpha == rhs.fullscreenDestinationAlpha) &&
	(pureDeviceMode == rhs.pureDeviceMode) && (vertexProcessingMode == rhs.vertexProcessingMode);
}


//====================================================================
void  VideoHints::Dump(std::ostream& os, uint tabLevel) const
{
	TabEffector te(tabLevel);

	os << te << "VSync: " << ToString(vSync) << nl;
	os << te << "FullscreenDestinationAlpha: " << ToString(fullscreenDestinationAlpha) << nl;
	os << te << "PureDeviceMode: " << ToString(pureDeviceMode) << nl;
	os << te << "VertexProcessingMode: " << FormatVertexProcessingMode(vertexProcessingMode) << nl;
}


//====================================================================
DisplayModeConfig::DisplayModeConfig(bool bFillWithCurrent)
{
	if(bFillWithCurrent) {
		platform_info::GetDesktopDisplayMode(&width, &height,
								&bitsPerPixel, &refreshFrequency);
	}
	else {
		width = 640;
		height = 480;
		bitsPerPixel = 32;
		refreshFrequency = 0;
	}
}


//====================================================================
DisplayModeConfig::DisplayModeConfig(uint _width, uint _height,
									uint _bitsPerPixel, uint _refreshFrequency)
{
	width = _width;
	height = _height;
	bitsPerPixel = _bitsPerPixel;
	refreshFrequency = _refreshFrequency;
}


//====================================================================
bool  DisplayModeConfig::operator==(const DisplayModeConfig& rhs) const
{
	return
	(width == rhs.width) && (height == rhs.height) &&
	(bitsPerPixel == rhs.bitsPerPixel) && (refreshFrequency == rhs.refreshFrequency);
}


//====================================================================
std::string  DisplayModeConfig::MakeDesc(const std::string& sSep,
										const std::string& sFreqSep) const
{
	std::ostringstream oss;
	oss << width << sSep << height << sSep << bitsPerPixel << sFreqSep << refreshFrequency;
	return oss.str();
}


//====================================================================
bool __fastcall  DisplayModeConfigFilter::IsSafeDisplayModeConfig(const DisplayModeConfig& dmc)
{
	return
	(dmc.width >= 640)  &&
	(dmc.height >= 480)  &&
	(dmc.bitsPerPixel >= 32)  &&
	(dmc.refreshFrequency >= 60);
}


//====================================================================
bool __fastcall  DisplayModeConfigFilter::IsStandardResolution(uint width, uint height)
{
	struct Resolution { uint w, h; };		
	Resolution r[] =
	{
		{ 640, 480 },
		{ 800, 600 },
		{ 1024, 768 },
		{ 1280, 1024 },
		{ 1600, 1200 }
	};
	const uint count = sizeof(r) / sizeof(Resolution);
	for(uint i = 0; i < count; i++) {
		if((width == r[i].w)  &&  (height == r[i].h)) return true;
	}
	return false;
}


//====================================================================
bool __fastcall  DisplayModeConfigFilter::IsSubDesktopResolution(uint width, uint height)
{
	DisplayModeConfig dmcDesktop(true);
	return
	(width <= dmcDesktop.width)  &&  (height <= dmcDesktop.height);
}


//====================================================================
uint  DisplayModeProperties::GetNumRefreshFrequencies() const
{
	return refreshFrequencies.size();
}


//====================================================================
uint  DisplayModeProperties::GetRefreshFrequency(uint i) const
{
	VW_ASSERT(i < refreshFrequencies.size(), "Index out of range");
	return refreshFrequencies[i];
}


//====================================================================
uint  DisplayModeProperties::GetLowestRefreshFrequency() const
{
	if(refreshFrequencies.empty()) return 0;
	uint min = UINT_MAX;
	for(uint i = 0; i < refreshFrequencies.size(); i++) {
		min = Min(min, refreshFrequencies[i]);
	}
	return min;
}


//====================================================================
uint  DisplayModeProperties::GetHighestRefreshFrequency() const
{
	uint max = 0;
	for(uint i = 0; i < refreshFrequencies.size(); i++) {
		max = Max(max, refreshFrequencies[i]);
	}
	return max;
}


//====================================================================
bool  DisplayModeProperties::IsValidRefreshFrequency(uint freq) const
{
	// 0 is valid, it means default
	if(freq == 0) return true;

	for(uint i = 0; i < refreshFrequencies.size(); i++) {
		if(refreshFrequencies[i] == freq) return true;
	}
	return false;
}


//====================================================================
bool  DisplayModeProperties::IsValid(const DisplayModeConfig& dmc) const
{
	return
	(dmc.width == width) && (dmc.height == height) &&
	(dmc.bitsPerPixel == bitsPerPixel) && IsValidRefreshFrequency(dmc.refreshFrequency);
}


//====================================================================
std::string  DisplayModeProperties::MakeDesc(const std::string& sSep) const
{
	std::ostringstream oss;
	oss << width << sSep << height << sSep << bitsPerPixel;
	return oss.str();
}


//====================================================================
void  DisplayModeProperties::SortRefreshFrequencies()
{
	// make unique
	std::set<uint>  setFreq;
	for(uint i = 0; i < refreshFrequencies.size(); i++) {
		setFreq.insert(refreshFrequencies[i]);
	}

	// extract sorted
	refreshFrequencies.clear();

	std::set<uint>::iterator it = setFreq.begin();
	while(it != setFreq.end()) {
		refreshFrequencies.push_back(*it++);
	}
}


//====================================================================
uint  DisplayModeProperties::GetClosestLowerRefreshFrequency(uint refFreq) const
{
	int iBestFreq = -1;
	int iAbsBestDiff = INT_MAX;

	for(uint i = 0; i < refreshFrequencies.size(); i++) {

		int diff = refFreq - refreshFrequencies[i];
		if(diff == 0) return refFreq;
		if(diff < 0) {
			int absDiff = Abs(diff);
			if(absDiff < iAbsBestDiff) {
				iAbsBestDiff = absDiff;
				iBestFreq = i;
			}
		}
	}
	if(iBestFreq >= 0) return refreshFrequencies[iBestFreq];
	return 0;
}


//====================================================================
FrameBufferConfig::FrameBufferConfig(bool _fullscreen,
									bool _hardwareRasterizer, uint _backBufferCount,
									uint _bitsDepth, uint _bitsStencil, uint _bitsAccum)
{
	fullscreen = _fullscreen;
	hardwareRasterizer = _hardwareRasterizer;
	backBufferCount = _backBufferCount;
	
	bitsRed   = 8;
	bitsGreen = 8;
	bitsBlue  = 8;
	bitsAlpha = 8;

	bitsDepth   = _bitsDepth;
	bitsStencil = _bitsStencil;

	bitsAccumRed   = _bitsAccum;
	bitsAccumGreen = _bitsAccum;
	bitsAccumBlue  = _bitsAccum;
	bitsAccumAlpha = _bitsAccum;
}


//====================================================================
bool  FrameBufferConfig::operator==(const FrameBufferConfig& rhs) const
{
	return
	(fullscreen == rhs.fullscreen) && (hardwareRasterizer == rhs.hardwareRasterizer) &&
	(backBufferCount == rhs.backBufferCount) &&
	(bitsRed == rhs.bitsRed) && (bitsGreen == rhs.bitsGreen) &&
	(bitsBlue == rhs.bitsBlue) && (bitsAlpha == rhs.bitsAlpha) &&
	(bitsDepth == rhs.bitsDepth) && (bitsStencil == rhs.bitsStencil) &&
	(bitsAccumRed == rhs.bitsAccumRed) && (bitsAccumGreen == rhs.bitsAccumGreen) &&
	(bitsAccumBlue == rhs.bitsAccumBlue) && (bitsAccumAlpha == rhs.bitsAccumAlpha);
}


//====================================================================
std::string  FrameBufferConfig::MakeLayoutDesc(const std::string& sBitSep,
										const std::string& sBufSep) const
{
	std::ostringstream  oss;
	oss << bitsRed << sBitSep << bitsGreen << sBitSep << bitsBlue << sBitSep << bitsAlpha;
	oss << sBufSep;
	oss << bitsDepth << sBitSep << bitsStencil;
	oss << sBufSep;
	oss << bitsAccumRed << sBitSep << bitsAccumGreen << sBitSep << bitsAccumBlue << sBitSep << bitsAccumAlpha;
	return oss.str();
}


//====================================================================
void  FrameBufferConfig::Dump(std::ostream& os, uint tabLevel) const
{
	const char dot = '.';
	TabEffector te(tabLevel);

	os << te << "Fullscreen: " << ToString(fullscreen) << nl;
	os << te << "HardwareRasterizer: " << ToString(hardwareRasterizer) << nl;
	os << te << "BackBufferCount: " << backBufferCount << nl;
	os << te << "BackBufferLayout: " << bitsRed << dot << bitsGreen << dot << bitsBlue << dot << bitsAlpha << nl;
	os << te << "DepthStencilBufferLayout: " << bitsDepth << dot << bitsStencil << nl;
	os << te << "AccumulationBufferLayout: " << bitsAccumRed << dot << bitsAccumGreen << dot << bitsAccumBlue << dot << bitsAccumAlpha << nl;
}


//====================================================================
VideoConfig::VideoConfig(bool _bResizable)
{
	resizable = _bResizable;
}


//====================================================================
bool  VideoConfig::operator==(const VideoConfig& rhs) const
{
	return
	(displayModeConfig == rhs.displayModeConfig) &&
	(frameBufferConfig == rhs.frameBufferConfig) &&
	(resizable == rhs.resizable) &&
	(videoHints == rhs.videoHints);
}


//====================================================================
bool  VideoConfig::IsEqualIgnoreVSync(const VideoConfig& rhs) const
{
	VideoConfig vcLHS = *this;
	VideoConfig vcRHS = rhs;

	vcLHS.videoHints.vSync = vcRHS.videoHints.vSync;

	if(rhs.frameBufferConfig.fullscreen) {
		vcLHS.resizable = vcRHS.resizable;
	}
	else {
		vcLHS.displayModeConfig.bitsPerPixel = vcRHS.displayModeConfig.bitsPerPixel;
		vcLHS.displayModeConfig.refreshFrequency = vcRHS.displayModeConfig.refreshFrequency;
	}
	return (vcLHS == vcRHS);
}


//====================================================================
void  VideoConfig::Dump(std::ostream& os, uint tabLevel) const
{
	TabEffector te(tabLevel);
	frameBufferConfig.Dump(os, tabLevel);
	os << te << displayModeConfig.MakeDesc() << nl;
	os << te << "Resizable: " << ToString(resizable) << nl;
	videoHints.Dump(os, tabLevel);
}


//====================================================================
const std::string&  VideoProfile::GetDriverDescription() const
{
	return driverDescription;
}


//====================================================================
const std::string&  VideoProfile::GetDriverVersion() const
{
	return driverVersion;
}


//====================================================================
const win_utils::Guid&  VideoProfile::GetDriverGuid() const
{
	return driverGuid;
}


//====================================================================
uint  VideoProfile::GetNumDisplayModeProperties() const
{
	return displayModeProperties.size();
}


//====================================================================
const DisplayModeProperties&  VideoProfile::GetDisplayModeProperties(uint i) const
{
	VW_ASSERT(i < displayModeProperties.size(), "Index out of range");
	return displayModeProperties[i];
}


//====================================================================
uint  VideoProfile::GetNumFrameBufferConfigs() const
{
	return frameBufferConfigs.size();
}


//====================================================================
const FrameBufferConfig&  VideoProfile::GetFrameBufferConfig(uint i) const
{
	VW_ASSERT(i < frameBufferConfigs.size(), "Index out of range");
	return frameBufferConfigs[i];
}


//====================================================================
bool  VideoProfile::IsValid(const DisplayModeConfig& dmc) const
{
	for(uint i = 0; i < displayModeProperties.size(); i++) {
		if(displayModeProperties[i].IsValid(dmc)) return true;
	}
	return false;
}


//====================================================================
bool  VideoProfile::IsValid(const FrameBufferConfig& fbc) const
{
	for(uint i = 0; i < frameBufferConfigs.size(); i++) {
		if(frameBufferConfigs[i] == fbc) return true;
	}
	return false;
}


//====================================================================
bool  VideoProfile::IsValid(const VideoConfig& vc) const
{
	if(! IsValid(vc.frameBufferConfig)) return false;

	if(vc.frameBufferConfig.fullscreen) {
		if(! IsValid(vc.displayModeConfig)) return false;
	}
	return true;
}


//====================================================================
const DisplayModeProperties*  VideoProfile::SelectClosestLowerDisplayModeProperties(
											const DisplayModeConfig& dmcRef) const
{
	int  iBestDMP = -1;
	uint iBestRank = UINT_MAX;

	for(uint i = 0; i < displayModeProperties.size(); i++) {
		const DisplayModeProperties& dmp = displayModeProperties[i];

		if((dmp.bitsPerPixel >= dmcRef.bitsPerPixel)  ||
			(dmp.width <= dmcRef.width)  &&  (dmp.height <= dmcRef.height))
		{
			uint rank = (dmp.bitsPerPixel - dmcRef.bitsPerPixel) * 1000 * 1000;
			rank += (dmcRef.width - dmp.width) * 1000;
			rank += (dmcRef.height - dmp.height);

			if(rank < iBestRank) {
				iBestRank = rank;
				iBestDMP = i;
			}
		}
	}

	if(iBestDMP >= 0) return &displayModeProperties[iBestDMP];
	return 0;
}


//====================================================================
const FrameBufferConfig*  VideoProfile::SelectClosestFrameBufferConfig(
											const FrameBufferConfig& fbcRef) const
{
	std::vector<FBCSelectionTarget>  selectionTargets;

	// rank points
	const int cPoints_Max                 = 1000000000;
	const int cPoints_Fullscreen          =  100000000;
	const int cPoints_Hardware            =  100000000;
	const int cPoints_BackBufferBit       =    1000000; // expected bit-count [32-64]
	const int cPoints_StencilBufferBit    =      10000; // expected bit-count [0-32]
	const int cPoints_DepthBufferBit      =        100; // expected bit-count [0-64]

	const int cPoints_AccumBufferBit      =   10000000; // expected bit-count [0-128]
	const int cPoints_BackBufferCountDiff =    1000000; // expected buffer-count [1-4]

	// collect valid targets
	{for(uint i = 0; i < frameBufferConfigs.size(); i++) {
		const FrameBufferConfig& fbc = frameBufferConfigs[i];

		// accumulate selection rank
		int rankPrimary = cPoints_Max;
		int rankSecondary = cPoints_Max;

		// switches
		if(fbc.fullscreen != fbcRef.fullscreen) rankPrimary -= cPoints_Fullscreen;
		if(fbc.hardwareRasterizer != fbcRef.hardwareRasterizer) rankPrimary -= cPoints_Hardware;

		// back-buffer bit-counts
		rankPrimary -= Abs<int>(fbc.bitsRed - fbcRef.bitsRed) * cPoints_BackBufferBit;
		rankPrimary -= Abs<int>(fbc.bitsGreen - fbcRef.bitsGreen) * cPoints_BackBufferBit;
		rankPrimary -= Abs<int>(fbc.bitsBlue - fbcRef.bitsBlue) * cPoints_BackBufferBit;
		rankPrimary -= Abs<int>(fbc.bitsAlpha - fbcRef.bitsAlpha) * cPoints_BackBufferBit;

		// depth-stencil-buffer bit-counts
		rankPrimary -= Abs<int>(fbc.bitsDepth - fbcRef.bitsDepth) * cPoints_DepthBufferBit;
		rankPrimary -= Abs<int>(fbc.bitsStencil - fbcRef.bitsStencil) * cPoints_StencilBufferBit;

		// accum-buffer bit-counts
		rankSecondary -= Abs<int>(fbc.bitsAccumRed - fbcRef.bitsAccumRed) * cPoints_AccumBufferBit;
		rankSecondary -= Abs<int>(fbc.bitsAccumGreen - fbcRef.bitsAccumGreen) * cPoints_AccumBufferBit;
		rankSecondary -= Abs<int>(fbc.bitsAccumBlue - fbcRef.bitsAccumBlue) * cPoints_AccumBufferBit;
		rankSecondary -= Abs<int>(fbc.bitsAccumAlpha - fbcRef.bitsAccumAlpha) * cPoints_AccumBufferBit;

		// back-buffer-count
		rankSecondary -= Abs<int>(fbc.backBufferCount - fbcRef.backBufferCount) * cPoints_BackBufferCountDiff;

		// store valid target
		FBCSelectionTarget target(&fbc);
		target.rankPrimary = rankPrimary;
		target.rankSecondary = rankSecondary;

		selectionTargets.push_back(target);
	}}

	// find highest rank target
	std::vector<FBCSelectionTarget>::iterator it = std::max_element(
							selectionTargets.begin(), selectionTargets.end());

	if(it != selectionTargets.end()) return it->fbc;
	return 0;
}


//====================================================================
bool  VideoProfile::ConstructClosestValidVideoConfig(const VideoConfig& vcRef, VideoConfig& vcOut) const
{
	vcOut = vcRef;

	// closest video-config
	const DisplayModeProperties* dmpClosest = SelectClosestLowerDisplayModeProperties(vcRef.displayModeConfig);
	if(! dmpClosest) return false;
	vcOut.displayModeConfig.width = dmpClosest->width;
	vcOut.displayModeConfig.height = dmpClosest->height;
	vcOut.displayModeConfig.bitsPerPixel = dmpClosest->bitsPerPixel;
	vcOut.displayModeConfig.refreshFrequency = dmpClosest->GetClosestLowerRefreshFrequency(vcRef.displayModeConfig.refreshFrequency);

	// closest frame buffer
	const FrameBufferConfig* fbcClosest = SelectClosestFrameBufferConfig(vcRef.frameBufferConfig);
	if(! fbcClosest) return false;
	vcOut.frameBufferConfig = *fbcClosest;

	return true;
}


//====================================================================
void  VideoProfile::Filter(VideoProfile& outProfile,
							DisplayModeConfigFilter& dmcFilter,
							FrameBufferConfigFilter& fbcFilter) const
{
	outProfile.Clear();

	// copy general-info
	outProfile.driverDescription = driverDescription;
	outProfile.driverVersion = driverVersion;
	outProfile.driverGuid = driverGuid;

	// filter display-modes
	for(uint iDMP = 0; iDMP < displayModeProperties.size(); iDMP++) {
		const DisplayModeProperties& dmp = displayModeProperties[iDMP];

		for(uint iRF = 0; iRF < dmp.refreshFrequencies.size(); iRF++) {
			DisplayModeConfig dmc(dmp.width, dmp.height, dmp.bitsPerPixel,
									dmp.refreshFrequencies[iRF]);
			if(dmcFilter.Filter(dmc)) {
				outProfile.Add(dmc);
			}
		}
	}
	// filter frame-buffer-configurations
	for(uint iFBC = 0; iFBC < frameBufferConfigs.size(); iFBC++) {
		const FrameBufferConfig& fbc = frameBufferConfigs[iFBC];

		if(fbcFilter.Filter(fbc)) {
			outProfile.frameBufferConfigs.push_back(fbc);
		}
	}
}


//====================================================================
void  VideoProfile::Clear()
{
	driverDescription.erase();
	driverVersion.erase();
	driverGuid = win_utils::Guid();
	displayModeProperties.clear();
	frameBufferConfigs.clear();
}


//====================================================================
void  VideoProfile::SetDriverInfo(const std::string& _driverDescription,
				const std::string& _driverVersion, const win_utils::Guid& _driverGuid)
{
	driverDescription = _driverDescription;
	driverVersion = _driverVersion;
	driverGuid = _driverGuid;
}


//====================================================================
void  VideoProfile::Add(const DisplayModeConfig& dmc)
{
	// discard duplicates
	DisplayModeProperties* pdmp = Find(dmc);
	if(pdmp) {
		// found mode, just add another available frequency
		if(std::find(pdmp->refreshFrequencies.begin(), pdmp->refreshFrequencies.end(),
								dmc.refreshFrequency) == pdmp->refreshFrequencies.end())
		{ pdmp->refreshFrequencies.push_back(dmc.refreshFrequency); }
		return;
	}
	// add new mode
	DisplayModeProperties dmp;
	dmp.width = dmc.width;
	dmp.height = dmc.height;
	dmp.bitsPerPixel = dmc.bitsPerPixel;
	dmp.refreshFrequencies.push_back(dmc.refreshFrequency);
	displayModeProperties.push_back(dmp);
}


//====================================================================
DisplayModeProperties*  VideoProfile::Find(const DisplayModeConfig& dmc)
{
	for(uint i = 0; i < displayModeProperties.size(); i++) {
		DisplayModeProperties& dmp = displayModeProperties[i];

		if(dmp.width == dmc.width  &&  dmp.height == dmc.height  &&
			dmp.bitsPerPixel == dmc.bitsPerPixel) return &dmp;
	}
	return 0;
}


//====================================================================
void  VideoProfile::Add(const FrameBufferConfig& fbc)
{
	// discard duplicates
	if(std::find(frameBufferConfigs.begin(), frameBufferConfigs.end(), fbc)
												== frameBufferConfigs.end())
	{
		frameBufferConfigs.push_back(fbc);
	}
}


//====================================================================
void  VideoProfile::BuildPostProcess()
{
	// process refresh frequencies
	for(uint i = 0; i < displayModeProperties.size(); i++) {
		displayModeProperties[i].SortRefreshFrequencies();
	}
}


//====================================================================
bool  VideoProfile::IsEmpty()
{
	if(! GetNumDisplayModeProperties()) return true;
	if(! GetNumFrameBufferConfigs()) return true;
	return false;
}


//====================================================================
void  VideoProfile::DumpGeneralInfo(std::ostream& os, uint tabLevel) const
{
	TabEffector te(tabLevel);

	// print general-info
	os << te << "DriverDescription: " << driverDescription << nl;
	os << te << "DriverVersion....: " << driverVersion << nl;
	os << te << "DriverGUID.......: " << driverGuid << nl;
}


//====================================================================
void  VideoProfile::DumpDisplayModeProperties(std::ostream& os, uint tabLevel) const
{
	TabEffector te(tabLevel);

	// print display-modes
	for(uint iDMP = 0; iDMP < displayModeProperties.size(); iDMP++) {
		const DisplayModeProperties& dmp = displayModeProperties[iDMP];

		os << te;
		os << std::setw(4) << dmp.width << " x ";
		os << std::setw(4) << dmp.height << " x ";
		os << std::setw(2) << dmp.bitsPerPixel << "  @ ";

		for(uint iRF = 0; iRF < dmp.refreshFrequencies.size(); iRF++) {
			os << " " << dmp.refreshFrequencies[iRF];
		}
		os << nl;
	}
	os << std::flush;
}


//====================================================================
void  VideoProfile::DumpFrameBufferConfigs(std::ostream& os,
									bool bDumpLegend, uint tabLevel) const
{
	TabEffector te(tabLevel);

	// print frame-buffer-configurations
	const uint  refSep = 4;
	const char* yes = "x";
	const char* no  = " ";
	const char* dot = ".";

	os << te;
	os << " FS" << fill(refSep) << "HWR" << fill(refSep);
	os << "BBC" << fill(refSep * 2) << "BBL" << fill(refSep * 2);
	os << "DSBL" << fill(refSep * 2) << "ACCBL" << nl;

	for(uint iFBC = 0; iFBC < frameBufferConfigs.size(); iFBC++) {
		const FrameBufferConfig& fbc = frameBufferConfigs[iFBC];

		os << te;
		os << spc;
		os << (fbc.fullscreen ? yes : no) << fill(refSep + 2);
		os << (fbc.hardwareRasterizer ? yes : no) << fill(refSep + 2);
		os << fbc.backBufferCount << fill(refSep + 3);
		os << fbc.bitsRed << dot << fbc.bitsGreen << dot;
		os << fbc.bitsBlue << dot << fbc.bitsAlpha << fill(refSep + 2);
		os << fbc.bitsDepth << dot << fbc.bitsStencil << fill(refSep + 2);
		os << fbc.bitsAccumRed << dot << fbc.bitsAccumGreen << dot;
		os << fbc.bitsAccumBlue << dot << fbc.bitsAccumAlpha;
		os << nl;
	}
	os << std::flush;

	if(bDumpLegend) {
		os << nl;
		os << te << "Legend:" << nl;
		os << te << "FS    = fullscreen" << nl;
		os << te << "HWR   = hardware-rasterizer" << nl;
		os << te << "BBC   = back-buffer-count" << nl;
		os << te << "BBL   = back-buffer-layout (r.g.b.a)" << nl;
		os << te << "DSBL  = depth-stencil-buffer-layout (d.s)" << nl;
		os << te << "ACCBL = accumulation-buffer-layout (r.g.b.a)" << nl;
		os << std::flush;
	}
}


//====================================================================
void  VideoProfile::Dump(std::ostream& os) const
{
	std::ostringstream oss;
	TitleEffector te("", 70, '=', true);

	os << nl << te("General information");
	DumpGeneralInfo(os);

	oss << "Display modes: " << displayModeProperties.size() << " found.";
	os << nl << te(oss.str());
	DumpDisplayModeProperties(os);

	oss.str("");
	oss << "Frame-buffer-configurations: " << frameBufferConfigs.size() << " found.";
	os << nl << te(oss.str());
	DumpFrameBufferConfigs(os, true);
	os << std::endl;
}


//********************************************************************
}//namespace vertexware


