

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\WinUtils.h"
#include "private\vw_common_exports.h"
#include <vector>
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class VW_COMMON_API VideoHints
{
public:
	enum eVertexProcessingMode
	{
		VertexProcessingMode_Default = 0,
		VertexProcessingMode_Software,
		VertexProcessingMode_Mixed,
		VertexProcessingMode_Hardware,

		VertexProcessingMode_Count,
		VertexProcessingMode_Force32Bits = 0x7FFFFFFF
	};

public:
	VideoHints(bool _vSync = true,
				bool _fullscreenDestinationAlpha = false,
				bool _pureDeviceMode = true,
				eVertexProcessingMode _vertexProcessingMode = VertexProcessingMode_Default);

	static std::string __fastcall FormatVertexProcessingMode(eVertexProcessingMode vpm);

	bool  operator==(const VideoHints& rhs) const;
	bool  operator!=(const VideoHints& rhs) const { return !(*this == rhs); }

	void  Dump(std::ostream& os, uint tabLevel = 0) const;

public:
	bool  vSync;
	bool  fullscreenDestinationAlpha;
	bool  pureDeviceMode;

	eVertexProcessingMode  vertexProcessingMode;
};


//====================================================================
class VW_COMMON_API DisplayModeConfig
{
public:
	DisplayModeConfig(bool bFillWithCurrent = false);
	DisplayModeConfig(uint width, uint height, uint bitsPerPixel, uint refreshFrequency);

	bool  operator==(const DisplayModeConfig& rhs) const;
	bool  operator!=(const DisplayModeConfig& rhs) const { return !(*this == rhs); }

	std::string  MakeDesc(const std::string& sSep = " x ",
						const std::string& sFreqSep = " @ ") const;


public:
	uint  width;
	uint  height;
	uint  bitsPerPixel;
	uint  refreshFrequency;
};


//====================================================================
class VW_COMMON_API DisplayModeConfigFilter
{
public:
	//----------------
	// Filter helpers.
	//----------------
	static bool __fastcall  IsSafeDisplayModeConfig(const DisplayModeConfig& dmc);
	static bool __fastcall  IsStandardResolution(uint width, uint height);
	static bool __fastcall  IsSubDesktopResolution(uint width, uint height);

public:
	virtual ~DisplayModeConfigFilter() {}

	//------------------------------------------------
	// Override; return false to reject configuration.
	//------------------------------------------------
	virtual bool Filter(const DisplayModeConfig& dmc)
	{
		return
		IsStandardResolution(dmc.width, dmc.height) &&
		IsSubDesktopResolution(dmc.width, dmc.height);
	}
};


//====================================================================
class VW_COMMON_API DisplayModeProperties
{
public:
	uint  GetNumRefreshFrequencies() const;
	uint  GetRefreshFrequency(uint i) const;

	uint  GetLowestRefreshFrequency() const;
	uint  GetHighestRefreshFrequency() const;

	bool  IsValidRefreshFrequency(uint freq) const;
	bool  IsValid(const DisplayModeConfig& dmc) const;

	std::string  MakeDesc(const std::string& sSep = " x ") const;

	void  SortRefreshFrequencies();
	uint  GetClosestLowerRefreshFrequency(uint refFreq) const; //returns 0 if not found

public:
	uint  width;
	uint  height;
	uint  bitsPerPixel;

	std::vector<uint>  refreshFrequencies;
};


//====================================================================
class VW_COMMON_API FrameBufferConfig
{
public:
	FrameBufferConfig(bool fullscreen = false,
						bool hardwareRasterizer = true,
						uint backBufferCount = 1,
						uint bitsDepth = 24,
						uint bitsStencil = 8,
						uint bitsAccum = 0);

	bool  operator==(const FrameBufferConfig& rhs) const;
	bool  operator!=(const FrameBufferConfig& rhs) const { return !(*this == rhs); }

	std::string  MakeLayoutDesc(const std::string& sBitSep = ".",
								const std::string& sBufSep = " / ") const;

	void  Dump(std::ostream& os, uint tabLevel = 0) const;

public:
	bool  fullscreen;
	bool  hardwareRasterizer;
	uint  backBufferCount;

	// back-buffer
	uint  bitsRed;
	uint  bitsGreen;
	uint  bitsBlue;
	uint  bitsAlpha;

	// depth-stencil buffer
	uint  bitsDepth;
	uint  bitsStencil;

	// accumulation buffer
	uint  bitsAccumRed;
	uint  bitsAccumGreen;
	uint  bitsAccumBlue;
	uint  bitsAccumAlpha;
};


//====================================================================
class VW_COMMON_API FrameBufferConfigFilter
{
public:
	virtual ~FrameBufferConfigFilter() {}

	//------------------------------------------------
	// Override; return false to reject configuration.
	//------------------------------------------------
	virtual bool Filter(const FrameBufferConfig& fbc)
	{
		return true;
	}
};


//====================================================================
// If <frameBufferConfig.fullscreen> is true
//    <resizable> is ignored
//    If <displayModeConfig.refreshFrequency> is 0
//        default frequency is used
// Else
//    <displayModeConfig.bitsPerPixel> is ignored
//    <displayModeConfig.refreshFrequency> is ignored
//====================================================================
class VW_COMMON_API VideoConfig
{
public:
	VideoConfig(bool bResizable = false);

	bool  operator==(const VideoConfig& rhs) const;
	bool  operator!=(const VideoConfig& rhs) const { return !(*this == rhs); }

	bool  IsEqualIgnoreVSync(const VideoConfig& rhs) const;

	void  Dump(std::ostream& os, uint tabLevel = 0) const;

public:
	FrameBufferConfig  frameBufferConfig;
	DisplayModeConfig  displayModeConfig;
	bool               resizable;
	VideoHints         videoHints;
};


//====================================================================
class VW_COMMON_API VideoProfile
{
public:
	//-----------------------
	// Information interface.
	//-----------------------
	const std::string&      GetDriverDescription() const;
	const std::string&      GetDriverVersion() const;
	const win_utils::Guid&  GetDriverGuid() const;

	uint                          GetNumDisplayModeProperties() const;
	const DisplayModeProperties&  GetDisplayModeProperties(uint i) const;

	uint                      GetNumFrameBufferConfigs() const;
	const FrameBufferConfig&  GetFrameBufferConfig(uint i) const;

	bool  IsValid(const DisplayModeConfig& dmc) const;
	bool  IsValid(const FrameBufferConfig& fbc) const;
	bool  IsValid(const VideoConfig& vc) const;

	// return 0 if failed
	const DisplayModeProperties*  SelectClosestLowerDisplayModeProperties(const DisplayModeConfig& dmcRef) const;
	const FrameBufferConfig*      SelectClosestFrameBufferConfig(const FrameBufferConfig& fbcRef) const;

	// returns false if failed
	bool  ConstructClosestValidVideoConfig(const VideoConfig& vcRef, VideoConfig& vcOut) const;

	//---------
	// Filters.
	//---------
	void  Filter(VideoProfile& outProfile,
				DisplayModeConfigFilter& dmcFilter = DisplayModeConfigFilter(),
				FrameBufferConfigFilter& fbcFilter = FrameBufferConfigFilter()) const;

	//--------------------
	// Building interface.
	//--------------------
	void  Clear();

	void  SetDriverInfo(const std::string& driverDescription,
						const std::string& driverVersion, const win_utils::Guid& driverGuid);

	void  Add(const DisplayModeConfig& dmc);
	void  Add(const FrameBufferConfig& fbc);

	void  BuildPostProcess();

	bool  IsEmpty();

	void  DumpGeneralInfo(std::ostream& os, uint tabLevel = 0) const;
	void  DumpDisplayModeProperties(std::ostream& os, uint tabLevel = 0) const;
	void  DumpFrameBufferConfigs(std::ostream& os, bool bDumpLegend, uint tabLevel = 0) const;
	void  Dump(std::ostream& os) const;

	//----------------------
	// Convenience typedefs.
	//----------------------
	typedef  std::vector<DisplayModeProperties>  DMP_VEC;
	typedef  std::vector<FrameBufferConfig>      FBC_VEC;


private:
	DisplayModeProperties*  Find(const DisplayModeConfig& dmc);


private:
	std::string      driverDescription;
	std::string      driverVersion;
	win_utils::Guid  driverGuid;

	DMP_VEC  displayModeProperties;
	FBC_VEC  frameBufferConfigs;
};


//********************************************************************
}//namespace vertexware


