

//--------------------------------------------------------------------
#include "Framework\VideoRenderer.h"


//====================================================================
#define  ASSERT_GUI_RENDERER_ATTACHED \
	VW_ASSERT(pGUIRenderer, "GUIRenderer not attached.");


//********************************************************************
namespace vertexware {


//====================================================================
VideoRendererInfo::VideoRendererInfo(
						const PluginInfo& _pluginInfo,
						const std::string& _vrInterfaceName,
						const VideoProfile& _videoProfile)
	:
	pluginInfo(_pluginInfo),
	vrInterfaceName(_vrInterfaceName),
	videoProfile(_videoProfile)
{
}


//====================================================================
bool  VideoRenderer::s_enableDynamicVSyncUpdate = true;


//====================================================================
VideoRenderer::VideoRenderer()
{
	pGUIRenderer = 0;
	pVideoSubsystem = 0;

	windowWidth  = 1;
	windowHeight = 1;
}


//====================================================================
VideoRenderer::~VideoRenderer()
{
	SafeDelete(pGUIRenderer);
}


//====================================================================
gui::Renderer&  VideoRenderer::GetGUIRenderer()
{
	ASSERT_GUI_RENDERER_ATTACHED;
	return *pGUIRenderer;
}


//====================================================================
const VideoConfig&  VideoRenderer::GetVideoConfig() const
{
	VW_ASSERT(pVideoSubsystem, "Video-subsystem not connected.");
	return pVideoSubsystem->GetVideoConfig();
}


//====================================================================
void  VideoRenderer::PresentBackBuffer()
{
	if(pVideoSubsystem) {
		pVideoSubsystem->PresentBackBuffer();
	}
}


//====================================================================
void  VideoRenderer::UpdateVSync(bool bEnable)
{
	if(pVideoSubsystem) {
		if(pVideoSubsystem->GetVideoConfig().videoHints.vSync == bEnable) return;

		// forward dynamic update to video-subsystem
		if(s_enableDynamicVSyncUpdate) {
			pVideoSubsystem->UpdateVSync(bEnable);
			Kernel::Instance().NotifyVSyncChanged();
		}
		// trigger video-reconfig with updated video-config
		else {
			VideoConfig vc = pVideoSubsystem->GetVideoConfig();
			vc.videoHints.vSync = bEnable;
			Kernel::Instance().RequestVideoReconfig(GetPluginInfo().GetPluginName(), vc);
		}
	}
}


//====================================================================
void  VideoRenderer::EnableDynamicVSyncUpdate(bool bEnable)
{
	s_enableDynamicVSyncUpdate = bEnable;
	Kernel::Instance().NotifyVSyncChanged();
}


//====================================================================
bool  VideoRenderer::CaptureFrameBuffer(AutoPtr<ImageBase>& out,
										uint frameBufferFlags,
										bool bForceColorSampleFormat)
{
	if(pVideoSubsystem) {
		return pVideoSubsystem->CaptureFrameBuffer(out,
							frameBufferFlags, bForceColorSampleFormat);
	}
	return false;
}


//====================================================================
void  VideoRenderer::AttachGUIRenderer(gui::Renderer& rend)
{
	VW_ASSERT(! pVideoSubsystem,
			"Cannot attach GUIRenderer while VideoSubsystem is connected.");

	pGUIRenderer = &rend;
}


//====================================================================
void  VideoRenderer::DetachGUIRenderer()
{
	VW_ASSERT(! pVideoSubsystem,
			"Cannot detach GUIRenderer while VideoSubsystem is connected.");

	pGUIRenderer = 0;
}


//====================================================================
void  VideoRenderer::ConnectVideoSubsystem(VideoSubsystem& vs)
{
	ASSERT_GUI_RENDERER_ATTACHED;

	pVideoSubsystem = &vs;
	pGUIRenderer->ConnectVideoSubsystem(vs);
}


//====================================================================
void  VideoRenderer::DisconnectVideoSubsystem()
{
	ASSERT_GUI_RENDERER_ATTACHED;

	pGUIRenderer->DisconnectVideoSubsystem();
	pVideoSubsystem = 0;
}


//====================================================================
void  VideoRenderer::CreateVideoResources(bool bPostReset)
{
	ASSERT_GUI_RENDERER_ATTACHED;

	pGUIRenderer->CreateVideoResources(bPostReset);
}


//====================================================================
void  VideoRenderer::ReleaseVideoResources(bool bPreReset)
{
	ASSERT_GUI_RENDERER_ATTACHED;

	pGUIRenderer->ReleaseVideoResources(bPreReset);
}


//====================================================================
void  VideoRenderer::OnWindowResize(uint width, uint height)
{
	ASSERT_GUI_RENDERER_ATTACHED;

	windowWidth  = ClampedLow<uint>(width, 1);
	windowHeight = ClampedLow<uint>(height, 1);

	pGUIRenderer->OnWindowResize(windowWidth, windowHeight);
}


//====================================================================
VideoRendererPluginDescriptor::VideoRendererPluginDescriptor(const PluginInfo& pi)
	:
	PluginDescriptor(PluginInfo(pi, PluginType_VideoRenderer))
{
}


//********************************************************************
}//namespace vertexware


