

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Exception.h"
#include "MathTools\Vector2.h"
#include "GUI\GUI_Renderer.h"
#include "Framework\Plugin.h"
#include "Framework\VideoSubsystem.h"
#include "private\vw_common_exports.h"


//********************************************************************
namespace vertexware {


//====================================================================
class Kernel;


//====================================================================
// Video-renderer info.
//====================================================================
class VW_COMMON_API VideoRendererInfo
{
public:
	VideoRendererInfo()
	{ }

	VideoRendererInfo(
				const PluginInfo& pluginInfo,
				const std::string& vrInterfaceName,
				const VideoProfile& videoProfile);

	const PluginInfo&    GetPluginInfo() const { return pluginInfo; }
	const std::string&   GetVideoRendererInterfaceName() const { return vrInterfaceName; }
	const VideoProfile&  GetVideoProfile() const { return videoProfile; }


private:
	PluginInfo    pluginInfo;
	std::string   vrInterfaceName;
	VideoProfile  videoProfile;
};


//====================================================================
// Video-renderer base-class.
//====================================================================
class VW_COMMON_API VideoRenderer : public Plugin
{
friend class Kernel;

public:
	virtual ~VideoRenderer();

	gui::Renderer&      GetGUIRenderer();
	const VideoConfig&  GetVideoConfig() const;
	void                PresentBackBuffer();
	void                UpdateVSync(bool bEnable);

	//----------------------------------------------------------------
	// If disabled, UpdateVSync() calls Kernel::RequestVideoReconfig()
	// with current video-config instead of forwarding dynamic update
	// to video-subsystem.
	//----------------------------------------------------------------
	static  void  EnableDynamicVSyncUpdate(bool bEnable = true);
	static  bool  IsEnabledDynamicVSyncUpdate() { return s_enableDynamicVSyncUpdate; }

	//------------------------------------------------------------------------
	// Exports buffer contents.
	// Returns false on error or unsupported.
	// Pixel-storage is auto-allocated and assigned to <out>.
	// Row-storage order is always top-to-bottom.
	// If <bForceColorSampleFormat> is true, special sample-formats
	// i.e. (z-buffer values) are internally converted to color sample-format.
	//------------------------------------------------------------------------
	bool  CaptureFrameBuffer(AutoPtr<ImageBase>& out,
							uint frameBufferFlags = FrameBufferFlags_Color,
							bool bForceColorSampleFormat = true);

	//----------------
	// Range [1, +inf>
	//----------------
	uint  GetWindowWidth() const { return windowWidth; }
	uint  GetWindowHeight() const { return windowHeight; }

	//------------------------
	// Exceptions NOT allowed.
	//------------------------
	virtual  std::string  GetVideoRendererInterfaceName() = 0;
	virtual  std::string  GetMatchingVideoSubsystemInterfaceName() = 0;


protected:
	VideoRenderer();

	VideoSubsystem*  GetVideoSubsystem() { return pVideoSubsystem; }

	//-------------------------------------------------------------------------------
	// Derived class must construct gui-renderer object in derived-class constructor,
	// then attach it to base-class. Attached gui-renderer object is auto-deleted
	// by base-class destructor. To avoid auto-delete detach object in
	// derived-class destructor. Attach/Detach is allowed only while
	// video-subsystem is disconnected (i.e. no run-time changes are allowed).
	//-------------------------------------------------------------------------------
	void  AttachGUIRenderer(gui::Renderer& rend);
	void  DetachGUIRenderer();

	//----------------------------------------------------
	// Kernel interface.
	// Base implementations must be called from overrides.
	// Exceptions NOT allowed.
	//----------------------------------------------------
	virtual  void  ConnectVideoSubsystem(VideoSubsystem& vs);
	virtual  void  DisconnectVideoSubsystem();

	virtual  void  CreateVideoResources(bool bPostReset);
	virtual  void  ReleaseVideoResources(bool bPreReset);

	virtual  void  OnWindowResize(uint width, uint height);


private:
	VideoSubsystem*  pVideoSubsystem;
	gui::Renderer*   pGUIRenderer;

	uint  windowWidth;
	uint  windowHeight;

	static  bool  s_enableDynamicVSyncUpdate;
};


//====================================================================
class VW_COMMON_API VideoRendererPluginDescriptor : public PluginDescriptor
{
public:
	virtual  Plugin*  CreatePlugin() {
		return CreateVideoRenderer();
	}

	virtual  VideoRenderer*  CreateVideoRenderer() = 0;


protected:
	VideoRendererPluginDescriptor(const PluginInfo& pi);
};


//====================================================================
// VideoRenderer plugin access-point declaration.
//====================================================================
#define  VW_DECLARE_PLUGIN_ACCESS_POINT_VIDEO_RENDERER( \
				ClassNamePlugin, StringPluginName, \
				UIntPluginVersionMajor, UIntPluginVersionMinor, \
				UIntPluginVersionRevision, UIntPluginVersionBuildFlags, \
				StringPluginDescription, StringPluginDeveloperInfo) \
\
struct ClassNamePlugin##__PluginDescriptor : public VideoRendererPluginDescriptor \
{ \
	ClassNamePlugin##__PluginDescriptor() \
		: VideoRendererPluginDescriptor(PluginInfo( \
				vertexware::library_info::LibraryVersion(), \
				PluginType_VideoRenderer, \
				StringPluginName, \
				vertexware::PluginVersion(UIntPluginVersionMajor, UIntPluginVersionMinor, \
										UIntPluginVersionRevision, UIntPluginVersionBuildFlags), \
				VW_BUILD_TIMESTAMP_STRING, \
				StringPluginDescription, \
				StringPluginDeveloperInfo)) \
	{ } \
\
	virtual  VideoRenderer*  CreateVideoRenderer() { \
		return new ClassNamePlugin; \
	} \
}; \
VW_DECLARE_PLUGIN_ACCESS_POINT( ClassNamePlugin##__PluginDescriptor )


//********************************************************************
}//namespace vertexware


