

//--------------------------------------------------------------------
#include "Framework\VideoSubsystem.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Debug.h"


//********************************************************************
namespace vertexware {


//====================================================================
VideoSubsystemWindow::VideoSubsystemWindow(bool _bFullscreen)
	:
	TopLevelWindow(false)
{
	pObserver = 0;
	bMoveHandlerActive = false;
	bResizeHandlerActive = false;
	bFullscreen = _bFullscreen;

	EnableSystemCommands(false, false, false);
}


//====================================================================
void  VideoSubsystemWindow::EnableSystemCommands(bool bCloseThroughAltF4,
								bool bMenuThroughF10, bool bMenuThroughAltSpace)
{
	TopLevelWindow::EnableCloseThroughAltF4(bCloseThroughAltF4);
	TopLevelWindow::EnableMenuThroughF10(bMenuThroughF10);
	TopLevelWindow::EnableMenuThroughAltSpace(bMenuThroughAltSpace);
}


//====================================================================
bool  VideoSubsystemWindow::MessageHandler(LRESULT& outResult,
											HWND hwnd, UINT msg,
											WPARAM wparam, LPARAM lparam)
{
	switch(msg) {
		case WM_SYSCOMMAND:
			switch(wparam) {
				case SC_MONITORPOWER:
				case SC_SCREENSAVE:
					// block windows-built-in handlers
					return true;
			}
			break;

		case WM_ENTERMENULOOP:
			if(GetObserver()) {
				GetObserver()->OnSystemMenuLoopActive(true);
			}
			return true;

		case WM_EXITMENULOOP:
			if(GetObserver()) {
				GetObserver()->OnSystemMenuLoopActive(false);
			}
			return true;

		case WM_ENTERSIZEMOVE:
			if(GetObserver()) {
				GetObserver()->OnSystemResizeMoveLoopActive(true);
			}
			return true;

		case WM_EXITSIZEMOVE:
			if(GetObserver()) {
				GetObserver()->OnSystemResizeMoveLoopActive(false);
			}
			return true;
	}
	return TopLevelWindow::MessageHandler(outResult, hwnd, msg, wparam, lparam);
}


//====================================================================
void  VideoSubsystemWindow::OnDisplaySettingsChanged()
{
	if(GetObserver()) {
		GetObserver()->OnDisplaySettingsChanged();
	}
}


//====================================================================
void  VideoSubsystemWindow::OnActivated(bool bActive)
{
	if(GetObserver()) {
		GetObserver()->OnVideoWindowActivated(bActive);
	}
}


//====================================================================
void  VideoSubsystemWindow::OnMoved()
{
	if(IsMoveHandlerActive()  &&  GetObserver()) {
		GetObserver()->OnVideoWindowMoved();
	}
}


//====================================================================
void  VideoSubsystemWindow::OnResized(eResizeAction action)
{
	if(IsResizeHandlerActive()  &&  GetObserver()) {
		GetObserver()->OnVideoWindowResized(action);
	}
}


//====================================================================
void  VideoSubsystemWindow::OnPaint()
{
	if(GetObserver()) {
		GetObserver()->OnVideoWindowPaint();
	}
}


//====================================================================
VideoSubsystem::VideoSubsystem()
{
	operationalState = OperationalState_WaitingInit;
}


//====================================================================
VideoSubsystem::~VideoSubsystem()
{
}


//====================================================================
void  VideoSubsystem::SetOperationalState(eOperationalState state)
{
	operationalState = state;
}


//====================================================================
bool  VideoSubsystem::BuildVideoProfile(const std::string& logName,
										const std::string& fileName,
										VideoProfile& outVideoProfile)
{
	const std::string sError("Error: ");

	// open log and clear current profile
	LogFileStream log(logName, fileName);
	outVideoProfile.Clear();

	// collect data
	log << "Collecting profile data." << nl;

	bool bValid = DoBuildVideoProfile(log, outVideoProfile);

	// post-process
	outVideoProfile.BuildPostProcess();

	// dump collected info
	outVideoProfile.Dump(log);

	// validate profile
	log << nl << "Validating profile." << nl;

	// check for derived-class errors
	if(! bValid) {
		log << sError << "Failed to collect all required information." << nl;
	}

	// check number of display-modes
	if(! outVideoProfile.GetNumDisplayModeProperties()) {
		bValid = false;
		log << sError << "No available display-modes." << nl;
	}

	// check number of frame-buffer-configs
	if(! outVideoProfile.GetNumFrameBufferConfigs()) {
		bValid = false;
		log << sError << "No available frame-buffer-configurations." << nl;
	}

	// report result
	log << "Validation result: " << (bValid ? "PASSED" : "FAILED") << nl;
	return bValid;
}


//====================================================================
VideoSubsystemPluginDescriptor::VideoSubsystemPluginDescriptor(const PluginInfo& pi)
	:
	PluginDescriptor(PluginInfo(pi, PluginType_VideoSubsystem))
{
}


//********************************************************************
}//namespace vertexware


