

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "Framework\Plugin.h"
#include "Framework\VideoProperties.h"
#include "BaseTools\TopLevelWindow.h"
#include "BaseTools\Memory.h"
#include "DataObjects\ImageBase.h"
#include "private\vw_common_exports.h"
#include <vector>
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class VW_COMMON_API VideoSubsystemWindowObserver
{
public:
	virtual  void  OnDisplaySettingsChanged() = 0;
	virtual  void  OnVideoWindowActivated(bool bActive) = 0;
	virtual  void  OnVideoWindowMoved() = 0;
	virtual  void  OnVideoWindowResized(TopLevelWindow::eResizeAction action) = 0;
	virtual  void  OnVideoWindowPaint() = 0;
	virtual  void  OnSystemMenuLoopActive(bool bLoopActive) = 0;
	virtual  void  OnSystemResizeMoveLoopActive(bool bLoopActive) = 0;
};


//====================================================================
class VW_COMMON_API VideoSubsystemWindow : public TopLevelWindow
{
public:
	void  SetObserver(VideoSubsystemWindowObserver* p = 0) { pObserver = p; }
	void  ActivateMoveHandler() { bMoveHandlerActive = true; }
	void  ActivateResizeHandler() { bResizeHandlerActive = true; }
	bool  IsFullscreen() { return bFullscreen; }

	void  EnableSystemCommands(bool bCloseThroughAltF4,
				bool bMenuThroughF10, bool bMenuThroughAltSpace);


protected:
	//-------------------------
	// Derived class interface.
	//-------------------------
	VideoSubsystemWindow(bool bFullscreen);

	virtual ~VideoSubsystemWindow()
	{ }

	VideoSubsystemWindowObserver*  GetObserver() { return pObserver; }
	bool  IsMoveHandlerActive() { return bMoveHandlerActive; }
	bool  IsResizeHandlerActive() { return bResizeHandlerActive; }


protected:
	//--------------------------
	// TopLevelWindow overrides.
	//--------------------------
	virtual  bool  MessageHandler(LRESULT& outResult, HWND hwnd, UINT msg,
									WPARAM wparam, LPARAM lparam);

	virtual  void  OnDisplaySettingsChanged();
	virtual  void  OnActivated(bool bActive);
	virtual  void  OnMoved();
	virtual  void  OnResized(eResizeAction action);
	virtual  void  OnPaint();


private:
	VideoSubsystemWindowObserver*  pObserver;
	bool  bMoveHandlerActive;
	bool  bResizeHandlerActive;
	bool  bFullscreen;
};


//====================================================================
// Frame-buffer flags.
//====================================================================
enum eFrameBufferFlags
{
	// components
	FrameBufferFlags_ComponentMask = 0x000000FF,
	FrameBufferFlags_Color         = 0x00000001,
	FrameBufferFlags_Depth         = 0x00000002,
	FrameBufferFlags_Stencil       = 0x00000004,

	// swap-chain
	FrameBufferFlags_SwapChainMask = 0x0000FF00,
	FrameBufferFlags_Back          = 0x00000000,
	FrameBufferFlags_Front         = 0x00000100,

	// mono/stereo
	FrameBufferFlags_StereoMask = 0x00FF0000,
	FrameBufferFlags_Mono       = 0x00000000,
	FrameBufferFlags_Left       = 0x00010000,
	FrameBufferFlags_Right      = 0x00020000,

	// not used
	FrameBufferFlags_Force32Bits = 0x7FFFFFFF
};


//====================================================================
// Video-subsystem handles low-level tasks such as rendering-window
// management, vertical-sync management, back-buffer-presentation etc.
//====================================================================
class VW_COMMON_API VideoSubsystem : public Plugin
{
public:
	//-----------------
	// Subsystem-state.
	//-----------------
	enum eOperationalState
	{
		OperationalState_WaitingInit = 0,
		OperationalState_Ready,
		OperationalState_WaitingReset,

		OperationalState_Count,
		OperationalState_Force32Bits = 0x7FFFFFFF
	};


public:
	virtual ~VideoSubsystem();

	bool  BuildVideoProfile(const std::string& logName,
							const std::string& fileName,
							VideoProfile& outVideoProfile);

	// current config & state queries
	eOperationalState   GetOperationalState() const { return operationalState; }
	const VideoConfig&  GetVideoConfig() const { return videoConfig; }

	//-------------------------------------
	// Exceptions allowed only where noted.
	//-------------------------------------
	virtual  std::string  GetVideoSubsystemInterfaceName() = 0;

	// return false on error
	virtual  bool  DoBuildVideoProfile(std::ostream& log,
										VideoProfile& outVideoProfile) = 0;

	// On error throw VideoSubsystemException if recovery is impossible,
	// otherwise throw VideoSubsystemRecoverableException.
	virtual  VideoSubsystemWindow*  InitializeWindow(const VideoConfig& vc,
													bool bAutoShowWindow) = 0;

	virtual  VideoSubsystemWindow*  GetWindow() = 0;

	virtual  void  CleanupWindow() = 0;

	virtual  void  PresentBackBuffer() = 0;

	// do explicit error-check and setup operational-state accordingly
	virtual  void  ErrorCheck() = 0;

	// throw exception if recovery is impossible
	virtual  void  Reset() = 0;

	virtual  void  UpdateVSync(bool bEnable) = 0;

	// see notes for delegation-method in video-renderer
	virtual  bool  CaptureFrameBuffer(AutoPtr<ImageBase>& out,
										uint frameBufferFlags,
										bool bForceColorSampleFormat) = 0;


private:
	eOperationalState  operationalState;


protected:
	//------------------------
	// Derived-class interface
	//------------------------
	VideoSubsystem();

	void  SetOperationalState(eOperationalState state);


protected:
	//--------------------------------------
	// Derived-class accessible data members
	//--------------------------------------
	VideoConfig  videoConfig; //update on any change
};


//====================================================================
class VW_COMMON_API VideoSubsystemPluginDescriptor : public PluginDescriptor
{
public:
	virtual  Plugin*  CreatePlugin() {
		return CreateVideoSubsystem();
	}

	virtual  VideoSubsystem*  CreateVideoSubsystem() = 0;


protected:
	VideoSubsystemPluginDescriptor(const PluginInfo& pi);
};


//====================================================================
// VideoSubsystem plugin access-point declaration.
//====================================================================
#define  VW_DECLARE_PLUGIN_ACCESS_POINT_VIDEO_SUBSYSTEM( \
				ClassNamePlugin, StringPluginName, \
				UIntPluginVersionMajor, UIntPluginVersionMinor, \
				UIntPluginVersionRevision, UIntPluginVersionBuildFlags, \
				StringPluginDescription, StringPluginDeveloperInfo) \
\
struct ClassNamePlugin##__PluginDescriptor : public VideoSubsystemPluginDescriptor \
{ \
	ClassNamePlugin##__PluginDescriptor() \
		: VideoSubsystemPluginDescriptor(PluginInfo( \
				vertexware::library_info::LibraryVersion(), \
				PluginType_VideoSubsystem, \
				StringPluginName, \
				vertexware::PluginVersion(UIntPluginVersionMajor, UIntPluginVersionMinor, \
										UIntPluginVersionRevision, UIntPluginVersionBuildFlags), \
				VW_BUILD_TIMESTAMP_STRING, \
				StringPluginDescription, \
				StringPluginDeveloperInfo)) \
	{ } \
\
	virtual  VideoSubsystem*  CreateVideoSubsystem() { \
		return new ClassNamePlugin; \
	} \
}; \
VW_DECLARE_PLUGIN_ACCESS_POINT( ClassNamePlugin##__PluginDescriptor )


//********************************************************************
}//namespace vertexware


