

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\Exception.h"
#include "MathTools\Vector2.h"
#include "Framework\FrameworkExceptions.h"
#include "HIDControllers\Keyboard.h"
#include "HIDControllers\Mouse.h"
#include "GUI\GUI_Renderer.h"
#include "GUI\GUI_CommonID.h"
#include <vector>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//====================================================================
struct Message;
class  MessageHandler;
class  VideoResourceUser;
class  Component;
class  InterleavedRenderArea;
class  Cursor;
class  BaseShader;
class  Manager;
class  ParamRegistry;


//====================================================================
typedef  std::vector<Component*>               COMPONENT_PTR_VECTOR;
typedef  COMPONENT_PTR_VECTOR::iterator        COMPONENT_PTR_ITER;
typedef  COMPONENT_PTR_VECTOR::const_iterator  COMPONENT_PTR_CITER;


//====================================================================
enum eCommandArea
{
	CommandArea_Void = 0,
	CommandArea_Client,
	CommandArea_Move,
	CommandArea_ResizeN,
	CommandArea_ResizeS,
	CommandArea_ResizeW,
	CommandArea_ResizeE,
	CommandArea_ResizeNW,
	CommandArea_ResizeNE,
	CommandArea_ResizeSW,
	CommandArea_ResizeSE,

	CommandArea_Count,
	CommandArea_Force32Bits = 0x7FFFFFFF
};


//====================================================================
enum eGeometryUpdateEvent
{
	GeometryUpdateEvent_Null = 0,

	GeometryUpdateEvent_QueryBegin,

	GeometryUpdateEvent_MoveBegin,
	GeometryUpdateEvent_Moving,
	GeometryUpdateEvent_MoveEnd,

	GeometryUpdateEvent_ResizeBegin,
	GeometryUpdateEvent_Resizing,
	GeometryUpdateEvent_ResizeEnd,

	GeometryUpdateEvent_Count,
	GeometryUpdateEvent_Force32Bits = 0x7FFFFFFF
};


//====================================================================
enum eScrollDistance
{
	ScrollDistance_Line = 0,
	ScrollDistance_Page,
	ScrollDistance_ToEnd,
	ScrollDistance_Offset,

	ScrollDistance_Count,
	ScrollDistance_Force32Bits = 0x7FFFFFFF
};


//====================================================================
class VW_COMMON_API VideoResourceUser
{
friend class Manager;

public:
	virtual ~VideoResourceUser() {}


protected:
	VideoResourceUser() {}

	// <vr> is 0 when video-renderer is disconnected,
	// use this to check & cache video-renderer type if necessary
	virtual  void  OnVideoRendererChanged(VideoRenderer* vr) {}

	virtual  void  OnTileTextureListChanged(gui::Manager& gm) {}
	virtual  void  OnFontListChanged(gui::Manager& gm) {}

	// return false if required tile-texture is missing
	virtual  bool  CacheTileTextureHandles(gui::Manager& gm) { return true; }
	virtual  void  PostCacheTileTextureHandles() {}

	void  RequestTileTextureRecache();

	// return false if required font is missing
	virtual  bool  CacheFontHandles(gui::Manager& gm) { return true; }
	virtual  void  PostCacheFontHandles() {}

	void  RequestFontRecache();
};


//====================================================================
// Message.
// Do not add data members in derived classes.
//====================================================================
struct VW_COMMON_API Message
{
	uint             messageID;
	MessageHandler*  target;

	union {
		ubyte  data[64];

		struct {
			void*   vData[4];
			int     iData[4];
			float   fData[4];
			double  dData[2];
		};
	};

	explicit Message(uint _messageID = MID_Null)
	{
		messageID = _messageID;
		target = 0;
	}

	~Message()
	{ }
};


//====================================================================
template<typename TMessage>
struct MessageFieldAccessor
{
	__forceinline TMessage&        MFACast() { return static_cast<TMessage&>(*this); }
	__forceinline const TMessage&  MFACast() const { return static_cast<const TMessage&>(*this); }
};


//====================================================================
// Message-casts.
//
// Example usage:
//     DerivedMessage mOriginal;
//     Message& m = mOriginal;
//     MsgPtr<DerivedMessage> dm(m);
//
//     const DerivedMessage mOriginal;
//     const Message& cm = mOriginal;
//     ConstMsgPtr<DerivedMessage> cdm(cm);
//====================================================================
template<typename TTargetMessage>
struct MsgPtr
{
	TTargetMessage* ptr;

	__forceinline operator TTargetMessage*() { return ptr; }
	__forceinline TTargetMessage* operator->() { return ptr; }
	__forceinline TTargetMessage& operator*() { return *ptr; }

	__forceinline  MsgPtr(Message& message)
	{
		ptr = static_cast<TTargetMessage*>(&message);
	}
	__forceinline  MsgPtr(Message* message = 0)
	{
		ptr = static_cast<TTargetMessage*>(message);
	}
};


//====================================================================
template<typename TTargetMessage>
struct ConstMsgPtr
{
	const TTargetMessage* ptr;

	__forceinline operator const TTargetMessage*() { return ptr; }
	__forceinline const TTargetMessage* operator->() { return ptr; }
	__forceinline const TTargetMessage& operator*() { return *ptr; }

	__forceinline  ConstMsgPtr(const Message& message)
	{
		ptr = static_cast<const TTargetMessage*>(&message);
	}
	__forceinline  ConstMsgPtr(const Message* message = 0)
	{
		ptr = static_cast<const TTargetMessage*>(message);
	}
};


//====================================================================
// Standard messages
//====================================================================
struct VW_COMMON_API Message_ServiceBase : public Message
{
	uint  GetServiceID() const { return iData[0]; }
	void  SetServiceID(uint serviceID) { iData[0] = serviceID; }


protected:
	explicit Message_ServiceBase(uint messageID, uint serviceID = SID_Null)
		:
		Message(messageID)
	{
		SetServiceID(serviceID);
	}
};


//====================================================================
struct VW_COMMON_API Message_ServiceProviderChanged : public Message_ServiceBase
{
	explicit Message_ServiceProviderChanged(uint serviceID = SID_Null)
		:
		Message_ServiceBase(MID_ServiceProviderChanged, serviceID)
	{ }
};


//====================================================================
struct VW_COMMON_API Message_ServiceCommand : public Message_ServiceBase
{
	uint  GetServiceCommandID() const { return iData[1]; }
	void  SetServiceCommandID(uint serviceCommandID) { iData[1] = serviceCommandID; }

	explicit Message_ServiceCommand(uint serviceID = SID_Null,
									uint serviceCommandID = SCID_Null)
		:
		Message_ServiceBase(MID_ServiceCommand, serviceID)
	{
		SetServiceCommandID(serviceCommandID);
	}
};


//====================================================================
struct VW_COMMON_API Message_Command : public Message
{
	uint  GetCommandID() const { return iData[0]; }
	void  SetCommandID(uint commandID) { iData[0] = commandID; }

	explicit Message_Command(uint commandID)
		:
		Message(MID_Command)
	{
		SetCommandID(commandID);
	}
};


//====================================================================
struct VW_COMMON_API Message_GeometryUpdateByManager : public Message
{
	eGeometryUpdateEvent  GetUpdateEvent() const { return eGeometryUpdateEvent(iData[0]); }
	void                  SetUpdateEvent(eGeometryUpdateEvent ev) { iData[0] = ev; }

	eCommandArea  GetCommandArea() const { return eCommandArea(iData[1]); }
	void          SetCommandArea(eCommandArea ca) { iData[1] = ca; }

	// applies to events { QueryBegin, MoveBegin, ResizeBegin }
	eMouseButton  GetButtonCode() const { return eMouseButton(iData[2]); }
	void              SetButtonCode(eMouseButton bc) { iData[2] = bc; }

	// applies to events { Moving, Resizing }
	vec2  GetNewPosition() const {
		return vec2(fData[0], fData[1]);
	}
	void  SetNewPosition(const vec2& newPosition) {
		fData[0] = newPosition.x;
		fData[1] = newPosition.y;
	}
	vec2  GetNewSize() const {
		return vec2(fData[2], fData[3]);
	}
	void  SetNewSize(const vec2& newSize) {
		fData[2] = newSize.x;
		fData[3] = newSize.y;
	}

	//-----------------------------------------------------------------
	// For events { Moving, Resizing } confirm means that manager will
	// actually apply new-position & new-size.
	// For events { QueryBegin } confirm means that manager
	// will actually start the moving/resizing procedure.
	//-----------------------------------------------------------------
	bool  GetConfirmEvent() const { return ToBool(iData[3]); }
	void  SetConfirmEvent(bool bConfirmEvent) { iData[3] = bConfirmEvent; }

	explicit Message_GeometryUpdateByManager(
						eGeometryUpdateEvent updateEvent = GeometryUpdateEvent_QueryBegin,
						eCommandArea commandArea = CommandArea_Move,
						eMouseButton buttonCode = MouseButton_Left,
						const vec2& newPosition = vec2::ZERO,
						const vec2& newSize = vec2::ONE,
						bool bConfirmEvent = true)
		:
		Message(MID_GeometryUpdateByManager)
	{
		SetUpdateEvent(updateEvent);
		SetCommandArea(commandArea);
		SetButtonCode(buttonCode);
		SetNewPosition(newPosition);
		SetNewSize(newSize);
		SetConfirmEvent(bConfirmEvent);
	}
};


//====================================================================
template<typename TMessage>
struct MessageFieldAccessor_MousePosition : public MessageFieldAccessor<TMessage>
{
	// handler's coord-system
	vec2  GetMousePosition() const {
		return vec2(MFACast().fData[0], MFACast().fData[1]);
	}
	void  SetMousePosition(const vec2& mousePosition) {
		MFACast().fData[0] = mousePosition.x;
		MFACast().fData[1] = mousePosition.y;
	}
};


//====================================================================
struct VW_COMMON_API Message_Activate
	:
	public Message,
	public MessageFieldAccessor_MousePosition<Message_Activate>
{
	bool  GetActivateByMouse() const { return ToBool(iData[0]); }
	void  SetActivateByMouse(bool bActivateByMouse) { iData[0] = bActivateByMouse; }

	eMouseButton  GetButtonCode() const { return eMouseButton(iData[1]); }
	void              SetButtonCode(eMouseButton bc) { iData[1] = bc; }

	Component*  GetClickedDescendant() const { return reinterpret_cast<Component*>(vData[0]); }
	void        SetClickedDescendant(Component* pClickedDescendant) { vData[0] = pClickedDescendant; }

	bool  GetAutoBringToFront() const { return ToBool(iData[2]); }
	void  SetAutoBringToFront(bool bAutoBringToFront) { iData[2] = bAutoBringToFront; }

	bool  GetAutoActivate() const { return ToBool(iData[3]); }
	void  SetAutoActivate(bool bAutoActivate) { iData[3] = bAutoActivate; }

	explicit Message_Activate(bool bActivateByMouse = false,
							eMouseButton buttonCode = MouseButton_Left,
							const vec2& mousePosition = vec2::ZERO,
							Component* pClickedDescendant = 0,
							bool bAutoBringToFront = true,
							bool bAutoActivate = true)
		:
		Message(MID_Activate)
	{
		SetActivateByMouse(bActivateByMouse);
		SetButtonCode(buttonCode);
		SetMousePosition(mousePosition);
		SetClickedDescendant(pClickedDescendant);
		SetAutoBringToFront(bAutoBringToFront);
		SetAutoActivate(bAutoActivate);
	}
};


//====================================================================
struct VW_COMMON_API Message_Deactivate : public Message
{
	bool  GetDeactivateByMouse() const { return ToBool(iData[0]); }
	void  SetDeactivateByMouse(bool bDeactivateByMouse) { iData[0] = bDeactivateByMouse; }

	explicit Message_Deactivate(bool bDeactivateByMouse = false)
		:
		Message(MID_Deactivate)
	{
		SetDeactivateByMouse(bDeactivateByMouse);
	}
};


//====================================================================
struct VW_COMMON_API Message_MouseCapture : public Message
{
	// else capture-end
	bool  GetCaptureBegin() const { return ToBool(iData[0]); }
	void  SetCaptureBegin(bool bCaptureBegin) { iData[0] = bCaptureBegin; }

	explicit Message_MouseCapture(bool bCaptureBegin = true)
		:
		Message(MID_MouseCapture)
	{
		SetCaptureBegin(bCaptureBegin);
	}
};


//====================================================================
struct VW_COMMON_API Message_MouseEnterExit
	:
	public Message,
	public MessageFieldAccessor_MousePosition<Message_MouseEnterExit>
{
	// otherwise mouse-exit
	bool  GetMouseEnter() const { return ToBool(iData[0]); }
	void  SetMouseEnter(bool bMouseEnter) { iData[0] = bMouseEnter; }

	// applies to mouse-enter only
	bool  GetMouseOverDirect() const { return ToBool(iData[1]); }
	void  SetMouseOverDirect(bool bMouseOverDirect) { iData[1] = bMouseOverDirect; }

	explicit Message_MouseEnterExit(bool bMouseEnter = true,
									bool bMouseOverDirect = true,
									const vec2& mousePosition = vec2::ZERO)
		:
		Message(MID_MouseEnterExit)
	{
		SetMouseEnter(bMouseEnter);
		SetMouseOverDirect(bMouseOverDirect);
		SetMousePosition(mousePosition);
	}
};


//====================================================================
struct VW_COMMON_API Message_MouseQueryCursor
	:
	public Message,
	public MessageFieldAccessor_MousePosition<Message_MouseQueryCursor>
{
	eCommandArea  GetCommandArea() const { return eCommandArea(iData[0]); }
	void          SetCommandArea(eCommandArea commandArea) { iData[0] = commandArea; }

	const Cursor* GetCursor() const { return reinterpret_cast<Cursor*>(vData[0]); }
	void          SetCursor(const Cursor* pCursor = 0) { vData[0] = const_cast<Cursor*>(pCursor); }

	explicit Message_MouseQueryCursor(eCommandArea commandArea = CommandArea_Void,
									const vec2& mousePosition = vec2::ZERO,
									const Cursor* pCursor = 0)
		:
		Message(MID_MouseQueryCursor)
	{
		SetCommandArea(commandArea);
		SetCursor(pCursor);
		SetMousePosition(mousePosition);
	}
};


//====================================================================
struct VW_COMMON_API Message_MouseMove
	:
	public Message,
	public MessageFieldAccessor_MousePosition<Message_MouseMove>
{
	// motionless
	bool  GetHover() const { return ToBool(iData[0]); }
	void  SetHover(bool bHover) { iData[0] = bHover; }

	// applies to hover only
	double  GetHoverTimeSeconds() const { return dData[0]; }
	void    SetHoverTimeSeconds(double hoverTimeSeconds) { dData[0] = hoverTimeSeconds; }

	explicit Message_MouseMove(bool bHover = false,
								double hoverTimeSeconds = 0,
								const vec2& mousePosition = vec2::ZERO)
		:
		Message(MID_MouseMove)
	{
		SetHover(bHover);
		SetHoverTimeSeconds(hoverTimeSeconds);
		SetMousePosition(mousePosition);
	}
};


//====================================================================
struct VW_COMMON_API Message_MouseWheel
	:
	public Message,
	public MessageFieldAccessor_MousePosition<Message_MouseWheel>
{
	// roll-towards-user = negative
	int   GetWheelDelta() const { return iData[0]; }
	void  SetWheelDelta(int wheelDelta) { iData[0] = wheelDelta; }

	explicit Message_MouseWheel(int wheelDelta = 0,
								const vec2& mousePosition = vec2::ZERO)
		:
		Message(MID_MouseWheel)
	{
		SetWheelDelta(wheelDelta);
		SetMousePosition(mousePosition);
	}
};


//====================================================================
struct VW_COMMON_API Message_MouseButton
	:
	public Message,
	public MessageFieldAccessor_MousePosition<Message_MouseButton>
{
	eMouseButton  GetButtonCode() const { return eMouseButton(iData[0]); }
	void              SetButtonCode(eMouseButton bc) { iData[0] = bc; }

	// otherwise button-up
	bool  GetButtonDown() const { return ToBool(iData[1]); }
	void  SetButtonDown(bool bButtonDown) { iData[1] = bButtonDown; }

	// applies to button-down only
	bool  GetDoubleClick() const { return ToBool(iData[2]); }
	void  SetDoubleClick(bool bDoubleClick) { iData[2] = bDoubleClick; }

	// applies to button-down only
	uint  GetRepeatCount() const { return iData[3]; }
	void  SetRepeatCount(uint repeatCount) { iData[3] = repeatCount; }

	explicit Message_MouseButton(eMouseButton buttonCode = MouseButton_Left,
								bool bButtonDown = true,
								bool bDoubleClick = false,
								uint repeatCount = 0,
								const vec2& mousePosition = vec2::ZERO)
		:
		Message(MID_MouseButton)
	{
		SetButtonCode(buttonCode);
		SetButtonDown(bButtonDown);
		SetDoubleClick(bDoubleClick);
		SetRepeatCount(repeatCount);
		SetMousePosition(mousePosition);
	}
};


//====================================================================
struct VW_COMMON_API Message_KeyboardQueryFocus : public Message
{
	bool  GetFocusByMouse() const { return ToBool(iData[0]); }
	void  SetFocusByMouse(bool bFocusByMouse) { iData[0] = bFocusByMouse; }

	eMouseButton  GetButtonCode() const { return eMouseButton(iData[1]); }
	void              SetButtonCode(eMouseButton bc) { iData[1] = bc; }

	bool  GetWantFocus() const { return ToBool(iData[2]); }
	void  SetWantFocus(bool bWantFocus) { iData[2] = bWantFocus; }

	bool  GetNeedCharacterData() const { return ToBool(iData[3]); }
	void  SetNeedCharacterData(bool bNeedCharacterData) { iData[3] = bNeedCharacterData; }

	explicit Message_KeyboardQueryFocus(bool bFocusByMouse = true,
										eMouseButton buttonCode = MouseButton_Left,
										bool bWantFocus = true,
										bool bNeedCharacterData = false)
		:
		Message(MID_KeyboardQueryFocus)
	{
		SetFocusByMouse(bFocusByMouse);
		SetButtonCode(buttonCode);
		SetWantFocus(bWantFocus);
		SetNeedCharacterData(bNeedCharacterData);
	}
};


//====================================================================
struct VW_COMMON_API Message_KeyboardFocus : public Message
{
	// otherwise lost-focus
	bool  GetGotFocus() const { return ToBool(iData[0]); }
	void  SetGotFocus(bool bGotFocus) { iData[0] = bGotFocus; }

	explicit Message_KeyboardFocus(bool bGotFocus = true)
		:
		Message(MID_KeyboardFocus)
	{
		SetGotFocus(bGotFocus);
	}
};


//====================================================================
struct VW_COMMON_API Message_KeyboardKey : public Message
{
	eKey  GetKeyCode() const { return eKey(iData[0]); }
	void      SetKeyCode(eKey kc) { iData[0] = kc; }

	// otherwise key-up
	bool  GetKeyDown() const { return ToBool(iData[1]); }
	void  SetKeyDown(bool bKeyDown) { iData[1] = bKeyDown; }

	// applies only to key-down
	uint  GetRepeatCount() const { return iData[2]; }
	void  SetRepeatCount(uint repeatCount) { iData[2] = repeatCount; }

	// applies only to key-down
	char  GetCharData() const { return iData[3]; }
	void  SetCharData(char ch) { iData[3] = ch; }

	explicit Message_KeyboardKey(eKey keyCode = eKey(0),
									bool bKeyDown = true,
									uint repeatCount = 0,
									char charData = 0)
		:
		Message(MID_KeyboardKey)
	{
		SetKeyCode(keyCode);
		SetKeyDown(bKeyDown);
		SetRepeatCount(repeatCount);
		SetCharData(charData);
	}
};


//====================================================================
struct VW_COMMON_API Message_ShowBase : public Message
{
	bool  GetShow() const { return ToBool(iData[0]); }
	void  SetShow(bool bShow) { iData[0] = bShow; }

protected:
	explicit Message_ShowBase(uint messageID, bool bShow = true)
		:
		Message(messageID)
	{
		SetShow(bShow);
	}
};


//====================================================================
struct VW_COMMON_API Message_Show : public Message_ShowBase
{
	explicit Message_Show(bool bShow = true)
		:
		Message_ShowBase(MID_Show, bShow)
	{ }
};


//====================================================================
struct VW_COMMON_API Message_ShowChildren : public Message_ShowBase
{
	explicit Message_ShowChildren(bool bShow = true)
		:
		Message_ShowBase(MID_ShowChildren, bShow)
	{ }
};


//====================================================================
struct VW_COMMON_API Message_ShowDescendants : public Message_ShowBase
{
	explicit Message_ShowDescendants(bool bShow = true)
		:
		Message_ShowBase(MID_ShowDescendants, bShow)
	{ }
};


//====================================================================
struct VW_COMMON_API Message_Enable : public Message
{
	bool  GetEnable() const { return ToBool(iData[0]); }
	void  SetEnable(bool bEnable) { iData[0] = bEnable; }

	explicit Message_Enable(bool bEnable = true)
		:
		Message(MID_Enable)
	{
		SetEnable(bEnable);
	}
};


//====================================================================
struct VW_COMMON_API Message_Flags : public Message
{
	uint  GetFlags() const { return iData[0]; }
	void  SetFlags(uint flags) { iData[0] = flags; }

	explicit Message_Flags(uint flags = 0)
		:
		Message(MID_Flags)
	{
		SetFlags(flags);
	}
};


//====================================================================
struct VW_COMMON_API Message_FlagsOnOff : public Message
{
	uint  GetFlags() const { return iData[0]; }
	void  SetFlags(uint flags) { iData[0] = flags; }

	bool  GetOn() const { return ToBool(iData[1]); }
	void  SetOn(bool bOn) { iData[1] = bOn; }

	explicit Message_FlagsOnOff(uint flags = 0, bool bOn = true)
		:
		Message(MID_FlagsOnOff)
	{
		SetFlags(flags);
		SetOn(bOn);
	}
};


//====================================================================
struct VW_COMMON_API Message_OnOff : public Message
{
	bool  GetOn() const { return ToBool(iData[0]); }
	void  SetOn(bool bOn) { iData[0] = bOn; }

	explicit Message_OnOff(bool bOn = false)
		:
		Message(MID_OnOff)
	{
		SetOn(bOn);
	}
};


//====================================================================
struct VW_COMMON_API Message_Select : public Message
{
	bool  GetSelect() const { return ToBool(iData[0]); }
	void  SetSelect(bool bSelect) { iData[0] = bSelect; }

	uint  GetGroupID() const { return iData[1]; }
	void  SetGroupID(uint groupID) { iData[1] = groupID; }

	explicit Message_Select(bool bSelect = false, uint groupID = 0)
		:
		Message(MID_Select)
	{
		SetSelect(bSelect);
		SetGroupID(groupID);
	}
};


//====================================================================
struct VW_COMMON_API Message_Scroll : public Message
{
	bool  GetForward() const { return ToBool(iData[0]); }
	void  SetForward(bool bForward) { iData[0] = bForward; }

	eScrollDistance  GetDistance() const { return eScrollDistance(iData[1]); }
	void             SetDistance(eScrollDistance d) { iData[1] = d; }

	int   GetOffset() const { return iData[2]; }
	void  SetOffset(int offset) { iData[2] = offset; }

	explicit Message_Scroll(bool bForward = true,
							eScrollDistance distance = ScrollDistance_Line,
							int offset = 0)
		:
		Message(MID_Scroll)
	{
		SetForward(bForward);
		SetDistance(distance);
		SetOffset(offset);
	}
};


//====================================================================
struct VW_COMMON_API Message_PreCommandExecute : public Message
{
	uint  GetCommandID() const { return iData[0]; }
	void  SetCommandID(uint commandID) { iData[0] = commandID; }

	explicit Message_PreCommandExecute(uint commandID = 0)
		:
		Message(MID_PreCommandExecute)
	{
		SetCommandID(commandID);
	}
};


//====================================================================
struct VW_COMMON_API Message_SendCommand : public Message
{
	explicit Message_SendCommand()
		:
		Message(MID_SendCommand)
	{ }
};


//====================================================================
// ToolTip service messages.
//====================================================================
struct VW_COMMON_API ToolTipParams
{
	bool         bDisplay;
	bool         bAutoPosition;
	std::string  text;
	vec2      position;
	vec2      mouseMoveIgnoreAreaPosition;
	vec2      mouseMoveIgnoreAreaSize;
	vec2      minSize;
};


//====================================================================
struct VW_COMMON_API Message_ServiceCommand_ToolTip : public Message_ServiceCommand
{
	// only reference to params is stored
	ToolTipParams*  GetToolTipParams() const { return reinterpret_cast<ToolTipParams*>(vData[0]); }
	void            SetToolTipParams(ToolTipParams* ttp) { vData[0] = ttp; }

	explicit Message_ServiceCommand_ToolTip(uint serviceCommandID = SCID_Null,
											ToolTipParams* ttp = 0)
		:
		Message_ServiceCommand(SID_ToolTip, serviceCommandID)
	{
		SetToolTipParams(ttp);
	}
};


//====================================================================
class VW_COMMON_API MessageHandler
{
friend class Manager;

public:
	virtual ~MessageHandler() {}

	// message dispatch
	bool  DispatchMessage(Message& msg);

	//-----------------------------------------------------------------
	// Highest-priority hook has lowest index (also called first-hook).
	// Original message-target handles message only if it's not
	// handled by any of installed hooks.
	//-----------------------------------------------------------------
	uint                   GetNumMessageHooks() const { return hooks.size(); }
	const MessageHandler&  GetMessageHook(uint i) const { return *hooks[i]; }
	MessageHandler&        GetMessageHook(uint i) { return *hooks[i]; }

	//-------------------------------------------------------------------------
	// If <bInstallAsFirst> is true, <hook> is installed as highest-priority
	// hook. If <bInstallAsFirst> is false, <hook> is installed before
	// <pInstallBefore> hook. That is, <hook> will handle messages before
	// <pInstallBefore>. If <pInstallBefore> is 0, <hook> is installed as
	// last hook, i.e. it will handle messages just before the original-target.
	// Hook can return true from HandleMessage() in order to prevent further
	// propagation of message, or simply examine or modify the message, then
	// return false to continue message propagation to lowest priority hooks
	// and target. If MessageHandler is installed as hook it will probably
	// receive messages for multiple targets incuding itself, so in this case
	// hook's HandleMessage() should check the <target> field of the message,
	// before deciding on appropriate action.
	//-------------------------------------------------------------------------
	void  InstallMessageHook(MessageHandler& hook,
							bool bInstallAsFirst = true,
							const MessageHandler* pInstallBefore = 0);

	void  UninstallMessageHook(MessageHandler& hook);
	void  UninstallAllMessageHooks();

	// message sending
	bool  SendMessageToServiceProvider(uint serviceID, Message& msg);
	bool  BroadcastMessage(Message& msg);
	bool  SendMessageTo(MessageHandler& mh, Message& msg);
	bool  SendMessageToSelf(Message& msg);

	// service-provider interface
	bool  IsRegisteredAsServiceProvider(uint serviceID) const;
	void  RegisterAsServiceProvider(uint serviceID);
	void  UnregisterAsServiceProvider(uint serviceID);


protected:
	MessageHandler() {}

	// return true if message is handled
	virtual  bool  HandleMessage(gui::Message& msg) { return false; }

	virtual  void  OnRegisteredAsServiceProvider(uint serviceID) {}
	virtual  void  OnUnregisteredAsServiceProvider(uint serviceID) {}

	virtual  void  ServiceProviderPreEventDispatch(double secsTimeDelta) {}
	virtual  void  ServiceProviderPostEventDispatch(double secsTimeDelta) {}


private:
	typedef  std::vector<MessageHandler*>      MESSAGE_HANDLER_VECTOR;
	typedef  MESSAGE_HANDLER_VECTOR::iterator  MESSAGE_HANDLER_ITER;

	MESSAGE_HANDLER_VECTOR  hooks;
};


//====================================================================
class VW_COMMON_API CommandSender
{
public:
	virtual ~CommandSender() {}

	MessageHandler*  GetCommandTarget() const { return pCommandTarget; }
	void             SetCommandTarget(MessageHandler* target = 0) { pCommandTarget = target; }

	uint  GetCommandID() const { return commandID; }
	void  SetCommandID(uint id = CID_Null) { commandID = id; }

	bool  SendMessageToTarget(Message& msg);
	bool  SendCommandMessageToTarget();


protected:
	CommandSender(MessageHandler* pCommandTarget = 0, uint commandID = 0);


private:
	MessageHandler*  pCommandTarget;
	uint             commandID;
};


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


