

//--------------------------------------------------------------------
#include "GUI\GUI_BehaviorBases.h"
#include "GUI\GUI_Manager.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//====================================================================
Thumb::Thumb(bool _bReportVerticalOffset,
			gui::Component* _parent)
{
	bReportVerticalOffset = _bReportVerticalOffset;
	bGrabbed = false;

	SetActionButton();

	SetCanDisable(true);

	if(_parent) AttachTo(*_parent);
}


//====================================================================
bool  Thumb::OnMouseCapture(bool bCaptureBegin)
{
	bGrabbed = bCaptureBegin;
	if(bCaptureBegin) {
		ProcessThumbGrabBegin();
	}
	else {
		ProcessThumbGrabEnd();
	}
	return true;
}


//====================================================================
bool  Thumb::OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos)
{
	if(IsMouseCaptureTarget()) {

		// report thumb-move
		float offset;
		if(bReportVerticalOffset) {
			offset = mousePos.y + GetAbsolutePosition().y - grabPosAbs.y;
		}
		else {
			offset = mousePos.x + GetAbsolutePosition().x - grabPosAbs.x;
		}
		if(offset) {
			ProcessThumbGrabOffset(offset);
		}
		return true;
	}
	return false;
}


//====================================================================
bool  Thumb::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(bDown  &&  ! repeatCount  &&  buttonCode == actionButton) {
		if(RequestMouseCapture()) {
			grabPosAbs = mousePos + GetAbsolutePosition();
		}
		return true;
	}
	return false;
}


//====================================================================
PressButton::PressButton(gui::Component* _parent)
{
	bPressed = false;

	SetActionButton();

	SetCanDisable(true);

	if(_parent) AttachTo(*_parent);
}


//====================================================================
bool  PressButton::OnDeactivate(bool bDeactivateByMouse)
{
	bPressed = false;
	return true;
}


//====================================================================
bool  PressButton::OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos)
{
	bPressed = false;
	return true;
}


//====================================================================
bool  PressButton::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(buttonCode == actionButton) {
		bPressed = bDown;
		if(bDown) {
			ProcessButtonPress();
		}
		return true;
	}
	return false;
}


//====================================================================
PushButton::PushButton(gui::Component* _parent)
{
	bPressed = false;

	SetActionButton();
	SetCanDisable(true);

	if(_parent) AttachTo(*_parent);
}


//====================================================================
bool  PushButton::OnDeactivate(bool bDeactivateByMouse)
{
	bPressed = false;
	return true;
}


//====================================================================
bool  PushButton::OnMouseCapture(bool bCaptureBegin)
{
	if(bCaptureBegin) {
		bPressed = true;
		return true;
	}
	if(bPressed) {
		HandlePressRelease();
		return true;
	}
	return false;
}


//====================================================================
bool  PushButton::OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos)
{
	if(! bHover  &&  IsMouseCaptureTarget()) {
		bPressed = HitTestSelf(mousePos);
		return true;
	}
	return false;
}


//====================================================================
bool  PushButton::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(buttonCode != actionButton  ||  repeatCount) return false;

	if(bDown) {
		RequestMouseCapture();
		return true;
	}
	if(bPressed) {
		ReleaseMouseCapture();
		HandlePressRelease();
		return true;
	}
	return false;
}


//====================================================================
void  PushButton::HandlePressRelease()
{
	bPressed = false;
	ProcessButtonRelease();
}


//====================================================================
Button::Button(gui::Component* _parent,
				gui::MessageHandler* _commandTarget,
				uint _commandID,
				eStyle _style,
				uint _groupID)
{
	style = _style;
	groupID = _groupID;

	bPressed = false;

	SetCommandTarget(_commandTarget);
	SetCommandID(_commandID);
	SetActionButton();
	SetCanDisable(true);

	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  Button::SetStyle(eStyle _style)
{
	style = _style;

	SetToggled(false);
}


//====================================================================
void  Button::SetGroupID(uint _groupID)
{
	groupID = _groupID;
}


//====================================================================
void  Button::SetOn(bool bOn, bool bSendMessage)
{
	if(style == Style_OnOff) {
		SetToggled(bOn);

		if(bSendMessage) {
			SendCommandMessageToTarget();
		}
	}
}


//====================================================================
void  Button::SetSelected(bool bSelected, bool bSendMessageToGroup, bool bSendMessageToTarget)
{
	if(style == Style_Selection) {
		SetToggled(bSelected);

		if(bSendMessageToGroup) {
			SendMessageToSiblings(gui::Message_Select(false, groupID));
		}
		if(bSendMessageToTarget) {
			SendCommandMessageToTarget();
		}
	}
}


//====================================================================
bool  Button::OnDeactivate(bool bDeactivateByMouse)
{
	bPressed = false;
	return true;
}


//====================================================================
bool  Button::OnMouseCapture(bool bCaptureBegin)
{
	// handle press-style
	if(style == Style_Press) return false;

	// handle other styles
	if(bCaptureBegin) {
		bPressed = true;
		return true;
	}
	if(bPressed) {
		HandlePressRelease();
		return true;
	}
	return false;
}


//====================================================================
bool  Button::OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos)
{
	// handle press-style
	if(style == Style_Press) {
		if(! bEnter) bPressed = false;
		return true;
	}

	// handle other styles
	return false;
}


//====================================================================
bool  Button::OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos)
{
	// handle press-style
	if(style == Style_Press) return false;

	// handle other styles
	if(! bHover  &&  IsMouseCaptureTarget()) {
		bPressed = HitTestSelf(mousePos);
		return true;
	}
	return false;
}


//====================================================================
bool  Button::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(buttonCode != actionButton) return false;

	// handle press-style
	if(style == Style_Press) {
		bPressed = bDown;
		if(bDown) SendCommandMessageToTarget();
		return true;
	}

	// handle other styles
	if(repeatCount) return false;

	if(bDown) {
		RequestMouseCapture();
		return true;
	}
	if(bPressed) {
		ReleaseMouseCapture();
		HandlePressRelease();
		return true;
	}
	return false;
}


//====================================================================
bool  Button::OnMsgOnOff(bool bOn)
{
	if(style == Style_OnOff) {
		SetToggled(bOn);
		return true;
	}
	return false;
}


//====================================================================
bool  Button::OnMsgSelect(bool bSelect, uint _groupID)
{
	if(style == Style_Selection  &&  groupID == _groupID) {
		SetToggled(bSelect);
		return true;
	}
	return false;
}


//====================================================================
void  Button::HandlePressRelease()
{
	bPressed = false;

	switch(style) {
		case Style_Press:
		case Style_Push:
			SendCommandMessageToTarget();
			break;

		case Style_OnOff:
			SetOn(! IsOn(), true);
			break;

		case Style_Selection:
			if(IsSelected()) return;
			SetSelected(true, true, true);
			break;
	}
}


//====================================================================
RubberBand::RubberBand(gui::MessageHandler* _commandTarget,
						uint _commandID)
{
	SetVisible(false);

	SetCommandTarget(_commandTarget);
	SetCommandID(_commandID);
}


//====================================================================
void  RubberBand::Begin(const vec2& _position)
{
	if(RequestMouseCapture()) {
		SetPosition(vec2::ZERO);
		UpdateGeometry();

		beginPosition = _position;
		topLeft = beginPosition;
		areaSize = vec2::ZERO;

		ShowTopLevel(true);
		SendCommandMessageToTarget();
	}
}


//====================================================================
void  RubberBand::GetLastArea(gui::Rectangle& r) const
{
	r.Set(topLeft, areaSize);
}


//====================================================================
bool  RubberBand::OnMouseCapture(bool bCaptureBegin)
{
	SetActive(bCaptureBegin);
	if(! bCaptureBegin) {
		ShowTopLevel(false);
		SendCommandMessageToTarget();
	}
	return true;
}


//====================================================================
bool  RubberBand::OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos)
{
	topLeft = ComponentMin(beginPosition, mousePos);
	areaSize = ComponentAbs(mousePos - beginPosition);
	SendCommandMessageToTarget();
	return true;
}


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


