

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Component.h"
#include "GUI\GUI_Shader.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//====================================================================
// Feature support:
// CanDisable.
//====================================================================
class VW_COMMON_API Thumb : public gui::Component
{
public:
	//-----------------
	// Thumb interface.
	//-----------------
	bool  IsReportingVerticalOffset() const { return bReportVerticalOffset; }
	bool  IsGrabbed() const { return bGrabbed; }

	void  ReleaseGrab() { ReleaseMouseCapture(); }


protected:
	Thumb(bool bReportVerticalOffset,
			gui::Component* parent = 0);

	void          SetActionButton(eMouseButton ab = MouseButton_Left) { actionButton = ab; }
	eMouseButton  GetActionButton() const { return actionButton; }

	//---------------------
	// Component overrides.
	//---------------------
	virtual  bool  OnMouseCapture(bool bCaptureBegin);
	virtual  bool  OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);

	//--------------------------------
	// Derived-class template-methods.
	//--------------------------------
	virtual  void  ProcessThumbGrabBegin() = 0;
	virtual  void  ProcessThumbGrabOffset(float offset) = 0;
	virtual  void  ProcessThumbGrabEnd() = 0;


protected:
	//------------
	// Properties.
	//------------
	bool          bReportVerticalOffset;
	eMouseButton  actionButton;

	//-------
	// Cache.
	//-------
	bool  bGrabbed;
	vec2  grabPosAbs;
};


//====================================================================
// Feature support:
// CanDisable.
//====================================================================
class VW_COMMON_API PressButton : public gui::Component
{
public:
	//-----------------------
	// PressButton interface.
	//-----------------------
	void  SetPressed(bool _bPressed) { bPressed = _bPressed; }
	bool  IsPressed() const { return bPressed; }


protected:
	PressButton(gui::Component* parent = 0);

	void          SetActionButton(eMouseButton ab = MouseButton_Left) { actionButton = ab; }
	eMouseButton  GetActionButton() const { return actionButton; }

	//---------------------
	// Component overrides.
	//---------------------
	virtual  bool  OnDeactivate(bool bDeactivateByMouse);
	virtual  bool  OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);

	//--------------------------------
	// Derived-class template-methods.
	//--------------------------------
	virtual  void  ProcessButtonPress() = 0;


protected:
	//------------
	// Properties.
	//------------
	eMouseButton  actionButton;

	//--------
	// States.
	//--------
	bool  bPressed;
};


//====================================================================
// Feature support:
// CanDisable.
//====================================================================
class VW_COMMON_API PushButton : public gui::Component
{
public:
	//----------------------
	// PushButton interface.
	//----------------------
	void  SetPressed(bool _bPressed) { bPressed = _bPressed; }
	bool  IsPressed() const { return bPressed; }


protected:
	PushButton(gui::Component* parent = 0);

	void          SetActionButton(eMouseButton ab = MouseButton_Left) { actionButton = ab; }
	eMouseButton  GetActionButton() const { return actionButton; }

	//---------------------
	// Component overrides.
	//---------------------
	virtual  bool  OnDeactivate(bool bDeactivateByMouse);
	virtual  bool  OnMouseCapture(bool bCaptureBegin);
	virtual  bool  OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);

	//------------------
	// Internal helpers.
	//------------------
	void  HandlePressRelease();

	//--------------------------------
	// Derived-class template-methods.
	//--------------------------------
	virtual  void  ProcessButtonRelease() = 0;


protected:
	//------------
	// Properties.
	//------------
	eMouseButton  actionButton;

	//--------
	// States.
	//--------
	bool  bPressed;
};


//====================================================================
// Feature support:
// CanDisable.
//====================================================================
class VW_COMMON_API Button : public gui::Component, public gui::CommandSender
{
public:
	//---------------
	// Button styles.
	//---------------
	enum eStyle
	{
		Style_Press = 0,
		Style_Push,
		Style_OnOff,
		Style_Selection,

		Style_Count,
		Style_Force32Bits = 0x7FFFFFFF
	};

	//-------
	// Flags.
	//-------
	enum eFlag
	{
		Flag_Toggled = 1
	};


public:
	//------------------
	// Button interface.
	//------------------
	void    SetStyle(eStyle style = Style_Push);
	eStyle  GetStyle() const { return style; }

	void    SetGroupID(uint groupID = 0);
	uint    GetGroupID() const { return groupID; }

	void    SetOn(bool bOn, bool bSendMessage = false);
	bool    IsOn() const { return IsToggled(); }

	void    SetSelected(bool bSelected, bool bSendMessageToGroup = false, bool bSendMessageToTarget = false);
	bool    IsSelected() const { return IsToggled(); }
	void    Select(bool bSendMessageToTarget = true) { SetSelected(true, true, bSendMessageToTarget); }

	void    SetPressed(bool _bPressed) { bPressed = _bPressed; }
	bool    IsPressed() const { return bPressed; }

	void    SetToggled(bool bToggled) { SetFlagsOnOff(Flag_Toggled, bToggled); }
	bool    IsToggled() const { return ToBool(IsFlagOn(Flag_Toggled)); }


protected:
	Button(gui::Component* parent = 0,
			gui::MessageHandler* commandTarget = 0,
			uint commandID = 0,
			eStyle style = Style_Push,
			uint groupID = 0);

	void          SetActionButton(eMouseButton ab = MouseButton_Left) { actionButton = ab; }
	eMouseButton  GetActionButton() const { return actionButton; }

	//---------------------
	// Component overrides.
	//---------------------
	virtual  bool  OnDeactivate(bool bDeactivateByMouse);
	virtual  bool  OnMouseCapture(bool bCaptureBegin);
	virtual  bool  OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos);
	virtual  bool  OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  bool  OnMsgOnOff(bool bOn);
	virtual  bool  OnMsgSelect(bool bSelect, uint groupID);

	//------------------
	// Internal helpers.
	//------------------
	void  HandlePressRelease();


protected:
	//------------
	// Properties.
	//------------
	eStyle        style;
	uint          groupID;
	eMouseButton  actionButton;

	//--------
	// States.
	//--------
	bool  bPressed;
};


//====================================================================
// Sends messages to target on each begin/end/mouse-move.
// Uses screen coords. Call GetLastArea() to check active area.
//====================================================================
class VW_COMMON_API RubberBand : public gui::Component, public gui::CommandSender
{
public:
	//----------------------
	// RubberBand interface.
	//----------------------
	void  Begin(const vec2& position = vec2::ZERO);
	void  GetLastArea(gui::Rectangle& r) const;


protected:
	RubberBand(gui::MessageHandler* commandTarget = 0,
				uint commandID = 0);

	//---------------------
	// Component overrides.
	//---------------------
	virtual  bool  OnMouseCapture(bool bCaptureBegin);
	virtual  bool  OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos);


protected:
	//-------
	// Cache.
	//-------
	vec2  beginPosition;
	vec2  topLeft;
	vec2  areaSize;
};


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


