

//--------------------------------------------------------------------
#include "GUI\GUI_Component.h"
#include "GUI\GUI_Manager.h"
#include "BaseTools\STLExtensions.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//====================================================================
bool  Component::s_bDefaultEnableShaderHitTest = false;


//====================================================================
Component::Component()
{
	layer = 0;
	parent = 0;

	absolutePosition = vec2::ZERO;
	position = vec2::ZERO;
	size = vec2::ONE;
	minSize = vec2::ZERO;
	maxSize = vec2::INFINITY;

	bEnableShaderHitTest = s_bDefaultEnableShaderHitTest;
	bNotGhost = true;
	bRecalcLayoutDone = false;
	bCanDisable = false;
	bCanFocus = false;
	bOverlay = false;
	bVisible = true;
	bEnabled = true;
	bActive = false;
	bFocused = false;
	bMouseOver = false;
	bMouseOverDirect = false;
	flags = 0;

	autoPositionOnWindowResizeX = AutoPosition_KeepInsideWindow;
	autoPositionOnWindowResizeY = AutoPosition_KeepInsideWindow;
	autoPositionAlignOffsetX = 0;
	autoPositionAlignOffsetY = 0;

	defaultCommandArea = gui::CommandArea_Client;
	pClientAreaCursor = 0;
	pBaseShader = 0;

	SetActiveGeomUpdateEvent();

	gui::Manager::Instance().NotifyObjectCreateDelete(*this, true);
}


//====================================================================
Component::~Component()
{
	DetachFromParent();
	DeleteAllChildren();

	gui::Manager::Instance().NotifyObjectCreateDelete(*this, false);
}


//====================================================================
std::string  Component::MakeHierarchyTrace(const std::string& levelSeparator) const
{
	std::string s = VW_GET_OBJECT_NAME(*this);
	const Component* c = GetParent();
	while(c) {
		s.insert(0, levelSeparator).insert(0, VW_GET_OBJECT_NAME(*c));
		c = c->GetParent();
	}
	return s;
}


//====================================================================
void  Component::DumpHierarchy(std::ostream& os, uint tabLevel,
							bool bPrintDepth, bool bPrintAttribs,
							uint depth, bool bLimitDepth, uint maxDepth) const

{
	if(bLimitDepth  &&  (depth > maxDepth)) return;

	TabEffector  te(tabLevel);
	const char* sep = ", ";

	// dump self
	os << te;
	if(bPrintDepth) {
		os << depth << sep;
	}
	os << VW_GET_CLASS_NAME(*this) << sep << "0x" << this;

	if(bPrintAttribs) {
		os << sep << "Shader=" << shaderName;
		os << sep << "Flags=";
		if(! bVisible) os << "H";
	}
	os << nl;

	// dump children
	for(uint i = 0; i < children.size(); i++) {
		children[i]->DumpHierarchy(os, tabLevel + 1, bPrintDepth, bPrintAttribs,
									depth + 1, bLimitDepth, maxDepth);
	}
}


//====================================================================
void  Component::Attach(Component& c)
{
	if(&c == this) return;
	if(IsParentOf(c)) return;

	if(c.IsAncestorOf(*this)) {
		VW_THROW(GUIException,
			"Attaching ancestor: " + VW_GET_OBJECT_NAME(c) +
			" to descendant: " + VW_GET_OBJECT_NAME(*this));
	}
	c.DetachFromParent();

	UniqueContainerPushBack(children, &c);
	c.parent = this;

	gui::Manager::Instance().NotifyComponentHierarchyChanged(c);
	gui::Manager::Instance().NotifyComponentHierarchyChanged(*this);

	OnAttachedChild(c);
	c.OnAttachedToParent();
	c.PropagateChangedBranchVisible(IsBranchVisible());
}


//====================================================================
void  Component::Detach(Component& c)
{
	if(! IsParentOf(c)) {
		VW_THROW(GUIException,
			"Detaching non-child: " + VW_GET_OBJECT_NAME(c) +
			" from: " + VW_GET_OBJECT_NAME(*this));
	}
	EraseFromContainer(children, &c);
	c.parent = 0;

	gui::Manager::Instance().NotifyComponentHierarchyChanged(c);
	gui::Manager::Instance().NotifyComponentHierarchyChanged(*this);

	c.OnDetachedFromParent();
	OnDetachedChild(c);
	c.PropagateChangedBranchVisible(c.IsBranchVisible());
}


//====================================================================
void  Component::DetachAllChildren()
{
	while(children.size()) {
		Detach(*children.back());
	}
}


//====================================================================
void  Component::DeleteAllChildren()
{
	while(children.size()) {
		Component* child = children.front();
		Detach(*child);
		delete child;
	}
}


//====================================================================
Component&  Component::GetTopLevelAncestor()
{
	Component* c = this;
	while(c->parent) {
		c = c->parent;
	}
	return *c;
}


//====================================================================
bool  Component::IsDescendantOf(const Component& c) const
{
	if(&c == this) return false;

	const Component* d = parent;
	while(d) {
		if(d == &c) return true;
		d = d->parent;
	}
	return false;
}


//====================================================================
uint  Component::CalcDepth() const
{
	uint depth = 0;
	const Component* c = parent;
	while(c) {
		c = c->parent;
		depth++;
	}
	return depth;
}


//====================================================================
void  Component::SetLayer(uint _layer)
{
	if(layer != _layer) {
		layer = _layer;

		gui::Manager::Instance().NotifyComponentLayerChanged(*this);
	}
}


//====================================================================
void  Component::BringToFront()
{
	if(IsTopLevel()) {
		gui::Manager::Instance().BringToFront(*this);
	}
}


//====================================================================
void  Component::SendToBack()
{
	if(IsTopLevel()) {
		gui::Manager::Instance().SendToBack(*this);
	}
}


//====================================================================
void  Component::BringInFront(Component& c)
{
	if(IsTopLevel()) {
		gui::Manager::Instance().SetZOrder(*this, c);
	}
}


//====================================================================
void  Component::SendBehind(Component& c)
{
	if(IsTopLevel()) {
		gui::Manager::Instance().SetZOrder(c, *this);
	}
}


//====================================================================
void  Component::SetGhost(bool _bGhost)
{
	if(bNotGhost == _bGhost) {
		bNotGhost = !_bGhost;

		gui::Manager::Instance().ForceHitTestAtNextFrame();
	}
}


//====================================================================
void  Component::SetOverlay(bool _bOverlay)
{
	if(bOverlay != _bOverlay) {
		bOverlay = _bOverlay;

		gui::Manager::Instance().NotifyComponentOverlayChanged(*this);
	}
}


//====================================================================
void  Component::SetVisible(bool _bVisible)
{
	if(bVisible != _bVisible) {

		bVisible = _bVisible;
		OnChangedVisible();

		if(IsTopLevel()) {
			PropagateChangedBranchVisible(bVisible);
		}
		gui::Manager::Instance().NotifyComponentVisibilityChanged(*this);
		bRecalcLayoutDone = false;
	}
	if(! bVisible) {
		SetActive(false);
		PropagateActiveToDescendants(false);
	}
}


//====================================================================
bool  Component::IsBranchVisible() const
{
	const Component* c = this;
	while(c) {
		if(! c->IsVisible()) return false;
		c = c->parent;
	}
	return true;
}


//====================================================================
void  Component::PropagateVisibleToChildren(bool bVisible)
{
	for(uint i = 0; i < children.size(); i++) {
		children[i]->SetVisible(bVisible);
	}
}


//====================================================================
void  Component::PropagateVisibleToDescendants(bool bVisible)
{
	for(uint i = 0; i < children.size(); i++) {
		children[i]->SetVisible(bVisible);
		children[i]->PropagateVisibleToDescendants(bVisible);
	}
}


//====================================================================
void  Component::PropagateActiveToDescendants(bool bActive)
{
	for(uint i = 0; i < children.size(); i++) {
		children[i]->SetActive(bActive);
		children[i]->PropagateActiveToDescendants(bActive);
	}
}


//====================================================================
void  Component::SetEnabled(bool _bEnabled)
{
	if(! bCanDisable) {
		_bEnabled = true;
	}
	if(bEnabled != _bEnabled) {

		bEnabled = _bEnabled;
		OnChangedEnabled();

		if(! bEnabled) {
			SetActive(false);
			SetFocused(false);
		}
	}
}


//====================================================================
void  Component::SetFocused(bool _bFocused)
{
	if(! bEnabled  ||  ! bCanFocus) {
		_bFocused = false;
	}
	if(bFocused != _bFocused) {

		bFocused = _bFocused;
		OnChangedFocused();
	}
}


//====================================================================
void  Component::SetActive(bool _bActive)
{
	if(! bEnabled) {
		bActive = false;
	}
	if(bActive != _bActive) {

		bActive = _bActive;
		OnChangedActive();
	}
}


//====================================================================
void  Component::SetMouseOver(bool _bMouseOver, bool _bMouseOverDirect)
{
	bMouseOver = _bMouseOver;
	bMouseOverDirect = _bMouseOverDirect;
}


//====================================================================
void  Component::SetFlags(uint _flags)
{
	if(flags != _flags) {
		flags = _flags;
		OnChangedFlags();
	}
}


//====================================================================
bool  Component::IsMouseCaptureTarget()
{
	return (gui::Manager::Instance().GetMouseCaptureTarget() == this);
}


//====================================================================
bool  Component::RequestMouseCapture()
{
	return gui::Manager::Instance().RequestMouseCapture(*this);
}


//====================================================================
void  Component::ReleaseMouseCapture()
{
	gui::Manager::Instance().ReleaseMouseCapture(*this);
}


//====================================================================
bool  Component::IsKeyboardFocusTarget()
{
	return (gui::Manager::Instance().GetKeyboardFocusTarget() == this);
}


//====================================================================
bool  Component::RequestKeyboardFocus()
{
	return gui::Manager::Instance().RequestKeyboardFocus(*this);
}


//====================================================================
bool  Component::RequestKeyboardFocusToAncestor(bool bReleaseIfTopLevel)
{
	if(! IsKeyboardFocusTarget()) return false;

	gui::Manager& gm = gui::Manager::Instance();

	if(parent) {
		return gm.RequestKeyboardFocus(*parent);
	}
	if(bReleaseIfTopLevel) {
		gm.ReleaseKeyboardFocus(*this);
		return true;
	}
	return false;
}


//====================================================================
void  Component::ReleaseKeyboardFocus()
{
	gui::Manager::Instance().ReleaseKeyboardFocus(*this);
}


//====================================================================
void  Component::ShowTopLevel(bool bShow, bool bAutoBringToFront)
{
	SetVisible(bShow);

	if(IsTopLevel()  &&  bVisible  &&  bAutoBringToFront) {
		gui::Manager::Instance().BringToFront(*this);
	}
}


//====================================================================
void  Component::ActivateTopLevel(bool bAutoBringToFront)
{
	gui::Manager::Instance().ActivateTopLevel(*this, bAutoBringToFront);
}


//====================================================================
void  Component::RunModal()
{
	gui::Manager::Instance().RunModal(*this);
}


//====================================================================
void  Component::EndModal()
{
	gui::Manager::Instance().EndModal(*this);
}


//====================================================================
bool  Component::IsRunningModal()
{
	return (gui::Manager::Instance().GetModalTarget() == this);
}


//====================================================================
void  Component::SetAutoPositionOnWindowResize(eAutoPosition apx, eAutoPosition apy,
											float alignOffsetX, float alignOffsetY)
{
	autoPositionOnWindowResizeX = apx;
	autoPositionOnWindowResizeY = apy;

	autoPositionAlignOffsetX = alignOffsetX;
	autoPositionAlignOffsetY = alignOffsetY;
}


//====================================================================
void  Component::GetAutoPositionOnWindowResize(eAutoPosition& apx, eAutoPosition& apy,
											float& alignOffsetX, float& alignOffsetY) const
{
	apx = autoPositionOnWindowResizeX;
	apy = autoPositionOnWindowResizeY;

	alignOffsetX = autoPositionAlignOffsetX;
	alignOffsetY = autoPositionAlignOffsetY;
}


//====================================================================
void  Component::SetPositionX(float x)
{
	position.x = x;
	gui::Manager::Instance().ForceHitTestAtNextFrame();
}


//====================================================================
void  Component::SetPositionY(float y)
{
	position.y = y;
	gui::Manager::Instance().ForceHitTestAtNextFrame();
}


//====================================================================
void  Component::SetPosition(float x, float y)
{
	position.Set(x, y);
	gui::Manager::Instance().ForceHitTestAtNextFrame();
}


//====================================================================
void  Component::SetPosition(const vec2& _position)
{
	position = _position;
	gui::Manager::Instance().ForceHitTestAtNextFrame();
}


//====================================================================
void  Component::SetSizeX(float x)
{
	size.x = x;
	gui::Manager::Instance().ForceHitTestAtNextFrame();
}


//====================================================================
void  Component::SetSizeY(float y)
{
	size.y = y;
	gui::Manager::Instance().ForceHitTestAtNextFrame();
}


//====================================================================
void  Component::SetSize(float x, float y)
{
	size.Set(x, y);
	gui::Manager::Instance().ForceHitTestAtNextFrame();
}


//====================================================================
void  Component::SetSize(const vec2& _size)
{
	size = _size;
	gui::Manager::Instance().ForceHitTestAtNextFrame();
}


//====================================================================
void  Component::SetMinSize(const vec2& _minSize)
{
	minSize = ComponentClamp(_minSize, vec2::ZERO, maxSize);
}


//====================================================================
void  Component::SetMaxSize(const vec2& _maxSize)
{
	maxSize = ComponentClamp(_maxSize, minSize, vec2::INFINITY);
}


//====================================================================
vec2  Component::CalcMinMaxClampedSize(const vec2& _size) const
{
	return ComponentClamp(_size, minSize, maxSize);
}


//====================================================================
vec2  Component::GetParentSize() const
{
	return (parent ? parent->GetSize() : gui::Manager::Instance().GetWindowSize());
}


//====================================================================
vec2  Component::CalcRelativeMousePosition() const
{
	return gui::Manager::Instance().GetMousePosition() - absolutePosition;
}


//====================================================================
void  Component::ReportDirtyGeometry()
{
	gui::Manager::Instance().ReportDirtyGeometry(*this);
}


//====================================================================
void  Component::UpdateAbsolutePositionCache()
{
	if(! bVisible) return;

	// update own absolute-position
	if(GetParent()) {
		absolutePosition = position + GetParent()->GetAbsolutePosition();
	}
	else {
		absolutePosition = position;
	}

	// update child absolute-positions
	for(uint i = 0; i < children.size(); i++) {
		children[i]->UpdateAbsolutePositionCache();
	}
}


//====================================================================
void  Component::UpdateGeometry(bool bFromTopLevel)
{
	Component& c = (bFromTopLevel ? GetTopLevelAncestor() : *this);

	// pass 1: calculate size (bottom-to-top)
	c.PropagateRecalcSize();

	// pass 2: calculate layout (top-to-bottom)
	c.PropagateRecalcLayout();

	// pass 3: auto-reset recalc-layout flag
	c.PropagateResetRecalcLayoutFlag(false);

	// pass 4: update absolute-position cache
	c.UpdateAbsolutePositionCache();

	// pass 5: update component-specific geometry cache
	c.PropagateUpdateGeometryCache();
}


//====================================================================
void  Component::PropagateRecalcSize()
{
	if(! bVisible) return;

	// bottom-to-top
	for(uint i = 0; i < children.size(); i++) {
		children[i]->PropagateRecalcSize();
	}
	RecalcSize();
}


//====================================================================
void  Component::PropagateRecalcLayout()
{
	if(! bVisible) return;

	// top-to-bottom
	if(bRecalcLayoutDone) return;
	RecalcLayout();
	bRecalcLayoutDone = true;

	for(uint i = 0; i < children.size(); i++) {
		children[i]->PropagateRecalcLayout();
	}
}


//====================================================================
void  Component::PropagateResetRecalcLayoutFlag(bool bLayoutDone)
{
	if(! bVisible) return;

	bRecalcLayoutDone = bLayoutDone;

	for(uint i = 0; i < children.size(); i++) {
		children[i]->PropagateResetRecalcLayoutFlag(bLayoutDone);
	}
}


//====================================================================
void  Component::PropagateUpdateGeometryCache()
{
	if(! bVisible) return;

	UpdateGeometryCache();

	for(uint i = 0; i < children.size(); i++) {
		children[i]->PropagateUpdateGeometryCache();
	}
}


//====================================================================
void  Component::ForceTopLevelComponentInsideWindow()
{
	if(! IsTopLevel()) return;

	// force inside window
	if(! gui::Manager::Instance().IntersectsWindow(*this)) {

		vec2  newPos = (gui::Manager::Instance().GetWindowSize() - GetSize()) * 0.5f;
		SetPosition(ComponentMax(newPos, vec2::ZERO));
	}
}


//====================================================================
bool  Component::HitTest(const vec2& testPoint) const
{
	if(bNotGhost  &&  HitTestSelf(testPoint)) {
		if(bEnableShaderHitTest  &&  pBaseShader) {
			return pBaseShader->HitTest(*this, testPoint);
		}
		return true;
	}
	return false;
}


//====================================================================
gui::eCommandArea  Component::HitTestCommandArea(const vec2& testPoint) const
{
	gui::eCommandArea ca = HitTestCommandAreaSelf(testPoint);
	if(bEnableShaderHitTest  &&  pBaseShader) {
		return pBaseShader->HitTestCommandArea(ca, testPoint);
	}
	return ca;
}


//====================================================================
void  Component::SetClientAreaCursorName(const std::string& cursorName)
{
	clientAreaCursorName = cursorName;
	RequestCursorRecache();
}


//====================================================================
void  Component::SetShaderName(const std::string& s)
{
	shaderName = s;
	RequestShaderRecache();
}


//====================================================================
void  Component::OnWindowResize(uint width, uint height)
{
	if(! IsTopLevel()) return;

	vec2 newPos = GetPosition();

	// auto-position x
	switch(autoPositionOnWindowResizeX) {
		case AutoPosition_AlignNear:
			newPos.x = autoPositionAlignOffsetX;
			break;

		case AutoPosition_AlignFar:
			newPos.x = width - GetSize().x - autoPositionAlignOffsetX;
			break;

		case AutoPosition_Center:
			newPos.x = (width - GetSize().x) * 0.5f;
			break;

		case AutoPosition_KeepInsideWindow:
			if(GetPosition().x >= width) {
				newPos.x = ClampedLow<float>(float(width) - GetSize().x, 0);
			}
			break;
	}

	// auto-position y
	switch(autoPositionOnWindowResizeY) {
		case AutoPosition_AlignNear:
			newPos.y = autoPositionAlignOffsetY;
			break;

		case AutoPosition_AlignFar:
			newPos.y = height - GetSize().y - autoPositionAlignOffsetY;
			break;

		case AutoPosition_Center:
			newPos.y = (height - GetSize().y) * 0.5f;
			break;

		case AutoPosition_KeepInsideWindow:
			if(GetPosition().y >= height) {
				newPos.y = ClampedLow<float>(float(height) - GetSize().y, 0);
			}
			break;
	}

	SetPosition(newPos);
}


//====================================================================
bool  Component::HitTestSelf(const vec2& testPoint) const
{
	return
	(testPoint.x >= 0)  &&  (testPoint.x < size.x)  &&
	(testPoint.y >= 0)  &&  (testPoint.y < size.y);
}


//====================================================================
gui::eCommandArea  Component::HitTestCommandAreaSelf(const vec2& testPoint) const
{
	return defaultCommandArea;
}


//====================================================================
void  Component::DrawSelf(gui::Renderer& gr)
{
	if(pBaseShader) {
		pBaseShader->Shade(*this, gr);
	}
}


//====================================================================
bool  Component::CacheCursorHandles(gui::Manager& gm)
{
	if(clientAreaCursorName.empty()) return true;
	pClientAreaCursor = gm.GetCursor(clientAreaCursorName);
	return (pClientAreaCursor != 0);
}


//====================================================================
void  Component::RequestCursorRecache()
{
	gui::Manager::Instance().RequestCursorRecache(*this);
}


//====================================================================
bool  Component::CacheShaderHandles(gui::Manager& gm)
{
	const gui::BaseShader* pShader = gm.GetShader(shaderName);

	if(CheckShader(pShader)) {
		SetBaseShader(pShader);
		return true;
	}
	SetBaseShader(0);
	return false;
}


//====================================================================
void  Component::PostCacheShaderHandles()
{
	ReportDirtyGeometry();
}


//====================================================================
void  Component::RequestShaderRecache()
{
	gui::Manager::Instance().RequestShaderRecache(*this);
}


//====================================================================
bool  Component::SendMessageToParent(Message& msg)
{
	if(parent) {
		return parent->DispatchMessage(msg);
	}
	return false;
}


//====================================================================
bool  Component::SendMessageToSiblings(Message& msg)
{
	bool bHandled = false;
	if(parent) {
		for(uint i = 0; i < parent->children.size(); i++) {
			Component& sibling = *(parent->children[i]);
			if(&sibling == this) continue; //skip self
			if(sibling.DispatchMessage(msg)) bHandled = true;
		}
	}
	return bHandled;
}


//====================================================================
bool  Component::SendMessageToPreviousSibling(bool bAllowSendToSelf, Message& msg)
{
	Component* pSibling = 0;
	if(parent  &&  parent->children.size() >= 2) {

		for(uint i = 0; i < parent->children.size(); i++) {
			if(parent->children[i] == this) {
				pSibling = (i == 0) ? parent->children.back() : parent->children[i - 1];
			}
		}
	}
	if(! pSibling  &&  bAllowSendToSelf) {
		pSibling = this;
	}
	if(pSibling) {
		return SendMessageTo(*pSibling, msg);
	}
	return false;
}


//====================================================================
bool  Component::SendMessageToNextSibling(bool bAllowSendToSelf, Message& msg)
{
	Component* pSibling = 0;
	if(parent  &&  parent->children.size() >= 2) {

		for(uint i = 0; i < parent->children.size(); i++) {
			if(parent->children[i] == this) {
				pSibling = (i == parent->children.size() - 1) ?
							parent->children.front() : parent->children[i + 1];
			}
		}
	}
	if(! pSibling  &&  bAllowSendToSelf) {
		pSibling = this;
	}
	if(pSibling) {
		return SendMessageTo(*pSibling, msg);
	}
	return false;
}


//====================================================================
bool  Component::SendMessageToSelfAndSiblings(Message& msg)
{
	bool bHandled = false;
	if(SendMessageToSelf(msg)) bHandled = true;
	if(SendMessageToSiblings(msg)) bHandled = true;
	return bHandled;
}


//====================================================================
bool  Component::SendMessageToChildren(Message& msg)
{
	bool bHandled = false;
	for(uint i = 0; i < children.size(); i++) {
		if(children[i]->DispatchMessage(msg)) bHandled = true;
	}
	return bHandled;
}


//====================================================================
bool  Component::SendMessageToFirstChild(Message& msg)
{
	bool bHandled = false;
	if(children.size()) {
		if(children.front()->DispatchMessage(msg)) bHandled = true;
	}
	return bHandled;
}


//====================================================================
bool  Component::SendMessageToLastChild(Message& msg)
{
	bool bHandled = false;
	if(children.size()) {
		if(children.back()->DispatchMessage(msg)) bHandled = true;
	}
	return bHandled;
}


//====================================================================
bool  Component::HandleMessage(gui::Message& msg)
{
	// messages directed to other components
	// (received if installed as message-hook)
	if(msg.target != this) {
		return HandleMessageAsHook(msg);
	}

	// messages directed to this component
	using namespace gui;
	switch(msg.messageID) {

	case MID_ServiceCommand: {
		MsgPtr<Message_ServiceCommand> m(msg);

		// handle tooltip queries
		if(m->GetServiceID() == SID_ToolTip) {
			MsgPtr<Message_ServiceCommand_ToolTip> mtt(msg);

			if(m->GetServiceCommandID() == SCID_QueryParams) {
				ToolTipParams* ttp = mtt->GetToolTipParams();
				if(ttp) return OnQueryToolTipParams(*ttp);
			}
		}
		return false;
	}
	case MID_Command: {
		MsgPtr<Message_Command> m(msg);
		return OnCommand(m->GetCommandID());
	}
	case MID_GeometryUpdateByManager: {
		MsgPtr<Message_GeometryUpdateByManager> m(msg);
		bool bConfirmEvent = m->GetConfirmEvent();
		bool retVal = OnGeometryUpdateByManager(bConfirmEvent, m->GetUpdateEvent(),
							m->GetCommandArea(), m->GetButtonCode(),
							m->GetNewPosition(), m->GetNewSize());
		m->SetConfirmEvent(bConfirmEvent);
		return retVal;
	}
	case MID_Activate: {
		MsgPtr<Message_Activate> m(msg);
		bool bAutoBringToFront = m->GetAutoBringToFront();
		bool bAutoActivate = m->GetAutoActivate();
		bool retVal = OnActivate(m->GetActivateByMouse(), m->GetButtonCode(),
								m->GetMousePosition(), m->GetClickedDescendant(),
								bAutoBringToFront, bAutoActivate);
		m->SetAutoBringToFront(bAutoBringToFront);
		m->SetAutoActivate(bAutoActivate);
		return retVal;
	}
	case MID_Deactivate: {
		MsgPtr<Message_Deactivate> m(msg);
		return OnDeactivate(m->GetDeactivateByMouse());
	}
	case MID_MouseCapture: {
		MsgPtr<Message_MouseCapture> m(msg);
		return OnMouseCapture(m->GetCaptureBegin());
	}
	case MID_MouseEnterExit: {
		MsgPtr<Message_MouseEnterExit> m(msg);
		return OnMouseEnterExit(m->GetMouseEnter(), m->GetMouseOverDirect(), m->GetMousePosition());
	}
	case MID_MouseQueryCursor: {
		MsgPtr<Message_MouseQueryCursor> m(msg);
		const Cursor* pCursor = m->GetCursor();
		bool retVal = OnMouseQueryCursor(pCursor, m->GetCommandArea(), m->GetMousePosition());
		m->SetCursor(pCursor);
		return retVal;
	}
	case MID_MouseMove: {
		MsgPtr<Message_MouseMove> m(msg);
		return OnMouseMove(m->GetHover(), m->GetHoverTimeSeconds(), m->GetMousePosition());
	}
	case MID_MouseWheel: {
		MsgPtr<Message_MouseWheel> m(msg);
		return OnMouseWheel(m->GetWheelDelta(), m->GetMousePosition());
	}
	case MID_MouseButton: {
		MsgPtr<Message_MouseButton> m(msg);
		return OnMouseButton(m->GetButtonDown(), m->GetButtonCode(), m->GetDoubleClick(),
							m->GetRepeatCount(), m->GetMousePosition());
	}
	case MID_KeyboardQueryFocus: {
		MsgPtr<Message_KeyboardQueryFocus> m(msg);
		bool bWantFocus = m->GetWantFocus();
		bool bNeedCharacterData = m->GetNeedCharacterData();
		bool retVal = OnKeyboardQueryFocus(m->GetFocusByMouse(), m->GetButtonCode(), bWantFocus, bNeedCharacterData);
		m->SetWantFocus(bWantFocus);
		m->SetNeedCharacterData(bNeedCharacterData);
		return retVal;
	}
	case MID_KeyboardFocus: {
		MsgPtr<Message_KeyboardFocus> m(msg);
		return OnKeyboardFocus(m->GetGotFocus());
	}
	case MID_KeyboardKey: {
		MsgPtr<Message_KeyboardKey> m(msg);
		return OnKeyboardKey(m->GetKeyDown(), m->GetKeyCode(), m->GetRepeatCount(), m->GetCharData());
	}
	case MID_Show: {
		MsgPtr<Message_Show> m(msg);
		return OnMsgShow(m->GetShow());
	}
	case MID_ShowChildren: {
		MsgPtr<Message_ShowChildren> m(msg);
		return OnMsgShowChildren(m->GetShow());
	}
	case MID_ShowDescendants: {
		MsgPtr<Message_ShowDescendants> m(msg);
		return OnMsgShowDescendants(m->GetShow());
	}
	case MID_Enable: {
		MsgPtr<Message_Enable> m(msg);
		return OnMsgEnable(m->GetEnable());
	}
	case MID_Flags: {
		MsgPtr<Message_Flags> m(msg);
		return OnMsgFlags(m->GetFlags());
	}
	case MID_FlagsOnOff: {
		MsgPtr<Message_FlagsOnOff> m(msg);
		return OnMsgFlagsOnOff(m->GetFlags(), m->GetOn());
	}
	case MID_OnOff: {
		MsgPtr<Message_OnOff> m(msg);
		return OnMsgOnOff(m->GetOn());
	}
	case MID_Select: {
		MsgPtr<Message_Select> m(msg);
		return OnMsgSelect(m->GetSelect(), m->GetGroupID());
	}
	case MID_Scroll: {
		MsgPtr<Message_Scroll> m(msg);
		return OnMsgScroll(m->GetForward(), m->GetDistance(), m->GetOffset());
	}
	case MID_PreCommandExecute: {
		MsgPtr<Message_PreCommandExecute> m(msg);
		return OnMsgPreCommandExecute(m->GetCommandID());
	}
	case MID_SendCommand: {
		MsgPtr<Message_SendCommand> m(msg);
		return OnMsgSendCommand();
	}
	}
	return false;
}


//====================================================================
bool  Component::HandleMessageAsHook(gui::Message& msg)
{
	// messages directed to this component
	using namespace gui;
	switch(msg.messageID) {

	case MID_MouseWheel: {
		MsgPtr<Message_MouseWheel> m(msg);
		return OnMouseWheel(msg.target, m->GetWheelDelta(), m->GetMousePosition());
	}
	case MID_MouseButton: {
		MsgPtr<Message_MouseButton> m(msg);
		return OnMouseButton(msg.target, m->GetButtonDown(), m->GetButtonCode(),
							m->GetDoubleClick(), m->GetRepeatCount(), m->GetMousePosition());
	}
	case MID_KeyboardQueryFocus: {
		MsgPtr<Message_KeyboardQueryFocus> m(msg);
		bool bWantFocus = m->GetWantFocus();
		bool bNeedCharacterData = m->GetNeedCharacterData();
		bool retVal = OnKeyboardQueryFocus(msg.target, m->GetFocusByMouse(), m->GetButtonCode(), bWantFocus, bNeedCharacterData);
		m->SetWantFocus(bWantFocus);
		m->SetNeedCharacterData(bNeedCharacterData);
		return retVal;
	}
	case MID_KeyboardFocus: {
		MsgPtr<Message_KeyboardFocus> m(msg);
		return OnKeyboardFocus(msg.target, m->GetGotFocus());
	}
	case MID_KeyboardKey: {
		MsgPtr<Message_KeyboardKey> m(msg);
		return OnKeyboardKey(msg.target, m->GetKeyDown(), m->GetKeyCode(), m->GetRepeatCount(), m->GetCharData());
	}
	}
	return false;
}


//====================================================================
bool  Component::OnCommand(uint commandID)
{
	switch(commandID) {
		case CID_ControlValuesChanged: {
			// propagate notification to parent by default
			if(parent) return SendMessageToParent(Message_Command(commandID));
			return false;
		}
	}
	return false;
}


//====================================================================
bool  Component::OnGeometryUpdateByManager(bool& bConfirmEvent,
								gui::eGeometryUpdateEvent updateEvent,
								gui::eCommandArea commandArea, eMouseButton buttonCode,
								const vec2& newPosition, const vec2& newSize)
{
	if(updateEvent == GeometryUpdateEvent_QueryBegin) {
		bConfirmEvent = (buttonCode == MouseButton_Left);
		return true;
	}
	return false;
}


//====================================================================
bool  Component::OnActivate(bool bActivateByMouse, eMouseButton buttonCode,
							const vec2& mousePosition, gui::Component* pClickedDescendant,
							bool& bAutoBringToFront, bool& bAutoActivate)
{
	if(bActivateByMouse) {
		bAutoBringToFront = bAutoActivate = (buttonCode == MouseButton_Left);
	}
	return true;
}


//====================================================================
bool  Component::OnMouseQueryCursor(const gui::Cursor* &pCursor,
						gui::eCommandArea commandArea, const vec2& mousePos)
{
	if(commandArea == gui::CommandArea_Client) pCursor = pClientAreaCursor;
	return true;
}


//====================================================================
bool  Component::OnQueryToolTipParams(gui::ToolTipParams& ttp)
{
	if(toolTipText.length()) {
		ttp.bDisplay = true;
		ttp.text = toolTipText;
		ttp.mouseMoveIgnoreAreaPosition = absolutePosition;
		ttp.mouseMoveIgnoreAreaSize = size;
		return true;
	}
	return false;
}


//====================================================================
bool  Component::OnMsgShow(bool bShow)
{
	SetVisible(bShow);
	return true;
}


//====================================================================
bool  Component::OnMsgShowChildren(bool bShow)
{
	PropagateVisibleToChildren(bShow);
	return true;
}


//====================================================================
bool  Component::OnMsgShowDescendants(bool bShow)
{
	PropagateVisibleToDescendants(bShow);
	return true;
}


//====================================================================
bool  Component::OnMsgEnable(bool bEnable)
{
	SetEnabled(bEnable);
	return true;
}


//====================================================================
bool  Component::OnMsgFlags(uint flags)
{
	SetFlags(flags);
	return true;
}


//====================================================================
bool  Component::OnMsgFlagsOnOff(uint flags, bool bOn)
{
	SetFlagsOnOff(flags, bOn);
	return true;
}


//====================================================================
bool  Component::OnMsgPreCommandExecute(uint commandID)
{
	return SendMessageToParent(gui::Message_PreCommandExecute(commandID));
}


//====================================================================
void  Component::SetActiveGeomUpdateEvent(gui::eGeometryUpdateEvent ev, gui::eCommandArea ca)
{
	activeGeomUpdateEvent = ev;
	activeGeomUpdateEventCommandArea = ca;
}


//====================================================================
void  Component::PropagateChangedBranchVisible(bool bBranchVisible)
{
	OnChangedBranchVisible(bBranchVisible);
	for(uint i = 0; i < children.size(); i++) {
		children[i]->PropagateChangedBranchVisible(bBranchVisible);
	}
}


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


