

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Base.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//====================================================================
// Auto-positioning (top-level components)
//====================================================================
enum eAutoPosition
{
	AutoPosition_NoChange = 0,
	AutoPosition_AlignNear, //left, top
	AutoPosition_AlignFar,  //right, bottom
	AutoPosition_Center,
	AutoPosition_KeepInsideWindow,

	AutoPosition_Count,
	AutoPosition_Force32Bits = 0x7FFFFFFF
};


//====================================================================
// GUI component.
// All child-components are owned by parent-component.
//====================================================================
class VW_COMMON_API Component : public VideoResourceUser, public MessageHandler
{
friend class gui::Manager;

public:
	virtual ~Component();

	//----------
	// Hierarchy
	//----------
	std::string  MakeHierarchyTrace(const std::string& levelSeparator = "/") const;

	void  DumpHierarchy(std::ostream& os, uint tabLevel,
						bool bPrintDepth = true, bool bPrintAttribs = true,
						uint depth = 0, bool bLimitDepth = false, uint maxDepth = 0) const;

	void  Attach(Component& c);
	void  AttachTo(Component& c) { c.Attach(*this); }
	void  Detach(Component& c);
	void  DetachFromParent() { if(parent) parent->Detach(*this); }
	void  DetachAllChildren();
	void  DeleteAllChildren();

	Component*        GetParent() { return parent; }
	const Component*  GetParent() const { return parent; }

	Component&        GetTopLevelAncestor(); //returns self if there is no parent
	const Component&  GetTopLevelAncestor() const { return const_cast<Component*>(this)->GetTopLevelAncestor(); }

	uint              GetNumChildren() const { return children.size(); }
	const Component&  GetChild(uint i) const { return *children[i]; }
	Component&        GetChild(uint i) { return *children[i]; }

	bool  IsTopLevel() const { return (parent == 0); }
	bool  IsLeaf() const { return children.empty(); }
	bool  IsChildOf(const Component& c) const { return (parent == &c); }
	bool  IsParentOf(const Component& c) const { return (c.parent == this); }
	bool  IsDescendantOf(const Component& c) const;
	bool  IsAncestorOf(const Component& c) const { return c.IsDescendantOf(*this); }
	uint  CalcDepth() const; //0-based

	// top-level components only
	void  SetLayer(uint layer);
	uint  GetLayer() const { return layer; }

	void  BringToFront();
	void  SendToBack();
	void  BringInFront(Component& c);
	void  SendBehind(Component& c);

	void  SetGhost(bool bGhost);
	bool  IsGhost() const { return ! bNotGhost; }

	void  SetOverlay(bool bOverlay);
	bool  IsOverlay() const { return bOverlay; }

	void  SetVisible(bool bVisible);
	bool  IsVisible() const { return bVisible; }

	// returns true if all components upwards to top-level-ancestor are visible
	bool  IsBranchVisible() const;

	void  PropagateVisibleToChildren(bool bVisible);
	void  PropagateVisibleToDescendants(bool bVisible);
	void  PropagateActiveToDescendants(bool bActive);

	//-------------------
	// Interaction states
	//-------------------
	bool  CanDisable() const { return bCanDisable; }
	void  SetEnabled(bool bEnabled);
	bool  IsEnabled() const { return bEnabled; }

	bool  CanFocus() const { return bCanFocus; }
	void  SetFocused(bool bFocused);
	bool  IsFocused() const { return bFocused; }

	void  SetActive(bool bActive);
	bool  IsActive() const { return bActive; }

	void  SetMouseOver(bool bMouseOver, bool bMouseOverDirect);
	bool  IsMouseOver() const { return bMouseOver; }
	bool  IsMouseOverDirect() const { return bMouseOverDirect; }

	bool  IsHot() const { return (bEnabled  &&  bMouseOverDirect); }

	void  SetFlags(uint flags);
	void  SetFlagsOn(uint f) { SetFlags(flags | f); }
	void  SetFlagsOff(uint f) { SetFlags(flags & ~f); }
	void  SetFlagsOnOff(uint f, bool bOn) { bOn ? SetFlagsOn(f) : SetFlagsOff(f); }

	uint  GetFlags() const { return flags; }
	uint  IsFlagOn(uint f) const { return (flags & f); }
	uint  IsFlagOff(uint f) const { return ! IsFlagOn(f); }

	// forwarders
	bool  IsMouseCaptureTarget();
	bool  RequestMouseCapture();
	void  ReleaseMouseCapture();

	bool  IsKeyboardFocusTarget();
	bool  RequestKeyboardFocus();
	bool  RequestKeyboardFocusToAncestor(bool bReleaseIfTopLevel = true);
	void  ReleaseKeyboardFocus();

	// top-level only
	void  ShowTopLevel(bool bShow, bool bAutoBringToFront = true);
	void  ActivateTopLevel(bool bAutoBringToFront = true);
	void  RunModal();
	void  EndModal();
	bool  IsRunningModal();

	gui::eGeometryUpdateEvent  GetActiveGeomUpdateEvent() const { return activeGeomUpdateEvent; }
	gui::eCommandArea          GetActiveGeomUpdateEventCommandArea() const { return activeGeomUpdateEventCommandArea; }

	bool  IsMoving() const { return (activeGeomUpdateEvent == gui::GeometryUpdateEvent_Moving); }
	bool  IsResizing() const { return (activeGeomUpdateEvent == gui::GeometryUpdateEvent_Resizing); }

	//------------------------------------------------------
	// Geometry.
	// Position is specified in parent's coordinate-system,
	// absolute-position is specified in window coordinates.
	//------------------------------------------------------
	void  SetAutoPositionOnWindowResize(eAutoPosition apx = AutoPosition_KeepInsideWindow,
										eAutoPosition apy = AutoPosition_KeepInsideWindow,
										float alignOffsetX = 0, float alignOffsetY = 0);

	void  GetAutoPositionOnWindowResize(eAutoPosition& apx, eAutoPosition& apy,
										float& alignOffsetX, float& alignOffsetY) const;

	const vec2&  GetAbsolutePosition() const { return absolutePosition; }

	void         SetPositionX(float x = 0);
	void         SetPositionY(float y = 0);
	void         SetPosition(float x = 0, float y = 0);
	void         SetPosition(const vec2& position = vec2::ZERO);
	const vec2&  GetPosition() const { return position; }

	void         SetSizeX(float x = 1);
	void         SetSizeY(float y = 1);
	void         SetSize(float x = 1, float y = 1);
	void         SetSize(const vec2& size = vec2::ONE);
	const vec2&  GetSize() const { return size; }

	void         SetMinSize(const vec2& minSize = vec2::ZERO);
	const vec2&  GetMinSize() const { return minSize; }

	void         SetMaxSize(const vec2& maxSize = vec2::INFINITY);
	const vec2&  GetMaxSize() const { return maxSize; }

	vec2   CalcMinMaxClampedSize(const vec2& size) const;
	vec2   CalcEndPosition() const { return (position + size); }
	float  CalcEndPositionX() const { return (position.x + size.x); }
	float  CalcEndPositionY() const { return (position.y + size.y); }

	vec2  GetParentPosition() const { return (parent ? parent->GetPosition() : vec2::ZERO); }
	vec2  GetParentSize() const;

	vec2  CalcRelativeMousePosition() const;
	vec2  CalcParentRelativeMousePosition(const vec2& relativeMousePos) const { return relativeMousePos + position; }

	void  ReportDirtyGeometry();
	void  UpdateAbsolutePositionCache();
	void  UpdateGeometry(bool bFromTopLevel = true);
	void  PropagateRecalcSize();
	void  PropagateRecalcLayout();
	void  ResetRecalcLayoutFlag(bool bLayoutDone = false) { bRecalcLayoutDone = bLayoutDone; }
	void  PropagateResetRecalcLayoutFlag(bool bLayoutDone = false);
	void  PropagateUpdateGeometryCache();
	void  ForceTopLevelComponentInsideWindow();

	// self hit-tests, test-point is specified in component's coordinate-system
	void               SetDefaultCommandArea(gui::eCommandArea ca = gui::CommandArea_Client) { defaultCommandArea = ca; }
	gui::eCommandArea  GetDefaultCommandArea() const { return defaultCommandArea; }

	bool               HitTest(const vec2& testPoint) const;
	gui::eCommandArea  HitTestCommandArea(const vec2& testPoint) const;

	void  EnableShaderHitTest(bool bEnable) { bEnableShaderHitTest = bEnable; }
	bool  IsEnabledShaderHitTest() const { return bEnableShaderHitTest; }

	static void __fastcall SetDefaultEnableShaderHitTest(bool bEnable = false) { s_bDefaultEnableShaderHitTest = bEnable; }
	static bool __fastcall GetDefaultEnableShaderHitTest() { return s_bDefaultEnableShaderHitTest; }

	//--------------------
	// Resources & drawing
	//--------------------
	void                SetClientAreaCursorName(const std::string& cursorName = "");
	const std::string&  GetClientAreaCursorName() const { return clientAreaCursorName; }

	void                SetShaderName(const std::string& s = "");
	const std::string&  GetShaderName() const { return shaderName; }

	//--------------------------------------------------
	// Help system. Pass empty string to disable tooltip
	//--------------------------------------------------
	void                SetToolTipText(const std::string& s = "") { toolTipText = s; }
	const std::string&  GetToolTipText() const { return toolTipText; }


protected:
	Component();

	//--------------------
	// Interaction control
	//--------------------
	void  SetCanDisable(bool _bCanDisable = false) { bCanDisable = _bCanDisable; }
	void  SetCanFocus(bool _bCanFocus = false) { bCanFocus = _bCanFocus; }

	//----------
	// Hierarchy
	//----------
	COMPONENT_PTR_VECTOR&        GetChildren()       { return children; }
	const COMPONENT_PTR_VECTOR&  GetChildren() const { return children; }

	virtual  void  OnAttachedChild(gui::Component& child) {}
	virtual  void  OnDetachedChild(gui::Component& child) {}

	virtual  void  OnAttachedToParent() {}
	virtual  void  OnDetachedFromParent() {}

	//---------
	// Geometry
	//---------
	virtual  void  RecalcSize() {}
	virtual  void  RecalcLayout() {}
	virtual  void  UpdateGeometryCache() {}
	virtual  void  OnWindowResize(uint width, uint height);

	virtual  bool               HitTestSelf(const vec2& testPoint) const;
	virtual  gui::eCommandArea  HitTestCommandAreaSelf(const vec2& testPoint) const;

	//--------------------
	// Resources & drawing
	//--------------------
	// Return true if shader is not required or <pShader>'s type is acceptable.
	// <pShader> is 0, if shader with matching name is not found in registry.
	virtual  bool  CheckShader(const gui::BaseShader* pShader) const { return true; }

	void                    SetBaseShader(const gui::BaseShader* shader = 0) { pBaseShader = shader; }
	const gui::BaseShader*  GetBaseShader() const { return pBaseShader; }

	// translate transform based on absolute-position is already pushed
	// by manager, so draw from (0, 0) to (sizeX, sizeY)
	virtual  void  DrawSelf(gui::Renderer& gr);

	// return false if required cursor is missing
	virtual  bool  CacheCursorHandles(gui::Manager& gm);
	virtual  void  PostCacheCursorHandles() {}
	virtual  void  OnCursorListChanged(gui::Manager& gm) {}

	void  RequestCursorRecache();

	// return false if required shader is missing or shader-type is inadequate
	virtual  bool  CacheShaderHandles(gui::Manager& gm);
	virtual  void  PostCacheShaderHandles();
	virtual  void  OnShaderListChanged(gui::Manager& gm) {}

	void  RequestShaderRecache();

	//-----------------
	// Message handling
	//-----------------
	bool  SendMessageToParent(Message& msg);
	bool  SendMessageToSiblings(Message& msg);
	bool  SendMessageToPreviousSibling(bool bAllowSendToSelf, Message& msg);
	bool  SendMessageToNextSibling(bool bAllowSendToSelf, Message& msg);
	bool  SendMessageToSelfAndSiblings(Message& msg);
	bool  SendMessageToChildren(Message& msg);
	bool  SendMessageToFirstChild(Message& msg);
	bool  SendMessageToLastChild(Message& msg);

	//--------------------------------------------------------
	// Default implementation of HandleMessage() maps messages
	// to event handlers. To handle additional messages or to
	// change default behavior, override this method.
	// Returns true if message is handled.
	//--------------------------------------------------------
	virtual  bool  HandleMessage(gui::Message& msg);

	//-----------------------------------------------------
	// Called by default implementation of HandleMessage()
	// if <msg.target> != this. Return true to stop further
	// propagation of message through hook-chain.
	//-----------------------------------------------------
	virtual  bool  HandleMessageAsHook(gui::Message& msg);

	//----------------------------------------------
	// Event-handlers called through HandleMessage()
	//----------------------------------------------
	virtual  bool  OnCommand(uint commandID);

	virtual  bool  OnGeometryUpdateByManager(bool& bConfirmEvent,
								gui::eGeometryUpdateEvent updateEvent,
								gui::eCommandArea commandArea, eMouseButton buttonCode,
								const vec2& newPosition, const vec2& newSize);

	virtual  bool  OnActivate(bool bActivateByMouse, eMouseButton buttonCode,
							const vec2& mousePosition, gui::Component* pClickedDescendant,
							bool& bAutoBringToFront, bool& bAutoActivate);

	virtual  bool  OnDeactivate(bool bDeactivateByMouse) { return false; }

	virtual  bool  OnMouseCapture(bool bCaptureBegin) { return false; }
	virtual  bool  OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos) { return false; }
	virtual  bool  OnMouseQueryCursor(const gui::Cursor* &pCursor, gui::eCommandArea commandArea, const vec2& mousePos);
	virtual  bool  OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos) { return false; }
	virtual  bool  OnMouseWheel(int delta, const vec2& mousePos) { return false; }
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos) { return false; }

	//-------------------------------------------------------------------
	// <ch> is 0 if:
	// 1) <bNeedCharacterData> was set to false in OnKeyboardQueryFocus()
	// 2) key-to-character mapping failed
	//-------------------------------------------------------------------
	virtual  bool  OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData) { return false; }
	virtual  bool  OnKeyboardFocus(bool bGotFocus) { return false; }
	virtual  bool  OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch) { return false; }

	__forceinline  void  DefaultProcessQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData)
	{
		bWantFocus = ! bFocusByMouse  ||  (buttonCode == MouseButton_Left);
		bNeedCharacterData = false;
	}

	virtual  bool  OnQueryToolTipParams(gui::ToolTipParams& ttp);

	virtual  bool  OnMsgShow(bool bShow);
	virtual  bool  OnMsgShowChildren(bool bShow);
	virtual  bool  OnMsgShowDescendants(bool bShow);
	virtual  bool  OnMsgEnable(bool bEnable);
	virtual  bool  OnMsgFlags(uint flags);
	virtual  bool  OnMsgFlagsOnOff(uint flags, bool bOn);
	virtual  bool  OnMsgOnOff(bool bOn) { return false; }
	virtual  bool  OnMsgSelect(bool bSelect, uint groupID) { return false; }
	virtual  bool  OnMsgScroll(bool bForward, gui::eScrollDistance distance, int offset) { return false; }
	virtual  bool  OnMsgPreCommandExecute(uint commandID); //default propagates to parent
	virtual  bool  OnMsgSendCommand() { return false; }

	//-----------------------------------------------------------------------
	// Event-handlers called through HandleMessageAsHook().
	// Return true to stop further propagation of message through hook-chain.
	//-----------------------------------------------------------------------
	virtual  bool  OnMouseWheel(gui::MessageHandler* target, int delta, const vec2& mousePos) { return false; }
	virtual  bool  OnMouseButton(gui::MessageHandler* target, bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos) { return false; }

	virtual  bool  OnKeyboardQueryFocus(gui::MessageHandler* target, bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData) { return false; }
	virtual  bool  OnKeyboardFocus(gui::MessageHandler* target, bool bGotFocus) { return false; }
	virtual  bool  OnKeyboardKey(gui::MessageHandler* target, bool bDown, eKey keyCode, uint repeatCount, char ch) { return false; }

	//------------------------------------------
	// Event-handlers called directly by manager
	//------------------------------------------
	virtual  void  OnTimeDelta(double secs) {}

	//----------------------------------------------------
	// Event-handlers called directly by Component methods
	//----------------------------------------------------
	virtual  void  OnChangedBranchVisible(bool bBranchVisible) {}
	virtual  void  OnChangedVisible() {}
	virtual  void  OnChangedEnabled() {}
	virtual  void  OnChangedFocused() {}
	virtual  void  OnChangedActive() {}
	virtual  void  OnChangedFlags() {}

	//---------------------------
	// Framework internal helpers
	//---------------------------
	void  SetActiveGeomUpdateEvent(gui::eGeometryUpdateEvent ev = gui::GeometryUpdateEvent_Null,
									gui::eCommandArea ca = gui::CommandArea_Client);

	void  PropagateChangedBranchVisible(bool bBranchVisible);


private:
	//------------
	// Properties.
	//------------
	static  bool  s_bDefaultEnableShaderHitTest;

	uint                  layer;
	Component*            parent;
	COMPONENT_PTR_VECTOR  children;

	vec2  absolutePosition;
	vec2  position;
	vec2  size;
	vec2  minSize;
	vec2  maxSize;

	bool  bEnableShaderHitTest;
	bool  bNotGhost;
	bool  bRecalcLayoutDone;
	bool  bCanDisable;
	bool  bCanFocus;
	bool  bOverlay;
	bool  bVisible;
	bool  bEnabled;
	bool  bFocused;
	bool  bActive;
	bool  bMouseOver;
	bool  bMouseOverDirect;
	uint  flags;

	gui::eGeometryUpdateEvent  activeGeomUpdateEvent;
	gui::eCommandArea          activeGeomUpdateEventCommandArea;

	eAutoPosition  autoPositionOnWindowResizeX;
	eAutoPosition  autoPositionOnWindowResizeY;
	float          autoPositionAlignOffsetX;
	float          autoPositionAlignOffsetY;

	gui::eCommandArea       defaultCommandArea;
	std::string             clientAreaCursorName;
	const Cursor*           pClientAreaCursor;
	std::string             shaderName;
	const gui::BaseShader*  pBaseShader;

	std::string  toolTipText;
};


//====================================================================
template<typename TTargetComponent>
struct ComponentPtr
{
	TTargetComponent* ptr;

	__forceinline operator TTargetComponent*() { return ptr; }
	__forceinline TTargetComponent* operator->() { return ptr; }
	__forceinline TTargetComponent& operator*() { return *ptr; }

	__forceinline  ComponentPtr(gui::Component& component)
	{
		ptr = dynamic_cast<TTargetComponent*>(&component);
	}
	__forceinline  ComponentPtr(gui::Component* component = 0)
	{
		ptr = dynamic_cast<TTargetComponent*>(component);
	}
};


//====================================================================
template<typename TTargetComponent>
struct ConstComponentPtr
{
	const TTargetComponent* ptr;

	__forceinline operator const TTargetComponent*() { return ptr; }
	__forceinline const TTargetComponent* operator->() { return ptr; }
	__forceinline const TTargetComponent& operator*() { return *ptr; }

	__forceinline  ConstComponentPtr(const gui::Component& component)
	{
		ptr = dynamic_cast<const TTargetComponent*>(&component);
	}
	__forceinline  ConstComponentPtr(const gui::Component* component = 0)
	{
		ptr = dynamic_cast<const TTargetComponent*>(component);
	}
};


//====================================================================
template<typename TTargetComponent>
__forceinline TTargetComponent&  ComponentCast(gui::Component& c)
{
	return static_cast<TTargetComponent&>(c);
}


//====================================================================
template<typename TTargetComponent>
__forceinline const TTargetComponent&  ComponentCast(const gui::Component& c)
{
	return static_cast<const TTargetComponent&>(c);
}


//====================================================================
#define  VW_GUI_COMPONENT_DECLARE_SHADER(TShader) \
	public: class TShader : public gui::BaseShader { \
	protected: TShader(const std::string& name = "") : BaseShader(name) {} \
	};


//====================================================================
#define  VW_GUI_COMPONENT_DECLARE_SHADER_DLL(DllDeclSpec, TShader) \
	public: class DllDeclSpec TShader : public gui::BaseShader { \
	protected: TShader(const std::string& name = "") : BaseShader(name) {} \
	};


//====================================================================
#define  VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(TShader) \
	protected: virtual bool CheckShader(const gui::BaseShader* pShader) const \
	{ return (pShader  &&  dynamic_cast<const TShader*>(pShader)); }


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


