

//--------------------------------------------------------------------
#include "GUI\GUI_Cursor.h"
#include "GUI\GUI_Manager.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//====================================================================
vec2     Cursor::s_globalSize  = vec2(16, 16);
color4b  Cursor::s_globalColor = color4b::WHITE;


//====================================================================
void __fastcall Cursor::SetGlobalSize(const vec2& _size)
{
	s_globalSize = ComponentMax(_size, vec2(1, 1));
	Manager::Instance().NotifyGlobalCursorParamsChanged();
}


//====================================================================
void __fastcall Cursor::SetGlobalColor(color4b c)
{
	s_globalColor = c;
	Manager::Instance().NotifyGlobalCursorParamsChanged();
}


//====================================================================
Cursor::Cursor(const std::string& _name, const vec2& _hotSpot,
				const vec2& _privateSize, color4b _privateColor)
	:
	cursorName(_name)
{
	hotSpot = ComponentClamp(_hotSpot, vec2::ZERO, vec2(1, 1));
	privateSize = ComponentMax(_privateSize, vec2(1, 1));
	privateColor = _privateColor;

	pActiveSize = &s_globalSize;
	pActiveColor = &s_globalColor;

	gui::Manager::Instance().NotifyObjectCreateDelete(*this, true);
}


//====================================================================
Cursor::~Cursor()
{
	gui::Manager::Instance().NotifyObjectCreateDelete(*this, false);
}


//====================================================================
void  Cursor::SetName(const std::string& s)
{
	cursorName = s;
	gui::Manager::Instance().NotifyNameChanged(*this);
}


//====================================================================
void  Cursor::SetHotSpot(const vec2& _hotSpot)
{
	hotSpot = ComponentClamp(_hotSpot, vec2::ZERO, vec2(1, 1));
	UpdateGeometryCache();
}


//====================================================================
void  Cursor::SetPrivateSize(bool bUsePrivateSize, const vec2& _privateSize)
{
	privateSize = ComponentMax(_privateSize, vec2(1, 1));
	pActiveSize = bUsePrivateSize ? &privateSize : &s_globalSize;
	UpdateGeometryCache();
}


//====================================================================
void  Cursor::SetPrivateColor(bool bUsePrivateColor, color4b c)
{
	privateColor = c;
	pActiveColor = bUsePrivateColor ? &privateColor : &s_globalColor;
	UpdateColorCache();
}


//====================================================================
void  Cursor::DrawSelf(gui::Renderer& gr)
{
	const float offset = 0.7f;
	const color4b c = GetActiveColor();
	const vec2& s = GetActiveSize();

	VertexBufferC vb(3);

	vb[0].position = vec3(0, 0, 0);
	vb[0].color = c;
	vb[1].position = vec3(s.x * offset, s.y, 0);
	vb[1].color = c;
	vb[2].position = vec3(s.x, s.y * offset, 0);
	vb[2].color = c;

	gr.DrawPrimitive(vb, PrimitiveType_TriangleList, 1, 0);
}


//====================================================================
ImageCursor::ImageCursor(const std::string& _name, const gui::TileParams& _tileParams,
				const vec2& _hotSpot, const vec2& _privateSize, color4b _privateColor)
	:
	Cursor(_name, _hotSpot, _privateSize, _privateColor)
{
	pTileTexture = 0;

	tileParams = _tileParams;

	UpdateGeometryCache();
	UpdateColorCache();
}


//====================================================================
void  ImageCursor::SetTileParams(const gui::TileParams& _tileParams)
{
	tileParams = _tileParams;
	RequestTileTextureRecache();
}


//====================================================================
void  ImageCursor::UpdateColorCache()
{
	rDrawArea.SetColor(GetActiveColor());
}


//====================================================================
void  ImageCursor::UpdateGeometryCache()
{
	rDrawArea.Set(0, 0, GetActiveSize().x, GetActiveSize().y);
}


//====================================================================
void  ImageCursor::DrawSelf(gui::Renderer& gr)
{
	gr.DrawRectangle(*pTileTexture, rDrawArea);
}


//====================================================================
bool  ImageCursor::CacheTileTextureHandles(Manager& gm)
{
	pTileTexture = gm.GetTileTexture(tileParams.texName);
	return (pTileTexture != 0);
}


//====================================================================
void  ImageCursor::PostCacheTileTextureHandles()
{
	if(pTileTexture) {
		pTileTexture->GetTileCoords(tileParams.texRow, tileParams.texCol, rDrawArea);
		rDrawArea.SetTexTransform(tileParams.texTransform);
	}
}


//====================================================================
void  ImageCursor::InvalidateTileTextureHandles()
{
	pTileTexture = 0;
}


//********************************************************************
namespace std_cursors {


//====================================================================
VW_DEFINE_CONSTANT( char*, CursorName_Void,     "std/void" );
VW_DEFINE_CONSTANT( char*, CursorName_Client,   "std/client" );
VW_DEFINE_CONSTANT( char*, CursorName_Move,     "std/move" );
VW_DEFINE_CONSTANT( char*, CursorName_ResizeN,  "std/resizeN" );
VW_DEFINE_CONSTANT( char*, CursorName_ResizeS,  "std/resizeS" );
VW_DEFINE_CONSTANT( char*, CursorName_ResizeW,  "std/resizeW" );
VW_DEFINE_CONSTANT( char*, CursorName_ResizeE,  "std/resizeE" );
VW_DEFINE_CONSTANT( char*, CursorName_ResizeNW, "std/resizeNW" );
VW_DEFINE_CONSTANT( char*, CursorName_ResizeNE, "std/resizeNE" );
VW_DEFINE_CONSTANT( char*, CursorName_ResizeSW, "std/resizeSW" );
VW_DEFINE_CONSTANT( char*, CursorName_ResizeSE, "std/resizeSE" );
VW_DEFINE_CONSTANT( char*, CursorName_Stop,     "std/stop" );


//********************************************************************
}//namespace std_cursors


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


