

//--------------------------------------------------------------------
#include "GUI\GUI_Layouters.h"
#include "BaseTools\STLExtensions.h"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;
using namespace vertexware::gui;
using namespace vertexware::gui::layouters;


//====================================================================
struct Predicate_FindCellByAddress
{
	bool operator()(const Cell& c) { return (&c == target); }
	Predicate_FindCellByAddress(const Cell* _target) { target = _target; }
	Predicate_FindCellByAddress(const Cell& _target) { target = &_target; }
	const Cell* target;
};


//====================================================================
struct Predicate_FindCellByTarget
{
	bool operator()(const Cell& c) { return (c.target == target); }
	Predicate_FindCellByTarget(const Component* _target) { target = _target; }
	const Component* target;
};


//====================================================================
struct Predicate_FindRowByAddress
{
	bool operator()(const Row& r) { return (&r == target); }
	Predicate_FindRowByAddress(const Row* _target) { target = _target; }
	Predicate_FindRowByAddress(const Row& _target) { target = &_target; }
	const Row* target;
};


//====================================================================
struct Predicate_FindRowByTarget
{
	bool operator()(const Row& r) { return (r.target == target); }
	Predicate_FindRowByTarget(const Component* _target) { target = _target; }
	const Component* target;
};


//====================================================================
__forceinline float __fastcall GetTargetSizeX(gui::Component* c)
{
	if(c  &&  c->IsVisible()) return c->GetSize().x;
	return 0;
}


//====================================================================
__forceinline float __fastcall GetTargetSizeY(gui::Component* c)
{
	if(c  &&  c->IsVisible()) return c->GetSize().y;
	return 0;
}


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//********************************************************************
namespace layouters {


//====================================================================
Cell::Cell(Component* _target,
			eSizeFlag _sizeFlagX, float _sizeX,
			eAlignFlag _alignX, gui::Component* _alignTargetX,
			eSizeFlag _sizeFlagY, float _sizeY,
			eAlignFlag _alignY, gui::Component* _alignTargetY)
{
	target = _target;
	sizeFlagX = _sizeFlagX;
	sizeFlagY = _sizeFlagY;
	sizeX = _sizeX;
	sizeY = _sizeY;
	alignX = _alignX;
	alignY = _alignY;
	alignTargetX = _alignTargetX;
	alignTargetY = _alignTargetY;

	outputOffsetX = 0;
	outputSizeX = 0;
}


//====================================================================
Row::Row(Component* _target, eSizeFlag _sizeFlagY, float _sizeY)
{
	target = _target;
	sizeFlagY = _sizeFlagY;
	sizeY = _sizeY;

	outputOffsetY = 0;
	outputSizeY = 0;
}


//====================================================================
Cell*  Row::FindCell(Component& target)
{
	CONTAINER_ITER it = std::find_if(cells.begin(), cells.end(), Predicate_FindCellByTarget(&target));
	if(it != cells.end()) return &(*it);
	return 0;
}


//====================================================================
Row&  Row::Clear()
{
	cells.clear();
	return *this;
}


//====================================================================
Row&  Row::Remove(const Cell& c)
{
	EraseFromContainer_If(cells, Predicate_FindCellByAddress(&c));
	return *this;
}


//====================================================================
Row&  Row::Remove(uint iRemove)
{
	if(iRemove < cells.size()) {
		Remove(cells[iRemove]);
	}
	return *this;
}


//====================================================================
Row&  Row::Append(const Cell& c)
{
	cells.push_back(c);
	return *this;
}


//====================================================================
Row&  Row::Prepend(const Cell& c)
{
	cells.insert(cells.begin(), c);
	return *this;
}


//====================================================================
Row&  Row::InsertBefore(const Cell& c, const Cell& before)
{
	InsertBefore_If<Predicate_FindCellByAddress>(cells, c, before);
	return *this;
}


//====================================================================
Row&  Row::InsertBefore(const Cell& c, uint iBefore)
{
	if(iBefore < cells.size()) {
		InsertBefore(c, cells[iBefore]);
	}
	return *this;
}


//====================================================================
Row&  Row::InsertAfter(const Cell& c, const Cell& after)
{
	InsertAfter_If<Predicate_FindCellByAddress>(cells, c, after);
	return *this;
}


//====================================================================
Row&  Row::InsertAfter(const Cell& c, uint iAfter)
{
	if(iAfter < cells.size()) {
		InsertAfter(c, cells[iAfter]);
	}
	return *this;
}


//====================================================================
Row&  Row::Append(const Row& r)
{
	for(uint i = 0; i < r.NumCells(); i++) {
		Append(r[i]);
	}
	return *this;
}


//====================================================================
Row&  Row::Prepend(const Row& r)
{
	for(int i = r.NumCells() - 1; i >= 0; i--) {
		Prepend(r[i]);
	}
	return *this;
}


//====================================================================
Row&  Row::Target(Component* target, eSizeFlag sizeFlagX, float sizeX,
					eAlignFlag alignX, gui::Component* alignTargetX,
					eAlignFlag alignY, gui::Component* alignTargetY)
{
	Append(Cell(target, sizeFlagX, sizeX, alignX, alignTargetX, SF_Target, 1, alignY, alignTargetY));
	return *this;
}


//====================================================================
Row&  Row::Spacer(float sizeX, eSizeFlag sizeFlagX)
{
	Append(Cell(0, sizeFlagX, sizeX));
	return *this;
}


//====================================================================
Row&  Row::AutoSpacer()
{
	Append(Cell(0, SF_Auto));
	return *this;
}


//====================================================================
Cell&  Row::GetCell(uint i)
{
	if(i < cells.size()) return cells[i];
	AutoSpacer();
	return cells.back();
}


//====================================================================
Cell&  Row::LastCell()
{
	if(cells.empty()) {
		AutoSpacer();
	}
	return cells.back();
}


//====================================================================
void  Row::ProcessX(float totalWidth)
{
	uint numCells = cells.size();
	if(! numCells) return;

	// accumulate data
	float  remainingWidth = totalWidth;
	float  totalRelative = 0;
	float  numAuto = 0;

	{for(uint i = 0; i < numCells; i++) {
		Cell& c = cells[i];

		switch(c.sizeFlagX) {
			case SF_Target: {
				c.outputSizeX = GetTargetSizeX(c.target);
				remainingWidth -= c.outputSizeX;
				break;
			}
			case SF_TargetMul: {
				c.outputSizeX = GetTargetSizeX(c.target) * c.sizeX;
				remainingWidth -= c.outputSizeX;
				break;
			}
			case SF_TargetAdd: {
				c.outputSizeX = GetTargetSizeX(c.target) + c.sizeX;
				remainingWidth -= c.outputSizeX;
				break;
			}
			case SF_Fixed: {
				c.outputSizeX = c.sizeX;
				remainingWidth -= c.sizeX;
				break;
			}
			case SF_Variable: {
				totalRelative += c.sizeX;
				break;
			}
			case SF_Auto: {
				numAuto += 1;
				break;
			}
		}
	}}

	float autoWidth = ClampedLow<float>(remainingWidth * (1 - totalRelative), 0);
	autoWidth = (numAuto > 0) ? (autoWidth / numAuto) : 0;

	// output results
	float offsetH = 0;

	{for(uint i = 0; i < numCells; i++) {
		Cell& c = cells[i];

		c.outputOffsetX = offsetH;

		switch(c.sizeFlagX) {
			case SF_Variable: {
				c.outputSizeX = remainingWidth * c.sizeX;
				break;
			}
			case SF_Auto: {
				c.outputSizeX = autoWidth;
				break;
			}
		}
		offsetH += c.outputSizeX;
	}}
}


//====================================================================
RowLayouter::GeomParams::GeomParams()
{
	preferredSize = vec2(200, 200);
	bAutoFitToParentOnRecalcLayout = false;
	autoFitPaddingTL = vec2(20, 20);
	autoFitPaddingBR = autoFitPaddingTL;
}


//====================================================================
RowLayouter::RowLayouter(gui::Component* _parent)
{
	SetGhost(true);
	SetSize(200, 200);
	SetGeomParams();

	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  RowLayouter::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
Row*  RowLayouter::FindRow(Component& target)
{
	CONTAINER_ITER it = std::find_if(rows.begin(), rows.end(), Predicate_FindRowByTarget(&target));
	if(it != rows.end()) return &(*it);
	return 0;
}


//====================================================================
RowLayouter&  RowLayouter::Clear()
{
	rows.clear();
	return *this;
}


//====================================================================
RowLayouter&  RowLayouter::Remove(const Row& r)
{
	EraseFromContainer_If(rows, Predicate_FindRowByAddress(&r));
	return *this;
}


//====================================================================
RowLayouter&  RowLayouter::Append(const Row& r)
{
	rows.push_back(r);
	return *this;
}


//====================================================================
RowLayouter&  RowLayouter::Prepend(const Row& r)
{
	rows.insert(rows.begin(), r);
	return *this;
}


//====================================================================
RowLayouter&  RowLayouter::InsertBefore(const Row& r, const Row& before)
{
	InsertBefore_If<Predicate_FindRowByAddress>(rows, r, before);
	return *this;
}


//====================================================================
RowLayouter&  RowLayouter::InsertAfter(const Row& r, const Row& after)
{
	InsertAfter_If<Predicate_FindRowByAddress>(rows, r, after);
	return *this;
}


//====================================================================
Row&  RowLayouter::TargetHeightRow(gui::Component* target, eSizeFlag sizeFlagY, float sizeY)
{
	Append(Row(target, sizeFlagY, sizeY));
	return rows.back();
}


//====================================================================
Row&  RowLayouter::FixedHeightRow(float sizeY)
{
	Append(Row(0, SF_Fixed, sizeY));
	return rows.back();
}


//====================================================================
Row&  RowLayouter::VariableHeightRow(float sizeY)
{
	Append(Row(0, SF_Variable, sizeY));
	return rows.back();
}


//====================================================================
Row&  RowLayouter::AutoHeightRow()
{
	Append(Row());
	return rows.back();
}


//====================================================================
Row&  RowLayouter::GetRow(uint i)
{
	if(i < rows.size()) return rows[i];
	AutoHeightRow();
	return rows.back();
}


//====================================================================
Row&  RowLayouter::LastRow()
{
	if(rows.empty()) {
		AutoHeightRow();
	}
	return rows.back();
}


//====================================================================
void  RowLayouter::Process(float totalWidth, float totalHeight)
{
	// compute vertical output-data
	ProcessY(totalHeight);

	// compute horizontal output-data
	for(uint iV = 0; iV < rows.size(); iV++) {
		rows[iV].ProcessX(totalWidth);
	}
}


//====================================================================
void  RowLayouter::ProcessY(float totalHeight)
{
	uint numRows = rows.size();
	if(! numRows) return;

	// accumulate data
	float  remainingHeight = totalHeight;
	float  totalRelative = 0;
	float  numAuto = 0;

	{for(uint i = 0; i < numRows; i++) {
		Row& r = rows[i];

		switch(r.sizeFlagY) {
			case SF_Target: {
				r.outputSizeY = GetTargetSizeY(r.target);
				remainingHeight -= r.outputSizeY;
				break;
			}
			case SF_TargetMul: {
				r.outputSizeY = GetTargetSizeY(r.target) * r.sizeY;
				remainingHeight -= r.outputSizeY;
				break;
			}
			case SF_TargetAdd: {
				r.outputSizeY = GetTargetSizeY(r.target) + r.sizeY;
				remainingHeight -= r.outputSizeY;
				break;
			}
			case SF_Fixed: {
				r.outputSizeY = r.sizeY;
				remainingHeight -= r.sizeY;
				break;
			}
			case SF_Variable: {
				totalRelative += r.sizeY;
				break;
			}
			case SF_Auto: {
				numAuto += 1;
				break;
			}
		}
	}}

	float autoHeight = ClampedLow<float>(remainingHeight * (1 - totalRelative), 0);
	autoHeight = (numAuto > 0) ? (autoHeight / numAuto) : 0;

	// output results
	float offsetV = 0;

	{for(uint i = 0; i < numRows; i++) {
		Row& r = rows[i];

		r.outputOffsetY = offsetV;

		switch(r.sizeFlagY) {
			case SF_Variable: {
				r.outputSizeY = remainingHeight * r.sizeY;
				break;
			}
			case SF_Auto: {
				r.outputSizeY = autoHeight;
				break;
			}
		}
		offsetV += r.outputSizeY;
	}}
}


//====================================================================
void  RowLayouter::RecalcSize()
{
	SetSize(geomParams.preferredSize);
}


//====================================================================
void  RowLayouter::RecalcLayout()
{
	// auto resize
	if(geomParams.bAutoFitToParentOnRecalcLayout) {
		Component* c = GetParent();
		if(c) {
			SetSize(c->GetSize() - geomParams.autoFitPaddingTL - geomParams.autoFitPaddingBR);
			SetPosition(geomParams.autoFitPaddingTL);
		}
	}

	// compute output-data
	Process(GetSize().x, GetSize().y);

	// dispatch cell sizes to targets
	{for(uint iV = 0; iV < rows.size(); iV++) {
		Row& r = rows[iV];

		bool bRowHeightIsTargetIndependent = true;
		switch(r.sizeFlagY) {
			case SF_Target:
			case SF_TargetMul:
			case SF_TargetAdd:
				bRowHeightIsTargetIndependent = false;
				break;
		}

		for(uint iH = 0; iH < r.NumCells(); iH++) {
			Cell& c = r[iH];

			gui::Component* target = c.target;
			if(target  &&  target->IsVisible()) {

				// compute & apply size
				vec2 size(c.outputSizeX, target->GetSize().y);

				if(c.sizeFlagY == SF_Variable) {

					// row-height must not depend on this target,
					// otherwise resizing artifacts may occur
					if(bRowHeightIsTargetIndependent  ||  r.target != target) {
						size.y = r.outputSizeY * c.sizeY;
					}
				}
				target->SetSize(size);

				// compute & apply offset
				vec2 offset;

				switch(c.alignX) {
					case AF_Near:
						offset.x = c.outputOffsetX;
						break;
					case AF_Far:
						offset.x = c.outputOffsetX + c.outputSizeX - target->GetSize().x;
						break;
					case AF_Center:
						offset.x = c.outputOffsetX + (c.outputSizeX - target->GetSize().x) * 0.5f;
						break;
					case AF_Target:
						offset.x = c.alignTargetX ? c.alignTargetX->GetPosition().x : c.outputOffsetX;
						break;
				}
				switch(c.alignY) {
					case AF_Near:
						offset.y = r.outputOffsetY;
						break;
					case AF_Far:
						offset.y = r.outputOffsetY + r.outputSizeY - target->GetSize().y;
						break;
					case AF_Center:
						offset.y = r.outputOffsetY + (r.outputSizeY - target->GetSize().y) * 0.5f;
						break;
					case AF_Target:
						offset.y = c.alignTargetY ? c.alignTargetY->GetPosition().y : r.outputOffsetY;
						break;
				}
				target->SetPosition(offset);
			}
		}
	}}
}


//====================================================================
ArrayLayouter::GeomParams::GeomParams()
{
	bVertical = true;
	bResizeControlsToMax = false;
	autoSpacing = DefaultAutoSpacing;
	autoSpacingHead = DefaultAutoSpacingHead;
	autoSpacingTail = DefaultAutoSpacingTail;
}


//====================================================================
ArrayLayouter::ArrayLayouter(gui::Component* _parent)
{
	SetGhost(true);
	SetSize(200, 300);
	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  ArrayLayouter::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  ArrayLayouter::ClearSpacers()
{
	spacers.clear();
	ReportDirtyGeometry();
}


//====================================================================
void  ArrayLayouter::AddSpacer(const Spacer& s)
{
	UniqueContainerPushBack(spacers, s);
	ReportDirtyGeometry();
}


//====================================================================
void  ArrayLayouter::RemoveSpacer(const gui::Component* pBefore)
{
	EraseFromContainer(spacers, Spacer(pBefore));
	ReportDirtyGeometry();
}


//====================================================================
void  ArrayLayouter::Process()
{
	uint numCh = GetNumChildren();
	uint lastCh = numCh - 1;

	// accumulate maximum sizes
	vec2 maxSize = vec2::ZERO;
	for(uint i = 0; i < numCh; i++) {
		gui::Component& c = GetChild(i);
		maxSize = ComponentMax(maxSize, c.GetSize());
	}
	vec2 pos = vec2::ZERO;

	// process as vertical array
	if(geomParams.bVertical) {
		if(geomParams.autoSpacingHead) pos.y += geomParams.autoSpacingHead;

		for(uint i = 0; i < numCh; i++) {
			gui::Component& c = GetChild(i);

			Spacer* spacer = FindSpacer(c);
			if(spacer) pos.y += spacer->length;

			c.SetPositionY(pos.y);

			if(geomParams.bResizeControlsToMax) {
				c.SetSizeX(maxSize.x);
				c.SetPositionX(0);
			}
			else {
				c.SetPositionX((maxSize.x - c.GetSize().x) * 0.5f);
			}
			pos.y += c.GetSize().y;
			if(i != lastCh) pos.y += geomParams.autoSpacing;
		}
		if(geomParams.autoSpacingTail) pos.y += geomParams.autoSpacingTail;

		SetSize(maxSize.x, pos.y);
	}

	// process as horizontal array
	else {
		if(geomParams.autoSpacingHead) pos.x += geomParams.autoSpacingHead;

		for(uint i = 0; i < numCh; i++) {
			gui::Component& c = GetChild(i);

			Spacer* spacer = FindSpacer(c);
			if(spacer) pos.x += spacer->length;

			c.SetPositionX(pos.x);

			if(geomParams.bResizeControlsToMax) {
				c.SetSizeY(maxSize.y);
				c.SetPositionY(0);
			}
			else {
				c.SetPositionY((maxSize.y - c.GetSize().y) * 0.5f);
			}
			pos.x += c.GetSize().x;
			if(i != lastCh) pos.x += geomParams.autoSpacing;
		}
		if(geomParams.autoSpacingTail) pos.x += geomParams.autoSpacingTail;

		SetSize(pos.x, maxSize.y);
	}
}


//====================================================================
void  ArrayLayouter::RecalcLayout()
{
	Process();
}


//====================================================================
Spacer*  ArrayLayouter::FindSpacer(const gui::Component& before)
{
	for(uint i = 0; i < spacers.size(); i++) {
		if(spacers[i].pBefore == &before) return &spacers[i];
	}
	return 0;
}


//********************************************************************
}//namespace layouters


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


