

//--------------------------------------------------------------------
#include "GUI\GUI_Manager.h"
#include "GUI\GUI_Component.h"
#include "GUI\GUI_Cursor.h"
#include "GUI\GUI_Shader.h"
#include "BaseTools\STLExtensions.h"
#include "BaseTools\String.h"
#include "BaseTools\Stream.h"
#include "MathTools\TransformTools.h"
#include "Framework\Application.h"
#include "Framework\Kernel.h"
#include "Framework\KernelMacros.h"
#include <stack>
#include <algorithm>
#include <iterator>


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;
using namespace vertexware::gui;


//====================================================================
struct LayerSortPredicate
{
	bool operator()(Component* lhs, Component* rhs)
	{
		return (lhs->GetLayer() < rhs->GetLayer());
	}
};


//====================================================================
// struct TProcessor {
//     // return false to discard <c> and it's children
//     bool operator()(Component& c);
// };
//====================================================================
template<typename TProcessor>
void __fastcall ConstructBreadthFirstQueue(Component& topLevel,
										COMPONENT_PTR_VECTOR& out,
										TProcessor& processor)
{
	if(! processor(topLevel)) return;
	out.push_back(&topLevel);

	uint  endIndex = out.size();
	uint  firstIndex = endIndex - 1;

	while(firstIndex != endIndex) {

		for(uint iParent = firstIndex; iParent < endIndex; iParent++) {
			Component& parent = *out[iParent];

			for(uint iChild = 0; iChild < parent.GetNumChildren(); iChild++) {
				Component& child = parent.GetChild(iChild);

				if(processor(child)) {
					out.push_back(&child);
				}
			}
		}

		firstIndex = endIndex;
		endIndex = out.size();
	}
}


//====================================================================
template<typename TProcessor>
void __fastcall ConstructBreadthFirstQueue(
									COMPONENT_PTR_VECTOR& components,
									bool bReverse,
									COMPONENT_PTR_VECTOR& out,
									TProcessor& processor)
{
	out.clear();
	if(components.empty()) return;

	if(bReverse) {
		for(int i = components.size() - 1; i >= 0; i--) {
			ConstructBreadthFirstQueue(*components[i], out, processor);
		}
	}
	else {
		for(uint i = 0; i < components.size(); i++) {
			ConstructBreadthFirstQueue(*components[i], out, processor);
		}
	}
}


//====================================================================
// struct TProcessor {
//     // return false to discard <c> and it's children
//     bool operator()(Component& c);
// };
//====================================================================
template<typename TProcessor>
void __fastcall ConstructDepthFirstQueue(
								Component& topLevel,
								COMPONENT_PTR_VECTOR& out,
								Manager::DEPTHFIRSTPATHMARKER_VECTOR& tempPathMarkers,
								TProcessor& processor)
{
	tempPathMarkers.clear();
	tempPathMarkers.push_back(Manager::DepthFirstPathMarker(&topLevel));

	while(tempPathMarkers.size()) {
		Manager::DepthFirstPathMarker& marker = tempPathMarkers.back();

		if(marker.bUnprocessed) {
			if(processor(*marker.component)) {
				// accepted
				out.push_back(marker.component);
				marker.bUnprocessed = false;
			}
			else {
				// discarded
				tempPathMarkers.pop_back();
				continue;
			}
		}
		if(marker.iNextChild < marker.component->GetNumChildren()) {
			tempPathMarkers.push_back(
				Manager::DepthFirstPathMarker(
					&marker.component->GetChild(marker.iNextChild++)));
		}
		else {
			tempPathMarkers.pop_back();
		}
	}
}


//====================================================================
template<typename TProcessor>
void __fastcall ConstructDepthFirstQueue(
								COMPONENT_PTR_VECTOR& components,
								bool bReverse,
								COMPONENT_PTR_VECTOR& out,
								Manager::DEPTHFIRSTPATHMARKER_VECTOR& tempPathMarkers,
								TProcessor& processor)
{
	out.clear();
	if(components.empty()) return;

	if(bReverse) {
		for(int i = components.size() - 1; i >= 0; i--) {
			ConstructDepthFirstQueue(*components[i], out, tempPathMarkers, processor);
		}
	}
	else {
		for(uint i = 0; i < components.size(); i++) {
			ConstructDepthFirstQueue(*components[i], out, tempPathMarkers, processor);
		}
	}
}


//====================================================================
struct ComponentProcessor_All
{
	__forceinline  bool  operator()(Component& c)
	{ return true; }
};


//====================================================================
struct ComponentProcessor_VisibleOnly
{
	__forceinline  bool  operator()(Component& c)
	{ return c.IsVisible(); }
};


//====================================================================
struct ComponentProcessor_VisibleActiveOnly
{
	__forceinline  bool  operator()(Component& c)
	{ return c.IsVisible()  &&  c.IsActive(); }
};


//====================================================================
struct ComponentProcessor_VisibleOnly_ExtractOverlays
{
	COMPONENT_PTR_VECTOR&  outOverlays;
	Component*             pTopLevel;


	ComponentProcessor_VisibleOnly_ExtractOverlays(COMPONENT_PTR_VECTOR& _outOverlays)
		: outOverlays(_outOverlays), pTopLevel(0)
	{ }

	void  SetTopLevel(Component* c) {
		pTopLevel = c;
	}

	__forceinline  bool  operator()(Component& c)
	{
		if(c.IsVisible()) {
			if(c.IsOverlay()  &&  (&c != pTopLevel)) {
				outOverlays.push_back(&c);
				return false;
			}
			return true;
		}
		return false;
	}
};


//====================================================================
void __fastcall ConstructDrawQueue(
							COMPONENT_PTR_VECTOR& components,
							COMPONENT_PTR_VECTOR& out,
							COMPONENT_PTR_VECTOR& tempOverlays,
							Manager::DEPTHFIRSTPATHMARKER_VECTOR& tempPathMarkers)
{
	out.clear();
	if(components.empty()) return;

	ComponentProcessor_VisibleOnly_ExtractOverlays  overlayExtractor(tempOverlays);

	for(uint i = 0; i < components.size(); i++) {

		tempOverlays.clear();
		tempOverlays.push_back(components[i]);

		uint  iUnprocessedOverlay = 0;
		while(iUnprocessedOverlay < tempOverlays.size()) {

			Component* unprocessedOverlay = tempOverlays[iUnprocessedOverlay++];

			overlayExtractor.SetTopLevel(unprocessedOverlay);
			ConstructDepthFirstQueue(*unprocessedOverlay, out, tempPathMarkers, overlayExtractor);
		}
	}
}


//====================================================================
template<typename TResourceMap, typename TKeyContainer>
void __fastcall MapCreateVRResources(const TResourceMap& resMap,
										VideoRenderer& vr,
										const TKeyContainer& keys)
{
	typename TKeyContainer::const_iterator itKey = keys.begin();
	while(itKey != keys.end()) {

		typename TResourceMap::const_iterator itRes = resMap.find(*itKey++);
		if(itRes != resMap.end()) {
			itRes->second->CreateVideoRendererResource(vr);
		}
	}
}


//====================================================================
template<typename TResourceMap>
void __fastcall MapReleaseVRResources(const TResourceMap& resMap)
{
	typename TResourceMap::const_iterator it = resMap.begin();
	while(it != resMap.end()) {
		(*it++).second->ReleaseVideoRendererResource();
	}
}


//====================================================================
__forceinline vec2 __fastcall
CalcRelativeMousePos(Component& c, const vec2& absMousePos)
{
	return (absMousePos - c.GetAbsolutePosition());
}


//====================================================================
template<typename TMap>
void __fastcall DumpMapKeys(const TMap& m, std::ostream& os, uint tabLevel = 0)
{
	TabEffector  te(tabLevel);

	TMap::const_iterator it = m.begin();
	while(it != m.end()) {
		os << te << (*it++).first << nl;
	}
}


//====================================================================
template<typename T>
struct Predicate_FindByName
{
	std::string targetName;

	Predicate_FindByName(const std::string& name) : targetName(name) {}
	bool operator()(const T* p) const { return (p->GetName() == targetName); }
};


//====================================================================
template<typename T>
struct Predicate_SortByName
{
	bool operator()(const T* lhs, const T* rhs) const {
		return (lhs->GetName() < rhs->GetName());
	}
};


//====================================================================
std::string __fastcall MakeObjectName(const gui::VideoResourceUser& vru)
{
	const gui::Component* c = dynamic_cast<const gui::Component*>(&vru);
	if(c) return c->MakeHierarchyTrace();
	return VW_GET_OBJECT_NAME(vru);
}


//====================================================================
// Returns -1 if not found.
//====================================================================
template<typename TProcessor>
int __fastcall GetFirst(const COMPONENT_PTR_VECTOR& cv,
						int iStartSearchFrom, bool bForward,
						TProcessor& processor)
{
	if(iStartSearchFrom < 0) return -1;
	if(iStartSearchFrom >= cv.size()) return -1;

	int iEnd = iStartSearchFrom;
	int i = iStartSearchFrom;

	while(true) {
		Component* c = cv[i];
		if(processor(*c)) return i;

		if(bForward) {
			i++;
			if(i >= cv.size()) i = 0;
		}
		else {
			i--;
			if(i < 0) i = cv.size() - 1;
		}
		if(i == iEnd) break;
	}
	return -1;
}


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//====================================================================
VW_IMPLEMENT_SINGLETON(Manager);


//====================================================================
Manager::Manager()
{
	VW_IMPLEMENT_SINGLETON_CTOR(Manager);

	bAllowMainContainerUpdateAtNotifyComponent = false;
	bAllowMainContainerUpdateAtNotifyShader = false;
	bAllowMainContainerUpdateAtNotifyCursor = false;

	pNullTarget = new NullTarget;
	pComponentToApplicationAdapter = new ComponentToApplicationAdapter;
	pNullCursor = new Cursor;

	bAllowMainContainerUpdateAtNotifyComponent = true;
	bAllowMainContainerUpdateAtNotifyShader = true;
	bAllowMainContainerUpdateAtNotifyCursor = true;

	bComponentLayersChanged = false;
	bComponentHierarchyChanged = false;

	bTileTextureObjectsChanged = false;
	bFontObjectsChanged = false;
	bShaderObjectsChanged = false;
	bCursorObjectsChanged = false;

	bTileTextureListChanged = false;
	bFontListChanged = false;

	bNeedFullTileTextureRecache = false;
	bNeedFullFontRecache = false;
	bNeedFullShaderRecache = false;
	bNeedFullCursorRecache = false;
	bNeedInternalCursorRecache = true;
	bNeedGlobalCursorParamsUpdate = false;
	bNeedFullGeometryUpdate = true;
	bNeedFullWindowResizeUpdate = true;

	pVideoRenderer = 0;
	pGUIRenderer = 0;
	bVideoRendererReadyToCreateResources = false;
	bRendererCreatesManagedResourcesOnly = false;
	windowWidth = 1;
	windowHeight = 1;

	bShowGUICursor = false;
	bShowSystemCursor = true;
	bVisibleSystemCursor = true;
	pActiveCursor = pNullCursor;

	bTolerateMissingInputDevice = false;
	pKeyboard = 0;
	pMouse = 0;

	secsTimeDelta = 0;
	secsMouseHover = 0;

	pCommandTarget = 0;
	commandArea = CommandArea_Void;
	commandButton = MouseButton_Count;

	pModalTarget = 0;
	pMouseCaptureTarget = 0;
	pKeyboardFocusTarget = pNullTarget;
	pMouseButtonNotifyTarget = pNullTarget;
	pForcedKeyboardTarget = 0;

	mousePosLast = vec2(-1, -1);
	bLastMousePosInsideAppWindow = false;
	bIgnoreMouseButtonEvents = false;
	bForceHitTest = true;
	bAllowExternalTopLevelActivate = true;
	bTargetNeedsCharacterData = false;
	mFirstClickedButtonLast = MouseButton_Count;

	SetKeyboardRepeat();
	ResetKeyboardRepeat();

	SetMouseRepeat();
	ResetMouseRepeat();

	SetMouseDoubleClick();
	ResetMouseDoubleClick();

	SetBehaviorOnWindowActivate();

	SetCycleQueueKeys();
	SetCycleQueueProperties();
	CycleQueueRemove(*pNullTarget);

	Internal_InvalidateCursorHandles();
}


//====================================================================
Manager::~Manager()
{
	// make sure system-cursor is visible
	if(! bShowSystemCursor) {
		ShowSystemCursor(true);
	}

	// delete all gui objects
	DeleteAllComponents();
	DeleteAllShaders();
	DeleteAllCursors();

	bAllowMainContainerUpdateAtNotifyComponent = false;
	bAllowMainContainerUpdateAtNotifyShader = false;
	bAllowMainContainerUpdateAtNotifyCursor = false;

	SafeDelete(pNullCursor);
	SafeDelete(pComponentToApplicationAdapter);
	SafeDelete(pNullTarget);

	// unregister video-resources
	UnregisterAllFonts();
	UnregisterAllTileTextures();

	VW_IMPLEMENT_SINGLETON_DTOR(Manager);
}


//====================================================================
void  Manager::DumpComponentHierarchy(std::ostream& os, uint tabLevel,
						bool bTopLevelOnly, bool bPrintDepth, bool bPrintAttribs)
{
	for(uint i = 0; i < componentsTopLevel.size(); i++) {
		componentsTopLevel[i]->DumpHierarchy(os, tabLevel,
					bPrintDepth, bPrintAttribs, 0, bTopLevelOnly, 0);
	}
}


//====================================================================
void  Manager::DumpCycleQueue(std::ostream& os, uint tabLevel)
{
	for(uint i = 0; i < componentsCycleQueue.size(); i++) {
		componentsCycleQueue[i]->DumpHierarchy(os, tabLevel, false, false, 0, true, 0);
	}
}


//====================================================================
void  Manager::DumpRegisteredTileTextures(std::ostream& os, uint tabLevel)
{
	DumpMapKeys(mapTileTextures, os, tabLevel);
}


//====================================================================
void  Manager::DumpRegisteredFonts(std::ostream& os, uint tabLevel)
{
	DumpMapKeys(mapFonts, os, tabLevel);
}


//====================================================================
void  Manager::DumpRegisteredShaders(std::ostream& os, uint tabLevel)
{
	TabEffector te(tabLevel);
	for(uint i = 0; i < shaders.size(); i++) {
		os << te;
		if(shaders[i]->GetName().empty()) os << "[unnamed] @ " << shaders[i];
		else os << shaders[i]->GetName();
		os << nl;
	}
}


//====================================================================
void  Manager::DumpRegisteredCursors(std::ostream& os, uint tabLevel)
{
	TabEffector te(tabLevel);
	for(uint i = 0; i < cursors.size(); i++) {
		os << te;
		if(cursors[i]->GetName().empty()) os << "[unnamed] @ " << cursors[i];
		else os << cursors[i]->GetName();
		os << nl;
	}
}


//====================================================================
void  Manager::SyncObjectStates()
{
	// video-renderer resource updates
	PrepareVideoRendererResources(false, false);

	// component-queue updates
	UpdateComponentHierarchyChanges();

	// window-resize update
	NotifyWindowResize();

	// resource-list notifications
	if(bTileTextureListChanged) {
		bTileTextureListChanged = false;

		VRUser_TileTextureListChanged();
	}
	if(bFontListChanged) {
		bFontListChanged = false;

		VRUser_FontListChanged();
	}

	// resource-handle recache
	bool bTileTextureRecache = bNeedFullTileTextureRecache  ||  tileTextureRecacheQueue.size();
	bool bFontRecache = bNeedFullFontRecache  ||  fontRecacheQueue.size();

	if(bTileTextureRecache) {
		if(bNeedFullTileTextureRecache) {
			CollectAllVRUsers(tileTextureRecacheQueue);
			bNeedFullGeometryUpdate = true;
		}
		VRUser_CacheTileTextureHandles();
	}
	if(bFontRecache) {
		if(bNeedFullFontRecache) {
			CollectAllVRUsers(fontRecacheQueue);
			bNeedFullGeometryUpdate = true;
		}
		VRUser_CacheFontHandles();
	}

	// resource-handle post-cache
	if(bTileTextureRecache) {
		VRUser_PostCacheTileTextureHandles();

		bNeedFullTileTextureRecache = false;
		tileTextureRecacheQueue.clear();
	}
	if(bFontRecache) {
		VRUser_PostCacheFontHandles();

		bNeedFullFontRecache = false;
		fontRecacheQueue.clear();
	}

	// shader-list notifications
	if(bNeedFullShaderRecache) {
		Component_ShaderListChanged();
	}
	// shader-handle recache
	if(bNeedFullShaderRecache  ||  shaderRecacheQueue.size()) {

		if(bNeedFullShaderRecache) {
			shaderRecacheQueue = componentsAll;
			bNeedFullGeometryUpdate = true;
		}
		Component_CacheShaderHandles();
		Component_PostCacheShaderHandles();

		bNeedFullShaderRecache = false;
		shaderRecacheQueue.clear();
	}

	// cursor-list notifications
	if(bNeedFullCursorRecache) {
		Component_CursorListChanged();
	}
	// cursor-handle recache
	if(bNeedFullCursorRecache  ||  cursorRecacheQueue.size()) {
		bNeedInternalCursorRecache = true;

		if(bNeedFullCursorRecache) {
			cursorRecacheQueue = componentsAll;
			bNeedFullGeometryUpdate = true;
		}
		Component_CacheCursorHandles();
		Component_PostCacheCursorHandles();

		bNeedFullCursorRecache = false;
		cursorRecacheQueue.clear();
	}
	if(bNeedInternalCursorRecache) {
		bNeedInternalCursorRecache = false;

		Internal_CacheCursorHandles();
	}

	// cursor-params recache
	if(bNeedGlobalCursorParamsUpdate) {
		bNeedGlobalCursorParamsUpdate = false;
		bNeedFullGeometryUpdate = true;

		DispatchGlobalCursorParams();
	}

	// geometry updates
	if(bNeedFullGeometryUpdate) {
		bNeedFullGeometryUpdate = false;

		componentsDirtyGeometry = componentsTopLevel;
	}
	UpdateDirtyGeometry();
}


//====================================================================
void  Manager::DispatchInputEvents()
{
	SyncObjectStates();

	// dispatch time-delta
	secsTimeDelta = Kernel::Instance().GetFrameTimeDeltaSecs();
	NotifyTimeDelta();

	// auto end-modal if target is hidden
	if(pModalTarget  &&  ! pModalTarget->IsVisible()) {
		EndModal(*pModalTarget);
	}

	// service-provider pre-event update
	NotifyServiceProvidersPreEventDispatch();

	// Dispatch by device-type. Employ asynchronous read sentries;
	// if device is in synchronous mode it is expected that either
	// kernel or application will sample it.
	if(pMouse) {
		AsyncSampleReadSentry  asrs(*pMouse);
		DispatchMouseEvents();
	}
	if(pKeyboard) {
		AsyncSampleReadSentry  asrs(*pKeyboard);
		DispatchKeyboardEvents();
	}

	// service-provider post-event update
	NotifyServiceProvidersPostEventDispatch();
}


//====================================================================
void  Manager::Draw()
{
	SyncObjectStates();
	if(! pGUIRenderer) return;

	// auto-begin
	bool bCallBeginEnd = ! pGUIRenderer->IsInsideBeginEnd();
	if(bCallBeginEnd) pGUIRenderer->Begin();

	// draw component hierarchy
	mat4  matTransformLast, matTransform;

	// cache last transform
	pGUIRenderer->GetTransform(matTransformLast);

	for(uint iComp = 0; iComp < componentsDrawQueue.size(); iComp++) {
		Component& c = *componentsDrawQueue[iComp];

		// activate positional transform
		const vec2& pos = c.GetAbsolutePosition();
		GenerateTranslateVM(matTransform, pos.x, pos.y, 0);

		pGUIRenderer->SetTransform(matTransform);

		// draw component
		c.DrawSelf(*pGUIRenderer);
	}

	// draw cursor
	if(pMouse  &&  pMouse->IsPositionInsideAppWindowClientArea()) {
		if(bShowGUICursor  ||  ! pMouse->CanDisplaySystemCursor()) {

			// calc absolute hot-spot
			vec2  hotSpot = pActiveCursor->GetHotSpot() * pActiveCursor->GetActiveSize();

			GenerateTranslateVM(matTransform,
				pMouse->GetPositionX() - hotSpot.x, pMouse->GetPositionY() - hotSpot.y, 0);

			pGUIRenderer->SetTransform(matTransform);
			const_cast<Cursor*>(pActiveCursor)->DrawSelf(*pGUIRenderer);
		}
	}

	// restore last transform
	pGUIRenderer->SetTransform(matTransformLast);

	// auto-end
	if(bCallBeginEnd) pGUIRenderer->End();
}


//====================================================================
MessageHandler*  Manager::GetServiceProvider(uint serviceID)
{
	SERVICE_PROVIDER_ITER it = mapServiceProviders.find(serviceID);
	if(it != mapServiceProviders.end()) return it->second;
	return 0;
}


//====================================================================
bool  Manager::RegisterServiceProvider(uint serviceID, MessageHandler& mh)
{
	if(serviceID == SID_Null) return false;

	MessageHandler* sp = GetServiceProvider(serviceID);
	if(sp) {
		if(sp == &mh) return true;
		sp->OnUnregisteredAsServiceProvider(serviceID);
	}

	mapServiceProviders[ serviceID ] = &mh;
	mh.OnRegisteredAsServiceProvider(serviceID);

	BroadcastMessage(Message_ServiceProviderChanged(serviceID));
	return true;
}


//====================================================================
bool  Manager::UnregisterServiceProvider(uint serviceID, MessageHandler& mh)
{
	if(serviceID == SID_Null) return false;

	SERVICE_PROVIDER_ITER it = mapServiceProviders.find(serviceID);
	if(it != mapServiceProviders.end()) {
		if(&mh == it->second) {
			mapServiceProviders.erase(serviceID);
			mh.OnUnregisteredAsServiceProvider(serviceID);
			BroadcastMessage(Message_ServiceProviderChanged(serviceID));
			return true;
		}
	}
	return false;
}


//====================================================================
bool  Manager::UnregisterServiceProvider(uint serviceID)
{
	if(serviceID == SID_Null) return false;

	SERVICE_PROVIDER_ITER it = mapServiceProviders.find(serviceID);
	if(it != mapServiceProviders.end()) {
		return UnregisterServiceProvider(serviceID, *it->second);
	}
	return false;
}


//====================================================================
void  Manager::UnregisterAllServiceProviders()
{
	mapServiceProviders.clear();
}


//====================================================================
bool  Manager::SendMessageToServiceProvider(uint serviceID, Message& msg)
{
	SERVICE_PROVIDER_ITER it = mapServiceProviders.find(serviceID);
	if(it != mapServiceProviders.end()) {
		MessageHandler* mh = it->second;
		return mh->DispatchMessage(msg);
	}
	return false;
}


//====================================================================
bool  Manager::BroadcastMessage(Message& msg)
{
	bool bHandled = false;
	for(uint i = 0; i < componentsAll.size(); i++) {
		if(componentsAll[i]->DispatchMessage(msg)) bHandled = true;
	}
	return bHandled;
}


//====================================================================
void  Manager::DeleteAllComponents()
{
	bAllowMainContainerUpdateAtNotifyComponent = false;
	DeletePtrContainer(componentsTopLevel);
	bAllowMainContainerUpdateAtNotifyComponent = true;
}


//====================================================================
void  Manager::SetCycleQueueProperties(bool bEnableAutoCycle,
										bool bAutoInsertAtTopLevelCreate,
										bool bAutoInsertAtHierarchyChange)
{
	bCycleQueueEnableAutoCycle = bEnableAutoCycle;
	bCycleQueueAutoInsertAtCreate = bAutoInsertAtTopLevelCreate;
	bCycleQueueAutoInsertAtHierarchyChange = bAutoInsertAtHierarchyChange;
}


//====================================================================
void  Manager::GetCycleQueueProperties(bool& bEnableAutoCycle,
										bool& bAutoInsertAtTopLevelCreate,
										bool& bAutoInsertAtHierarchyChange)
{
	bEnableAutoCycle = bCycleQueueEnableAutoCycle;
	bAutoInsertAtTopLevelCreate = bCycleQueueAutoInsertAtCreate;
	bAutoInsertAtHierarchyChange = bCycleQueueAutoInsertAtHierarchyChange;
}


//====================================================================
void  Manager::CycleQueueClear()
{
	componentsCycleQueue.clear();
}


//====================================================================
void  Manager::CycleQueueInsert(Component& c)
{
	UniqueContainerPushBack(componentsCycleQueue, &c);
}


//====================================================================
void  Manager::CycleQueueRemove(Component& c)
{
	EraseFromContainer(componentsCycleQueue, &c);
}


//====================================================================
void  Manager::CycleQueueOrder(Component& before, Component& after)
{
	if(&before == &after) return;

	// find current queue-positions
	int iBefore = -1;
	int iAfter  = -1;

	for(uint i = 0; i < componentsCycleQueue.size(); i++) {
		Component* c = componentsCycleQueue[i];

		if(c == &before) iBefore = i;
		else if(c == &after) iAfter = i;
	}

	// swap positions if necessary
	if(iBefore >= 0  &&  iAfter >= 0) {
		if(iAfter < iBefore) {
			Swap(componentsCycleQueue[iBefore], componentsCycleQueue[iAfter]);
		}
	}
}


//====================================================================
void  Manager::CycleQueueOrderToEnd(Component& c)
{
	if(FindInContainer(componentsCycleQueue, &c)) {
		EraseFromContainer(componentsCycleQueue, &c);
		componentsCycleQueue.push_back(&c);
	}
}


//====================================================================
void  Manager::CycleQueueOrderToStart(Component& c)
{
	if(FindInContainer(componentsCycleQueue, &c)) {
		EraseFromContainer(componentsCycleQueue, &c);
		componentsCycleQueue.insert(componentsCycleQueue.begin(), &c);
	}
}


//====================================================================
void  Manager::CycleQueueInsert(Application& app)
{
	CycleQueueInsert(*pComponentToApplicationAdapter);
}


//====================================================================
void  Manager::CycleQueueRemove(Application& app)
{
	CycleQueueRemove(*pComponentToApplicationAdapter);
}


//====================================================================
void  Manager::CycleQueueOrder(Component& before, Application& app)
{
	CycleQueueOrder(before, *pComponentToApplicationAdapter);
}


//====================================================================
void  Manager::CycleQueueOrder(Application& app, Component& after)
{
	CycleQueueOrder(*pComponentToApplicationAdapter, after);
}


//====================================================================
void  Manager::CycleQueueOrderToEnd(Application& app)
{
	CycleQueueOrderToEnd(*pComponentToApplicationAdapter);
}


//====================================================================
void  Manager::CycleQueueOrderToStart(Application& app)
{
	CycleQueueOrderToStart(*pComponentToApplicationAdapter);
}


//====================================================================
void  Manager::CycleQueueActivate(bool bForward)
{
	// find first active
	int iActive = GetFirst(componentsCycleQueue, 0, true, ComponentProcessor_VisibleActiveOnly());

	// determine search position
	int iSearchStart = iActive + (bForward ? 1 : -1);

	if(iSearchStart < 0  ||  iSearchStart >= componentsCycleQueue.size()) {
		iSearchStart = bForward ? 0 : (componentsCycleQueue.size() - 1);
	}

	// find next available
	int iNext = GetFirst(componentsCycleQueue, iSearchStart, bForward, ComponentProcessor_VisibleOnly());

	// activate next
	if(iNext >= 0) {
		ActivateTopLevel(*componentsCycleQueue[iNext]);
	}
}


//====================================================================
void  Manager::BringToFront(Component& c)
{
	COMPONENT_PTR_ITER  it = std::find(componentsTopLevel.begin(), componentsTopLevel.end(), &c);
	if(it == componentsTopLevel.end()) return;
	if(componentsTopLevel.back() == &c) return; //already front

	if(componentsTopLevel.back()->GetLayer() != c.GetLayer()) {
		bComponentLayersChanged = true; //layer update required
	}

	componentsTopLevel.erase(it);
	componentsTopLevel.push_back(&c);

	bComponentHierarchyChanged = true;
}


//====================================================================
void  Manager::SendToBack(Component& c)
{
	COMPONENT_PTR_ITER  it = std::find(componentsTopLevel.begin(), componentsTopLevel.end(), &c);
	if(it == componentsTopLevel.end()) return;
	if(componentsTopLevel.front() == &c) return; //already back

	if(componentsTopLevel.front()->GetLayer() != c.GetLayer()) {
		bComponentLayersChanged = true; //layer update required
	}

	componentsTopLevel.erase(it);
	componentsTopLevel.insert(componentsTopLevel.begin(), &c);

	bComponentHierarchyChanged = true;
}


//====================================================================
void  Manager::SetZOrder(Component& front, Component& back)
{
	// only valid between components in the same layer
	if(front.GetLayer() != back.GetLayer()) return;

	// find current z-positions
	int iFront = -1;
	int iBack  = -1;

	for(uint i = 0; i < componentsTopLevel.size(); i++) {
		Component* c = componentsTopLevel[i];

		if(c == &front) iFront = i;
		else if(c == &back) iBack = i;
	}

	// swap z-positions if necessary
	if(iFront >= 0  &&  iBack >= 0) {
		if(iFront < iBack) {
			Swap(componentsTopLevel[iFront], componentsTopLevel[iBack]);

			bComponentHierarchyChanged = true;
		}
	}
}


//====================================================================
void  Manager::ReportDirtyGeometry(Component& c)
{
	Component& cTopLevel = c.GetTopLevelAncestor();
	UniqueContainerPushBack(componentsDirtyGeometry, &cTopLevel);
}


//====================================================================
void  Manager::UpdateDirtyGeometry()
{
	for(uint i = 0; i < componentsDirtyGeometry.size(); i++) {
		componentsDirtyGeometry[i]->UpdateGeometry();
	}
	componentsDirtyGeometry.clear();
}


//====================================================================
void  Manager::UpdateAllGeometry()
{
	UpdateComponentHierarchyChanges();

	componentsDirtyGeometry = componentsTopLevel;
	UpdateDirtyGeometry();
}


//====================================================================
bool  Manager::IntersectsWindow(Component& c)
{
	if(! pVideoRenderer) return true;

	Rectangle  rWindow(0, 0, windowWidth, windowHeight);
	Rectangle  rComp(c.GetAbsolutePosition().x, c.GetAbsolutePosition().y, c.GetSize().x, c.GetSize().y);

	return rWindow.Intersects(rComp);
}


//====================================================================
void  Manager::PrepareVideoRendererResources(bool bRecreateAllTileTextures,
											bool bRecreateAllFonts)
{
	if(bTileTextureObjectsChanged  ||  bRecreateAllTileTextures) {
		DoCreateVRRTileTextures(bRecreateAllTileTextures);
		bNeedFullTileTextureRecache = true;
	}
	if(bFontObjectsChanged  ||  bRecreateAllFonts) {
		DoCreateVRRFonts(bRecreateAllFonts);
		bNeedFullFontRecache = true;
	}
}


//====================================================================
void  Manager::RegisterTileTexture(const std::string& name, TileTexture* pTexture)
{
	if(name.empty()) {
		VW_THROW(GUIException, "Invalid argument: passed empty string for texture-name.");
	}
	if(! pTexture) {
		VW_THROW(GUIException, "Invalid argument: passed null for texture-object.");
	}
	std::string outName;
	if(IsRegisteredTileTexture(*pTexture, &outName)) {
		VW_THROW(GUIException, name + ": texture-object already registered under name: " + outName);
	}
	mapTileTextures[name] = TILETEX_APTR(pTexture);
	bTileTextureObjectsChanged = true;
	bTileTextureListChanged = true;
	vrrCreateListTileTextures.push_back(name);
}


//====================================================================
void  Manager::RegisterTileTexture(const std::string& name, AutoPtr<TileTexture>& apTexture)
{
	RegisterTileTexture(name, apTexture.Get());
	apTexture.Release();
}


//====================================================================
void  Manager::UnregisterTileTexture(const std::string& name)
{
	if(mapTileTextures.erase(name)) {
		bTileTextureObjectsChanged = true;
		bTileTextureListChanged = true;
	}
}


//====================================================================
void  Manager::UnregisterAllTileTextures()
{
	mapTileTextures.clear();
	bTileTextureObjectsChanged = true;
	bTileTextureListChanged = true;
}


//====================================================================
void  Manager::GetRegisteredTileTextureNames(std::vector<std::string>& outNames)
{
	MapExtractKeys(mapTileTextures, outNames);
}


//====================================================================
bool  Manager::IsRegisteredTileTexture(TileTexture& texture, std::string* pOutName)
{
	return (MapFindKeyByValue(mapTileTextures, &texture, pOutName) != mapTileTextures.end());
}


//====================================================================
void  Manager::AddTileTextureAlias(const std::string& alias, const std::string& refName)
{
	if(refName.empty()) return;
	mapTileTextureAliases[alias] = refName;
	bNeedFullTileTextureRecache = true;
	bTileTextureListChanged = true;
}


//====================================================================
void  Manager::ClearTileTextureAlias(const std::string& alias)
{
	if(mapTileTextureAliases.erase(alias)) {
		bNeedFullTileTextureRecache = true;
		bTileTextureListChanged = true;
	}
}


//====================================================================
void  Manager::ClearAllTileTextureAliases()
{
	mapTileTextureAliases.clear();
	bNeedFullTileTextureRecache = true;
	bTileTextureListChanged = true;
}


//====================================================================
void  Manager::GetTileTextureAliases(std::vector<std::string>& outAliases)
{
	MapExtractKeys(mapTileTextureAliases, outAliases);
}


//====================================================================
bool  Manager::ResolveTileTextureAlias(const std::string& alias, std::string& refName)
{
	return MapResolveAlias(mapTileTextureAliases, alias, refName);
}


//====================================================================
const TileTexture*  Manager::GetTileTexture(const std::string& name)
{
	if(name.empty()) return 0;

	std::string refName;
	ResolveTileTextureAlias(name, refName);

	TILETEX_ITER it = mapTileTextures.find(refName);
	if(it != mapTileTextures.end()) return it->second.Get();
	return 0;
}


//====================================================================
void  Manager::RequestTileTextureRecache(VideoResourceUser& vru)
{
	tileTextureRecacheQueue.push_back(&vru);
}


//====================================================================
void  Manager::RegisterFont(const std::string& name, Font* pFont)
{
	if(name.empty()) {
		VW_THROW(GUIException, "Invalid argument: passed empty string for font-name.");
	}
	if(! pFont) {
		VW_THROW(GUIException, "Invalid argument: passed null for font-object.");
	}
	std::string outName;
	if(IsRegisteredFont(*pFont, &outName)) {
		VW_THROW(GUIException, name + ": font-object already registered under name: " + outName);
	}
	mapFonts[name] = FONT_APTR(pFont);
	bFontObjectsChanged = true;
	bFontListChanged = true;
	vrrCreateListFonts.push_back(name);
}


//====================================================================
void  Manager::RegisterFont(const std::string& name, AutoPtr<Font>& apFont)
{
	RegisterFont(name, apFont.Get());
	apFont.Release();
}


//====================================================================
void  Manager::RegisterFont(const std::string& name, AutoPtr<FontTexture>& apFontTexture)
{
	RegisterFont(name, apFontTexture.Get());
	apFontTexture.Release();
}


//====================================================================
void  Manager::UnregisterFont(const std::string& name)
{
	if(mapFonts.erase(name)) {
		bFontObjectsChanged = true;
		bFontListChanged = true;
	}
}


//====================================================================
void  Manager::UnregisterAllFonts()
{
	mapFonts.clear();
	bFontObjectsChanged = true;
	bFontListChanged = true;
}


//====================================================================
void  Manager::GetRegisteredFontNames(std::vector<std::string>& outNames)
{
	MapExtractKeys(mapFonts, outNames);
}


//====================================================================
bool  Manager::IsRegisteredFont(Font& font, std::string* pOutName)
{
	return (MapFindKeyByValue(mapFonts, &font, pOutName) != mapFonts.end());
}


//====================================================================
void  Manager::AddFontAlias(const std::string& alias, const std::string& refName)
{
	if(refName.empty()) return;
	mapFontAliases[alias] = refName;
	bNeedFullFontRecache = true;
	bFontListChanged = true;
}


//====================================================================
void  Manager::ClearFontAlias(const std::string& alias)
{
	if(mapFontAliases.erase(alias)) {
		bNeedFullFontRecache = true;
		bFontListChanged = true;
	}
}


//====================================================================
void  Manager::ClearAllFontAliases()
{
	mapFontAliases.clear();
	bNeedFullFontRecache = true;
	bFontListChanged = true;
}


//====================================================================
void  Manager::GetFontAliases(std::vector<std::string>& outAliases)
{
	MapExtractKeys(mapFontAliases, outAliases);
}


//====================================================================
bool  Manager::ResolveFontAlias(const std::string& alias, std::string& refName)
{
	return MapResolveAlias(mapFontAliases, alias, refName);
}


//====================================================================
const Font*  Manager::GetFont(const std::string& name)
{
	if(name.empty()) return 0;

	std::string refName;
	ResolveFontAlias(name, refName);

	FONT_ITER it = mapFonts.find(refName);
	if(it != mapFonts.end()) return it->second.Get();
	return 0;
}


//====================================================================
void  Manager::RequestFontRecache(VideoResourceUser& vru)
{
	fontRecacheQueue.push_back(&vru);
}


//====================================================================
void  Manager::DeleteAllShaders()
{
	bAllowMainContainerUpdateAtNotifyShader = false;
	DeletePtrContainer(shaders);
	bAllowMainContainerUpdateAtNotifyShader = true;
}


//====================================================================
void  Manager::DeleteShader(const std::string& shaderName)
{
	SHADER_PTR_ITER it = std::find_if(shaders.begin(), shaders.end(),
						Predicate_FindByName<gui::BaseShader>(shaderName));
	if(it != shaders.end()) {
		delete *it;
	}
}


//====================================================================
void  Manager::AddShaderAlias(const std::string& alias, const std::string& refName)
{
	if(refName.empty()) return;
	mapShaderAliases[alias] = refName;
	bNeedFullShaderRecache = true;
}


//====================================================================
void  Manager::ClearShaderAlias(const std::string& alias)
{
	if(mapShaderAliases.erase(alias)) {
		bNeedFullShaderRecache = true;
	}
}


//====================================================================
void  Manager::ClearAllShaderAliases()
{
	mapShaderAliases.clear();
	bNeedFullShaderRecache = true;
}


//====================================================================
void  Manager::GetShaderAliases(std::vector<std::string>& outAliases)
{
	MapExtractKeys(mapShaderAliases, outAliases);
}


//====================================================================
bool  Manager::ResolveShaderAlias(const std::string& alias, std::string& refName)
{
	return MapResolveAlias(mapShaderAliases, alias, refName);
}


//====================================================================
const BaseShader*  Manager::GetShader(const std::string& name)
{
	if(name.empty()) return 0;

	std::string refName;
	ResolveShaderAlias(name, refName);

	SHADER_PTR_ITER it = std::find_if(shaders.begin(), shaders.end(),
						Predicate_FindByName<gui::BaseShader>(refName));
	if(it != shaders.end()) return *it;
	return 0;
}


//====================================================================
void  Manager::RequestShaderRecache(Component& c)
{
	shaderRecacheQueue.push_back(&c);
}


//====================================================================
void  Manager::DeleteAllCursors()
{
	bAllowMainContainerUpdateAtNotifyCursor = false;
	DeletePtrContainer(cursors);
	bAllowMainContainerUpdateAtNotifyCursor = true;
}


//====================================================================
void  Manager::DeleteCursor(const std::string& cursorName)
{
	CURSOR_PTR_ITER it = std::find_if(cursors.begin(), cursors.end(),
						Predicate_FindByName<gui::Cursor>(cursorName));
	if(it != cursors.end()) {
		delete *it;
	}
}


//====================================================================
void  Manager::AddCursorAlias(const std::string& alias, const std::string& refName)
{
	if(refName.empty()) return;
	mapCursorAliases[alias] = refName;
	bNeedFullCursorRecache = true;
}


//====================================================================
void  Manager::ClearCursorAlias(const std::string& alias)
{
	if(mapCursorAliases.erase(alias)) {
		bNeedFullCursorRecache = true;
	}
}


//====================================================================
void  Manager::ClearAllCursorAliases()
{
	mapCursorAliases.clear();
	bNeedFullCursorRecache = true;
}


//====================================================================
void  Manager::GetCursorAliases(std::vector<std::string>& outAliases)
{
	MapExtractKeys(mapCursorAliases, outAliases);
}


//====================================================================
bool  Manager::ResolveCursorAlias(const std::string& alias, std::string& refName)
{
	return MapResolveAlias(mapCursorAliases, alias, refName);
}


//====================================================================
const Cursor*  Manager::GetCursor(const std::string& name)
{
	if(name.empty()) return 0;

	std::string refName;
	ResolveCursorAlias(name, refName);

	CURSOR_PTR_ITER it = std::find_if(cursors.begin(), cursors.end(),
						Predicate_FindByName<gui::Cursor>(refName));
	if(it != cursors.end()) return *it;
	return 0;
}


//====================================================================
void  Manager::RequestCursorRecache(Component& c)
{
	cursorRecacheQueue.push_back(&c);
}


//====================================================================
void  Manager::DispatchGlobalCursorParams()
{
	for(uint i = 0; i < cursors.size(); i++) {
		cursors[i]->UpdateColorCache();
		cursors[i]->UpdateGeometryCache();
	}
}


//====================================================================
void  Manager::ShowGUICursor(bool bShow)
{
	bShowGUICursor = bShow;
}


//====================================================================
void  Manager::ShowSystemCursor(bool bShow)
{
	bShowSystemCursor = bShow;

	// avoid hiding the cursor while it is outside the application-window
	if(bShowSystemCursor  ||  bLastMousePosInsideAppWindow) {
		DisplaySystemCursor(bShowSystemCursor);
	}
}


//====================================================================
void  Manager::SetKeyboardRepeat(double secsDelay, double secsPause)
{
	fKeyboardRepeatDelay = ClampedLow<double>(secsDelay, 0);
	fKeyboardRepeatPause = ClampedLow<double>(secsPause, 0);
}


//====================================================================
void  Manager::GetKeyboardRepeat(double& secsDelay, double& secsPause)
{
	secsDelay = fKeyboardRepeatDelay;
	secsPause = fKeyboardRepeatPause;
}


//====================================================================
void  Manager::SetMouseRepeat(double secsDelay, double secsPause)
{
	fMouseRepeatDelay = ClampedLow<double>(secsDelay, 0);
	fMouseRepeatPause = ClampedLow<double>(secsPause, 0);
}


//====================================================================
void  Manager::GetMouseRepeat(double& secsDelay, double& secsPause)
{
	secsDelay = fMouseRepeatDelay;
	secsPause = fMouseRepeatPause;
}


//====================================================================
void  Manager::SetMouseDoubleClick(double secsMaxToSecondClick)
{
	fDoubleClickLimit = ClampedLow<double>(secsMaxToSecondClick, 0);
}


//====================================================================
void  Manager::GetMouseDoubleClick(double& secsMaxToSecondClick)
{
	secsMaxToSecondClick = fDoubleClickLimit;
}


//====================================================================
void  Manager::SetCycleQueueKeys(eKey kcCycle, eKeyModifier kmCycle, eKeyModifier kmBackward)
{
	kcCycleQueue = kcCycle;
	kmCycleQueue = kmCycle;
	kmCycleQueueBackward = kmBackward;
}


//====================================================================
void  Manager::GetCycleQueueKeys(eKey& kcCycle, eKeyModifier& kmCycle, eKeyModifier& kmBackward)
{
	kcCycle = kcCycleQueue;
	kmCycle = kmCycleQueue;
	kmBackward = kmCycleQueueBackward;
}


//====================================================================
void  Manager::SetBehaviorOnWindowActivate(bool _bIgnoreMouseClickAfterWindowActivate,
										bool _bIgnoreMouseClickAfterWindowCreate)
{
	bIgnoreMouseClickAfterWindowActivate = _bIgnoreMouseClickAfterWindowActivate;
	bIgnoreMouseClickAfterWindowCreate = _bIgnoreMouseClickAfterWindowCreate;
}


//====================================================================
void  Manager::GetBehaviorOnWindowActivate(bool& _bIgnoreMouseClickAfterWindowActivate,
										bool& _bIgnoreMouseClickAfterWindowCreate)
{
	_bIgnoreMouseClickAfterWindowActivate = bIgnoreMouseClickAfterWindowActivate;
	_bIgnoreMouseClickAfterWindowCreate = bIgnoreMouseClickAfterWindowCreate;
}


//====================================================================
void  Manager::EnableApplicationMessageDispatch(bool bEnable)
{
	pComponentToApplicationAdapter->EnableMessageDispatch(bEnable);
}


//====================================================================
void  Manager::TimeNotifyRegister(Component& c, bool bRegister)
{
	if(bRegister) {
		UniqueContainerPushBack(timeNotifyComponents, &c);
	}
	else {
		EraseFromContainer(timeNotifyComponents, &c);
	}
}


//====================================================================
void  Manager::TimeNotifyRegister(BaseShader& s, bool bRegister)
{
	if(bRegister) {
		UniqueContainerPushBack(timeNotifyShaders, &s);
	}
	else {
		EraseFromContainer(timeNotifyShaders, &s);
	}
}


//====================================================================
void  Manager::TimeNotifyRegister(Cursor& c, bool bRegister)
{
	if(bRegister) {
		UniqueContainerPushBack(timeNotifyCursors, &c);
	}
	else {
		EraseFromContainer(timeNotifyCursors, &c);
	}
}


//====================================================================
void  Manager::ResetMouseHoverTime(double _secsMouseHover)
{
	secsMouseHover = ClampedLow<double>(_secsMouseHover, 0);
}


//====================================================================
Component*  Manager::GetMouseOverTarget()
{
	if(hitRecordDeepCurr.component) return hitRecordDeepCurr.component;
	return pNullTarget;
}


//====================================================================
void  Manager::ActivateTopLevel(Component& c, bool bAutoBringToFront)
{
	if(bAllowExternalTopLevelActivate  &&  c.IsTopLevel()) {

		EndMouseCapture();
		EndCommandMode();

		// deactivate all except target
		NotifyDeactivate(componentsAll, &c, false);

		if(&c != pComponentToApplicationAdapter) {
			NotifyDeactivate(*pComponentToApplicationAdapter, false);
		}

		// activate target
		COMPONENT_PTR_VECTOR targets;
		targets.push_back(&c);
		SetupMouseClickedBranchCache(targets, MouseButton_Left);
		bForceHitTest = true;

		NotifyActivate(mouseClickedBranchLast, false, vec2::ZERO);

		// additional setup
		c.ShowTopLevel(true, bAutoBringToFront);
		ChangeKeyboardFocusTarget(c, false);
	}
}


//====================================================================
void  Manager::ActivateApplication()
{
	ActivateTopLevel(*pComponentToApplicationAdapter, false);
}


//====================================================================
Component*  Manager::GetActiveTopLevel()
{
	for(uint i = 0; i < componentsTopLevel.size(); i++) {
		Component* c = componentsTopLevel[i];
		if(c->IsVisible()  &&  c->IsActive()) return componentsTopLevel[i];
	}
	return 0;
}


//====================================================================
bool  Manager::IsApplicationActive()
{
	return (pComponentToApplicationAdapter->IsActive());
}


//====================================================================
void  Manager::RunModal(Component& c)
{
	if(c.IsTopLevel()) {
		pModalTarget = &c;
		modalTargetStack.push_back(pModalTarget);
		ActivateTopLevel(*pModalTarget, true);
	}
}


//====================================================================
void  Manager::EndModal(Component& c)
{
	if(&c == pModalTarget) {
		if(modalTargetStack.size()) {
			modalTargetStack.pop_back();
		}
		if(modalTargetStack.empty()) {
			pModalTarget = 0;
		}
		else {
			pModalTarget = modalTargetStack.back();

			if(pModalTarget == pComponentToApplicationAdapter) {
				ActivateApplication();
			}
			else {
				ActivateTopLevel(*pModalTarget, true);
			}
		}
	}
}


//====================================================================
void  Manager::RunModal(Application& app)
{
	pModalTarget = pComponentToApplicationAdapter;
	modalTargetStack.push_back(pModalTarget);
	ActivateApplication();
}


//====================================================================
void  Manager::EndModal(Application& app)
{
	EndModal(*pComponentToApplicationAdapter);
}


//====================================================================
bool  Manager::RequestMouseCapture(Component& c)
{
	// handle re-capture
	if(IsMouseCaptured()) {
		return (pMouseCaptureTarget == &c);
	}
	// cannot capture in command-mode
	if(pCommandTarget) return false;

	// only modal-target descendants or modal-target itself can capture
	if(pModalTarget) {
		if(! IsModalTargetDescendant(c)) return false;
	}

	// store target & send notification
	pMouseCaptureTarget = &c;
	pMouseCaptureTarget->DispatchMessage(Message_MouseCapture(true));
	return true;
}


//====================================================================
void  Manager::ReleaseMouseCapture(Component& c)
{
	if(pMouseCaptureTarget == &c) {
		EndMouseCapture();
	}
}


//====================================================================
bool  Manager::RequestMouseCapture(Application& app)
{
	return RequestMouseCapture(*pComponentToApplicationAdapter);
}


//====================================================================
void  Manager::ReleaseMouseCapture(Application& app)
{
	ReleaseMouseCapture(*pComponentToApplicationAdapter);
}


//====================================================================
bool  Manager::RequestKeyboardFocus(Component& c)
{
	// only command-target descendants or command-target itself can refocus
	if(pCommandTarget) {
		if(! pCommandTarget->IsAncestorOf(c)) return false;
	}

	// only modal-target descendants or modal-target itself can capture
	if(pModalTarget) {
		if(! IsModalTargetDescendant(c)) return false;
	}

	// mark forced target
	pForcedKeyboardTarget = &c;

	ChangeKeyboardFocusTarget(c, false);
	return true;
}


//====================================================================
void  Manager::ReleaseKeyboardFocus(Component& c)
{
	if(pKeyboardFocusTarget == &c) {
		ChangeKeyboardFocusTarget(*pComponentToApplicationAdapter, false);
	}
}


//====================================================================
bool  Manager::RequestKeyboardFocus(Application& app)
{
	return RequestKeyboardFocus(*pComponentToApplicationAdapter);
}


//====================================================================
void  Manager::BreakCommandMode()
{
	if(pCommandTarget) {
		EndCommandMode();
	}
}


//====================================================================
Manager::NullTarget::NullTarget()
{
	SetDefaultCommandArea(CommandArea_Void);
}


//====================================================================
Manager::ComponentToApplicationAdapter::ComponentToApplicationAdapter()
{
	pApp = 0;
	bEnableMessageDispatch = true;
	SetDefaultCommandArea(CommandArea_Void);
}


//====================================================================
void  Manager::ComponentToApplicationAdapter::ConnectApplication(bool bConnect)
{
	pApp = bConnect ? Kernel::Instance().GetApplicationPtr() : 0;
}


//====================================================================
void  Manager::ComponentToApplicationAdapter::EnableMessageDispatch(bool bEnable)
{
	bEnableMessageDispatch = bEnable;
}


//====================================================================
bool  Manager::ComponentToApplicationAdapter::HandleMessage(gui::Message& msg)
{
	if(bEnableMessageDispatch  &&  pApp) {
		return pApp->DispatchMessage(msg);
	}
	return false;
}


//====================================================================
void  Manager::DisplaySystemCursor(bool bDisplay)
{
	if(bVisibleSystemCursor != bDisplay) {
		bVisibleSystemCursor = bDisplay;

		// system keeps internal display-count
		::ShowCursor(bDisplay ? TRUE : FALSE);
	}
}


//====================================================================
void  Manager::NotifyVideoRendererChanged()
{
	UpdateComponentHierarchyChanges();

	VRUSER_PTR_VECTOR  vrUsers;
	CollectAllVRUsers(vrUsers);

	for(uint i = 0; i < vrUsers.size(); i++) {
		vrUsers[i]->OnVideoRendererChanged(pVideoRenderer);
	}
}


//====================================================================
void  Manager::NotifyWindowResize()
{
	// avoid sending bad window-size
	if(! pGUIRenderer) return;

	if(bNeedFullWindowResizeUpdate) {
		componentsWindowResize = componentsAll;
	}
	for(uint i = 0; i < componentsWindowResize.size(); i++) {
		componentsWindowResize[i]->OnWindowResize(windowWidth, windowHeight);
	}
	bNeedFullWindowResizeUpdate = false;
	componentsWindowResize.clear();
}


//====================================================================
void  Manager::NotifyTimeDelta()
{
	{for(uint i = 0; i < timeNotifyShaders.size(); i++) {
		timeNotifyShaders[i]->OnTimeDelta(secsTimeDelta);
	}}
	{for(uint i = 0; i < timeNotifyCursors.size(); i++) {
		timeNotifyCursors[i]->OnTimeDelta(secsTimeDelta);
	}}
	{for(uint i = 0; i < timeNotifyComponents.size(); i++) {
		timeNotifyComponents[i]->OnTimeDelta(secsTimeDelta);
	}}
}


//====================================================================
void  Manager::NotifyServiceProvidersPreEventDispatch()
{
	SERVICE_PROVIDER_ITER it = mapServiceProviders.begin();
	while(it != mapServiceProviders.end()) {
		MessageHandler* mh = (*it++).second;
		mh->ServiceProviderPreEventDispatch(secsTimeDelta);
	}
}


//====================================================================
void  Manager::NotifyServiceProvidersPostEventDispatch()
{
	SERVICE_PROVIDER_ITER it = mapServiceProviders.begin();
	while(it != mapServiceProviders.end()) {
		MessageHandler* mh = (*it++).second;
		mh->ServiceProviderPostEventDispatch(secsTimeDelta);
	}
}


//====================================================================
void  Manager::UpdateComponentHierarchyChanges()
{
	if(bComponentLayersChanged) {
		UpdateComponentLayers();
	}
	if(bComponentHierarchyChanged) {
		UpdateComponentQueues();
	}
}


//====================================================================
void  Manager::UpdateComponentLayers()
{
	bComponentLayersChanged = false;
	bComponentHierarchyChanged = true;

	std::stable_sort(componentsTopLevel.begin(), componentsTopLevel.end(), LayerSortPredicate());
}


//====================================================================
void  Manager::UpdateComponentQueues()
{
	bComponentHierarchyChanged = false;
	bForceHitTest = true;

	ConstructDepthFirstQueue(componentsTopLevel, false, componentsAll,
						tsDepthFirstPathMarkers, ComponentProcessor_All());

	ConstructDrawQueue(componentsTopLevel, componentsDrawQueue,
						tsOverlays, tsDepthFirstPathMarkers);
}


//====================================================================
void  Manager::DispatchKeyboardEvents()
{
	// auto-change focus if component disabled
	if(! pKeyboardFocusTarget->IsEnabled()) {

		if(pKeyboardFocusTarget->GetParent()) {
			ChangeKeyboardFocusTarget(*pKeyboardFocusTarget->GetParent(), false);
		}
		else {
			ChangeKeyboardFocusTarget(*pComponentToApplicationAdapter, false);
		}
	}
	// auto-change focus if component is hidden
	if(! pKeyboardFocusTarget->IsBranchVisible()) {
		ChangeKeyboardFocusTarget(*pComponentToApplicationAdapter, false);
	}

	// auto cycle-active
	if(bCycleQueueEnableAutoCycle) {

		bool bCycle = pKeyboard->KeyDown(kcCycleQueue)  &&
			(kmCycleQueue == KeyModifier_Count  ||  pKeyboard->KeyModifierPressed(kmCycleQueue));

		if(bCycle) {
			bool bForward = kmCycleQueueBackward == KeyModifier_Count  ||
							! pKeyboard->KeyModifierPressed(kmCycleQueueBackward);

			CycleQueueActivate(bForward);
			return;
		}
	}

	// event counters
	uint numKeyDown = 0;
	uint numKeyUp = 0;

	// accumulate events
	{for(uint i = 0; i < Key_Count; i++) {
		eKey kc = eKey(i);

		if(pKeyboard->KeyDown(kc)) {
			keyDown[numKeyDown] = kc;

			// map key-code to character
			char ch = 0;
			if(bTargetNeedsCharacterData) {
				if(! pKeyboard->MapKeyCodeToCharacter(kc, ch)) ch = 0;
			}
			keyDownChar[numKeyDown++] = ch;
			continue;
		}

		if(pKeyboard->KeyUp(kc)) {
			keyUp[numKeyUp] = kc;

			// map key-code to character
			char ch = 0;
			if(bTargetNeedsCharacterData) {
				if(! pKeyboard->MapKeyCodeToCharacter(kc, ch)) ch = 0;
			}
			keyUpChar[numKeyUp++] = ch;
		}
	}}

	// dispatch key-down events
	if(numKeyDown) {
		for(uint i = 0; i < numKeyDown; i++) {
			NotifyKeyboardKey(keyDown[i], true, 0, keyDownChar[i]);
		}
		// store first key-down
		kRepeatKeyCode = keyDown[0];
		kRepeatChar = keyDownChar[0];
		kRepeatCount = 0;
		kRepeatTimeLast = 0;
		kRepeatTimeAccum = 0;
	}

	// repeat last key
	else if(kRepeatKeyCode != Key_Count) {
		kRepeatTimeAccum += secsTimeDelta;

		if(pKeyboard->KeyPressed(kRepeatKeyCode)) {
			float testPause = kRepeatCount ? fKeyboardRepeatPause : fKeyboardRepeatDelay;

			if(kRepeatTimeAccum >= kRepeatTimeLast + testPause) {
				kRepeatTimeLast = kRepeatTimeAccum;
				kRepeatCount++;

				NotifyKeyboardKey(kRepeatKeyCode, true, kRepeatCount, kRepeatChar);
			}
		}
		else {
			kRepeatKeyCode = Key_Count;
		}
	}

	// dispatch key-up events
	{for(uint i = 0; i < numKeyUp; i++) {
		NotifyKeyboardKey(keyUp[i], false, 0, keyUpChar[i]);
	}}
}


//====================================================================
void  Manager::NotifyKeyboardKey(eKey keyCode, bool bDown, uint repeatCount, char ch)
{
	Message_KeyboardKey  msg(keyCode, bDown, repeatCount, ch);

	// dispatch message to branch
	Component* c = pKeyboardFocusTarget;
	while(c) {
		if(c->DispatchMessage(msg)) return;
		c = c->GetParent();
	}
}


//====================================================================
void  Manager::DispatchMouseEvents()
{
	// check mouse state
	if(pMouse->IsActiveRelativeMode()  ||  ! pMouse->IsPositionInsideAppWindowClientArea()) {

		// mouse-event stream lost, auto-exit command-mode
		if(pCommandTarget) {
			EndCommandMode();
		}

		// auto-exit capture-mode
		if(IsMouseCaptured()) {
			EndMouseCapture();
		}

		// broadcast mouse-exit
		if(bLastMousePosInsideAppWindow) {
			bLastMousePosInsideAppWindow = false;

			if(! pMouse->IsActiveRelativeMode()) {
				DisplaySystemCursor(true);
			}
			NotifyMouseExit(GetMousePosition());
		}
		return;
	}

	// mouse-event stream acquired, update some states
	if(! bLastMousePosInsideAppWindow) {
		bLastMousePosInsideAppWindow = true;

		DisplaySystemCursor(bShowSystemCursor);
		InvalidateHitTestCache();
	}

	// cache mouse-movements
	vec2  mousePosCurr  = GetMousePosition();
	vec2  mousePosDelta = mousePosCurr - mousePosLast;
	bool  bMouseMoved   = ! mousePosDelta.IsZero();

	mousePosLast = mousePosCurr;

	// cache mouse-button events
	ProcessMouseButtons();

	// handle capture-mode (auto-end on button-event, report mouse-movement only)
	if(IsMouseCaptured()) {
		if(mNumEvents) {
			EndMouseCapture();
			return;
		}
		if(bMouseMoved) {
			NotifyMouseMove(*pMouseCaptureTarget, mousePosCurr);
		}
		return;
	}

	// update hover-time
	if(bMouseMoved  ||  mNumClicks) {
		secsMouseHover = 0;
	}
	else {
		secsMouseHover += secsTimeDelta;
	}

	// Reset forced-keyboard-target cache before dispatching mouse-buttons.
	// Used to disable auto-change-keyboard-focus if hot component requested
	// keyboard-focus for another component while processing mouse-buttons.
	pForcedKeyboardTarget = 0;

	// handle command-mode
	if(pCommandTarget) {

		// allow double-click event on command-target
		if(mDoubleClickEvent) {
			pMouseButtonNotifyTarget = pCommandTarget;
			NotifyMouseButtons(mousePosCurr);
		}

		// update states changed by command while mouse-button is pressed
		if(pMouse->ButtonPressed(commandButton)) {

			if(! bMouseMoved) return;

			vec2  grabDelta = mousePosCurr - grabPosScreen;
			vec2  newSize   = pCommandTarget->GetSize();
			vec2  newPos    = pCommandTarget->GetPosition();

			switch(commandArea) {
				case CommandArea_Move: {
					newPos = grabPosTargetOriginal + grabDelta;
					break;
				}
				case CommandArea_ResizeN: {
					newSize = pCommandTarget->CalcMinMaxClampedSize(grabSizeTargetOriginal - vec2(0, grabDelta.y));
					newPos  = grabPosTargetOriginal + (grabSizeTargetOriginal - newSize);
					break;
				}
				case CommandArea_ResizeS: {
					newSize = pCommandTarget->CalcMinMaxClampedSize(grabSizeTargetOriginal + vec2(0, grabDelta.y));
					break;
				}
				case CommandArea_ResizeW: {
					newSize = pCommandTarget->CalcMinMaxClampedSize(grabSizeTargetOriginal - vec2(grabDelta.x, 0));
					newPos  = grabPosTargetOriginal + (grabSizeTargetOriginal - newSize);
					break;
				}
				case CommandArea_ResizeE: {
					newSize = pCommandTarget->CalcMinMaxClampedSize(grabSizeTargetOriginal + vec2(grabDelta.x, 0));
					break;
				}
				case CommandArea_ResizeNW: {
					newSize = pCommandTarget->CalcMinMaxClampedSize(grabSizeTargetOriginal - vec2(grabDelta.x, grabDelta.y));
					newPos  = grabPosTargetOriginal + (grabSizeTargetOriginal - newSize);
					break;
				}
				case CommandArea_ResizeNE: {
					newSize = pCommandTarget->CalcMinMaxClampedSize(grabSizeTargetOriginal + vec2(grabDelta.x, -grabDelta.y));
					newPos  = vec2(grabPosTargetOriginal.x, grabPosTargetOriginal.y + (grabSizeTargetOriginal.y - newSize.y));
					break;
				}
				case CommandArea_ResizeSW: {
					newSize = pCommandTarget->CalcMinMaxClampedSize(grabSizeTargetOriginal + vec2(-grabDelta.x, grabDelta.y));
					newPos  = vec2(grabPosTargetOriginal.x + (grabSizeTargetOriginal.x - newSize.x), grabPosTargetOriginal.y);
					break;
				}
				case CommandArea_ResizeSE: {
					newSize = pCommandTarget->CalcMinMaxClampedSize(grabSizeTargetOriginal + vec2(grabDelta.x, grabDelta.y));
					break;
				}
			}

			// send update-notification/get update-confirmation
			eGeometryUpdateEvent updateEvent =
					(commandArea == CommandArea_Move) ?
					GeometryUpdateEvent_Moving : GeometryUpdateEvent_Resizing;

			Message_GeometryUpdateByManager  msg(
					updateEvent, commandArea, commandButton, newPos, newSize, true);

			pCommandTarget->DispatchMessage(msg);

			// execute update if confirmed
			if(msg.GetConfirmEvent()) {

				pCommandTarget->SetSize(msg.GetNewSize());
				pCommandTarget->SetPosition(msg.GetNewPosition());

				if(commandArea == CommandArea_Move) {
					pCommandTarget->UpdateAbsolutePositionCache();
				}
				else {
					pCommandTarget->UpdateGeometry();
				}
			}
			return;
		}
		// exit command-mode on mouse-button release
		else {
			EndCommandMode();
		}
	}

	// update hit-test records
	if(bMouseMoved  ||  bForceHitTest) {
		bForceHitTest = false;
		bMouseMoved = true;

		HitTestCollect(mousePosCurr, hitRecordsCurr, hitRecordDeepCurr);

		if(hitRecordDeepCurr.component != hitRecordDeepLast.component) {
			ResetMouseRepeat();
			ResetMouseDoubleClick();
		}
	}

	// mouse-over / mouse-move notifications
	if(bMouseMoved) {
		ExtractHitRecordDifferences();
		NotifyMouseOver(mousePosCurr);
		NotifyMouseMove(*hitRecordDeepCurr.component, mousePosCurr);
	}
	else if(secsMouseHover) {
		NotifyMouseHover(*hitRecordDeepCurr.component, mousePosCurr);
	}

	// clicked/declicked branch notifications
	if(mNumClicks) {
		ExtractMouseClickedBranch();
		ExtractMouseClickedBranchDifferences();

		bAllowExternalTopLevelActivate = false;

		DispatchMouseClickToBranches(mousePosCurr);

		bAllowExternalTopLevelActivate = true;
	}

	// determine cursor-area & update active cursor
	eCommandArea cursorArea = hitRecordDeepCurr.component->HitTestCommandArea(
					CalcRelativeMousePos(*hitRecordDeepCurr.component, mousePosCurr));

	SetupActiveCursor(*hitRecordDeepCurr.component, cursorArea, mousePosCurr);

	// check for command-begin
	if(cursorArea != CommandArea_Client  &&  mFirstClickedButtonCurr != MouseButton_Count)
	{
		commandButton = mFirstClickedButtonCurr;

		Component& c = hitRecordDeepCurr.component->GetTopLevelAncestor();
		if(c.IsTopLevel()) {

			// command-begin query
			Message_GeometryUpdateByManager  msg(
				GeometryUpdateEvent_QueryBegin, cursorArea, commandButton, c.GetPosition(), c.GetSize(), false);

			c.DispatchMessage(msg);
			if(msg.GetConfirmEvent()) {

				// found command-target
				grabPosScreen = mousePosCurr;
				commandArea = cursorArea;
				pCommandTarget = &c;
			}
		}

		// execute command-begin
		if(pCommandTarget) {

			ResetKeyboardRepeat();
			ResetMouseRepeat();

			grabPosTargetOriginal = pCommandTarget->GetPosition();
			grabSizeTargetOriginal = pCommandTarget->GetSize();

			// command-begin notification
			switch(commandArea) {

				case CommandArea_Move: {
					pCommandTarget->DispatchMessage(Message_GeometryUpdateByManager(
											GeometryUpdateEvent_MoveBegin, commandArea));

					pCommandTarget->SetActiveGeomUpdateEvent(GeometryUpdateEvent_Moving, commandArea);
					break;
				}
				case CommandArea_ResizeN:
				case CommandArea_ResizeS:
				case CommandArea_ResizeW:
				case CommandArea_ResizeE:
				case CommandArea_ResizeNW:
				case CommandArea_ResizeNE:
				case CommandArea_ResizeSW:
				case CommandArea_ResizeSE: {
					pCommandTarget->DispatchMessage(Message_GeometryUpdateByManager(
											GeometryUpdateEvent_ResizeBegin, commandArea));

					pCommandTarget->SetActiveGeomUpdateEvent(GeometryUpdateEvent_Resizing, commandArea);
					break;
				}
			}

			// update keyboard focus
			if(! pForcedKeyboardTarget) {
				ChangeKeyboardFocusTarget(*hitRecordDeepCurr.component, true, commandButton);
			}
		}
	}

	// dispatch active-mouse-events
	if(! pCommandTarget) {

		NotifyMouseWheel(*hitRecordDeepCurr.component, mousePosCurr);

		pMouseButtonNotifyTarget = hitRecordDeepCurr.component;
		NotifyMouseButtons(mousePosCurr);

		// restore double-click
		if(mDoubleClickEvent) {
			mDoubleClickButtonCode = MouseButton_Count;
			mDoubleClickEvent = false;
		}
	}

	// auto-change keyboard focus on mouse-click
	if(mNumClicks  &&  ! pForcedKeyboardTarget) {
		if(hitRecordDeepCurr.component != pKeyboardFocusTarget) {
			ChangeKeyboardFocusTarget(*hitRecordDeepCurr.component, true, mFirstClickedButtonCurr);
		}
	}

	// store last results
	if(bMouseMoved) {
		hitRecordsLast = hitRecordsCurr;
		hitRecordDeepLast = hitRecordDeepCurr;
	}
	if(mNumClicks) {
		mouseClickedBranchLast = mouseClickedBranchCurr;
		mFirstClickedButtonLast = mFirstClickedButtonCurr;
	}
}


//====================================================================
void  Manager::NotifyMouseExit(const vec2& mousePos)
{
	for(uint i = 0; i < componentsAll.size(); i++) {
		Component& c = *componentsAll[i];
		ReportMouseOver(c, false, false, CalcRelativeMousePos(c, mousePos));
	}
	ReportMouseOver(*pComponentToApplicationAdapter, false, false, mousePos);
}


//====================================================================
void  Manager::NotifyMouseOver(const vec2& mousePos)
{
	{for(uint i = 0; i < hitRecordsDiffExit.size(); i++) {
		const HitRecord& hr = hitRecordsDiffExit[i];
		ReportMouseOver(*hr.component, false, false,
						CalcRelativeMousePos(*hr.component, mousePos));
	}}
	{for(uint i = 0; i < hitRecordsDiffEnter.size(); i++) {
		const HitRecord& hr = hitRecordsDiffEnter[i];
		ReportMouseOver(*hr.component, true, ! hr.bDescendantHit,
						CalcRelativeMousePos(*hr.component, mousePos));
	}}
}


//====================================================================
void  Manager::ReportMouseOver(Component& c, bool bOver, bool bOverDirect, const vec2& mousePos)
{
	bool bChanged = (bOver != c.IsMouseOver())  ||
					(bOver  &&  (bOverDirect != c.IsMouseOverDirect()));

	if(! bOver) bOverDirect = false;

	if(bChanged) {
		c.SetMouseOver(bOver, bOverDirect);
		c.DispatchMessage(Message_MouseEnterExit(bOver, bOverDirect, mousePos));
	}
}


//====================================================================
void  Manager::NotifyMouseMove(Component& c, const vec2& mousePos)
{
	c.DispatchMessage(Message_MouseMove(false, 0, CalcRelativeMousePos(c, mousePos)));
}


//====================================================================
void  Manager::NotifyMouseHover(Component& c, const vec2& mousePos)
{
	c.DispatchMessage(Message_MouseMove(true, secsMouseHover, CalcRelativeMousePos(c, mousePos)));
}


//====================================================================
void  Manager::NotifyMouseWheel(Component& target, const vec2& mousePos)
{
	int wheelDelta = pMouse->GetDeltaWheel();
	if(wheelDelta) {

		Message_MouseWheel msg(wheelDelta);

		Component* c = &target;
		while(c) {
			// disabled components cannot receive active-events
			if(c->IsEnabled()) {
				msg.SetMousePosition(CalcRelativeMousePos(*c, mousePos));
				if(c->DispatchMessage(msg)) return;
			}
			c = c->GetParent();
		}
	}
}


//====================================================================
void  Manager::NotifyMouseButtons(const vec2& mousePos)
{
	// disabled components cannot receive active-events
	if(! pMouseButtonNotifyTarget->IsEnabled()) return;

	Message_MouseButton msg(MouseButton_Left, true, false, 0,
				CalcRelativeMousePos(*pMouseButtonNotifyTarget, mousePos));

	// dispatch mouse-button-down
	if(mRepeatCount) {
		msg.SetButtonCode(mRepeatButtonCode);
		msg.SetRepeatCount(mRepeatCount);
		pMouseButtonNotifyTarget->DispatchMessage(msg);
		return;
	}
	if(mDoubleClickEvent) {
		mDoubleClickEvent = false;
		msg.SetButtonCode(mDoubleClickButtonCode);
		msg.SetDoubleClick(true);
		pMouseButtonNotifyTarget->DispatchMessage(msg);
		return;
	}

	{for(uint i = 0; i < MouseButton_Count; i++) {
		if(mButtonDown[i]) {
			msg.SetButtonCode(eMouseButton(i));
			pMouseButtonNotifyTarget->DispatchMessage(msg);
		}
	}}

	// dispatch mouse-button-up
	msg.SetButtonDown(false);

	{for(uint i = 0; i < MouseButton_Count; i++) {
		if(mButtonUp[i]) {
			msg.SetButtonCode(eMouseButton(i));
			pMouseButtonNotifyTarget->DispatchMessage(msg);
		}
	}}
}


//====================================================================
void  Manager::NotifyActivate(const COMPONENT_PTR_VECTOR& components,
							bool bByMouse, const vec2& mousePos)
{
	for(uint i = 0; i < components.size(); i++) {
		Component& c = *components[i];

		// activate notify/query
		Message_Activate msg(
				bByMouse,
				mFirstClickedButtonCurr,
				CalcRelativeMousePos(c, mousePos),
				(&c != hitRecordDeepCurr.component) ? hitRecordDeepCurr.component : 0,
				true, true);

		c.DispatchMessage(msg);

		// auto-activate
		if(msg.GetAutoActivate()) {
			c.SetActive(true);
		}

		// auto bring-to-front top-level-components
		if(c.IsTopLevel()  &&  msg.GetAutoBringToFront()) {
			BringToFront(c);
		}
	}
}


//====================================================================
void  Manager::NotifyDeactivate(const COMPONENT_PTR_VECTOR& components,
								Component* cIgnore, bool bByMouse)
{
	Message_Deactivate  msg(bByMouse);

	for(uint i = 0; i < components.size(); i++) {
		Component* c = components[i];

		if(c == cIgnore) continue;

		c->DispatchMessage(msg);
		c->SetActive(false);
	}
}


//====================================================================
void  Manager::NotifyDeactivate(Component& c, bool bByMouse)
{
	c.DispatchMessage(Message_Deactivate(bByMouse));
	c.SetActive(false);
}


//====================================================================
bool  Manager::IsModalTargetDescendant(Component& c)
{
	return (&c.GetTopLevelAncestor() == pModalTarget);
}


//====================================================================
void  Manager::EndMouseCapture()
{
	if(pMouseCaptureTarget) {
		pMouseCaptureTarget->DispatchMessage(Message_MouseCapture(false));
		pMouseCaptureTarget = 0;
	}

	bForceHitTest = true;
	secsMouseHover = 0;
}


//====================================================================
void  Manager::EndCommandMode()
{
	if(pCommandTarget) {
		pCommandTarget->SetActiveGeomUpdateEvent();

		// command-end notification
		switch(commandArea) {

			case CommandArea_Move:
				pCommandTarget->DispatchMessage(Message_GeometryUpdateByManager(
									GeometryUpdateEvent_MoveEnd, commandArea));
				break;

			case CommandArea_ResizeN:
			case CommandArea_ResizeS:
			case CommandArea_ResizeW:
			case CommandArea_ResizeE:
			case CommandArea_ResizeNW:
			case CommandArea_ResizeNE:
			case CommandArea_ResizeSW:
			case CommandArea_ResizeSE:
				pCommandTarget->DispatchMessage(Message_GeometryUpdateByManager(
									GeometryUpdateEvent_ResizeEnd, commandArea));
				break;
		}
	}

	pCommandTarget = 0;
	commandArea = CommandArea_Client;

	bForceHitTest = true;
	secsMouseHover = 0;

	ResetKeyboardRepeat();
	ResetMouseRepeat();
}


//====================================================================
void  Manager::ChangeKeyboardFocusTarget(Component& c, bool bByMouse, eMouseButton buttonCode)
{
	if(pKeyboardFocusTarget == &c) return;

	// search focus-branch for first enabled component that wants focus
	Component* pDeepTarget = 0;
	bool       bNeedCharacterData = false;

	Message_KeyboardQueryFocus msg(bByMouse, buttonCode, false, false);

	Component* pTarget = &c;
	while(pTarget) {
		if(pTarget->IsEnabled()) {

			// Assume the following defaults:
			// 1) application wants focus
			// 2) other components do not want focus

			msg.SetWantFocus(pTarget == pComponentToApplicationAdapter);
			msg.SetNeedCharacterData(false);

			pTarget->DispatchMessage(msg);

			if(msg.GetWantFocus()) {
				if(! pDeepTarget) {
					pDeepTarget = pTarget;
				}
				if(msg.GetNeedCharacterData()) {
					bNeedCharacterData = true;
				}
			}
		}
		pTarget = pTarget->GetParent();
	}

	// if branch does not want focus
	if(! pDeepTarget) {

		// leave it as is if both targets have same top-level-ancestor
		if(&c.GetTopLevelAncestor() == &pKeyboardFocusTarget->GetTopLevelAncestor()) {
			pDeepTarget = pKeyboardFocusTarget;
		}
		// otherwise fallback to null-target
		else {
			pDeepTarget = pNullTarget;
			bNeedCharacterData = false;
		}
	}
	// if modal-target is active and deep-target is not in active modal-chain, fallback to null-target
	if(pModalTarget  &&  ! IsModalTargetDescendant(*pDeepTarget)) {
		pDeepTarget = pNullTarget;
		bNeedCharacterData = false;
	}
	if(pDeepTarget == pKeyboardFocusTarget) return;

	// lost-focus notification
	if(pKeyboardFocusTarget) {
		pKeyboardFocusTarget->SetFocused(false);
		pKeyboardFocusTarget->DispatchMessage(Message_KeyboardFocus(false));
	}

	// do change target
	pKeyboardFocusTarget = pDeepTarget;
	bTargetNeedsCharacterData = bNeedCharacterData;

	// got-focus notification
	if(pKeyboardFocusTarget) {
		pKeyboardFocusTarget->SetFocused(true);
		pKeyboardFocusTarget->DispatchMessage(Message_KeyboardFocus(true));
	}

	// additional keyboard setup
	SetupKeyboardForFocusTarget();
}


//====================================================================
void  Manager::SetupKeyboardForFocusTarget()
{
	if(pKeyboard) {
		pKeyboard->EnableCharacterMapping(bTargetNeedsCharacterData);
	}
	ResetKeyboardRepeat();
}


//====================================================================
void  Manager::ResetKeyboardRepeat()
{
	kRepeatKeyCode = Key_Count;
	kRepeatChar = 0;
	kRepeatCount = 0;
	kRepeatTimeLast = 0;
	kRepeatTimeAccum = 0;
}


//====================================================================
void  Manager::ResetMouseRepeat()
{
	mRepeatButtonCode = MouseButton_Count;
	mRepeatCount = 0;
	mRepeatTimeLast = 0;
	mRepeatTimeAccum = 0;
}


//====================================================================
void  Manager::ResetMouseDoubleClick()
{
	mDoubleClickButtonCode = MouseButton_Count;
	mDoubleClickEvent = false;
	mDoubleClickTimeAccum = 0;
}


//====================================================================
void  Manager::ResetMouseButtonEvents()
{
	mNumEvents = 0;
	mNumClicks = 0;
	mFirstClickedButtonCurr = MouseButton_Count;

	for(uint i = 0; i < MouseButton_Count; i++) {
		mButtonDown[i] = 0;
		mButtonUp[i] = 0;
	}

	ResetMouseRepeat();
	ResetMouseDoubleClick();
}


//====================================================================
void  Manager::ProcessMouseButtons()
{
	mNumEvents = 0;
	mNumClicks = 0;
	mFirstClickedButtonCurr = MouseButton_Count;

	// accumulate events
	for(uint i = 0; i < MouseButton_Count; i++) {
		eMouseButton mc = eMouseButton(i);

		mButtonDown[i] = pMouse->ButtonDown(mc);
		mButtonUp[i] = pMouse->ButtonUp(mc);

		if(mButtonDown[i]) {
			if(! mNumClicks) mFirstClickedButtonCurr = mc;
			mNumClicks++;
			mNumEvents++;
		}
		else if(mButtonUp[i]) {
			mNumEvents++;
		}
	}

	// handle repeat/double-click
	mDoubleClickTimeAccum += secsTimeDelta;

	if(mNumClicks) {
		// reset button-repeat
		mRepeatButtonCode = mFirstClickedButtonCurr;
		mRepeatCount = 0;
		mRepeatTimeLast = 0;
		mRepeatTimeAccum = 0;

		// track double-click
		if(mNumClicks == 1) {
			if(mDoubleClickButtonCode == mFirstClickedButtonCurr  &&
				mDoubleClickTimeAccum <= fDoubleClickLimit)
			{
				// generate double-click event
				mDoubleClickEvent = true;
				mDoubleClickTimeAccum = fDoubleClickLimit;
			}
			else {
				// restart double-click timing
				mDoubleClickButtonCode = mFirstClickedButtonCurr;
				mDoubleClickEvent = false;
				mDoubleClickTimeAccum = 0;
			}
		}
		else {
			// reset double-click tracking
			mDoubleClickButtonCode = MouseButton_Count;
			mDoubleClickEvent = false;
		}
	}

	// update button-repeat
	else if(mRepeatButtonCode != MouseButton_Count) {
		mRepeatTimeAccum += secsTimeDelta;

		if(pMouse->ButtonPressed(mRepeatButtonCode)) {
			float testPause = mRepeatCount ? fMouseRepeatPause : fMouseRepeatDelay;

			// increment repeat-count
			if(mRepeatTimeAccum >= mRepeatTimeLast + testPause) {
				mRepeatTimeLast = mRepeatTimeAccum;
				mRepeatCount++;
			}
		}
		else {
			// reset repeat-count
			mRepeatButtonCode = MouseButton_Count;
			mRepeatCount = 0;
		}
	}

	// ignore until first button-up
	if(bIgnoreMouseButtonEvents) {
		for(uint i = 0; i < MouseButton_Count; i++) {
			if(mButtonUp[i]) {
				bIgnoreMouseButtonEvents = false;
				break;
			}
		}
		ResetMouseButtonEvents();
	}
}


//====================================================================
void  Manager::InvalidateHitTestCache()
{
	bForceHitTest = true;

	hitRecordsCurr.clear();
	hitRecordsLast.clear();
	hitRecordDeepCurr.Clear();
	hitRecordDeepLast.Clear();

	hitRecordsDiffEnter.clear();
	hitRecordsDiffExit.clear();
}


//====================================================================
void  Manager::HitTestCollect(const vec2& testPointAbs,
								HITRECORD_VECTOR& out, HitRecord& outDeep)
{
	out.clear();
	outDeep.Clear();

	// test modal target first
	if(pModalTarget) {
		HitTestCollect(testPointAbs, out, outDeep, *pModalTarget,
						tsHitRecords, tsDepthFirstPathMarkersHitTest);

		if(out.empty()) {
			// use null-target
			outDeep = HitRecord(pNullTarget, 0, false);
			out.push_back(outDeep);
		}
		return;
	}

	// test layers in order
	for(int i = componentsTopLevel.size() - 1; i >= 0; i--) {

		HitTestCollect(testPointAbs, out, outDeep, *componentsTopLevel[i],
						tsHitRecords, tsDepthFirstPathMarkersHitTest);

		if(out.size()) return;
	}

	// use application-adapter
	outDeep = HitRecord(pComponentToApplicationAdapter, 0, false);
	out.push_back(outDeep);
}


//====================================================================
void  Manager::HitTestCollect(const vec2& testPointAbs,
								HITRECORD_VECTOR& out, HitRecord& outDeep,
								Component& topLevel, HITRECORD_VECTOR& tempHitRecords,
								DEPTHFIRSTPATHMARKERHITTEST_VECTOR& tempPathMarkersHitTest)
{
	HitRecord  hitRecord;
	hitRecord.bDescendantHit = false;

	int  depthLevel = -1;
	int  overlayLevel = 0;
	int  maxHitOverlayLevel = 0;
	int  iLastOverlayHitRecord = -1;

	tempPathMarkersHitTest.clear();
	tempPathMarkersHitTest.push_back(Manager::DepthFirstPathMarkerHitTest(&topLevel, -1));

	while(tempPathMarkersHitTest.size()) {
		Manager::DepthFirstPathMarkerHitTest& marker = tempPathMarkersHitTest.back();

		if(marker.bUnprocessed) {
			// discard invisible branches
			if(marker.component->IsVisible()) {

				// update levels
				depthLevel++;
				if(marker.component->IsOverlay()) overlayLevel++;

				// single component hit-test
				if(marker.component->HitTest(
					testPointAbs - marker.component->GetAbsolutePosition()))
				{
					// update parent hit info
					if(marker.iParentHitRecord >= 0) {
						out[marker.iParentHitRecord].bDescendantHit = true;
					}

					// fill & store hit info
					hitRecord.component = marker.component;
					hitRecord.depth = depthLevel;
					out.push_back(hitRecord);

					// redirect parent index to own hit-record
					marker.iParentHitRecord = out.size() - 1;

					// restrict further testing to deepest overlay branch
					if(marker.component->IsOverlay()) {
						if(overlayLevel > maxHitOverlayLevel) {
							maxHitOverlayLevel = overlayLevel;
							iLastOverlayHitRecord = marker.iParentHitRecord;
						}
					}
				}
				marker.bUnprocessed = false;
			}
			else {
				// backtrack
				tempPathMarkersHitTest.pop_back();
				continue;
			}
		}
		if((overlayLevel >= maxHitOverlayLevel)  &&
			(marker.iNextChild < marker.component->GetNumChildren()))
		{
			// descend to next child
			tempPathMarkersHitTest.push_back(Manager::DepthFirstPathMarkerHitTest(
					&marker.component->GetChild(marker.iNextChild), marker.iParentHitRecord));

			marker.iNextChild++;
		}
		else {
			// update levels
			if(marker.component->IsOverlay()) overlayLevel--;
			depthLevel--;

			// backtrack
			tempPathMarkersHitTest.pop_back();
		}
	}

	// extract deepest overlay branch plus ancestors
	if(iLastOverlayHitRecord >= 0) {
		tempHitRecords.clear();

		// deep branch
		{for(int i = iLastOverlayHitRecord; i < out.size(); i++) {
			tempHitRecords.push_back(out[i]);
		}}

		// ancestors
		Component* parent = out[iLastOverlayHitRecord].component->GetParent();
		int i = iLastOverlayHitRecord - 1;
		while(parent  &&  i >= 0) {
			if(out[i].component == parent) {
				tempHitRecords.push_back(out[i]);
				parent = out[i].component->GetParent();
			}
			i--;
		}
		out = tempHitRecords;
	}

	// extract deep hit-record
	int maxDepth = -1;

	for(uint i = 0; i < out.size(); i++) {
		const HitRecord& hr = out[i];

		if(hr.depth > maxDepth) {
			maxDepth = hr.depth;
			outDeep = hr;
		}
	}
}


//====================================================================
void  Manager::ExtractHitRecordDifferences()
{
	// extract enter-event targets
	hitRecordsDiffEnter.clear();
	hitRecordsDiffEnter.reserve(hitRecordsCurr.size());

	{for(uint i = 0; i < hitRecordsCurr.size(); i++) {
		const HitRecord& hrCurr = hitRecordsCurr[i];
		bool bDiff = true;

		for(uint j = 0; j < hitRecordsLast.size(); j++) {
			const HitRecord& hrLast = hitRecordsLast[j];

			if(hrCurr.component == hrLast.component) {
				// enter-event must also be generated when number-of-descedant-hits changes
				if(hrCurr.bDescendantHit == hrLast.bDescendantHit) {
					bDiff = false;
				}
				break;
			}
		}
		if(bDiff) {
			hitRecordsDiffEnter.push_back(hitRecordsCurr[i]);
		}
	}}

	// extract exit-event targets
	hitRecordsDiffExit.clear();
	hitRecordsDiffExit.reserve(hitRecordsLast.size());

	{for(uint i = 0; i < hitRecordsLast.size(); i++) {
		Component* comp = hitRecordsLast[i].component;
		bool bDiff = true;

		for(uint j = 0; j < hitRecordsCurr.size(); j++) {
			if(comp == hitRecordsCurr[j].component) {
				bDiff = false;
				break;
			}
		}
		if(bDiff) {
			hitRecordsDiffExit.push_back(hitRecordsLast[i]);
		}
	}}
}


//====================================================================
void  Manager::InvalidateMouseClickedBranchCache()
{
	mouseClickedBranchCurr.clear();
	mouseClickedBranchLast.clear();
	mouseClickedBranchDiffCurrNotLast.clear();
	mouseClickedBranchDiffLastNotCurr.clear();

	mFirstClickedButtonLast = MouseButton_Count;
}


//====================================================================
void  Manager::SetupMouseClickedBranchCache(COMPONENT_PTR_VECTOR& clickedTargets,
											eMouseButton clickedButton)
{
	InvalidateMouseClickedBranchCache();

	// store clicked components
	mouseClickedBranchLast = clickedTargets;
	mouseClickedBranchCurr = mouseClickedBranchLast;
	mFirstClickedButtonLast = clickedButton;
}


//====================================================================
void  Manager::ExtractMouseClickedBranch()
{
	mouseClickedBranchCurr.clear();

	Component* c = hitRecordDeepCurr.component;
	while(c) {
		mouseClickedBranchCurr.push_back(c);
		c = c->GetParent();
	}
}


//====================================================================
void  Manager::ExtractMouseClickedBranchDifferences()
{
	// extract new-clicked targets
	if(mFirstClickedButtonCurr == mFirstClickedButtonLast) {

		mouseClickedBranchDiffCurrNotLast.clear();
		mouseClickedBranchDiffCurrNotLast.reserve(mouseClickedBranchCurr.size());

		for(uint i = 0; i < mouseClickedBranchCurr.size(); i++) {
			Component* c = mouseClickedBranchCurr[i];
			bool bDiff = true;

			for(uint j = 0; j < mouseClickedBranchLast.size(); j++) {
				if(c == mouseClickedBranchLast[j]) {
					bDiff = false;
					break;
				}
			}
			if(bDiff) {
				mouseClickedBranchDiffCurrNotLast.push_back(c);
			}
		}
	}
	else {
		mouseClickedBranchDiffCurrNotLast = mouseClickedBranchCurr;
	}

	// extract de-clicked targets
	mouseClickedBranchDiffLastNotCurr.clear();
	mouseClickedBranchDiffLastNotCurr.reserve(mouseClickedBranchLast.size());

	{for(uint i = 0; i < mouseClickedBranchLast.size(); i++) {
		Component* c = mouseClickedBranchLast[i];
		bool bDiff = true;

		for(uint j = 0; j < mouseClickedBranchCurr.size(); j++) {
			if(c == mouseClickedBranchCurr[j]) {
				bDiff = false;
				break;
			}
		}
		if(bDiff) {
			mouseClickedBranchDiffLastNotCurr.push_back(c);
		}
	}}
}


//====================================================================
void  Manager::DispatchMouseClickToBranches(const vec2& mousePos)
{
	// notify declicked components
	NotifyDeactivate(mouseClickedBranchDiffLastNotCurr, 0, true);

	// notify new-clicked components
	NotifyActivate(mouseClickedBranchDiffCurrNotLast, true, mousePos);
}


//====================================================================
void  Manager::EraseFromHitTestCache(Component& c)
{
	NullifyHitTestCache(hitRecordsLast, c);
	NullifyHitTestCache(hitRecordsCurr, c);
	NullifyHitTestCache(hitRecordsDiffExit, c);
	NullifyHitTestCache(hitRecordsDiffEnter, c);
	NullifyHitTestCache(hitRecordDeepLast, c);
	NullifyHitTestCache(hitRecordDeepCurr, c);
}


//====================================================================
void  Manager::NullifyHitTestCache(HitRecord& hr, Component& c)
{
	if(hr.component == &c) {
		hr.component = pNullTarget;
	}
}


//====================================================================
void  Manager::NullifyHitTestCache(HITRECORD_VECTOR& hrv, Component& c)
{
	for(uint i = 0; i < hrv.size(); i++) {
		NullifyHitTestCache(hrv[i], c);
	}
}


//====================================================================
void  Manager::EraseFromMouseClickedBranchCache(Component& c)
{
	EraseFromContainer(mouseClickedBranchLast, &c);
	EraseFromContainer(mouseClickedBranchCurr, &c);
	EraseFromContainer(mouseClickedBranchDiffLastNotCurr, &c);
	EraseFromContainer(mouseClickedBranchDiffCurrNotLast, &c);
}


//====================================================================
void  Manager::SetupActiveCursor(Component& c, eCommandArea cursorArea, const vec2& mousePos)
{
	// check if component wants to use custom cursor
	Message_MouseQueryCursor  msg(
		cursorArea, CalcRelativeMousePos(c, mousePos), 0);

	c.DispatchMessage(msg);
	pActiveCursor = msg.GetCursor();

	if(! pActiveCursor) {
		// use default
		pActiveCursor = cursorHandles[cursorArea];
	}
}


//====================================================================
void  Manager::NotifyObjectCreateDelete(BaseShader& s, bool bCreate)
{
	if(bCreate) {
		if(bAllowMainContainerUpdateAtNotifyShader) {
			UniqueContainerPushBack(shaders, &s);
			std::sort(shaders.begin(), shaders.end(), Predicate_SortByName<BaseShader>());
		}
		UniqueContainerPushBack(tileTextureRecacheQueue, &s);
		UniqueContainerPushBack(fontRecacheQueue, &s);
	}
	else {
		if(bAllowMainContainerUpdateAtNotifyShader) {
			EraseFromContainer(shaders, &s);
		}
		EraseFromContainer(tileTextureRecacheQueue, &s);
		EraseFromContainer(fontRecacheQueue, &s);
		TimeNotifyRegister(s, false);
	}
	bNeedFullShaderRecache = true;
}


//====================================================================
void  Manager::NotifyObjectCreateDelete(Cursor& c, bool bCreate)
{
	if(bCreate) {
		if(bAllowMainContainerUpdateAtNotifyCursor) {
			UniqueContainerPushBack(cursors, &c);
			std::sort(cursors.begin(), cursors.end(), Predicate_SortByName<Cursor>());
		}
		UniqueContainerPushBack(tileTextureRecacheQueue, &c);
		UniqueContainerPushBack(fontRecacheQueue, &c);
	}
	else {
		if(bAllowMainContainerUpdateAtNotifyCursor) {
			EraseFromContainer(cursors, &c);
		}
		EraseFromContainer(tileTextureRecacheQueue, &c);
		EraseFromContainer(fontRecacheQueue, &c);
		TimeNotifyRegister(c, false);
	}
	bNeedFullCursorRecache = true;
}


//====================================================================
void  Manager::NotifyObjectCreateDelete(Component& c, bool bCreate)
{
	bComponentLayersChanged = true;
	bComponentHierarchyChanged = true;

	if(bCreate) {
		if(bAllowMainContainerUpdateAtNotifyComponent) {
			UniqueContainerPushBack(componentsTopLevel, &c);
		}
		if(bCycleQueueAutoInsertAtCreate) {
			UniqueContainerPushBack(componentsCycleQueue, &c);
		}
		UniqueContainerPushBack(tileTextureRecacheQueue, &c);
		UniqueContainerPushBack(fontRecacheQueue, &c);
		UniqueContainerPushBack(componentsWindowResize, &c);
		ReportDirtyGeometry(c);
	}
	else {
		if(bAllowMainContainerUpdateAtNotifyComponent) {
			EraseFromContainer(componentsTopLevel, &c);
		}
		TimeNotifyRegister(c, false);
		EraseFromHitTestCache(c);
		EraseFromMouseClickedBranchCache(c);
		EraseFromContainer(tileTextureRecacheQueue, &c);
		EraseFromContainer(fontRecacheQueue, &c);
		EraseFromContainer(shaderRecacheQueue, &c);
		EraseFromContainer(cursorRecacheQueue, &c);
		EraseFromContainer(componentsAll, &c);
		EraseFromContainer(componentsDrawQueue, &c);
		EraseFromContainer(componentsDirtyGeometry, &c);
		EraseFromContainer(componentsWindowResize, &c);
		EraseFromContainer(componentsCycleQueue, &c);

		EraseFromContainer(modalTargetStack, &c);
		if(pModalTarget == &c) {
			if(modalTargetStack.size()) {
				pModalTarget = modalTargetStack.back();
			}
			else {
				pModalTarget = 0;
			}
		}
		if(pCommandTarget == &c) {
			pCommandTarget = 0;
		}
		if(pKeyboardFocusTarget == &c) {
			pKeyboardFocusTarget = pNullTarget;
		}
		if(pMouseButtonNotifyTarget == &c) {
			pMouseButtonNotifyTarget = pNullTarget;
		}
	}
}


//====================================================================
void  Manager::NotifyComponentHierarchyChanged(Component& c)
{
	bComponentHierarchyChanged = true;

	// main container update
	if(bAllowMainContainerUpdateAtNotifyComponent) {
		if(c.IsTopLevel()) {
			UniqueContainerPushBack(componentsTopLevel, &c);
		}
		else {
			EraseFromContainer(componentsTopLevel, &c);
		}
	}

	// cycle-queue update
	if(c.IsTopLevel()) {
		if(bCycleQueueAutoInsertAtHierarchyChange) {
			UniqueContainerPushBack(componentsCycleQueue, &c);
		}
	}
	else {
		EraseFromContainer(componentsCycleQueue, &c);
	}
	ReportDirtyGeometry(c);
}


//====================================================================
void  Manager::NotifyComponentLayerChanged(Component& c)
{
	bComponentLayersChanged = true;
}


//====================================================================
void  Manager::NotifyComponentOverlayChanged(Component& c)
{
	bComponentHierarchyChanged = true;
	ReportDirtyGeometry(c);
}


//====================================================================
void  Manager::NotifyComponentVisibilityChanged(Component& c)
{
	bComponentHierarchyChanged = true;
	ReportDirtyGeometry(c);

	if(bAllowExternalTopLevelActivate) {
		// auto-activate next if active top-level is hidden
		if(c.IsTopLevel()  &&  c.IsActive()  &&  ! c.IsVisible()) {
			CycleQueueActivate(true);
		}
	}
}


//====================================================================
void  Manager::NotifyNameChanged(BaseShader& s)
{
	bNeedFullShaderRecache = true;
}


//====================================================================
void  Manager::NotifyNameChanged(Cursor& c)
{
	bNeedFullCursorRecache = true;
}


//====================================================================
void  Manager::NotifyGlobalCursorParamsChanged()
{
	bNeedGlobalCursorParamsUpdate = true;
}


//====================================================================
void  Manager::ConnectApplication(bool bConnect)
{
	pComponentToApplicationAdapter->ConnectApplication(bConnect);
}


//====================================================================
void  Manager::NotifyApplicationWindowActive(bool bActive)
{
	bIgnoreMouseButtonEvents = bIgnoreMouseClickAfterWindowActivate;

	if(bActive) {
		ResetMouseRepeat();
		ResetMouseDoubleClick();
		ResetKeyboardRepeat();
	}
}


//====================================================================
void  Manager::NotifyApplicationWindowCreated()
{
	bIgnoreMouseButtonEvents = bIgnoreMouseClickAfterWindowCreate;
}


//====================================================================
void  Manager::BeginConnectHumanInterfaceDevices()
{
}


//====================================================================
void  Manager::ConnectHumanInterfaceDevice(HumanInterfaceDevice& hid)
{
	// attach only system-keyboard and system-mouse
	if(! pKeyboard) {
		Keyboard* k = InterfaceCast<Keyboard>(hid.FindController(VW_INTERFACE_NAME_Keyboard));
		if(k  &&  k->IsSystemKeyboard()) pKeyboard = k;
	}
	if(! pMouse) {
		Mouse* m = InterfaceCast<Mouse>(hid.FindController(VW_INTERFACE_NAME_Mouse));
		if(m  &&  m->IsSystemMouse()) pMouse = m;
	}
}


//====================================================================
void  Manager::EndConnectHumanInterfaceDevices()
{
	if(! pKeyboard) {
		if(bTolerateMissingInputDevice) {
			olog << e_warning << "gui-manager: missing keyboard-device.\n" << e_info;
		}
		else {
			VW_THROW(GUIException, "gui-manager cannot continue without keyboard-device.");
		}
	}
	if(! pMouse) {
		if(bTolerateMissingInputDevice) {
			olog << e_warning << "gui-manager: missing mouse-device.\n" << e_info;
		}
		else {
			VW_THROW(GUIException, "gui-manager cannot continue without mouse-device.");
		}
	}

	ResetMouseRepeat();
	ResetMouseDoubleClick();
	SetupKeyboardForFocusTarget();
}


//====================================================================
void  Manager::DisconnectHumanInterfaceDevices()
{
	ResetKeyboardRepeat();
	ResetMouseRepeat();
	ResetMouseDoubleClick();

	pKeyboard = 0;
	pMouse = 0;
}


//====================================================================
void  Manager::SetTolerateMissingInputDevice(bool bTolerate)
{
	bTolerateMissingInputDevice = bTolerate;
	olog << e_warning << "Changed gui-manager variable: TolerateMissingInputDevice = " << bTolerateMissingInputDevice << endpara << e_info;
}


//====================================================================
void  Manager::ConnectVideoRenderer(VideoRenderer& vr)
{
	VW_ASSERT(! pVideoRenderer, "Video-renderer already connected to GUI-manager.");

	pVideoRenderer = &vr;
	pGUIRenderer = &pVideoRenderer->GetGUIRenderer();

	bRendererCreatesManagedResourcesOnly = pGUIRenderer->QueryCreatesManagedResourcesOnly();
	olog << "gui-renderer-interface creates managed resources only: " << ToString(bRendererCreatesManagedResourcesOnly) << std::endl;

	NotifyVideoRendererChanged();
}


//====================================================================
void  Manager::DisconnectVideoRenderer()
{
	pVideoRenderer = 0;
	bRendererCreatesManagedResourcesOnly = false;

	NotifyVideoRendererChanged();
}


//====================================================================
void  Manager::CreateVideoResources(bool bPostReset)
{
	VW_ASSERT(pVideoRenderer, "Video-renderer is not connected to GUI-manager.");

	bVideoRendererReadyToCreateResources = true;
	if(bPostReset  &&  bRendererCreatesManagedResourcesOnly) return;

	PrepareVideoRendererResources(true, true);
}


//====================================================================
void  Manager::ReleaseVideoResources(bool bPreReset)
{
	VW_ASSERT(pVideoRenderer, "Video-renderer is not connected to GUI-manager.");

	bVideoRendererReadyToCreateResources = false;
	if(bPreReset  &&  bRendererCreatesManagedResourcesOnly) return;

	MapReleaseVRResources(mapTileTextures);
	MapReleaseVRResources(mapFonts);
}


//====================================================================
void  Manager::OnWindowResize(uint width, uint height)
{
	windowWidth  = ClampedLow<uint>(width, 1);
	windowHeight = ClampedLow<uint>(height, 1);

	pComponentToApplicationAdapter->SetPosition(vec2::ZERO);
	pComponentToApplicationAdapter->SetSize(windowWidth, windowHeight);
	pComponentToApplicationAdapter->UpdateGeometry();

	bNeedFullWindowResizeUpdate = true;
	bNeedFullGeometryUpdate = true;
}


//====================================================================
void  Manager::DoCreateVRRTileTextures(bool bRecreateAll)
{
	if(! bVideoRendererReadyToCreateResources) return;

	if(bRecreateAll) {
		vrrCreateListTileTextures.clear();
		GetRegisteredTileTextureNames(vrrCreateListTileTextures);
	}
	MapCreateVRResources(mapTileTextures, *pVideoRenderer, vrrCreateListTileTextures);
	vrrCreateListTileTextures.clear();
	bTileTextureObjectsChanged = false;
}


//====================================================================
void  Manager::DoCreateVRRFonts(bool bRecreateAll)
{
	if(! bVideoRendererReadyToCreateResources) return;

	if(bRecreateAll) {
		vrrCreateListFonts.clear();
		GetRegisteredFontNames(vrrCreateListFonts);
	}
	MapCreateVRResources(mapFonts, *pVideoRenderer, vrrCreateListFonts);
	vrrCreateListFonts.clear();
	bFontObjectsChanged = false;
}


//====================================================================
void  Manager::CollectAllVRUsers(VRUSER_PTR_VECTOR& vru)
{
	vru.clear();
	std::back_insert_iterator<VRUSER_PTR_VECTOR> it(vru);
	std::copy(shaders.begin(), shaders.end(), it);
	std::copy(cursors.begin(), cursors.end(), it);
	std::copy(componentsAll.begin(), componentsAll.end(), it);
}


//====================================================================
void  Manager::VRUser_CacheTileTextureHandles()
{
	olog << e_debug << "Caching tile-texture handles\n" << e_info;

	for(uint i = 0; i < tileTextureRecacheQueue.size(); i++) {
		VideoResourceUser* vru = tileTextureRecacheQueue[i];

		if(! vru->CacheTileTextureHandles(*this)) {
			VW_THROW(GUIException, "Missing tile-texture for: " + MakeObjectName(*vru));
		}
	}
}


//====================================================================
void  Manager::VRUser_PostCacheTileTextureHandles()
{
	for(uint i = 0; i < tileTextureRecacheQueue.size(); i++) {
		tileTextureRecacheQueue[i]->PostCacheTileTextureHandles();
	}
}


//====================================================================
void  Manager::VRUser_TileTextureListChanged()
{
	VRUSER_PTR_VECTOR vrUsers;
	CollectAllVRUsers(vrUsers);

	for(uint i = 0; i < vrUsers.size(); i++) {
		vrUsers[i]->OnTileTextureListChanged(*this);
	}
}


//====================================================================
void  Manager::VRUser_CacheFontHandles()
{
	olog << e_debug << "Caching font handles\n" << e_info;

	for(uint i = 0; i < fontRecacheQueue.size(); i++) {
		VideoResourceUser* vru = fontRecacheQueue[i];

		if(! vru->CacheFontHandles(*this)) {
			VW_THROW(GUIException, "Missing font for: " + MakeObjectName(*vru));
		}
	}
}


//====================================================================
void  Manager::VRUser_PostCacheFontHandles()
{
	for(uint i = 0; i < fontRecacheQueue.size(); i++) {
		fontRecacheQueue[i]->PostCacheFontHandles();
	}
}


//====================================================================
void  Manager::VRUser_FontListChanged()
{
	VRUSER_PTR_VECTOR vrUsers;
	CollectAllVRUsers(vrUsers);

	for(uint i = 0; i < vrUsers.size(); i++) {
		vrUsers[i]->OnFontListChanged(*this);
	}
}


//====================================================================
void  Manager::Component_CacheCursorHandles()
{
	olog << e_debug << "Caching cursor handles\n" << e_info;

	for(uint i = 0; i < cursorRecacheQueue.size(); i++) {
		Component* c = cursorRecacheQueue[i];

		if(! c->CacheCursorHandles(*this)) {
			VW_THROW(GUIException, "Missing cursor for: " + MakeObjectName(*c));
		}
	}
}


//====================================================================
void  Manager::Component_PostCacheCursorHandles()
{
	for(uint i = 0; i < cursorRecacheQueue.size(); i++) {
		cursorRecacheQueue[i]->PostCacheCursorHandles();
	}
}


//====================================================================
void  Manager::Component_CursorListChanged()
{
	for(uint i = 0; i < componentsAll.size(); i++) {
		componentsAll[i]->OnCursorListChanged(*this);
	}
}


//====================================================================
void  Manager::Component_CacheShaderHandles()
{
	olog << e_debug << "Caching shader handles\n" << e_info;

	for(uint i = 0; i < shaderRecacheQueue.size(); i++) {
		Component* c = shaderRecacheQueue[i];

		if(! c->CacheShaderHandles(*this)) {
			VW_THROW(GUIException, "Missing shader for: " + MakeObjectName(*c));
		}
	}
}


//====================================================================
void  Manager::Component_PostCacheShaderHandles()
{
	for(uint i = 0; i < shaderRecacheQueue.size(); i++) {
		shaderRecacheQueue[i]->PostCacheShaderHandles();
	}
}


//====================================================================
void  Manager::Component_ShaderListChanged()
{
	for(uint i = 0; i < componentsAll.size(); i++) {
		componentsAll[i]->OnShaderListChanged(*this);
	}
}


//====================================================================
void  Manager::Internal_CacheCursorHandles()
{
	using namespace std_cursors;

	cursorHandles[CommandArea_Void]     = GetCursor(CursorName_Void);
	cursorHandles[CommandArea_Client]   = GetCursor(CursorName_Client);
	cursorHandles[CommandArea_Move]     = GetCursor(CursorName_Move);
	cursorHandles[CommandArea_ResizeN]  = GetCursor(CursorName_ResizeN);
	cursorHandles[CommandArea_ResizeS]  = GetCursor(CursorName_ResizeS);
	cursorHandles[CommandArea_ResizeW]  = GetCursor(CursorName_ResizeW);
	cursorHandles[CommandArea_ResizeE]  = GetCursor(CursorName_ResizeE);
	cursorHandles[CommandArea_ResizeNW] = GetCursor(CursorName_ResizeNW);
	cursorHandles[CommandArea_ResizeNE] = GetCursor(CursorName_ResizeNE);
	cursorHandles[CommandArea_ResizeSW] = GetCursor(CursorName_ResizeSW);
	cursorHandles[CommandArea_ResizeSE] = GetCursor(CursorName_ResizeSE);

	// for each cursor: if not found use null-cursor
	for(uint i = 0; i < CommandArea_Count; i++) {
		if(! cursorHandles[i]) {
			cursorHandles[i] = pNullCursor;
		}
	}
}


//====================================================================
void  Manager::Internal_InvalidateCursorHandles()
{
	for(uint i = 0; i < CommandArea_Count; i++) {
		cursorHandles[i] = pNullCursor;
	}
}


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


