

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Memory.h"
#include "BaseTools\Color.h"
#include "MathTools\Vector2.h"
#include "MathTools\Vector3.h"
#include "MathTools\Matrix44.h"
#include "VideoRendererObjects\VRO_Textures.h"
#include "DataObjects\Image.h"
#include "Framework\FrameworkExceptions.h"


//********************************************************************
namespace vertexware {


//====================================================================
class VideoSubsystem;
class VideoRenderer;


//********************************************************************
namespace gui {


//====================================================================
class Renderer;


//====================================================================
// Position, Color.
//====================================================================
class VW_COMMON_API VertexC
{
public:
	vec3  position;
	color4b  color;

public:
	__forceinline VertexC()	{}

	__forceinline VertexC(const vec3& _position, color4b _color)
	{
		position = _position;
		color    = _color;
	}
};


//====================================================================
// Position, Color, TextureCoords.
//====================================================================
class VW_COMMON_API VertexCT
{
public:
	vec3  position;
	color4b  color;
	vec2  texCoord;

public:
	__forceinline VertexCT() {}

	__forceinline VertexCT(const vec3& _position, color4b _color, const vec2& _texCoord)
	{
		position = _position;
		color    = _color;
		texCoord = _texCoord;
	}
};


//====================================================================
// GUI vertex buffers.
//====================================================================
class VW_COMMON_API VertexBufferC : public Buffer<VertexC>
{
public:
	VertexBufferC(uint vertexCount = 0) : Buffer<VertexC>(vertexCount) {}
};


//====================================================================
class VW_COMMON_API VertexBufferCT : public Buffer<VertexCT>
{
public:
	VertexBufferCT(uint vertexCount = 0) : Buffer<VertexCT>(vertexCount) {}
};


//====================================================================
// Primitive type.
//====================================================================
enum ePrimitiveType
{
	PrimitiveType_TriangleList = 0,
	PrimitiveType_TriangleStrip,
	PrimitiveType_TriangleFan,

	PrimitiveType_Count,
	PrimitiveType_Force32Bits = 0x7FFFFFFF
};


//====================================================================
class VW_COMMON_API Rectangle
{
public:
	float  left;
	float  top;
	float  width;
	float  height;

public:
	Rectangle() {}

	Rectangle(float left, float top, float width, float height) { Set(left, top, width, height); }
	Rectangle(const vec2& position, const vec2& size) { Set(position, size); }

	void  Set(float left, float top, float width, float height);
	void  Set(const vec2& position, const vec2& size);

	void  SetPosition(float left, float top);
	void  SetPosition(const vec2& positionTopLeft);

	void  SetSize(float width, float height);
	void  SetSize(const vec2& size);

	void  Offset(float x, float y);
	void  Offset(const vec2& off);

	void  Grow(float delta);
	void  Shrink(float delta);

	void  CopyGeometryParamsFrom(const Rectangle& r);

	bool  HitTest(float x, float y) const;
	bool  HitTest(const vec2& testPoint) const { return HitTest(testPoint.x, testPoint.y); }

	bool  Intersects(const Rectangle& r) const;
	bool  Contains(const Rectangle& r) const;
};


//====================================================================
class VW_COMMON_API RectangleC : public Rectangle
{
public:
	color4b  colorTL;
	color4b  colorTR;
	color4b  colorBL;
	color4b  colorBR;

public:
	RectangleC() {}

	RectangleC(float left, float top, float width, float height) : Rectangle(left, top, width, height) {}
	RectangleC(const vec2& position, const vec2& size) : Rectangle(position, size) {}

	void  SetColor(color4b c);
	void  SetColorLeft(color4b c);
	void  SetColorRight(color4b c);
	void  SetColorTop(color4b c);
	void  SetColorBottom(color4b c);

	void  CopyColorParamsFrom(const RectangleC& r);
};


//====================================================================
class VW_COMMON_API RectangleCT : public RectangleC
{
public:
	// rotation is 2D counter-clockwise
	enum eTexTransform
	{
		TexTransform_Identity = 0,
		TexTransform_FlipH,
		TexTransform_FlipV,
		TexTransform_Rotate90,
		TexTransform_Rotate180,
		TexTransform_Rotate270,

		TexTransform_Count,
		TexTransform_Force32Bits = 0x7FFFFFFF
	};

public:
	float  texCoordLeft;
	float  texCoordRight;
	float  texCoordTop;
	float  texCoordBottom;

	eTexTransform  texTransform;


public:
	RectangleCT() { texTransform = TexTransform_Identity; }

	RectangleCT(float left, float top, float width, float height) : RectangleC(left, top, width, height)
	{ texTransform = TexTransform_Identity; }

	RectangleCT(const vec2& position, const vec2& size) : RectangleC(position, size)
	{ texTransform = TexTransform_Identity; }

	void  SetStandardTexCoords();

	void  SetTexTransform(eTexTransform tt = TexTransform_Identity) { texTransform = tt; }

	void  FillTransformedTexCoords(VertexCT& vertexTL, VertexCT& vertexTR,
									VertexCT& vertexBL, VertexCT& vertexBR) const;

	void  CopyTexCoordParamsFrom(const RectangleCT& r);
};


//====================================================================
struct VW_COMMON_API TextMetrics
{
	float  textHeight;
	bool   bMonospace;
	bool   bUseFontDefaultGlyphSpacing;
	float  customGlyphSpacing;
	float  scaleGlyphWidth; //relative to text-height
	float  scaleGlyphSpacing; //relative to text-height

	TextMetrics(float textHeight = 20,
				bool bMonospace = false,
				bool bUseFontDefaultGlyphSpacing = true,
				float customGlyphSpacing = 0,
				float scaleGlyphWidth = 1,
				float scaleGlyphSpacing = 1);
};


//====================================================================
struct VW_COMMON_API TextMetricsL : public TextMetrics
{
	float  lineHeight; //relative to text-height

	TextMetricsL(const TextMetrics& tm = TextMetrics(),
				float lineHeight = 1);

	float  AbsLineHeight() const { return (textHeight * lineHeight); }
};


//====================================================================
struct VW_COMMON_API TextAttributes
{
	std::string   fontName;
	TextMetricsL  textMetricsL;

	TextAttributes(const std::string& fontName = "",
				const TextMetricsL& textMetricsL = TextMetricsL());
};


//====================================================================
struct VW_COMMON_API DrawTextParams
{
	color4b      color;
	vec2         positionTopLeft;
	TextMetrics  tm;

	DrawTextParams(color4b color = color4b::WHITE,
					const vec2& positionTopLeft = vec2::ZERO,
					const gui::TextMetrics& textMetrics = gui::TextMetrics());
};


//====================================================================
struct VW_COMMON_API TileParams
{
	std::string                      texName;
	uint                             texRow;
	uint                             texCol;
	gui::RectangleCT::eTexTransform  texTransform;

	TileParams(const std::string& _texName = "", uint _texRow = 0, uint _texCol = 0,
				gui::RectangleCT::eTexTransform _texTransform = gui::RectangleCT::TexTransform_Identity)
		:
		texName(_texName), texRow(_texRow), texCol(_texCol), texTransform(_texTransform)
	{ }
};


//====================================================================
// Parameters are specified in standard-RHS.
// Rotations are counter-clockwise specified in degrees.
// Operations are performed in following order:
//
// 1) scaling
// 2) rotation around y-axis
// 3) rotation around x-axis
// 4) rotation around z-axis
// 5) translation
//
// If results are to be passed to gui-renderer, generate the matrix
// by using GenerateVM(). For convenience or quick-tests,
// the structure can be passed directly to gui::Renderer::SetTransform().
//====================================================================
class VW_COMMON_API Transform
{
public:
	vec3  translate;
	vec3  scale;
	vec3  rotate;


public:
	Transform(const vec3& _translate = vec3::ZERO,
				const vec3& _scale = vec3::ONE,
				const vec3& _rotate = vec3::ZERO)
		:
		translate(_translate), scale(_scale), rotate(_rotate)
	{ }

	void  GenerateMV(mat4& out) const;
	void  GenerateVM(mat4& out) const;

	operator mat4() const
	{
		mat4 mTemp;
		GenerateVM(mTemp);
		return mTemp;
	}


private:
	__forceinline static void __fastcall ConcatToOut(mat4& out, mat4& temp, mat4*& pOut)
	{
		if(pOut == &temp) out *= temp;
		else pOut = &temp;
	}
};


//====================================================================
class VW_COMMON_API VideoResource : public NoCopy
{
public:
	virtual ~VideoResource() {}

	// throw exception on error
	virtual  void  CreateVideoRendererResource(VideoRenderer& vr) = 0;

	// exceptions not allowed
	virtual  void  ReleaseVideoRendererResource() = 0;


protected:
	VideoResource();
};


//====================================================================
// Allowed sample-formats: X8R8G8B8, A8R8G8B8.
//====================================================================
class VW_COMMON_API Texture : virtual public VideoResource
{
public:
	enum eFilter
	{
		Filter_Invalid = -1,

		Filter_Off = 0,
		Filter_Nearest,
		Filter_Linear,

		Filter_Count,
		Filter_Force32Bits = 0x7FFFFFFF
	};


public:
	Texture(const ImageBase& srcImage, bool bMultiLevel);
	Texture(eSampleFormat sampleFormat, bool bMultiLevel, uint height, uint width);

	eSampleFormat  GetSampleFormat() const { return sampleFormat; }
	bool           IsMultiLevel() const { return bMultiLevel; }
	uint           GetHeight() const { return height; }
	uint           GetWidth() const { return width; }

	void     SetMipFilter(eFilter f = Filter_Linear) { mipFilter = f; }
	eFilter  GetMipFilter() const { return mipFilter; }

	//-----------------------------------------------------------------------
	// Internal object is constructed from <pSrcImage>, dimensions must match
	//-----------------------------------------------------------------------
	void  SetSourceImage(const ImageBase* pSrcImage = 0, bool bLoadBottomToTop = false);
	void  SetSourceImage(const std::string& fileName = "", bool bLoadBottomToTop = false);

	//-----------------------------
	// VideoResource implementation
	//-----------------------------
	virtual  void  CreateVideoRendererResource(VideoRenderer& vr);
	virtual  void  ReleaseVideoRendererResource();

	void                                UpdateVideoRendererResource();
	video_renderer_objects::Texture2D*  GetVideoRendererResource() const { return apVideoRendererResource.Get(); }


protected:
	// NOTES: override as pair
	virtual  ImageBase*  GenerateSourceImage(bool& bBottomToTop);
	virtual  void        ReleaseSourceImage(ImageBase* pGeneratedSourceImage);


private:
	void  CheckParameters();


private:
	eSampleFormat  sampleFormat;
	bool           bMultiLevel;
	uint           height;
	uint           width;
	eFilter        mipFilter;

	bool                bLoadBottomToTop;
	std::string         srcFileName;
	AutoPtr<ImageBase>  apSrcImage;

	mutable AutoPtr<video_renderer_objects::Texture2D>  apVideoRendererResource;
};


//====================================================================
class VW_COMMON_API TileTexture : public Texture
{
public:
	TileTexture(const ImageBase& srcImage, bool bMultiLevel,
				uint numTileRows, uint numTileColumns);

	TileTexture(eSampleFormat sampleFormat, bool bMultiLevel, uint height, uint width, 
				uint numTileRows, uint numTileColumns);

	uint   GetNumTileRows() const { return numTileRows; }
	uint   GetNumTileColumns() const { return numTileColumns; }

	float  GetTileSizeU() const { return tileSizeU; }
	float  GetTileSizeV() const { return tileSizeV; }

	float  GetTilePixelSizeU() const { return tilePixelSizeU; }
	float  GetTilePixelSizeV() const { return tilePixelSizeV; }

	__forceinline  float  GetTileCoordU(float tileColumn) const { return tileSizeU * tileColumn; }
	__forceinline  float  GetTileCoordV(float tileRow) const { return tileSizeV * tileRow; }

	void  GetTileCoords(float tileRow, float tileColumn,
						float& left, float& right, float& top, float& bottom) const;

	void  GetTileCoords(float tileRow, float tileColumn, RectangleCT& r) const;


private:
	void  CalcParameters();


private:
	uint   numTileRows;
	uint   numTileColumns;

	float  tileSizeU;
	float  tileSizeV;

	float  tilePixelSizeU;
	float  tilePixelSizeV;
};


//====================================================================
// If font is ascii-subset-only, valid character range is [0, 127],
// otherwise it is [0, 255]. If character out of valid range is
// passed to renderer, renderer must draw glyph 0 instead of
// offending character.
//====================================================================
class VW_COMMON_API Font : virtual public VideoResource
{
public:
	//----------------
	// Base properties
	//----------------
	void                SetFamilyName(const std::string& s = "Arial") { familyName = s; }
	const std::string&  GetFamilyName() const { return familyName; }

	void  SetBold(bool _bBold = false) { bBold = _bBold; }
	bool  IsBold() const { return bBold; }

	void  SetItalic(bool _bItalic = false) { bItalic = _bItalic; }
	bool  IsItalic() const { return bItalic; }

	void  SetAsciiSubsetOnly(bool bAsciiSubsetOnly = true) { numGlyphs = (bAsciiSubsetOnly ? 128 : 256); }
	bool  IsAsciiSubsetOnly() const { return (numGlyphs < 256); }

	uint  GetNumGlyphs() const { return numGlyphs; }

	//-----------------------------------
	// Properties relative to text-height
	//-----------------------------------
	void   SetDefaultSpacingMono(float _spacingMono = 0) { defaultSpacingMono = _spacingMono; }
	void   SetDefaultSpacing(float _spacing = 0) { defaultSpacing = _spacing; }

	float  GetDefaultSpacingMono() const { return defaultSpacingMono; }
	float  GetDefaultSpacing() const { return defaultSpacing; }
	float  GetDefaultSpacing(bool bMonospace) const { return (bMonospace ? defaultSpacingMono : defaultSpacing); }

	//--------------------------------------------------------
	// Single-line only (special characters are not processed)
	//--------------------------------------------------------
	virtual  void  DrawText(gui::Renderer& gr,
							uint textLen, const char* pText = 0,
							const gui::DrawTextParams& dtp = gui::DrawTextParams()) const = 0;

	//------------------------
	// Text-width calculations
	//------------------------
	float  CalcSpacingWidth(const TextMetrics& tm) const;

	virtual  float  CalcGlyphWidth(const TextMetrics& tm,
								char ch = 0) const = 0;

	virtual  float  CalcTextWidth(const TextMetrics& tm,
								uint textLen, const char* pText = 0) const = 0;

	virtual  void  CalcTextWidths(std::vector<float>& outWidths,
								const TextMetrics& tm,
								uint textLen, const char* pText = 0) const = 0;

	virtual  void  CalcMaxVisibleText(uint& outLen, float& outWidth, float maxWidth,
								const TextMetrics& tm,
								uint textLen, const char* pText = 0) const = 0;

	//-----------------------
	// std::string forwarders
	//-----------------------
	__forceinline  void  DrawText(gui::Renderer& gr,
								const std::string& text,
								const gui::DrawTextParams& dtp) const
	{ DrawText(gr, text.length(), text.c_str(), dtp); }

	__forceinline  float  CalcTextWidth(const TextMetrics& tm,
									const std::string& text) const
	{ return CalcTextWidth(tm, text.length(), text.c_str()); }

	__forceinline  void  CalcTextWidths(std::vector<float>& outWidths,
									const TextMetrics& tm,
									const std::string& text) const
	{ CalcTextWidths(outWidths, tm, text.length(), text.c_str()); }

	__forceinline  void  CalcMaxVisibleText(uint& outLen, float& outWidth, float maxWidth,
									const TextMetrics& tm,
									const std::string& text) const
	{ CalcMaxVisibleText(outLen, outWidth, maxWidth, tm, text.length(), text.c_str()); }


protected:
	Font();


private:
	std::string  familyName;
	bool         bBold;
	bool         bItalic;
	uint         numGlyphs;

	float  defaultSpacingMono;
	float  defaultSpacing;
};


//====================================================================
// Texture's sample-format will be set to A8R8G8B8.
// <textureWidth> must be { 256, 512, 1024 }.
// If <bAsciiSubsetOnly> is true only ascii glyphs [0-127]
// are generated, otherwise all ascii glyphs are generated [0-255].
// Glyph widths & heights are relative to tile-size i.e. in range [0, 1].
//====================================================================
class VW_COMMON_API FontTexture : public Font, public TileTexture
{
public:
	enum { NumGlyphColumns = 16 };

	static uint __fastcall FixTextureWidth(uint textureWidth);

	//------------------------------
	// Indirect creation parameters.
	//------------------------------
	struct VW_COMMON_API CreationParams
	{
		CreationParams();

		void  Dump(std::ostream& os, uint tabLevel = 0) const;

		// base properties
		std::string  familyName;
		bool         bBold;
		bool         bItalic;

		// properties relative to text-height
		float  defaultSpacingMono;
		float  defaultSpacing;

		// texture
		bool  bAsciiSubsetOnly;
		uint  textureWidth;
		bool  bMultiLevel;

		// image generation parameters
		float  igScaleEmHeight;
		float  igGlyphOffsetH;
		float  igGlyphOffsetV;
		float  igGlyphSizeModifyH;
		float  igGlyphSizeModifyV;

		// glyph properties; relative to glyph's pixel-size, range [0, 1]
		float  autoCalcWidthEmptyGlyph;
		float  autoCalcAddPixelsLeft;
		float  autoCalcAddPixelsRight;
		float  autoCalcAddPixelsTop;
		float  autoCalcAddPixelsBottom;
	};

	//----------------------------------------
	// Creates font-texture object indirectly.
	// Returns 0 on error.
	//----------------------------------------
	static FontTexture* __fastcall CreateFontTexture(const CreationParams& cp);


public:
	void  GetCreationParameters(CreationParams& cp) const;

	FontTexture(bool bAsciiSubsetOnly = true,
				uint textureWidth = 512,
				bool bMultiLevel = false);

	uint  GetNumGlyphRows() const { return (GetNumGlyphs() >> 4); }
	uint  GetNumGlyphColumns() const { return NumGlyphColumns; }

	__forceinline  float  GetGlyphPixelSizeU() const { return GetTilePixelSizeU(); }
	__forceinline  float  GetGlyphPixelSizeV() const { return GetTilePixelSizeV(); }

	//---------------------------------------------------
	// Font-image generation control.
	// If <bGenerateImageFromSystemFont> is false,
	// source-image from Texture base-class will be used.
	//---------------------------------------------------
	void  SetGenerateImageFromSystemFont(bool bGenerateImageFromSystemFont = true);
	bool  GetGenerateImageFromSystemFont() const { return bGenerateImageFromSystemFont; }

	ImageBase*  GenerateImageFromSystemFont();

	void   SetImageGenerationParameters(float scaleEmHeight = 1,
							float glyphOffsetH = 0, float glyphOffsetV = 0,
							float glyphSizeModifyH = 0, float glyphSizeModifyV = 0);

	void   GetImageGenerationParameters(float& scaleEmHeight,
							float& glyphOffsetH, float& glyphOffsetV,
							float& glyphSizeModifyH, float& glyphSizeModifyV);

	//-----------------------------------------------------------------------
	// Texture-glyph properties; relative to glyph's pixel-size; range [0, 1]
	//-----------------------------------------------------------------------
	void   ClearGlyphProperties();

	void   CalcGlyphPropertiesFromImage(float widthEmptyGlyph = 0.5f,
										float addPixelsLeft = 1.0f, float addPixelsRight = 1.0f,
										float addPixelsTop = 1.0f, float addPixelsBottom = 1.0f);

	void   SetGlyphHeight(float h = 1.0f);
	float  GetGlyphHeight() const { return glyphHeight; }

	void   SetGlyphVOffset(float o = 0.0f);
	float  GetGlyphVOffset() const { return glyphVOffset; }

	void   SetGlyphWidthMono(float w = 1.0f);
	float  GetGlyphWidthMono() const { return glyphWidthMono; }

	void   SetGlyphHOffsetMono(float o = 0.0f);
	float  GetGlyphHOffsetMono() const { return glyphHOffsetMono; }

	void   SetGlyphWidths(const float widths[256]);
	void   SetGlyphWidths(float w = 1.0f);

	__forceinline  float  GetGlyphWidth(ubyte ch) const { return glyphWidths[ch]; }
	__forceinline  float  GetGlyphWidth(byte ch) const { return GetGlyphWidth(static_cast<ubyte>(ch)); }

	void   SetGlyphHOffsets(const float offsets[256]);
	void   SetGlyphHOffsets(float o = 0.0f);

	__forceinline  float  GetGlyphHOffset(ubyte ch) const { return glyphHOffsets[ch]; }
	__forceinline  float  GetGlyphHOffset(byte ch) const { return GetGlyphHOffset(static_cast<ubyte>(ch)); }

	__forceinline  float  CalcGlyphWidthMultiplier(const TextMetrics& tm) const
	{
		return (oneOverGlyphHeight * tm.textHeight * tm.scaleGlyphWidth);
	}

	//---------------------
	// Texture coordinates.
	//---------------------
	__forceinline  void  GetGlyphTileCoords(ubyte ch, float& left, float& right, float& top, float& bottom) const
	{
		left = (ch % NumGlyphColumns) * GetTileSizeU();
		right = left + GetTileSizeU();
		top = (ch / NumGlyphColumns) * GetTileSizeV();
		bottom = top + GetTileSizeV();
	}

	__forceinline  void  GetGlyphTileCoords(byte ch, float& left, float& right, float& top, float& bottom) const
	{
		GetGlyphTileCoords(static_cast<ubyte>(ch), left, right, top, bottom);
	}

	__forceinline  void  GetGlyphCoordsMono(ubyte ch, float& left, float& right, float& top, float& bottom) const
	{
		left = glyphCoordsLeftMono[ch];
		right = glyphCoordsRightMono[ch];
		top = glyphCoordsTop[ch];
		bottom = glyphCoordsBottom[ch];
	}

	__forceinline  void  GetGlyphCoordsMono(byte ch, float& left, float& right, float& top, float& bottom) const
	{
		GetGlyphCoordsMono(static_cast<ubyte>(ch), left, right, top, bottom);
	}

	__forceinline  void  GetGlyphCoords(ubyte ch, float& left, float& right, float& top, float& bottom) const
	{
		left = glyphCoordsLeft[ch];
		right = glyphCoordsRight[ch];
		top = glyphCoordsTop[ch];
		bottom = glyphCoordsBottom[ch];
	}

	__forceinline  void  GetGlyphCoords(byte ch, float& left, float& right, float& top, float& bottom) const
	{
		GetGlyphCoords(static_cast<ubyte>(ch), left, right, top, bottom);
	}

	//--------------------
	// Font implementation
	//--------------------
	virtual  void  DrawText(gui::Renderer& gr,
							uint textLen, const char* pText = 0,
							const gui::DrawTextParams& dtp = gui::DrawTextParams()) const;

	virtual  float  CalcGlyphWidth(const TextMetrics& tm,
								char ch = 0) const;

	virtual  float  CalcTextWidth(const TextMetrics& tm,
								uint textLen, const char* pText = 0) const;

	virtual  void  CalcTextWidths(std::vector<float>& outWidths,
								const TextMetrics& tm,
								uint textLen, const char* pText = 0) const;

	virtual  void  CalcMaxVisibleText(uint& outLen, float& outWidth, float maxWidth,
								const TextMetrics& tm,
								uint textLen, const char* pText = 0) const;

	//---------------------------------------------------------------
	// VideoResource implementation (avoid inheritance via dominance)
	//---------------------------------------------------------------
	virtual  void  CreateVideoRendererResource(VideoRenderer& vr) { TileTexture::CreateVideoRendererResource(vr); }
	virtual  void  ReleaseVideoRendererResource() { TileTexture::ReleaseVideoRendererResource(); }


protected:
	//-----------------
	// Internal helpers
	//-----------------
	void  CalcVerticalGlyphCoords();
	void  CalcHorizontalGlyphCoords();

	//-----------------------
	// Texture implementation
	//-----------------------
	virtual  ImageBase*  GenerateSourceImage(bool& bBottomToTop);
	virtual  void        ReleaseSourceImage(ImageBase* pGeneratedSourceImage);


private:
	bool   bGenerateImageFromSystemFont;

	float  fIGScaleEmHeight;
	float  fIGGlyphOffsetH;
	float  fIGGlyphOffsetV;
	float  fIGGlyphSizeModifyH;
	float  fIGGlyphSizeModifyV;

	float  glyphHeight;
	float  oneOverGlyphHeight;
	float  glyphVOffset;

	float  glyphWidthMono;
	float  glyphHOffsetMono;

	float  autoCalcWidthEmptyGlyph;
	float  autoCalcAddPixelsLeft;
	float  autoCalcAddPixelsRight;
	float  autoCalcAddPixelsTop;
	float  autoCalcAddPixelsBottom;

	float  glyphWidths[256];
	float  glyphHOffsets[256];

	float  glyphCoordsLeftMono[256];
	float  glyphCoordsRightMono[256];
	float  glyphCoordsLeft[256];
	float  glyphCoordsRight[256];
	float  glyphCoordsTop[256];
	float  glyphCoordsBottom[256];
};


//====================================================================
// GUI renderer.
// Uses screen-space right-handed-coord-system.
// (+x left-to-right, +y top-to-bottom, +z away-from-the-viewer)
// Face orientation is counter-clockwise.
// Transforms are specified for VM multiplication order.
// Visible XY-plane of screen-space is defined by grid from
// (0, 0) to (winWidth, winHeight) inclusive,
// so there is total of (winWidth + 1) x (winHeight + 1)
// grid-points. Coords (0, 0) represent top-left corner.
// Primitive that covers area [0, 0] to [1, 1] will fill pixel (0, 0).
// Primitive that covers area [0, 0] to [2, 2] will fill pixels
// (0, 0), (0, 1), (1, 0), (1, 1).
// Allowed texture sample-formats: X8R8G8B8, A8R8G8B8.
// Texture-addressing-mode is always clamp-to-edge.
// Base-filtering-mode is always linear. If texture is multi-level,
// mip-filtering can be controlled through texture-object.
// Text drawing rules: if character out of valid range for given
// font is passed to renderer, it must draw glyph 0 instead of
// offending character.
//====================================================================
class VW_COMMON_API Renderer
{
friend class VideoRenderer;

public:
	//-----------------------------
	struct VW_COMMON_API Statistics
	{
		Statistics();

		void  Clear(uint val = 0);
		void  Accum(const Statistics& s);
		void  Average(const Statistics& s, uint frameCount);
		void  Min(const Statistics& s1, const Statistics& s2);
		void  Max(const Statistics& s1, const Statistics& s2);
		void  Dump(std::ostream& os, uint tabLevel = 0) const;

		// call counts
		uint  numClearBuffers;
		uint  numSetViewport;
		uint  numBegin;
		uint  numEnd;
		uint  numBeginInterleaved;
		uint  numEndInterleaved;
		uint  numSetDepthParams;
		uint  numSetTransform;
		uint  numDrawPrimitiveColored;
		uint  numDrawPrimitiveTextured;
		uint  numDrawText;
		uint  numDrawRectangleColored;
		uint  numDrawRectangleTextured;
		uint  numDrawRectangleFrameColored;

		// additional
		uint  numTextureChanges;
		uint  numPrimitivesColored;
		uint  numPrimitivesTextured;
		uint  numTextGlyphs;
	};

	//---------------------------------
	struct VW_COMMON_API RendererParams
	{
		RendererParams();

		bool  bTextureRowOrderBottomToTop;
		bool  bTextureLayerOrderBackToFront;
	};

	//------------------------------
	struct VW_COMMON_API PixelParams
	{
		PixelParams(float offsetX = 0, float offsetY = 0);

		// range [-1, 1]
		float  offsetX; //default: implementation dependant
		float  offsetY; //default: implementation dependant
	};

	//------------------------------
	struct VW_COMMON_API ClearParams
	{
		ClearParams(bool bClearColor = true, color4b color = color4b(0, 0, 128),
					bool bClearDepth = true, float depth = 1,
					bool bClearStencil = false, uint stencil = 0,
					bool bUseRectangle = false,
					const gui::Rectangle& r = gui::Rectangle(0, 0, 0, 0));

		bool  bClearColor;
		bool  bClearDepth;
		bool  bClearStencil;
		bool  bUseRectangle;

		color4b    color;
		float      depth; //range: [0, 1]
		uint       stencil;
		Rectangle  rectangle;
	};

	//---------------------------------
	struct VW_COMMON_API ViewportParams
	{
		ViewportParams(const vec2& topLeft = vec2::ZERO,
						const vec2& size = vec2(200, 200),
						float renderDepthNear = 0,
						float renderDepthFar = 1);

		vec2   topLeft;
		vec2   size;
		float  renderDepthNear;
		float  renderDepthFar;
	};

	//------------------------------
	struct VW_COMMON_API DepthParams
	{
		DepthParams(float clippingPlaneNear = -100, float clippingPlaneFar = 100,
					float renderDepthNear = 0, float renderDepthFar = 1,
					bool bEnableDepthTest = false, bool bEnableDepthWrite = false,
					bool bUseWBuffering = false);

		// range [-inf, inf]
		float  clippingPlaneNear;
		float  clippingPlaneFar;

		// range [0, 1]
		float  renderDepthNear;
		float  renderDepthFar;

		bool  bEnableDepthTest;
		bool  bEnableDepthWrite;
		bool  bUseWBuffering;
	};

	//-----------------------------------
	class VW_COMMON_API DepthParamsSentry
	{
	public:
		DepthParamsSentry(Renderer& r, const DepthParams& depthParams) : rend(r)
		{
			rend.GetDepthParams(dpLast);
			rend.SetDepthParams(depthParams);
		}
		~DepthParamsSentry()
		{
			rend.SetDepthParams(dpLast);
		}
	private:
		Renderer&    rend;
		DepthParams  dpLast;
	};

	//--------------------------------------------------
	// If <bPush> is true, <transform> is pre-multiplied
	// with curent transform.
	//--------------------------------------------------
	class VW_COMMON_API TransformSentry
	{
	public:
		TransformSentry(Renderer& r, const mat4& transform, bool bPush = true) : rend(r)
		{
			rend.GetTransform(tLast);
			if(bPush) {
				rend.SetTransform(transform * tLast);
			}
			else {
				rend.SetTransform(transform);
			}
		}
		~TransformSentry()
		{
			rend.SetTransform(tLast);
		}
	private:
		Renderer&  rend;
		mat4       tLast;
	};


public:
	Renderer(VideoRenderer& vrParent);
	virtual ~Renderer();

	//---------
	// Various.
	//---------
	VideoRenderer&  GetParentVideoRenderer() { return parentVideoRenderer; }

	void                              ClearStats() { statistics.Clear(); }
	gui::Renderer::Statistics&        Stats() { return statistics; }
	const gui::Renderer::Statistics&  Stats() const { return statistics; }

	// Range [1, +inf>
	uint  GetWindowWidth() const { return windowWidth; }
	uint  GetWindowHeight() const { return windowHeight; }
	vec2  GetWindowSize() const { return vec2(windowWidth, windowHeight); }

	virtual  void  GetRendererParams(gui::Renderer::RendererParams& rp) const = 0;

	//--------------------------------------
	// Methods valid only outside Begin/End.
	//--------------------------------------

	// Returns false if low-level-api cannot handle the given viewport (crosses window boundaries)
	virtual  bool  SetViewport(const ViewportParams& vp) = 0;

	// If clearing with <cp.bUseRectangle> == false, method internally
	// sets the viewport to full window area
	virtual  void  ClearBuffers(const ClearParams& cp) = 0;

	virtual  void  SetPixelParams(const gui::Renderer::PixelParams& pp) = 0;
	virtual  void  GetPixelParams(gui::Renderer::PixelParams& pp) const = 0;

	// Base parameters are automatically restored after call to Begin
	virtual  void  SetBaseDepthParams(const DepthParams& dp) = 0;
	virtual  void  GetBaseDepthParams(DepthParams& dp) const = 0;

	// Returns true if all created gui-resources are managed,
	// (they do not have to be recreated after reset)
	virtual  bool  QueryCreatesManagedResourcesOnly() = 0;

	// Resource creation (client is responsible for resource destruction)
	virtual  video_renderer_objects::Texture2D*  CreateResourceTexture2D(
					eSampleFormat sampleFormat, bool bMultiLevel, uint height, uint width) = 0;

	//-----------
	// Begin/End.
	//-----------
	virtual  bool  IsInsideBeginEnd() = 0;
	virtual  void  Begin() = 0;
	virtual  void  End() = 0;

	// Use if low-level-api calls need to be interleaved with gui. Valid between Begin/End.
	virtual  bool  IsInsideBeginEndInterleaved() = 0;
	virtual  void  BeginInterleaved() = 0;
	virtual  void  EndInterleaved() = 0;

	//-------------------------------------
	// Methods valid only inside Begin/End.
	//-------------------------------------
	virtual  void  SetDepthParams(const DepthParams& dp) = 0;
	virtual  void  GetDepthParams(DepthParams& dp) const = 0;

	virtual  void  SetTransform(const mat4& t = mat4::IDENTITY) = 0;
	virtual  void  GetTransform(mat4& t) const = 0;

	virtual  void  DrawPrimitive(const gui::VertexBufferC& vertexBuffer,
								gui::ePrimitiveType primitiveType,
								uint numPrimitives,
								uint startVertex) = 0;

	virtual  void  DrawPrimitive(const gui::Texture& texture,
								const gui::VertexBufferCT& vertexBuffer,
								gui::ePrimitiveType primitiveType,
								uint numPrimitives,
								uint startVertex) = 0;

	//---------------------------------------------------------
	// Single-line only (special characters are not processed).
	//---------------------------------------------------------
	virtual  void  DrawText(const gui::FontTexture& fontTexture,
							uint textLen, const char* pText,
							const gui::DrawTextParams& dtp) = 0;

	__forceinline  void  DrawText(const gui::FontTexture& fontTexture,
								const std::string& text,
								const gui::DrawTextParams& dtp)
	{ DrawText(fontTexture, text.length(), text.c_str(), dtp); }

	//----------------------------------------
	// Additional methods.
	// Default implementations use core calls.
	//----------------------------------------
	virtual  void  DrawRectangle(const gui::RectangleC& rect);

	virtual  void  DrawRectangle(const gui::Texture& texture,
								const gui::RectangleCT& rect);

	virtual  void  DrawRectangleFrame(const gui::RectangleC& rectOuter,
									float frameWidth = 1.0f);

	__forceinline  void  DrawText(const gui::Font& font,
								uint textLen, const char* pText,
								const gui::DrawTextParams& dtp)
	{ font.DrawText(*this, textLen, pText, dtp); }

	__forceinline  void  DrawText(const gui::Font& font,
								const std::string& text,
								const gui::DrawTextParams& dtp)
	{ font.DrawText(*this, text, dtp); }


private:
	VideoRenderer&             parentVideoRenderer;
	gui::Renderer::Statistics  statistics;

	uint  windowWidth;
	uint  windowHeight;

	gui::VertexBufferC   tempVertexBufferC;
	gui::VertexBufferCT  tempVertexBufferCT;


protected:
	//-------------------------------
	// Video-subsystem notifications.
	//-------------------------------
	virtual  void  ConnectVideoSubsystem(VideoSubsystem& vs) = 0;
	virtual  void  DisconnectVideoSubsystem() = 0;

	virtual  void  CreateVideoResources(bool bPostReset) = 0;
	virtual  void  ReleaseVideoResources(bool bPreReset) = 0;

	virtual  void  OnWindowResize(uint width, uint height);
};


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


