

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Base.h"
#include "GUI\GUI_StdObjects.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui {


//====================================================================
class VW_COMMON_API BaseShader : public VideoResourceUser
{
friend class Manager;

public:
	//----------------------
	// BaseShader interface.
	//----------------------
	virtual ~BaseShader();

	void                SetName(const std::string& s = "");
	const std::string&  GetName() const { return shaderName; }

	static gui::TextAttributes __fastcall DefaultTextAttributes();

	void                        SetTextAttributes(const gui::TextAttributes& ta = gui::TextAttributes());
	const gui::TextAttributes&  GetTextAttributes() const { return textAtts; }

	const gui::Font*          GetFont() const { return pFont; }
	const gui::TextMetricsL&  GetTextMetricsL() const { return textAtts.textMetricsL; }
	gui::TextMetricsL         GetTextMetricsL(float textHeight, float lineHeight = 1) const;

	//------------------------------
	// Implemented by derived-class.
	//------------------------------
	virtual  void  OnTimeDelta(double secs) {}

	virtual  void  Shade(const gui::Component& c, gui::Renderer& gr) const = 0;

	//------------------------------------------------------
	// <testPoint> is already tested by component against
	// basic shape, shader can make additional tests against
	// bitmap mask. Return true on hit.
	//------------------------------------------------------
	virtual  bool  HitTest(const gui::Component& c, const vec2& testPoint) const
	{ return true; }

	//---------------------------------------------------
	// <commandArea> is result of test by component-test.
	//---------------------------------------------------
	virtual  gui::eCommandArea  HitTestCommandArea(gui::eCommandArea commandArea,
													const vec2& testPoint) const
	{ return commandArea; }


protected:
	BaseShader(const std::string& name = "");

	//-----------------------------
	// VideoResourceUser overrides.
	//-----------------------------
	virtual  bool  CacheFontHandles(gui::Manager& gm);

	//-------------------------
	// Derived-class interface.
	//-------------------------
	__forceinline  gui::TextMetricsL&  GetTextMetricsLCache(float textHeight, float lineHeight = 1) const
	{
		textMetricsLCache.textHeight = textHeight;
		textMetricsLCache.lineHeight = lineHeight;
		return textMetricsLCache;
	}


protected:
	//------------
	// Properties.
	//------------
	std::string          shaderName;
	gui::TextAttributes  textAtts;

	//-------
	// Cache.
	//-------
	const gui::Font*           pFont;
	mutable gui::TextMetricsL  textMetricsLCache;
};


//====================================================================
template<typename TTargetShader>
struct ShaderPtr
{
	TTargetShader* ptr;

	__forceinline operator TTargetShader*() { return ptr; }
	__forceinline TTargetShader* operator->() { return ptr; }
	__forceinline TTargetShader& operator*() { return *ptr; }

	__forceinline  ShaderPtr(gui::BaseShader& shader)
	{
		ptr = dynamic_cast<TTargetShader*>(&shader);
	}
	__forceinline  ShaderPtr(gui::BaseShader* shader = 0)
	{
		ptr = dynamic_cast<TTargetShader*>(shader);
	}
};


//====================================================================
template<typename TTargetShader>
struct ConstShaderPtr
{
	const TTargetShader* ptr;

	__forceinline operator const TTargetShader*() { return ptr; }
	__forceinline const TTargetShader* operator->() { return ptr; }
	__forceinline const TTargetShader& operator*() { return *ptr; }

	__forceinline  ConstShaderPtr(const gui::BaseShader& shader)
	{
		ptr = dynamic_cast<const TTargetShader*>(&shader);
	}
	__forceinline  ConstShaderPtr(const gui::BaseShader* shader = 0)
	{
		ptr = dynamic_cast<const TTargetShader*>(shader);
	}
};


//********************************************************************
}//namespace gui


//********************************************************************
}//namespace vertexware


