

//--------------------------------------------------------------------
#include "HIDControllers\GameController.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"
#include "BaseTools\Constructs.h"
#include "MathTools\MathBase.h"


//********************************************************************
namespace vertexware {


//====================================================================
GameController::GameController()
{
	ReportHardwareCaps_Axes(false, false, false, false, false, false);
	ReportHardwareCaps_Gadgets(0, 0, 0);

	InitOperationModifiers();
	ResetInputReadings();
}


//====================================================================
GameController::~GameController()
{
}


//====================================================================
std::string  GameController::GetControllerInterfaceName() const
{
	return VW_INTERFACE_NAME_GameController;
}


//====================================================================
void  GameController::DumpControllerInfo(std::ostream& os, uint tabLevel) const
{
	TabEffector  te(tabLevel);
	os << te << "AxisX.....: " << ToString(bHasAxis[Axis_X]) << nl;
	os << te << "AxisY.....: " << ToString(bHasAxis[Axis_Y]) << nl;
	os << te << "AxisZ.....: " << ToString(bHasAxis[Axis_Z]) << nl;
	os << te << "AxisRX....: " << ToString(bHasAxis[Axis_RX]) << nl;
	os << te << "AxisRY....: " << ToString(bHasAxis[Axis_RY]) << nl;
	os << te << "AxisRZ....: " << ToString(bHasAxis[Axis_RZ]) << nl;
	os << te << "NumSliders: " << ToString(numSliders) << nl;
	os << te << "NumPOVHats: " << ToString(numPOVHats) << nl;
	os << te << "NumButtons: " << ToString(numButtons) << nl;
}


//====================================================================
void  GameController::ResetInputReadings()
{
	ResetAxisStates();
	ResetSliderStates();
	ResetPOVHatStates();
	ResetButtonStates();
}


//====================================================================
void  GameController::PreSampleUpdate()
{
	if(GetParentHID().IsAsyncSamplingActive()) {
		ClearCurrentButtonStates();
	}
	else {
		ShiftButtonStates();
	}
}


//====================================================================
void  GameController::PostSampleUpdate()
{
}


//====================================================================
void  GameController::OnBeginAsyncSampleRead()
{
}


//====================================================================
void  GameController::OnEndAsyncSampleRead()
{
	ShiftCopyButtonStates();
}


//====================================================================
void  GameController::InitOperationModifiers()
{
	{for(uint i = 0; i < Axis_Count; i++) {
		SetAxisReverse(eAxis(i), false);
		axisDeadZones[eAxis(i)] = 0;
		axisSaturations[eAxis(i)] = 1;
	}}	

	{for(uint i = 0; i < MaxSliders; i++) {
		SetSliderReverse(i, false);
		sliderDeadZones[i] = 0;
		sliderSaturations[i] = 1;
	}}	
}


//====================================================================
void  GameController::ResetAxisProperties()
{
	for(uint i = 0; i < Axis_Count; i++) {
		SetAxisProperties(eAxis(i));
	}
}


//====================================================================
void  GameController::SetAxisProperties(eAxis a, float deadZone, float saturation)
{
	axisDeadZones[a] = Clamped<float>(deadZone, 0, 1);
	axisSaturations[a] = Clamped<float>(saturation, axisDeadZones[a], 1);

	OnInputModifiersChanged();
}


//====================================================================
void  GameController::ResetSliderProperties()
{
	for(uint i = 0; i < GetNumSliders(); i++) {
		SetSliderProperties(i);
	}
}


//====================================================================
void  GameController::SetSliderProperties(uint sliderIndex, float deadZone, float saturation)
{
	sliderDeadZones[sliderIndex] = Clamped<float>(deadZone, 0, 1);
	sliderSaturations[sliderIndex] = Clamped<float>(saturation, sliderDeadZones[sliderIndex], 1);

	OnInputModifiersChanged();
}


//====================================================================
void  GameController::SetAxisReverse(eAxis a, bool bReverse)
{
	axisReverseMods[a] = bReverse ? -1 : 1;
}


//====================================================================
void  GameController::SetSliderReverse(uint sliderIndex, bool bReverse)
{
	sliderReverseMods[sliderIndex] = bReverse ? -1 : 1;
}


//====================================================================
void  GameController::ReportHardwareCaps_Axes(
					bool _bHasX, bool _bHasY, bool _bHasZ,
					bool _bHasRX, bool _bHasRY, bool _bHasRZ)
{
	bHasAxis[Axis_X] = _bHasX;
	bHasAxis[Axis_Y] = _bHasY;
	bHasAxis[Axis_Z] = _bHasZ;
	bHasAxis[Axis_RX] = _bHasRX;
	bHasAxis[Axis_RY] = _bHasRY;
	bHasAxis[Axis_RZ] = _bHasRZ;
}


//====================================================================
void  GameController::ReportHardwareCaps_Gadgets(
					uint _numSliders, uint _numPOVHats, uint _numButtons)
{
	numSliders = ClampedHigh<uint>(_numSliders, MaxSliders);
	numPOVHats = ClampedHigh<uint>(_numPOVHats, MaxPOVHats);
	numButtons = ClampedHigh<uint>(_numButtons, MaxButtons);
}


//====================================================================
void  GameController::ResetAxisStates()
{
	for(uint i = 0; i < Axis_Count; i++) {
		axisPositions[i] = 0;
	}
}


//====================================================================
void  GameController::ResetSliderStates()
{
	for(uint i = 0; i < MaxSliders; i++) {
		sliderPositions[i] = 0;
	}
}


//====================================================================
void  GameController::ResetPOVHatStates()
{
	for(uint i = 0; i < MaxPOVHats; i++) {

		// negative value means centered
		povHatPositions[i] = -1;
	}
}


//====================================================================
void  GameController::ResetButtonStates()
{
	for(uint i = 0; i < MaxButtons; i++) {
		buttonStates[i] = 0;
	}
}


//====================================================================
void  GameController::ClearCurrentButtonStates()
{
	for(uint i = 0; i < MaxButtons; i++) {
		buttonStates[i] &= CurrentStateInv;
	}
}


//====================================================================
void  GameController::ShiftButtonStates()
{
	for(uint i = 0; i < MaxButtons; i++) {
		buttonStates[i] >>= StateBitShift;
	}
}


//====================================================================
void  GameController::ShiftCopyButtonStates()
{
	for(uint i = 0; i < MaxButtons; i++) {
		ubyte currState = buttonStates[i]  &  CurrentStateBit;
		buttonStates[i] = currState | (currState >> StateBitShift);
	}
}


//********************************************************************
}//namespace vertexware


