

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "Framework\HIDController.h"


//********************************************************************
namespace vertexware {


//====================================================================
// All axes report absolute position in range [-1, 1],
// with 0 indicating center position.
// X-Axis: move-left = negative, move-right = positive.
// Y-Axis: move-away-from-user = negative, move-towards-user = positive.
// Z-Axis: move-high = negative, move-low = positive.
// XYZ-axes comprise RHS, so XYZ-Rotations are RHS clockwise
// in range [-1, 1], with 0 indicating center position.
// Since sliders are axes, they also report absolute position
// in range [-1, 1], with 0 indicating center position.
// Slider values should increase from position-near-the-user
// to position-far-from-user.
// PointOfViewHats report direction specified in degrees measured
// clockwise from north (north is 0 degrees). Negative POV-Hat value
// indicates centered position (no direction).
// DeadZone/Saturation range is [0, 1], DeadZone <= Saturation.
//====================================================================
class VW_COMMON_API GameController : public HIDController
{
public:
	//----------------------
	// Limits and constants.
	//----------------------
	enum
	{
		MaxSliders = 8,
		MaxPOVHats = 4,
		MaxButtons = 32
	};

	enum eAxis
	{
		Axis_X = 0,
		Axis_Y,
		Axis_Z,
		Axis_RX,
		Axis_RY,
		Axis_RZ,

		Axis_Count,
		Axis_Force32Bits = 0x7FFFFFFF
	};


public:
	virtual ~GameController();

	//------------------------------
	// HIDController implementation.
	//------------------------------
	virtual  std::string  GetControllerInterfaceName() const;
	virtual  void         DumpControllerInfo(std::ostream& os, uint tabLevel) const;
	virtual  void         ResetInputReadings();
	virtual  void         PreSampleUpdate();
	virtual  void         PostSampleUpdate();
	virtual  void         OnBeginAsyncSampleRead();
	virtual  void         OnEndAsyncSampleRead();


	//--------------------------
	// GameController interface.
	//--------------------------

	// hardware capabilities
	__forceinline  bool  HasAxis(eAxis a) const { return bHasAxis[a]; }
	__forceinline  uint  GetNumSliders() const { return numSliders; }
	__forceinline  uint  GetNumPOVHats() const { return numPOVHats; }
	__forceinline  uint  GetNumButtons() const { return numButtons; }

	// operation modifiers
	bool  IsAxisReversed(eAxis a) const { return (axisReverseMods[a] < 0); }
	void  SetAxisReverse(eAxis a, bool bReverse);

	bool  IsSliderReversed(uint sliderIndex) const { return (sliderReverseMods[sliderIndex] < 0); }
	void  SetSliderReverse(uint sliderIndex, bool bReverse);

	// axis properties, range [0, 1]
	void  ResetAxisProperties();
	void  SetAxisProperties(eAxis a, float deadZone = 0, float saturation = 1);

	__forceinline  float  GetAxisDeadZone(eAxis a) const { return axisDeadZones[a]; }
	__forceinline  float  GetAxisSaturation(eAxis a) const { return axisSaturations[a]; }

	// slider properties, range [0, 1]
	void  ResetSliderProperties();
	void  SetSliderProperties(uint sliderIndex, float deadZone = 0, float saturation = 1);

	__forceinline  float  GetSliderDeadZone(uint sliderIndex) const { return sliderDeadZones[sliderIndex]; }
	__forceinline  float  GetSliderSaturation(uint sliderIndex) const { return sliderSaturations[sliderIndex]; }

	// standard axes
	__forceinline  float  GetAxisPosition(eAxis a) const { return axisPositions[a]; }

	// sliders (additional axes)
	__forceinline  float  GetSliderPosition(uint sliderIndex) const { return sliderPositions[sliderIndex]; }

	// point-of-view hats, negative value indicates center
	__forceinline  float  GetPOVHatPosition(uint povHatIndex) const { return povHatPositions[povHatIndex]; }
	__forceinline  bool   IsPOVHatCentered(uint povHatIndex) const { return (povHatPositions[povHatIndex] < 0); }

	// tests current button-state
	__forceinline  ubyte  ButtonPressed(uint buttonIndex) const {
		return IsPressed(buttonStates[buttonIndex]);
	}

	// tests button-state transitions
	__forceinline  ubyte  ButtonDown(uint buttonIndex) const {
		ubyte state = buttonStates[buttonIndex];
		return IsPressed(state)  &&  ! WasPressed(state);
	}

	__forceinline  ubyte  ButtonUp(uint buttonIndex) const {
		ubyte state = buttonStates[buttonIndex];
		return WasPressed(state)  &&  ! IsPressed(state);
	}


private:
	//------------------
	// Internal helpers.
	//------------------
	enum {
		CurrentStateBit = 0x80,
		CurrentStateInv = 0x7F,
		LastStateBit    = 0x01,
		StateBitShift   = 0x07,
		Force8Bits      = 0x7F //not used
	};

	__forceinline static ubyte  IsPressed(ubyte buttonState) {
		return buttonState & CurrentStateBit;
	}
	__forceinline static ubyte  WasPressed(ubyte buttonState) {
		return buttonState & LastStateBit;
	}

	void  InitOperationModifiers();


private:
	// hardware capabilities
	bool  bHasAxis[Axis_Count];
	uint  numSliders;
	uint  numPOVHats;
	uint  numButtons;

	// operation modifiers
	float  axisReverseMods[Axis_Count];
	float  sliderReverseMods[MaxSliders];

	float  axisDeadZones[Axis_Count];
	float  sliderDeadZones[MaxSliders];

	float  axisSaturations[Axis_Count];
	float  sliderSaturations[MaxSliders];

	// input readings
	float  axisPositions[Axis_Count];
	float  sliderPositions[MaxSliders];
	float  povHatPositions[MaxPOVHats];
	ubyte  buttonStates[MaxButtons];


protected:
	//-------------------------
	// Derived-class interface.
	//-------------------------
	GameController();

	// report capabilities
	void  ReportHardwareCaps_Axes(
					bool bHasX, bool bHasY, bool bHasZ,
					bool bHasRX, bool bHasRY, bool bHasRZ);

	void  ReportHardwareCaps_Gadgets(
					uint numSliders, uint numPOVHats, uint numButtons);

	// report data for standard axes
	__forceinline  void  SetAxisPosition(eAxis a, float val) {
		axisPositions[a] = val * axisReverseMods[a];
	}

	// report slider data
	__forceinline  void  SetSliderPosition(uint sliderIndex, float val) {
		sliderPositions[sliderIndex] = val * sliderReverseMods[sliderIndex];
	}

	// report point-of-view-hat data
	__forceinline  void  SetPOVHatPosition(uint povHatIndex, float val) {
		povHatPositions[povHatIndex] = val;
	}

	// report button states
	__forceinline  void  SetButtonPressed(uint buttonIndex) {
		buttonStates[buttonIndex] |= CurrentStateBit;
	}

	void  ResetAxisStates();
	void  ResetSliderStates();
	void  ResetPOVHatStates();

	void  ResetButtonStates();
	void  ClearCurrentButtonStates();
	void  ShiftButtonStates();
	void  ShiftCopyButtonStates();

	//------------------
	// Template methods.
	//------------------
	virtual  void  OnInputModifiersChanged() = 0;
};


//====================================================================
#define  VW_INTERFACE_NAME_GameController  "GameController"


//********************************************************************
}//namespace vertexware


