

//--------------------------------------------------------------------
#include "HIDControllers\Glove.h"
#include "BaseTools\String.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Constructs.h"


//====================================================================
#define  EQUAL_VALUE_FLEX_NORM            0.5

#define  FLEX_TO_ANGLE_SCALE_WRIST      180
#define  FLEX_TO_ANGLE_BIAS_WRIST       -90

#define  FLEX_TO_ANGLE_SCALE_FINGER      80
#define  FLEX_TO_ANGLE_BIAS_FINGER        0

#define  FLEX_TO_ANGLE_SCALE_ABDUCTION   40
#define  FLEX_TO_ANGLE_BIAS_ABDUCTION     0


//********************************************************************
namespace vertexware {


//====================================================================
std::string __fastcall Glove::GetSensorName(eGloveSensor sensor)
{
	switch(sensor) {
		case GloveSensor_WristYaw: return "WristYaw";
		case GloveSensor_WristPitch: return "WristPitch";
		case GloveSensor_WristRoll: return "WristRoll";

		case GloveSensor_AbductionThumbIndex: return "AbductionThumbIndex";
		case GloveSensor_AbductionIndexMiddle: return "AbductionIndexMiddle";
		case GloveSensor_AbductionMiddleRing: return "AbductionMiddleRing";
		case GloveSensor_AbductionRingPinky: return "AbductionRingPinky";

		case GloveSensor_ThumbRoll: return "ThumbRoll";
		case GloveSensor_ThumbJointInner: return "ThumbJointInner";
		case GloveSensor_ThumbJointOuter: return "ThumbJointOuter";

		case GloveSensor_IndexJointInner: return "IndexJointInner";
		case GloveSensor_IndexJointMiddle: return "IndexJointMiddle";
		case GloveSensor_IndexJointOuter: return "IndexJointOuter";

		case GloveSensor_MiddleJointInner: return "MiddleJointInner";
		case GloveSensor_MiddleJointMiddle: return "MiddleJointMiddle";
		case GloveSensor_MiddleJointOuter: return "MiddleJointOuter";

		case GloveSensor_RingJointInner: return "RingJointInner";
		case GloveSensor_RingJointMiddle: return "RingJointMiddle";
		case GloveSensor_RingJointOuter: return "RingJointOuter";

		case GloveSensor_PinkyJointInner: return "PinkyJointInner";
		case GloveSensor_PinkyJointMiddle: return "PinkyJointMiddle";
		case GloveSensor_PinkyJointOuter: return "PinkyJointOuter";
	}
	return "Unknown";
}


//====================================================================
Glove::Glove()
{
	ResetInputReadings();

	ReportHandedness(false);
	ReportHardwareCaps_Measurement(false);
	for(uint i = 0; i < GloveSensor_Count; i++) {
		ReportHardwareCaps_Sensor(eGloveSensor(i), false);
	}

	SetDefaultFlexureToAngleMaps();
	SetDefaultEqualValueFlexureNormalizations();

	// auto-calibrate by default
	ActivateFlexureAutoCalibration(true, true);
}


//====================================================================
Glove::~Glove()
{
}


//====================================================================
std::string  Glove::GetControllerInterfaceName() const
{
	return VW_INTERFACE_NAME_Glove;
}


//====================================================================
void  Glove::DumpControllerInfo(std::ostream& os, uint tabLevel) const
{
	TabEffector te(tabLevel);
	const char padChar = '.';

	uint maxNameLen = 0;
	{for(uint i = 0; i < GloveSensor_Count; i++) {
		maxNameLen = Max(maxNameLen, GetSensorName(eGloveSensor(i)).length());
	}}

	std::string  sAngleMeasurement("AngleMeasurement");
	PadToLength(sAngleMeasurement, maxNameLen, padChar);
	os << te << sAngleMeasurement << ": " << ToString(IsSupportedAngleMeasurement()) << nl;

	{for(uint i = 0; i < GloveSensor_Count; i++) {
		std::string sensorName = GetSensorName(eGloveSensor(i));
		PadToLength(sensorName, maxNameLen, padChar);
		os << te << sensorName << ": " << ToString(IsSupportedSensor(eGloveSensor(i))) << nl;
	}}
}


//====================================================================
void  Glove::ResetInputReadings()
{
	for(uint i = 0; i < GloveSensor_Count; i++) {
		angles[i] = 0.0f;
		normalizedFlexures[i] = 0.0f;
		flexures[i] = 0.0f;
		rawValues[i] = 0.0f;
	}
}


//====================================================================
void  Glove::PreSampleUpdate()
{
}


//====================================================================
void  Glove::PostSampleUpdate()
{
	if(bFlexureAutoCalibration) {
		UpdateFlexureNormalizers();
	}
	CalculateNormalizedFlexures();
}


//====================================================================
void  Glove::OnBeginAsyncSampleRead()
{
}


//====================================================================
void  Glove::OnEndAsyncSampleRead()
{
}


//====================================================================
void  Glove::CalculateNormalizedFlexures()
{
	for(uint i = 0; i < GloveSensor_Count; i++) {
		float oneOverRange = oneOverRangeFlexures[i];
		if(oneOverRange) {
			normalizedFlexures[i] = (flexures[i] - minFlexures[i]) * oneOverRange;
		}
		else {
			// equal-value normalization
			normalizedFlexures[i] = equalValueFlexureNorm[i];
		}
	}
}


//====================================================================
void  Glove::UpdateFlexureNormalizers()
{
	for(uint i = 0; i < GloveSensor_Count; i++) {
		UpdateFlexureNormalizer(eGloveSensor(i));
	}
}


//====================================================================
void  Glove::UpdateFlexureNormalizer(eGloveSensor sensor)
{
	minFlexures[sensor] = Min(minFlexures[sensor], flexures[sensor]);
	maxFlexures[sensor] = Max(maxFlexures[sensor], flexures[sensor]);

	float range = maxFlexures[sensor] - minFlexures[sensor];
	if(range != 0.0f) range = 1.0f / range;

	oneOverRangeFlexures[sensor] = range;
}


//====================================================================
void  Glove::ResetFlexureAutoCalibration()
{
	for(uint i = 0; i < GloveSensor_Count; i++) {

		minFlexures[i] = 1.0f;
		maxFlexures[i] = 0.0f;
		oneOverRangeFlexures[i] = 1.0f;
	}
}


//====================================================================
void  Glove::ActivateFlexureAutoCalibration(bool bActivate, bool bResetOnActivate)
{
	bFlexureAutoCalibration = bActivate;

	if(bActivate  &&  bResetOnActivate) {
		ResetFlexureAutoCalibration();
	}
}


//====================================================================
void  Glove::SetMinFlexure(eGloveSensor sensor, float minFlexure)
{
	minFlexures[sensor] = minFlexure;
	UpdateFlexureNormalizer(sensor);
}


//====================================================================
void  Glove::SetMaxFlexure(eGloveSensor sensor, float maxFlexure)
{
	maxFlexures[sensor] = maxFlexure;
	UpdateFlexureNormalizer(sensor);
}


//====================================================================
void  Glove::SetDefaultEqualValueFlexureNormalizations()
{
	for(uint i = 0; i < GloveSensor_Count; i++) {
		equalValueFlexureNorm[i] = EQUAL_VALUE_FLEX_NORM;
	}
}


//====================================================================
void  Glove::SetEqualValueFlexureNormalization(eGloveSensor sensor, float norm)
{
	equalValueFlexureNorm[sensor] = norm;
}


//====================================================================
void  Glove::SetDefaultFlexureToAngleMaps()
{
	flexureToAngleScale[GloveSensor_WristYaw] = FLEX_TO_ANGLE_SCALE_WRIST;
	flexureToAngleBias[GloveSensor_WristYaw]  = FLEX_TO_ANGLE_BIAS_WRIST;

	flexureToAngleScale[GloveSensor_WristPitch] = FLEX_TO_ANGLE_SCALE_WRIST;
	flexureToAngleBias[GloveSensor_WristPitch]  = FLEX_TO_ANGLE_BIAS_WRIST;

	flexureToAngleScale[GloveSensor_WristRoll] = FLEX_TO_ANGLE_SCALE_WRIST;
	flexureToAngleBias[GloveSensor_WristRoll]  = FLEX_TO_ANGLE_BIAS_WRIST;

	flexureToAngleScale[GloveSensor_AbductionThumbIndex] = FLEX_TO_ANGLE_SCALE_ABDUCTION;
	flexureToAngleBias[GloveSensor_AbductionThumbIndex]  = FLEX_TO_ANGLE_BIAS_ABDUCTION;

	flexureToAngleScale[GloveSensor_AbductionIndexMiddle] = FLEX_TO_ANGLE_SCALE_ABDUCTION;
	flexureToAngleBias[GloveSensor_AbductionIndexMiddle]  = FLEX_TO_ANGLE_BIAS_ABDUCTION;

	flexureToAngleScale[GloveSensor_AbductionMiddleRing] = FLEX_TO_ANGLE_SCALE_ABDUCTION;
	flexureToAngleBias[GloveSensor_AbductionMiddleRing]  = FLEX_TO_ANGLE_BIAS_ABDUCTION;

	flexureToAngleScale[GloveSensor_AbductionRingPinky] = FLEX_TO_ANGLE_SCALE_ABDUCTION;
	flexureToAngleBias[GloveSensor_AbductionRingPinky]  = FLEX_TO_ANGLE_BIAS_ABDUCTION;

	flexureToAngleScale[GloveSensor_ThumbRoll] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_ThumbRoll]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_ThumbJointInner] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_ThumbJointInner]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_ThumbJointOuter] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_ThumbJointOuter]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_IndexJointInner] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_IndexJointInner]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_IndexJointMiddle] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_IndexJointMiddle]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_IndexJointOuter] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_IndexJointOuter]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_MiddleJointInner] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_MiddleJointInner]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_MiddleJointMiddle] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_MiddleJointMiddle]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_MiddleJointOuter] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_MiddleJointOuter]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_RingJointInner] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_RingJointInner]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_RingJointMiddle] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_RingJointMiddle]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_RingJointOuter] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_RingJointOuter]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_PinkyJointInner] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_PinkyJointInner]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_PinkyJointMiddle] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_PinkyJointMiddle]  = FLEX_TO_ANGLE_BIAS_FINGER;

	flexureToAngleScale[GloveSensor_PinkyJointOuter] = FLEX_TO_ANGLE_SCALE_FINGER;
	flexureToAngleBias[GloveSensor_PinkyJointOuter]  = FLEX_TO_ANGLE_BIAS_FINGER;
}


//====================================================================
void  Glove::SetFlexureToAngleScale(eGloveSensor sensor, float scaleDeg)
{
	flexureToAngleScale[sensor] = scaleDeg;
}


//====================================================================
void  Glove::SetFlexureToAngleBias(eGloveSensor sensor, float biasDeg)
{
	flexureToAngleBias[sensor] = biasDeg;
}


//====================================================================
void  Glove::ReportHandedness(bool _bLeftHanded)
{
	bLeftHanded = _bLeftHanded;
}


//====================================================================
void  Glove::ReportHardwareCaps_Measurement(bool _bSupportsAngleMeasurement)
{
	bSupportsAngleMeasurement = _bSupportsAngleMeasurement;
}


//====================================================================
void  Glove::ReportHardwareCaps_Sensor(eGloveSensor sensor, bool bSupported)
{
	bSupportsSensor[sensor] = bSupported;
}


//====================================================================
void  Glove::DeriveAnglesFromFlexures()
{
	for(uint iSensor = 0; iSensor < GloveSensor_Count; iSensor++) {
		angles[iSensor] =
			flexureToAngleScale[iSensor] * flexures[iSensor] + flexureToAngleBias[iSensor];
	}
}


//********************************************************************
}//namespace vertexware


