

//--------------------------------------------------------------------
#include "HIDControllers\GloveGestures.h"
#include "BaseTools\Constructs.h"
#include "BaseTools\STLExtensions.h"


//====================================================================
#define  DEFAULT_THRESH_UNFLEXED  0.4f
#define  DEFAULT_THRESH_FLEXED    0.6f
#define  DEFAULT_THRESH_CENTER    0.5f


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
struct UnaryPredicateFindGestureByID
{
	uint  id;

	UnaryPredicateFindGestureByID(uint _id) : id(_id) {}

	bool operator()(const static_gestures::Gesture& g) const { return (g.GetID() == id); }
};


//====================================================================
struct UnaryPredicateFindGestureByName
{
	std::string  name;

	UnaryPredicateFindGestureByName(const std::string& _name) : name(_name) {}

	bool operator()(const static_gestures::Gesture& g) const { return (g.GetName() == name); }
};


//====================================================================
struct UnaryPredicateFindConditionBySensor
{
	eGloveSensor  sensor;

	UnaryPredicateFindConditionBySensor(eGloveSensor _sensor) : sensor(_sensor) {}

	bool operator()(const static_gestures::Condition& cond) const { return (cond.sensor == sensor); }
};


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//********************************************************************
namespace static_gestures {


//====================================================================
Thresholds::Thresholds()
{
	ResetDefaultValues();
}


//====================================================================
void  Thresholds::SetValues(eGloveSensor sensor,
				float threshUnflexed, float threshFlexed, float threshCenter)
{
	unflexed[sensor] = Clamped<float>(threshUnflexed, 0, 1);
	flexed[sensor] = Clamped<float>(threshFlexed, unflexed[sensor], 1);
	center[sensor] = Clamped<float>(threshCenter, 0, 1);
}


//====================================================================
void  Thresholds::ResetDefaultValues()
{
	for(uint i = 0; i < GloveSensor_Count; i++) {
		SetValues(eGloveSensor(i),
			DEFAULT_THRESH_UNFLEXED, DEFAULT_THRESH_FLEXED, DEFAULT_THRESH_CENTER);
	}
}


//====================================================================
Gesture::Gesture(uint _id, const std::string& _gestureName)
{
	id = _id;
	gestureName = _gestureName;
}


//====================================================================
void  Gesture::RemoveCondition(eGloveSensor sensor)
{
	EraseFromContainer_If(conditions, UnaryPredicateFindConditionBySensor(sensor));
}


//====================================================================
void  Gesture::AddCondition(const Condition& cond)
{
	RemoveCondition(cond.sensor);
	conditions.push_back(cond);
}


//====================================================================
GestureAnalyzer::GestureAnalyzer()
{
	nextTestIndex = 0;

	RegisterDefaultGestures();
}


//====================================================================
void  GestureAnalyzer::SetThresholds(const Thresholds& t)
{
	thresholds = t;
}


//====================================================================
void  GestureAnalyzer::RegisterDefaultGestures()
{
	nextTestIndex = 0;
	using namespace standard_gestures;

	// 5-finger
	Register(Grab());
	Register(Stop());
	Register(ThumbUp());

	// 4-finger
	Register(Fist());
	Register(FlatHand());
	Register(IndexPoint());
	Register(NotIndex());
	Register(DoublePoint());
	Register(TriplePoint());
	Register(PinkyPoint());
	Register(UpYours());
	Register(NotUpYours());
	Register(Howzit());
	Register(NotHowzit());
	Register(NotRing());

	// wrist
	Register(RollNeutral());
	Register(RollLeft());
	Register(RollRight());
	Register(PitchNeutral());
	Register(PitchUp());
	Register(PitchDown());
}


//====================================================================
void  GestureAnalyzer::Register(const Gesture& g)
{
	nextTestIndex = 0;
	Unregister(g.GetID());
	gestures.push_back(g);
}


//====================================================================
void  GestureAnalyzer::Unregister(uint gestureID)
{
	nextTestIndex = 0;
	EraseFromContainer_If(gestures, UnaryPredicateFindGestureByID(gestureID));
}


//====================================================================
void  GestureAnalyzer::Unregister(const std::string& gestureName)
{
	nextTestIndex = 0;
	EraseFromContainer_If(gestures, UnaryPredicateFindGestureByName(gestureName));
}


//====================================================================
void  GestureAnalyzer::UnregisterAll()
{
	nextTestIndex = 0;
	gestures.clear();
}


//====================================================================
const Gesture*  GestureAnalyzer::FindRegisteredGesture(uint gestureID) const
{
	for(uint i = 0; i < gestures.size(); i++) {
		if(gestures[i].GetID() == gestureID) return &gestures[i];
	}
	return 0;
}


//====================================================================
const Gesture*  GestureAnalyzer::FindRegisteredGesture(const std::string& gestureName) const
{
	for(uint i = 0; i < gestures.size(); i++) {
		if(gestures[i].GetName() == gestureName) return &gestures[i];
	}
	return 0;
}


//====================================================================
bool  GestureAnalyzer::IsGestureActive(Glove& glove, const Gesture& gesture) const
{
	// test all conditions
	for(uint iCond = 0; iCond < gesture.GetNumConditions(); iCond++) {
		const Condition& cond = gesture.GetCondition(iCond);

		eGloveSensor  sensor   = cond.sensor;
		float         normFlex = glove.GetNormalizedFlexure(sensor);

		switch(cond.flexure)
		{
			case Flexure_Unflexed:
				if(normFlex > thresholds.GetUnflexed(sensor)) return false;
				break;

			case Flexure_Neutral:
				if(normFlex <= thresholds.GetUnflexed(sensor)  ||
					normFlex >= thresholds.GetFlexed(sensor)) return false;
				break;

			case Flexure_Flexed:
				if(normFlex < thresholds.GetFlexed(sensor)) return false;
				break;

			case Flexure_Low:
				if(normFlex >= thresholds.GetCenter(sensor)) return false;
				break;

			case Flexure_High:
				if(normFlex < thresholds.GetCenter(sensor)) return false;
				break;

			default:
				return false;
		}
	}
	return true;
}


//====================================================================
bool  GestureAnalyzer::IsGestureActive(Glove& glove, uint gestureID) const
{
	const Gesture* gesture = FindRegisteredGesture(gestureID);
	if(gesture) return IsGestureActive(glove, *gesture);
	return false;
}


//====================================================================
bool  GestureAnalyzer::IsGestureActive(Glove& glove, const std::string& gestureName) const
{
	const Gesture* gesture = FindRegisteredGesture(gestureName);
	if(gesture) return IsGestureActive(glove, *gesture);
	return false;
}


//====================================================================
const Gesture*  GestureAnalyzer::GetFirstActiveGesture(Glove& glove) const
{
	nextTestIndex = 0;
	return GetNextActiveGesture(glove);
}


//====================================================================
const Gesture*  GestureAnalyzer::GetNextActiveGesture(Glove& glove) const
{
	while(nextTestIndex < gestures.size()) {
		uint index = nextTestIndex++;
		if(IsGestureActive(glove, gestures[index])) return &gestures[index];
	}
	return 0;
}


//********************************************************************
namespace standard_gestures {


//====================================================================
Grab::Grab(const std::string& name) : Gesture(ID_Grab, name)
{
	AddCondition(GloveSensor_ThumbJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_IndexJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Flexed);
}


//====================================================================
Stop::Stop(const std::string& name) : Gesture(ID_Stop, name)
{
	AddCondition(GloveSensor_ThumbJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_IndexJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Unflexed);
}


//====================================================================
ThumbUp::ThumbUp(const std::string& name) : Gesture(ID_ThumbUp, name)
{
	AddCondition(GloveSensor_ThumbJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_IndexJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Flexed);
}


//====================================================================
Fist::Fist(const std::string& name) : Gesture(ID_Fist, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Flexed);
}


//====================================================================
FlatHand::FlatHand(const std::string& name) : Gesture(ID_FlatHand, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Unflexed);
}


//====================================================================
IndexPoint::IndexPoint(const std::string& name) : Gesture(ID_IndexPoint, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Flexed);
}


//====================================================================
NotIndex::NotIndex(const std::string& name) : Gesture(ID_NotIndex, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Unflexed);
}


//====================================================================
DoublePoint::DoublePoint(const std::string& name) : Gesture(ID_DoublePoint, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Flexed);
}


//====================================================================
TriplePoint::TriplePoint(const std::string& name) : Gesture(ID_TriplePoint, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Flexed);
}


//====================================================================
PinkyPoint::PinkyPoint(const std::string& name) : Gesture(ID_PinkyPoint, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Unflexed);
}


//====================================================================
UpYours::UpYours(const std::string& name) : Gesture(ID_UpYours, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Flexed);
}


//====================================================================
NotUpYours::NotUpYours(const std::string& name) : Gesture(ID_NotUpYours, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Unflexed);
}


//====================================================================
Howzit::Howzit(const std::string& name) : Gesture(ID_Howzit, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Unflexed);
}


//====================================================================
NotHowzit::NotHowzit(const std::string& name) : Gesture(ID_NotHowzit, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Flexed);
}


//====================================================================
NotRing::NotRing(const std::string& name) : Gesture(ID_NotRing, name)
{
	AddCondition(GloveSensor_IndexJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_MiddleJointInner, Flexure_Unflexed);
	AddCondition(GloveSensor_RingJointInner, Flexure_Flexed);
	AddCondition(GloveSensor_PinkyJointInner, Flexure_Unflexed);
}


//====================================================================
RollNeutral::RollNeutral(const std::string& name) : Gesture(ID_RollNeutral, name)
{
	AddCondition(GloveSensor_WristRoll, Flexure_Neutral);
}


//====================================================================
RollLeft::RollLeft(const std::string& name) : Gesture(ID_RollLeft, name)
{
	AddCondition(GloveSensor_WristRoll, Flexure_Unflexed);
}


//====================================================================
RollRight::RollRight(const std::string& name) : Gesture(ID_RollRight, name)
{
	AddCondition(GloveSensor_WristRoll, Flexure_Flexed);
}


//====================================================================
PitchNeutral::PitchNeutral(const std::string& name) : Gesture(ID_PitchNeutral, name)
{
	AddCondition(GloveSensor_WristPitch, Flexure_Neutral);
}


//====================================================================
PitchUp::PitchUp(const std::string& name) : Gesture(ID_PitchUp, name)
{
	AddCondition(GloveSensor_WristPitch, Flexure_Unflexed);
}


//====================================================================
PitchDown::PitchDown(const std::string& name) : Gesture(ID_PitchDown, name)
{
	AddCondition(GloveSensor_WristPitch, Flexure_Flexed);
}


//********************************************************************
}//namespace standard_gestures


//********************************************************************
}//namespace static_gestures


//********************************************************************
}//namespace vertexware


