

//--------------------------------------------------------------------
#include "HIDControllers\Keyboard.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"


//********************************************************************
namespace vertexware {


//====================================================================
Keyboard::Keyboard(bool _bSystemKeyboard)
{
	bSystemKeyboard = _bSystemKeyboard;
	bEnableCharacterMapping = false;
	ResetInputReadings();
}


//====================================================================
Keyboard::~Keyboard()
{
}


//====================================================================
std::string  Keyboard::GetControllerInterfaceName() const
{
	return VW_INTERFACE_NAME_Keyboard;
}


//====================================================================
void  Keyboard::DumpControllerInfo(std::ostream& os, uint tabLevel) const
{
	TabEffector  te(tabLevel);
	os << te << "IsSystemKeyboard: " << ToString(bSystemKeyboard) << nl;
}


//====================================================================
void  Keyboard::ResetInputReadings()
{
	ResetKeyStates();
	ResetKeyModifierCache();
	ResetNumKeysCache();
}


//====================================================================
void  Keyboard::PreSampleUpdate()
{
	if(GetParentHID().IsAsyncSamplingActive()) {
		ClearCurrentKeyStates();
	}
	else {
		ShiftKeyStates();
	}
}


//====================================================================
void  Keyboard::PostSampleUpdate()
{
	if(! GetParentHID().IsAsyncSamplingActive()) {
		UpdateKeyModifierCache();
		UpdateNumKeysCache();
	}
}


//====================================================================
void  Keyboard::OnBeginAsyncSampleRead()
{
	UpdateKeyModifierCache();
	UpdateNumKeysCache();
}


//====================================================================
void  Keyboard::OnEndAsyncSampleRead()
{
	ShiftCopyKeyStates();
}


//====================================================================
void  Keyboard::EnableCharacterMapping(bool bEnable)
{
	if(bEnableCharacterMapping != bEnable) {
		bEnableCharacterMapping = bEnable;

		OnChangedProcessingMode();
	}
}


//====================================================================
void  Keyboard::ResetKeyStates()
{
	for(uint i = 0; i < KeyStateStorageSize; i++) {
		keyStates[i] = 0;
	}
}


//====================================================================
void  Keyboard::ClearCurrentKeyStates()
{
	for(uint i = 0; i < KeyStateStorageSize; i++) {
		keyStates[i] &= CurrentStateInv;
	}
}


//====================================================================
void  Keyboard::ShiftKeyStates()
{
	for(uint i = 0; i < KeyStateStorageSize; i++) {
		keyStates[i] >>= StateBitShift;
	}
}


//====================================================================
void  Keyboard::ShiftCopyKeyStates()
{
	for(uint i = 0; i < KeyStateStorageSize; i++) {
		ubyte currState = keyStates[i]  &  CurrentStateBit;
		keyStates[i] = currState | (currState >> StateBitShift);
	}
}


//====================================================================
void  Keyboard::ResetKeyModifierCache()
{
	for(uint i = 0; i < KeyModifierStateStorageSize; i++) {
		keyModifierStates[i] = 0;
	}
}


//====================================================================
void  Keyboard::UpdateKeyModifierCache()
{
	keyModifierStates[KeyModifier_Shift] = KeyPressed(Key_LeftShift) | KeyPressed(Key_RightShift);
	keyModifierStates[KeyModifier_Control] = KeyPressed(Key_LeftControl) | KeyPressed(Key_RightControl);
	keyModifierStates[KeyModifier_Alt] = KeyPressed(Key_LeftAlt) | KeyPressed(Key_RightAlt);

	numKeyModifiersPressed = 0;
	for(uint i = 0; i < KeyModifierStateStorageSize; i++) {
		if(keyModifierStates[i]) {
			numKeyModifiersPressed++;
		}
	}
}


//====================================================================
void  Keyboard::ResetNumKeysCache()
{
	numKeyModifiersPressed = 0;
	numKeysPressed = 0;
	numKeysDown = 0;
	numKeysUp = 0;
}


//====================================================================
void  Keyboard::UpdateNumKeysCache()
{
	numKeysPressed = 0;
	{for(uint i = 0; i < KeyStateStorageSize; i++) {
		if(KeyPressed(eKey(i))) {
			numKeysPressed++;
		}
	}}

	numKeysDown = 0;
	{for(uint i = 0; i < KeyStateStorageSize; i++) {
		if(KeyDown(eKey(i))) {
			numKeysDown++;
		}
	}}

	numKeysUp = 0;
	{for(uint i = 0; i < KeyStateStorageSize; i++) {
		if(KeyUp(eKey(i))) {
			numKeysUp++;
		}
	}}
}


//********************************************************************
}//namespace vertexware


