

//--------------------------------------------------------------------
#include "HIDControllers\Mouse.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"
#include "BaseTools\Constructs.h"
#include "MathTools\MathBase.h"


//********************************************************************
namespace vertexware {


//====================================================================
Mouse::Mouse(bool _bSystemMouse)
{
	bSystemMouse = _bSystemMouse;
	bCanDisplaySystemCursor = false;
	bCanSyncWithSystemCursor = false;

	bActiveSyncWithSystemCursor = true;
	bActiveRelativeMode = false;

	SetReverseDeltaX(false);
	SetReverseDeltaY(false);
	SetReverseDeltaWheel(false);

	SetRelativeModeSensitivity();
	SetAbsoluteModeSpeed();
	SetAbsoluteModeAccelerationThresholds();
	SetAbsoluteModeAccelerationLevel();

	ResetInputReadings();
}


//====================================================================
Mouse::~Mouse()
{
}


//====================================================================
std::string  Mouse::GetControllerInterfaceName() const
{
	return VW_INTERFACE_NAME_Mouse;
}


//====================================================================
void  Mouse::DumpControllerInfo(std::ostream& os, uint tabLevel) const
{
	TabEffector  te(tabLevel);
	os << te << "IsSystemMouse: " << ToString(bSystemMouse) << nl;
}


//====================================================================
void  Mouse::ResetInputReadings()
{
	ResetAxisStates();
	ResetButtonStates();
	ResetNumButtonsCache();
}


//====================================================================
void  Mouse::PreSampleUpdate()
{
	if(GetParentHID().IsAsyncSamplingActive()) {
		ClearCurrentButtonStates();
	}
	else {
		ShiftButtonStates();
		ResetDelta();
		ResetWheelDelta();
	}
}


//====================================================================
void  Mouse::PostSampleUpdate()
{
	if(! GetParentHID().IsAsyncSamplingActive()) {
		UpdateNumButtonsCache();
	}
}


//====================================================================
void  Mouse::OnBeginAsyncSampleRead()
{
	UpdateNumButtonsCache();
}


//====================================================================
void  Mouse::OnEndAsyncSampleRead()
{
	ShiftCopyButtonStates();
	ResetDelta();
	ResetWheelDelta();
}


//====================================================================
void  Mouse::SetReverseDeltaX(bool bReverse)
{
	iRMReverseX = bReverse ? -1 : 1;
}


//====================================================================
void  Mouse::SetReverseDeltaY(bool bReverse)
{
	iRMReverseY = bReverse ? -1 : 1;
}


//====================================================================
void  Mouse::SetReverseDeltaWheel(bool bReverse)
{
	iRMReverseWheel = bReverse ? -1 : 1;
}


//====================================================================
void  Mouse::SetRelativeModeSensitivity(float x, float y)
{
	fRMSensitivityX = x;
	fRMSensitivityY = y;
}


//====================================================================
void  Mouse::GetRelativeModeSensitivity(float& x, float& y) const
{
	x = fRMSensitivityX;
	y = fRMSensitivityY;
}


//====================================================================
void  Mouse::SetAbsoluteModeAccelerationLevel(uint level)
{
	iAMAccelerationLevel = Clamped<uint>(level, 0, 2);
}


//====================================================================
void  Mouse::SetAbsoluteModeSpeed(float speed)
{
	fAMSpeed = ClampedLow<float>(speed, 0);
}


//====================================================================
void  Mouse::SetAbsoluteModeAccelerationThresholds(float t1, float t2)
{
	fAMAccelerationThresholds[0] = ClampedLow<float>(t1, 0);
	fAMAccelerationThresholds[1] = ClampedLow<float>(t2, fAMAccelerationThresholds[0]);
}


//====================================================================
void  Mouse::GetAbsoluteModeAccelerationThresholds(float& t1, float& t2) const
{
	t1 = fAMAccelerationThresholds[0];
	t2 = fAMAccelerationThresholds[1];
}


//====================================================================
void  Mouse::ActivateSyncWithSystemCursor(bool bActivate)
{
	bActiveSyncWithSystemCursor = bActivate;

	OnSyncWithSystemCursorChanged();
}


//====================================================================
void  Mouse::ActivateRelativeMode(bool bActivate)
{
	bActiveRelativeMode = bActivate;

	ResetAxisStates();
	OnInputModeChanged();
}


//====================================================================
void  Mouse::ApplyAbsoluteModeModifiers(float rawX, float rawY, float& finalX, float& finalY)
{
	uint level = 0;
	while(level < iAMAccelerationLevel) {
		float thresh = fAMAccelerationThresholds[level++];
		if(Abs(rawX) > thresh  ||  Abs(rawY) > thresh) {
			rawX *= 2;
			rawY *= 2;
		}
	}
	finalX = rawX * fAMSpeed;
	finalY = rawY * fAMSpeed;
}


//====================================================================
void  Mouse::ApplyRelativeModeModifiers(float rawX, float rawY, float& finalX, float& finalY)
{
	finalX = rawX * fRMSensitivityX;
	finalY = rawY * fRMSensitivityY;
}


//====================================================================
void  Mouse::ResetAxisStates()
{
	bPositionInsideClientArea = false;
	xPosition = 0;
	yPosition = 0;
	xDelta = 0;
	yDelta = 0;
	wheelDelta = 0;
}


//====================================================================
void  Mouse::ResetButtonStates()
{
	for(uint i = 0; i < ButtonStateStorageSize; i++) {
		buttonStates[i] = 0;
	}
}


//====================================================================
void  Mouse::ClearCurrentButtonStates()
{
	for(uint i = 0; i < ButtonStateStorageSize; i++) {
		buttonStates[i] &= CurrentStateInv;
	}
}


//====================================================================
void  Mouse::ShiftButtonStates()
{
	for(uint i = 0; i < ButtonStateStorageSize; i++) {
		buttonStates[i] >>= StateBitShift;
	}
}


//====================================================================
void  Mouse::ShiftCopyButtonStates()
{
	for(uint i = 0; i < ButtonStateStorageSize; i++) {
		ubyte currState = buttonStates[i]  &  CurrentStateBit;
		buttonStates[i] = currState | (currState >> StateBitShift);
	}
}


//====================================================================
void  Mouse::ResetNumButtonsCache()
{
	numButtonsPressed = 0;
	numButtonsDown = 0;
	numButtonsUp = 0;
}


//====================================================================
void  Mouse::UpdateNumButtonsCache()
{
	numButtonsPressed = 0;
	{for(uint i = 0; i < ButtonStateStorageSize; i++) {
		if(ButtonPressed(eMouseButton(i))) {
			numButtonsPressed++;
		}
	}}

	numButtonsDown = 0;
	{for(uint i = 0; i < ButtonStateStorageSize; i++) {
		if(ButtonDown(eMouseButton(i))) {
			numButtonsDown++;
		}
	}}

	numButtonsUp = 0;
	{for(uint i = 0; i < ButtonStateStorageSize; i++) {
		if(ButtonUp(eMouseButton(i))) {
			numButtonsUp++;
		}
	}}
}


//********************************************************************
}//namespace vertexware


