

//--------------------------------------------------------------------
#include "HIDControllers\OrientationTracker.h"
#include "MathTools\TransformTools.h"
#include "BaseTools\String.h"
#include "BaseTools\Stream.h"


//********************************************************************
namespace vertexware {


//====================================================================
OrientationTracker::OrientationTracker()
{
	ResetInputReadings();

	EnableReverseModifiers(true);
	SetReverseHead(false);
	SetReversePitch(false);
	SetReverseRoll(false);

	SetTransform();

	ReportHardwareCaps_ResetControls(false, false, false, false);
	for(uint i = 0; i < SamplingMode_Count; i++) {
		ReportHardwareCaps_SamplingMode(eSamplingMode(i), false);
	}

	samplingMode = SamplingMode_Unknown;

	EnableHPRModifier(false);
}


//====================================================================
OrientationTracker::~OrientationTracker()
{
}


//====================================================================
std::string  OrientationTracker::GetControllerInterfaceName() const
{
	return VW_INTERFACE_NAME_OrientationTracker;
}


//====================================================================
void  OrientationTracker::DumpControllerInfo(std::ostream& os, uint tabLevel) const
{
	TabEffector te(tabLevel);
	os << te << "ResetOrientation: " << ToString(IsSupportedResetOrientation()) << nl;
	os << te << "ResetHead.......: " << ToString(IsSupportedResetHead()) << nl;
	os << te << "ResetPitch......: " << ToString(IsSupportedResetPitch()) << nl;
	os << te << "ResetRoll.......: " << ToString(IsSupportedResetRoll()) << nl;
	os << te << "EulerMode.......: " << ToString(IsSupportedSamplingMode(SamplingMode_Euler)) << nl;
	os << te << "QuaternionMode..: " << ToString(IsSupportedSamplingMode(SamplingMode_Quaternion)) << nl;
}


//====================================================================
void  OrientationTracker::ResetInputReadings()
{
	finalUnitU = vec3::UNIT_X;
	finalUnitV = vec3::UNIT_Y;
	finalUnitN = vec3::UNIT_Z;

	rawUnitU = vec3::UNIT_X;
	rawUnitV = vec3::UNIT_Y;
	rawUnitN = vec3::UNIT_Z;

	rawHead  = 0;
	rawPitch = 0;
	rawRoll  = 0;

	rawQuaternion.Set(vec3::UNIT_Z, 0);
}


//====================================================================
void  OrientationTracker::PreSampleUpdate()
{
}


//====================================================================
void  OrientationTracker::PostSampleUpdate()
{
}


//====================================================================
void  OrientationTracker::OnBeginAsyncSampleRead()
{
}


//====================================================================
void  OrientationTracker::OnEndAsyncSampleRead()
{
}


//====================================================================
bool  OrientationTracker::ResetOrientation()
{
	if(! bSupportsResetOrientation) return false;
	return DoResetOrientation();
}


//====================================================================
bool  OrientationTracker::ResetHead()
{
	if(! bSupportsResetHead) return false;
	return DoResetHead();
}


//====================================================================
bool  OrientationTracker::ResetPitch()
{
	if(! bSupportsResetPitch) return false;
	return DoResetPitch();
}


//====================================================================
bool  OrientationTracker::ResetRoll()
{
	if(! bSupportsResetRoll) return false;
	return DoResetRoll();
}


//====================================================================
bool  OrientationTracker::SetSamplingMode(eSamplingMode mode)
{
	ResetInputReadings();
	if(mode == samplingMode) return true;
	if(! IsSupportedSamplingMode(mode)) return false;
	if(DoChangeSamplingMode(samplingMode)) {
		samplingMode = mode;
		return true;
	}
	return false;
}


//====================================================================
void  OrientationTracker::EnableHPRModifier(bool bEnable)
{
	bEnableHPRModifier = bEnable;
}


//====================================================================
void  OrientationTracker::EnableReverseModifiers(bool bEnable)
{
	bEnableReverseModifiers = bEnable;
}


//====================================================================
void  OrientationTracker::SetReverseHead(bool bReverse)
{
	fReverseHead = bReverse ? -1 : 1;
}


//====================================================================
void  OrientationTracker::SetReversePitch(bool bReverse)
{
	fReversePitch = bReverse ? -1 : 1;
}


//====================================================================
void  OrientationTracker::SetReverseRoll(bool bReverse)
{
	fReverseRoll = bReverse ? -1 : 1;
}


//====================================================================
void  OrientationTracker::SetTransform(const mat3& _transformMV)
{
	transformMV = _transformMV;
}


//====================================================================
void  OrientationTracker::ReportHardwareCaps_ResetControls(
										bool _bSupportsResetOrientation,
										bool _bSupportsResetHead,
										bool _bSupportsResetPitch,
										bool _bSupportsResetRoll)
{
	bSupportsResetOrientation = _bSupportsResetOrientation;
	bSupportsResetHead = _bSupportsResetHead;
	bSupportsResetPitch = _bSupportsResetPitch;
	bSupportsResetRoll = _bSupportsResetRoll;
}


//====================================================================
void  OrientationTracker::ReportHardwareCaps_SamplingMode(
										eSamplingMode mode, bool bSupported)
{
	bSupportsSamplingMode[mode] = bSupported;
}


//====================================================================
void  OrientationTracker::SetOrientation(const vec3& axisV, const vec3& axisN)
{
	// update raw data
	rawUnitV = axisV.Unit();
	rawUnitN = axisN.Unit();
	rawUnitU.Cross(rawUnitV, rawUnitN);

	// calculate final orientation, apply reverse modifiers if requested
	if(bEnableReverseModifiers) {

		// analyze rotation matrix
		mat3  mRot;
		GenerateRotateMV(mRot, rawUnitU, rawUnitV, rawUnitN);
		float h, p, r;
		FactorEulerHPR_MV(mRot, h, p, r);

		// build final matrix
		mat3  mFinal = transformMV;
		mat3  temp;

		GenerateRotateMV(temp, fReverseRoll * r, Axis_Z);
		mFinal *= temp;

		GenerateRotateMV(temp, fReversePitch * p, Axis_X);
		mFinal *= temp;

		GenerateRotateMV(temp, fReverseHead * h, Axis_Y);
		mFinal *= temp;

		// analyze final matrix
		FactorRotateMV(mFinal, finalUnitU, finalUnitV, finalUnitN);
	}
	else {
		finalUnitU = transformMV * rawUnitU;
		finalUnitV = transformMV * rawUnitV;
		finalUnitN = transformMV * rawUnitN;
	}
}


//====================================================================
void  OrientationTracker::SetOrientation(const vec3& direction, float rollDeg)
{
	// force valid range
	RemapAngleTo180DegRange(rollDeg);

	// convert to UVN
	float  h, p;
	EulerHPFromDirection(direction.Unit(), h, p);

	mat3  mRot;
	GenerateEulerHPR_MV(mRot, h, p, rollDeg);

	vec3  u, v, n;
	FactorRotateMV(mRot, u, v, n);
	SetOrientation(v, n);
}


//====================================================================
void  OrientationTracker::SetOrientation(float headDeg, float pitchDeg, float rollDeg)
{
	// copy raw data
	rawHead  = headDeg;
	rawPitch = pitchDeg;
	rawRoll  = rollDeg;

	// force valid range
	RemapAngleTo180DegRange(headDeg);
	RemapAngleTo180DegRange(rollDeg);
	RemapAngleTo90DegRange(pitchDeg);

	// convert from PHR-order to HPR-order
	if(bEnableHPRModifier) {
		mat3  mRot;
		GenerateEulerPHR_MV(mRot, headDeg, pitchDeg, rollDeg);
		FactorEulerHPR_MV(mRot, headDeg, pitchDeg, rollDeg);
	}

	// convert to UVN
	mat3  mRot;
	GenerateEulerHPR_MV(mRot, headDeg, pitchDeg, rollDeg);

	vec3  u, v, n;
	FactorRotateMV(mRot, u, v, n);
	SetOrientation(v, n);
}


//====================================================================
void  OrientationTracker::SetOrientation(const quat& quat)
{
	// copy raw data
	rawQuaternion = quat;

	// convert to UVN
	mat3  mRot;
	quat.Unit().ToRotationMatrixMV(mRot);

	vec3  u, v, n;
	FactorRotateMV(mRot, u, v, n);
	SetOrientation(v, n);
}


//********************************************************************
}//namespace vertexware


