

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "Framework\HIDController.h"
#include "MathTools\Vector3.h"
#include "MathTools\Matrix33.h"
#include "MathTools\Quaternion.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Drivers must report orientation in right-handed-coordinate-system.
// The orientation is reported in HeadPitchRoll, Quaternion
// or UVN (CoordSysAxes) format.
// Default sampling format is Euler if available (otherwise Unknown).
// Reference orientation:
//    HPR(0, 0, 0)
//    QUAT((0, 0, 0), 1)
//    DRA((0, 0, 1), 0)
//    UVN((1, 0, 0), (0, 1, 0), (0, 0, 1))
// which means the viewer is looking straight ahead towards +zInfinity.
//
// Positive Head - viewer turns head left
// Negative Head - viewer turns head right
// Valid range: [-180.0, 180.0] degrees
//
// Positive Pitch - viewer tilts head down
// Negative Pitch - viewer tilts head up
// Valid range: [-90.0, 90.0] degrees
//
// Positive Roll - viewer twists head right
// Negative Roll - viewer twists head left
// Valid range: [-180.0, 180.0] degrees
//
// All angles must be specified in degrees.
// All reported values must be within valid ranges.
// When the tracking-device is properly attached to the tracked
// object (e.g. user's head or hmd), driver must return
// the data-values according to given rules. If the actual hardware
// behaves differently make the necessary adjustments to match
// these rules.
//
//====================================================================
class VW_COMMON_API OrientationTracker : public HIDController
{
public:
	enum eSamplingMode
	{
		SamplingMode_Unknown = 0,
		SamplingMode_Euler,
		SamplingMode_Quaternion,

		SamplingMode_Count,
		SamplingMode_Force32Bits = 0x7FFFFFFF
	};


public:
	virtual ~OrientationTracker();

	//------------------------------
	// HIDController implementation.
	//------------------------------
	virtual  std::string  GetControllerInterfaceName() const;
	virtual  void         DumpControllerInfo(std::ostream& os, uint tabLevel) const;
	virtual  void         ResetInputReadings();
	virtual  void         PreSampleUpdate();
	virtual  void         PostSampleUpdate();
	virtual  void         OnBeginAsyncSampleRead();
	virtual  void         OnEndAsyncSampleRead();


	//------------------------------
	// OrientationTracker interface.
	//------------------------------

	// hardware capabilities
	bool  IsSupportedResetOrientation() const { return bSupportsResetOrientation; }
	bool  IsSupportedResetHead() const { return bSupportsResetHead; }
	bool  IsSupportedResetPitch() const { return bSupportsResetPitch; }
	bool  IsSupportedResetRoll() const { return bSupportsResetRoll; }
	bool  IsSupportedSamplingMode(eSamplingMode mode) const { return bSupportsSamplingMode[mode]; }

	// reset controls
	bool  ResetOrientation();
	bool  ResetHead();
	bool  ResetPitch();
	bool  ResetRoll();

	// sampling-mode control
	eSamplingMode  GetSamplingMode() const { return samplingMode; }
	bool           SetSamplingMode(eSamplingMode mode);

	// processing of raw-euler-angles (changes PHR to HPR order)
	bool  IsEnabledHPRModifier() const { return bEnableHPRModifier; }
	void  EnableHPRModifier(bool bEnable); //default is set by device-driver

	// processing of raw-orientation-axes
	bool  AreEnabledReverseModifiers() const { return bEnableReverseModifiers; }
	void  EnableReverseModifiers(bool bEnable); //enabled by default

	bool  IsReversedHead() const { return (fReverseHead < 0); }
	void  SetReverseHead(bool bReverse);

	bool  IsReversedPitch() const { return (fReversePitch < 0); }
	void  SetReversePitch(bool bReverse);

	bool  IsReversedRoll() const { return (fReverseRoll < 0); }
	void  SetReverseRoll(bool bReverse);

	void  SetTransform(const mat3& transformMV = mat3::IDENTITY);

	// final orientation (valid for all sampling-modes)
	void  GetOrientation(vec3& axisUnitU, vec3& axisUnitV, vec3& axisUnitN) const
	{
		axisUnitU = finalUnitU;
		axisUnitV = finalUnitV;
		axisUnitN = finalUnitN;
	}

	// raw sampled orientation (valid for all sampling-modes)
	void  GetOrientationRaw(vec3& axisUnitU, vec3& axisUnitV, vec3& axisUnitN) const
	{
		axisUnitU = rawUnitU;
		axisUnitV = rawUnitV;
		axisUnitN = rawUnitN;
	}

	// raw sampled values (valid if sampling-mode is Euler)
	void  GetOrientationRaw(float& headDeg, float& pitchDeg, float& rollDeg) const
	{
		headDeg  = rawHead;
		pitchDeg = rawPitch;
		rollDeg  = rawRoll;
	}

	// raw sampled values (valid if sampling-mode is Quaternion)
	void  GetOrientationRaw(quat& quat) const
	{
		quat = rawQuaternion;
	}


private:
	// hardware capabilities
	bool  bSupportsResetOrientation;
	bool  bSupportsResetHead;
	bool  bSupportsResetPitch;
	bool  bSupportsResetRoll;
	bool  bSupportsSamplingMode[SamplingMode_Count];

	// operational states
	eSamplingMode  samplingMode;

	bool   bEnableReverseModifiers;
	float  fReverseHead;
	float  fReversePitch;
	float  fReverseRoll;

	bool  bEnableHPRModifier;

	mat3  transformMV;

	// final orientation axes
	vec3  finalUnitU;
	vec3  finalUnitV;
	vec3  finalUnitN;

	// raw orientation axes: all sampling modes
	vec3  rawUnitU;
	vec3  rawUnitV;
	vec3  rawUnitN;

	// raw values: SamplingMode_Euler
	float  rawHead;
	float  rawPitch;
	float  rawRoll;

	// raw values: SamplingMode_Quaternion
	quat  rawQuaternion;


protected:
	//-------------------------
	// Derived-class interface.
	//-------------------------
	OrientationTracker();

	void  ReportHardwareCaps_ResetControls(bool bSupportsResetOrientation,
											bool bSupportsResetHead,
											bool bSupportsResetPitch,
											bool bSupportsResetRoll);

	void  ReportHardwareCaps_SamplingMode(eSamplingMode mode, bool bSupported);

	void  SetOrientation(const vec3& axisV, const vec3& axisN);
	void  SetOrientation(const vec3& direction, float rollDeg);
	void  SetOrientation(float headDeg, float pitchDeg, float rollDeg);
	void  SetOrientation(const quat& quat);

	//------------------
	// Template methods.
	//------------------
	virtual  bool  DoResetOrientation() = 0;
	virtual  bool  DoResetHead() = 0;
	virtual  bool  DoResetPitch() = 0;
	virtual  bool  DoResetRoll() = 0;
	virtual  bool  DoChangeSamplingMode(eSamplingMode mode) = 0;
};


//====================================================================
#define  VW_INTERFACE_NAME_OrientationTracker  "OrientationTracker"


//********************************************************************
}//namespace vertexware


