

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "Framework\HIDController.h"
#include "MathTools\Vector3.h"
#include "MathTools\Matrix44.h"


//********************************************************************
namespace vertexware {


//====================================================================
// Drivers must report position in right-handed-coordinate-system.
// All values must be reported in meters.
//====================================================================
class VW_COMMON_API PositionTracker : public HIDController
{
public:
	virtual ~PositionTracker();

	//------------------------------
	// HIDController implementation.
	//------------------------------
	virtual  std::string  GetControllerInterfaceName() const;
	virtual  void         DumpControllerInfo(std::ostream& os, uint tabLevel) const;
	virtual  void         ResetInputReadings();
	virtual  void         PreSampleUpdate();
	virtual  void         PostSampleUpdate();
	virtual  void         OnBeginAsyncSampleRead();
	virtual  void         OnEndAsyncSampleRead();


	//---------------------------
	// PositionTracker interface.
	//---------------------------

	// hardware capabilities
	bool  IsSupportedResetPosition() const { return bSupportsResetPosition; }
	bool  IsSupportedResetX() const { return bSupportsResetX; }
	bool  IsSupportedResetY() const { return bSupportsResetY; }
	bool  IsSupportedResetZ() const { return bSupportsResetZ; }

	// reset control
	bool  ResetPosition();
	bool  ResetX();
	bool  ResetY();
	bool  ResetZ();

	// processing of raw-position
	bool  IsReversedX() const { return (fReverseX < 0); }
	void  SetReverseX(bool bReverse);

	bool  IsReversedY() const { return (fReverseY < 0); }
	void  SetReverseY(bool bReverse);

	bool  IsReversedZ() const { return (fReverseZ < 0); }
	void  SetReverseZ(bool bReverse);

	void  SetTransform(const mat4& transformMV = mat4::IDENTITY);

	// final position
	const vec3&  GetPosition() const { return finalPosition; }

	// raw sampled position
	const vec3&  GetPositionRaw() const { return rawPosition; }


private:
	// hardware capabilities
	bool  bSupportsResetPosition;
	bool  bSupportsResetX;
	bool  bSupportsResetY;
	bool  bSupportsResetZ;

	// operational states
	float     fReverseX;
	float     fReverseY;
	float     fReverseZ;
	mat4  transformMV;

	// final position
	vec3  finalPosition;

	// raw position
	vec3  rawPosition;


protected:
	//-------------------------
	// Derived-class interface.
	//-------------------------
	PositionTracker();

	void  ReportHardwareCaps(bool bSupportsResetPosition,
							bool bSupportsResetX,
							bool bSupportsResetY,
							bool bSupportsResetZ);

	void  SetPosition(const vec3& position);

	//------------------
	// Template methods.
	//------------------
	virtual  bool  DoResetPosition() = 0;
	virtual  bool  DoResetX() = 0;
	virtual  bool  DoResetY() = 0;
	virtual  bool  DoResetZ() = 0;
};


//====================================================================
#define  VW_INTERFACE_NAME_PositionTracker  "PositionTracker"


//********************************************************************
}//namespace vertexware


