

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "MathTools\MatrixBase.h"
#include "MathTools\Vector2.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class mat3;
class mat4;


//====================================================================
// 2 x 2 matrix.
// Storage form is row-major.
//====================================================================
class VW_COMMON_API mat2
{
public:
	//------
	// Data.
	//------
	union {
		float  mm[2][2];
		float  m[4];
		struct {
			float  m00, m01;
			float  m10, m11;
		};
	};

	//-----------------------
	// Convenience constants.
	//-----------------------
	static const mat2  ZERO;
	static const mat2  IDENTITY;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline mat2()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline mat2(float _00, float _01, float _10, float _11) {
		m00 = _00;
		m01 = _01;
		m10 = _10;
		m11 = _11;
	}

	__forceinline mat2(const float* a) {
		m[0] = a[0];
		m[1] = a[1];
		m[2] = a[2];
		m[3] = a[3];
	}

	__forceinline mat2(const vec2& row0, const vec2& row1) {
		m00 = row0.x;
		m01 = row0.y;
		m10 = row1.x;
		m11 = row1.y;
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline mat2(const mat2& rhs) {
		m00 = rhs.m00;
		m01 = rhs.m01;
		m10 = rhs.m10;
		m11 = rhs.m11;
	}

	__forceinline mat2(const mat3& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m10 = reinterpret_cast<const float*>(&rhs)[3];
		m11 = reinterpret_cast<const float*>(&rhs)[4];
	}

	__forceinline mat2(const mat4& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m10 = reinterpret_cast<const float*>(&rhs)[4];
		m11 = reinterpret_cast<const float*>(&rhs)[5];
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline mat2& operator=(const mat2& rhs) {
		m00 = rhs.m00;
		m01 = rhs.m01;
		m10 = rhs.m10;
		m11 = rhs.m11;
		return *this;
	}

	mat2& operator=(const mat3& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m10 = reinterpret_cast<const float*>(&rhs)[3];
		m11 = reinterpret_cast<const float*>(&rhs)[4];
		return *this;
	}

	mat2& operator=(const mat4& rhs) {
		m00 = reinterpret_cast<const float*>(&rhs)[0];
		m01 = reinterpret_cast<const float*>(&rhs)[1];
		m10 = reinterpret_cast<const float*>(&rhs)[4];
		m11 = reinterpret_cast<const float*>(&rhs)[5];
		return *this;
	}

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(float _00, float _01, float _10, float _11) {
		m00 = _00;
		m01 = _01;
		m10 = _10;
		m11 = _11;
	}

	__forceinline void  Set(const float* a) {
		m[0] = a[0];
		m[1] = a[1];
		m[2] = a[2];
		m[3] = a[3];
	}

	__forceinline void  Set(const vec2& row0, const vec2& row1) {
		m00 = row0.x;
		m01 = row0.y;
		m10 = row1.x;
		m11 = row1.y;
	}

	//---------------
	// Special forms.
	//---------------
	__forceinline void  SetIdentity() {
		m00 = 1;
		m01 = 0;
		m10 = 0;
		m11 = 1;
	}

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline float*        ToPtr()       { return &m00; }
	__forceinline const float*  ToPtr() const { return &m00; }

	__forceinline operator float*()             { return ToPtr(); }
	__forceinline operator const float*() const { return ToPtr(); }

	//------------------
	// Grid-like access.
	//------------------
	__forceinline float*        operator[](uint i)       { return &mm[i][0]; }
	__forceinline const float*  operator[](uint i) const { return &mm[i][0]; }

	__forceinline float*        operator[](int i)       { return &mm[i][0]; }
	__forceinline const float*  operator[](int i) const { return &mm[i][0]; }

	//------------
	// Row access.
	//------------
	__forceinline vec2&        Row0()       { return reinterpret_cast<vec2&>(mm[0][0]); }
	__forceinline const vec2&  Row0() const { return reinterpret_cast<const vec2&>(mm[0][0]); }

	__forceinline vec2&        Row1()       { return reinterpret_cast<vec2&>(mm[1][0]); }
	__forceinline const vec2&  Row1() const { return reinterpret_cast<const vec2&>(mm[1][0]); }

	__forceinline vec2&        Row(uint i)       { return reinterpret_cast<vec2&>(mm[i][0]); }
	__forceinline const vec2&  Row(uint i) const { return reinterpret_cast<const vec2&>(mm[i][0]); }

	__forceinline vec2&        Row(int i)       { return reinterpret_cast<vec2&>(mm[i][0]); }
	__forceinline const vec2&  Row(int i) const { return reinterpret_cast<const vec2&>(mm[i][0]); }

	//-------------------
	// Column generators.
	//-------------------
	__forceinline vec2  Column0() const {
		return vec2(m00, m10);
	}
	__forceinline void  Column0(vec2& dst) const {
		dst.Set(m00, m10);
	}

	__forceinline vec2  Column1() const {
		return vec2(m01, m11);
	}
	__forceinline void  Column1(vec2& dst) const {
		dst.Set(m01, m11);
	}

	__forceinline vec2  Column(uint i) const {
		return vec2(mm[0][i], mm[1][i]);
	}
	__forceinline void  Column(uint i, vec2& dst) const {
		dst.Set(mm[0][i], mm[1][i]);
	}

	__forceinline vec2  Column(int i) const {
		return vec2(mm[0][i], mm[1][i]);
	}
	__forceinline void  Column(int i, vec2& dst) const {
		dst.Set(mm[0][i], mm[1][i]);
	}

	//----------
	// Addition.
	//----------
	__forceinline mat2&  operator+=(const mat2& rhs) {
		m00 += rhs.m00;
		m01 += rhs.m01;
		m10 += rhs.m10;
		m11 += rhs.m11;
		return *this;
	}

	__forceinline mat2&  Add(const mat2& rhs) {
		m00 += rhs.m00;
		m01 += rhs.m01;
		m10 += rhs.m10;
		m11 += rhs.m11;
		return *this;
	}

	__forceinline mat2  operator+(const mat2& rhs) const {
		return mat2(m00 + rhs.m00, m01 + rhs.m01, m10 + rhs.m10, m11 + rhs.m11);
	}

	__forceinline mat2&  Add(const mat2& lhs, const mat2& rhs) {
		m00 = lhs.m00 + rhs.m00;
		m01 = lhs.m01 + rhs.m01;
		m10 = lhs.m10 + rhs.m10;
		m11 = lhs.m11 + rhs.m11;
		return *this;
	}

	//-------------
	// Subtraction.
	//-------------
	__forceinline mat2&  operator-=(const mat2& rhs) {
		m00 -= rhs.m00;
		m01 -= rhs.m01;
		m10 -= rhs.m10;
		m11 -= rhs.m11;
		return *this;
	}

	__forceinline mat2&  Sub(const mat2& rhs) {
		m00 -= rhs.m00;
		m01 -= rhs.m01;
		m10 -= rhs.m10;
		m11 -= rhs.m11;
		return *this;
	}

	__forceinline mat2  operator-(const mat2& rhs) const {
		return mat2(m00 - rhs.m00, m01 - rhs.m01, m10 - rhs.m10, m11 - rhs.m11);
	}

	__forceinline mat2&  Sub(const mat2& lhs, const mat2& rhs) {
		m00 = lhs.m00 - rhs.m00;
		m01 = lhs.m01 - rhs.m01;
		m10 = lhs.m10 - rhs.m10;
		m11 = lhs.m11 - rhs.m11;
		return *this;
	}

	//--------------------------
	// Multiplication by matrix.
	//--------------------------
	__forceinline mat2&  operator*=(const mat2& rhs) {
		return this->Mul(mat2(*this), rhs);
	}

	__forceinline mat2  operator*(const mat2& rhs) const {
		mat2 temp;
		temp.Mul(*this, rhs);
		return temp;
	}

	__forceinline mat2&  PreMul(const mat2& lhs) {
		return this->Mul(lhs, mat2(*this));
	}

	__forceinline mat2&  PostMul(const mat2& rhs) {
		return this->Mul(mat2(*this), rhs);
	}

	mat2&  Mul(const mat2& lhs, const mat2& rhs);

	//--------------------------
	// Multiplication by scalar.
	//--------------------------
	__forceinline mat2&  operator*=(float num) {
		m00 *= num;
		m01 *= num;
		m10 *= num;
		m11 *= num;
		return *this;
	}

	__forceinline mat2&  Mul(float num) {
		m00 *= num;
		m01 *= num;
		m10 *= num;
		m11 *= num;
		return *this;
	}

	__forceinline mat2  operator*(float num) const {
		return mat2(m00 * num, m01 * num, m10 * num, m11 * num);
	}

	__forceinline mat2&  Mul(const mat2& lhs, float num) {
		m00 = lhs.m00 * num;
		m01 = lhs.m01 * num;
		m10 = lhs.m10 * num;
		m11 = lhs.m11 * num;
		return *this;
	}

	//----------------------------------------------
	// Division by scalar (no divide-by-zero check).
	//----------------------------------------------
	__forceinline mat2&  operator/=(float num) {
		return this->Mul(1 / num);
	}

	__forceinline mat2&  Div(float num) {
		return this->Mul(1 / num);
	}

	__forceinline mat2  operator/(float num) const {
		num = 1 / num;
		return mat2(m00 * num, m01 * num, m10 * num, m11 * num);
	}

	__forceinline mat2&  Div(const mat2& lhs, float num) {
		return this->Mul(lhs, 1 / num);
	}

	//----------
	// Negation.
	//----------
	__forceinline mat2&  Neg() {
		return this->Mul(*this, -1);
	}

	__forceinline mat2&  Neg(const mat2& src) {
		return this->Mul(src, -1);
	}

	//------------------
	// Unary operations.
	//------------------
	__forceinline mat2  operator-() const {
		return mat2(-m00, -m01, -m10, -m11);
	}

	__forceinline mat2&        operator+()       { return *this; }
	__forceinline const mat2&  operator+() const { return *this; }

	//-----------
	// Transpose.
	//-----------
	__forceinline mat2&  Transpose() {
		Swap(m10, m01);
		return *this;
	}

	__forceinline mat2  Transposed() const {
		return mat2(m00, m10, m01, m11);
	}

	__forceinline mat2&  Transposed(const mat2& src) {
		Set(src.m00, src.m10, src.m01, src.m11);
		return *this;
	}

	//---------
	// Inverse.
	//---------
	__forceinline mat2&  Invert(bool* pFailed = 0) {
		return this->Inverse(*this, pFailed);
	}

	__forceinline mat2  Inverse(bool* pFailed = 0) const {
		mat2 temp;
		return temp.Inverse(*this, pFailed);
	}

	mat2&  Inverse(const mat2& src, bool* pFailed = 0);

	//------------------------------
	// Inverse (exception versions).
	//------------------------------
	__forceinline mat2&  InvertE() {
		return this->InverseE(*this);
	}

	__forceinline mat2  InverseE() const {
		mat2 temp;
		return temp.InverseE(*this);
	}

	__forceinline mat2&  InverseE(const mat2& src) {
		bool failed;
		this->Inverse(src, &failed);
		if(failed) {
			VW_THROW(MatrixNotInvertibleException, "Uninvertible matrix.");
		}
		return *this;
	}

	//-------
	// Trace.
	//-------
	__forceinline float  Trace() const {
		return  (m00 + m11);
	}

	//-------------
	// Determinant.
	//-------------
	float  Determinant() const {
		return  (m00 * m11 - m01 * m10);
	}

	//-------------
	// Comparisons.
	//-------------
	__forceinline bool  operator==(const mat2& rhs) const {
		return ((m00 == rhs.m00) && (m01 == rhs.m01) && (m10 == rhs.m10) && (m11 == rhs.m11));
	}
	__forceinline bool  operator!=(const mat2& rhs) const {
		return !(*this == rhs);
	}

	//----------------------------------------------------
	// Noise filtering (zero-out all elements close to 0).
	//----------------------------------------------------
	__forceinline mat2&  NoiseFilter(const mat2& src, float tolerance =
										math_constants::MatrixNoiseFilterDefaultTolerance) {
		vertexware::NoiseFilter(this->ToPtr(), 4, tolerance);
		return *this;
	}

	__forceinline  mat2 __fastcall NoiseFiltered(float tolerance =
											math_constants::MatrixNoiseFilterDefaultTolerance) const {
		mat2 temp;
		vertexware::NoiseFilter(temp.ToPtr(), this->ToPtr(), 4, tolerance);
		return temp;
	}
};


//====================================================================
// Multiplication by scalar (redirection).
//====================================================================
__forceinline  mat2 __fastcall operator*(float num, const mat2& rhs) {
	return rhs * num;
}


//====================================================================
// Row-column dot product.
//====================================================================
__forceinline float __fastcall Dot(const mat2& lhs, uint row, const mat2& rhs, uint column)
{
	return
	lhs.mm[row][0] * rhs.mm[0][column] + lhs.mm[row][1] * rhs.mm[1][column];
}


//====================================================================
// Row-column dot product.
// Usage example:
//    mat2 a, b;
//    float d = Dot01(a, b); // takes dot of row0 from a & column1 from b
//====================================================================
#define  VW_IMPLEMENT_MATRIX22_DOT(row, column) \
	__forceinline float __fastcall Dot##row##column(const mat2& lhs, const mat2& rhs) { \
		return \
		lhs.mm[row][0] * rhs.mm[0][column] + lhs.mm[row][1] * rhs.mm[1][column]; \
	}


VW_IMPLEMENT_MATRIX22_DOT(0, 0)
VW_IMPLEMENT_MATRIX22_DOT(0, 1)
VW_IMPLEMENT_MATRIX22_DOT(1, 0)
VW_IMPLEMENT_MATRIX22_DOT(1, 1)


#undef  VW_IMPLEMENT_MATRIX22_DOT


//====================================================================
// Multiplication matrix * vector.
//====================================================================
__forceinline vec2& __fastcall Mul(vec2& dst, const mat2& m, const vec2& v)
{
	dst.Set(m.Row0().Dot(v), m.Row1().Dot(v));
	return dst;
}


//====================================================================
__forceinline vec2 __fastcall operator*(const mat2& m, const vec2& v)
{
	return vec2(m.Row0().Dot(v), m.Row1().Dot(v));
}


//====================================================================
// Multiplication vector * matrix.
//====================================================================
__forceinline vec2& __fastcall Mul(vec2& dst, const vec2& v, const mat2& m)
{
	dst.Set(v.x * m.m00 + v.y * m.m10,
			v.x * m.m01 + v.y * m.m11);
	return dst;
}


//====================================================================
__forceinline vec2 __fastcall operator*(const vec2& v, const mat2& m)
{
	return
	vec2(v.x * m.m00 + v.y * m.m10,
			v.x * m.m01 + v.y * m.m11);
}


//====================================================================
// Element-access indirectors. Useful for MV/VM matrix analyzers.
//====================================================================
template<typename TTranspose>
struct Matrix22Accessor
{ };


//====================================================================
template<>
struct Matrix22Accessor<no_transpose_t>
{
	mat2& m;

	Matrix22Accessor(mat2& _m) : m(_m) {}
	__forceinline float&  operator()(uint row, uint column) { return m.mm[row][column]; };
};


//====================================================================
template<>
struct Matrix22Accessor<transpose_t>
{
	mat2& m;

	Matrix22Accessor(mat2& _m) : m(_m) {}
	__forceinline float&  operator()(uint row, uint column) { return m.mm[column][row]; };
};


//====================================================================
template<typename TTranspose>
struct Matrix22ConstAccessor
{ };


//====================================================================
template<>
struct Matrix22ConstAccessor<no_transpose_t>
{
	const mat2& m;

	Matrix22ConstAccessor(const mat2& _m) : m(_m) {}
	__forceinline float  operator()(uint row, uint column) { return m.mm[row][column]; };
};


//====================================================================
template<>
struct Matrix22ConstAccessor<transpose_t>
{
	const mat2& m;

	Matrix22ConstAccessor(const mat2& _m) : m(_m) {}
	__forceinline float  operator()(uint row, uint column) { return m.mm[column][row]; };
};


//====================================================================
// Fill indirectors. Useful for MV/VM matrix generators.
//====================================================================
template<typename TTranspose>
__forceinline void __fastcall Fill(const TTranspose&, mat2& m,
									float _00, float _01,
									float _10, float _11);


//====================================================================
template<>
__forceinline void __fastcall Fill(const no_transpose_t&, mat2& m,
										float _00, float _01,
										float _10, float _11)
{
	m.m00 = _00;
	m.m01 = _01;
	m.m10 = _10;
	m.m11 = _11;
}


//====================================================================
template<>
__forceinline void __fastcall Fill(const transpose_t&, mat2& m,
										float _00, float _01,
										float _10, float _11)
{
	m.m00 = _00;
	m.m01 = _10;
	m.m10 = _01;
	m.m11 = _11;
}


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const mat2& m);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, mat2& m);


//********************************************************************
}//namespace vertexware



