

//--------------------------------------------------------------------
#include "MathTools\Matrix44.h"
#include "MathTools\Matrix33.h"
#include "BaseTools\Serializers.h"


//====================================================================
// Define to compile with optimized inverse (about 33% faster).
//====================================================================
#define  VW_MATRIX44_INVERSE_OPTIMIZE


//********************************************************************
namespace vertexware {


//====================================================================
const mat4  mat4::ZERO     = mat4(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
const mat4  mat4::IDENTITY = mat4(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);


//====================================================================
mat4&  mat4::Mul(const mat4& lhs, const mat4& rhs)
{
	m00 = Dot00(lhs, rhs);
	m01 = Dot01(lhs, rhs);
	m02 = Dot02(lhs, rhs);
	m03 = Dot03(lhs, rhs);
	m10 = Dot10(lhs, rhs);
	m11 = Dot11(lhs, rhs);
	m12 = Dot12(lhs, rhs);
	m13 = Dot13(lhs, rhs);
	m20 = Dot20(lhs, rhs);
	m21 = Dot21(lhs, rhs);
	m22 = Dot22(lhs, rhs);
	m23 = Dot23(lhs, rhs);
	m30 = Dot30(lhs, rhs);
	m31 = Dot31(lhs, rhs);
	m32 = Dot32(lhs, rhs);
	m33 = Dot33(lhs, rhs);
	return *this;
}


//====================================================================
// Inverse optimization helpers.
//====================================================================
#ifdef  VW_MATRIX44_INVERSE_OPTIMIZE


//--------------------------------------------------------------------
// Subtract scaled.
//--------------------------------------------------------------------
__forceinline  void __fastcall Sub_YZW(vec4& d, const vec4& s, float a) {
	d.y -= s.y * a;
	d.z -= s.z * a;
	d.w -= s.w * a;
}

//--------------------------------------------------------------------
__forceinline  void __fastcall Sub_ZW(vec4& d, const vec4& s, float a) {
	d.z -= s.z * a;
	d.w -= s.w * a;
}

//--------------------------------------------------------------------
__forceinline  void __fastcall Sub_W(vec4& d, const vec4& s, float a) {
	d.w -= s.w * a;
}

//--------------------------------------------------------------------
// Multiply.
//--------------------------------------------------------------------
__forceinline  void __fastcall Mul_YZW(vec4& d, float a) {
	d.y *= a;
	d.z *= a;
	d.w *= a;
}

//--------------------------------------------------------------------
__forceinline  void __fastcall Mul_ZW(vec4& d, float a) {
	d.z *= a;
	d.w *= a;
}

//--------------------------------------------------------------------
__forceinline  void __fastcall Mul_W(vec4& d, float a) {
	d.w *= a;
}

//--------------------------------------------------------------------
// Swap.
//--------------------------------------------------------------------
__forceinline  void __fastcall Swap_YZW(vec4& v1, vec4& v2) {
	float  t;
	Swap(v1.y, v2.y, t);
	Swap(v1.z, v2.z, t);
	Swap(v1.w, v2.w, t);
}

//--------------------------------------------------------------------
__forceinline  void __fastcall Swap_ZW(vec4& v1, vec4& v2) {
	float  t;
	Swap(v1.z, v2.z, t);
	Swap(v1.w, v2.w, t);
}


//--------------------------------------------------------------------
#endif //VW_MATRIX44_INVERSE_OPTIMIZE


//====================================================================
mat4&  mat4::InverseNC(mat4& src, bool* pFailed)
{
	// NOTES: uses gaussian elimination with partial-pivoting

	// prepare source-matrix alias
	mat4& m = src;

	// prepare result-matrix
	mat4& r = *this;
	r = mat4::IDENTITY;

	// cache global parameter
	float  dividerTolerance = float(math_constants::MatrixInverseDefaultDividerTolerance);

	//--------------------
	// loop through pivots
	for(uint piv = 0; piv < 4; piv++) {

		//---------------------------
		// find max element in column
		float    maxVal = m[piv][piv];
		uint maxRow = piv;

		{for(uint row = piv + 1; row < 4; row++) {
			float  val = m[row][piv];
			if(Abs(val) > Abs(maxVal)) {
				maxVal = val;
				maxRow = row;
			}
		}}

		//------------------------------
		// check for uninvertible matrix
		if(Abs(maxVal) < dividerTolerance) {
			*this = mat4::IDENTITY;
			SafeAssign(pFailed, true);
			return *this;
		}

		//------------
		// reduce rows

#ifdef  VW_MATRIX44_INVERSE_OPTIMIZE

		// common calculations
		float  invMaxVal = 1 / maxVal;
		if(maxRow != piv) {
			Swap(r.Row(piv), r.Row(maxRow));
		}

		// pivot specific processing
		switch(piv) {
			case 0: {
				if(maxRow != 0) {
					Swap(m.Row(0), m.Row(maxRow));
				}
				r.Row(0) *= invMaxVal;
				r.Row(1) -= r.Row(0) * m[1][0];
				r.Row(2) -= r.Row(0) * m[2][0];
				r.Row(3) -= r.Row(0) * m[3][0];

				Mul_YZW(m.Row(0), invMaxVal);
				Sub_YZW(m.Row(1), m.Row(0), m[1][0]);
				Sub_YZW(m.Row(2), m.Row(0), m[2][0]);
				Sub_YZW(m.Row(3), m.Row(0), m[3][0]);
			} break;

			case 1: {
				if(maxRow != 1) {
					Swap_YZW(m.Row(1), m.Row(maxRow));
				}
				r.Row(1) *= invMaxVal;
				r.Row(0) -= r.Row(1) * m[0][1];
				r.Row(2) -= r.Row(1) * m[2][1];
				r.Row(3) -= r.Row(1) * m[3][1];

				Mul_ZW(m.Row(1), invMaxVal);
				Sub_ZW(m.Row(0), m.Row(1), m[0][1]);
				Sub_ZW(m.Row(2), m.Row(1), m[2][1]);
				Sub_ZW(m.Row(3), m.Row(1), m[3][1]);
			} break;

			case 2: {
				if(maxRow != 2) {
					Swap_ZW(m.Row(2), m.Row(maxRow));
				}
				r.Row(2) *= invMaxVal;
				r.Row(0) -= r.Row(2) * m[0][2];
				r.Row(1) -= r.Row(2) * m[1][2];
				r.Row(3) -= r.Row(2) * m[3][2];

				Mul_W(m.Row(2), invMaxVal);
				Sub_W(m.Row(0), m.Row(2), m[0][2]);
				Sub_W(m.Row(1), m.Row(2), m[1][2]);
				Sub_W(m.Row(3), m.Row(2), m[3][2]);
			} break;

			case 3: {
				r.Row(3) *= invMaxVal;
				r.Row(0) -= r.Row(3) * m[0][3];
				r.Row(1) -= r.Row(3) * m[1][3];
				r.Row(2) -= r.Row(3) * m[2][3];
			} break;
		}

#else //NO_OPTIMIZATION

		// put row with maximum value at pivot-column as current pivot-row
		if(maxRow != piv) {
			vec4  temp;
			Swap(m.Row(piv), m.Row(maxRow), temp);
			Swap(r.Row(piv), r.Row(maxRow), temp);
		}

		// divide current pivot-row by maximum value
		float  invMaxVal = 1 / maxVal;
		m.Row(piv) *= invMaxVal;
		r.Row(piv) *= invMaxVal;

		// scale and subtract
		{for(uint row = 0; row < 4; row++) {
			if(row == piv) continue;
			float  scale = m[row][piv];
			m.Row(row) -= m.Row(piv) * scale;
			r.Row(row) -= r.Row(piv) * scale;
		}}

#endif //VW_MATRIX44_INVERSE_OPTIMIZE

	}//end pivot-loop

	// success
	SafeAssign(pFailed, false);
	return *this;
}


//====================================================================
float  mat4::Determinant() const
{
	mat3  cofactorX(m11, m12, m13,
						m21, m22, m23,
						m31, m32, m33);

	mat3  cofactorY(m10, m12, m13,
						m20, m22, m23,
						m30, m32, m33);

	mat3  cofactorZ(m10, m11, m13,
						m20, m21, m23,
						m30, m31, m33);

	mat3  cofactorW(m10, m11, m12,
						m20, m21, m22,
						m30, m31, m32);

	return  ( m00 * cofactorX.Determinant() - m01 * cofactorY.Determinant()
			+ m02 * cofactorZ.Determinant() - m03 * cofactorW.Determinant());
}


//====================================================================
bool  mat4::operator==(const mat4& rhs) const
{
	return
	(m00 == rhs.m00)  &&
	(m01 == rhs.m01)  &&
	(m02 == rhs.m02)  &&
	(m03 == rhs.m03)  &&
	(m10 == rhs.m10)  &&
	(m11 == rhs.m11)  &&
	(m12 == rhs.m12)  &&
	(m13 == rhs.m13)  &&
	(m20 == rhs.m20)  &&
	(m21 == rhs.m21)  &&
	(m22 == rhs.m22)  &&
	(m23 == rhs.m23)  &&
	(m30 == rhs.m30)  &&
	(m31 == rhs.m31)  &&
	(m32 == rhs.m32)  &&
	(m33 == rhs.m33);
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const mat4& m)
{
	for(uint i = 0; i < 16; i++) {
		os <= m.m[i];
	}
	return os;
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, mat4& m)
{
	for(uint i = 0; i < 16; i++) {
		is >= m.m[i];
	}
	return is;
}


//********************************************************************
}//namespace vertexware


