

//--------------------------------------------------------------------
#include "MathTools\Quaternion.h"
#include "MathTools\TransformTools.h"
#include "BaseTools\Serializers.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace math_constants {


//====================================================================
VW_DEFINE_CONSTANT( float, QuaternionNoiseFilterDefaultTolerance, 1e-6);


//********************************************************************
}//namespace math_constants


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
#define  TO_ROTATION_MATRIX_PROLOG \
	float  x_2   = q.x + q.x; \
	float  y_2   = q.y + q.y; \
	float  z_2   = q.z + q.z; \
	float  x_x_2 = x_2 * q.x; \
	float  x_y_2 = x_2 * q.y; \
	float  x_z_2 = x_2 * q.z; \
	float  x_w_2 = x_2 * q.w; \
	float  y_y_2 = y_2 * q.y; \
	float  y_z_2 = y_2 * q.z; \
	float  y_w_2 = y_2 * q.w; \
	float  z_z_2 = z_2 * q.z; \
	float  z_w_2 = z_2 * q.w;


#define  TO_ROTATION_MATRIX_M00   (1 - y_y_2 - z_z_2)
#define  TO_ROTATION_MATRIX_M01   (x_y_2 - z_w_2)
#define  TO_ROTATION_MATRIX_M02   (x_z_2 + y_w_2)
#define  TO_ROTATION_MATRIX_M10   (x_y_2 + z_w_2)
#define  TO_ROTATION_MATRIX_M11   (1 - x_x_2 - z_z_2)
#define  TO_ROTATION_MATRIX_M12   (y_z_2 - x_w_2)
#define  TO_ROTATION_MATRIX_M20   (x_z_2 - y_w_2)
#define  TO_ROTATION_MATRIX_M21   (x_w_2 + y_z_2)
#define  TO_ROTATION_MATRIX_M22   (1 - x_x_2 - y_y_2)


//====================================================================
template<typename TTranspose>
__forceinline mat3&  ToRotationMatrix(const TTranspose& tt,
										const quat& q, mat3& dst)
{
	// MV-native
	TO_ROTATION_MATRIX_PROLOG;

	Fill(tt, dst,
		TO_ROTATION_MATRIX_M00, TO_ROTATION_MATRIX_M01, TO_ROTATION_MATRIX_M02,
		TO_ROTATION_MATRIX_M10, TO_ROTATION_MATRIX_M11, TO_ROTATION_MATRIX_M12,
		TO_ROTATION_MATRIX_M20, TO_ROTATION_MATRIX_M21, TO_ROTATION_MATRIX_M22);

	return dst;
}


//====================================================================
template<typename TTranspose>
__forceinline mat4&  ToRotationMatrix(const TTranspose& tt,
										const quat& q, mat4& dst)
{
	// MV-native
	TO_ROTATION_MATRIX_PROLOG;

	Fill(tt, dst,
		TO_ROTATION_MATRIX_M00, TO_ROTATION_MATRIX_M01, TO_ROTATION_MATRIX_M02, 0,
		TO_ROTATION_MATRIX_M10, TO_ROTATION_MATRIX_M11, TO_ROTATION_MATRIX_M12, 0,
		TO_ROTATION_MATRIX_M20, TO_ROTATION_MATRIX_M21, TO_ROTATION_MATRIX_M22, 0,
		0,                      0,                      0,                      1);

	return dst;
}


//====================================================================
template<typename TMatrixConstAccessor>
__forceinline void  FromRotationMatrixBody(quat& q, TMatrixConstAccessor& m)
{
	float trace = m.m.Trace();

	// find largest term for division
	if(trace > 0) {
		// use w for division
		q.w = 0.5f * Sqrt(trace + 1);

		float OneOver4W = 0.25f / q.w;

		q.x = (m(2, 1) - m(1, 2)) * OneOver4W;
		q.y = (m(0, 2) - m(2, 0)) * OneOver4W;
		q.z = (m(1, 0) - m(0, 1)) * OneOver4W;
	}
	else {
		if(m(0, 0) >= m(1, 1)  &&  m(0, 0) >= m(2, 2)) {
			// use x for division
			q.x = 0.5f * Sqrt(m(0, 0) - m(1, 1) - m(2, 2) + 1);

			float OneOver4X = 0.25f / q.x;

			q.y = (m(1, 0) + m(0, 1)) * OneOver4X;
			q.z = (m(2, 0) + m(0, 2)) * OneOver4X;
			q.w = (m(2, 1) - m(1, 2)) * OneOver4X;
		}
		else if(m(1, 1) >= m(2, 2)) {
			// use y for division
			q.y = 0.5f * Sqrt(m(1, 1) - m(0, 0) - m(2, 2) + 1);

			float OneOver4Y = 0.25f / q.y;

			q.x = (m(0, 1) + m(1, 0)) * OneOver4Y;
			q.z = (m(2, 1) + m(1, 2)) * OneOver4Y;
			q.w = (m(0, 2) - m(2, 0)) * OneOver4Y;
		}
		else {
			// use z for division
			q.z = 0.5f * Sqrt(m(2, 2) - m(0, 0) - m(1, 1) + 1);

			float OneOver4Z = 0.25f / q.z;

			q.x = (m(0, 2) + m(2, 0)) * OneOver4Z;
			q.y = (m(1, 2) + m(2, 1)) * OneOver4Z;
			q.w = (m(1, 0) - m(0, 1)) * OneOver4Z;
		}
	}
}


//====================================================================
template<typename TTranspose>
__forceinline quat&  FromRotationMatrix(const TTranspose&,
											quat& q, const mat3& src)
{
	// MV-native
	typedef  Matrix33ConstAccessor<TTranspose>  MatrixConstAccessor;

	MatrixConstAccessor  m(src);
	FromRotationMatrixBody(q, m);

	return q;
}


//====================================================================
template<typename TTranspose>
__forceinline quat&  FromRotationMatrix(const TTranspose&,
											quat& q, const mat4& src)
{
	// MV-native
	typedef  Matrix44ConstAccessor<TTranspose>  MatrixConstAccessor;

	MatrixConstAccessor  m(src);
	FromRotationMatrixBody(q, m);

	return q;
}


//********************************************************************
}//namespace


//====================================================================
const quat  quat::ZERO     = quat(0, 0, 0, 0);
const quat  quat::IDENTITY = quat(0, 0, 0, 1);
const quat  quat::INFINITY = quat(math_constants::PlusInfinity,
													math_constants::PlusInfinity,
													math_constants::PlusInfinity,
													math_constants::PlusInfinity);


//====================================================================
void  quat::ToRotationAxisAngle(vec3& unitRotAxis, float& rotAngleDeg) const
{
	// for unit-quat norm is 1, so if (w == 1) then (x * x + y * y + z * z == 0)
	float  lenSqr = x * x + y * y + z * z;
	if(lenSqr > 0) {
		// w != 1
		lenSqr = 1 / Sqrt(lenSqr);
		unitRotAxis.Set(x * lenSqr, y * lenSqr, z * lenSqr);
		rotAngleDeg = 2 * ArcCosDeg(w);
	}
	else {
		// w == 1, export 0-angle and arbitrary-axis
		unitRotAxis.Set(0, 0, 1);
		rotAngleDeg = 0;
	}
}


//====================================================================
void  quat::FromRotationAxisAngle(const vec3& unitRotAxis, float rotAngleDeg)
{
	float  sinHalf;
	SinCosDeg(rotAngleDeg * 0.5f, sinHalf, w);
	x = unitRotAxis.x * sinHalf;
	y = unitRotAxis.y * sinHalf;
	z = unitRotAxis.z * sinHalf;
}


//====================================================================
mat3&  quat::ToRotationMatrixMV(mat3& dst) const
{
	return ToRotationMatrix(no_transpose_t(), *this, dst);
}


//====================================================================
quat&  quat::FromRotationMatrixMV(const mat3& src)
{
	return FromRotationMatrix(no_transpose_t(), *this, src);
}


//====================================================================
mat3&  quat::ToRotationMatrixVM(mat3& dst) const
{
	return ToRotationMatrix(transpose_t(), *this, dst);
}


//====================================================================
quat&  quat::FromRotationMatrixVM(const mat3& src)
{
	return FromRotationMatrix(transpose_t(), *this, src);
}


//====================================================================
mat4&  quat::ToRotationMatrixMV(mat4& dst) const
{
	return ToRotationMatrix(no_transpose_t(), *this, dst);
}


//====================================================================
quat&  quat::FromRotationMatrixMV(const mat4& src)
{
	return FromRotationMatrix(no_transpose_t(), *this, src);
}


//====================================================================
mat4&  quat::ToRotationMatrixVM(mat4& dst) const
{
	return ToRotationMatrix(transpose_t(), *this, dst);
}


//====================================================================
quat&  quat::FromRotationMatrixVM(const mat4& src)
{
	return FromRotationMatrix(transpose_t(), *this, src);
}


//====================================================================
void  quat::ToCoordSystemAxes(vec3& unitU, vec3& unitV, vec3& unitN) const
{
	mat3  rot;
	ToRotationMatrixMV(rot);
	FactorRotateMV(rot, unitU, unitV, unitN);
}


//====================================================================
void  quat::FromCoordSystemAxes(const vec3& unitU, const vec3& unitV, const vec3& unitN)
{
	mat3  rot;
	GenerateRotateMV(rot, unitU, unitV, unitN);
	FromRotationMatrixMV(rot);
}


//====================================================================
vec3&  quat::Rotate(vec3& dst, const vec3& src) const
{
	mat3  rot;
	ToRotationMatrixMV(rot);

	vertexware::Mul(dst, rot, src);

	return dst;
}


//====================================================================
quat&  quat::ShortestArcRotation(const vec3& unitInitial,
											const vec3& unitTarget)
{
	vec3 rotAxis;
	rotAxis.Cross(unitInitial, unitTarget);

	float  d = unitInitial.Dot(unitTarget);
	float  s = Sqrt(2 * (d + 1));
	float  OneOverS = 1 / s;

	x = rotAxis.x * OneOverS;
	y = rotAxis.y * OneOverS;
	z = rotAxis.z * OneOverS;
	w = s * 0.5f;

	return *this;
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const quat& q)
{
	os <= q.x <= q.y <= q.z <= q.w;
	return os;
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, quat& q)
{
	is >= q.x >= q.y >= q.z >= q.w;
	return is;
}


//********************************************************************
}//namespace vertexware

