

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "MathTools\Vector3.h"
#include "MathTools\Matrix33.h"
#include "MathTools\Matrix44.h"
#include "MathTools\MathProcessors.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace math_constants {


//====================================================================
VW_COMMON_API  VW_DECLARE_CONSTANT( float, QuaternionNoiseFilterDefaultTolerance );


//********************************************************************
}//namespace math_constants


//====================================================================
// Quaternion.
// q = sin(angle/2)(xi + yj + zk) + cos(angle/2)
// represents rotation by (angle) radians around the axis v(x, y, z).
//====================================================================
class VW_COMMON_API quat
{
public:
	//------
	// Data.
	//------
	float  x;
	float  y;
	float  z;
	float  w;


	//-----------------------
	// Convenience constants.
	//-----------------------
	static const quat  ZERO;
	static const quat  IDENTITY;
	static const quat  INFINITY;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline quat()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline quat(float _x, float _y, float _z, float _w) {
		x = _x;
		y = _y;
		z = _z;
		w = _w;
	}

	__forceinline explicit quat(const float* a) {
		x = a[0];
		y = a[1];
		z = a[2];
		w = a[3];
	}

	__forceinline quat(const vec3& unitRotAxis, float rotAngleDeg) {
		FromRotationAxisAngle(unitRotAxis, rotAngleDeg);
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline quat(const quat& rhs) {
		this->operator=(rhs);
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline quat& operator=(const quat& rhs) {
		x = rhs.x;
		y = rhs.y;
		z = rhs.z;
		w = rhs.w;
		return *this;
	}

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(float _x, float _y, float _z, float _w) {
		x = _x;
		y = _y;
		z = _z;
		w = _w;
	}

	__forceinline void  Set(const float* a) {
		x = a[0];
		y = a[1];
		z = a[2];
		w = a[3];
	}

	__forceinline void  Set(const vec3& unitRotAxis, float rotAngleDeg) {
		FromRotationAxisAngle(unitRotAxis, rotAngleDeg);
	}

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline float*        ToPtr()       { return &x; }
	__forceinline const float*  ToPtr() const { return &x; }

	__forceinline operator float*()             { return ToPtr(); }
	__forceinline operator const float*() const { return ToPtr(); }

	//----------
	// Addition.
	//----------
	__forceinline quat&  operator+=(const quat& rhs) {
		x += rhs.x;
		y += rhs.y;
		z += rhs.z;
		w += rhs.w;
		return *this;
	}

	__forceinline quat&  Add(const quat& rhs) {
		x += rhs.x;
		y += rhs.y;
		z += rhs.z;
		w += rhs.w;
		return *this;
	}

	__forceinline quat  operator+(const quat& rhs) const {
		return quat(x + rhs.x, y + rhs.y, z + rhs.z, w + rhs.w);
	}

	__forceinline quat&  Add(const quat& lhs, const quat& rhs) {
		x = lhs.x + rhs.x;
		y = lhs.y + rhs.y;
		z = lhs.z + rhs.z;
		w = lhs.w + rhs.w;
		return *this;
	}

	//-------------
	// Subtraction.
	//-------------
	__forceinline quat&  operator-=(const quat& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		z -= rhs.z;
		w -= rhs.w;
		return *this;
	}

	__forceinline quat&  Sub(const quat& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		z -= rhs.z;
		w -= rhs.w;
		return *this;
	}

	__forceinline quat  operator-(const quat& rhs) const {
		return quat(x - rhs.x, y - rhs.y, z - rhs.z, w - rhs.w);
	}

	__forceinline quat&  Sub(const quat& lhs, const quat& rhs) {
		x = lhs.x - rhs.x;
		y = lhs.y - rhs.y;
		z = lhs.z - rhs.z;
		w = lhs.w - rhs.w;
		return *this;
	}

	//----------------------------------
	// Multiplication (not commutative).
	//----------------------------------
	__forceinline quat&  operator*=(const quat& rhs) {
		return this->Mul(quat(*this), rhs);
	}

	__forceinline quat&  Mul(const quat& rhs) {
		return this->Mul(quat(*this), rhs);
	}

	__forceinline quat  operator*(const quat& rhs) const {
		return quat(
					y * rhs.z - z * rhs.y + x * rhs.w + w * rhs.x,
					z * rhs.x - x * rhs.z + y * rhs.w + w * rhs.y,
					x * rhs.y - y * rhs.x + z * rhs.w + w * rhs.z,
					w * rhs.w - x * rhs.x - y * rhs.y - z * rhs.z);
	}

	__forceinline quat&  Mul(const quat& lhs, const quat& rhs) {
		x = lhs.y * rhs.z - lhs.z * rhs.y + lhs.x * rhs.w + lhs.w * rhs.x;
		y = lhs.z * rhs.x - lhs.x * rhs.z + lhs.y * rhs.w + lhs.w * rhs.y;
		z = lhs.x * rhs.y - lhs.y * rhs.x + lhs.z * rhs.w + lhs.w * rhs.z;
		w = lhs.w * rhs.w - lhs.x * rhs.x - lhs.y * rhs.y - lhs.z * rhs.z;
		return *this;
	}

	//--------------------------
	// Multiplication by scalar.
	//--------------------------
	__forceinline quat&  operator*=(float num) {
		x *= num;
		y *= num;
		z *= num;
		w *= num;
		return *this;
	}

	__forceinline quat&  Mul(float num) {
		x *= num;
		y *= num;
		z *= num;
		w *= num;
		return *this;
	}

	__forceinline quat  operator*(float num) const {
		return quat(x * num, y * num, z * num, w * num);
	}

	__forceinline quat&  Mul(const quat& lhs, float num) {
		x = lhs.x * num;
		y = lhs.y * num;
		z = lhs.z * num;
		w = lhs.w * num;
		return *this;
	}

	//----------------------------------------------
	// Division by scalar (no divide-by-zero check).
	//----------------------------------------------
	__forceinline quat&  operator/=(float num) {
		num = 1 / num;
		x *= num;
		y *= num;
		z *= num;
		w *= num;
		return *this;
	}

	__forceinline quat&  Div(float num) {
		num = 1 / num;
		x *= num;
		y *= num;
		z *= num;
		w *= num;
		return *this;
	}

	__forceinline quat  operator/(float num) const {
		num = 1 / num;
		return quat(x * num, y * num, z * num, w * num);
	}

	__forceinline quat&  Div(const quat& lhs, float num) {
		num = 1 / num;
		x = lhs.x * num;
		y = lhs.y * num;
		z = lhs.z * num;
		w = lhs.w * num;
		return *this;
	}

	//----------
	// Negation.
	//----------
	__forceinline quat&  Neg() {
		x *= -1;
		y *= -1;
		z *= -1;
		w *= -1;
		return *this;
	}

	__forceinline quat&  Neg(const quat& src) {
		x = - src.x;
		y = - src.y;
		z = - src.z;
		w = - src.w;
		return *this;
	}

	//------------------
	// Unary operations.
	//------------------
	__forceinline quat  operator-() const {
		return quat(-x, -y, -z, -w);
	}

	__forceinline quat&        operator+()       { return *this; }
	__forceinline const quat&  operator+() const { return *this; }

	//-------------
	// Dot product.
	//-------------
	__forceinline float  Dot(const quat& rhs) const {
		return (x * rhs.x + y * rhs.y + z * rhs.z + w * rhs.w);
	}

	//-----------------------
	// Norm (squared length).
	//-----------------------
	__forceinline float  Norm() const {
		return (x * x + y * y + z * z + w * w);
	}

	//--------
	// Length.
	//--------
	__forceinline float  Length() const {
		return Sqrt<float>(x * x + y * y + z * z + w * w);
	}

	//-----------
	// Conjugate.
	//-----------
	__forceinline quat&  Conjugate(const quat& src) {
		x = -src.x;
		y = -src.y;
		z = -src.z;
		w = +src.w;
		return  *this;
	}

	__forceinline quat  Conjugate() const {
		return  quat(-x, -y, -z, +w);
	}

	//---------
	// Unitize.
	//---------
	__forceinline quat&  Unitize() {
		float len = Length();
		if(len > 0) {
			len = 1 / len;
			x *= len;
			y *= len;
			z *= len;
			w *= len;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	__forceinline quat  Unit() const {
		float len = Length();
		if(len > 0) {
			len = 1 / len;
			return quat(x * len, y * len, z * len, w * len);
		}
		return INFINITY;
	}

	__forceinline quat&  Unit(const quat& src) {
		float len = Length();
		if(len > 0) {
			len = 1 / len;
			x = src.x * len;
			y = src.y * len;
			z = src.z * len;
			w = src.w * len;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	//------------------------
	// Multiplicative inverse.
	//------------------------
	__forceinline quat&  Invert() {
		float norm = Norm();
		if(norm > 0) {
			norm = -1 / norm;
			x *= norm;
			y *= norm;
			z *= norm;
			w *= - norm;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	__forceinline quat  Inverse() const {
		float norm = Norm();
		if(norm > 0) {
			norm = -1 / norm;
			return quat(x * norm, y * norm, z * norm, - w * norm);
		}
		return INFINITY;
	}

	__forceinline quat&  Inverse(const quat& src) {
		float norm = src.Norm();
		if(norm > 0) {
			norm = -1 / norm;
			x = src.x * norm;
			y = src.y * norm;
			z = src.z * norm;
			w = src.w * - norm;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	// U-version is faster, but can be used only for unit quaternions.
	// U stands for "Unit-quat".
	__forceinline quat&  InvertU() {
		x *= -1;
		y *= -1;
		z *= -1;
		return *this;
	}

	__forceinline quat  InverseU() const {
		return quat(-x, -y, -z, +w);
	}

	__forceinline quat&  InverseU(const quat& unitSrc) {
		x = - unitSrc.x;
		y = - unitSrc.y;
		z = - unitSrc.z;
		w = + unitSrc.w;
		return *this;
	}

	//-------------
	// Comparisons.
	//-------------
	__forceinline bool  operator==(const quat& rhs) const {
		return ((x == rhs.x) && (y == rhs.y) && (z == rhs.z) && (w == rhs.w));
	}
	__forceinline bool  operator!=(const quat& rhs) const {
		return !(*this == rhs);
	}

	//----------------------------------------------------
	// Noise filtering (zero-out all elements close to 0).
	//----------------------------------------------------
	__forceinline quat&  NoiseFilter(const quat& src, float tolerance =
										math_constants::QuaternionNoiseFilterDefaultTolerance) {
		vertexware::NoiseFilter(this->ToPtr(), 4, tolerance);
		return *this;
	}

	__forceinline  quat __fastcall NoiseFiltered(float tolerance =
										math_constants::QuaternionNoiseFilterDefaultTolerance) const {
		quat temp;
		vertexware::NoiseFilter(temp.ToPtr(), this->ToPtr(), 4, tolerance);
		return temp;
	}

	//-----------------------------------------------------
	// Rotation axis/angle <=> unit-quat conversions.
	//-----------------------------------------------------
	void  ToRotationAxisAngle(vec3& unitRotAxis, float& rotAngleDeg) const;
	void  FromRotationAxisAngle(const vec3& unitRotAxis, float rotAngleDeg);

	//-------------------------------------------------
	// Rotation-matrix <=> unit-quat conversions.
	//-------------------------------------------------
	mat3&    ToRotationMatrixMV(mat3& dst) const;
	quat&  FromRotationMatrixMV(const mat3& src);

	mat3&    ToRotationMatrixVM(mat3& dst) const;
	quat&  FromRotationMatrixVM(const mat3& src);

	mat4&    ToRotationMatrixMV(mat4& dst) const;
	quat&  FromRotationMatrixMV(const mat4& src);

	mat4&    ToRotationMatrixVM(mat4& dst) const;
	quat&  FromRotationMatrixVM(const mat4& src);

	//--------------------------------------------------------
	// Coordinate-system-axes <=> unit-quat conversions.
	//--------------------------------------------------------
	void  ToCoordSystemAxes(vec3& unitU, vec3& unitV, vec3& unitN) const;
	void  FromCoordSystemAxes(const vec3& unitU, const vec3& unitV, const vec3& unitN);

	//--------------------------------------------------
	// Combinations for rotation.
	// Combine() constructs rotation-quat which
	// first applies <rot1> then <rot2>.
	// PreCombine() construct rotation-quat which
	// first applies <rot1> then <this>.
	// PostCombine() construct rotation-quat which
	// first applies <this> then <rot2>.
	//--------------------------------------------------
	__forceinline quat&  Combine(const quat& rot1, const quat& rot2) {
		return this->Mul(rot2, rot1);
	}
	__forceinline quat&  PreCombine(const quat& rot1) {
		return this->Mul(quat(*this), rot1);
	}
	__forceinline quat&  PostCombine(const quat& rot2) {
		return this->Mul(rot2, quat(*this));
	}

	//-----------------------------------------------------------
	// Apply rotation-quat to vector (clockwise in LHS,
	// counter-clockwise in RHS). <this> must be unit-quat.
	// Actual operation: q * v * q_inverse
	//
	// PERFORMANCE WARNING:
	// Do not use this method when rotating more than one vector.
	// Instead, export quat to rotation matrix and use it
	// to transform vectors.
	//-----------------------------------------------------------
	vec3&  Rotate(vec3& dst, const vec3& src) const;

	__forceinline vec3  Rotate(const vec3& src) const {
		vec3 temp;
		return this->Rotate(temp, src);
	}

	__forceinline  vec3 operator*(const vec3& src) const {
		vec3 temp;
		return this->Rotate(temp, src);
	}

	//--------------------------------------------------------
	// Creates shortest-arc quat used for rotation from
	// initial-direction-vector to target-direction-vector.
	// <unitInitial> & <unitTarget> must be of unit-length.
	// Routine is numerically stable for all cases except when
	// <unitInitial> approaches <unitTarget> * -1.
	//--------------------------------------------------------
	quat&  ShortestArcRotation(const vec3& unitInitial,
										const vec3& unitTarget);
};


//====================================================================
// Multiplication by scalar (redirection).
//====================================================================
__forceinline  quat __fastcall operator*(float num, const quat& rhs) {
	return rhs * num;
}


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const quat& q);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, quat& q);


//********************************************************************
}//namespace vertexware


