

//--------------------------------------------------------------------
#include "MathTools\TransformTools.h"
#include "BaseTools\Constructs.h"


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
__forceinline void __fastcall RemapAngleTo360DegRange(float& angle)
{
	int iNumFullturns = int(angle) / 360;
	if(iNumFullturns != 0) angle -= iNumFullturns * 360;
}


//====================================================================
template<typename TTranspose>
__forceinline void __fastcall GenerateTranslate(const TTranspose& tt, mat3& m,
												float trX, float trY)
{
	// MV-native
	Fill(tt, m,
		1, 0, trX,
		0, 1, trY,
		0, 0, 1);
}


//====================================================================
template<typename TTranspose>
__forceinline void __fastcall GenerateTranslate(const TTranspose& tt, mat4& m,
												float trX, float trY, float trZ)
{
	// MV-native
	Fill(tt, m,
		1, 0, 0, trX,
		0, 1, 0, trY,
		0, 0, 1, trZ,
		0, 0, 0, 1);
}


//====================================================================
template<typename TTranspose>
__forceinline void __fastcall GenerateRotate(const TTranspose& tt, mat2& m,
											float angleDeg)
{
	// MV-native
	float  s, c;
	SinCosDeg(angleDeg, s, c);

	Fill(tt, m,
		 c, s,
		-s, c);
}


//====================================================================
template<typename TTranspose>
__forceinline void __fastcall GenerateRotate(const TTranspose& tt, mat3& m,
											float angleDeg, eAxis axis)
{
	// MV-native
	float  s, c;
	SinCosDeg(angleDeg, s, c);

	switch(axis) {

	// rotate around x-axis
	case Axis_X:

		Fill(tt, m,
			1, 0, 0,
			0, c, -s,
			0, s, c);

		return;

	// rotate around y-axis
	case Axis_Y:

		Fill(tt, m,
			 c, 0, s,
			 0, 1, 0,
			-s, 0, c);

		return;

	// rotate around z-axis
	case Axis_Z:

		Fill(tt, m,
			c, -s, 0,
			s,  c, 0,
			0,  0, 1);

		return;
	}
}


//====================================================================
template<typename TTranspose>
__forceinline void __fastcall GenerateRotate(const TTranspose& tt, mat4& m,
											float angleDeg, eAxis axis)
{
	// MV-native
	float  s, c;
	SinCosDeg(angleDeg, s, c);

	switch(axis) {

	// rotate around x-axis
	case Axis_X:

		Fill(tt, m,
			1, 0,  0, 0,
			0, c, -s, 0,
			0, s,  c, 0,
			0, 0,  0, 1);

		return;

	// rotate around y-axis
	case Axis_Y:

		Fill(tt, m,
			 c, 0, s, 0,
			 0, 1, 0, 0,
			-s, 0, c, 0,
			 0, 0, 0, 1);

		return;

	// rotate around z-axis
	case Axis_Z:

		Fill(tt, m,
			c, -s, 0, 0,
			s,  c, 0, 0,
			0,  0, 1, 0,
			0,  0, 0, 1);

		return;
	}
}


//====================================================================
#define  GENERATE_ROTATE_ANGLE_AXIS_PROLOG \
	float  s, c; \
	SinCosDeg(angleDeg, s, c); \
	\
	float  x = unitAxis.x; \
	float  y = unitAxis.y; \
	float  z = unitAxis.z; \
	\
	float  omc = 1 - c; \
	\
	float  xx = x * x * omc; \
	float  xy = x * y * omc; \
	float  xz = x * z * omc; \
	float  yy = y * y * omc; \
	float  yz = y * z * omc; \
	float  zz = z * z * omc; \
	\
	float  xs = x * s; \
	float  ys = y * s; \
	float  zs = z * s;


#define  GENERATE_ROTATE_ANGLE_AXIS_M00   (xx + c)
#define  GENERATE_ROTATE_ANGLE_AXIS_M01   (xy - zs)
#define  GENERATE_ROTATE_ANGLE_AXIS_M02   (xz + ys)
#define  GENERATE_ROTATE_ANGLE_AXIS_M10   (xy + zs)
#define  GENERATE_ROTATE_ANGLE_AXIS_M11   (yy + c)
#define  GENERATE_ROTATE_ANGLE_AXIS_M12   (yz - xs)
#define  GENERATE_ROTATE_ANGLE_AXIS_M20   (xz - ys)
#define  GENERATE_ROTATE_ANGLE_AXIS_M21   (yz + xs)
#define  GENERATE_ROTATE_ANGLE_AXIS_M22   (zz + c)


//====================================================================
template<typename TTranspose>
__forceinline void __fastcall GenerateRotate(const TTranspose& tt, mat3& m,
											float angleDeg, const vec3& unitAxis)
{
	// MV-native
	GENERATE_ROTATE_ANGLE_AXIS_PROLOG;

	Fill(tt, m,
		GENERATE_ROTATE_ANGLE_AXIS_M00, GENERATE_ROTATE_ANGLE_AXIS_M01, GENERATE_ROTATE_ANGLE_AXIS_M02,
		GENERATE_ROTATE_ANGLE_AXIS_M10, GENERATE_ROTATE_ANGLE_AXIS_M11, GENERATE_ROTATE_ANGLE_AXIS_M12,
		GENERATE_ROTATE_ANGLE_AXIS_M20, GENERATE_ROTATE_ANGLE_AXIS_M21, GENERATE_ROTATE_ANGLE_AXIS_M22);
}


//====================================================================
template<typename TTranspose>
__forceinline void __fastcall GenerateRotate(const TTranspose& tt, mat4& m,
											float angleDeg, const vec3& unitAxis)
{
	// MV-native
	GENERATE_ROTATE_ANGLE_AXIS_PROLOG;

	Fill(tt, m,
		GENERATE_ROTATE_ANGLE_AXIS_M00, GENERATE_ROTATE_ANGLE_AXIS_M01, GENERATE_ROTATE_ANGLE_AXIS_M02, 0,
		GENERATE_ROTATE_ANGLE_AXIS_M10, GENERATE_ROTATE_ANGLE_AXIS_M11, GENERATE_ROTATE_ANGLE_AXIS_M12, 0,
		GENERATE_ROTATE_ANGLE_AXIS_M20, GENERATE_ROTATE_ANGLE_AXIS_M21, GENERATE_ROTATE_ANGLE_AXIS_M22, 0,
		0,                              0,                              0,                              1);
}


//====================================================================
template<typename TMatrixConstAccessor>
__forceinline void __fastcall FactorRotateBody(TMatrixConstAccessor& m,
										vec3& unitAxisU, vec3& unitAxisV, vec3& unitAxisN)
{
	// MV-native
	unitAxisU.x = m(0, 0);
	unitAxisU.y = m(1, 0);
	unitAxisU.z = m(2, 0);

	unitAxisV.x = m(0, 1);
	unitAxisV.y = m(1, 1);
	unitAxisV.z = m(2, 1);

	unitAxisN.x = m(0, 2);
	unitAxisN.y = m(1, 2);
	unitAxisN.z = m(2, 2);
}


//====================================================================
#define  GENERATE_EULER_HPR_PROLOG \
	float  sH, cH, sP, cP, sR, cR; \
	\
	SinCosDeg(headDeg, sH, cH); \
	SinCosDeg(pitchDeg, sP, cP); \
	SinCosDeg(rollDeg, sR, cR); \
	\
	float  cRcH = cR * cH; \
	float  sRcH = sR * cH; \
	float  cRsH = cR * sH; \
	float  sRsH = sR * sH;


#define  GENERATE_EULER_HPR_M00   (cRcH - sRsH * sP)
#define  GENERATE_EULER_HPR_M01   (- sR * cP)
#define  GENERATE_EULER_HPR_M02   (cRsH + sRcH * sP)
#define  GENERATE_EULER_HPR_M10   (sRcH + cRsH * sP)
#define  GENERATE_EULER_HPR_M11   (cR * cP)
#define  GENERATE_EULER_HPR_M12   (sRsH - cRcH * sP)
#define  GENERATE_EULER_HPR_M20   (- cP * sH)
#define  GENERATE_EULER_HPR_M21   (sP)
#define  GENERATE_EULER_HPR_M22   (cP * cH)


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall GenerateEulerHPR(const TTranspose& tt, mat3& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	// MV-native
	GENERATE_EULER_HPR_PROLOG;

	Fill(tt, m,
		GENERATE_EULER_HPR_M00, GENERATE_EULER_HPR_M01, GENERATE_EULER_HPR_M02,
		GENERATE_EULER_HPR_M10, GENERATE_EULER_HPR_M11, GENERATE_EULER_HPR_M12,
		GENERATE_EULER_HPR_M20, GENERATE_EULER_HPR_M21, GENERATE_EULER_HPR_M22);
}


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall GenerateEulerHPR(const TTranspose& tt, mat4& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	// MV-native
	GENERATE_EULER_HPR_PROLOG;

	Fill(tt, m,
		GENERATE_EULER_HPR_M00, GENERATE_EULER_HPR_M01, GENERATE_EULER_HPR_M02, 0,
		GENERATE_EULER_HPR_M10, GENERATE_EULER_HPR_M11, GENERATE_EULER_HPR_M12, 0,
		GENERATE_EULER_HPR_M20, GENERATE_EULER_HPR_M21, GENERATE_EULER_HPR_M22, 0,
		0,                      0,                      0,                      1);
}


//====================================================================
template<typename TMatrixConstAccessor>
__forceinline void  FactorEulerHPRBody(TMatrixConstAccessor& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	if(m(2, 1) < 1.0f) {
		if(m(2, 1) > -1.0f) {
			headDeg  = Atan2Deg(- m(2, 0), m(2, 2));
			pitchDeg = ArcSinDeg(m(2, 1));
			rollDeg  = Atan2Deg(- m(0, 1), m(1, 1));
		}
		else {
			// no unique solution (kill roll)
			headDeg  = - Atan2Deg(m(0, 2), m(0, 0));
			pitchDeg = -90.0f;
			rollDeg  = 0.0f;
		}
	}
	else {
		// no unique solution (kill roll)
		headDeg  = Atan2Deg(m(0, 2), m(0, 0));
		pitchDeg = 90.0f;
		rollDeg  = 0.0f;
	}
}


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall FactorEulerHPR(const TTranspose&, const mat3& _m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	// MV-native
	typedef  Matrix33ConstAccessor<TTranspose>  MatrixConstAccessor;

	MatrixConstAccessor  m(_m);
	FactorEulerHPRBody(m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall FactorEulerHPR(const TTranspose&, const mat4& _m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	// MV-native
	typedef  Matrix44ConstAccessor<TTranspose>  MatrixConstAccessor;

	MatrixConstAccessor  m(_m);
	FactorEulerHPRBody(m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
#define  GENERATE_EULER_PHR_PROLOG \
	float  sH, cH, sP, cP, sR, cR; \
	\
	SinCosDeg(headDeg, sH, cH); \
	SinCosDeg(pitchDeg, sP, cP); \
	SinCosDeg(rollDeg, sR, cR); \
	\
	float  cRsP = cR * sP; \
	float  cRcP = cR * cP; \
	float  sRcP = sR * cP; \
	float  sRsP = sR * sP;


#define  GENERATE_EULER_PHR_M00   (cH * cR)
#define  GENERATE_EULER_PHR_M01   (cRsP * sH - sRcP)
#define  GENERATE_EULER_PHR_M02   (cRcP * sH + sRsP)
#define  GENERATE_EULER_PHR_M10   (cH * sR)
#define  GENERATE_EULER_PHR_M11   (cRcP + sRsP * sH)
#define  GENERATE_EULER_PHR_M12   (sRcP * sH - cRsP)
#define  GENERATE_EULER_PHR_M20   (- sH)
#define  GENERATE_EULER_PHR_M21   (cH * sP)
#define  GENERATE_EULER_PHR_M22   (cH * cP)


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall GenerateEulerPHR(const TTranspose& tt, mat3& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	// MV-native
	GENERATE_EULER_PHR_PROLOG;

	Fill(tt, m,
		GENERATE_EULER_PHR_M00, GENERATE_EULER_PHR_M01, GENERATE_EULER_PHR_M02,
		GENERATE_EULER_PHR_M10, GENERATE_EULER_PHR_M11, GENERATE_EULER_PHR_M12,
		GENERATE_EULER_PHR_M20, GENERATE_EULER_PHR_M21, GENERATE_EULER_PHR_M22);
}


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall GenerateEulerPHR(const TTranspose& tt, mat4& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	// MV-native
	GENERATE_EULER_PHR_PROLOG;

	Fill(tt, m,
		GENERATE_EULER_PHR_M00, GENERATE_EULER_PHR_M01, GENERATE_EULER_PHR_M02, 0,
		GENERATE_EULER_PHR_M10, GENERATE_EULER_PHR_M11, GENERATE_EULER_PHR_M12, 0,
		GENERATE_EULER_PHR_M20, GENERATE_EULER_PHR_M21, GENERATE_EULER_PHR_M22, 0,
		0,                      0,                      0,                      1);
}


//====================================================================
template<typename TMatrixConstAccessor>
__forceinline void  FactorEulerPHRBody(TMatrixConstAccessor& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	if(m(2, 0) < 1.0f) {
		if(m(2, 0) > -1.0f) {
			headDeg  = ArcSinDeg(-m(2, 0));
			pitchDeg = Atan2Deg(m(2, 1), m(2, 2));
			rollDeg  = Atan2Deg(m(1, 0), m(0, 0));
		}
		else {
			// no unique solution (kill roll)
			headDeg  = 90.0f;
			pitchDeg = - Atan2Deg(m(0, 1), m(0, 2));
			rollDeg  = 0.0f;
		}
	}
	else {
		// no unique solution (kill roll)
		headDeg  = -90.0f;
		pitchDeg = Atan2Deg(-m(0, 1), -m(0, 2));
		rollDeg  = 0.0f;
	}
}


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall FactorEulerPHR(const TTranspose&, const mat3& _m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	// MV-native
	typedef  Matrix33ConstAccessor<TTranspose>  MatrixConstAccessor;

	MatrixConstAccessor  m(_m);
	FactorEulerPHRBody(m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall FactorEulerPHR(const TTranspose&, const mat4& _m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	// MV-native
	typedef  Matrix44ConstAccessor<TTranspose>  MatrixConstAccessor;

	MatrixConstAccessor  m(_m);
	FactorEulerPHRBody(m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
template<typename TTranspose>
__forceinline
void __fastcall GenerateProjectionOrtho(const TTranspose& tt, mat4& m,
								bool bEyeSpaceRHS, bool bClipSpace0To1,
								float xLeft, float xRight,
								float yBottom, float yTop,
								float zNear, float zFar)
{
	// MV-native
	float OneOverRMinL = xRight - xLeft;
	if(! OneOverRMinL) return;
	OneOverRMinL = 1.0f / OneOverRMinL;

	float OneOverTMinB = yTop - yBottom;
	if(! OneOverTMinB) return;
	OneOverTMinB = 1.0f / OneOverTMinB;

	float OneOverFMinN = zFar - zNear;
	if(! OneOverFMinN) return;
	OneOverFMinN = 1.0f / OneOverFMinN;

	float ModifierRHS = bEyeSpaceRHS ? -1 : 1;
	float ModifierM22;
	float ModifierM23;

	if(bClipSpace0To1) {
		ModifierM22 = ModifierRHS;
		ModifierM23 = - zNear;
	}
	else {
		ModifierM22 = ModifierRHS * 2;
		ModifierM23 = -(zFar + zNear);
	}

	Fill(tt, m,
		2 * OneOverRMinL,   0,   0,   -(xRight + xLeft) * OneOverRMinL,
		0,   2 * OneOverTMinB,   0,   -(yTop + yBottom) * OneOverTMinB,
		0,   0,   ModifierM22 * OneOverFMinN,   ModifierM23 * OneOverFMinN,
		0,   0,   0,   1);
}


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
void __fastcall RemapAngleTo180DegRange(float& angle)
{
	RemapAngleTo360DegRange(angle);

	if(angle <= -180)    angle += 360;
	else if(angle > 180) angle -= 360;
}


//====================================================================
void __fastcall RemapAngleTo90DegRange(float& angle)
{
	RemapAngleTo360DegRange(angle);

	if(angle < -90) angle += 360;
	else if(angle > 270) angle -= 360;

	if(angle > 90) angle = 180 - angle;
}


//====================================================================
void __fastcall GenerateUniformScale(mat2& m, float scale)
{
	Fill(no_transpose_t(), m,
		scale, 0,
		0, scale);
}


//====================================================================
void __fastcall GenerateUniformScale(mat3& m, float scale)
{
	Fill(no_transpose_t(), m,
		scale, 0, 0,
		0, scale, 0,
		0, 0, scale);
}


//====================================================================
void __fastcall GenerateUniformScale(mat4& m, float scale)
{
	Fill(no_transpose_t(), m,
		scale, 0, 0, 0,
		0, scale, 0, 0,
		0, 0, scale, 0,
		0, 0,     0, 1);
}


//====================================================================
void __fastcall GenerateNonUniformScale(mat2& m, float scaleX, float scaleY)
{
	Fill(no_transpose_t(), m,
		scaleX, 0,
		0, scaleY);
}


//====================================================================
void __fastcall GenerateNonUniformScale(mat3& m, float scaleX, float scaleY, float scaleZ)
{
	Fill(no_transpose_t(), m,
		scaleX, 0, 0,
		0, scaleY, 0,
		0, 0, scaleZ);
}


//====================================================================
void __fastcall GenerateNonUniformScale(mat4& m, float scaleX, float scaleY, float scaleZ)
{
	Fill(no_transpose_t(), m,
		scaleX, 0, 0, 0,
		0, scaleY, 0, 0,
		0, 0, scaleZ, 0,
		0, 0,      0, 1);
}


//====================================================================
void __fastcall GenerateTranslateMV(mat3& m, float trX, float trY)
{
	GenerateTranslate(no_transpose_t(), m, trX, trY);
}


//====================================================================
void __fastcall GenerateTranslateVM(mat3& m, float trX, float trY)
{
	GenerateTranslate(transpose_t(), m, trX, trY);
}


//====================================================================
void __fastcall GenerateTranslateMV(mat4& m, float trX, float trY, float trZ)
{
	GenerateTranslate(no_transpose_t(), m, trX, trY, trZ);
}


//====================================================================
void __fastcall GenerateTranslateVM(mat4& m, float trX, float trY, float trZ)
{
	GenerateTranslate(transpose_t(), m, trX, trY, trZ);
}


//====================================================================
void __fastcall GenerateRotateMV(mat2& m, float angleDeg)
{
	GenerateRotate(no_transpose_t(), m, angleDeg);
}


//====================================================================
void __fastcall GenerateRotateVM(mat2& m, float angleDeg)
{
	GenerateRotate(transpose_t(), m, angleDeg);
}


//====================================================================
void __fastcall GenerateRotateMV(mat3& m, float angleDeg, eAxis axis)
{
	GenerateRotate(no_transpose_t(), m, angleDeg, axis);
}


//====================================================================
void __fastcall GenerateRotateVM(mat3& m, float angleDeg, eAxis axis)
{
	GenerateRotate(transpose_t(), m, angleDeg, axis);
}


//====================================================================
void __fastcall	GenerateRotateMV(mat3& m, float angleDeg, const vec3& unitAxis)
{
	GenerateRotate(no_transpose_t(), m, angleDeg, unitAxis);
}


//====================================================================
void __fastcall	GenerateRotateVM(mat3& m, float angleDeg, const vec3& unitAxis)
{
	GenerateRotate(transpose_t(), m, angleDeg, unitAxis);
}


//====================================================================
void __fastcall GenerateRotateMV(mat4& m, float angleDeg, eAxis axis)
{
	GenerateRotate(no_transpose_t(), m, angleDeg, axis);
}


//====================================================================
void __fastcall GenerateRotateVM(mat4& m, float angleDeg, eAxis axis)
{
	GenerateRotate(transpose_t(), m, angleDeg, axis);
}


//====================================================================
void __fastcall	GenerateRotateMV(mat4& m, float angleDeg, const vec3& unitAxis)
{
	GenerateRotate(no_transpose_t(), m, angleDeg, unitAxis);
}


//====================================================================
void __fastcall	GenerateRotateVM(mat4& m, float angleDeg, const vec3& unitAxis)
{
	GenerateRotate(transpose_t(), m, angleDeg, unitAxis);
}


//====================================================================
void __fastcall	GenerateRotateMV(mat3& m, const vec3& unitAxisU,
								const vec3& unitAxisV, const vec3& unitAxisN)
{
	Fill(no_transpose_t(), m,
		unitAxisU.x, unitAxisV.x, unitAxisN.x,
		unitAxisU.y, unitAxisV.y, unitAxisN.y,
		unitAxisU.z, unitAxisV.z, unitAxisN.z);
}


//====================================================================
void __fastcall	GenerateRotateVM(mat3& m, const vec3& unitAxisU,
								const vec3& unitAxisV, const vec3& unitAxisN)
{
	Fill(transpose_t(), m,
		unitAxisU.x, unitAxisV.x, unitAxisN.x,
		unitAxisU.y, unitAxisV.y, unitAxisN.y,
		unitAxisU.z, unitAxisV.z, unitAxisN.z);
}


//====================================================================
void __fastcall	GenerateRotateMV(mat4& m, const vec3& unitAxisU,
								const vec3& unitAxisV, const vec3& unitAxisN)
{
	Fill(no_transpose_t(), m,
		unitAxisU.x, unitAxisV.x, unitAxisN.x, 0,
		unitAxisU.y, unitAxisV.y, unitAxisN.y, 0,
		unitAxisU.z, unitAxisV.z, unitAxisN.z, 0,
		0,           0,           0,           1);
}


//====================================================================
void __fastcall	GenerateRotateVM(mat4& m, const vec3& unitAxisU,
								const vec3& unitAxisV, const vec3& unitAxisN)
{
	Fill(transpose_t(), m,
		unitAxisU.x, unitAxisV.x, unitAxisN.x, 0,
		unitAxisU.y, unitAxisV.y, unitAxisN.y, 0,
		unitAxisU.z, unitAxisV.z, unitAxisN.z, 0,
		0,           0,           0,           1);
}


//====================================================================
void __fastcall	FactorRotateMV(const mat3& _m, vec3& unitAxisU,
								vec3& unitAxisV, vec3& unitAxisN)
{
	Matrix33ConstAccessor<no_transpose_t>  m(_m);
	FactorRotateBody(m, unitAxisU, unitAxisV, unitAxisN);
}


//====================================================================
void __fastcall	FactorRotateVM(const mat3& _m, vec3& unitAxisU,
								vec3& unitAxisV, vec3& unitAxisN)
{
	Matrix33ConstAccessor<transpose_t>  m(_m);
	FactorRotateBody(m, unitAxisU, unitAxisV, unitAxisN);
}


//====================================================================
void __fastcall	FactorRotateMV(const mat4& _m, vec3& unitAxisU,
								vec3& unitAxisV, vec3& unitAxisN)
{
	Matrix44ConstAccessor<no_transpose_t>  m(_m);
	FactorRotateBody(m, unitAxisU, unitAxisV, unitAxisN);
}


//====================================================================
void __fastcall	FactorRotateVM(const mat4& _m, vec3& unitAxisU,
								vec3& unitAxisV, vec3& unitAxisN)
{
	Matrix44ConstAccessor<transpose_t>  m(_m);
	FactorRotateBody(m, unitAxisU, unitAxisV, unitAxisN);
}


//====================================================================
void __fastcall DirectionFromEulerHP(vec3& unitDirection,
								float headDeg, float pitchDeg)
{
	// x =   cos(p) * sin(h)
	// y = - sin(p)
	// z =   cos(p) * cos(h)

	float sH, cH, sP, cP;
	SinCosDeg(headDeg, sH, cH);
	SinCosDeg(pitchDeg, sP, cP);

	unitDirection.x = cP * sH;
	unitDirection.y = - sP;
	unitDirection.z = cP * cH;
}


//====================================================================
void __fastcall EulerHPFromDirection(const vec3& unitDirection,
								float& headDeg, float& pitchDeg)
{
	// tan(h) = sin(h) / cos(h) = x / z
	headDeg = Atan2Deg(unitDirection.x, unitDirection.z);

	// p = asin(-y)
	pitchDeg = ArcSinDeg(- unitDirection.y);
}


//====================================================================
void __fastcall GenerateEulerHPR_MV(mat3& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	GenerateEulerHPR(no_transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall GenerateEulerHPR_VM(mat3& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	GenerateEulerHPR(transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall GenerateEulerHPR_MV(mat4& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	GenerateEulerHPR(no_transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall GenerateEulerHPR_VM(mat4& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	GenerateEulerHPR(transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall FactorEulerHPR_MV(const mat3& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	FactorEulerHPR(no_transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall FactorEulerHPR_VM(const mat3& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	FactorEulerHPR(transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall FactorEulerHPR_MV(const mat4& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	FactorEulerHPR(no_transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall FactorEulerHPR_VM(const mat4& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	FactorEulerHPR(transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall GenerateEulerPHR_MV(mat3& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	GenerateEulerPHR(no_transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall GenerateEulerPHR_VM(mat3& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	GenerateEulerPHR(transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall GenerateEulerPHR_MV(mat4& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	GenerateEulerPHR(no_transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall GenerateEulerPHR_VM(mat4& m,
						float headDeg, float pitchDeg, float rollDeg)
{
	GenerateEulerPHR(transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall FactorEulerPHR_MV(const mat3& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	FactorEulerPHR(no_transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall FactorEulerPHR_VM(const mat3& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	FactorEulerPHR(transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall FactorEulerPHR_MV(const mat4& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	FactorEulerPHR(no_transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall FactorEulerPHR_VM(const mat4& m,
						float& headDeg, float& pitchDeg, float& rollDeg)
{
	FactorEulerPHR(transpose_t(), m, headDeg, pitchDeg, rollDeg);
}


//====================================================================
void __fastcall GenerateProjectionOrthoMV(mat4& m,
								bool bEyeSpaceRHS, bool bClipSpace0To1,
								float xLeft, float xRight,
								float yBottom, float yTop,
								float zNear, float zFar)
{
	GenerateProjectionOrtho(no_transpose_t(), m, bEyeSpaceRHS, bClipSpace0To1,
								xLeft, xRight, yBottom, yTop, zNear, zFar);
}


//====================================================================
void __fastcall GenerateProjectionOrthoVM(mat4& m,
								bool bEyeSpaceRHS, bool bClipSpace0To1,
								float xLeft, float xRight,
								float yBottom, float yTop,
								float zNear, float zFar)
{
	GenerateProjectionOrtho(transpose_t(), m, bEyeSpaceRHS, bClipSpace0To1,
								xLeft, xRight, yBottom, yTop, zNear, zFar);
}


//********************************************************************
}//namespace vertexware


