

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "MathTools\Matrix22.h"
#include "MathTools\Matrix33.h"
#include "MathTools\Matrix44.h"
#include "private\vw_common_exports.h"


//********************************************************************
namespace vertexware {


//--------------------------------------------------------------------
// NOTES:
// 1) Use MV-generators to create matrices suitable for
//    matrix * vector multiplication order.
// 2) Use VM-generators to create matrices suitable for
//    vector * matrix multiplication order.
// 3) If specific generator does not have MV/VM suffix, the generated
//    matrix is suitable for use with both multiplication orders.
//--------------------------------------------------------------------


//====================================================================
// Remaps angle to <-180, 180]
//====================================================================
VW_COMMON_API
void __fastcall RemapAngleTo180DegRange(float& angle);


//====================================================================
// Remaps angle to [-90, 90]
//====================================================================
VW_COMMON_API
void __fastcall RemapAngleTo90DegRange(float& angle);


//====================================================================
// Uniform scaling.
//====================================================================
VW_COMMON_API
void __fastcall GenerateUniformScale(mat2& m, float scale);


//====================================================================
VW_COMMON_API
void __fastcall GenerateUniformScale(mat3& m, float scale);


//====================================================================
VW_COMMON_API
void __fastcall GenerateUniformScale(mat4& m, float scale);


//====================================================================
// Non-uniform scaling.
//====================================================================
VW_COMMON_API
void __fastcall GenerateNonUniformScale(mat2& m, float scaleX, float scaleY);


//====================================================================
VW_COMMON_API
void __fastcall GenerateNonUniformScale(mat3& m, float scaleX, float scaleY, float scaleZ);


//====================================================================
VW_COMMON_API
void __fastcall GenerateNonUniformScale(mat4& m, float scaleX, float scaleY, float scaleZ);


//====================================================================
// Translation.
//====================================================================
VW_COMMON_API
void __fastcall GenerateTranslateMV(mat3& m, float trX, float trY);


//====================================================================
VW_COMMON_API
void __fastcall GenerateTranslateVM(mat3& m, float trX, float trY);


//====================================================================
VW_COMMON_API
void __fastcall GenerateTranslateMV(mat4& m, float trX, float trY, float trZ);


//====================================================================
VW_COMMON_API
void __fastcall GenerateTranslateVM(mat4& m, float trX, float trY, float trZ);


//====================================================================
// 2D-Rotation (clockwise).
//====================================================================
VW_COMMON_API
void __fastcall GenerateRotateMV(mat2& m, float angleDeg);


//====================================================================
VW_COMMON_API
void __fastcall GenerateRotateVM(mat2& m, float angleDeg);


//====================================================================
// 3D-Rotation (clockwise in LHS, counter-clockwise in RHS).
//====================================================================
enum eAxis
{
	Axis_X = 0,
	Axis_Y,
	Axis_Z,

	Axis_Count,
	Axis_Force32Bits = 0x7FFFFFFF
};


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateMV(mat3& m, float angleDeg, eAxis axis);


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateVM(mat3& m, float angleDeg, eAxis axis);


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateMV(mat3& m, float angleDeg, const vec3& unitAxis);


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateVM(mat3& m, float angleDeg, const vec3& unitAxis);


//====================================================================
VW_COMMON_API
void __fastcall GenerateRotateMV(mat4& m, float angleDeg, eAxis axis);


//====================================================================
VW_COMMON_API
void __fastcall GenerateRotateVM(mat4& m, float angleDeg, eAxis axis);


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateMV(mat4& m, float angleDeg, const vec3& unitAxis);


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateVM(mat4& m, float angleDeg, const vec3& unitAxis);


//====================================================================
// Rotation matrix <===> coordinate-system axes.
// N is the direction-vector.
// V is the up-vector.
// U = V x N
//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateMV(mat3& m, const vec3& unitAxisU,
								const vec3& unitAxisV, const vec3& unitAxisN);


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateVM(mat3& m, const vec3& unitAxisU,
								const vec3& unitAxisV, const vec3& unitAxisN);


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateMV(mat4& m, const vec3& unitAxisU,
								const vec3& unitAxisV, const vec3& unitAxisN);


//====================================================================
VW_COMMON_API
void __fastcall	GenerateRotateVM(mat4& m, const vec3& unitAxisU,
								const vec3& unitAxisV, const vec3& unitAxisN);


//====================================================================
VW_COMMON_API
void __fastcall	FactorRotateMV(const mat3& m, vec3& unitAxisU,
								vec3& unitAxisV, vec3& unitAxisN);


//====================================================================
VW_COMMON_API
void __fastcall	FactorRotateVM(const mat3& m, vec3& unitAxisU,
								vec3& unitAxisV, vec3& unitAxisN);


//====================================================================
VW_COMMON_API
void __fastcall	FactorRotateMV(const mat4& m, vec3& unitAxisU,
								vec3& unitAxisV, vec3& unitAxisN);


//====================================================================
VW_COMMON_API
void __fastcall	FactorRotateVM(const mat4& m, vec3& unitAxisU,
								vec3& unitAxisV, vec3& unitAxisN);


//====================================================================
// Euler Head-Pitch-Roll tools.
// Positive head, pitch and roll are coordinate-system native.
// Reference orientation is HPR(0, 0, 0) or Direction(0, 0, 1)/Roll(0)
// Ranges:
//    Head  [-180, 180]
//    Pitch [- 90,  90]
//    Roll  [-180, 180]
//====================================================================
VW_COMMON_API
void __fastcall DirectionFromEulerHP(vec3& unitDirection,
								float headDeg, float pitchDeg);


//====================================================================
VW_COMMON_API
void __fastcall EulerHPFromDirection(const vec3& unitDirection,
								float& headDeg, float& pitchDeg);


//====================================================================
// Head-Pitch-Roll transform order
//====================================================================
VW_COMMON_API
void __fastcall GenerateEulerHPR_MV(mat3& m,
						float headDeg, float pitchDeg, float rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall GenerateEulerHPR_VM(mat3& m,
						float headDeg, float pitchDeg, float rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall GenerateEulerHPR_MV(mat4& m,
						float headDeg, float pitchDeg, float rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall GenerateEulerHPR_VM(mat4& m,
						float headDeg, float pitchDeg, float rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall FactorEulerHPR_MV(const mat3& m,
						float& headDeg, float& pitchDeg, float& rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall FactorEulerHPR_VM(const mat3& m,
						float& headDeg, float& pitchDeg, float& rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall FactorEulerHPR_MV(const mat4& m,
						float& headDeg, float& pitchDeg, float& rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall FactorEulerHPR_VM(const mat4& m,
						float& headDeg, float& pitchDeg, float& rollDeg);


//====================================================================
// Pitch-Head-Roll transform order
//====================================================================
VW_COMMON_API
void __fastcall GenerateEulerPHR_MV(mat3& m,
						float headDeg, float pitchDeg, float rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall GenerateEulerPHR_VM(mat3& m,
						float headDeg, float pitchDeg, float rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall GenerateEulerPHR_MV(mat4& m,
						float headDeg, float pitchDeg, float rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall GenerateEulerPHR_VM(mat4& m,
						float headDeg, float pitchDeg, float rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall FactorEulerPHR_MV(const mat3& m,
						float& headDeg, float& pitchDeg, float& rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall FactorEulerPHR_VM(const mat3& m,
						float& headDeg, float& pitchDeg, float& rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall FactorEulerPHR_MV(const mat4& m,
						float& headDeg, float& pitchDeg, float& rollDeg);


//====================================================================
VW_COMMON_API
void __fastcall FactorEulerPHR_VM(const mat4& m,
						float& headDeg, float& pitchDeg, float& rollDeg);


//====================================================================
// Orthographic projections (from eye-space to LHS clip-space).
// Set <bEyeSpaceRHS> to true if eye-space is RHS.
// Set <bClipSpace0To1> to true if clip-space expects transformed
// z-coords in range [0, 1] (used by D3D).
// Set <bClipSpace0To1> to false if clip-space expects transformed
// z-coords in range [-1, 1] (used by OpenGL).
//====================================================================
VW_COMMON_API
void __fastcall GenerateProjectionOrthoMV(mat4& m,
								bool bEyeSpaceRHS, bool bClipSpace0To1,
								float xLeft, float xRight,
								float yBottom, float yTop,
								float zNear, float zFar);


//====================================================================
VW_COMMON_API
void __fastcall GenerateProjectionOrthoVM(mat4& m,
								bool bEyeSpaceRHS, bool bClipSpace0To1,
								float xLeft, float xRight,
								float yBottom, float yTop,
								float zNear, float zFar);


//********************************************************************
}//namespace vertexware


