

//--------------------------------------------------------------------
#include "MathTools\Vector3.h"
#include "BaseTools\Serializers.h"


//********************************************************************
namespace vertexware {


//====================================================================
const vec3  vec3::ZERO     = vec3(0, 0, 0);
const vec3  vec3::ONE      = vec3(1, 1, 1);
const vec3  vec3::INFINITY = vec3(math_constants::PlusInfinity,
											math_constants::PlusInfinity,
											math_constants::PlusInfinity);
const vec3  vec3::UNIT_X   = vec3(1, 0, 0);
const vec3  vec3::UNIT_Y   = vec3(0, 1, 0);
const vec3  vec3::UNIT_Z   = vec3(0, 0, 1);


//====================================================================
bool  vec3::IsZero(float tolerance) const
{
	return
	InRange(x, -tolerance, tolerance)  &&
	InRange(y, -tolerance, tolerance)  &&
	InRange(z, -tolerance, tolerance);
}


//====================================================================
bool  vec3::HasZeroComponent(float tolerance) const
{
	return
	InRange(x, -tolerance, tolerance)  ||
	InRange(y, -tolerance, tolerance)  ||
	InRange(z, -tolerance, tolerance);
}


//====================================================================
vec3&  vec3::RotationAroundUnit(const vec3& axisUnit,
										float angleDeg, const vec3& src)
{
	// 1) axis is unit-vector
	// 2) decompose vector to components: parallel to axis & perpendicular to axis
	//    para = (axis . src) * axis
	//    perp = src - para
	// 3) para remains unchanged after rotation
	//    Rot(para) = para
	// 4) perp is rotated in plane perpendicular to axis
	//    V = vector perpendicular to perp, lying in the plane perpendicular to axis
	//    V = axis  x  perp  =  axis  x  src
	//    Rot(perp) = cos(angle) * perp  +  sin(angle) * V
	// 5) Rot(src)  = Rot(para) + Rot(perp)
	//    Rot(src)  = cos(angle) * src + (1 - cos(angle)) * axis * (axis . src)
	//                                 + sin(angle) * (axis x src)

	float sinA, cosA;
	SinCosDeg(angleDeg, sinA, cosA);

	this->Mul( src, cosA );
	this->Add( axisUnit * ((1 - cosA) * axisUnit.Dot(src)) );
	this->Add( axisUnit.Cross(src) * sinA );

	return *this;
}


//====================================================================
void __fastcall OrthoNormalize(vec3& v0, vec3& v1, vec3& v2)
{
	v0.Unitize();

	v1 -= v1.ProjectionTo(v0);
	v1.Unitize();

	v2 -= v2.ProjectionTo(v0) + v2.ProjectionTo(v1);
	v2.Unitize();
}


//====================================================================
std::ostream& __fastcall operator<=(std::ostream &os, const vec3& v)
{
	os <= v.x <= v.y <= v.z;
	return os;
}


//====================================================================
std::istream& __fastcall operator>=(std::istream &is, vec3& v)
{
	is >= v.x >= v.y >= v.z;
	return is;
}


//********************************************************************
}//namespace vertexware


