

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "MathTools\VectorBase.h"
#include "MathTools\MathProcessors.h"
#include "private\vw_common_exports.h"
#include <iosfwd>


//********************************************************************
namespace vertexware {


//====================================================================
class vec2;
class vec4;


//====================================================================
// 3-element vector.
//====================================================================
class VW_COMMON_API vec3
{
public:
	//------
	// Data.
	//------
	union {
		float  v[3];
		struct {
			float  x;
			float  y;
			float  z;
		};
	};

	//-----------------------
	// Convenience constants.
	//-----------------------
	static const vec3  ZERO;
	static const vec3  ONE;
	static const vec3  INFINITY;
	static const vec3  UNIT_X;
	static const vec3  UNIT_Y;
	static const vec3  UNIT_Z;


public:
	//-------------------------
	// No-overhead constructor.
	//-------------------------
	__forceinline vec3()
	{ }

	//---------------------------
	// Initializing constructors.
	//---------------------------
	__forceinline vec3(float _x, float _y, float _z) {
		x = _x;
		y = _y;
		z = _z;
	}

	__forceinline explicit vec3(const float* a) {
		x = a[0];
		y = a[1];
		z = a[2];
	}

	//------------------
	// Copy constructor.
	//------------------
	__forceinline vec3(const vec3& rhs) {
		x = rhs.x;
		y = rhs.y;
		z = rhs.z;
	}

	__forceinline vec3(const vec2& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		z = 0;
	}

	__forceinline vec3(const vec4& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		z = reinterpret_cast<const float*>(&rhs)[2];
	}

	//-----------------
	// Copy assignment.
	//-----------------
	__forceinline vec3& operator=(const vec3& rhs) {
		x = rhs.x;
		y = rhs.y;
		z = rhs.z;
		return *this;
	}

	vec3& operator=(const vec2& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		z = 0;
		return *this;
	}

	vec3& operator=(const vec4& rhs) {
		x = reinterpret_cast<const float*>(&rhs)[0];
		y = reinterpret_cast<const float*>(&rhs)[1];
		z = reinterpret_cast<const float*>(&rhs)[2];
		return *this;
	}

	//----------------------
	// One-call assignments.
	//----------------------
	__forceinline void  Set(float _x, float _y, float _z) {
		x = _x;
		y = _y;
		z = _z;
	}
	__forceinline void  Set(const float* a) {
		x = a[0];
		y = a[1];
		z = a[2];
	}

	//------------------------
	// Conversion to pointers.
	//------------------------
	__forceinline float*        ToPtr()       { return &x; }
	__forceinline const float*  ToPtr() const { return &x; }

	__forceinline operator float*()             { return ToPtr(); }
	__forceinline operator const float*() const { return ToPtr(); }

	//----------
	// Addition.
	//----------
	__forceinline vec3&  operator+=(const vec3& rhs) {
		x += rhs.x;
		y += rhs.y;
		z += rhs.z;
		return *this;
	}

	__forceinline vec3&  Add(const vec3& rhs) {
		x += rhs.x;
		y += rhs.y;
		z += rhs.z;
		return *this;
	}

	__forceinline vec3  operator+(const vec3& rhs) const {
		return vec3(x + rhs.x, y + rhs.y, z + rhs.z);
	}

	__forceinline vec3&  Add(const vec3& lhs, const vec3& rhs) {
		x = lhs.x + rhs.x;
		y = lhs.y + rhs.y;
		z = lhs.z + rhs.z;
		return *this;
	}

	//-------------
	// Subtraction.
	//-------------
	__forceinline vec3&  operator-=(const vec3& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		z -= rhs.z;
		return *this;
	}

	__forceinline vec3&  Sub(const vec3& rhs) {
		x -= rhs.x;
		y -= rhs.y;
		z -= rhs.z;
		return *this;
	}

	__forceinline vec3  operator-(const vec3& rhs) const {
		return vec3(x - rhs.x, y - rhs.y, z - rhs.z);
	}

	__forceinline vec3&  Sub(const vec3& lhs, const vec3& rhs) {
		x = lhs.x - rhs.x;
		y = lhs.y - rhs.y;
		z = lhs.z - rhs.z;
		return *this;
	}

	//----------------
	// Multiplication.
	//----------------
	__forceinline vec3&  operator*=(const vec3& rhs) {
		x *= rhs.x;
		y *= rhs.y;
		z *= rhs.z;
		return *this;
	}

	__forceinline vec3&  Mul(const vec3& rhs) {
		x *= rhs.x;
		y *= rhs.y;
		z *= rhs.z;
		return *this;
	}

	__forceinline vec3  operator*(const vec3& rhs) const {
		return vec3(x * rhs.x, y * rhs.y, z * rhs.z);
	}

	__forceinline vec3&  Mul(const vec3& lhs, const vec3& rhs) {
		x = lhs.x * rhs.x;
		y = lhs.y * rhs.y;
		z = lhs.z * rhs.z;
		return *this;
	}

	//------------------------------------
	// Division (no divide-by-zero check).
	//------------------------------------
	__forceinline vec3&  operator/=(const vec3& rhs) {
		x /= rhs.x;
		y /= rhs.y;
		z /= rhs.z;
		return *this;
	}

	__forceinline vec3&  Div(const vec3& rhs) {
		x /= rhs.x;
		y /= rhs.y;
		z /= rhs.z;
		return *this;
	}

	__forceinline vec3  operator/(const vec3& rhs) const {
		return vec3(x / rhs.x, y / rhs.y, z / rhs.z);
	}

	__forceinline vec3&  Div(const vec3& lhs, const vec3& rhs) {
		x = lhs.x / rhs.x;
		y = lhs.y / rhs.y;
		z = lhs.z / rhs.z;
		return *this;
	}

	//--------------------------
	// Multiplication by scalar.
	//--------------------------
	__forceinline vec3&  operator*=(float num) {
		x *= num;
		y *= num;
		z *= num;
		return *this;
	}

	__forceinline vec3&  Mul(float num) {
		x *= num;
		y *= num;
		z *= num;
		return *this;
	}

	__forceinline vec3  operator*(float num) const {
		return vec3(x * num, y * num, z * num);
	}

	__forceinline vec3&  Mul(const vec3& lhs, float num) {
		x = lhs.x * num;
		y = lhs.y * num;
		z = lhs.z * num;
		return *this;
	}

	//----------------------------------------------
	// Division by scalar (no divide-by-zero check).
	//----------------------------------------------
	__forceinline vec3&  operator/=(float num) {
		num = 1 / num;
		x *= num;
		y *= num;
		z *= num;
		return *this;
	}

	__forceinline vec3&  Div(float num) {
		num = 1 / num;
		x *= num;
		y *= num;
		z *= num;
		return *this;
	}

	__forceinline vec3  operator/(float num) const {
		num = 1 / num;
		return vec3(x * num, y * num, z * num);
	}

	__forceinline vec3&  Div(const vec3& lhs, float num) {
		num = 1 / num;
		x = lhs.x * num;
		y = lhs.y * num;
		z = lhs.z * num;
		return *this;
	}

	//----------
	// Negation.
	//----------
	__forceinline vec3&  Neg() {
		x *= -1;
		y *= -1;
		z *= -1;
		return *this;
	}

	__forceinline vec3&  Neg(const vec3& src) {
		x = - src.x;
		y = - src.y;
		z = - src.z;
		return *this;
	}

	//------------------
	// Unary operations.
	//------------------
	__forceinline vec3  operator-() const {
		return vec3(-x, -y, -z);
	}

	__forceinline vec3&        operator+()       { return *this; }
	__forceinline const vec3&  operator+() const { return *this; }

	//----------------------------
	// Absolute minimum component.
	//----------------------------
	__forceinline float  AbsMin() const {
		return Min(Abs(x), Abs(y), Abs(z));
	}

	//----------------------------
	// Absolute maximum component.
	//----------------------------
	__forceinline float  AbsMax() const {
		return Max(Abs(x), Abs(y), Abs(z));
	}

	//--------
	// Length.
	//--------
	__forceinline float  Length() const {
		return Sqrt<float>(x * x + y * y + z * z);
	}

	//----------------
	// Squared length.
	//----------------
	__forceinline float  LengthSquared() const {
		return (x * x + y * y + z * z);
	}

	//---------
	// Unitize.
	//---------
	__forceinline vec3&  Unitize() {
		float len = Length();
		if(len > 0) {
			len = 1 / len;
			x *= len;
			y *= len;
			z *= len;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	__forceinline vec3  Unit() const {
		float len = Length();
		if(len > 0) {
			len = 1 / len;
			return vec3(x * len, y * len, z * len);
		}
		return INFINITY;
	}

	__forceinline vec3&  Unit(const vec3& src) {
		float len = src.Length();
		if(len > 0) {
			len = 1 / len;
			x = src.x * len;
			y = src.y * len;
			z = src.z * len;
			return *this;
		}
		*this = INFINITY;
		return *this;
	}

	//-------------
	// Dot product.
	//-------------
	__forceinline float  Dot(const vec3& rhs) const {
		return (x * rhs.x + y * rhs.y + z * rhs.z);
	}

	//---------------
	// Cross product.
	//---------------
	__forceinline vec3  Cross(const vec3& rhs) const {
		return vec3(y * rhs.z - rhs.y * z, rhs.x * z - x * rhs.z, x * rhs.y - rhs.x * y);
	}
	__forceinline vec3&  Cross(const vec3& lhs, const vec3& rhs) {
		x = lhs.y * rhs.z - rhs.y * lhs.z;
		y = rhs.x * lhs.z - lhs.x * rhs.z;
		z = lhs.x * rhs.y - rhs.x * lhs.y;
		return *this;
	}

	//------------
	// Projection.
	//------------
	__forceinline vec3&  ProjectToUnit(const vec3& to) {
		*this = to * this->Dot(to);
		return *this;
	}
	__forceinline vec3   ProjectionToUnit(const vec3& to) const {
		return to * this->Dot(to);
	}
	__forceinline vec3&   ProjectionToUnit(const vec3& to, const vec3& src) {
		return this->Mul(to, src.Dot(to));
	}

	__forceinline vec3&  ProjectTo(const vec3& to) {
		*this = to * (this->Dot(to) / to.LengthSquared());
		return *this;
	}
	__forceinline vec3   ProjectionTo(const vec3& to) const {
		return to * (this->Dot(to) / to.LengthSquared());
	}
	__forceinline vec3&   ProjectionTo(const vec3& to, const vec3& src) {
		return this->Mul(to, src.Dot(to) / to.LengthSquared());
	}

	//----------
	// Distance.
	//----------
	__forceinline float  Distance(const vec3& rhs) const {
		return (*this - rhs).Length();
	}

	//------------------
	// Squared distance.
	//------------------
	__forceinline float  DistanceSquared(const vec3& rhs) const {
		return (*this - rhs).LengthSquared();
	}

	//------------
	// Zero tests.
	//------------
	__forceinline bool  IsZero() const {
		return ((x == 0) && (y == 0) && (z == 0));
	}
	bool  IsZero(float tolerance) const;

	__forceinline bool  HasZeroComponent() const {
		return ((x == 0) || (y == 0) || (z == 0));
	}
	bool  HasZeroComponent(float tolerance) const;

	//-------------
	// Angle tests.
	//-------------
	__forceinline bool  Orthogonal(const vec3& rhs, float maxCosAngle =
									math_constants::VectorOrthogonalTestDefaultMaxCosAngle) const {
		return (Abs(this->Dot(rhs)) <= maxCosAngle);
	}
	__forceinline bool  Parallel(const vec3& rhs, float minCosAngle =
									math_constants::VectorParallelTestDefaultMinCosAngle) const {
		return (Abs(this->Dot(rhs)) >= minCosAngle);
	}

	//-------------
	// Comparisons.
	//-------------
	__forceinline bool  operator==(const vec3& rhs) const {
		return ((x == rhs.x) && (y == rhs.y) && (z == rhs.z));
	}
	__forceinline bool  operator!=(const vec3& rhs) const {
		return !(*this == rhs);
	}
	__forceinline bool  Equal(const vec3& rhs, float toleranceRadius =
								math_constants::VectorEqualTestDefaultToleranceRadius) const {
		return (DistanceSquared(rhs) <= toleranceRadius * toleranceRadius);
	}

	//----------------------------------------------------
	// Noise filtering (zero-out all elements close to 0).
	//----------------------------------------------------
	__forceinline vec3&  NoiseFilter(const vec3& src, float tolerance =
										math_constants::VectorNoiseFilterDefaultTolerance) {
		vertexware::NoiseFilter(this->ToPtr(), 3, tolerance);
		return *this;
	}

	__forceinline  vec3 __fastcall NoiseFiltered(float tolerance =
										math_constants::VectorNoiseFilterDefaultTolerance) const {
		vec3 temp;
		vertexware::NoiseFilter(temp.ToPtr(), this->ToPtr(), 3, tolerance);
		return temp;
	}

	//-------------------------------------------------------
	// Rotation (clockwise in LHS, counter-clockwise in RHS).
	//-------------------------------------------------------
	__forceinline vec3&  RotateAroundUnit(const vec3& axisUnit, float angleDeg) {
		return  this->RotationAroundUnit(axisUnit, angleDeg, vec3(*this));
	}
	__forceinline vec3  RotationAroundUnit(const vec3& axisUnit, float angleDeg) const {
		vec3  temp;
		temp.RotationAroundUnit(axisUnit, angleDeg, *this);
		return  temp;
	}
	vec3&  RotationAroundUnit(const vec3& axisUnit, float angleDeg, const vec3& src);


	__forceinline vec3&  RotateAround(const vec3& axis, float angleDeg) {
		return  this->RotationAround(axis, angleDeg, vec3(*this));
	}
	__forceinline vec3  RotationAround(const vec3& axis, float angleDeg) const {
		vec3  temp;
		temp.RotationAround(axis, angleDeg, *this);
		return  temp;
	}
	__forceinline vec3&  RotationAround(const vec3& axis, float angleDeg, const vec3& src) {
		vec3 axisUnit;
		axisUnit.Unit(axis);
		return  this->RotationAroundUnit(axisUnit, angleDeg, src);
	}
};


//====================================================================
// Clamping.
//====================================================================
__forceinline  vec3 __fastcall ComponentAbs(const vec3& val)
{
	return vec3(Abs(val.x), Abs(val.y), Abs(val.z));
}


//====================================================================
__forceinline  vec3 __fastcall ComponentMin(const vec3& v1, const vec3& v2) {
	return vec3(Min(v1.x, v2.x), Min(v1.y, v2.y), Min(v1.z, v2.z));
}


//====================================================================
__forceinline  vec3 __fastcall ComponentMax(const vec3& v1, const vec3& v2) {
	return vec3(Max(v1.x, v2.x), Max(v1.y, v2.y), Max(v1.z, v2.z));
}


//====================================================================
__forceinline  vec3 __fastcall ComponentClamp(const vec3& v, const vec3& low, const vec3& high) {
	return ComponentMax(ComponentMin(v, high), low);
}


//====================================================================
// Multiplication by scalar (redirection).
//====================================================================
__forceinline  vec3 __fastcall operator*(float num, const vec3& rhs) {
	return rhs * num;
}


//====================================================================
// Orthonormalization.
//====================================================================
VW_COMMON_API
void __fastcall OrthoNormalize(vec3& v0, vec3& v1, vec3& v2);


//====================================================================
// Serialization / deserialization.
//====================================================================
VW_COMMON_API
std::ostream& __fastcall operator<=(std::ostream &os, const vec3& v);
VW_COMMON_API
std::istream& __fastcall operator>=(std::istream &is, vec3& v);


//********************************************************************
}//namespace vertexware


