

//--------------------------------------------------------------------
#include "VideoRendererObjects\VRO_Textures.h"
#include "BaseTools\Memory.h"
#include "MathTools\MathBase.h"
#include "DataObjects\ImageOperations.h"
#include "DataObjects\VolumeOperations.h"
#include "DataObjects\MipmapChains.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace video_renderer_objects {


//====================================================================
bool __fastcall TextureBase::CheckDimension(uint dim, bool bMustBePowerOf2)
{
	if(! dim) return false;
	if(bMustBePowerOf2  &&  ! IsPowerOf2(dim)) return false;
	return true;
}


//====================================================================
TextureBase::TextureBase(eResourceType resourceType, eSampleFormat _sampleFormat,
						bool bMultiLevel, uint maxDimension)
	:
	Resource(resourceType)
{
	sampleFormat = _sampleFormat;

	numLevels = 1;
	if(bMultiLevel) {
		numLevels += Log2<uint>(ClampedLow<uint>(maxDimension, 1));
	}
	baseLevel = 0;
}


//====================================================================
TextureBase::~TextureBase()
{
}


//====================================================================
bool  TextureBase::SetBaseLevel(uint newBaseLevel)
{
	if(baseLevel == newBaseLevel) return true;
	baseLevel = ClampedHigh<uint>(newBaseLevel, numLevels - 1);

	return UpdateBaseLevel();
}


//====================================================================
Texture1D::Texture1D(eSampleFormat sampleFormat, bool bMultiLevel,
						uint _widthLevel0)
	:
	TextureBase(ResourceType_Texture1D, sampleFormat,
				bMultiLevel, _widthLevel0)
{
	widthLevel0  = _widthLevel0;
}


//====================================================================
bool  Texture1D::LoadLevel(uint level, const SignalBase& src,
							uint dataWidth,
							uint dstOffsetX,
							uint srcOffsetX)
{
	if(! dataWidth) dataWidth = src.GetNumSamples();

	VW_ASSERT(level < GetNumLevels(), "Level-index out of range.");
	VW_ASSERT(GetSampleFormat() == src.GetSampleFormat(), "Sample-format mismatch.");
	VW_ASSERT(src.GetNumSamples(), "Source empty.");
	VW_ASSERT(dataWidth, "Data-range empty.");
	VW_ASSERT(dstOffsetX + dataWidth <= GetWidth(level),
				"Destination-offset-x / data-width out of range.");
	VW_ASSERT(srcOffsetX + dataWidth <= src.GetNumSamples(),
				"Source-offset-x / data-width out of range.");

	return
	DoLoadLevel(level, src, dataWidth, dstOffsetX, srcOffsetX);
}


//====================================================================
bool  Texture1D::LoadLevels(const SignalMipmapChain& src)
{
	VW_ASSERT(src.GetSampleFormat() == GetSampleFormat(), "Sample-format mismatch.");
	VW_ASSERT(src.GetNumLevels() == GetNumLevels(), "Level-count mismatch.");
	VW_ASSERT(src.GetLevel(0).GetNumSamples() == GetWidth(0), "Level-0 dimension mismatch.");

	for(uint level = 0; level < src.GetNumLevels(); level++) {
		const SignalBase& srcLevel = src.GetLevel(level);

		if(! DoLoadLevel(level, srcLevel, srcLevel.GetNumSamples(), 0, 0)) return false;
	}
	return true;
}


//====================================================================
bool  Texture1D::LoadLevels(const SignalBase& srcLevel0)
{
	VW_ASSERT(srcLevel0.GetNumSamples() == GetWidth(0), "Level-0 dimension mismatch.");

	if(GetNumLevels() == 1) {
		AutoPtr<SignalBase>  apSrcLevel0(const_cast<SignalBase*>(&srcLevel0), false);
		if(GetSampleFormat() != srcLevel0.GetSampleFormat()) {
			apSrcLevel0.Assign(ManufactureSignal(GetSampleFormat()));
			if(! TransferSamples1D(*apSrcLevel0, srcLevel0)) return false;
		}
		return LoadLevel(0, *apSrcLevel0);
	}

	SignalMipmapChain  mipmapChain(GetSampleFormat());
	if(! mipmapChain.Build(srcLevel0, false)) return false;

	return LoadLevels(mipmapChain);
}


//====================================================================
bool  Texture1D::LoadLevels(ubyte val)
{
	AutoPtr<SignalBase>  apSrcLevel0(ManufactureSignal(GetSampleFormat(), GetWidth(0)));
	ProcessSamples_FillNBytes(*apSrcLevel0, val);

	return LoadLevels(*apSrcLevel0);
}


//====================================================================
bool  Texture2D::s_bSourceRowOrderBottomToTop = false;


//====================================================================
Texture2D::Texture2D(eSampleFormat sampleFormat, bool bMultiLevel,
						uint _heightLevel0, uint _widthLevel0)
	:
	TextureBase(ResourceType_Texture2D, sampleFormat,
				bMultiLevel, Max(_heightLevel0, _widthLevel0))
{
	heightLevel0 = _heightLevel0;
	widthLevel0  = _widthLevel0;
}


//====================================================================
bool  Texture2D::LoadLevel(uint level, const ImageBase& src,
						uint dataHeight, uint dataWidth,
						uint dstOffsetY, uint dstOffsetX,
						uint srcOffsetY, uint srcOffsetX)
{
	if(! dataHeight) dataHeight = src.GetNumRows();
	if(! dataWidth) dataWidth = src.GetNumColumns();

	VW_ASSERT(level < GetNumLevels(), "Level-index out of range.");
	VW_ASSERT(GetSampleFormat() == src.GetSampleFormat(), "Sample-format mismatch.");
	VW_ASSERT(src.GetNumSamples(), "Source empty.");
	VW_ASSERT(dataHeight && dataWidth, "Data-range empty.");
	VW_ASSERT(dstOffsetY + dataHeight <= GetHeight(level),
				"Destination-offset-y / data-height out of range.");
	VW_ASSERT(dstOffsetX + dataWidth <= GetWidth(level),
				"Destination-offset-x / data-width out of range.");
	VW_ASSERT(srcOffsetY + dataHeight <= src.GetNumRows(),
				"Source-offset-y / data-height out of range.");
	VW_ASSERT(srcOffsetX + dataWidth <= src.GetNumColumns(),
				"Source-offset-x / data-width out of range.");

	return
	DoLoadLevel(level, src, IsSourceRowOrderBottomToTop(),
				dataHeight, dataWidth,
				dstOffsetY, dstOffsetX,
				srcOffsetY, srcOffsetX);
}


//====================================================================
bool  Texture2D::LoadLevels(const ImageMipmapChain& src)
{
	VW_ASSERT(src.GetSampleFormat() == GetSampleFormat(), "Sample-format mismatch.");
	VW_ASSERT(src.GetNumLevels() == GetNumLevels(), "Level-count mismatch.");
	VW_ASSERT(src.GetLevel(0).GetNumRows() == GetHeight(0) &&
			src.GetLevel(0).GetNumColumns() == GetWidth(0), "Level-0 dimension mismatch.");

	for(uint level = 0; level < src.GetNumLevels(); level++) {
		const ImageBase& srcLevel = src.GetLevel(level);

		if(! DoLoadLevel(level, srcLevel,
					IsSourceRowOrderBottomToTop(),
					srcLevel.GetNumRows(), srcLevel.GetNumColumns(),
					0, 0, 0, 0)) return false;
	}
	return true;
}


//====================================================================
bool  Texture2D::LoadLevels(const ImageBase& srcLevel0)
{
	VW_ASSERT(srcLevel0.GetNumRows() == GetHeight(0) &&
			srcLevel0.GetNumColumns() == GetWidth(0), "Level-0 dimension mismatch.");

	if(GetNumLevels() == 1) {
		AutoPtr<ImageBase>  apSrcLevel0(const_cast<ImageBase*>(&srcLevel0), false);
		if(GetSampleFormat() != srcLevel0.GetSampleFormat()) {
			apSrcLevel0.Assign(ManufactureImage(GetSampleFormat()));
			if(! TransferSamples2D(*apSrcLevel0, srcLevel0)) return false;
		}
		return LoadLevel(0, *apSrcLevel0);
	}

	ImageMipmapChain  mipmapChain(GetSampleFormat());
	if(! mipmapChain.Build(srcLevel0, false)) return false;

	return LoadLevels(mipmapChain);
}


//====================================================================
bool  Texture2D::LoadLevels(ubyte val)
{
	AutoPtr<ImageBase>  apSrcLevel0(ManufactureImage(GetSampleFormat(), GetHeight(0), GetWidth(0)));
	ProcessSamples_FillNBytes(*apSrcLevel0, val);

	return LoadLevels(*apSrcLevel0);
}


//====================================================================
bool  Texture3D::s_bSourceLayerOrderBackToFront = false;
bool  Texture3D::s_bSourceRowOrderBottomToTop = false;


//====================================================================
Texture3D::Texture3D(eSampleFormat sampleFormat, bool bMultiLevel,
						uint _depthLevel0, uint _heightLevel0, uint _widthLevel0)
	:
	TextureBase(ResourceType_Texture3D, sampleFormat,
				bMultiLevel, Max(_depthLevel0, _heightLevel0, _widthLevel0))
{
	depthLevel0  = _depthLevel0;
	heightLevel0 = _heightLevel0;
	widthLevel0  = _widthLevel0;
}


//====================================================================
bool  Texture3D::LoadLevel(uint level, const VolumeBase& src,
							uint dataDepth, uint dataHeight, uint dataWidth,
							uint dstOffsetZ, uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetZ, uint srcOffsetY, uint srcOffsetX)
{
	if(! dataDepth) dataDepth = src.GetNumLayers();
	if(! dataHeight) dataHeight = src.GetNumRows();
	if(! dataWidth) dataWidth = src.GetNumColumns();

	VW_ASSERT(level < GetNumLevels(), "Level-index out of range.");
	VW_ASSERT(GetSampleFormat() == src.GetSampleFormat(), "Sample-format mismatch.");
	VW_ASSERT(src.GetNumSamples(), "Source empty.");
	VW_ASSERT(dataDepth && dataHeight && dataWidth, "Data-range empty.");
	VW_ASSERT(dstOffsetZ + dataDepth <= GetDepth(level),
				"Destination-offset-z / data-depth out of range.");
	VW_ASSERT(dstOffsetY + dataHeight <= GetHeight(level),
				"Destination-offset-y / data-height out of range.");
	VW_ASSERT(dstOffsetX + dataWidth <= GetWidth(level),
				"Destination-offset-x / data-width out of range.");
	VW_ASSERT(srcOffsetZ + dataDepth <= src.GetNumLayers(),
				"Source-offset-z / data-depth out of range.");
	VW_ASSERT(srcOffsetY + dataHeight <= src.GetNumRows(),
				"Source-offset-y / data-height out of range.");
	VW_ASSERT(srcOffsetX + dataWidth <= src.GetNumColumns(),
				"Source-offset-x / data-width out of range.");

	return
	DoLoadLevel(level, src,
				IsSourceLayerOrderBackToFront(), IsSourceRowOrderBottomToTop(),
				dataDepth, dataHeight, dataWidth,
				dstOffsetZ, dstOffsetY, dstOffsetX,
				srcOffsetZ, srcOffsetY, srcOffsetX);
}


//====================================================================
bool  Texture3D::LoadLevels(const VolumeMipmapChain& src)
{
	VW_ASSERT(src.GetSampleFormat() == GetSampleFormat(), "Sample-format mismatch.");
	VW_ASSERT(src.GetNumLevels() == GetNumLevels(), "Level-count mismatch.");
	VW_ASSERT(src.GetLevel(0).GetNumLayers() == GetDepth(0) &&
			src.GetLevel(0).GetNumRows() == GetHeight(0) &&
			src.GetLevel(0).GetNumColumns() == GetWidth(0), "Level-0 dimension mismatch.");

	for(uint level = 0; level < src.GetNumLevels(); level++) {
		const VolumeBase& srcLevel = src.GetLevel(level);

		if(! DoLoadLevel(level, srcLevel,
				IsSourceLayerOrderBackToFront(), IsSourceRowOrderBottomToTop(),
				srcLevel.GetNumLayers(), srcLevel.GetNumRows(), srcLevel.GetNumColumns(),
				0, 0, 0, 0, 0, 0)) return false;
	}
	return true;
}


//====================================================================
bool  Texture3D::LoadLevels(const VolumeBase& srcLevel0)
{
	VW_ASSERT(srcLevel0.GetNumLayers() == GetDepth(0) &&
			srcLevel0.GetNumRows() == GetHeight(0) &&
			srcLevel0.GetNumColumns() == GetWidth(0), "Level-0 dimension mismatch.");

	if(GetNumLevels() == 1) {
		AutoPtr<VolumeBase>  apSrcLevel0(const_cast<VolumeBase*>(&srcLevel0), false);
		if(GetSampleFormat() != srcLevel0.GetSampleFormat()) {
			apSrcLevel0.Assign(ManufactureVolume(GetSampleFormat()));
			if(! TransferSamples3D(*apSrcLevel0, srcLevel0)) return false;
		}
		return LoadLevel(0, *apSrcLevel0);
	}

	VolumeMipmapChain  mipmapChain(GetSampleFormat());
	if(! mipmapChain.Build(srcLevel0, false)) return false;

	return LoadLevels(mipmapChain);
}


//====================================================================
bool  Texture3D::LoadLevels(ubyte val)
{
	AutoPtr<VolumeBase>  apSrcLevel0(ManufactureVolume(GetSampleFormat(),
								GetDepth(0), GetHeight(0), GetWidth(0)));
	ProcessSamples_FillNBytes(*apSrcLevel0, val);

	return LoadLevels(*apSrcLevel0);
}


//********************************************************************
}//namespace video_renderer_objects


//********************************************************************
}//namespace vertexware


