

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "VideoRendererObjects\VRO_Resources.h"
#include "BaseTools\Constructs.h"
#include "DataObjects\ImageBase.h"
#include "DataObjects\VolumeBase.h"
#include "DataObjects\MipmapChains.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace video_renderer_objects {


//====================================================================
// Level-0 contains highest resolution data.
// Dimensions are not checked by base-classes.
//====================================================================
class VW_COMMON_API TextureBase : public Resource
{
public:
	__forceinline static uint __fastcall ComputeDimensionAtLevel(uint level, uint dimLevel0) {
		return ClampedLow<uint>((dimLevel0 >> level), 1);
	}

	static bool __fastcall CheckDimension(uint dim, bool bMustBePowerOf2);


public:
	virtual ~TextureBase();

	eSampleFormat  GetSampleFormat() const { return sampleFormat; }

	uint  GetNumLevels() const { return numLevels; }

	bool  SetBaseLevel(uint newBaseLevel = 0);
	uint  GetBaseLevel() const { return baseLevel; }


protected:
	TextureBase(eResourceType resourceType, eSampleFormat sampleFormat,
				bool bMultiLevel, uint maxDimension);

	// return true on success
	virtual  bool  UpdateBaseLevel() = 0;


private:
	eSampleFormat  sampleFormat;
	uint           numLevels;
	uint           baseLevel;
};


//====================================================================
class VW_COMMON_API Texture1D : public TextureBase
{
public:
	__forceinline uint  GetWidth(uint level = 0) const {
		return ComputeDimensionAtLevel(level, widthLevel0);
	}


	//-------------------------------------------------
	// Load one level from signal.
	// Returns false on error.
	// Source and texture sample-format must match.
	// Source and texture-level dimensions must match.
	// Offsets and sizes must be in valid range.
	// If <dataWidth> is 0 full source-width is loaded.
	//-------------------------------------------------
	bool  LoadLevel(uint level, const SignalBase& src,
					uint dataWidth = 0,
					uint dstOffsetX = 0,
					uint srcOffsetX = 0);

	//------------------------------------------------
	// Load all levels from mipmap source.
	// Returns false on error.
	// Source and texture sample-format must match.
	// Source and texture level dimensions must match.
	//------------------------------------------------
	bool  LoadLevels(const SignalMipmapChain& src);

	//--------------------------------------------------
	// Load all levels from source for level0.
	// Returns false on error.
	// Error is returned if source-sample-format to
	// texture-sample-format conversion is undefined.
	// Source and texture-level-0 dimensions must match.
	//--------------------------------------------------
	bool  LoadLevels(const SignalBase& srcLevel0);

	//-----------------------------------
	// Load all levels with single-value.
	//-----------------------------------
	bool  LoadLevels(ubyte val = 255);


protected:
	Texture1D(eSampleFormat sampleFormat, bool bMultiLevel,
				uint widthLevel0);


	//--------------------------------------------------------
	// Parameters are checked by caller.
	// Source-sample-format matches the texture-sample-format.
	//--------------------------------------------------------
	virtual  bool  DoLoadLevel(uint level, const SignalBase& src,
							uint dataWidth,
							uint dstOffsetX,
							uint srcOffsetX) = 0;


private:
	uint  widthLevel0;
};


//====================================================================
class VW_COMMON_API Texture2D : public TextureBase
{
public:
	// default: false
	static void __fastcall SetSourceRowOrderBottomToTop(bool bBottomToTop) { s_bSourceRowOrderBottomToTop = bBottomToTop; }
	static bool __fastcall IsSourceRowOrderBottomToTop() { return s_bSourceRowOrderBottomToTop; }

	// row-order sentry
	class SourceRowOrderSentry {
	public:
		SourceRowOrderSentry(bool bBottomToTop) {
			bWasBottomToTop = IsSourceRowOrderBottomToTop();
			SetSourceRowOrderBottomToTop(bBottomToTop);
		}
		~SourceRowOrderSentry() {
			SetSourceRowOrderBottomToTop(bWasBottomToTop);
		}
	private:
		bool bWasBottomToTop;
	};


public:
	__forceinline uint  GetHeight(uint level = 0) const {
		return ComputeDimensionAtLevel(level, heightLevel0);
	}
	__forceinline uint  GetWidth(uint level = 0) const {
		return ComputeDimensionAtLevel(level, widthLevel0);
	}


	//---------------------------------------------------
	// Load one level from image.
	// Returns false on error.
	// Source and texture sample-format must match.
	// Source and texture-level dimensions must match.
	// Offsets and sizes must be in valid range.
	// If <dataHeight> is 0 full source-height is loaded.
	// If <dataWidth> is 0 full source-width is loaded.
	// Offsets always use top-left convention.
	//---------------------------------------------------
	bool  LoadLevel(uint level, const ImageBase& src,
					uint dataHeight = 0, uint dataWidth = 0,
					uint dstOffsetY = 0, uint dstOffsetX = 0,
					uint srcOffsetY = 0, uint srcOffsetX = 0);

	//------------------------------------------------
	// Load all levels from mipmap source.
	// Returns false on error.
	// Source and texture sample-format must match.
	// Source and texture level dimensions must match.
	//------------------------------------------------
	bool  LoadLevels(const ImageMipmapChain& src);

	//--------------------------------------------------
	// Load all levels from source for level0.
	// Returns false on error.
	// Error is returned if source-sample-format to
	// texture-sample-format conversion is undefined.
	// Source and texture-level-0 dimensions must match.
	//--------------------------------------------------
	bool  LoadLevels(const ImageBase& srcLevel0);

	//-----------------------------------
	// Load all levels with single-value.
	//-----------------------------------
	bool  LoadLevels(ubyte val = 255);


protected:
	Texture2D(eSampleFormat sampleFormat, bool bMultiLevel,
				uint heightLevel0, uint widthLevel0);


	//----------------------------------------------------------------
	// Parameters are checked by caller.
	// Source-sample-format matches the texture-sample-format.
	// <bSrcIsStoredBottomToTop> indicates that rows are ordered bottom-to-top.
	// Offsets always use top-left convention.
	//----------------------------------------------------------------
	virtual  bool  DoLoadLevel(uint level, const ImageBase& src,
							bool bSrcIsStoredBottomToTop,
							uint dataHeight, uint dataWidth,
							uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetY, uint srcOffsetX) = 0;


private:
	uint  heightLevel0;
	uint  widthLevel0;

	static  bool  s_bSourceRowOrderBottomToTop;
};


//====================================================================
class VW_COMMON_API Texture3D : public TextureBase
{
public:
	// default: false
	static void __fastcall SetSourceLayerOrderBackToFront(bool bBackToFront) { s_bSourceLayerOrderBackToFront = bBackToFront; }
	static bool __fastcall IsSourceLayerOrderBackToFront() { return s_bSourceLayerOrderBackToFront; }

	// layer-order sentry
	class SourceLayerOrderSentry {
	public:
		SourceLayerOrderSentry(bool bBackToFront) {
			bWasBackToFront = IsSourceLayerOrderBackToFront();
			SetSourceLayerOrderBackToFront(bBackToFront);
		}
		~SourceLayerOrderSentry() {
			SetSourceLayerOrderBackToFront(bWasBackToFront);
		}
	private:
		bool bWasBackToFront;
	};

	// default: false
	static void __fastcall SetSourceRowOrderBottomToTop(bool bBottomToTop) { s_bSourceRowOrderBottomToTop = bBottomToTop; }
	static bool __fastcall IsSourceRowOrderBottomToTop() { return s_bSourceRowOrderBottomToTop; }

	// row-order sentry
	class SourceRowOrderSentry {
	public:
		SourceRowOrderSentry(bool bBottomToTop) {
			bWasBottomToTop = IsSourceRowOrderBottomToTop();
			SetSourceRowOrderBottomToTop(bBottomToTop);
		}
		~SourceRowOrderSentry() {
			SetSourceRowOrderBottomToTop(bWasBottomToTop);
		}
	private:
		bool bWasBottomToTop;
	};


public:
	__forceinline uint  GetDepth(uint level = 0) const {
		return ComputeDimensionAtLevel(level, depthLevel0);
	}
	__forceinline uint  GetHeight(uint level = 0) const {
		return ComputeDimensionAtLevel(level, heightLevel0);
	}
	__forceinline uint  GetWidth(uint level = 0) const {
		return ComputeDimensionAtLevel(level, widthLevel0);
	}


	//---------------------------------------------------
	// Load one level from volume.
	// Returns false on error.
	// Source and texture sample-format must match.
	// Source and texture-level dimensions must match.
	// Offsets and sizes must be in valid range.
	// If <dataDepth> is 0 full source-depth is loaded.
	// If <dataHeight> is 0 full source-height is loaded.
	// If <dataWidth> is 0 full source-width is loaded.
	// Offsets always use front-top-left convention.
	//---------------------------------------------------
	bool  LoadLevel(uint level, const VolumeBase& src,
					uint dataDepth = 0, uint dataHeight = 0, uint dataWidth = 0,
					uint dstOffsetZ = 0, uint dstOffsetY = 0, uint dstOffsetX = 0,
					uint srcOffsetZ = 0, uint srcOffsetY = 0, uint srcOffsetX = 0);

	//------------------------------------------------
	// Load all levels from mipmap source.
	// Returns false on error.
	// Source and texture sample-format must match.
	// Source and texture level dimensions must match.
	//------------------------------------------------
	bool  LoadLevels(const VolumeMipmapChain& src);

	//--------------------------------------------------
	// Load all levels from source for level0.
	// Returns false on error.
	// Error is returned if source-sample-format to
	// texture-sample-format conversion is undefined.
	// Source and texture-level-0 dimensions must match.
	//--------------------------------------------------
	bool  LoadLevels(const VolumeBase& srcLevel0);

	//-----------------------------------
	// Load all levels with single-value.
	//-----------------------------------
	bool  LoadLevels(ubyte val = 255);


protected:
	Texture3D(eSampleFormat sampleFormat, bool bMultiLevel,
				uint depthLevel0, uint heightLevel0, uint widthLevel0);


	//----------------------------------------------------------------
	// Parameters are checked by caller.
	// Source-sample-format matches the texture-sample-format.
	// <bSrcIsStoredBackToFront> indicates that layers are ordered back-to-front.
	// <bSrcIsStoredBottomToTop> indicates that rows are ordered bottom-to-top.
	// Offsets always use front-top-left convention.
	//----------------------------------------------------------------
	virtual  bool  DoLoadLevel(uint level, const VolumeBase& src,
							bool bSrcIsStoredBackToFront, bool bSrcIsStoredBottomToTop,
							uint dataDepth, uint dataHeight, uint dataWidth,
							uint dstOffsetZ, uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetZ, uint srcOffsetY, uint srcOffsetX) = 0;


private:
	uint  depthLevel0;
	uint  heightLevel0;
	uint  widthLevel0;

	static  bool  s_bSourceLayerOrderBackToFront;
	static  bool  s_bSourceRowOrderBottomToTop;
};


//********************************************************************
}//namespace video_renderer_objects


//********************************************************************
}//namespace vertexware


