

//--------------------------------------------------------------------
#include "D3D8Tools\Textures.h"
#include "D3D8Tools\ErrorHandling.h"
#include "DataObjects\ImageOperations.h"
#include "BaseTools\Memory.h"
#include "BaseTools\WinUtils.h"
#include <sstream>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace d3d8_tools {


//====================================================================
uint __fastcall TextureStage::GetNumTextureStages(IDirect3DDevice8& device)
{
	D3DCAPS8 caps;
	if(device.GetDeviceCaps(&caps) != D3D_OK) return 1;
	return caps.MaxTextureBlendStages;
}


//====================================================================
bool __fastcall TextureBase::AreTextureCapsSupported(IDirect3DDevice8& device, DWORD flags)
{
	D3DCAPS8 caps;
	if(device.GetDeviceCaps(&caps) != D3D_OK) return false;
	return ((caps.TextureCaps & flags) == flags);
}


//====================================================================
bool __fastcall TextureBase::IsD3DInternalFormatSupported(D3DFORMAT internalFormat,
									IDirect3DDevice8& device, D3DRESOURCETYPE rType, DWORD usage)
{
	IDirect3D8*  pD3DObject = 0;
	if(device.GetDirect3D(&pD3DObject) != D3D_OK) return false;

	D3DDEVICE_CREATION_PARAMETERS  creationParams;
	if(device.GetCreationParameters(&creationParams) != D3D_OK) return false;

	D3DDISPLAYMODE  displayMode;
	if(device.GetDisplayMode(&displayMode) != D3D_OK) return false;

	HRESULT hr = pD3DObject->CheckDeviceFormat(creationParams.AdapterOrdinal,
								creationParams.DeviceType, displayMode.Format,
								usage, rType, internalFormat);
	win_utils::SafeRelease(pD3DObject);

	return (hr == D3D_OK);
}


//====================================================================
bool __fastcall TextureBase::IsSampleFormatSupported(eSampleFormat sampleFormat,
								IDirect3DDevice8& device, D3DRESOURCETYPE rType, DWORD usage)
{
	D3DFORMAT internalFormat = MapSampleFormatToD3DInternalFormat(sampleFormat);
	if(internalFormat == D3DFMT_UNKNOWN) return false;
	return IsD3DInternalFormatSupported(internalFormat, device, rType, usage);
}


//====================================================================
D3DFORMAT __fastcall TextureBase::MapSampleFormatToD3DInternalFormat(eSampleFormat sampleFormat)
{
	switch(sampleFormat) {
		case SampleFormat_L8: return D3DFMT_L8;
		case SampleFormat_A8: return D3DFMT_A8;
		case SampleFormat_A8L8: return D3DFMT_A8L8;
		case SampleFormat_A4R4G4B4: return D3DFMT_A4R4G4B4;
		case SampleFormat_A1R5G5B5: return D3DFMT_A1R5G5B5;
		case SampleFormat_X8R8G8B8: return D3DFMT_X8R8G8B8;
		case SampleFormat_A8R8G8B8: return D3DFMT_A8R8G8B8;
	}
	return D3DFMT_UNKNOWN;
}


//====================================================================
eSampleFormat __fastcall TextureBase::MapD3DInternalFormatToSampleFormat(D3DFORMAT internalFormat)
{
	switch(internalFormat) {
		case D3DFMT_L8: return SampleFormat_L8;
		case D3DFMT_A8: return SampleFormat_A8;
		case D3DFMT_A8L8: return SampleFormat_A8L8;
		case D3DFMT_A4R4G4B4: return SampleFormat_A4R4G4B4;
		case D3DFMT_A1R5G5B5: return SampleFormat_A1R5G5B5;
		case D3DFMT_X8R8G8B8: return SampleFormat_X8R8G8B8;
		case D3DFMT_A8R8G8B8: return SampleFormat_A8R8G8B8;
	}
	return SampleFormat_Invalid;
}


//====================================================================
TextureBase::TextureBase(D3DFORMAT _internalFormat)
{
	internalFormat  = _internalFormat;
	iD3DBaseTexture = 0;
}


//====================================================================
TextureBase::~TextureBase()
{
	win_utils::SafeRelease(iD3DBaseTexture);
}


//====================================================================
void  TextureBase::ValidateInternalFormat()
{
	if(GetInternalFormat() == 0) {
		VW_THROW(D3D8Exception, "Cannot map sample-format to D3D8-internal-format.");
	}
}


//====================================================================
void  TextureBase::RealizePriority(uint newPriority)
{
	VW_ASSERT(iD3DBaseTexture, "Internal error, ID3DBaseTexture-pointer not available.");
	iD3DBaseTexture->SetPriority(newPriority);
}


//====================================================================
bool  TextureBase::RealizeBaseLevel(uint newBaseLevel)
{
	VW_ASSERT(iD3DBaseTexture, "Internal error, ID3DBaseTexture-pointer not available.");
	iD3DBaseTexture->SetLOD(newBaseLevel);
	return true;
}


//====================================================================
bool __fastcall Texture1D::IsD3DInternalFormatSupported(D3DFORMAT internalFormat, IDirect3DDevice8& device)
{
	return TextureBase::IsD3DInternalFormatSupported(internalFormat, device, D3DRTYPE_TEXTURE, 0);
}


//====================================================================
bool __fastcall Texture1D::IsSampleFormatSupported(eSampleFormat sampleFormat, IDirect3DDevice8& device)
{
	return TextureBase::IsSampleFormatSupported(sampleFormat, device, D3DRTYPE_TEXTURE, 0);
}


//====================================================================
Texture1D::Texture1D(IDirect3DDevice8& device,
					eSampleFormat sampleFormat, bool bMultiLevel,
					uint widthLevel0)
	:
	video_renderer_objects::Texture1D(sampleFormat, bMultiLevel,
							widthLevel0),
	d3d8_tools::TextureBase(MapSampleFormatToD3DInternalFormat(sampleFormat))
{
	ValidateInternalFormat();
	Allocate(device);
}


//====================================================================
Texture1D::Texture1D(IDirect3DDevice8& device,
					D3DFORMAT internalFormat, bool bMultiLevel,
					uint widthLevel0)
	:
	video_renderer_objects::Texture1D(MapD3DInternalFormatToSampleFormat(internalFormat), bMultiLevel,
							widthLevel0),
	d3d8_tools::TextureBase(internalFormat)
{
	Allocate(device);
}


//====================================================================
void  Texture1D::Allocate(IDirect3DDevice8& device)
{
	if(! CheckDimension(GetWidth(0), true)) {
		VW_THROW(D3D8Exception, "Texture dimensions are 0 or not power-of-2.");
	}
	if(GetNumLevels() > 1  &&  ! AreTextureCapsSupported(device, D3DPTEXTURECAPS_MIPMAP)) {
		VW_THROW(D3D8Exception, "1D-mipmaps not supported by active Direct3D device.");
	}
	if(! IsD3DInternalFormatSupported(GetInternalFormat(), device)) {
		VW_THROW(D3D8Exception, "1D-texture-format not supported by active Direct3D device.");
	}

	// allocate memory resources for all mipmap levels
	pD3DTexture = 0;

	HRESULT hr = device.CreateTexture(GetWidth(0), 1,
							(GetNumLevels() > 1 ? 0 : 1), 0, GetInternalFormat(),
							D3DPOOL_MANAGED, &pD3DTexture);

	std::string  errDesc;
	if(D3DErrorCheck(hr, "IDirect3DDevice8::CreateTexture", "", &errDesc)) {

		std::ostringstream oss;
		oss << " (w=" << GetWidth(0) << ").";
		errDesc.append(oss.str());

		VW_THROW(D3D8Exception, errDesc);
	}
	VW_ASSERT(GetNumLevels() == pD3DTexture->GetLevelCount(),
					"Internal error, level-count mismatch.");

	SetD3DBaseTexture(pD3DTexture);
	UpdatePriority();
}


//====================================================================
bool  Texture1D::DoLoadLevel(uint level, const SignalBase& src,
							uint dataWidth, uint dstOffsetX, uint srcOffsetX)
{
	// parameters already checked by caller
	RECT* pLockRect = 0;
	RECT  lockRect;

	if(dataWidth != GetWidth(level)) {

		pLockRect = &lockRect;

		lockRect.left   = dstOffsetX;
		lockRect.right  = dstOffsetX + dataWidth;
		lockRect.top    = 0;
		lockRect.bottom = 1;
	}

	// upload pixels to D3D-memory
	D3DLOCKED_RECT  lockDesc;
	if(pD3DTexture->LockRect(level, &lockDesc, pLockRect, 0) != D3D_OK) return false;

	::CopyMemory(lockDesc.pBits, src.SampleB(dstOffsetX), dataWidth * src.GetSampleByteSize());

	return (pD3DTexture->UnlockRect(level) == D3D_OK);
}


//====================================================================
bool __fastcall Texture2D::IsD3DInternalFormatSupported(D3DFORMAT internalFormat, IDirect3DDevice8& device)
{
	return TextureBase::IsD3DInternalFormatSupported(internalFormat, device, D3DRTYPE_TEXTURE, 0);
}


//====================================================================
bool __fastcall Texture2D::IsSampleFormatSupported(eSampleFormat sampleFormat, IDirect3DDevice8& device)
{
	return TextureBase::IsSampleFormatSupported(sampleFormat, device, D3DRTYPE_TEXTURE, 0);
}


//====================================================================
Texture2D::Texture2D(IDirect3DDevice8& device,
					eSampleFormat sampleFormat, bool bMultiLevel,
					uint heightLevel0, uint widthLevel0)
	:
	video_renderer_objects::Texture2D(sampleFormat, bMultiLevel,
							heightLevel0, widthLevel0),
	d3d8_tools::TextureBase(MapSampleFormatToD3DInternalFormat(sampleFormat))
{
	ValidateInternalFormat();
	Allocate(device);
}


//====================================================================
Texture2D::Texture2D(IDirect3DDevice8& device,
					D3DFORMAT internalFormat, bool bMultiLevel,
					uint heightLevel0, uint widthLevel0)
	:
	video_renderer_objects::Texture2D(MapD3DInternalFormatToSampleFormat(internalFormat), bMultiLevel,
							heightLevel0, widthLevel0),
	d3d8_tools::TextureBase(internalFormat)
{
	Allocate(device);
}


//====================================================================
void  Texture2D::Allocate(IDirect3DDevice8& device)
{
	if(! CheckDimension(GetHeight(0), true)  ||  ! CheckDimension(GetWidth(0), true)) {
		VW_THROW(D3D8Exception, "Texture dimensions are 0 or not power-of-2.");
	}
	if(GetNumLevels() > 1  &&  ! AreTextureCapsSupported(device, D3DPTEXTURECAPS_MIPMAP)) {
		VW_THROW(D3D8Exception, "2D-mipmaps not supported by active Direct3D device.");
	}
	if(! IsD3DInternalFormatSupported(GetInternalFormat(), device)) {
		VW_THROW(D3D8Exception, "2D-texture-format not supported by active Direct3D device.");
	}

	// allocate memory resources for all mipmap levels
	pD3DTexture = 0;

	HRESULT hr = device.CreateTexture(GetWidth(0), GetHeight(0),
							(GetNumLevels() > 1 ? 0 : 1), 0, GetInternalFormat(),
							D3DPOOL_MANAGED, &pD3DTexture);

	std::string  errDesc;
	if(D3DErrorCheck(hr, "IDirect3DDevice8::CreateTexture", "", &errDesc)) {

		std::ostringstream oss;
		oss << " (h=" << GetHeight(0) << " w=" << GetWidth(0) << ").";
		errDesc.append(oss.str());

		VW_THROW(D3D8Exception, errDesc);
	}
	VW_ASSERT(GetNumLevels() == pD3DTexture->GetLevelCount(),
					"Internal error, level-count mismatch.");

	SetD3DBaseTexture(pD3DTexture);
	UpdatePriority();
}


//====================================================================
bool  Texture2D::DoLoadLevel(uint level, const ImageBase& src,
							bool bSrcIsStoredBottomToTop,
							uint dataHeight, uint dataWidth,
							uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetY, uint srcOffsetX)
{
	// parameters already checked by caller
	int  srcRowStride = src.GetRowByteSize();

	if(bSrcIsStoredBottomToTop) {

		// Source-image is stored bottom-to-top.
		// Source-row-offsets are specified in top-to-bottom convention.
		srcOffsetY = src.GetNumRows() - 1 - srcOffsetY;
		FlipSign(srcRowStride);
	}

	RECT* pLockRect = 0;
	RECT  lockRect;

	if(dataHeight != GetHeight(level)  ||  dataWidth != GetWidth(level)) {

		pLockRect = &lockRect;

		lockRect.left   = dstOffsetX;
		lockRect.right  = dstOffsetX + dataWidth;
		lockRect.top    = dstOffsetY;
		lockRect.bottom = dstOffsetY + dataHeight;
	}

	// upload pixels to D3D-memory
	D3DLOCKED_RECT  lockDesc;
	if(pD3DTexture->LockRect(level, &lockDesc, pLockRect, 0) != D3D_OK) return false;

	ubyte*       pDstRow        = static_cast<ubyte*>(lockDesc.pBits);
	ubyte*       pDstRowEnd     = pDstRow + lockDesc.Pitch * dataHeight;
	const ubyte* pSrcRow        = src.SampleB(srcOffsetY, srcOffsetX);
	uint         numBytesPerRow = src.GetSampleByteSize() * dataWidth;

	while(pDstRow != pDstRowEnd) {
		::CopyMemory(PostIncrement(pDstRow, lockDesc.Pitch),
					PostIncrement(pSrcRow, srcRowStride), numBytesPerRow);
	}

	return (pD3DTexture->UnlockRect(level) == D3D_OK);
}


//====================================================================
bool __fastcall Texture3D::IsD3DInternalFormatSupported(D3DFORMAT internalFormat, IDirect3DDevice8& device)
{
	return TextureBase::IsD3DInternalFormatSupported(internalFormat, device, D3DRTYPE_VOLUMETEXTURE, 0);
}


//====================================================================
bool __fastcall Texture3D::IsSampleFormatSupported(eSampleFormat sampleFormat, IDirect3DDevice8& device)
{
	return TextureBase::IsSampleFormatSupported(sampleFormat, device, D3DRTYPE_VOLUMETEXTURE, 0);
}


//====================================================================
Texture3D::Texture3D(IDirect3DDevice8& device,
					eSampleFormat sampleFormat, bool bMultiLevel,
					uint depthLevel0, uint heightLevel0, uint widthLevel0)
	:
	video_renderer_objects::Texture3D(sampleFormat, bMultiLevel,
							depthLevel0, heightLevel0, widthLevel0),
	d3d8_tools::TextureBase(MapSampleFormatToD3DInternalFormat(sampleFormat))
{
	ValidateInternalFormat();
	Allocate(device);
}


//====================================================================
Texture3D::Texture3D(IDirect3DDevice8& device,
					D3DFORMAT internalFormat, bool bMultiLevel,
					uint depthLevel0, uint heightLevel0, uint widthLevel0)
	:
	video_renderer_objects::Texture3D(MapD3DInternalFormatToSampleFormat(internalFormat), bMultiLevel,
							depthLevel0, heightLevel0, widthLevel0),
	d3d8_tools::TextureBase(internalFormat)
{
	Allocate(device);
}


//====================================================================
void  Texture3D::Allocate(IDirect3DDevice8& device)
{
	if(! CheckDimension(GetDepth(0), true)  ||  ! CheckDimension(GetHeight(0), true)  ||
		! CheckDimension(GetWidth(0), true))
	{
		VW_THROW(D3D8Exception, "Texture dimensions are 0 or not power-of-2.");
	}
	if(! AreTextureCapsSupported(device, D3DPTEXTURECAPS_VOLUMEMAP)) {
		VW_THROW(D3D8Exception, "3D-textures not supported by active Direct3D device.");
	}
	if(GetNumLevels() > 1  &&  ! AreTextureCapsSupported(device, D3DPTEXTURECAPS_MIPVOLUMEMAP)) {
		VW_THROW(D3D8Exception, "3D-mipmaps not supported by active Direct3D device.");
	}
	if(! IsD3DInternalFormatSupported(GetInternalFormat(), device)) {
		VW_THROW(D3D8Exception, "3D-texture-format not supported by active Direct3D device.");
	}

	// allocate memory resources for all mipmap levels
	pD3DVolumeTexture = 0;

	HRESULT hr = device.CreateVolumeTexture(GetWidth(0), GetHeight(0), GetDepth(0),
							(GetNumLevels() > 1 ? 0 : 1), 0, GetInternalFormat(),
							D3DPOOL_MANAGED, &pD3DVolumeTexture);

	std::string  errDesc;
	if(D3DErrorCheck(hr, "IDirect3DDevice8::CreateVolumeTexture", "", &errDesc)) {

		std::ostringstream oss;
		oss << " (d=" << GetDepth(0) << " h=" << GetHeight(0) << " w=" << GetWidth(0) << ").";
		errDesc.append(oss.str());

		VW_THROW(D3D8Exception, errDesc);
	}
	VW_ASSERT(GetNumLevels() == pD3DVolumeTexture->GetLevelCount(),
					"Internal error, level-count mismatch.");

	SetD3DBaseTexture(pD3DVolumeTexture);
	UpdatePriority();
}


//====================================================================
bool  Texture3D::DoLoadLevel(uint level, const VolumeBase& src,
							bool bSrcIsStoredBackToFront, bool bSrcIsStoredBottomToTop,
							uint dataDepth, uint dataHeight, uint dataWidth,
							uint dstOffsetZ, uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetZ, uint srcOffsetY, uint srcOffsetX)
{
	// parameters already checked by caller
	int  srcLayerStride = src.GetLayerByteSize();
	int  srcRowStride   = src.GetRowByteSize();

	if(bSrcIsStoredBackToFront) {

		// Source-volume is stored back-to-front.
		// Source-layer-offsets are specified in front-to-back convention.
		srcOffsetZ = src.GetNumLayers() - 1 - srcOffsetZ;
		FlipSign(srcLayerStride);
	}

	if(bSrcIsStoredBottomToTop) {

		// Source-volume is stored bottom-to-top.
		// Source-row-offsets are specified in top-to-bottom convention.
		srcOffsetY = src.GetNumRows() - 1 - srcOffsetY;
		FlipSign(srcRowStride);
	}

	D3DBOX* pLockBox = 0;
	D3DBOX  lockBox;

	if(dataDepth != GetDepth(level)  ||  dataHeight != GetHeight(level)  ||
		dataWidth != GetWidth(level))
	{
		pLockBox = &lockBox;

		lockBox.Left   = dstOffsetX;
		lockBox.Right  = dstOffsetX + dataWidth;
		lockBox.Top    = dstOffsetY;
		lockBox.Bottom = dstOffsetY + dataHeight;
		lockBox.Front  = dstOffsetZ;
		lockBox.Back   = dstOffsetZ + dataDepth;
	}

	// upload pixels to D3D-memory
	D3DLOCKED_BOX  lockDesc;
	if(pD3DVolumeTexture->LockBox(level, &lockDesc, pLockBox, 0) != D3D_OK) return false;

	ubyte*       pDstLayer      = static_cast<ubyte*>(lockDesc.pBits);
	ubyte*       pDstLayerEnd   = pDstLayer + lockDesc.SlicePitch * dataDepth;
	int          dstSkipRows    = lockDesc.RowPitch * dataHeight;
	const ubyte* pSrcLayer      = src.SampleB(srcOffsetZ, srcOffsetY, srcOffsetX);
	uint         numBytesPerRow = src.GetSampleByteSize() * dataWidth;

	while(pDstLayer != pDstLayerEnd) {

		ubyte*       pDstRow    = pDstLayer;
		ubyte*       pDstRowEnd = pDstRow + dstSkipRows;
		const ubyte* pSrcRow    = pSrcLayer;

		while(pDstRow != pDstRowEnd) {
			::CopyMemory(PostIncrement(pDstRow, lockDesc.RowPitch),
						PostIncrement(pSrcRow, srcRowStride), numBytesPerRow);
		}

		pDstLayer += lockDesc.SlicePitch;
		pSrcLayer += srcLayerStride;
	}

	return (pD3DVolumeTexture->UnlockBox(level) == D3D_OK);
}


//********************************************************************
}//namespace d3d8_tools


//********************************************************************
}//namespace vertexware


