

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Types.h"
#include "BaseTools\NoCopy.h"
#include "VideoRendererObjects\VRO_Textures.h"
#include "private\vw_direct3d8_exports.h"
#include <d3d8.h>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace d3d8_tools {


//====================================================================
class VW_DIRECT3D8_API TextureStage
{
public:
	static uint __fastcall GetNumTextureStages(IDirect3DDevice8& device);
};


//====================================================================
class VW_DIRECT3D8_API TextureBase : public NoCopy
{
public:
	static bool __fastcall AreTextureCapsSupported(IDirect3DDevice8& device, DWORD flags);

	static bool __fastcall IsD3DInternalFormatSupported(D3DFORMAT internalFormat,
								IDirect3DDevice8& device, D3DRESOURCETYPE rType, DWORD usage);

	static bool __fastcall IsSampleFormatSupported(eSampleFormat sampleFormat,
								IDirect3DDevice8& device, D3DRESOURCETYPE rType, DWORD usage);

	// returns D3DFMT_UNKNOWN if mapping is undefined
	static D3DFORMAT __fastcall MapSampleFormatToD3DInternalFormat(eSampleFormat sampleFormat);

	// returns SampleFormat_Invalid if mapping is undefined
	static eSampleFormat __fastcall MapD3DInternalFormatToSampleFormat(D3DFORMAT internalFormat);


public:
	virtual ~TextureBase();

	D3DFORMAT               GetInternalFormat() const { return internalFormat; }
	IDirect3DBaseTexture8*  GetD3DBaseTexture() const { return iD3DBaseTexture; }


protected:
	TextureBase(D3DFORMAT internalFormat);

	// throws exception (should be called from ctor)
	void  ValidateInternalFormat();

	// TextureBase will release the given interface in dtor if it's != 0. To override
	// this behavior, call this method from derived-dtor with argument set to 0.
	void  SetD3DBaseTexture(IDirect3DBaseTexture8* iBaseTexture = 0) { iD3DBaseTexture = iBaseTexture; }

	void  RealizePriority(uint newPriority);
	bool  RealizeBaseLevel(uint newBaseLevel);


private:
	D3DFORMAT               internalFormat;
	IDirect3DBaseTexture8*  iD3DBaseTexture;
};


//====================================================================
class VW_DIRECT3D8_API Texture1D : public video_renderer_objects::Texture1D,
									public d3d8_tools::TextureBase
{
public:
	static bool __fastcall IsD3DInternalFormatSupported(D3DFORMAT internalFormat,
													IDirect3DDevice8& device);

	static bool __fastcall IsSampleFormatSupported(eSampleFormat sampleFormat,
													IDirect3DDevice8& device);


public:
	// sample-format mapping to D3D-internal-format is enforced
	Texture1D(IDirect3DDevice8& device,
				eSampleFormat sampleFormat, bool bMultiLevel,
				uint widthLevel0);

	// D3D-internal-format mapping to sample-format is NOT enforced
	Texture1D(IDirect3DDevice8& device,
				D3DFORMAT internalFormat, bool bMultiLevel,
				uint widthLevel0);


	IDirect3DTexture8*  GetD3DTexture() const { return pD3DTexture; }


protected:
	//--------------------------------------
	// video_renderer_objects::Texture1D implementation.
	//--------------------------------------
	virtual  void  UpdatePriority() { RealizePriority(GetPriority()); }
	virtual  bool  UpdateBaseLevel() { return RealizeBaseLevel(GetBaseLevel()); }

	virtual  bool  DoLoadLevel(uint level, const SignalBase& src,
								uint dataWidth,
								uint dstOffsetX,
								uint srcOffsetX);


private:
	void  Allocate(IDirect3DDevice8& device);


private:
	IDirect3DTexture8*  pD3DTexture;
};


//====================================================================
class VW_DIRECT3D8_API Texture2D : public video_renderer_objects::Texture2D,
									public d3d8_tools::TextureBase
{
public:
	static bool __fastcall IsD3DInternalFormatSupported(D3DFORMAT internalFormat,
													IDirect3DDevice8& device);

	static bool __fastcall IsSampleFormatSupported(eSampleFormat sampleFormat,
													IDirect3DDevice8& device);


public:
	// sample-format mapping to D3D-internal-format is enforced
	Texture2D(IDirect3DDevice8& device,
				eSampleFormat sampleFormat, bool bMultiLevel,
				uint heightLevel0, uint widthLevel0);

	// D3D-internal-format mapping to sample-format is NOT enforced
	Texture2D(IDirect3DDevice8& device,
				D3DFORMAT internalFormat, bool bMultiLevel,
				uint heightLevel0, uint widthLevel0);


	IDirect3DTexture8*  GetD3DTexture() const { return pD3DTexture; }


protected:
	//--------------------------------------
	// video_renderer_objects::Texture2D implementation.
	//--------------------------------------
	virtual  void  UpdatePriority() { RealizePriority(GetPriority()); }
	virtual  bool  UpdateBaseLevel() { return RealizeBaseLevel(GetBaseLevel()); }

	virtual  bool  DoLoadLevel(uint level, const ImageBase& src,
								bool bSrcIsStoredBottomToTop,
								uint dataHeight, uint dataWidth,
								uint dstOffsetY, uint dstOffsetX,
								uint srcOffsetY, uint srcOffsetX);


private:
	void  Allocate(IDirect3DDevice8& device);


private:
	IDirect3DTexture8*  pD3DTexture;
};


//====================================================================
class VW_DIRECT3D8_API Texture3D : public video_renderer_objects::Texture3D,
									public d3d8_tools::TextureBase
{
public:
	static bool __fastcall IsD3DInternalFormatSupported(D3DFORMAT internalFormat,
													IDirect3DDevice8& device);

	static bool __fastcall IsSampleFormatSupported(eSampleFormat sampleFormat,
													IDirect3DDevice8& device);


public:
	// sample-format mapping to D3D-internal-format is enforced
	Texture3D(IDirect3DDevice8& device,
				eSampleFormat sampleFormat, bool bMultiLevel,
				uint depthLevel0, uint heightLevel0, uint widthLevel0);

	// D3D-internal-format mapping to sample-format is NOT enforced
	Texture3D(IDirect3DDevice8& device,
				D3DFORMAT internalFormat, bool bMultiLevel,
				uint depthLevel0, uint heightLevel0, uint widthLevel0);


	IDirect3DVolumeTexture8*  GetD3DVolumeTexture() const { return pD3DVolumeTexture; }


protected:
	//--------------------------------------
	// video_renderer_objects::Texture3D implementation.
	//--------------------------------------
	virtual  void  UpdatePriority() { RealizePriority(GetPriority()); }
	virtual  bool  UpdateBaseLevel() { return RealizeBaseLevel(GetBaseLevel()); }

	virtual  bool  DoLoadLevel(uint level, const VolumeBase& src,
							bool bSrcIsStoredBackToFront, bool bSrcIsStoredBottomToTop,
							uint dataDepth, uint dataHeight, uint dataWidth,
							uint dstOffsetZ, uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetZ, uint srcOffsetY, uint srcOffsetX);


private:
	void  Allocate(IDirect3DDevice8& device);


private:
	IDirect3DVolumeTexture8*  pD3DVolumeTexture;
};


//********************************************************************
}//namespace d3d8_tools


//********************************************************************
}//namespace vertexware


