

//--------------------------------------------------------------------
#include "Framework\GUIRendererDirect3D8.h"
#include "BaseTools\Debug.h"
#include "MathTools\Vector3.h"
#include "MathTools\TransformTools.h"
#include "D3D8Tools\Textures.h"
#include "D3D8Tools\Transforms.h"
#include "D3D8Tools\ErrorHandling.h"


//********************************************************************
namespace vertexware {


//====================================================================
#define  DEFAULT_PIXEL_OFFSET_X                       -0.5f
#define  DEFAULT_PIXEL_OFFSET_Y                       -0.5f
#define  APP_POOL_VERTEX_BUFFER_MIN_SIZE             100
#define  APP_POOL_VERTEX_BUFFER_MIN_STABLE_CYCLES    100


//====================================================================
#define  ASSERT_INSIDE_BEGIN_END \
	VW_ASSERT(bInsideBeginEnd, "Invalid call outside gui::Renderer::Begin/End.");


//====================================================================
#define  ASSERT_INSIDE_BEGIN_END_INTERLEAVED \
	VW_ASSERT(bInsideBeginEndInterleaved, "Invalid call outside gui::Renderer::BeginInterleaved/EndInterleaved.");


//====================================================================
#define  ASSERT_OUTSIDE_BEGIN_END \
	VW_ASSERT(! bInsideBeginEnd, "Invalid call inside gui::Renderer::Begin/End.");


//====================================================================
#define  ASSERT_OUTSIDE_BEGIN_END_INTERLEAVED \
	VW_ASSERT(! bInsideBeginEndInterleaved, "Invalid call inside gui::Renderer::BeginInterleaved/EndInterleaved.");


//====================================================================
#define  VERTEX_SHADER_FVF_C   (D3DFVF_XYZ | D3DFVF_DIFFUSE)
#define  VERTEX_SHADER_FVF_CT  (D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0))


//====================================================================
GUIRendererDirect3D8::GUIRendererDirect3D8(VideoRendererDirect3D8& vrParent)
	:
	gui::Renderer(vrParent),
	appPoolVertexBuffer(APP_POOL_VERTEX_BUFFER_MIN_SIZE, APP_POOL_VERTEX_BUFFER_MIN_STABLE_CYCLES)
{
	pD3DVideoSubsystem = 0;
	pdev = 0;

	bInsideBeginEnd = false;
	bInsideBeginEndInterleaved = false;
	bTexturingEnabled = false;
	bAlphaTestEnabled = false;

	InvalidateLastTexParams();

	pixelParams.offsetX = DEFAULT_PIXEL_OFFSET_X;
	pixelParams.offsetY = DEFAULT_PIXEL_OFFSET_Y;

	depthParamsBase = DepthParams();
	depthParams = depthParamsBase;
	matTransform = mat4::IDENTITY;
	ConstructInternalTransform();
}


//====================================================================
GUIRendererDirect3D8::~GUIRendererDirect3D8()
{
}


//====================================================================
void  GUIRendererDirect3D8::ConnectVideoSubsystem(VideoSubsystem& vs)
{
	pD3DVideoSubsystem = InterfaceCast<VideoSubsystemDirect3D8>(&vs);
	if(! pD3DVideoSubsystem) {
		VW_THROW(VideoRendererException,
				"Wrong video-subsystem connected to: " + VW_GET_CLASS_NAME(*this));
	}
	pdev = pD3DVideoSubsystem->GetD3DDevice();

	// cache capabilities
	numTextureStages = 1;
	bSupportsWBuffering = false;
	bSupportsMipFilter = false;

	D3DCAPS8 caps;
	if(pdev->GetDeviceCaps(&caps) == D3D_OK) {

		numTextureStages = caps.MaxTextureBlendStages;
		bSupportsWBuffering = ToBool(caps.RasterCaps & D3DPRASTERCAPS_WBUFFER);

		bSupportsMipFilter =
			ToBool(caps.TextureFilterCaps & D3DPTFILTERCAPS_MIPFPOINT)  &&
			ToBool(caps.TextureFilterCaps & D3DPTFILTERCAPS_MIPFLINEAR);
	}
}


//====================================================================
void  GUIRendererDirect3D8::DisconnectVideoSubsystem()
{
	pD3DVideoSubsystem = 0;
	pdev = 0;
}


//====================================================================
void  GUIRendererDirect3D8::CreateVideoResources(bool bPostReset)
{
	apSB_Begin.Assign(new d3d8_tools::StateBlock(*pdev));
	apSB_End.Assign(new d3d8_tools::StateBlock(*pdev));

	CompileStateBlock_BeginEnd(apSB_Begin);
	CompileStateBlock_BeginEnd(apSB_End);
}


//====================================================================
void  GUIRendererDirect3D8::ReleaseVideoResources(bool bPreReset)
{
	apSB_Begin.SafeDelete();
	apSB_End.SafeDelete();
}


//====================================================================
void  GUIRendererDirect3D8::OnWindowResize(uint width, uint height)
{
	Renderer::OnWindowResize(width, height);

	ConstructInternalTransform();
	CompileStateBlock_BeginEnd(apSB_Begin);
}


//====================================================================
void  GUIRendererDirect3D8::GetRendererParams(RendererParams& rp) const
{
	rp.bTextureRowOrderBottomToTop = false;
	rp.bTextureLayerOrderBackToFront = false;
}


//====================================================================
bool  GUIRendererDirect3D8::SetViewport(const ViewportParams& vp)
{
	ASSERT_OUTSIDE_BEGIN_END;

	Stats().numSetViewport++;

	// Viewport must be contained in window, otherwise D3D8 fails
	gui::Rectangle  rViewport(vp.topLeft, vp.size);
	gui::Rectangle  rWindow(vec2::ZERO, GetWindowSize());

	if(! rWindow.Contains(rViewport)) return false;

	HRESULT hr = d3d8_tools::Viewport(*pdev,
						vp.topLeft.x, vp.topLeft.y, vp.size.x, vp.size.y,
						vp.renderDepthNear, vp.renderDepthFar);

	return (hr == D3D_OK);
}


//====================================================================
void  GUIRendererDirect3D8::ClearBuffers(const ClearParams& cp)
{
	ASSERT_OUTSIDE_BEGIN_END;

	Stats().numClearBuffers++;

	DWORD clearFlags = 0;

	// check available buffers to prevent D3D8-method failure
	const FrameBufferConfig& fbc = GetParentVideoRenderer().GetVideoConfig().frameBufferConfig;

	if(cp.bClearColor) clearFlags |= D3DCLEAR_TARGET;
	if(cp.bClearDepth  &&  fbc.bitsDepth) clearFlags |= D3DCLEAR_ZBUFFER;
	if(cp.bClearStencil  &&  fbc.bitsStencil) clearFlags |= D3DCLEAR_STENCIL;
	if(! clearFlags) return;

	float depth = Clamped<float>(cp.depth, 0.0f, 1.0f);

	if(cp.bUseRectangle) {
		// rectangle clear
		const gui::Rectangle& r = cp.rectangle;

		D3DRECT clearRect;

		clearRect.x1 = r.left;
		clearRect.y1 = r.top;
		clearRect.x2 = clearRect.x1 + r.width;
		clearRect.y2 = clearRect.y1 + r.height;

		ClampHigh<LONG>(clearRect.x2, GetWindowWidth());
		ClampHigh<LONG>(clearRect.y2, GetWindowHeight());

		pdev->Clear(1, &clearRect, clearFlags, cp.color, depth, cp.stencil);
	}
	else {
		// full-window clear
		d3d8_tools::Viewport(*pdev, 0, 0, GetWindowWidth(), GetWindowHeight(), 0, 1);
		pdev->Clear(0, 0, clearFlags, cp.color, depth, cp.stencil);
	}
}


//====================================================================
void  GUIRendererDirect3D8::SetPixelParams(const PixelParams& pp)
{
	ASSERT_OUTSIDE_BEGIN_END;

	pixelParams = pp;

	ConstructInternalTransform();
	CompileStateBlock_BeginEnd(apSB_Begin);
}


//====================================================================
void  GUIRendererDirect3D8::GetPixelParams(PixelParams& pp) const
{
	pp = pixelParams;
}


//====================================================================
void  GUIRendererDirect3D8::SetBaseDepthParams(const DepthParams& dp)
{
	ASSERT_OUTSIDE_BEGIN_END;

	depthParamsBase = dp;
	CompileStateBlock_BeginEnd(apSB_Begin);
}


//====================================================================
void  GUIRendererDirect3D8::GetBaseDepthParams(DepthParams& dp) const
{
	dp = depthParamsBase;
}


//====================================================================
bool  GUIRendererDirect3D8::QueryCreatesManagedResourcesOnly()
{
	return true;
}


//====================================================================
video_renderer_objects::Texture2D*  GUIRendererDirect3D8::CreateResourceTexture2D(
							eSampleFormat sampleFormat, bool bMultiLevel, uint height, uint width)
{
	ASSERT_OUTSIDE_BEGIN_END;

	switch(sampleFormat) {
		case SampleFormat_X8R8G8B8:
		case SampleFormat_A8R8G8B8:
			// throws exception on error
			return new d3d8_tools::Texture2D(*pdev, sampleFormat, bMultiLevel, height, width);
	}
	return 0;
}


//====================================================================
void  GUIRendererDirect3D8::Begin()
{
	ASSERT_OUTSIDE_BEGIN_END;
	ASSERT_OUTSIDE_BEGIN_END_INTERLEAVED;

	Stats().numBegin++;

	apSB_End->Capture();
	apSB_Begin->Apply();
	pdev->BeginScene();

	bInsideBeginEnd = true;
	bTexturingEnabled = false;
	bAlphaTestEnabled = false;

	depthParams = depthParamsBase;

	InvalidateLastTexParams();

	appPoolVertexBuffer.BeginReserveCycle();
}


//====================================================================
void  GUIRendererDirect3D8::End()
{
	ASSERT_OUTSIDE_BEGIN_END_INTERLEAVED;
	ASSERT_INSIDE_BEGIN_END;

	Stats().numEnd++;

	pdev->SetTexture(0, 0);
	pdev->EndScene();
	apSB_End->Apply();

	bInsideBeginEnd = false;
	bTexturingEnabled = false;
	bAlphaTestEnabled = false;

	InvalidateLastTexParams();

	appPoolVertexBuffer.EndReserveCycle();
}


//====================================================================
void  GUIRendererDirect3D8::BeginInterleaved()
{
	ASSERT_INSIDE_BEGIN_END;

	Stats().numBeginInterleaved++;

	End();
	bInsideBeginEndInterleaved = true;
}


//====================================================================
void  GUIRendererDirect3D8::EndInterleaved()
{
	ASSERT_INSIDE_BEGIN_END_INTERLEAVED;

	Stats().numEndInterleaved++;

	bInsideBeginEndInterleaved = false;
	Begin();
}


//====================================================================
void  GUIRendererDirect3D8::SetDepthParams(const DepthParams& dp)
{
	ASSERT_INSIDE_BEGIN_END;

	Stats().numSetDepthParams++;

	depthParams = dp;
	Send_DepthParams(depthParams);
}


//====================================================================
void  GUIRendererDirect3D8::GetDepthParams(DepthParams& dp) const
{
	dp = depthParams;
}


//====================================================================
void  GUIRendererDirect3D8::SetTransform(const mat4& m)
{
	ASSERT_INSIDE_BEGIN_END;

	Stats().numSetTransform++;

	matTransform = m;
	Send_Transform();
}


//====================================================================
void  GUIRendererDirect3D8::GetTransform(mat4& m) const
{
	m = matTransform;
}


//====================================================================
void  GUIRendererDirect3D8::DrawPrimitive(const gui::VertexBufferC& vertexBuffer,
									gui::ePrimitiveType primitiveType,
									uint numPrimitives, uint startVertex)
{
	Stats().numDrawPrimitiveColored++;
	Stats().numPrimitivesColored += numPrimitives;

	ASSERT_INSIDE_BEGIN_END;

	Send_EnableAlphaTest(false);
	Send_DisableTexturing();

	uint vertexCount;
	D3DPRIMITIVETYPE primType = TranslatePrimitiveParams(primitiveType, numPrimitives, vertexCount);

	VW_ASSERT(vertexBuffer.size() >= (startVertex + vertexCount), "Insufficient gui-vertex-buffer size.");

	// send vertices
	pdev->DrawPrimitiveUP(primType, numPrimitives,
						vertexBuffer + startVertex, sizeof(gui::VertexC));
}


//====================================================================
void  GUIRendererDirect3D8::DrawPrimitive(const gui::Texture& texture,
									const gui::VertexBufferCT& vertexBuffer,
									gui::ePrimitiveType primitiveType,
									uint numPrimitives, uint startVertex)
{
	Stats().numDrawPrimitiveTextured++;
	Stats().numPrimitivesTextured += numPrimitives;

	ASSERT_INSIDE_BEGIN_END;

	Send_EnableAlphaTest(false);
	Send_BindTextureAndEnableTexturing(texture);

	uint vertexCount;
	D3DPRIMITIVETYPE primType = TranslatePrimitiveParams(primitiveType, numPrimitives, vertexCount);

	VW_ASSERT(vertexBuffer.size() >= (startVertex + vertexCount), "Insufficient gui-vertex-buffer size.");

	// send vertices
	pdev->DrawPrimitiveUP(primType, numPrimitives,
						vertexBuffer + startVertex, sizeof(gui::VertexCT));
}


//====================================================================
void  GUIRendererDirect3D8::DrawText(const gui::FontTexture& fontTexture,
									uint textLen, const char* pText,
									const gui::DrawTextParams& dtp)
{
	Stats().numDrawText++;
	Stats().numPrimitivesTextured += textLen * 2;
	Stats().numTextGlyphs += textLen;

	ASSERT_INSIDE_BEGIN_END;
	if(! textLen) return;

	Send_EnableAlphaTest(true);
	Send_BindTextureAndEnableTexturing(fontTexture);

	// construct triangle list
	uint vertexCount = textLen * 6;
	appPoolVertexBuffer.ReserveSpace(vertexCount);

	gui::VertexCT* pVertex = appPoolVertexBuffer;

	const  gui::TextMetrics& tm = dtp.tm;

	bool   bForceAsciiRange = (fontTexture.GetNumGlyphs() < 256);
	float  spacing = fontTexture.CalcSpacingWidth(tm);
	float  gWidth = tm.bMonospace ? fontTexture.CalcGlyphWidth(tm, 0) : fontTexture.CalcGlyphWidthMultiplier(tm);

	vec3  posTopLeft = dtp.positionTopLeft;
	vec3  posBottomLeft(posTopLeft.x, posTopLeft.y + tm.textHeight, posTopLeft.z);
	vec3  posTopRight = posTopLeft;
	vec3  posBottomRight = posBottomLeft;

	for(uint i = 0; i < textLen; i++) {
		char ch = pText[i];

		// force 0-127 character range
		if(bForceAsciiRange  &&  ch < 0) ch = 0;

		// compute glyph's font-texture coords
		float  texLeft, texRight, texTop, texBottom;

		if(tm.bMonospace) {
			fontTexture.GetGlyphCoordsMono(ch, texLeft, texRight, texTop, texBottom);
		}
		else {
			fontTexture.GetGlyphCoords(ch, texLeft, texRight, texTop, texBottom);
		}

		// compute glyph's right-side positions
		posTopRight.x = posTopLeft.x + (tm.bMonospace ? gWidth : (gWidth * fontTexture.GetGlyphWidth(ch)));
		posBottomRight.x = posTopRight.x;

		// left-triangle: top-left point
		pVertex->position = posTopLeft;
		pVertex->color = dtp.color;
		pVertex->texCoord.Set(texLeft, texTop);
		pVertex++;

		// left-triangle: bottom-left point
		pVertex->position = posBottomLeft;
		pVertex->color = dtp.color;
		pVertex->texCoord.Set(texLeft, texBottom);
		pVertex++;

		// left-triangle: top-right point
		pVertex->position = posTopRight;
		pVertex->color = dtp.color;
		pVertex->texCoord.Set(texRight, texTop);
		pVertex++;

		// right-triangle: top-right point
		pVertex->position = posTopRight;
		pVertex->color = dtp.color;
		pVertex->texCoord.Set(texRight, texTop);
		pVertex++;

		// right-triangle: bottom-left point
		pVertex->position = posBottomLeft;
		pVertex->color = dtp.color;
		pVertex->texCoord.Set(texLeft, texBottom);
		pVertex++;

		// right-triangle: bottom-right point
		pVertex->position = posBottomRight;
		pVertex->color = dtp.color;
		pVertex->texCoord.Set(texRight, texBottom);
		pVertex++;

		// shift positions to next-glyph
		posTopLeft.x = posTopRight.x + spacing;
		posBottomLeft.x = posTopLeft.x;
	}

	// send vertices
	pdev->DrawPrimitiveUP(D3DPT_TRIANGLELIST, textLen * 2,
						appPoolVertexBuffer, sizeof(AppPoolVertexBuffer::VertexType));
}


//====================================================================
void  GUIRendererDirect3D8::CompileStateBlock_BeginEnd(
										AutoPtr<d3d8_tools::StateBlock>& apSB)
{
	if(! apSB) return;
	d3d8_tools::StateBlockCompileSentry  sbcs(*apSB);

	pdev->SetPixelShader(0);
	pdev->SetVertexShader(VERTEX_SHADER_FVF_C);
	pdev->SetRenderState(D3DRS_LIGHTING, FALSE);
	pdev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
	pdev->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
	pdev->SetRenderState(D3DRS_EDGEANTIALIAS, FALSE);
	pdev->SetRenderState(D3DRS_ZBIAS, 0);
	pdev->SetRenderState(D3DRS_DITHERENABLE, FALSE);
	pdev->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD);
	pdev->SetRenderState(D3DRS_FOGENABLE, FALSE);
	pdev->SetRenderState(D3DRS_ALPHATESTENABLE, FALSE);
	pdev->SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_GREATER);
	d3d8_tools::SetRenderState_AlphaRef(*pdev, 0);
	pdev->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
	pdev->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
	pdev->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
	pdev->SetRenderState(D3DRS_WRAP0, 0);

	pdev->SetTexture(0, 0);
	pdev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);
	if(numTextureStages >= 2) {
		pdev->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
	}
	pdev->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
	pdev->SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
	pdev->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
	pdev->SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_CLAMP);
	pdev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	pdev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	pdev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);

	pdev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	pdev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT);
	pdev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
	pdev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_CURRENT);

	pdev->SetTransform(D3DTS_VIEW, d3d8_tools::ReinterpretCast(&mat4::IDENTITY));
	Send_Transform();
	Send_DepthParams(depthParamsBase);
}


//====================================================================
void  GUIRendererDirect3D8::Send_DepthParams(const DepthParams& dp)
{
	mat4 projection;
	GenerateProjectionOrthoVM(projection, false, true,
							0, GetWindowWidth(), 0, GetWindowHeight(),
							dp.clippingPlaneNear, dp.clippingPlaneFar);

	pdev->SetTransform(D3DTS_PROJECTION, d3d8_tools::ReinterpretCast(&projection));

	d3d8_tools::Viewport(*pdev, 0, 0, GetWindowWidth(), GetWindowHeight(),
						dp.renderDepthNear, dp.renderDepthFar);

	// z-buffer writes work only if D3DRS_ZENABLE is TRUE
	pdev->SetRenderState(D3DRS_ZENABLE,
				(dp.bEnableDepthWrite || dp.bEnableDepthTest) ?
				((bSupportsWBuffering && dp.bUseWBuffering) ? D3DZB_USEW : D3DZB_TRUE)
				: D3DZB_FALSE);

	pdev->SetRenderState(D3DRS_ZFUNC, dp.bEnableDepthTest ? D3DCMP_LESSEQUAL : D3DCMP_ALWAYS);
	pdev->SetRenderState(D3DRS_ZWRITEENABLE, dp.bEnableDepthWrite ? TRUE : FALSE);
}


//====================================================================
void  GUIRendererDirect3D8::Send_EnableAlphaTest(bool bEnable)
{
	if(bAlphaTestEnabled != bEnable) {
		bAlphaTestEnabled = bEnable;

		d3d8_tools::SetRenderState_Enable(*pdev, D3DRS_ALPHATESTENABLE, bEnable);
	}
}


//====================================================================
void  GUIRendererDirect3D8::Send_DisableTexturing()
{
	if(bTexturingEnabled) {
		bTexturingEnabled = false;

		InvalidateLastTexParams();

		pdev->SetVertexShader(VERTEX_SHADER_FVF_C);
		pdev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);
	}
}


//====================================================================
void  GUIRendererDirect3D8::Send_BindTextureAndEnableTexturing(const gui::Texture& texture)
{
	// enable texturing
	if(! bTexturingEnabled) {
		bTexturingEnabled = true;

		pdev->SetVertexShader(VERTEX_SHADER_FVF_CT);
	}

	// bind texture
	if(pLastBoundTexture != &texture) {
		pLastBoundTexture = &texture;

		const d3d8_tools::Texture2D* pResource =
			static_cast<const d3d8_tools::Texture2D*>(texture.GetVideoRendererResource());

		VW_ASSERT(pResource, "Using gui::Texture without created video-renderer-resource.");
		if(pResource) {
			pdev->SetTexture(0, pResource->GetD3DBaseTexture());
		}

		Stats().numTextureChanges++;
	}

	// setup component operations
	if(lastTexSampleFormat != texture.GetSampleFormat()) {
		lastTexSampleFormat = texture.GetSampleFormat();

		switch(texture.GetSampleFormat()) {
			case SampleFormat_X8R8G8B8:
				pdev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
				pdev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2);
				break;

			case SampleFormat_A8R8G8B8:
				pdev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
				pdev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
				break;

			default:
				pdev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);
				break;
		}
	}

	// setup mipmap operations
	if(lastTexMipFilter != texture.GetMipFilter()) {
		lastTexMipFilter =
			(bSupportsMipFilter  &&  texture.IsMultiLevel()) ?
			texture.GetMipFilter() : gui::Texture::Filter_Off;

		D3DTEXTUREFILTERTYPE filter = D3DTEXF_NONE;

		switch(lastTexMipFilter) {
			case gui::Texture::Filter_Nearest:
				filter = D3DTEXF_POINT;
				break;

			case gui::Texture::Filter_Linear:
				filter = D3DTEXF_LINEAR;
				break;
		}
		pdev->SetTextureStageState(0, D3DTSS_MIPFILTER, filter);
	}
}


//====================================================================
void  GUIRendererDirect3D8::ConstructInternalTransform()
{
	mat4  matTemp;

	matInternalTransform.SetIdentity();

	// apply pixel-offset
	GenerateTranslateVM(matTemp, pixelParams.offsetX, pixelParams.offsetY, 0);
	matInternalTransform *= matTemp;

	// convert coords from Screen-RHS to D3D-LHS
	GenerateNonUniformScale(matTemp, 1, -1, 1);
	matInternalTransform *= matTemp;

	GenerateTranslateVM(matTemp, 0, GetWindowHeight(), 0);
	matInternalTransform *= matTemp;
}


//====================================================================
D3DPRIMITIVETYPE __fastcall GUIRendererDirect3D8::TranslatePrimitiveParams(gui::ePrimitiveType pt,
												uint numPrimitives, uint& outVertexCount)
{
	VW_ASSERT(numPrimitives > 0, "Primitive-count must be > 0.");

	switch(pt) {
		case gui::PrimitiveType_TriangleList:
			outVertexCount = numPrimitives * 3;
			return D3DPT_TRIANGLELIST;

		case gui::PrimitiveType_TriangleStrip:
			outVertexCount = numPrimitives + 2;
			return D3DPT_TRIANGLESTRIP;

		case gui::PrimitiveType_TriangleFan:
			outVertexCount = numPrimitives + 2;
			return D3DPT_TRIANGLEFAN;
	}
	VW_ASSERT(0, "Unsupported gui-primitive-type.");
	return D3DPRIMITIVETYPE(0);
}


//********************************************************************
}//namespace vertexware


