

//--------------------------------------------------------------------
#include "GUI_CS0\ComboBox.h"
#include "GUI_CS0\ShaderNames.h"
#include "GUI\GUI_Manager.h"
#include "BaseTools\STLExtensions.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
ComboBoxItem::ComboBoxItem(const std::string& _text, int _id)
	:
	text(_text)
{
	id = _id;
}


//====================================================================
ComboBoxPaneButton::ComboBoxPaneButton(ComboBoxPane& _comboBoxPane)
	:
	PressButton(&_comboBoxPane),
	comboBoxPane(_comboBoxPane)
{
	bClickedBeforeAutoScroll = false;
	bFirstAutoScrollEmit = true;
	secsAccumTime = 0;
}


//====================================================================
bool  ComboBoxPaneButton::OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos)
{
	PressButton::OnMouseEnterExit(bEnter, bMouseOverDirect, mousePos);

	bClickedBeforeAutoScroll = false;
	bFirstAutoScrollEmit = true;
	secsAccumTime = 0;
	return true;
}


//====================================================================
void  ComboBoxPaneButton::OnTimeDelta(double secs)
{
	if(IsHot()) {
		if(bClickedBeforeAutoScroll) return;

		double autoScrollDelta = comboBoxPane.comboBox.GetAutoScrollDelay();
		if(autoScrollDelta < 0) return;

		if(! bFirstAutoScrollEmit) {
			autoScrollDelta = comboBoxPane.comboBox.GetAutoScrollRepeat();
		}
		secsAccumTime += secs;

		if(secsAccumTime >= autoScrollDelta) {
			bFirstAutoScrollEmit = false;
			secsAccumTime = 0;
			comboBoxPane.ScrollItems(*this);
		}
	}
}


//====================================================================
void  ComboBoxPaneButton::ProcessButtonPress()
{
	if(bFirstAutoScrollEmit) {
		bClickedBeforeAutoScroll = true;
	}
	bFirstAutoScrollEmit = true;
	secsAccumTime = 0;
	comboBoxPane.ScrollItems(*this);
}


//====================================================================
ComboBoxPane::ComboBoxPane(ComboBox& _comboBox)
	:
	comboBox(_comboBox)
{
	SetOverlay(true);

	pButtonForward = new ComboBoxPaneButton(*this);
	pButtonForward->SetVisible(false);

	pButtonBackward = new ComboBoxPaneButton(*this);
	pButtonBackward->SetVisible(false);

	AttachTo(comboBox);
}


//====================================================================
bool  ComboBoxPane::GetVisibleItemRange(int& first, int& last) const
{
	first = firstVisibleItem;
	last = lastVisibleItem;
	return (first <= last);
}


//====================================================================
void  ComboBoxPane::CalcVisibleItemRange()
{
	firstVisibleItem = 0;

	if(comboBox.GetNumItems() > comboBox.GetGeomParams().maxVisibleItems) {
		int iSel = comboBox.GetSelectedItemIndex();
		int half = comboBox.GetGeomParams().maxVisibleItems / 2;

		firstVisibleItem = ClampedLow<int>(iSel - half, 0);
	}
	UpdateVisibleItemRange(0);
}


//====================================================================
void  ComboBoxPane::UpdateVisibleItemRange(int scroll)
{
	int numVisible = Min(comboBox.GetGeomParams().maxVisibleItems, comboBox.GetNumItems());

	firstVisibleItem += scroll;
	Clamp<int>(firstVisibleItem, 0, comboBox.GetNumItems() - numVisible);

	lastVisibleItem = firstVisibleItem + numVisible - 1;

	bool bNeedsBackward = (firstVisibleItem > 0);
	pButtonBackward->SetEnabled(bNeedsBackward);

	bool bNeedsForward = (lastVisibleItem < comboBox.GetNumItems() - 1);
	pButtonForward->SetEnabled(bNeedsForward);
}


//====================================================================
void  ComboBoxPane::ScrollItems(const ComboBoxPaneButton& button)
{
	if(&button == pButtonBackward) {
		UpdateVisibleItemRange(-1);
	}
	else if(&button == pButtonForward) {
		UpdateVisibleItemRange(1);
	}
}


//====================================================================
int  ComboBoxPane::CalcHotItem(const vec2& mousePos)
{
	if(! comboBox.GetNumItems()) return -1;
	float itemY = mousePos.y - (pButtonBackward->IsVisible() ? pButtonBackward->GetSize().y : 0);
	int   iHot = itemY / comboBox.GetGeomParams().itemHeightPane;
	Clamp<int>(iHot, 0, comboBox.GetNumItems() - 1);
	return iHot;
}


//====================================================================
void  ComboBoxPane::RecalcSize()
{
	const ComboBox::GeomParams& gp = comboBox.GetGeomParams();

	vec2  buttonSize(comboBox.GetSize().x, gp.paneButtonHeight);

	pButtonBackward->SetSize(buttonSize);
	pButtonForward->SetSize(buttonSize);

	float  totalItemHeight = gp.itemHeightPane * Min(comboBox.GetNumItems(), gp.maxVisibleItems);
	float  totalHeight = totalItemHeight + 2 * buttonSize.y;

	if(! comboBox.GetNumItems()) {
		pButtonBackward->SetVisible(false);
		pButtonForward->SetVisible(false);

		SetSize(buttonSize.x, gp.itemHeightPane * (gp.panePosition == ComboBox::PanePosition_Center ? 1.5f : 0.5f));
	}
	else if(comboBox.GetNumItems() > gp.maxVisibleItems) {
		pButtonBackward->SetVisible(true);
		pButtonForward->SetVisible(true);

		SetSize(buttonSize.x, totalHeight);
	}
	else {
		pButtonBackward->SetVisible(false);
		pButtonForward->SetVisible(false);

		SetSize(buttonSize.x, totalItemHeight);
	}
	UpdateVisibleItemRange(0);
}


//====================================================================
void  ComboBoxPane::RecalcLayout()
{
	pButtonBackward->SetPosition(vec2::ZERO);
	pButtonForward->SetPosition(0, GetSize().y - pButtonForward->GetSize().y);
}


//====================================================================
void  ComboBoxPane::UpdateGeometryCache()
{
	CalcVisibleItemRange();
}


//====================================================================
void  ComboBoxPane::DrawSelf(gui::Renderer& gr)
{
	if(comboBox.pComboBoxShader) {
		comboBox.pComboBoxShader->Shade(*this, gr);
	}
}


//====================================================================
bool  ComboBoxPane::OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos)
{
	if(bEnter  &&  bMouseOverDirect) {
		hotItem = CalcHotItem(mousePos);
	}
	else {
		hotItem = -1;
	}
	return true;
}


//====================================================================
bool  ComboBoxPane::OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos)
{
	hotItem = CalcHotItem(mousePos);
	return true;
}


//====================================================================
bool  ComboBoxPane::OnMouseWheel(int delta, const vec2& mousePos)
{
	UpdateVisibleItemRange(-delta);
	return true;
}


//====================================================================
bool  ComboBoxPane::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(! bDown  ||  repeatCount) return false;

	comboBox.OpenPane(false);
	if(! comboBox.GetNumItems()) return true;

	int hotItem = CalcHotItem(mousePos);
	int newSelectedItem = GetHotItem();

	if(newSelectedItem >= 0) {
		comboBox.SelectItem(newSelectedItem);
		comboBox.SendCommandMessageToTarget();
	}
	return true;
}


//====================================================================
void  ComboBoxPane::OnChangedVisible()
{
	gui::Manager& gm = gui::Manager::Instance();
	bool bVisible = IsVisible();

	gm.TimeNotifyRegister(*pButtonForward, bVisible);
	gm.TimeNotifyRegister(*pButtonBackward, bVisible);
}


//====================================================================
ComboBoxButton::ComboBoxButton(ComboBox& _comboBox)
	:
	PushButton(&_comboBox),
	comboBox(_comboBox)
{
}


//====================================================================
void  ComboBoxButton::ProcessButtonRelease()
{
	comboBox.OpenPane(! comboBox.IsPaneOpened());
}


//====================================================================
ComboBox::GeomParams::GeomParams()
{
	maxVisibleItems = 5;
	panePosition = PanePosition_Bottom;
	preferredSize = vec2(100, 24);
	openButtonWidth = 24;
	bOpenButtonRightAlign = true;
	paneButtonHeight = 12;
	textHeightBox = 18;
	textPaddingBox = 4;
	textHeightPane = 18;
	textPaddingPane = 4;
	itemHeightPane = 24;
}


//====================================================================
ComboBox::ComboBox(gui::Component* _parent,
					gui::MessageHandler* _commandTarget,
					uint _commandID)
{
	pComboBoxShader = 0;

	pSelectedItem = 0;
	iSelectedItem = -1;
	bNeedsSelectedItemPopup = false;
	bShowSelectedItemPopup = false;

	SetCanDisable(true);
	SetCommandTarget(_commandTarget);
	SetCommandID(_commandID);

	pButtonOpen = new ComboBoxButton(*this);

	pPane = new ComboBoxPane(*this);
	pPane->SetVisible(false);

	SetGeomParams();
	SetAutoScrollDelay();
	SetAutoScrollRepeat();
	SetPopupHoverDelay();
	SetShaderName(DefaultShaderName(Shader_ComboBox));

	if(_parent) AttachTo(*_parent);
}


//====================================================================
ComboBox::~ComboBox()
{
	pSelectedItem = 0;
	DeletePtrContainer(items);
}


//====================================================================
void  ComboBox::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ClampLow<uint>(geomParams.maxVisibleItems, 1);

	ReportDirtyGeometry();
}


//====================================================================
void  ComboBox::OpenPane(bool bOpen)
{
	if(bOpen == IsPaneOpened()) return;
	pPane->SetVisible(bOpen);
}


//====================================================================
void  ComboBox::SelectItem(ComboBoxItem& cbi)
{
	if(! OwnsItem(cbi)) return;
	pSelectedItem = &cbi;
	UpdateSelectedItemCache();
}


//====================================================================
void  ComboBox::SelectItem(uint i)
{
	if(i < GetNumItems()) {
		SelectItem(GetItem(i));
	}
}


//====================================================================
void  ComboBox::SelectPreviousItem()
{
	int iSel = GetSelectedItemIndex();
	if(iSel < 0) return;
	int prevSel = iSel - 1;
	if(prevSel < 0) prevSel = GetNumItems() - 1;
	if(iSel != prevSel) {
		SelectItem(prevSel);
	}
}


//====================================================================
void  ComboBox::SelectNextItem()
{
	int iSel = GetSelectedItemIndex();
	if(iSel < 0) return;
	uint nextSel = (iSel + 1) % GetNumItems();
	if(iSel != nextSel) {
		SelectItem(nextSel);
	}
}


//====================================================================
void  ComboBox::AddItem(ComboBoxItem& cbi)
{
	OpenPane(false);
	if(OwnsItem(cbi)) return;

	UniqueContainerPushBack(items, &cbi);

	// auto-select first added
	if(GetNumItems() == 1) {
		SelectItem(cbi);
	}
}


//====================================================================
void  ComboBox::RemoveItem(ComboBoxItem& cbi)
{
	OpenPane(false);
	if(! OwnsItem(cbi)) return;

	EraseFromContainer(items, &cbi);

	// if selected-item is removed auto-select first available
	if(IsSelected(cbi)) {
		SelectItem(0);
	}
	if(! GetNumItems()) {
		pSelectedItem = 0;
		UpdateSelectedItemCache();
	}
}


//====================================================================
void  ComboBox::RemoveAllItems()
{
	OpenPane(false);

	pSelectedItem = 0;
	items.clear();

	UpdateSelectedItemCache();
}


//====================================================================
void  ComboBox::DeleteItem(ComboBoxItem& cbi)
{
	OpenPane(false);
	if(! OwnsItem(cbi)) return;

	DeleteFromPtrContainer(items, &cbi);

	// if selected-item is removed auto-select first available
	if(IsSelected(cbi)) {
		SelectItem(0);
	}
	if(! GetNumItems()) {
		pSelectedItem = 0;
		UpdateSelectedItemCache();
	}
}


//====================================================================
void  ComboBox::DeleteAllItems()
{
	OpenPane(false);

	pSelectedItem = 0;
	DeletePtrContainer(items);

	UpdateSelectedItemCache();
}


//====================================================================
bool  ComboBox::OwnsItem(const ComboBoxItem& cbi) const
{
	return (std::find(items.begin(), items.end(), &cbi) != items.end());
}


//====================================================================
void  ComboBox::RecalcSize()
{
	SetSize(geomParams.preferredSize);
}


//====================================================================
void  ComboBox::RecalcLayout()
{
	pButtonOpen->SetSize(geomParams.openButtonWidth, GetSize().y);
	pButtonOpen->SetPosition((geomParams.bOpenButtonRightAlign ? (GetSize().x - pButtonOpen->GetSize().x) : 0), 0);

	switch(geomParams.panePosition) {
		case PanePosition_Top:
			pPane->SetPosition(0, - pPane->GetSize().y);
			break;

		case PanePosition_Bottom:
			pPane->SetPosition(0, GetSize().y);
			break;

		case PanePosition_Center:
			pPane->SetPosition(0, -(pPane->GetSize().y - GetSize().y) * 0.5);
			break;
	}
}


//====================================================================
void  ComboBox::UpdateGeometryCache()
{
	maxWidthSelectedText = GetSize().x - pButtonOpen->GetSize().x - 2 * geomParams.textPaddingBox;

	UpdateSelectedItemCache();
	UpdatePaneItemCache();
}


//====================================================================
void  ComboBox::DrawSelf(gui::Renderer& gr)
{
	if(pComboBoxShader) {
		pComboBoxShader->Shade(*this, gr);
	}
}


//====================================================================
void  ComboBox::PostCacheShaderHandles()
{
	pComboBoxShader = gui::ConstShaderPtr<ComboBox::Shader>(GetBaseShader());

	Component::PostCacheShaderHandles();
}


//====================================================================
bool  ComboBox::OnDeactivate(bool bDeactivateByMouse)
{
	bShowSelectedItemPopup = false;
	OpenPane(false);
	return true;
}


//====================================================================
bool  ComboBox::OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos)
{
	bShowSelectedItemPopup = false;
	return true;
}


//====================================================================
bool  ComboBox::OnMouseMove(bool bHover, double hoverTimeSecs, const vec2& mousePos)
{
	if(bHover  &&  hoverTimeSecs >= popupHoverDelay) {
		bShowSelectedItemPopup = true;
	}
	return true;
}


//====================================================================
bool  ComboBox::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(! bDown  ||  repeatCount) return false;

	OpenPane(false);
	if(! GetNumItems()) return true;

	if(buttonCode == MouseButton_Left) {
		SelectNextItem();
		SendCommandMessageToTarget();
	}
	else if(buttonCode == MouseButton_Right) {
		SelectPreviousItem();
		SendCommandMessageToTarget();
	}
	return true;
}


//====================================================================
void  ComboBox::OnChangedEnabled()
{
	pButtonOpen->SetEnabled(IsEnabled());
}


//====================================================================
void  ComboBox::UpdateSelectedItemCache()
{
	// update index
	iSelectedItem = -1;
	if(pSelectedItem) {
		for(uint i = 0; i < GetNumItems(); i++) {
			if(pSelectedItem == &GetItem(i)) {
				iSelectedItem = i;
				break;
			}
		}
	}

	// update text
	selectedItemText.erase();
	selectedItemVisibleTextLength = 0;
	fullSelectedTextWidth = 0;
	bNeedsSelectedItemPopup = false;

	if(pSelectedItem) {
		selectedItemText = pSelectedItem->GetText();

		gui::ConstShaderPtr<ComboBox::Shader> pShader = GetBaseShader();
		if(! pShader) return;

		const gui::Font* pFont = pShader->GetFont();
		if(! pFont) return;

		fullSelectedTextWidth = pFont->CalcTextWidth(pShader->GetTextMetricsL(geomParams.textHeightBox), selectedItemText);

		float visWidth;
		pFont->CalcMaxVisibleText(selectedItemVisibleTextLength, visWidth, maxWidthSelectedText,
					pShader->GetTextMetricsL(geomParams.textHeightBox), selectedItemText);

		if(selectedItemText.length() > selectedItemVisibleTextLength) {
			bNeedsSelectedItemPopup = true;
		}
	}
}


//====================================================================
void  ComboBox::UpdatePaneItemCache()
{
	// auto-init to 0
	itemVisibleTextLengthsPane.resize(GetNumItems());
	itemFullTextWidthsPane.resize(GetNumItems());

	gui::ConstShaderPtr<ComboBox::Shader> pShader = GetBaseShader();
	if(! pShader) return;
	const gui::Font* pFont = pShader->GetFont();
	if(! pFont) return;

	float maxTextWidth = pPane->GetSize().x - 2 * geomParams.textPaddingPane;

	for(uint i = 0; i < items.size(); i++) {
		gui::TextMetricsL tml = pShader->GetTextMetricsL(geomParams.textHeightPane);

		float visWidth;
		pFont->CalcMaxVisibleText(itemVisibleTextLengthsPane[i], visWidth, maxTextWidth,
								tml, items[i]->GetText());

		itemFullTextWidthsPane[i] = pFont->CalcTextWidth(tml, items[i]->GetText());
	}
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


