

//--------------------------------------------------------------------
#include "GUI_CS0\Console.h"
#include "GUI\GUI_Manager.h"
#include "GUI\GUI_StdObjects.h"
#include "Framework\KernelMacros.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
Console::GeomParams::GeomParams()
{
	outputTextPaddingHorz = 4;
	outputTextPaddingTop = 4;
	outputTextPaddingBottom = 2;
	scrollbarWidth = 20;
	buttonConsoleControlsTextHeight = 20;
	buttonConsoleControlsPadding = 4;
}


//====================================================================
Console::DrawParams::DrawParams()
{
	outputTextTypeColors[OutputStream::OutputType_Info] = color4b::GREEN;
	outputTextTypeColors[OutputStream::OutputType_Warning] = color4b(255, 128, 0);
	outputTextTypeColors[OutputStream::OutputType_Error] = color4b::RED;
	outputTextTypeColors[OutputStream::OutputType_Attention] = color4b::YELLOW;
	outputTextTypeColors[OutputStream::OutputType_Query] = color4b::GREEN;
	outputTextTypeColors[OutputStream::OutputType_Echo] = color4b::WHITE;
	outputTextTypeColors[OutputStream::OutputType_Debug] = color4b::CYAN;

	textAtts.fontName = gui::std_fonts::FontName_Console;
	textAtts.textMetricsL.bMonospace = true;
	textAtts.textMetricsL.textHeight = 20;
}


//====================================================================
Console::Console(const std::string& _title)
	:
	TopLevelFrame(_title)
{
	editBoxObserver.pConsole = this;

	pOutputTextBgrShader = 0;
	pFont = 0;

	textHeight = 20;
	minTextHeight = 12;
	maxTextHeight = 32;
	sPrompt = ">";
	bHScrollAtTop = true;

	pCommandInterpreter = 0;
	pConsoleBuffers = 0;
	pTextSource = 0;
	bHookedToConsoleBuffers = false;
	iLastTotalNumCols = 0;
	iCommandHistoryIndex = 0;

	// Create components
	pButtonHelpInfo = new TextButton(this, this, CID_HelpInfo, gui::Button::Style_Push, 0, "?");
	pButtonHorzBarPos = new TextButton(this, this, CID_HorzBarPos, gui::Button::Style_Push, 0, "=");
	pButtonTextInc = new TextButton(this, this, CID_TextHeightInc, gui::Button::Style_Push, 0, ">");
	pButtonTextDec = new TextButton(this, this, CID_TextHeightDec, gui::Button::Style_Push, 0, "<");

	pEditBox = new EditBox(this, this, CID_Input);
	pEditBox->SetObserver(&editBoxObserver);
	pEditBox->SetShiftFocusToAncestor(false, false);
	pEditBox->SetSendCommand(true, false);
	pEditBox->SetResetCaretPositionOnKeyboardFocusLost(false);

	pScrollBarV = new ScrollBar(this, this, CID_ScrollVert, true);
	pScrollBarV->ActivateAutoEnableDisable(false);
	pScrollBarV->EnableMaxPixelsPerStep(false);

	pScrollBarH = new ScrollBar(this, this, CID_ScrollHorz, false);
	pScrollBarH->ActivateAutoEnableDisable(false);
	pScrollBarH->EnableMaxPixelsPerStep(false);

	ResetScrollBars();
	SetConsoleSubcomponentToolTipText();
	SetConsoleKeys();

	TopLevelFrame::GeomParams gpTLF;
	gpTLF.bResizable = true;
	gpTLF.minSize = vec2(300, 200);
	gpTLF.maxSize = vec2::INFINITY;
	TopLevelFrame::SetGeomParams(gpTLF);

	SetTextHeight();
	SetGeomParams();
	SetSize(400, 300);

	SetDrawParams();
	SetConsoleShaderNames();

	pScrollBarH->InstallMessageHook(*this);
	pScrollBarV->InstallMessageHook(*this);
	pEditBox->InstallMessageHook(*this);

	// auto-connect kernel-buffers
	ConnectConsoleBuffers(Kernel::Instance().GetConsoleBuffers());
}


//====================================================================
Console::~Console()
{
	// auto-disconnect buffers
	UnhookFromConsoleBuffers();
}


//====================================================================
void  Console::ConnectCommandInterpreter(CommandInterpreter& commandInterpreter)
{
	pCommandInterpreter = &commandInterpreter;
}


//====================================================================
void  Console::DisconnectCommandInterpreter()
{
	pCommandInterpreter = 0;
}


//====================================================================
void  Console::ConnectConsoleBuffers(ConsoleBuffers& consoleBuffers)
{
	UnhookFromConsoleBuffers();
	pConsoleBuffers = &consoleBuffers;
	HookToConsoleBuffers();
}


//====================================================================
void  Console::DisconnectConsoleBuffers()
{
	UnhookFromConsoleBuffers();
	pConsoleBuffers = 0;
}


//====================================================================
void  Console::SetConsoleShaderNames(const std::string& outputBgr,
									const std::string& textButton,
									const std::string& editBox,
									const std::string& scrollBar)
{
	shaderNameOutputBgr = outputBgr;
	pButtonHelpInfo->SetShaderName(textButton);
	pButtonHorzBarPos->SetShaderName(textButton);
	pButtonTextInc->SetShaderName(textButton);
	pButtonTextDec->SetShaderName(textButton);
	pEditBox->SetShaderName(editBox);
	pScrollBarV->SetShaderName(scrollBar);
	pScrollBarH->SetShaderName(scrollBar);
	RequestShaderRecache();
}


//====================================================================
void  Console::SetConsoleSubcomponentToolTipText(
								const std::string& buttonHelpInfo,
								const std::string& buttonHorzBarPosition,
								const std::string& buttonTextHeightInc,
								const std::string& buttonTextHeightDec)
{
	pButtonHelpInfo->SetToolTipText(buttonHelpInfo);
	pButtonHorzBarPos->SetToolTipText(buttonHorzBarPosition);
	pButtonTextInc->SetToolTipText(buttonTextHeightInc);
	pButtonTextDec->SetToolTipText(buttonTextHeightDec);
}


//====================================================================
void  Console::SetConsoleKeys(eKey _kcHelpInfo, eKey _kcHorzBarPos,
								eKey _kcTextHeightDec, eKey _kcTextHeightInc)
{
	kcHelpInfo = _kcHelpInfo;
	kcHorzBarPos = _kcHorzBarPos;
	kcTextHeightDec = _kcTextHeightDec;
	kcTextHeightInc = _kcTextHeightInc;
}


//====================================================================
void  Console::SetPrompt(const std::string& _sPrompt)
{
	sPrompt = _sPrompt;
}


//====================================================================
void  Console::SetHorzScrollBarAtTop(bool bAtTop)
{
	bHScrollAtTop = bAtTop;
	ReportDirtyGeometry();
}


//====================================================================
void  Console::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;

	TextButton::GeomParams  gpTextButtons;
	gpTextButtons.minSize = vec2::ZERO;
	gpTextButtons.textHeight = geomParams.buttonConsoleControlsTextHeight;
	gpTextButtons.padding = geomParams.buttonConsoleControlsPadding;

	pButtonHelpInfo->SetGeomParams(gpTextButtons);
	pButtonHorzBarPos->SetGeomParams(gpTextButtons);
	pButtonTextInc->SetGeomParams(gpTextButtons);
	pButtonTextDec->SetGeomParams(gpTextButtons);

	ScrollBar::GeomParams  gpScrollbar;
	gpScrollbar.buttonLength = geomParams.scrollbarWidth;
	gpScrollbar.preferredSize = vec2(gpScrollbar.buttonLength, 200);
	pScrollBarV->SetGeomParams(gpScrollbar);

	gpScrollbar.preferredSize = vec2(200, gpScrollbar.buttonLength);
	pScrollBarH->SetGeomParams(gpScrollbar);

	ReportDirtyGeometry();
}


//====================================================================
void  Console::SetDrawParams(const DrawParams& dp)
{
	drawParams = dp;
	drawParams.textAtts.textMetricsL.textHeight = textHeight;

	RequestShaderRecache();
	RequestFontRecache();

	ReportDirtyGeometry();
}


//====================================================================
void  Console::SetTextHeight(uint th)
{
	textHeight = Clamped<float>(th, minTextHeight, maxTextHeight);

	drawParams.textAtts.textMetricsL.textHeight = textHeight;

	EditBox::GeomParams  gpEditBox = pEditBox->GetGeomParams();
	float yDiff = gpEditBox.preferredSize.y - gpEditBox.textHeight;
	gpEditBox.textHeight = textHeight;
	gpEditBox.preferredSize.y = textHeight + yDiff;
	pEditBox->SetGeomParams(gpEditBox);

	ReportDirtyGeometry();
}


//====================================================================
void  Console::SetMinTextHeight(uint th)
{
	minTextHeight = ClampedLow<float>(th, 1);
	SetTextHeight(textHeight);
}


//====================================================================
void  Console::SetMaxTextHeight(uint th)
{
	maxTextHeight = ClampedLow<float>(th, minTextHeight);
	SetTextHeight(textHeight);
}


//====================================================================
bool  Console::CacheFontHandles(gui::Manager& gm)
{
	if(! TopLevelFrame::CacheFontHandles(gm)) return false;

	pFont = gm.GetFont(drawParams.textAtts.fontName);
	return (pFont != 0);
}


//====================================================================
void  Console::PostCacheFontHandles()
{
	ReportDirtyGeometry();
}


//====================================================================
void  Console::RecalcLayout()
{
	TopLevelFrame::RecalcLayout();

	vec2 size, tl, tr, bl, br;
	CalcTLFMainArea(size, tl, tr, bl, br);

	pEditBox->SetSizeX(size.x);
	pScrollBarV->SetSizeY(size.y - pEditBox->GetSize().y);
	pScrollBarH->SetSizeX(size.x - pScrollBarV->GetSize().x);

	pEditBox->SetPosition(bl.x, bl.y - pEditBox->GetSize().y);
	pScrollBarV->SetPosition(tr.x - pScrollBarV->GetSize().x, tr.y);

	if(bHScrollAtTop) {
		pScrollBarH->SetPosition(tl);
	}
	else {
		pScrollBarH->SetPosition(bl.x, pEditBox->GetPosition().y - pScrollBarH->GetSize().y);
	}
}


//====================================================================
void  Console::UpdateGeometryCache()
{
	TopLevelFrame::UpdateGeometryCache();

	UpdateOutputTextAreaParams();
}


//====================================================================
void  Console::OnChangedVisible()
{
	if(IsVisible()) {
		HookToConsoleBuffers();
	}
	else {
		UnhookFromConsoleBuffers();
	}
}


//====================================================================
void  Console::DrawSelf(gui::Renderer& gr)
{
	//------------
	// Draw frame.
	//------------
	TopLevelFrame::DrawSelf(gr);

	//-----------------------------
	// Draw output-text background.
	//-----------------------------
	if(pOutputTextBgrShader) {
		pOutputTextBgrShader->Shade(*this, gr);
	}

	//-------------------
	// Print output text.
	//-------------------
	if(! pFont) return;
	if(! pTextSource) return;

	pTextSource->Lock();

	gui::DrawTextParams  dtp(color4b::WHITE, vec2::ZERO, drawParams.textAtts.textMetricsL);
	float fAbsLineHeight = drawParams.textAtts.textMetricsL.AbsLineHeight();

	vec2  curPos(rTextArea.left + geomParams.outputTextPaddingHorz,
					rTextArea.top + geomParams.outputTextPaddingTop);

	uint  hPosition = pScrollBarH->GetScrollPosition();
	uint  numVisibleCols = pScrollBarH->GetVisibleScrollSize();

	uint lastLine = pScrollBarV->GetLastVisibleScrollPosition();
	for(uint i = pScrollBarV->GetScrollPosition(); i <= lastLine; i++)
	{
		// get one line
		uint outputType;
		const std::string& s = pTextSource->GetLine(i, outputType);

		if(s.length() > hPosition) {
			const char* firstCharToDraw = s.c_str() + hPosition;
			uint numCharsToDraw  = ClampedHigh<uint>(s.length() - hPosition, numVisibleCols);

			if(outputType >= OutputStream::OutputType_Count) outputType = OutputStream::OutputType_Info;
			dtp.color = drawParams.outputTextTypeColors[outputType];

			// draw one line
			dtp.positionTopLeft = curPos;
			gr.DrawText(*pFont, numCharsToDraw, firstCharToDraw, dtp);
		}
		curPos.y += fAbsLineHeight;
	}
	pTextSource->Unlock();
}


//====================================================================
bool  Console::CacheShaderHandles(gui::Manager& gm)
{
	// cache shader used by top-level-frame to draw it's part
	if(! TopLevelFrame::CacheShaderHandles(gm)) return false;

	// cache output-text-area shader
	pOutputTextBgrShader = gui::ConstShaderPtr<OutputBgrShader>(gm.GetShader(shaderNameOutputBgr));
	return (pOutputTextBgrShader != 0);
}


//====================================================================
bool  Console::OnCommand(uint commandID)
{
	switch(commandID) {
		case CID_Input: {
			FocusOnLastLine();
			const std::string& sCommand = pEditBox->GetText();
			olog << e_echo << sPrompt << sCommand << nl << e_info;
			if(sCommand.length()) {
				ResetCommandHistoryWalk();
				InterpretCommandLine(sCommand);
				pEditBox->SetText();
			}
			return true;
		}
		case CID_HelpInfo: {
			DumpConsoleHelp();
			DumpConsoleParams();
			return true;
		}
		case CID_HorzBarPos: {
			SetHorzScrollBarAtTop(! IsHorzScrollBarAtTop());
			return true;
		}
		case CID_TextHeightInc: {
			SetTextHeight(textHeight + 1);
			return true;
		}
		case CID_TextHeightDec: {
			SetTextHeight(textHeight - 1);
			return true;
		}
	}
	return TopLevelFrame::OnCommand(commandID);
}


//====================================================================
bool  Console::OnMouseWheel(int delta, const vec2& mousePos)
{
	ScrollBar* scrollTarget =
		gui::Manager::Instance().IsKeyModifierPressed(KeyModifier_Control) ?
		static_cast<ScrollBar*>(pScrollBarH) : static_cast<ScrollBar*>(pScrollBarV);

	return scrollTarget->DispatchMessage(gui::Message_MouseWheel(delta, mousePos));
}


//====================================================================
bool  Console::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	// switch to edit-box on double-click
	if(bDoubleClick  &&  buttonCode == MouseButton_Left) {
		gui::Manager::Instance().RequestKeyboardFocus(*pEditBox);
		return true;
	}
	return false;
}


//====================================================================
bool  Console::OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData)
{
	DefaultProcessQueryFocus(bFocusByMouse, buttonCode, bWantFocus, bNeedCharacterData);
	bNeedCharacterData = true;
	return true;
}


//====================================================================
bool  Console::OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch)
{
	if(! bDown) return false;
	gui::Manager& gm = gui::Manager::Instance();

	ScrollBar* scrollTarget = gm.IsKeyModifierPressed(KeyModifier_Control) ?
		static_cast<ScrollBar*>(pScrollBarV) : static_cast<ScrollBar*>(pScrollBarH);

	// process fixed keys
	switch(keyCode) {
		case Key_Tab:
			if(! repeatCount) {
				gm.RequestKeyboardFocus(*pEditBox);
			}
			return true;

		case Key_Enter:
		case Key_NumEnter:
			if(! repeatCount) {
				gm.RequestKeyboardFocus(*pEditBox);
			}
			return true;

		case Key_Home:
			if(! repeatCount) {
				SendMessageTo(*scrollTarget, gui::Message_Scroll(false, gui::ScrollDistance_ToEnd));
			}
			return true;

		case Key_End:
			if(! repeatCount) {
				SendMessageTo(*scrollTarget, gui::Message_Scroll(true, gui::ScrollDistance_ToEnd));
			}
			return true;

		case Key_PageUp:
			SendMessageTo(*pScrollBarV, gui::Message_Scroll(false, gui::ScrollDistance_Page));
			return true;

		case Key_PageDown:
			SendMessageTo(*pScrollBarV, gui::Message_Scroll(true, gui::ScrollDistance_Page));
			return true;

		case Key_ArrowUp:
			SendMessageTo(*pScrollBarV, gui::Message_Scroll(false,
							gm.IsKeyModifierPressed(KeyModifier_Control) ?
							gui::ScrollDistance_Page : gui::ScrollDistance_Line));
			return true;

		case Key_ArrowDown:
			SendMessageTo(*pScrollBarV, gui::Message_Scroll(true,
							gm.IsKeyModifierPressed(KeyModifier_Control) ?
							gui::ScrollDistance_Page : gui::ScrollDistance_Line));
			return true;

		case Key_ArrowLeft:
			SendMessageTo(*pScrollBarH, gui::Message_Scroll(false,
							gm.IsKeyModifierPressed(KeyModifier_Control) ?
							gui::ScrollDistance_Page : gui::ScrollDistance_Line));
			return true;

		case Key_ArrowRight:
			SendMessageTo(*pScrollBarH, gui::Message_Scroll(true,
							gm.IsKeyModifierPressed(KeyModifier_Control) ?
							gui::ScrollDistance_Page : gui::ScrollDistance_Line));
			return true;
	}

	// process configurable keys
	if(! repeatCount) {
		if(keyCode == kcHelpInfo) {
			SendMessageToSelf(gui::Message_Command(CID_HelpInfo));
			return true;
		}
		if(keyCode == kcHorzBarPos) {
			SendMessageToSelf(gui::Message_Command(CID_HorzBarPos));
			return true;
		}
		if(keyCode == kcTextHeightInc) {
			SendMessageToSelf(gui::Message_Command(CID_TextHeightInc));
			return true;
		}
		if(keyCode == kcTextHeightDec) {
			SendMessageToSelf(gui::Message_Command(CID_TextHeightDec));
			return true;
		}
	}

	// propagate to base-class
	if(TopLevelFrame::OnKeyboardKey(bDown, keyCode, repeatCount, ch)) {
		return true;
	}

	// auto-redirect focus to edit-box if character is typed
	if(ch != 0) {
		if(gm.RequestKeyboardFocus(*pEditBox)) {
			SendMessageTo(*pEditBox, gui::Message_KeyboardKey(keyCode, bDown, repeatCount, ch));
		}
		return true;
	}
	return false;
}


//====================================================================
bool  Console::OnKeyboardKey(gui::MessageHandler* target, bool bDown, eKey keyCode, uint repeatCount, char ch)
{
	if(target == pEditBox) {
		if(! bDown) return true;

		gui::Manager& gm = gui::Manager::Instance();

		switch(keyCode) {
			// erase input if any / otherwise change focus
			case Key_Escape: {
				if(repeatCount) return true;
				if(pEditBox->GetText().empty()) {
					RequestKeyboardFocus();
				}
				else {
					ResetCommandHistoryWalk();
					pEditBox->SetText("");
				}
				return true;
			}
			// complete command / switch focus to frame
			case Key_Tab: {
				if(repeatCount) return true;
				if(gm.IsKeyModifierPressed(KeyModifier_Control)) {
					RequestKeyboardFocus();
				}
				else {
					std::string  sPrefix = pEditBox->ExtractTextBeforeCaret();
					std::string  sCompleted;
					if(CompleteCommandLine(sPrefix, sCompleted)) {
						pEditBox->SetText(sCompleted);
						ResetCommandHistoryWalk();
					}
				}
				return true;
			}
			// get previous command from history
			case Key_ArrowUp: {
				std::string s;
				if(GetPreviousCommandFromHistory(s)) {
					pEditBox->SetText(s);
				}
				return true;
			}
			// get next command from history
			case Key_ArrowDown: {
				std::string s;
				if(GetNextCommandFromHistory(s)) {
					pEditBox->SetText(s);
				}
				return true;
			}
		}
		// break handler recursion
		return false;
	}

	// redirect other targets to standard handler
	return OnKeyboardKey(bDown, keyCode, repeatCount, ch);
}


//====================================================================
void  Console::OnTLFControlsLayouterUpdate()
{
	using namespace gui::layouters;
	RowLayouter& l = *pLayouterTLFControls;

	Row r;
	r.AutoSpacer().Target(pButtonTextDec).Target(pButtonTextInc);
	r.Target(pButtonHorzBarPos).Target(pButtonHelpInfo);
	r.Spacer(TopLevelFrame::GetGeomParams().buttonFrameControlsSpacing);
	l.LastRow().Remove(0).Prepend(r);
}


//====================================================================
void  Console::OnChanged(const ConsoleOutputBuffer& cob)
{
	bool  bFocusOnLastLine = pScrollBarV->IsScrollPositionAtMaximum();

	cob.Lock();
	pScrollBarH->SetTotalScrollSize(cob.GetMaxColumns());
	pScrollBarV->SetTotalScrollSize(cob.GetNumLines());
	cob.Unlock();

	if(bFocusOnLastLine) {
		pScrollBarV->SetScrollPositionToMax();
	}

	// show/hide horizontal bar
	if(iLastTotalNumCols != pScrollBarH->GetTotalScrollSize()) {
		iLastTotalNumCols = pScrollBarH->GetTotalScrollSize();

		UpdateOutputTextAreaParams();
	}
}


//====================================================================
void  Console::UpdateOutputTextAreaParams()
{
	// initial text-area size
	vec2 tl = CalcTLFMainAreaOffset();
	vec2 size = CalcTLFMainAreaSize();

	rTextArea.Set(tl.x,
			(bHScrollAtTop ? (pScrollBarH->GetPosition().y + pScrollBarH->GetSize().y) : tl.y),
			size.x - pScrollBarV->GetSize().x,
			size.y - pScrollBarH->GetSize().y - pEditBox->GetSize().y);

	if(! pFont) return;

	// compute new scroll params, hide horizontal bar if possible
	bool  bFocusOnLastLine = pScrollBarV->IsScrollPositionAtMaximum();

	float fTextAreaReduceX = geomParams.outputTextPaddingHorz * 2;
	float fTextAreaReduceY = geomParams.outputTextPaddingTop + geomParams.outputTextPaddingBottom;

	float fRowHeight = ClampedLow<float>(drawParams.textAtts.textMetricsL.AbsLineHeight(), 1);
	float fColWidth = ClampedLow<float>(pFont->CalcGlyphWidth(drawParams.textAtts.textMetricsL, 0), 1);

	int numCols = (rTextArea.width - fTextAreaReduceX) / fColWidth;
	int numRows = (rTextArea.height - fTextAreaReduceY) / fRowHeight;

	// update text-area size based on whether horizontal bar is necessary
	pScrollBarH->SetVisibleScrollSize(numCols);

	if(pScrollBarH->IsScrollSizeVisibleLessThanTotal()) {
		pScrollBarH->SetVisible(true);
	}
	else {
		pScrollBarH->SetVisible(false);

		if(bHScrollAtTop) {
			rTextArea.top -= pScrollBarH->GetSize().y;
		}
		rTextArea.height += pScrollBarH->GetSize().y;

		// recompute
		numRows = (rTextArea.height - fTextAreaReduceY) / fRowHeight;
	}
	pScrollBarV->SetVisibleScrollSize(numRows);

	// update vertical scroll position
	if(bFocusOnLastLine) {
		pScrollBarV->SetScrollPositionToMax();
	}
}


//====================================================================
void  Console::HookToConsoleBuffers()
{
	if(pConsoleBuffers) {
		if(! bHookedToConsoleBuffers) {
			bHookedToConsoleBuffers = true;
			iLastTotalNumCols = 0;

			pTextSource = &pConsoleBuffers->GetOutputBuffer();
			pTextSource->AddObserver(*this);

			ResetCommandHistoryWalk();
			UpdateOutputTextAreaParams();
		}
	}
}


//====================================================================
void  Console::UnhookFromConsoleBuffers()
{
	if(pConsoleBuffers) {
		if(bHookedToConsoleBuffers) {
			bHookedToConsoleBuffers = false;
			iLastTotalNumCols = 0;

			pTextSource->RemoveObserver(*this);
			pTextSource = 0;

			ResetCommandHistoryWalk();
			ResetScrollBars();
			UpdateOutputTextAreaParams();
		}
	}
}


//====================================================================
void  Console::ResetScrollBars()
{
	pScrollBarH->ResetLogicalScrollParameters();
	pScrollBarV->ResetLogicalScrollParameters();
}


//====================================================================
void  Console::FocusOnLastLine(bool bForceScrollToLeft)
{
	pScrollBarV->SetScrollPositionToMax();

	if(bForceScrollToLeft) {
		pScrollBarH->SetScrollPosition(0);
	}
}


//====================================================================
void  Console::DumpConsoleParams()
{
	FocusOnLastLine();

	olog << caption(VW_GET_CLASS_NAME(*this) + " parameters");
	olog << "TextHeight........: " << textHeight << nl;
	olog << "FontName..........: " << drawParams.textAtts.fontName;
	if(pFont) {
		olog << " ("  << pFont->GetFamilyName() << ")";
	}
	olog << nl;
	olog << "CommandInterpreter: ";
	if(pCommandInterpreter) {
		olog << VW_GET_CLASS_NAME(*pCommandInterpreter);
		olog << " (" << pCommandInterpreter->GetInterpreterDesc() << ")" << nl;
	}
	else {
		olog << "not connected" << nl;
	}
	olog << std::endl;
}


//====================================================================
void  Console::DumpConsoleHelp()
{
	FocusOnLastLine();

	olog << caption(VW_GET_CLASS_NAME(*this) + " help");
	olog << "Escape - clear command/change focus/close console\n";
	olog << "Tab, Control+Tab - change focus/command completion\n";
	olog << "Enter - change focus/execute command\n";
	olog << "Backspace, Delete - edit command\n";
	olog << "Any character - change focus/edit command\n";
	olog << "Arrows, Control+Arrows, PageUp, PageDown, Home, End - previous/next command, scroll text, move caret\n";
	olog << "MouseWheel, Control+MouseWheel - scroll text\n";
	olog << GetKeyName(TopLevelFrame::kcClose) << " - close\n";
	olog << GetKeyName(TopLevelFrame::kcMaximizeRestore) << " - maximize/restore\n";
	olog << GetKeyName(kcHelpInfo) << " - dump help & parameters\n";
	olog << GetKeyName(kcHorzBarPos) << " - horizontal scrollbar top/bottom\n";
	olog << GetKeyName(kcTextHeightDec) << " - decrease text-height\n";
	olog << GetKeyName(kcTextHeightInc) << " - increase text-height\n";
	olog << std::endl;
}


//====================================================================
void  Console::ResetCommandHistoryWalk()
{
	iCommandHistoryIndex = 0;
	commandBeforeHistoryWalk.erase();
}


//====================================================================
bool  Console::GetPreviousCommandFromHistory(std::string& s)
{
	if(! pConsoleBuffers) return false;
	bool bChanged = false;
	pConsoleBuffers->LockCommandHistory();
	uint numCommands = pConsoleBuffers->GetNumCommandsInHistory();
	if(numCommands) {
		if(iCommandHistoryIndex) {
			if(iCommandHistoryIndex < numCommands) {
				iCommandHistoryIndex = ClampedHigh<uint>(iCommandHistoryIndex + 1, numCommands);
				bChanged = true;
			}
		}
		else {
			commandBeforeHistoryWalk = pEditBox->GetText();
			iCommandHistoryIndex = 1;
			bChanged = true;
		}
	}
	if(bChanged) {
		s = pConsoleBuffers->GetCommandFromHistory(numCommands - iCommandHistoryIndex);
	}
	pConsoleBuffers->UnlockCommandHistory();
	return bChanged;
}


//====================================================================
bool  Console::GetNextCommandFromHistory(std::string& s)
{
	if(! pConsoleBuffers) return false;
	bool bChanged = false;
	pConsoleBuffers->LockCommandHistory();
	uint numCommands = pConsoleBuffers->GetNumCommandsInHistory();
	if(numCommands) {
		if(iCommandHistoryIndex > 1) {
			iCommandHistoryIndex = ClampedHigh<uint>(iCommandHistoryIndex - 1, numCommands);
			s = pConsoleBuffers->GetCommandFromHistory(numCommands - iCommandHistoryIndex);
			bChanged = true;
		}
		else if(iCommandHistoryIndex == 1) {
			s = commandBeforeHistoryWalk;
			iCommandHistoryIndex = 0;
			bChanged = true;
		}
	}
	pConsoleBuffers->UnlockCommandHistory();
	return bChanged;
}


//====================================================================
bool  Console::InterpretCommandLine(const std::string& s)
{
	if(pCommandInterpreter) {
		return pCommandInterpreter->InterpretCommandLine(s, olog);
	}
	olog << e_error << "! CommandInterpreter not connected to gui-console.\n" << e_info;
	return false;
}


//====================================================================
bool  Console::CompleteCommandLine(const std::string& prefix, std::string& complete)
{
	if(pCommandInterpreter) {
		return pCommandInterpreter->CompleteCommandLine(prefix, complete, olog);
	}
	return false;
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


