

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Layouters.h"
#include "GUI_CS0\TopLevelFrame.h"
#include "GUI_CS0\ScrollBar.h"
#include "GUI_CS0\EditBox.h"
#include "GUI_CS0\Buttons.h"
#include "BaseTools\ConsoleBuffers.h"
#include "BaseTools\CommandInterpreter.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
class VW_GUI_CS0_API Console : public gui_cs0::TopLevelFrame,
								public ConsoleOutputBuffer::Observer
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, OutputBgrShader);


public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		float  outputTextPaddingHorz;
		float  outputTextPaddingTop;
		float  outputTextPaddingBottom;
		float  scrollbarWidth;
		float  buttonConsoleControlsTextHeight;
		float  buttonConsoleControlsPadding;

		GeomParams();
	};

	//------------------------------------------
	// <textHeight> in <textAtts> is ignored
	// (must be set by separate console-method).
	//------------------------------------------
	struct VW_GUI_CS0_API DrawParams
	{
		color4b              outputTextTypeColors[OutputStream::OutputType_Count];
		gui::TextAttributes  textAtts;

		DrawParams();
	};


public:
	//-------------------
	// Console interface.
	//-------------------
	Console(const std::string& title = "Console");

	virtual ~Console();

	void  ConnectCommandInterpreter(CommandInterpreter& ci);
	void  DisconnectCommandInterpreter();

	//------------------------------------------------------
	// Auto-connects kernel-console-buffers at construction,
	// auto-disconnects console-buffers at destruction.
	//------------------------------------------------------
	void  ConnectConsoleBuffers(ConsoleBuffers& cb);
	void  DisconnectConsoleBuffers();

	void  SetConsoleShaderNames(
				const std::string& outputBgr = DefaultShaderName(Shader_Console_OutputBgr),
				const std::string& textButton = DefaultShaderName(Shader_TopLevelFrame_TextButton_FrameControl_Push),
				const std::string& editBox = DefaultShaderName(Shader_Console_EditBox),
				const std::string& scrollBar = DefaultShaderName(Shader_Console_ScrollBar));

	void  SetConsoleSubcomponentToolTipText(
				const std::string& buttonHelpInfo = "Help/Info",
				const std::string& buttonHorzBarPosition = "Horizontal scroll-bar\ntop/bottom",
				const std::string& buttonTextHeightInc = "Increase text-height",
				const std::string& buttonTextHeightDec = "Decrease text-height");

	void  SetConsoleKeys(eKey kcHelpInfo = Key_F1,
				eKey kcHorzBarPos = Key_F2,
				eKey kcTextHeightDec = Key_F3,
				eKey kcTextHeightInc = Key_F4);

	void                SetPrompt(const std::string& sPrompt);
	const std::string&  GetPrompt() const { return sPrompt; }

	void  SetHorzScrollBarAtTop(bool bAtTop = true);
	bool  IsHorzScrollBarAtTop() const { return bHScrollAtTop; }

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	void               SetDrawParams(const DrawParams& dp = DrawParams());
	const DrawParams&  GetDrawParams() const { return drawParams; }

	void  SetTextHeight(uint th = 20);
	uint  GetTextHeight() const { return textHeight; }

	void  SetMinTextHeight(uint th = 12);
	uint  GetMinTextHeight() const { return minTextHeight; }

	void  SetMaxTextHeight(uint th = 32);
	uint  GetMaxTextHeight() const { return maxTextHeight; }

	//------------------------------------------------
	// Additional methods intended for use by shaders.
	//------------------------------------------------
	void  GetOutputTextAreaRectangle(gui::Rectangle& r) const { r = rTextArea; }


protected:
	//-----------------------------
	// VideoResourceUser overrides.
	//-----------------------------
	virtual  bool  CacheFontHandles(gui::Manager& gm);
	virtual  void  PostCacheFontHandles();

	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcLayout();
	virtual  void  UpdateGeometryCache();
	virtual  void  OnChangedVisible();
	virtual  void  DrawSelf(gui::Renderer& gr);
	virtual  bool  CacheShaderHandles(gui::Manager& gm);
	virtual  bool  OnCommand(uint commandID);
	virtual  bool  OnMouseWheel(int delta, const vec2& mousePos);
	virtual  bool  OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos);
	virtual  bool  OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData);
	virtual  bool  OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch);
	virtual  bool  OnKeyboardKey(gui::MessageHandler* target, bool bDown, eKey keyCode, uint repeatCount, char ch);

	//-------------------------
	// TopLevelFrame overrides.
	//-------------------------
	virtual  void  OnTLFControlsLayouterUpdate();

	//----------------------------------------------
	// ConsoleOutputBuffer::Observer implementation.
	//----------------------------------------------
	virtual  void  OnChanged(const ConsoleOutputBuffer& cob);

	//---------------------
	// Command identifiers.
	//---------------------
	enum {
		CID_Input = gui::CID_User,
		CID_ScrollVert,
		CID_ScrollHorz,
		CID_HelpInfo,
		CID_HorzBarPos,
		CID_TextHeightInc,
		CID_TextHeightDec,
		CID_Last
	};

	//------------------
	// Internal helpers.
	//------------------
	void  UpdateOutputTextAreaParams();
	void  HookToConsoleBuffers();
	void  UnhookFromConsoleBuffers();
	void  ResetScrollBars();
	void  FocusOnLastLine(bool bForceScrollToLeft = true);
	void  DumpConsoleParams();
	void  DumpConsoleHelp();
	void  ResetCommandHistoryWalk();
	bool  GetPreviousCommandFromHistory(std::string& s);
	bool  GetNextCommandFromHistory(std::string& s);
	bool  InterpretCommandLine(const std::string& s);
	bool  CompleteCommandLine(const std::string& prefix, std::string& complete);

	//----------------------------------
	// EditBox::Observer implementation.
	//----------------------------------
	struct VW_GUI_CS0_API EditBoxObserver : public EditBox::Observer
	{
		Console* pConsole;

		EditBoxObserver() { pConsole = 0; }

		virtual  void  OnTextChanged(const std::string& newText) {
			if(pConsole  &&  pConsole->iCommandHistoryIndex) pConsole->ResetCommandHistoryWalk();
		}
	};
	friend struct EditBoxObserver;


protected:
	EditBoxObserver  editBoxObserver;

	//----------------
	// Sub-components.
	//----------------
	TextButton*  pButtonHelpInfo;
	TextButton*  pButtonHorzBarPos;
	TextButton*  pButtonTextInc;
	TextButton*  pButtonTextDec;
	EditBox*     pEditBox;
	ScrollBar*   pScrollBarH;
	ScrollBar*   pScrollBarV;

	//------------
	// Properties.
	//------------
	GeomParams   geomParams;
	DrawParams   drawParams;
	std::string  shaderNameOutputBgr;
	uint         minTextHeight;
	uint         maxTextHeight;
	std::string  sPrompt;
	bool         bHScrollAtTop;
	eKey     kcHelpInfo;
	eKey     kcHorzBarPos;
	eKey     kcTextHeightDec;
	eKey     kcTextHeightInc;

	//--------
	// States.
	//--------
	CommandInterpreter*   pCommandInterpreter;
	ConsoleBuffers*       pConsoleBuffers;
	ConsoleOutputBuffer*  pTextSource;
	bool                  bHookedToConsoleBuffers;
	uint                  iLastTotalNumCols;
	uint                  iCommandHistoryIndex;
	std::string           commandBeforeHistoryWalk;
	float                 textHeight;

	//-------
	// Cache.
	//-------
	const OutputBgrShader*  pOutputTextBgrShader;
	const gui::Font*        pFont;
	gui::Rectangle          rTextArea;
};


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


