

//--------------------------------------------------------------------
#include "GUI_CS0\EditBox.h"
#include "GUI_CS0\ShaderNames.h"
#include "GUI\GUI_Manager.h"
#include "BaseTools\Stream.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
EditBox::GeomParams::GeomParams()
{
	preferredSize = vec2(100, 24);
	textPadding = 4;
	textHeight = 18;
}


//====================================================================
EditBox::EditBox(gui::Component* _parent,
				gui::MessageHandler* _commandTarget,
				uint _commandID)
{
	pEditBoxShader = 0;
	pObserver = 0;

	maxChars = 0;
	SetSendCommand();
	SetShiftFocusToAncestor();
	SetResetCaretPositionOnKeyboardFocusLost();

	bChangedSinceLastConfirm = false;
	caretPos = 0;
	firstVisible = 0;
	maxVisible = 0;

	SetCanDisable(true);
	SetCanFocus(true);
	SetShaderName(DefaultShaderName(Shader_EditBox));
	SetGeomParams();

	SetCommandTarget(_commandTarget);
	SetCommandID(_commandID);

	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  EditBox::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  EditBox::SetText(const std::string& _text)
{
	if(maxChars > 0) {
		text.assign(_text, 0, maxChars);
	}
	else {
		text = _text;
	}
	caretPos = text.length();

	UpdateTextAndCaretDrawCache();
}


//====================================================================
std::string  EditBox::ExtractTextBeforeCaret() const
{
	if(caretPos) {
		return text.substr(0, caretPos);
	}
	return "";
}


//====================================================================
std::string  EditBox::ExtractTextAfterCaret() const
{
	if(caretPos < text.length()) {
		return text.substr(caretPos);
	}
	return "";
}


//====================================================================
void  EditBox::SetCaretPosition(uint pos)
{
	caretPos = Clamped<uint>(pos, 0, text.length());

	UpdateTextAndCaretDrawCache();
}


//====================================================================
void  EditBox::SetMaxChars(uint _maxChars)
{
	maxChars = _maxChars;
	SetText(text);
}


//====================================================================
void  EditBox::SetShiftFocusToAncestor(bool bShiftOnEnter, bool bShiftOnTab)
{
	bShiftFocusToAncestorOnEnter = bShiftOnEnter;
	bShiftFocusToAncestorOnTab = bShiftOnTab;
}


//====================================================================
void  EditBox::GetShiftFocusToAncestor(bool& bShiftOnEnter, bool& bShiftOnTab) const
{
	bShiftOnEnter = bShiftFocusToAncestorOnEnter;
	bShiftOnTab = bShiftFocusToAncestorOnTab;
}


//====================================================================
void  EditBox::SetSendCommand(bool bSendOnEnterRepeat, bool bSendOnKeyboardFocusLost)
{
	bSendCommandOnEnterRepeat = bSendOnEnterRepeat;
	bSendCommandOnKeyboardFocusLost = bSendOnKeyboardFocusLost;
}


//====================================================================
void  EditBox::GetSendCommand(bool& bSendOnEnterRepeat, bool& bSendOnKeyboardFocusLost) const
{
	bSendOnEnterRepeat = bSendCommandOnEnterRepeat;
	bSendOnKeyboardFocusLost = bSendCommandOnKeyboardFocusLost;
}


//====================================================================
void  EditBox::SetObserver(EditBox::Observer* observer)
{
	pObserver = observer;
}


//====================================================================
void  EditBox::GetVisibleTextInfo(uint& _firstVisible, uint& _maxVisible) const
{
	_firstVisible = firstVisible;
	_maxVisible = maxVisible; 
}


//====================================================================
void  EditBox::RecalcSize()
{
	SetSize(geomParams.preferredSize);
}


//====================================================================
void  EditBox::UpdateGeometryCache()
{
	fMaxVisibleTextWidth = GetSize().x - 2 * geomParams.textPadding;
	UpdateTextAndCaretDrawCache();
}


//====================================================================
void  EditBox::PostCacheShaderHandles()
{
	Component::PostCacheShaderHandles();

	pEditBoxShader = gui::ConstShaderPtr<EditBox::Shader>(GetBaseShader());
}


//====================================================================
bool  EditBox::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(bDown  &&  buttonCode == MouseButton_Left  &&  ! repeatCount) {
		if(IsKeyboardFocusTarget()) {
			SetCaretPosition(mousePos);
		}
		return true;
	}
	return false;
}


//====================================================================
bool  EditBox::OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData)
{
	DefaultProcessQueryFocus(bFocusByMouse, buttonCode, bWantFocus, bNeedCharacterData);
	bNeedCharacterData = true;
	return true;
}


//====================================================================
bool  EditBox::OnKeyboardFocus(bool bGotFocus)
{
	if(bGotFocus) {
		bChangedSinceLastConfirm = false;
	}
	else {
		if(bChangedSinceLastConfirm) {
			bChangedSinceLastConfirm = false;

			if(bSendCommandOnKeyboardFocusLost) {
				SendCommandMessageToTarget();
			}
		}
		if(bResetCaretPositionOnKeyboardFocusLost) {
			caretPos = text.length();
			UpdateTextAndCaretDrawCache();
		}
	}
	return true;
}


//====================================================================
bool  EditBox::OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch)
{
	if(! bDown) return false;

	const char* wsChars = " \t";

	bool bTestChar = false;
	bool bConfirmed = false;
	uint lastTextLen = text.length();

	// process key commands
	gui::Manager& gm = gui::Manager::Instance();

	switch(keyCode) {
		case Key_Tab: {
			if(bShiftFocusToAncestorOnTab) {
				RequestKeyboardFocusToAncestor();
				return true;
			}
		}
		case Key_Enter:
		case Key_NumEnter: {
			bConfirmed = bSendCommandOnEnterRepeat  ||  (! repeatCount);
			break;
		}
		case Key_Backspace: {
			if(caretPos > 0) {

				uint newCaretPos = caretPos - 1;
				std::string newText = text;
				newText.erase(newCaretPos, 1);

				if(ConfirmTextChange(text, newText)) {
					caretPos = newCaretPos;
					text = newText;
				}
			}
			break;
		}
		case Key_Delete: {
			if(gm.IsKeyModifierPressed(KeyModifier_Control)) {
				if(ConfirmTextChange(text, "")) {
					caretPos = 0;
					text.erase();
				}
				break;
			}
			if(caretPos < text.length()) {

				std::string newText = text;
				newText.erase(caretPos, 1);

				if(ConfirmTextChange(text, newText)) {
					text = newText;
				}
			}
			break;
		}
		case Key_Home: {
			caretPos = 0;
			break;
		}
		case Key_End: {
			caretPos = text.length();
			break;
		}
		case Key_ArrowUp:
		case Key_ArrowLeft: {
			if(caretPos > 0) {
				if(gm.IsKeyModifierPressed(KeyModifier_Control)) {
					caretPos = text.find_last_not_of(wsChars, caretPos - 1);
					if(caretPos != std::string::npos) {
						caretPos = text.find_last_of(wsChars, caretPos);
						if(caretPos != std::string::npos) {
							caretPos++;
						}
					}
					if(caretPos == std::string::npos) {
						caretPos = 0;
					}
				}
				else {
					caretPos--;
				}
			}
			break;
		}
		case Key_ArrowDown:
		case Key_ArrowRight: {
			if(caretPos != text.length()) {
				if(gm.IsKeyModifierPressed(KeyModifier_Control)) {
					caretPos = text.find_first_of(wsChars, caretPos);
					if(caretPos != std::string::npos) {
						caretPos = text.find_first_not_of(wsChars, caretPos);
					}
				}
				else {
					caretPos++;
				}
			}
			break;
		}
		default: {
			bTestChar = true;
			break;
		}
	}

	// process character
	if(bTestChar) {
		if(ch == 0) return false;

		Clamp<uint>(caretPos, 0, text.length());

		if((! maxChars)  ||  (text.length() < maxChars)) {

			std::string newText = text;
			newText.insert(caretPos, 1, ch);
			uint newCaretPos = caretPos + 1;

			if(ConfirmTextChange(text, newText)) {
				caretPos = newCaretPos;
				text = newText;
			}
		}
	}

	// force valid caret-position (observer can change text)
	Clamp<uint>(caretPos, 0, text.length());

	// geometry-update & notify
	UpdateTextAndCaretDrawCache();

	if(lastTextLen != text.length()) {
		bChangedSinceLastConfirm = true;
		NotifyTextChanged(text);
	}
	if(bConfirmed) {
		bChangedSinceLastConfirm = false;

		SendCommandMessageToTarget();

		if(bShiftFocusToAncestorOnEnter) {
			RequestKeyboardFocusToAncestor();
		}
	}
	return true;
}


//====================================================================
void  EditBox::SetCaretPosition(const vec2& mousePos)
{
	if(textWidths.empty()) return;

	float fFirstVisibleDrawPos = firstVisible ? textWidths[firstVisible - 1] : 0;
	float fClickedPos = mousePos.x - geomParams.textPadding + fFirstVisibleDrawPos;

	// find lower limit
	uint lastVisible = firstVisible + maxVisible - 1;
	for(uint i = firstVisible; i < lastVisible; i++) {
		if(textWidths[i] > fClickedPos) break;
	}

	// fine-tune to closest spacing
	caretPos = i;
	if(caretPos < textWidths.size()) {
		float deltaCurr = Abs(((caretPos > 0) ? textWidths[caretPos - 1] : 0) - fClickedPos);
		float deltaNext = Abs(textWidths[caretPos] - fClickedPos);
		if(deltaNext < deltaCurr) {
			caretPos++;
		}
	}

	UpdateTextAndCaretDrawCache();
}


//====================================================================
void  EditBox::UpdateTextAndCaretDrawCache()
{
	// update widths
	if(! pEditBoxShader) return;
	const gui::Font* pFont = pEditBoxShader->GetFont();
	if(! pFont) return;

	pFont->CalcTextWidths(textWidths, pEditBoxShader->GetTextMetricsL(geomParams.textHeight), text);

	// handle 0-length text
	if(textWidths.empty()) {
		firstVisible = 0;
		maxVisible = 0;
		fCaretDrawPos = geomParams.textPadding;
		return;
	}

	// calc drawing positions
	ClampHigh<uint>(caretPos, textWidths.size());
	ClampHigh<uint>(firstVisible, textWidths.size() - 1);

	float fFirstVisibleDrawPos = firstVisible ? textWidths[firstVisible - 1] : 0;
	float fLastVisibleDrawPos = fFirstVisibleDrawPos + fMaxVisibleTextWidth;
	fCaretDrawPos = caretPos ? textWidths[caretPos - 1] : 0;

	// adjust drawing positions
	if(textWidths.back() > fMaxVisibleTextWidth) {

		// no virtual-space
		if(fLastVisibleDrawPos > textWidths.back()) {
			fLastVisibleDrawPos = textWidths.back();
			fFirstVisibleDrawPos = fLastVisibleDrawPos - fMaxVisibleTextWidth;
		}

		// keep caret in visible area
		if(fCaretDrawPos > fLastVisibleDrawPos) {
			fFirstVisibleDrawPos = fCaretDrawPos - fMaxVisibleTextWidth;
		}
		else if(fCaretDrawPos < fFirstVisibleDrawPos) {
			fFirstVisibleDrawPos = fCaretDrawPos;
		}
		fLastVisibleDrawPos = fFirstVisibleDrawPos + fMaxVisibleTextWidth;

		// find first completely-visible character
		uint iFV = 0;
		while((iFV < textWidths.size())  &&  (textWidths[iFV] <= fFirstVisibleDrawPos)) {
			iFV++;
		}
		if(iFV > 0) {
			if(fFirstVisibleDrawPos > textWidths[iFV - 1]) iFV++;
		}
		else {
			if(fFirstVisibleDrawPos > 0) iFV++;
		}
		firstVisible = ClampedHigh<uint>(iFV, textWidths.size() - 1);

		// find last completely-visible character
		uint lastVisible = firstVisible;
		for(int iLV = textWidths.size() - 1; iLV >= firstVisible; iLV--) {
			if(textWidths[iLV] <= fLastVisibleDrawPos) {
				lastVisible = iLV;
				break;
			}
		}
		maxVisible = lastVisible - firstVisible + 1;

		fFirstVisibleDrawPos = firstVisible ? textWidths[firstVisible - 1] : 0;
	}
	else {
		firstVisible = 0;
		maxVisible = textWidths.size();
	}

	// update drawing cache
	fCaretDrawPos = geomParams.textPadding + fCaretDrawPos - fFirstVisibleDrawPos;
}


//====================================================================
bool  EditBox::ConfirmTextChange(const std::string& lastText, const std::string& newText)
{
	if(pObserver) {
		return pObserver->ConfirmTextChange(lastText, newText);
	}
	return true;
}


//====================================================================
void  EditBox::NotifyTextChanged(const std::string& newText)
{
	if(pObserver) {
		pObserver->OnTextChanged(newText);
	}
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


