

//--------------------------------------------------------------------
#include "GUI_CS0\FontTest.h"
#include "GUI\GUI_Manager.h"
#include "GUI\GUI_StdObjects.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
FontTest::GeomParams::GeomParams()
{
	cellSize = 32;
}


//====================================================================
FontTest::DrawParams::DrawParams()
{
	colorBackground = color4b(0, 0, 128);
	colorGrid = color4b(128, 128, 128);
	colorBoxMonospace = color4b::RED;
	colorBoxVarPitch = color4b::GREEN;
	colorFontGlyph = color4b::WHITE;
}


//====================================================================
FontTest::FontTest()
{
	pFont = 0;
	pFontTexture = 0;

	SetVisible(false);
	SetCanFocus(true);
	SetDefaultCommandArea(gui::CommandArea_Move);

	SetGeomParams();
	SetDrawParams();

	SetBoxDisplayMode();
	SetDrawGrid();
	ShowNextFont();
}


//====================================================================
void  FontTest::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	Clamp<float>(geomParams.cellSize, 8, 48);

	float len = geomParams.cellSize * 16;
	SetSize(len, len);

	ReportDirtyGeometry();
}


//====================================================================
void  FontTest::SetDrawParams(const DrawParams& dp)
{
	drawParams = dp;

	rBackground.SetColor(dp.colorBackground);
	rGridCell.SetColor(dp.colorGrid);
	rGlyphCell.SetColor(dp.colorFontGlyph);
	rBoxMono.SetColor(dp.colorBoxMonospace);
	rBoxVar.SetColor(dp.colorBoxVarPitch);
}


//====================================================================
void  FontTest::SetBoxDisplayMode(eBoxDisplayMode mode)
{
	boxDisplayMode = mode;
	if(boxDisplayMode >= BoxDisplayMode_Count) {
		boxDisplayMode = eBoxDisplayMode(0);
	}
}


//====================================================================
void  FontTest::SetDrawGrid(bool _bDrawGrid)
{
	bDrawGrid = _bDrawGrid;
}


//====================================================================
void  FontTest::ShowNextFont()
{
	activeFontName = GetNextRegisteredFontName();
	SetupToolTip();
	ResetToolTip();
	RequestFontRecache();
}


//====================================================================
void  FontTest::UpdateGeometryCache()
{
	glyphSize.Set(geomParams.cellSize, geomParams.cellSize);
	rBackground.Set(vec2::ZERO, GetSize());
	rGridCell.Set(vec2::ZERO, glyphSize);
	rGlyphCell.CopyGeometryParamsFrom(rGridCell);
}


//====================================================================
void  FontTest::DrawSelf(gui::Renderer& gr)
{
	// draw background
	gr.DrawRectangle(rBackground);

	if(! pFont) return;

	// draw grid
	if(bDrawGrid) {

		rGridCell.top = 0;
		for(uint r = 0; r < 16; r++) {

			rGridCell.left = 0;
			for(uint c = 0; c < 16; c++) {

				gr.DrawRectangleFrame(rGridCell);
				rGridCell.left += geomParams.cellSize;
			}
			rGridCell.top += geomParams.cellSize;
		}
	}

	//-----------------------
	// Generic gui::Font type
	//-----------------------
	if(! pFontTexture) {

		VideoRenderer& vr = gr.GetParentVideoRenderer();

		gui::DrawTextParams dtp(
				drawParams.colorFontGlyph, vec2::ZERO,
				gui::TextMetrics(geomParams.cellSize, false, false, 0, 1, 1));

		uint numRows = pFont->GetNumGlyphs() / 16;

		// draw glyphs
		char ch = 0;
		for(uint r = 0; r < numRows; r++) {

			float xPos = 0;
			dtp.positionTopLeft.x = 0;
			for(uint c = 0; c < 16; c++) {

				float offset = (geomParams.cellSize - pFont->CalcGlyphWidth(dtp.tm, ch)) * 0.5f;
				dtp.positionTopLeft.x += offset;

				gr.DrawText(*pFont, 1, &ch, dtp);

				xPos += geomParams.cellSize;
				dtp.positionTopLeft.x = xPos;

				ch++;
			}
			dtp.positionTopLeft.y += geomParams.cellSize;
		}
		return;
	}

	//---------------------------
	// gui::FontTexture type only
	//---------------------------

	// draw glyphs
	char ch = 0;

	rGlyphCell.top = 0;
	{for(uint r = 0; r < pFontTexture->GetNumTileRows(); r++) {

		rGlyphCell.left = 0;
		for(uint c = 0; c < pFontTexture->GetNumTileColumns(); c++) {

			pFontTexture->GetTileCoords(r, c, rGlyphCell);
			gr.DrawRectangle(*pFontTexture, rGlyphCell);
			rGlyphCell.left += geomParams.cellSize;
			ch++;
		}
		rGlyphCell.top += geomParams.cellSize;
	}}

	// draw tex-coord boxes
	float  boxW = GetSize().x;
	float  boxH = GetSize().y * ((pFontTexture->GetNumTileRows() < 16) ? 0.5f : 1.0f);

	switch(boxDisplayMode) {
		case BoxDisplayMode_Monospace: {
			for(uint i = 0; i < pFontTexture->GetNumGlyphs(); i++) {

				pFontTexture->GetGlyphCoordsMono(ubyte(i),
					rBoxMono.left, rBoxMono.width, rBoxMono.top, rBoxMono.height);

				rBoxMono.width -= rBoxMono.left;
				rBoxMono.height -= rBoxMono.top;
				rBoxMono.left *= boxW;
				rBoxMono.width *= boxW;
				rBoxMono.top *= boxH;
				rBoxMono.height *= boxH;

				gr.DrawRectangleFrame(rBoxMono, 1);
			}
			break;
		}
		case BoxDisplayMode_VarPitch: {
			for(uint i = 0; i < pFontTexture->GetNumGlyphs(); i++) {

				pFontTexture->GetGlyphCoords(ubyte(i),
					rBoxVar.left, rBoxVar.width, rBoxVar.top, rBoxVar.height);

				rBoxVar.width -= rBoxVar.left;
				rBoxVar.height -= rBoxVar.top;
				rBoxVar.left *= boxW;
				rBoxVar.width *= boxW;
				rBoxVar.top *= boxH;
				rBoxVar.height *= boxH;

				gr.DrawRectangleFrame(rBoxVar, 1);
			}
			break;
		}
	}
}


//====================================================================
void  FontTest::OnFontListChanged(gui::Manager& gm)
{
	if(! gm.GetFont(activeFontName)) {
		ShowNextFont();
	}
}


//====================================================================
bool  FontTest::CacheFontHandles(gui::Manager& gm)
{
	bool bOK = false;
	if(activeFontName.empty()) {
		pFont = 0;
		pFontTexture = 0;
		bOK = true;
	}
	else {
		pFont = gm.GetFont(activeFontName);
		pFontTexture = dynamic_cast<const gui::FontTexture*>(pFont);
		bOK = (pFont != 0);
	}
	SetupToolTip();
	return bOK;
}


//====================================================================
bool  FontTest::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(! bDown  ||  repeatCount) return false;

	if(buttonCode == MouseButton_Right) {
		SetBoxDisplayMode(eBoxDisplayMode(boxDisplayMode + 1));
		ResetToolTip();
		return true;
	}
	return false;
}


//====================================================================
bool  FontTest::OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData)
{
	DefaultProcessQueryFocus(bFocusByMouse, buttonCode, bWantFocus, bNeedCharacterData);
	return true;
}


//====================================================================
bool  FontTest::OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch)
{
	if(! bDown) return false;

	switch(keyCode) {
		case Key_Escape: {
			if(repeatCount  ||  ! IsTopLevel()) break;
			ShowTopLevel(false);
			return true;
		}
		case Key_G: {
			if(repeatCount) break;
			SetDrawGrid(! bDrawGrid);
			return true;
		}
		case Key_Space: {
			if(repeatCount) break;
			ShowNextFont();
			return true;
		}
		case Key_NumAdd:
		case Key_NumSubtract: {
			GeomParams gp = geomParams;
			gp.cellSize += (keyCode == Key_NumAdd) ? 1 : -1;
			SetGeomParams(gp);
			UpdateGeometry();
			SetupToolTip();
			ResetToolTip();
			return true;
		}
	}
	return false;
}


//====================================================================
bool  FontTest::OnQueryToolTipParams(gui::ToolTipParams& ttp)
{
	Component::OnQueryToolTipParams(ttp);
	ttp.mouseMoveIgnoreAreaSize.Set(geomParams.cellSize, geomParams.cellSize);
	ttp.mouseMoveIgnoreAreaPosition = ttp.position - ttp.mouseMoveIgnoreAreaSize * 0.5f;
	return true;
}


//====================================================================
std::string  FontTest::GetNextRegisteredFontName()
{
	std::vector<std::string> regFontNames;
	gui::Manager::Instance().GetRegisteredFontNames(regFontNames);

	if(regFontNames.empty()) return activeFontName;

	uint iNext = 0;
	for(uint i = 0; i < regFontNames.size(); i++) {
		if(regFontNames[i] == activeFontName) {
			if(i == (regFontNames.size() - 1)) iNext = 0;
			else iNext = i + 1;
			break;
		}
	}
	return regFontNames[iNext];
}


//====================================================================
void  FontTest::SetupToolTip()
{
	std::string sClassName;
	std::string sInfo;
	std::string sCommands;

	if(pFont) {
		sClassName = VW_GET_CLASS_NAME(*pFont);

		sInfo =  pFont->GetFamilyName() + ", ";
		sInfo += ToString(pFont->GetNumGlyphs()) + " glyphs";
	}
	if(pFontTexture) {
		sCommands = " RMB - glyph boundaries [hide/monospace/varpitch]";
	}

	SetToolTipText(
		"FontName: " + activeFontName + "\n"
		"ClassName: " + sClassName + "\n"
		"FontInfo: " + sInfo + "\n"
		"CellSize: " + ToString(geomParams.cellSize) + "\n"
		"Commands:\n Space - next font\n Esc - close\n"
		" NumPad-/+ - adjust cell-size\n G - toggle grid\n"
		+ sCommands);
}


//====================================================================
void  FontTest::ResetToolTip()
{
	if(IsMouseOverDirect()) {
		SendMessageToServiceProvider(gui::SID_ToolTip, gui::Message_ServiceCommand_ToolTip(gui::SCID_Reset));
	}
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


