

//--------------------------------------------------------------------
#include "GUI_CS0\HIDConfigDialog.h"
#include "Framework\Application.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
HIDConfigDialog::HIDConfigDialog(const std::string& _title,
								uint _hidReconfigCode)
	:
	Dialog(_title)
{
	SetHIDReconfigCode(_hidReconfigCode);

	pLayouter = new gui::layouters::RowLayouter(this);

	pHorzTextSeparatorDInput = new HorzTextSeparator(pLayouter, "DirectInput");
	pCheckBoxEnableDInput = new CheckBox(pLayouter, this, CID_DIEnable);
	pLabelEnableDInput = new Label(pLayouter, 0, 0, "Enable DirectInput", pCheckBoxEnableDInput);
	pCheckBoxLastVersion = new CheckBox(pLayouter, this, CID_DILastVersion);
	pLabelLastVersion = new Label(pLayouter, 0, 0, "Use last version", pCheckBoxLastVersion);
	pCheckBoxExclusiveMode = new CheckBox(pLayouter, this, CID_DIExclusiveMode);
	pLabelExclusiveMode = new Label(pLayouter, 0, 0, "Exclusive mode", pCheckBoxExclusiveMode);
	pCheckBoxUseAtInit = new CheckBox(pLayouter, this, CID_DIUseAtInit);
	pLabelUseAtInit = new Label(pLayouter, 0, 0, "Use at device init", pCheckBoxUseAtInit);
	pLabelVersion = new Label(pLayouter, 0, 0, "Version:");
	pComboBoxVersion = new ComboBox(pLayouter, this, CID_DIVersion);

	// RevisionCheckPoint:
	// Add DirectInput versions.
	pComboBoxVersion->AddItem(new ComboBoxItem("8", 8)); //version 8

	SetHIDConfigDialogShaderNames();
	SetHIDConfigDialogSubcomponentToolTipText();
	SetSize(250, 280);

	OnDialogOpened();
	Kernel::Instance().RegisterReconfigObserver(*this);

	// Specify layout
	SetWorkAreaLayouter(pLayouter);

	using namespace gui::layouters;
	RowLayouter& l = *pLayouter;

	l.Clear();
	l.TargetHeightRow(pHorzTextSeparatorDInput).Target(pHorzTextSeparatorDInput, SF_Variable);
	l.FixedHeightRow().Target(pCheckBoxEnableDInput).Spacer().Target(pLabelEnableDInput);
	l.FixedHeightRow().Target(pCheckBoxLastVersion).Spacer().Target(pLabelLastVersion);
	l.FixedHeightRow().Target(pLabelVersion).Spacer().Target(pComboBoxVersion);
	l.FixedHeightRow().Target(pCheckBoxExclusiveMode).Spacer().Target(pLabelExclusiveMode);
	l.FixedHeightRow().Target(pCheckBoxUseAtInit).Spacer().Target(pLabelUseAtInit);
}


//====================================================================
void  HIDConfigDialog::SetHIDConfigDialogShaderNames(
									const std::string& label,
									const std::string& labelCB,
									const std::string& checkBox,
									const std::string& comboBox,
									const std::string& horzTextSeparator)
{
	pHorzTextSeparatorDInput->SetShaderName(horzTextSeparator);
	pLabelEnableDInput->SetShaderName(labelCB);
	pLabelLastVersion->SetShaderName(labelCB);
	pLabelExclusiveMode->SetShaderName(labelCB);
	pLabelUseAtInit->SetShaderName(labelCB);
	pLabelVersion->SetShaderName(label);
	pCheckBoxEnableDInput->SetShaderName(checkBox);
	pCheckBoxLastVersion->SetShaderName(checkBox);
	pCheckBoxExclusiveMode->SetShaderName(checkBox);
	pCheckBoxUseAtInit->SetShaderName(checkBox);
	pComboBoxVersion->SetShaderName(comboBox);
}


//====================================================================
void  HIDConfigDialog::SetHIDConfigDialogSubcomponentToolTipText(
									const std::string& sEnableDInput,
									const std::string& sLastVersion,
									const std::string& sExclusiveMode,
									const std::string& sUseAtInit,
									const std::string& sSelectVersion)
{
	pCheckBoxEnableDInput->SetToolTipText(sEnableDInput);
	pCheckBoxLastVersion->SetToolTipText(sLastVersion);
	pCheckBoxExclusiveMode->SetToolTipText(sExclusiveMode);
	pCheckBoxUseAtInit->SetToolTipText(sUseAtInit);
	pComboBoxVersion->SetToolTipText(sSelectVersion);
}


//====================================================================
void  HIDConfigDialog::OnHIDReconfig()
{
	OnDialogOpened();
}


//====================================================================
bool  HIDConfigDialog::OnCommand(uint commandID)
{
	switch(commandID) {

		case CID_DIEnable:
		case CID_DILastVersion:
		case CID_DIVersion:
		case CID_DIExclusiveMode:

			UpdateControlDependencies();
			pButtonApply->SetEnabled(true);
			return true;
	}
	return Dialog::OnCommand(commandID);
}


//====================================================================
void  HIDConfigDialog::OnDialogOpened()
{
	MatchActiveConfig();
	UpdateControlDependencies();
	pButtonApply->SetEnabled(false);
}


//====================================================================
void  HIDConfigDialog::OnApplySettings()
{
	// fill new configuration
	DirectInputConfig diConfig;

	diConfig.enableDirectInput = pCheckBoxEnableDInput->IsOn();
	diConfig.useLastVersion = pCheckBoxLastVersion->IsOn();
	diConfig.enableFullscreenExclusiveMode = pCheckBoxExclusiveMode->IsOn();
	diConfig.directInputVersion = 0;
	diConfig.useAtDeviceInit = pCheckBoxUseAtInit->IsOn();

	const ComboBoxItem* cbi = pComboBoxVersion->GetSelectedItem();
	if(cbi) diConfig.directInputVersion = cbi->GetID();

	// give derived class a chance to override new settings
	OnPreApplySettings(diConfig);

	DirectInputConfig diConfigActive;
	Kernel::Instance().GetActiveDirectInputConfig(diConfigActive);

	// do reconfig
	if(! diConfig.IsLogicallyEqual(diConfigActive)) {
		Kernel::Instance().RequestHIDReconfig(diConfig, iHIDReconfigCode);
	}
}


//====================================================================
void  HIDConfigDialog::MatchActiveConfig()
{
	DirectInputConfig diConfigActive;
	Kernel::Instance().GetActiveDirectInputConfig(diConfigActive);

	pCheckBoxEnableDInput->SetOn(diConfigActive.enableDirectInput);
	pCheckBoxLastVersion->SetOn(diConfigActive.useLastVersion);
	pCheckBoxExclusiveMode->SetOn(diConfigActive.enableFullscreenExclusiveMode);
	pCheckBoxUseAtInit->SetOn(diConfigActive.useAtDeviceInit);

	for(uint i = 0; i < pComboBoxVersion->GetNumItems(); i++) {
		const ComboBoxItem& cbi = pComboBoxVersion->GetItem(i);
		if(cbi.GetID() == diConfigActive.directInputVersion) {
			pComboBoxVersion->SelectItem(i);
			break;
		}
	}
}


//====================================================================
void  HIDConfigDialog::UpdateControlDependencies()
{
	bool bDIEnable = pCheckBoxEnableDInput->IsOn();

	pCheckBoxLastVersion->SetEnabled(bDIEnable);
	pLabelLastVersion->SetEnabled(pCheckBoxLastVersion->IsEnabled());

	pCheckBoxExclusiveMode->SetEnabled(bDIEnable);
	pLabelExclusiveMode->SetEnabled(pCheckBoxExclusiveMode->IsEnabled());

	pCheckBoxUseAtInit->SetEnabled(bDIEnable);
	pLabelUseAtInit->SetEnabled(pCheckBoxUseAtInit->IsEnabled());

	pComboBoxVersion->SetEnabled(bDIEnable  &&  (! pCheckBoxLastVersion->IsOn()));
	pLabelVersion->SetEnabled(pComboBoxVersion->IsEnabled());
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


