

//--------------------------------------------------------------------
#include "GUI_CS0\HIDControllerPanes.h"


//====================================================================
#define   DRAW_ONE_LINE(s)  DrawOneLine(s, gr, font, dtp, lineSpacing);


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
HIDControllerPane::DrawParams::DrawParams()
{
	textOffsetTopLeft.Set(10, 26);
	textSpacingH = 10;
	textSpacingV = 6;
	colorText = color4b(255, 255, 255, 160);
	colorTextActive = color4b::WHITE;
	colorTextHot = color4b::WHITE;
}


//====================================================================
color4b  HIDControllerPane::DrawParams::ChooseTextColor(const HIDControllerPane& pane) const
{
	if(pane.IsMouseOver()) return colorTextHot;
	if(pane.IsActive()) return colorTextActive;
	return colorText;
}


//====================================================================
HIDControllerPane::HIDControllerPane(const std::string& _sControllerInterfaceName,
									uint _iControllerOrdinalNumber,
									const std::string& _controllerName,
									const DrawParams& _dp,
									const gui::TextAttributes& _ta)
	:
	TopLevelPane("")
{
	pController = 0;

	SetTextAttributes(_ta);
	SetDrawParams(_dp);
	SetHIDControllerName(_controllerName);
	SetControllerInterfaceName(_sControllerInterfaceName);
	SetControllerOrdinalNumber(_iControllerOrdinalNumber);

	Kernel::Instance().RegisterReconfigObserver(*this);
}


//====================================================================
void  HIDControllerPane::SetDrawParams(const DrawParams& dp)
{
	drawParams = dp;
}


//====================================================================
void  HIDControllerPane::SetHIDControllerName(const std::string& s)
{
	controllerName = s;
	UpdatePaneTitle();
}


//====================================================================
void  HIDControllerPane::SetControllerOrdinalNumber(uint iOrdinalNumber)
{
	iControllerOrdinalNumber = iOrdinalNumber;
	UpdatePaneTitle();
}


//====================================================================
void  HIDControllerPane::PostCacheFontHandles()
{
	TopLevelPane::PostCacheFontHandles();
	RecalcInfoTextCache();
}


//====================================================================
void  HIDControllerPane::DrawSelf(gui::Renderer& gr)
{
	TopLevelPane::DrawSelf(gr);

	if(! pFont) return;

	// setup/print common info
	gui::DrawTextParams  dtp(
		drawParams.ChooseTextColor(*this), drawParams.textOffsetTopLeft, textAtts.textMetricsL);

	float lineSpacing = drawParams.textSpacingV + textAtts.textMetricsL.textHeight;

	if(pController) {
		HumanInterfaceDevice& hid = pController->GetParentHID();

		DrawOneLine("Using DInput: " + ToString(hid.IsUsingDirectInput()),
					gr, *pFont, dtp, lineSpacing);

		std::string sAsyncRate = "Async rate: ";
		if(hid.IsAsyncSamplingActive()) {
			sAsyncRate.append(ToStringFix(hid.GetAverageSampleRate(), 0, 2));
		}
		else {
			sAsyncRate.append("NA");
		}
		DrawOneLine(sAsyncRate, gr, *pFont, dtp, lineSpacing);
	}

	// check for async read
	bool bAsync = (pController  &&  pController->GetParentHID().IsAsyncSamplingActive());
	if(bAsync) pController->GetParentHID().LockSampledData();

	DrawInfoText(gr, *pFont, dtp, drawParams, lineSpacing);

	if(bAsync) pController->GetParentHID().UnlockSampledData();
}


//====================================================================
void  HIDControllerPane::OnHIDReconfig()
{
	pController = 0;

	HIDController* hidc = Kernel::Instance().GetHIDController(
							sControllerInterfaceName, iControllerOrdinalNumber);

	if(CacheHIDController(hidc)) {
		pController = hidc;
	}
}


//====================================================================
void  HIDControllerPane::UpdatePaneTitle()
{
	SetTitleText(controllerName + " (" + ToString(iControllerOrdinalNumber) + ") State");
}


//====================================================================
MouseStatePane::MouseStatePane(const std::string& _controllerName,
								uint _iMouseOrdinalNumber,
								const DrawParams& _dp,
								const gui::TextAttributes& _ta)
	:
	HIDControllerPane(VW_INTERFACE_NAME_Mouse, _iMouseOrdinalNumber, _controllerName, _dp, _ta)
{
	SetSize(320, 250);

	pMouse = 0;
}


//====================================================================
bool  MouseStatePane::CacheHIDController(HIDController* hidc)
{
	pMouse = InterfaceCast<Mouse>(hidc);
	return (pMouse != 0);
}


//====================================================================
void  MouseStatePane::DrawInfoText(gui::Renderer& gr,
									const gui::Font& font, gui::DrawTextParams& dtp,
									const DrawParams& dp, float lineSpacing)
{
	if(! pMouse) {
		gr.DrawText(font, "Mouse not found", dtp);
		return;
	}

	// print pressed buttons
	std::string s = "Buttons: ";
	if(pMouse->ButtonPressed(MouseButton_Left)) s.append("L ");
	if(pMouse->ButtonPressed(MouseButton_Right)) s.append("R ");
	if(pMouse->ButtonPressed(MouseButton_Middle)) s.append("M ");
	if(pMouse->ButtonPressed(MouseButton_X1)) s.append("X1 ");
	if(pMouse->ButtonPressed(MouseButton_X2)) s.append("X2");

	DRAW_ONE_LINE(s);

	// print wheel
	DRAW_ONE_LINE("Wheel: " + ToString(pMouse->GetDeltaWheel()));

	// relative-mode specific
	if(pMouse->IsActiveRelativeMode()) {
		DRAW_ONE_LINE("DeltaX: " + ToString(pMouse->GetDeltaX()));
		DRAW_ONE_LINE("DeltaY: " + ToString(pMouse->GetDeltaY()));

		float sx, sy;
		pMouse->GetRelativeModeSensitivity(sx, sy);
		DRAW_ONE_LINE("SensitivityX: " + ToString(sx));
		DRAW_ONE_LINE("SensitivityY: " + ToString(sy));
	}
	// absolute-mode specific
	else {
		DRAW_ONE_LINE("PositionX: " + ToString(pMouse->GetPositionX()));
		DRAW_ONE_LINE("PositionY: " + ToString(pMouse->GetPositionY()));
		DRAW_ONE_LINE("SysCursorSync: " + ToString(pMouse->IsActiveSyncWithSystemCursor()));
		DRAW_ONE_LINE("Speed: " + ToString(pMouse->GetAbsoluteModeSpeed()));
		DRAW_ONE_LINE("AccelLevel: " + ToString(pMouse->GetAbsoluteModeAccelerationLevel()));

		float t1, t2;
		pMouse->GetAbsoluteModeAccelerationThresholds(t1, t2);
		DRAW_ONE_LINE("AccelThresh: " + ToString(t1) + " " + ToString(t2));
	}
}


//====================================================================
KeyboardStatePane::KeyboardStatePane(const std::string& _controllerName,
								uint _iKeyboardOrdinalNumber,
								const DrawParams& _dp,
								const gui::TextAttributes& _ta)
	:
	HIDControllerPane(VW_INTERFACE_NAME_Keyboard, _iKeyboardOrdinalNumber, _controllerName, _dp, _ta)
{
	SetSize(450, 210);

	pKeyboard = 0;
}


//====================================================================
bool  KeyboardStatePane::CacheHIDController(HIDController* hidc)
{
	pKeyboard = InterfaceCast<Keyboard>(hidc);
	return (pKeyboard != 0);
}


//====================================================================
void  KeyboardStatePane::DrawInfoText(gui::Renderer& gr,
									const gui::Font& font, gui::DrawTextParams& dtp,
									const DrawParams& dp, float lineSpacing)
{
	if(! pKeyboard) {
		gr.DrawText(font, "Keyboard not found", dtp);
		return;
	}

	// print character mapping state
	DRAW_ONE_LINE("CharacterMappingEnabled: " + ToString(pKeyboard->IsEnabledCharacterMapping()));

	// print event counts
	uint numPressed = pKeyboard->NumKeysPressed();
	DRAW_ONE_LINE("NumPressed: " + ToString(numPressed));
	DRAW_ONE_LINE("NumDown: " + ToString(pKeyboard->NumKeysDown()));
	DRAW_ONE_LINE("NumUp: " + ToString(pKeyboard->NumKeysUp()));
	DRAW_ONE_LINE("NumKeyModifiersPressed: " + ToString(pKeyboard->NumKeyModifiersPressed()));

	// print first pressed key
	if(numPressed) {
		for(uint i = 0; i < Key_Count; i++) {
			eKey kc = eKey(i);
			if(pKeyboard->KeyPressed(kc)) {
				char ch = 0;
				pKeyboard->MapKeyCodeToCharacter(kc, ch);
				DRAW_ONE_LINE("Pressed: " + GetKeyName(kc) + ", " + std::string(1, ch));
				break;
			}
		}
	}
	else {
		DRAW_ONE_LINE("Pressed: ");
	}
}


//====================================================================
PositionTrackerStatePane::PositionTrackerStatePane(const std::string& _controllerName,
									uint _iPositionTrackerOrdinalNumber,
									const DrawParams& _dp,
									const gui::TextAttributes& _ta)
	:
	HIDControllerPane(VW_INTERFACE_NAME_PositionTracker, _iPositionTrackerOrdinalNumber, _controllerName, _dp, _ta)
{
	SetSize(350, 140);

	pTracker = 0;
}


//====================================================================
void  PositionTrackerStatePane::DrawInfoText(gui::Renderer& gr,
								const gui::Font& font, gui::DrawTextParams& dtp,
								const DrawParams& dp, float lineSpacing)
{
	if(! pTracker) {
		gr.DrawText(font, "Tracker not found", dtp);
		return;
	}

	// print positions
	const vec3& pos = pTracker->GetPosition();

	uint width = 12;
	uint precision = 4;

	DRAW_ONE_LINE("X: " + ToStringFix(pos.x, width, precision, true, true));
	DRAW_ONE_LINE("Y: " + ToStringFix(pos.y, width, precision, true, true));
	DRAW_ONE_LINE("Z: " + ToStringFix(pos.z, width, precision, true, true));
}


//====================================================================
bool  PositionTrackerStatePane::CacheHIDController(HIDController* hidc)
{
	pTracker = InterfaceCast<PositionTracker>(hidc);
	return (pTracker != 0);
}


//====================================================================
GloveStatePane::GloveStatePane(const std::string& _controllerName,
								uint _iGloveOrdinalNumber,
								const DrawParams& _dp,
								const gui::TextAttributes& _ta)
	:
	HIDControllerPane(VW_INTERFACE_NAME_Glove, _iGloveOrdinalNumber, _controllerName, _dp, _ta)
{
	SetSize(350, 230);

	pGlove = 0;
}


//====================================================================
void  GloveStatePane::DrawInfoText(gui::Renderer& gr,
								const gui::Font& font, gui::DrawTextParams& dtp,
								const DrawParams& dp, float lineSpacing)
{
	if(! pGlove) {
		gr.DrawText(font, "Glove not found", dtp);
		return;
	}

	// print general info
	DRAW_ONE_LINE("LeftHanded: " + ToString(pGlove->IsLeftHanded()));
	DRAW_ONE_LINE("AngleMeasurement: " + ToString(pGlove->IsSupportedAngleMeasurement()));

	// print gestures
	DRAW_ONE_LINE("ActiveGestures:");
	dtp.positionTopLeft.x += dp.textSpacingH * 2;

	uint maxActive = 4;
	uint numActive = 0;

	const static_gestures::Gesture* pActiveGesture = gestureAnalyzer.GetFirstActiveGesture(*pGlove);
	while(pActiveGesture) {
		numActive++;
		if(numActive > maxActive) break;
		DRAW_ONE_LINE(pActiveGesture->GetName());
		pActiveGesture = gestureAnalyzer.GetNextActiveGesture(*pGlove);
	}
	if(! numActive) {
		DRAW_ONE_LINE("[no active gestures]");
	}
}


//====================================================================
bool  GloveStatePane::CacheHIDController(HIDController* hidc)
{
	pGlove = InterfaceCast<Glove>(hidc);
	return (pGlove != 0);
}


//====================================================================
GameControllerStatePane::GameControllerStatePane(const std::string& _controllerName,
								uint _iGameControllerOrdinalNumber,
								const DrawParams& _dp,
								const gui::TextAttributes& _ta)
	:
	HIDControllerPane(VW_INTERFACE_NAME_GameController, _iGameControllerOrdinalNumber, _controllerName, _dp, _ta)
{
	SetSize(350, 320);

	pGameController = 0;
}


//====================================================================
bool  GameControllerStatePane::CacheHIDController(HIDController* hidc)
{
	pGameController = InterfaceCast<GameController>(hidc);
	return (pGameController != 0);
}


//====================================================================
void  GameControllerStatePane::DrawInfoText(gui::Renderer& gr,
									const gui::Font& font, gui::DrawTextParams& dtp,
									const DrawParams& dp, float lineSpacing)
{
	if(! pGameController) {
		gr.DrawText(font, "GameController not found", dtp);
		return;
	}

	// print some pressed buttons
	std::string s = "Buttons: ";
	uint maxButtons = Min<uint>(pGameController->GetNumButtons(), 8);
	for(uint i = 0; i < maxButtons; i++) {
		if(pGameController->ButtonPressed(i)) s.append(ToString(i));
	}
	DRAW_ONE_LINE(s);

	uint width = 10;
	uint precision = 4;

	// print standard axes
	DRAW_ONE_LINE(" X: " + ToStringFix(pGameController->GetAxisPosition(GameController::Axis_X), width, precision, true, true));
	DRAW_ONE_LINE(" Y: " + ToStringFix(pGameController->GetAxisPosition(GameController::Axis_Y), width, precision, true, true));
	DRAW_ONE_LINE(" Z: " + ToStringFix(pGameController->GetAxisPosition(GameController::Axis_Z), width, precision, true, true));
	DRAW_ONE_LINE("Rx: " + ToStringFix(pGameController->GetAxisPosition(GameController::Axis_RX), width, precision, true, true));
	DRAW_ONE_LINE("Ry: " + ToStringFix(pGameController->GetAxisPosition(GameController::Axis_RY), width, precision, true, true));
	DRAW_ONE_LINE("Rz: " + ToStringFix(pGameController->GetAxisPosition(GameController::Axis_RZ), width, precision, true, true));

	// print some sliders
	DRAW_ONE_LINE("Slider0: " + ToStringFix(pGameController->GetSliderPosition(0), width, precision, true, true));
	DRAW_ONE_LINE("Slider1: " + ToStringFix(pGameController->GetSliderPosition(1), width, precision, true, true));

	// print some pov-hats
	DRAW_ONE_LINE("POVHat0: " + ToStringFix(pGameController->GetPOVHatPosition(0), width, precision, true, true));
	DRAW_ONE_LINE("POVHat1: " + ToStringFix(pGameController->GetPOVHatPosition(1), width, precision, true, true));
}


//====================================================================
OrientationTrackerStatePane::OrientationTrackerStatePane(const std::string& _controllerName,
									uint _iOrientationTrackerOrdinalNumber,
									const DrawParams& _dp,
									const gui::TextAttributes& _ta)
	:
	HIDControllerPane(VW_INTERFACE_NAME_OrientationTracker, _iOrientationTrackerOrdinalNumber, _controllerName, _dp, _ta)
{
	SetSize(380, 270);

	bUseToolTip = true;
	pTracker = 0;

	SetToolTipText("R - reset orientation\nH - reset head\n"
					"S - change sampling mode\nT - tooltip on/off");
}


//====================================================================
bool  OrientationTrackerStatePane::OnQueryToolTipParams(gui::ToolTipParams& ttp)
{
	if(bUseToolTip) {
		return HIDControllerPane::OnQueryToolTipParams(ttp);
	}
	return false;
}


//====================================================================
bool  OrientationTrackerStatePane::OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch)
{
	if(bDown  &&  ! repeatCount) {
		switch(keyCode) {
			case Key_R: {
				if(pTracker) pTracker->ResetOrientation();
				return true;
			}
			case Key_T: {
				Flip(bUseToolTip);
				return true;
			}
			case Key_S: {
				if(pTracker) {
					OrientationTracker::eSamplingMode mode = pTracker->GetSamplingMode();
					if(mode == OrientationTracker::SamplingMode_Euler) {
						mode = OrientationTracker::SamplingMode_Quaternion;
					}
					else {
						mode = OrientationTracker::SamplingMode_Euler;
					}
					pTracker->SetSamplingMode(mode);
				}
				return true;
			}
			case Key_H: {
				if(pTracker) pTracker->ResetHead();
				return true;
			}
		}
	}
	return HIDControllerPane::OnKeyboardKey(bDown, keyCode, repeatCount, ch);
}


//====================================================================
void  OrientationTrackerStatePane::DrawInfoText(gui::Renderer& gr,
								const gui::Font& font, gui::DrawTextParams& dtp,
								const DrawParams& dp, float lineSpacing)
{
	if(! pTracker) {
		gr.DrawText(font, "Tracker not found", dtp);
		return;
	}

	// print caps
	DRAW_ONE_LINE("CanSampleEuler: " + ToString(pTracker->IsSupportedSamplingMode(OrientationTracker::SamplingMode_Euler)));
	DRAW_ONE_LINE("CanSampleQuaternion: " + ToString(pTracker->IsSupportedSamplingMode(OrientationTracker::SamplingMode_Quaternion)));
	DRAW_ONE_LINE("CanResetAll: " + ToString(pTracker->IsSupportedResetOrientation()));
	DRAW_ONE_LINE("CanResetHead: " + ToString(pTracker->IsSupportedResetHead()));

	std::string s = "SamplingMode: ";
	switch(pTracker->GetSamplingMode()) {
		case OrientationTracker::SamplingMode_Unknown:
			s.append("unknown");
			break;
		case OrientationTracker::SamplingMode_Euler:
			s.append("euler");
			break;
		case OrientationTracker::SamplingMode_Quaternion:
			s.append("quaternion");
			break;
	}
	DRAW_ONE_LINE(s);

	// print raw orientation
	switch(pTracker->GetSamplingMode()) {
		case OrientationTracker::SamplingMode_Euler: {
			uint width = 7;
			uint precision = 2;
			float h, p, r;
			pTracker->GetOrientationRaw(h, p, r);
			DRAW_ONE_LINE("RawHead:  " + ToStringFix(h, width, precision, true, true));
			DRAW_ONE_LINE("RawPitch: " + ToStringFix(p, width, precision, true, true));
			DRAW_ONE_LINE("RawRoll:  " + ToStringFix(r, width, precision, true, true));
			break;
		}
		case OrientationTracker::SamplingMode_Quaternion: {
			uint width = 8;
			uint precision = 5;
			quat q;
			pTracker->GetOrientationRaw(q);
			DRAW_ONE_LINE("RawQuatX: " + ToStringFix(q.x, width, precision, true, true));
			DRAW_ONE_LINE("RawQuatY: " + ToStringFix(q.y, width, precision, true, true));
			DRAW_ONE_LINE("RawQuatZ: " + ToStringFix(q.z, width, precision, true, true));
			DRAW_ONE_LINE("RawQuatW: " + ToStringFix(q.w, width, precision, true, true));
			break;
		}
	}
}


//====================================================================
bool  OrientationTrackerStatePane::CacheHIDController(HIDController* hidc)
{
	pTracker = InterfaceCast<OrientationTracker>(hidc);
	return (pTracker != 0);
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


