

//--------------------------------------------------------------------
#include "GUI_CS0\Menu.h"
#include "GUI\GUI_Manager.h"
#include "GUI\GUI_StdObjects.h"
#include "BaseTools\Constructs.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
MenuBar::GeomParams::GeomParams()
{
	height = 28;
	paddingLeft = 12;
	paddingRight = 4;
	spacingButtonTitle = 20;
	spacingTitles = 12;
	dockSensorWidth = 28;
}


//====================================================================
IconButton::GeomParams __fastcall MenuBar::DefaultGeomParams_IconButton()
{
	IconButton::GeomParams gp;
	gp.preferredSize = vec2(18, 18);
	gp.iconSize = vec2(16, 16);
	return gp;
}


//====================================================================
MenuBar::MenuBar()
{
	bEnableUndock = true;
	bDocked = false;
	bMovingInDockSensorArea = false;

	using namespace gui::std_tile_textures;

	pButtonClose = new IconButton(this, this, gui::CID_Close, gui::Button::Style_Push);
	pButtonClose->SetGeomParams(DefaultGeomParams_IconButton());
	pButtonClose->SetTileParams(MakeTileParams(Tile_X));

	SetMenuBarShaderNames();
	SetMenuBarSubcomponentToolTipText();
	SetMenuBarSubcomponentLayout();
	SetGeomParams();
	SetMenuBarPosition(true);

	pButtonClose->InstallMessageHook(*this);
}


//====================================================================
void  MenuBar::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  MenuBar::SetMenuBarShaderNames(const std::string& menuBar,
									const std::string& iconButtonBarControl)
{
	SetShaderName(menuBar);
	pButtonClose->SetShaderName(iconButtonBarControl);
}


//====================================================================
void  MenuBar::SetMenuBarSubcomponentLayout(bool bShowCloseButton)
{
	pButtonClose->SetVisible(bShowCloseButton);
	ReportDirtyGeometry();
}


//====================================================================
void  MenuBar::SetMenuBarSubcomponentToolTipText(const std::string& buttonClose)
{
	pButtonClose->SetToolTipText(buttonClose);
}


//====================================================================
void  MenuBar::EnableUndock(bool bEnable)
{
	bEnableUndock = bEnable;
	RealizeDock(bDocked, Component::GetPosition());
}


//====================================================================
void  MenuBar::SetMenuBarPosition(bool _bDock, const vec2& _position)
{
	RealizeDock(_bDock, _position);
}


//====================================================================
void  MenuBar::RecalcSize()
{
	float totalWidth = geomParams.paddingLeft;
	if(pButtonClose->IsVisible()) {
		totalWidth += geomParams.spacingButtonTitle + geomParams.paddingRight;
	}
	else {
		totalWidth += geomParams.paddingLeft;
	}

	if(bDocked) {
		// use full window width
		totalWidth = gui::Manager::Instance().GetWindowWidth();
	}
	else {
		// accumulate child sizes
		for(uint i = 0; i < GetNumChildren(); i++) {
			gui::Component& child = GetChild(i);
			if(child.IsVisible()) {
				totalWidth += child.GetSize().x;
			}
		}
		// add padding between menu-titles (ignore close button)
		if(GetNumChildren() >= 3) {
			totalWidth += (GetNumChildren() - 2) * geomParams.spacingTitles;
		}
	}
	SetSize(totalWidth, geomParams.height);
}


//====================================================================
void  MenuBar::RecalcLayout()
{
	float posX = geomParams.paddingLeft;

	for(uint i = 0; i < GetNumChildren(); i++) {
		gui::Component& child = GetChild(i);

		// handle close-button
		if(&child == pButtonClose) {
			child.SetPosition(GetSize().x - pButtonClose->GetSize().x - geomParams.paddingRight,
							(GetSize().y - child.GetSize().y) * 0.5f);
		}
		// handle menu-titles
		else {
			child.SetPosition(posX, (GetSize().y - child.GetSize().y) * 0.5f);
			posX += child.GetSize().x + geomParams.spacingTitles;
		}
	}
}


//====================================================================
void  MenuBar::UpdateGeometryCache()
{
	rDockSensor.Set(- geomParams.dockSensorWidth, - geomParams.dockSensorWidth,
					2 * geomParams.dockSensorWidth, 2 * geomParams.dockSensorWidth);
}


//====================================================================
void  MenuBar::OnWindowResize(uint width, uint height)
{
	// keep menu-bar inside window
	RealizeDock(bDocked, GetPosition());
}


//====================================================================
gui::eCommandArea  MenuBar::HitTestCommandAreaSelf(const vec2& testPoint) const
{
	if(bEnableUndock) return gui::CommandArea_Move;
	return gui::CommandArea_Client;
}


//====================================================================
bool  MenuBar::OnCommand(uint commandID)
{
	switch(commandID) {
		case gui::CID_Close: {
			ShowTopLevel(false);
			return true;
		}
	}
	return false;
}


//====================================================================
bool  MenuBar::OnGeometryUpdateByManager(bool& bConfirmEvent,
								gui::eGeometryUpdateEvent updateEvent,
								gui::eCommandArea commandArea, eMouseButton buttonCode,
								const vec2& newPosition, const vec2& newSize)
{
	switch(updateEvent) {
		case gui::GeometryUpdateEvent_MoveBegin: {
			CloseMenu();
			bMovingInDockSensorArea = rDockSensor.HitTest(GetPosition());
			break;
		}
		case gui::GeometryUpdateEvent_Moving: {
			bMovingInDockSensorArea = rDockSensor.HitTest(GetPosition());
			return true;
		}
		case gui::GeometryUpdateEvent_MoveEnd: {
			bMovingInDockSensorArea = false;
			bool bDock = rDockSensor.HitTest(GetPosition());
			RealizeDock(bDock, GetPosition());
			return true;
		}
	}
	return Component::OnGeometryUpdateByManager(bConfirmEvent, updateEvent,
							commandArea, buttonCode, newPosition, newSize);
}


//====================================================================
bool  MenuBar::OnActivate(bool bActivateByMouse, eMouseButton buttonCode,
							const vec2& mousePosition, gui::Component* pClickedDescendant,
							bool& bAutoBringToFront, bool& bAutoActivate)
{
	CloseMenu();
	return
	Component::OnActivate(bActivateByMouse, buttonCode, mousePosition,
					pClickedDescendant, bAutoBringToFront, bAutoActivate);
}


//====================================================================
bool  MenuBar::OnDeactivate(bool bDeactivateByMouse)
{
	CloseMenu();
	return true;
}


//====================================================================
bool  MenuBar::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(bDown  &&  buttonCode == MouseButton_Left  &&  ! repeatCount) {
		CloseMenu();
		return true;
	}
	return false;
}


//====================================================================
bool  MenuBar::OnMouseButton(gui::MessageHandler* target, bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(target == pButtonClose) {
		if(bDown) CloseMenu();
	}
	return false;
}


//====================================================================
void  MenuBar::OnChangedVisible()
{
	if(! IsVisible()) {
		CloseMenu();
	}
}


//====================================================================
void  MenuBar::CloseMenu()
{
	SendMessageToChildren(gui::Message_FlagsOnOff(MenuTitle::Flag_TitleOpened, false));
}


//====================================================================
void  MenuBar::RealizeDock(bool bDock, const vec2& undockedPosition)
{
	bDocked = bEnableUndock ? bDock : true;

	if(! bDocked) {
		SetPosition(undockedPosition);
	}

	// force dock at sensor
	if(rDockSensor.HitTest(GetPosition())) {
		bDocked = true;
	}

	// force safe boundaries
	if(! bDocked) {
		UpdateGeometry();

		if(! gui::Manager::Instance().IntersectsWindow(*this)) {
			bDocked = true;
		}
	}

	// update docked position
	if(bDocked) {
		SetPosition(vec2::ZERO);
	}

	// force update
	UpdateGeometry();
}


//====================================================================
MenuTitle::GeomParams::GeomParams()
{
	textHeight = 20;
	paneOffset = vec2(0, -1);
}


//====================================================================
MenuTitle::MenuTitle(MenuBar* _parent, const std::string& _text)
{
	SetShaderName(DefaultShaderName(Shader_MenuTitle));

	text = _text;

	SetCanDisable(true);
	SetGeomParams();

	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  MenuTitle::AttachToMenuBar(MenuBar& menuBar)
{
	Component::AttachTo(menuBar);
}


//====================================================================
void  MenuTitle::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  MenuTitle::SetText(const std::string& _text)
{
	text = _text;
	ReportDirtyGeometry();
}


//====================================================================
void  MenuTitle::RecalcSize()
{
	gui::ConstShaderPtr<MenuTitle::Shader> pShader = GetBaseShader();
	if(! pShader) return;

	const gui::Font* pFont = pShader->GetFont();
	if(! pFont) return;

	SetSize(pFont->CalcTextWidth(pShader->GetTextMetricsL(geomParams.textHeight), text), geomParams.textHeight);
}


//====================================================================
void  MenuTitle::RecalcLayout()
{
	float barHeight = GetParentSize().y;
	vec2  pos = vec2(0, (barHeight - GetSize().y) * 0.5f + GetSize().y);

	// update child-pane position
	if(GetNumChildren()) {
		GetChild(0).SetPosition(pos + geomParams.paneOffset);
	}
}


//====================================================================
bool  MenuTitle::OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos)
{
	if(bEnter) {
		// change active menu-title
		if(IsEnabled()  &&  IsTitleOpened()  &&  ! IsTitleActive()) {
			SendMessageToSiblings(gui::Message_FlagsOnOff(MenuTitle::Flag_TitleActive, false));
			SendMessageToSelf(gui::Message_FlagsOnOff(MenuTitle::Flag_TitleActive, true));
			return true;
		}
	}
	return false;
}


//====================================================================
bool  MenuTitle::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(bDown  &&  buttonCode == MouseButton_Left  &&  ! repeatCount) {
		// open/close menu
		SendMessageToSelfAndSiblings(gui::Message_FlagsOnOff(MenuTitle::Flag_TitleOpened, ! IsTitleOpened()));

		// update active menu-title
		if(IsTitleOpened()) {
			SendMessageToSiblings(gui::Message_FlagsOnOff(MenuTitle::Flag_TitleActive, false));
			SendMessageToSelf(gui::Message_FlagsOnOff(MenuTitle::Flag_TitleActive, true));
		}
		return true;
	}
	return false;
}


//====================================================================
bool  MenuTitle::OnMsgFlagsOnOff(uint flags, bool bOn)
{
	if(flags  &  MenuTitle::Flag_TitleOpened) {
		if(! bOn) flags |= MenuTitle::Flag_TitleActive;
	}
	return Component::OnMsgFlagsOnOff(flags, bOn);
}


//====================================================================
bool  MenuTitle::OnMsgPreCommandExecute(uint commandID)
{
	// close menu
	SendMessageToSelfAndSiblings(gui::Message_FlagsOnOff(MenuTitle::Flag_TitleOpened, false));
	return true;
}


//====================================================================
void  MenuTitle::OnChangedFlags()
{
	if(GetNumChildren()) {
		bool bShow = IsTitleOpened()  &&  IsTitleActive();
		// show/hide MenuPane attached to MenuTitle
		GetChild(0).SetVisible(bShow);
	}
}


//====================================================================
MenuPane::GeomParams::GeomParams()
{
	padding = vec2(2, 2);
	minWidth = 100;
}


//====================================================================
MenuPane::MenuPane()
{
	Init();
}


//====================================================================
MenuPane::MenuPane(MenuTitle& parent)
{
	Init();
	AttachToMenuTitle(parent);
}


//====================================================================
MenuPane::MenuPane(MenuItem& parent)
{
	Init();
	AttachToMenuItem(parent);
}


//====================================================================
void  MenuPane::Init()
{
	SetShaderName(DefaultShaderName(Shader_MenuPane));

	SetGeomParams();
	SetVisible(false);
}


//====================================================================
void  MenuPane::AttachToMenuTitle(MenuTitle& menuTitle)
{
	if(menuTitle.GetNumChildren()) {
		VW_THROW(GUIException, "Attaching MenuPane to occupied MenuTitle.");
	}
	Component::AttachTo(menuTitle);
	SetVisible(false);
}


//====================================================================
void  MenuPane::AttachToMenuItem(MenuItem& menuItem)
{
	if(menuItem.GetNumChildren()) {
		VW_THROW(GUIException, "Attaching MenuPane to occupied MenuItem.");
	}
	Component::AttachTo(menuItem);
	SetVisible(false);
}


//====================================================================
void  MenuPane::ShowAsContextMenu(bool bShow, const vec2& position)
{
	if(IsTopLevel()) {
		SetVisible(bShow);
		if(IsVisible()) {
			ActivateTopLevel(true);
			SetPosition(position);
			ReportDirtyGeometry();
		}
	}
}


//====================================================================
void  MenuPane::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  MenuPane::RecalcSize()
{
	// accumulate child sizes
	vec2 totalSize = vec2::ZERO;

	{for(uint i = 0; i < GetNumChildren(); i++) {
		gui::Component& child = GetChild(i);

		totalSize.x = Max(totalSize.x, child.GetSize().x);
		totalSize.y += child.GetSize().y;
	}}

	// calculate own size
	totalSize += 2 * geomParams.padding;
	totalSize.x = Max(totalSize.x, geomParams.minWidth);

	SetSize(totalSize);

	float fMaxItemWidth = totalSize.x - 2 * geomParams.padding.x;

	// adjust child sizes
	{for(uint i = 0; i < GetNumChildren(); i++) {
		GetChild(i).SetSizeX(fMaxItemWidth);
	}}
}


//====================================================================
void  MenuPane::RecalcLayout()
{
	// adjust child positions
	vec2 pos = geomParams.padding;

	for(uint i = 0; i < GetNumChildren(); i++) {
		gui::Component& child = GetChild(i);
		child.SetPosition(pos);
		pos.y += child.GetSize().y;
	}
}


//====================================================================
void  MenuPane::OnWindowResize(uint width, uint height)
{
	// auto-hide context-menu pane
	if(IsTopLevel()) {
		ShowTopLevel(false);
	}
}


//====================================================================
void  MenuPane::OnChangedVisible()
{
	if(IsVisible()) {
		// hide descendant MenuPanes attached to MenuItems
		SendMessageToChildren(gui::Message_ShowChildren(false));

		ReportDirtyGeometry();
	}
}


//====================================================================
bool  MenuPane::OnDeactivate(bool bDeactivateByMouse)
{
	if(IsTopLevel()) {
		// auto-hide if used as context menu
		ShowTopLevel(false);
		return true;
	}
	return Component::OnDeactivate(bDeactivateByMouse);
}


//====================================================================
bool  MenuPane::OnMsgPreCommandExecute(uint commandID)
{
	if(IsTopLevel()) {
		// auto-hide if used as context menu
		ShowTopLevel(false);
		return true;
	}
	return Component::OnMsgPreCommandExecute(commandID);
}


//====================================================================
MenuSeparator::GeomParams::GeomParams()
{
	height = 10;
}


//====================================================================
MenuSeparator::MenuSeparator(MenuPane* _parent)
{
	SetShaderName(DefaultShaderName(Shader_MenuSeparator));

	SetGeomParams();

	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  MenuSeparator::AttachToMenuPane(MenuPane& menuPane)
{
	Component::AttachTo(menuPane);
}


//====================================================================
void  MenuSeparator::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  MenuSeparator::RecalcSize()
{
	SetSize(0, geomParams.height);
}


//====================================================================
MenuItem::GeomParams::GeomParams()
{
	textHeight = 20;
	itemHeight = 24;
	padding = 1;
	spacingIconText = 2;
	iconSize = 20;
	paneOffset = vec2(-1, 0);
}


//====================================================================
MenuItem::MenuItem(MenuPane* _parent,
					gui::MessageHandler* _commandTarget,
					uint _commandID,
					const std::string& _text,
					eStyle _style,
					uint _groupID)
{
	SetShaderName(DefaultShaderName(Shader_MenuItem));

	style = _style;
	groupID = _groupID;

	SetCanDisable(true);
	SetCommandTarget(_commandTarget);
	SetCommandID(_commandID);

	SetGeomParams();
	SetText(_text);

	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  MenuItem::AttachTo(MenuPane& menuPane)
{
	Component::AttachTo(menuPane);
}


//====================================================================
void  MenuItem::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  MenuItem::SetText(const std::string& _text)
{
	text = _text;
	ReportDirtyGeometry();
}


//====================================================================
void  MenuItem::SetStyle(eStyle _style)
{
	if(GetNumChildren()) _style = Style_Default;

	style = _style;

	SetOn(false);
	SetSelected(false);
}


//====================================================================
void  MenuItem::SetOn(bool bOn)
{
	if(style != Style_OnOff  ||  GetNumChildren()) {
		bOn = false;
	}
	SetFlagsOnOff(Flag_On, bOn);
	SetFlagsOnOff(Flag_Selected, false);
}


//====================================================================
void  MenuItem::SetSelected(bool bSelected)
{
	if(style != Style_Selection ||  GetNumChildren()) {
		bSelected = false;
	}
	SetFlagsOnOff(Flag_Selected, bSelected);
	SetFlagsOnOff(Flag_On, false);
}


//====================================================================
void  MenuItem::OnAttachedChild(gui::Component& child)
{
	SetStyle(Style_Default);
}


//====================================================================
void  MenuItem::RecalcSize()
{
	gui::ConstShaderPtr<MenuItem::Shader> pShader = GetBaseShader();
	if(! pShader) return;

	const gui::Font* pFont = pShader->GetFont();
	if(! pFont) return;

	SetSize(
		(geomParams.padding + geomParams.iconSize + geomParams.spacingIconText) * 2 +
		pFont->CalcTextWidth(pShader->GetTextMetricsL(geomParams.textHeight), text),
		geomParams.itemHeight);
}


//====================================================================
void  MenuItem::RecalcLayout()
{
	// update child-pane position
	if(GetNumChildren()) {
		MenuPane* pPane = dynamic_cast<MenuPane*>(GetParent());

		if(pPane) {
			float offsetX = pPane->GetSize().x - pPane->GetGeomParams().padding.x;
			GetChild(0).SetPosition(vec2(offsetX, 0) + geomParams.paneOffset);
		}
	}
}


//====================================================================
bool  MenuItem::OnMouseEnterExit(bool bEnter, bool bMouseOverDirect, const vec2& mousePos)
{
	if(bEnter  &&  IsEnabled()) {
		// show MenuPane attached to self & hide MenuPanes attached to sibling-MenuItems
		SendMessageToSelf(gui::Message_ShowChildren(true));
		SendMessageToSiblings(gui::Message_ShowChildren(false));
		return true;
	}
	return false;
}


//====================================================================
bool  MenuItem::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(! bDown  &&  buttonCode == MouseButton_Left) {
		if(! GetNumChildren()  &&  IsEnabled()) {

			switch(style) {
				case Style_OnOff:
					SetOn(! IsOn());
					break;

				case Style_Selection:
					SendMessageToSelf(gui::Message_Select(true, groupID));
					SendMessageToSiblings(gui::Message_Select(false, groupID));
					break;
			}
			SendMessageToParent(gui::Message_PreCommandExecute(GetCommandID()));
			SendCommandMessageToTarget();
			return true;
		}
	}
	return false;
}


//====================================================================
bool  MenuItem::OnMsgOnOff(bool bOn)
{
	if(style == Style_OnOff) {
		SetOn(bOn);
		return true;
	}
	return false;
}


//====================================================================
bool  MenuItem::OnMsgSelect(bool bSelect, uint _groupID)
{
	if(style == Style_Selection  &&  groupID == _groupID) {
		SetSelected(bSelect);
		return true;
	}
	return false;
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


