

//--------------------------------------------------------------------
#include "GUI_CS0\PropertyFrame.h"
#include "GUI_CS0\ShaderNames.h"
#include "GUI\GUI_Manager.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
PropertyFrame::GeomParams::GeomParams()
{
	tabOffsetLeft = 2;
	tabOffsetLeftSelected = 0;
	tabSpacing = 2;
	tabSpacingSelected = 0;
	tabSelectedExtend = vec2(4, 2);
	tabTextHeight = 18;
	tabPadding = 8;
	tabMinSize = vec2(0, 24);
}


//====================================================================
PropertyFrame::PropertyFrame(gui::Component* _parent)
{
	EnableAutoUpdateControlsOnSelect();

	SetGhost(true);
	SetGeomParams();
	SetSize(300, 200);

	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  PropertyFrame::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  PropertyFrame::UpdateControlDependencies(bool bSelectedOnly)
{
	for(uint i = 0; i < GetNumChildren(); i++) {
		gui::ComponentPtr<PropertyTab> pTab(GetChild(i));
		if(pTab  &&  (pTab->IsSelected()  ||  ! bSelectedOnly)) {
			pTab->UpdateControlDependencies();
		}
	}
}


//====================================================================
void  PropertyFrame::ApplyControlValues(bool bSelectedOnly)
{
	for(uint i = 0; i < GetNumChildren(); i++) {
		gui::ComponentPtr<PropertyTab> pTab(GetChild(i));
		if(pTab  &&  (pTab->IsSelected()  ||  ! bSelectedOnly)) {
			pTab->ApplyControlValues();
		}
	}
}


//====================================================================
void  PropertyFrame::OnAttachedChild(gui::Component& child)
{
	gui::ComponentPtr<PropertyTab> pTab(child);
	if(pTab) {
		pTab->Select();
	}
	ReportDirtyGeometry();
}


//====================================================================
void  PropertyFrame::RecalcLayout()
{
	vec2  pos = vec2::ZERO;
	bool  bPrevSelected = false;
	float spacing = 0;

	for(uint i = 0; i < GetNumChildren(); i++) {
		gui::Component& c = GetChild(i);

		if(c.IsFlagOn(PropertyTab::Flag_Selected)) {
			bPrevSelected = true;
			spacing = geomParams.tabSpacingSelected;
		}
		else if(bPrevSelected) {
			bPrevSelected = false;
			spacing = geomParams.tabSpacingSelected;
		}
		else {
			spacing = geomParams.tabSpacing;
		}

		if(i == 0) {
			spacing = c.IsFlagOn(PropertyTab::Flag_Selected) ?
						geomParams.tabOffsetLeftSelected : geomParams.tabOffsetLeft;
		}

		pos.x += spacing;
		pos.y = c.IsFlagOn(PropertyTab::Flag_Selected) ? 0 : geomParams.tabSelectedExtend.y;

		GetChild(i).SetPosition(pos);

		pos.x += GetChild(i).GetSize().x;
	}
}


//====================================================================
PropertyTab::PropertyTab(PropertyFrame* _parent,
						gui::MessageHandler* _commandTarget,
						uint _commandID,
						const std::string& _text)
{
	pFrame = 0;
	pPage = 0;

	SetShaderName(DefaultShaderName(Shader_PropertyTab));
	SetText(_text);

	SetCommandTarget(_commandTarget);
	SetCommandID(_commandID);

	if(_parent) AttachToPropertyFrame(*_parent);
}


//====================================================================
void  PropertyTab::SetText(const std::string& _text)
{
	text = _text;
	ReportDirtyGeometry();
}


//====================================================================
void  PropertyTab::Select()
{
	if(! IsSelected()) {
		SendMessageToSiblings(gui::Message_Select(false));
		SendMessageToSelf(gui::Message_Select(true));

		ReportDirtyGeometry();

		// notify target about tab-selection
		SendCommandMessageToTarget();
	}
	if(pFrame  &&  pFrame->IsEnabledAutoUpdateControlsOnSelect()) {
		UpdateControlDependencies();
	}
}


//====================================================================
void  PropertyTab::UpdateControlDependencies()
{
	if(pPage) {
		pPage->UpdateControlDependencies();
	}
}


//====================================================================
void  PropertyTab::ApplyControlValues()
{
	if(pPage) {
		// apply then automatically pull-back
		pPage->ApplyControlValues();
		pPage->UpdateControlDependencies();
	}
}


//====================================================================
void  PropertyTab::OnAttachedToParent()
{
	pFrame = gui::ComponentPtr<PropertyFrame>(GetParent());
}


//====================================================================
void  PropertyTab::OnDetachedFromParent()
{
	pFrame = 0;
}


//====================================================================
void  PropertyTab::OnAttachedChild(gui::Component& child)
{
	pPage = gui::ComponentPtr<PropertyPage>(child);
	if(pPage) {
		pPage->SetVisible(IsSelected());
	}
}


//====================================================================
void  PropertyTab::OnDetachedChild(gui::Component& child)
{
	if(pPage == &child) pPage = 0;
}


//====================================================================
void  PropertyTab::RecalcSize()
{
	if(! pFrame) return;
	const PropertyFrame::GeomParams& gp = pFrame->GetGeomParams();

	gui::ConstShaderPtr<PropertyTab::Shader> pShader = GetBaseShader();
	if(! pShader) return;

	// calc & cache text-size
	const gui::Font* pFont = pShader->GetFont();
	if(! pFont) return;

	gui::TextMetricsL tml = pShader->GetTextMetricsL(gp.tabTextHeight);
	textSize.Set(pFont->CalcTextWidth(tml, text), tml.textHeight);

	unselectedSize = ComponentMax(gp.tabMinSize, textSize + vec2(2 * gp.tabPadding, 0));

	vec2 size = unselectedSize;
	if(IsSelected()) {
		size += gp.tabSelectedExtend;
	}
	SetSize(size);
}


//====================================================================
void  PropertyTab::RecalcLayout()
{
	if(! pPage || ! pFrame) return;

	const PropertyFrame::GeomParams& gp = pFrame->GetGeomParams();

	vec2 pagePos(- GetPosition().x, GetSize().y);
	vec2 pageSize = GetParentSize();
	pageSize.y -= unselectedSize.y + gp.tabSelectedExtend.y;

	pPage->SetSize(pageSize);
	pPage->SetPosition(pagePos);
	pPage->SetTabPosition(GetPosition().x, GetSize().x);
}


//====================================================================
bool  PropertyTab::OnCommand(uint commandID)
{
	// forward command when title-label is pushed
	if(commandID == gui::CID_SendCommand) {
		SendCommandMessageToTarget();
		return true;
	}
	return Component::OnCommand(commandID);
}


//====================================================================
bool  PropertyTab::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(bDown  &&  ! repeatCount  &&  buttonCode == MouseButton_Left) {
		Select();
	}
	return false;
}


//====================================================================
bool  PropertyTab::OnMsgSelect(bool bSelect, uint groupID)
{
	SetSelected(bSelect);

	if(GetNumChildren()) {
		GetChild(0).SetVisible(IsSelected());
	}
	return true;
}


//====================================================================
PropertyPage::PropertyPage(PropertyTab* _parent)
{
	SetShaderName(DefaultShaderName(Shader_PropertyPage));
	SetTabPosition(0, 0);

	if(_parent) AttachToPropertyTab(*_parent);
}


//====================================================================
void  PropertyPage::SetTabPosition(float _tabOffset, float _tabLength)
{
	tabOffset = _tabOffset;
	tabLength = _tabLength;
}


//====================================================================
void  PropertyPage::GetTabPosition(float& _tabOffset, float& _tabLength) const
{
	_tabOffset = tabOffset;
	_tabLength = tabLength;
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


