

//--------------------------------------------------------------------
#include "GUI_CS0\ScrollBar.h"
#include "GUI_CS0\ShaderNames.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
ScrollBarThumb::ScrollBarThumb(gui_cs0::ScrollBar& _sbar, bool _bReportVerticalOffset)
	:
	Thumb(_bReportVerticalOffset, &_sbar),
	sbar(_sbar)
{
}


//====================================================================
void  ScrollBarThumb::ProcessThumbGrabBegin()
{
	sbar.ReportThumbGrabBegin();
}


//====================================================================
void  ScrollBarThumb::ProcessThumbGrabOffset(float offset)
{
	sbar.ReportThumbGrabOffset(offset);
}


//====================================================================
ScrollBarButton::ScrollBarButton(ScrollBar& _sbar, eButtonID _id)
	:
	PressButton(&_sbar),
	sbar(_sbar)
{
	buttonID = _id;
}


//====================================================================
void  ScrollBarButton::ProcessButtonPress()
{
	SendMessageTo(sbar, gui::Message_Scroll(IsButtonForward(), gui::ScrollDistance_Line));
}


//====================================================================
ScrollBar::GeomParams::GeomParams(bool bVertical)
{
	if(bVertical) {
		preferredSize = vec2(28, 240);
	}
	else {
		preferredSize = vec2(240, 28);
	}
	buttonLength = 28;
	minThumbLength = 10;
}


//====================================================================
ScrollBar::ScrollBar(gui::Component* _parent,
					gui::MessageHandler* _commandTarget,
					uint _commandID,
					bool _bVertical)
{
	SetShaderName(DefaultShaderName(Shader_ScrollBar));

	bVertical = _bVertical;
	bActiveAutoEnableDisable = true;
	deltaPerWheelUnit = 3;

	SetCanDisable(true);
	SetCanFocus(true);

	pButtonBackward = new ScrollBarButton(*this, bVertical ? ScrollBarButton::ButtonID_Up : ScrollBarButton::ButtonID_Left);
	pButtonForward = new ScrollBarButton(*this, bVertical ? ScrollBarButton::ButtonID_Down : ScrollBarButton::ButtonID_Right);
	pThumb = new ScrollBarThumb(*this, bVertical);

	ScrollGadget::SetTotalSize(100);
	ScrollGadget::SetVisibleSize(40);
	ScrollGadget::SetPosition(0);

	SetCommandTarget(_commandTarget);
	SetCommandID(_commandID);

	GeomParams gp(bVertical);
	SetGeomParams(gp);

	if(_parent) AttachTo(*_parent);
}


//====================================================================
void  ScrollBar::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ClampLow<float>(geomParams.buttonLength, 1);
	ClampLow<float>(geomParams.minThumbLength, 1);

	ScrollGadget::SetMinThumbLength(geomParams.minThumbLength);

	ReportDirtyGeometry();
}


//====================================================================
void  ScrollBar::EnableMaxPixelsPerStep(bool bEnable)
{
	ScrollGadget::EnableMaxPixelsPerStep(bEnable);
	UpdateBar();
}


//====================================================================
void  ScrollBar::ActivateAutoEnableDisable(bool bActivate)
{
	bActiveAutoEnableDisable = bActivate;
	UpdateBar();
}


//====================================================================
void  ScrollBar::SetDeltaPerMouseWheelUnit(int delta)
{
	deltaPerWheelUnit = Abs(delta);
}


//====================================================================
void  ScrollBar::ResetLogicalScrollParameters()
{
	ScrollGadget::ResetLogicalParameters();
	UpdateBar();
}


//====================================================================
void  ScrollBar::SetTotalScrollSize(uint total)
{
	ScrollGadget::SetTotalSize(total);
	UpdateBar();
}


//====================================================================
void  ScrollBar::SetVisibleScrollSize(uint visible)
{
	ScrollGadget::SetVisibleSize(visible);
	UpdateBar();
}


//====================================================================
void  ScrollBar::SetScrollPosition(int pos)
{
	ScrollGadget::SetPosition(pos);
	UpdateThumbGeometry();
}


//====================================================================
void  ScrollBar::SetScrollPositionToMax()
{
	ScrollGadget::SetPositionToMax();
	UpdateThumbGeometry();
}


//====================================================================
void  ScrollBar::OffsetScrollPosition(int offset)
{
	ScrollGadget::OffsetPosition(offset);
	UpdateThumbGeometry();
}


//====================================================================
void  ScrollBar::GetDrawRectangles(gui::Rectangle& rBase,
								gui::Rectangle& rThumb,
								gui::Rectangle& rButtonBackward,
								gui::Rectangle& rButtonForward) const
{
	rBase.Set(vec2::ZERO, GetSize());
	rThumb.Set(pThumb->GetPosition(), pThumb->GetSize());
	rButtonBackward.Set(pButtonBackward->GetPosition(), pButtonBackward->GetSize());
	rButtonForward.Set(pButtonForward->GetPosition(), pButtonForward->GetSize());
}


//====================================================================
void  ScrollBar::RecalcSize()
{
	SetSize(geomParams.preferredSize);
}


//====================================================================
void  ScrollBar::RecalcLayout()
{
	vec2  buttonSize;
	if(bVertical) {
		buttonSize.Set(GetSize().x, geomParams.buttonLength);
	}
	else {
		buttonSize.Set(geomParams.buttonLength, GetSize().y);
	}
	pButtonBackward->SetSize(buttonSize);
	pButtonBackward->SetPosition(vec2::ZERO);

	pButtonForward->SetSize(buttonSize);
	if(bVertical) {
		pButtonForward->SetPosition(0, GetSize().y - pButtonForward->GetSize().y);
	}
	else {
		pButtonForward->SetPosition(GetSize().x - pButtonForward->GetSize().x, 0);
	}
}


//====================================================================
void  ScrollBar::UpdateGeometryCache()
{
	UpdateBar();
}


//====================================================================
bool  ScrollBar::OnMouseWheel(int delta, const vec2& mousePos)
{
	SendMessageToSelf(gui::Message_Scroll(true, gui::ScrollDistance_Offset,
					- delta * GetDeltaPerMouseWheelUnit()));
	return true;
}


//====================================================================
bool  ScrollBar::OnMouseButton(bool bDown, eMouseButton buttonCode, bool bDoubleClick, uint repeatCount, const vec2& mousePos)
{
	if(bDown  &&  buttonCode == MouseButton_Left) {

		if(bVertical) {
			const vec2& thumbPos = pThumb->GetPosition();
			const vec2& thumbSize = pThumb->GetSize();

			if(mousePos.y < thumbPos.y) {
				SendMessageToSelf(gui::Message_Scroll(false, gui::ScrollDistance_Page));
			}
			else if(mousePos.y > thumbPos.y + thumbSize.y) {
				SendMessageToSelf(gui::Message_Scroll(true, gui::ScrollDistance_Page));
			}
		}
		else {
			const vec2& thumbPos = pThumb->GetPosition();
			const vec2& thumbSize = pThumb->GetSize();

			if(mousePos.x < thumbPos.x) {
				SendMessageToSelf(gui::Message_Scroll(false, gui::ScrollDistance_Page));
			}
			else if(mousePos.x > thumbPos.x + thumbSize.x) {
				SendMessageToSelf(gui::Message_Scroll(true, gui::ScrollDistance_Page));
			}
		}
		return true;
	}
	return false;
}


//====================================================================
bool  ScrollBar::OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData)
{
	DefaultProcessQueryFocus(bFocusByMouse, buttonCode, bWantFocus, bNeedCharacterData);
	return true;
}


//====================================================================
bool  ScrollBar::OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch)
{
	if(! bDown) return false;

	if(pThumb->IsGrabbed()) {
		pThumb->ReleaseGrab();
	}

	switch(keyCode) {
		case Key_Home:
			SendMessageToSelf(gui::Message_Scroll(false, gui::ScrollDistance_ToEnd));
			return true;

		case Key_End:
			SendMessageToSelf(gui::Message_Scroll(true, gui::ScrollDistance_ToEnd));
			return true;

		case Key_PageUp:
			SendMessageToSelf(gui::Message_Scroll(false, gui::ScrollDistance_Page));
			return true;

		case Key_PageDown:
			SendMessageToSelf(gui::Message_Scroll(true, gui::ScrollDistance_Page));
			return true;

		case Key_ArrowUp:
		case Key_ArrowLeft:
			SendMessageToSelf(gui::Message_Scroll(false, gui::ScrollDistance_Line));
			return true;

		case Key_ArrowDown:
		case Key_ArrowRight:
			SendMessageToSelf(gui::Message_Scroll(true, gui::ScrollDistance_Line));
			return true;
	}
	return false;
}


//====================================================================
void  ScrollBar::OnChangedEnabled()
{
	if(bActiveAutoEnableDisable) return; //break recursive-call
	DoEnable(IsEnabled()); //update sub-components
	UpdateBar();
}


//====================================================================
bool  ScrollBar::OnMsgScroll(bool bForward, gui::eScrollDistance distance, int offset)
{
	uint lastScrollPos = ScrollGadget::GetPosition();

	switch(distance) {
		case gui::ScrollDistance_Line: {
			ScrollGadget::OffsetPosition(bForward ? 1 : -1);
			break;
		}
		case gui::ScrollDistance_Page: {
			ScrollGadget::OffsetPosition((bForward ? 1 : -1) * ScrollGadget::GetVisibleSize());
			break;
		}
		case gui::ScrollDistance_ToEnd: {
			if(bForward) {
				ScrollGadget::SetPositionToMax();
			}
			else {
				ScrollGadget::SetPosition(0);
			}
			break;
		}
		case gui::ScrollDistance_Offset: {
			ScrollGadget::OffsetPosition(bForward ? offset : -offset);
			break;
		}
	}

	if(ScrollGadget::GetPosition() != lastScrollPos) {
		UpdateThumbGeometry();
		SendCommandMessageToTarget();
	}
	return true;
}


//====================================================================
void  ScrollBar::ReportThumbGrabBegin()
{
	if(bVertical) {
		thumbPixelOffsetAtGrab = pThumb->GetPosition().y - geomParams.buttonLength;
	}
	else {
		thumbPixelOffsetAtGrab = pThumb->GetPosition().x - geomParams.buttonLength;
	}
}


//====================================================================
void  ScrollBar::ReportThumbGrabOffset(float offset)
{
	uint  lastScrollPos = ScrollGadget::GetPosition();

	// move thumb by pixel-offset accumulated by mouse,
	// then convert new thumb-pixel-offset to logical scrollbar position
	float maxThumbOffset = ScrollGadget::GetBarLength() - ScrollGadget::GetThumbLength();
	float thumbOffset = Clamped<int>(thumbPixelOffsetAtGrab + offset, 0, maxThumbOffset);
	float scrollPos = maxThumbOffset ? (thumbOffset / (maxThumbOffset - 1)) : 0;

	// move thumb to new pixel-offset
	if(bVertical) {
		pThumb->SetPosition(0, thumbOffset + geomParams.buttonLength);
	}
	else {
		pThumb->SetPosition(thumbOffset + geomParams.buttonLength, 0);
	}

	UpdateAbsolutePositionCache();

	// update logical position
	ScrollGadget::SetPosition(scrollPos * GetMaxPosition());
	if(ScrollGadget::GetPosition() != lastScrollPos) {
		SendCommandMessageToTarget();
	}
}


//====================================================================
void  ScrollBar::DoEnable(bool bEnable)
{
	pButtonBackward->SetEnabled(bEnable);
	pButtonForward->SetEnabled(bEnable);
	pThumb->SetEnabled(bEnable);

	Component::SetEnabled(bEnable);
}


//====================================================================
void  ScrollBar::UpdateBar()
{
	// update bar geometry-length
	if(bVertical) {
		SetBarLength(GetSize().y - geomParams.buttonLength * 2);
	}
	else {
		SetBarLength(GetSize().x - geomParams.buttonLength * 2);
	}

	// realize auto-enable/disable
	if(IsActiveAutoEnableDisable()) {
		if(IsEnabled()) {
			if(! ScrollGadget::IsVisibleLessThanTotal()) DoEnable(false);
		}
		else {
			if(ScrollGadget::IsVisibleLessThanTotal()) DoEnable(true);
		}
	}

	// update thumb geometry
	UpdateThumbGeometry();
}


//====================================================================
void  ScrollBar::UpdateThumbGeometry()
{
	ScrollGadget::UpdateThumbParameters();

	bool bForceMaximize = ! IsEnabled();

	if(bVertical) {
		if(bForceMaximize) {
			pThumb->SetSize(GetSize().x, GetSize().y - geomParams.buttonLength * 2);
			pThumb->SetPosition(0, geomParams.buttonLength);
		}
		else {
			vec2  thumbPos;

			if(IsScrollPositionAtMaximum()) {
				thumbPos.Set(0, GetSize().y - geomParams.buttonLength - GetThumbLength());
			}
			else {
				thumbPos.Set(0, geomParams.buttonLength + GetThumbOffset());
			}
			pThumb->SetSize(GetSize().x, GetThumbLength());
			pThumb->SetPosition(thumbPos);
		}
	}
	else {
		if(bForceMaximize) {
			pThumb->SetSize(GetSize().x - geomParams.buttonLength * 2, GetSize().y);
			pThumb->SetPosition(geomParams.buttonLength, 0);
		}
		else {
			vec2  thumbPos;

			if(IsScrollPositionAtMaximum()) {
				thumbPos.Set(GetSize().x - geomParams.buttonLength - GetThumbLength(), 0);
			}
			else {
				thumbPos.Set(geomParams.buttonLength + GetThumbOffset(), 0);
			}
			pThumb->SetSize(GetThumbLength(), GetSize().y);
			pThumb->SetPosition(thumbPos);
		}
	}
	UpdateAbsolutePositionCache();
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


