

//--------------------------------------------------------------------
#include "GUI_CS0\ToolTip.h"
#include "GUI_CS0\ShaderNames.h"
#include "GUI\GUI_Manager.h"
#include "GUI\GUI_Tools.h"


//====================================================================
#define  ADDITIONAL_VERTICAL_OFFSET_TOP   2


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
ToolTip::GeomParams::GeomParams()
{
	textPadding = vec2(4, 4);
}


//====================================================================
ToolTip::ToolTip()
{
	SetShaderName(DefaultShaderName(Shader_ToolTip));

	animState = AnimationState_Stable;
	alphaCurr = 0;
	alphaCurrB = 0;
	mouseMoveIgnoreArea.Set(0, 0, 0, 0);
	pLastHoverCheckTarget = 0;
	pLastQueryTarget = 0;

	SetMinHoverTime();
	SetFadeTimes();

	SetGhost(true);
	SetGeomParams();
	SetText("ToolTip");

	Hide();
	gui::Manager::Instance().CycleQueueRemove(*this);
}


//====================================================================
void  ToolTip::SetMinHoverTime(double _secsMinHoverTime)
{
	secsMinHoverTime = _secsMinHoverTime;
}


//====================================================================
void  ToolTip::SetFadeTimes(double _secsFadeIn, double _secsFadeOut)
{
	secsFadeIn = ClampedLow<double>(_secsFadeIn, 0);
	secsFadeOut = ClampedLow<double>(_secsFadeOut, 0);

	alphaPerSecFadeIn = secsFadeIn ? (1 / secsFadeIn) : -1;
	alphaPerSecFadeOut = secsFadeOut ? (1 / secsFadeOut) : -1;
}


//====================================================================
void  ToolTip::GetFadeTimes(double& _secsFadeIn, double& _secsFadeOut)
{
	_secsFadeIn = secsFadeIn;
	_secsFadeOut = secsFadeOut;
}


//====================================================================
void  ToolTip::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;
	ReportDirtyGeometry();
}


//====================================================================
void  ToolTip::SetText(const std::string& _text)
{
	text = _text;
	BreakIntoLines(text, textLines);
	textLineWidths.resize(textLines.size());
	RecalcSize();
}


//====================================================================
void  ToolTip::Show(const vec2& _position,
					const vec2& _mouseMoveIgnoreAreaPosition,
					const vec2& _mouseMoveIgnoreAreaSize)
{
	mouseMoveIgnoreArea.Set(_mouseMoveIgnoreAreaPosition, _mouseMoveIgnoreAreaSize);

	// force into window
	vec2  windowSize = gui::Manager::Instance().GetWindowSize();
	vec2  position = _position;

	if(position.x + GetSize().x > windowSize.x) {
		position.x = windowSize.x - GetSize().x;
	}
	ClampLow<float>(position.x, 0);

	if(position.y + GetSize().y > windowSize.y) {
		position.y = windowSize.y - GetSize().y;
	}
	ClampLow<float>(position.y, 0);

	SetPosition(position);
	UpdateAbsolutePositionCache();

	// setup animation state
	animState = AnimationState_FadeIn;

	if(! IsVisible()) {
		alphaCurr = 0;
	}
	UpdateTranslucencyCache();

	// do show
	Component::ShowTopLevel(true, true);
}


//====================================================================
void  ToolTip::Hide()
{
	Component::ShowTopLevel(false);
}


//====================================================================
void  ToolTip::ServiceProviderPreEventDispatch(double secsTimeDelta)
{
	if(! IsVisible()) return;

	gui::Manager& gm = gui::Manager::Instance();

	switch(animState) {
		case AnimationState_FadeIn: {
			if(alphaPerSecFadeIn < 0) {
				alphaCurr = 1;
			}
			else {
				alphaCurr = ClampedHigh<float>(alphaCurr + alphaPerSecFadeIn * secsTimeDelta, 1);
			}
			if(alphaCurr >= 1) {
				animState = AnimationState_Stable;
			}
			UpdateTranslucencyCache();
			// fall through
		}
		case AnimationState_Stable: {
			// watch for mouse & keyboard events
			if(gm.IsKeyDown()  ||  gm.IsButtonDown()  ||
				! mouseMoveIgnoreArea.HitTest(gm.GetMousePosition()))
			{
				animState = AnimationState_FadeOut;
			}
			break;
		}
		case AnimationState_FadeOut: {
			if(alphaPerSecFadeOut < 0) {
				alphaCurr = 0;
			}
			else {
				alphaCurr = ClampedLow<float>(alphaCurr - alphaPerSecFadeOut * secsTimeDelta, 0);
			}
			if(alphaCurr <= 0) {
				Hide();
			}
			UpdateTranslucencyCache();
			break;
		}
	}
}


//====================================================================
void  ToolTip::ServiceProviderPostEventDispatch(double secsTimeDelta)
{
	gui::Manager& gm = gui::Manager::Instance();

	// check for hover target change
	gui::MessageHandler* pTarget = gm.GetMouseOverTarget();

	if(! pTarget) {
		pLastHoverCheckTarget = 0;
		pLastQueryTarget = 0;
		return;
	}
	if(pTarget != pLastHoverCheckTarget) {
		pLastQueryTarget = 0;
	}
	if(pTarget == pLastQueryTarget) return;
	pLastHoverCheckTarget = pTarget;

	if(! IsVisible()) {
		if(gm.GetMouseHoverTime() < secsMinHoverTime) return;
	}
	pLastQueryTarget = pTarget;

	// get tool-tip parameters from target
	vec2  mousePos = gm.GetMousePosition();

	ttParams.bDisplay = false;
	ttParams.bAutoPosition = true;
	ttParams.position = mousePos;
	ttParams.mouseMoveIgnoreAreaPosition = mousePos;
	ttParams.mouseMoveIgnoreAreaSize = vec2::ONE;
	ttParams.minSize = vec2::ZERO;

	if(! SendMessageTo(*pLastQueryTarget, gui::Message_ServiceCommand_ToolTip(
										gui::SCID_QueryParams, &ttParams))) return;
	if(! ttParams.bDisplay) return;

	// apply new text (auto-recalc new size)
	SetText(ttParams.text);
	SetSize(ComponentMax(GetSize(), ttParams.minSize));

	if(ttParams.bAutoPosition) {
		ttParams.position.x = mousePos.x;

		// apply vertical offset depending on screen position
		const gui::Cursor& cursor = gui::Manager::Instance().GetActiveCursor();

		float offsetY = (1 - cursor.GetHotSpot().y) * cursor.GetActiveSize().y;
		ttParams.position.y = mousePos.y + offsetY;

		if(ttParams.position.y + GetSize().y >= gui::Manager::Instance().GetWindowHeight()) {
			ttParams.position.y = mousePos.y - GetSize().y - ADDITIONAL_VERTICAL_OFFSET_TOP;
		}
	}

	// show tool-tip
	Show(ttParams.position, ttParams.mouseMoveIgnoreAreaPosition, ttParams.mouseMoveIgnoreAreaSize);
}


//====================================================================
bool  ToolTip::HandleMessage(gui::Message& msg)
{
	using namespace gui;

	// handle service commands
	if(msg.messageID == MID_ServiceCommand) {
		MsgPtr<Message_ServiceCommand> m(msg);

		if(m->GetServiceID() == SID_ToolTip) {
			MsgPtr<Message_ServiceCommand_ToolTip> mtt(msg);

			// reset hover-target cache
			if(mtt->GetServiceCommandID() == SCID_Reset) {
				pLastHoverCheckTarget = 0;
				pLastQueryTarget = 0;
				return true;
			}
		}
	}
	return Component::HandleMessage(msg);
}


//====================================================================
void  ToolTip::RecalcSize()
{
	gui::ConstShaderPtr<ToolTip::Shader> pShader = GetBaseShader();
	if(! pShader) return;

	// calc & cache text-size
	const gui::Font* pFont = pShader->GetFont();
	if(! pFont) return;

	gui::CalcTextSize(*pFont, pShader->GetTextMetricsL(), textLines,
						textSize, &textLineWidths);

	// update size
	SetSize(textSize + geomParams.textPadding * 2);
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


