

//--------------------------------------------------------------------
#include "GUI_CS0\TopLevelFrame.h"
#include "GUI\GUI_Manager.h"
#include "GUI\GUI_Tools.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
TopLevelFrame::GeomParams::GeomParams()
{
	bResizable = false;
	minSize = vec2(250, 100);
	maxSize = vec2(640, 480);
	titleBarHeight = 28;
	titleTextHeight = 20;
	titleOffsetLeft = 4;
	frameBorderWidth = 4;
	resizeAreaWidth = 4;
	resizeCornerExtents = vec2(20, 20);
	workAreaPaddingH = 8;
	workAreaPaddingT = 8;
	workAreaPaddingB = 8;
	buttonFrameControlsOffsetRight = 4;
	buttonFrameControlsSpacing = 2;
	buttonFrameControlsSize = 18;
	buttonFrameControlsIconSize = 16;
}


//====================================================================
TopLevelFrame::TopLevelFrame(const std::string& _title)
{
	bMaximized = false;
	sizePreMaximized = GetSize();
	positionPreMaximized = GetPosition();

	SetVisible(false);
	SetCanFocus(true);

	// Create components
	pLayouterTLFControls = new gui::layouters::RowLayouter(this);

	pButtonClose = new IconButton(pLayouterTLFControls, this, gui::CID_Close, gui::Button::Style_Push);
	pButtonMaximize = new IconButton(pLayouterTLFControls, this, gui::CID_Maximize, gui::Button::Style_Push);
	pButtonRestore = new IconButton(pLayouterTLFControls, this, gui::CID_Restore, gui::Button::Style_Push);

	SetSize(400, 300);
	SetGeomParams();
	SetTitleText(_title);
	SetTLFSubcomponentLayout();
	SetTLFSubcomponentToolTipText();
	SetTLFSubcomponentTileParams();
	SetTLFShaderNames();
	SetCloseFrameOnKey();
	SetMaximizeRestoreFrameOnKey();
}


//====================================================================
void  TopLevelFrame::SetGeomParams(const GeomParams& gp)
{
	geomParams = gp;

	// process frame params
	geomParams.minSize = ComponentMax(geomParams.minSize,
			vec2(2 * geomParams.frameBorderWidth, geomParams.titleBarHeight + geomParams.frameBorderWidth));

	geomParams.maxSize = ComponentMax(geomParams.minSize, geomParams.maxSize);

	vec2 size = GetSize();
	ComponentClamp(size, geomParams.minSize, geomParams.maxSize);

	SetMinSize(geomParams.minSize);
	SetMaxSize(geomParams.maxSize);

	if(bMaximized) {
		sizePreMaximized = size;
	}
	else {
		SetSize(size);
	}

	// process control params
	IconButton::GeomParams  gpIconButton;
	gpIconButton.preferredSize = vec2(geomParams.buttonFrameControlsSize, geomParams.buttonFrameControlsSize);
	gpIconButton.iconSize = vec2(geomParams.buttonFrameControlsIconSize, geomParams.buttonFrameControlsIconSize);

	pButtonClose->SetGeomParams(gpIconButton);
	pButtonMaximize->SetGeomParams(gpIconButton);
	pButtonRestore->SetGeomParams(gpIconButton);

	UpdateMaximizeRestoreButtonVisibility();
	ReportDirtyGeometry();
}


//====================================================================
void  TopLevelFrame::SetResizable(bool bResizable)
{
	GeomParams gp;
	gp.bResizable = bResizable;
	SetGeomParams(gp);
}


//====================================================================
void  TopLevelFrame::SetPositionTopLevelFrame(const vec2& _position)
{
	if(bMaximized) {
		positionPreMaximized = _position;
	}
	else {
		SetPosition(_position);
	}
}


//====================================================================
void  TopLevelFrame::SetTLFShaderNames(const std::string& topLevelFrame,
										const std::string& iconButtonFrameControl)
{
	SetShaderName(topLevelFrame);
	pButtonClose->SetShaderName(iconButtonFrameControl);
	pButtonMaximize->SetShaderName(iconButtonFrameControl);
	pButtonRestore->SetShaderName(iconButtonFrameControl);
}


//====================================================================
void  TopLevelFrame::SetTLFSubcomponentLayout(bool _bShowTitle,
											bool _bShowMaximizeRestoreButton,
											bool _bShowCloseButton)
{
	bShowTitle = _bShowTitle;
	bShowMaximizeRestoreButton = _bShowMaximizeRestoreButton;
	bShowCloseButton = _bShowCloseButton;

	pButtonClose->SetVisible(bShowCloseButton);
	UpdateMaximizeRestoreButtonVisibility();
}


//====================================================================
void  TopLevelFrame::SetTLFSubcomponentToolTipText(
									const std::string& buttonClose,
									const std::string& buttonMaximize,
									const std::string& buttonRestore)
{
	pButtonClose->SetToolTipText(buttonClose);
	pButtonMaximize->SetToolTipText(buttonMaximize);
	pButtonRestore->SetToolTipText(buttonRestore);
}


//====================================================================
void  TopLevelFrame::SetTLFSubcomponentTileParams(
									const gui::TileParams& buttonClose,
									const gui::TileParams& buttonMaximize,
									const gui::TileParams& buttonRestore)
{
	pButtonClose->SetTileParams(buttonClose);
	pButtonMaximize->SetTileParams(buttonMaximize);
	pButtonRestore->SetTileParams(buttonRestore);
}


//====================================================================
void  TopLevelFrame::SetCloseFrameOnKey(bool bCloseOnKey, eKey kc)
{
	bCloseFrameOnKey = bCloseOnKey;
	kcClose = kc;
}


//====================================================================
void  TopLevelFrame::SetMaximizeRestoreFrameOnKey(bool bMaxRestoreOnKey, eKey kc)
{
	bMaximizeRestoreFrameOnKey = bMaxRestoreOnKey;
	kcMaximizeRestore = kc;
}


//====================================================================
vec2  TopLevelFrame::CalcTLFFrameSize() const
{
	return vec2(geomParams.frameBorderWidth * 2, geomParams.titleBarHeight + geomParams.frameBorderWidth);
}


//====================================================================
vec2  TopLevelFrame::CalcTLFWorkAreaFrameSize() const
{
	return CalcTLFFrameSize() +
	vec2(geomParams.workAreaPaddingH * 2, geomParams.workAreaPaddingT + geomParams.workAreaPaddingB);
}


//====================================================================
vec2  TopLevelFrame::CalcTLFMainAreaOffset() const
{
	return vec2(geomParams.frameBorderWidth, geomParams.titleBarHeight);
}


//====================================================================
vec2  TopLevelFrame::CalcTLFMainAreaSize() const
{
	return GetSize() -
	vec2(geomParams.frameBorderWidth * 2, geomParams.titleBarHeight + geomParams.frameBorderWidth);
}


//====================================================================
void  TopLevelFrame::CalcTLFMainArea(vec2& size,
							vec2& topLeft, vec2& topRight,
							vec2& bottomLeft, vec2& bottomRight) const
{
	size = CalcTLFMainAreaSize();
	topLeft = CalcTLFMainAreaOffset();
	gui::CalcPointsFromSizeAndTopLeft(size, topLeft, topRight, bottomLeft, bottomRight);
}


//====================================================================
vec2  TopLevelFrame::CalcTLFWorkAreaOffset() const
{
	return vec2(geomParams.frameBorderWidth + geomParams.workAreaPaddingH,
				geomParams.titleBarHeight + geomParams.workAreaPaddingT);
}


//====================================================================
vec2  TopLevelFrame::CalcTLFWorkAreaSize() const
{
	return GetSize() -
	vec2((geomParams.frameBorderWidth + geomParams.workAreaPaddingH) * 2,
		geomParams.titleBarHeight + geomParams.frameBorderWidth + geomParams.workAreaPaddingT + geomParams.workAreaPaddingB);
}


//====================================================================
void  TopLevelFrame::CalcTLFWorkArea(vec2& size,
							vec2& topLeft, vec2& topRight,
							vec2& bottomLeft, vec2& bottomRight) const
{
	size = CalcTLFWorkAreaSize();
	topLeft = CalcTLFWorkAreaOffset();
	gui::CalcPointsFromSizeAndTopLeft(size, topLeft, topRight, bottomLeft, bottomRight);
}


//====================================================================
void  TopLevelFrame::RecalcLayout()
{
	pLayouterTLFControls->SetPosition(0, 0);
	pLayouterTLFControls->SetSize(GetSize().x, geomParams.titleBarHeight);
}


//====================================================================
void  TopLevelFrame::UpdateGeometryCache()
{
	rTitleBar.Set(vec2::ZERO, vec2(GetSize().x, geomParams.titleBarHeight));

	titleOffset.Set(geomParams.titleOffsetLeft,
				(geomParams.titleBarHeight - geomParams.titleTextHeight) * 0.5f);
}


//====================================================================
void  TopLevelFrame::OnWindowResize(uint width, uint height)
{
	if(bMaximized) {
		SetSize(width, height);
		SetPosition(vec2::ZERO);
		return;
	}
	Component::OnWindowResize(width, height);
}


//====================================================================
gui::eCommandArea  TopLevelFrame::HitTestCommandAreaSelf(const vec2& testPoint) const
{
	if(bMaximized) {
		return gui::CommandArea_Client;
	}
	if(geomParams.bResizable) {
		const vec2& ce = geomParams.resizeCornerExtents;

		if(testPoint.x < geomParams.resizeAreaWidth) {
			if(testPoint.y < ce.y) return gui::CommandArea_ResizeNW;
			if(testPoint.y >= GetSize().y - ce.y) return gui::CommandArea_ResizeSW;
			return gui::CommandArea_ResizeW;
		}
		else if(testPoint.x >= GetSize().x - geomParams.resizeAreaWidth) {
			if(testPoint.y < ce.y) return gui::CommandArea_ResizeNE;
			if(testPoint.y >= GetSize().y - ce.y) return gui::CommandArea_ResizeSE;
			return gui::CommandArea_ResizeE;
		}
		else if(testPoint.y < geomParams.resizeAreaWidth) {
			if(testPoint.x < ce.x) return gui::CommandArea_ResizeNW;
			if(testPoint.x >= GetSize().x - ce.x) return gui::CommandArea_ResizeNE;
			return gui::CommandArea_ResizeN;
		}
		else if(testPoint.y >= GetSize().y - geomParams.resizeAreaWidth) {
			if(testPoint.x < ce.x) return gui::CommandArea_ResizeSW;
			if(testPoint.x >= GetSize().x - ce.x) return gui::CommandArea_ResizeSE;
			return gui::CommandArea_ResizeS;
		}
	}
	if(rTitleBar.HitTest(testPoint)) {
		return gui::CommandArea_Move;
	}
	return gui::CommandArea_Client;
}


//====================================================================
bool  TopLevelFrame::OnCommand(uint commandID)
{
	switch(commandID) {

		case gui::CID_Close: {
			ShowTopLevel(false);
			OnFrameClose();
			return true;
		}
		case gui::CID_Maximize: {
			if(geomParams.bResizable  &&  bShowMaximizeRestoreButton) {
				SetMaximized(true);
				OnFrameMaximize();
				return true;
			}
			break;
		}
		case gui::CID_Restore: {
			if(geomParams.bResizable  &&  bShowMaximizeRestoreButton) {
				SetMaximized(false);
				OnFrameRestore();
				return true;
			}
			break;
		}
	}
	return false;
}


//====================================================================
bool  TopLevelFrame::OnGeometryUpdateByManager(bool& bConfirmEvent,
								gui::eGeometryUpdateEvent updateEvent,
								gui::eCommandArea commandArea, eMouseButton buttonCode,
								const vec2& newPosition, const vec2& newSize)
{
	if(updateEvent == gui::GeometryUpdateEvent_QueryBegin) {
		if(geomParams.bResizable  ||  commandArea == gui::CommandArea_Move) {
			bConfirmEvent = (buttonCode == MouseButton_Left);
		}
		else {
			bConfirmEvent = false;
		}
		return true;
	}
	return false;
}


//====================================================================
bool  TopLevelFrame::OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData)
{
	DefaultProcessQueryFocus(bFocusByMouse, buttonCode, bWantFocus, bNeedCharacterData);
	return true;
}


//====================================================================
bool  TopLevelFrame::OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch)
{
	if(! bDown  ||  repeatCount) return false;

	if(keyCode == kcClose  &&  bCloseFrameOnKey) {
		SendMessageToSelf(gui::Message_Command(gui::CID_Close));
		return true;
	}
	if(keyCode == kcMaximizeRestore  &&  bMaximizeRestoreFrameOnKey  &&  bShowMaximizeRestoreButton) {
		SendMessageToSelf(gui::Message_Command(bMaximized ? gui::CID_Restore : gui::CID_Maximize));
		return true;
	}
	return false;
}


//====================================================================
void  TopLevelFrame::SetMaximized(bool _bMaximized)
{
	if(! geomParams.bResizable) _bMaximized = false;

	// store pre-maximized position
	if(! bMaximized) {
		sizePreMaximized = GetSize();
		positionPreMaximized = GetPosition();
	}

	// update flag
	bMaximized = _bMaximized;

	// update states
	if(bMaximized) {
		SetSize(gui::Manager::Instance().GetWindowSize());
		SetPosition(vec2::ZERO);
	}
	else {
		SetSize(CalcMinMaxClampedSize(sizePreMaximized));
		SetPosition(positionPreMaximized);
	}
	UpdateMaximizeRestoreButtonVisibility();
	UpdateGeometry();
}


//====================================================================
void  TopLevelFrame::UpdateMaximizeRestoreButtonVisibility()
{
	if(geomParams.bResizable  &&  bShowMaximizeRestoreButton) {
		pButtonMaximize->SetVisible(! bMaximized);
		pButtonRestore->SetVisible(bMaximized);
	}
	else {
		pButtonMaximize->SetVisible(false);
		pButtonRestore->SetVisible(false);
	}
	BuildLayout();
}


//====================================================================
void  TopLevelFrame::BuildLayout()
{
	using namespace gui::layouters;
	RowLayouter& l = *pLayouterTLFControls;

	l.Clear();
	l.FixedHeightRow(geomParams.titleBarHeight).AutoSpacer();
	Row& lastRow = l.LastRow();

	bool bClose = pButtonClose->IsVisible();
	bool bMaximize = pButtonMaximize->IsVisible();
	bool bRestore = pButtonRestore->IsVisible();

	if(bMaximize) {
		lastRow.Target(pButtonMaximize);
	}
	if(bRestore) {
		lastRow.Target(pButtonRestore);
	}
	if(bClose) {
		if(bMaximize || bRestore) {
			lastRow.Spacer(geomParams.buttonFrameControlsSpacing);
		}
		lastRow.Target(pButtonClose);
	}
	lastRow.Spacer(geomParams.buttonFrameControlsOffsetRight);

	OnTLFControlsLayouterUpdate();
	ReportDirtyGeometry();
}


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


