

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Layouters.h"
#include "GUI_CS0\Buttons.h"
#include "GUI_CS0\Label.h"
#include "GUI_CS0\ShaderNames.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
// Feature support:
// CanFocus.
//
// Internal hierarchy:
// * TopLevelFrame
//   * IconButton (Close)
//   * IconButton (Maximize)
//   * IconButton (Restore)
//====================================================================
class VW_GUI_CS0_API TopLevelFrame : public gui::Component
{
	VW_GUI_COMPONENT_DECLARE_SHADER_DLL(VW_GUI_CS0_API, Shader);
	VW_GUI_COMPONENT_DECLARE_CHECK_SHADER(Shader);


public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		bool   bResizable;
		vec2   minSize;
		vec2   maxSize;
		float  titleBarHeight;
		float  titleTextHeight;
		float  titleOffsetLeft;
		float  frameBorderWidth;
		float  resizeAreaWidth;
		vec2   resizeCornerExtents;
		float  workAreaPaddingH;
		float  workAreaPaddingT;
		float  workAreaPaddingB;
		float  buttonFrameControlsOffsetRight;
		float  buttonFrameControlsSpacing;
		float  buttonFrameControlsSize;
		float  buttonFrameControlsIconSize;

		GeomParams();
		VW_DECLARE_GUI_PARAMS(GeomParams);
	};


public:
	//-------------------------
	// TopLevelFrame interface.
	//-------------------------
	TopLevelFrame(const std::string& title = "TopLevelFrame");

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() const { return geomParams; }

	void  SetResizable(bool bResizable);

	//-----------------------------------------------------------------
	// Use SetPositionTopLevelFrame() instead of standard SetPosition()
	// to stay in sync with current maximized/restored states.
	//-----------------------------------------------------------------
	void  SetPositionTopLevelFrame(const vec2& position = vec2::ZERO);
	void  SetPositionTopLevelFrame(float x = 0, float y = 0) { SetPositionTopLevelFrame(vec2(x, y)); }

	void                SetTitleText(const std::string& text = "TopLevelFrame") { titleText = text; }
	const std::string&  GetTitleText() const { return titleText; }

	bool         IsTitleVisible() const { return bShowTitle; }
	const vec2&  GetTitleOffset() const { return titleOffset; }

	void  SetTLFShaderNames(
				const std::string& topLevelFrame = DefaultShaderName(Shader_TopLevelFrame),
				const std::string& iconButtonFrameControl = DefaultShaderName(Shader_TopLevelFrame_IconButton_FrameControl_Push));

	void  SetTLFSubcomponentLayout(
				bool bShowTitle = true,
				bool bShowMaximizeRestoreButton = true,
				bool bShowCloseButton = true);

	void  SetTLFSubcomponentToolTipText(
				const std::string& buttonClose = "Close",
				const std::string& buttonMaximize = "Maximize",
				const std::string& buttonRestore = "Restore");

	void  SetTLFSubcomponentTileParams(
				const gui::TileParams& buttonClose = gui::std_tile_textures::MakeTileParams(gui::std_tile_textures::Tile_X),
				const gui::TileParams& buttonMaximize = gui::std_tile_textures::MakeTileParams(gui::std_tile_textures::Tile_Maximize),
				const gui::TileParams& buttonRestore = gui::std_tile_textures::MakeTileParams(gui::std_tile_textures::Tile_Restore));

	void  SetCloseFrameOnKey(bool bCloseOnKey = true, eKey kc = Key_Escape);
	void  SetMaximizeRestoreFrameOnKey(bool bMaxRestoreOnKey = true, eKey kc = Key_F12);

	void  SetMaximized(bool bMaximized);
	bool  IsMaximized() const { return bMaximized; }

	// area used by title-bar & border
	vec2  CalcTLFFrameSize() const;

	// area used by title-bar, border & work-area-padding
	vec2  CalcTLFWorkAreaFrameSize() const;

	// area not used by title-bar & border
	vec2  CalcTLFMainAreaOffset() const;
	vec2  CalcTLFMainAreaSize() const;

	void  CalcTLFMainArea(vec2& size,
						vec2& topLeft, vec2& topRight,
						vec2& bottomLeft, vec2& bottomRight) const;

	// area not used by title-bar, border & work-area-padding
	vec2  CalcTLFWorkAreaOffset() const;
	vec2  CalcTLFWorkAreaSize() const;

	void  CalcTLFWorkArea(vec2& size,
						vec2& topLeft, vec2& topRight,
						vec2& bottomLeft, vec2& bottomRight) const;


protected:
	//---------------------
	// Component overrides.
	//---------------------
	virtual  void  RecalcLayout();
	virtual  void  UpdateGeometryCache();
	virtual  void  OnWindowResize(uint width, uint height);
	virtual  gui::eCommandArea  HitTestCommandAreaSelf(const vec2& testPoint) const;
	virtual  bool  OnCommand(uint commandID);

	virtual  bool  OnGeometryUpdateByManager(bool& bConfirmEvent,
								gui::eGeometryUpdateEvent updateEvent,
								gui::eCommandArea commandArea, eMouseButton buttonCode,
								const vec2& newPosition, const vec2& newSize);

	virtual  bool  OnKeyboardQueryFocus(bool bFocusByMouse, eMouseButton buttonCode, bool& bWantFocus, bool& bNeedCharacterData);
	virtual  bool  OnKeyboardKey(bool bDown, eKey keyCode, uint repeatCount, char ch);

	//------------------
	// Internal helpers.
	//------------------
	void  UpdateMaximizeRestoreButtonVisibility();

	//----------------
	// Layout control.
	//----------------
	void           BuildLayout();
	virtual  void  OnTLFControlsLayouterUpdate() {}

	//--------------------------------
	// Derived-class template-methods.
	//--------------------------------
	virtual  void  OnFrameClose() {}
	virtual  void  OnFrameMaximize() {}
	virtual  void  OnFrameRestore() {}


protected:
	//----------------
	// Sub-components.
	//----------------
	gui::layouters::RowLayouter*  pLayouterTLFControls;

	IconButton*  pButtonClose;
	IconButton*  pButtonMaximize;
	IconButton*  pButtonRestore;

	//------------
	// Properties.
	//------------
	GeomParams   geomParams;
	std::string  titleText;
	bool         bShowTitle;
	bool         bShowMaximizeRestoreButton;
	bool         bShowCloseButton;
	bool         bCloseFrameOnKey;
	eKey         kcClose;
	bool         bMaximizeRestoreFrameOnKey;
	eKey         kcMaximizeRestore;

	//--------
	// States.
	//--------
	bool  bMaximized;

	//-------
	// Cache.
	//-------
	gui::Rectangle  rTitleBar;
	vec2            titleOffset;
	vec2            sizePreMaximized;
	vec2            positionPreMaximized;
};


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


