

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "GUI\GUI_Layouters.h"
#include "GUI_CS0\Dialog.h"
#include "Framework\VideoProperties.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_cs0 {


//====================================================================
class VW_GUI_CS0_API VideoConfigDialog : public gui_cs0::Dialog, public ReconfigObserver
{
public:
	//----------------
	// Parameter sets.
	//----------------
	struct VW_GUI_CS0_API GeomParams
	{
		float  smallCheckBoxSize;
		float  smallLabelTextHeight;

		GeomParams();
	};


public:
	//-----------------------------
	// VideoConfigDialog interface.
	//-----------------------------
	VideoConfigDialog(const std::string& title = "Video Configuration",
						uint videoReconfigCode = 0);

	void               SetGeomParams(const GeomParams& gp = GeomParams());
	const GeomParams&  GetGeomParams() { return geomParams; }

	void  SetVideoConfigDialogShaderNames(
				const std::string& label = DefaultShaderName(Shader_Label_Static),
				const std::string& labelCB = DefaultShaderName(Shader_Label_Active),
				const std::string& checkBox = DefaultShaderName(Shader_CheckBox),
				const std::string& comboBox = DefaultShaderName(Shader_ComboBox),
				const std::string& horzTextSeparator = DefaultShaderName(Shader_HorzTextSeparator));

	void  SetVideoConfigDialogSubcomponentLayout(bool bShowDeveloperControls);

	void  SetVideoConfigDialogSubcomponentToolTipText(
				const std::string& sRenderer = "Select video-renderer",
				const std::string& sHardware = "Enable hardware rendering",
				const std::string& sDisplayMode = "Select display mode\nor window size",
				const std::string& sRefreshFreq = "Select refresh rate\n(fullscreen only)",
				const std::string& sFullscreen = "Fullscreen/windowed",
				const std::string& sTripple = "Use tripple buffering",
				const std::string& sVSync = "Use vertical synchronization (hint)",
				const std::string& sFrameBuffer = "Select frame-buffer configuration\n(RGBA / DepthStencil / AccumRGBA)",
				const std::string& sVertProc = "Select vertex-processing mode (hint)",
				const std::string& sResizable = "Enable window resize\n(windowed-mode only)",
				const std::string& sPure = "Use pure-device mode (hint)",
				const std::string& sDestAlpha = "Enable fullscreen\ndestination alpha (hint)",
				const std::string& sDynamicVSync = "Enable dynamic\nvsync update");

	// code used when calling Kernel::RequestVideoReconfig()
	void  SetVideoReconfigCode(uint rc = 0) { iVideoReconfigCode = rc; }


protected:
	//----------------------------
	// ReconfigObserver overrides.
	//----------------------------
	virtual  void  OnVideoReconfig();
	virtual  void  OnVSyncReconfig(bool bVSyncEnabled);

	//---------------------
	// Component overrides.
	//---------------------
	virtual  bool  OnCommand(uint commandID);

	//------------------
	// Dialog overrides.
	//------------------
	virtual  void  OnStdDialogControlsLayouterUpdate();
	virtual  void  OnDialogOpened();
	virtual  void  OnDialogClosed() {}
	virtual  void  OnApplySettings();

	//---------------------
	// Command identifiers.
	//---------------------
	enum {
		CID_Renderer = gui::CID_User,
		CID_Hardware,
		CID_Fullscreen,
		CID_Tripple,
		CID_VSync,
		CID_DisplayMode,
		CID_Refresh,
		CID_FrameBuffer,
		CID_VertProc,
		CID_Resizable,
		CID_Pure,
		CID_DestAlpha,
		CID_DynamicVSync
	};

	//------------------
	// Internal helpers.
	//------------------
	bool  GetActiveRendererInfo(VideoRendererInfo& vri, VideoConfig& vc);
	bool  GetSelectedRendererVideoProfile(VideoProfile& vp);
	void  RefillRenderers();
	void  RefillDisplayModes();
	void  RefillRefreshFreq();
	void  RefillFrameBuffer();
	void  SelectActiveRenderer();
	void  SelectSwitchesClosestToActive();
	void  SelectDisplayModeClosestToActive();
	void  SelectRefreshClosestToActive();
	void  SelectFrameBufferClosestToActive();
	void  SelectVertProcClosestToActive();
	void  UpdateStandardButtons();

	//-------------------------
	// Derived-class interface.
	//-------------------------
	virtual  void  OnPreApplySettings(std::string& vrPluginName, VideoConfig& vc) {}


private:
	//------------
	// Properties.
	//------------
	uint        iVideoReconfigCode;
	GeomParams  geomParams;

	//----------------
	// Sub-components.
	//----------------
	gui::layouters::RowLayouter*  pLayouter;
	Label*                        pLabelDynamicVSync;
	CheckBox*                     pCheckBoxDynamicVSync;

	// renderer controls
	HorzTextSeparator*  pHorzTextSeparatorRenderer;
	Label*              pLabelHardware;
	CheckBox*           pCheckBoxHardware;
	ComboBox*           pComboBoxRenderer;

	// display controls
	HorzTextSeparator*  pHorzTextSeparatorDisplay;
	Label*              pLabelDisplayMode;
	Label*              pLabelRefreshFreq;
	Label*              pLabelFullscreen;
	Label*              pLabelVSync;
	Label*              pLabelTripple;
	CheckBox*           pCheckBoxFullscreen;
	CheckBox*           pCheckBoxVSync;
	CheckBox*           pCheckBoxTripple;
	ComboBox*           pComboBoxDisplayMode;
	ComboBox*           pComboBoxRefreshFreq;

	// developer controls
	HorzTextSeparator*  pHorzTextSeparatorDeveloper;
	Label*              pLabelFrameBuffer;
	Label*              pLabelVertProc;
	Label*              pLabelResizable;
	Label*              pLabelPure;
	Label*              pLabelDestAlpha;
	CheckBox*           pCheckBoxResizable;
	CheckBox*           pCheckBoxPure;
	CheckBox*           pCheckBoxDestAlpha;
	ComboBox*           pComboBoxFrameBuffer;
	ComboBox*           pComboBoxVertProc;
};


//********************************************************************
}//namespace gui_cs0


//********************************************************************
}//namespace vertexware


