

//--------------------------------------------------------------------
#include "GUI_SS0\CheckBox.h"
#include "GUI\GUI_Manager.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_ss0 {


//====================================================================
CS0_CheckBoxShader_Opaque::ShaderParams::ShaderParams()
{
	colorDisabled = color4b::LIGHT_GRAY;
	colorDisabledBgr = color4b::GRAY;
	colorEnabledBorder = color4b::BLACK;
	colorEnabledIcon = color4b::BLACK;
	colorEnabledBgrTop = color4b(128, 128, 128);
	colorEnabledBgrBottom = color4b(192, 192, 192);
	colorPressedBgrTop = color4b(64, 64, 64);
	colorPressedBgrBottom = color4b(128, 128, 128);
	colorHotBorder = color4b(255, 255, 0, 255);
	iconPadding = vec2(3, 3);
	pressOffset = vec2(1, 1);
	mainBorderWidth = 1;
	hotBorderWidth = 2;
	tileParams = gui::std_tile_textures::MakeTileParams(gui::std_tile_textures::Tile_Check);
}


//====================================================================
void  CS0_CheckBoxShader_Opaque::ShaderParams::ChooseParams(
									const gui_cs0::CheckBox& checkBox,
									color4b& cTop, color4b& cBottom,
									color4b& cIcon, color4b& cMainBorder,
									vec2& pressOff) const
{
	// choose offset
	pressOff = checkBox.IsPressed() ? pressOffset : vec2::ZERO;

	// choose colors
	if(checkBox.IsEnabled()) {
		if(checkBox.IsPressed()) {
			cTop = colorPressedBgrTop;
			cBottom = colorPressedBgrBottom;
		}
		else {
			cTop = colorEnabledBgrTop;
			cBottom = colorEnabledBgrBottom;
		}
		cIcon = colorEnabledIcon;
		cMainBorder = colorEnabledBorder;
	}
	else {
		cTop = colorDisabledBgr;
		cBottom = colorDisabledBgr;
		cIcon = colorDisabled;
		cMainBorder = colorDisabled;
	}
}


//====================================================================
CS0_CheckBoxShader_Opaque::CS0_CheckBoxShader_Opaque(const std::string& shaderName,
												const ShaderParams& sp)
	:
	gui_cs0::CheckBox::Shader(shaderName)
{
	pTileTexture = 0;
	SetShaderParams(sp);
}


//====================================================================
void  CS0_CheckBoxShader_Opaque::SetShaderParams(const ShaderParams& sp)
{
	shaderParams = sp;
	iconWidth = shaderParams.iconPadding * 2;
	RequestTileTextureRecache();
}


//====================================================================
void  CS0_CheckBoxShader_Opaque::Shade(const gui::Component& c, gui::Renderer& gr) const
{
	const gui_cs0::CheckBox& checkBox = gui::ComponentCast<gui_cs0::CheckBox>(c);

	if(! pTileTexture) return;

	// choose params
	color4b  colorTop, colorBottom, colorIcon, colorMainBorder;
	vec2     pressOffset;

	shaderParams.ChooseParams(checkBox, colorTop, colorBottom, colorIcon,
								colorMainBorder, pressOffset);

	gui::RectangleC  rBorder(pressOffset, checkBox.GetSize());

	// draw background
	rBorder.SetColorTop(colorTop);
	rBorder.SetColorBottom(colorBottom);

	gr.DrawRectangle(rBorder);

	// draw icon
	if(checkBox.IsOn()) {

		rTile.Set(pressOffset + shaderParams.iconPadding, checkBox.GetSize() - iconWidth);
		rTile.SetColor(colorIcon);
		gr.DrawRectangle(*pTileTexture, rTile);
	}

	// draw main border
	rBorder.SetColor(colorMainBorder);
	gr.DrawRectangleFrame(rBorder, shaderParams.mainBorderWidth);

	// draw hot border
	if(checkBox.IsPressed()) return;

	if(checkBox.IsHot()) {
		rBorder.SetColor(shaderParams.colorHotBorder);
		rBorder.Shrink(shaderParams.mainBorderWidth);
		gr.DrawRectangleFrame(rBorder, shaderParams.hotBorderWidth);
	}
}


//====================================================================
bool  CS0_CheckBoxShader_Opaque::CacheTileTextureHandles(gui::Manager& gm)
{
	pTileTexture = gm.GetTileTexture(shaderParams.tileParams.texName);
	return (pTileTexture != 0);
}


//====================================================================
void  CS0_CheckBoxShader_Opaque::PostCacheTileTextureHandles()
{
	if(pTileTexture) {
		pTileTexture->GetTileCoords(shaderParams.tileParams.texRow, shaderParams.tileParams.texCol, rTile);
	}
	rTile.SetTexTransform(shaderParams.tileParams.texTransform);
}


//====================================================================
CS0_CheckBoxShader_Translucent::ShaderParams::ShaderParams()
{
	colorDisabled = color4b::GRAY;
	colorEnabled = color4b::WHITE;
	colorHot = color4b::YELLOW;
	alphaPressedBgr = 160;
	alphaHotBgr = 80;
	iconPadding = vec2(3, 3);
	borderWidth = 2;
	tileParams = gui::std_tile_textures::MakeTileParams(gui::std_tile_textures::Tile_Check);
}


//====================================================================
void  CS0_CheckBoxShader_Translucent::ShaderParams::ChooseParams(
										const gui_cs0::CheckBox& checkBox,
										color4b& cIcon, color4b& cBgr) const
{
	if(checkBox.IsEnabled()) {
		cIcon = checkBox.IsHot() ? colorHot : colorEnabled;
	}
	else {
		cIcon = colorDisabled;
	}
	cBgr = cIcon;

	if(checkBox.IsPressed()) {
		cBgr.a = alphaPressedBgr;
	}
	else if(checkBox.IsHot()) {
		cBgr.a = alphaHotBgr;
	}
}


//====================================================================
CS0_CheckBoxShader_Translucent::CS0_CheckBoxShader_Translucent(const std::string& shaderName,
												const ShaderParams& sp)
	:
	gui_cs0::CheckBox::Shader(shaderName)
{
	pTileTexture = 0;
	SetShaderParams(sp);
}


//====================================================================
void  CS0_CheckBoxShader_Translucent::SetShaderParams(const ShaderParams& sp)
{
	shaderParams = sp;
	iconWidth = shaderParams.iconPadding * 2;
	RequestTileTextureRecache();
}


//====================================================================
void  CS0_CheckBoxShader_Translucent::Shade(const gui::Component& c, gui::Renderer& gr) const
{
	const gui_cs0::CheckBox& checkBox = gui::ComponentCast<gui_cs0::CheckBox>(c);

	if(! pTileTexture) return;

	// choose params
	color4b  colorIcon, colorBgr;

	shaderParams.ChooseParams(checkBox, colorIcon, colorBgr);

	gui::RectangleC  rBorder(vec2::ZERO, checkBox.GetSize());

	// draw background
	if(checkBox.IsPressed()  ||  checkBox.IsHot()) {
		rBorder.SetColor(colorBgr);
		gr.DrawRectangle(rBorder);
	}

	// draw icon
	if(checkBox.IsOn()) {

		rTile.Set(shaderParams.iconPadding, checkBox.GetSize() - iconWidth);
		rTile.SetColor(colorIcon);

		gr.DrawRectangle(*pTileTexture, rTile);
	}

	// draw border
	rBorder.SetColor(colorIcon);
	gr.DrawRectangleFrame(rBorder, shaderParams.borderWidth);
}


//====================================================================
bool  CS0_CheckBoxShader_Translucent::CacheTileTextureHandles(gui::Manager& gm)
{
	pTileTexture = gm.GetTileTexture(shaderParams.tileParams.texName);
	return (pTileTexture != 0);
}


//====================================================================
void  CS0_CheckBoxShader_Translucent::PostCacheTileTextureHandles()
{
	if(pTileTexture) {
		pTileTexture->GetTileCoords(shaderParams.tileParams.texRow, shaderParams.tileParams.texCol, rTile);
	}
	rTile.SetTexTransform(shaderParams.tileParams.texTransform);
}


//********************************************************************
}//namespace gui_ss0


//********************************************************************
}//namespace vertexware


