

//--------------------------------------------------------------------
#include "GUI_SS0\Label.h"
#include "GUI\GUI_Manager.h"
#include "GUI\GUI_StdObjects.h"
#include "GUI\GUI_Tools.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gui_ss0 {


//====================================================================
CS0_LabelShader_Static::ShaderParams::ShaderParams()
{
	colorDisabled = color4b::GRAY;
	colorEnabled = color4b::BLACK;
	colorHot = color4b::BLACK;
}


//====================================================================
CS0_LabelShader_Static::CS0_LabelShader_Static(const std::string& shaderName,
										const ShaderParams& sp,
										const gui::TextAttributes& ta)
	:
	gui_cs0::Label::Shader(shaderName),
	shaderParams(sp)
{
	SetTextAttributes(ta);
}


//====================================================================
void  CS0_LabelShader_Static::Shade(const gui::Component& c, gui::Renderer& gr) const
{
	if(! pFont) return;

	const gui_cs0::Label& label = gui::ComponentCast<gui_cs0::Label>(c);
	const gui_cs0::Label::GeomParams& gp = label.GetGeomParams();

	color4b color;
	if(label.IsEnabled()) {
		color = label.IsHot() ? shaderParams.colorHot : shaderParams.colorEnabled;
	}
	else {
		color = shaderParams.colorDisabled;
	}

	gui::DrawTextLines(gr, *pFont, color,
				0.5f * (label.GetSize() - label.GetTextSize()),
				GetTextMetricsLCache(gp.textHeight, gp.lineHeight),
				label.GetTextLines());
}


//====================================================================
CS0_LabelShader_Active::ShaderParams::ShaderParams()
{
	colorDisabledText = color4b::GRAY;
	colorEnabledText = color4b::BLACK;
	colorHotBorder = color4b(0, 0, 0, 64);
	colorPressedBorder = color4b(0, 0, 0, 160);
	borderWidth = 1;
	borderOffset = vec2(2, 1);
}


//====================================================================
CS0_LabelShader_Active::CS0_LabelShader_Active(const std::string& shaderName,
										const ShaderParams& sp,
										const gui::TextAttributes& ta)
	:
	gui_cs0::Label::Shader(shaderName),
	shaderParams(sp)
{
	SetTextAttributes(ta);
}


//====================================================================
void  CS0_LabelShader_Active::Shade(const gui::Component& c, gui::Renderer& gr) const
{
	if(! pFont) return;

	const gui_cs0::Label& label = gui::ComponentCast<gui_cs0::Label>(c);
	const gui_cs0::Label::GeomParams& gp = label.GetGeomParams();

	// draw text
	vec2 textOffset = 0.5f * (label.GetSize() - label.GetTextSize());

	gui::DrawTextLines(gr, *pFont,
			label.IsEnabled() ? shaderParams.colorEnabledText : shaderParams.colorDisabledText,
			textOffset, GetTextMetricsLCache(gp.textHeight, gp.lineHeight),
			label.GetTextLines());

	// draw border
	if(label.IsHot()) {
		gui::RectangleC  r(
				textOffset - shaderParams.borderOffset,
				label.GetSize() - (textOffset - shaderParams.borderOffset) * 2);

		r.SetColor(label.IsPressed() ? shaderParams.colorPressedBorder : shaderParams.colorHotBorder);
		gr.DrawRectangleFrame(r, shaderParams.borderWidth);
	}
}


//====================================================================
CS0_LabelShader_Link::ShaderParams::ShaderParams()
{
	colorDisabled = color4b::GRAY;
	colorEnabled = color4b::BLACK;
	colorHotText = color4b::RED;
	colorHotUnderline = color4b::RED;
	colorPressedText = color4b::BLUE;
	colorPressedUnderline = color4b::BLUE;
	underlineWidth = 1;
	underlineOffset = -1;
}


//====================================================================
void  CS0_LabelShader_Link::ShaderParams::ChooseParams(
							const gui_cs0::Label& label,
							color4b& cText, color4b& cUnderline, bool& bDrawUnderline) const
{
	bDrawUnderline = false;

	if(label.IsEnabled()) {
		if(label.IsPressed()) {
			cText = colorPressedText;
			cUnderline = colorPressedUnderline;
			bDrawUnderline = true;
		}
		else if(label.IsHot()) {
			cText = colorHotText;
			cUnderline = colorHotUnderline;
			bDrawUnderline = true;
		}
		else {
			cText = colorEnabled;
		}
	}
	else {
		cText = colorDisabled;
	}
}


//====================================================================
CS0_LabelShader_Link::CS0_LabelShader_Link(const std::string& shaderName,
											const ShaderParams& sp,
											const gui::TextAttributes& ta)
	:
	gui_cs0::Label::Shader(shaderName),
	shaderParams(sp)
{
	SetTextAttributes(ta);
}


//====================================================================
void  CS0_LabelShader_Link::Shade(const gui::Component& c, gui::Renderer& gr) const
{
	if(! pFont) return;

	const gui_cs0::Label& label = gui::ComponentCast<gui_cs0::Label>(c);
	const gui_cs0::Label::GeomParams& gp = label.GetGeomParams();

	gui::TextMetricsL& tml = GetTextMetricsLCache(gp.textHeight, gp.lineHeight);

	// choose params
	color4b  colorText, colorUnderline;
	bool     bDrawUnderline;

	shaderParams.ChooseParams(label, colorText, colorUnderline, bDrawUnderline);

	// draw text
	vec2 textOffset = 0.5f * (label.GetSize() - label.GetTextSize());

	gui::DrawTextLines(gr, *pFont, colorText, textOffset, tml, label.GetTextLines());

	// draw link-underlines
	if(! bDrawUnderline) return;

	float fAbsLineHeight = tml.AbsLineHeight();

	gui::RectangleC  rUnderline;
	rUnderline.SetColor(colorUnderline);

	rUnderline.left = textOffset.x;
	rUnderline.top = tml.textHeight + shaderParams.underlineOffset;
	rUnderline.height = shaderParams.underlineWidth;

	for(uint i = 0; i < label.GetNumTextLines(); i++) {

		rUnderline.width = label.GetTextLineWidth(i);
		gr.DrawRectangle(rUnderline);
		rUnderline.top += fAbsLineHeight;
	}
}


//********************************************************************
}//namespace gui_ss0


//********************************************************************
}//namespace vertexware


