

//--------------------------------------------------------------------
#include "private\HIDManager.h"
#include "private\vw_kernel.h"
#include "private\SystemKeyboard.h"
#include "private\SystemMouse.h"
#include "BaseTools\Debug.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"
#include "GUI\GUI_Manager.h"


//====================================================================
HIDManager::HIDManager()
{
	iReconfigCount = 0;
	reconfigCode = 0;

	bDevicesInitialized = false;
	bDevicesConnectedToWindow = false;

	pDirectInputProxy = 0;

	pSystemKeyboard = 0;
	pSystemMouse = 0;
}


//====================================================================
HIDManager::~HIDManager()
{
}


//====================================================================
void  HIDManager::SetReconfigParams(const DirectInputConfig& diConfig,
									uint _reconfigCode)
{
	reconfigCode = _reconfigCode;
	diConfigReconfig = diConfig;
}


//====================================================================
void  HIDManager::InitializeDevices()
{
	if(bDevicesInitialized) return;

	olog << caption("human-interface-device init");

	// reconfig init
	if(iReconfigCount) {
		directInputConfig = diConfigReconfig;
	}
	// first-time init
	else {
		directInputConfig.enableDirectInput = vw_kernel::Instance().GetDefaultEnableDirectInput();
		directInputConfig.useLastVersion = true;
		directInputConfig.directInputVersion = 0;
		directInputConfig.enableFullscreenExclusiveMode = true;
		directInputConfig.useAtDeviceInit = true;
	}

	// query application to override direct-input-configuration
	olog << "Querying application for hid-configuration." << nl;
	vw_kernel::Instance().GetApplication().QueryHIDConfig(
									directInputConfig, iReconfigCount, reconfigCode);
	iReconfigCount++;

	olog << "Requested configuration:\n";
	directInputConfig.Dump(olog, 1);
	olog << nl;

	// configure direct-input
	VW_ASSERT(! pDirectInputProxy, "Kernel internal error.");
	pDirectInputProxy = 0;

	if(directInputConfig.enableDirectInput) {
		if(directInputConfig.useLastVersion) {
			directInputConfig.directInputVersion = 8;
		}
		DirectInputProxyPlugin* dipp = vw_kernel::Instance().GetPluginManager().GetDirectInputProxyPlugin(directInputConfig.directInputVersion);
		if(dipp) {
			pDirectInputProxy = dipp->GetDirectInputProxy();

			olog << "Initializing DirectInput version " << directInputConfig.directInputVersion << endpara;
			pDirectInputProxy->Initialize(); //exception-throw allowed
			pDirectInputProxy->SetAllowFullscreenExclusiveMode(directInputConfig.enableFullscreenExclusiveMode);
		}
		else {
			olog << e_warning << "DirectInput version " << directInputConfig.directInputVersion;
			olog << " not available" << endpara << e_info;

			directInputConfig.directInputVersion = 0;
		}
	}
	else {
		directInputConfig.directInputVersion = 0;
	}
	HumanInterfaceDevice::SetUseDirectInputAtInit(directInputConfig.useAtDeviceInit);

	// enumerate devices compatible with direct-input version used
	EnumerateDirectInputDevices();

	// create system devices
	olog << "Creating devices handled by: " << KERNEL_NAME << std::endl;

	pSystemKeyboard = CreateSystemKeyboard();
	deviceLinks.push_back(DeviceLink(0, 0, pSystemKeyboard));

	pSystemMouse = CreateSystemMouse();
	deviceLinks.push_back(DeviceLink(0, 0, pSystemMouse));

	olog << nl;

	// create attached devices handled by device-drivers
	PluginManager& pluginManager = vw_kernel::Instance().GetPluginManager();

	{for(uint i = 0; i < pluginManager.GetNumHIDDriverPlugins(); i++) {

		HIDDriverPlugin* hiddp = pluginManager.GetHIDDriverPlugin(i);
		HIDDriver* hidd = hiddp->GetHIDDriver();

		// skip drivers incompatible with active DirectInput version
		if(hiddp->directInputVersion != 0  &&  hiddp->directInputVersion != directInputConfig.directInputVersion) {
			continue;
		}
		olog << "Creating devices handled by: " << hiddp->GetPluginInfo().GetPluginName() << std::endl;

		// check with application wheter to ignore the driver
		bool  bIgnoreDriver = vw_kernel::Instance().GetApplication().QueryIgnoreHIDDriver(
												hiddp->GetPluginInfo().GetPluginName());
		if(bIgnoreDriver) {
			olog << "Driver ignored by application.\n\n";
			continue;
		}

		// probe devices
		uint  numDevices = ProbeDevices(*hidd);
		olog << "Found " << numDevices << " device(s).\n";

		// create available devices
		for(uint iDev = 0; iDev < numDevices; iDev++) {

			// create device and store device-link
			HumanInterfaceDevice* hid = CreateDevice(*hidd, iDev);
			if(hid) {
				deviceLinks.push_back(DeviceLink(hiddp, hidd, hid));
			}
			else {
				olog << e_error << "Failed to create device " << iDev << nl << e_info;
			}
		}

		// clear cached-information possibly stored by driver
		hidd->ClearCache();
		olog << nl;
	}}

	// setup creation-id / creator-driver-name
	{for(uint i = 0; i < deviceLinks.size(); i++) {
		HIDDriverPlugin* hiddp = deviceLinks[i].GetPlugin();
		HumanInterfaceDevice& hid = *(deviceLinks[i].GetDevice());

		HumanInterfaceDeviceInfo& hidInfo = const_cast<HumanInterfaceDeviceInfo&>(hid.GetDeviceInfo());

		hidInfo.SetCreationID(i);
		hidInfo.SetCreatorPluginName(hiddp ? hiddp->GetPluginInfo().GetPluginName() : KERNEL_NAME);
	}}

	// dump & confirm hid-s
	vw_kernel::Instance().GetApplication().BeginConfirmHumanInterfaceDevices();

	std::vector<DeviceLink>  acceptedDeviceLinks;
	std::vector<DeviceLink>  rejectedDeviceLinks;

	{for(uint i = 0; i < deviceLinks.size(); i++) {
		HIDDriverPlugin* hiddp = deviceLinks[i].GetPlugin();
		HumanInterfaceDevice& hid = *(deviceLinks[i].GetDevice());

		olog << e_attention << "HumanInterfaceDevice " << hid.GetDeviceInfo().GetCreationID() << nl << e_info;
		hid.GetDeviceInfo().Dump(olog, 0);

		olog << "Device caps:\n";
		hid.DumpDeviceCaps(olog, 1);

		olog << "Exposed controllers:\n";
		{for(uint i = 0; i < hid.GetNumControllers(); i++) {
			HIDController& hidc = hid.GetController(i);
			olog << e_attention << "\tHIDController " << i << nl << e_info;
			olog << "\tControllerInterfaceName: " << hidc.GetControllerInterfaceName() << nl;
			olog << "\tController info:\n";
			hidc.DumpControllerInfo(olog, 2);
		}}

		// always accept system-devices
		if((deviceLinks[i].GetDevice() == pSystemKeyboard) ||
			(deviceLinks[i].GetDevice() == pSystemMouse))
		{
			acceptedDeviceLinks.push_back(deviceLinks[i]);
		}
		else {
			// give application a chance to reject non-system devices
			if(vw_kernel::Instance().GetApplication().ConfirmHumanInterfaceDevice(hid)) {
				acceptedDeviceLinks.push_back(deviceLinks[i]);
			}
			else {
				rejectedDeviceLinks.push_back(deviceLinks[i]);
				olog << "Rejected by application.\n";
			}
		}
		olog << nl;
	}}

	// remove unused links
	deviceLinks = acceptedDeviceLinks;

	// release unused devices
	if(rejectedDeviceLinks.size()) {
		olog << caption("releasing unused human-interface-devices");

		for(int i = rejectedDeviceLinks.size() - 1; i >= 0; i--) {
			CleanupDevice(rejectedDeviceLinks[i]);
		}
		olog << nl;
	}

	// end-confirm notify
	vw_kernel::Instance().GetApplication().EndConfirmHumanInterfaceDevices();

	// connect devices to gui-manager
	olog << "Connecting human-interface-devices to gui-manager.\n";
	vw_kernel::Instance().GUIManager_BeginConnectHumanInterfaceDevices();

	{for(uint i = 0; i < deviceLinks.size(); i++) {
		HumanInterfaceDevice& hid = *(deviceLinks[i].GetDevice());
		vw_kernel::Instance().GUIManager_ConnectHumanInterfaceDevice(hid);
	}}
	vw_kernel::Instance().GUIManager_EndConnectHumanInterfaceDevices();

	// connect devices to application
	olog << "Connecting human-interface-devices to application.\n";
	vw_kernel::Instance().GetApplication().BeginConnectHumanInterfaceDevices(directInputConfig);

	{for(uint i = 0; i < deviceLinks.size(); i++) {
		HumanInterfaceDevice& hid = *(deviceLinks[i].GetDevice());
		vw_kernel::Instance().GetApplication().ConnectHumanInterfaceDevice(hid);
	}}
	vw_kernel::Instance().GetApplication().EndConnectHumanInterfaceDevices();

	olog << nl;

	// devices created
	ExtractNonSystemDevices();
	bDevicesInitialized = true;
}


//====================================================================
void  HIDManager::CleanupDevices()
{
	if(! bDevicesInitialized) return;

	olog << caption("human-interface-device cleanup");

	// disconnect hid-s from application
	olog << "Disconnecting human-interface-devices from application.\n";
	vw_kernel::Instance().GetApplication().DisconnectHumanInterfaceDevices();

	olog << "Disconnecting human-interface-devices from gui-manager.\n";
	vw_kernel::Instance().GUIManager_DisconnectHumanInterfaceDevices();

	// release all devices
	for(int i = deviceLinks.size() - 1; i >= 0; i--) {
		CleanupDevice(deviceLinks[i]);
	}
	deviceLinks.clear();

	nonSystemDevices.clear();

	pSystemKeyboard = 0;
	pSystemMouse = 0;

	// shutdown direct-input-proxy
	if(pDirectInputProxy) {
		olog << "\nShutting-down DirectInput\n";
		pDirectInputProxy->Shutdown();
		pDirectInputProxy = 0;
	}
	olog << nl;

	// devices not created
	bDevicesInitialized = false;
}


//====================================================================
void  HIDManager::ConnectDevicesToWindow(VideoSubsystemWindow& window)
{
	if(bDevicesConnectedToWindow) return;

	olog << caption("connecting human-interface-devices to video-window");

	for(int i = 0; i < deviceLinks.size(); i++) {
		olog << "Connecting: " << MakeShortDesc(deviceLinks[i]) << nl;

		// pass window-pointer only to devices created by kernel
		TopLevelWindow* pAppWindow = deviceLinks[i].GetDriver() ? 0 : &window;

		deviceLinks[i].GetDevice()->ConnectToAppWindow(pAppWindow,
										window.GetHWND(), window.IsFullscreen());
	}
	olog << nl;

	bDevicesConnectedToWindow = true;

	// force window-geometry notification
	OnAppWindowGeometryChanged();

	// setup async-hid sampling
	olog << caption("human-interface-device async-sampling setup");
	vw_kernel::Instance().GetApplication().SetupAsyncHIDSampling();
	olog << nl;
}


//====================================================================
void  HIDManager::DisconnectDevicesFromWindow()
{
	if(! bDevicesConnectedToWindow) return;

	olog << caption("disconnecting human-interface-devices from video-window");

	for(int i = deviceLinks.size() - 1; i >= 0; i--) {
		olog << "Disconnecting: " << MakeShortDesc(deviceLinks[i]) << nl;

		HumanInterfaceDevice* hid = deviceLinks[i].GetDevice();

		hid->StopAsyncSampling();
		hid->DisconnectFromAppWindow();
	}
	olog << nl;

	bDevicesConnectedToWindow = false;
}


//====================================================================
void  HIDManager::OnAppWindowActive(bool bActive)
{
	if(! bDevicesConnectedToWindow) return;

	for(uint i = 0; i < deviceLinks.size(); i++) {
		deviceLinks[i].GetDevice()->OnAppWindowActive(bActive);
	}
}


//====================================================================
void  HIDManager::OnAppWindowGeometryChanged()
{
	if(! bDevicesConnectedToWindow) return;

	for(uint i = 0; i < deviceLinks.size(); i++) {
		deviceLinks[i].GetDevice()->OnAppWindowGeometryChanged();
	}
}


//====================================================================
void  HIDManager::GetActiveDirectInputConfig(DirectInputConfig& diConfig)
{
	diConfig = directInputConfig;
}


//====================================================================
uint  HIDManager::GetNumDevices()
{
	return deviceLinks.size();
}


//====================================================================
HumanInterfaceDevice*  HIDManager::GetDevice(uint i)
{
	VW_ASSERT(i < deviceLinks.size(), "Device index out-of-range.");
	return deviceLinks[i].GetDevice();
}


//====================================================================
void  HIDManager::SetAutoSamplingConfig(const HIDAutoSamplingConfig& asc)
{
	autoSamplingConfig = asc;
}


//====================================================================
void  HIDManager::AutoSampleDevices()
{
	if(autoSamplingConfig.bAutoSampleSystemKeyboard  &&  pSystemKeyboard) {
		if(! pSystemKeyboard->IsAsyncSamplingActive()) {
			pSystemKeyboard->SampleInputData();
		}
	}
	if(autoSamplingConfig.bAutoSampleSystemMouse  &&  pSystemMouse) {
		if(! pSystemMouse->IsAsyncSamplingActive()) {
			pSystemMouse->SampleInputData();
		}
	}
	if(autoSamplingConfig.bAutoSampleNonSystemDevices) {
		for(uint i = 0; i < nonSystemDevices.size(); i++) {
			HumanInterfaceDevice* hid = nonSystemDevices[i];
			if(! hid->IsAsyncSamplingActive()) {
				hid->SampleInputData();
			}
		}
	}
}


//====================================================================
std::string  HIDManager::MakeShortDesc(DeviceLink& deviceLink)
{
	return deviceLink.GetDevice()->GetDeviceInfo().MakeShortDesc();
}


//====================================================================
void  HIDManager::CleanupDevice(DeviceLink& deviceLink)
{
	olog << "Releasing: " << MakeShortDesc(deviceLink) << nl;
	deviceLink.GetDevice()->StopAsyncSampling();
	deviceLink.DeleteDevice();
}


//====================================================================
void  HIDManager::ExtractNonSystemDevices()
{
	for(uint i = 0; i < deviceLinks.size(); i++) {
		HumanInterfaceDevice* hid = deviceLinks[i].GetDevice();
		if(hid != pSystemKeyboard  &&  hid != pSystemMouse) {
			nonSystemDevices.push_back(hid);
		}
	}
}


//====================================================================
void  HIDManager::EnumerateDirectInputDevices()
{
	// init system-device-handler states
	bFoundDISysKeyboard = false;
	bFoundDISysMouse = false;

	// proxy-version dependent enumeration
	DirectInputProxy8* dip8 = InterfaceCast<DirectInputProxy8>(pDirectInputProxy);
	if(dip8) {
		dip8->GetDirectInputObject()->EnumDevices(
			DI8DEVCLASS_ALL, DIEnumDevicesCallback8, this, DIEDFL_ATTACHEDONLY);
	}
}


//====================================================================
uint  HIDManager::ProbeDevices(HIDDriver& hidd)
{
	HIDDriver0* hidd0 = InterfaceCast<HIDDriver0>(&hidd);
	if(hidd0) {
		olog << "Probing attached devices..." << std::endl;
		return hidd0->ProbeDevices();
	}

	HIDDriver8* hidd8 = InterfaceCast<HIDDriver8>(&hidd);
	if(hidd8) {
		return hidd8->GetNumDevices();
	}
	return 0;
}


//====================================================================
HumanInterfaceDevice*  HIDManager::CreateDevice(HIDDriver& hidd, uint i)
{
	HIDDriver0* hidd0 = InterfaceCast<HIDDriver0>(&hidd);
	if(hidd0) {
		return hidd0->CreateDevice(i);
	}

	HIDDriver8* hidd8 = InterfaceCast<HIDDriver8>(&hidd);
	if(hidd8) {
		DirectInputProxy8* dip8 = InterfaceCast<DirectInputProxy8>(pDirectInputProxy);
		VW_ASSERT(dip8, "Kernel internal-error.");
		return hidd8->CreateDevice(i, *dip8);
	}
	return 0;
}


//====================================================================
HumanInterfaceDevice*  HIDManager::CreateSystemKeyboard()
{
	DirectInputProxy8* dip8 = InterfaceCast<DirectInputProxy8>(pDirectInputProxy);
	if(dip8) {
		if(! bFoundDISysKeyboard) dip8 = 0;
		return new SystemKeyboard8(dip8, didiiSysKeyboard);
	}
	return new SystemKeyboard8(0, didiiSysKeyboard);
}


//====================================================================
HumanInterfaceDevice*  HIDManager::CreateSystemMouse()
{
	DirectInputProxy8* dip8 = InterfaceCast<DirectInputProxy8>(pDirectInputProxy);
	if(dip8) {
		if(! bFoundDISysMouse) dip8 = 0;
		return new SystemMouse8(dip8, didiiSysMouse);
	}
	return new SystemMouse8(0, didiiSysMouse);
}


//====================================================================
BOOL  HIDManager::DIEnumDevicesCallback8(const DIDEVICEINSTANCE* didi, void* v)
{
	reinterpret_cast<HIDManager*>(v)->DispatchDirectInputDevice8(*didi);
	return DIENUM_CONTINUE; //keep enumerating
}


//====================================================================
void  HIDManager::DispatchDirectInputDevice8(const DirectInputDeviceInstanceInfo8& didii)
{
	DirectInputProxy8* dip8 = InterfaceCast<DirectInputProxy8>(pDirectInputProxy);
	VW_ASSERT(dip8, "Kernel internal-error.");

	// dump device info
	olog << "Found attached DirectInput8 device:\n";
	didii.DumpInfo(olog, 1);

	switch(GET_DIDEVICE_TYPE(didii.dwDevType))
	{
		// keyboard is handled internally
		case DI8DEVTYPE_KEYBOARD: {
			// ignore non-system keyboards
			if(didii.guidInstance == dip8->GetGuidSysKeyboard()) {
				bFoundDISysKeyboard = true;
				didii.CopyInfo(didiiSysKeyboard);
				olog << "\tHandled by: " << KERNEL_NAME << nl;
			}
			break;
		}
		// mouse is handled internally
		case DI8DEVTYPE_MOUSE: {
			// ignore non-system mices
			if(didii.guidInstance == dip8->GetGuidSysMouse()) {
				bFoundDISysMouse = true;
				didii.CopyInfo(didiiSysMouse);
				olog << "\tHandled by: " << KERNEL_NAME << nl;
			}
			break;
		}
		// other device types
		default: {
			PluginManager& pluginManager = vw_kernel::Instance().GetPluginManager();

			bool bHandled = false;
			for(uint i = 0; i < pluginManager.GetNumHIDDriverPlugins(); i++) {

				HIDDriverPlugin* hiddp = pluginManager.GetHIDDriverPlugin(i);
				HIDDriver* hidd = hiddp->GetHIDDriver();

				// find first handler, skip other candidates
				if(CanHandleDevice8(*hidd, didii)) {
					olog << "\tHandled by: " << hiddp->GetPluginInfo().GetPluginName() << nl;
					bHandled = true;
					break;
				}
			}
			if(! bHandled) {
				// handler not found
				olog << e_warning << "\tUnhandled\n" << e_info;
			}
			break;
		}
	}
	olog << nl;
}


//====================================================================
bool  HIDManager::CanHandleDevice8(HIDDriver& hidd,
									const DirectInputDeviceInstanceInfo8& didii8)
{
	HIDDriver0* hidd0 = InterfaceCast<HIDDriver0>(&hidd);
	HIDDriver8* hidd8 = InterfaceCast<HIDDriver8>(&hidd);

	if(hidd0) {
		DirectInputDeviceInstanceInfo  didii(didii8.guidInstance, didii8.tszInstanceName,
											didii8.guidProduct, didii8.tszProductName);
		return hidd0->CanHandleDevice(didii);
	}
	if(hidd8) {
		return hidd8->CanHandleDevice(didii8);
	}
	return false;
}


