

//--------------------------------------------------------------------
#include "BaseTools\STLExtensions.h"
#include "private\vw_kernel.h"
#include "private\AppSubsysManager.h"
#include "private\VideoManager.h"
#include "private\HIDManager.h"
#include "private\PluginManager.h"


//--------------------------------------------------------------------
#define  PREFIX_KERNEL_COMMAND  "KernelCommand: "


//--------------------------------------------------------------------
#define  ASSERT_ALLOWED_CHANGE_FRAME_OBSERVERS \
	VW_ASSERT(bAllowChangeFrameObservers, "Register/Unregister frame observer not allowed.")


//--------------------------------------------------------------------
#define  ASSERT_ALLOWED_CHANGE_RECONFIG_OBSERVERS \
	VW_ASSERT(bAllowChangeReconfigObservers, "Register/Unregister reconfig observer not allowed.")


//====================================================================
void  vw_kernel::SetExecutionParams(const KernelExecutionParams& kep)
{
	KernelExecutionParams kepLast = executionParams;
	executionParams = kep;
	executionParams.bDeveloperMode = kepLast.bDeveloperMode;

	pHIDManager->SetAutoSamplingConfig(executionParams.hidAutoSamplingConfig);

	// report
	std::ostream& os = *pLogOut;
	os << PREFIX_KERNEL_COMMAND << "SetExecutionParams\n";
	executionParams.Dump(os, 1);
	os << std::endl;
}


//====================================================================
void  vw_kernel::GetExecutionParams(KernelExecutionParams& kep)
{
	kep = executionParams;
}


//====================================================================
void  vw_kernel::SetKernelThreadInfo(const KernelThreadInfo& kti)
{
	// update states
	::SetThreadPriority(kernelThreadInfo.threadHandle, kti.priority);

	BOOL bBoostDisabled = kti.bEnablePriorityBoost ? FALSE : TRUE;
	::SetThreadPriorityBoost(kernelThreadInfo.threadHandle, bBoostDisabled);

	// get current states
	kernelThreadInfo.priority = ::GetThreadPriority(kernelThreadInfo.threadHandle);

	bBoostDisabled = FALSE;
	::GetThreadPriorityBoost(kernelThreadInfo.threadHandle, &bBoostDisabled);
	kernelThreadInfo.bEnablePriorityBoost = (bBoostDisabled ? false : true);

	// report
	std::ostream& os = *pLogOut;
	os << PREFIX_KERNEL_COMMAND << "SetKernelThreadInfo\n";
	kernelThreadInfo.Dump(os, 1);
	os << std::endl;
}


//====================================================================
void  vw_kernel::GetKernelThreadInfo(KernelThreadInfo& kti)
{
	kti = kernelThreadInfo;
}


//====================================================================
void  vw_kernel::SetMainWindowProperties(const MainWindowProperties& mwp)
{
	VideoSubsystemWindow* w = pVideoManager->GetVideoWindow();
	if(w) {
		w->SetIcon(mwp.hIcon);
		w->SetText(mwp.text);
	}
}


//====================================================================
void  vw_kernel::GetMainWindowProperties(MainWindowProperties& mwp)
{
	VideoSubsystemWindow* w = pVideoManager->GetVideoWindow();
	if(w) {
		mwp.hIcon = w->GetIcon();
		mwp.text = w->GetText();
	}
}


//====================================================================
void  vw_kernel::RequestShutdown()
{
	std::ostream& os = *pLogOut;
	os << PREFIX_KERNEL_COMMAND << "RequestShutdown" << std::endl;

	if(CheckCommandInsideFrameLoop()) {
		bShutdown = true;
		RequestFrameLoopExit();
	}
}


//====================================================================
void  vw_kernel::RequestHIDReconfig(const DirectInputConfig& diConfig,
									uint reconfigCode)
{
	std::ostream& os = *pLogOut;
	os << PREFIX_KERNEL_COMMAND << "RequestHIDReconfig" << std::endl;

	if(CheckCommandInsideFrameLoop()) {
		pHIDManager->SetReconfigParams(diConfig, reconfigCode);
		bHIDReconfigPending = true;
		RequestFrameLoopExit();
	}
}


//====================================================================
void  vw_kernel::RequestVideoReconfig(const std::string& vrPluginName,
									const VideoConfig& vc, uint reconfigCode)
{
	std::ostream& os = *pLogOut;
	os << PREFIX_KERNEL_COMMAND << "RequestVideoReconfig" << std::endl;

	if(CheckCommandInsideFrameLoop()) {
		pVideoManager->SetReconfigParams(vrPluginName, vc, reconfigCode);
		bVideoReconfigPending = true;
		RequestFrameLoopExit();
	}
}


//====================================================================
void  vw_kernel::RequestSubsystemReconfig(uint reconfigCode)
{
	std::ostream& os = *pLogOut;
	os << PREFIX_KERNEL_COMMAND << "RequestSubsystemReconfig" << std::endl;

	if(CheckCommandInsideFrameLoop()) {
		pAppSubsysManager->SetReconfigParams(reconfigCode);
		bSubsysReconfigPending = true;
		RequestFrameLoopExit();
	}
}


//====================================================================
void  vw_kernel::RequestGenericReconfig(uint reconfigCode)
{
	std::ostream& os = *pLogOut;
	os << PREFIX_KERNEL_COMMAND << "RequestGenericReconfig" << std::endl;

	if(CheckCommandInsideFrameLoop()) {
		iGenericReconfigCode = reconfigCode;
		bGenericReconfigPending = true;
		RequestFrameLoopExit();
	}
}


//====================================================================
bool  vw_kernel::SetPluginParameter(const std::string& pluginName, const NamedParameter& np)
{
	return pPluginManager->SetPluginParameter(pluginName, np);
}


//====================================================================
bool  vw_kernel::GetPluginParameter(const std::string& pluginName, const std::string& parameterName, NamedParameter& out)
{
	return pPluginManager->GetPluginParameter(pluginName, parameterName, out);
}


//====================================================================
uint  vw_kernel::GetNumGenericExtensions()
{
	return pPluginManager->GetNumAppUsedGenericExtensionPlugins();
}


//====================================================================
GenericExtension&  vw_kernel::GetGenericExtension(uint i)
{
	VW_ASSERT(i < pPluginManager->GetNumAppUsedGenericExtensionPlugins(),
					"Generic extension index out-of-range.");

	return *pPluginManager->GetAppUsedGenericExtensionPlugin(i)->GetGenericExtension();
}


//====================================================================
GenericExtension*  vw_kernel::FindGenericExtension(const std::string* pPluginName,
								const std::string* pGenericExtensionInterfaceName)
{
	GenericExtensionPlugin* p =
		pPluginManager->FindAppUsedGenericExtensionPlugin(
							pPluginName, pGenericExtensionInterfaceName);

	if(p) return p->GetGenericExtension();
	return 0;
}


//====================================================================
uint  vw_kernel::GetNumVideoRenderers()
{
	return pPluginManager->GetNumVideoRendererPlugins();
}


//====================================================================
void  vw_kernel::GetVideoRendererInfo(uint i, VideoRendererInfo& vrInfo)
{
	VW_ASSERT(i < GetNumVideoRenderers(), "Video-renderer index out-of-range.");

	const VideoRendererPlugin* vrp = pPluginManager->GetVideoRendererPlugin(i);
	const VideoSubsystemPlugin* vsp = pPluginManager->FindVideoSubsystemPlugin(
												vrp->videoSubsystemInterfaceName);

	vrInfo = VideoRendererInfo(vrp->GetPluginInfo(),
				vrp->videoRendererInterfaceName, vsp->videoProfileFiltered);
}


//====================================================================
void  vw_kernel::GetVideoRendererUnfilteredProfile(uint i, VideoProfile& profile)
{
	VW_ASSERT(i < GetNumVideoRenderers(), "Video-renderer index out-of-range.");

	const VideoRendererPlugin* vrp = pPluginManager->GetVideoRendererPlugin(i);
	const VideoSubsystemPlugin* vsp = pPluginManager->FindVideoSubsystemPlugin(
												vrp->videoSubsystemInterfaceName);

	profile = vsp->videoProfileUnfiltered;
}


//====================================================================
VideoRenderer*  vw_kernel::GetActiveVideoRendererPtr()
{
	return pVideoManager->GetVideoRenderer();
}


//====================================================================
bool  vw_kernel::GetActiveVideoRendererInfo(VideoRendererInfo& vrInfo)
{
	VideoRenderer* pvr = pVideoManager->GetVideoRenderer();
	if(! pvr) return false;

	const VideoSubsystemPlugin* vsp = pPluginManager->FindVideoSubsystemPlugin(pvr->GetMatchingVideoSubsystemInterfaceName());

	vrInfo = VideoRendererInfo(pvr->GetPluginInfo(),
					pvr->GetVideoRendererInterfaceName(), vsp->videoProfileFiltered);
	return true;
}


//====================================================================
bool  vw_kernel::GetActiveVideoRendererUnfilteredProfile(VideoProfile& profile)
{
	profile.Clear();

	VideoRenderer* pvr = pVideoManager->GetVideoRenderer();
	if(! pvr) return false;

	const VideoSubsystemPlugin* vsp = pPluginManager->FindVideoSubsystemPlugin(pvr->GetMatchingVideoSubsystemInterfaceName());

	profile = vsp->videoProfileUnfiltered;
	return true;
}


//====================================================================
uint  vw_kernel::GetNumHIDDrivers()
{
	return pPluginManager->GetNumHIDDriverPlugins();
}


//====================================================================
void  vw_kernel::GetHIDDriverInfo(uint i, PluginInfo& pi)
{
	VW_ASSERT(i < GetNumHIDDrivers(), "HID-driver index out-of-range.");
	HIDDriverPlugin* hiddp = pPluginManager->GetHIDDriverPlugin(i);
	pi = hiddp->GetPluginInfo();
}


//====================================================================
uint  vw_kernel::GetNumHumanInterfaceDevices()
{
	return pHIDManager->GetNumDevices();
}


//====================================================================
HumanInterfaceDevice&  vw_kernel::GetHumanInterfaceDevice(uint i)
{
	VW_ASSERT(i < GetNumHumanInterfaceDevices(), "Human-interface-device index out-of-range.");
	return *pHIDManager->GetDevice(i);
}


//====================================================================
void  vw_kernel::GetActiveDirectInputConfig(DirectInputConfig& diConfig)
{
	pHIDManager->GetActiveDirectInputConfig(diConfig);
}


//====================================================================
void  vw_kernel::ResetFrameNumber(uint frameNumber)
{
	iFrameNumber = frameNumber;
}


//====================================================================
void  vw_kernel::RegisterFrameObserver(FrameObserver& fo)
{
	ASSERT_ALLOWED_CHANGE_FRAME_OBSERVERS;
	UniqueContainerPushBack(*pFrameObservers, VEC_FRAME_OBSERVER::value_type(&fo));
	pFrameObserversTemp->reserve(pFrameObservers->size());
}


//====================================================================
void  vw_kernel::UnregisterFrameObserver(FrameObserver& fo)
{
	ASSERT_ALLOWED_CHANGE_FRAME_OBSERVERS;
	EraseFromContainer(*pFrameObservers, VEC_FRAME_OBSERVER::value_type(&fo));
}


//====================================================================
void  vw_kernel::UnregisterAllFrameObservers()
{
	ASSERT_ALLOWED_CHANGE_FRAME_OBSERVERS;
	pFrameObservers->clear();
}


//====================================================================
void  vw_kernel::RegisterReconfigObserver(ReconfigObserver& ro)
{
	ASSERT_ALLOWED_CHANGE_RECONFIG_OBSERVERS;
	UniqueContainerPushBack(*pReconfigObservers, VEC_RECONFIG_OBSERVER::value_type(&ro));
}


//====================================================================
void  vw_kernel::UnregisterReconfigObserver(ReconfigObserver& ro)
{
	ASSERT_ALLOWED_CHANGE_RECONFIG_OBSERVERS;
	EraseFromContainer(*pReconfigObservers, VEC_RECONFIG_OBSERVER::value_type(&ro));
}


//====================================================================
void  vw_kernel::UnregisterAllReconfigObservers()
{
	ASSERT_ALLOWED_CHANGE_RECONFIG_OBSERVERS;
	pReconfigObservers->clear();
}


//====================================================================
void  vw_kernel::NotifyVSyncChanged()
{
	bool bVSyncEnabled = GetActiveVideoRenderer().GetVideoConfig().videoHints.vSync;

	NotifyVSyncReconfigObservers(bVSyncEnabled);
	pApplication->OnVSyncUpdate(bVSyncEnabled);
}


