

//--------------------------------------------------------------------
#include "private\vw_kernel.h"
#include "private\LogWindowThread.h"
#include "private\PluginManager.h"
#include "BaseTools\FileSystem.h"
#include "BaseTools\PlatformInfo.h"


//====================================================================
void  vw_kernel::Level0()
{
	KernelLevelSentry  kls(*this, KernelLevel_KernelSetup);

	//---------------------
	// Create base-objects.
	//---------------------
	ExtractThreadHandle();

	// callback storage
	AutoPtr<VEC_FRAME_OBSERVER>  apFrameObservers(new VEC_FRAME_OBSERVER);
	pFrameObservers = apFrameObservers;

	AutoPtr<VEC_FRAME_OBSERVER>  apFrameObserversTemp(new VEC_FRAME_OBSERVER);
	pFrameObserversTemp = apFrameObserversTemp;

	AutoPtr<VEC_RECONFIG_OBSERVER>  apReconfigObservers(new VEC_RECONFIG_OBSERVER);
	pReconfigObservers = apReconfigObservers;

	// timers
	AutoPtr<HighResolutionTimer>  apHighResTimer(new HighResolutionTimer);
	pHighResTimer = apHighResTimer;

	// profiler
	AutoPtr<FrameworkProfiler>  apFrameworkProfiler(new FrameworkProfiler);
	pFrameworkProfiler = apFrameworkProfiler;

	// string-storage
	AutoPtr<StringStorage>  apStringStorage(new StringStorage);
	pStringStorage = apStringStorage;
	StringStorage& ss = *pStringStorage;

	// get command-line from os
	ss.commandLine = ::GetCommandLine();

	// create command-line parser
	AutoPtr<CommandLineParser>  apCmdLineParser(new CommandLineParser);
	pCmdLineParser = apCmdLineParser;

	// check and parse command-line
	pCmdLineParser->SetParseFlags(true, true);
	pCmdLineParser->Parse(ss.commandLine); //throws exception on error
	pCmdLineParser->ClearLastCommandLine();

	// filter command-line for application
	RemoveKernelCLArguments(*pCmdLineParser, ss.filteredCommandLine);

	// exe-directory
	if(! file_system::GetProcessDirectory(ss.exeDir, true)) {
		VW_THROW(KernelException, "OS failed to report process-directory.");
	}

	// plugin-directory
	ss.pluginDir = ss.exeDir;

	if(! file_system::DirectoryExists(ss.pluginDir)) {
		VW_THROW(KernelException, std::string("Cannot find plugin-directory: ") + ss.pluginDir);
	}

	// user directories
	if(! file_system::GetUserHomeDirectory(ss.userHomeDir, true)) {
		VW_THROW(KernelException, "OS failed to report user-home directory.");
	}
	if(! file_system::GetUserProfileDirectory(ss.userProfileDir, true)) {
		VW_THROW(KernelException, "OS failed to report user-profile directory.");
	}

	// get temporary-file directory from OS
	if(! file_system::GetTempDirectory(ss.tempDir, true)) {
		VW_THROW(KernelException, "OS failed to report temporary-file directory.");
	}

	// create temp log-file
	if(! file_system::CreateTempFile(ss.tempLogPath, ss.tempDir, TEMP_LOG_FILE_PREFIX)) {
		VW_THROW(KernelException, "OS failed to create temporary-file.");
	}

	// setup default app-paths in case app-plugin cannot start
	ss.appProcessDir = ss.pluginDir;
	ss.appProfileDir = ss.pluginDir;
	ss.appLogFilename = DEFAULT_LOG_FILE_NAME;
	ss.appLogPath = ss.appProfileDir + ss.appLogFilename;

	// create & setup console-buffers
	AutoPtr<ConsoleBuffers>  apConsoleBuffers(new ConsoleBuffers);
	pConsoleBuffers = apConsoleBuffers;

	pConsoleBuffers->SetMaxOutputLines(CONBUF_MAX_OUTPUT_LINES);
	pConsoleBuffers->SetMaxCommandsInHistory(CONBUF_MAX_COMMAND_HISTORY);

	// create console-output-stream
	AutoPtr<ConsoleOutputStream>  apConOut(pConsoleBuffers->CreateOutputStream());
	pConOut = apConOut;

	// create log-file
	AutoPtr<LogFileStream>  apLogFile(new LogFileStream(DEFAULT_LOG_NAME, ss.tempLogPath));
	pLogFile = apLogFile;

	// create log-output-stream & setup output targets
	AutoPtr<ControlledOutputStream>  apLogOut(new ControlledOutputStream);
	pLogOut = apLogOut;

	pLogOut->AddOutputTarget(*pLogFile);
	pLogOut->AddOutputTarget(*pConOut);

	// register output-streams
	SetOutputStreams(*pConOut, *pLogOut);

	// create log-window-thread and connect it to console-buffer
	AutoPtr<LogWindowThread>  apLogWindowThread(new LogWindowThread);
	pLogWindowThread = apLogWindowThread;

	// create plugin-manager
	AutoPtr<PluginManager>  apPluginManager(new PluginManager);
	pPluginManager = apPluginManager;

	// message hooks
	AutoPtr<MessageHookImp>  apMessageHook(new MessageHookImp);
	pMessageHook = apMessageHook;

	// collect startup options
	bCleanExit = true;

	bDefaultEnableDirectInput = ! pCmdLineParser->GetOption(claStrings[CLA_NoDirectInput]);
	bReportDLLLoadErrors = pCmdLineParser->GetOption(claStrings[CLA_ReportDLLLoadErrors]);

	executionParams.bDeveloperMode = pCmdLineParser->GetOption(claStrings[CLA_DeveloperMode]);
	executionParams.bEnableHardCleanupConfigLoop = ! pCmdLineParser->GetOption(claStrings[CLA_NoHardCleanupConfigLoop]);
	executionParams.bEnableHardRethrowExit = executionParams.bDeveloperMode;

	if(pCmdLineParser->GetOption(claStrings[CLA_HardRethrowExit])) executionParams.bEnableHardRethrowExit = true;
	if(pCmdLineParser->GetOption(claStrings[CLA_NoHardRethrowExit])) executionParams.bEnableHardRethrowExit = false;

	bool bHelpOnly = pCmdLineParser->GetOption(claStrings[CLA_Help]);
	bool bShowLog = bHelpOnly  ||  pCmdLineParser->GetOption(claStrings[CLA_ShowLog]);

	executionParams.bWaitCloseConfirmationOnExit = bHelpOnly  ||  pCmdLineParser->GetOption(claStrings[CLA_WaitClose]);

	LogWindowThread::eWindowPosition logWinPos = LogWindowThread::WindowPosition_Center;
	if(pCmdLineParser->GetOption(claStrings[CLA_LogTopLeft])) logWinPos = LogWindowThread::WindowPosition_TopLeft;
	if(pCmdLineParser->GetOption(claStrings[CLA_LogTopRight])) logWinPos = LogWindowThread::WindowPosition_TopRight;

	// configure and start log-window-thread
	pLogWindowThread->SetTitle(KERNEL_NAME);
	pLogWindowThread->EnableCloseGadget(false, false);
	pLogWindowThread->SetWindowPosition(logWinPos);

	if(bShowLog) {
		pLogWindowThread->UnsplashWindow();
	}
	else {
		pLogWindowThread->SplashWindow();
	}
	pLogWindowThread->Run();
	pLogWindowThread->WaitReady();

	// store log-window handle
	pMessageHook->SetLogWindowHandle(pLogWindowThread->GetWindow()->GetHWND());
	pMessageHook->SetLog(pLogOut);

	//---------------------------
	// Check compiler parameters.
	//---------------------------
	VW_ASSERT(CheckBuiltInTypeSizes(*pLogOut), "Unexpected built-in type sizes.");

	//------------------------
	// Dump help if requested.
	//------------------------
	if(bHelpOnly  ||  executionParams.bDeveloperMode) {
		DumpHelp();
	}

	//-------------------
	// Dump startup info.
	//-------------------
	olog << e_info;
	olog << "LogWindowHandle: 0x" << pLogWindowThread->GetWindow()->GetHWND() << endpara;

	// print received command-line
	olog << "Command-line: " << nl;
	olog << FitToLineLengthBreakAtWhitespaceOnly(ss.commandLine, 100) << endpara;

	// print kernel info
	olog << caption("Kernel info");
	olog << "KernelName....: " << KERNEL_NAME  << nl;
	olog << "KernelVersion.: " << KERNEL_VERSION_MAJOR << '.' << KERNEL_VERSION_MINOR << '.' << KERNEL_VERSION_REVISION << nl;
	olog << "BuildTimestamp: " << VW_BUILD_TIMESTAMP_STRING << nl;
	olog << "LibraryVersion: " << library_info::LibraryVersion().MakeString() << endpara;

	// print library info
	olog << caption("Library info");
	olog << "LibraryName...: " << library_info::GetLibraryName() << nl;
	olog << "LibraryVersion: " << library_info::GetLibraryVersion().MakeString() << nl;
	olog << "BuildTimestamp: " << library_info::GetBuildTimestamp() << endpara;

	olog << "Copyright:\n";
	olog << tab << library_info::GetCopyrightNotice() << endpara;

	olog << "Developers:\n";
	{for(uint i = 0; i < library_info::GetNumDevelopers(); i++) {
		olog << tab << library_info::GetDeveloperInfo(i) << nl;
	}}
	olog << nl;

	// print platform info
	olog << caption("platform info");
	olog << "Machine name....: " << platform_info::GetMachineName() << nl;
	olog << "User name.......: " << platform_info::GetUserName() << nl;
	olog << "Operating system: " << platform_info::GetOperatingSystemDescription() << nl;
	olog << "Desktop settings: " << platform_info::GetDesktopDisplayModeDescription() << endpara;

	// print startup options
	olog << caption("startup options");
	olog << "DirectInput................: " << ToString(bDefaultEnableDirectInput) << nl;
	olog << "HIDDriver plugins..........: " << ToString(pCmdLineParser->GetOption(claStrings[CLA_NoHIDDPlugins])) << nl;
	olog << "Developer mode.............: " << ToString(executionParams.bDeveloperMode) << nl;
	olog << "Wait close confirmation....: " << ToString(executionParams.bWaitCloseConfirmationOnExit) << nl;
	olog << "Hard cleanup by config loop: " << ToString(executionParams.bEnableHardCleanupConfigLoop) << nl;
	olog << "Hard rethrow exit..........: " << ToString(executionParams.bEnableHardRethrowExit) << nl;
	olog << nl;

	//-----------------------------------------------
	// If help is explicitly requested, skip Level-1.
	//-----------------------------------------------
	if(! bHelpOnly) {

		//---------------
		// Level-1 frame.
		//---------------
		try {
			// check kernel/library version match
			if(library_info::GetLibraryVersion() != library_info::LibraryVersion()) {
				VW_THROW(KernelException, "Kernel/Library version mismatch.");
			}
			//---------------
			// Enter Level-1.
			//---------------
			Level1();
		}
		catch(Exception& e) {
			ReportException(e);
		}
		catch(std::exception& e) {
			ReportException(e);
		}
		catch(...) {
			// This is most likely a hardware-level exception,
			// so pass it to default handler unless explicitly requested.

			ReportExceptionUnknown("Hard rethrow enabled: "
						+ ToString(executionParams.bEnableHardRethrowExit));

			if(executionParams.bEnableHardRethrowExit) {

				// write event to log-file & exit log-window-thread
				olog << e_info << "Terminating log-window thread.\n";
				TerminateLogWindowThread();

				// force log-stream flush before actual rethrow
				ReportHardRethrow();
				CloseLogFile();
				throw;
			}
		}
	}

	//------------------
	// Dump exit-status.
	//------------------
	olog << caption("exit status");
	if(bCleanExit) {
		olog << "Clean exit." << std::endl;
	}
	else {
		olog << "Error occured." << std::endl;

		executionParams.bWaitCloseConfirmationOnExit = true;
	}

	//---------------------
	// Shutdown log-window.
	//---------------------
	if(executionParams.bWaitCloseConfirmationOnExit) {
		olog << "Waiting close confirmation." << std::endl;

		pLogWindowThread->EnableCloseGadget(true, false);
		pLogWindowThread->UnsplashWindow();

		if(! pLogWindowThread->IsRunning()) {
			pLogWindowThread->SetTitle(KERNEL_NAME);
			pLogWindowThread->Run();
			pLogWindowThread->WaitReady();
		}
		pLogWindowThread->FocusOnLastLine();
		pLogWindowThread->WaitExit();
	}
	else {
		olog << "Auto-closing." << std::endl;

		TerminateLogWindowThread();
	}

	//------------------------------------------------------------------
	// Cleanup of Level0-objects automatically handled by auto-pointers.
	//------------------------------------------------------------------
	CloseLogFile();
	CloseThreadHandle();
}


