

//--------------------------------------------------------------------
#include "private\vw_kernel.h"
#include "private\LogWindowThread.h"
#include "private\PluginManager.h"
#include "private\HIDManager.h"
#include "private\VideoManager.h"
#include "private\AppSubsysManager.h"
#include "DataObjects\DataCodecManager.h"
#include "DataObjects\SignalCodecVWDP.h"
#include "DataObjects\SignalCodecGdiplus.h"
#include "DataObjects\SignalIO.h"
#include "DataObjects\SignalOperations.h"
#include "Framework\LibWrapperDXErr8.h"
#include "GUI\GUI_Manager.h"


//====================================================================
// NOTES:
// All objects that directly or indirectly create/destroy other
// objects of any type defined by plugin must be created after
// the plugin-manager (i.e. at level-1). This prevents plugins
// from unloading before destruction-code located within plugin-module
// is executed.
//====================================================================
void  vw_kernel::Level1()
{
	KernelLevelSentry  kls(*this, KernelLevel_PluginSetup);

	StringStorage& ss = *pStringStorage;

	//-----------------------
	// Init execution-states.
	//-----------------------
	bShutdown = false;
	bHIDReconfigPending = true;
	bVideoReconfigPending = true;
	bSubsysReconfigPending = true;
	bGenericReconfigPending = true;
	bFrameInProgress = false;
	bSpinFrameLoop = false;
	iFrameLoopIteration = 0;
	iFrameNumber = 0;
	bAppWindowActive = false;

	//---------------------------------------
	// Create library-singletons and globals.
	//---------------------------------------
	AutoPtr<VariableRegistry>  apVariableRegistry(new VariableRegistry(GLOBAL_VARIABLE_REGISTRY_NAME));
	pVariableRegistry = apVariableRegistry;

	AutoPtr<VideoManager>  apVideoManager(new VideoManager);
	pVideoManager = apVideoManager;

	AutoPtr<HIDManager>  apHIDManager(new HIDManager);
	pHIDManager = apHIDManager;
	pHIDManager->SetAutoSamplingConfig(executionParams.hidAutoSamplingConfig);

	AutoPtr<AppSubsysManager>  apAppSubsysManager(new AppSubsysManager);
	pAppSubsysManager = apAppSubsysManager;

	AutoPtr<SignalOperationsManager>  apSignalOperationsManager(new SignalOperationsManager());
	AutoPtr<DataCodecManager>         apDataCodecManager(new DataCodecManager());
	AutoPtr<gui::Manager>             apGUIManager(new gui::Manager());

	// pass external directives to gui-manager
	if(pCmdLineParser->GetOption(claStrings[CLA_GMTolerateMissingDevice])) {
		GUIManager_SetTolerateMissingInputDevice(true);
	}

	//------------------------------------
	// Load and analyze available plugins.
	//------------------------------------
	if(pCmdLineParser->GetOption(claStrings[CLA_NoHIDDPlugins])) {
		pPluginManager->DisableAllHIDDriverPlugins();
	}

	for(uint i = 0; i < pCmdLineParser->GetNumArguments(); i++) {
		const CommandLineParser::Argument& arg = pCmdLineParser->GetArgument(i);

		if(arg == CommandLineParser::ArgumentType_Assignment) {
			if(arg.Assignment().Name() == claStrings[CLA_DisablePlugin]) {

				pPluginManager->DisablePlugin(arg.Assignment().Value());
			}
		}
	}

	pPluginManager->LoadAvailablePlugins(ss.pluginDir);

	//--------------------------
	// Setup application-plugin.
	//--------------------------
	std::string  requestedAppPluginName;
	pCmdLineParser->GetAssignment(claStrings[CLA_Application], requestedAppPluginName);

	pPluginManager->ScanApplicationPlugins(requestedAppPluginName);
	pApplication = pPluginManager->GetApplicationPlugin()->GetApplication();

	ss.appDLLBasename = pPluginManager->GetApplicationPlugin()->GetBaseName();
	ss.appName = pPluginManager->GetApplicationPlugin()->GetPluginInfo().GetPluginName();

	// ensure single instance of particular application
	ss.appFlagname = library_info::GetLibraryName();
	ss.appFlagname.append(ss.appName);
	ss.appFlagname.append(pApplication->GetApplicationUID().ToString());

	SingleInstanceFlagSentry  appInstanceFlag(ss.appFlagname);
	if(! appInstanceFlag.IsAcquired()) {
		olog << e_error << "Another instance of \"" << ss.appName;
		olog << "\" is already running. Exiting." << endpara << e_info;

		// already running, silent exit
		bUseProfileDir = false;
		return;
	}

	// construct default application paths
	ss.appProfileDir = ss.userProfileDir + ss.appDLLBasename;
	file_system::ForceTrailingPathSeparator(ss.appProfileDir);
	ss.appLogFilename = ss.appDLLBasename + DEFAULT_LOG_FILE_EXTENSION;
	ss.appLogPath = ss.appProfileDir + ss.appLogFilename;

	// connect application to gui-manager
	AutoPtr<GUIManagerAppConnectSentry>  apGUIManagerAppConnectSentry(new GUIManagerAppConnectSentry(*this));

	//----------------------------------------------
	// Application-specific startup-info processing.
	//----------------------------------------------
	bool bAutoCreateProfileDir = true;
	bool bShowLogAndExit = false;

	try {
		olog << nl << caption("begin startup");
		olog << "Kernel-thread ID: " << kernelThreadInfo.threadID << endpara;

		olog << "Filtered command-line:\n";
		olog << FitToLineLengthBreakAtWhitespaceOnly(ss.filteredCommandLine, 100) << endpara;

		Application::StartupInfo  startupInfo;
		startupInfo.commandLine = ss.filteredCommandLine;
		startupInfo.fullCommandLine = ss.commandLine;
		startupInfo.processDir = ss.appProcessDir;
		startupInfo.userHomeDir = ss.userHomeDir;
		startupInfo.userProfileDir = ss.userProfileDir;
		startupInfo.appProfileDir = ss.appProfileDir;
		startupInfo.appLogFilename = ss.appLogFilename;

		olog << "StartupInfo paths:\n";
		startupInfo.DumpPaths(olog, 1);
		olog << nl;

		// setup startup result
		Application::StartupResult  startupResult;
		startupResult.processDir = startupInfo.processDir;
		startupResult.appProfileDir = startupInfo.appProfileDir;
		startupResult.appLogFilename = startupInfo.appLogFilename;
		startupResult.appWindowText = ss.appName;

		// do process info
		pApplication->BeginStartup(startupInfo, startupResult);

		bUseProfileDir = startupResult.bUseProfileDir;
		bAutoCreateProfileDir = startupResult.bAutoCreateProfileDir;
		bShowLogAndExit = startupResult.bShowLogAndExit;

		// update log-window title
		ss.appWindowText = startupResult.appWindowText;
		pLogWindowThread->SetTitle(ss.appWindowText);

		// dump startup result
		olog << "StartupResult paths:\n";
		startupResult.DumpPaths(olog, 1);
		olog << nl;

		// process startup result
		std::string  finalProcessDir;
		if(! file_system::ExpandPathName(startupResult.processDir, finalProcessDir, true)) {
			VW_THROW(KernelException, "Application specified invalid process directory.");
		}
		if(! file_system::DirectoryExists(finalProcessDir)) {
			VW_THROW(KernelException, "Application specified invalid process directory.");
		}

		std::string  finalProfileDir;
		if(bUseProfileDir) {
			if(! file_system::ExpandPathName(startupResult.appProfileDir, finalProfileDir, true)) {
				VW_THROW(KernelException, "Application specified invalid profile directory.");
			}
		}

		std::string  finalLogPath;
		if(bUseProfileDir) {
			std::string  pathName, baseName, extName;
			file_system::SplitFileName(finalProfileDir + startupResult.appLogFilename, pathName, baseName, extName);
			if(! file_system::ExpandPathName(finalProfileDir + baseName + "." + extName, finalLogPath, false)) {
				VW_THROW(KernelException, "Application specified invalid log-filename.");
			}
			file_system::RemoveTrailingPathSeparator(finalLogPath);
		}

		// update kernel-internal app-paths
		ss.appProcessDir = finalProcessDir;

		if(bUseProfileDir) {
			ss.appProfileDir = finalProfileDir;
			ss.appLogFilename = startupResult.appLogFilename;
			ss.appLogPath = finalLogPath;
		}
	}
	catch(Exception& e) {
		ReportException(e);
		VW_THROW(KernelException, "Application incorrectly processed startup-info.");
	}

	// create profile directory if necessary
	if(bUseProfileDir) {
		if(! file_system::DirectoryExists(ss.appProfileDir)) {
			olog << "Application-specified profile directory does not exist.\n";

			try {
				if(bAutoCreateProfileDir) {
					olog << "Creating: " << ss.appProfileDir << endpara;

					if(! file_system::CreateDirectoryPath(ss.appProfileDir)) {
						VW_THROW(KernelException, "Failed to create application profile-directory.");
					}
				}
				else {
					VW_THROW(KernelException, "Application profile-directory does not exist.");
				}
			}
			catch(Exception&) {
				// fallback to default profile directory
				ss.appProfileDir = ss.pluginDir;
				ss.appLogFilename = DEFAULT_LOG_FILE_NAME;
				ss.appLogPath = ss.appProfileDir + ss.appLogFilename;

				throw;
			}
		}
	}

	// cannot execute multiple kernel instances in same profile-directory
	SingleInstanceFlagSentry  profileDirFlag("Local\\" + ss.appProfileDir + KERNEL_NAME);
	if(! profileDirFlag.IsAcquired()) {

		olog << e_error << "Cannot execute multiple kernel instances ";
		olog << "in same application-profile directory. Exiting." << endpara << e_info;

		// silent exit
		bUseProfileDir = false;
		return;
	}

	// check if application requested show-log-only
	if(bShowLogAndExit) {
		executionParams.bWaitCloseConfirmationOnExit = true;
		return;
	}

	//--------------------------
	// Change process directory.
	//--------------------------
	olog << "Changing process-directory to: " << ss.appProcessDir << endpara;
	if(! file_system::SetProcessDirectory(ss.appProcessDir)) {
		VW_THROW(KernelException, "Failed to change process directory.");
	}

	//----------------------------
	// Application initialization.
	//----------------------------
	try {
		gui::Manager&     gm  = gui::Manager::Instance();
		DataCodecManager& dcm = DataCodecManager::Instance();

		// auto-init
		olog << caption("auto init");

		Application::AutoInitInfo aii;
		aii.bDeveloperMode = executionParams.bDeveloperMode;

		pApplication->QueryAutoInit(aii);

		// register signal-codecs
		if(aii.bAutoRegisterStandardSignalCodecs) {
			olog << "Registering signal-codecs\n";

			dcm.RegisterDataCategory(DataCategory_Signal);
			RegisterDefaultSignalFileExtensions();

			dcm.SetEncoderDefaultParamList(DataCategory_Signal, SignalCodecParamFactory());
			dcm.SetDecoderDefaultParamList(DataCategory_Signal, SignalCodecParamFactory());

			dcm.RegisterEncoder(new SignalEncoderVWDP);
			dcm.RegisterDecoder(new SignalDecoderVWDP);
		}

		// register image-codecs
		if(aii.bAutoRegisterStandardImageCodecs) {
			olog << "Registering image-codecs\n";

			dcm.RegisterDataCategory(DataCategory_Image);
			RegisterDefaultImageFileExtensions();

			dcm.SetEncoderDefaultParamList(DataCategory_Image, SignalCodecParamFactory());
			dcm.SetDecoderDefaultParamList(DataCategory_Image, SignalCodecParamFactory());

			dcm.RegisterEncoder(new ImageEncoderGdiplus);
			dcm.RegisterDecoder(new ImageDecoderGdiplus);
			dcm.RegisterEncoder(new ImageEncoderVWDP);
			dcm.RegisterDecoder(new ImageDecoderVWDP);
		}

		// register volume-codecs
		if(aii.bAutoRegisterStandardVolumeCodecs) {
			olog << "Registering volume-codecs\n";

			dcm.RegisterDataCategory(DataCategory_Volume);
			RegisterDefaultVolumeFileExtensions();

			dcm.SetEncoderDefaultParamList(DataCategory_Volume, SignalCodecParamFactory());
			dcm.SetDecoderDefaultParamList(DataCategory_Volume, SignalCodecParamFactory());

			dcm.RegisterEncoder(new VolumeEncoderVWDP);
			dcm.RegisterDecoder(new VolumeDecoderVWDP);
		}

		// register standard resources
		if(aii.bAutoRegisterStandardTileTextures) {
			olog << "Registering standard tile-textures\n";
			gui::std_tile_textures::RegisterTileTexture_Set0(olog);
		}
		if(aii.bAutoRegisterStandardFonts) {
			olog << "Registering standard fonts\n";
			gui::std_fonts::RegisterFont_Default(olog);
			gui::std_fonts::RegisterFont_Console(olog);
		}
		if(aii.bAutoRegisterStandardCursors) {
			olog << "Registering standard cursors\n";
			gui::std_cursors::RegisterCursors(olog);
		}
		olog << nl;

		// dump available fonts
		if(aii.bDumpSystemFonts) {
			std::vector<std::string>  fontFamilyNames;
			win_utils::GdiplusGetFontFamilyNames(fontFamilyNames);

			olog << "Installed system-fonts:\n";
			for(uint i = 0; i < fontFamilyNames.size(); i++) {
				olog << tab << fontFamilyNames[i] << nl;
			}
		}

		// app-specific init
		olog << caption("application init");
		pApplication->Initialize();

		// dump various init-info
		olog << nl << caption("post-init info");

		if(aii.bDumpSignalCodecs) {
			olog << "\nRegistered signal-codecs:\n";
			dcm.DumpRegisteredEncoders(DataCategory_Signal, olog);
			dcm.DumpRegisteredDecoders(DataCategory_Signal, olog);
		}
		if(aii.bDumpImageCodecs) {
			olog << "\nRegistered image-codecs:\n";
			dcm.DumpRegisteredEncoders(DataCategory_Image, olog);
			dcm.DumpRegisteredDecoders(DataCategory_Image, olog);
		}
		if(aii.bDumpVolumeCodecs) {
			olog << "\nRegistered volume-codecs:\n";
			dcm.DumpRegisteredEncoders(DataCategory_Volume, olog);
			dcm.DumpRegisteredDecoders(DataCategory_Volume, olog);
		}
		if(aii.bDumpTileTextures) {
			olog << "\nRegistered tile-textures:\n";
			gm.DumpRegisteredTileTextures(olog, 1);
		}
		if(aii.bDumpFonts) {
			olog << "\nRegistered fonts:\n";
			gm.DumpRegisteredFonts(olog, 1);
		}
		if(aii.bDumpCursors) {
			olog << "\nCursors:\n";
			gm.DumpRegisteredCursors(olog, 1);
		}
		if(aii.bDumpShaders) {
			olog << "\nShaders:\n";
			gm.DumpRegisteredShaders(olog, 1);
		}
		if(aii.bDumpComponentHierarchy) {
			olog << "\nComponent hierarchy:\n";
			gm.DumpComponentHierarchy(olog, 1);
		}
		if(aii.bDumpCycleQueue) {
			olog << "\nCycle-queue:\n";
			gm.DumpCycleQueue(olog, 1);
		}

		// continue auto-init
		olog << "\nForcing initial gui sync\n";
		gm.SyncObjectStates();
		gm.ShowGUICursor(aii.bShowGUICursor);
		gm.ShowSystemCursor(aii.bShowSystemCursor);

		olog << nl;

		// update developer-mode state
		if(executionParams.bDeveloperMode != aii.bDeveloperMode) {
			executionParams.bDeveloperMode = aii.bDeveloperMode;
			olog << "\nDeveloperMode changed to: " << ToString(executionParams.bDeveloperMode) << nl;
		}
	}
	catch(Exception& e) {
		ReportException(e);
		VW_THROW(KernelException, "Application failed to initialize.");
	}

	// update log-window title
	if(executionParams.bDeveloperMode) {
		pLogWindowThread->SetTitle(KERNEL_NAME);
	}

	// notify gui-manager
	GUIManager_ConnectApplication(true);

	//---------------------------------
	// Setup generic-extension plugins.
	//---------------------------------
	pPluginManager->ScanGenericExtensionPlugins();

	olog << caption("setting-up generic-extensions used by kernel");

	// create lib-wrapper-dxerr8
	const std::string& interfaceNameLibWrapperDXErr8 = VW_INTERFACE_NAME_LibWrapperDXErr8;

	GenericExtensionPlugin* pLibWrapperDXErr8Plugin =
		pPluginManager->FindKernelUsedGenericExtensionPlugin(0, &interfaceNameLibWrapperDXErr8);

	if(pLibWrapperDXErr8Plugin) {
		LibWrapperDXErr8* lw = InterfaceCast<LibWrapperDXErr8>(pLibWrapperDXErr8Plugin->GetGenericExtension());
		if(lw) {
			olog << "Initializing LibWrapperDXErr8\n";
			lw->HookProcAddresses();
		}
	}
	olog << nl;

	// signal application to connect generic extensions
	olog << caption("setting-up generic-extensions used by application");

	pApplication->ConnectGenericExtensions(); //exception throw allowed
	olog << nl;

	//---------------------------------
	// Setup direct-input-proxy plugin.
	//---------------------------------
	pPluginManager->ScanDirectInputProxyPlugins();

	//---------------------------------------------
	// Setup human-interface-device-driver plugins.
	//---------------------------------------------
	pPluginManager->ScanHIDDriverPlugins();

	//---------------------
	// Setup video plugins.
	//---------------------
	pPluginManager->ScanVideoSubsystemPlugins();
	pPluginManager->ScanVideoRendererPlugins();

	//--------------------
	// End startup notify.
	//--------------------
	olog << caption("end startup");
	pApplication->EndStartup();
	olog << nl;

	//---------------
	// Enter Level-2.
	//---------------
	Level2();

	//------------------------------
	// Application-specific cleanup.
	//------------------------------
	olog << caption("application cleanup");
	olog << "CleanExitFlag = " << bCleanExit << nl;

	if(! bCleanExit) {
		olog << "Failure origin at kernel-level: " << failureDesc.origin << nl;
	}

	pApplication->Cleanup(bCleanExit, failureDesc);
	olog << nl;

	//------------------------------------------------------------------
	// Cleanup of Level1-objects automatically handled by auto-pointers.
	//------------------------------------------------------------------
}

