

//--------------------------------------------------------------------
#include "private\vw_kernel.h"
#include "private\LogWindowThread.h"
#include "private\VideoManager.h"
#include "private\HIDManager.h"
#include "private\AppSubsysManager.h"
#include "GUI\GUI_Manager.h"
#include "GUI\GUI_StdObjects.h"


//====================================================================
void  vw_kernel::Level2()
{
	KernelLevelSentry  kls(*this, KernelLevel_ConfigLoop);

	StringStorage& ss = *pStringStorage;

	//---------------
	// Level-2 frame.
	//---------------
	try {
		//--------------------
		// Configuration-loop.
		//--------------------
		while(! bShutdown) {
			// disconnect hid-s from window
			if(bHIDReconfigPending  ||  bVideoReconfigPending) {
				pHIDManager->DisconnectDevicesFromWindow();
			}

			// disconnect app-specific subsystems from window
			if(bSubsysReconfigPending  ||  bVideoReconfigPending) {
				pAppSubsysManager->DisconnectSubsystemsFromAppWindow();
			}

			// video configuration
			bool bWindowRecreated = false;

			if(bVideoReconfigPending) {
				bVideoReconfigPending = false;
				bWindowRecreated = true;

				pVideoManager->CleanupWindow();
				pVideoManager->InitializeWindow(); //exception-throw allowed

				NotifyVideoReconfigObservers();

				// Video-manager does not send first activate-event by design;
				// Reseting flag here forces ReportAppWindowActive() to dispatch
				// activate-event before entering Level3.

				bAppWindowActive = false;
			}
			VideoSubsystemWindow& videoWindow = *pVideoManager->GetVideoWindow();
			videoWindow.EnableSystemCommands(false, false, false);

			// hid configuration
			if(bHIDReconfigPending) {
				bHIDReconfigPending = false;

				pHIDManager->CleanupDevices();
				pHIDManager->InitializeDevices(); //exception-throw allowed

				NotifyHIDReconfigObservers();
			}

			// connect hid-s to window
			pHIDManager->ConnectDevicesToWindow(videoWindow);
			pHIDManager->OnAppWindowActive(bAppWindowActive);

			// app-specific subsystem configuration
			if(bSubsysReconfigPending) {
				bSubsysReconfigPending = false;

				pAppSubsysManager->CleanupSubsystems();
				pAppSubsysManager->InitializeSubsystems(); //exception-throw allowed
			}

			// connect app-specific subsystems to window
			pAppSubsysManager->ConnectSubsystemsToAppWindow(videoWindow);

			// generic configuration
			if(bGenericReconfigPending) {
				bGenericReconfigPending = false;

				// exception-throw allowed
				pApplication->GenericReconfig(iGenericReconfigCount, iGenericReconfigCode);
				iGenericReconfigCount++;
			}

			// move video-window to foreground
			videoWindow.Show();
			videoWindow.Activate();

			// remove log-window in non-developer mode
			if(! executionParams.bDeveloperMode) {
				if(pLogWindowThread->IsRunning()) {
					TerminateLogWindowThread();
				}
			}

			// prepare execution-states for main-loop
			if(executionParams.bDeveloperMode) {
				pLogWindowThread->EnableCloseGadget(true, true);
			}
			ReportAppWindowActive(videoWindow.IsActive());

			// notify gui-manager about window creation
			// (must be called after activate-event dispatch)
			if(bWindowRecreated) {
				GUIManager_NotifyApplicationWindowCreated();
			}

			// spin frame-loop
			olog << nl << caption("FrameLoop start " + ToString(iFrameLoopIteration));

			Level3();

			olog << nl << caption("FrameLoop end " + ToString(iFrameLoopIteration++)) << std::endl;

			// restore states for configuration-loop
			if(executionParams.bDeveloperMode) {
				pLogWindowThread->EnableCloseGadget(false, false);
			}
		}
	}

	//----------------
	// Forced cleanup.
	//----------------
	catch(Exception& e) {
		ReportException(e, true);
		PrepareLogWindowForCleanup();
		Level2Cleanup(true);
		return;
	}
	catch(std::exception& e) {
		ReportException(e, true);
		PrepareLogWindowForCleanup();
		Level2Cleanup(true);
		return;
	}
	catch(...) {
		ReportExceptionUnknown("Hard cleanup enabled: " +
					ToString(executionParams.bEnableHardCleanupConfigLoop));

		PrepareLogWindowForCleanup();

		if(executionParams.bEnableHardCleanupConfigLoop) {
			Level2Cleanup(true);
		}
		ReportHardRethrow();
		throw;
	}

	//-----------------
	// Regular cleanup.
	//-----------------
	PrepareLogWindowForCleanup();
	Level2Cleanup(false);
}


//====================================================================
void  vw_kernel::Level2Cleanup(bool bForced)
{
	if(bForced) {
		olog << e_warning << "Forcing subsystem-cleanup.\n\n" << e_info;
	}
	else {
		olog << "Regular subsystem-cleanup.\n" << std::endl;
	}

	//-------------------
	// Subsystem cleanup.
	//-------------------
	pAppSubsysManager->DisconnectSubsystemsFromAppWindow();
	pAppSubsysManager->CleanupSubsystems();

	pHIDManager->DisconnectDevicesFromWindow();
	pHIDManager->CleanupDevices();

	pVideoManager->CleanupWindow();
}


//====================================================================
void  vw_kernel::NotifyVideoReconfigObservers()
{
	olog << "Notifying video-reconfig observers.\n";

	bAllowChangeReconfigObservers = false;
	for(uint i = 0; i < pReconfigObservers->size(); i++) {
		(*pReconfigObservers)[i]->OnVideoReconfig();
	}
	bAllowChangeReconfigObservers = true;

	olog << nl;
}


//====================================================================
void  vw_kernel::NotifyHIDReconfigObservers()
{
	olog << "Notifying hid-reconfig observers.\n";

	bAllowChangeReconfigObservers = false;
	for(uint i = 0; i < pReconfigObservers->size(); i++) {
		(*pReconfigObservers)[i]->OnHIDReconfig();
	}
	bAllowChangeReconfigObservers = true;

	olog << nl;
}


//====================================================================
void  vw_kernel::NotifyVSyncReconfigObservers(bool bVSyncEnabled)
{
	olog << "Notifying vsync-reconfig observers.\n";

	bAllowChangeReconfigObservers = false;
	for(uint i = 0; i < pReconfigObservers->size(); i++) {
		(*pReconfigObservers)[i]->OnVSyncReconfig(bVSyncEnabled);
	}
	bAllowChangeReconfigObservers = true;

	olog << nl;
}

