

//--------------------------------------------------------------------
#include "private\vw_kernel.h"
#include "private\VideoManager.h"
#include "private\HIDManager.h"
#include "BaseTools\STLExtensions.h"
#include "GUI\GUI_Manager.h"


//====================================================================
void  vw_kernel::Level3()
{
	KernelLevelSentry  kls(*this, KernelLevel_FrameLoop);

	bSpinFrameLoop = true;
	ResetFrameTimer();

	//------------------------------
	// Spin until exit is requested.
	//------------------------------
	while(bSpinFrameLoop) {
		VW_PROFILE_ROOT_BLOCK;

		//-----------------------------------------------
		// Priority task: process all available messages.
		//-----------------------------------------------
		{
		VW_PROFILE_BLOCK("MessageDispatch");

		MSG msg;
		while(::PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {

			// auto-exit on standard quit-message
			if(msg.message == WM_QUIT) {

				RequestShutdown();

				// skip additional cycle
				goto labelFrameLoopExit;
			}

			// process message
			::TranslateMessage(&msg);
			::DispatchMessage(&msg);
		}
		}
		//---------------------------------------------
		// No messages currently available, application
		// is active. Use idle-time.
		//---------------------------------------------
		if(bAppWindowActive) {
			VW_PROFILE_BLOCK("AppWindowActive");

			// check subsystems
			pVideoManager->ErrorCheck();
			bool bVideoOK = pVideoManager->IsReadyToRender();
			bool bSubsystemsOK = pApplication->CheckSubsystems();
			bool bAllOK = bVideoOK  &&  bSubsystemsOK;

			// execute frame-sequence
			if(bAllOK) {
				VW_PROFILE_BLOCK("SubsystemsReady");

				//---------------------------------------------------------------
				// Compute frame times.
				// Loss of precision occurs with large time-differences, which
				// eventually leads to frame-delta of 0 (at least theoretically).
				// Nevertheless, greater precision is always desirable,
				// so reset the timer periodically.
				//---------------------------------------------------------------
				double secsFrameStart = pHighResTimer->GetElapsedSeconds();
				secsFrameDelta = secsFrameStart - secsLastFrameStart;
				secsLastFrameStart = secsFrameStart;

				if(secsFrameDelta > 1) {
					ResetFrameTimer();
				}

				// notifications
				NotifyFrameObservers(secsFrameDelta);

				// signal frame-in-progress
				{
					VW_PROFILE_BLOCK("FrameBegin");
					bFrameInProgress = true;
					pApplication->FrameBegin(secsFrameDelta);
				}

				// sample input devices
				{
					VW_PROFILE_BLOCK("SampleInput");
						{
						VW_PROFILE_BLOCK("AutoSampleDevices");
						pHIDManager->AutoSampleDevices();
						}
						{
						VW_PROFILE_BLOCK("FrameSampleInput");
						pApplication->FrameSampleInput();
						}
				}

				// auto-dispatch gui-events (pre-update-states)
				bool bGUIEventsDispatched = false;

				if(executionParams.bAutoDispatchGUIEvents  &&
					executionParams.bGUIEventsPreFrameUpdateStates)
				{
					VW_PROFILE_BLOCK("AutoDispatchGUIEvents");
					gui::Manager::Instance().DispatchInputEvents();
					bGUIEventsDispatched = true;
				}

				// per-frame state update
				{
					VW_PROFILE_BLOCK("FrameUpdateStates");
					pApplication->FrameUpdateStates(secsFrameDelta);

					// auto-dispatch gui-events (post-update-states)
					if(executionParams.bAutoDispatchGUIEvents  &&  ! bGUIEventsDispatched)
					{
						VW_PROFILE_BLOCK("AutoDispatchGUIEvents");
						gui::Manager::Instance().DispatchInputEvents();
					}
				}

				// pre render-video
				{
					VW_PROFILE_BLOCK("FramePreRenderVideo");
					pApplication->FramePreRenderVideo();
				}

				// reset gui-stats
				if(executionParams.bAutoClearGUIStats) {

					pVideoManager->GetVideoRenderer()->GetGUIRenderer().ClearStats();
				}

				// perform video-rendering
				pVideoManager->Render(bAutoPresentBackBuffer);

				// post render-video
				{
					VW_PROFILE_BLOCK("FramePostRenderVideo");
					pApplication->FramePostRenderVideo();
				}

				// release frame-in-progress
				{
					VW_PROFILE_BLOCK("FrameEnd");
					pApplication->FrameEnd();
				}
				bFrameInProgress = false;

				// next-frame
				iFrameNumber++;
			}

			// fix broken subsystems
			else {
				VW_PROFILE_BLOCK("FixBrokenSubsystems");

				if(! bVideoOK) {
					VW_PROFILE_BLOCK("FixVideoSubsystem");
					pVideoManager->Fix();
				}
				if(! bSubsystemsOK) {
					VW_PROFILE_BLOCK("AppFixSubsystems");
					pApplication->FixSubsystems();
				}
			}
		}

		//---------------------------------------------
		// No messages currently available, application
		// is minimized or background. Wait activation.
		//---------------------------------------------
		else {
			VW_PROFILE_BLOCK("AppWindowInactive");
			::WaitMessage();
		}
	}

	//------------
	// Exit-point.
	//------------
	labelFrameLoopExit:

	bSpinFrameLoop = false;
}


//====================================================================
void  vw_kernel::NotifyFrameObservers(double secsFrameTimeDelta)
{
	bAllowChangeFrameObservers = false;

	// frame notification
	for(uint i = 0; i < pFrameObservers->size(); i++) {
		FrameObserver* fo = (*pFrameObservers)[i];

		bool bAutoUnregister = false;
		fo->OnFrame(bAutoUnregister, iFrameNumber, secsFrameTimeDelta);

		if(bAutoUnregister) {
			pFrameObserversTemp->push_back(fo);
		}
	}

	// auto-unregistration
	if(pFrameObserversTemp->size()) {
		while(pFrameObserversTemp->size()) {
			EraseFromContainer(*pFrameObservers, pFrameObserversTemp->back());
			pFrameObserversTemp->pop_back();
		}
	}
	bAllowChangeFrameObservers = true;
}


