

//--------------------------------------------------------------------
#include "private\LogWindowThread.h"
#include "BaseTools\PlatformInfo.h"
#include "BaseTools\ConsoleBuffers.h"
#include "Framework\KernelMacros.h"

using namespace vertexware;


//====================================================================
#define   WINPOS_TOP_RIGHT_OFFSET_Y   32


//====================================================================
LogWindowThread::LogWindowThread()
{
	pConOutWin = 0;

	sTitle = VW_GET_CLASS_NAME(*this);
	bEnableCloseGadget = true;
	bCallKernelShutdownOnClose = false;
	windowPosition = WindowPosition_Center;
	bSplashWindow = false;
	bUnsplashWindow = false;

	bUpdateTitle = false;
	bUpdateCloseGadget = false;
	bUpdateLayout = false;
	bAutoActivate = false;
}


//====================================================================
void  LogWindowThread::SetTitle(const std::string& s)
{
	DynamicUpdateSentry  dus(*this);
	sTitle = s;
	bUpdateTitle = true;
}


//====================================================================
void  LogWindowThread::EnableCloseGadget(bool _bEnableCloseGadget, bool _bCallKernelShutdownOnClose)
{
	DynamicUpdateSentry  dus(*this);
	bEnableCloseGadget = _bEnableCloseGadget;
	bCallKernelShutdownOnClose = _bCallKernelShutdownOnClose;
	if(! bEnableCloseGadget) bCallKernelShutdownOnClose = false;
	bUpdateCloseGadget = true;
}


//====================================================================
void  LogWindowThread::SetWindowPosition(eWindowPosition pos)
{
	windowPosition = pos;
}


//====================================================================
void  LogWindowThread::SplashWindow()
{
	DynamicUpdateSentry  dus(*this);

	bSplashWindow = true;
	bUpdateLayout = true;
}


//====================================================================
void  LogWindowThread::UnsplashWindow()
{
	DynamicUpdateSentry  dus(*this);

	if(pConOutWin  &&  ! pConOutWin->IsMinimized()  &&  ! pConOutWin->IsSplashSized()) {
		bAutoActivate = true;
	}
	else {
		bUnsplashWindow = true;
		bUpdateLayout = true;
	}
}


//====================================================================
void  LogWindowThread::FocusOnLastLine()
{
	DynamicUpdateSentry  dus(*this);
	if(pConOutWin) {
		pConOutWin->FocusOnLastLine();
	}
}


//====================================================================
uint  LogWindowThread::DoWork()
{
	// create window
	pConOutWin = new ConsoleOutputWindow;
	pConOutWin->SetTextSource(&Kernel::Instance().GetConsoleBuffers().GetOutputBuffer());

	RealizeTitle();
	RealizeCloseGadget();
	RealizeLayout();

	// present window
	pConOutWin->Show();

	// signal ready state
	SignalReady();

	// message loop
	bool bSpin = true;
	while(bSpin) {

		MSG  msg;
		switch(win_utils::DispatchOneMessage(msg)) {

			case win_utils::DispatchResult_MessageNotAvailable: {
				::WaitMessage();
				break;
			}
			case win_utils::DispatchResult_FoundQuitMessage: {
				if(bCallKernelShutdownOnClose) {
					Kernel::Instance().RequestShutdown();
				}
				else {
					bSpin = false;
				}
				break;
			}
		}
		if(bUpdateTitle) {
			RealizeTitle();
		}
		if(bUpdateCloseGadget) {
			RealizeCloseGadget();
		}
		if(bUpdateLayout) {
			RealizeLayout();
		}
		if(bAutoActivate) {
			RealizeAutoActivate();
		}
	}

	// guard object destruction
	CriticalSectionSentry css(cs);

	// cleanup window
	SafeDelete(pConOutWin);
	return 0;
}


//====================================================================
void  LogWindowThread::RealizeTitle()
{
	bUpdateTitle = false;
	if(! pConOutWin) return;

	pConOutWin->SetTitleText(sTitle);
	pConOutWin->SetText(sTitle);
}


//====================================================================
void  LogWindowThread::RealizeCloseGadget()
{
	bUpdateCloseGadget = false;
	if(! pConOutWin) return;

	pConOutWin->EnableCloseGadget(bEnableCloseGadget);
}


//====================================================================
void  LogWindowThread::RealizeLayout()
{
	bUpdateLayout = false;
	if(! pConOutWin) return;

	// auto-activate
	pConOutWin->Restore();
	pConOutWin->Activate();

	// setup geometry
	uint desktopW, desktopH;
	platform_info::GetDesktopDisplayMode(&desktopW, &desktopH, 0, 0);

	uint minW, minH;
	pConOutWin->GetMinSize(minW, minH);

	if(bSplashWindow) {
		switch(windowPosition) {

			case WindowPosition_Center:
				pConOutWin->CenterOnDesktop(0, 0);
				break;

			case WindowPosition_TopLeft:
				pConOutWin->SetPositionAndSize(0, 0, minW, minH);
				break;

			case WindowPosition_TopRight:
				pConOutWin->SetPositionAndSize(desktopW - minW, WINPOS_TOP_RIGHT_OFFSET_Y, minW, minH);
				break;
		}
	}
	else if(bUnsplashWindow) {
		switch(windowPosition) {

			case WindowPosition_Center:
				pConOutWin->SetPositionAndSize(0.0625f * desktopW, 0.25f * desktopH,
												0.875f * desktopW, 0.5f * desktopH);
				break;

			case WindowPosition_TopLeft:
				pConOutWin->SetPositionAndSize(0, 0, minW, 0.75f * desktopH);
				break;

			case WindowPosition_TopRight:
				pConOutWin->SetPositionAndSize(desktopW - minW, WINPOS_TOP_RIGHT_OFFSET_Y, minW, 0.75f * desktopH);
				break;
		}
	}
	bSplashWindow = false;
	bUnsplashWindow = false;
}


//====================================================================
void  LogWindowThread::RealizeAutoActivate()
{
	bAutoActivate = false;
	if(! pConOutWin) return;

	// auto-activate
	if(pConOutWin->IsMinimized()) {
		pConOutWin->Restore();
	}
	pConOutWin->Activate();
}


//====================================================================
void  LogWindowThread::EnterDynamicUpdate()
{
	CriticalSectionSentry css(cs);

	if(IsRunning()) {
		Suspend();
	}
}


//====================================================================
void  LogWindowThread::LeaveDynamicUpdate()
{
	CriticalSectionSentry css(cs);

	if(IsRunning()) {
		Resume();
	}
	if(pConOutWin) {
		pConOutWin->PostNullMessage();
	}
}

