

//--------------------------------------------------------------------
#include "private\PluginManager.h"
#include "private\vw_kernel.h"
#include "BaseTools\Stream.h"
#include "BaseTools\String.h"
#include "BaseTools\Memory.h"
#include "BaseTools\STLExtensions.h"
#include "Framework\DirectInputProxy8.h"
#include "Framework\HIDDriver0.h"
#include "Framework\HIDDriver8.h"
#include <algorithm>

using namespace vertexware;


//====================================================================
#define   PLUGIN_EXTENSION   "dll"


//********************************************************************
namespace {


//====================================================================
template<typename ContainerType>
void  StorePluginPointer(ContainerType& c, VertexwarePlugin* vwp)
{
	c.push_back(static_cast<ContainerType::value_type>(vwp));
}


//====================================================================
GenericExtensionPlugin*  FindGenericExtensionPlugin(
								std::vector<GenericExtensionPlugin*>& container,
								const std::string* pPluginName,
								const std::string* pGenericExtensionInterfaceName)
{
	if(! pPluginName  &&  ! pGenericExtensionInterfaceName) return 0;
	for(uint i = 0; i < container.size(); i++) {
		GenericExtensionPlugin* gep = container[i];
		bool bFound = true;
		if(pPluginName  &&  *pPluginName != gep->GetPluginInfo().GetPluginName()) bFound = false;
		if(pGenericExtensionInterfaceName  &&  *pGenericExtensionInterfaceName != gep->genericExtensionInterfaceName) bFound = false;
		if(bFound) return gep;
	}
	return 0;
}


//====================================================================
struct DriverPrioritySortPredicate
{
	bool operator()(const HIDDriverPlugin* hiddp1,
					const HIDDriverPlugin* hiddp2) const
	{
		return (hiddp1->finalDriverPriority < hiddp2->finalDriverPriority);
	}
};


//********************************************************************
}//namespace


//====================================================================
PluginManager::PluginManager()
{
	bDisableAllHIDDriverPlugins = false;
}


//====================================================================
PluginManager::~PluginManager()
{
	// main-container owns all plugins
	DeletePtrContainer(plugins);
}


//====================================================================
void  PluginManager::DisablePlugin(const std::string& pluginName)
{
	disabledPluginNames.push_back(LowerCase(pluginName));
}


//====================================================================
void  PluginManager::DisableAllHIDDriverPlugins()
{
	bDisableAllHIDDriverPlugins = true;
}


//====================================================================
void  PluginManager::LoadAvailablePlugins(const std::string& pluginDirectory)
{
	bool bReportDLLErrors = vw_kernel::Instance().GetReportDLLLoadErrors();

	olog << caption("loading available plugins");
	olog << "Report DLL load errors: " << ToString(bReportDLLErrors) << nl;
	olog << "Plugin directory......: " << pluginDirectory << endpara;

	// setup OS critical-error message-box
	if(! bReportDLLErrors) {
		::SetErrorMode(SEM_FAILCRITICALERRORS);
	}

	// do plugin-directory walk
	Walk(pluginDirectory);
	CheckForDuplicatePlugins();
	ClassifyPlugins();

	disabledPluginNames.clear();
	bDisableAllHIDDriverPlugins = false;

	// restore OS critical-error message-box
	if(! bReportDLLErrors) {
		::SetErrorMode(0);
	}
}


//====================================================================
void  PluginManager::OnFile(uint depth, uint fileAttributes,
							const std::string& fileName, const std::string& path)
{
	if(LowerCase(file_system::GetFileNameExtension(fileName)) == PLUGIN_EXTENSION) {
		olog << "Found DLL: " << fileName << nl;

		// load library
		VertexwarePlugin vwp(path);
		if(vwp.Load()) {

			if(vwp.ImportSymbols()) {
				vwp.ClearImportTable();
				StorePlugin(vwp);
			}
			else {
				// this one is not vertexware plugin
				olog << "Not recognized.\n";
			}
		}
		else {
			// not a valid DLL or load-error
			olog << e_error << "Cannot load. Check DLL dependencies.\n" << e_info;
		}
		olog << nl;
	}
}


//====================================================================
void  PluginManager::StorePlugin(VertexwarePlugin& vwp)
{
	olog << e_attention << vwp.GetFormattedName() << nl << e_info;

	// check vw-version & dump info
	if(! vwp.GetPluginInfo().GetLibraryVersion().IsCompatibleWith(library_info::LibraryVersion())) {
		olog << e_error << "Library-version mismatch: ";
		olog << vwp.GetPluginInfo().GetLibraryVersion().MakeString() << nl;
		olog << e_info;
		return;
	}
	vwp.GetPluginInfo().Dump(olog, 1);

	// ignore specified plugins, ignored-names are already lowercase
	std::string  vwpPluginName = LowerCase(vwp.GetPluginInfo().GetPluginName());

	if(std::find(disabledPluginNames.begin(), disabledPluginNames.end(), vwpPluginName)
		!= disabledPluginNames.end())
	{
		olog << "Disabled by command-line.\n";
		return;
	}

	// create exact plugin-type
	VertexwarePlugin* pPlugin = 0;

	switch(vwp.GetPluginInfo().GetPluginType()) {

		case PluginType_GenericExtension:
			pPlugin = new GenericExtensionPlugin;
			break;

		case PluginType_Application:
			pPlugin = new ApplicationPlugin;
			break;

		case PluginType_DirectInputProxy:
			pPlugin = new DirectInputProxyPlugin;
			break;

		case PluginType_HIDDriver:
			if(bDisableAllHIDDriverPlugins) {
				olog << "Disabled by command-line.\n";
				return;
			}
			pPlugin = new HIDDriverPlugin;
			break;

		case PluginType_VideoSubsystem:
			pPlugin = new VideoSubsystemPlugin;
			break;

		case PluginType_VideoRenderer:
			pPlugin = new VideoRendererPlugin;
			break;

		default:
			olog << e_error << "Unsupported plugin type.\n" << e_info;
			return;
	}

	// Create plugin-object.
	// Unloading the dll before exception is caught causes access-violations,
	// so catch them here, report and continue.
	olog << "Creating plugin-object.\n";
	try {
		pPlugin->SafeCopyFrom(vwp);

		if(! pPlugin->CreatePluginObject()) {
			VW_THROW(KernelException, "Plugin failed to create plugin-object.");
		}

		// store plugin
		pPlugin->EnableAutoDeletePluginObject(true);
		plugins.push_back(pPlugin);
	}
	catch(Exception& e) {
		vw_kernel::Instance().ReportException(e);

		// destroy plugin
		delete pPlugin;
	}
}


//====================================================================
void  PluginManager::CheckForDuplicatePlugins()
{
	if(plugins.empty()) return;

	// PluginName is unique identifier (case-insensitive)
	bool bOK = true;
	for(uint i = 0; i < plugins.size() - 1; i++) {
		const VertexwarePlugin& vwp1 = *plugins[i];
		const std::string vwp1PluginName = LowerCase(vwp1.GetPluginInfo().GetPluginName());

		for(uint j = i + 1; j < plugins.size(); j++) {
			const VertexwarePlugin& vwp2 = *plugins[j];

			if(vwp1PluginName == LowerCase(vwp2.GetPluginInfo().GetPluginName())) {
				olog << e_error << vwp1.GetFormattedName() << " and " << vwp2.GetFormattedName();
				olog << " are duplicates.\n" << e_info;
				bOK = false;
			}
		}
	}
	olog << nl;
	if(! bOK) {
		VW_THROW(KernelException, "Found duplicate plugins.");
	}
}


//====================================================================
void  PluginManager::ClassifyPlugins()
{
	for(uint i = 0; i < plugins.size(); i++) {
		VertexwarePlugin* vwp = plugins[i];

		switch(vwp->GetPluginInfo().GetPluginType()) {

			case PluginType_GenericExtension:
				StorePluginPointer(kernelUsedGenericExtensionPlugins, vwp);
				break;

			case PluginType_Application:
				StorePluginPointer(applicationPlugins, vwp);
				break;

			case PluginType_DirectInputProxy:
				StorePluginPointer(directInputProxyPlugins, vwp);
				break;

			case PluginType_HIDDriver:
				StorePluginPointer(hidDriverPlugins, vwp);
				break;

			case PluginType_VideoSubsystem:
				StorePluginPointer(videoSubsystemPlugins, vwp);
				break;

			case PluginType_VideoRenderer:
				StorePluginPointer(videoRendererPlugins, vwp);
				break;
		}
	}
}


//====================================================================
void  PluginManager::DeletePlugin(VertexwarePlugin* pPlugin)
{
	std::vector<VertexwarePlugin*>::iterator it = plugins.begin();
	while(it != plugins.end()) {
		VertexwarePlugin* vwp = *it;
		if(vwp == pPlugin) {
			delete vwp;
			plugins.erase(it);
			return;
		}
		++it;
	}
}


//====================================================================
bool  PluginManager::SetPluginParameter(const std::string& pluginName,
										const NamedParameter& np)
{
	VertexwarePlugin* vwp = FindPlugin(pluginName);
	if(! vwp) return false;
	return vwp->GetPluginObject()->SetPluginParameter(np);
}


//====================================================================
bool  PluginManager::GetPluginParameter(const std::string& pluginName,
										const std::string& parameterName, NamedParameter& out)
{
	VertexwarePlugin* vwp = FindPlugin(pluginName);
	if(! vwp) return 0;
	return vwp->GetPluginObject()->GetPluginParameter(parameterName, out);
}


//====================================================================
void  PluginManager::ScanGenericExtensionPlugins()
{
	olog << caption("scanning generic-extensions");

	Application& app = Kernel::Instance().GetApplication();

	std::vector<GenericExtensionPlugin*>  acceptedPlugins;
	std::vector<GenericExtensionPlugin*>  rejectedPlugins;

	// preprocess generic-extensions
	{for(uint i = 0; i < kernelUsedGenericExtensionPlugins.size(); i++) {

		GenericExtensionPlugin* gep = kernelUsedGenericExtensionPlugins[i];
		GenericExtension* ge = gep->GetGenericExtension();

		gep->genericExtensionInterfaceName = ge->GetGenericExtensionInterfaceName();

		// dump plugin-specific info
		olog << e_attention << gep->GetFormattedName() << nl << e_info;
		olog << "\tGenericExtensionInterfaceName: " << gep->genericExtensionInterfaceName << nl;

		// classify by usage
		gep->bUsedByKernel = false;
		gep->bUsedByApp = false;

		if(vw_kernel::Instance().ConfirmGenericExtension(gep->GetPluginInfo().GetPluginName(), gep->genericExtensionInterfaceName)) {
			gep->bUsedByKernel = true;
		}
		if(! gep->bUsedByKernel  &&
			app.ConfirmGenericExtension(gep->GetPluginInfo().GetPluginName(), gep->genericExtensionInterfaceName))
		{
			gep->bUsedByApp = true;
		}
		if(gep->bUsedByKernel  ||  gep->bUsedByApp) {
			acceptedPlugins.push_back(gep);
		}
		else {
			rejectedPlugins.push_back(gep);
		}
		olog << nl;
	}}

	// discard unused
	{for(uint i = 0; i < rejectedPlugins.size(); i++) {
		DeletePlugin(rejectedPlugins[i]);
	}}

	// store by usage
	appUsedGenericExtensionPlugins.clear();
	kernelUsedGenericExtensionPlugins.clear();

	{for(uint i = 0; i < acceptedPlugins.size(); i++) {
		GenericExtensionPlugin* gep = acceptedPlugins[i];

		if(gep->bUsedByApp) appUsedGenericExtensionPlugins.push_back(gep);
		if(gep->bUsedByKernel) kernelUsedGenericExtensionPlugins.push_back(gep);
	}}

	// dump used generic-extensions
	olog << "Used by kernel:\n";
	{for(uint i = 0; i < kernelUsedGenericExtensionPlugins.size(); i++) {
		olog << tab << kernelUsedGenericExtensionPlugins[i]->GetFormattedName() << nl;
	}}
	olog << "\nUsed by application:\n";
	{for(uint i = 0; i < appUsedGenericExtensionPlugins.size(); i++) {
		olog << tab << appUsedGenericExtensionPlugins[i]->GetFormattedName() << nl;
	}}
	olog << nl;
}


//====================================================================
uint  PluginManager::GetNumAppUsedGenericExtensionPlugins()
{
	return appUsedGenericExtensionPlugins.size();
}


//====================================================================
GenericExtensionPlugin*  PluginManager::GetAppUsedGenericExtensionPlugin(uint i)
{
	return appUsedGenericExtensionPlugins[i];
}


//====================================================================
GenericExtensionPlugin*  PluginManager::FindAppUsedGenericExtensionPlugin(
												const std::string* pPluginName,
												const std::string* pGenericExtensionInterfaceName)
{
	return FindGenericExtensionPlugin(appUsedGenericExtensionPlugins, pPluginName, pGenericExtensionInterfaceName);
}


//====================================================================
uint  PluginManager::GetNumKernelUsedGenericExtensionPlugins()
{
	return kernelUsedGenericExtensionPlugins.size();
}


//====================================================================
GenericExtensionPlugin*  PluginManager::GetKernelUsedGenericExtensionPlugin(uint i)
{
	return kernelUsedGenericExtensionPlugins[i];
}


//====================================================================
GenericExtensionPlugin*  PluginManager::FindKernelUsedGenericExtensionPlugin(
												const std::string* pPluginName,
												const std::string* pGenericExtensionInterfaceName)
{
	return FindGenericExtensionPlugin(kernelUsedGenericExtensionPlugins, pPluginName, pGenericExtensionInterfaceName);
}


//====================================================================
void  PluginManager::ScanApplicationPlugins(const std::string& requestedPluginName)
{
	olog << caption("scanning applications");

	// dump available applications
	{for(uint i = 0; i < applicationPlugins.size(); i++) {
		olog << e_attention << applicationPlugins[i]->GetFormattedName() << nl << e_info;
	}}
	olog << nl;

	// select one application
	ApplicationPlugin*  pSelection = 0;

	switch(applicationPlugins.size()) {
		case 0: {
			olog << e_error << "No applications available." << nl << e_info;
			break;
		}
		case 1: {
			pSelection = applicationPlugins[0];
			break;
		}
		default: {
			olog << e_warning << "Multiple applications available." << nl << e_info;
			std::string  appPluginName = LowerCase(requestedPluginName);

			for(uint i = 0; i < applicationPlugins.size(); i++) {
				if(LowerCase(applicationPlugins[i]->GetPluginInfo().GetPluginName()) == appPluginName) {
					pSelection = applicationPlugins[i];
				}
			}
		}
	}

	// keep selection and discard others
	{for(uint i = 0; i < applicationPlugins.size(); i++) {
		ApplicationPlugin* p = applicationPlugins[i];
		if(p != pSelection) {
			DeletePlugin(p);
		}
	}}
	applicationPlugins.clear();

	if(! pSelection) {
		VW_THROW(KernelException, "Cannot select application plugin.");
	}
	applicationPlugins.push_back(pSelection);

	// report selection-result
	olog << "Selected application:\n" << pSelection->GetFormattedName() << endpara;
}


//====================================================================
ApplicationPlugin*  PluginManager::GetApplicationPlugin()
{
	if(applicationPlugins.size()) return applicationPlugins[0];
	return 0;
}


//====================================================================
void  PluginManager::ScanDirectInputProxyPlugins()
{
	olog << caption("scanning direct-input-proxies");

	Application& app = Kernel::Instance().GetApplication();

	std::vector<DirectInputProxyPlugin*>  acceptedPlugins;
	std::vector<DirectInputProxyPlugin*>  rejectedPlugins;

	{for(uint i = 0; i < directInputProxyPlugins.size(); i++) {
		DirectInputProxyPlugin* dipp = directInputProxyPlugins[i];
		DirectInputProxy* dip = dipp->GetDirectInputProxy();

		// cache additional info
		dipp->directInputProxyInterfaceName = dip->GetDirectInputProxyInterfaceName();

		bool bDIVersionValid = ExtractDirectInputVersion(*dipp);

		// give application a chance to discard direct-input proxy
		bool bAcceptedByApp = bDIVersionValid  &&
					app.ConfirmDirectInputVersion(dipp->directInputVersion);

		// dump plugin-specific info
		olog << e_attention << dipp->GetFormattedName() << nl << e_info;
		olog << "\tDirectInputProxyInterfaceName: " << dipp->directInputProxyInterfaceName << nl;
		olog << "\tDirectInputVersion...........: " << dipp->directInputVersion << nl;

		if(bDIVersionValid) {
			if(bAcceptedByApp) {
				acceptedPlugins.push_back(dipp);
			}
			else {
				rejectedPlugins.push_back(dipp);
				olog << "\tRejected by application\n";
			}
		}
		else {
			olog << e_error << "\tUnsupported DirectInputProxy interface\n" << e_info;
			rejectedPlugins.push_back(dipp);
		}
		olog << nl;
	}}
	if(directInputProxyPlugins.empty()) {
		olog << nl;
	}

	// discard unusable
	{for(uint i = 0; i < rejectedPlugins.size(); i++) {
		DeletePlugin(rejectedPlugins[i]);
	}}
	directInputProxyPlugins = acceptedPlugins;
}


//====================================================================
DirectInputProxyPlugin*  PluginManager::GetDirectInputProxyPlugin(uint directInputVersion)
{
	// find first match
	for(uint i = 0; i < directInputProxyPlugins.size(); i++) {
		DirectInputProxyPlugin* dipp = directInputProxyPlugins[i];
		if(dipp->directInputVersion == directInputVersion) {
			return dipp;
		}
	}
	return 0;
}


//====================================================================
void  PluginManager::ScanHIDDriverPlugins()
{
	olog << caption("scanning human-interface-device-drivers");

	Application& app = Kernel::Instance().GetApplication();

	std::vector<HIDDriverPlugin*>  acceptedPlugins;
	std::vector<HIDDriverPlugin*>  rejectedPlugins;

	// preprocess hid-drivers
	{for(uint i = 0; i < hidDriverPlugins.size(); i++) {

		HIDDriverPlugin* hiddp = hidDriverPlugins[i];
		HIDDriver* hidd = hiddp->GetHIDDriver();

		// cache additional info
		hiddp->hidDriverInterfaceName = hidd->GetHIDDriverInterfaceName();
		hiddp->defaultDriverPriority = hidd->GetDefaultDriverPriority();
		hiddp->finalDriverPriority = hiddp->defaultDriverPriority;

		bool bDIVersionValid = ExtractDirectInputVersion(*hiddp);

		// give application a chance to change driver-priority or discard driver
		bool bAcceptedByApp = bDIVersionValid  &&
				app.ConfirmHIDDriver(hiddp->GetPluginInfo().GetPluginName(),
									hiddp->defaultDriverPriority, hiddp->finalDriverPriority);

		// dump plugin-specific info
		olog << e_attention << hiddp->GetFormattedName() << nl << e_info;
		olog << "\tHIDDriverInterfaceName: " << hiddp->hidDriverInterfaceName << nl;
		olog << "\tDirectInputVersion....: " << hiddp->directInputVersion << nl;
		olog << "\tDefaultPriority.......: " << hiddp->defaultDriverPriority << nl;
		olog << "\tFinalPriority.........: " << hiddp->finalDriverPriority << nl;

		// reject unusable plugins
		if(bDIVersionValid) {
			if(bAcceptedByApp) {
				acceptedPlugins.push_back(hiddp);
			}
			else {
				rejectedPlugins.push_back(hiddp);
				olog << "\tRejected by application\n";
			}
		}
		else {
			olog << e_error << "\tUnsupported HIDDriver interface\n" << e_info;
			rejectedPlugins.push_back(hiddp);
		}
		olog << nl;
	}}

	// driver-priority must be unique
	CheckForDuplicateDriverPriorities(hidDriverPlugins);

	// discard unusable
	{for(uint i = 0; i < rejectedPlugins.size(); i++) {
		DeletePlugin(rejectedPlugins[i]);
	}}
	hidDriverPlugins = acceptedPlugins;

	// sort by driver-priority
	std::sort(hidDriverPlugins.begin(), hidDriverPlugins.end(),
				DriverPrioritySortPredicate());

	// dump driver queue
	olog << "Final human-interface-device-driver queue:\n";
	{for(uint i = 0; i < hidDriverPlugins.size(); i++) {
		olog << hidDriverPlugins[i]->GetFormattedName() << nl;
	}}
	olog << nl;
}


//====================================================================
void  PluginManager::CheckForDuplicateDriverPriorities(
									std::vector<HIDDriverPlugin*>& iddPlugins)
{
	if(iddPlugins.empty()) return;

	// driver-priorities must be unique
	bool bOK = true;
	for(uint i = 0; i < iddPlugins.size() - 1; i++) {
		const HIDDriverPlugin& iddp1 = *iddPlugins[i];

		for(uint j = i + 1; j < iddPlugins.size(); j++) {
			const HIDDriverPlugin& iddp2 = *iddPlugins[j];

			if(iddp1.finalDriverPriority == iddp2.finalDriverPriority) {
				olog << e_error << iddp1.GetFormattedName() << " and " << iddp2.GetFormattedName();
				olog << " have equal driver-priority.\n" << e_info;
				bOK = false;
			}
		}
	}
	if(! bOK) {
		VW_THROW(KernelException, "Found duplicate input-device-driver priorities.");
	}
}


//====================================================================
bool  PluginManager::ExtractDirectInputVersion(DirectInputProxyPlugin& dipp)
{
	if(dipp.directInputProxyInterfaceName == VW_INTERFACE_NAME_DirectInputProxy8) {
		dipp.directInputVersion = 8;
		return ToBool(InterfaceCast<DirectInputProxy8>(dipp.GetDirectInputProxy()));
	}
	dipp.directInputVersion = 0x7FFFFFFF;
	return false; //unsupported
}


//====================================================================
bool  PluginManager::ExtractDirectInputVersion(HIDDriverPlugin& hiddp)
{
	if(hiddp.hidDriverInterfaceName == VW_INTERFACE_NAME_HIDDriver0) {
		hiddp.directInputVersion = 0;
		return ToBool(InterfaceCast<HIDDriver0>(hiddp.GetHIDDriver()));
	}
	if(hiddp.hidDriverInterfaceName == VW_INTERFACE_NAME_HIDDriver8) {
		hiddp.directInputVersion = 8;
		return ToBool(InterfaceCast<HIDDriver8>(hiddp.GetHIDDriver()));
	}
	hiddp.directInputVersion = 0x7FFFFFFF;
	return false; //unsupported
}


//====================================================================
uint  PluginManager::GetNumHIDDriverPlugins()
{
	return hidDriverPlugins.size();
}


//====================================================================
HIDDriverPlugin*  PluginManager::GetHIDDriverPlugin(uint i)
{
	return hidDriverPlugins[i];
}


//====================================================================
void  PluginManager::ScanVideoSubsystemPlugins()
{
	olog << caption("scanning video-subsystems");

	Application& app = Kernel::Instance().GetApplication();

	std::vector<VideoSubsystemPlugin*>  acceptedPlugins;
	std::vector<VideoSubsystemPlugin*>  rejectedPlugins;

	// preprocess video-subsystems
	for(uint i = 0; i < videoSubsystemPlugins.size(); i++) {

		VideoSubsystemPlugin* vsp = videoSubsystemPlugins[i];
		VideoSubsystem* vs = vsp->GetVideoSubsystem();

		olog << e_attention << vsp->GetFormattedName() << nl << e_info;

		// cache additional info
		vsp->videoSubsystemInterfaceName = vs->GetVideoSubsystemInterfaceName();

		// construct profile-names
		std::string  profileLogName(vsp->GetPluginInfo().GetPluginName());
		profileLogName.append(" profile");

		std::string  profileFileName;
		if(vw_kernel::Instance().IsUsingProfileDirectory()) {
			profileFileName = vw_kernel::Instance().GetProfileDirectory();
			profileFileName.append(vsp->GetBaseName()).append(vw_kernel::Instance().GetLogFilenameExtension());
		}

		// build video-profile
		if(vs->BuildVideoProfile(profileLogName, profileFileName, vsp->videoProfileUnfiltered)) {

			// filter video-profile
			vsp->videoProfileUnfiltered.Filter(vsp->videoProfileFiltered);
			app.FilterVideoProfile(vsp->videoProfileUnfiltered, vsp->videoProfileFiltered);

			// dump plugin-specific info
			olog << "\tVideoSubsystemInterfaceName: " << vsp->videoSubsystemInterfaceName << nl;

			olog << "\n\tDriver-info:\n";
			vsp->videoProfileFiltered.DumpGeneralInfo(olog, 1);

			olog << "\n\tApplication-filtered display-modes:\n";
			vsp->videoProfileFiltered.DumpDisplayModeProperties(olog, 1);

			olog << "\n\tApplication-filtered frame-buffer-configurations:\n";
			vsp->videoProfileFiltered.DumpFrameBufferConfigs(olog, false, 1);

			// reject unusable plugins
			if(vsp->videoProfileFiltered.IsEmpty()) {
				rejectedPlugins.push_back(vsp);
				olog << e_warning << "\n\tVideo-subsystem filtered-out by application.\n" << e_info;
			}
			else {
				acceptedPlugins.push_back(vsp);
			}
		}
		else {
			olog << e_error << "\tCannot build video-profile.\n" << e_info;
		}
		olog << nl;
	}

	// discard unusable
	{for(uint i = 0; i < rejectedPlugins.size(); i++) {
		VideoSubsystemPlugin* vsp = rejectedPlugins[i];
		DeletePlugin(vsp);
	}}
	videoSubsystemPlugins = acceptedPlugins;

	// must have at least one video-subsystem
	if(videoSubsystemPlugins.empty()) {
		VW_THROW(KernelException, "No video-subsystem found that application can use.");
	}

	// dump usable video-subsystems
	olog << "Usable video-subsystems:\n";
	{for(uint i = 0; i < videoSubsystemPlugins.size(); i++) {
		olog << videoSubsystemPlugins[i]->GetFormattedName() << nl;
	}}
	olog << nl;
}


//====================================================================
VideoSubsystemPlugin*  PluginManager::FindVideoSubsystemPlugin(const std::string& interfaceName)
{
	for(uint i = 0; i < videoSubsystemPlugins.size(); i++) {
		VideoSubsystemPlugin* vsp = videoSubsystemPlugins[i];
		if(vsp->videoSubsystemInterfaceName == interfaceName) {
			return vsp;
		}
	}
	return 0; //not found
}


//====================================================================
void  PluginManager::ScanVideoRendererPlugins()
{
	olog << caption("scanning video-renderers");

	Application& app = Kernel::Instance().GetApplication();

	std::vector<VideoRendererPlugin*>  acceptedPlugins;
	std::vector<VideoRendererPlugin*>  rejectedPlugins;

	// preprocess video-renderers
	for(uint i = 0; i < videoRendererPlugins.size(); i++) {

		VideoRendererPlugin* vrp = videoRendererPlugins[i];
		VideoRenderer* vr = vrp->GetVideoRenderer();

		olog << e_attention << vrp->GetFormattedName() << nl << e_info;

		// cache additional info
		vrp->videoRendererInterfaceName  = vr->GetVideoRendererInterfaceName();
		vrp->videoSubsystemInterfaceName = vr->GetMatchingVideoSubsystemInterfaceName();

		// dump plugin-specific info
		olog << "\tVideoRendererInterfaceName.: " << vrp->videoRendererInterfaceName << nl;
		olog << "\tVideoSubsystemInterfaceName: " << vrp->videoSubsystemInterfaceName << nl;

		// for each loaded video-renderer, matching video-subsystem must be loaded
		bool bAccepted = false;
		if(FindVideoSubsystemPlugin(vrp->videoSubsystemInterfaceName)) {

			if(app.ConfirmVideoRenderer(vrp->GetPluginInfo().GetPluginName(),
										vrp->videoRendererInterfaceName))
			{
				bAccepted = true;
			}
			else {
				olog << "\tRejected by application.\n";
			}
		}
		else {
			olog << e_error << "\tCannot find matching video-subsystem.\n" << e_info;
		}

		// classify
		if(bAccepted) {
			acceptedPlugins.push_back(vrp);
		}
		else {
			rejectedPlugins.push_back(vrp);
		}
		olog << nl;
	}

	// discard unusable video-renderers
	{for(uint i = 0; i < rejectedPlugins.size(); i++) {
		VideoRendererPlugin* vrp = rejectedPlugins[i];
		DeletePlugin(vrp);
	}}
	videoRendererPlugins = acceptedPlugins;

	// must have at least one video-renderer
	if(videoRendererPlugins.empty()) {
		VW_THROW(KernelException, "No video-renderer found that application can use.");
	}

	// dump used video-renderers & collect used video-subsystems
	std::vector<VideoSubsystemPlugin*>  usedVideoSubsystems;

	olog << "Used video-renderers:\n";
	{for(uint i = 0; i < videoRendererPlugins.size(); i++) {
		VideoRendererPlugin* vrp = videoRendererPlugins[i];

		olog << vrp->GetFormattedName() << nl;
		UniqueContainerPushBack(usedVideoSubsystems,
					FindVideoSubsystemPlugin(vrp->videoSubsystemInterfaceName));
	}}
	olog << nl;

	// discard unused video-subsystems
	olog << "Discarding unused video-subsystems:\n";
	{for(uint i = 0; i < videoSubsystemPlugins.size(); i++) {
		VideoSubsystemPlugin* vsp = videoSubsystemPlugins[i];

		std::vector<VideoSubsystemPlugin*>::iterator it =
			std::find(usedVideoSubsystems.begin(), usedVideoSubsystems.end(), vsp);

		// if not found, it is unused
		if(it == usedVideoSubsystems.end()) {
			olog << vsp->GetFormattedName() << nl;
			DeletePlugin(vsp);
		}
	}}
	olog << nl;

	videoSubsystemPlugins = usedVideoSubsystems;
}


//====================================================================
VideoRendererPlugin*  PluginManager::FindVideoRendererPlugin(const std::string& pluginName)
{
	for(uint i = 0; i < videoRendererPlugins.size(); i++) {
		VideoRendererPlugin* vrp = videoRendererPlugins[i];
		if(vrp->GetPluginInfo().GetPluginName() == pluginName) {
			return vrp;
		}
	}
	return 0;
}


//====================================================================
uint  PluginManager::GetNumVideoRendererPlugins()
{
	return videoRendererPlugins.size();
}


//====================================================================
VideoRendererPlugin*  PluginManager::GetVideoRendererPlugin(uint i)
{
	return videoRendererPlugins[i];
}


//====================================================================
VertexwarePlugin*  PluginManager::FindPlugin(const std::string& pluginName)
{
	for(uint i = 0; i < plugins.size(); i++) {
		if(plugins[i]->GetPluginInfo().GetPluginName() == pluginName) return plugins[i];
	}
	return 0;
}

