

//--------------------------------------------------------------------
#include "private\SystemMouse.h"
#include "Framework\KernelMacros.h"
#include "BaseTools\WinUtils.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Debug.h"
#include "BaseTools\PlatformInfo.h"

using namespace vertexware;


//====================================================================
#define   RELATIVE_MODE_GUARD_BAND    2


//====================================================================
SystemMouse8::SystemMouse8(DirectInputProxy8* dip8,
						const DirectInputDeviceInstanceInfo& didii)
	:
	Mouse(true)
{
	// init parent-hid
	SetParentHID(*this);

	olog << "Creating SystemMouse.\n";

	pDirectInputProxy8 = dip8;
	pAppWindow = 0;
	iDIDevice = 0;
	bDIDeviceConnectedToWindow = false;
	bDIDeviceActive = false;
	DIWheelGranularity = 1;
	wheelAbsMax = 0;
	bUsingExclusiveMode = false;
	bAppWindowFullscreen = false;
	bAppWindowActiveCycle = false;
	bRelativeModeBeforeAppWindowActiveCycle = false;

	SetCanDisplaySystemCursor(true);
	CreateDIDevice();

	DeviceInfo().SetDeviceDescription("SystemMouse");
	DeviceInfo().SetDirectInputOnly(false);
	DeviceInfo().SetCanUseDirectInput(pDirectInputProxy8 != 0);
	if(dip8) {
		DeviceInfo().SetDirectInputDeviceInstanceInfo(didii);
	}

	UpdateUsingDirectInput(GetUseDirectInputAtInit() ? ToBool(pDirectInputProxy8) : false);
}


//====================================================================
SystemMouse8::~SystemMouse8()
{
	DisconnectFromAppWindow();
	ReleaseDIDevice();
}


//====================================================================
void  SystemMouse8::ConnectToAppWindow(TopLevelWindow* _pAppWindow,
									HWND hwndAppWindow, bool bFullscreen)
{
	VW_ASSERT(_pAppWindow, "Invalid argument.");

	// setup win32 hooks
	pAppWindow = _pAppWindow;
	bAppWindowFullscreen = bFullscreen;
	bUsingExclusiveMode = false;
	SetupMouseMessageDispatch(false);

	// setup direct-input device behavior
	if(iDIDevice) {
		bDIDeviceConnectedToWindow = false;

		iDIDevice->Unacquire();

		DWORD dwFlags = DISCL_FOREGROUND;
		if(bAppWindowFullscreen  &&  pDirectInputProxy8->IsAllowedFullscreenExclusiveMode()) {
			dwFlags |= DISCL_EXCLUSIVE;
		}
		else {
			dwFlags |= DISCL_NONEXCLUSIVE;
		}
		olog << "Setting DirectInput cooperative level: " << pDirectInputProxy8->GetCooperativeLevelDesc(dwFlags) << nl;

		HRESULT hr = iDIDevice->SetCooperativeLevel(pAppWindow->GetHWND(), dwFlags);
		if(SUCCEEDED(hr)) {
			bUsingExclusiveMode = ToBool(dwFlags & DISCL_EXCLUSIVE);
			bDIDeviceConnectedToWindow = true;
		}
		else {
			olog << e_error << "SetCooperativeLevel failed.\n" << e_info;
		}
	}
	OnUseDirectInput(IsUsingDirectInput());
}


//====================================================================
void  SystemMouse8::DisconnectFromAppWindow()
{
	// disable cursor-clip
	::ClipCursor(0);

	// remove win32 hooks
	ResetMouseMessageDispatch();
	pAppWindow = 0;

	// disconnect direct-input device from window
	ActivateDIDevice(false);
	bDIDeviceConnectedToWindow = false;
}


//====================================================================
void  SystemMouse8::OnAppWindowActive(bool bActive)
{
	if(bActive) {
		// exit cycle
		bAppWindowActiveCycle = false;
		ActivateRelativeMode(bRelativeModeBeforeAppWindowActiveCycle);
	}
	else {
		// enter cycle
		bRelativeModeBeforeAppWindowActiveCycle = IsActiveRelativeMode();
		ActivateRelativeMode(false);
		bAppWindowActiveCycle = true; //change after ActivateRelativeMode()
	}
}


//====================================================================
void  SystemMouse8::OnAppWindowGeometryChanged()
{
	if(! pAppWindow) return;
	pAppWindow->GetVisibleClientAreaScreenCoords(appClientRect);

	appClientSize.Set(appClientRect.right - appClientRect.left,
						appClientRect.bottom - appClientRect.top);

	appClientRectGuardBand = appClientRect;

	appClientRectGuardBand.left   += RELATIVE_MODE_GUARD_BAND;
	appClientRectGuardBand.right  -= RELATIVE_MODE_GUARD_BAND;
	appClientRectGuardBand.top    += RELATIVE_MODE_GUARD_BAND;
	appClientRectGuardBand.bottom -= RELATIVE_MODE_GUARD_BAND;

	appClientCenter.Set((appClientRect.left + appClientRect.right) / 2,
						(appClientRect.top + appClientRect.bottom) / 2);

	SetupCursorStates();
}


//====================================================================
void  SystemMouse8::OnUseDirectInput(bool bUseDI)
{
	ActivateDIDevice(bUseDI);

	HumanInterfaceDevice::UpdateUsingDirectInput(bDIDeviceActive);

	bool bSystemCursorOperational = ! bDIDeviceActive  ||  ! bUsingExclusiveMode;
	SetCanDisplaySystemCursor(bSystemCursorOperational);
	SetCanSyncWithSystemCursor(bSystemCursorOperational);
}


//====================================================================
void  SystemMouse8::OnInputModeChanged()
{
	if(bAppWindowActiveCycle) return;
	SetupCursorStates();
}


//====================================================================
void  SystemMouse8::OnSyncWithSystemCursorChanged()
{
}


//====================================================================
void  SystemMouse8::ClipPosition(win_utils::Point& pos)
{
	if(bAppWindowFullscreen) {
		Clamp<LONG>(pos.x, 0, appClientSize.x - 1);
		Clamp<LONG>(pos.y, 0, appClientSize.y - 1);
	}
	else {
		Clamp<LONG>(pos.x, -1, appClientSize.x);
		Clamp<LONG>(pos.y, -1, appClientSize.y);
	}
}


//====================================================================
bool  SystemMouse8::IsPositionInsideClientArea(const win_utils::Point& pos)
{
	if(pos.x < 0  ||  pos.x >= appClientSize.x) return false;
	if(pos.y < 0  ||  pos.y >= appClientSize.y) return false;
	return true;
}


//====================================================================
void  SystemMouse8::SetupCursorStates()
{
	if(IsActiveRelativeMode()  &&  ! bAppWindowFullscreen) {
		::ClipCursor(&appClientRectGuardBand);
	}
	else {
		::ClipCursor(0);
	}
	lastCursorPos = appClientCenter;
}


//====================================================================
bool  SystemMouse8::SampleInputData()
{
	PreSampleUpdate();

	//----------------------
	// Use DirectInputDevice
	//----------------------
	if(bDIDeviceActive) {

		// get direct-input mouse state
		DIMOUSESTATE2 diMouseState2;
		HRESULT hr = iDIDevice->GetDeviceState(sizeof(diMouseState2), &diMouseState2);

		// handle errors
		if(FAILED(hr)) {
			switch(hr) {
				case DIERR_NOTACQUIRED:
				case DIERR_INPUTLOST:
					// device is not acquired or data stream was lost, try to reacquire
					iDIDevice->Unacquire();
					iDIDevice->Acquire();
					break;

				case E_PENDING:
					break;
			}
			ResetButtonStates();
			return false;
		}

		// relative-mode
		if(IsActiveRelativeMode()) {

			float finalX, finalY;
			ApplyRelativeModeModifiers(diMouseState2.lX, diMouseState2.lY, finalX, finalY);

			AccumDelta(finalX, finalY);
		}

		// absolute-mode
		else {
			// always sync with system-cursor in windowed mode,
			// but only when requested and possible in fullscreen mode
			bool bSync = bAppWindowFullscreen ?
				(IsActiveSyncWithSystemCursor()  &&  ! bUsingExclusiveMode) : true;

			if(bSync) {
				::GetCursorPos(&lastCursorPos);
				::ScreenToClient(pAppWindow->GetHWND(), &lastCursorPos);
			}
			else {
				float finalX, finalY;
				ApplyAbsoluteModeModifiers(diMouseState2.lX, diMouseState2.lY, finalX, finalY);
				lastCursorPos.x += finalX;
				lastCursorPos.y += finalY;
			}
			ClipPosition(lastCursorPos);
			SetPosition(lastCursorPos.x, lastCursorPos.y, IsPositionInsideClientArea(lastCursorPos));
		}

		// report axis states
		AccumWheelDelta(diMouseState2.lZ / DIWheelGranularity);

		// report button states
		for(uint i = 0; i < MouseButton_Count; i++) {
			if(diMouseState2.rgbButtons[ i ] & 0x80) {
				SetButtonPressed(eMouseButton(i));
			}
		}
	}

	//--------------------
	// Use Win32 functions
	//--------------------
	else {
		// read win32 cursor position
		win_utils::Point  cursorPos;
		::GetCursorPos(&cursorPos);

		// relative-mode
		if(IsActiveRelativeMode()) {
			win_utils::Point  deltaPos = cursorPos - lastCursorPos;

			if(deltaPos.x  ||  deltaPos.y) {
				::SetCursorPos(appClientCenter.x, appClientCenter.y);
			}

			float finalX, finalY;
			ApplyRelativeModeModifiers(deltaPos.x, deltaPos.y, finalX, finalY);

			AccumDelta(finalX, finalY);
		}

		// absolute-mode
		else {
			if(! bAppWindowFullscreen) {
				::ScreenToClient(pAppWindow->GetHWND(), &cursorPos);
			}
			ClipPosition(cursorPos);
			SetPosition(cursorPos.x, cursorPos.y, IsPositionInsideClientArea(cursorPos));
		}

		// report axis states
		AccumWheelDelta(wheelAbsMax);
		wheelAbsMax = 0;

		// report button states
		const SHORT mask = SHORT(0x8000);
		if(::GetAsyncKeyState(VK_LBUTTON) & mask) SetButtonPressed(MouseButton_Left);
		if(::GetAsyncKeyState(VK_RBUTTON) & mask) SetButtonPressed(MouseButton_Right);
		if(::GetAsyncKeyState(VK_MBUTTON) & mask) SetButtonPressed(MouseButton_Middle);
		if(::GetAsyncKeyState(VK_XBUTTON1) & mask) SetButtonPressed(MouseButton_X1);
		if(::GetAsyncKeyState(VK_XBUTTON2) & mask) SetButtonPressed(MouseButton_X2);
	}

	// sampling succeeded
	PostSampleUpdate();
	return true;
}


//====================================================================
bool  SystemMouse8::OnMouseMove(const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
{
	// not used
	return true;
}


//====================================================================
bool  SystemMouse8::OnMouseWheel(int delta, const win_utils::Point& mousePosClient,
								WPARAM keyFlags)
{
	// delta is already normalized
	if(delta >= 0) {
		wheelAbsMax = Max(wheelAbsMax, delta);
	}
	else {
		wheelAbsMax = Min(wheelAbsMax, delta);
	}
	return true;
}


//====================================================================
bool  SystemMouse8::OnMouseButtonLeft(bool bDown, const win_utils::Point& mousePosClient,
									WPARAM keyFlags)
{
	// not used
	return true;
}


//====================================================================
bool  SystemMouse8::OnMouseButtonMiddle(bool bDown, const win_utils::Point& mousePosClient,
										WPARAM keyFlags)
{
	// not used
	return true;
}


//====================================================================
bool  SystemMouse8::OnMouseButtonRight(bool bDown, const win_utils::Point& mousePosClient,
										WPARAM keyFlags)
{
	// not used
	return true;
}


//====================================================================
void  SystemMouse8::CreateDIDevice()
{
	iDIDevice = 0;
	if(! pDirectInputProxy8) return;

	// create direct-input system-mouse
	HRESULT hr = pDirectInputProxy8->GetDirectInputObject()->CreateDevice(
							pDirectInputProxy8->GetGuidSysMouse(), &iDIDevice, 0);

	if(SUCCEEDED(hr)) {
		hr = iDIDevice->SetDataFormat(&pDirectInputProxy8->GetDataFormatMouse2());

		if(SUCCEEDED(hr)) {
			ExtractHardwareProperties();
			return;
		}
		olog << e_error << "Cannot set DirectInput data-format.\n" << e_info;
	}
	else {
		iDIDevice = 0;
		olog << e_error << "Cannot create DirectInput device.\n" << e_info;
	}
	win_utils::SafeRelease(iDIDevice);
}


//====================================================================
void  SystemMouse8::ReleaseDIDevice()
{
	if(iDIDevice) {
		iDIDevice->Unacquire();
	}
	win_utils::SafeRelease(iDIDevice);
}


//====================================================================
void  SystemMouse8::ActivateDIDevice(bool bActivate)
{
	bDIDeviceActive = false;
	SetupMouseMessageDispatch(false);

	if(iDIDevice  &&  bDIDeviceConnectedToWindow) {
		if(bActivate) {
			bDIDeviceActive = true;
			iDIDevice->Acquire();
			SetupMouseMessageDispatch(true);
		}
		else {
			iDIDevice->Unacquire();
		}
	}
}


//====================================================================
void  SystemMouse8::ResetMouseMessageDispatch()
{
	if(! pAppWindow) return;

	pAppWindow->SetMouseMessageHook(0);
	pAppWindow->EnableMessageDispatchMouseMove(true);
	pAppWindow->EnableMessageDispatchMouseWheel(true);
	pAppWindow->EnableMessageDispatchMouseButtons(true);
}


//====================================================================
void  SystemMouse8::SetupMouseMessageDispatch(bool bUsingDI)
{
	if(! pAppWindow) return;

	pAppWindow->SetMouseMessageHook(this);
	pAppWindow->EnableMessageDispatchMouseMove(false);
	pAppWindow->EnableMessageDispatchMouseButtons(false);

	if(bUsingDI) {
		pAppWindow->EnableMessageDispatchMouseWheel(false);
	}
	else {
		pAppWindow->EnableMessageDispatchMouseWheel(true);
	}
}


//====================================================================
void  SystemMouse8::ExtractHardwareProperties()
{
	if(! iDIDevice) return;

	// wheel granularity
	DIPropDWord dipdwWheelGranularity(DIMOFS_Z, DIPH_BYOFFSET);
	if(SUCCEEDED(iDIDevice->GetProperty(DIPROP_GRANULARITY, dipdwWheelGranularity))) {
		DIWheelGranularity = ClampedLow<int>(dipdwWheelGranularity.dwData, 1);
	}
}

