

//--------------------------------------------------------------------
#include "private\VideoManager.h"
#include "private\vw_kernel.h"
#include "private\PluginTypes.h"
#include "private\PluginManager.h"
#include "BaseTools\Memory.h"
#include "GUI\GUI_Manager.h"


//--------------------------------------------------------------------
#define  MIN_VIDEO_WINDOW_WIDTH   160
#define  MIN_VIDEO_WINDOW_HEIGHT  160


//====================================================================
VideoManager::VideoManager()
{
	iReconfigCount = 0;
	reconfigCode = 0;

	bSystemCursorVisible = false;
	bWindowInitialized = false;
	bResourcesInitialized = false;
	iWindowCreateCount = 0;

	bResizeMoveSystemLoopActive = false;
	iResetCount = 0;

	pVideoRenderer = 0;
	pVideoSubsystem = 0;
	pVideoWindow = 0;

	bLastKnownGoodConfigValid = false;
}


//====================================================================
VideoManager::~VideoManager()
{
}


//====================================================================
void  VideoManager::SetReconfigParams(const std::string& vrPluginName,
							const VideoConfig& vc, uint _reconfigCode)
{
	reconfigCode = _reconfigCode;
	vrPluginNameReconfig = vrPluginName;
	videoConfigReconfig = vc;
}


//====================================================================
void  VideoManager::OnDisplaySettingsChanged()
{
	vw_kernel::Instance().ReportAppWindowGeometryChanged();
}


//====================================================================
void  VideoManager::OnVideoWindowActivated(bool bActive)
{
	// Since OS keeps internal show-cursor count, assume that system-cursor is hidden
	// while main-window is active, and make sure it is shown when window is deactivated
	ShowSystemCursor(! bActive);

	vw_kernel::Instance().ReportAppWindowActive(bActive);
}


//====================================================================
void  VideoManager::OnVideoWindowMoved()
{
	vw_kernel::Instance().ReportAppWindowGeometryChanged();
}


//====================================================================
void  VideoManager::OnVideoWindowResized(TopLevelWindow::eResizeAction action)
{
	switch(action) {
		case TopLevelWindow::ResizeAction_Minimize:
			// do nothing
			olog << "Video-window minimized" << std::endl;
			break;

		case TopLevelWindow::ResizeAction_Resize:
			// avoid dynamic resize, only restore should be handled here
			if(! bResizeMoveSystemLoopActive) {
				ReportWindowResize();
			}
			break;

		case TopLevelWindow::ResizeAction_Maximize:
			ReportWindowResize();
			break;
	}
}


//====================================================================
void  VideoManager::OnVideoWindowPaint()
{
	// ignore; async rendering only introduces unnecessary
	// complications for this type of application
}


//====================================================================
void  VideoManager::OnSystemMenuLoopActive(bool bLoopActive)
{
	vw_kernel::Instance().ReportAppWindowActive(! bLoopActive);
}


//====================================================================
void  VideoManager::OnSystemResizeMoveLoopActive(bool bLoopActive)
{
	vw_kernel::Instance().ReportAppWindowActive(! bLoopActive);
	bResizeMoveSystemLoopActive = bLoopActive;

	// cache current window-size
	if(bResizeMoveSystemLoopActive) {
		lastWindowWidth  = pVideoWindow->GetClientAreaWidth();
		lastWindowHeight = pVideoWindow->GetClientAreaHeight();
	}
	// change only if window is resized (window-move also causes this notification)
	else {
		if((pVideoWindow->GetClientAreaWidth() != lastWindowWidth)  ||
			(pVideoWindow->GetClientAreaHeight() != lastWindowHeight))
		{
			ReportWindowResize();
		}
	}
}


//====================================================================
void  VideoManager::InitializeWindow()
{
	// NOTES: if anything fails, ensure that video-renderer is released
	if(bWindowInitialized) return;

	olog << caption("video-window init");

	std::string  videoRendererPluginName;
	VideoConfig  videoConfig;

	// reconfig init
	if(iReconfigCount) {
		videoRendererPluginName = vrPluginNameReconfig;
		videoConfig = videoConfigReconfig;
	}
	// first-time init
	else {
		VideoConfig  refVideoConfig;

		refVideoConfig.resizable = true;
		refVideoConfig.displayModeConfig.width = 640;
		refVideoConfig.displayModeConfig.height = 480;
		refVideoConfig.displayModeConfig.bitsPerPixel = 32;
		refVideoConfig.displayModeConfig.refreshFrequency = 60;
		refVideoConfig.frameBufferConfig = FrameBufferConfig(false, true, 1, 24, 8, 0);

		// auto-construct closest safe
		Kernel::Instance().ConstructClosestSafeVideoConfig("", refVideoConfig, videoRendererPluginName, videoConfig);
	}

	// query video properties
	olog << "Querying application for video-configuration." << nl;
	vw_kernel::Instance().GetApplication().QueryVideoConfig(
						videoRendererPluginName, videoConfig, iReconfigCount, reconfigCode);
	iReconfigCount++;

	// do init
	while(true) {
		try {
			// dump requested configuration
			olog << "Requested configuration:\n";
			olog << tab << videoRendererPluginName << nl;
			videoConfig.Dump(olog, 1);

			// find plugins
			olog << "Seeking video-renderer plugin." << nl;
			VideoRendererPlugin* vrp = vw_kernel::Instance().GetPluginManager().FindVideoRendererPlugin(videoRendererPluginName);
			if(! vrp) {
				VW_THROW(KernelException,
						"Incorrectly filled video-renderer plugin-name.");
			}
			VideoSubsystemPlugin* vsp = vw_kernel::Instance().GetPluginManager().FindVideoSubsystemPlugin(
															vrp->videoSubsystemInterfaceName);

			// setup pointers
			pVideoRenderer = vrp->GetVideoRenderer();
			pVideoSubsystem = vsp->GetVideoSubsystem();

			// create video-subsystem-window (exception-throw allowed)
			olog << "Creating video-subsystem-window.\n";

			pVideoWindow = pVideoSubsystem->InitializeWindow(videoConfig, false);
			if(! pVideoWindow) {
				VW_THROW(KernelException,
						"Video-subsystem failed to initialize video-window.");
			}

			// store good-config
			videoConfig = pVideoSubsystem->GetVideoConfig();
			lastKnownGoodConfig = videoConfig;
			lastKnownGoodPluginName = videoRendererPluginName;
			bLastKnownGoodConfigValid = true;

			// creation succeeded
			break;
		}
		catch(VideoSubsystemRecoverableException& e) {

			olog << nl << e_error << caption("Exception caught");
			olog << e.MakeDescription() << nl << e_info;

			// try to recover if last-known-good-config is available
			if(bLastKnownGoodConfigValid) {
				olog << "Restoring last-known-good-configuration.\n\n";

				videoConfig = lastKnownGoodConfig;
				videoRendererPluginName = lastKnownGoodPluginName;
				bLastKnownGoodConfigValid = false;
			}
			else {
				olog << "Last-known-good-configuration not available. Rethrowing.\n\n";
				throw;
			}
		}
	}

	// window-ready
	bWindowInitialized = true;

	// set default window properties
	MainWindowProperties  mwp;
	mwp.text = vw_kernel::Instance().GetApplicationWindowText();

	MainWindowGeometry  mwg;
	GetMainWindowGeometry(mwg);

	// store main-window handle
	vw_kernel::Instance().GetMessageHook().SetMainWindowHandle(pVideoWindow->GetHWND());
	olog << "MainWindowHandle: 0x" << pVideoWindow->GetHWND() << nl;

	// query application to update window properties
	olog << "Post-create main-window (CreateCount=" << iWindowCreateCount << ")" << nl;

	vw_kernel::Instance().GetApplication().PostCreateMainWindow(mwp, mwg,
							videoRendererPluginName, videoConfig, iWindowCreateCount);
	iWindowCreateCount++;

	// update window properties
	Kernel::Instance().SetMainWindowProperties(mwp);
	SetMainWindowGeometry(mwg);

	// make window visible
	if(mwg.bMaximized) {
		pVideoWindow->ShowMaximized();
	}
	else {
		pVideoWindow->Show();
	}

	// flush initial reset caused by multiple window-resizes
	if(NeedsReset()) {
		olog << "Flushing initial video-subsystem reset.\n";
		ResetVideoSubsystem();
	}

	// window is stable, activate handlers
	olog << "Activating window-geometry handlers.\n";

	pVideoWindow->SetObserver(this);
	pVideoWindow->ActivateResizeHandler();
	pVideoWindow->ActivateMoveHandler();

	// dump realized configuration
	olog << "Realized configuration:\n";
	pVideoSubsystem->GetVideoConfig().Dump(olog, 1);

	// connect video-subsystem to video-renderer
	olog << "Connecting video-subsystem to video-renderer." << nl;
	vw_kernel::Instance().VideoRenderer_ConnectVideoSubsystem(*pVideoRenderer, *pVideoSubsystem);

	// connect video-renderer to gui-manager
	olog << "Connecting video-renderer to gui-manager." << nl;
	vw_kernel::Instance().GUIManager_ConnectVideoRenderer(*pVideoRenderer);

	// connect video-renderer to application
	olog << "Connecting video-renderer to application." << nl;
	vw_kernel::Instance().GetApplication().ConnectVideoRenderer(*pVideoRenderer);

	// handle video-renderer-resources
	InitializeResources(false); //exception-throw-allowed

	// init video-states & send synthetic window-resize event
	ReportInitVideoRendererStates(false);
	ReportWindowResize();

	olog << std::endl;
}


//====================================================================
void  VideoManager::CleanupWindow()
{
	if(! bWindowInitialized) return;

	olog << caption("video-window cleanup");

	// handle video-renderer-resources
	CleanupResources(false);

	// window-not-ready
	bWindowInitialized = false;

	// disconnect video-renderer from application
	olog << "Disconnecting video-renderer from application." << nl;
	vw_kernel::Instance().GetApplication().DisconnectVideoRenderer();

	// disconnect video-renderer from gui-manager
	olog << "Disconnecting video-renderer from gui-manager." << nl;
	vw_kernel::Instance().GUIManager_DisconnectVideoRenderer();

	// disconnect video-subsystem from video-renderer
	olog << "Disconnecting video-subsystem from video-renderer." << nl;
	vw_kernel::Instance().VideoRenderer_DisconnectVideoSubsystem(*pVideoRenderer);

	// notify application about window destroy
	MainWindowGeometry  mwg;
	GetMainWindowGeometry(mwg);

	olog << "Pre-destroy main-window." << nl;
	vw_kernel::Instance().GetApplication().PreDestroyMainWindow(mwg,
			pVideoRenderer->GetPluginInfo().GetPluginName(), pVideoSubsystem->GetVideoConfig());

	// deactivate handlers
	olog << "Deactivating window-geometry handlers.\n";

	pVideoWindow->SetObserver(0);

	// enable debug-logging
	KernelExecutionParams  kep;

	MessageHookImp& mhi = vw_kernel::Instance().GetMessageHook();
	if(kep.bLogWinMsgOnMainWindowCleanup) mhi.Enable(true);

	// release video-subsystem-window
	olog << "Releasing video-subsystem-window." << nl;
	pVideoSubsystem->CleanupWindow();

	// disable debug-logging & invalidate main-window hook
	if(kep.bLogWinMsgOnMainWindowCleanup) mhi.Enable(false);
	mhi.SetMainWindowHandle(0);

	// invalidate pointers
	pVideoWindow = 0;
	pVideoRenderer = 0;
	pVideoSubsystem = 0;

	olog << std::endl;
}


//====================================================================
void  VideoManager::InitializeResources(bool bPostReset)
{
	if(! bWindowInitialized) return;
	if(bResourcesInitialized) return;

	// signal components that video-resources should be created
	olog << "Creating video-resources (PostResetFlag = " << bPostReset << ")\n";

	olog << "Creating application managed video-resources" << nl;
	vw_kernel::Instance().GetApplication().CreateVideoResources(bPostReset); //exception-throw-allowed

	olog << "Creating gui-manager managed video-resources" << nl;
	vw_kernel::Instance().GUIManager_CreateVideoResources(bPostReset); //exception-throw-allowed

	olog << "Creating video-renderer managed video-resources" << nl;
	vw_kernel::Instance().VideoRenderer_CreateVideoResources(*pVideoRenderer, bPostReset); //exception-throw-allowed

	bResourcesInitialized = true;
	olog << std::flush;
}


//====================================================================
void  VideoManager::CleanupResources(bool bPreReset)
{
	if(! bWindowInitialized) return;
	if(! bResourcesInitialized) return;

	// signal components that video-resources should be released
	olog << "Releasing video-resources (PreResetFlag = " << bPreReset << ")\n";

	olog << "Releasing application managed video-resources" << nl;
	vw_kernel::Instance().GetApplication().ReleaseVideoResources(bPreReset);

	olog << "Releasing gui-manager managed video-resources" << nl;
	vw_kernel::Instance().GUIManager_ReleaseVideoResources(bPreReset);

	olog << "Releasing video-renderer managed video-resources" << nl;
	vw_kernel::Instance().VideoRenderer_ReleaseVideoResources(*pVideoRenderer, bPreReset);

	bResourcesInitialized = false;
	olog << std::flush;
}


//====================================================================
void  VideoManager::ErrorCheck()
{
	pVideoSubsystem->ErrorCheck();
}


//====================================================================
bool  VideoManager::IsReadyToRender()
{
	return
	(pVideoSubsystem->GetOperationalState() == VideoSubsystem::OperationalState_Ready);
}


//====================================================================
void  VideoManager::Render(const KernelExecutionParams& kep)
{
	if(IsReadyToRender()) {
		VW_PROFILE_BLOCK("FrameRenderVideo");
		vw_kernel::Instance().GetApplication().FrameRenderVideo();

		// auto-draw-gui
		if(kep.bAutoDrawGUI)
		{
			VW_PROFILE_BLOCK("AutoDrawGUI");
			gui::Manager::Instance().Draw();
		}

		// auto-end gui-renderer
		if(kep.bAutoEndGUIRenderer)
		{
			VW_PROFILE_BLOCK("AutoEndGUIRenderer");
			gui::Renderer& gr = pVideoRenderer->GetGUIRenderer();
			if(gr.IsInsideBeginEnd()) gr.End();
		}

		// auto-present-back-buffer
		if(kep.bAutoPresentBackBuffer)
		{
			VW_PROFILE_BLOCK("AutoPresentBackBuffer");
			pVideoSubsystem->PresentBackBuffer();
		}
	}
}


//====================================================================
void  VideoManager::Fix()
{
	if(IsReadyToRender()) return;

	// prerequisite for successful reset
	CleanupResources(true);

	// try reset
	if(ResetVideoSubsystem()) {

		// success, restore video-resources
		InitializeResources(true); //exception-throw-allowed

		// restore video-states & send synthetic window-resize event
		ReportInitVideoRendererStates(true);
		ReportWindowResize();
	}
}


//====================================================================
void  VideoManager::SetMainWindowGeometry(const MainWindowGeometry& mwgRef)
{
	if(! pVideoWindow  ||  pVideoWindow->IsFullscreen()) return;
	MainWindowGeometry mwg = mwgRef;

	// setup absolute minimum
	ClampLow<uint>(mwg.minSizeX, MIN_VIDEO_WINDOW_WIDTH);
	ClampLow<uint>(mwg.minSizeY, MIN_VIDEO_WINDOW_HEIGHT);
	pVideoWindow->SetMinSize(mwg.minSizeX, mwg.minSizeY);

	// update position
	if(mwg.bMaximized) {
		pVideoWindow->SetMaximizedPosition(0, 0);
	}
	else {
		pVideoWindow->SetPosition(mwg.positionX, mwg.positionY);
	}
}


//====================================================================
void  VideoManager::GetMainWindowGeometry(MainWindowGeometry& mwg)
{
	if(pVideoWindow) {
		mwg.bMaximized = (! pVideoWindow->IsFullscreen()  &&  pVideoWindow->IsMaximized());
		mwg.positionX = mwg.bMaximized ? 0 : pVideoWindow->GetPositionX();
		mwg.positionY = mwg.bMaximized ? 0 : pVideoWindow->GetPositionY();
		pVideoWindow->GetMinSize(mwg.minSizeX, mwg.minSizeY);
	}
	else {
		mwg.bMaximized = false;
		mwg.positionX = 0;
		mwg.positionY = 0;
		mwg.minSizeX = MIN_VIDEO_WINDOW_WIDTH;
		mwg.minSizeY = MIN_VIDEO_WINDOW_HEIGHT;
	}
}


//====================================================================
bool  VideoManager::NeedsReset()
{
	return
	(pVideoSubsystem->GetOperationalState() == VideoSubsystem::OperationalState_WaitingReset);
}


//====================================================================
bool  VideoManager::ResetVideoSubsystem()
{
	if(! NeedsReset()) {
		VW_THROW(KernelException, "ResetVideoSubsystem(): expected OperationalState_WaitingReset.");
	}

	// log reset start
	if(iResetCount == 0) {
		olog << "Reseting video-subsystem." << std::endl;
	}
	iResetCount ++;

	// try reset
	try {
		pVideoSubsystem->Reset();
	}
	catch(Exception& e) {
		vw_kernel::Instance().ReportException(e);

		// cannot recover
		olog << "Video reset failed. Reset count: " << iResetCount << std::endl;
		iResetCount = 0;

		VW_THROW(KernelException, "Cannot reset video-subsystem.");
	}

	// check state
	if(IsReadyToRender()) {

		// reset successful
		olog << "Video reset successful. Reset count: ";
		olog << iResetCount << std::endl;

		iResetCount = 0;
		return true;
	}
	return false;
}


//====================================================================
void  VideoManager::ReportInitVideoRendererStates(bool bPostReset)
{
	if(IsReadyToRender()) {
		olog << "Init video-renderer states." << std::endl;
		vw_kernel::Instance().GetApplication().InitVideoRendererStates(bPostReset);
	}
}


//====================================================================
void  VideoManager::ReportWindowResize()
{
	if(IsReadyToRender()) {
		uint width  = pVideoWindow->GetClientAreaWidth();
		uint height = pVideoWindow->GetClientAreaHeight();

		olog << "Video-window resize: " << width << " x " << height << std::endl;
		vw_kernel::Instance().VideoRenderer_OnWindowResize(*pVideoRenderer, width, height);
		vw_kernel::Instance().GUIManager_OnWindowResize(width, height);
		vw_kernel::Instance().GetApplication().OnWindowResize(width, height);
	}
	vw_kernel::Instance().ReportAppWindowGeometryChanged();
}


//====================================================================
void  VideoManager::ShowSystemCursor(bool bShow)
{
	if(bSystemCursorVisible != bShow) {
		bSystemCursorVisible = bShow;

		::ShowCursor(bShow ? TRUE : FALSE);
	}
}

