

//--------------------------------------------------------------------
#include "private\vw_kernel.h"
#include "private\LogWindowThread.h"
#include "private\VideoManager.h"
#include "private\HIDManager.h"
#include "Framework\LibWrapperDXErr8.h"
#include <algorithm>


//====================================================================
const CLAStrings  claStrings;


//====================================================================
CLAStrings::CLAStrings()
{
	for(uint i = 0; i < CLA_Count; i++) {
		szValues[i] = "";
	}

	szValues[CLA_Help]                    = "vwk-help";
	szValues[CLA_ShowLog]                 = "vwk-showlog";
	szValues[CLA_LogTopLeft]              = "vwk-logtl";
	szValues[CLA_LogTopRight]             = "vwk-logtr";
	szValues[CLA_NoHIDDPlugins]           = "vwk-nohidd";
	szValues[CLA_DisablePlugin]           = "vwk-dplugin";
	szValues[CLA_Application]             = "vwk-app";
	szValues[CLA_NoDirectInput]           = "vwk-nodinput";
	szValues[CLA_DeveloperMode]           = "vwk-devmode";
	szValues[CLA_WaitClose]               = "vwk-waitclose";
	szValues[CLA_ReportDLLLoadErrors]     = "vwk-dllerror";
	szValues[CLA_NoHardCleanupConfigLoop] = "vwk-nohccl";
	szValues[CLA_HardRethrowExit]         = "vwk-hre";
	szValues[CLA_NoHardRethrowExit]       = "vwk-nohre";
	szValues[CLA_GMTolerateMissingDevice] = "vwk-gm-tolerate-missing-device";
}


//====================================================================
void  MessageHookImp::Enable(bool bEnable)
{
	TopLevelWindow::SetMessageHook(bEnable ? this : 0);
	if(os) {
		*os << "Windows message logging enabled: " << bEnable << std::endl;
	}
}


//====================================================================
void  MessageHookImp::Process(eDispatchPoint dp, const WinMsg& wm,
							LRESULT& result, bool& bProcessed)
{
	if(wm.hwnd == hwndLogWindow) {
		LogMsg("LogWindow", dp, wm, result, bProcessed);
		return;
	}
	if(wm.hwnd == hwndMainWindow) {
		LogMsg("MainWindow", dp, wm, result, bProcessed);
		return;
	}
}


//====================================================================
void  MessageHookImp::LogMsg(const std::string& sWindowName, eDispatchPoint dp,
							const WinMsg& wm, LRESULT& result, bool& bProcessed)
{
	if(! os) return;
	std::ostream& out = *os;

	out << "WinMsg: " << sWindowName << ", ";

	switch(dp) {
		case DispatchPoint_PreMessageHandler:
			out << "PreMessageHandler";
			break;
		case DispatchPoint_PostMessageHandler:
			out << "PostMessageHandler";
			break;
		case DispatchPoint_PreDefWindowProc:
			out << "PreDefWindowProc";
			break;
		case DispatchPoint_PostDefWindowProc:
			out << "PostDefWindowProc";
			break;
	}
	out << ", Processed=" << bProcessed;
	out << ", Result=" << result;
	out << ", Hwnd=0x" << wm.hwnd;
	out << ", Msg=0x" << ToStringHex(wm.msg);
	out << ", Wparam=0x" << wm.wparam;
	out << ", Lparam=0x" << wm.lparam;
	out << std::endl;
}


//====================================================================
VW_IMPLEMENT_SINGLETON( vw_kernel );


//====================================================================
vw_kernel::vw_kernel()
	:
	//Important: forces standard init (kernel is not constructed yet)
	executionParams(false) 
{
	VW_IMPLEMENT_SINGLETON_CTOR( vw_kernel );

	pFrameObservers = 0;
	pFrameObserversTemp = 0;
	pReconfigObservers = 0;

	pHighResTimer = 0;
	pFrameworkProfiler = 0;
	pStringStorage = 0;
	pVariableRegistry = 0;
	pCmdLineParser = 0;
	pConsoleBuffers = 0;
	pConOut = 0;
	pLogFile = 0;
	pLogOut = 0;
	pLogWindowThread = 0;
	pPluginManager = 0;
	pApplication = 0;
	pHIDManager = 0;
	pVideoManager = 0;
	pAppSubsysManager = 0;

	bAllowChangeFrameObservers = true;
	bAllowChangeReconfigObservers = true;

	bUseProfileDir = false;
	bAutoPresentBackBuffer = true;

	secsLastFrameStart = 0;
	secsFrameDelta = 0;
	activeKernelLevel = KernelLevel_KernelSetup;

	iGenericReconfigCount = 0;
	iGenericReconfigCode = 0;
}


//====================================================================
vw_kernel::~vw_kernel()
{
	VW_IMPLEMENT_SINGLETON_DTOR( vw_kernel );
}


//====================================================================
int  vw_kernel::Run()
{
	try {
		//---------------------------------------------
		// Memory-allocation control.
		// This should be the first constructed object,
		// before any dynamic allocation occurs.
		//---------------------------------------------
		NewThrowsExceptionSentry  ntes;

		//---------------
		// Enter Level-0.
		//---------------
		Level0();
	}
	catch(Exception& e) {
		ShowException(e);
	}
	return 0;
}


//====================================================================
void  vw_kernel::SetLevel(eKernelLevel _kl)
{
	if(activeKernelLevel != _kl) {
		activeKernelLevel = _kl;

		// avoid reporting low kernel levels to application
		if(activeKernelLevel < KernelLevel_ConfigLoop) return;

		if(pApplication) {
			pApplication->OnKernelLevelChanged(activeKernelLevel);
		}
	}
}


//====================================================================
void  vw_kernel::ReportException(Exception& e, bool bStoreFailureDesc)
{
	bCleanExit = false;

	if(bStoreFailureDesc) {
		failureDesc = FailureDesc(activeKernelLevel, &e, 0);
	}

	olog << e_error << nl << caption("Exception caught");
	olog << e.MakeDescription() << endpara << e_info;
}


//====================================================================
void  vw_kernel::ReportException(std::exception& e, bool bStoreFailureDesc)
{
	bCleanExit = false;

	if(bStoreFailureDesc) {
		failureDesc = FailureDesc(activeKernelLevel, 0, &e);
	}

	olog << e_error << nl << caption("Exception caught");
	olog << VW_GET_CLASS_NAME(e) << "\n! " << e.what() << endpara << e_info;
}


//====================================================================
void  vw_kernel::ReportExceptionUnknown(const std::string& msg)
{
	bCleanExit = false;

	olog << e_error << nl;
	olog << caption("Hard exception caught at kernel-level " + ToString(activeKernelLevel));
	olog << msg << endpara << e_info;
}


//====================================================================
void  vw_kernel::ReportHardRethrow()
{
	olog << e_error << "Rethrowing hard-exception from kernel-level: ";
	olog << activeKernelLevel << endpara << e_info;
}


//====================================================================
void  vw_kernel::NudgeFrameLoop()
{
	VideoSubsystemWindow* pVideoWindow = pVideoManager->GetVideoWindow();
	if(pVideoWindow) {
		pVideoWindow->PostNullMessage();
	}
}


//====================================================================
void  vw_kernel::ReportAppWindowGeometryChanged()
{
	pHIDManager->OnAppWindowGeometryChanged();
}


//====================================================================
void  vw_kernel::ReportAppWindowActive(bool bActive)
{
	if(bAppWindowActive != bActive) {
		bAppWindowActive = bActive;

		olog << "AppWindowActive: " << ToString(bAppWindowActive) << std::endl;

		// dispatch notifications
		GUIManager_NotifyApplicationWindowActive(bAppWindowActive);
		pHIDManager->OnAppWindowActive(bAppWindowActive);
		pApplication->OnWindowActive(bAppWindowActive);
	}
	if(bActive) {
		ResetFrameTimer();
	}
}


//====================================================================
bool  vw_kernel::ConfirmGenericExtension(const std::string& pluginName,
						const std::string& genericExtensionInterfaceName)
{
	if(genericExtensionInterfaceName == VW_INTERFACE_NAME_LibWrapperDXErr8) return true;
	return false;
}


//====================================================================
std::string  vw_kernel::GetLogFilenameExtension()
{
	return DEFAULT_LOG_FILE_EXTENSION;
}


//====================================================================
void  vw_kernel::CloseLogFile()
{
	StringStorage& ss = *pStringStorage;

	// flush buffers
	pLogFile->Close();

	// move temp-log-file to log-directory
	if(bUseProfileDir) {
		if(file_system::CopyFile(ss.tempLogPath, ss.appLogPath, true)) {
			file_system::DeleteFile(ss.tempLogPath);
		}
	}
}


//====================================================================
void  vw_kernel::PrepareLogWindowForCleanup()
{
	pLogWindowThread->EnableCloseGadget(false, false);
}


//====================================================================
void  vw_kernel::TerminateLogWindowThread()
{
	pLogWindowThread->EnableCloseGadget(false, false);
	pLogWindowThread->WaitMessageQueueReady();
	pLogWindowThread->PostStandardQuitMessage();
	pLogWindowThread->WaitExit();
}


//====================================================================
void  vw_kernel::ExtractThreadHandle()
{
	::DuplicateHandle(GetCurrentProcess(), GetCurrentThread(), GetCurrentProcess(),
					&kernelThreadInfo.threadHandle, 0, FALSE, DUPLICATE_SAME_ACCESS);

	kernelThreadInfo.threadID = ::GetCurrentThreadId();
}


//====================================================================
void  vw_kernel::CloseThreadHandle()
{
	::CloseHandle(kernelThreadInfo.threadHandle);
	kernelThreadInfo.threadHandle = 0;
	kernelThreadInfo.threadID = 0;
}


//====================================================================
void  vw_kernel::RequestFrameLoopExit()
{
	bSpinFrameLoop = false;
	NudgeFrameLoop();
}


//====================================================================
bool  vw_kernel::CheckCommandInsideFrameLoop()
{
	if(bSpinFrameLoop) return true;
	*pLogOut << e_error << "Command not allowed outside of frame-loop. Ignoring.\n" << e_info;
	return false;
}


//====================================================================
void  vw_kernel::DumpHelp()
{
	const char* valueChoiceOpen  = "[";
	const char* valueChoiceSep   = " | ";
	const char* valueChoiceClose = "]";

	std::string value;

	// print usage
	olog << e_attention;
	olog << caption(std::string(KERNEL_NAME) + " command-line arguments");

	// kernel help
	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_Help], "Print kernel-help only.") << nl;

	// log
	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_ShowLog],
								"Auto-expand log-window at startup.") << nl;

	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_LogTopLeft],
								"Position log-window at top-left corner of the screen.") << nl;

	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_LogTopRight],
								"Position log-window at top-right corner of the screen.") << nl;

	// plugins
	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_NoHIDDPlugins],
								"Disable all HIDDriver plugins.") << nl;

	olog << pCmdLineParser->FormatAssignmentHelp(claStrings[CLA_DisablePlugin], "plugin-name",
								"Disable specified plugin (case-insensitive).") << nl;

	olog << pCmdLineParser->FormatAssignmentHelp(claStrings[CLA_Application], "application-plugin-name",
								"Select among available applications (case-insensitive).") << nl;

	// direct-input
	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_NoDirectInput],
								"Disable DirectInput by default.") << nl;

	// execution control
	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_DeveloperMode],
								"Execute in developer-mode.") << nl;

	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_WaitClose],
								"Wait close confirmation.") << nl;

	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_ReportDLLLoadErrors],
								"Enable OS critical-error message-box when loading plugins.") << nl;

	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_NoHardCleanupConfigLoop],
								"Disable forced cleanup by kernel-config-loop after hard-exception.") << nl;

	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_HardRethrowExit],
								"Enable hard-exception rethrow at kernel-level-0.") << nl;

	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_NoHardRethrowExit],
								"Disable hard-exception rethrow at kernel-level-0.") << nl;

	// gui-manager control
	olog << pCmdLineParser->FormatOptionHelp(claStrings[CLA_GMTolerateMissingDevice],
								"Force gui-manager to tolerate missing input-devices.") << nl;

	olog << e_info << std::endl;
}


//====================================================================
void  vw_kernel::FillKernelCLArgumentNames(std::vector<std::string>& kernelArgNames)
{
	kernelArgNames.clear();
	kernelArgNames.reserve(claStrings.GetSize());

	for(uint i = 0; i < claStrings.GetSize(); i++) {
		kernelArgNames.push_back(claStrings[eCLA(i)]);
	}
}


//====================================================================
void  vw_kernel::RemoveKernelCLArguments(const CommandLineParser& clp, std::string& out)
{
	out.erase();

	std::vector<std::string>  kernelArgNames;
	FillKernelCLArgumentNames(kernelArgNames);

	for(uint i = 0; i < clp.GetNumArguments(); i++) {
		const CommandLineParser::Argument& arg = clp.GetArgument(i);
		if(std::find(kernelArgNames.begin(), kernelArgNames.end(), arg.String()) == kernelArgNames.end()) {
			// not found, add to output
			out.append(arg.MakeFullString());
			out.append(" ");
		}
	}
}


