

//--------------------------------------------------------------------
#pragma once


//--------------------------------------------------------------------
#include "BaseTools\Time.h"
#include "BaseTools\Stream.h"
#include "BaseTools\CommandLineParser.h"
#include "BaseTools\ConsoleBuffers.h"
#include "BaseTools\CommandLineParser.h"
#include "BaseTools\VariableProxy.h"
#include "Framework\Application.h"
#include "Framework\Kernel.h"
#include "Framework\KernelMacros.h"
#include "Framework\DirectInputProxy.h"
#include "Framework\FrameworkProfiler.h"
#include <string>
#include <vector>

using namespace vertexware;


//====================================================================
#define    KERNEL_NAME                "vw_kernel"
#define    KERNEL_VERSION_MAJOR       1
#define    KERNEL_VERSION_MINOR       0
#define    KERNEL_VERSION_REVISION    0


//====================================================================
#define    DEFAULT_LOG_NAME                KERNEL_NAME
#define    DEFAULT_LOG_FILE_NAME           "vw_kernel.log"
#define    DEFAULT_LOG_FILE_EXTENSION      ".log"
#define    TEMP_LOG_FILE_PREFIX            "vwk"
#define    CONBUF_MAX_OUTPUT_LINES         1000
#define    CONBUF_MAX_COMMAND_HISTORY      50
#define    GLOBAL_VARIABLE_REGISTRY_NAME   "Global"


//====================================================================
class LogWindowThread;
class PluginManager;
class HIDManager;
class VideoManager;
class AppSubsysManager;


//====================================================================
// Kernel command-line arguments.
//====================================================================
enum eCLA
{
	CLA_Help = 0,
	CLA_ShowLog,
	CLA_LogTopLeft,
	CLA_LogTopRight,
	CLA_NoHIDDPlugins,
	CLA_DisablePlugin,
	CLA_Application,
	CLA_NoDirectInput,
	CLA_DeveloperMode,
	CLA_WaitClose,
	CLA_ReportDLLLoadErrors,
	CLA_NoHardCleanupConfigLoop,
	CLA_HardRethrowExit,
	CLA_NoHardRethrowExit,
	CLA_GMTolerateMissingDevice,

	CLA_Count,
	CLA_Force32Bits = 0x7FFFFFFF
};


//====================================================================
class CLAStrings
{
private:
	const char*  szValues[CLA_Count];

public:
	CLAStrings();

	uint  GetSize() const { return CLA_Count; }
	const char*  operator[](eCLA cla) const { return szValues[cla]; }
};

extern const CLAStrings  claStrings;


//====================================================================
struct StringStorage
{
	std::string  exeDir;
	std::string  pluginDir;
	std::string  userHomeDir;
	std::string  userProfileDir;
	std::string  tempDir;
	std::string  tempLogPath;

	std::string  commandLine;
	std::string  filteredCommandLine;

	std::string  appDLLBasename;
	std::string  appName;
	std::string  appFlagname;

	std::string  appProcessDir;
	std::string  appProfileDir;
	std::string  appLogFilename;
	std::string  appLogPath;

	std::string  appWindowText;
};


//====================================================================
typedef  std::vector<FrameObserver*>     VEC_FRAME_OBSERVER;
typedef  std::vector<ReconfigObserver*>  VEC_RECONFIG_OBSERVER;


//====================================================================
class MessageHookImp : public MessageHook
{
public:
	MessageHookImp()
	{
		os = 0;
		hwndLogWindow = 0;
		hwndMainWindow = 0;
	}

	void  SetLog(std::ostream* _os = 0) { os = _os; }

	void  SetLogWindowHandle(HWND hwnd = 0) { hwndLogWindow = hwnd; }
	void  SetMainWindowHandle(HWND hwnd = 0) { hwndMainWindow = hwnd; }

	void  Enable(bool bEnable = false);

	virtual  void  Process(eDispatchPoint dp, const WinMsg& wm,
							LRESULT& result, bool& bProcessed);


private:
	void  LogMsg(const std::string& sWindowName, eDispatchPoint dp,
					const WinMsg& wm, LRESULT& result, bool& bProcessed);


private:
	std::ostream*  os;
	HWND           hwndLogWindow;
	HWND           hwndMainWindow;
};


//====================================================================
// Kernel implementation
//====================================================================
class vw_kernel : public Kernel
{
VW_DECLARE_SINGLETON( vw_kernel );

public:
	//---------------------------------
	// Kernel-interface implementation.
	//---------------------------------
	virtual  void  SetExecutionParams(const KernelExecutionParams& kep);
	virtual  void  GetExecutionParams(KernelExecutionParams& kep);

	virtual  void  SetKernelThreadInfo(const KernelThreadInfo& kti);
	virtual  void  GetKernelThreadInfo(KernelThreadInfo& kti);

	virtual  eKernelLevel  GetActiveKernelLevel() { return activeKernelLevel; }
	virtual  bool          IsAppWindowActive() { return bAppWindowActive; }

	virtual  void  SetMainWindowProperties(const MainWindowProperties& mwp);
	virtual  void  GetMainWindowProperties(MainWindowProperties& mwp);

	virtual  void  RequestShutdown();
	virtual  void  RequestHIDReconfig(const DirectInputConfig& diConfig, uint reconfigCode);
	virtual  void  RequestVideoReconfig(const std::string& vrPluginName, const VideoConfig& vc, uint reconfigCode);
	virtual  void  RequestSubsystemReconfig(uint reconfigCode);
	virtual  void  RequestGenericReconfig(uint reconfigCode);

	virtual  bool  SetPluginParameter(const std::string& pluginName, const NamedParameter& np);
	virtual  bool  GetPluginParameter(const std::string& pluginName, const std::string& parameterName, NamedParameter& out);

	virtual  uint               GetNumGenericExtensions();
	virtual  GenericExtension&  GetGenericExtension(uint i);

	virtual  GenericExtension*  FindGenericExtension(const std::string* pPluginName,
										const std::string* pGenericExtensionInterfaceName);

	virtual  Application*  GetApplicationPtr() { return pApplication; }

	virtual  uint  GetNumVideoRenderers();
	virtual  void  GetVideoRendererInfo(uint i, VideoRendererInfo& vrInfo);
	virtual  void  GetVideoRendererUnfilteredProfile(uint i, VideoProfile& profile);

	virtual  VideoRenderer*  GetActiveVideoRendererPtr();
	virtual  bool            GetActiveVideoRendererInfo(VideoRendererInfo& vrInfo);
	virtual  bool            GetActiveVideoRendererUnfilteredProfile(VideoProfile& profile);

	virtual  uint  GetNumHIDDrivers();
	virtual  void  GetHIDDriverInfo(uint i, PluginInfo& pi);

	virtual  uint                   GetNumHumanInterfaceDevices();
	virtual  HumanInterfaceDevice&  GetHumanInterfaceDevice(uint i);

	virtual  void   GetActiveDirectInputConfig(DirectInputConfig& diConfig);

	virtual  void   ResetFrameNumber(uint frameNumber);
	virtual  uint   GetFrameNumber() { return iFrameNumber; }
	virtual  bool   IsFrameInProgress() { return bFrameInProgress; }
	virtual  double GetFrameTimeDeltaSecs() { return secsFrameDelta; }

	virtual  FrameworkProfiler&  GetFrameworkProfiler() { return *pFrameworkProfiler; }
	virtual  ConsoleBuffers&     GetConsoleBuffers() { return *pConsoleBuffers; }
	virtual  VariableRegistry&   GetVariableRegistry() { return *pVariableRegistry; }

	virtual  void  RegisterFrameObserver(FrameObserver& fo);
	virtual  void  UnregisterFrameObserver(FrameObserver& fo);
	virtual  void  UnregisterAllFrameObservers();

	virtual  void  RegisterReconfigObserver(ReconfigObserver& ro);
	virtual  void  UnregisterReconfigObserver(ReconfigObserver& ro);
	virtual  void  UnregisterAllReconfigObservers();

	virtual  void  NotifyVSyncChanged();


	//------------------------------
	// Internal vw_kernel interface.
	//------------------------------
	vw_kernel();
	virtual ~vw_kernel();

	int           Run();
	void          SetLevel(eKernelLevel kl);
	eKernelLevel  GetLevel() { return activeKernelLevel; }

	void  ReportException(Exception& e, bool bStoreFailureDesc = false);
	void  ReportException(std::exception& e, bool bStoreFailureDesc = false);
	void  ReportExceptionUnknown(const std::string& msg = "");
	void  ReportHardRethrow();

	void  NudgeFrameLoop();

	void  ReportAppWindowGeometryChanged();
	void  ReportAppWindowActive(bool bActive);

	bool  ConfirmGenericExtension(const std::string& pluginName,
					const std::string& genericExtensionInterfaceName);

	PluginManager&      GetPluginManager() { return *pPluginManager; }
	MessageHookImp&     GetMessageHook() { return *pMessageHook; }
	bool                GetReportDLLLoadErrors() { return bReportDLLLoadErrors; }
	bool                GetDefaultEnableDirectInput() { return bDefaultEnableDirectInput; }
	std::string         GetLogFilenameExtension();
	const std::string&  GetProfileDirectory() { return pStringStorage->appProfileDir; }
	bool                IsUsingProfileDirectory() { return bUseProfileDir; }
	const std::string&  GetApplicationName() { return pStringStorage->appName; }
	const std::string&  GetApplicationWindowText() { return pStringStorage->appWindowText; }

	// gui-manager forwarders
	void  GUIManager_ConnectApplication(bool bConnect) { Kernel::GUIManager_ConnectApplication(bConnect); }
	void  GUIManager_NotifyApplicationWindowActive(bool bActive) { Kernel::GUIManager_NotifyApplicationWindowActive(bActive); }
	void  GUIManager_NotifyApplicationWindowCreated() { Kernel::GUIManager_NotifyApplicationWindowCreated(); }

	void  GUIManager_BeginConnectHumanInterfaceDevices() { Kernel::GUIManager_BeginConnectHumanInterfaceDevices(); }
	void  GUIManager_ConnectHumanInterfaceDevice(HumanInterfaceDevice& hid) { Kernel::GUIManager_ConnectHumanInterfaceDevice(hid); }
	void  GUIManager_EndConnectHumanInterfaceDevices() { Kernel::GUIManager_EndConnectHumanInterfaceDevices(); }
	void  GUIManager_DisconnectHumanInterfaceDevices() { Kernel::GUIManager_DisconnectHumanInterfaceDevices(); }
	void  GUIManager_SetTolerateMissingInputDevice(bool bTolerate) { Kernel::GUIManager_SetTolerateMissingInputDevice(bTolerate); }

	void  GUIManager_ConnectVideoRenderer(VideoRenderer& vr) { Kernel::GUIManager_ConnectVideoRenderer(vr); }
	void  GUIManager_DisconnectVideoRenderer() { Kernel::GUIManager_DisconnectVideoRenderer(); }
	void  GUIManager_CreateVideoResources(bool bPostReset) { Kernel::GUIManager_CreateVideoResources(bPostReset); }
	void  GUIManager_ReleaseVideoResources(bool bPreReset) { Kernel::GUIManager_ReleaseVideoResources(bPreReset); }
	void  GUIManager_OnWindowResize(uint width, uint height) { Kernel::GUIManager_OnWindowResize(width, height); }

	// video-renderer forwarders
	void  VideoRenderer_ConnectVideoSubsystem(VideoRenderer& vr, VideoSubsystem& vs) { Kernel::VideoRenderer_ConnectVideoSubsystem(vr, vs); }
	void  VideoRenderer_DisconnectVideoSubsystem(VideoRenderer& vr) { Kernel::VideoRenderer_DisconnectVideoSubsystem(vr); }
	void  VideoRenderer_CreateVideoResources(VideoRenderer& vr, bool bPostReset) { Kernel::VideoRenderer_CreateVideoResources(vr, bPostReset); }
	void  VideoRenderer_ReleaseVideoResources(VideoRenderer& vr, bool bPreReset) { Kernel::VideoRenderer_ReleaseVideoResources(vr, bPreReset); }
	void  VideoRenderer_OnWindowResize(VideoRenderer& vr, uint width, uint height) { Kernel::VideoRenderer_OnWindowResize(vr, width, height); }


private:
	//-------------------
	// Base-object setup.
	//-------------------
	void  Level0();

	//--------------
	// Plugin setup.
	//--------------
	void  Level1();

	//--------------------
	// Configuration loop.
	//--------------------
	void  Level2();
	void  Level2Cleanup(bool bForced);
	void  NotifyVideoReconfigObservers();
	void  NotifyHIDReconfigObservers();
	void  NotifyVSyncReconfigObservers(bool bVSyncEnabled);

	//------------
	// Frame loop.
	//------------
	void  Level3();
	void  NotifyFrameObservers(double secsFrameTimeDelta);

	//------------------
	// Internal helpers.
	//------------------
	void  CloseLogFile();
	void  PrepareLogWindowForCleanup();
	void  TerminateLogWindowThread();
	void  ExtractThreadHandle();
	void  CloseThreadHandle();
	void  RequestFrameLoopExit();
	bool  CheckCommandInsideFrameLoop();
	void  DumpHelp();
	void  FillKernelCLArgumentNames(std::vector<std::string>& kernelArgNames);
	void  RemoveKernelCLArguments(const CommandLineParser& clp, std::string& out);

	__forceinline  void  ResetFrameTimer()
	{
		pHighResTimer->Start();
		secsLastFrameStart = pHighResTimer->GetElapsedSeconds();
		secsFrameDelta = 0;
	}


private:
	// base-objects
	VEC_FRAME_OBSERVER*     pFrameObservers;
	VEC_FRAME_OBSERVER*     pFrameObserversTemp;
	VEC_RECONFIG_OBSERVER*  pReconfigObservers;

	HighResolutionTimer*     pHighResTimer;
	FrameworkProfiler*       pFrameworkProfiler;
	StringStorage*           pStringStorage;
	VariableRegistry*        pVariableRegistry;
	CommandLineParser*       pCmdLineParser;
	ConsoleBuffers*          pConsoleBuffers;
	ConsoleOutputStream*     pConOut;
	LogFileStream*           pLogFile;
	ControlledOutputStream*  pLogOut;
	LogWindowThread*         pLogWindowThread;
	PluginManager*           pPluginManager;
	MessageHookImp*          pMessageHook;

	// plugin objects & managers
	Application*       pApplication;
	HIDManager*        pHIDManager;
	VideoManager*      pVideoManager;
	AppSubsysManager*  pAppSubsysManager;

	// execution states
	bool                    bReportDLLLoadErrors;
	bool                    bDefaultEnableDirectInput;
	bool                    bCleanExit;
	bool                    bUseProfileDir;
	KernelExecutionParams   executionParams;
	KernelThreadInfo        kernelThreadInfo;
	FailureDesc             failureDesc;

	eKernelLevel  activeKernelLevel;
	bool          bAppWindowActive;
	bool          bAutoPresentBackBuffer;
	bool          bShutdown;
	bool          bHIDReconfigPending;
	bool          bVideoReconfigPending;
	bool          bSubsysReconfigPending;
	bool          bGenericReconfigPending;
	bool          bAllowChangeFrameObservers;
	bool          bAllowChangeReconfigObservers;
	uint          iFrameLoopIteration;
	bool          bSpinFrameLoop;
	uint          iFrameNumber;
	bool          bFrameInProgress;
	double        secsLastFrameStart;
	double        secsFrameDelta;

	// reconfig-request contexts
	uint  iGenericReconfigCount;
	uint  iGenericReconfigCode;
};


//====================================================================
struct GUIManagerAppConnectSentry
{
	vw_kernel& k;

	GUIManagerAppConnectSentry(vw_kernel& _k) : k(_k)
	{
		k.GUIManager_ConnectApplication(true);
	}
	~GUIManagerAppConnectSentry()
	{
		k.GUIManager_ConnectApplication(false);
	}
};


//====================================================================
struct KernelLevelSentry
{
	vw_kernel&    k;
	eKernelLevel  klLast;

	KernelLevelSentry(vw_kernel& _k, eKernelLevel _kl) : k(_k)
	{
		klLast = k.GetLevel();
		k.SetLevel(_kl);
	}
	~KernelLevelSentry()
	{
		k.SetLevel(klLast);
	}
};


