

//--------------------------------------------------------------------
#include "Framework\FrameworkCommandsOpenGL.h"
#include "GLTools\Extensions.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace framework_commands_opengl {


//====================================================================
void __fastcall CreateAndRegisterAllCommands(
								std_command_interpreter::Interpreter& ci)
{
	ci.RegisterExternalCommand(new framework_commands_opengl::VROpenGLCmd);
}


//====================================================================
const char* VROpenGLCmd::s_szDefaultName                    = "vr-opengl";
const char* VROpenGLCmd::s_szOptionDumpDriverInfoAll        = "drv";
const char* VROpenGLCmd::s_szOptionDumpDriverInfoVendor     = "drvv";
const char* VROpenGLCmd::s_szOptionDumpDriverInfoExtensions = "drve";
const char* VROpenGLCmd::s_szOptionDumpDriverInfoAdditional = "drva";
const char* VROpenGLCmd::s_szOptionRefreshExtensionPointers = "r";
const char* VROpenGLCmd::s_szOptionEnableAllExtensions      = "ea";
const char* VROpenGLCmd::s_szOptionDisableAllExtensions     = "da";
const char* VROpenGLCmd::s_szAssignEnableExtensions         = "e";
const char* VROpenGLCmd::s_szAssignDisableExtensions        = "d";


//====================================================================
VROpenGLCmd::VROpenGLCmd()
{
	SetDefaultName(s_szDefaultName);
	SetShortDesc("Configure active OpenGL video-renderer");
}


//====================================================================
bool  VROpenGLCmd::ParseAndExecute(std::ostream& os, const CommandLineParser& clp,
									const std::string& rawCommandLine)
{
	VideoRendererOpenGL* pvr = InterfaceCast<VideoRendererOpenGL>(
									Kernel::Instance().GetActiveVideoRendererPtr());

	// must have active opengl renderer
	if(! pvr) {
		os << e_error << "Cannot find active OpenGL renderer.\n" << e_info;
		return false;
	}

	// driver info
	bool  bPrintVendor = clp.GetOption(s_szOptionDumpDriverInfoVendor);
	bool  bPrintExt = clp.GetOption(s_szOptionDumpDriverInfoExtensions);
	bool  bPrintAdd = clp.GetOption(s_szOptionDumpDriverInfoAdditional);

	if(clp.GetOption(s_szOptionDumpDriverInfoAll)) {
		bPrintVendor = bPrintExt = bPrintAdd = true;
	}
	if(bPrintVendor || bPrintExt || bPrintAdd) {
		os << "OpenGL driver info:\n";
		pvr->DumpDriverInfo(os, 1, bPrintVendor, bPrintExt, bPrintAdd);
		return true;
	}

	// enable/disable all extensions
	if(clp.GetOption(s_szOptionEnableAllExtensions)) {
		pvr->EnableAllExtensions(true);
	}
	else if(clp.GetOption(s_szOptionDisableAllExtensions)) {
		pvr->EnableAllExtensions(false);
	}

	// enable/disable extension list
	std::vector<std::string>  extNameList;

	if(clp.GetPackedAssignment(s_szAssignEnableExtensions, extNameList)) {
		EnableExtensions(*pvr, true, extNameList);
	}
	else if(clp.GetPackedAssignment(s_szAssignDisableExtensions, extNameList)) {
		EnableExtensions(*pvr, false, extNameList);
	}

	// print disabled extensions
	os << "Disabled OpenGL extensions:\n";
	pvr->DumpDisabledExtensions(os);

	// refresh pointers
	if(clp.GetOption(s_szOptionRefreshExtensionPointers)) {
		pvr->RefreshExtensionPointers();
	}
	return true;
}


//====================================================================
void  VROpenGLCmd::DumpHelp(std::ostream& os, const CommandLineParser& clp)
{
	os << "Options:\n";
	os << clp.FormatOptionHelp(s_szOptionDumpDriverInfoAll, "print driver info\n", 1);
	os << clp.FormatOptionHelp(s_szOptionDumpDriverInfoVendor, "print driver vendor info\n", 1);
	os << clp.FormatOptionHelp(s_szOptionDumpDriverInfoExtensions, "print driver extension info\n", 1);
	os << clp.FormatOptionHelp(s_szOptionDumpDriverInfoAdditional, "print driver additional info\n", 1);

	os << clp.FormatOptionHelp(s_szOptionRefreshExtensionPointers, "refresh extension-pointers\n", 1);
	os << clp.FormatOptionHelp(s_szOptionEnableAllExtensions, "enable all extensions\n", 1);
	os << clp.FormatOptionHelp(s_szOptionDisableAllExtensions, "disable all extensions\n", 1);

	os << "Assignments:\n";
	std::string sepBy = std::string("extension-names separated by {") +
						CommandLineParser::s_szDefaultPackedArgumentSeparators + "}";
	os << clp.FormatAssignmentHelp(s_szAssignEnableExtensions, sepBy, "enable specified extensions\n", 1);
	os << clp.FormatAssignmentHelp(s_szAssignDisableExtensions, sepBy, "disable specified extensions\n", 1);

	os << "Default behavior:\n\tprint disabled extensions\n";
}


//====================================================================
void  VROpenGLCmd::EnableExtensions(VideoRendererOpenGL& vr, bool bEnable,
									const std::vector<std::string>& extNameList)
{
	for(uint i = 0; i < extNameList.size(); i++) {

		gl_ext::eExtension e = gl_ext::GetExtensionFromName(extNameList[i]);
		if(e != gl_ext::Extension_Invalid) {

			vr.EnableExtension(e, bEnable);
		}
	}
}


//********************************************************************
}//namespace framework_commands_opengl


//********************************************************************
}//namespace vertexware


