

//--------------------------------------------------------------------
#include "Framework\GUIRendererOpenGL.h"
#include "BaseTools\Debug.h"
#include "GLTools\States.h"
#include "GLTools\Stacks.h"
#include "GLTools\Extensions.h"
#include "GLTools\ErrorHandling.h"
#include "MathTools\TransformTools.h"


//********************************************************************
namespace vertexware {


//====================================================================
#define  DEFAULT_PIXEL_OFFSET_X                     -0.125f
#define  DEFAULT_PIXEL_OFFSET_Y                     -0.125f
#define  APP_POOL_VERTEX_BUFFER_MIN_SIZE             100
#define  APP_POOL_VERTEX_BUFFER_MIN_STABLE_CYCLES    100


//====================================================================
#define  ASSERT_INSIDE_BEGIN_END \
	VW_ASSERT(bInsideBeginEnd, "Invalid call outside gui::Renderer::Begin/End.");


//====================================================================
#define  ASSERT_INSIDE_BEGIN_END_INTERLEAVED \
	VW_ASSERT(bInsideBeginEndInterleaved, "Invalid call outside gui::Renderer::BeginInterleaved/EndInterleaved.");


//====================================================================
#define  ASSERT_OUTSIDE_BEGIN_END \
	VW_ASSERT(! bInsideBeginEnd, "Invalid call inside gui::Renderer::Begin/End.");


//====================================================================
#define  ASSERT_OUTSIDE_BEGIN_END_INTERLEAVED \
	VW_ASSERT(! bInsideBeginEndInterleaved, "Invalid call inside gui::Renderer::BeginInterleaved/EndInterleaved.");


//====================================================================
GUIRendererOpenGL::GUIRendererOpenGL(VideoRendererOpenGL& vrParent)
	:
	gui::Renderer(vrParent),
	appPoolVertexBuffer(APP_POOL_VERTEX_BUFFER_MIN_SIZE, APP_POOL_VERTEX_BUFFER_MIN_STABLE_CYCLES)
{
	bInsideBeginEnd = false;
	bInsideBeginEndInterleaved = false;
	bTexturingEnabled = false;
	bAlphaTestEnabled = false;

	InvalidateLastTexParams();

	pixelParams.offsetX = DEFAULT_PIXEL_OFFSET_X;
	pixelParams.offsetY = DEFAULT_PIXEL_OFFSET_Y;

	depthParamsBase = DepthParams();
	depthParams = depthParamsBase;
	matTransform = mat4::IDENTITY;
	ConstructInternalTransform();
}


//====================================================================
GUIRendererOpenGL::~GUIRendererOpenGL()
{
}


//====================================================================
void  GUIRendererOpenGL::ConnectVideoSubsystem(VideoSubsystem& vs)
{
	if(! InterfaceCast<VideoSubsystemOpenGL>(&vs)) {
		VW_THROW(VideoRendererException,
				"Wrong video-subsystem connected to: " + VW_GET_CLASS_NAME(*this));
	}

	apDL_Begin.Assign(new gl_tools::DisplayList);
	apDL_End.Assign(new gl_tools::DisplayList);
	apDL_PrepareTextureParams.Assign(new gl_tools::DisplayList);

	Compile_Begin();
	Compile_End();
	Compile_PrepareTextureParams();
}


//====================================================================
void  GUIRendererOpenGL::DisconnectVideoSubsystem()
{
	apDL_Begin.SafeDelete();
	apDL_End.SafeDelete();
	apDL_PrepareTextureParams.SafeDelete();
}


//====================================================================
void  GUIRendererOpenGL::CreateVideoResources(bool bPostReset)
{
}


//====================================================================
void  GUIRendererOpenGL::ReleaseVideoResources(bool bPreReset)
{
}


//====================================================================
void  GUIRendererOpenGL::OnWindowResize(uint width, uint height)
{
	Renderer::OnWindowResize(width, height);

	ConstructInternalTransform();
	Compile_Begin();
}


//====================================================================
void  GUIRendererOpenGL::GetRendererParams(RendererParams& rp) const
{
	rp.bTextureRowOrderBottomToTop = true;
	rp.bTextureLayerOrderBackToFront = false;
}


//====================================================================
bool  GUIRendererOpenGL::SetViewport(const ViewportParams& vp)
{
	ASSERT_OUTSIDE_BEGIN_END;

	Stats().numSetViewport++;

	glViewport(vp.topLeft.x,
				GetWindowHeight() - (vp.topLeft.y + vp.size.y),
				vp.size.x, vp.size.y);

	glDepthRange(vp.renderDepthNear, vp.renderDepthFar);

	return true;
}


//====================================================================
void  GUIRendererOpenGL::ClearBuffers(const ClearParams& cp)
{
	ASSERT_OUTSIDE_BEGIN_END;

	Stats().numClearBuffers++;

	GLbitfield clearMask = 0;

	if(cp.bClearColor) {
		clearMask |= GL_COLOR_BUFFER_BIT;
		color4f c = cp.color;
		glClearColor(c.r, c.g, c.b, c.a);
	}
	if(cp.bClearDepth) {
		clearMask |= GL_DEPTH_BUFFER_BIT;
		float d = Clamped<float>(cp.depth, 0.0f, 1.0f);
		glClearDepth(d);
	}
	if(cp.bClearStencil) {
		clearMask |= GL_STENCIL_BUFFER_BIT;
		glClearStencil(cp.stencil);
	}
	if(! clearMask) return;

	if(cp.bUseRectangle) {
		// rectangle clear
		const gui::Rectangle& r = cp.rectangle;

		glScissor(r.left, GetWindowHeight() - (r.height + r.top), r.width, r.height);
		glEnable(GL_SCISSOR_TEST);
	}
	else {
		// full-window clear
		glViewport(0, 0, GetWindowWidth(), GetWindowHeight());
	}
	glClear(clearMask);

	if(cp.bUseRectangle) {
		glDisable(GL_SCISSOR_TEST);
	}
}


//====================================================================
void  GUIRendererOpenGL::SetPixelParams(const PixelParams& pp)
{
	ASSERT_OUTSIDE_BEGIN_END;

	pixelParams = pp;

	ConstructInternalTransform();
	Compile_Begin();
}


//====================================================================
void  GUIRendererOpenGL::GetPixelParams(PixelParams& pp) const
{
	pp = pixelParams;
}


//====================================================================
void  GUIRendererOpenGL::SetBaseDepthParams(const DepthParams& dp)
{
	ASSERT_OUTSIDE_BEGIN_END;

	depthParamsBase = dp;
	Compile_Begin();
}


//====================================================================
void  GUIRendererOpenGL::GetBaseDepthParams(DepthParams& dp) const
{
	dp = depthParamsBase;
}


//====================================================================
bool  GUIRendererOpenGL::QueryCreatesManagedResourcesOnly()
{
	return true;
}


//====================================================================
video_renderer_objects::Texture2D*  GUIRendererOpenGL::CreateResourceTexture2D(
						eSampleFormat sampleFormat, bool bMultiLevel, uint height, uint width)
{
	ASSERT_OUTSIDE_BEGIN_END;

	switch(sampleFormat) {
		case SampleFormat_X8R8G8B8:
		case SampleFormat_A8R8G8B8:
			// throws exception on error
			return new gl_tools::Texture2D(sampleFormat, bMultiLevel, height, width);
	}
	return 0;
}


//====================================================================
void  GUIRendererOpenGL::Begin()
{
	ASSERT_OUTSIDE_BEGIN_END;
	ASSERT_OUTSIDE_BEGIN_END_INTERLEAVED;

	Stats().numBegin++;

	// setup server states
	apDL_Begin->Execute();

	// setup client states
	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

	glDisableClientState(GL_EDGE_FLAG_ARRAY);
	glDisableClientState(GL_INDEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	gl_tools::TextureCoordArray::DisableAllArraysAndActivateArray0();
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	// update states
	bInsideBeginEnd = true;
	bTexturingEnabled = false;
	bAlphaTestEnabled = false;

	depthParams = depthParamsBase;

	InvalidateLastTexParams();

	appPoolVertexBuffer.BeginReserveCycle();
	appPoolVertexBuffer.SetAutoHookPointersToGLOnRealloc(true);
	appPoolVertexBuffer.HookPointersToGL();
}


//====================================================================
void  GUIRendererOpenGL::End()
{
	ASSERT_OUTSIDE_BEGIN_END_INTERLEAVED;
	ASSERT_INSIDE_BEGIN_END;

	Stats().numEnd++;

	// setup server states
	apDL_End->Execute();

	// setup client states
	glPopClientAttrib();

	// update states
	bInsideBeginEnd = false;
	bTexturingEnabled = false;
	bAlphaTestEnabled = false;

	InvalidateLastTexParams();

	appPoolVertexBuffer.SetAutoHookPointersToGLOnRealloc(false);
	appPoolVertexBuffer.EndReserveCycle();
}


//====================================================================
void  GUIRendererOpenGL::BeginInterleaved()
{
	ASSERT_INSIDE_BEGIN_END;

	Stats().numBeginInterleaved++;

	End();
	bInsideBeginEndInterleaved = true;
}


//====================================================================
void  GUIRendererOpenGL::EndInterleaved()
{
	ASSERT_INSIDE_BEGIN_END_INTERLEAVED;

	Stats().numEndInterleaved++;

	bInsideBeginEndInterleaved = false;
	Begin();
}


//====================================================================
void  GUIRendererOpenGL::SetDepthParams(const DepthParams& dp)
{
	ASSERT_INSIDE_BEGIN_END;

	Stats().numSetDepthParams++;

	depthParams = dp;
	Send_DepthParamsFlags(depthParams);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glPushMatrix();

	Send_DepthParamsRanges(depthParams);

	glMatrixMode(GL_MODELVIEW);
}


//====================================================================
void  GUIRendererOpenGL::GetDepthParams(DepthParams& dp) const
{
	dp = depthParams;
}


//====================================================================
void  GUIRendererOpenGL::SetTransform(const mat4& t)
{
	ASSERT_INSIDE_BEGIN_END;

	Stats().numSetTransform++;

	matTransform = t;
	Send_Transform();
}


//====================================================================
void  GUIRendererOpenGL::GetTransform(mat4& t) const
{
	t = matTransform;
}


//====================================================================
void  GUIRendererOpenGL::DrawPrimitive(const gui::VertexBufferC& vertexBuffer,
								gui::ePrimitiveType primitiveType,
								uint numPrimitives, uint startVertex)
{
	Stats().numDrawPrimitiveColored++;
	Stats().numPrimitivesColored += numPrimitives;

	ASSERT_INSIDE_BEGIN_END;

	Send_EnableAlphaTest(false);
	Send_DisableTexturing();

	uint vertexCount;
	GLenum primType = TranslatePrimitiveParams(primitiveType, numPrimitives, vertexCount);

	VW_ASSERT(vertexBuffer.size() >= (startVertex + vertexCount), "Insufficient gui-vertex-buffer size.");

	// convert vertices to GL format
	const gui::VertexC* pVertex    = vertexBuffer + startVertex;
	const gui::VertexC* pVertexEnd = pVertex + vertexCount;

	appPoolVertexBuffer.ReserveSpace(vertexCount);
	gui::VertexCT* pAppPoolVertex = appPoolVertexBuffer;

	while(pVertex != pVertexEnd) {
		pAppPoolVertex->position = pVertex->position;
		pAppPoolVertex->color = pVertex->color.FlippedRGBOrder();

		pAppPoolVertex++;
		pVertex++;
	}

	// send vertices
	glDrawArrays(primType, 0, vertexCount);
}


//====================================================================
void  GUIRendererOpenGL::DrawPrimitive(const gui::Texture& texture,
								const gui::VertexBufferCT& vertexBuffer,
								gui::ePrimitiveType primitiveType,
								uint numPrimitives, uint startVertex)
{
	Stats().numDrawPrimitiveTextured++;
	Stats().numPrimitivesTextured += numPrimitives;

	ASSERT_INSIDE_BEGIN_END;

	Send_EnableAlphaTest(false);
	Send_BindTextureAndEnableTexturing(texture);

	uint vertexCount;
	GLenum primType = TranslatePrimitiveParams(primitiveType, numPrimitives, vertexCount);

	VW_ASSERT(vertexBuffer.size() >= (startVertex + vertexCount), "Insufficient gui-vertex-buffer size.");

	// convert vertices to GL format
	const gui::VertexCT* pVertex    = vertexBuffer + startVertex;
	const gui::VertexCT* pVertexEnd = pVertex + vertexCount;

	appPoolVertexBuffer.ReserveSpace(vertexCount);
	gui::VertexCT* pAppPoolVertex = appPoolVertexBuffer;

	while(pVertex != pVertexEnd) {
		pAppPoolVertex->position = pVertex->position;
		pAppPoolVertex->color = pVertex->color.FlippedRGBOrder();
		pAppPoolVertex->texCoord.Set(pVertex->texCoord.x, 1.0f - pVertex->texCoord.y);

		pAppPoolVertex++;
		pVertex++;
	}

	// send vertices
	glDrawArrays(primType, 0, vertexCount);
}


//====================================================================
void  GUIRendererOpenGL::DrawText(const gui::FontTexture& fontTexture,
								uint textLen, const char* pText,
								const gui::DrawTextParams& dtp)
{
	Stats().numDrawText++;
	Stats().numPrimitivesTextured += textLen * 2;
	Stats().numTextGlyphs += textLen;

	ASSERT_INSIDE_BEGIN_END;
	if(! textLen) return;

	Send_EnableAlphaTest(true);
	Send_BindTextureAndEnableTexturing(fontTexture);

	// construct quads
	uint vertexCount = textLen * 4;
	appPoolVertexBuffer.ReserveSpace(vertexCount);

	gui::VertexCT* pAppPoolVertex = appPoolVertexBuffer;

	const  gui::TextMetrics& tm = dtp.tm;

	bool     bForceAsciiRange = (fontTexture.GetNumGlyphs() < 256);
	color4b  color = dtp.color.FlippedRGBOrder();
	float    spacing = fontTexture.CalcSpacingWidth(tm);
	float    gWidth = tm.bMonospace ? fontTexture.CalcGlyphWidth(tm, 0) : fontTexture.CalcGlyphWidthMultiplier(tm);

	vec3  positionTop = dtp.positionTopLeft;
	vec3  positionBottom(positionTop.x, positionTop.y + tm.textHeight, positionTop.z);

	for(uint i = 0; i < textLen; i++) {
		char ch = pText[i];

		// force 0-127 character range
		if(bForceAsciiRange  &&  ch < 0) ch = 0;

		// compute glyph's font-texture coords
		float  texLeft, texRight, texTop, texBottom;

		if(tm.bMonospace) {
			fontTexture.GetGlyphCoordsMono(ch, texLeft, texRight, texTop, texBottom);
		}
		else {
			fontTexture.GetGlyphCoords(ch, texLeft, texRight, texTop, texBottom);
		}

		// convert to OpenGL texture-coords style
		texTop = 1.0f - texTop;
		texBottom = 1.0f - texBottom;

		// top-left point
		pAppPoolVertex->position = positionTop;
		pAppPoolVertex->color = color;
		pAppPoolVertex->texCoord.Set(texLeft, texTop);
		pAppPoolVertex++;

		// bottom-left point
		pAppPoolVertex->position = positionBottom;
		pAppPoolVertex->color = color;
		pAppPoolVertex->texCoord.Set(texLeft, texBottom);
		pAppPoolVertex++;

		// shift position to glyph's right edge
		positionTop.x += (tm.bMonospace ? gWidth : (gWidth * fontTexture.GetGlyphWidth(ch)));
		positionBottom.x = positionTop.x;

		// bottom-right point
		pAppPoolVertex->position = positionBottom;
		pAppPoolVertex->color = color;
		pAppPoolVertex->texCoord.Set(texRight, texBottom);
		pAppPoolVertex++;

		// top-right point
		pAppPoolVertex->position = positionTop;
		pAppPoolVertex->color = color;
		pAppPoolVertex->texCoord.Set(texRight, texTop);
		pAppPoolVertex++;

		// shift position to next-glyph's left edge
		positionTop.x += spacing;
		positionBottom.x = positionTop.x;
	}

	// send vertices
	glDrawArrays(GL_QUADS, 0, vertexCount);
}


//====================================================================
void  GUIRendererOpenGL::Compile_Begin()
{
	if(! apDL_Begin) return;
	gl_tools::DisplayListCompileSentry  dlcs(*apDL_Begin);

	glPushAttrib(
		GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ENABLE_BIT | GL_HINT_BIT |
		GL_LIGHTING_BIT | GL_POLYGON_BIT | GL_TEXTURE_BIT | GL_SCISSOR_BIT |
		GL_TRANSFORM_BIT | GL_VIEWPORT_BIT);

	glDisable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glDisable(GL_POLYGON_STIPPLE);
	glDisable(GL_POLYGON_SMOOTH);
	glDisable(GL_POLYGON_OFFSET_FILL);
	glDisable(GL_DITHER);
	glDisable(GL_SCISSOR_TEST);
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glDisable(GL_FOG);
	glDisable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0);

	Send_DepthParamsFlags(depthParamsBase);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	gl_tools::TextureUnit::DisableAllUnitsAndActivateUnit0();
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glDisable(GL_TEXTURE_GEN_Q);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	Send_DepthParamsRanges(depthParamsBase);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	Send_Transform();

	glViewport(0, 0, GetWindowWidth(), GetWindowHeight());
}


//====================================================================
void  GUIRendererOpenGL::Compile_End()
{
	if(! apDL_End) return;
	gl_tools::DisplayListCompileSentry  dlcs(*apDL_End);

	glPopAttrib();

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}


//====================================================================
void  GUIRendererOpenGL::Compile_PrepareTextureParams()
{
	if(! apDL_PrepareTextureParams) return;
	gl_tools::DisplayListCompileSentry  dlcs(*apDL_PrepareTextureParams);

	GLint texAddressingMode = GL_CLAMP;
	if(gl_ext::IsSupported(gl_ext::Extension_GL_CORE_12)) {
		texAddressingMode = GL_CLAMP_TO_EDGE;
	}

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, texAddressingMode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, texAddressingMode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}


//====================================================================
void  GUIRendererOpenGL::Send_DepthParamsRanges(const DepthParams& dp)
{
	glLoadIdentity();

	glOrtho(0, GetWindowWidth(), 0, GetWindowHeight(),
			dp.clippingPlaneNear, dp.clippingPlaneFar);

	glDepthRange(dp.renderDepthNear, dp.renderDepthFar);
}


//====================================================================
void  GUIRendererOpenGL::Send_DepthParamsFlags(const DepthParams& dp)
{
	// z-buffer writes work only if GL_DEPTH_TEST is enabled
	gl_tools::Enable(GL_DEPTH_TEST, (dp.bEnableDepthWrite || dp.bEnableDepthTest));
	glDepthFunc(dp.bEnableDepthTest ? GL_LEQUAL : GL_ALWAYS);
	glDepthMask(dp.bEnableDepthWrite ? GL_TRUE : GL_FALSE);
}


//====================================================================
void  GUIRendererOpenGL::Send_EnableAlphaTest(bool bEnable)
{
	if(bAlphaTestEnabled != bEnable) {
		bAlphaTestEnabled = bEnable;

		gl_tools::Enable(GL_ALPHA_TEST, bEnable);
	}
}


//====================================================================
void  GUIRendererOpenGL::Send_DisableTexturing()
{
	if(bTexturingEnabled) {
		bTexturingEnabled = false;

		InvalidateLastTexParams();

		glDisable(GL_TEXTURE_2D);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
}


//====================================================================
void  GUIRendererOpenGL::Send_BindTextureAndEnableTexturing(const gui::Texture& texture)
{
	// enable texturing
	if(! bTexturingEnabled) {
		bTexturingEnabled = true;

		glEnable(GL_TEXTURE_2D);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	}

	// must bind texture before setting parameters
	if(pLastBoundTexture != &texture) {
		pLastBoundTexture = &texture;

		const gl_tools::Texture2D* pResource =
			static_cast<const gl_tools::Texture2D*>(texture.GetVideoRendererResource());

		VW_ASSERT(pResource, "Using gui::Texture without created video-renderer-resource.");
		if(pResource) {
			pResource->BindToActiveTextureUnit();
		}

		// setup common texture parameters
		apDL_PrepareTextureParams->Execute();
		lastTexMipFilter = gui::Texture::Filter_Invalid;

		Stats().numTextureChanges++;
	}

	// setup minification/mipmap filter
	if(lastTexMipFilter != texture.GetMipFilter()) {
		lastTexMipFilter = texture.IsMultiLevel() ?
				texture.GetMipFilter() : gui::Texture::Filter_Off;

		GLint filter = GL_LINEAR;

		switch(lastTexMipFilter) {
			case gui::Texture::Filter_Nearest:
				filter = GL_LINEAR_MIPMAP_NEAREST;
				break;

			case gui::Texture::Filter_Linear:
				filter = GL_LINEAR_MIPMAP_LINEAR;
				break;
		}
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);
	}
}


//====================================================================
void  GUIRendererOpenGL::ConstructInternalTransform()
{
	// OpenGL uses MV order but requires column-major matrices,
	// so compute as VM to avoid additional transpose later.
	mat4  matTemp;

	matInternalTransform.SetIdentity();

	// apply pixel-offset
	GenerateTranslateVM(matTemp, pixelParams.offsetX, pixelParams.offsetY, 0);
	matInternalTransform *= matTemp;

	// convert coords from Screen-RHS to GL-RHS
	GenerateNonUniformScale(matTemp, 1, -1, -1);
	matInternalTransform *= matTemp;

	GenerateTranslateVM(matTemp, 0, GetWindowHeight(), 0);
	matInternalTransform *= matTemp;
}


//====================================================================
GLenum __fastcall GUIRendererOpenGL::TranslatePrimitiveParams(gui::ePrimitiveType pt,
													uint numPrimitives, uint& outVertexCount)
{
	VW_ASSERT(numPrimitives > 0, "Primitive-count must be > 0.");

	switch(pt) {
		case gui::PrimitiveType_TriangleList:
			outVertexCount = numPrimitives * 3;
			return GL_TRIANGLES;

		case gui::PrimitiveType_TriangleStrip:
			outVertexCount = numPrimitives + 2;
			return GL_TRIANGLE_STRIP;

		case gui::PrimitiveType_TriangleFan:
			outVertexCount = numPrimitives + 2;
			return GL_TRIANGLE_FAN;
	}
	VW_ASSERT(0, "Unsupported gui-primitive-type.");
	return 0;
}


//====================================================================
void  GUIRendererOpenGL::AppPoolVertexBuffer::DoHookPointersToGL(const gui::VertexCT* pVertexZero) const
{
	glVertexPointer(3, GL_FLOAT, sizeof(AppPoolVertexBuffer::VertexType), &pVertexZero->position);
	glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(AppPoolVertexBuffer::VertexType), &pVertexZero->color);
	glTexCoordPointer(2, GL_FLOAT, sizeof(AppPoolVertexBuffer::VertexType), &pVertexZero->texCoord);
}


//********************************************************************
}//namespace vertexware


