

//--------------------------------------------------------------------
#include "Framework\VideoRendererOpenGL.h"
#include "GLTools\Extensions.h"
#include "GLTools\Stacks.h"
#include "GLTools\Textures.h"
#include "BaseTools\String.h"
#include "BaseTools\Stream.h"
#include "BaseTools\Debug.h"
#include <list>


//********************************************************************
namespace {


//--------------------------------------------------------------------
using namespace vertexware;


//====================================================================
class ExtensionProc
{
private:
	std::string  extensionProcName;
	void**       pProcPointer;


public:
	ExtensionProc(const std::string& _extensionProcName, void** _pProcPointer) :
		extensionProcName(_extensionProcName), pProcPointer(_pProcPointer)
	{ }

	bool  HookProcPointer(VideoRendererOpenGL& vr) {
		*pProcPointer = gl_tools::GetExtensionProc(extensionProcName.c_str());
		return (*pProcPointer != 0);
	}

	void  InvalidateProcPointer() {
		*pProcPointer = 0;
	}
};


//====================================================================
class ExtensionDesc
{
private:
	typedef  std::list<ExtensionProc>  EXTPROC_LIST;
	typedef  EXTPROC_LIST::iterator    EXTPROC_LIST_ITER;

	EXTPROC_LIST        extensionProcs;
	std::string         name;
	gl_ext::eExtension  extension;
	bool                bCheckExtensionName;
	bool                bCheckVersion;
	uint                minVersionMajor;
	uint                minVersionMinor;

	bool  bIsSupported;


public:
	ExtensionDesc(gl_ext::eExtension _extension,
					bool _bCheckExtensionName = true,
					bool _bCheckVersion = false,
					uint _minVersionMajor = 0,
					uint _minVersionMinor = 0)
		:
		extension(_extension),
		bCheckExtensionName(_bCheckExtensionName),
		bCheckVersion(_bCheckVersion),
		minVersionMajor(_minVersionMajor),
		minVersionMinor(_minVersionMinor)
	{
		name = gl_ext::GetExtensionName(extension);
		SetSupported(false);
	}

	// extension info
	const std::string&  GetName() { return name; }
	bool                IsSupported() { return bIsSupported; }

	// register extension-proc
	void  AddExtensionProc(const std::string& extensionProcName, void** pProcPointer)
	{
		extensionProcs.push_back(ExtensionProc(extensionProcName, pProcPointer));
	}

	// check support and hook all registered pointers
	void  Validate(VideoRendererOpenGL& vr)
	{
		VW_ASSERT(! bIsSupported, "Internal error");

		if(! vr.IsExtensionEnabled(extension)) return;

		if(bCheckExtensionName) {
			if(! gl_tools::IsExtensionSupported(vr.GetExtensionsString(), name.c_str())) return;
		}
		if(bCheckVersion) {
			if(gl_ext::GetVersionGLMajor() < minVersionMajor ||
				gl_ext::GetVersionGLMinor() < minVersionMinor) return;
		}

		SetSupported(true);

		EXTPROC_LIST_ITER it = extensionProcs.begin();
		while(it != extensionProcs.end()) {
			if(! (*it++).HookProcPointer(vr)) {
				bIsSupported = false;
				break;
			}
		}
		// auto-invalidate on error
		if(! bIsSupported) {
			Invalidate();
		}
	}

	// reset support-flags and invalidate registered pointers
	void  Invalidate()
	{
		SetSupported(false);

		EXTPROC_LIST_ITER it = extensionProcs.begin();
		while(it != extensionProcs.end()) {
			(*it++).InvalidateProcPointer();
		}
	}


private:
	void  SetSupported(bool bSupported)
	{
		bIsSupported = bSupported;
		gl_ext::SetSupported(extension, bSupported);
	}
};


//====================================================================
#define  ADD_EXT_PROC(extensionDesc, pointerNamespace, extensionProc) \
	extensionDesc.AddExtensionProc(#extensionProc, \
					reinterpret_cast<void**>(&pointerNamespace::extensionProc));


//====================================================================
class ExtensionDescList
{
private:
	typedef  std::list<ExtensionDesc>  EXTDESC_LIST;
	typedef  EXTDESC_LIST::iterator    EXTDESC_LIST_ITER;

	EXTDESC_LIST  extensionDescs;


public:
	void  AddExtensionDesc(const ExtensionDesc& ed) {
		extensionDescs.push_back(ed);
	}

	void  Validate(VideoRendererOpenGL& vr) {
		EXTDESC_LIST_ITER it = extensionDescs.begin();
		while(it != extensionDescs.end()) {
			(*it++).Validate(vr);
		}
	}

	void  Invalidate() {
		EXTDESC_LIST_ITER it = extensionDescs.begin();
		while(it != extensionDescs.end()) {
			(*it++).Invalidate();
		}
	}
};


//********************************************************************
}//namespace


//********************************************************************
namespace vertexware {


//====================================================================
VideoRendererOpenGL::VideoRendererOpenGL()
{
	pVideoSubsystemOpenGL = 0;
}


//====================================================================
VideoRendererOpenGL::~VideoRendererOpenGL()
{
}


//====================================================================
std::string  VideoRendererOpenGL::GetVideoRendererInterfaceName()
{
	return VW_INTERFACE_NAME_VideoRendererOpenGL;
}


//====================================================================
std::string  VideoRendererOpenGL::GetMatchingVideoSubsystemInterfaceName()
{
	return VW_INTERFACE_NAME_VideoSubsystemOpenGL;
}


//====================================================================
void  VideoRendererOpenGL::ConnectVideoSubsystem(VideoSubsystem& vs)
{
	VideoRenderer::ConnectVideoSubsystem(vs);

	pVideoSubsystemOpenGL = InterfaceCast<VideoSubsystemOpenGL>(&vs);

	gl_ext::SetGLReady(true);
	gl_ext::ClearImplementationInfo();
	SetupGLVersion(true);
	SetupGLExtensions(true);

	std::ostream& os = GetLogOut();
	os << "\nOpenGL driver info:\n";
	DumpDriverInfo(os);
	os << nl;

	RefreshExtensionPointers();
}


//====================================================================
void  VideoRendererOpenGL::DisconnectVideoSubsystem()
{
	gl_ext::SetGLReady(false);
	gl_ext::ClearImplementationInfo();
	SetupGLVersion(false);
	SetupGLExtensions(false);

	pVideoSubsystemOpenGL = 0;

	VideoRenderer::DisconnectVideoSubsystem();
}


//====================================================================
void  VideoRendererOpenGL::CreateVideoResources(bool bPostReset)
{
	VideoRenderer::CreateVideoResources(bPostReset);
}


//====================================================================
void  VideoRendererOpenGL::ReleaseVideoResources(bool bPreReset)
{
	VideoRenderer::ReleaseVideoResources(bPreReset);
}


//====================================================================
bool  VideoRendererOpenGL::IsExtensionEnabled(gl_ext::eExtension e)
{
	return (disabledExtensions.find(e) == disabledExtensions.end());
}


//====================================================================
void  VideoRendererOpenGL::EnableExtension(gl_ext::eExtension e, bool bEnable)
{
	if(bEnable) {
		disabledExtensions.erase(e);
	}
	else {
		disabledExtensions.insert(e);
	}
}


//====================================================================
void  VideoRendererOpenGL::EnableAllExtensions(bool bEnable)
{
	if(bEnable) {
		disabledExtensions.clear();
	}
	else {
		for(uint i = 0; i < gl_ext::Extension_Count; i++) {
			disabledExtensions.insert(gl_ext::eExtension(i));
		}
	}
}


//====================================================================
void  VideoRendererOpenGL::RefreshExtensionPointers()
{
	if(pVideoSubsystemOpenGL) {
		GetLogOut() << "Refreshing OpenGL extension-pointers.\n";
		SetupGLExtensions(true);
	}
}


//====================================================================
void  VideoRendererOpenGL::DumpDriverInfo(std::ostream& os, uint tabLevel,
						bool bVendor, bool bExtensions, bool bAdditional)
{
	TabEffector te(tabLevel);

	if(bVendor) {
		os << te << "Vendor..: " << glGetString(GL_VENDOR) << nl;
		os << te << "Renderer: " << glGetString(GL_RENDERER) << nl;
		os << te << "Version.: " << glGetString(GL_VERSION) << nl;
	}
	if(bExtensions) {
		if(bVendor) os << nl;
		os << te << "Extensions:\n";
		os << AddLinePrefix(FitToLineLengthBreakAtWhitespaceOnly(extensionsString, 0),
							std::string(tabLevel + 1, '\t')) << nl;
	}
	if(bAdditional) {
		if(bVendor || bExtensions) os << nl;
		os << te << "NumTextureUnits..........: " << gl_tools::TextureUnit::GetNumUnits() << nl;
		os << te << "MaxTextureSize...........: " << gl_tools::TextureBase::GetMaxTextureSize() << nl;
		os << te << "ModelViewStackMaxDepth...: " << gl_tools::MatrixStack::GetMaxStackDepth(GL_MODELVIEW) << nl;
		os << te << "ProjectionStackMaxDepth..: " << gl_tools::MatrixStack::GetMaxStackDepth(GL_PROJECTION) << nl;
		os << te << "TextureStackMaxDepth.....: " << gl_tools::MatrixStack::GetMaxStackDepth(GL_TEXTURE) << nl;
		os << te << "ColorStackMaxDepth.......: " << gl_tools::MatrixStack::GetMaxStackDepth(GL_COLOR) << nl;
		os << te << "AttribStackMaxDepth......: " << gl_tools::AttribStack::GetMaxStackDepth() << nl;
		os << te << "ClientAttribStackMaxDepth: " << gl_tools::ClientAttribStack::GetMaxStackDepth() << nl;
	}
}


//====================================================================
void  VideoRendererOpenGL::DumpDisabledExtensions(std::ostream& os, uint tabLevel)
{
	TabEffector te(tabLevel);

	for(uint i = 0; i < gl_ext::Extension_Count; i++) {
		if(disabledExtensions.find(gl_ext::eExtension(i)) != disabledExtensions.end()) {
			os << te << gl_ext::GetExtensionName(gl_ext::eExtension(i)) << nl;
		}
	}
}


//====================================================================
void  VideoRendererOpenGL::SetupGLVersion(bool bInit)
{
	uint  major = 0;
	uint  minor = 0;
	uint  release = 0;

	if(bInit) {
		gl_tools::ExtractVersionNumbers(major, minor, release);
	}

	gl_ext::SetVersionGLMajor(major);
	gl_ext::SetVersionGLMinor(minor);
	gl_ext::SetVersionGLRelease(release);
}


//====================================================================
void  VideoRendererOpenGL::SetupGLExtensions(bool bInit)
{
	ExtensionDescList edl;
	FillExtensionDescList(&edl);

	if(bInit) {
		gl_tools::GetExtensionsString(GetVideoSubsystem()->GetWindow()->GetHDC(), extensionsString);
		edl.Validate(*this);
	}
	else {
		extensionsString.erase();
		edl.Invalidate();
	}
}


//====================================================================
void  VideoRendererOpenGL::FillExtensionDescList(void* v)
{
	ExtensionDescList* edl = reinterpret_cast<ExtensionDescList*>(v);

	//-----------
	// GL_CORE_12
	//-----------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_CORE_12, false, true, 1, 2);
	ADD_EXT_PROC(ed, gl_ext, glDrawRangeElements);
	ADD_EXT_PROC(ed, gl_ext, glTexImage3D);
	ADD_EXT_PROC(ed, gl_ext, glTexSubImage3D);
	ADD_EXT_PROC(ed, gl_ext, glCopyTexSubImage3D);
	edl->AddExtensionDesc(ed);
	}

	//-----------
	// GL_CORE_13
	//-----------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_CORE_13, false, true, 1, 3);
	ADD_EXT_PROC(ed, gl_ext, glActiveTexture);
	ADD_EXT_PROC(ed, gl_ext, glClientActiveTexture);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1d);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1dv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1f);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1fv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1i);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1iv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1s);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1sv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2d);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2dv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2f);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2fv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2i);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2iv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2s);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2sv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3d);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3dv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3f);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3fv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3i);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3iv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3s);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3sv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4d);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4dv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4f);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4fv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4i);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4iv);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4s);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4sv);
	ADD_EXT_PROC(ed, gl_ext, glLoadTransposeMatrixf);
	ADD_EXT_PROC(ed, gl_ext, glLoadTransposeMatrixd);
	ADD_EXT_PROC(ed, gl_ext, glMultTransposeMatrixf);
	ADD_EXT_PROC(ed, gl_ext, glMultTransposeMatrixd);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexImage1D);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexImage2D);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexImage3D);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexSubImage1D);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexSubImage2D);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexSubImage3D);
	ADD_EXT_PROC(ed, gl_ext, glGetCompressedTexImage);
	ADD_EXT_PROC(ed, gl_ext, glSampleCoverage);
	edl->AddExtensionDesc(ed);
	}

	//---------------
	// GL_ARB_imaging
	//---------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_imaging);
	ADD_EXT_PROC(ed, gl_ext, glBlendColor);
	ADD_EXT_PROC(ed, gl_ext, glBlendEquation);
	ADD_EXT_PROC(ed, gl_ext, glColorTable);
	ADD_EXT_PROC(ed, gl_ext, glColorTableParameterfv);
	ADD_EXT_PROC(ed, gl_ext, glColorTableParameteriv);
	ADD_EXT_PROC(ed, gl_ext, glCopyColorTable);
	ADD_EXT_PROC(ed, gl_ext, glGetColorTable);
	ADD_EXT_PROC(ed, gl_ext, glGetColorTableParameterfv);
	ADD_EXT_PROC(ed, gl_ext, glGetColorTableParameteriv);
	ADD_EXT_PROC(ed, gl_ext, glColorSubTable);
	ADD_EXT_PROC(ed, gl_ext, glCopyColorSubTable);
	ADD_EXT_PROC(ed, gl_ext, glConvolutionFilter1D);
	ADD_EXT_PROC(ed, gl_ext, glConvolutionFilter2D);
	ADD_EXT_PROC(ed, gl_ext, glConvolutionParameterf);
	ADD_EXT_PROC(ed, gl_ext, glConvolutionParameterfv);
	ADD_EXT_PROC(ed, gl_ext, glConvolutionParameteri);
	ADD_EXT_PROC(ed, gl_ext, glConvolutionParameteriv);
	ADD_EXT_PROC(ed, gl_ext, glCopyConvolutionFilter1D);
	ADD_EXT_PROC(ed, gl_ext, glCopyConvolutionFilter2D);
	ADD_EXT_PROC(ed, gl_ext, glGetConvolutionFilter);
	ADD_EXT_PROC(ed, gl_ext, glGetConvolutionParameterfv);
	ADD_EXT_PROC(ed, gl_ext, glGetConvolutionParameteriv);
	ADD_EXT_PROC(ed, gl_ext, glGetSeparableFilter);
	ADD_EXT_PROC(ed, gl_ext, glSeparableFilter2D);
	ADD_EXT_PROC(ed, gl_ext, glGetHistogram);
	ADD_EXT_PROC(ed, gl_ext, glGetHistogramParameterfv);
	ADD_EXT_PROC(ed, gl_ext, glGetHistogramParameteriv);
	ADD_EXT_PROC(ed, gl_ext, glGetMinmax);
	ADD_EXT_PROC(ed, gl_ext, glGetMinmaxParameterfv);
	ADD_EXT_PROC(ed, gl_ext, glGetMinmaxParameteriv);
	ADD_EXT_PROC(ed, gl_ext, glHistogram);
	ADD_EXT_PROC(ed, gl_ext, glMinmax);
	ADD_EXT_PROC(ed, gl_ext, glResetHistogram);
	ADD_EXT_PROC(ed, gl_ext, glResetMinmax);
	edl->AddExtensionDesc(ed);
	}

	//----------------------
	// GL_ARB_matrix_palette
	//----------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_matrix_palette);
	ADD_EXT_PROC(ed, gl_ext, glCurrentPaletteMatrixARB);
	ADD_EXT_PROC(ed, gl_ext, glMatrixIndexubvARB);
	ADD_EXT_PROC(ed, gl_ext, glMatrixIndexusvARB);
	ADD_EXT_PROC(ed, gl_ext, glMatrixIndexuivARB);
	ADD_EXT_PROC(ed, gl_ext, glMatrixIndexPointerARB);
	edl->AddExtensionDesc(ed);
	}

	//-------------------
	// GL_ARB_multisample
	//-------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_multisample);
	ADD_EXT_PROC(ed, gl_ext, glSampleCoverageARB);
	edl->AddExtensionDesc(ed);
	}

	//--------------------
	// GL_ARB_multitexture
	//--------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_multitexture);
	ADD_EXT_PROC(ed, gl_ext, glActiveTextureARB);
	ADD_EXT_PROC(ed, gl_ext, glClientActiveTextureARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1dARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1dvARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1fARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1fvARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1iARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1ivARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1sARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord1svARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2dARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2dvARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2fARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2fvARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2iARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2ivARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2sARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord2svARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3dARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3dvARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3fARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3fvARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3iARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3ivARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3sARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord3svARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4dARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4dvARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4fARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4fvARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4iARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4ivARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4sARB);
	ADD_EXT_PROC(ed, gl_ext, glMultiTexCoord4svARB);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_ARB_point_parameters
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_point_parameters);
	ADD_EXT_PROC(ed, gl_ext, glPointParameterfARB);
	ADD_EXT_PROC(ed, gl_ext, glPointParameterfvARB);
	edl->AddExtensionDesc(ed);
	}

	//---------------------------
	// GL_ARB_texture_compression
	//---------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_texture_compression);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexImage1DARB);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexImage2DARB);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexImage3DARB);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexSubImage1DARB);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexSubImage2DARB);
	ADD_EXT_PROC(ed, gl_ext, glCompressedTexSubImage3DARB);
	ADD_EXT_PROC(ed, gl_ext, glGetCompressedTexImageARB);
	edl->AddExtensionDesc(ed);
	}

	//-----------------------
	// GL_ARB_texture_env_add
	//-----------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_texture_env_add);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_ARB_texture_env_dot3
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_texture_env_dot3);
	edl->AddExtensionDesc(ed);
	}

	//---------------------------
	// GL_ARB_texture_env_combine
	//---------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_texture_env_combine);
	edl->AddExtensionDesc(ed);
	}

	//----------------------------
	// GL_ARB_texture_env_crossbar
	//----------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_texture_env_crossbar);
	edl->AddExtensionDesc(ed);
	}

	//----------------------------
	// GL_ARB_texture_border_clamp
	//----------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_texture_border_clamp);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_ARB_texture_cube_map
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_texture_cube_map);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_ARB_transpose_matrix
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_transpose_matrix);
	ADD_EXT_PROC(ed, gl_ext, glLoadTransposeMatrixfARB);
	ADD_EXT_PROC(ed, gl_ext, glLoadTransposeMatrixdARB);
	ADD_EXT_PROC(ed, gl_ext, glMultTransposeMatrixfARB);
	ADD_EXT_PROC(ed, gl_ext, glMultTransposeMatrixdARB);
	edl->AddExtensionDesc(ed);
	}

	//--------------------
	// GL_ARB_vertex_blend
	//--------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ARB_vertex_blend);
	ADD_EXT_PROC(ed, gl_ext, glWeightbvARB);
	ADD_EXT_PROC(ed, gl_ext, glWeightsvARB);
	ADD_EXT_PROC(ed, gl_ext, glWeightivARB);
	ADD_EXT_PROC(ed, gl_ext, glWeightfvARB);
	ADD_EXT_PROC(ed, gl_ext, glWeightdvARB);
	ADD_EXT_PROC(ed, gl_ext, glWeightubvARB);
	ADD_EXT_PROC(ed, gl_ext, glWeightusvARB);
	ADD_EXT_PROC(ed, gl_ext, glWeightuivARB);
	ADD_EXT_PROC(ed, gl_ext, glWeightPointerARB);
	ADD_EXT_PROC(ed, gl_ext, glVertexBlendARB);
	edl->AddExtensionDesc(ed);
	}

	//------------
	// GL_EXT_abgr
	//------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_abgr);
	edl->AddExtensionDesc(ed);
	}

	//------------
	// GL_EXT_bgra
	//------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_bgra);
	edl->AddExtensionDesc(ed);
	}

	//-----------------------------
	// GL_EXT_compiled_vertex_array
	//-----------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_compiled_vertex_array);
	ADD_EXT_PROC(ed, gl_ext, glLockArraysEXT);
	ADD_EXT_PROC(ed, gl_ext, glUnlockArraysEXT);
	edl->AddExtensionDesc(ed);
	}

	//-----------------
	// GL_EXT_fog_coord
	//-----------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_fog_coord);
	ADD_EXT_PROC(ed, gl_ext, glFogCoordfEXT);
	ADD_EXT_PROC(ed, gl_ext, glFogCoordfvEXT);
	ADD_EXT_PROC(ed, gl_ext, glFogCoorddEXT);
	ADD_EXT_PROC(ed, gl_ext, glFogCoorddvEXT);
	ADD_EXT_PROC(ed, gl_ext, glFogCoordPointerEXT);
	edl->AddExtensionDesc(ed);
	}

	//-------------------------
	// GL_EXT_multi_draw_arrays
	//-------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_multi_draw_arrays);
	ADD_EXT_PROC(ed, gl_ext, glMultiDrawArraysEXT);
	ADD_EXT_PROC(ed, gl_ext, glMultiDrawElementsEXT);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_EXT_point_parameters
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_point_parameters);
	ADD_EXT_PROC(ed, gl_ext, glPointParameterfEXT);
	ADD_EXT_PROC(ed, gl_ext, glPointParameterfvEXT);
	edl->AddExtensionDesc(ed);
	}

	//-----------------------
	// GL_EXT_secondary_color
	//-----------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_secondary_color);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3bEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3bvEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3dEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3dvEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3fEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3fvEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3iEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3ivEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3sEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3svEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3ubEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3ubvEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3uiEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3uivEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3usEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColor3usvEXT);
	ADD_EXT_PROC(ed, gl_ext, glSecondaryColorPointerEXT);
	edl->AddExtensionDesc(ed);
	}

	//--------------------
	// GL_EXT_stencil_wrap
	//--------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_stencil_wrap);
	edl->AddExtensionDesc(ed);
	}

	//----------------------------------
	// GL_EXT_texture_filter_anisotropic
	//----------------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_texture_filter_anisotropic);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_EXT_texture_lod_bias
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_texture_lod_bias);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_EXT_vertex_weighting
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_EXT_vertex_weighting);
	ADD_EXT_PROC(ed, gl_ext, glVertexWeightfEXT);
	ADD_EXT_PROC(ed, gl_ext, glVertexWeightfvEXT);
	ADD_EXT_PROC(ed, gl_ext, glVertexWeightPointerEXT);
	edl->AddExtensionDesc(ed);
	}

	//--------------------
	// GL_ATI_pn_triangles
	//--------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_ATI_pn_triangles);
	ADD_EXT_PROC(ed, gl_ext, glPNTrianglesiATI);
	ADD_EXT_PROC(ed, gl_ext, glPNTrianglesfATI);
	edl->AddExtensionDesc(ed);
	}

	//-------------------
	// GL_NV_blend_square
	//-------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_blend_square);
	edl->AddExtensionDesc(ed);
	}

	//--------------------------
	// GL_NV_copy_depth_to_color
	//--------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_copy_depth_to_color);
	edl->AddExtensionDesc(ed);
	}

	//-----------------
	// GL_NV_evaluators
	//-----------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_evaluators);
	ADD_EXT_PROC(ed, gl_ext, glMapControlPointsNV);
	ADD_EXT_PROC(ed, gl_ext, glMapParameterivNV);
	ADD_EXT_PROC(ed, gl_ext, glMapParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetMapControlPointsNV);
	ADD_EXT_PROC(ed, gl_ext, glGetMapParameterivNV);
	ADD_EXT_PROC(ed, gl_ext, glGetMapParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetMapAttribParameterivNV);
	ADD_EXT_PROC(ed, gl_ext, glGetMapAttribParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glEvalMapsNV);
	edl->AddExtensionDesc(ed);
	}

	//------------
	// GL_NV_fence
	//------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_fence);
	ADD_EXT_PROC(ed, gl_ext, glGenFencesNV);
	ADD_EXT_PROC(ed, gl_ext, glDeleteFencesNV);
	ADD_EXT_PROC(ed, gl_ext, glSetFenceNV);
	ADD_EXT_PROC(ed, gl_ext, glTestFenceNV);
	ADD_EXT_PROC(ed, gl_ext, glFinishFenceNV);
	ADD_EXT_PROC(ed, gl_ext, glIsFenceNV);
	ADD_EXT_PROC(ed, gl_ext, glGetFenceivNV);
	edl->AddExtensionDesc(ed);
	}

	//-------------------
	// GL_NV_fog_distance
	//-------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_fog_distance);
	edl->AddExtensionDesc(ed);
	}

	//-------------------------
	// GL_NV_light_max_exponent
	//-------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_light_max_exponent);
	edl->AddExtensionDesc(ed);
	}

	//---------------------------
	// GL_NV_packed_depth_stencil
	//---------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_packed_depth_stencil);
	edl->AddExtensionDesc(ed);
	}

	//-------------------------
	// GL_NV_register_combiners
	//-------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_register_combiners);
	ADD_EXT_PROC(ed, gl_ext, glCombinerParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glCombinerParameterfNV);
	ADD_EXT_PROC(ed, gl_ext, glCombinerParameterivNV);
	ADD_EXT_PROC(ed, gl_ext, glCombinerParameteriNV);
	ADD_EXT_PROC(ed, gl_ext, glCombinerInputNV);
	ADD_EXT_PROC(ed, gl_ext, glCombinerOutputNV);
	ADD_EXT_PROC(ed, gl_ext, glFinalCombinerInputNV);
	ADD_EXT_PROC(ed, gl_ext, glGetCombinerInputParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetCombinerInputParameterivNV);
	ADD_EXT_PROC(ed, gl_ext, glGetCombinerOutputParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetCombinerOutputParameterivNV);
	ADD_EXT_PROC(ed, gl_ext, glGetFinalCombinerInputParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetFinalCombinerInputParameterivNV);
	edl->AddExtensionDesc(ed);
	}

	//--------------------------
	// GL_NV_register_combiners2
	//--------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_register_combiners2);
	ADD_EXT_PROC(ed, gl_ext, glCombinerStageParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetCombinerStageParameterfvNV);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_NV_texgen_reflection
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_texgen_reflection);
	edl->AddExtensionDesc(ed);
	}

	//---------------------------
	// GL_NV_texture_env_combine4
	//---------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_texture_env_combine4);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_NV_texture_rectangle
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_texture_rectangle);
	edl->AddExtensionDesc(ed);
	}

	//---------------------
	// GL_NV_texture_shader
	//---------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_texture_shader);
	edl->AddExtensionDesc(ed);
	}

	//----------------------
	// GL_NV_texture_shader2
	//----------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_texture_shader2);
	edl->AddExtensionDesc(ed);
	}

	//-------------------------
	// GL_NV_vertex_array_range
	//-------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_vertex_array_range);
	ADD_EXT_PROC(ed, gl_ext, glFlushVertexArrayRangeNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexArrayRangeNV);
	edl->AddExtensionDesc(ed);
	}

	//--------------------------
	// GL_NV_vertex_array_range2
	//--------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_vertex_array_range2);
	edl->AddExtensionDesc(ed);
	}
 
	//---------------------
	// GL_NV_vertex_program
	//---------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_NV_vertex_program);
	ADD_EXT_PROC(ed, gl_ext, glBindProgramNV);
	ADD_EXT_PROC(ed, gl_ext, glDeleteProgramsNV);
	ADD_EXT_PROC(ed, gl_ext, glExecuteProgramNV);
	ADD_EXT_PROC(ed, gl_ext, glGenProgramsNV);
	ADD_EXT_PROC(ed, gl_ext, glAreProgramsResidentNV);
	ADD_EXT_PROC(ed, gl_ext, glRequestResidentProgramsNV);
	ADD_EXT_PROC(ed, gl_ext, glGetProgramParameterfvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetProgramParameterdvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetProgramivNV);
	ADD_EXT_PROC(ed, gl_ext, glGetProgramStringNV);
	ADD_EXT_PROC(ed, gl_ext, glGetTrackMatrixivNV);
	ADD_EXT_PROC(ed, gl_ext, glGetVertexAttribdvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetVertexAttribfvNV);
	ADD_EXT_PROC(ed, gl_ext, glGetVertexAttribivNV);
	ADD_EXT_PROC(ed, gl_ext, glGetVertexAttribPointervNV);
	ADD_EXT_PROC(ed, gl_ext, glIsProgramNV);
	ADD_EXT_PROC(ed, gl_ext, glLoadProgramNV);
	ADD_EXT_PROC(ed, gl_ext, glProgramParameter4fNV);
	ADD_EXT_PROC(ed, gl_ext, glProgramParameter4dNV);
	ADD_EXT_PROC(ed, gl_ext, glProgramParameter4dvNV);
	ADD_EXT_PROC(ed, gl_ext, glProgramParameter4fvNV);
	ADD_EXT_PROC(ed, gl_ext, glProgramParameters4dvNV);
	ADD_EXT_PROC(ed, gl_ext, glProgramParameters4fvNV);
	ADD_EXT_PROC(ed, gl_ext, glTrackMatrixNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribPointerNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib1sNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib1fNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib1dNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib2sNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib2fNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib2dNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib3sNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib3fNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib3dNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib4sNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib4fNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib4dNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib4ubNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib1svNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib1fvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib1dvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib2svNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib2fvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib2dvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib3svNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib3fvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib3dvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib4svNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib4fvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib4dvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttrib4ubvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs1svNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs1fvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs1dvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs2svNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs2fvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs2dvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs3svNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs3fvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs3dvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs4svNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs4fvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs4dvNV);
	ADD_EXT_PROC(ed, gl_ext, glVertexAttribs4ubvNV);
	edl->AddExtensionDesc(ed);
	}

	//------------------------
	// GL_SGIS_generate_mipmap
	//------------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_SGIS_generate_mipmap);
	edl->AddExtensionDesc(ed);
	}

	//---------------
	// GL_SGIX_shadow
	//---------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_SGIX_shadow);
	edl->AddExtensionDesc(ed);
	}

	//----------------------
	// GL_SGIX_depth_texture
	//----------------------
	{
	ExtensionDesc ed(gl_ext::Extension_GL_SGIX_depth_texture);
	edl->AddExtensionDesc(ed);
	}
}


//********************************************************************
}//namespace vertexware


