

//--------------------------------------------------------------------
#include "GLTools\Stacks.h"
#include "GLTools\States.h"
#include "GLTools\Extensions.h"
#include "GLTools\Textures.h"
#include "BaseTools\Debug.h"
#include "BaseTools\Stream.h"


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gl_tools {


//====================================================================
std::string __fastcall MatrixStack::GetStackName(GLenum stack)
{
	switch(stack) {
		case GL_MODELVIEW:  return "ModelViewMatrixStack";
		case GL_PROJECTION: return "ProjectionMatrixStack";
		case GL_TEXTURE:    return "TextureMatrixStack";
		case GL_COLOR:      return "ColorMatrixStack";
	}
	return "UnknownMatrixStack";
}


//====================================================================
uint __fastcall MatrixStack::GetMaxStackDepth(GLenum stack)
{
	switch(stack) {
		case GL_MODELVIEW:
			return gl_tools::GetInteger(GL_MAX_MODELVIEW_STACK_DEPTH);

		case GL_PROJECTION:
			return gl_tools::GetInteger(GL_MAX_PROJECTION_STACK_DEPTH);

		case GL_TEXTURE:
			return gl_tools::GetInteger(GL_MAX_TEXTURE_STACK_DEPTH);

		case GL_COLOR:
			return (gl_ext::IsSupported(gl_ext::Extension_GL_ARB_imaging) ?
				gl_tools::GetInteger(GL_MAX_COLOR_MATRIX_STACK_DEPTH) : 0);

	}
	return 0;
}


//====================================================================
uint __fastcall MatrixStack::GetStackDepth(GLenum stack)
{
	switch(stack) {
		case GL_MODELVIEW:
			return gl_tools::GetInteger(GL_MODELVIEW_STACK_DEPTH);

		case GL_PROJECTION:
			return gl_tools::GetInteger(GL_PROJECTION_STACK_DEPTH);

		case GL_TEXTURE:
			return gl_tools::GetInteger(GL_TEXTURE_STACK_DEPTH);

		case GL_COLOR:
			return (gl_ext::IsSupported(gl_ext::Extension_GL_ARB_imaging) ?
				gl_tools::GetInteger(GL_COLOR_MATRIX_STACK_DEPTH) : 0);
	}
	return 0;
}


//====================================================================
uint __fastcall MatrixStack::GetAvailableStackSpace(GLenum stack)
{
	return (GetMaxStackDepth(stack) - GetStackDepth(stack));
}


//====================================================================
void  MatrixStack::Push()
{
	VW_ASSERT(GetAvailableStackSpace(stack), GetStackName(stack) + " overflow imminent.");
	glMatrixMode(stack);
	glPushMatrix();
}


//====================================================================
void  MatrixStack::Pop()
{
	VW_ASSERT(GetStackDepth(stack), GetStackName(stack) + " underflow imminent.");
	glMatrixMode(stack);
	glPopMatrix();
}


//====================================================================
MatrixStackSentry::MatrixStackSentry(GLenum _stack)
{
	VW_ASSERT(MatrixStack::GetAvailableStackSpace(_stack),
						MatrixStack::GetStackName(_stack) + " overflow imminent.");
	stack = _stack;
	glMatrixMode(stack);
	glPushMatrix();
}


//====================================================================
MatrixStackSentry::~MatrixStackSentry()
{
	VW_ASSERT(MatrixStack::GetStackDepth(stack),
						MatrixStack::GetStackName(stack) + " underflow imminent.");
	glMatrixMode(stack);
	glPopMatrix();
}


//====================================================================
uint __fastcall AttribStack::GetAvailableStackSpace()
{
	return (GetMaxStackDepth() - GetStackDepth());
}


//====================================================================
void  AttribStack::Push(GLbitfield attribMask)
{
	VW_ASSERT(GetAvailableStackSpace(), "AttributeStack overflow imminent.");
	glPushAttrib(attribMask);
}


//====================================================================
void  AttribStack::Pop()
{
	VW_ASSERT(GetStackDepth(), "AttributeStack underflow imminent.");
	glPopAttrib();
}


//====================================================================
AttribStackSentry::AttribStackSentry(GLbitfield attribMask)
{
	VW_ASSERT(AttribStack::GetAvailableStackSpace(), "AttributeStack overflow imminent.");
	glPushAttrib(attribMask);
}


//====================================================================
AttribStackSentry::~AttribStackSentry()
{
	VW_ASSERT(AttribStack::GetStackDepth(), "AttributeStack underflow imminent.");
	glPopAttrib();
}


//====================================================================
uint __fastcall ClientAttribStack::GetAvailableStackSpace()
{
	return (GetMaxStackDepth() - GetStackDepth());
}


//====================================================================
void  ClientAttribStack::Push(GLbitfield clientAttribMask)
{
	VW_ASSERT(GetAvailableStackSpace(), "ClientAttributeStack overflow imminent.");
	glPushClientAttrib(clientAttribMask);
}


//====================================================================
void  ClientAttribStack::Pop()
{
	VW_ASSERT(GetStackDepth(), "ClientAttributeStack underflow imminent.");
	glPopClientAttrib();
}


//====================================================================
ClientAttribStackSentry::ClientAttribStackSentry(GLbitfield clientAttribMask)
{
	VW_ASSERT(ClientAttribStack::GetAvailableStackSpace(), "ClientAttributeStack overflow imminent.");
	glPushClientAttrib(clientAttribMask);
}


//====================================================================
ClientAttribStackSentry::~ClientAttribStackSentry()
{
	VW_ASSERT(ClientAttribStack::GetStackDepth(), "ClientAttributeStack underflow imminent.");
	glPopClientAttrib();
}


//====================================================================
void __fastcall DumpAllStacks(std::ostream& os)
{
	const char* sep = " / ";
	os << caption("OpenGL stack dump");

	// modelview
	os << "\tModelView...: " << MatrixStack::GetStackDepth(GL_MODELVIEW);
	os << sep << MatrixStack::GetMaxStackDepth(GL_MODELVIEW) << nl;

	// projection
	os << "\tProjection..: " << MatrixStack::GetStackDepth(GL_PROJECTION);
	os << sep << MatrixStack::GetMaxStackDepth(GL_PROJECTION) << nl;

	// texture units
	TextureUnit::ActiveUnitSentry  aus(TextureUnit::GetActiveUnitIndex());

	uint numTexUnits = gl_tools::TextureUnit::GetNumUnits();
	for(uint tu = 0; tu < numTexUnits; tu++) {
		TextureUnit::ActiveUnitSentry  aus(tu);

		os << "\tTexture" << tu;
		if(tu < 10) os << ".";
		os << "...: " << MatrixStack::GetStackDepth(GL_TEXTURE);
		os << sep << MatrixStack::GetMaxStackDepth(GL_TEXTURE) << nl;
	}

	// color
	os << "\tColor.......: " << MatrixStack::GetStackDepth(GL_COLOR);
	os << sep << MatrixStack::GetMaxStackDepth(GL_COLOR) << nl;

	// server attributes
	os << "\tAttrib......: " << AttribStack::GetStackDepth();
	os << sep << AttribStack::GetMaxStackDepth() << nl;

	// client attributes
	os << "\tClientAttrib: " << ClientAttribStack::GetStackDepth();
	os << sep << ClientAttribStack::GetMaxStackDepth() << endpara;
}


//********************************************************************
}//namespace gl_tools


//********************************************************************
}//namespace vertexware


