

//--------------------------------------------------------------------
#include "GLTools\Textures.h"
#include "GLTools\Extensions.h"
#include "GLTools\States.h"
#include "GLTools\ErrorHandling.h"
#include "MathTools\MathBase.h"
#include "BaseTools\Constructs.h"
#include "DataObjects\ImageOperations.h"
#include "DataObjects\VolumeOperations.h"
#include <sstream>


//********************************************************************
namespace vertexware {


//********************************************************************
namespace gl_tools {


//====================================================================
TextureBindingSentry::TextureBindingSentry(GLenum _target, GLenum bindingPName,
											GLuint textureNameToBind)
{
	target = _target;
	lastBoundName = gl_tools::GetInteger(bindingPName);
	if(lastBoundName != textureNameToBind) {
		glBindTexture(_target, textureNameToBind);
	}
	else {
		target = 0;
	}
}


//====================================================================
TextureBindingSentry::~TextureBindingSentry()
{
	if(target) {
		glBindTexture(target, lastBoundName);
	}
}


//====================================================================
TextureBindingSentry3D::TextureBindingSentry3D(GLuint textureName)
	:
	TextureBindingSentry(GL_TEXTURE_3D, GL_TEXTURE_BINDING_3D, textureName)
{
	if(! gl_ext::IsSupported(gl_ext::Extension_GL_CORE_12)) {
		VW_THROW(OpenGLException,
				"3D textures are not supported by active GL-driver.");
	}
}


//====================================================================
uint __fastcall TextureUnit::GetNumUnits()
{
	int  numTextureUnits = 1;
	if(gl_ext::glActiveTexture) {
		numTextureUnits = gl_tools::GetInteger(GL_MAX_TEXTURE_UNITS);
	}
	else if(gl_ext::glActiveTextureARB) {
		numTextureUnits = gl_tools::GetInteger(GL_MAX_TEXTURE_UNITS_ARB);
	}
	return numTextureUnits;
}


//====================================================================
uint __fastcall TextureUnit::GetActiveUnitIndex()
{
	int  activeTextureUnitName = GL_TEXTURE0;
	if(gl_ext::glActiveTexture) {
		activeTextureUnitName = gl_tools::GetInteger(GL_ACTIVE_TEXTURE);
	}
	else if(gl_ext::glActiveTextureARB) {
		activeTextureUnitName = gl_tools::GetInteger(GL_ACTIVE_TEXTURE_ARB);
	}
	return (activeTextureUnitName - GL_TEXTURE0);
}


//====================================================================
void __fastcall TextureUnit::ActivateUnit(uint unitIndex)
{
	VW_ASSERT(unitIndex < GetNumUnits(), "Texture-unit index out of range.");

	if(gl_ext::glActiveTexture) {
		gl_ext::glActiveTexture(GL_TEXTURE0 + unitIndex);
		return;
	}
	if(gl_ext::glActiveTextureARB) {
		gl_ext::glActiveTextureARB(GL_TEXTURE0_ARB + unitIndex);
	}
}


//====================================================================
void __fastcall TextureUnit::DisableActiveUnit()
{
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	if(gl_ext::IsSupported(gl_ext::Extension_GL_CORE_12)) {
		glDisable(GL_TEXTURE_3D);

		if(gl_ext::IsSupported(gl_ext::Extension_GL_CORE_13)) {
			glDisable(GL_TEXTURE_CUBE_MAP);
		}
	}
}


//====================================================================
void __fastcall TextureUnit::EnableUnit(uint unitIndex, GLenum target)
{
	VW_ASSERT(unitIndex < GetNumUnits(), "Texture-unit index out of range.");

	ActiveUnitSentry  aus(unitIndex);
	EnableActiveUnit(target);
}


//====================================================================
void __fastcall TextureUnit::DisableUnit(uint unitIndex)
{
	VW_ASSERT(unitIndex < GetNumUnits(), "Texture-unit index out of range.");

	ActiveUnitSentry  aus(unitIndex);
	DisableActiveUnit();
}


//====================================================================
void __fastcall TextureUnit::DisableAllUnitsAndActivateUnit0()
{
	uint numUnits = GetNumUnits();

	if(numUnits == 1) {
		DisableActiveUnit();
		return;
	}
	for(uint unitIndex = 0; unitIndex < numUnits; unitIndex++) {
		ActivateUnit(unitIndex);
		DisableActiveUnit();
	}
	ActivateUnit(0);
}


//====================================================================
TextureUnit::ActiveUnitSentry::ActiveUnitSentry(uint newActiveUnitIndex)
{
	lastActiveUnitIndex = TextureUnit::GetActiveUnitIndex();
	bNeedsChange = (newActiveUnitIndex != lastActiveUnitIndex);
	if(bNeedsChange) {
		TextureUnit::ActivateUnit(newActiveUnitIndex);
	}
}


//====================================================================
TextureUnit::ActiveUnitSentry::~ActiveUnitSentry()
{
	if(bNeedsChange) {
		TextureUnit::ActivateUnit(lastActiveUnitIndex);
	}
}


//====================================================================
TextureUnit::TextureUnit(uint textureUnitIndex)
{
	unitIndex = textureUnitIndex;

	if(unitIndex >= GetNumUnits()) {
		VW_THROW(OpenGLException, "Texture-unit-index out of range.");
	}
	unitName = GL_TEXTURE0 + unitIndex;
}


//====================================================================
uint __fastcall TextureCoordArray::GetActiveArrayIndex()
{
	int  activeTextureArrayName = GL_TEXTURE0;
	if(gl_ext::glActiveTexture) {
		activeTextureArrayName = gl_tools::GetInteger(GL_CLIENT_ACTIVE_TEXTURE);
	}
	else if(gl_ext::glActiveTextureARB) {
		activeTextureArrayName = gl_tools::GetInteger(GL_CLIENT_ACTIVE_TEXTURE_ARB);
	}
	return (activeTextureArrayName - GL_TEXTURE0);
}


//====================================================================
void __fastcall TextureCoordArray::ActivateArray(uint arrayIndex)
{
	VW_ASSERT(arrayIndex < GetNumArrays(), "Texture-coord-array index out of range.");

	if(gl_ext::glClientActiveTexture) {
		gl_ext::glClientActiveTexture(GL_TEXTURE0 + arrayIndex);
		return;
	}
	if(gl_ext::glClientActiveTextureARB) {
		gl_ext::glClientActiveTextureARB(GL_TEXTURE0_ARB + arrayIndex);
	}
}


//====================================================================
void __fastcall TextureCoordArray::EnableArray(uint arrayIndex)
{
	VW_ASSERT(arrayIndex < GetNumArrays(), "Texture-coord-array index out of range.");

	ActiveArraySentry  aas(arrayIndex);
	EnableActiveArray();
}


//====================================================================
void __fastcall TextureCoordArray::DisableArray(uint arrayIndex)
{
	VW_ASSERT(arrayIndex < GetNumArrays(), "Texture-coord-array index out of range.");

	ActiveArraySentry  aas(arrayIndex);
	DisableActiveArray();
}


//====================================================================
void __fastcall TextureCoordArray::DisableAllArraysAndActivateArray0()
{
	uint numArrays = GetNumArrays();

	if(numArrays == 1) {
		DisableActiveArray();
		return;
	}
	for(uint arrayIndex = 0; arrayIndex < numArrays; arrayIndex++) {
		ActivateArray(arrayIndex);
		DisableActiveArray();
	}
	ActivateArray(0);
}


//====================================================================
TextureCoordArray::ActiveArraySentry::ActiveArraySentry(uint newActiveArrayIndex)
{
	lastActiveArrayIndex = TextureCoordArray::GetActiveArrayIndex();
	bNeedsChange = (newActiveArrayIndex != lastActiveArrayIndex);
	if(bNeedsChange) {
		TextureCoordArray::ActivateArray(newActiveArrayIndex);
	}
}


//====================================================================
TextureCoordArray::ActiveArraySentry::~ActiveArraySentry()
{
	if(bNeedsChange) {
		TextureCoordArray::ActivateArray(lastActiveArrayIndex);
	}
}


//====================================================================
TextureCoordArray::TextureCoordArray(uint textureArrayIndex)
{
	arrayIndex = textureArrayIndex;

	if(arrayIndex >= GetNumArrays()) {
		VW_THROW(OpenGLException, "Texture-coord-array-index out of range.");
	}
}


//====================================================================
uint __fastcall TextureBase::GetMaxTextureSize()
{
	return gl_tools::GetInteger(GL_MAX_TEXTURE_SIZE);
}


//====================================================================
GLint __fastcall TextureBase::MapSampleFormatToGLInternalFormat(eSampleFormat sampleFormat)
{
	switch(sampleFormat) {
		case SampleFormat_L8: return GL_LUMINANCE8;
		case SampleFormat_A8: return GL_ALPHA8;
		case SampleFormat_A8L8: return GL_LUMINANCE8_ALPHA8;
		case SampleFormat_A4R4G4B4: return GL_RGBA4;
		case SampleFormat_A1R5G5B5: return GL_RGB5_A1;
		case SampleFormat_X8R8G8B8: return GL_RGB8;
		case SampleFormat_A8R8G8B8: return GL_RGBA8;
	}
	return 0;
}


//====================================================================
eSampleFormat __fastcall TextureBase::MapGLInternalFormatToSampleFormat(GLint internalFormat)
{
	switch(internalFormat) {
		case GL_LUMINANCE8: return SampleFormat_L8;
		case GL_ALPHA8: return SampleFormat_A8;
		case GL_LUMINANCE8_ALPHA8: return SampleFormat_A8L8;
		case GL_RGBA4: return SampleFormat_A4R4G4B4;
		case GL_RGB5_A1: return SampleFormat_A1R5G5B5;
		case GL_RGB8: return SampleFormat_X8R8G8B8;
		case GL_RGBA8: return SampleFormat_A8R8G8B8;
	}
	return SampleFormat_Invalid;
}


//====================================================================
// Assumes BGRA extension
//====================================================================
bool __fastcall TextureBase::MapSampleFormatToGLUploadParameters(eSampleFormat sampleFormat,
											GLenum& outDataFormat, GLenum& outDataType,
											bool& outNeedsConversionToARGB32)
{
	switch(sampleFormat) {
		case SampleFormat_L8:
			outNeedsConversionToARGB32 = false;
			outDataFormat = GL_LUMINANCE;
			outDataType = GL_UNSIGNED_BYTE;
			return true;

		case SampleFormat_A8:
			outNeedsConversionToARGB32 = false;
			outDataFormat = GL_ALPHA;
			outDataType = GL_UNSIGNED_BYTE;
			return true;

		case SampleFormat_A8L8:
			outNeedsConversionToARGB32 = false;
			outDataFormat = GL_LUMINANCE_ALPHA;
			outDataType = GL_UNSIGNED_BYTE;
			return true;

		case SampleFormat_A4R4G4B4:
			outNeedsConversionToARGB32 = ! gl_ext::IsSupported(gl_ext::Extension_GL_CORE_12);
			outDataFormat = GL_BGRA;
			outDataType = outNeedsConversionToARGB32 ?
				GL_UNSIGNED_BYTE : GL_UNSIGNED_SHORT_4_4_4_4_REV;
			return true;

		case SampleFormat_A1R5G5B5:
			outNeedsConversionToARGB32 = ! gl_ext::IsSupported(gl_ext::Extension_GL_CORE_12);
			outDataFormat = GL_BGRA;
			outDataType = outNeedsConversionToARGB32 ?
				GL_UNSIGNED_BYTE : GL_UNSIGNED_SHORT_1_5_5_5_REV;
			return true;

		case SampleFormat_X8R8G8B8:
		case SampleFormat_A8R8G8B8:
			outNeedsConversionToARGB32 = false;
			outDataFormat = GL_BGRA;
			outDataType = gl_ext::IsSupported(gl_ext::Extension_GL_CORE_12) ?
				GL_UNSIGNED_INT_8_8_8_8_REV : GL_UNSIGNED_BYTE;
			return true;
	}
	return false;
}


//====================================================================
TextureBase::TextureBase(GLenum _target, GLenum _bindingPName, GLint _internalFormat)
{
	target = _target;
	bindingPName = _bindingPName;
	internalFormat = _internalFormat;

	// allocate name
	name = 0;
	glGenTextures(1, &name);
	if(! name) {
		VW_THROW(OpenGLException, "Cannot allocate GL-texture-name.");
	}
}


//====================================================================
TextureBase::~TextureBase()
{
	glDeleteTextures(1, &name);
}


//====================================================================
bool __fastcall TextureBase::CheckRealizedTextureWidth(GLenum target, uint level, uint expectedWidth)
{
	GLint testWidth;
	glGetTexLevelParameteriv(target, level, GL_TEXTURE_WIDTH, &testWidth);

	return (testWidth == expectedWidth);
}


//====================================================================
void  TextureBase::ValidateInternalFormat()
{
	if(GetInternalFormat() == 0) {
		VW_THROW(OpenGLException,
				"Cannot map texture sample-format to GL-internal-format.");
	}
}


//====================================================================
void  TextureBase::RealizePriority(uint newPriority)
{
	GLclampf fPriority = float(newPriority) / float(video_renderer_objects::Resource::HighestPriority);
	glPrioritizeTextures(1, &name, &fPriority);
}


//====================================================================
bool  TextureBase::RealizeBaseLevel(uint newBaseLevel)
{
	if(gl_ext::IsSupported(gl_ext::Extension_GL_CORE_12)) {
		TextureBindingSentry tbs(target, bindingPName, GetName());
		glTexParameteri(target, GL_TEXTURE_BASE_LEVEL, newBaseLevel);
		return true;
	}
	return false;
}


//====================================================================
Texture1D::Texture1D(eSampleFormat sampleFormat, bool bMultiLevel,
					uint widthLevel0)
	:
	video_renderer_objects::Texture1D(sampleFormat, bMultiLevel,
							widthLevel0),
	gl_tools::TextureBase(GL_TEXTURE_1D, GL_TEXTURE_BINDING_1D,
							MapSampleFormatToGLInternalFormat(sampleFormat))
{
	ValidateInternalFormat();
	Allocate();
}


//====================================================================
Texture1D::Texture1D(GLint internalFormat, bool bMultiLevel,
					uint widthLevel0)
	:
	video_renderer_objects::Texture1D(MapGLInternalFormatToSampleFormat(internalFormat),
							bMultiLevel, widthLevel0),
	gl_tools::TextureBase(GL_TEXTURE_1D, GL_TEXTURE_BINDING_1D,
							internalFormat)
{
	Allocate();
}


//====================================================================
void  Texture1D::Allocate()
{
	if(! CheckDimension(GetWidth(0), true)) {
		VW_THROW(OpenGLException, "Texture dimensions are 0 or not power-of-2.");
	}

	TextureBindingSentry1D  tbs(GetName());

	// check if implementation can handle given parameters &
	// allocate memory resources for all mipmap levels
	for(uint level = 0; level < GetNumLevels(); level++) {

		uint levelWidth = GetWidth(level);

		// query for current mipmap level
		glTexImage1D(GL_PROXY_TEXTURE_1D,
					level, GetInternalFormat(),
					levelWidth, 0,
					GL_RGBA, GL_UNSIGNED_BYTE, 0);

		if(! CheckRealizedTextureWidth(GL_PROXY_TEXTURE_1D, level, levelWidth)) {
			std::ostringstream oss;
			oss << "Cannot allocate GL-texture-resources (w=" << GetWidth(0) << ").";
			VW_THROW(OpenGLException, oss.str());
		}

		// allocate current mipmap level
		glTexImage1D(GL_TEXTURE_1D,
					level, GetInternalFormat(),
					levelWidth, 0,
					GL_RGBA, GL_UNSIGNED_BYTE, 0);
	}

	UpdatePriority();
}


//====================================================================
bool  Texture1D::DoLoadLevel(uint level, const SignalBase& src,
							uint dataWidth,
							uint dstOffsetX,
							uint srcOffsetX)
{
	// parameters already checked by caller
	GLenum  dataFormat, dataType;
	bool    bNeedsConversionToARGB32;

	if(! MapSampleFormatToGLUploadParameters(src.GetSampleFormat(),
				dataFormat, dataType, bNeedsConversionToARGB32)) return false;

	const SignalBase*   pSrcSignal = &src;
	AutoPtr<SignalBase> apARGB32Signal;

	if(bNeedsConversionToARGB32) {
		apARGB32Signal.Assign(new Signal32(SampleFormat_A8R8G8B8, src.GetNumSamples()));
		if(! TransferSamples1D(*apARGB32Signal, src, false)) return false;
		pSrcSignal = apARGB32Signal;
	}

	// upload pixels to GL-memory
	TextureBindingSentry1D  tbs(GetName());

	SetupPixelUnpack(false, 1);

	glTexSubImage1D(GL_TEXTURE_1D, level,
					dstOffsetX, dataWidth,
					dataFormat, dataType, pSrcSignal->SampleB(srcOffsetX));
	return true;
}


//====================================================================
Texture2D::Texture2D(eSampleFormat sampleFormat, bool bMultiLevel,
					uint heightLevel0, uint widthLevel0)
	:
	video_renderer_objects::Texture2D(sampleFormat, bMultiLevel,
							heightLevel0, widthLevel0),
	gl_tools::TextureBase(GL_TEXTURE_2D, GL_TEXTURE_BINDING_2D,
							MapSampleFormatToGLInternalFormat(sampleFormat))
{
	ValidateInternalFormat();
	Allocate();
}


//====================================================================
Texture2D::Texture2D(GLint internalFormat, bool bMultiLevel,
					uint heightLevel0, uint widthLevel0)
	:
	video_renderer_objects::Texture2D(MapGLInternalFormatToSampleFormat(internalFormat),
							bMultiLevel, heightLevel0, widthLevel0),
	gl_tools::TextureBase(GL_TEXTURE_2D, GL_TEXTURE_BINDING_2D,
							internalFormat)
{
	Allocate();
}


//====================================================================
void  Texture2D::Allocate()
{
	if(! CheckDimension(GetHeight(0), true)  ||  ! CheckDimension(GetWidth(0), true)) {
		VW_THROW(OpenGLException, "Texture dimensions are 0 or not power-of-2.");
	}

	TextureBindingSentry2D  tbs(GetName());

	// check if implementation can handle given parameters &
	// allocate memory resources for all mipmap levels
	for(uint level = 0; level < GetNumLevels(); level++) {

		uint levelHeight = GetHeight(level);
		uint levelWidth  = GetWidth(level);

		// query for current mipmap level
		glTexImage2D(GL_PROXY_TEXTURE_2D,
					level, GetInternalFormat(),
					levelWidth, levelHeight, 0,
					GL_RGBA, GL_UNSIGNED_BYTE, 0);

		if(! CheckRealizedTextureWidth(GL_PROXY_TEXTURE_2D, level, levelWidth)) {
			std::ostringstream oss;
			oss << "Cannot allocate GL-texture-resources (h=" << GetHeight(0);
			oss << " w=" << GetWidth(0) << ").";
			VW_THROW(OpenGLException, oss.str());
		}

		// allocate current mipmap level
		glTexImage2D(GL_TEXTURE_2D,
					level, GetInternalFormat(),
					levelWidth, levelHeight, 0,
					GL_RGBA, GL_UNSIGNED_BYTE, 0);
	}

	UpdatePriority();
}


//====================================================================
bool  Texture2D::DoLoadLevel(uint level, const ImageBase& src,
							bool bSrcIsStoredBottomToTop,
							uint dataHeight, uint dataWidth,
							uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetY, uint srcOffsetX)
{
	// parameters already checked by caller
	GLenum  dataFormat, dataType;
	bool    bNeedsConversionToARGB32;

	if(! MapSampleFormatToGLUploadParameters(src.GetSampleFormat(),
				dataFormat, dataType, bNeedsConversionToARGB32)) return false;

	const ImageBase*   pSrcImage = &src;
	AutoPtr<ImageBase> apARGB32Image;

	if(bNeedsConversionToARGB32) {
		apARGB32Image.Assign(new Image32(SampleFormat_A8R8G8B8, src.GetNumRows(), src.GetNumColumns()));
		if(! TransferSamples2D(*apARGB32Image, src, false)) return false;
		pSrcImage = apARGB32Image;
	}

	// convert data-order to GL-native bottom-to-top if necessary
	AutoPtr<ImageBase> apFlipedImage;

	if(! bSrcIsStoredBottomToTop) {
		apFlipedImage.Assign(ManufactureImage(pSrcImage->GetSampleFormat(),
							dataHeight, dataWidth));

		// Source-image is stored top-to-bottom.
		// Source-row-offsets are specified in top-to-bottom convention.
		CopySamples2D(*apFlipedImage, *pSrcImage, dataHeight, dataWidth,
								0, 0, srcOffsetY, srcOffsetX, true);
		srcOffsetY = 0;
		srcOffsetX = 0;

		pSrcImage = apFlipedImage;
	}

	// upload pixels to GL-memory
	TextureBindingSentry2D  tbs(GetName());

	uint rowLengthInPixels = (dataWidth != pSrcImage->GetNumColumns()) ? pSrcImage->GetNumColumns() : 0;

	SetupPixelUnpack(false, 1, rowLengthInPixels);

	// row-offsets are specified in top-to-bottom convention, change to bottom-to-top convention
	glTexSubImage2D(GL_TEXTURE_2D, level,
			dstOffsetX, GetHeight(level) - (dstOffsetY + dataHeight),
			dataWidth, dataHeight,
			dataFormat, dataType,
			pSrcImage->SampleB(pSrcImage->GetNumRows() - (srcOffsetY + dataHeight), srcOffsetX));

	return true;
}


//====================================================================
Texture3D::Texture3D(eSampleFormat sampleFormat, bool bMultiLevel,
					uint depthLevel0, uint heightLevel0, uint widthLevel0)
	:
	video_renderer_objects::Texture3D(sampleFormat, bMultiLevel,
							depthLevel0, heightLevel0, widthLevel0),
	gl_tools::TextureBase(GL_TEXTURE_3D, GL_TEXTURE_BINDING_3D,
							MapSampleFormatToGLInternalFormat(sampleFormat))
{
	ValidateInternalFormat();
	Allocate();
}


//====================================================================
Texture3D::Texture3D(GLint internalFormat, bool bMultiLevel,
					uint depthLevel0, uint heightLevel0, uint widthLevel0)
	:
	video_renderer_objects::Texture3D(MapGLInternalFormatToSampleFormat(internalFormat),
							bMultiLevel, depthLevel0, heightLevel0, widthLevel0),
	gl_tools::TextureBase(GL_TEXTURE_3D, GL_TEXTURE_BINDING_3D,
							internalFormat)
{
	Allocate();
}


//====================================================================
void  Texture3D::Allocate()
{
	if(! CheckDimension(GetDepth(0), true)  ||  ! CheckDimension(GetHeight(0), true)  ||
		! CheckDimension(GetWidth(0), true))
	{
		VW_THROW(OpenGLException, "Texture dimensions are 0 or not power-of-2.");
	}

	// sentry checks implementation's support for 3D textures
	TextureBindingSentry3D  tbs(GetName());

	// check if implementation can handle given parameters &
	// allocate memory resources for all mipmap levels
	for(uint level = 0; level < GetNumLevels(); level++) {

		uint levelDepth  = GetDepth(level);
		uint levelHeight = GetHeight(level);
		uint levelWidth  = GetWidth(level);

		// query for current mipmap level
		gl_ext::glTexImage3D(GL_PROXY_TEXTURE_3D,
						level, GetInternalFormat(),
						levelWidth, levelHeight, levelDepth, 0,
						GL_RGBA, GL_UNSIGNED_BYTE, 0);

		if(! CheckRealizedTextureWidth(GL_PROXY_TEXTURE_3D, level, levelWidth)) {
			std::ostringstream oss;
			oss << "Cannot allocate GL-texture-resources (d=" << GetDepth(0);
			oss << " h=" << GetHeight(0) << " w=" << GetWidth(0) << ").";
			VW_THROW(OpenGLException, oss.str());
		}

		// allocate current mipmap level
		gl_ext::glTexImage3D(GL_TEXTURE_3D,
						level, GetInternalFormat(),
						levelWidth, levelHeight, levelDepth, 0,
						GL_RGBA, GL_UNSIGNED_BYTE, 0);
	}

	UpdatePriority();
}


//====================================================================
bool  Texture3D::DoLoadLevel(uint level, const VolumeBase& src,
							bool bSrcIsStoredBackToFront, bool bSrcIsStoredBottomToTop,
							uint dataDepth, uint dataHeight, uint dataWidth,
							uint dstOffsetZ, uint dstOffsetY, uint dstOffsetX,
							uint srcOffsetZ, uint srcOffsetY, uint srcOffsetX)
{
	// parameters already checked by caller
	GLenum  dataFormat, dataType;
	bool    bNeedsConversionToARGB32;

	if(! MapSampleFormatToGLUploadParameters(src.GetSampleFormat(),
				dataFormat, dataType, bNeedsConversionToARGB32)) return false;

	const VolumeBase*   pSrcVolume = &src;
	AutoPtr<VolumeBase> apARGB32Volume;

	if(bNeedsConversionToARGB32) {
		apARGB32Volume.Assign(new Volume32(SampleFormat_A8R8G8B8, src.GetNumLayers(),
											src.GetNumRows(), src.GetNumColumns()));
		if(! TransferSamples3D(*apARGB32Volume, src, false)) return false;
		pSrcVolume = apARGB32Volume;
	}

	// convert data-order to GL-native front-to-back/bottom-to-top if necessary
	AutoPtr<VolumeBase> apFlipedVolume;

	bool  bNeedsLayerFlip = (bSrcIsStoredBackToFront != false);
	bool  bNeedsRowFlip   = (bSrcIsStoredBottomToTop != true);

	if(bNeedsLayerFlip  ||  bNeedsRowFlip) {
		apFlipedVolume.Assign(ManufactureVolume(pSrcVolume->GetSampleFormat(),
											dataDepth, dataHeight, dataWidth));

		// If <bNeedsRowFlip> is true, source-volume is stored top-to-bottom.
		// Source-row-offsets are specified in top-to-bottom convention.
		// If <bNeedsLayerFlip> is true, source-volume is stored back-to-front.
		// Source-layer-offsets are specified in front-to-back convention.
		CopySamples3D(*apFlipedVolume, *pSrcVolume,
					dataDepth, dataHeight, dataWidth,
					0, 0, 0,
					(bNeedsLayerFlip ? pSrcVolume->GetNumLayers() - (srcOffsetZ + dataDepth) : srcOffsetZ),
					srcOffsetY, srcOffsetX,
					bNeedsLayerFlip, bNeedsRowFlip);

		srcOffsetZ = 0;
		srcOffsetY = 0;
		srcOffsetX = 0;

		pSrcVolume = apFlipedVolume;
	}

	// upload pixels to GL-memory
	TextureBindingSentry3D  tbs(GetName());

	uint imageHeightInPixels = (dataHeight != pSrcVolume->GetNumRows()) ? pSrcVolume->GetNumRows() : 0;
	uint rowLengthInPixels = (dataWidth != pSrcVolume->GetNumColumns()) ? pSrcVolume->GetNumColumns() : 0;

	SetupPixelUnpack(false, 1, rowLengthInPixels, imageHeightInPixels);

	// row-offsets are specified in top-to-bottom convention, change to bottom-to-top convention
	gl_ext::glTexSubImage3D(GL_TEXTURE_3D, level,
						dstOffsetX,
						GetHeight(level) - (dstOffsetY + dataHeight),
						dstOffsetZ,
						dataWidth, dataHeight, dataDepth,
						dataFormat, dataType,
						pSrcVolume->SampleB(
								srcOffsetZ,
								pSrcVolume->GetNumRows() - (srcOffsetY + dataHeight),
								srcOffsetX));
	return true;
}


//********************************************************************
}//namespace gl_tools


//********************************************************************
}//namespace vertexware


